/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: CMS.java,v 1.4 2001/06/02 21:57:16 kuldeep Exp $
 *
 */

package net.jxta.share;

import java.lang.reflect.InvocationTargetException;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;

import net.jxta.discovery.Discovery;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.Document;
import net.jxta.document.MimeMediaType;
import net.jxta.endpoint.Message;
import net.jxta.exception.PeerGroupException;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.PipeID;
import net.jxta.pipe.InputPipe;
import net.jxta.peergroup.PeerGroup;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.service.Service;

/**
 * This class implements the Content Management Service.
 * The class is started with one of the startApp methods.
 * The command pipe and persistent directory is passed to the startApp method, else
 * default values are used.
 * A remote client interacts with the service by finding the request PipeAdvertisement
 * and sending it LIST and GET requests. The requests are handled by the 
 * ListMessageProcessor and GetMessageProcessor.
 */

public class CMS implements Service {

    public static final String DEFAULT_DIR = "cms";
    public static final String PIPE_ADV_FILENAME = "cms_pipe.xml";
    public static final String PIPE_ADV_NAME = "jxta:CMSPipe";

    // Message tag's
    public static final String LIST_REQUEST = "LIST_REQ";
    public static final String LIST_RESULT = "LIST_RES";
    public static final String GET_REQUEST = "GET_REQ";
    public static final String GET_RESULT = "GET_RES";

    public static final String MESSAGE_TYPE = "msgType";
    public static final String REQUEST_ID = "RID";
    public static final String CONTENT_ID = "CID";
    public static final String PIPE_ADVERTISEMENT = "PIPE";
    public static final String CONTENT_ADVERTISEMENT = "CADV";
    public static final String CONTENT_LENGTH = "CLEN";
    public static final String ADVERTISEMENT_COUNT = "COUNT";
    public static final String CHUNK_OFFSET = "OFFSET";
    public static final String CHUNK_SIZE = "SIZE";
    public static final String CHUNK_DATA = "DATA";

    public static final String QUERY_SUBSTRING = "QSUBSTR";

    public static int pipeTimeout = 30000; // 30 second timeout
    public static MimeMediaType encodeAs = new MimeMediaType("text", "xml");

    private PeerGroup group = null;
    private InputPipe requestInPipe = null;
    private PipeAdvertisement requestPipeAdv = null;
    private ListMessageProcessor listQueue = null;
    private GetMessageProcessor getQueue = null;
    private ContentManager contentManager = null;
    private File cmsDir = null;

    public CMS() {
        // System.out.println("CMS()");
    }

    /**
     * Initialize the application passing it ist peer group handle.
     *
     * @param group PeerGroup this application is started from
     *
     * @exception PeerGroupException failure to initialize the peer group application
     *
     * @since   JXTA 1.0
     */
    public void init(PeerGroup group, Advertisement adv)
        throws PeerGroupException {
        this.group = group;
    }

    /**
     * Start the service with the given parameters.
     * The parameters are in the form of name=value pairs.
     * There is one optional parameter, the directory for storing
     * persistent data about the shared content.  
     *
     * @param arg String[] of parameters
     *
     * @return int status indication. 0 if ok, -1 if error
     */
    public int startApp(String[] arg) {
        PipeAdvertisement pipeAdv = null;
        File cmsDir = null;

        for (int i=0; i<arg.length; i++) {
            // Directory parameter
            if (arg[i].startsWith("dir=")) {
                cmsDir = new File(arg[i].substring(4));
            }
        }

        return startApp(cmsDir);
    }

    /**
     * Start the service with the given parameters.
     *
     * @param rPipeAdv PipeAdvertisement of the request pipe
     * @param cmsDir File pointing to the persistence directory
     *
     * @return int status indication. 0 if ok, -1 if error.
     */
    public int startApp(File dir) {
        // check if a cmsDir was passed
        cmsDir = dir;
        if (null == cmsDir) {
            cmsDir = new File(DEFAULT_DIR);
        }

        // Check if the cms directory exists
        if (false == cmsDir.exists()) {
            // create the directory
            cmsDir.mkdir();
        }

        // load the pipe advertisement for the service
        try {
            loadPipeAdvertisement();
        } catch (IOException e) {
	    //System.err.println("Could not initialize content manager");
	    //e.printStackTrace();
            return (-1);
        }

        // Initialize the content manager
        try {
            contentManager = new ContentManagerImpl(cmsDir);
        } catch (IOException e) {
	    //System.err.println("Could not initialize content manager");
	    //e.printStackTrace();
            return (-1);
        }

        // initialize the message processor queues
        listQueue = new ListMessageProcessor(this);
        getQueue = new GetMessageProcessor(this);

        // start the command listener
        try {
            listen();
        } catch(IOException e) {
        }

        // publish the pipe advertisement
        try {
            group.getDiscovery().publish(getRequestPipeAdvertisement(), Discovery.ADV);
            group.getDiscovery().remotePublish(getRequestPipeAdvertisement(), Discovery.ADV);
        } catch (IOException e) {
	    //System.err.println("Could not initialize content manager");
	    //e.printStackTrace();
            return (-1);
        }

        return 0;
    }

    /**
     * Stops the service
     */
    public void stopApp() {
    }

    /**
     * Service objects are not manipulated directly to protect usage
     * of the service. A Service interface is returned to access the service
     * methods.
     *
     * @return Service public interface of the service
     *
     * @version $Revision: 1.4 $
     * @since JXTA 1.0 
     */
    public Service getInterface() {
        return this;
    }

    /**
     * Returns the advertisment for the service.
     *
     * @return Advertisement the advertisement.
     *
     * @version $Revision: 1.4 $
     * @since JXTA 1.0 
     */
    public Advertisement getAdvertisement() {
        return null;
    }

    /**
     * Returns the pipe service for this group.
     *
     * @return Pipe the pipe service.
     *
     * @version $Revision: 1.4 $
     * @since JXTA 1.0 
     */
    public Pipe getPipe() {
        return group.getPipe();
    }

    /**
     * Returns the request pipe advertisment for this service.
     *
     * @return PipeAdvertisement the request pipe advertisement.
     *
     * @version $Revision: 1.4 $
     * @since JXTA 1.0 
     */
    public PipeAdvertisement getRequestPipeAdvertisement() {
        return requestPipeAdv;
    }

    /**
     * Returns the Content Manager for this service.
     *
     * @return ContentManager the Content Manager.
     *
     * @version $Revision: 1.4 $
     * @since JXTA 1.0 
     */
    public ContentManager getContentManager() {
        return contentManager;
    }

    /**
     * Returns the String contained in the given tag in the given Message.
     *
     * @return String from given tag in given message.
     *
     * @version $Revision: 1.4 $
     * @since JXTA 1.0 
     */
    protected static String popString(Message message, String tag) throws IOException {
        InputStream ip = message.pop(tag);
        byte[] buffer = new byte[ip.available()];
        ip.read(buffer);
        String str = new String(buffer);
        
        return str;
    }

    private void loadPipeAdvertisement() throws IOException {
        // check if the pipe advertisement file exist
        File file = new File(cmsDir, PIPE_ADV_FILENAME);
        if (file.exists()) {
            // load the pipe advertisement
            try {
                Advertisement adv = 
                    AdvertisementFactory.newAdvertisement(CMS.encodeAs, 
                        new FileInputStream(file));
                if (adv instanceof PipeAdvertisement) {
                    requestPipeAdv = (PipeAdvertisement)adv;
                }
            } catch (InvocationTargetException e) {
	      //e.printStackTrace();
            }
        }

        if (null == requestPipeAdv) {
            // create a new Pipe Advertisement for this service
            try {
                Advertisement adv = 
                    AdvertisementFactory.newAdvertisement(PipeAdvertisement.getAdvertisementType());
                if (adv instanceof PipeAdvertisement) {
                    requestPipeAdv = (PipeAdvertisement)adv;
                }
            } catch (InvocationTargetException e) {
	        //System.err.println("Warning: could not create PipeAdvertisement");
	        //e.printStackTrace();
                return;
            }

            requestPipeAdv.setPipeID(new PipeID(group.getID()));
            requestPipeAdv.setName(PIPE_ADV_NAME);

            // save the new pipe advertisement
            Document doc = null;
            try {
                doc = requestPipeAdv.getDocument(CMS.encodeAs);
            } catch (Exception e) {
	        //e.printStackTrace();
                return;
            }

            FileOutputStream out = new FileOutputStream(file);
            doc.sendToStream(out);
            out.close();
        }
    }

    private void listen() throws IOException {
        // create the input pipe
        requestInPipe = group.getPipe().createInputPipe(requestPipeAdv);

        Thread thread = new Thread(new Listener());
        thread.start();
    }

    class Listener implements Runnable {
        public void run() {
	    //System.out.println("CMS.Listener.run()");
        
            Message message = null;
            while (true) {
                try {
                    message = requestInPipe.waitForMessage();
                } catch (IOException e) {
                    break;
                }

                //System.out.println("CMS.Listener got message");
                if (message == null) {
                    if (Thread.interrupted()) {
                        // We have been asked to stop
                        break;
                    }
                } else {
                    // Handle the message
                    try {
                        String messageType = popString(message, MESSAGE_TYPE);

                        //System.out.println("handleMessage type = " + messageType);
                        if (LIST_REQUEST.equals(messageType)) {
                            // queue up on the list command queue
                            listQueue.push(message);
                        } else if (GET_REQUEST.equals(messageType)) {
                            // queue up on the get command queue
                            getQueue.push(message);
                        }
                    } catch (IOException e) {
                    }
                }
            }

            if (null != requestInPipe) {
                requestInPipe.close();
                requestInPipe = null;
            }
        }
    }
}
