/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ContentAdvertisementImpl.java,v 1.3 2001/03/27 05:15:55 dac Exp $
 *
 */

package net.jxta.share;

import net.jxta.document.Advertisement;
import net.jxta.document.Document;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.StructuredDocument;
import net.jxta.document.TextElement;
import net.jxta.document.Element;
import net.jxta.document.MimeMediaType;

import java.io.InputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;
import java.lang.reflect.InvocationTargetException;

/*
 * ContentAdvertisement implementation for content manager.
 */
class ContentAdvertisementImpl extends ContentAdvertisement {
    private String name;	// content name (required)
    private ContentId cid;	// content id (required)
    private long length = -1;	// content length (optional)
    private String type;	// content type (optional)
    private String description; // content description (optional)

    /*
     * Creates a new ContentAdvertisement for the specified content name,
     * id, length, type, and description
     * @param name the content name
     * @param cid the content id
     * @param length the content length, or -1 if not specified
     * @param type the content mime type, or null if not specified
     * @param description the content description, or null if not specified
     */
    ContentAdvertisementImpl(String name, ContentId cid, long length,
			 String type, String description)
    {
	if (name == null) {
	    throw new NullPointerException("name");
	}
	if (cid == null) {
	    throw new NullPointerException("cid");
	}
	this.name = name;
	this.cid = cid;
	this.type = type;
	this.length = length;
	this.description = description;
    }

    /**
     * Creates a new ContentAdvertisement from the specified InputStream.
     * @param is the InputStream containing the advertisement
     * @exception IllegalArgumentException if the advertisement was invalid
     * @exception IOException if there was an I/O error
     */
    public ContentAdvertisementImpl(InputStream is) throws IOException {
	StructuredTextDocument doc;
	try {
	    doc = (StructuredTextDocument)
		StructuredDocumentFactory.newStructuredDocument(
		    new MimeMediaType("text", "xml"), is);
	} catch (InvocationTargetException e) {
	    Throwable t = e.getTargetException();
	    if (t instanceof IOException) {
		throw (IOException)t;
	    } else {
		t.printStackTrace();
		throw new IllegalArgumentException(
		    "Invalid content advertisement format");
	    }
	}
	init(doc);
    }

    /**
     * Creates a new ContentAdvertisement from the specified TextElement.
     * @exception 
     */
    public ContentAdvertisementImpl(TextElement te) {
	init(te);
    }

    // reads advertisement from specified TextElement
    private void init(TextElement root) {
	if (!root.getName().equals(getAdvertisementType())) {
	    throw new IllegalArgumentException(
		"Not a ContentAdvertisement element: " + root.getName());
	}
	Enumeration e = root.getChildren();
	while (e.hasMoreElements()) {
	    TextElement te = (TextElement)e.nextElement();
	    String s = te.getName();
	    String t = te.getTextValue();
	    if (t != null) {
		if (s.equals(EN_NAME)) {
		    name = t;
		} else if (s.equals(EN_CID)) {
		    try {
			cid = new ContentIdImpl(t);
		    } catch (IllegalArgumentException ex) {
			throw new IllegalArgumentException(
			    "Invalid content id: " + t);
		    }
		} else if (s.equals(EN_LENGTH)) {
		    try {
			length = Long.parseLong(t);
		    } catch (NumberFormatException ex) {
			throw new IllegalArgumentException(
			    "Invalid content length: " + t);
		    }
		} else if (s.equals(EN_TYPE)) {
		    type = t;
		} else if (s.equals(EN_DESCRIPTION)) {
		    description = t;
		}
	    }
	}
	if (name == null) {
	    throw new IllegalArgumentException(
		"Missing required 'name' element");
	}
	if (cid == null) {
	    throw new IllegalArgumentException(
		"Missing required 'cid' element");
	}
    }

    /**
     * Returns the advertisement type.
     */
    public static String getAdvertisementType() {
	return "jxta:ContentAdvertisement";
    }

    /**
     * Returns the name of the content.
     */
    public String getName() {
	return name;
    }

    /**
     * Returns the content id.
     */
    public ContentId getContentId() {
	return cid;
    }

    /**
     * Returns the length of the content, or -1 if unknown
     */
    public long getLength() {
	return length;
    }

    /**
     * Returns the mime type of the content, or null if unknown
     */
    public String getType() {
	return type;
    }

    /**
     * Returns the description of the content or null if none.
     */
    public String getDescription() {
	return description;
    }

    /**
     * Returns the document representation of this advertisement.
     */
    public Document getDocument(MimeMediaType mtype) throws Exception {
	StructuredDocument doc = (StructuredDocument)
	    StructuredDocumentFactory.newStructuredDocument(
		mtype, getAdvertisementType());
	appendElements(doc, doc);
	return doc;
    }

    /**
     * Appends this content advertisement to the specified structured
     * document element.
     */ 
    public void appendDocument(StructuredDocument doc, Element el) {
	Element e = doc.createElement(getAdvertisementType());
	el.appendChild(e);
	appendElements(doc, e);
    }

    // append elements to specified document element
    private void appendElements(StructuredDocument doc, Element e) {
	e.appendChild(doc.createElement(EN_NAME, name));
	e.appendChild(doc.createElement(EN_CID, cid.toString()));
	if (length != -1) {
	    e.appendChild(doc.createElement(EN_LENGTH, Long.toString(length)));
	}
	if (type != null) {
	    e.appendChild(doc.createElement(EN_TYPE, type));
	}
	if (description != null) {
	    e.appendChild(doc.createElement(EN_DESCRIPTION, description));
	}
    }

    /**
     * Returns a string representation of this advertisement.
     */
    public String toString() {
	StringBuffer sb = new StringBuffer();
	sb.append("{name=");
	sb.append(name);
	sb.append(";cid=");
	sb.append(cid.toString());
	if (length != -1) {
	    sb.append(";length=");
	    sb.append(length);
	}
	if (type != null) {
	    sb.append(";type=");
	    sb.append(type);
	}
	if (description != null) {
	    sb.append(";description=");
	    sb.append(description);
	}
	sb.append('}');
	return sb.toString();
    }
}
