/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: GetContentRequest.java,v 1.3 2001/06/02 21:39:13 kuldeep Exp $
 *
 */

package net.jxta.share;

import java.lang.reflect.InvocationTargetException;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.BitSet;

import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.endpoint.Message;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.PipeID;
import net.jxta.protocol.PipeAdvertisement;

/**
 * This class implements Get Content Request.
 * It request the given content from the given CMS request pipe.
 */

public class GetContentRequest implements Runnable {

    private Pipe pipe = null;
    private PipeAdvertisement requestPipe = null;
    private PipeAdvertisement resultPipe = null;
    private ContentId cid = null;
    private String rid = null;
    private File destFile = null;

    private Thread thread = null;

    private boolean isDone = false;
    private boolean hasFailed = false;

    private long contentLength = 0;
    private int chunkSize = 0;
    private int numberChunks = 0;
    private int numberReceived = 0;
    private BitSet bitSet = null;

    public GetContentRequest(PeerGroup group, PipeAdvertisement inRequest, 
            ContentId inCid, File inFile) throws InvocationTargetException {
        cid = inCid;
        pipe = group.getPipe();
        requestPipe = inRequest;
        destFile = inFile;

        rid = new UUID().toString();

        // create a new PipeAdvertisement
        Advertisement adv = 
            AdvertisementFactory.newAdvertisement(PipeAdvertisement.getAdvertisementType());
        if (adv instanceof PipeAdvertisement) {
            resultPipe = (PipeAdvertisement)adv;
        }
        resultPipe.setPipeID(new PipeID(group.getID()));
        resultPipe.setName("ListContentRequest");

        thread = new Thread(this);
        thread.start();
    }

    public GetContentRequest(PeerGroup group, PipeAdvertisement inRequest, 
            PipeAdvertisement inResult, ContentId inCid, File inFile) {
        cid = inCid;
        pipe = group.getPipe();
        requestPipe = inRequest;
        resultPipe = inResult;
        destFile = inFile;

        rid = new UUID().toString();

        Thread thread = new Thread(this);
        thread.start();
    }

    public void run() {
        // create the input pipe
        InputPipe inPipe = null;
        try {
            inPipe = pipe.createInputPipe(resultPipe);
        } catch (IOException e) {
            System.err.println("Could not create InputPipe");
            e.printStackTrace();
        }        

        // send the request
        try {
            sendRequest();
        } catch(Exception e) {
            e.printStackTrace();
            isDone = true;
            notifyFailure();
        }

        // wait for result
        Message message = null;
        while (!isDone) {
            try {
                message = inPipe.waitForMessage();
            } catch (IOException e) {
                inPipe.close();
                break;
            }

            if (message == null) {
                if (Thread.interrupted()) {
                    // We have been asked to stop
		  //System.err.println("ListContentRequest.run() terminated");
                    isDone = true;
                    break;
                }
            } else {
                // Handle the message
                String messageType = null;
                String mRid = null;
                String mCid = null;
                try {
                    messageType = CMS.popString(message, CMS.MESSAGE_TYPE);
                    mRid = CMS.popString(message, CMS.REQUEST_ID);
                    mCid = CMS.popString(message, CMS.CONTENT_ID);
                } catch (IOException e) {
                    e.printStackTrace();
                }

                if (CMS.GET_RESULT.equals(messageType) && rid.equals(mRid) && 
                    cid.toString().equals(mCid)) {
                    handleMessage(message);
                }
            }
        }

        if (null != inPipe) {
            inPipe.close();
            inPipe = null;
        }

        // call the done notify routine
        notifyDone();
    }

    public void cancel() {
        isDone = true;
        hasFailed = true;
        if (null != thread) {
            thread.interrupt();
        }
    }

    public boolean isDone() {
        return isDone;
    }

    public boolean hasFailed() {
        return hasFailed;
    }

    public int getPercentDone() {
        if (0 == numberChunks) {
            return 0;
        }

        return (numberReceived*100)/numberChunks;
    }

    /**
     * These notify methods are to be overridden by a subclass that 
     * wants to track the progress of the request
     */
    public void notifyDone() {
    }

    public void notifyFailure() {
    }

    public void notifyUpdate(int percentage) {
    }

    public File getFile() {
        return destFile;
    }

    private void sendRequest() throws Exception {
        // create the message
        Message message = pipe.createMessage();

        message.push(CMS.MESSAGE_TYPE, new ByteArrayInputStream(CMS.GET_REQUEST.getBytes()));
        message.push(CMS.REQUEST_ID, new ByteArrayInputStream(rid.getBytes()));
        message.push(CMS.CONTENT_ID, new ByteArrayInputStream(cid.toString().getBytes()));
        message.push(CMS.PIPE_ADVERTISEMENT, resultPipe.getDocument(CMS.encodeAs).getStream());

        // Create output pipe
        OutputPipe outPipe = null;

        outPipe = pipe.createOutputPipe(requestPipe, Pipe.NonBlocking, CMS.pipeTimeout);
        outPipe.send(message);
        outPipe.close();
    }

    private void handleMessage(Message message) {
        long length = 0;
        long offset = 0;
        int size = 0;

        // extract the content length
        try {
            length = popLong(message, CMS.CONTENT_LENGTH);
        } catch (IOException e) {
	  //e.printStackTrace();
            return;
        } catch (NumberFormatException e) {
	  //e.printStackTrace();
            return;
        }

        // extract the content offset
        try {
            offset = popLong(message, CMS.CHUNK_OFFSET);
        } catch (IOException e) {
	  //e.printStackTrace();
            return;
        } catch (NumberFormatException e) {
	  //e.printStackTrace();
            return;
        }

        // extract the chunk size
        try {
            size = popInt(message, CMS.CHUNK_SIZE);
        } catch (IOException e) {
	  //e.printStackTrace();
            return;
        } catch (NumberFormatException e) {
	  //e.printStackTrace();
            return;
        }

        // check if the bitSet has been created
        if (bitSet == null) {
            numberChunks = (int)(length / size);
            // check for case where there is a particle last chunk

            if (size * numberChunks < length) {
                numberChunks++;
            }

            // this is the first chunk
            bitSet = new BitSet(numberChunks);

            contentLength = length;
            chunkSize = size;
        }

        // check if this chunk has been received already
        int chunkNumber = (int)(offset / chunkSize);
        if (bitSet.get(chunkNumber)) {
            // already seen this chunk
            return;
        }

        // set the chunk as seen
        bitSet.set(chunkNumber);

        InputStream chunkStream = null;
        try {
            chunkStream = message.pop(CMS.CHUNK_DATA);
        } catch (IOException e) {
	  //e.printStackTrace();
            return;
        }

        RandomAccessFile op = null;
        try {
            op = new RandomAccessFile(destFile, "rw");
        } catch (Exception e) {
	  //e.printStackTrace();
            return;
        }

        try {
            int available = chunkStream.available();
            byte[] buffer = new byte[available];

            int res = chunkStream.read(buffer);
            op.seek(offset);
            op.write(buffer, 0, res);
        } catch (IOException e) {
	  //e.printStackTrace();
            return;
        }

        try {
            op.close();
            chunkStream.close();
        } catch (IOException e) {
	  //e.printStackTrace();
        }

        numberReceived++;
        if (numberReceived >= numberChunks) {
            isDone = true;
        } else {
            notifyUpdate(getPercentDone());
        }
    }

    private static long popLong(Message message, String tag) throws IOException, NumberFormatException {
        String str = CMS.popString(message, tag);
        return Long.parseLong(str);
    }

    private static int popInt(Message message, String tag) throws IOException, NumberFormatException {
        String str = CMS.popString(message, tag);
        return Integer.parseInt(str);
    }
}
