/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: GetMessageProcessor.java,v 1.1 2001/05/15 15:37:01 chaywood Exp $
 *
 */

package net.jxta.share;

import java.lang.reflect.InvocationTargetException;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.Vector;

import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.endpoint.Message;
import net.jxta.pipe.OutputPipe;
import net.jxta.pipe.Pipe;
import net.jxta.protocol.PipeAdvertisement;

/**
 * This class implements Get Message Processor.
 * It handles all Get requests from clients for this service.
 */

public class GetMessageProcessor extends MessageProcessor {

    private static int defaultChunkSize = 64*1024;

    public GetMessageProcessor(CMS cms) {
        super(cms);
    }

    protected void process(Message message) {
        // get result PipeAdvertisement
        InputStream pipeAdvStream = null;
        PipeAdvertisement resultPipe = null;
        String rid = null;
        String cid = null;

        try {
            pipeAdvStream = message.pop(CMS.PIPE_ADVERTISEMENT);
            rid = CMS.popString(message, CMS.REQUEST_ID);
            cid = CMS.popString(message, CMS.CONTENT_ID);
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }

        try {
            Advertisement adv = 
                AdvertisementFactory.newAdvertisement(CMS.encodeAs, pipeAdvStream);
            if (adv instanceof PipeAdvertisement) {
                resultPipe = (PipeAdvertisement)adv;
            }
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            return;
        }

        // Create output pipe
        OutputPipe outPipe = null;
        try {
            outPipe = getPipe().createOutputPipe(resultPipe, Pipe.NonBlocking, CMS.pipeTimeout);
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }

        // find the content
        Content content = getContent(cid);
        if (null == content) {
            return;
        }

        // get the InputStream for the content
        InputStream contentStream = null;
        try {
            contentStream = content.getInputStream();
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }

        long clength = content.getContentAdvertisement().getLength();
        String clengthStr = Long.toString(clength);        
        String chunkSizeStr = Integer.toString(defaultChunkSize);

        // send a chunk of the file at a time
        long offset = 0;
        int bytesRead = 0;
        byte[] buffer;
        while (offset < clength) {
            // create a new result message
            Message resultMsg = getPipe().createMessage();

            bytesRead = 0;
            try {
                int avail = contentStream.available();
                if (avail < defaultChunkSize) {
                    buffer = new byte[avail];
                } else {
                    buffer = new byte[defaultChunkSize];
                }
                bytesRead = contentStream.read(buffer);
            } catch (IOException e) {
                e.printStackTrace();
                break;
            }

            String offsetStr = Long.toString(offset);
            offset += bytesRead;

            try {
                resultMsg.push(CMS.MESSAGE_TYPE, new ByteArrayInputStream(CMS.GET_RESULT.getBytes()));
                resultMsg.push(CMS.REQUEST_ID, new ByteArrayInputStream(rid.getBytes()));
                resultMsg.push(CMS.CONTENT_ID, new ByteArrayInputStream(cid.getBytes()));
                resultMsg.push(CMS.CONTENT_LENGTH, new ByteArrayInputStream(clengthStr.getBytes()));
                resultMsg.push(CMS.CHUNK_OFFSET, new ByteArrayInputStream(offsetStr.getBytes()));
                resultMsg.push(CMS.CHUNK_SIZE, new ByteArrayInputStream(chunkSizeStr.getBytes()));
                resultMsg.push(CMS.CHUNK_DATA, new ByteArrayInputStream(buffer));
            } catch (IOException e) {
                e.printStackTrace();
                return;
            }

            // send the result
            try {
                outPipe.send(resultMsg);
                outPipe.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        // close the output pipe
        if (null != outPipe) {
            outPipe.close();
            outPipe = null;
        }
    }

    private Content getContent(String cids) {
        ContentId cid;

        try {
            cid = new ContentIdImpl(cids);
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
            return null;
        }

        Content[] clist = getContentManager().getContent(cid);
        if (clist != null && clist.length > 0) {
            return clist[0];
        }

        return null;
    }
}
