/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ListContentRequest.java,v 1.3 2001/05/25 17:50:23 kuldeep Exp $
 *
 */

package net.jxta.share;

import java.lang.reflect.InvocationTargetException;
import java.io.ByteArrayInputStream;
import java.io.IOException;

import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.endpoint.Message;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.PipeID;
import net.jxta.protocol.PipeAdvertisement;

/**
 * This class implements List Content Request.
 * It request a list of content from the given CMS request pipe.
 */

public class ListContentRequest implements Runnable {

    private Pipe pipe = null;
    private PipeAdvertisement requestPipe = null;
    private PipeAdvertisement resultPipe = null;
    private String rid = null;
    private String subString = null;

    private Thread thread = null;

    private boolean isDone = false;
    private ContentAdvertisement[] result = null;

    public ListContentRequest(PeerGroup group, PipeAdvertisement inRequest, 
            String inSubStr) throws InvocationTargetException {
        pipe = group.getPipe();
        requestPipe = inRequest;
        subString = inSubStr;

        rid = new UUID().toString();

        // create a new PipeAdvertisement
        Advertisement adv = 
            AdvertisementFactory.newAdvertisement(PipeAdvertisement.getAdvertisementType());
        if (adv instanceof PipeAdvertisement) {
            resultPipe = (PipeAdvertisement)adv;
        }
        resultPipe.setPipeID(new PipeID(group.getID()));
        resultPipe.setName("ListContentRequest");

        thread = new Thread(this);
        thread.start();
    }

    public ListContentRequest(PeerGroup group, PipeAdvertisement inRequest, 
            PipeAdvertisement inResult, String inSubStr) {
        pipe = group.getPipe();
        requestPipe = inRequest;
        resultPipe = inResult;
        subString = inSubStr;

        rid = new UUID().toString();

        Thread thread = new Thread(this);
        thread.start();
    }

    public void run() {
        // create the input pipe
        InputPipe inPipe = null;
        try {
            inPipe = pipe.createInputPipe(resultPipe);
        } catch (IOException e) {
            System.err.println("Could not create InputPipe");
            e.printStackTrace();
        }        

        // send the request
        try {
            sendRequest();
        } catch(Exception e) {
	    //e.printStackTrace();
	  //isDone = true;
	  //XXX To check with Carl on this change
            notifyFailure();
	    if (null != inPipe) {
	      inPipe.close();
	      inPipe = null;
	    }
	    return;
        }

        // wait for result
        Message message = null;
        while (!isDone) {
            try {
                message = inPipe.waitForMessage();
            } catch (IOException e) {
                inPipe.close();
                break;
            }

            if (message == null) {
                if (Thread.interrupted()) {
                    // We have been asked to stop
                    System.err.println("ListContentRequest.run() terminated");
                    isDone = true;
                    break;
                }
            } else {
                // Handle the message
                String messageType = null;
                String id = null;
                try {
                    messageType = CMS.popString(message, CMS.MESSAGE_TYPE);
                    id = CMS.popString(message, CMS.REQUEST_ID);
                } catch (IOException e) {
                    e.printStackTrace();
                }

                if (CMS.LIST_RESULT.equals(messageType) && rid.equals(id)) {
                    handleMessage(message);
                }
            }
        }

        if (null != inPipe) {
            inPipe.close();
            inPipe = null;
        }

        // call the done notify routine
        notifyDone();
    }

    public boolean isDone() {
        return isDone;
    }

    public void notifyDone() {
    }

    public void notifyFailure() {
    }

    public void cancel() {
        isDone = true;
        if (null != thread) {
            thread.interrupt();
        }
    }

    public ContentAdvertisement[] getResults() {
        return result;
    }

    public PipeAdvertisement getRequestPipe() {
        return requestPipe;
    }

    private void sendRequest() throws Exception {
        // create the message
        Message message = pipe.createMessage();

        message.push(CMS.MESSAGE_TYPE, new ByteArrayInputStream(CMS.LIST_REQUEST.getBytes()));
        message.push(CMS.REQUEST_ID, new ByteArrayInputStream(rid.getBytes()));
        message.push(CMS.PIPE_ADVERTISEMENT, resultPipe.getDocument(CMS.encodeAs).getStream());
        if (null != subString) {
            message.push(CMS.QUERY_SUBSTRING, new ByteArrayInputStream(subString.getBytes()));
        }

        // Create output pipe
        OutputPipe outPipe = null;

        outPipe = pipe.createOutputPipe(requestPipe, Pipe.NonBlocking, CMS.pipeTimeout);
        if ( outPipe != null ) {
	  outPipe.send(message);
	  outPipe.close();
	}
    }

    private void handleMessage(Message message) {
        String countStr = null;
        int count = 0;

        try {
            countStr = CMS.popString(message, CMS.ADVERTISEMENT_COUNT);
        } catch (IOException e) {
            e.printStackTrace();
        }

        try {
            count = Integer.parseInt(countStr);
        } catch (NumberFormatException e) {
            e.printStackTrace();
        }

        isDone = true;

        result = new ContentAdvertisement[count];

        // copy the result ContentAdvertisements to the result array
        String tag;
        for (int i=0; i<count; i++) {
            tag = CMS.CONTENT_ADVERTISEMENT + i;

            try {
                result[i] = new ContentAdvertisementImpl(message.pop(tag));
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }
}
