/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ListMessageProcessor.java,v 1.3 2001/05/25 19:53:55 kuldeep Exp $
 *
 */

package net.jxta.share;

import java.lang.reflect.InvocationTargetException;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.Vector;

import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.endpoint.Message;
import net.jxta.pipe.OutputPipe;
import net.jxta.pipe.Pipe;
import net.jxta.protocol.PipeAdvertisement;

/**
 * This class implements List Message Processor.
 * It handles all List requests from clients for this service.
 */

public class ListMessageProcessor extends MessageProcessor {

    private String subString = null;

    public ListMessageProcessor(CMS cms) {
        super(cms);
    }

    protected void process(Message message) {
        // get result PipeAdvertisement
        InputStream pipeAdvStream = null;
        PipeAdvertisement resultPipe = null;
        try {
            pipeAdvStream = message.pop(CMS.PIPE_ADVERTISEMENT);
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }

        try {
            Advertisement adv = 
                AdvertisementFactory.newAdvertisement(CMS.encodeAs, pipeAdvStream);
            if (adv instanceof PipeAdvertisement) {
                resultPipe = (PipeAdvertisement)adv;
            }
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            return;
        }

        // create a new result message
        Message resultMsg = getPipe().createMessage();

        // get the list of shared content
        Content[] clist = getContentManager().getContent();

        // load the query parameters of the list request
        getQueryParameters(message);

        int count = 0;
        ContentAdvertisement adv = null;
        String tag;
        for (int i = 0; i < clist.length; i++) {
            // get the content advertisement
            adv = clist[i].getContentAdvertisement();

            // check if the advertisement matches the query
            if (isMatch(adv)) {
                tag = CMS.CONTENT_ADVERTISEMENT + count++;
                
                try {
                    resultMsg.push(tag, adv.getDocument(CMS.encodeAs).getStream());
                } catch (Exception e) {
                    //e.printStackTrace();
                }
            }
        }

        try {
            resultMsg.push(CMS.MESSAGE_TYPE, new ByteArrayInputStream(CMS.LIST_RESULT.getBytes()));
            resultMsg.push(CMS.REQUEST_ID, message.pop(CMS.REQUEST_ID));
            resultMsg.push(CMS.ADVERTISEMENT_COUNT, new ByteArrayInputStream(Integer.toString(count).getBytes()));
        } catch (IOException e) {
            //e.printStackTrace();
            return;
        }

        // send the result
        try {
            // Create output pipe
            OutputPipe outPipe = null;

            outPipe = getPipe().createOutputPipe(resultPipe, Pipe.NonBlocking, CMS.pipeTimeout);
            outPipe.send(resultMsg);
            outPipe.close();
        } catch (IOException e) {
            //e.printStackTrace();
        }
    }

    protected void getQueryParameters(Message message) {
        try {
            subString = CMS.popString(message, CMS.QUERY_SUBSTRING);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    protected boolean isMatch(ContentAdvertisement cadv) {
        if (subString == null) {
            return true;
        }

        // check if the name contains the substring
        String name = cadv.getName();
        if (name != null && name.indexOf(subString) != -1) {
            return true;
        }

        // check if the name contains the substring
        String desc = cadv.getDescription();
        if (desc != null && desc.indexOf(subString) != -1) {
            return true;
        }
        
        return false;
    }
}
