/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ContentInfoImpl.java,v 1.1 2001/04/15 05:10:18 dac Exp $
 *
 */

package net.jxta.cms.impl;

import net.jxta.cms.ContentInfo;

import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.Document;
import net.jxta.document.TextDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentUtils;
import net.jxta.document.TextElement;
import net.jxta.document.Element;
import net.jxta.document.MimeMediaType;
import net.jxta.protocol.PipeAdvertisement;

import java.io.InputStream;
import java.io.IOException;
import java.io.Serializable;
import java.io.StringWriter;
import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.lang.reflect.InvocationTargetException;

/*
 * ContentInfo implementation for the ContentManager service.
 */
class ContentInfoImpl extends ContentInfo {
    private String name;	    // content name (required)
    private String cid;		    // content identifier (required)
    private PipeAdvertisement pipe; // content request pipe adv (optional)
    private long length = -1;	    // content length (optional)
    private String type;	    // content type (optional)
    private String description;     // content description (optional)

    private static final String DOC_ROOT_NAME = "jxta:ContentInfo";
    private static final String DOC_EN_NAME = "name";
    private static final String DOC_EN_CID = "cid";
    private static final String DOC_EN_PIPE = "pipe";
    private static final String DOC_EN_TYPE = "type";
    private static final String DOC_EN_LENGTH = "length";
    private static final String DOC_EN_DESCRIPTION = "description";

    private static final MimeMediaType MMT_TEXT_XML =
	new MimeMediaType("text", "xml");

    /*
     * Creates a new ContentInfoImpl object for the specified content id,
     * name, and optional pipe advertisement, length, type, and description
     * @param name content name
     * @param cid content id
     * @param pipe content request pipe advertisement, or null if none
     * @param length content length, or -1 if not specified
     * @param type content mime type, or null if not specified
     * @param description content description, or null if not specified
     */
    ContentInfoImpl(String name, String cid, PipeAdvertisement pipe,
		    long length, String type, String description)
    {
	if (name == null) {
	    throw new NullPointerException("name");
	}
	if (cid == null) {
	    throw new NullPointerException("cid");
	}
	this.name = name;
	this.cid = cid;
	this.pipe = pipe;
	this.length = length;
	this.type = type;
	this.description = description;
    }

    /**
     * Creates a new ContentInfoImpl object from the specified InputStream.
     * @param is the InputStream containing the advertisement
     * @exception IllegalArgumentException if the advertisement was invalid
     * @exception IOException if there was an I/O error
     */
    public ContentInfoImpl(InputStream is) throws IOException {
	StructuredTextDocument doc;
	try {
	    doc = (StructuredTextDocument)
		StructuredDocumentFactory.newStructuredDocument(
		    MMT_TEXT_XML, is);
	} catch (InvocationTargetException e) {
	    Throwable t = e.getTargetException();
	    if (t instanceof IOException) {
		throw (IOException)t;
	    } else {
		t.printStackTrace(); // DEBUG
		throw new IllegalArgumentException(
		    "Invalid content advertisement format: " + t);
	    }
	} catch (ClassCastException e) {
	    throw new InternalError("StructuredDocumentFactory is broken");
	}
	init(doc);
    }

    /**
     * Creates a new ContentInfoImpl object from the specified TextElement.
     */
    public ContentInfoImpl(TextElement te) {
	init(te);
    }

    // reads advertisement from specified TextElement
    private void init(TextElement root) {
	if (!root.getName().equals(DOC_ROOT_NAME)) {
	    throw new IllegalArgumentException(
		"Not a ContentInfo document: " + root.getName());
	}
	Enumeration e = root.getChildren();
	while (e.hasMoreElements()) {
	    TextElement te = (TextElement)e.nextElement();
	    String s = te.getName();
	    if (s.equals(DOC_EN_NAME)) {
		name = te.getTextValue();
	    } else if (s.equals(DOC_EN_CID)) {
		cid = te.getTextValue();
	    } else if (s.equals(DOC_EN_PIPE)) {
		try {
		    pipe = (PipeAdvertisement)
			AdvertisementFactory.newAdvertisement(
			    (TextElement)te.getValue());
		} catch (Exception ex) {
		    throw new IllegalArgumentException(
			"Invalid pipe advertisement: " + te.getTextValue());
		}
	    } else if (s.equals(DOC_EN_LENGTH)) {
		try {
		    length = Long.parseLong(te.getTextValue());
		} catch (Exception ex) {
		    throw new IllegalArgumentException(
			"Invalid content length: " + te.getTextValue());
		}
	    } else if (s.equals(DOC_EN_TYPE)) {
		type = te.getTextValue();
	    } else if (s.equals(DOC_EN_DESCRIPTION)) {
		description = te.getTextValue();
	    }
	}
	if (name == null) {
	    throw new IllegalArgumentException(
		"Missing required 'name' element");
	}
	if (cid == null) {
	    throw new IllegalArgumentException(
		"Missing required 'cid' element");
	}
    }

    /**
     * Returns the name of the content.
     */
    public String getName() {
	return name;
    }

    /**
     * Returns the content id.
     */
    public String getID() {
	return cid;
    }

    /**
     * Returns the request pipe advertisement.
     */
    public PipeAdvertisement getPipeAdvertisement() {
	return pipe;
    }

    /**
     * Returns the length of the content, or -1 if unknown
     */
    public long getLength() {
	return length;
    }

    /**
     * Returns the mime type of the content, or null if unknown
     */
    public String getType() {
	return type;
    }

    /**
     * Returns the description of the content or null if none.
     */
    public String getDescription() {
	return description;
    }

    /**
     * Returns the document representation of this advertisement.
     */
    public Document getDocument(MimeMediaType mtype) {
	StructuredDocument doc;
	try {
	    doc = (StructuredDocument)
		StructuredDocumentFactory.newStructuredDocument(
		    mtype, DOC_ROOT_NAME);
	} catch (Exception e) {
	    throw new IllegalStateException(
		"Error while generating document: " + e);
	}
	appendElements(doc, doc);
	return doc;
    }

    /**
     * Appends this content advertisement to the specified structured
     * document element.
     */ 
    public void appendDocument(StructuredDocument doc, Element el) {
	Element e = doc.createElement(DOC_ROOT_NAME);
	el.appendChild(e);
	appendElements(doc, e);
    }

    // append elements to specified document element
    private void appendElements(StructuredDocument doc, Element e) {
	e.appendChild(doc.createElement(DOC_EN_NAME, name));
	e.appendChild(doc.createElement(DOC_EN_CID, cid));
	if (pipe != null) {
	    StructuredDocument adv;
	    try {
		adv = (StructuredDocument)pipe.getDocument(MMT_TEXT_XML);
	    } catch (Exception ex) {
		ex.printStackTrace(); // DEBUG
		throw new IllegalStateException(
		    "Error creating pipe advertisement: " + ex);
	    }
	    StructuredDocumentUtils.copyElements(doc, e, adv);
	}
	if (length != -1) {
	    e.appendChild(doc.createElement(DOC_EN_LENGTH,
					    Long.toString(length)));
	}
	if (type != null) {
	    e.appendChild(doc.createElement(DOC_EN_TYPE, type));
	}
	if (description != null) {
	    e.appendChild(doc.createElement(DOC_EN_DESCRIPTION, description));
	}
    }

    /**
     * Returns a string representation of this advertisement.
     */
    public String toString() {
	TextDocument doc = (TextDocument)getDocument(MMT_TEXT_XML);
	StringWriter sw = new StringWriter();
	try {
	    doc.sendToWriter(sw);
	} catch (IOException e) {
	    throw new IllegalStateException(
		"I/O error while writing document: " + e);
	}
	return sw.toString();
    }
}
