/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ContentRequestImpl.java,v 1.2 2001/04/13 03:21:21 dac Exp $
 *
 */

package net.jxta.cms.impl;

import net.jxta.cms.ContentRequest;

import net.jxta.document.AdvertisementFactory;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.TextElement;
import net.jxta.document.MimeMediaType;
import net.jxta.protocol.PipeAdvertisement;

import java.io.InputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.lang.reflect.InvocationTargetException;

/*
 * ContentRequest implementation which parses a content request
 * message.
 */
class ContentRequestImpl extends ContentRequest {
    private String rid; // request id
    private String cid;	// content id
    private PipeAdvertisement pipe; // response pipe advertisement

    static final String MSG_DATA_NAME = "ContentRequest";
    private static final String DOC_ROOT_NAME = "jxta:ContentRequest";
    private static final String DOC_EN_RID = "rid";
    private static final String DOC_EN_CID = "cid";
    private static final String DOC_EN_PIPE = "pipe";

    private static final MimeMediaType MMT_TEXT_XML =
	new MimeMediaType("text", "xml");

    /*
     * Contructs a new ContentRequest by parsing the request message
     * from the specified input stream.
     * @exception IOException if there was an I/O error
     * @exception IllegalArgumentException if the message was invalid
     */
    ContentRequestImpl(InputStream is) throws IOException {
	StructuredTextDocument doc;
	try {
	    doc = (StructuredTextDocument)
		StructuredDocumentFactory.newStructuredDocument(
		    MMT_TEXT_XML, is);
	} catch (InvocationTargetException e) {
	    Throwable t = e.getTargetException();
	    if (t instanceof IOException) {
		throw (IOException)t;
	    } else {
		t.printStackTrace(); // DEBUG
		throw new IllegalArgumentException(
		    "Invalid content request message format: " + t);
	    }
	}
	init(doc);
    }

    /*
     * Contructs a ContentRequest from the specified document element.
     */
    ContentRequestImpl(TextElement root) {
	init(root);
    }

    // parses request from specified text element
    private void init(TextElement root) {
	if (!root.getName().equals(DOC_ROOT_NAME)) {
	    throw new IllegalArgumentException(
		"Not a ContentRequest root element: " + root.getName());
	}
	Enumeration e = root.getChildren();
	while (e.hasMoreElements()) {
	    TextElement te = (TextElement)e.nextElement();
	    String name = te.getName();
	    if (name.equals(DOC_EN_RID)) {
		rid = te.getTextValue();
	    } else if (name.equals(DOC_EN_CID)) {
		cid = te.getTextValue();
	    } else if (name.equals(DOC_EN_PIPE)) {
		try {
		    pipe = (PipeAdvertisement)
			AdvertisementFactory.newAdvertisement(te);
		} catch (Exception ex) {
		    ex.printStackTrace(); // DEBUG
		    throw new IllegalArgumentException(
			"Invalid pipe advertisement: " + te.getTextValue());
		}
	    }
	}
	if (rid == null) {
	    throw new IllegalArgumentException(
		"Missing required '" + DOC_EN_RID + "' element");
	}
	if (cid == null) {
	    throw new IllegalArgumentException(
		"Missing required '" + DOC_EN_CID + "' element");
	}
	if (pipe == null) {
	    throw new IllegalArgumentException(
		"Missing required '" + DOC_EN_PIPE + "' element");
	}
    }

    /**
     * Returns the request id.
     */
    public String getRequestID() {
	return rid;
    }

    /**
     * Returns the content id.
     */
    public String getContentID() {
	return cid;
    }

    /**
     * Returns the response pipe advertisement.
     */
    public PipeAdvertisement getPipeAdvertisement() {
	return pipe;
    }
}
