/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ContentResponseImpl.java,v 1.2 2001/04/13 03:21:21 dac Exp $
 *
 */

package net.jxta.cms.impl;

import net.jxta.cms.ContentResponse;

import net.jxta.endpoint.Message;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.document.Document;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.MimeMediaType;

import java.io.InputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;

/*
 * ContentResponse implementation for ContentServiceImpl.
 */
class ContentResponseImpl extends ContentResponse {
    private Pipe pipe;		// pipe service
    private String rid;	        // request id
    private long offset;        // response data offset
    private int length;	        // response data length
    private InputStream is;     // response data input stream

    private static final String MSG_DATA_NAME = "ContentResponse";
    private static final String DOC_ROOT_NAME = "jxta:ContentResponse";
    private static final String DOC_EN_RID = "rid";
    private static final String DOC_EN_OFFSET = "offset";
    private static final String DOC_EN_LENGTH = "length";

    private static final MimeMediaType MMT_TEXT_XML =
	new MimeMediaType("text", "xml");

    /*
     * Creates a new ContentResponseImpl for the specified ContentServiceImpl
     * and request id.
     */
    ContentResponseImpl(ContentServiceImpl cs) {
	pipe = cs.getPeerGroup().getPipe();
    }

    /**
     * Sets the request id.
     */
    public void setRequestID(String rid) {
	this.rid = rid;
    }

    /**
     * Sets the offset of the response data.
     */
    public void setOffset(long offset) {
	this.offset = offset;
    }

    /**
     * Sets the length of the response data.
     */
    public void setLength(int length) {
	this.length = length;
    }

    /**
     * Sets the input stream which will be used to read the response data.
     */
    public void setInputStream(InputStream is) {
	this.is = is;
    }

    /**
     * Sends this response to the specified output pipe.
     */
    public void send(OutputPipe op) throws IOException {
	Message msg = pipe.createMessage();
	Document doc = getDocument(MMT_TEXT_XML);
	msg.push(MSG_DATA_NAME, doc.getStream());
	msg.push(rid, is);
	op.send(msg);
    }

    // Returns the document for this response message
    private Document getDocument(MimeMediaType type) {
	StructuredDocument doc;
	try {
	    doc = (StructuredDocument)
		StructuredDocumentFactory.newStructuredDocument(
		    type, DOC_ROOT_NAME);
	} catch (InvocationTargetException e) {
	    Throwable t = e.getTargetException();
	    t.printStackTrace(); // DEBUG
	    throw new IllegalStateException(
		"Error creating content response document: " + t);
	}
	doc.appendChild(doc.createElement(DOC_EN_RID, rid));
	doc.appendChild(doc.createElement(DOC_EN_OFFSET,
					  Long.toString(offset)));
	doc.appendChild(doc.createElement(DOC_EN_LENGTH,
					  Integer.toString(length)));
	return doc;
    }
}
