/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ContentServiceImpl.java,v 1.3 2001/04/14 01:27:28 dac Exp $
 *
 */

package net.jxta.cms.impl;

import net.jxta.cms.Content;
import net.jxta.cms.ContentService;
import net.jxta.cms.ContentRequest;
import net.jxta.cms.ContentResponse;
import net.jxta.cms.ContentHandler;
import net.jxta.cms.ContentEvent;
import net.jxta.cms.ContentListener;
import net.jxta.cms.util.FileReader;
import net.jxta.cms.util.EventDispatcher;

import net.jxta.service.Service;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.InputPipe;
import net.jxta.endpoint.Message;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.EventObject;
import java.util.EventListener;

/*
 * The ContentService is used for sharing arbitrary content within a peer
 * group. Each Content is reference by a unique content identifier string.
 */
public class ContentServiceImpl extends ContentService implements Runnable {
    private InputPipe requestPipe; // input pipe for request
    private Hashtable shares;	   // shared Content objects
    private File sharesFile;	   // shares index file
    private Vector handlers; 	   // active handler threads
    private boolean stopped;	   // true if service was stopped
    private ContentEventDispatcher dispatcher; // event dispatcher

    // shared content index file name
    private static final String SHARES_SER = "shares.ser";

    // max number of concurrent request handler threads
    private static final int MAX_THREADS = 8; 

    /**
     * Creates a new, uninitialized ContentServiceImpl object.
     */
    public ContentServiceImpl() {
    }

    /**
     * Initializes the ContentService with the specified peer group,
     * request input pipe, and content store directory.
     */
    public void init(InputPipe requestPipe, File dir)
	throws IOException
    {
	if (!dir.isDirectory()) {
	    throw new IOException("Directory not found: " + dir);
	}
	this.requestPipe = requestPipe;
	sharesFile = new File(dir, SHARES_SER);
	if (sharesFile.isFile()) {
	    load();
	} else {
	    shares = new Hashtable();
	}
	handlers = new Vector();
	dispatcher = new ContentEventDispatcher();
    }

    /**
     * Starts the content service.
     */
    public void start() {
	dispatcher.start();
    }

    /**
     * Stops the content service.
     */
    public void stop() {
	stopped = true;
	dispatcher.stop();
    }

    /**
     * Shares the specified ContentHandler and returns its associated
     * Content object.
     */
    public Content share(ContentHandler handler) {
	Content content = new ContentImpl(handler);
	String id = content.getContentID();
	synchronized (this) {
	    shares.put(id, content);
	    try {
		save();
	    } catch (IOException e) {
		// If there was an error while trying to save the index
		// then restore the original index.
		shares.remove(id);
		e.printStackTrace();
		throw new IllegalStateException(
		    "Error while saving shared content index");
	    }
	    content.getContentHandler().shared(this);
	}
	dispatcher.post(content, ContentEvent.SHARED, null);
	return content;
    }

    /**
     * Unshares the specified Content object.
     * @exception IllegalArgumentException
     *            if the specified content object is not currently shared
     */
    public void unshare(Content content) {
	synchronized (this) {
	    String id = content.getContentID();
	    Content old = (Content)shares.remove(id);
	    if (content != old) {
		// if content not found or stored object is different
		shares.put(id, old);
		throw new IllegalArgumentException(
		    "Content not shared: " + content);
	    }
	    try {
		save();
	    } catch (IOException e) {
		// If there was an error while trying to save the index
		// then restore the original index.
		shares.put(id, old);
		e.printStackTrace();
		throw new IllegalStateException(
		    "Error while saving shared content index");
	    }
	    content.getContentHandler().unshared();
	}
	dispatcher.post(content, ContentEvent.UNSHARED, null);
    }

    /**
     * If the specified content id is shared, then return its corresponding
     * content object, otherwise return null.
     */
    public Content getContent(String id) {
	synchronized (this) {
	    return (Content)shares.get(id);
	}
    }

    /**
     * Returns an Enumeration of currently shared Content objects.
     */
    public Enumeration getContent() {
	synchronized (this) {
	    return shares.elements();
	}
    }

    /**
     * Returns the number of currently shared content items.
     */
    public int size() {
	synchronized (this) {
	    return shares.size();
	}
    }

    /**
     * Adds a new ContentListener.
     */
    public void addListener(ContentListener listener) {
	dispatcher.addListener(listener);
    }

    /**
     * Removes the specified ContentListener.
     */
    public void removeListener(ContentListener listener) {
	dispatcher.removeListener(listener);
    }

    /**
     * Creates a new ContentResponse object.
     */
    public ContentResponse newContentResponse() {
	return new ContentResponseImpl(this);
    }

    /**
     * Runs the request handler thread.
     */
    public void run() {
	while (!stopped) {
	    try {
		Message msg = requestPipe.waitForMessage();
		for (;;) {
		    InputStream is = msg.pop(ContentRequestImpl.MSG_DATA_NAME);
		    if (is == null) {
			break;
		    }
		    ContentRequest req = new ContentRequestImpl(is);
		    handleRequest(req);
		}
	    } catch (IOException e) {
		e.printStackTrace(); // DEBUG
		dispatcher.post(null, ContentEvent.ERROR, e);
	    }
	}
    }

    // handle content request
    private void handleRequest(ContentRequest request) {
	// first lookup the content object
	String cid = request.getContentID();
	Content content;
	synchronized (this) {
	    content = (Content)shares.get(cid);
	}
	if (content == null) {
	    System.err.println("Content not found for id: " + cid); // DEBUG
	    dispatcher.post(null, ContentEvent.ERROR,
			    new NoSuchElementException(
				"Content not found for id: " + cid));
	
	    return;
	}
	// if found, then create the handler thread
	Thread handler;
	synchronized (handlers) {
	    while (handlers.size() > MAX_THREADS) {
		try {
		    handlers.wait();
		} catch (InterruptedException e) {
		}
	    }
	    handler = new HandlerThread(request, content);
	    handlers.addElement(handler);
	}
	dispatcher.post(content, ContentEvent.REQUEST_STARTED, null);
	handler.start();
    }

    // content handler thread
    private class HandlerThread extends Thread {
	private ContentRequest request;
	private Content content;

	HandlerThread(ContentRequest request, Content content) {
	    this.request = request;
	    this.content = content;
	}

	public void run() {
	    ContentHandler handler = content.getContentHandler();
	    try {
		handler.handleRequest(request);
	    } catch (Throwable e) {
		e.printStackTrace(); // DEBUG
		dispatcher.post(content, ContentEvent.ERROR, e);
	    }
	    dispatcher.post(content, ContentEvent.REQUEST_ENDED, null);
	    synchronized (handlers) {
		handlers.remove(this);
		handlers.notify();
	    }
	}
    }

    // saves content index to file
    private void save() throws IOException {
	File tmp = tmpFile(sharesFile);
	BufferedOutputStream os =
	    new BufferedOutputStream(new FileOutputStream(tmp));
	try {
	    ObjectOutputStream oos = new ObjectOutputStream(os);
	    oos.writeObject(shares);
	    oos.close();
	} catch (IOException e) {
	    os.close();
	    tmp.delete();
	    throw e;
	}
	sharesFile.delete();
	if (!tmp.renameTo(sharesFile)) {
	    throw new IOException("Could not rename file: " + tmp);
	}
    }

    // load content index from file
    private void load() throws IOException {
	// load the content index file
	BufferedInputStream is =
	    new BufferedInputStream(new FileInputStream(sharesFile));
	try {
	    ObjectInputStream ois = new ObjectInputStream(is);
	    shares = (Hashtable)ois.readObject();
	    ois.close();
	} catch (IOException e) {
	    is.close();
	    throw e;
	} catch (ClassNotFoundException e) {
	    e.printStackTrace();
	    throw new IllegalStateException(
		"Deserialization of shared content index failed");
	}
	// reshare all the content handlers
	Enumeration e = shares.elements();
	while (e.hasMoreElements()) {
	    Content content = (Content)e.nextElement();
	    content.getContentHandler().shared(this);
	}
    }

    // generate temp file name
    private static File tmpFile(File file) {
	return new File(file.getPath() + ".tmp");
    }

    // content event dispatcher
    private class ContentEventDispatcher extends EventDispatcher {
	ContentEventDispatcher() {
	    super("content event dispatcher");
	}
	// post content event only if there are listeners
	void post(Content c, int type, Throwable e) {
	    if (hasListeners()) {
		post(new ContentEvent(c, type, e));
	    }
	}
	// dispatch event to listener
	protected void dispatch(EventObject event, EventListener listener) {
	    ((ContentListener)listener).contentEvent((ContentEvent)event);
	}
    }
}
