/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: EventDispatcher.java,v 1.2 2001/04/14 02:56:52 dac Exp $
 *
 */

package net.jxta.cms.util;

import java.util.EventObject;
import java.util.EventListener;
import java.util.Vector;

/**
 * Generic event dispatcher. Can be used to implement almost any type of
 * asynchronous event dispatcher.
 */
public abstract class EventDispatcher {
    private Thread dispatcher;	 // dispatcher thread
    private Vector listeners;	 // event listeners
    private Queue queue;	 // event queue
    private int maxSize = -1;	 // maximum size of event queue
    private boolean nonBlocking; // true if non-blocking mode is enabled
    private boolean stopped;	 // true if the dispatcher has been stopped

    /**
     * Creates a new EventDispatcherThread using the specified
     * dispatcher thread name.
     * @param name the name for the dispatcher thread
     */
    public EventDispatcher(String name) {
	dispatcher = new DispatcherThread(name);
	listeners = new Vector();
	queue = new Queue();
    }

    /**
     * Creates a new EventDispatcher.
     */
    public EventDispatcher() {
	this("event dispatcher");
    }

    /**
     * Sets the maximum size of the event queue. By default, the event
     * queue has no maximum size.
     */
    public void setMaxSize(int maxSize) {
	this.maxSize = maxSize;
    }

    /**
     * Sets the non-blocking mode for posting events. If non-blocking
     * mode is enabled, then post() will return immediately if the
     * queue is full rather than waiting until the event can be
     * posted.
     */
    public void setNonBlocking(boolean nonBlocking) {
	this.nonBlocking = nonBlocking;
    }

    /**
     * Posts a new event to the dispatcher thread. The event
     * will be asynchronously delivered to all the registered
     * listeners. If the queue is currently full and non-blocking
     * mode has been set then this method will return false.
     * Otherwise, the method will block until the queue is no
     * longer full and the event can be posted.
     * @return true if the event could be posted, false otherwise
     */
    public boolean post(EventObject event) {
	int maxSize = this.maxSize;
	synchronized (queue) {
	    if (stopped) {
		return false;
	    }
	    if (maxSize != -1 && queue.size() >= maxSize) {
		if (nonBlocking) {
		    return false;
		}
		do {
		    try {
			queue.wait();
		    } catch (InterruptedException e) {
		    }
		    if (stopped) {
			return false;
		    }
		} while (queue.size() >= maxSize);
	    }
	    queue.add(event);
	    queue.notify();
	    return true;
	}
    }

    /**
     * Adds a new listener to this event dispatcher.
     */
    public void addListener(EventListener listener) {
	synchronized (listeners) {
	    if (!listeners.contains(listener)) {
		listeners.addElement(listener);
	    }
	}
    }

    /**
     * Removes a listener from this event dispatcher.
     */
    public void removeListener(EventListener listener) {
	synchronized (listeners) {
	    listeners.removeElement(listener);
	}
    }

    /**
     * Returns true if this event dispatcher currently has any listeners.
     */
    public boolean hasListeners() {
	synchronized (listeners) {
	    return listeners.isEmpty();
	}
    }

    /**
     * Starts the event dispatcher.
     */
    public void start() {
	if (stopped) {
	    throw new IllegalStateException("Event dispatcher already stopped");
	}
	dispatcher.start();
    }

    /**
     * Stops the event dispatcher. Any pending events will be discarded.
     */
    public void stop() {
	synchronized (queue) {
	    stopped = true;
	    queue.notify();
	}
    }

    /**
     * This method will be called in order to dispatch an event object
     * to an event listener. It should be overridden by implementations
     * to handle a more specific event type and listener object.
     * @param event the event to be delivered
     * @param listener the event listener
     */
    protected abstract void dispatch(EventObject event, EventListener listener);

    /*
     * Event dispatcher thread. Currently, we use a single thread for
     * dispatching events to all listeners.
     */
    private class DispatcherThread extends Thread {
	DispatcherThread(String name) {
	    super(name);
	    setDaemon(true);
	    //setPriority(MAX_PRIORITY);
	}
	public void run() {
	    for (;;) {
		EventObject event;
		synchronized (queue) {
		    if (stopped) {
			return;
		    }
		    while (queue.isEmpty()) {
			try {
			    queue.wait();
			} catch (InterruptedException e) {
			}
			if (stopped) {
			    return;
			}
		    }
		    event = (EventObject)queue.remove();
		    queue.notify();
		}
		synchronized (listeners) {
		    int len = listeners.size();
		    for (int i = 0; i < len; i++) {
			dispatch(event, (EventListener)listeners.elementAt(i));
		    }
		}
	    }
	}
    }
}
