/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: FileContentHandler.java,v 1.2 2001/04/13 03:21:21 dac Exp $
 *
 */

package net.jxta.cms.util;

import net.jxta.cms.ContentHandler;
import net.jxta.cms.ContentRequest;
import net.jxta.cms.ContentResponse;
import net.jxta.cms.ContentService;

import net.jxta.pipe.Pipe;
import net.jxta.pipe.OutputPipe;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.BufferedInputStream;
import java.io.IOException;

/**
 * FileContentHandler implementation for handling file content requests.
 */
public class FileContentHandler extends ContentHandler {
    private File file;
    private transient ContentService cs;
    private transient Pipe pipe;

    private static final int PIPE_TIMEOUT = 10000; // output pipe timeout
    private static final int CHUNK_SIZE = 64*1024; // data chunk size
    private static final int MAX_ACTIVE = 8;	   // max active file streams

    /**
     * Creates a new FileContentHandler for the specified File.
     * @param cs the ContentService under which this file will be shared
     * @param file the file to be shared
     */
    public FileContentHandler(File file) throws IOException {
	if (!file.isFile()) {
	    throw new FileNotFoundException(file.getPath());
	}
	this.file = file;
    }

    /**
     * The shared() method will be called when this Handler is shared.
     */
    public void shared(ContentService cs) {
	this.cs = cs;
	pipe = cs.getPeerGroup().getPipe();
    }

    /**
     * Handles file content request.
     */
    public void handleRequest(ContentRequest req) throws Exception {
	OutputPipe op = pipe.createOutputPipe(req.getPipeAdvertisement(),
					      Pipe.NonBlocking, PIPE_TIMEOUT);
	try {
	    send(file, req, op);
	} finally {
	    op.close();
	}
    }

    // Sends contents of specified file to output pipe
    private void send(File file, ContentRequest req, OutputPipe op)
	throws IOException
    {
	FileReader fr = new FileReader(file);
	String rid = req.getRequestID();
	long len = file.length();
	long off = 0;
	while (off < len) {
	    ContentResponse res = cs.newContentResponse();
	    res.setRequestID(rid);
	    res.setOffset(off);
	    int size = (int)Math.min((long)CHUNK_SIZE, len - off);
	    res.setLength(size);
	    BufferedInputStream is =
		new BufferedInputStream(fr.getInputStream(off, size));
	    res.setInputStream(is);
	    res.send(op);
	    if (fr.activeCount() > MAX_ACTIVE) {
		// Wait until enough enough active streams have been
		// cleared before continuing.
		synchronized (fr) {
		    while (fr.activeCount() > MAX_ACTIVE) {
			try {
			    fr.wait();
			} catch (InterruptedException e) {
			}
		    }
		}
	    }
	    off += size;
	}
	// Wait until all active streams have been cleared before returning.
	synchronized (fr) {
	    while (fr.activeCount() > 0) {
		try {
		    fr.wait();
		} catch (InterruptedException e) {
		}
	    }
	}
	fr.close();
    }
}
