/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: FileReader.java,v 1.2 2001/04/13 03:21:21 dac Exp $
 *
 */

package net.jxta.cms.util;

import java.util.Vector;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.InputStream;
import java.io.IOException;

/**
 * Utility class for simultaneously reading from multiple regions of a file.
 * Currently uses a single instance of RandomAccessFile but could be optimized
 * by using multiple instances, assuming there are enough available file
 * descriptors on the platform.
 */
public class FileReader {
    private File file;		  // file being read
    private RandomAccessFile raf; // random access file
    private Vector streams;	  // currently active streams
    private boolean closed;	  // true if file is closed

    /**
     * Creates a new FileReader for the specified file.
     * @exception IOException if the file could not be opened for reading
     */
    public FileReader(File file) throws IOException {
	this.file = file;
	raf = new RandomAccessFile(file, "r");
	streams = new Vector();
    }

    /**
     * Returns an input stream for reading the specified file region.
     * @param offset the offset of the region in bytes
     * @param length the size of the region in bytes
     * @exception IndexOutOfBoundsException
     *            if the specified region offset and/or length was invalid
     * @return an unbuffered input stream for reading the file region
     */
    public InputStream getInputStream(long offset, int length)
	throws IOException
    {
	if (offset < 0 || offset > file.length()) {
	    throw new IndexOutOfBoundsException("offset");
	}
	if (length < 0 || offset + length > file.length()) {
	    throw new IndexOutOfBoundsException("length");
	}
	InputStream is = new FileReaderInputStream(offset, length);
	synchronized (this) {
	    streams.addElement(is);
	}
	return is;
    }

    /**
     * Returns the number of currently active region input streams.
     * An input stream is considered still active if it has not been
     * completely read or has not been explicitly closed. This FileReader
     * object will be notified each time the active count is changed.
     */
    public int activeCount() {
	synchronized (this) {
	    return streams.size();
	}
    }

    /**
     * Closes the FileReader and closes any active streams.
     */
    public void close() throws IOException {
	synchronized (this) {
	    raf.close();
	    int count = streams.size();
	    for (int i = 0; i < count; i++) {
		FileReaderInputStream is =
		    (FileReaderInputStream)streams.elementAt(i);
		is.closed = true;
	    }
	    streams.clear();
	    closed = true;
	}
    }

    /*
     * The input stream for reading a region of bytes within the file.
     */
    private class FileReaderInputStream extends InputStream {
	private long offset;    // current file position
	private int length;     // remaining bytes for this stream
	boolean closed;		// true if stream is closed

	/*
         * Creates a new FileReaderInputStream for the specified region
         * offset and length in bytes.
         */
	FileReaderInputStream(long offset, int length) throws IOException {
	    this.offset = offset;
	    this.length = length;
	}

	/**
         * Reads from the file region into the specified byte array.
         * Returns the number of bytes read or -1 if the end of the
         * region was reached.
	 */
	public int read(byte[] b, int off, int len) throws IOException {
	    if (b == null) {
		throw new NullPointerException("b");
	    }
	    if (b.length == 0 || len == 0) {
		return 0;
	    }
	    if (off < 0 || off >= b.length) {
		throw new IndexOutOfBoundsException("off");
	    }
	    if (len < 0 || off + len > b.length) {
		throw new IndexOutOfBoundsException("len");
	    }
	    if (closed == true) {
		throw new IOException("Stream is closed");
	    }
	    if (length == 0) {
		return -1;
	    }
	    if (len > length) {
		len = length;
	    }
	    synchronized (FileReader.this) {
		if (FileReader.this.closed) {
		    throw new IOException("Stream is closed");
		}
		raf.seek(offset);
		raf.readFully(b, off, len);
		offset += len;
		length -= len;
		if (length == 0) {
		    if (!streams.remove(this)) {
			throw new IllegalStateException();
		    }
		    FileReader.this.notify();
		}
	    }
	    return len;
	}

	/**
	 * Reads a single byte from the file region. Returns the byte
         * read or -1 if the end of the region was reached.
         */
	public int read() throws IOException {
	    byte[] b = new byte[1];
	    return read(b, 0, 1) == 1 ? b[0] & 0xff : -1;
	}

	/**
	 * Skips over the specified number of bytes in the region. Returns
	 * the actual number of bytes skipped.
	 */
	public long skip(long n) throws IOException {
	    if (closed) {
		throw new IOException("Stream is closed");
	    }
	    if (length == 0) {
		return 0;
	    }
	    if (n > length) {
		n = length;
	    }
	    offset += n;
	    length -= n;
	    if (length == 0) {
		synchronized (FileReader.this) {
		    if (FileReader.this.closed) {
			throw new IOException("Stream is closed");
		    }
		    if (!streams.remove(this)) {
			throw new IllegalStateException();
		    }
		    FileReader.this.notify();
		}
	    }
	    return n;
	}

	/**
	 * Returns the number of bytes which can be read from the
	 * region without blocking.
	 */
	public int available() throws IOException {
	    if (closed) {
		throw new IOException("Stream is closed");
	    }
	    return length;
	}

	/**
	 * Closes the input stream.
	 */
	public void close() {
	    closed = true;
	    synchronized (FileReader.this) {
		if (!FileReader.this.closed) {
		    if (!streams.remove(this)) {
			throw new IllegalStateException();
		    }
		    FileReader.this.notify();
		}
	    }
	}
    }
}
