/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Queue.java,v 1.4 2001/04/14 02:56:52 dac Exp $
 *
 */

package net.jxta.cms.util;

import java.util.NoSuchElementException;

/**
 * This class implements a simple queue using an array of variable size.
 */
public class Queue {
    private Object[] entries;	    // queue entries
    private int hIndex, tIndex;	    // head, tail index
    private int size;		    // number of entries
    private int capacityIncrement;  // capacity increment

    /**
     * Creates a new Queue with the specified initial capacity
     * and capacity increment.
     * @param capacity the initial capacity of the queue
     * @param capacityIncrement the amount by which the capacity is
     *                          increased when the queue overflows
     */
    public Queue(int capacity, int capacityIncrement) {
	entries = new Object[capacity];
	this.capacityIncrement = capacityIncrement;
    }

    /**
     * Creates a new Queue with the specified initial capacity
     * and capacity increment of 0.
     * @param capacity the initial capacity of the queue
     */
    public Queue(int capacity) {
	this(capacity, 0);
    }

    /**
     * Creates a new Queue with initial capacity of <tt>10</tt>.
     */
    public Queue() {
	this(10);
    }

    /**
     * Adds an object to the tail of the queue.
     * @param obj the object to be added
     */
    public void add(Object obj) {
	if (size >= entries.length) {
	    ensureCapacity(size + 1);
	}
	entries[tIndex] = obj;
	tIndex = (tIndex + 1) % entries.length;
	size++;
    }

    /**
     * Removes the object at the head of the queue and returns it.
     * @exception NoSuchElementException if the queue is empty
     */
    public Object remove() {
	if (size <= 0) {
	    throw new NoSuchElementException("Queue is empty");
	}
	Object obj = entries[hIndex];
	hIndex = (hIndex + 1) % entries.length;
	--size;
	return obj;
    }

    /**
     * Returns the object which is at the head of the queue without
     * removing it.
     * @exception NoSuchElementException if the queue is empty
     */
    public Object peek() {
	if (size <= 0) {
	    throw new NoSuchElementException("Queue is empty");
	}
	return entries[hIndex];
    }

    /**
     * Returns the number of entries in the queue.
     */
    public int size() {
	return size;
    }

    /**
     * Returns true if the queue is empty.
     */
    public boolean isEmpty() {
	return size == 0;
    }

    /**
     * Removes all the elements from this queue.
     */
    public void clear() {
	hIndex = tIndex = size = 0;
    }

    /**
     * Increases the size of the queue, if necessary, to accomodate
     * the number of entries specified by the minCapacity argument.
     * If the current capacity of this queue is less than minCapacity
     * then its capacity will be increased. The capacity of the new
     * queue will be the old capacity plus the capacity increment,
     * unless the capacity increment is less than or equal to zero
     * or was never specified, in which case the new capacity will
     * be twice the old capacity. In any case, if the new capacity
     * is less than minCapacity, then the new capacity will be set
     * to minCapacity.
     * @param minCapacity the desired minimum capacity
     */
    public void ensureCapacity(int minCapacity) {
	if (minCapacity > entries.length) {
	    int capacity = capacityIncrement > 0 ?
		entries.length + capacityIncrement : entries.length * 2;
	    if (capacity < minCapacity) {
		capacity = minCapacity;
	    }
	    resize(capacity);
	}
    }

    /**
     * Trims the capacity of this queue to the queue's current size.
     */
    public void trimToSize() {
	if (size < entries.length) {
	    resize(size);
	}
    }

    /*
     * Resizes the queue to accomodate the specified number of
     * entries.
     */
    private void resize(int capacity) {
	Object[] tmp = new Object[capacity];
	if (size > 0) {
	    copyInto(tmp);
	}
	entries = tmp;
	hIndex = 0;
	tIndex = size;
    }
    
    /**
     * Copies the elements of this queue into the specified array,
     * starting with the entry at the head of the queue.
     * @param objs the array into which the elements are to be copied
     */
    public void copyInto(Object[] objs) {
	if (tIndex > hIndex) {
	    int count = tIndex - hIndex;
	    System.arraycopy(entries, hIndex, objs, 0, count);
	} else if (tIndex <= hIndex) {
	    int count = entries.length - hIndex;
	    System.arraycopy(entries, hIndex, objs, 0, count);
	    if (tIndex > 0) {
		System.arraycopy(entries, 0, objs, count, tIndex);
	    }
	}
    }

}
