#!/usr/bin/env python
#  This is Genny Test, an implementation of the Genny protocol.
#  Copyright (C) 2000 Tom Goulet
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#  
#  You can contact me at:  tomg@nova.yi.org
#  More information is available at:  http://www.gnutelladev.com/
"""
Protocol bytes:
0-3:  Client Identifier
4-7:  Server Identifier
8:  Broadcast Identifier
9:  Time To Live/Sequence
10-11:  Service
"""
import sys # I need to do this print can work with my funky stdout

from sys import version

from os.path import exists, isfile, expanduser, join, isdir
path_join = join ; del join

from os import remove, kill, waitpid, WNOHANG, mkdir

from socket import socket, AF_UNIX, SOCK_DGRAM, AF_INET, error, gethostname
socket_error = error ; del error

from errno import ECONNREFUSED

from string import strip, split, join

from select import select

from time import time

from struct import unpack, pack

from gennylib import hostorip2rawip, get_random_bytes, raw_send, hostify, \
	genny_conn, GENNY, RWELC, rebroadcast, PORT

from genny_basic import GENNY_CONTROL, GENNY_INFO, GENNY_ECHO, GENNY_DAYTIME, \
	GENNY_HOSTNAME, handle_command_echo, handle_command_update, \
	handle_command_announce, handle_command_daytime, handle_command_who, \
	do_foo_bars, handle_genny_control, handle_genny_info, \
	handle_genny_echo, handle_genny_daytime, handle_genny_hostname, \
	control, handle_command_myip, GENNY_CONTROL_NOSERVICE, \
	GENNY_CONTROL_TOOSMALL

from genny_chat import GENNY_CHAT, handle_command_nick, handle_command_say, \
	handle_command_act, handle_command_msgbeep, handle_genny_chat

from genny_file import GENNY_FILE, handle_command_find, \
	handle_command_response, handle_command_get, handle_command_lclear, \
	handle_command_share, handle_command_webserve, handle_genny_file, \
	handle_command_push, handle_command_webip, handle_command_webport, \
	handle_command_downloadpath

# Nonportable, two lines
from curses_ui import initialise_ui, display_prompt, handle_command_clear, \
	shutdown_ui

#from boring_ui import initialise_ui, display_prompt, handle_command_clear, \
#	shutdown_ui

STATUS = 'Genny Test v0.31 Ready -- type help for help'  # Our status line.
PROMPT = '> '				# Our prompt

# Sum up all the services for easy access.
GENNY_SERVICES = GENNY_CONTROL, GENNY_INFO, GENNY_FILE, GENNY_ECHO, \
	GENNY_DAYTIME, GENNY_CHAT, GENNY_HOSTNAME

def main():
	try:
		s, config, connections, identifiers = initialise()
		main_loop(s, config, connections, identifiers)
	except:
		if 's' not in dir():
			s = None
		if 'config' not in dir():
			config = None
		if 'connections' not in dir():
			connections = None
		shutdown(s, config, connections)
		raise

def initialise():
	#
	# Python 1.5.1 and sooner do not have <list_object>.pop
	if version[0:5] == '1.5.1':
		raise RuntimeError, 'Genny Test needs Python 1.5.2 or later'
	#
	# This dictionary will hold most runtime information, both static
	# and volatile
	config = {}
	#
	# Set up playground directory.
	playground = expanduser("~/.genny")
	if not isdir(playground):
		mkdir(playground)
	#
	# When genny forks it should close every file/socket in this list.
	config['closeme'] = []
	#
	# The webserver and fetching processes write logging information to
	# this socket
	logsockpath = path_join(playground, "log_sock")
	if exists(logsockpath):
		remove(logsockpath)
	# Nonportable, two lines.
	config['log_sock'] = socket(AF_UNIX, SOCK_DGRAM)
	config['log_sock'].bind(logsockpath)
	config['closeme'].append(config['log_sock'])
	#
	config['results'] = []
	config['transfers'] = {}
	config['hosts'] = []
	config['loglevel'] = 1
	#
	# This is the dict of hosts currently thought of as connected.
	connections = []
	#
	class identifiers:
		client = {}
		server = {}
		broadcast = {}
	#
	# User interface
	initialise_ui(STATUS, PROMPT)
	#
	# I *heart* UDP.
	s = socket(AF_INET, SOCK_DGRAM)
	s.setblocking(0)
	# No need to listen(), go straight to bind...
	try:
		s.bind(('', PORT))
	except socket_error, (errno, strerror):
		if errno == 98:
			print "PORT is already bound, trying PORT+1"
			s.bind(('', PORT+1))
			config['myport'] = PORT+1
	config['closeme'].append(s)
	#
	# Read and parse the config file.
	# Nonportable, two lines.
	if isfile(path_join(playground, "rc")):
		f = open(path_join(playground, "rc"), 'r')
		for raw_command in f.readlines():
			if not raw_command:
				break
			command = strip(raw_command)
			handle_command(s, config, connections, identifiers, \
				command)
		f.close()
	#
	if not config.has_key('myip'):
		try:
			config['myip'] = hostorip2rawip(gethostname())
		except socket_error:
			config['myip'] = hostorip2rawip('0.0.0.0')
	if isfile(path_join(playground, "nid")):
		f = open(path_join(playground, "nid"), 'r')
		config['nid'] = int(f.read())
	else:
		f = open(path_join(playground, "nid"), 'w')
		config['nid'] = unpack('l', get_random_bytes(4))[0]
		f.write(str(config['nid']))
	f.close()
	config['bid'] = 0
	# File Service
	if not config.has_key('share'):
		config['share'] = ()
	if not config.has_key('webport'):
		config['webport'] = 8000
	if not config.has_key('webip'):
		config['webip'] = config['myip']
	if not config.has_key('downloadpath'):
		config['downloadpath'] = '.'
	# Chat Service
	if not config.has_key('nick'):
		config['nick'] = 'unnamed'	# Yes, QuakeWorld.
	if not config.has_key('msgbeep'):
		config['msgbeep'] = 0
	#
	config['myport'] = PORT
	config['ignores'] = []
	config['hexbit'] = 1
	#
	return s, config, connections, identifiers

def shutdown(s, config, connections):
	try:
		shutdown_ui()
	except:
		pass
	if config:
		print "config exists"
		print "killing children, if any",
		if config.has_key('webserverpid'):
			try:
				print 'webserver',
				pid = config['webserverpid']
				kill(pid, 15)
				print waitpid(pid, 0)
			except:
				pass
		if config.has_key('transfers'):
			for pid in config['transfers'].keys():
				try:
					print 'transfer...',
					kill(pid, 15)
					print waitpid(pid, 0)
				except:
					pass
	if connections:
		print "connections exists, closing connections"
		for conn in connections:
			try:
				spacket = 'GENNY DISCONNECT\nQuitting\n\n'
				conn.send(spacket)
				print "Disconnected %s: " % conn + `spacket`
			except:
				pass
	if s:
		s.close()
		print "closed socket"
	return

def main_loop(s, config, connections, identifiers):
	# Nonportable, one line
	log_sock = config['log_sock']
	display_prompt()
	last_foo_bar = int(time())
	while 1:
		# Nonportable, seventeen lines
		# Check for input from keyboard or socket.
		for i in select([sys.stdin, s, log_sock], [], [], 30)[0]:
			# A command has been typed, handle it ...
			if i is sys.stdin:
				raw_command = sys.stdin.readline()
				if raw_command:
					handle_command(s, config, connections, \
						identifiers, strip(raw_command))
			# A packet has been received, handle it ...
			elif i is s:
				handle_packet(s, config, connections, \
					identifiers)
			elif i is log_sock:
				log_message = log_sock.recv(512)
				if config['loglevel'] > 1:
					print log_message
			display_prompt()
		#
		garbage_collect(s, config, connections, identifiers)
		#
		now = int(time())
		if last_foo_bar < now - 30:
			do_foo_bars(config, connections)
			last_foo_bar = now
	# Never reached.
	return

def garbage_collect(s, config, connections, identifiers):
	# Broadcast identifiers.
	hourago = int(time()) - 60 * 60
	for i in identifiers.broadcast.keys():
		if identifiers.broadcast[i] < hourago:
			del identifiers.broadcast[i]
	# Forked transfer processes.
	# Nonportable, seven lines.
	for pid, transfer in config['transfers'].items():
		exitedpid, status = waitpid(pid, WNOHANG)
		if exitedpid > 0:
			del config['transfers'][pid]
			if config['loglevel'] > 0:
				print "%s finished, status: %s" % \
					(transfer, status)
	return

def handle_command(s, config, connections, identifiers, command):
	if ' ' in command:
		command_name, arguments = split(command, ' ', 1)
	else:
		command_name, arguments = command, None
	if command_name == '' or command_name[0] == '#':
		pass
	# 
	# Genny Basic Service commands
	elif command_name == 'echo':
		handle_command_echo(config, connections, identifiers, arguments)
	elif command_name == 'update':
		handle_command_update(config, connections, identifiers)
	elif command_name == 'announce':
		handle_command_announce(config, connections, identifiers)
	elif command_name == 'daytime':
		handle_command_daytime(config, connections, identifiers)
	elif command_name == 'who':
		handle_command_who(config, connections, identifiers)
	elif command_name == 'hosts':
		print `config['hosts']`
	elif command_name == 'myip':
		handle_command_myip(config, arguments)
	#
	# Genny Chat Service commands
	elif command_name == 'nick':
		handle_command_nick(config, arguments)
	elif command_name == 's' or command_name == 'say':
		handle_command_say(config, connections, identifiers, arguments)
	elif command_name == 'me':
		handle_command_act(config, connections, identifiers, arguments)
	elif command_name == 'msgbeep':
		handle_command_msgbeep(config, arguments)
	#
	# Genny File Service commands
	elif command_name == 'find':
		handle_command_find(config, connections, identifiers, arguments)
	elif command_name == 'response':
		handle_command_response(config, arguments)
	elif command_name == 'get':
		handle_command_get(config, arguments)
	elif command_name == 'lclear':
		handle_command_lclear(config, arguments)
	elif command_name == 'share':
		handle_command_share(config, arguments)
	elif command_name == 'push':
		handle_command_push(config, identifiers, arguments)
	elif command_name == 'webip':
		handle_command_webip(config, arguments)
	elif command_name == 'webport':
		handle_command_webport(config, arguments)
	elif command_name == 'webserve':
		handle_command_webserve(config, arguments)
	elif command_name == 'downloadpath':
		handle_command_downloadpath(config, arguments)
	#
	elif command_name == 'help':
		if not arguments or ' ' in arguments:
			print """Help is available for the following subjects.
Type "help subject" to see its contents.
  internal - Connecting, quitting, etc
  basic - Info and daytime services, etc
  chat - Chat service commands
  file - File service commands"""
		elif arguments == 'internal': 
			print """Internal commands:
  list - List current connections
  hosts - Show hosts cachter
  open host[:port] - Open a connection
  close i|host[:port] [reason] - Close a connection
  ignore [hexstring] - Drop all packets from this node
  hexbit [on|off] - Toggle showing of remote node identifiers
  eval, exec - Crash this program
  help [subject] - Show help
  log loglevel - Set the level of logging
  quit - Quit Genny Test"""
		elif arguments == 'basic':
			print """Commands for the basic services:
  myip address - Set the servant's idea of its own IP address
  echo - Send an echo broadcast
  update - Send an info ping broadcast
  announce - Send an info pong broadcast
  daytime - Send a daytime broadcast
  who - Get the hostnames of computers"""
		elif arguments == 'chat':
			print """Commands for the chat service:
  say something - Send a chat broadcast
  s something - Alias for say
  me something - Send an action broadcast
  nick nickname - Set your nickname
  msgbeep [on|off] - Set beep-on-message on or off"""
		elif arguments == 'file':
			print """Commands for the file service:
  find terms - Search the file service
  response - Shows file search responses
  get num - Get the file matched with num
  push num - Send a push request for the file matched with num
  lclear - Clear search results
  downloadpath directory - Set the directory where files will be download to
  share path - Allocate files to be shared
  webip address - Set the servant's idea of its webserver's IP address
  webport port - Set the servant's idea of its webserver's TCP port
  webserve [port] - run the webserver, the default port is set with webport"""
		else:
			print "No help for %s." % arguments
	#
	elif command_name == 'list':
		print "Connections:"
		for conn in connections:
			i = connections.index(conn)
			print "%s %s %s" % (i, conn, conn.flag)
	#
	elif command_name == 'open':
		if not arguments or ' ' in arguments:
			print "usage:  open host[:port]"
		else:
			try:
				addr = hostify(arguments)
				for conn in connections:
					if conn.addr == addr:
						# leave conn for later
						break
				else:
					conn = genny_conn(s, addr)
					connections.append(conn)
					if config['loglevel'] > 0:
						print "Added connection: %s" % \
							conn
				spacket = 'GENNY CONNECT/0.8\n\n'
				conn.send(spacket)
				conn.flag = RWELC
				if config['loglevel'] > 1:
					print "Handshook %s: " % conn + \
						`spacket`
			except socket_error, message:
				print message
	#
	elif command_name == 'close':
		handle_command_close(s, config, connections, arguments)
	#
	elif command_name == 'quit':
		# sys.exit() actually does this anyways
		# I just want to make it clear that the exception is raised
		# and handled in main()
		raise SystemExit
	#
	elif command_name == 'log':
		if not arguments:
			print "loglevel is currently %s." % config['loglevel']
		else:
			try:
				config['loglevel'] = int(arguments)
				print "loglevel set to %s." % arguments
			except:
				print "usage:  log [level]"
	#
	elif command_name == 'eval':
		try:
			print `eval(arguments)`
		except:
			print "Exception raised in eval."
	elif command_name == 'exec':
		exec arguments
		print "exec happened."
	#
	elif command_name == 'clear':
		handle_command_clear()
	#
	elif command_name == 'ignore':
		if not arguments:
			output = []
			for hexbit in config['ignores']:
				output.append("%8X" % hexbit)
			print join(output)
		else:
			try:
				for hexbit in split(arguments):
					intbit = eval("0x" + hexbit)
					config['ignores'].append(intbit)
			except:
				print "exception raised, probably not a " \
					"valid hexbit"
	#
	elif command_name == 'hexbit':
		if not arguments:
			if config['hexbit']:
				print "hexbit is on"
			else:
				print "hexbit is off"
		elif arguments == "on":
			config['hexbit'] = 1
			print "hexbit is on"
		elif arguments == "off":
			config['hexbit'] = 0
			print "hexbit is off"
		else:
			print "usage:  hexbit [on|off]"
	#
	elif command_name == 'ids':
		if arguments == "b":
			print eval("identifiers.broadcast")
		elif arguments == "c":
			print eval("identifiers.client")
		elif arguments == "s":
			print eval("identifiers.server")
		else:
			print eval("identifiers.broadcast")
			print eval("identifiers.client")
			print eval("identifiers.server")
	else:
		print "Unknown command: %s" % command_name
	#
	return

def handle_command_close(s, config, connections, arguments):
	if not arguments:
		print "usage:  close i|host[:port] [reason]"
		return
	if ' ' in arguments:
		host, reason = split(arguments, ' ', 1)
		s_packet = 'GENNY DISCONNECT\n' + reason + '\n\n'
	else:
		host, reason = arguments, None
		s_packet = 'GENNY DISCONNECT\n\n'
	try:
		conn = connections[int(host)]
	except:
		try:
			s_addr = hostify(host)
		except socket_error, message:
			print message
			return
		for conn in connections:
			if s_addr == conn.addr:
				break
		else:
			conn = None
	if not conn:
		raw_send(s, s_packet, s_addr)
		if config['loglevel'] > 1:
			print "Disconnected %s:%s: " % s_addr + `s_packet`
		return
	conn.send(s_packet)
	if config['loglevel'] > 1:
		print "Disconnected %s: " % conn + `s_packet`
	connections.remove(conn)
	if config['loglevel'] > 0:
		print "Removed connection %s: " % conn + `reason`
	return

def handle_packet(s, config, connections, identifiers):
	# I recv the whole possible UDP datagram in any case.
	try:
		r_packet, r_addr = s.recvfrom(65536)
	except socket_error, (errno, strerror):
		if errno == ECONNREFUSED:
			print "Connection refused raised."
			return
		else:
			raise
	#
	if config['loglevel'] > 3:
		print "Received %s:%s: " % r_addr + `r_packet`
	#
	for conn in connections:
		if r_addr == conn.addr:
			break
	else:
		conn = None
	#
	# Begin packet decision
	if r_packet == 'GENNY CONNECT/0.8\n\n':
		if config['loglevel'] > 1:
			print "Was Handshook %s:%s: " % r_addr + `r_packet`
		if not conn:
			conn = genny_conn(s, r_addr)
			connections.append(conn)
			if config['loglevel'] > 0:
				print "Added connection: %s" % conn
		s_packet = "GENNY OK\n\n"
		conn.send(s_packet)
		if config['loglevel'] > 1:
			print "Welcomed %s: " % conn + `s_packet`
		conn.flag = GENNY
	#
	elif not conn:
		if config['loglevel'] > 2:
			print "Dropped packet from unconnected %s:%s: " % \
				r_addr + `r_packet`
		pass
	#
	elif r_packet == 'GENNY OK\n\n':
		if conn.flag == RWELC:
			if config['loglevel'] > 1:
				print "Was Welcomed %s: " % conn + `r_packet`
			conn.flag = GENNY
		else:
			if config['loglevel'] > 1:
				print "Was extraly Welcomed: %s " % conn + \
					`r_packet`
	#
	elif r_packet[:17] == 'GENNY DISCONNECT\n':
		if config['loglevel'] > 1:
			print "Was Disconnected %s: " % conn + `r_packet`
		if len(r_packet) == 18:
			if config['loglevel'] > 0:
				print "Removed connection %s:" % conn + `None`
		else:
			if config['loglevel'] > 0:
				print "Removed connection %s: " % \
					conn + `r_packet[17:-2]`
		connections.remove(conn)
	#
	elif conn.flag != GENNY:
		if config['loglevel'] > 0:
			print "Dropped from not fully connected %s: " % \
				conn + `r_packet`
		pass
	#
	elif len(r_packet) < 12:
		control(config, conn, None, pack('B', GENNY_CONTROL_TOOSMALL))
	#
	else:
		handle_genny_packet(config, connections, identifiers, conn, \
			r_packet)
	# End packet decision
	return

def handle_genny_packet(config, connections, identifiers, r_conn, r_packet):
	# Once upon a time, the routing logic of Genny was 8 if statements.
	#
	# If set to true, the packet is not handled according to its service.
	bogus = 0
	# Begin routing.
	cid, sid, bid, ttl, service = unpack('llBBH', r_packet[0:12])
	if config['loglevel'] > 3:
		print "%s, %s, %s, %s, %s, %s" % (cid, sid, bid, ttl, service, \
			r_conn)
	# Garbage filter
	if service not in GENNY_SERVICES:
		payload = pack('BH', GENNY_CONTROL_NOSERVICE, service)
		control(config, r_conn, None, payload)
		bogus = 1
	# Control Packets
	elif service == GENNY_CONTROL:
		# Control packets cannot enter the routing code.
		handle_genny_control(config, identifiers, r_conn, r_packet)
		bogus = 1
	# Ignore checking
	elif cid in config['ignores'] or sid in config['ignores']:
		pass
		bogus = 1
	# Broadcasts.
	elif ttl <= 7:
		# Don't override older client identifiers
		if not identifiers.client.has_key(cid):
			identifiers.client[cid] = r_conn
		conn = identifiers.client[cid]
		cbid = (cid, bid)
		if not identifiers.broadcast.has_key(cbid):
			identifiers.broadcast[cbid] = int(time())
			# Make sure we aren't the source
			if conn:
				rebroadcast(config, connections, conn, r_packet)
			else:
				# Packet is from me, don't handle it
				bogus = 1
		else:
			# We have already seen this packet
			bogus = 1
	# Replies.
	elif ttl == 255:
		# The packet must have a Broadcast Precursor
		if identifiers.client.has_key(cid):
			# Don't override older server identifiers
			if not identifiers.server.has_key(sid):
				identifiers.server[sid] = r_conn
			s_conn = identifiers.client[cid]
			# Make sure we're not the destination
			if s_conn:
				s_conn.send(r_packet)
		else:
			# The packet has no known client identifier.
			bogus = 1
	# ReReplies.
	elif ttl == 254:
		# The packet must have a Reply Precursor
		if identifiers.server.has_key(sid):
			s_conn = identifiers.server[sid]
			# Make sure we're not the destination
			if s_conn:
				s_conn.send(r_packet)
		else:
			# The packet has no known server identifier.
			bogus = 1
	# Ignore packets with other TTLs.
	else:
		bogus = 1
	# Routing done.
	#
	if not bogus:
		# This looks monotonous, doesn't it?
		if service == GENNY_INFO:
			handle_genny_info(config, identifiers, r_packet)
		elif service == GENNY_ECHO:
			handle_genny_echo(config, identifiers, r_packet)
		elif service == GENNY_DAYTIME:
	 		handle_genny_daytime(config, identifiers, r_packet)
		elif service == GENNY_HOSTNAME:
	 		handle_genny_hostname(config, identifiers, r_packet)
		elif service == GENNY_CHAT:
	 		handle_genny_chat(config, identifiers, r_packet)
		elif service == GENNY_FILE:
			handle_genny_file(config, identifiers, r_packet)
	return

if __name__ == '__main__':
	main()

