/*
 *  Gnubile - Gnutella software for Unix
 *  Copyright (C) 2000  Spencer Kimball and Gene Kan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Interface to the admin functionality
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gnome-xml/parser.h>
#include <gnome-xml/tree.h>

#include "interface.h"
#include "support.h"
#include "globals.h"
#include "admin.h"
#include "network.h"

/*
 *  Static variables
 */
static gint   settings_maintaincount = DEFAULT_MAINTAIN_COUNT;
static gint   settings_linkspeed     = 28800;
static gchar *settings_downloaddir   = "/tmp";

static GList *known_nodes = NULL;
static GList *connected_nodes = NULL;
static GList *hosted_files = NULL;

static GHashTable *node_hash = NULL;


/*
 *  Statistics
 */
gulong totalhosts = 0;
gulong totalfiles = 0;
gulong totalbytes = 0;
gulong servedfiles = 0;
gulong servedbytes = 0;
gulong downloadfiles = 0;
gulong downloadbytes = 0;


/*
 *  Static function declarations
 */
static gchar *admin_settings_file         (void);
static gint   admin_settings_file_exists  (void);
static void   parse_settings              (xmlDocPtr   xml_doc);
static void   parse_node                  (xmlDocPtr   xml_doc,
					   xmlNodePtr  xml_node,
					   ServerNode *node);
static void   parse_file                  (xmlDocPtr   xml_doc,
					   xmlNodePtr  xml_node,
					   ServerFile *file);
static void   admin_add_file              (ServerFile *file);
static void   admin_remove_file           (ServerFile *file);

static gchar *admin_node_status_text      (NodeStatus status);
static void   admin_add_connected_node    (ServerNode *node);
static void   admin_remove_connected_node (ServerNode *node);
static void   admin_add_known_node        (ServerNode *node);
static void   admin_remove_known_node     (ServerNode *node);
static void   admin_remove_node           (GtkCList   *clist,
					   ServerNode *node);
static guint  admin_node_hash_hash        (gconstpointer  key);
static gint   admin_node_hash_compare     (gconstpointer  a,
					   gconstpointer  b);



/*
 *  Save preferences and exit the application
 */
void admin_exit (void)
{
  admin_save ();
  gtk_exit (0);
}

/*
 *  Load the settings from the user's .gnucleus file
 */
void admin_load (void)
{
  gchar     *pref_file = NULL;
  xmlDocPtr  xml_doc;

  /*
   *  If the file doesn't exist, write a new one, finished
   */
  if (admin_settings_file_exists () == 0)
    {
      g_print ("created new user preferences\n");
      admin_save ();
      return;
    }

  /*
   *  Get the XML prefs file
   */
  pref_file = admin_settings_file ();
  if ((xml_doc = xmlParseFile (pref_file)) == NULL)
    {
      g_print ("Error reading preference file: %s\n", pref_file);
      gtk_exit (1);
      return;
    }
  
  /*
   *  Parse the XML file
   */
  parse_settings (xml_doc);

  /*
   *  Free the XML tree
   */
  xmlFreeDoc (xml_doc);
}

/*
 *  Save the current settings in the user's .gnucleus file
 */
void admin_save (void)
{
  GList      *node_list;
  GList      *file_list;
  ServerNode *node;
  ServerFile *file;
  gchar      *pref_file = admin_settings_file ();
  gchar       int_str[32];
  xmlDocPtr   xml_doc;
  xmlNodePtr  xml_node;
  xmlNodePtr  xml_subnode;

  /*
   *  Create the xml tree
   */
  xml_doc = xmlNewDoc ("1.0");
  xml_doc->root = xmlNewDocNode (xml_doc, NULL, "GNUCLEUSPREFS", NULL);
  xml_node = xmlNewChild (xml_doc->root, NULL, "downloaddir",
			  admin_get_downloaddir ());
  sprintf (int_str, "%d", admin_get_linkspeed ());
  xml_node = xmlNewChild (xml_doc->root, NULL, "linkspeed", int_str);
  sprintf (int_str, "%d", admin_get_maintaincount ());
  xml_node = xmlNewChild (xml_doc->root, NULL, "maintaincount", int_str);

  /*
   *  Create all nodes
   */
  node_list = known_nodes;
  while (node_list != NULL)
    {
      node = (ServerNode *) node_list->data;
      node_list = g_list_next (node_list);

      xml_node = xmlNewChild (xml_doc->root, NULL, "node", NULL);
      xml_subnode = xmlNewChild (xml_node, NULL, "server", node->server);
      sprintf (int_str, "%d", node->port);
      xml_subnode = xmlNewChild (xml_node, NULL, "port", int_str);
    }

  /*
   *  Create all files
   */
  file_list = hosted_files;
  while (file_list != NULL)
    {
      file = (ServerFile *) file_list->data;
      file_list = g_list_next (file_list);

      xml_node = xmlNewChild (xml_doc->root, NULL, "file", NULL);
      xml_subnode = xmlNewChild (xml_node, NULL, "filename", file->filename);
      sprintf (int_str, "%d", file->size);
      xml_subnode = xmlNewChild (xml_node, NULL, "size", int_str);
    }

  /*
   *  Save the tree
   */
  xmlSaveFile(pref_file, xml_doc);

  /*
   *  Free the tree
   */
  xmlFreeDoc (xml_doc);
}

/*
 *  These functions should be called from the server
 *   to retrieve current administrative settings
 */
gchar *admin_get_downloaddir (void)
{
  return settings_downloaddir;
}

gint admin_get_linkspeed (void)
{
  return settings_linkspeed;
}

gint admin_get_maintaincount (void)
{
  return settings_maintaincount;
}


/*
 *  This function should be called from the interface to
 *   set the download directory (the final resting place of
 *   files)
 */
void admin_set_downloaddir (gchar *downloaddir,
			    gint   update_interface)
{
  gint           err;
  struct stat    filestat;

  /* Stat the supplied file */
  err = stat (downloaddir, &filestat);

  /* Handle depending on type... */
  if (err != 0)
    {
      gtk_widget_show (create_direrror ());
      g_print ("error accessing directory: %s\n", downloaddir);
      return;
    }
  else if (S_ISDIR (filestat.st_mode))
    {
      /*  Set the downloaddir  */
      settings_downloaddir = downloaddir;

      /*  Only update the interface if requested (avoid infinite loops) */
      if (update_interface)
	{
	  GtkEntry *entry = GTK_ENTRY (lookup_widget (admin, "downloadentry"));
	  gtk_entry_set_text (entry, downloaddir);
	}
    }
  else /* Non-directory file type */
    {
      gtk_widget_show (create_direrror ());
      g_print ("specified file is not a directory: %s\n", downloaddir);
      return;
    }
}

/*
 *  This function should be called by the interface to set
 *   the link speed
 */
void admin_set_linkspeed (gint linkspeed,
			  gint update_interface)
{
  /*  Set the linkspeed  */
  settings_linkspeed = linkspeed;

  /*  Only update the interface if requested (avoid infinite loops) */
  if (update_interface)
    {
      gchar text_str[32];
      GtkEntry *entry = GTK_ENTRY (lookup_widget (admin, "speedcomboentry"));
      sprintf (text_str, "%.1f kbps", linkspeed / 1000.0);
      gtk_entry_set_text (entry, text_str);
    }
}

/*
 *  This function should be called by the interface to set
 *   the maintained connected node count
 */
void admin_set_maintaincount (gint maintaincount,
			      gint update_interface)
{
  /*  Set the maintaincount  */
  settings_maintaincount = maintaincount;

  /*  Only update the interface if requested (avoid infinite loops) */
  if (update_interface)
    {
      gchar text_str[32];
      GtkEntry *entry = GTK_ENTRY (lookup_widget (admin, "maintaincountentry"));
      sprintf (text_str, "%d", maintaincount);
      gtk_entry_set_text (entry, text_str);
    }
}

/*
 *  This function should be called by either the interface
 *   or the server to add a file to the served file set.
 */
void admin_file_add (gchar *filename)
{
  ServerFile    *file;
  gint           err;
  gchar         *entry_filename;
  DIR           *dir;
  struct dirent *dir_ent;
  struct stat    filestat;

  /* Stat the supplied file */
  err = stat (filename, &filestat);

  /* Handle depending on type... */
  if (err != 0)
    {
      gtk_widget_show (create_fileerror ());
      g_print ("error accessing file: %s\n", filename);
      return;
    }
  else if (S_ISLNK (filestat.st_mode) ||
	   S_ISREG (filestat.st_mode))
    {
      file = g_new (ServerFile, 1);
      file->filename = g_strdup (filename);
      file->size = filestat.st_size;
      admin_add_file (file);

      /*
       *  Add the file to the hosted file list
       */
      hosted_files = g_list_append (hosted_files, file);
    }
  else if (S_ISDIR (filestat.st_mode))
    {
      /* Open directory */
      dir = opendir (filename);

      if (!dir)
	{
	  gtk_widget_show (create_fileerror ());
	  g_print ("error reading directory: %s\n", filename);
	  return;
	}
      else
	{
	  while ((dir_ent = readdir (dir)))
	    {
	      if (strcmp (dir_ent->d_name, ".") != 0 &&
		  strcmp (dir_ent->d_name, "..") != 0)
		{ 
		  entry_filename = g_malloc (strlen(filename) + 
					     strlen (dir_ent->d_name) + 2);

		  if (filename[strlen (filename) - 1] == '/')
		    sprintf (entry_filename, "%s%s", 
			     filename, dir_ent->d_name);
		  else
		    sprintf (entry_filename, "%s/%s", 
			     filename, dir_ent->d_name);

		  admin_file_add (entry_filename);

		  g_free (entry_filename);
		}
	    }

	  closedir (dir);
	}
    }
  else /* Unknown file type */
    {
      gtk_widget_show (create_fileerror ());
      g_print ("unknown file type: %s\n", filename);
      return;
    }
}

/*
 *  This function should be called by either the interface
 *   or the server to remove a file from the served file set
 */
void admin_file_remove (ServerFile *file)
{
  admin_remove_file (file);

  /*
   *  Remove the file from the hosted file list
   */
  hosted_files = g_list_remove (hosted_files, file);
}


/*
 *  This function adds hosts from the known table to
 *   connected status in an attempt to keep up with the
 *   targeted node maintainance setting
 */
void admin_node_balance (void)
{
  ServerNode *node;
  GList      *cons;
  int         good_state_count = 0;
  int         i;

  cons = connected_nodes;
  while (cons != NULL)
    {
      node = (ServerNode *) cons->data;
      if (node->status == NODE_CONNECTING ||
	  node->status == NODE_INBOUND ||
          node->status == NODE_PROTECTED ||
	  node->status == NODE_OUTBOUND)
	good_state_count += 1;
      cons = cons->next;
    }

  cons = known_nodes;
  for (i = good_state_count; i < settings_maintaincount && cons != NULL; i++)
    {
      node = (ServerNode *) cons->data;
      cons = cons->next;
      
      if (node->status == NODE_DISCONNECTED &&
	  g_list_find (connected_nodes, node) == NULL)
	admin_node_connect (node, 0);
    }
}

/*
 *  This function should be called by the server
 *   to register a new admin
 */
ServerNode* admin_node_add (ServerNode *node)
{
  /*
   *  Make sure the node isn't already known
   */
  gchar cache_str[sizeof("XXX.XXX.XXX.XXX:XXXXX")];
  ServerNode *tnode;

  if (node_hash == NULL)
    node_hash = g_hash_table_new (admin_node_hash_hash, 
				  admin_node_hash_compare);
  sprintf (cache_str, "%s:%d", node->server, node->port);
  tnode = g_hash_table_lookup (node_hash, (gpointer) cache_str);
  if (tnode)
      return tnode;

  g_hash_table_insert (node_hash, g_strdup(cache_str), node);

  /*
   *  Update statistics
   */
  admin_set_statistic ("totalhostslabel", &totalhosts, 1);
  admin_set_statistic ("totalfileslabel", &totalfiles, node->files);
  admin_set_statistic ("totalbyteslabel", &totalbytes, node->bytes);

  /*
   *  Add the node to the known host list
   */
  admin_add_known_node (node);
  
  /*
   *  Add the node to the known node list
   */
  known_nodes = g_list_append (known_nodes, node);

  return NULL;
}

ServerNode* admin_node_find (ServerNode *node)
{
  gchar cache_str[sizeof("XXX.XXX.XXX.XXX:XXXXX")];

  if (node_hash == NULL)
    node_hash = g_hash_table_new (admin_node_hash_hash, 
				  admin_node_hash_compare);

  sprintf (cache_str, "%s:%d", node->server, node->port);

  return g_hash_table_lookup (node_hash, (gpointer) cache_str);
}

/*
 *  This function is called by the interface to
 *   connect a known node
 */
void admin_node_connect (ServerNode *node, gint protect)
{
  if(protect) node->protected = 1;
  /*
   *  Attempt to connect
   */
  node->status = NODE_CONNECTING;
  if (network_start_client (node->server, node->port, node) != 0)
    {
      node->status = NODE_FAILURE;

      /*
       *  Log the connection error
       */
      g_print ("Error connecting to server %s:%d\n", node->server, node->port);
    }

  /*
   *  Add to connected node list
   */
  admin_add_connected_node (node);


  /*
   *  Add the node to the connected node list
   */
  connected_nodes = g_list_append (connected_nodes, node);

}


/*
 *  This function is called by the interface to
 *   disconnect a connected node
 */
void admin_node_disconnect (ServerNode *node)
{
  /*
   *  Remove from current list, add to complete list
   */
  admin_remove_connected_node (node);
  node->status = NODE_DISCONNECTED;

  /*
   *  Remove the node from the connected node list
   */
  connected_nodes = g_list_remove (connected_nodes, node);
}


/*
 *  This function is called by the interface to
 *   delete a known node
 */
void admin_node_delete (ServerNode *node)
{
  /*
   *  Remove from known list
   */
  admin_remove_known_node (node);
  admin_remove_connected_node (node);

  /*
   *  Remove the node from the connected node list
   */
  known_nodes = g_list_remove (known_nodes, node);
  connected_nodes = g_list_remove (connected_nodes, node);

  /*
   *  Update statistics
   */
  admin_set_statistic ("totalhostslabel", &totalhosts, -1);
  admin_set_statistic ("totalfileslabel", &totalfiles, -node->files);
  admin_set_statistic ("totalbyteslabel", &totalbytes, -node->bytes);

  /*
   *  Remove from the node hash
   */
  if (node_hash != NULL)
    {
      gchar cache_str[sizeof("XXX.XXX.XXX.XXX:XXXXX")];
      sprintf (cache_str, "%s:%d", node->server, node->port);
      g_hash_table_remove (node_hash, cache_str);
    }
}

/*
 *  This function is called by the server to update the
 *   connected status of a server node
 */
void admin_node_update (ServerNode *node, 
			NodeStatus  status)
{
  GtkCList *clist = GTK_CLIST (lookup_widget (admin, "chostlist"));
  int row = gtk_clist_find_row_from_data (clist, node);
  if ((node->protected == 1) && (status == NODE_OUTBOUND)) status = NODE_PROTECTED;

  gtk_clist_set_text (clist, row, 1, admin_node_status_text (status));

  /*  If the status is failure or disconnect remove from the connected list
   */
  if (status == NODE_FAILURE || 
      status == NODE_DISCONNECTED)
    admin_node_disconnect (node);

  node->status = status;
}

int admin_query_protected (ServerNode *node)
{
  return node->protected;
}

/*
 *  Set the label for a particular statistic
 */
void admin_set_statistic (gchar  *statlabel,
			  gulong *old_value,
			  gulong  addition)
{
  GtkLabel *label = GTK_LABEL (lookup_widget (app, statlabel));
  gchar     label_str[32];

  *old_value += addition;

  if (strcmp (statlabel, "totalbyteslabel") == 0)
    sprintf (label_str, "(%ld MB)", (*old_value / 1024));
  else
    sprintf (label_str, "%ld", *old_value);

  gtk_label_set_text (label, label_str);
}

/*
 *  Static functions
 */

/*
 *  Get the user's settings file
 */
static gchar *admin_settings_file (void)
{
  gchar *home_dir;
  gchar *filename;
  
  home_dir = getenv ("HOME");
  filename = g_new (gchar, strlen (home_dir) + strlen (".gnucleus") + 2);

  sprintf (filename, "%s/.gnucleus", home_dir);
  
  return filename;
}

static gint admin_settings_file_exists (void)
{
  gchar         *filename = admin_settings_file ();
  struct stat    filestat;

  /* Stat the supplied file */
  return (stat (filename, &filestat) == 0);
}

static void parse_settings (xmlDocPtr xml_doc)
{
  xmlNodePtr xml_node;

  xml_node = xml_doc->root->childs;
  while (xml_node != NULL)
    {
      if (strcmp (xml_node->name, "downloaddir") == 0)
	{
	  gchar *downloaddir = 
	    xmlNodeListGetString (xml_doc, xml_node->childs, 1);
	  admin_set_downloaddir (downloaddir, TRUE);
	}
      else if (strcmp (xml_node->name, "linkspeed") == 0)
	{
	  gchar *linkspeed = 
	    xmlNodeListGetString (xml_doc, xml_node->childs, 1);
	  gint   int_linkspeed = atoi (linkspeed);

	  admin_set_linkspeed (int_linkspeed, TRUE);
	}
      else if (strcmp (xml_node->name, "maintaincount") == 0)
	{
	  gchar *maintaincount = 
	    xmlNodeListGetString (xml_doc, xml_node->childs, 1);
	  gint   int_maintaincount = atoi (maintaincount);

	  admin_set_maintaincount (int_maintaincount, TRUE);
	}
      else if (strcmp (xml_node->name, "node") == 0)
	{
	  ServerNode *node = g_new (ServerNode, 1);
	  parse_node (xml_doc, xml_node, node);

	  /*  Add node  */
	  admin_node_add (node);
	}
      else if (strcmp (xml_node->name, "file") == 0)
	{
	  ServerFile *file = g_new (ServerFile, 1);
	  parse_file (xml_doc, xml_node, file);

	  /*  Add file  */
	  admin_file_add (file->filename);

	  /*  Delete this file structure (a new one is created when added)  */
	  g_free (file);
	}

      xml_node = xml_node->next;
    }
}

/*
 *  Parse a server node object
 */
static void parse_node (xmlDocPtr   xml_doc,
			xmlNodePtr  xml_node,
			ServerNode *node)
{
  node->server = NULL;
  node->port = 0;
  node->status = NODE_DISCONNECTED;
  node->files = 0;
  node->bytes = 0;

  xml_node = xml_node->childs;
  while (xml_node != NULL)
    {
      if (strcmp (xml_node->name, "server") == 0)
	node->server =
	  xmlNodeListGetString (xml_doc, xml_node->childs, 1);
      else if (strcmp (xml_node->name, "port") == 0)
	node->port =
	  atoi (xmlNodeListGetString (xml_doc, xml_node->childs, 1));

      xml_node = xml_node->next;
    }
}

/*
 *  Parse a server file object
 */
static void parse_file (xmlDocPtr   xml_doc,
			xmlNodePtr  xml_node,
			ServerFile *file)
{
  file->filename = NULL;
  file->size = 0;

  xml_node = xml_node->childs;
  while (xml_node != NULL)
    {
      if (strcmp (xml_node->name, "filename") == 0)
	file->filename =
	  xmlNodeListGetString (xml_doc, xml_node->childs, 1);
      if (strcmp (xml_node->name, "size") == 0)
	{
	  gchar *size_str =
	    xmlNodeListGetString (xml_doc, xml_node->childs, 1);
	  file->size = atoi (size_str);
	}

      xml_node = xml_node->next;
    }
}

/*
 *  Add a file to the list of served files
 */
static void admin_add_file (ServerFile *file)
{
  GtkCList *list = GTK_CLIST (lookup_widget (admin, "filelist"));
  gchar *text[2];
  gchar  size_str[32];

  /*  Format column values  */
  text[0] = g_strdup (file->filename);
  sprintf (size_str, "%d", file->size);
  text[1] = g_strdup (size_str);

  /*  Insert into list  */
  gtk_clist_append (list, text);
  gtk_clist_set_row_data (list, list->rows - 1, file);
}

/*
 *  Remove a file from the columned list of served files
 */
static void admin_remove_file (ServerFile *file)
{
  GtkCList *list = GTK_CLIST (lookup_widget (admin, "filelist"));
  gint row;

  row = gtk_clist_find_row_from_data (list, file);
  gtk_clist_remove (list, row);
}

/*
 *  Get the textual message for a server status code
 */
static gchar *admin_node_status_text (NodeStatus status)
{
  gchar *text;

  switch (status)
    {
    case NODE_DISCONNECTED: text = "Disconnected"; break;
    case NODE_FAILURE:      text = "Failure"; break;
    case NODE_CONNECTING:   text = "Connecting..."; break;
    case NODE_INBOUND:      text = "Inbound"; break;
    case NODE_OUTBOUND:     text = "Outbound"; break;
    case NODE_PROTECTED:    text = "Protected"; break;
    default:                abort(); break;
    }

  return text;
}

/*
 *  Add connected node
 */
static void admin_add_connected_node (ServerNode *node)
{
  GtkCList *list = GTK_CLIST (lookup_widget (admin, "chostlist"));
  gchar *text[2];
  gchar  server_name[256];

  /*  Format column values  */
  sprintf (server_name, "%s:%d", node->server, node->port);
  text[0] = g_strdup (server_name);
  text[1] = admin_node_status_text (node->status);

  /*  Insert into list  */
  gtk_clist_append (list, text);
  gtk_clist_set_row_data (list, list->rows - 1, node);
}

/*
 *  Remove connected node
 */
static void admin_remove_connected_node (ServerNode *node)
{
  GtkCList *list = GTK_CLIST (lookup_widget (admin, "chostlist"));
  admin_remove_node (list, node);
}

/*
 *  Add known node
 */
static void admin_add_known_node (ServerNode *node)
{
  GtkCList *list = GTK_CLIST (lookup_widget (admin, "khostlist"));
  gchar *text[1];
  gchar  server_name[256];

  /*  Format column values  */
  sprintf (server_name, "%s:%d", node->server, node->port);
  text[0] = g_strdup (server_name);

  /*  Insert into list  */
  gtk_clist_append (list, text);
  gtk_clist_set_row_data (list, list->rows - 1, node);
}

/*
 *  Remove known node
 */
static void admin_remove_known_node (ServerNode *node)
{
  GtkCList *list = GTK_CLIST (lookup_widget (admin, "khostlist"));
  admin_remove_node (list, node);
}

/*
 *  Remove row from list
 */
static void admin_remove_node (GtkCList   *list,
			       ServerNode *node)
{
  gint row;

  row = gtk_clist_find_row_from_data (list, node);
  gtk_clist_remove (list, row);
}

/*
 *  Hash functions
 */
static guint admin_node_hash_hash (gconstpointer  key)
{
  gchar *string;
  guint result;
  int c;

  /*
   * I tried a zillion different hash functions and asked many other
   * people for advice.  Many people had their own favorite functions,
   * all different, but no-one had much idea why they were good ones.
   * I chose the one below (multiply by 9 and add new character)
   * because of the following reasons:
   *
   * 1. Multiplying by 10 is perfect for keys that are decimal strings,
   *    and multiplying by 9 is just about as good.
   * 2. Times-9 is (shift-left-3) plus (old).  This means that each
   *    character's bits hang around in the low-order bits of the
   *    hash value for ever, plus they spread fairly rapidly up to
   *    the high-order bits to fill out the hash value.  This seems
   *    works well both for decimal and non-decimal strings.
   *
   * tclHash.c --
   *
   *      Implementation of in-memory hash tables for Tcl and Tcl-based
   *      applications.
   *
   * Copyright (c) 1991-1993 The Regents of the University of California.
   * Copyright (c) 1994 Sun Microsystems, Inc.
   */

  string = (gchar *) key;
  result = 0;
  while (1)
    {
      c = *string;
      string++;
      if (c == 0)
	break;
      result += (result << 3) + c;
    }

  return result;
}

static gint admin_node_hash_compare (gconstpointer a,
				     gconstpointer b)
{
  return (strcmp ((char *) a, (char *) b) == 0);
}
