/*
 *  Gnubile - Gnutella software for Unix
 *  Copyright (C) 2000  Spencer Kimball and Gene Kan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Interface to the download functionality
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "globals.h"
#include "search.h"
#include "download.h"
#include "admin.h"
#include "network.h"

/*
 *  Static function declarations
 */
static gchar *download_get_size_text      (gint read, gint total);
static gchar *download_get_status_text    (DownloadStatus status);
static void download_add_active_file      (DownloadFile *dfile);
static void download_remove_active_file   (DownloadFile *dfile);
static void download_add_complete_file    (DownloadFile *dfile);
static void download_remove_complete_file (DownloadFile *dfile);
static void download_add_file             (GtkCList     *clist,
					   DownloadFile *dfile);
static void download_remove_file          (GtkCList     *clist,
					   DownloadFile *dfile);


/*
 *  This function should be called by the server
 *   to register a new download
 */
void download_register (SearchResult *result)
{
  DownloadFile *dfile = g_new (DownloadFile, 1);

  dfile->filename = result->filename;
  dfile->size = result->size;
  dfile->read = 0;
  dfile->status = DOWNLOAD_CONNECTING;
  dfile->link_speed = result->link_speed;
    
  /*
   *  Register the download with the active download
   *   list
   */
  download_add_active_file (dfile);

  /*
   *  Request the file from the network
   */
  if (network_get_file (result->server, result->port, result->index,
			dfile->filename, admin_get_downloaddir (), dfile) != 0)
    {
      /*  Create an error condition  */
      download_update (dfile, DOWNLOAD_ERROR);
    }
}

/*
 *  This function is called by the interface to
 *   remove an entry from the complete file list
 */
void download_clear (DownloadFile *dfile)
{
  download_remove_complete_file (dfile);
}

/*
 *  This function should be called by the server
 *   to update the status of a particular
 *   download in progress (bytes read, status)
 */
void download_update (DownloadFile   *dfile,
		      DownloadStatus  new_status)
{
  /*
   *  Force the bytes read to total length
   */
  if (new_status == DOWNLOAD_COMPLETE)
    {
      dfile->read = dfile->size;

      /*
       *  Update statistics
       */
      admin_set_statistic ("downloadfileslabel", &downloadfiles, 1);
      admin_set_statistic ("downloadbyteslabel", &downloadbytes, dfile->size);
    }      

  /*
   *  If status changed from ACTIVE OR CONNECTING to > ACTIVE
   *   Remove dfile from current list and add to complete list
   */
  if ((dfile->status == DOWNLOAD_ACTIVE ||
       dfile->status == DOWNLOAD_CONNECTING) &&
      (new_status == DOWNLOAD_COMPLETE ||
       new_status == DOWNLOAD_ABORTED ||
       new_status == DOWNLOAD_ERROR))
    {
      dfile->status = new_status;
      download_remove_active_file (dfile);
      download_add_complete_file (dfile);
    }
  else
    {
      GtkCList *clist = GTK_CLIST (lookup_widget (app, "adlist"));
      int row = gtk_clist_find_row_from_data (clist, dfile);

      /*
       *  Update the status and bytes read fields
       */
      dfile->status = new_status;

      gtk_clist_set_text (clist, row, 1, 
			  download_get_size_text (dfile->read, dfile->size));
      gtk_clist_set_text (clist, row, 2, 
			  download_get_status_text (dfile->status));
    }
}

/*
 *  This function is called by the interface to
 *   abort the download of a file
 */
void download_abort (DownloadFile *dfile)
{
  if (dfile->status != DOWNLOAD_ACTIVE)
    return;

  /*
   *  Remove from current list, add to complete list
   */
  download_remove_active_file (dfile);
  dfile->status = DOWNLOAD_ABORTED;
  download_add_complete_file (dfile);
}


/*
 *  Static functions
 */

/*
 *  Add active file
 */
static void download_add_active_file (DownloadFile *dfile)
{
  GtkCList *active_list = GTK_CLIST (lookup_widget (app, "adlist"));
  download_add_file (active_list, dfile);
}

/*
 *  Remove active file
 */
static void download_remove_active_file (DownloadFile *dfile)
{
  GtkCList *active_list = GTK_CLIST (lookup_widget (app, "adlist"));
  download_remove_file (active_list, dfile);
}

/*
 *  Add complete file
 */
static void download_add_complete_file (DownloadFile *dfile)
{
  GtkCList *active_list = GTK_CLIST (lookup_widget (app, "cdlist"));
  download_add_file (active_list, dfile);
}

/*
 *  Remove complete file
 */
static void download_remove_complete_file (DownloadFile *dfile)
{
  GtkCList *active_list = GTK_CLIST (lookup_widget (app, "cdlist"));
  download_remove_file (active_list, dfile);
}

/*
 *  Remove row from clist
 */
static void download_remove_file (GtkCList     *clist,
				  DownloadFile *dfile)
{
  gint row;

  row = gtk_clist_find_row_from_data (clist, dfile);
  gtk_clist_remove (clist, row);
}

/*
 *  Get the text for bytes read/ total bytes
 */
static gchar *download_get_size_text (gint read,
				      gint size)
{
  gchar size_str[32];

  sprintf (size_str, "%d of %d (%.1f%%)", (read + 1023) / 1024, 
	   (size + 1023) / 1024, 
	   ((gdouble) read / (gdouble) size) * 100);

  return g_strdup (size_str);
}

/*
 *  Get the text for the status
 */
static gchar *download_get_status_text (DownloadStatus status)
{
  gchar *text;

  switch (status)
    {
    case DOWNLOAD_CONNECTING: text = "Connecting..."; break;
    case DOWNLOAD_ACTIVE:     text = "Downloading"; break;
    case DOWNLOAD_COMPLETE:   text = "Complete"; break;
    case DOWNLOAD_ABORTED:    text = "Aborted"; break;
    case DOWNLOAD_ERROR:      text = "Failure"; break;
    default:                  text = "Unknown"; break;
    }

  return text;
}

/*
 *  Add row to clist
 */
static void download_add_file (GtkCList     *clist,
			       DownloadFile *dfile)
{
  gchar *text[3];

  /*  Format column values  */
  text[0] = g_strdup (dfile->filename);
  text[1] = download_get_size_text (dfile->read, dfile->size);
  text[2] = download_get_status_text (dfile->status);
      
  /*  Insert into clist  */
  gtk_clist_prepend (clist, text);
  gtk_clist_set_row_data (clist, 0, dfile);
}
