/*
 *  Gnubile - Gnutella software for Unix
 *  Copyright (C) 2000  Spencer Kimball and Gene Kan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "stream_config.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <dirent.h>
#include <assert.h>
#include <glib.h>


#include "stream_sched.h"
#include "stream_inet.h"
#include "stream_std.h"

#include "uuid.h"
#include "textutils.h"
#include "ip.h"
#include "admin.h"
#include "search.h"
#include "monitor.h"
#include "download.h"
#include "log.h"
#include "hexdump.h"

#define FREE(func,x) (func)(x), (x) = NULL

#define MAX_DROPPED               1000
#define MAX_TTL                     10
#define MAX_MSGCACHE            500000
#define MAX_CLIENTIDCACHE        10000
#define MAX_QUERYCACHE            1000
#define MAX_GUARANTEEQUEUELEN       20
#define MAX_UNGUARANTEEQUEUELEN      5

/* Protocol objects. */

typedef struct {
	uuid      msgid;
	u_int8_t  f;
#define GNUT_F_PING  0x00
#define GNUT_F_PONG  0x01
#define GNUT_F_PUSH  0x40
#define GNUT_F_QUERY 0x80
#define GNUT_F_HITS  0x81
	u_int8_t  ttl;
	u_int8_t  hops;

	int32_t   len;
} gnut_hdr;

#define gnut_hdr_r_format "%16d%1s%1s%1s%4s"
#define gnut_hdr_w_format "%16d%1s%1s%1s%4s"

typedef struct {
	u_int16_t port;
	u_int32_t ip;
	u_int32_t nfiles;
	u_int32_t nbytes;
} gnut_pong;

#define gnut_pong_r_format "%2s%4d%4s%4s"
#define gnut_pong_w_format "%2s%4d%4s%4s"

typedef struct {
	uuid      clientid;
	u_int32_t index;
	u_int32_t ip;
	u_int16_t port;
} gnut_push;

#define gnut_push_r_format "%16d%4s%4d%2s"
#define gnut_push_w_format "%16d%4s%4d%2s"

typedef struct {
	u_int16_t minspeed;
	u_int8_t* keywords;
} gnut_query;

#define gnut_query_r_format "%2s%p"
#define gnut_query_w_format "%2s%?p"

typedef struct {
	u_int32_t index;
	u_int32_t size;
	u_int8_t* name;
} gnut_hit;

#define gnut_hit_r_format "%4s%4s%p"
#define gnut_hit_w_format "%4s%4s%?p"

typedef struct {
	u_int8_t  n;
	u_int16_t port;
	u_int32_t ip;
	u_int32_t speed;
	u_int8_t* hits;
} gnut_hitset;

#define gnut_hitset_r_format "%1s%2s%4d%4s%p"
#define gnut_hitset_w_format "%1s%2s%4d%4s%?p"

/* Private objects and their "member functions". */

typedef struct {
	int      len;
	u_int8_t payload[0];
} gnut_msg;
	static gnut_msg* new_gnut_msg (u_int8_t* _buf, int _len) {
		gnut_msg* msg;

		msg = (gnut_msg*) g_malloc(sizeof(gnut_msg) + _len);
		memcpy(msg->payload, _buf, _len);
		msg->len = _len;

		return msg;
	}
	static void free_gnut_msg (gnut_msg* _msg) {
		g_free(_msg);
	}

typedef struct {
	stream_host*  host;
	GList*        msgq;
	GList*        seenqueries;

	int sent;
	int received;
	int dropped;
	int bigttl;
	int evil;
        int protected;
} connectedhost;
	static connectedhost* new_connectedhost (stream_host* _host) {
		connectedhost* host;

		host = (connectedhost*) g_malloc(sizeof(connectedhost));
		host->host        = _host;
		host->msgq        = NULL;
		host->seenqueries = NULL;

		host->sent     = 0;
		host->received = 0;
		host->dropped  = 0;
		host->bigttl   = 0;
		host->evil     = 0;
                host->protected= 0;

		return host;
	}
	static void free_connectedhost (connectedhost* _host) {
		GList*    li;

		/* Free queued messages. */
		for (li = _host->msgq; li; li = g_list_next(li))
			FREE(free_gnut_msg, (gnut_msg*) li->data);
		g_list_free(_host->msgq);

		/* Free seen queries. */
		for (li = _host->seenqueries; li; li = g_list_next(li))
			FREE(g_free, (char*) li->data);
		g_list_free(_host->seenqueries);

		FREE(g_free, _host);
	}

typedef struct {
	connectedhost* fromhost;
	uuid           msgid;
	u_int8_t       f;
} gnut_seenmsg;
	static gnut_seenmsg* new_gnut_seenmsg (connectedhost* _fromhost, uuid _msgid, u_int8_t _f) {
		gnut_seenmsg* seenmsg;

		seenmsg = (gnut_seenmsg*) g_malloc(sizeof(gnut_seenmsg));
		seenmsg->fromhost = _fromhost;
		seenmsg->msgid    = _msgid;
		seenmsg->f        = _f;

		return seenmsg;
	}
	static void free_gnut_seenmsg (gnut_seenmsg* _seenmsg) {
		FREE(g_free, _seenmsg);
	}

typedef struct {
	connectedhost* chost;
	uuid           clientid;
} gnut_seenclient;
	static gnut_seenclient* new_gnut_seenclient (connectedhost* _chost, uuid _clientid) {
		gnut_seenclient* seenclient;

		seenclient = (gnut_seenclient*) g_malloc(sizeof(gnut_seenclient));
		seenclient->chost    = _chost;
		seenclient->clientid = _clientid;

		return seenclient;
	}
	static void free_gnut_seenclient (gnut_seenclient* _seenclient) {
		FREE(g_free, _seenclient);
	}

/* Just a few globals. */

GList* g_connectedhosts     = NULL;
GHashTable* g_seenmsg_hash  = NULL;
GList* g_seenmsg_list       = NULL;
GHashTable* g_clientid_hash = NULL;
GList* g_clientid_list      = NULL;

static void report_kill (connectedhost* _chost) {
	char msg[128];

	sprintf(msg,
			"killed: %s (snd=%d, rcv=%d, drp=%d, ttl=%d, evil=%d)",
			ip2str(_chost->host->theirip),
			_chost->sent,
			_chost->received,
			_chost->dropped,
			_chost->bigttl,
			_chost->evil);
	log_append(msg);
}

static int enqueue_msg_private (connectedhost* _chost, gnut_msg* _msg, int _tolerance) {
	int len;

	len = g_list_length(_chost->msgq);

	if (len > _tolerance) {
		_chost->dropped++;
		FREE(free_gnut_msg, _msg);
		if (_chost->dropped > MAX_DROPPED) {
			report_kill(_chost);
			stream_sched_kill(_chost->host->s);
		}
		return 1;
	}

	if (_chost->dropped >= 0)
		_chost->dropped--;

	if (!len)
		stream_sched_call(_chost->host->s);

	_chost->msgq = g_list_append(_chost->msgq, _msg);
	return 0;
}

static int enqueue_msg_guaranteed (connectedhost* _chost, gnut_msg* _msg) {
	/* Even guarantees can't be guaranteed! */
	return enqueue_msg_private(_chost, _msg, MAX_GUARANTEEQUEUELEN);
}

static int enqueue_msg (connectedhost* _chost, gnut_msg* _msg) {
	return enqueue_msg_private(_chost, _msg, MAX_UNGUARANTEEQUEUELEN);
}

static gnut_seenmsg* seenthismsg (uuid _msgid) {
	return (gnut_seenmsg*) g_hash_table_lookup(g_seenmsg_hash, &_msgid);
}

static void rememberthismsg (connectedhost* _fromhost, uuid _msgid, int _f) {
	gnut_seenmsg* seenmsg;
	GList* li;

	/* Make sure the cache is of reasonable size. */
	if (g_hash_table_size(g_seenmsg_hash) > MAX_MSGCACHE) {
		seenmsg = (gnut_seenmsg*) g_seenmsg_list->data;

		li = g_seenmsg_list;
		g_seenmsg_list = g_list_remove_link(g_seenmsg_list, g_seenmsg_list);
		FREE(g_list_free_1, li);

		g_hash_table_remove(g_seenmsg_hash, &seenmsg->msgid);

		FREE(free_gnut_seenmsg, seenmsg);
	}

	/* If the msgid is not already in the cache, put it in there. */
	if (!g_hash_table_lookup(g_seenmsg_hash, &_msgid)) {
		seenmsg = new_gnut_seenmsg(_fromhost, _msgid, _f);

		g_hash_table_insert(g_seenmsg_hash,
		                    g_memdup(&_msgid, sizeof(_msgid)),
		                    seenmsg);

		g_seenmsg_list = g_list_append(g_seenmsg_list, seenmsg);
	}
}

static connectedhost* seenclientid (uuid _clientid) {
	gnut_seenclient* seenclient;

	seenclient = (gnut_seenclient*) g_hash_table_lookup(g_clientid_hash,
	                                                    &_clientid);
	return (seenclient ? seenclient->chost : NULL);
}

static void rememberclientid (connectedhost* _fromhost, uuid _clientid) {
	gnut_seenclient* seenclient;
	GList* li;

	/* Remove the _clientid from the cache. */
	if ((seenclient = g_hash_table_lookup(g_clientid_hash, &_clientid))) {
		g_hash_table_remove(g_clientid_hash, &_clientid);

		li = g_list_find(g_clientid_list, seenclient);
		g_clientid_list = g_list_remove_link(g_clientid_list, li);

		FREE(free_gnut_seenclient, seenclient);
	}

	/* Make sure the cache is of reasonable size. */
	if (g_hash_table_size(g_clientid_hash) > MAX_CLIENTIDCACHE) {
		seenclient = (gnut_seenclient*) g_clientid_list->data;

		li = g_clientid_list;
		g_clientid_list = g_list_remove_link(g_clientid_list, g_clientid_list);
		FREE(g_list_free_1, li);

		g_hash_table_remove(g_clientid_hash, &seenclient->clientid);

		FREE(free_gnut_seenclient, seenclient);
	}

	/* Update the _clientid in the cache. */
	seenclient = new_gnut_seenclient(_fromhost, _clientid);

	g_hash_table_insert(g_clientid_hash,
		                g_memdup(&_clientid, sizeof(_clientid)),
		                seenclient);
	g_clientid_list = g_list_append(g_clientid_list, seenclient);
}

static char* canonicalize_query (const char* _query) {
	static char buf[4096];
	char tmpquery[4096];
	char* p;
	char* tok;
	char* last;

	strcpy(tmpquery, _query);
	strcpy(buf, "");

	for (p = tmpquery; *p; p++) {
		if (!isalnum(*p))
			*p = '.';
		else if (isupper(*p))
			*p = tolower(*p);
	}
	for (tok = strtok_r(tmpquery, ".", &last);
		 tok;
		 tok = strtok_r(NULL, ".", &last)) {
		strcat(buf, tok);
	}

	return buf;
}

static int seenthisquery (connectedhost* _chost, const char* _query) {
	return !!g_list_find_custom(_chost->seenqueries, canonicalize_query(_query), (GCompareFunc) strcmp);
}

static void rememberthisquery (connectedhost* _chost, const char* _query) {
	/* Make sure the cache is of reasonable size. */
	if (g_list_length(_chost->seenqueries) > MAX_QUERYCACHE) {
		GList* li;

		FREE(g_free, _chost->seenqueries->data);
		li = _chost->seenqueries;
		_chost->seenqueries =
			g_list_remove_link(_chost->seenqueries, _chost->seenqueries);
		FREE(g_list_free_1, li);
	}
	/* Insert the latest query into the cache. */
	_chost->seenqueries =
		g_list_append(_chost->seenqueries, g_strdup(canonicalize_query(_query)));
}

/* Broadcasting is the most innovative but dangerous thing that
   Gnutella does.  This function should be scrutinized above all
   others.  It's a minor offense to inappropriately drop a
   packet or incorrectly route a packet: clients will just drop
   your mistake.  But it's a major offense to inappropriately
   broadcast a packet.  Or to broadcast an inappropriate packet.
   The network will take a little while to sort that out, and
   meanwhile, everyone will suffer. */
static void forward_broadcast_packet (connectedhost* _fromhost, gnut_hdr* _hdr, void* _payload) {
	int            nbytes;
	char*          buf;
	GList*         li;
	connectedhost* chost;

	if (!_hdr->ttl || _hdr->ttl > MAX_TTL)
		return;

	_hdr->ttl--;
	_hdr->hops++;

	nbytes = formatsize(gnut_hdr_w_format) + _hdr->len;
	buf = (char*) g_malloc(nbytes);
	sprintbstream(buf,
				  gnut_hdr_w_format,
	              &_hdr->msgid,
				  _hdr->f,
				  _hdr->ttl,
				  _hdr->hops,
				  _hdr->len);
	memcpy(buf + formatsize(gnut_hdr_w_format), _payload, _hdr->len);

	for (li = g_connectedhosts; li; li = g_list_next(li)) {
		chost = (connectedhost*) li->data;

		if (chost == _fromhost)
			continue;
                if (chost->protected == 1)
                        continue;
		enqueue_msg(chost, new_gnut_msg(buf, nbytes));
	}

	FREE(g_free, buf);
}

static void forward_routed_packet (connectedhost* _host, gnut_hdr* _hdr, void* _payload) {
	int   nbytes;
	char* buf;

	if (!_hdr->ttl || _hdr->ttl > MAX_TTL)
		return;

	_hdr->ttl--;
	_hdr->hops++;

	nbytes = formatsize(gnut_hdr_w_format) + _hdr->len;
	buf = (char*) g_malloc(nbytes);
	sprintbstream(buf,
				  gnut_hdr_w_format,
	              &_hdr->msgid,
				  _hdr->f,
				  _hdr->ttl,
				  _hdr->hops,
				  _hdr->len);
	memcpy(buf + formatsize(gnut_hdr_w_format), _payload, _hdr->len);

	enqueue_msg(_host, new_gnut_msg(buf, nbytes));

	FREE(g_free, buf);
}

static void originate_broadcast_packet (void* _packet, int _len) {
	GList*         li;
	connectedhost* chost;

	for (li = g_connectedhosts; li; li = g_list_next(li)) {
		chost = (connectedhost*) li->data;
		enqueue_msg_guaranteed(chost, new_gnut_msg(_packet, _len));
	}
}

static void originate_routed_packet (connectedhost* _host, void* _packet, int _len) {
	enqueue_msg_guaranteed(_host, new_gnut_msg(_packet, _len));
}

static void dispatch (connectedhost* _fromhost, gnut_hdr* _hdr, u_int8_t* _payload) {
	int n;

	/* Just alert the user. We drop this packet in the routing. */
	if (_hdr->ttl > MAX_TTL) {
		_fromhost->bigttl++;
		/*printf("big ttl: %s:%d: %d\n",
		       ip2str(_fromhost->host->theirip),
		       _fromhost->host->theirport,
		       _hdr->ttl);*/
	}

	_fromhost->received++;

	switch(_hdr->f) {
		/*
		PING:
			Forward pings to everyone.  If this msgid was seen already,
			drop it.
		*/
		case GNUT_F_PING: {
			if (seenthismsg(_hdr->msgid))
				break;

			rememberthismsg(_fromhost, _hdr->msgid, _hdr->f);

			/* PING packets should not have a payload. */
			if (_hdr->len) {
				printf("ping with %d bytes payload dropped.\n", _hdr->len);
				break;
			}

			forward_broadcast_packet(_fromhost, _hdr, _payload);

			if (ipisroutable(_fromhost->host->ourip)) {
				char msgbuf[64];
				int  len;
				uuid msgid;
				extern int g_listenport;

				len = sprintbstream(msgbuf,
				                    gnut_hdr_w_format gnut_pong_w_format,
									&msgid,
									GNUT_F_PONG,
									_hdr->hops,
									0,
									formatsize(gnut_pong_w_format),
									g_listenport, /* port */
									_fromhost->host->ourip, /* ip */
									0, /* nfiles */
									0); /* nbytes */
						  
				originate_routed_packet(_fromhost, msgbuf, len);
			}

			break;
		}
		/*
		PONG:
			Forward pong down the pipe its ping came from.  Toss it if we
			didn't see the ping.  Why?  If we didn't see the ping yet, that
			means there was a routing mistake by some other server.
		*/
		case GNUT_F_PONG: {
			gnut_pong     pong;
			gnut_seenmsg* seenmsg;

			n = sscanbstream(_payload,
						     _hdr->len,
			                 gnut_pong_r_format,
						     &pong.port,
						     &pong.ip,
						     &pong.nfiles,
						     &pong.nbytes);
			assert(n == formatsize(gnut_pong_r_format));

			/*  Notify the front end of a known host  */
			if (ipisroutable(pong.ip) && pong.port != 0) {
				ServerNode *node = g_new (ServerNode, 1);
			  
				node->server = g_strdup(ip2str(pong.ip));
				node->port   = pong.port;
				node->status = NODE_DISCONNECTED;
				node->files  = pong.nfiles;
				node->bytes  = pong.nbytes;
                                /* If the pong is from a protected client, protect the originator */
                                if (_fromhost->protected == 1)
                                  node->protected = 1;

				/*  If the add fails, this means that the
				 *   server is already known...free up the
				 *   node data structure in this case
				 */
				if (admin_node_add (node))
					g_free (node);
			}
			/*else
				printf("unroutable host pong: %s:%d\n",
				       ip2str(pong.ip), pong.port);*/

			if (!(seenmsg = seenthismsg(_hdr->msgid)))
				break;
			if (seenmsg->f != GNUT_F_PING) 
				break;
			if (!seenmsg->fromhost)
				break;

			forward_routed_packet(seenmsg->fromhost, _hdr, _payload);

			break;
		}
		/*
		PUSH:
			Route the push along the path its hit came from.
		*/
		case GNUT_F_PUSH: {
			gnut_push      push;
			connectedhost* chost;

			/* Not really sure about the routing here.  Perhaps
			   I should make sure I've seen the query and/or query
			   hits related to this push request? Email
			   gene@wego.com if you have any thoughts. */

			n = sscanbstream(_payload,
						     _hdr->len,
						     gnut_push_r_format,
						     &push.clientid,
						     &push.index,
						     &push.ip,
						     &push.port);
			if (n != formatsize(gnut_push_r_format)) {
				printf("evil push packet was %d short\n", n);
				_fromhost->evil++;
				break;
			}

			if ((chost = seenclientid(push.clientid)))
				forward_routed_packet(chost, _hdr, _payload);

			break;
		}
		/*
		QUERY:
			Forward queries to everyone.
		*/
		case GNUT_F_QUERY: {
			gnut_query query;

			/* XXX do something to defend against packets that
				   lack the terminating NULL */

			n = sscanbstream(_payload,
						     _hdr->len,
			                 gnut_query_r_format,
						     &query.minspeed,
						     &query.keywords);

			if (n < formatsize(gnut_query_r_format)) {
				printf("evil query packet was %d short\n", n);
				_fromhost->evil++;
				break;
			}

			if (seenthismsg(_hdr->msgid))
				break;

			rememberthismsg(_fromhost, _hdr->msgid, GNUT_F_QUERY);

			/* Weed out query abuse. */
			if (_hdr->len > 128)
				break;
			if (!strlen(canonicalize_query(query.keywords)))
				break;

			if (seenthisquery(_fromhost, query.keywords)) {
				printf("duplicated query: %s:%d: '%s'\n",
				       ip2str(_fromhost->host->theirip),
				       _fromhost->host->theirport,
				       query.keywords);
				break;
			}
			else
				rememberthisquery(_fromhost, query.keywords);

			monitor_append (query.keywords);

			forward_broadcast_packet(_fromhost, _hdr, _payload);

			break;
		}
		/*
		HITS:
			Forward hits down the pipe its query came from.  Toss it if we
			didn't see the query.  Why?  If we didn't see the query yet,
			that means there was a routing mistake by some other server.
		*/
		case GNUT_F_HITS: {
			gnut_hitset   hitset;
			gnut_hit      hit;
			gnut_seenmsg* seenmsg;
			int           i;
			uuid          clientid;

			if (!(seenmsg = seenthismsg(_hdr->msgid)))
				break;
			if (seenmsg->f != GNUT_F_QUERY)
				break;
			/* Weed out hit abuse. */
			if (_hdr->len > 255*255)
				break;

			sscanbstream(_payload + _hdr->len - sizeof(uuid),
						 sizeof(uuid),
						 "%16d",
						 &clientid);
			rememberclientid(_fromhost, clientid);

			if (!seenmsg->fromhost) {
				n = sscanbstream(_payload,
							     _hdr->len,
							     gnut_hitset_r_format, 
							     &hitset.n,
							     &hitset.port,
							     &hitset.ip,
							     &hitset.speed,
							     &hitset.hits);

				if (n < 0) {
					printf("evil hitset packet was %d short\n", n);
					_fromhost->evil++;
					break;
				}

				_hdr->len -= n;

				for (i = 0; i < hitset.n; i++) {
					SearchResult *result = g_new (SearchResult, 1);
					  
					n = sscanbstream(hitset.hits,
								     _hdr->len,
								     gnut_hit_r_format,
								     &hit.index,
								     &hit.size,
								     &hit.name);

					if (n < 0) {
						printf("evil hit was %d short\n", n);
						_fromhost->evil++;
						break;
					}

					result->search_id = _hdr->msgid;
					result->filename = g_strdup (hit.name);
					result->size = hit.size;
					result->server = g_strdup(ip2str(hitset.ip));
					result->port = hitset.port;
					result->link_speed = hitset.speed;
					result->index = hit.index;
					search_result (result);

					_hdr->len   -= n + strlen(hit.name) + 2;
					hitset.hits += n + strlen(hit.name) + 2;
				}

				break;
			}

			forward_routed_packet(seenmsg->fromhost, _hdr, _payload);
			break;
		}
	}
}

static gboolean seenmsg_hash_clean
	(gpointer _key, gpointer _value, gpointer _privdata) {

	gnut_seenmsg* value = (gnut_seenmsg*) _value;

	if (value->fromhost == (connectedhost*) _privdata) {
		FREE(free_gnut_seenmsg, value);
		return TRUE;
	}

	return FALSE;
}

static gboolean clientid_hash_clean
	(gpointer _key, gpointer _value, gpointer _privdata) {

	gnut_seenclient* value = (gnut_seenclient*) _value;

	if (value->chost == (connectedhost*) _privdata) {
		FREE(free_gnut_seenclient, value);
		return TRUE;
	}

	return FALSE;
}

typedef struct {
	int sndstate;
#define CLI_SND_WAIT_HANDSHAKE   0
#define CLI_SND_BEGIN   1
#define CLI_SND_CMPL    2

	int rcvstate;
#define CLI_RCV_WELCOME 0
#define CLI_RCV_CONNECT 1
#define CLI_RCV_HDR     2
#define CLI_RCV_PAYLOAD 3
	u_int8_t  rbuf[131072];
	gnut_hdr  hdr;

	connectedhost* chost;

	void *ui_data;
} host_privdata;

/* This callback is shared between client and server.
   This makes the code much smaller at the expense of making it much harder to read. */
static void host_cb
(int _s, short _events,
 char* _rbuf, int _rbytes,
 char* _wbuf, int _wbytes,
 void* _privdata) {
	stream_host*   host     = (stream_host*) _privdata;
	host_privdata* privdata = (host_privdata*) host->privdata;

	if (_events & STREAMKILL)
		goto killconn;

	if (_events & STREAMHUP)
		goto killconn;

	/* This only happens for outbound connects.  The inbound STREAMCONNECT
	   signal is delivered to the listen callback. */
	if (_events & STREAMCONNECT) {
		u_int8_t  msgbuf[64];
		int       len;
		uuid      msgid;

		privdata->chost = new_connectedhost(host);
                if(admin_query_protected(privdata->ui_data) == 1) privdata->chost->protected = 1;
		g_connectedhosts = g_list_append(g_connectedhosts, privdata->chost);

		msgid = uuid_create();

		len = sprintbstream(msgbuf,
		                    gnut_hdr_w_format,
							&msgid,
							GNUT_F_PING,
							7,
							0,
							0);

		originate_routed_packet(privdata->chost, msgbuf, len);

		rememberthismsg(NULL, msgid, GNUT_F_PING);

		stream_sched_post_write(host->s,
								"GNUTELLA CONNECT/0.4\n\n",
								strlen("GNUTELLA CONNECT/0.4\n\n"));
		stream_sched_post_read(host->s,
		                       privdata->rbuf,
		                       strlen("GNUTELLA OK\n\n"));

		/*  Update the UI  */
                admin_node_update (privdata->ui_data, NODE_OUTBOUND);
	}

	if (_events & STREAMIN) {
		switch (privdata->rcvstate) {
			case CLI_RCV_CONNECT:
				if (_rbytes != strlen("GNUTELLA CONNECT/0.4\n\n"))
					break;

				if (memcmp(privdata->rbuf,
						   "GNUTELLA CONNECT/0.4",
						   strlen("GNUTELLA CONNECT/0.4")))
					goto killconn;

				stream_sched_post_write(host->s,
				                        "GNUTELLA OK\n\n",
				                        strlen("GNUTELLA OK\n\n"));
				privdata->sndstate = CLI_SND_BEGIN;
				stream_sched_post_read(host->s,
				                       privdata->rbuf,
				                       formatsize(gnut_hdr_r_format));
				privdata->rcvstate = CLI_RCV_HDR;

				break;
			case CLI_RCV_WELCOME:
				if (_rbytes != strlen("GNUTELLA OK\n\n"))
					break;

				if (memcmp(privdata->rbuf,
						   "GNUTELLA OK", strlen("GNUTELLA OK")))
					goto killconn;

				stream_sched_post_read(host->s,
				                       privdata->rbuf,
				                       formatsize(gnut_hdr_r_format));
				privdata->rcvstate = CLI_RCV_HDR;

				stream_sched_call(host->s);
				privdata->sndstate = CLI_SND_BEGIN;

				break;
			case CLI_RCV_HDR:
				if (_rbytes != formatsize(gnut_hdr_r_format))
					break;

				sscanbstream(privdata->rbuf,
							 _rbytes,
				             gnut_hdr_r_format,
				             &privdata->hdr.msgid,
				             &privdata->hdr.f,
				             &privdata->hdr.ttl,
				             &privdata->hdr.hops,
				             &privdata->hdr.len);

				if (privdata->hdr.len == 0) {
					if (privdata->hdr.f == GNUT_F_PING)
						dispatch(privdata->chost, &privdata->hdr, NULL);
					stream_sched_post_read(host->s,
					                       privdata->rbuf,
					                       formatsize(gnut_hdr_r_format));
					break;
				}

				if (privdata->hdr.len < 0 || privdata->hdr.len > sizeof(privdata->rbuf))
					goto killconn;

				stream_sched_post_read(host->s, privdata->rbuf, privdata->hdr.len);
				privdata->rcvstate = CLI_RCV_PAYLOAD;
				break;
			case CLI_RCV_PAYLOAD:
				if (_rbytes != privdata->hdr.len)
					break;

				dispatch(privdata->chost, &privdata->hdr, privdata->rbuf);

				stream_sched_post_read(host->s,
				                       privdata->rbuf,
				                       formatsize(gnut_hdr_r_format));
				privdata->rcvstate = CLI_RCV_HDR;
				break;
		}
	}

	if (_events & STREAMOUT) {
		switch (privdata->sndstate) {
			case CLI_SND_WAIT_HANDSHAKE:
				break;
			case CLI_SND_BEGIN: {
				gnut_msg* msg;

				if (!privdata->chost->msgq)
					break;

				msg = (gnut_msg*) privdata->chost->msgq->data;
				stream_sched_post_write(host->s, msg->payload, msg->len);

				privdata->sndstate = CLI_SND_CMPL;
				break;
			}
			case CLI_SND_CMPL: {
				GList* li;

				privdata->chost->sent++;

				/* Clean up the message we just sent. */
				FREE(free_gnut_msg, (gnut_msg*) privdata->chost->msgq->data);

				li = privdata->chost->msgq;
				privdata->chost->msgq =
					g_list_remove_link(privdata->chost->msgq,
					                   privdata->chost->msgq);
				FREE(g_list_free_1, li);

				stream_sched_call(host->s);
				privdata->sndstate = CLI_SND_BEGIN;

				break;
			}
		}
	}

	return;

killconn: {
	GList* li;
	GList* dead;

	/*  Update the UI  */
	admin_node_update (privdata->ui_data, NODE_FAILURE);

	/* Clear out the seen messages.  Otherwise, if we try to route a packet
	   on a dead host, we might get a segfault. */
	for (li = g_seenmsg_list; li; ) {
		dead = li;
		li = g_list_next(li);
		if (privdata->chost == ((gnut_seenmsg*) dead->data)->fromhost) {
			g_seenmsg_list = g_list_remove_link(g_seenmsg_list, dead);
			FREE(g_list_free_1, dead);
		}
	}
	g_hash_table_foreach_remove(g_seenmsg_hash,
	                            seenmsg_hash_clean,
	                            privdata->chost);

	/* Clear out the clientids.  Otherwise, if we try to route a packet
	   on a dead host, we might get a segfault. */
	for (li = g_clientid_list; li; ) {
		dead = li;
		li = g_list_next(li);
		if (privdata->chost == ((gnut_seenclient*) dead->data)->chost) {
			g_clientid_list = g_list_remove_link(g_clientid_list, dead);
			FREE(g_list_free_1, dead);
		}
	}
	g_hash_table_foreach_remove(g_clientid_hash,
								clientid_hash_clean,
								privdata->chost);

	/* Remove the host from the list of connected hosts.  The host will
	   only be in this list if the host actually connected. */
	if (privdata->chost) {
		report_kill(privdata->chost);
		assert(li = g_list_find(g_connectedhosts, privdata->chost));
		FREE(free_connectedhost, (connectedhost*) li->data);
		g_connectedhosts = g_list_remove_link(g_connectedhosts, li);
		FREE(g_list_free_1, li);
	}

	stream_host_close(host);
	FREE(g_free, privdata);

	return;
	}
}

static void listen_cb
(int _s, short _events,
 char* _rbuf, int _rbytes,
 char* _wbuf, int _wbytes,
 void* _privdata) {
	stream_host*   host     = (stream_host*) _privdata;
	host_privdata* privdata = (host_privdata*) host->privdata;

	privdata = (host_privdata*) g_malloc(sizeof(host_privdata));

	host = stream_serv_accept(_s,
	                          host_cb,
	                          privdata);
	if (!host) {
		stream_err("dead connect");
		FREE(g_free, privdata);
		return;
	}

	privdata->chost = new_connectedhost(host);
	g_connectedhosts = g_list_append(g_connectedhosts, privdata->chost);

	privdata->rcvstate = CLI_RCV_CONNECT;
	privdata->sndstate = CLI_SND_WAIT_HANDSHAKE;
	stream_sched_post_read(host->s,
						   privdata->rbuf,
						   strlen("GNUTELLA CONNECT/0.4\n\n"));

	{
	ServerNode *node = g_new (ServerNode, 1);

	node->server = g_strdup(ip2str(host->theirip));
	node->port   = host->theirport;
	node->status = NODE_CONNECTING;
	node->files  = 0;
	node->bytes  = 0;

	privdata->ui_data = admin_node_add(node);

	if (privdata->ui_data)
		g_free (node);
	else
		privdata->ui_data = node;

	admin_node_update (privdata->ui_data, NODE_INBOUND);
	}
}

int network_start_client (char* _dns, int _port, void *ui_data) {
	stream_host*   host;
	host_privdata* privdata;

	privdata = (host_privdata*) g_malloc(sizeof(host_privdata));
	privdata->ui_data = ui_data;
	privdata->chost   = NULL;

	if (!(host = stream_cli_new(_dns, _port, privdata, host_cb)))
		return 1;
	privdata->sndstate = CLI_SND_WAIT_HANDSHAKE;
	privdata->rcvstate = CLI_RCV_WELCOME;

	return 0;
}

int network_start_server (int _port) {
	stream_host* host;

	if (!(host = stream_serv_new(_port, 10, NULL, listen_cb)))
		return 1;

	return 0;
}

uuid network_send_query (char* _query) {
	uuid           msgid;
	int            len;
	u_int8_t       msgbuf[1024];

	msgid = uuid_create();

	len = sprintbstream(msgbuf,
						gnut_hdr_w_format gnut_query_w_format,
						&msgid,
						GNUT_F_QUERY,
						7,
						0,
						formatsize(gnut_query_w_format) + strlen(_query) + 1,
						0,
						strlen(_query) + 1,
						_query);

	originate_broadcast_packet(msgbuf, len);

	rememberthismsg(NULL, msgid, GNUT_F_QUERY);

	return msgid;
}

void network_init (void) {
	struct sigaction sa;

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;

	sa.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sa, NULL);
	sigaction(SIGTRAP, &sa, NULL);

	stream_openlog("gnucleus");

	uuid_init();

	g_connectedhosts = NULL;
	g_seenmsg_hash = g_hash_table_new(uuidphash, uuidpcmp);
	g_seenmsg_list = NULL;
	g_clientid_hash = g_hash_table_new(uuidphash, uuidpcmp);
	g_clientid_list = NULL;

	stream_sched_init();
}
