/*
 *  Gnubile - Gnutella software for Unix
 *  Copyright (C) 2000  Spencer Kimball and Gene Kan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "stream_config.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <netdb.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "stream_sched.h"
#include "stream_inet.h"
#include "stream_std.h"

typedef struct _stream_buf {
	char* buf;
	int bytes;
	int len;
} stream_buf;

typedef struct _stream_fd {
	stream_sched_cb cb;
	void* privdata;
} stream_fd;

static stream_fd fdtable[FD_SETSIZE];
static stream_buf wbufs[FD_SETSIZE], rbufs[FD_SETSIZE];
static char fdtypes[FD_SETSIZE];
#define FDTYPE_NORMAL	0
#define FDTYPE_LISTEN	1
#define FDTYPE_CONNECT	2
static int kills[FD_SETSIZE];
static int nkills = 0;
static int calls[FD_SETSIZE];
static int ncalls = 0;

static stream_sched sched;

/*
 * Due to the way sockets work, we NEVER turn off STREAMIN.  That's the
 * only reliable way to be notified of a hangup.  Well...if you really
 * know what you're doing...  STREAMIN can be used to throttle.
 */

void stream_sched_init (void) {
	sched.nfds = 0;
	sched.maxfd = 0;

	FD_ZERO(&sched.r);
	FD_ZERO(&sched.w);
}

int stream_sched_run(int _to_ms) {
	struct timeval tv;
	short events;
	fd_set tr, tw;

	int i;
	int n;
	int nactive;

	int bytes;

	memcpy(&tr, &sched.r, sizeof(fd_set));
	memcpy(&tw, &sched.w, sizeof(fd_set));

	tv.tv_sec = _to_ms / 1000;
	tv.tv_usec = (_to_ms - tv.tv_sec * 1000) * 1000;
	nactive = select(sched.maxfd, &tr, &tw, NULL,
	                 (_to_ms == -1) ? NULL : &tv);
	if (nactive == -1 && errno != EINTR) {
		stream_err("select");
		return -1;
	}

	for (i = 0, n = 0; n < nactive && i < sched.maxfd; i++) {
		events = 0;
		if (FD_ISSET(i, &tr)) {
			if (fdtypes[i] == FDTYPE_LISTEN) {
				events |= STREAMCONNECT;
				goto activate;
			}

			bytes = read(i,
			             rbufs[i].buf + rbufs[i].bytes,
			             rbufs[i].len - rbufs[i].bytes);
			switch (bytes) {
				case -1:
					if (errno != EAGAIN) {
						stream_err("read error");
						events |= STREAMHUP;
					}
					break;
				case 0:
					events |= STREAMHUP;
					break;
				default:
					rbufs[i].bytes += bytes;
					events |= STREAMIN;
					break;
			}
		}
		if (FD_ISSET(i, &tw)) {
			if (fdtypes[i] == FDTYPE_CONNECT) {
				struct sockaddr_in addr;
				int addrlen;

				fdtypes[i] = FDTYPE_NORMAL;
				stream_sched_set_fd_ev(i, 0);
				events |= STREAMCONNECT;

				addrlen = sizeof(addr);
				if (getsockname(i, (struct sockaddr*) &addr, &addrlen) == -1) {
					assert(errno == ENOTCONN || errno == EINVAL);
					events = STREAMHUP;
					goto activate;
				}
				((stream_host*) fdtable[i].privdata)->ourip = ntohl(addr.sin_addr.s_addr);
				((stream_host*) fdtable[i].privdata)->ourport = ntohs(addr.sin_port);

				addrlen = sizeof(addr);
				if (getpeername(i, (struct sockaddr*) &addr, &addrlen) == -1) {
					assert(errno == ENOTCONN || errno == EINVAL);
					events = STREAMHUP;
					goto activate;
				}
				((stream_host*) fdtable[i].privdata)->theirip = ntohl(addr.sin_addr.s_addr);
				((stream_host*) fdtable[i].privdata)->theirport = ntohs(addr.sin_port);

				if (!wbufs[i].len)
					goto activate;
			}

			bytes = write(i,
						  wbufs[i].buf + wbufs[i].bytes,
						  wbufs[i].len - wbufs[i].bytes);
			switch (bytes) {
				case -1:
					if (errno != EAGAIN) {
						stream_err("write error");
						events |= STREAMHUP;
					}
					break;
				case 0:
					if (wbufs[i].len)
						stream_msg("Strange. 0 lengthed write.");
					break;
				default:
					wbufs[i].bytes += bytes;
					break;
			}

			if (wbufs[i].bytes == wbufs[i].len) {
				events |= STREAMOUT;
				stream_sched_sub_fd_ev(i, STREAMOUT);
			}
		}

		if (!events)
			continue;

activate:
		n++;
		(*fdtable[i].cb)(i, events,
		                 rbufs[i].buf, rbufs[i].bytes,
		                 wbufs[i].buf, wbufs[i].bytes,
		                 fdtable[i].privdata);
	}

	while (nkills || ncalls) {
		int tnkills;
		int tncalls;

		tnkills = nkills;
		nkills = 0;

		/* Process explicit kills. */
		for (i = 0; i < tnkills; i++) {
			/* Make sure we didn't kill this host already. */
			if (!fdtable[kills[i]].cb)
				continue;

			(*fdtable[kills[i]].cb)(kills[i], STREAMKILL,
		                            rbufs[kills[i]].buf, rbufs[kills[i]].bytes,
		                            wbufs[kills[i]].buf, wbufs[kills[i]].bytes,
		                            fdtable[kills[i]].privdata);
		}

		/* Do this here because ncalls could have changed in
		   the kill-processing loop. */

		tncalls = ncalls;
		ncalls = 0;

		/* Historically, explicit calls were done with by:
		   stream_sched_post_write(s, "", 0);
		   That works pretty well, but it's expensive.  This
		   is riskier, but it's cheaper.  Careful. */

		/* Process explicit calls. */
		for (i = 0; i < tncalls; i++) {
			/* Make sure we didn't kill this host already. */
			if (!fdtable[calls[i]].cb)
				continue;

			(*fdtable[calls[i]].cb)(calls[i], STREAMCALL | STREAMOUT,
		                            rbufs[calls[i]].buf, rbufs[calls[i]].bytes,
		                            wbufs[calls[i]].buf, wbufs[calls[i]].bytes,
		                            fdtable[calls[i]].privdata);
		}

		/* Now, repeat this cycle until we are satisfied. */
	}

	return nactive;
}

/*
 * Returns 0 if okay or -1 if error.  The only error is if the fd is out of
 * range.
 */
static int private_add_fd
(int _fd, int _fdtype, stream_sched_cb _cb,
 void* _privdata) {
	/* Protect ourselves. */
	if (_fd < 0 || _fd >= FD_SETSIZE)
		return -1;

	fdtypes[_fd] = _fdtype;

	fdtable[_fd].cb = _cb;
	fdtable[_fd].privdata = _privdata;

	rbufs[_fd].buf = NULL; rbufs[_fd].len = 0;
	wbufs[_fd].buf = NULL; wbufs[_fd].len = 0;

	return 0;
}

int stream_sched_add_fd
(int _fd, stream_sched_cb _cb, void* _privdata) {
	return private_add_fd(_fd, FDTYPE_NORMAL, _cb, _privdata);
}

int stream_sched_add_listen_fd
(int _fd, stream_sched_cb _cb, void* _privdata) {
	if (private_add_fd(_fd, FDTYPE_LISTEN, _cb, _privdata) == -1)
		return -1;

	stream_sched_set_fd_ev(_fd, STREAMIN);

	return 0;
}

int stream_sched_add_connect_fd
(int _fd, stream_sched_cb _cb, void* _privdata) {
	if (private_add_fd(_fd, FDTYPE_CONNECT, _cb, _privdata) == -1)
		return -1;

	stream_sched_set_fd_ev(_fd, STREAMOUT);

	return 0;
}

void stream_sched_del_fd (int _fd) {
	/* Protect ourselves. */
	if (_fd < 0 || _fd >= FD_SETSIZE)
		return;

	rbufs[_fd].buf = NULL;
	wbufs[_fd].buf = NULL;

	fdtable[_fd].cb = NULL;
	fdtable[_fd].privdata = NULL;

	stream_sched_set_fd_ev(_fd, 0);
}

void stream_sched_set_fd_ev (int _fd, short _events) {
	/*int i;*/

	/* Protect ourselves. */
	if (_fd < 0 || _fd >= FD_SETSIZE)
		return;

	FD_CLR(_fd, &sched.r);
	FD_CLR(_fd, &sched.w);

	if (_events & STREAMIN) FD_SET(_fd, &sched.r);
	if (_events & STREAMOUT) FD_SET(_fd, &sched.w);

	/*_sched->maxfd = 0;
	for (i = 0; i < FD_SETSIZE; i++) {
		if (FD_ISSET(i, &_sched->r) || FD_ISSET(i, &_sched->w))
			_sched->maxfd = i + 1;
	}*/

	sched.maxfd = FD_SETSIZE;
}

short stream_sched_get_fd_ev (int _fd) {
	short events;

	events = 0;

	if (FD_ISSET(_fd, &sched.r)) events |= STREAMIN;
	if (FD_ISSET(_fd, &sched.w)) events |= STREAMOUT;

	return events;
}

void stream_sched_add_fd_ev (int _fd, short _events) {
	stream_sched_set_fd_ev(_fd,
						   _events | stream_sched_get_fd_ev(_fd));
}

void stream_sched_sub_fd_ev (int _fd, short _events) {
	stream_sched_set_fd_ev(_fd,
						   ~_events & stream_sched_get_fd_ev(_fd));
}

void stream_sched_post_read
(int _fd, char* _buf, int _len) {
	rbufs[_fd].buf = _buf;
	rbufs[_fd].len = _len;
	rbufs[_fd].bytes = 0;

	stream_sched_add_fd_ev(_fd, STREAMIN);
}

void stream_sched_consume
(int _fd, int _n) {
	rbufs[_fd].bytes -= _n;
	assert(rbufs[_fd].bytes >= 0);
	memmove(rbufs[_fd].buf, rbufs[_fd].buf + _n, rbufs[_fd].bytes);
}

void stream_sched_post_write
(int _fd, char* _buf, int _len) {
	wbufs[_fd].buf = _buf;
	wbufs[_fd].len = _len;
	wbufs[_fd].bytes = 0;

	stream_sched_add_fd_ev(_fd, STREAMOUT);
}

void stream_sched_kill (int _fd) {
	kills[nkills++] = _fd;
}

void stream_sched_call (int _fd) {
	calls[ncalls++] = _fd;
}
