/*
 *  Gnubile - Gnutella software for Unix
 *  Copyright (C) 2000  Spencer Kimball and Gene Kan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "stream_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include "stream_sched.h"
#include "stream_inet.h"
#include "stream_std.h"

stream_host* stream_serv_new
(u_int16_t _port, int _backlog, void* _privdata, stream_sched_cb _cb) {
	stream_host* server;
	struct sockaddr_in sin;
	int one = 1;
	struct linger linger = { 0, 0 };

	server = (stream_host*) malloc(sizeof(stream_host));

	server->s = socket(PF_INET, SOCK_STREAM, 0);
	if (server->s == -1) {
		stream_err("socket");
		goto bad;
	}
	server->privdata = _privdata;

	if (fcntl(server->s, F_SETFL, O_NONBLOCK) == -1) {
		stream_err("fcntl");
		goto bad1;
	}

	if (setsockopt(server->s, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
		stream_err("setsockopt reuseaddr");
		goto bad1;
	}
	if (setsockopt(server->s, SOL_SOCKET, SO_OOBINLINE, &one, sizeof(one)) == -1) {
		stream_err("setsockopt oobinline");
		goto bad1;
	}
	if (setsockopt(server->s, SOL_SOCKET, SO_KEEPALIVE, &one, sizeof(one)) == -1) {
		stream_err("setsockopt keepalive");
		goto bad1;
	}
	if (setsockopt(server->s, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger)) == -1) {
		stream_err("setsockopt linger");
		goto bad1;
	}

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = htons(_port);
	if (bind(server->s, (struct sockaddr*) &sin, sizeof(sin)) == -1) {
		stream_err("bind");
		goto bad1;
	}

	if (listen(server->s, _backlog) == -1) {
		stream_err("listen");
		goto bad1;
	}

	if (stream_sched_add_listen_fd(server->s, _cb, server) == -1) {
		stream_err("add_listen_fd");
		goto bad1;
	}

	return server;

bad1:	close(server->s);
bad:	free(server);
	return NULL;
}

void stream_host_close (stream_host* _host) {
	stream_sched_del_fd(_host->s);
	close(_host->s);
	free(_host);
}

stream_host* stream_serv_accept
(int _s, stream_sched_cb _cb, void* _privdata) {
	int news;
	stream_host* server;
	struct sockaddr_in addr;
	int addrlen;

	server = (stream_host*) malloc(sizeof(stream_host));

	addrlen = sizeof(addr);
	news = accept(_s, (struct sockaddr*) &addr, &addrlen);

	if (news == -1)
		goto bad;

	server->s = news;
	if (stream_sched_add_fd(server->s, _cb, server) == -1) {
		stream_msg("out of file descriptors\n");
		goto bad1;
	}

	server->privdata = _privdata;

	server->theirip = ntohl(addr.sin_addr.s_addr);
	server->theirport = ntohs(addr.sin_port);

	addrlen = sizeof(addr);
	if (getsockname(server->s, (struct sockaddr*) &addr, &addrlen) == -1) {
		stream_err("getsockname");
		goto bad1;
	}
	server->ourip = ntohl(addr.sin_addr.s_addr);
	server->ourport = ntohs(addr.sin_port);


	return server;

bad1:
	close(server->s);
bad:
	free(server);
	return NULL;
}
