/*
 *  Gnubile - Gnutella software for Unix
 *  Copyright (C) 2000  Spencer Kimball and Gene Kan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* * * * * * * * * * * * * * * * * * * *
 * textutils.c
 *
 * Various functions to manipulate text strings.
 *
 * 7-95 Version 1
 * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <assert.h>
#include <netinet/in.h>

#include "textutils.h"

#ifndef MAX
#define MAX(x,y) (((x) > (y)) ? (x) : (y))
#endif

/* chop */
/* Chop the end character off a string. */
/* string: string to chop. */
/* return: string. */

char *chop (char *string)
   {
   if (!string)
      return NULL;
   else
      {
      string[MAX(strlen (string) - 1, 0)] = 0;
      return string;
      }
   }

/* cut */
/* Cut the first character off a string. */
/* string: string to cut. */
/* return: string. */

char *cut (char *string)
   {
   if (!string)
      return NULL;
   else
      {
      strcpy (string, string + 1);
      return string;
      }
   }

/* stripspaces */
/* Strips LEADING and/or TRAILING spaces from a string depending on mode of
   operation. */
/* s: the string to operate on. */
/* mode: LEADING, TRAILING.  Can be binary OR'd to do both. */
/* return: s. */ 

char *stripspaces (char *s, int mode)
   {
   if (s && strlen (s))
      {
      if (mode & LEADING)
         {
         while (strlen (s) && isspace (s[0]))
            cut (s);
	 }
      if (mode & TRAILING)
         {
         while (strlen (s) && isspace (s[strlen (s) - 1]))
            chop (s);
	 }
      }
   return s;
   }

/* hashpjw */
/* PJW's hash function.  Very good...use it. */
/* s: string to hash. */
/* return: virgin hash value. */

u_int32_t hashpjw (const void* s, int len)
   {
   const char *p;
   unsigned long int h;
   unsigned long topnibble;

   h = 479001599; /* start things with a very large prime. */
		   
   for (p = (char*) s; len; p++, len--)
      {
      h = (h << 4) + *p;
      topnibble = h & 0xf0000000;     
      h = h ^ (topnibble >> 24) ^ topnibble;
      }
   return h;
   }

char* strjoindup (const char* _joinsep, int _n, ...) {
	va_list ap;
	int joinseplen;
	int totlen;
	char* dst;
	int i;

	joinseplen = strlen(_joinsep);

	va_start(ap, _n);
	for (i = 0, totlen = 0; i < _n; i++)
		totlen += strlen(va_arg(ap, char*) ? : "") + joinseplen;
	totlen++;
	va_end(ap);

	dst = (char*) malloc(totlen);

	va_start(ap, _n);
	for (i = 0, strcpy(dst, ""); i < _n; i++) {
		strcat(dst, va_arg(ap, char*) ? : "");
		if (i < _n - 1)
			strcat(dst, _joinsep);
	}
	va_end(ap);

	return dst;
}

int formatsize (const char* _format) {
	const char* p;
	      char* ep;
	      int   n;
	      int   tot;

	tot = 0;
	for (p = _format; *p; p++) {
		assert(*p == '%');

		p++;

		n = strtol(p, &ep, 10);
		p = ep;

		if (*p == '?') {
			p++;
			assert(*p == 'p');
			n = 0;
		}

		tot += n;
	}

	return tot;
}

#define get1byte(buf) ((u_int8_t) (buf)[0])
#define get2bytes(buf) ((u_int16_t) \
	(((buf)[0] << 8) & 0xff00) | \
	(((buf)[1] << 0) & 0x00ff)   \
)
#define get4bytes(buf) ((u_int32_t) \
	(((buf)[0] << 24) & 0xff000000) | \
	(((buf)[1] << 16) & 0x00ff0000) | \
	(((buf)[2] <<  8) & 0x0000ff00) | \
	(((buf)[3] <<  0) & 0x000000ff)   \
)

#define put1byte(dst,src) (((u_int8_t*) dst)[0] = ((u_int8_t*) src)[0])
#define put2bytes(dst,src) { \
	((u_int8_t*) dst)[0] = ((u_int8_t*) src)[1]; \
	((u_int8_t*) dst)[1] = ((u_int8_t*) src)[0]; \
}
#define put4bytes(dst,src) { \
	((u_int8_t*) dst)[0] = ((u_int8_t*) src)[3]; \
	((u_int8_t*) dst)[1] = ((u_int8_t*) src)[2]; \
	((u_int8_t*) dst)[2] = ((u_int8_t*) src)[1]; \
	((u_int8_t*) dst)[3] = ((u_int8_t*) src)[0]; \
}

/* Format is "(%[1-9]+[sd])+".  The digits mean the length of the field,
   the s or d means to _s_wap or _d_on't swap the byte order.  * means just
   throw it away.  p means set a pointer into the stream.  We always convert
   the stuff to host byte order because we're _scan_ning a stream.
   _streamlen is the number of bytes in the _stream.

   Returns the number of bytes, or the negation of the number of bytes
   required to fulfill the the format.  (Total of the digits.) */
int sscanbstream (u_int8_t* _stream, int _streamlen, const char* _format, ...) {
	const u_int8_t* p;
	      u_int8_t* ep;
	      int       n;
	      int       tot;
	      va_list   ap;

	va_start(ap, _format);

	tot = formatsize(_format);
	if (tot > _streamlen)
		return (_streamlen - tot);

	tot = 0;
	for (p = _format; *p; p++) {
		assert(*p == '%');

		p++;

		n = strtol(p, (char**) &ep, 10);
		p = ep;

		assert(*p == 's' || *p == 'd' || *p == '*' || *p == 'p');

		if (*p == 'p')
			*(va_arg(ap, void**)) = _stream;
		else if (*p != '*') {
			switch (n) {
				case 1:
					*(va_arg(ap, u_int8_t*)) = get1byte(_stream);
					break;
				case 2:
					*(va_arg(ap, u_int16_t*)) =
						((*p == 's') ? ntohs(get2bytes(_stream)) : get2bytes(_stream));
					break;
				case 4:
					*(va_arg(ap, u_int32_t*)) =
						((*p == 's') ? ntohl(get4bytes(_stream)) : get4bytes(_stream));
					break;
				default:
					memcpy(va_arg(ap, u_int8_t*), _stream, n);
					break;
			}
		}

		_stream += n;
		tot += n;
	}

	va_end(ap);

	return tot;
}

/* Format is "(%[1-9]+[sd])+".  The digits mean the length of the field,
   the s or d means _s_wap or _d_on't swap the byte order.  We always
   convert the stuff to network byte order because we're _print_ing a
   stream.

   Returns the number of bytes.  (Total of the digits.) */
int vsprintbstream (u_int8_t* _stream, const char* _format, va_list _ap) {
	const u_int8_t* p;
	      u_int8_t* ep;
	      int       n;
	      int       tot;
	      void*     voidp;

		  u_int16_t h;
		  u_int32_t l;

	tot = 0;
	for (p = _format; *p; p++) {
		assert(*p == '%');

		p++;

		n = strtol(p, (char**) &ep, 10);
		p = ep;

		assert(*p == 's' || *p == 'd' || *p == 'p' || *p == '?');

		if (*p == '?') {
			p++;
			assert(*p == 'p');
			n = va_arg(_ap, int);
			voidp = va_arg(_ap, void*);
			memcpy(_stream, voidp, n);
		}
		else {
			switch (n) {
				case 1:
					put1byte(_stream, &va_arg(_ap, u_int8_t));
					break;
				case 2:
					h = va_arg(_ap, u_int16_t);
					if (*p == 's')
						h = htons(h);
					put2bytes(_stream, &h);
					break;
				case 4:
					l = va_arg(_ap, u_int32_t);
					if (*p == 's')
						l = htonl(l);
					put4bytes(_stream, &l);
					break;
				default:
					memcpy(_stream, va_arg(_ap, u_int8_t*), n);
					break;
			}
		}

		_stream += n;
		tot += n;
	}

	return tot;
}

/* Format is "(%[1-9]+[sd])+".  The digits mean the length of the field, the s or d
   means _s_wap or _d_on't swap the byte order.  We always convert
   the stuff to network byte order because we're _print_ing a stream.

   Returns the number of bytes.  (Total of the digits.) */
int sprintbstream (u_int8_t* _stream, const char* _format, ...) {
	va_list ap;
	int     n;

	va_start(ap, _format);
	n = vsprintbstream(_stream, _format, ap);
	va_end(ap);

	return n;
}
