/* Josh Pieper, (c) 2000
   Robert Munafo, (c) 2001

   This file is distributed under the GPL, see file COPYING for details */

/* Proposed optimization of host cache:

         packets ---.
                     v
   full cache --> LRU cache --> new connections
    (50000)        (1000)

 - full cache is maintained just as before: hosts are added by host_restore()
   and by sniffing IP's from packets, and hosts are removed when a connection
   closes or fails to open.

 - LRU cache is filled two ways:
    - any IP seen in a packet, if not currently in the LRU cache,
      is copied and added. If this makes the LRU cache overfull,
      the oldest entries are deleted
    - if LRU cache is less than half-full, random hosts from the
      full cache are copied and added (but left in the full cache)
 - each time a new random connection is started, it gets its IP from
   the end (oldest) entry on the LRU cache, and that entry is deleted

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#endif

#include "conf.h"
#include "host.h"
#include "net.h"
#include "list.h"
#include "hash.h"

Gnut_Hash *host_list = 0;

uint hosts_num = 0;
uint64 hosts_files = 0;
uint64 hosts_size = 0;

#define HLRUSIZE 1024
host_entry * host_lru;
int next_lrup;

#ifndef PTHREADS_DRAFT4
pthread_mutex_t host_mutex=PTHREAD_MUTEX_INITIALIZER;
#else
pthread_mutex_t host_mutex;
#endif

void fre_hen(host_entry **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

int host_class(uchar *ip)
{
  /* The VPN filtering needs to be conditional: If a VPN host
   * is found, we can try connecting to it and if successful, can use it.
   * But we shouldn't blindly add all VPN addresses because a lot come
   * in from other VPNs elsewhere on the Internet and those should be
   * ignored */
  if (ip[0] == 10) {
    /* Class A virtual private network */
    return VPN_A;
  }

  if (ip[0] == 172) {
    if ((ip[1] >= 16) && (ip[1] <= 31)) {
      /* Class B virtual private network */
      return VPN_B;
    }
  }

  if ((ip[0] == 192) && (ip[1] == 168)) {
    /* Class C virtual private network */
    return VPN_C;
  }

  if ((ip[0] == 169) && (ip[1] == 254)) {
    /* APIPA address */
    return VPN_P;
  }

  if (ip[0] == 127) {
    /* Local loopback range */
    return BAD_ADDR;
  }

  if (ip[0] == 0) {
    /* class A network 0 -- old-style broadcast */
    return BAD_ADDR;
  }

  if ((ip[0] == 192) && (ip[1] == 0) && (ip[2] == 2)) {
    /* TEST-NET */
    return BAD_ADDR;
  }

  if (ip[0] >= 224) {
    /* Class D multicast and class E reserved
     * %%% not sure about this -- parts of this range might be another type
     * of VPN address within certain institutional networks */
    return BAD_ADDR;
  }

  return 0;
}

int can_connect(uchar *from_ip, uchar *to_ip)
{
  if (gc_no_rfc_1597) {
    return 1;
  }
  if (host_class(to_ip)) {
    /* to_ip is a VPN address: they're connectable only if in the same
     * VPN class as from_ip (this happens when both are within the same
     * institution's private IP network) */
    if (host_class(from_ip) & host_class(to_ip)) {
      return 1;
    }
    return 0;
  }
  /* Else: to_ip is a normal address */
  return 1;
}

int host_ok_for_get(uchar *ip)
{
  return(can_connect(net_local_ip(), ip));
}

int host_ok_for_push(uchar *ip)
{
  return(can_connect(ip, net_local_ip()));
}

int host_ok_for_either(uchar *ip)
{
  return(can_connect(net_local_ip(), ip) || can_connect(ip, net_local_ip()));
}

int host_ok_for_both(uchar *ip)
{
  return(can_connect(net_local_ip(), ip) && can_connect(ip, net_local_ip()));
}

int host_ok_for_neither(uchar *ip)
{
  if(can_connect(net_local_ip(), ip) || can_connect(ip, net_local_ip())) {
    return 0;
  }
  return 1;
}

uchar host_hash(void *a)
{
  int i;
  uchar b=0;
  
  host_entry *he;
  he=a;
  
  for (i=0;i<4;i++) {
    b ^= he->ip[i];
  }
  b ^= he->port[0];
  b ^= he->port[1];
  return b;
}

int host_compare(void *a, void *b)
{
  host_entry *he1;
  host_entry *he2;
  he1 = (host_entry *) a;
  he2 = (host_entry *) b;

  if ((memcmp(he1->ip, he2->ip, 4) == 0)
      && (memcmp(he1->port, he2->port, 2) == 0)) {
    return 0;
  }
  return -1;
}

void clear_he(host_entry * he)
{
  he->ip[0] = 0;
  he->ip[1] = 0;
  he->ip[2] = 0;
  he->ip[3] = 0;
  he->port[0] = 0;
  he->port[1] = 0;
  he->files = 0;
  he->bytes = 0;
}

int host_init()
{
  int i;

  gd_s(3, "host_init entering\n");
  host_list = gnut_hash_new(host_hash, host_compare);

  host_lru = ymaloc(HLRUSIZE * sizeof(host_entry), 154);

  for(i=0; i<HLRUSIZE; i++) {
    clear_he(&(host_lru[i]));
  }

  return 0;
}

/* Gets a random host_entry from the recent-hosts array */
void host_get_lru(host_entry *he)
{
  int going;
  int countdown;

  going = 1;
  countdown = 20;
  while(going) {
    next_lrup = (next_lrup + (rand() & 0xff)) % HLRUSIZE;
    if (host_lru[next_lrup].ip[0]) {
      *he = host_lru[next_lrup];
      clear_he(&(host_lru[next_lrup]));
      going = 0;
    }
    if (countdown <= 0) {
      clear_he(he);
      going = 0;
    }
    countdown--;
  }
}

/* Store a host_entry in the recent-hosts array */
void host_put_lru(host_entry *he)
{
  uint32 crc;
  char *c;
  int i;

  c = (char *) he;

  crc32_start(&crc);
  for(i=0; i<6; i++) {
    crc32_add8(&crc, c[i]);
  }
  i = crc % HLRUSIZE;
  host_lru[i] = *he;
}

int host_add(host_entry *he)
{
  gd_s(3, "host_add entering host_list=");
  gd_p(3, host_list);
  gd_s(3, " he=");
  gd_p(3, he);
  gd_s(3, "\n");

  pthread_mutex_lock(&host_mutex);

  if (gnut_hash_find(host_list, he)) {
    /* Already there */
    /* printf("host_add skip duplicate %i.%i.%i.%i\n", he->ip[0], he->ip[1],
       he->ip[2], he->ip[3]); */
  } else {
    /* It's new */
    gnut_hash_insert(host_list, he, 499);

    hosts_num++;
    hosts_files += he->files;
    hosts_size += (he->bytes/1024);
  }

  pthread_mutex_unlock(&host_mutex);
  
  gd_s(3,"host_add returning\n");
  return 0;
}

/* host_restore loads the host list file (usually ~/.gnut_hosts)
 * by calling host.add() on each entry to add the host IPs to the
 * host IP hashtable. */
int host_restore(char *fname)
{
  FILE *fp;
  host_entry *he;
  char buf[40];
  int a[4];
  char *ptr;
  uint16 st;

  he = 0;

  gd_s(2, "host_restore entering\n");

  fp = fopen(fname,"r");

  if (fp == 0) {
    /* perror("host_restore, fopen"); */
    gd_s(1, "host_restore couldn't open file: ");
    gd_s(1, fname);
    gd_s(1, "\n");
    return -1;
  }

  host_clear();

  gd_s(3, "host_restore host_list=");
  gd_p(3, host_list);
  gd_s(3, "\n");
  
  gd_s(2,"host_restore starting to read hosts...\n");
  while ((hosts_num < MAX_HOST_RESTORE) && (fgets(buf, sizeof(buf), fp))) {
    if (he) {
      fre_hen(&he, 495);
    }

    he = (host_entry *) ycaloc(sizeof(host_entry), 1,
			       330); /* There are two 330's on purpose! */
    ptr = buf;
    if ((ptr = strchr(buf, ' '))) {
      *(ptr++) = 0;
    } else if ((ptr = strchr(buf, ':'))) {
      *(ptr++) = 0;
    } else {
      printf("In ~/.gnut_hosts, ignoring %s\n",buf);
      continue;
    }

    if (4 != sscanf(buf, "%i.%i.%i.%i", &a[0], &a[1], &a[2], &a[3])) {
      printf("In ~/.gnut_hosts, no ip in %s %s", buf, ptr);
      continue;
    }

    st = atoi(ptr);
    if ((st == 0) && (*ptr != '0')) {
      printf("In ~/.gnut_hosts, no port in %s %s\n",buf,ptr);
      continue;
    }

    he->ip[0]=a[0];
    he->ip[1]=a[1];
    he->ip[2]=a[2];
    he->ip[3]=a[3];
    st=atoi(ptr);
    st=GUINT16_TO_LE(st);
    memcpy(he->port, &st, 2);

    if (host_ok_for_gnet(he->ip)) {
      host_add(he);
      he = 0;
    }
  }
  if (he) {
    fre_hen(&he, 496);
  }

  fclose(fp);

  printf("Got %i addresses from hosts file.\n", hosts_num);

  gd_s(2,"host_restore returning\n");
  return 0;
}

int host_write_line(void *a, void *b)
{
  host_entry *he;
  FILE *fp;
  uint16 st;

  he = a;
  fp = b;

  memcpy(&st, he->port, 2);
  st = GUINT16_FROM_LE(st);
  fprintf(fp, "%i.%i.%i.%i %i\n", he->ip[0], he->ip[1],
	  he->ip[2], he->ip[3], st);
  return 0;
}

int host_save(char *fname)
{
  FILE *fp;
  int i;
  host_entry *he;
  uint16 st;
  
  /* if nothing to save, we want to keep old list */
  if (!hosts_num) {
    return 0;
  }

  unlink(fname);

  fp=fopen(fname, "w");
  if (fp == 0) {
    return -1;
  }

  gd_s(3, "host_save opened file\n");

  for(i=0; i<HLRUSIZE; i++) {
    he = host_lru + i;
    if (he->ip[0]) {
      host_write_line(he, fp);
    }

    if (0) {
      if (he->ip[0]) {
	memcpy(&st, he->port, 2);
	st = GUINT16_FROM_LE(st);
	fprintf(fp, "%i.%i.%i.%i %i\n", he->ip[0], he->ip[1],
		he->ip[2], he->ip[3], st);
      }
    }
  }

  pthread_mutex_lock(&host_mutex);

  gnut_hash_foreach(host_list, host_write_line, fp);
  gd_s(3, "host_save locked mutex\n");

  pthread_mutex_unlock(&host_mutex);

  fclose(fp);
  return 0;
}

int host_remove(uchar ip[4], uchar port[2])
{
  host_entry he;
  host_entry *he2;

  gd_s(1,"host_remove entering ");
  gd_i(1, ip[0]);
  gd_s(1, ".");
  gd_i(1, ip[1]);
  gd_s(1, ".");
  gd_i(1, ip[2]);
  gd_s(1, ".");
  gd_i(1, ip[3]);
  gd_s(1, "\n");

  pthread_mutex_lock(&host_mutex);

  memcpy(he.ip, ip, 4);
  memcpy(he.port, port, 2);

  he2 = gnut_hash_find(host_list, &he);
  if (!he2) {
    pthread_mutex_unlock(&host_mutex);
    /* This happens if the same host was connected twice and both
     * connections close */
    gd_s(1, "host_remove  host not found!\n");
    return -1;
  }

  gnut_hash_remove(host_list, he2);

  fre_hen(&he2, 497);

  hosts_num--;

  pthread_mutex_unlock(&host_mutex);

  return 0;
}

int host_totals(uint *num, uint64 *files, uint64 *size)
{
  gd_s(2, "host_totals num=");
  gd_i(2, hosts_num);
  gd_s(2, " files=");
  gd_li(2, hosts_files);
  gd_s(2, " size=");
  gd_li(2, hosts_size);
  gd_s(2, "\n");
  *num = hosts_num;
  *files = hosts_files;
  *size = hosts_size;
  return 0;
}

int host_find(uchar ip[4], uchar port[2])
{
  host_entry he,*ret;
  gd_s(3,"host_find entering ");
  gd_i(3, ip[0]);
  gd_s(3, ".");
  gd_i(3, ip[1]);
  gd_s(3, ".");
  gd_i(3, ip[2]);
  gd_s(3, ".");
  gd_i(3, ip[3]);
  gd_s(3, "\n");
  
  pthread_mutex_lock(&host_mutex);
  
  memcpy(he.ip, ip, 4);
  memcpy(he.port, port, 2);

  ret = gnut_hash_find(host_list,&he);

  pthread_mutex_unlock(&host_mutex);
  
  if (ret) {
    gd_s(2,"found ip ");
    gd_i(2, ret->ip[0]);
    gd_s(2, ".");
    gd_i(2, ret->ip[1]);
    gd_s(2, ".");
    gd_i(2, ret->ip[2]);
    gd_s(2, ".");
    gd_i(2, ret->ip[3]);
    gd_s(2, "\n");
    return 0;
  }
  return -1;
}

int host_clear()
{
  pthread_mutex_lock(&host_mutex);
  
  gd_s(3, "host_clear entering\n");
  gnut_hash_fre(host_list);
  
  host_list=gnut_hash_new(host_hash, host_compare);

  gd_s(3, "host_clear new host_list=");
  gd_p(3, host_list);
  gd_s(3, "\n");
    
  hosts_num = 0;
  hosts_files = 0;
  hosts_size = 0;
  
  pthread_mutex_unlock(&host_mutex);
  return 0;
}

int host_lock()
{
  pthread_mutex_lock(&host_mutex);
  return 0;
}

int host_unlock()
{
  pthread_mutex_unlock(&host_mutex);
  return 0;
}

Gnut_Hash * host_retrieve()
{
  return host_list;
}
