/* Josh Pieper, (c) 2000
   Robert Munafo, (c) 2001

   This file is distributed under the GPL, see file COPYING for details */

/* Define YTAGS to enable tagging of all memory blocks to trace leaks. This
 * works, but is slow and causes a 4-byte per block memory overhead, so
 * I disable it in release versions */
#define YTAGS 0

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif

#ifdef MCHECK
#include <mcheck.h>
#endif

#include "lib.h"
#include "threads.h"

/* Debug queue, really a circular array. I record numbers into this
 * array for debugging purposes, and dump it when there is a segfault. */
uint16 dq[DQ_SIZE];
uint16 dq_copy[DQ_SIZE];  /* This is for quickly copying the array
                           * to minimize overwrite by other threads
                           * when we are ready to dump */
int32 dqp;
int32 dq_dumped;

uint16 dqt[DQT_CHUNK_SIZE * DQT_CHUNK_NUM];
int32 dqtp[DQT_CHUNK_NUM];
uint16 dqtt;

int32 dq_nth;

/* dq_init is called once when the program starts up */
void dq_init(void)
{
  int32 i;

  for(i=0; i<DQ_SIZE; i++) {
    dq[i] = 0;
  }
  for(i=0; i<DQT_CHUNK_SIZE * DQT_CHUNK_NUM; i++) {
    dqt[i] = 0;
  }
  for(i=0; i<DQT_CHUNK_NUM; i++) {
    dqtp[i] = 0;
  }
  dqp = 0;
  dq_dumped = 0;
  dq_nth = 0;
}

/* dq_start is called by each thread when it starts up */
void dq_start(char *thtype, uint16 dqinum)
{
  uint16 i;

  dq_nth++;
  /* printf("Thread %s %3i %i\n", thtype, DQ_TID, dq_nth); */

  for(i=0; i<DQT_CHUNK_SIZE; i++) {
    dqt[(DQ_TID * DQT_CHUNK_SIZE) + i] = 0;
  }

  dqi(dqinum);
}

/* dq_end is called by a thread when it wants to exit normally */
void dq_end(void * retval)
{
  dq_nth--;
  /* printf(" exit     %3i %i\n", DQ_TID, dq_nth); */

  pthread_exit(retval);
}

#define DQ_DUMP_MAX 4

void dqd_dump1(FILE *dumpfile, uint16 bufsize, uint16 *buf, uint16 startloc)
{
  int32 i, j, k;

  k = 0; j = startloc;
  for(i=0; i<bufsize; i++) {
    fprintf(dumpfile, "%4hx", buf[j]);
    k++;
    if (k >= 16) {
      fprintf(dumpfile, "\n");
      k = 0;
    } else {
      fprintf(dumpfile, " ");
    }

    j++;
    if (j >= bufsize) {
      j = 0;
    }
  }
  if (k > 0) {
    fprintf(dumpfile, "\n");
  }
}

void dq_dump(char *msg)
{
  int32 i, j;
  uint16 *a;
  uint16 *b;
  FILE *dumpfile;

  if (dq_dumped >= DQ_DUMP_MAX) {
    return;
  }

  /* Flash-copy the global debug queue */
  j = dqp;
  a = dq; b = dq_copy;
  for(i=0; i<DQ_SIZE; i++) {
    *b++ = *a++;
  }

  if (dq_dumped == 0) {
    /* First time */
    dumpfile = fopen("gnut-segfault-info", "w");
  } else {
    dumpfile = fopen("gnut-segfault-info", "a");
  }
  /* Dump the global debug queue */
  dqd_dump1(dumpfile, DQ_SIZE, dq_copy, j);
  fprintf(dumpfile, " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");

  /* Dump the thread-specific debug queue */
  dqd_dump1(dumpfile, DQT_CHUNK_SIZE, dqt + (DQ_TID * DQT_CHUNK_SIZE),
	    dqtp[DQ_TID]);
  fprintf(dumpfile, "thread %i %s\n", DQ_TID, msg);
  fprintf(dumpfile, " -----------------------------------------------------------------------------\n");

  fclose(dumpfile);

  dq_dumped++;
}

int _gnut_mprobe(void *d)
{
#ifdef MCHECK
  return mprobe(d);
#else
  return 0;
#endif
}

#define LEAKIDX 700
long leakers[LEAKIDX];
long leak_d[LEAKIDX];
long reset_count;
long reset_thres;

/* Init the tables used to keep track of leaks */
void init_leaks()
{
  int i;

  for(i=0; i<LEAKIDX; i++) {
    leakers[i] = 0;
    leak_d[i] = 0;
  }
  leakers[1] = 1;
  reset_count = 0;
  reset_thres = 100;
}

/* Reset delta counts for leak detection */
void reset_leaks()
{
  int i;

  for(i=0; i<LEAKIDX; i++) {
    leak_d[i] = 0;
  }
}

/* Find and report the memory-block-type that is currently believed to be
   the biggest suspect for leaking */
void tell_leaks()
{
  int i;
  int max, maxi, maxd;

  max = 0; maxi = 0; maxd = 0;
  for(i=0; i<LEAKIDX; i++) {
    if (leak_d[i] > maxd) {
      maxd = leak_d[i];
      max = leakers[i];
      maxi = i;
    }
  }
  printf("ymaloc(..., %d) == %d, delta %d\n", maxi, max, maxd);
}

/* Every time a block is allocated via ymaloc, this routine is called.
   It resets the delta counts periodically but with an exponentially
   lenghthening frequency. The effect of this is that the delta count
   will measure allocations that have taken place during the last 5%
   (approximately) of the history of this invocation of gnut. That's
   the most useful way to catch leaks, because it filters out short-term
   spikes as well at the big initial allocations associated with startup.
 */
void check_reset()
{
  reset_count++;
  if (reset_count > reset_thres) {
    reset_count = 0;
    reset_thres += (reset_thres / 20);
    reset_leaks();
  }
}

/* Leak-detection version of free(3) */
void yfre(void **a, int tracking_number)
{
#ifdef YTAGS
  char *p1;
  int tsize;
#endif
  int *p2;
  int tn;

  gd_s(5, "yfre a=");
  gd_p(5, a);
  gd_s(5, " *a=");
  gd_p(5, *a);
  gd_s(5, "\n");
  if (*a == 0) {
    if (tracking_number) {
      printf("Double-fr""ee error, case %d\n", tracking_number);
      printf("Please email gnut@mrob.com and include 'case %d' in subject\n",
	     tracking_number);
    }
    return;
  }

#ifdef YTAGS
  tsize = sizeof(int);
  p1 = (char *) (*a);
  p1 -= tsize;
  p2 = (int *) p1;
  tn = *p2;
  leakers[tn]--;
  leak_d[tn]--;
#else
  tn = 0;
  p2 = (int *) (*a);
#endif

  /*  printf("yfre %d source %d\n", tracking_number, tn); */

  free((void *) p2);
  *a=0;
}

void fre_v(void **x, int bugnum)
{
  yfre(x, bugnum);
}

void fre_strl(char ***x, int bugnum)
{
  yfre((void **) x, bugnum);
}

void fre_str(char **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

/* Leak-detection version of malloc(3) */
void *ymaloc(int size, int tracking_number)
{
  char *ptr;
#ifdef YTAGS
  int *p2;
#endif
  int tsiz;

#ifdef YTAGS
  tsiz = sizeof(int);
#else
  tsiz = 0;
#endif

  ptr = (char *) malloc(size + tsiz);

  if (!ptr) {
    if (tracking_number) {
      printf("Failed al""loc case %d (%d bytes)\n", tracking_number, size);
      printf("Please email gnut@mrob.com and include 'case %d' in subject\n",
	     tracking_number);
    } else {
      printf("Failed al""loc %d bytes\n", size);
    }
    exit(1);
  }

#ifdef YTAGS
  p2 = (int *) ptr;
  *p2 = tracking_number;
  ptr += tsiz;
#endif

  leakers[tracking_number]++;
  leak_d[tracking_number]++;

  check_reset();

  /*  printf("ymaloc src %d size %d\n", tracking_number, size + tsiz); */

  return ptr;
}

/* Leak-detection version of strdup(3) */
char *ystdup(const char *s, int tracking_number)
{
#ifdef YTAGS
  char * rv;
  int len;

  len = strlen(s) + 1;

  rv = (char *) ymaloc(len, tracking_number);

  memcpy(rv, s, len);

  return rv;
#else
  return(strdup(s));
#endif
}

/* Leak-detection version of calloc(3) */
void * ycaloc(size_t nmemb, size_t size, int tracking_number)
{
#ifdef YTAGS
  size_t ts;
  char * rv;
  int i;

  ts = nmemb * size;
  rv = ymaloc(ts, tracking_number);

  for(i=0; i<ts; i++) {
    rv[i] = 0;
  }

  return rv;
#else
  return calloc(nmemb, size);
#endif
}

#ifdef WIN32
char *path_separator = ";";
char *path_slash = "\\";
#else
char *path_separator = ":";
char *path_slash = "/";
#endif

char *expand_path(char *a)
{
  char *b;
  if (strncmp(a,"~/", 2)==0) {
#ifndef WIN32
    char *c;
    c=getenv("HOME");
#else
    char c[MAX_PATH];
	/*    GetWindowsDirectory(c, MAX_PATH); */
    /* I did this because too many people couldn't find
     * the .gnut* files under win32  -jp */
    c[0]=0;  
#endif
    b=(char *)ymaloc(strlen(a)+strlen(c)+3, 321);
    b[0]=0;
    if (strlen(c)>1) {
      strcpy(b, c);
      strcat(b, path_slash);
    }
    strcat(b, &a[2]);
  } else {
    b=(char *) ymaloc(strlen(a)+3, 322);
    strcpy(b,a);
  }

  if (strlen(b) > 1) {
    if (b[strlen(b)-1] == path_slash[0]) {
      b[strlen(b)-1]=0;
    }
  }

  return b;   
}                                               

int gnut_lib_debug=0;

#ifndef PTHREADS_DRAFT4
pthread_mutex_t _g_debug_mutex=PTHREAD_MUTEX_INITIALIZER;
#else
pthread_mutex_t _g_debug_mutex;
#endif

/* These gd_xxx routines are for debug printf's. There is one for each
   type of thing you might want to print, like strings, hex bytes,
   integers, etc. Originally printf's with multiple arguments were
   used, but I found some bugs in Josh's gnut that resulted from
   failure to pass the right parameters on the stack. I eliminated
   those bugs by removing all use of var_args routines including
   printf */
void gd_s(int level, char * str)
{
  if (gnut_lib_debug >= level) {
    pthread_mutex_lock(&_g_debug_mutex);
#ifndef WIN32
    fflush(stdout);
#endif
    fprintf(stderr, "%s", str);
#ifndef WIN32
    fflush(stderr);
#endif
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}

void gd_02x(int level, uchar c)
{
  if (gnut_lib_debug >= level) {
    pthread_mutex_lock(&_g_debug_mutex);
#ifndef WIN32
    fflush(stdout);
#endif
    fprintf(stderr, "%02x", c);
#ifndef WIN32
    fflush(stderr);
#endif
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}

void gd_i(int level, int32 x)
{
  if (gnut_lib_debug >= level) {
    pthread_mutex_lock(&_g_debug_mutex);
#ifndef WIN32
    fflush(stdout);
#endif
    fprintf(stderr, "%i", x);
#ifndef WIN32
    fflush(stderr);
#endif
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}

void gd_x(int level, int32 x)
{
  if (gnut_lib_debug >= level) {
    pthread_mutex_lock(&_g_debug_mutex);
#ifndef WIN32
    fflush(stdout);
#endif
    fprintf(stderr, "%x", x);
#ifndef WIN32
    fflush(stderr);
#endif
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}

void gd_li(int level, int32 x)
{
  if (gnut_lib_debug >= level) {
    pthread_mutex_lock(&_g_debug_mutex);
#ifndef WIN32
    fflush(stdout);
#endif
    fprintf(stderr, "%li", (long int) x);
#ifndef WIN32
    fflush(stderr);
#endif
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}

void gd_lu(int level, uint32 x)
{
  if (gnut_lib_debug >= level) {
    pthread_mutex_lock(&_g_debug_mutex);
#ifndef WIN32
    fflush(stdout);
#endif
    fprintf(stderr, "%lu", (long unsigned int) x);
#ifndef WIN32
    fflush(stderr);
#endif
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}

void gd_p(int level, void *p)
{
  if (gnut_lib_debug >= level) {
    pthread_mutex_lock(&_g_debug_mutex);
#ifndef WIN32
    fflush(stdout);
#endif
    fprintf(stderr, "%p", p);
#ifndef WIN32
    fflush(stderr);
#endif
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}

char *gnut_strdelimit(char *string, char *delim, char new_delim)
{
  char *c;
  
  for (c=string;*c;c++) {
    if (strchr(delim,*c)) *c=new_delim;
  }
  
  return string;
}

/* remove any trailing whitespace */
void gnut_strstrip(char *string)
{
  int i;
  for (i=(strlen(string)-1);i>0 && isspace(string[i]);i--) string[i]=0;
}

char *munge_time(int secs)
{
  /* these buffers should be 'large enough' */
  char *s;
  char buf[10];
  int days,hours,minutes,seconds;
  int fmt;
  
  s = ymaloc(40*sizeof(char), 323);
  days = hours = minutes = seconds = 0;
  strcpy(s,"");
  
  if (secs==0) {
    strcat(s,"0s");
    return s;
  }
  
  if (secs<0) {
    /* handle negatives correctly */
    strcat(s,"-");
    secs=-secs;
  }
  
  seconds=secs;
  
  if (seconds > 60) {
    /* how many minutes? */
    minutes=seconds / 60;
    seconds=seconds % 60;
  }
  
  if (minutes >= 60) {
    /* hours? */
    hours=minutes / 60;
    minutes = minutes %60;
  }
  
  if (hours >= 24) {
    /* and the days */
    days=hours/24;
    hours=hours % 24;
  }
  
  /* build the string */
  fmt = 0;
  if (days) {
	sprintf(buf, "%dd", days);
	strcat(s, buf);
	fmt = 1;
  }
  if (hours || fmt) {
	sprintf(buf, "%dh", hours);
	strcat(s, buf); 
	fmt = 1;
  }
  if (minutes || fmt) {
	sprintf(buf, "%dm", minutes);
	strcat(s, buf);
	fmt = 1;
  }
  if (seconds || fmt) {
	sprintf(buf, "%ds", seconds);
	strcat(s, buf);
	fmt = 1;
  }
  
  return s;
}

char *munge_bytes(uint64 bytes)
{
  /* these buffers should be 'large enough' */
  char *s;
  
  s = ymaloc(50*sizeof(char), 324);
  
  if (bytes<1024) {
    sprintf(s, "%Li", bytes);
    return s;
  }
  
  /* here are some defines to make the following code cleaner */
  /* note that somewhere along the line, the defn of an MB changed */
  /* these could probably be moved to a .h somehwere */
  
#define MEGABYTE (uint64)(1024 * 1024)
#define GIGABYTE (uint64)(1000 * MEGABYTE)
#define TERABYTE (uint64)(1000 * GIGABYTE)

  if ((bytes >= 1024) && (bytes < MEGABYTE)) {
    /* we only need one digit after the decimal for kilobytes */
    sprintf(s, "%.2fK", ((float)bytes/1024.0));
    return s;
  }
  
  if ((bytes >= MEGABYTE) && (bytes < GIGABYTE)) {
    sprintf(s,"%.3fM", ((float)bytes/MEGABYTE));
    return s;
  }
  
  if ((bytes >= GIGABYTE) && (bytes < TERABYTE)) {
    sprintf(s,"%.3fG", ((float)bytes/GIGABYTE));
    return s;
  }
  
  sprintf(s,"%.3fT",((float)bytes/TERABYTE));
  return s;
  
#undef MEGABYTE
#undef GIGABYTE
#undef TERABYTE
}

int strlen_an(char *s)
{
  int rv;

  for(rv=0; *s; s++) {
    if ((*s >= '0') && (*s <= '9')) {
      rv++;
    } else if ((*s >= 'a') && (*s <= 'z')) {
      rv++;
    } else if ((*s >= 'A') && (*s <= 'Z')) {
      rv++;
    }
  }
  return rv;
}

/* CRC-16

The constant 0x1021 is based on the X25 standard CRC polynomial
x^16 + x^12 + x^5 + 1. The highest bit (x616) doesn't need to be
included; the other three bits (12, 5, 1) form the value 0x1021.

The following illustration is from http://www.io.com/~ritter/ARTS/CRCMYST.HTM

       Polynomial = x^5 + x^4 + x^2 + 1 = 110101

       dn   x^4         x^3     x^2         x^1     x^0
       v   +---+       +---+   +---+       +---+   +---+
   +<-XOR<-|Q D|<-XOR<-|Q D|<--|Q D|<-XOR<-|Q D|<--|Q D|<-+
   |       |   |   ^   |   |   |   |   ^   |   |   |   |  |
   |       +---+   |   +---+   +---+   |   +---+   +---+  |
   |               |                   |                  |
   +---------------+-------------------+------------------+

   FIGURE 2.  CRC Hardware

crc = 0xffff;
for each data bit {
  if (high bit of CRC ^ new data bit == 1) {
    crc = (crc << 1) ^ 0x1021;
  } else {
    crc = crc << 1;
  }
}

for CRC-32 (used in e.g. Ethernet) we use 0x04c11db7L, or in binary:
1 0000 0100 1100 0001 0001 1101 1011 0111

There is a long explanation at http://www.snippets.org/CRC.TXT

See also http://cell-relay.indiana.edu/cell-relay/publications/software/CRC/32bitCRC.c.html

If any of these URLs don't work, contact the author Robert Munafo
(www.mrob.com)

*/

#define CRC32_MASK 0x04C11DB7L;

uint32 crc32_conv_8[256];
int crc32_init = 0;

void crc32_start(uint32 *crc)
{
  int i, j;
  uint32 cctemp;
  
  *crc = 0xffffffffL;
  if (crc32_init == 0) {
	for(i=0; i<256; i++) {
	  cctemp = ((uint32) i) << 24L;
	  for(j=0; j<8; j++) {
		if (cctemp & 0x80000000L) {
		  cctemp = (cctemp << 1L) ^ CRC32_MASK;
		} else {
		  cctemp <<= 1L;
		}
	  }
	  crc32_conv_8[i] = cctemp;
	}
	crc32_init = 1;
  }
}

void crc32_add8(uint32 *crc, uchar data)
{
  uchar index;
  
  index = (*crc >> 24L) ^ data;
  *crc = (*crc << 8L) ^ crc32_conv_8[index];
}

uint32 crc32_string(char *data)
{
  uint32 rv;
  char *s;

  crc32_start(&rv);
  s = data;
  while(*s) {
	crc32_add8(&rv, *s);
	s++;
  }
  return rv;
}

/* CRC an arbitrary block of memory */
uint32 crc32_block(void * data, uint32 len)
{
  uint32 crc;
  uint32 i;
  uchar *k;

  /* Generate 32-bit CRC from key */
  crc32_start(&crc);
  k = (uchar *) data;
  for(i=0; i<len; i++) {
	crc32_add8(&crc, k[i]);
  }

  return crc;
}

void print_ascii(uchar *s)
{
  while(*s) {
	if ((*s < 32) || (*s > 127)) {
	  putchar('#');
	} else {
	  putchar(*s);
	}
	s++;
  }
}

void print_asc_n(uchar *s, int len)
{
  int i;

  for(i=0; i<len; i++) {
	if ((s[i] < 32) || (s[i] > 127)) {
	  putchar('#');
	} else {
	  putchar(s[i]);
	}
  }
}

int writestr(int sock, char *str)
{
  int rv;

  rv = write(sock, str, strlen(str));
  return(rv);
}

int trycmd(char *cmd, char *pat)
{
  FILE *pipe;
  char inbuf[100];
  int rv;

  gd_s(2, "trycmd '");
  gd_s(2, cmd);
  gd_s(2, "'\n");

  pipe = popen(cmd, "r");
  if (pipe == 0) {
    gd_s(2, "trycmd openpipe failed\n");
    return(0);
  }

  /* Read results of command. NOTE: If command not found, error goes to
   * STDERR and doesn't get read. Anything to STDOUT is taken as proof
   * that the command exists. */
  rv = 0;
  while (fgets(inbuf, 100, pipe)) {
    gd_s(3, "trycmd read '");
    gd_s(3, inbuf);
    gd_s(3, "'\n");
    if(strstr(inbuf, pat)) {
      rv = 1;
    }
  }
  pclose(pipe);
  return rv;
}

void strcatlim(char *dest, char *src, int maxlen)
{
  int l1, l2, i;
  char *s;

  l1 = strlen(dest);
  l2 = strlen(src);
  if (l1 >= maxlen) {
    /* Don't catenate anything */
    return;
  } else if (l1 + l2 < maxlen) {
    /* All set */
  } else {
    /* Enforce l1 + l2 < maxlen */
    l2 = maxlen - l1 - 1;
  }

  s = dest + l1;
  for(i=0; i<l2; i++) {
    s[i] = src[i];
  }
  s[i] = 0;
}
