/* prompt.c - all things prompt-related */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <pthread.h>

#ifdef HAVE_REGEX_H
#include <regex.h>
#endif

#ifdef HAVE_TERMIO_H
#include <termio.h>
#endif

#ifdef HAVE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include "lib.h"
#include "threads.h"
#include "list.h"
#include "hash.h"
#include "conf.h"
#include "connection.h"
#include "gnut.h"
#include "host.h"
#include "share.h"
#include "cache.h"
#include "qry.h"
#include "monitor.h"
#include "transfer.h"
#include "net.h"
#include "cli_output.h"
#include "cli.h"
#include "prompt.h"
#if GNUT_HTTP_FILE_LIST || GNUT_HTTP_SEARCH
# include "http.h"
#endif

Variable_Info var_list[] = {
  {"HOSTS",		prompt_host_stats},
  {"FILES",		prompt_host_stats},
  {"SIZE",		prompt_host_stats},
  {"MSGS_R",		prompt_net_stats},
  {"MSGS_S",		prompt_net_stats},
  {"BYTES_R",		prompt_net_stats},
  {"BYTES_S",		prompt_net_stats},
  {"QUERIES",		prompt_query_stats},
  {"QUERY_RESP",		prompt_query_stats},
  {"NUM_SHARED",		prompt_share_stats},
  {"SIZE_SHARED",		prompt_share_stats},
  {"RESPONSES",		prompt_results_stats},
  {NULL,			NULL},
};

char result[BUF_MAX];

/* prompt_str() - returns a prompt string, post-interpolation
 */
char * prompt_str(void)
{
  static char prompt[PROMPT_STR_LEN];
  char *chr, *position;
  
  bzero((void *)prompt, PROMPT_STR_LEN);
  chr = conf_get_str("prompt");
  
  do {
	position = strstr(chr, "${");
	
	if(position) {					/* variable interpolation */
	  strncat(prompt, chr, position - chr);
	  strcat(prompt, interp_var(position, &chr));
	}
	else {						/* no interpolation remaining */
	  strcat(prompt, chr);			/* TODO: make this not a buffer overflow */
	}
  } while(position);
  
  return prompt;
}

/* char *interp_var(char *start, char **finish)
 *
 * start is a pointer to string that goes something like "${foobar}check one two"
 * finish ends up a pointer to string that goes something like "check one two"
 * return value is a pointer to the value of the variable
 */
char * interp_var(char *start, char **finish)
{
  char *value, *end, *var_name;
  
  /* sanity check */
  if( !start ||
	  start[0] != '$' || start[1] != '{' ||
	  !(end = strchr(start, '}'))
	  )
	{	*finish = NULL;
	return "<Error>";
	}
  
  if(end - start == 2) {				/* I like it when ${} expands to "\n"	*/
	*finish = end + 1;
	return "\n";
  }
  
  start += 2; end--;				/* start and finish bound the variable	*/
  var_name = ymaloc(end - start + 2, 303);
  bzero(var_name, end - start + 2);
  strncpy(var_name, start, (end - start) + 1);
  /* fprintf(stderr, "Var is '%s'\n", var_name); */
  if(var_name[0] >= 'a' && var_name[0] <= 'z')	/* lowercase - config variable		*/
	value = conf_get_str(var_name);
  else {						/* uppercase - gnut info stuff		*/
	Variable_Info *a;
	
	for(a = var_list; a && strcmp(a->name, var_name); a++);
	if(a) {					/* variable match */
	  (a->useful_function)(var_name);
	  value = result;			/* TODO: i don't like this global variable crap - jhs */
	}
	else
	  value = NULL;
  }
  *finish = end + 2;			/* just after the ${...} b.s.		*/
  fre_str(&var_name, 374);
  return (value) ? value : "<Error>";
}

/* these functions all look up stuff for prompt variables */

/* prompt_host_stats() - returns info based on var_name
 * HOSTS) total hosts on network
 * FILES) total files on network
 * SIZE)  total size of files on network
 */
void prompt_host_stats(char *var_name)			/* works like info_hosts() */
{
  uint num;
  uint64 size, files;
  
  int munge = conf_get_int("munge");
  
#ifndef FNET
  host_totals(&num, &files, &size);
#else
  num = 0;
  files = 0;
  size = 0;
#endif
  if(strcmp(var_name, "HOSTS") == 0)
	sprintf(result, "%d", num);
  else if(strcmp(var_name, "FILES") == 0)
	if(munge)
	  sprintf(result, "%s", munge_bytes(files));
	else
	  sprintf(result, "%Li", files);
  else if(strcmp(var_name, "SIZE") == 0)
	if(munge)
	  sprintf(result, "%s", munge_bytes(size * (uint64)1024000));
	else
	  sprintf(result, "%Li K", size);
  else
	sprintf(result, "<Error>");
}

/* prompt_net_stats() - returns info based on var_name
 * MSGS_R)  messages received
 * MSGS_S)  messages sent
 * BYTES_R) bytes received
 * BYTES_S) bytes sent
 * BLACKLIST_S) messages blacklisted
 */
void prompt_net_stats(char *var_name)			/* works like info_net() */
{
  int messages, sent;
  uint64 rcv_bytes, sent_bytes;
  uint64 blacklisted;
  
  int munge = conf_get_int("munge");
  
#ifndef FNET
  gnut_connection_net_totals(&messages, &sent, &rcv_bytes, &sent_bytes, &blacklisted);
#else
  messages = 0;
  sent = 0;
  rcv_bytes = 0;
  sent_bytes = 0;
  blacklisted = 0;
#endif
  if(strcmp(var_name, "MSGS_R") == 0)
	if(munge)
	  sprintf(result, "%s", munge_bytes(messages));
	else
	  sprintf(result, "%d", messages);
  else if(strcmp(var_name, "MSGS_S") == 0)
	if(munge)
	  sprintf(result, "%s", munge_bytes(sent));
	else
	  sprintf(result, "%d", sent);
  else if(strcmp(var_name, "BYTES_R") == 0)
	if(munge)
	  sprintf(result, "%s", munge_bytes(rcv_bytes));
	else
	  sprintf(result, "%Li", rcv_bytes);
  else if(strcmp(var_name, "BYTES_S") == 0)
	if(munge)
	  sprintf(result, "%s", munge_bytes(sent_bytes));
	else
	  sprintf(result, "%Li", sent_bytes);
  else if(strcmp(var_name, "BLACKLIST_S") == 0)
	if(munge)
	  sprintf(result, "%s", munge_bytes(blacklisted));
	else
	  sprintf(result, "%qi", blacklisted);
  else
	sprintf(result, "<Error>");
}

/* prompt_query_stats - returns query stats:
 * QUERIES)    messages received
 * QUERY_RESP) bytes sent
 */
void prompt_query_stats(char *var_name)			/* works like info_net() */
{
	int queries, responses;

#ifndef FNET
	gnut_connection_query_totals(&queries, &responses);
#else
	queries = 0;
	responses = 0;
#endif

	if(strcmp(var_name, "QUERIES") == 0)
			sprintf(result, "%d", queries);
	else if(strcmp(var_name, "QUERY_RESP") == 0)
			sprintf(result, "%d", responses);
	else
		sprintf(result, "<Error>");
}

/* prompt_share_stats - returns query stats:
 * NUM_SHARED)  number of files shared
 * SIZE_SHARED) total bytes of files shared
 */
void prompt_share_stats(char *var_name)			/* works like info_shares() */
{
	int num, size;

	int munge = conf_get_int("munge");

#ifndef FNET
	share_totals(&num, &size);
#else
	num = 0;
	size = 0;
#endif
	if(strcmp(var_name, "NUM_SHARED") == 0)
			sprintf(result, "%d", num);
	else if(strcmp(var_name, "SIZE_SHARED") == 0)
		if(munge)
			sprintf(result, "%s", munge_bytes((uint64)size * (uint64)1024));
		else
			sprintf(result, "%d", size);
	else
		sprintf(result, "<Error>");
}

/* prompt_results_stats - returns the number of find responses received
 */
void prompt_results_stats(char *var_name)			/* works like info_shares() */
{
#ifndef FNET
	sprintf(result, "%d", query_num());
#else
	strcpy(result, "0");
#endif
}
