/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Chat.java,v 1.36 2001/06/03 01:19:19 kuldeep Exp $
 */


package net.jxta.instantp2p;

import java.io.*;
import java.util.*;

import net.jxta.exception.*;
import net.jxta.discovery.*;
import net.jxta.endpoint.*;
import net.jxta.pipe.*;
import net.jxta.impl.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.document.*;
import net.jxta.id.*;
import net.jxta.protocol.*;

import jxta.security.exceptions.CryptoException;
import jxta.security.util.Util;
import jxta.security.util.URLBase64;
import jxta.security.util.GetOpt;

import jxta.security.impl.random.JRandom;
import jxta.security.impl.jxtasec.JxtaSec1;

/**
 * Chat utility in InstantP2P: send and receive message from other users.
 *
 */

public class Chat implements PeerListener{

    private Discovery disco = null;

    protected PipeAdvertisement myPipeAdvt = null;

    public  static final String CHATNAMETAG   = "IP2PName";
    private static final String SENDERNAME    = "IP2PSenderName";
    private static final String SENDERMESSAGE = "IP2PSenderMessage";
    private static final String PIPEADVT      = "IP2PPipeAdvt";
    private static final String PUBLICKEY     = "IP2PPublicKey";
    private static final String SESSIONKEY    = "IP2PSessionKey";
    private static final String MESSAGEID     = "IP2PMessageType";
    private static final String CHATID        = "0";
    private static final String PUBLICID      = "1";
    private static final String SESSIONID     = "2";
    private static final String SCHATID       = "4";

    private   Thread            thread = null;
    protected PeerGroup         group;
    protected MessageBoard      messageBoard;   // UI for display
    protected Pipe              pipes;
    protected String            myName;         // name used for chatting 
    private   String            toName = null;  // receiver's name
    protected OutputPipe        outputPipe = null;

    PeerGroupAdvertisement peerAdv;

    public    static final int    HUNKYDORY     = 1;
    public    static final int    CHAT_ERROR    = -1;

    private   static final int    WaitingTime   = 2000;
    protected static final int    PipeTimeout   = 10000;
    protected ChatServer          chatServer = null;   // Chat Server thread

    protected boolean  secured = false;
    private boolean debug = false;
    private JxtaSec1 js1 = null;

    /**
     * Chat Constructor
     *
     * @param PeerGroup group - chat group
     * @param MessageBoard messageBoard - to display messages
     */
    public Chat( PeerGroup group, MessageBoard messageBoard ) {
        this.group = group;
	//rootGroup = group;
        //groupAdvt = group.getAdvertisement ();
        this.messageBoard = messageBoard;
	
	pipes = group.getPipe ();
	disco = group.getDiscovery();
    }

    /**
     * groupChanged - acts on change of peer group
     * @param PeerGroup group - changed group object
     *
     * currently, Chat is provided in NetPeerGroup only,
     * so that everyone can chat with everyone.
     */
    public void groupChanged ( PeerGroup group ) {
    }

    /**
     * peerChanged - acts on change of peer for chat
     * @param PeerGroupAdvertisement pgAdv - changed group Advt object
     *
     */
    public void peerChanged ( PeerGroupAdvertisement pgAdv ) {

      //  if this user is not yet registered for Chat, do not act
        if ( !chatServerRunning()) 
	  return;

	if ( peerAdv == pgAdv )
	    return; // same peer- no need to re create pipes 

        peerAdv = pgAdv;
	toName = pgAdv.getPeerName() ;
	
        if ( outputPipe != null ) {
	    outputPipe.close();          // close the older pipe
            outputPipe = null;
        }
       
	if ( createOutputPipe() == HUNKYDORY ) {
	    if ( secured ) {
		sendPublicKey ();
	    }
	    else 
	      messageBoard.info ( "Now you can start chatting with " + 
				  toName + "\n");
	}
    }
    
    /**
     * makes sure that a user is logged in
     */
    private boolean chatServerRunning( ) {
        return ( chatServer != null );
    }
    

    public int login( String name, boolean secured ) {

        this.secured = secured;	

        // check if the user is already logged in
	if (chatServerRunning ()) {
	    messageBoard.debug_MSG ("chat: user " + name + " is already listening");
	    return CHAT_ERROR; 
	}
	
        // search for peer advts for this user
	myPipeAdvt = registerUser ( name );
	if ( myPipeAdvt == null ) {
	    messageBoard.debug_MSG("chat: user " + name + " is not a registered user");
	    return CHAT_ERROR;
	}
	
        // set the user name and start a server thread for this user name
        myName = name;
	runChatServer ();
	
	return HUNKYDORY; 
    }
    
    /**
     * logout a user
     * 
     * @param String name - name used for logging
     */
    public int logout () {
	
        // check if the server is running for this user
	if (!chatServerRunning ()) {
	    return CHAT_ERROR; 
	}
	stopChatServer ();
	js1 = null;
	return HUNKYDORY; 
    }

    private int sendPublicKey () {
	byte[] advPublicKey = null;
	Message msg = null;
	InputStream ip = null;

	js1 = new JxtaSec1(); // security object
	
	if (debug) js1.setDebug();	// debugging output
	
	// @Peer1 and @Peer2: Both initialize object
	if (!js1.jxtaSecInit()) {
	    StatusPanel.displayStatus("JXTA SEC Init failure");
	    return CHAT_ERROR;
	}

	// All generated byte arrays are URL-Base64 encoded strings
	// @Peer1:
	// Get our public key to place it in our pipe 
	// advertisement:
	try {
	    advPublicKey = js1.getOurRSAPublicKey();
	} catch (CryptoException cex) {
	    StatusPanel.displayStatus("GET Public Key: " + cex.toString());
	    return CHAT_ERROR;
	}

	if ( advPublicKey == null ) {
	    return CHAT_ERROR;
	}
	
	if (debug) {
	    String adv = new String(advPublicKey); // URL-Base64
	    System.out.println("PK Encrypted:\n" + "          " + adv);
	}  
	
	try {
	    // Build a JXTA message
	    msg = pipes.createMessage();
	    
	    // push public key
	    ip = new ByteArrayInputStream ( advPublicKey );
	    msg.push ( PUBLICKEY, ip );
	    
	    // push my PipeAdvt
            StringWriter out = new StringWriter ();
            MimeMediaType displayAs = new MimeMediaType( "text/xml" );
            StructuredTextDocument doc = (StructuredTextDocument) 
	      myPipeAdvt.getDocument( displayAs );
            doc.sendToWriter( out );
            String pipeAdvString = out.toString ();

	    ip = new ByteArrayInputStream ( pipeAdvString.getBytes() );
	    msg.push ( PIPEADVT, ip );
	    
	    // push Message ID/tag
	    ip = new ByteArrayInputStream ( PUBLICID.getBytes() );
	    msg.push ( MESSAGEID, ip );
	    
	    outputPipe.send ( msg );
	    
	} catch ( Exception e ) {
	    //e.printStackTrace();
	    messageBoard.error ( "Failed sending public key to " + toName );
	    return CHAT_ERROR; 
	} 
	return HUNKYDORY;
    }

    private String getName ( PipeAdvertisement pipeAdv ) {
        String name = pipeAdv.getName();
	if ( name != null ) {
	    return name.substring ( name.lastIndexOf(':') + 1 );
	}
	return null;
    }

    private int sendSessionKey ( byte [] pipeAdvString, 
				 byte [] publicKey ) {

        InputStream ip = new ByteArrayInputStream ( pipeAdvString );
        PipeAdvertisement pipeAdv = null;

        if ( ip == null)
	    return CHAT_ERROR;

	// create pipeadvt object from document
        try{
	    pipeAdv = (PipeAdvertisement)
	      AdvertisementFactory.newAdvertisement( new MimeMediaType( "text/xml" ), ip );
        } catch ( Exception e ) {
	    //e.printStackTrace ();
            return CHAT_ERROR;
        }

	// popup dialog box if the user wants to accept the chat
        // proposal
	secured = false;
	boolean accept = messageBoard.displayDialog ( getName ( pipeAdv ),
						      pipeAdv.toString() );
	
	// if the propsal is rejected, do not even bother to
        // reply back.
	if ( accept ) {
	  secured = true;
	  //OutputPipe outputPipe = createOutputPipeByAdv ( pipeAdv );
	  outputPipe = createOutputPipeByAdv ( pipeAdv );
	  toName = getName ( pipeAdv );
	  if ( outputPipe == null )
	    return CHAT_ERROR;
	  
	  if ( js1 == null ){
            js1 = new JxtaSec1(); // security object
	    
	    if (debug) js1.setDebug();	// debugging output
	    
	    // @Peer1 and @Peer2: Both initialize object
	    if (!js1.jxtaSecInit()) {
	      StatusPanel.displayStatus("JXTA SEC Init failure");
	      return CHAT_ERROR;
	    }
	  }
	  
	  if (debug) {
	    String adv = new String(publicKey); // URL-Base64
	    System.out.println("PK Encrypted:\n" + "          " + adv);
	  }  
	  
	  byte[] sessionKey = null;
	  try {
	    sessionKey = js1.getSessionKey( publicKey );
	  } catch (CryptoException cex) {
	    StatusPanel.displayStatus("GET Session Key: " + cex.toString());
	    return CHAT_ERROR;
	  }
	  
	  if (debug) {
	    String skey = new String(sessionKey); // URL-Base64
	    System.out.println("**Encrypted generated session Key: " + skey);
	  }
	  
	  Message msg = null;
	  
	  try {
	    if ( sessionKey == null ) {
	      return CHAT_ERROR;
	    }
	    // Build a JXTA message
	    msg = pipes.createMessage();
	    
	    // push session key
	    ip = new ByteArrayInputStream ( sessionKey );
	    msg.push ( SESSIONKEY, ip );
	    
	    // push pipe advt.
	    ip = new ByteArrayInputStream ( myPipeAdvt.toString().getBytes() );
	    msg.push ( PIPEADVT, ip );
	    
	    // push message id/type
	    ip = new ByteArrayInputStream ( SESSIONID.getBytes() );
	    msg.push ( MESSAGEID, ip );
	    
	    outputPipe.send ( msg );
	    
	  } catch ( Exception e ) {
	    //e.printStackTrace();
	    messageBoard.error ( "Failed sending session key to " + toName );
	    return CHAT_ERROR; 
	  } 
	    messageBoard.info ( "You are connected with " + toName + " on secured channel\n");
	}
	  return HUNKYDORY;
    }


    private int setSessionKey ( byte[] pipeAdvString, 
				byte [] sessionKey ) {

        if ( js1 == null ){
            js1 = new JxtaSec1(); // security object

	    if (debug) js1.setDebug();	// debugging output

	    // @Peer1 and @Peer2: Both initialize object
	    if (!js1.jxtaSecInit()) {
	        StatusPanel.displayStatus ("JXTA SEC Init failure");
	        return CHAT_ERROR;
	    }
	}

	if (debug) {
	    String skey = new String(sessionKey); // URL-Base64
	    System.out.println("**Encrypted generated session Key: " + skey);
	}

	// @Peer1
	// We have received a pipe advertisement from @Peer2. It contains
	// the symmetric session key we will use encrypted in our advertised
	// pipe. 
	try {
 	    js1.setSessionKey(sessionKey);
	} catch (CryptoException cex) {
	    StatusPanel.displayStatus("SET Session Key: " + cex.toString());
	    return CHAT_ERROR;
	}	
	messageBoard.info ( "Now you can securedly chat with " + 
			    toName + "\n");
	return HUNKYDORY;
    }


    private OutputPipe createOutputPipeByAdv ( PipeAdvertisement pipeAdv ) {
      
        OutputPipe outputPipe;

	if ( pipeAdv == null ) {
	    messageBoard.error( "Not a Valid Chat User" );
	    return null; 
	}
	
	try {
	    outputPipe = 
		pipes.createOutputPipe ( pipeAdv, Pipe.NonBlocking, PipeTimeout );
	    if ( outputPipe == null ) {
		messageBoard.error ( toName + " is not chatting at this time. try later");
		return null; 
	    }
	} catch ( Exception e ) {
	    //e.printStackTrace ();
	    messageBoard.error ( toName + " is not listening");
	    return null;
	}
	
        return outputPipe;
    }


    /**
     * sendMessageToPeers - sends message to listening peer(s).
     * @param String message - to be sent
     */
    public int sendMessageToPeers ( String message ) {
        // check if the Chat server is running
        if ( !chatServerRunning()) 
	    return CHAT_ERROR;
	messageBoard.displayMessage ( message + "\n");
	
	if ( outputPipe != null )
	    return sendMessage ( message );
	return CHAT_ERROR;
    }

    private int createOutputPipe () {

        // check if the name is registered
        if ( !chatServerRunning()) {
            messageBoard.error( myName + " is not active" );
            return CHAT_ERROR; 
        }

	PipeAdvertisement toAdvt = discoverPeerAdvt ( toName );
	if ( toAdvt == null ) {
	    messageBoard.error( toName + " is not a registered Chat user" );
	    return CHAT_ERROR; 
	}

	try {
	    outputPipe = 
		pipes.createOutputPipe ( toAdvt, Pipe.NonBlocking, PipeTimeout );
	    if ( outputPipe == null ) {
		messageBoard.error ( toName + " is not chatting at this time. try later");
		return CHAT_ERROR; 
	    }
	} catch ( Exception e ) {
	    //e.printStackTrace ();
	    messageBoard.error ( toName + " is not listening");
	    return CHAT_ERROR;
	}
	
        return HUNKYDORY;
    }
    /**
     * Sends messages to the receiver end
     * 
     * @param String message - message to be dispatched
     */
    public int sendMessage ( String message ) {

	//messageBoard.debug_MSG("You are connected to " + toName);

	Message msg = null;
	InputStream ip = null;

	try {
	    if ( message == null ) {
		return CHAT_ERROR;
	    }
	    // Build a JXTA message
	    msg = pipes.createMessage();


	    // Push message - encrypted or raw
	    if ( secured ) {
	      try {
		byte [] encMessage = js1.encryptData(message.getBytes());
		ip = new ByteArrayInputStream ( encMessage );
		if ( debug )
		  messageBoard.displayMessage ( myName + " ENCRYPTED> " + new String ( encMessage ) + "\n");

	      } catch (CryptoException cex) {
		StatusPanel.displayStatus("Encrypt Data: " + cex.toString());
		return CHAT_ERROR;
	      }      
	    }else 	      
	      ip = new ByteArrayInputStream ( message.getBytes() );
	    msg.push ( SENDERMESSAGE, ip );
	    
	    // push my Name
	    ip = new ByteArrayInputStream ( myName.getBytes() );
	    msg.push ( SENDERNAME, ip );

	    // push message tyep
	    if (secured )
	      ip = new ByteArrayInputStream ( SCHATID.getBytes() );
	    else
	      ip = new ByteArrayInputStream ( CHATID.getBytes() );
	    msg.push ( MESSAGEID, ip );
	    
	    outputPipe.send ( msg );
	    
	} catch ( Exception e ) {
	    //e.printStackTrace();
	    messageBoard.error ( "Failed sending message to " + toName );
	    return CHAT_ERROR; 
	} 	
	return HUNKYDORY; 
    }
    
    
    /**
     * Registers a new user. Publishes its advertisement
     *
     * @param String name - user name
     */
    public PipeAdvertisement registerUser ( String name ) {

	// Check if there is already a registered user of the
	PipeAdvertisement advt = discoverPeerAdvt ( name );
	
	if ( advt != null ) {
	    return advt; 
	}
	
        try {
    	    // Create a pipe advertisement for this pipe.
	    advt = ( PipeAdvertisement )
		AdvertisementFactory.newAdvertisement( 
		    PipeAdvertisement.getAdvertisementType() );
        }
        catch( Exception e ) {
	    //e.printStackTrace();
            messageBoard.error( "Unable to regitser " + name );	
            return null; 
        }
	
	advt.setPipeID( new PipeID( group.getID() ) );
	advt.setName( CHATNAMETAG + ":" + name );

   	try {
	    // Publish the output pipe advertisement
	    disco.publish( advt, Discovery.ADV );
	    disco.remotePublish( advt, Discovery.ADV );
	} catch ( Exception e ) {
	    //e.printStackTrace();
	    messageBoard.error( "Unable to publish channel" );	
            return null;
	}
	
	messageBoard.debug_MSG( name + " is registered now");
	
	return advt; 
    }
    
    
    private boolean checkUserAdvt ( PipeAdvertisement pipeAdv, String name ) {

	if ( pipeAdv == null ) 
            return false;
	
	if ( pipeAdv.getName() == null ) return false;
	if ( pipeAdv.getName().equals( CHATNAMETAG + ":" + name)) return true;
	
	return false;
    }


    private PipeAdvertisement discoverPeerAdvt ( String name ) {
	
	Enumeration enum = null; 

	// Initiate remote discovery first
	/*
	  Current Discovery cannot hangs or takes a long time to
          resolve peerID if the peer is not active. So, its better
          to provide a better user experience than a better feature :(
	if ( peerAdv != null ) {
            StatusPanel.displayStatus ( "Peer Discovering" + ChatNameTag + "*" );
	    disco.getRemoteAdvertisements( peerAdv.getPid(), Discovery.ADV,
					 "name", CHATNAMETAG + "*",
					 4 );
        }
	else 
	*/
	{
            StatusPanel.displayStatus ( "Discovering" + CHATNAMETAG + "*" );
  	    disco.getRemoteAdvertisements( null, Discovery.ADV,
					 "name", CHATNAMETAG + "*",
					 4 );
        }
        StatusPanel.displayStatus ( "Discovering" + CHATNAMETAG + "* DONE" );

	// then look locally
	try {
	    enum = disco.getLocalAdvertisements ( disco.ADV,
						  "name", CHATNAMETAG + ":" + name );
	    if ( (enum != null) && (enum.hasMoreElements()) ) {
		PipeAdvertisement pipeAdv = null;
		
		while (enum.hasMoreElements()) {
		    
		    try {
			pipeAdv = ( PipeAdvertisement ) enum.nextElement();
		    } catch(Exception e) {
			continue;
		    }
		    
		    if ( checkUserAdvt ( pipeAdv, name )) {
			return pipeAdv ;
		    }
		}
		
	    }
	} catch (Exception e) {
	    return null;
	}
	return null;
    }
    
    private void runChatServer( ) {
	chatServer = new ChatServer ( myName );
	chatServer.start();
    }
    

    private void stopChatServer () {
        if ( chatServer != null ) {
            chatServer.interrupt();
	    chatServer.stopServer ();
        }
        chatServer = null;
    }
       
    class ChatServer extends Thread {
	String myName;
	InputPipe inputPipe = null;
        Thread    thisThread;

	public ChatServer ( String myName ) {
	    this.myName = myName;
	    thisThread = Thread.currentThread ();
	}
	
	public void run (){
	    if ( createInputPipe () == HUNKYDORY)
		receiveMessage ();
	}
	
	public synchronized void stopServer () {
	    thisThread = null;
	}
		
	private int createInputPipe () {
	    // First create the input pipe advertisement 
	    
	    messageBoard.debug_MSG("Creating input pipe for " + myName );
	    
	    try {
		inputPipe = pipes.createInputPipe ( myPipeAdvt );
	    } catch ( Exception e ) {
		return CHAT_ERROR;
	    }
	    if ( inputPipe == null ) {
		messageBoard.error ("cannot open Input pipe");
		return CHAT_ERROR;
	    }

	    messageBoard.info( "*** To chat, select a peer from \"Peers\" list ***\7 \n");
	    
	    return HUNKYDORY;
	}
	
	private void receiveMessage () {
	    Message msg = null;
	    
	    while ( thisThread != null ) {
		
		try {
		    msg = inputPipe.waitForMessage();
		    if ( msg == null ) {
			if ( Thread.interrupted ()) {
			    
			    messageBoard.error( "Chatserver going down" );
			    break;
			}
		    }
		} catch ( Exception e ) {
		    messageBoard.info( "Good Bye ;-)\n" );
		    break;
		}
		
		// Get the first tag
		InputStream ip = null;
		Enumeration enum = msg.getNames();
		
		if (( enum == null ) || ( !enum.hasMoreElements ())) {
		    // Empty message
		    continue;
		}
              
		int res = 0;

		String messageID = "";
		byte[] buffer = null;
		
		// Get sender information
		try {
		    ip = msg.pop ( MESSAGEID );
		    if ( ip != null ) {
			buffer = new byte [ ip.available() ];
			res = ip.read ( buffer );
			messageID = new String (buffer);
		    }
		} catch (Exception e) {
		    // No user information
		    messageBoard.error ( e.toString());
		}

		if ( messageID.equals ( PUBLICID )) {

		    byte [] pipeAdvString = null;
		    
		     // Get sender information
		     try {
		       ip = msg.pop ( PIPEADVT );
		       if ( ip != null ) {
			 buffer = new byte [ ip.available() ];
			 res = ip.read ( buffer );
			 pipeAdvString = buffer ;
		       }
		     } catch (Exception e) {
		       // No user information
		       messageBoard.error ( e.toString());
		     }
		     
		     byte[] publicKey = null;
		     
		     // Get sender information
		     try {
		       ip = msg.pop ( PUBLICKEY );
		       if ( ip != null ) {
			 buffer = new byte [ ip.available() ];
			 res = ip.read ( buffer );
			 publicKey =  buffer ;
		       }
		     } catch (Exception e) {
		       // No user information
		       messageBoard.error ( e.toString());
		     }
		     sendSessionKey ( pipeAdvString, publicKey );
		}
		else if ( messageID.equals ( SESSIONID )) {
		      
		    byte[] pipeAdvString = null;
		     
		    // Get sender information
		    try {
		        ip = msg.pop ( PIPEADVT );
			if ( ip != null ) {
			    buffer = new byte [ ip.available() ];
			    res = ip.read ( buffer );
			    pipeAdvString = buffer;
			}
		    } catch (Exception e) {
		        // No user information
		        messageBoard.error ( e.toString());
		    }
			
		    byte[] sessionKey = null;
			
		    // Get sender information
		    try {
		        ip = msg.pop ( SESSIONKEY );
			if ( ip != null ) {
			    buffer = new byte [ ip.available() ];
			    res = ip.read ( buffer );
			    sessionKey =  buffer ;
			}
		    } catch (Exception e) {
		        // No user information
		      messageBoard.error ( e.toString());
		    }
		    setSessionKey ( pipeAdvString, sessionKey );
		}
		else {
		  boolean encryptedMessage = false;
		  if ( messageID.equals ( SCHATID )) 
                      encryptedMessage = true;

		   String sender = "Anonymous";
		     
		     // Get sender information
		     try {
		       ip = msg.pop ( SENDERNAME );
		       if ( ip != null ) {
			 buffer = new byte [ ip.available() ];
			 res = ip.read ( buffer );
			 sender = new String ( buffer );
		       }
		     } catch (Exception e) {
		       // No user information
		       messageBoard.error ( e.toString());
		     }
		     
		     String sentMessage = null;
		     // Get message
		     try {
		       ip = msg.pop ( SENDERMESSAGE );
		       if ( ip != null ) {
			 buffer = new byte [ ip.available() ];
			 res = ip.read ( buffer );
			 sentMessage = new String ( buffer );
		       } else {
			 messageBoard.displayMessage( sender + "> " + "empty message" + "\n" );
			 continue;
		       }
		     } catch (Exception e) {
		       messageBoard.displayMessage( sender + "> " + e.toString () + "\n" );
		     }

		     if ( encryptedMessage ) {
		       if ( debug )
			 messageBoard.displayMessage( sender + " ENCRYPTED> " + sentMessage + "\n" );
		       try {
			 String recvStr = new String (js1.decryptData( sentMessage.getBytes() ));
			 messageBoard.displayMessage( sender + "> " +  recvStr + "\n" );
		       } catch (CryptoException cex) {
			 //System.out.println("Decrypt Data: " + cex.toString());
			 continue;
		       }
		     }
		     else {
		       messageBoard.displayMessage( sender + "> " + sentMessage + "\n" );
		     }
		}
	    }
	    inputPipe.close ();
	    messageBoard.info( "*** " + myName  + " has left ***\n" );
	}
    }
}


