/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: GroupChat.java,v 1.9 2001/06/02 17:07:16 kuldeep Exp $
 */


package net.jxta.instantp2p;

import java.awt.*;
import java.io.*;
import java.util.*;

import net.jxta.endpoint.*;
import net.jxta.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.document.*;
import net.jxta.id.*;
import net.jxta.protocol.*;
import net.jxta.discovery.*;
import net.jxta.exception.*;


/**
 * chat is a P2P bases Instant Messaging
 *
 */

public class GroupChat extends Chat implements PeerListener {

    public  static final String ChatNameTag   = "IP2PUserName";
    public  static final String ChatUUIDTag   = "IP2PPipeUUID";
    private static final String SenderName    = "IP2PSenderName";
    private static final String SenderMessage = "IP2PSenderMessage";


    private ChatServer        chatServer = null;   // Chat Server thread
    private DiscoverPeers     discovery = null;   // discovery thread
    private PeerGroup         rootGroup;
    private PeerGroupAdvertisement groupAdvt = null;
    private Discover          discover;
    private Vector            peers;          // a vector of pipe advt of peers

    /**
     * Chat Constructor
     *
     * @param PeerGroup group - chat group
     * @param MessageBoard messageBoard - to display messages
     */
    public GroupChat( PeerGroup group, MessageBoard messageBoard ) {
	super ( group, messageBoard );
	rootGroup = group;
        groupAdvt = group.getAdvertisement ();
	peers = new Vector ();
    }

    public void groupChanged ( PeerGroup peerGroup ) {
        if ( peerGroup == null || peerGroup.equals ( group )) {
	    return;
	}
	group = peerGroup;
	groupAdvt = group.getAdvertisement();
	StatusPanel.displayStatus ( "Changing group to " + groupAdvt.getName());

        if ( chatServerRunning ()) {
            logout ();
	    login ( myName, false );
        }
    }

    public void peerChanged ( PeerGroupAdvertisement pgAdv ) {
    }
    /**
     * Logs ins the user
     *
     */
    public int login ( String myName, boolean secured ) {

        discover = new Discover ( group );
	discover.getRemoteAdvertisements( null, "name", ChatNameTag + "*");

	this.myName = myName;

	// If I can successfully log in to the group, let me start my own
	// chat daemon to listen to others
	if (( myPipeAdvt = registerUser ( myName )) != null ) 
	    runChatServer ();
	
	// ummm.... its time to find peers in the group
	discoverPeerAdvt ();
	messageBoard.info ( "Peers: " );
	
	// how about displaying the list of peers to me...
	for ( int i = 0; i < peers.size(); i++) {
	    PipeAdvertisement pipeAdv = ( PipeAdvertisement ) peers.elementAt(i);
	    String name = pipeAdv.getName().substring (
				 pipeAdv.getName().lastIndexOf(':') + 1);
		
	    messageBoard.info ( name + " ");
	}
	messageBoard.info ( "\n" );
	// let me start my discovery thread in the background. This would help
	// me know about others who join the group
	startLocalDiscovery ();	
	
	return 0;
    }

    /**
     * logout a user
     * 
     * @param String name - name used for logging
     */
    public int logout () {
        peers.removeAllElements();
	
	if ( chatServer == null || chatServer.isInterrupted())
	    return CHAT_ERROR;
	
	stopChatServer ();
	chatServer = null;
	stopLocalDiscovery();
	return HUNKYDORY; 
    }
    
   /**
     * makes sure chat server is running
     */
    private boolean chatServerRunning( ) {
        return ( chatServer != null );
    }

    public int sendMessageToPeers ( String message ) {
        // check if the Chat server is running
	if ( chatServer == null || chatServer.isInterrupted())
	    return CHAT_ERROR;
	
	// if so, let me create my chat channels (pipes) to them
	for ( int i = 0 ; i < peers.size(); i++ ) {
	    OutputPipe pipe = 
		createOutputPipe ( (PipeAdvertisement) peers.elementAt(i) );
	    
	    if ( pipe != null) {
		// wow!! now I can send messages...
		sendMessage ( pipe, message );
		// let me cleanup the dirt before, I talk to others.
		// else, I might run out of resources...
		pipe.close();
	    }
	    else {
		peers.removeElementAt ( i );
	    }
	}
	return HUNKYDORY;
    }
    
    private OutputPipe createOutputPipe ( PipeAdvertisement toAdvt ) {
	
	OutputPipe outputPipe = null;
	
	// I should not be talking to guys I do not know or the ones that
	// don't exist...
	if ( toAdvt == null ) {
	    messageBoard.error( toAdvt + " bad user" );
	    return null; 
	}
	
	try {
	    outputPipe = pipes.createOutputPipe ( toAdvt, Pipe.NonBlocking, PipeTimeout );
	    
	    if ( outputPipe == null ) {
		messageBoard.error ( toAdvt.getName() + " is not listening at this time. try later");
		return null; 
	    }
	} catch ( Exception e ) {
	    //e.printStackTrace ();
	    messageBoard.error ( toAdvt.getName() + " is not listening");
	    return null;
	}
	
        return outputPipe;
    }
    
    /**
     * Sends messages to the receiver end
     * 
     * @param String message - message to be dispatched
     */
    public int sendMessage ( OutputPipe outputPipe, String message ) {
	
	Message msg = null;
	InputStream ip = null;
	
	try {
	    if ( message == null ) {
		return CHAT_ERROR;
	    }
	    // Build a JXTA message
	    msg = pipes.createMessage();
	    
	    ip = new ByteArrayInputStream ( message.getBytes() );
	    msg.push ( SenderMessage, ip );
	    
	    if ( myName != null ) {
		ip = new ByteArrayInputStream ( myName.getBytes() );
		msg.push ( SenderName, ip );
	    }
	    
	    outputPipe.send ( msg );
	    
	} catch ( Exception e ) {
	    //e.printStackTrace();
	    messageBoard.error ( "Failed sending message to " );
	    return CHAT_ERROR; 
	} 
	
	return HUNKYDORY; 
    }
    
    
    /**
     * Registers a new user. Publishes its advertisement
     *
     * @param String name - user name
     *
     * @return PipeAdvertisement - for the user
     */
    public PipeAdvertisement registerUser ( String name ) {
	
	// Check if there is already a registered user of the
	PipeAdvertisement advt = discoverPeerByName ( ChatNameTag + ":" + name );

	if ( advt != null ) {
	    messageBoard.info ( "*** " + name + " has joined group " + groupAdvt.getName() + " ***\n");
	    return advt; 
	}
	
        try {
    	    // Create a pipe advertisement for this pipe.
	    advt = ( PipeAdvertisement )
		AdvertisementFactory.newAdvertisement( PipeAdvertisement.getAdvertisementType() ); 
        }
        catch( Exception e ) {
	    //e.printStackTrace();
            messageBoard.error( "Unable to regitser " + name );	
	    return null; 
        }
	
	advt.setPipeID( new PipeID( group.getID() ));
	advt.setName( ChatNameTag + ":" + name );
	
   	try {
	    Discovery disco = group.getDiscovery();
	    
	    // Publish the output pipe advertisement
	    disco.publish( advt, Discovery.ADV );
	    disco.remotePublish( advt, Discovery.ADV );
	} catch ( Exception e ) {
	    //e.printStackTrace();
	    messageBoard.error( "Unable to publish advertisement" );	
            return null;
	}
	
	messageBoard.debug_MSG( name + " is registered now");	
	
	return advt; 
    }
    
    
    private PipeAdvertisement discoverPeerByName ( String name ) {
	
	Enumeration enum; 
	
	enum = discover.getAdvertisements ( null, "name", name );
	
	try {
	    if ( (enum != null) && (enum.hasMoreElements ())) {
		PipeAdvertisement pipeAdv = null;
		
		while (enum.hasMoreElements()) {
		    
		    try {
			pipeAdv = ( PipeAdvertisement ) enum.nextElement();
		    } catch(Exception e) {
			continue;
		    }
		    if ( pipeAdv.getName().equals ( name ))
		        return pipeAdv;
		}
	    }
	} catch (Exception e) {
	    //e.printStackTrace ();
	    return null;
	}
	
	return null;
    }
    
    
    private void discoverPeerAdvt () {
	
	try {
	    
	    Enumeration enum = discover.getAdvertisements( null, "name", ChatNameTag + "*");
	    if ( (enum != null) && (enum.hasMoreElements ())) {
		PipeAdvertisement pipeAdv = null;
		
		while (enum.hasMoreElements()) {
		    
		    try {
			pipeAdv = ( PipeAdvertisement ) enum.nextElement();
			//System.out.println ("pipeAdv: " + pipeAdv.getName());
		    } catch(Exception e) {
			//e.printStackTrace ();
			continue;
		    }
		    
		    int i;
		    		    
		    for ( i = 0; i < peers.size(); i++ ) {
			if ( pipeAdv.getPipeID().equals ( ((PipeAdvertisement) peers.elementAt(i)).getPipeID()))
			    break;
		    }
		    if (peers.size() == 0 || i == peers.size()) {
			peers.addElement (pipeAdv);
			String name = pipeAdv.getName().substring (
								   pipeAdv.getName().lastIndexOf(':') + 1);
			messageBoard.info ( "*** " + name + " has joined " + groupAdvt.getName() + " ***\n");
		    }
		}
	    }
	} catch (Exception e) {
	    // e.printStackTrace ();
	    return;
	}
	
	return;
    }
    
    private void runChatServer( ) {
	chatServer = new ChatServer ( myName );
	if ( chatServer != null )
	    chatServer.start();
    }
    
    private void stopChatServer () {
	if ( chatServer != null ) {
            chatServer.interrupt();
	    chatServer.stopServer ();
        }
        chatServer = null;
    }
    
    
    class ChatServer extends Thread {
	String myName;
	InputPipe inputPipe = null;
	Thread    thisThread;
	
	public ChatServer ( String myName ) {
	    this.myName = myName;
	    thisThread = Thread.currentThread ();
	}
	
	public void run (){
	    if ( createInputPipe () == HUNKYDORY)
		receiveMessage ();
	}
	
	public synchronized void stopServer () {
	    thisThread = null;
	}
	
	
	private int createInputPipe () {
	    // First create the input pipe advertisement 
	    
	    //System.out.println("Creating input pipe for " + myName );
	    
	    try {
		inputPipe = pipes.createInputPipe ( myPipeAdvt );
	    } catch ( Exception e ) {
		return CHAT_ERROR;
	    }
	    if ( inputPipe == null ) {
		messageBoard.error ("cannot open InputPipe channel");
		return CHAT_ERROR;
	    }
	    
	    return HUNKYDORY;
	}
	
	private void receiveMessage () {
	    Message msg = null;
	    
	    while ( thisThread != null ) {
		
		try {
		    msg = inputPipe.waitForMessage();
		    if ( msg == null ) {
			continue;
		    }
		    if ( Thread.interrupted ()) {
			
			messageBoard.error( "Chatserver going down" );
			break;
		    }
		} catch ( Exception e ) {
		    //e.printStackTrace();
		    messageBoard.error( "Chatserver going down on exception" );
		    break;
		}
		
		// Get the first tag
		InputStream ip = null;
		Enumeration enum = msg.getNames();
		
		if (( enum == null ) || ( !enum.hasMoreElements ())) {
		    // Empty message
		    continue;
		}
		
		int res = 0;
		String sender = "Anonymous";
		byte[] buffer = null;
		
		// Get sender information
		try {
		    ip = msg.pop ( SenderName );
		    if ( ip != null ) {
			buffer = new byte [ ip.available() ];
			res = ip.read ( buffer );
			sender = new String ( buffer );
		    }
		} catch (Exception e) {
		    // No user information
		    messageBoard.error ( e.toString());
		}
		
		String sentMessage = null;
		// Get message
		try {
		    ip = msg.pop ( SenderMessage );
		    if ( ip != null ) {
			buffer = new byte [ ip.available() ];
			res = ip.read ( buffer );
			sentMessage = new String ( buffer );
		    } else {
			messageBoard.displayMessage( sender + "> " + "empty message" + "\n" );
			continue;
		    }
		} catch (Exception e) {
		    messageBoard.displayMessage( sender + "> " + e.toString () + "\n" );
		}
		messageBoard.displayMessage( sender + "> " + sentMessage + "\n" );
		//System.out.println ("Message received");
	    }
	    inputPipe.close ();
	    messageBoard.info( "*** " + myName  + " has left ***\n" );
	    
	}
    }
    
    private void startLocalDiscovery( ) {
	
	discovery = new DiscoverPeers ();
	discovery.start();    
    }
    
    private void stopLocalDiscovery () {
	if ( discovery != null )
	    discovery.stopDiscovery ();
        discovery = null;
    }
    
    
    // This is a temporary class and should go away once
    // Discovery provides call backs
    class DiscoverPeers extends Thread {
      int waitingTime = 10000; // 10 secs.
      int MAXWAITIME  = 60 * waitingTime;
	Thread thisThread;
	
	public DiscoverPeers () {
	    thisThread = Thread.currentThread ();
	}
	
	public synchronized void stopDiscovery () {
	    thisThread = null;
	}
	
	public void run (){
            while ( thisThread != null ) {
	        try {
	            Thread.sleep ( waitingTime );
	        } catch ( Exception e ) {
	        }
                discoverPeerAdvt();
		if ( waitingTime < MAXWAITIME )
		    waitingTime *= 2;
                //System.out.println ("Discovery...");
            } 
	    //System.out.println ( "discovery stopped");
        }
    }
}


