/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerGroupManager.java,v 1.25 2001/06/02 05:38:11 kuldeep Exp $
 *
 */

package net.jxta.instantp2p;

import java.lang.reflect.InvocationTargetException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;

import net.jxta.discovery.Discovery;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.Document;
import net.jxta.document.MimeMediaType;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.id.*;
import net.jxta.pipe.PipeID;
import net.jxta.peergroup.PeerGroup;
import net.jxta.peergroup.PeerGroupID;
import net.jxta.peergroup.PeerGroupFactory;
import net.jxta.pipe.Pipe;
import net.jxta.protocol.PeerGroupAdvertisement;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.protocol.ServiceAdvertisement;

public class PeerGroupManager {

    private PeerGroup group = null;
    private PeerGroup rootPeerGroup;
    private String groupName = null;
    private File groupDir = null;
    private Discovery myDiscoverService = null;
    private Pipe      myPipeService = null;

    private boolean pollDiscoveryFlag = false;
    private boolean runRemoteDiscoveryFlag = false;

    private Discover discover = null;

    public PeerGroupManager ( PeerGroup group ) throws PeerGroupException {

        // save PeerGroup object
        this.group = group;
	rootPeerGroup = group;
        groupName = group.getAdvertisement().getName();

	discover = new Discover ( group );
	myDiscoverService = discover.getDiscovery();
        if ( myDiscoverService == null ) {
            // fatal error
            throw new PeerGroupException();
        }

        // find the PipeService
	myPipeService = group.getPipe ();
        if (myPipeService == null) {
            // fatal error
            throw new PeerGroupException();
        }

        // Start the AppPolicies
	//        startService();
    }

    /**
     *  PeerGroupApp Interface Methods
     */

    /**
     *  Returns the PeerGroup for this PeerGroupApp
     *
     * @return the PeerGroup
     * @since   JXTA 1.0
     */
    public PeerGroup getPeerGroup() {
        return group;
    }

    public PeerGroup getRootPeerGroup() {
        return rootPeerGroup;
    }

    /**
     *  Returns the PeerGroupAdvertisement for this PeerGroupApp
     *
     * @return the PeerGroupAdvertisement
     * @since   JXTA 1.0
     */
    public PeerGroupAdvertisement getAdvertisement() {
        return getPeerGroup().getAdvertisement();
    }

    /**
     *  Returns the Group name for this PeerGroupApp
     *
     * @return the name for this group
     * @since   JXTA 1.0
     */
    public String getGroupName()   {
        return groupName;
    }

    /**
     *  Joins the PeerGroup with the given gid
     *  The group is only joined if the group is in the list and 
     *  has not already been joined
     *
     * @param gid Gid of the requested Group to join
     * @return the JxtaApp for the joined group
     * @since   JXTA 1.0
     */
    public PeerGroup joinGroup( PeerGroupAdvertisement pgAdv ) {
	
	PeerGroup group;

        if ( pgAdv == null ) 
	    return null;

        // create the peergroup
        group = PeerGroupFactory.newPeerGroup();
	
        if ( group != null ) {

	    try {
		// initialize the peergroup
		group.init( rootPeerGroup, pgAdv );
	    } catch ( PeerGroupException e ) {
		//e.printStackTrace();
		return null;
	    }
	    //group.startApp( null );
	}
	
        return group;
    }

    /**
     *  Creates a new PeerGroup with the given name and Joins the group
     *
     * @param gid Gid of the requested Group to join
     * @return the JxtaApp for the new group
     * @since   JXTA 1.0
     */
    public PeerGroupAdvertisement createGroup( String name )
	throws PeerGroupException {
	
        PeerGroupAdvertisement pgAdv = createGroupAdvertisement( name );
			
        return pgAdv;
    }


    /**
     *  check results of discovery request
     *
     * @since   JXTA 1.0
     */
    public void checkDiscoverResults( PeerGroup group ) {
	if ( group == null )
	    return;
	discover.setPeerGroup ( group );
	StatusPanel.displayStatus ( "Discovering peers.... ");
        getPeers ();
	StatusPanel.displayStatus ( "Discovering peers Done ");
    }

    public Enumeration getPeers() {

	return ( discover.getPeers ()) ;
    }
    
    public Enumeration getGroups() {

	return ( discover.getGroups ()) ;
    }


    private void discoverRemote() {
        // System.out.println("PeerGroupApp.discoveryRemote()");

        // start Remote Discovery of Peer Advertisements
        discover.getRemotePeers( 10 );

        // start Remote Discovery of Group Advertisements
        discover.getRemoteGroups( 10 );
    }

    private void discoverFlush() {
        // System.out.println("PeerGroupApp.discoverFlush()");

        discover.flush ();
    }


  // This is not the correct place for this method. Need to separate
  // out UI from guts.
    public synchronized void selectPeer(PeerGroupAdvertisement pgAdv) {
      //        remoteContentTab.selectPeer(pgAdv);
    }

    
    private PeerGroupAdvertisement createGroupAdvertisement ( String name ) 
	throws PeerGroupException {
	
        PeerGroupAdvertisement pgAdv = null;

        try {
            pgAdv = (PeerGroupAdvertisement) 
                AdvertisementFactory.newAdvertisement(PeerGroupAdvertisement.getAdvertisementType());
        } catch( InvocationTargetException e ) {
            throw new PeerGroupException();
        }
	
        String gid = new PeerGroupID().toString(); // create new GroupId
        Hashtable services = (Hashtable)getAdvertisement().getServiceAdvertisements().clone();

        ServiceAdvertisement app = createShareAppAdvertisement();
        if (app == null) {
            throw new PeerGroupException();
        }

        // pgAdv.setPid (group.getPeerID().toString());   // use parent defined PID.
        pgAdv.setPid(null);                                 // clear the Pid
        pgAdv.setGid(gid);                                  // use the new Gid
        pgAdv.setName (name);
        pgAdv.setKeywords("");                              // Keyword
        pgAdv.setServiceAdvertisements(services);            // same ones
        pgAdv.setEndpointAdvertisements(new Hashtable());   // no private endpoints
        pgAdv.setApp(app);
        pgAdv.setGroupImpl(getAdvertisement().getGroupImpl());

        return pgAdv;
    }

  // This is not the correct place for this method. Need to separate
  // out UI from guts.
    private ServiceAdvertisement createShareAppAdvertisement() {
        // create a new Service for the initial app
        ServiceAdvertisement appService = null;
        try {
            appService = (ServiceAdvertisement)
                AdvertisementFactory.newAdvertisement(ServiceAdvertisement.getAdvertisementType());
        } catch (InvocationTargetException e) {
            return null;
        }

        appService.setName("jxta:/service/GroupApp");
        appService.setVersion("JXTA1.0");
        appService.setUri("http:/www.jxta.org/download/instantp2p.jar");
        appService.setProvider("sun.com");
        appService.setSecurity("TBD");
        appService.setCode("net.jxta.instantp2p.InstantP2P");

        return appService;
    }
    
    public Enumeration getGroupByName ( String groupName ) {
	return ( discover.getGroups ( null, "Name", groupName ));
    }

    public void publishGroup ( PeerGroupAdvertisement pgAdv ) {
	// publish the new group to the PeerGroup
        try {
            myDiscoverService.publish( pgAdv, Discovery.GROUP );
            myDiscoverService.remotePublish( pgAdv, Discovery.GROUP );
        } catch ( Exception e ) {
            StatusPanel.displayStatus("Could not publish PeerAdvertisement");
            //e.printStackTrace();
        }
    }
}
