/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Search.java,v 1.8 2001/06/02 22:25:33 kuldeep Exp $
 *
 */

package net.jxta.instantp2p;

import java.lang.reflect.InvocationTargetException;
import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import net.jxta.document.Advertisement;
import net.jxta.discovery.Discovery;
import net.jxta.exception.PeerGroupException;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.PipeID;
import net.jxta.protocol.PeerGroupAdvertisement;
import net.jxta.protocol.PipeAdvertisement;

import net.jxta.share.*;

public class Search implements Runnable {

    public static final String DEFAULT_DIR = "search";

    private PeerGroup group = null;
    private File dir = null;
    private Hashtable pipes = new Hashtable();
    private Vector results = new Vector();
    private boolean doQuit = false;

    private ListRequestor [] request = null;
    private SearchListener listener = null;

    public Search(PeerGroup inGroup, File inDir) {
        group = inGroup;
        dir = inDir;

        Thread thread = new Thread(this);
        thread.start();
    }

    public void search(String subString) {
        // stop the old search
        cancel();

        request = new ListRequestor[pipes.size()];

        Enumeration enum = pipes.elements();
        PipeAdvertisement pipeAdv = null;
        for (int i=0; i<request.length; i++) {
            pipeAdv = (PipeAdvertisement)enum.nextElement();
            if (null != pipeAdv) {
                try {
                    request[i] = new ListRequestor(group, pipeAdv, subString);
                } catch (InvocationTargetException e) {
		  //e.printStackTrace();
                }
            }
        }
    }

    public void cancel() {
        if (null != request) {
            for (int i=0; i<request.length; i++) {
                if (null != request[i]) {
                    request[i].cancel();
                }
            }
            request = null;
            results.removeAllElements();
        }

    }

    public void addListener(SearchListener l) {
        listener = l;
    }

    public void removeListener(SearchListener l) {
        listener = null;
    }

    public SearchResult[] getResults() {
        SearchResult[] res = new SearchResult[results.size()];
        results.copyInto(res);

        return res;
    }

    public void run() {
        int waitingTime = 10000; // 10 secs.
        int MAXWAITIME  = 2 * 60 * waitingTime;

        // get the Discovery Service
        Discovery discover = (Discovery) group.getDiscovery();

        // start Remote Discovery of Pipe Advertisements
        discover.getRemoteAdvertisements(null, Discovery.ADV, "name", CMS.PIPE_ADV_NAME, 10);

        try {
            Thread.sleep(2*1000);
        } catch (InterruptedException e) {
            //e.printStackTrace();
        }

        int discoverCount = 0;
        PipeAdvertisement pipeAdv = null;
        while (!doQuit) {
            Enumeration advs = null;

            //System.err.println("Getting local advertisements");
            try {
                advs = discover.getLocalAdvertisements(Discovery.ADV, "name", CMS.PIPE_ADV_NAME);
            } catch (IOException e) {
                //System.err.println("Couldn't get local Advertisements");
	      //e.printStackTrace();
            }

            //System.err.println("Got local advertisements");
            if (advs != null) {
                while (advs.hasMoreElements()) {
                    Advertisement adv = (Advertisement)advs.nextElement();
                    if (adv instanceof PipeAdvertisement && 
                        CMS.PIPE_ADV_NAME.equals(((PipeAdvertisement)adv).getName())) {
                        // try adding the pipe to the list
                        pipeAdv = (PipeAdvertisement)adv;
                        String key = pipeAdv.getPipeID().toString();
                        if (null == pipes.get(key)) {
                            pipes.put(key, pipeAdv);
                        }
                    }
                }
            }

            discoverCount++;
            if (discoverCount >= 6) {
                discoverCount = 0;

                //System.out.println("discover.getRemoteAdvertisements()");

                // start Remote Discovery of Peer Advertisements
                discover.getRemoteAdvertisements(null, Discovery.ADV, "name", CMS.PIPE_ADV_NAME, 10);
		if ( waitingTime < MAXWAITIME )
		  waitingTime *= 2;
            }

            try {
                Thread.sleep( waitingTime );
            } catch (InterruptedException e) {
	      //e.printStackTrace();
            }
        }
    }

    class ListRequestor extends ListContentRequest {
        public ListRequestor(PeerGroup group, PipeAdvertisement inRequest, 
                String inSubStr) throws InvocationTargetException {
            super(group, inRequest, inSubStr);
        }

        public void notifyDone() {
	  //System.out.println("Requestor done");
            ContentAdvertisement[] result = this.getResults();
            
            for (int i=0; i<result.length; i++) {
	        SearchResult searchResult = new SearchResult(getRequestPipe(), result[i]);

		if ( searchResult != null )
                results.addElement( searchResult );
            }

            if (null != listener) {
                listener.searchUpdate();
            }
        }

        public void notifyFailure() {
	    //System.out.println("Requestor failure");
	    // XXX: Kuldip believes that there is no need to call searchupdate
	    //      need to verify with Carl.
	    /* 
            if (null != listener) {
                listener.searchUpdate();
            }
	   */ 
        }

    }
}
