/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TabButton.java,v 1.5 2001/05/25 01:39:11 kuldeep Exp $
 */

package net.jxta.widget;

import java.awt.*;
import java.awt.event.*;

/**
 * TabButton - a class that produces a lightweight button as tabs.
 *
 */
public class TabButton extends Component {

  ActionListener actionListener;     // Post action events to listeners
  String label;                      // The Button's text
  protected boolean pressed = false; // true if the button is detented.
  private static final int PADX = 5; // padding for x direction
  private static final int PADY = 5; // padding for y direction
  private Font   font = new Font ("Dialog", Font.BOLD, 12);
  private Color  bgColor = Color.lightGray;
  private Color  fgColor = Color.black;

  /**
   * Constructs a TabButton with no label.
   */
  public TabButton() {
      this("");
  }

  /**
   * Constructs a TabButton with the specified label.
   * @param label the label of the button
   */
  public TabButton(String label) {
      this.label = label;
      enableEvents(AWTEvent.MOUSE_EVENT_MASK);
  }

  /**
   * gets the label
   * @see setLabel
   */
  public String getLabel() {
      return label;
  }
  
  /**
   * sets the tab label
   * @see getLabel
   */
  public void setLabel(String label) {
      this.label = label;
      invalidate();
      repaint();
  }
    /*
  public Color getBackground () {
      return bgColor;
  }

  public Color getForeground () {
      return fgColor;
  }
    */

  public void setBackground ( Color bgColor ) {
     this.bgColor = bgColor;
  }

  public void setForeground ( Color fgColor ) {
     this.fgColor = fgColor;
  }

  public void setFont( Font font) {
      this.font = font;
  }
  
  /**
   * paints the TabButton
   */
  public void paint(Graphics g) {
      Dimension size =  getSize();
      int width = size.width;
      int height = size.height;
      
      // draw the perimeter of the button
      // top left side first
      //g.setColor(getBackground().brighter().brighter().brighter());
      g.setColor(SystemColor.controlLtHighlight);
      g.fillRect (0, 0, width-2, height-2);

      // bottom right next
      g.setColor(getBackground().darker().darker().darker());
      g.fillRect (1, 1, width-1, height-2);
     
      // paint the interior of the button
      if(pressed) {
	  g.setColor(getBackground().darker().darker());
      } else {
	  g.setColor(getBackground());
      }
      g.fillRect (1, 1, width-3, height-1); 
      
      // draw the label centered in the button
      if(font != null) {
	  FontMetrics fm = getFontMetrics(font);
	  g.setColor(getForeground());
	  g.drawString(label,
		       ( width - fm.stringWidth(label))/2,
		       height/2 + fm.getMaxDescent());
      }
  }
  
  /**
   * The preferred size of the button. 
   */
  public Dimension getPreferredSize() {
      if(font != null) {
	  FontMetrics fm = getFontMetrics(font);
	  Dimension size = new Dimension(fm.stringWidth(label) + PADX, fm.getHeight() + PADY);
         return size;
      } else {
	  return new Dimension(25, 25);
      }
  }
  
  /**
   * The minimum size of the button. 
   */
  public Dimension getMinimumSize() {
      return getPreferredSize();
  }

  /**
   * Adds the specified action listener to receive action events
   * from this button.
   * @param listener the action listener
   */
   public void addActionListener(ActionListener listener) {
       actionListener = AWTEventMulticaster.add(actionListener, listener);
       enableEvents(AWTEvent.MOUSE_EVENT_MASK);
   }
 
   /**
    * Removes the specified action listener so it no longer receives
    * action events from this button.
    * @param listener the action listener
    */
   public void removeActionListener(ActionListener listener) {
       actionListener = AWTEventMulticaster.remove(actionListener, listener);
   }

  /**
   * Determine if click was inside tab button.
   */
   public boolean contains(int x, int y) {
       Dimension size = getSize();
       int width = size.width;
       int height = size.height;
       return ((x > 0 && x < width) && (y > 0 && y < height));
   }
   
   /**
    * Paints the button and distribute an action event to all listeners.
    */
   public void processMouseEvent(MouseEvent e) {
       Graphics g;
       switch(e.getID()) {
          case MouseEvent.MOUSE_PRESSED:
	    // render myself inverted....
	    pressed = true;
	    repaint(); 
	    break;
          case MouseEvent.MOUSE_RELEASED:
	    if(actionListener != null) {
	       actionListener.actionPerformed(new ActionEvent(
		   this, ActionEvent.ACTION_PERFORMED, label));
	    }
	    // render myself normal again
	    if(pressed == true) {
		pressed = false;
		repaint();
	    }
	    break;
          case MouseEvent.MOUSE_ENTERED:

	    break;
          case MouseEvent.MOUSE_EXITED:
	    if(pressed == true) {
		pressed = false;

		repaint();
	    }
	    break;
       }
       super.processMouseEvent(e);
   }
}

