/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TabbedPanel.java,v 1.7 2001/05/25 01:39:11 kuldeep Exp $
 */

package net.jxta.widget;

import java.awt.*;
import java.awt.event.*;

public class TabbedPanel extends Panel implements ActionListener {

    private TabPanel       	tabPanel;       // Panel with Tabs

    private Panel               cards;          // Panel for card layout

    private int                 top;            // top of cards panel
    private Frame               frame;

    private GridBagLayout gridbag;
    private GridBagConstraints c;

    /**
     * Constructs a TabbedPanel 
     */
    public TabbedPanel(Frame frame)  {
        this.frame = frame;

        cards = new Panel();
        cards.setLayout(new CardLayout());

        gridbag = new GridBagLayout();
        c = new GridBagConstraints();
        setLayout(gridbag);

        // lay the Tabs Panel on top of cards
        c.gridx = 0;
        c.gridy = 0;
        c.weightx = 1.0;
        c.weighty = 0.0;
        c.anchor = GridBagConstraints.SOUTHWEST;
        c.fill = GridBagConstraints.HORIZONTAL;
        c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = 1;//GridBagConstraints.RELATIVE;
        c.insets = new Insets (0, 0, 0, 0);

        tabPanel = new TabPanel();
        gridbag.setConstraints(tabPanel, c);
        add (tabPanel, c);

        c.anchor = GridBagConstraints.NORTHWEST;
        c.weightx = 1.0;
        c.weighty = 1.0;
        c.gridx = 0;
        c.gridy = 1; //GridBagConstraints.RELATIVE;
        c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = GridBagConstraints.REMAINDER;
        c.insets = new Insets (10, 10, 10, 10);
        c.fill = GridBagConstraints.BOTH;

        gridbag.setConstraints(cards, c);
        add(cards);
    }

    /**
     * Add Tabs to TabbedPanel
     * @param String label
     * @param Component component to added
     */
    public void addTab (String label, Component component) {
        cards.add(label, component);   // add card to cards panel

        tabPanel.addTab (label);       // add tab to tabs panel
    }

    public void actionPerformed ( ActionEvent evt ) {
	tabPanel.actionPerformed ( evt );
    }

    /**
     * Sets font for Tabs
     *
     * @param Font font
     * Default font is Dialog
     */
    public void setFont (Font font) {
        tabPanel.setFont (font);
    }

    public void setBackground ( Color bgColor ) {
        tabPanel.setBackground ( bgColor );
    }

    public void setForeground ( Color fgColor ) {
        tabPanel.setForeground ( fgColor );
    }    

    public void paint (Graphics g) {
        int thickness = 1;

	Dimension size = getSize();
        Color topLeft = SystemColor.controlLtHighlight;//getBackground().brighter().brighter().brighter();
        Color botRight = SystemColor.controlShadow;//getBackground().darker().darker().darker();

        Rectangle border = getBounds();
        int w = border.width; 
        int h = border.height; 
        int x = 0;
        int y = 0;

        g.setColor (topLeft);
        // draw borderline along the left side of this panel
        g.fillRect ( x, y, thickness, h);

        g.setColor (botRight);
        // draw bordeline along right side of panel
        g.fillRect ( x + w - thickness, y + top, 
                     thickness, h - top);
        // draw border line at the bottom of this panel
        g.fillRect ( x, y + h - thickness, w, thickness);

	super.paint(g);  // ensures lightweights are painted
    }

    /**
     * TabPanel - a lightweight Panel that holds all the Tabs
     *
     */
    class TabPanel extends Container implements ActionListener{

        private boolean pressed = false;     // true if the button is depressed
        private static final int PADX = 5;   // padding for x direction
        private static final int PADY = 5;   // padding for y direction
        private static final int HGAP = 1;   // padding between tabs
        private static final int MAXTABS = 10;
        private TabButton[] tabs = new TabButton[MAXTABS]; 

        private int    tabCount = 0;
        private int    activeTab = 0;
        // default font settings
        private Font   font = new Font ("Dialog", Font.BOLD, 12);
        // dividers at the bottom of tabs
        private Divider ruleLeft, ruleRight;
        private int thickness = 2;           // default thickness of dividers

  
        /**
         * Constructs a TabPanel 
         */
        public TabPanel() {
            setLayout(null);

            ruleLeft = new Divider();     // divider for left of active tab
            ruleRight = new Divider();    // right of active tab
            add(ruleLeft);
            add(ruleRight);
        }

        public void actionPerformed(ActionEvent event) {
            CardLayout cl = (CardLayout)(cards.getLayout());
            cl.show(cards, (String)event.getActionCommand());

            for (int i = 0; i < tabCount; i++) {
	      if (event.getActionCommand().equals (tabs[i].getLabel()))
                    activeTab = i;
            }
            adjustRulers();
        }

        /**
         * Adds tabs to panel
         * 
         * @param String label - tab label
         */
        public void addTab (String label) {
            TabButton button = new TabButton(label);
            if (tabCount==0){
                button.setLocation(0,0);
            } 
            else
                button.setLocation(tabs[tabCount-1].getBounds().x + 
                      tabs[tabCount-1].getSize().width + HGAP, 0);
            add(button);
            button.setSize(button.getPreferredSize()); 
            button.addActionListener(this);
            tabs[tabCount] = button;
            tabCount++;
            repaint();
        }

        /**
         * sets tab fonts
         * 
         * @param Font font
         * Default font is dialog
         */
        public void setFont (Font font) {
            this.font = font;
        }


	public void setBackground ( Color bgColor ) {
	    for ( int i = 0; i < tabCount; i++ )
		tabs[i].setBackground ( bgColor );
	}
	
	public void setForeground ( Color fgColor ) {
	    for ( int i = 0; i < tabCount; i++ )
		tabs[i].setForeground ( fgColor );
	}    

        /**
         * The preferred size of the button. 
         */
        public Dimension getPreferredSize() {
            if(font != null) {
	        FontMetrics fm = getFontMetrics(font);
	        Dimension size = new Dimension(fm.stringWidth("abc") + PADX, 
                                     fm.getHeight() + PADY);
               return size;
            } else {
	        return new Dimension(25, 25);
            }
        }
  
        /**
         * The minimum size of the button. 
         */
        public Dimension getMinimumSize() {
            return getPreferredSize();
        }

        private void adjustRulers() {
             Rectangle rt = tabs[activeTab].getBounds();
             int myWidth = getSize().width;

             ruleLeft.setBounds (0, rt.height - thickness, rt.x, thickness);
             ruleRight.setBounds (rt.x + rt.width-thickness, rt.height - thickness, 
                      myWidth - ( rt.x + rt.width), thickness);
        }

        public void setSize(int w, int h) {
            setBounds (getLocation().x, getLocation().y, w, h);
        }

        public void setBounds(int x, int y, int w, int h) {
            super.setBounds (x, y, w, h);

            top = y + h;  // top of the Cards Panel
            adjustRulers();
        }

    }
}

