/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: viewfile.cpp,v 1.2 2001/06/19 02:44:56 akhil Exp $
 */

/*
 * viewfile.exe, for Win32 systems.
 *
 * This program will find and launch a viewer for a given data file.
 * The end result will the same as if the user had double-clicked on the
 * file in the Windows Explorer.
 *
 * Syntax: viewfile [-s] filename
 *    -s : suppress warnings about executable files.
 *
 * The filename parm may be a complete or relative path to the data file.
 *
 */

#include <windows.h>
#include <process.h>
#include <stdio.h>
#include <stdlib.h>

#define VIEWFILE_VERSION "ViewFile 1.0"


// Error codes returned by this program.
// Returns zero for success.
enum ErrorList {
    ERROR_NO_FILENAME_PARM = 1,
    ERROR_USER_CANCEL,
    ERROR_STARTING_OPENWITH_DLG,
    ERROR_UNABLE_TO_OPEN_FILE,
};

// Error messages for errors returned by the ShellExecute() function. 
// To Do: these strings should be moved to a string table in a resource file.
//        (Will help if we ever internationalize it.)
//
#define STR_ERROR_MEMORY_RESOURCES "The operating system is out of memory or resources."
#define STR_ERROR_FILE_NOT_FOUND   "The specified file was not found."
#define STR_ERROR_PATH_NOT_FOUND   "The specified path was not found."
#define STR_ERROR_BAD_FORMAT       "The .EXE file is invalid (non-Win32 .EXE or error in .EXE image)."
#define STR_SE_ERR_ACCESSDENIED    "The operating system denied access to the specified file."
#define STR_SE_ERR_ASSOCINCOMPLETE "The filename association is incomplete or invalid."
#define STR_SE_ERR_DDEBUSY         "The DDE transaction could not be completed because other DDE transactions were being processed"
#define STR_SE_ERR_DDEFAIL         "The DDE transaction failed."
#define STR_SE_ERR_DDETIMEOUT      "The DDE transaction could not be completed because the request timed out."
#define STR_SE_ERR_DLLNOTFOUND     "The specified dynamic-link library was not found."
#define STR_SE_ERR_OOM             "There was not enough memory to complete the operation."
#define STR_SE_ERR_SHARE           "A sharing violation occurred."


/*****************************************************************************
*
* ReportShellExecuteError()
*
* Report errors returned by the ShellExecute() function.
*
*****************************************************************************/

void ReportShellExecuteError(int errCode)
{
    char errMsg [128];

    switch (errCode) {
    case 0:                      strcpy(errMsg, STR_ERROR_MEMORY_RESOURCES); break;
    case ERROR_FILE_NOT_FOUND:   strcpy(errMsg, STR_ERROR_FILE_NOT_FOUND);   break;
    case ERROR_PATH_NOT_FOUND:   strcpy(errMsg, STR_ERROR_PATH_NOT_FOUND);   break;
    case ERROR_BAD_FORMAT:       strcpy(errMsg, STR_ERROR_BAD_FORMAT);       break;
    case SE_ERR_ACCESSDENIED:    strcpy(errMsg, STR_SE_ERR_ACCESSDENIED);    break;
    case SE_ERR_ASSOCINCOMPLETE: strcpy(errMsg, STR_SE_ERR_ASSOCINCOMPLETE); break;
    case SE_ERR_DDEBUSY:         strcpy(errMsg, STR_SE_ERR_DDEBUSY);         break;
    case SE_ERR_DDEFAIL:         strcpy(errMsg, STR_SE_ERR_DDEFAIL);         break;
    case SE_ERR_DDETIMEOUT:      strcpy(errMsg, STR_SE_ERR_DDETIMEOUT);      break;
    case SE_ERR_DLLNOTFOUND:     strcpy(errMsg, STR_SE_ERR_DLLNOTFOUND);     break;
    case SE_ERR_OOM:             strcpy(errMsg, STR_SE_ERR_OOM);             break;
    case SE_ERR_SHARE:           strcpy(errMsg, STR_SE_ERR_SHARE);           break;
    }

    MessageBox(NULL, errMsg, VIEWFILE_VERSION, MB_ICONEXCLAMATION | MB_OK);

}

/*****************************************************************************
*
* WinMain()
*
*****************************************************************************/

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
    int rc = NO_ERROR;

    char filePath [MAX_PATH];
    char fileExt  [MAX_PATH];
    BOOL warnExe = TRUE;
    int  currArg = 1;

    // Check for the -s option.
    if (__argc > 1  &&  !stricmp(__argv[1], "-s")) {
        warnExe = FALSE;
        currArg++;
    }

    // Check for existance of filename on command-line.
    if (__argc < currArg + 1) {
        MessageBox(NULL, "Error: No filename given.\n\n"
                   "Syntax: viewfile [-s] filename\n"
                   "-s: suppress warnings about executable files.",
                   VIEWFILE_VERSION, MB_ICONEXCLAMATION | MB_OK);

        return ERROR_NO_FILENAME_PARM;
    }


    // Get the filename to view.
    memset(filePath, 0, sizeof(filePath));
    strncpy(filePath, __argv[currArg], sizeof(filePath) - 1);

    // Check for executable files.  We'll warn the user before running them,
    // unless the -s option was used.
    if (warnExe) {
        _splitpath(filePath, NULL, NULL, NULL, fileExt);
        if (!stricmp(fileExt, ".exe")  ||
            !stricmp(fileExt, ".com")  ||
            !stricmp(fileExt, ".bat")  ||
            !stricmp(fileExt, ".vbs")) {

            char warning [256];
            sprintf(warning, "Warning: about to execute %s.\nContinue?", filePath);
            int yn = MessageBox(NULL, warning, VIEWFILE_VERSION, 
                                MB_ICONQUESTION | MB_YESNO);
            if (yn == IDNO)
                return ERROR_USER_CANCEL;
        }
    }
            

    // The ShellExecute() Windows API call will open the data file using
    // the appropriate program.  If the data file is an executable, it
    // will be run.  ShellExecute will return a value greater than 32 if
    // it succeeds.
    //
    int hInst = (int)ShellExecute(NULL, NULL, filePath, NULL, NULL, SW_SHOW);
    if (hInst <= 32) {

        // If there's no association registered for this file extension,
        // trigger the Windows shell's "Open With" dialog, to select an
        // application.
        if (hInst == SE_ERR_NOASSOC) {
            int spawnRc = _spawnlp(_P_NOWAIT, "rundll32.exe", "rundll32.exe",
                                   "shell32.dll,OpenAs_RunDLL", filePath, NULL);
            if (spawnRc == -1)
                rc = ERROR_STARTING_OPENWITH_DLG;
        }
        else {
            ReportShellExecuteError(hInst);
            rc = ERROR_UNABLE_TO_OPEN_FILE;
        }
    }

    return rc;
}
