package freenet;

use strict;
use IO::Socket;
use Digest::SHA1 qw(sha1_hex);

$freenet::VERSION = "0.01";
$freenet::PROTOVER = "1.2";

sub configure {
  $freenet::node = shift;
  $freenet::port = shift;
  $freenet::dh = shift;
  $freenet::dl = shift;
  $freenet::hop = shift;
  $freenet::hash = shift;
}


sub request {
  my $key = shift;

  my ($iaddr, $paddr, $proto, $id, $message, $messagetype, %fields, $field,
      $depth, @lines);

  local *SOCK;

  #
  # convert service name to port, if necessary
  #

  $freenet::port = getservbyname($freenet::port, 'tcp') if ($freenet::port =~ /\D/);

  if (!$freenet::port) {
    return "Could not convert service name to port number.";
  }

  $iaddr = inet_aton($freenet::node);

  if (!$iaddr) {
    return "No such host: $freenet::node";
  }

  $paddr = sockaddr_in($freenet::port, $iaddr);
  $proto = getprotobyname('tcp');

  #
  # now open the server and send the HandshakeRequest
  #

  if (!socket(SOCK, PF_INET, SOCK_STREAM, $proto)) {
    return "Unable to create socket. ($!)";
  }

  #
  # no buffering on socket
  #

  my $oldfh = select(SOCK); $| = 1; select($oldfh);

  #
  # connect to freenet node
  #

  if (!connect(SOCK, $paddr)) {
    return "Unable to connect. ($!)";
  }

  #
  # assemble handshake message and send to node
  #

  $id = mk_random_id();

  $message = "HandshakeRequest\n" .
             "UniqueID=$id\n" .
	     "Depth=1\n" .
	     "HopsToLive=1\n" .
             "KeepAlive=true\n" .
             "EndMessage\n";

  print SOCK $message;

  #
  # get returned message type
  #

  $messagetype = <SOCK>;

  #
  # read in message contents
  #

  while (defined($field = <SOCK>) && ($field =~ /=/)) {
    my($fieldname, $fieldvalue) = split('=', $field);

    chomp $fieldvalue;
    $fields{$fieldname} = $fieldvalue;
  }

  #
  # returned message should be HandshakeReply
  #

  if ($messagetype =~ /HandshakeReply/) {
    if ($fields{'Version'} < $freenet::PROTOVER) {
      return "Server running version $fields{'Version'}, expected $freenet::PROTOVER or better.";
    }
  } else {
    return "Received invalid response: expected handshake.";
  }

  #
  # hash key for the request
  #

  if ($freenet::hash eq 'SHA1') {    
    $key = sha1_hex($key);    
    $key =~ tr/[a-z]/[A-Z]/;
  } else {
    return "Unknown hash function: $freenet::hash";
  }

  #
  # assemble DataRequest message
  #

  $id = mk_random_id();
  $depth = (int(rand($freenet::dh - $freenet::dl + 1)) + $freenet::dl);

  $message = "DataRequest\n" .
             "SearchKey=$key\n" . 
             "UniqueID=$id\n" . 
             "Depth=$depth\n" . 
             "HopsToLive=$freenet::hop\n" . 
             "KeepAlive=true\n" . 
             "EndMessage\n";

  print SOCK $message;

  do {

    #
    # get returned message type
    #

    $messagetype = <SOCK>;

    #
    # read message contents
    #

    %fields = ();

    while (defined($field = <SOCK>) && ($field =~ /=/)) {
      my($fieldname, $fieldvalue) = split('=', $field);

      chomp $fieldvalue;
      $fields{$fieldname} = $fieldvalue;
    }

  } while ($messagetype =~ /QueryRestarted/);

  #
  # message reply should be DataReply
  #

  if ($messagetype =~ /DataReply/) {
    if ($fields{'UniqueID'} eq $id) {

      return ("Success", *SOCK);

    }
  } elsif ($messagetype =~ /RequestFailed/) {
    if ($fields{'UniqueID'} eq $id) {
      close (SOCK);
      return "Requested key not found.";
    }
  } elsif ($messagetype =~ /TimedOut/) {
    if ($fields{'UniqueID'} eq $id) {
      close (SOCK);
      return "Request timed out.";
    }
  } else {
    if ($fields{'UniqueID'} eq $id) {
      close (SOCK);
      return "Received unexpected reply: $messagetype\n%fields";
    }
  }

  #
  # terminate connection to node
  #
  #if (!close(SOCK)) {
  #  return "Unable to close socket. ($!)";
  #}

}

sub mk_random_id() {
  my($id) = ();

  for (1 .. 16) { 
    $id .= sprintf("%x", int(rand(16)));
  }

  return $id;
}

1;
