# Steve's key index
#
# Shamelessly written by Steven Hazel
#
# This code is in the public domain.  As if you care.

package keyindex;

$keyindex::VERSION = "0.01";

use strict;
use freenet;
use LWP::Simple;
use Fcntl ':flock'; # import LOCK_* constants


sub add_key {
  my $indexfile = shift;
  my $lockfile = "$indexfile.lock";
  my $key = shift;
  my $line;
  my $found = 0;

  $key =~ s/^\s*//;
  $key =~ s/\s*$//;
  if ($key eq "") { return };

  open(LF, ">$lockfile")
    or die "Can't open lockfile: $lockfile";
  flock(LF,LOCK_EX);

  open(KR, "$indexfile")
    or die "Can't open keyindex: $indexfile";

  while ($line = <KR>) {
    if ($line eq "$key\n") {
      $found = 1;
      last;
    }
  }

  close(KR);

  if ($found == 0) {
    open(KW, ">>$indexfile")
      or die "Can't open keyindex: $indexfile";
    print KW "$key\n";
    close(KW);
  }

  flock(LF,LOCK_UN);
  close(LF);
}


sub del_key {
  my $indexfile = shift;
  my $lockfile = "$indexfile.lock";
  my $purgefile = "$indexfile.purged";
  my $key = shift;
  my $line;
  my @lines = ();
  my $found = 0;

  $key =~ s/^\s*//;
  $key =~ s/\s*$//;
  if ($key eq "") { return };

  open(LF, ">$lockfile")
    or die "Can't open lockfile: $lockfile";
  flock(LF,LOCK_EX);

  open(KR, "$indexfile")
    or die "Can't open keyindex: $indexfile";

  while ($line = <KR>) {
    if ($line ne "$key\n") {
      push(@lines, $line);
    } else {
      $found = 1;
    }
  }

  close(KR);

  if ($found == 1) {
    open(KW, ">$indexfile")
      or die "Can't open keyindex: $indexfile";
    print KW @lines;
    close(KW);
    open(PW, ">>$purgefile")
      or die "Can't open purge file: $purgefile";
    print PW "$key\n";
    close(PW);
  }

  flock(LF,LOCK_UN);
  close(LF);
}


sub get_keys {
  my $indexfile = shift;
  my $lockfile = "$indexfile.lock";
  my $line;
  my @lines = ();

  open(LF, ">$lockfile")
    or die "Can't open lockfile: $lockfile";
  flock(LF,LOCK_EX);

  open(KR, "$indexfile")
    or die "Can't open keyindex: $indexfile";

  while ($line = <KR>) {
    push(@lines, $line);
  }

  close(KR);

  flock(LF,LOCK_UN);
  close(LF);

  return @lines;
}


sub clean {
  my $indexfile = shift;
  my $lockfile = "$indexfile.lock";
  my $line;
  my $previous;
  my @lines = ();

  open(LF, ">$lockfile")
    or die "Can't open lockfile: $lockfile";
  flock(LF,LOCK_EX);

  open(KR, "$indexfile")
    or die "Can't open keyindex: $indexfile";

  while ($line = <KR>) {
    push(@lines, $line);
  }

  close(KR);

  @lines = sort(@lines);

  open(KW, ">$indexfile")
    or die "Can't open keyindex: $indexfile";
  $previous = "";
  for $line (@lines) {
    $line =~ s/^\s*//;
    $line =~ s/\s*$//;
    if (($line ne $previous) and ($line ne "")) {
      print KW "$line\n";
      $previous = $line;
    }
  }
  close(KW);

  flock(LF,LOCK_UN);
  close(LF);
}


sub merge {
  my $indexfile = shift;
  my $lockfile = "$indexfile.lock";
  my $keyindex = shift;
  my $line;
  my @lines = ();

  @lines = split(/\n/, get($keyindex));

  if (@lines) {
    open(LF, ">$lockfile")
      or die "Can't open lockfile: $lockfile";
    flock(LF,LOCK_EX);

    open(KW, ">>$indexfile")
      or die "Can't open keyindex: $indexfile";
    
    for $line (@lines) {
      $line =~ s/^\s*//;
      $line =~ s/\s*$//;
      print KW "$line\n";
    }

    close(KW);

    flock(LF,LOCK_UN);
    close(LF);
  }

  clean($indexfile);

}


# slow
sub refile {
  my $indexfile = shift;
  my $lockfile = "$indexfile.lock";
  my $purgefile = "$indexfile.purged";
  my $node = shift;
  my $port = shift;
  my $dh = shift;
  my $dl = shift;
  my $hop = shift;
  my $hash = shift;
  my $line;
  my $previous;
  my @lines = ();
  my $status;
  my $sock;

  freenet::configure($node, $port, $dh, $dl, $hop, $hash);

  open(LF, ">$lockfile")
    or die "Can't open lockfile: $lockfile";
  flock(LF,LOCK_EX);

  open(KR, "$indexfile")
    or die "Can't open keyindex: $indexfile";

  while ($line = <KR>) {
    push(@lines, $line);
  }

  close(KR);

  open(PR, "$purgefile")
    or die "Can't open purge file: $purgefile";

  while ($line = <PR>) {
    push(@lines, $line);
  }

  close(PR);

  @lines = sort(@lines);

  open(KW, ">$indexfile")
    or die "Can't open keyindex: $indexfile";
  open(PW, ">$purgefile")
    or die "Can't open purge file: $purgefile";
  $previous = "";
  for $line (@lines) {
    $line =~ s/^\s*//;
    $line =~ s/\s*$//;
    if (($line ne $previous) and ($line ne "")) {
      ($status, $sock) = freenet::request($line);
      if ($status eq "Success") { 
	close($sock);
	print KW "$line\n";
      } else {
	print PW "$line\n";
      }
      $previous = $line;
    }
  }
  close(KW);
  close(PW);

  flock(LF,LOCK_UN);
  close(LF);
}

1;
