// Jason Rohrer
// ClickableJList.java

/**
*
*	JList that responds to double click events by firing
*   ListChoiceEvents.
* 
*	Created 8-10-2000
*	Mods:
*		Jason Rohrer 8-11-2000	Added other constructors 
*								for completness.  Changed so that
*								it fires a vector of selected items.
*								Added support for choosing items
*								when Enter key is hit.
*		Jason Rohrer 8-16-2000	Changed add() to addElement() to support 
*								java 1.1
*/

package konspire.client;

import javax.swing.JList;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyAdapter;

import konspire.common.Vector;

/**
 * <code>JList</code> that responds to double click events by firing
 * <code>ListChoiceEvents</code>.
 *
 * @author Jason Rohrer
 */
public class ClickableJList extends JList {
	
	private Vector mListenerList = new Vector();
	
	
	
	/**
	 * Virtual key code for key that chooses selected items.
	 * <p>
	 * Default is the enter key.
	 */
	private int mChoiceKeyCode = KeyEvent.VK_ENTER;
	
	
	
	/**
	 * Constructs a <code>ClickableJList</code>.
	 */
	public ClickableJList() {
		super();
		setup();
		}
	
	
	
	/**
	 * Constructs a <code>ClickableJList</code>.  Provides
	 * functionality of similar <code>JList</code> constructor.
	 */
	public ClickableJList( javax.swing.ListModel dataModel ) {
		super( dataModel );
		setup();
		}
		
		
		
	/**
	 * Constructs a <code>ClickableJList</code>.  Provides
	 * functionality of similar <code>JList</code> constructor.
	 */
	public ClickableJList( Object[] listData ) {
		super( listData );
		setup();
		} 
		
		
		
	/**
	 * Constructs a <code>ClickableJList</code>.  Provides
	 * functionality of similar <code>JList</code> constructor.
	 */
	public ClickableJList( java.util.Vector listData ) {
		super( listData );
		setup();
		} 
	
	
	
	/**
	 * Common setup code shared by all constructors.
	 */
	private void setup() {
		addMouseListener( new DoubleClickListener() );
		addKeyListener( new EnterKeyListener() );
		}
	
	
	
	
	/**
	 * Adds a <code>ListChoiceListener</code> to this JList.
	 * <p>
	 * Each listener will receive a <code>ListChoiceEvent</code>
	 * every time an item from the list is double clicked.
	 *
	 * @param inListener the listener to add.
	 */
	public void addListChoiceListener( ListChoiceListener inListener ) {
		mListenerList.addElement( inListener );
		}

	
	
	/**
	 * Removes a <code>ListChoiceListener</code> from this JList.
	 *
	 * @param inListener the listener to remove.
	 */
	public void removeListChoiceListener( ListChoiceListener inListener ) {
		mListenerList.removeElement( inListener );
		}
	
	
	
	/**
	 * Sets the virtual key code for the key used to choose items
	 * from the list (as an alternate to double clicking).
	 *
	 * @param inVirtualKeyCode virtual key code for the choice key.
	 *  Valid values are constants defined in {@link #KeyEvent}.
	 */
	public void setChoiceKey( int inVirtualKeyCode ) {
		mChoiceKeyCode = inVirtualKeyCode;
		}
	 
	
	
	/**
	 * Listens for <code>DoubleClickEvent</code>s and fires them to
	 * all registered listeners.
	 */
	private class DoubleClickListener extends MouseAdapter {
		
		// overrides MouseAdapter mouseClicked( MouseEvent ) method
		public void mouseClicked( MouseEvent inME ) {
			if( inME.getClickCount() == 2 ) {
				// make sure clicked item is still selected
				//int index = locationToIndex( e.getPoint() );
				//setSelectedIndex( index );
				
				Object selectedObjects[] = getSelectedValues();
				
				
				ListChoiceEvent clickEvent = 
					new ListChoiceEvent( selectedObjects );
				
				int numListeners = mListenerList.size();
				for( int i=0; i<numListeners; i++ ) {
					ListChoiceListener listener =
						(ListChoiceListener)( mListenerList.elementAt( i ) );
					listener.listItemChosen( clickEvent );
					}
				}
			}
		}	
	
	
	
	/**
	 * Listens for enter key strokes and fires ListChoiceEvents
	 * to all registered listeners.
	 */
	private class EnterKeyListener extends KeyAdapter {
		
		// overrides KeyAdapter keyReleased( KeyEvent ) method
		public void keyReleased( KeyEvent inKE ) {
			
			if( inKE.getKeyCode() == mChoiceKeyCode ) {
				Object selectedObjects[] = getSelectedValues();
				
				
				ListChoiceEvent clickEvent = 
					new ListChoiceEvent( selectedObjects );
				
				int numListeners = mListenerList.size();
				for( int i=0; i<numListeners; i++ ) {
					ListChoiceListener listener =
						(ListChoiceListener)( mListenerList.elementAt( i ) );
					listener.listItemChosen( clickEvent );
					}
				}
			}
		}	
	
	
	}
