// Jason Rohrer
// Client.java

/**
*
*	konspire client 
*
*	Created 5-11-2000
*	Mods:
*		Jason Rohrer	6-16-2000	Client now randomizes order of server list vector
*									before trying to connect. 
*		Jason Rohrer	6-25-2000	Fixed bug that caused client to start rejecting connections
*									number of connected clients was being incremented instead
*									of decremented when a client disconnnected.
*		Jason Rohrer	7-4-2000	Changed so that client looks for a "good" local
*									address in the server list and replaces
*									its local address with the new address.
*									Client is responsible for updating its FileHost
*									message so that all file descriptors have the correct
*									address.
*		Jason Rohrer	7-23-2000	Changed so that all spawned threads are set to 
*									minimum priority to make sure that the GUI
*									is always completely responsive.  Note that this
*									priority setting doesn't usually increase GUI responsiveness
*									noticibly, since most threads are handling network connections
*									and are therefore blocked often enough so that the GUI thread
*									has ample time to run.  However, on faster network connections,
*									e.g., gigabit ethernet, the network threads will not
*									be blocked as often and could cause GUI responsiveness
*									to be sporadic.  We're assuming that as long as the GUI
*									isn't being manipulated or updated, MIN_PRIORITY thread
*									performance will not be affected.  This should be verified.
*		Jason Rohrer	7-24-2000	Fixed 1.1 compatability bug in menu handle.  Was using getParentFile(),
*									needed to use getParent().
*		Jason Rohrer	8-3-2000	Changed layout of main frame to GridBagLayout to
*									make sure search results table is as large as possible.
*		Jason Rohrer	8-23-2000	Added JLog logging support.
*		Jason Rohrer	8-27-2000	Added more verbose detail logging.
*									Synchronized client connection counter.
*		Jason Rohrer	9-22-2000	Changed socket connections to server from
*									TwoWayMessageSockets to
*									CheckedMessageSockets to participate in live 
*									host detection scheme.  Note that client-to-
*									client sockets are still ordinary 
*									TwoWayMessageSockets because they are used 
*									to send binary data.
*		Jason Rohrer	9-23-2000	Added support for refreshing server list.
*		Jason Rohrer	9-29-2000	Added support for saving window position and
*									location in ClientSettings file.
*		Jason Rohrer	9-30-2000	Added support for a multi-directory file
*									path.
*									Cleaned up older, commented out code.
*/

package konspire.client;

import konspire.common.*;
import konspire.common.log.AppLog;
import konspire.common.log.LogManager;

import konspire.common.FileDescriptor;

import java.net.*;
import java.io.*;
import java.awt.Event;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.WindowEvent;
import java.awt.FlowLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Dimension;
import javax.swing.*;


import java.util.Enumeration;
import java.util.NoSuchElementException;

import java.util.Random;


/**
 * Main <code>JFrame</code> for konspire client.  Contains main method.
 *
 * @author Jason Rohrer
 */
public class Client extends JFrame implements DownloadController {
	
	static int defaultPort = 6433;
	static int defaultMaxClients = 10;
	
	static Random randGen = new Random();
	
	int port;

	int maxClients;
	int maxConnections;
	
	int numClients = 0;
	
	ClientSettingsFile mSettingsFile = new ClientSettingsFile();
	ClientSettings mSettings = mSettingsFile.mClientSettings;
	
	
	static String hostsFileName = "servers.dat";
	static String serverListTextFileName = "servers.txt";
	ServerListFile sListFile;
	
	Vector serverVector;	// vector of Hosts
	
	boolean connected = false;		// connected to server?
	
	Host currentServer = null;
	
	CheckedMessageSocket messageSock = null;	// socket connection to server
	
	// client info
	
	Host localAddress;
	
	FileHost fileListMessage;	// message to send to server on connect accept
	
	WindowCloseListener closeListener;
	MenuActionListener menuListener;
	
	JPanel infoPanel;	// contains HardenedCriminal info

	JPanel infoNamePanel;
	JPanel infoAddressPanel;
	
	
	JLabel nameLabel;
	JLabel addressLabel;
	
	StatePanel stPanel;
	SearchPanel srchPanel;
	ServerListPanel sListPanel;
	
	SearchResultsTable srTable;
	DownloadTable dlTable;
	UploadTable ulTable;
	
	JPanel topPanel;
	
	JPanel topTopPanel;
	JPanel topBottomPanel;
	
	JPanel tablesPanel;
	JPanel transferTablesPanel;
	
	JFrame transferFrame;	// separate frame for transfers

	JMenuBar mainMenuBar = new JMenuBar();
	
	JMenu fileMenu = new JMenu( "file" );
	JMenuItem selectFilesDirectoryMenuItem = 
		new JMenuItem( "select files directories" );
	JMenuItem rescanFilesMenuItem = new JMenuItem( "rescan files" );
	JMenuItem saveTextServerListMenuItem = new JMenuItem( "save text server list" );
	JMenuItem quitMenuItem = new JMenuItem( "quit" );

	// menu to control network functionality
	JMenu networkMenu = new JMenu( "network" );
	JMenuItem refreshServerListMenuItem = 
		new JMenuItem( "refresh server list" );
	
	// menu to control which windows are showing
	JMenu windowMenu = new JMenu( "window" );
	JMenuItem transfersMenuItem = new JMenuItem( "transfers" );

	
	/**
	 * Constructs a client.
	 *
	 * @param prt port client will listen on
	 * @param maxCli maximum number of other clients that can connect
	 * to this client simultaneously
	 */
	public Client( int prt, int maxCli ) {
		
		super("konspire..client");
		
		GuiMessage.setParent( this );	// take over parenting for gui messages	
		
		port = prt;
		
		// get local host address
		try {
			InetAddress iAddress = InetAddress.getLocalHost();
			localAddress = new Host( iAddress.getHostAddress(), port );
			}
		catch( UnknownHostException e ) {
			AppLog.criticalError( "failed to retrieve local host IP address.",
				true );
			ErrorHandler.report( e );
			System.exit(1);
			}
		
		// setup local file host message
		fileListMessage = new FileHost( localAddress, 
			mSettings.mDirectoryPath );
	
		if( ! fileListMessage.downloadDirectoryExists() ) {
			GuiMessage.show( "download directory \"" 
				+ mSettings.mDirectoryPath.getDownloadDirectory().getName() 
				+ "\" does not exist." );
			}
		
		maxClients = maxCli;
		maxConnections =  2 * maxClients;
		
		setSize( mSettings.mMainWindowSize );
		setLocation( mSettings.mMainWindowLocation );
		
		infoPanel = new JPanel();
		infoPanel.setLayout( new BoxLayout( infoPanel, BoxLayout.X_AXIS ) );
		
		infoNamePanel = new JPanel( new FlowLayout( FlowLayout.LEFT ) );
		infoAddressPanel = new JPanel( new FlowLayout( FlowLayout.RIGHT ) );
		
		
		nameLabel = new JLabel( "HardenedCriminal software" );
		addressLabel = new JLabel( "www.hardenedcriminal.n3.net" );
		
		
		infoNamePanel.add( nameLabel );
		infoAddressPanel.add( addressLabel );
		
		infoPanel.add( infoNamePanel );
		infoPanel.add( infoAddressPanel );
		
		topPanel = new JPanel();
		topTopPanel = new JPanel();
		topBottomPanel = new JPanel();
		
		topPanel.setLayout( new BoxLayout( topPanel, BoxLayout.Y_AXIS ) );
		topTopPanel.setLayout( new BoxLayout( topTopPanel, BoxLayout.X_AXIS ) );
		topBottomPanel.setLayout( new BoxLayout( topBottomPanel, BoxLayout.X_AXIS ) );
		
		topPanel.add( infoPanel );
		topPanel.add( topTopPanel );
		topPanel.add( topBottomPanel );
		
		// need to init this first so it can be passed into srchPanel
		srTable = new SearchResultsTable( this );
		
		stPanel = new StatePanel( this );
		srchPanel = new SearchPanel( srTable, this );
		sListPanel = new ServerListPanel();
		
		topTopPanel.add( stPanel );
		topTopPanel.add( sListPanel );
		
		topBottomPanel.add( srchPanel );
		
        
		tablesPanel = new JPanel();
		
		tablesPanel.setLayout( new BoxLayout( tablesPanel, BoxLayout.Y_AXIS ) );
		
		
		transferTablesPanel = new JPanel();
		
		transferTablesPanel.setLayout( new BoxLayout( transferTablesPanel, BoxLayout.Y_AXIS ) );
		
		dlTable = new DownloadTable();
		ulTable = new UploadTable();
		
		tablesPanel.add( srTable );
		transferTablesPanel.add( dlTable );
		transferTablesPanel.add( ulTable );
		

		getContentPane().setLayout( new GridBagLayout() );
		GridBagConstraints gbc = new GridBagConstraints();


		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.gridwidth = 1;
		gbc.gridheight = 1;

		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.gridx = 0;
		gbc.gridy = 0;
		gbc.weightx = 1.0;
		gbc.weighty = 0;
		getContentPane().add( topPanel, gbc );

		gbc.gridx = 0;
		gbc.gridy = 1;
		gbc.weightx = 1.0;
		gbc.weighty = 1.0;
		gbc.fill = GridBagConstraints.BOTH;
		getContentPane().add( tablesPanel, gbc );
		
		transferFrame = new JFrame( "konspire transfers" );
		
		transferFrame.setSize( mSettings.mTransferWindowSize );
		transferFrame.setLocation( mSettings.mTransferWindowLocation );	
		
		transferFrame.getContentPane().add( transferTablesPanel );
		
		
		
		setJMenuBar( mainMenuBar );
		
		selectFilesDirectoryMenuItem.setMnemonic( 'd' );
		rescanFilesMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'R', Event.CTRL_MASK, false ) );
		rescanFilesMenuItem.setMnemonic( 'r' );
		selectFilesDirectoryMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'D', Event.CTRL_MASK, false ) );
		saveTextServerListMenuItem.setMnemonic( 'l' );
		saveTextServerListMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'L', Event.CTRL_MASK, false ) );
		quitMenuItem.setMnemonic( 'q' );
		quitMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'Q', Event.CTRL_MASK, false ) );
		
		fileMenu.add( selectFilesDirectoryMenuItem );
		fileMenu.add( rescanFilesMenuItem );
		fileMenu.add( saveTextServerListMenuItem );
		fileMenu.addSeparator();
		fileMenu.add( quitMenuItem );
		
		fileMenu.setMnemonic( 'f' );
		
		networkMenu.add( refreshServerListMenuItem );
		networkMenu.setMnemonic( 'n' );
		refreshServerListMenuItem.setMnemonic( 'r' );
		refreshServerListMenuItem.setAccelerator( 
			KeyStroke.getKeyStroke( 'U', Event.CTRL_MASK, false ) );
		
		transfersMenuItem.setMnemonic( 't' );
		transfersMenuItem.setAccelerator( 
			KeyStroke.getKeyStroke( 'T', Event.CTRL_MASK, false ) );
		
		windowMenu.add( transfersMenuItem );
		windowMenu.setMnemonic( 'w' );
		
		mainMenuBar.add( fileMenu );
		mainMenuBar.add( networkMenu );
		mainMenuBar.add( windowMenu );
		
		
		menuListener = new MenuActionListener();
		transfersMenuItem.addActionListener( menuListener );
		selectFilesDirectoryMenuItem.addActionListener( menuListener );
		rescanFilesMenuItem.addActionListener( menuListener );
		saveTextServerListMenuItem.addActionListener( menuListener );
		quitMenuItem.addActionListener( menuListener );
		refreshServerListMenuItem.addActionListener( menuListener );
		
		closeListener = new WindowCloseListener();
		
		addWindowListener( closeListener );
		
		
		// load server list from file
		sListFile = new ServerListFile( hostsFileName );
		serverVector = sListFile.readHostsVectorFromFile();
		sListPanel.setServerList( serverVector );
		
		show();
		transferFrame.show();
		}
		
	
	
	/**
	 * Rescan files directory and resend file list.
	 */
	private void rescanFiles() {
		fileListMessage = new FileHost( localAddress, 
			mSettings.mDirectoryPath );
		if( ! fileListMessage.downloadDirectoryExists() ) {
			GuiMessage.show( "download directory \"" 
				+ mSettings.mDirectoryPath.getDownloadDirectory().getName() 
				+ "\" does not exist." );
			}
		// if connected, send new file list to server
		if( connected ) {
			try {
				try {
					messageSock.obtainLock();
					}
				catch( InterruptedException inIE ) {
					AppLog.error( "Interrupted while waiting to " +
						"access socket. " );
					ErrorHandler.report( inIE );
					messageSock.closeTotal();
	            	serverConnectionBroken();
					return;
					}
				messageSock.send( fileListMessage );
				messageSock.releaseLock();
				}
			catch( IOException ioe ) {
				AppLog.error( "Connection broken to server: " + ioe.toString() );
				ErrorHandler.report( ioe );
	            serverConnectionBroken();
				}
			}	
		}
	
	
	
	/**
	 * Request a new server list.
	 */
	private void refreshServerList() {
		AppLog.info( "Trying to refresh server list." );
		// if connected, send a server list request
		if( connected ) {
			try {
				try {
					messageSock.obtainLock();
					}
				catch( InterruptedException inIE ) {
					AppLog.error( "Interrupted while waiting to " +
						"access socket. " );
					ErrorHandler.report( inIE );
					messageSock.closeTotal();
	            	serverConnectionBroken();
					return;
					}
				messageSock.send( new ServerListRequest() );
				Message m = messageSock.receive();
				messageSock.releaseLock();
				
				if( m instanceof ServerList ) {
					AppLog.info( "New server list received." );
					ServerList sList = (ServerList)m;
					serverVector = sList.getHostVector();
					sListPanel.setServerList( sList );
					
					processStats( sList );
					
					// write new server list out to file
					sListFile.writeHostsVectorToFile( serverVector );
					}
				else {
					unexpectedServerMessage( m );
					}
				}
			catch( IOException ioe ) {
				AppLog.error( "Connection broken to server: " + ioe.toString() );
				ErrorHandler.report( ioe );
	            serverConnectionBroken();
				}
			}	
		}
	
	
	/**
	 * Listener for menu events.
	 */
	private class MenuActionListener implements ActionListener {
		public void actionPerformed( ActionEvent e ) {
			if( e.getSource() == transfersMenuItem ) {
				transferFrame.show();
				}
			else if( e.getSource() == selectFilesDirectoryMenuItem ) {
				// let user modify file directory path
				// this constructor blocks until dialog box is dismissed
				new FileDirectoryDialog( 
					mSettings.mDirectoryPath, Client.this );
				
				// now write settings to file
				mSettingsFile.writeToFile();
				}
			else if( e.getSource() == rescanFilesMenuItem ) {
				rescanFiles();
				}
			else if( e.getSource() == refreshServerListMenuItem ) {
				refreshServerList();
				}
			else if( e.getSource() == saveTextServerListMenuItem ) {
				// default location for server list text file is download dir
				File startDir = mSettings.mDirectoryPath.getDownloadDirectory();
				File textServerListFile = 
					FileGuiUtils.launchSaveAs( startDir, Client.this );
				if( textServerListFile != null ) {
					sListFile.writeHostsVectorToTextFile( serverVector, textServerListFile );
					}
				}
			else if( e.getSource() == quitMenuItem ) {
				quitCleanup();
				System.exit(0);
				}
			}
		}
	
	
	/**
	 * Main client run loop.  Trys to connect to a server.
	 * Listens for incoming connections and creates
	 * a ReceiveThread for each one.
	 */
	public void run() {
		
		// first, try connecting to a server from the server list
		if( serverVector.size() >= 0 ) {
			// spawn a thread to try the connection
			ConnectThread cThread = new ConnectThread();
			cThread.start();
			}
		
		
		ServerSocket servSock;
		try {
			servSock = new ServerSocket( localAddress.getPort(), maxConnections );
			}
		catch( IOException e ) {
			AppLog.criticalError( "server socket failed to initialize.", true );
			ErrorHandler.report( e );
			return;
			}
		
		while( true ) {
			try {
				AppLog.info( "Listening for connections." );
				Socket sock = servSock.accept();
				AppLog.info( "Connection received." );
				
				// connection received
				// pass it to receiver thread
				ReceiveThread rThread = new ReceiveThread( sock, this );
				
				rThread.start();
				
				// collect garbage
				System.gc();
				}
			catch( IOException e ) {
				AppLog.info( "Connection broken on server accept." );
				ErrorHandler.report( e );
				}
			}
		}	// end run()
	
	
    static Client mainClient;
	
	/**
	 * Constructs and launches a client at launch.
	 *
	 * @param args array of arguments passed in at launch
	 */
	public static void main( String args[] ) {
		
		Thread.currentThread().setName( "Client.main" );
		LogManager.setApplicationName( "konspireClient" );
		AppLog.setMessageHandler( new GuiMessage() );
		
		randGen.setSeed( System.currentTimeMillis() );
		
		int cPort = defaultPort;
		int maxCli = defaultMaxClients;
		
		if( args.length >= 1 ) {
			// port passed in
			try{
				cPort = Integer.parseInt(args[0]);
				}
			catch( NumberFormatException e ) {
				usage();
				}
			}
		if( args.length >= 2 ) {
			// max clients passed in
			try{
				maxCli = Integer.parseInt(args[1]);
				}
			catch( NumberFormatException e ) {
				usage();
				}
			}
		
			
		mainClient = new Client( cPort, maxCli );
		
		mainClient.run();

		}
	/**
	 * Prints usage help message and exits.
	 */
	private static void usage() {				

		System.out.print("usage:\n");
		System.out.print("Client [port_number] [max_connect]\n");
		
		System.out.print("examples:\n\t");
		System.out.print("Client\n\t");
		System.out.print("Client 6433\n\t");
		System.out.print("Client 6433 15\n\t");
		
		System.exit(1);
		}
	
	// implements the download controller interface
	
	public void startDownload( FileDescriptor f ) {
		// make sure download directory exists before
		// allowing a download to happen
		File downloadDirectory =
			mSettings.mDirectoryPath.getDownloadDirectory();
		if( downloadDirectory.exists() ) {
			DownloadThread dlThread = new DownloadThread( f, 
				downloadDirectory );
			dlTable.addThread( dlThread );
			}
		else {
			GuiMessage.show( "download directory \"" 
				+ mSettings.mDirectoryPath.getDownloadDirectory().getName() 
				+ "\" does not exist." );
			}
		}
	
	/**
	 * Called by components if connection to server breaks.
	 */
	public void serverConnectionBroken() {
		AppLog.info( "connection broken to server: " + currentServer );
		serverVector.removeElement( currentServer );
		currentServer = null;
		messageSock = null;
		connected = false;
		
		srchPanel.setSearchSocket( messageSock );
		stPanel.setLiveServer( currentServer );
		
		ConnectThread cThread = new ConnectThread();
		cThread.start();
		}
	
	// should server vector be obtained from next server connected to?
	boolean serverVectorNeedsUpdate = true;
	boolean serverVectorBeenUpdated = false;
	// this lets loop through servers know whether to restart enumeration or not
	
	
	/**
	 * Tries to connect to a server on the server list.
	 * Tries all until list exhausted.
	 *
	 * @return true iff connected to a server on the list successfully
	 */
	public boolean tryConnecting() {
		shuffleVector( serverVector );
		Enumeration em = serverVector.elements();
		try {
	
			Host failedHost = localAddress;	// set to dummy address for now
			
			serverVectorNeedsUpdate = true;
			serverVectorBeenUpdated = false;
			
			while( em.hasMoreElements() ) {
				
				Host s = (Host)( em.nextElement() );
				
				if( connectToServer( s ) ) {
					currentServer = s;
					connected();
					return true;
					}
				else {	// need to refresh enumeration due to new server vector
					if( serverVectorNeedsUpdate && serverVectorBeenUpdated ) {
						em = serverVector.elements();
						serverVectorNeedsUpdate = false;
						}
					}					
				}
			AppLog.detail( "No servers left to try in server list." );
			}
		catch( NoSuchElementException ex ) {
			AppLog.error( "Iterating through server vector failed." );
			ErrorHandler.report( ex );
			}
		return false;	// no connection established
		}

	/**
	 * Tries to connect to a server.
	 *
	 * @return true iff connected to server successfully
	 */
	public boolean connectToServer( Host s ) {
		AppLog.info( "Trying to connect to server: " + s );
		try {
			// create a new message socket
			messageSock = new CheckedMessageSocket( s );
			
			// send connect request
			messageSock.send( new ClientRequestConnect( localAddress ) );
			
			// now receive server list from server
			Message m = messageSock.receive();
			
			if( m instanceof ServerList ) {
				ServerList sList = (ServerList) m;
				Host correctLocalAddress = sList.getReceiverCorrectAddress();
				if( !correctLocalAddress.equals( localAddress ) ) {
					// we retrieved or local address incorrectly
					localAddress = correctLocalAddress;
					fileListMessage.setHost( localAddress );
					AppLog.info( "Received a corrected local address from server: " 
						+ localAddress );
					}
				if( serverVectorNeedsUpdate ) {
					serverVector = sList.getHostVector();
					sListPanel.setServerList( sList );
					serverVectorBeenUpdated = true;
					
					// write new server list out to file
					sListFile.writeHostsVectorToFile( serverVector );
					}
				processStats( sList );
				}
			else {
				unexpectedServerMessage( m );
				messageSock.close();
				messageSock = null;
				return false;			
				}
			
			// next receive connection authorization from server
			m = messageSock.receive();
			if( m instanceof ConnectResponse ) {
				ConnectResponse cr = (ConnectResponse) m;
				if( cr.isAccepted() ) {
					// accepted by this server
					AppLog.info( "Accepted by server: " + s );
					
					// send our file list
					messageSock.send( fileListMessage );
					
					return true;
					}
				else {
					// rejected
					AppLog.info( "Rejected by server: " + s );
					messageSock.close();
					messageSock = null;
					return false;
					}
				}
			else {
				unexpectedServerMessage( m );
				messageSock.close();
				messageSock = null;
				return false;			
				}
			}
		catch( IOException e ) {
			AppLog.error( "Connection broken to server: " + e.toString() );
			ErrorHandler.report( e );
			messageSock = null;
			return false;
			}
		
		
		}
	
	/**
	 * Informs various components about current connection.
	 */
	private void connected() {
		srchPanel.setSearchSocket( messageSock );
		stPanel.setLiveServer( currentServer );
		connected = true;
		}
	
	/**
	 * Handles receipt of an unexpected message type.
	 *
	 * @param m message that was of an unexpected type
	 */	
	private void unexpectedServerMessage( Message m ) {
		AppLog.error( "Server responded to search with an unexpected message type: " 
			+ m.getClass().toString() );
		AppLog.info( "perhaps a new version of konspire is available.", true );
		}
	
	/**
	 * Handles receipt of a message that contains updated system stats.
	 *
	 * @param m message sent to client containing updated system stats
	 */	
	public void processStats( ToClientMessage m ) {
		stPanel.setOnlineStats( m );
		}
	
	/**
	 * Tries to connect to seed server.
	 *
	 * @param s <code>Host</code> to connect to as a seed server
	 */
	public void trySeed( Host s ) {
		AppLog.info( "Adding seed server: " + s );
		serverVector.addElement( s );
		
		if( !connected ) {
			ConnectThread cThread = new ConnectThread();
			cThread.start();
			}
		}
	
	/**
	 * Called by client handling thread whenever a client disconnects.
	 */
	public synchronized void clientDisconnected() {
		numClients--;
		}
	
	/**
	 * Checks if an open client connection slot is available and adds
	 * a client connection if it is.
	 *
	 * @return true iff a slot was available and a connection was added.
	 */
	public synchronized boolean addClientConnection() {		
		if( numClients < maxClients ) {
			// accept client connection
			numClients++;
			return true;
			}
		return false;
		}
	
	
	/**
	 * Shuffles a <code>Vector</code> by switching its starting point.
	 *
	 * @param v <code>Vector</code> to shuffle
	 */
	private void shuffleVector( Vector v ) {
		Object array[] = v.toArray();
		
		if( array.length > 0 ) {
			v.removeAllElements();
			
			int startIndex = (int)( Math.round( randGen.nextFloat() * ( array.length - 1 ) ) );
			
			for( int i=startIndex; i< array.length; i++ ) {
				v.addElement( array[i] );	
				}
			
			for( int i=0; i<startIndex; i++ ) {
				v.addElement( array[i] );
				}
			}	
		}
	
	/**
 	 * Runs cleanup code.  Should be called before all non-error exits.
	 */
	private void quitCleanup() {
		mSettings.mMainWindowSize = getSize();
		mSettings.mMainWindowLocation = getLocation();

		mSettings.mTransferWindowSize = transferFrame.getSize();
		mSettings.mTransferWindowLocation = transferFrame.getLocation();			
		
		mSettingsFile.writeToFile();
		}

	/**
	 * Listens for window close event
	 */
	private class WindowCloseListener extends DefaultWindowListener {
	
		public void windowClosing( WindowEvent e ) {
			AppLog.info( "Exiting" );
			quitCleanup();
			System.exit(0);
			}	
		}

	// these monotonically increasing ThreadIDs provide unique
	// string tags for thread messages in the log.
	
	private static long nextConnectThreadID = 0;

	/**
	 * Increments <code>nextConnectThreadID</code> in a thread-safe manner.
	 *
	 * @return the value of the ID before it was incremented.
	 */
	private synchronized static long incrementConnectThreadID() {
		long oldID = nextConnectThreadID;
		nextConnectThreadID++;
		return oldID;
		}

	private static long nextReceiveThreadID = 0;

	/**
	 * Increments <code>nextReceiveThreadID</code> in a thread-safe manner.
	 *
	 * @return the value of the ID before it was incremented.
	 */
	private synchronized static long incrementReceiveThreadID() {
		long oldID = nextReceiveThreadID;
		nextReceiveThreadID++;
		return oldID;
		}

	
	/**
	 * Thread that handles establishing a connection to a server from the server list
	 */
	class ConnectThread extends NamedErrorHandlingThread {
		
		/**
		 * Constructs a <code>ConnectThread</code>.
		 */	
		public ConnectThread() {
			super( "ConnectThread" + incrementConnectThreadID() );
			setPriority( Thread.MIN_PRIORITY );
			}
		
		// overrides ErrorHandlingThread runMethod() method
		public void runMethod() {
			boolean connectionEstablished = tryConnecting();
			if( !connectionEstablished ) {
				stPanel.setLiveServer( null );
				// empty out server list, since none are valid
				// this is done so user not confused
				// user must enter a new seed server to connect
				serverVector = new Vector();
				sListPanel.setServerList( serverVector );
				AppLog.info( "connections to all servers on server list " +
					"failed.", true );
				}
			}
		}
		
			
	/**
	 * Thread for receiving and processing initial messages from a 
	 * newly received socket connection.
	 * <p>
     * Spawns other threads as appropriate.
     */
	class ReceiveThread extends NamedErrorHandlingThread {
	
		Socket sock;
		
		// Don't change to a CheckedMessageSocket because this is a client-
		// to-client connection.
		TwoWayMessageSocket messageSock;
	
		Client parent;
		
		/**
		 * Constructs a <code>ReceiveThread</code>
		 *
		 * @param s the socket that connection has arrived on
		 * @param par the client parent of this thread
		 */
		public ReceiveThread( Socket s, Client par ) {
			super( "ReceiveThread" + incrementReceiveThreadID() );
			sock = s;
			parent = par;
			setPriority( Thread.MIN_PRIORITY );
			}
	
		// overrides ErrorHandlingThread runMethod() method
		public void runMethod() {
			
			try {
				messageSock = new TwoWayMessageSocket( sock );
				}
			catch( IOException ioe ) {
				AppLog.error( "Connection broken while receive thread " + 
					"setting up messsage socket." );
				ErrorHandler.report( ioe );
				return;
				}
			
			try {
				Message m = messageSock.receive();
				AppLog.detail( "Received message: " + m );
				
				if( m instanceof ClientFileRequest ) {
					// client wants to dl a file
					ClientFileRequest fileReq = (ClientFileRequest) m;
					
					boolean accept = addClientConnection();
					
					
					// create thread and add it to the upload table	
					UploadThread ulThread = new UploadThread( fileReq,
						accept, messageSock, parent );
					ulTable.addThread( ulThread );
					return;
					}
				// ignore other message types
				AppLog.error( "Client received an unexpected message type: " 
					+ m.getClass().toString() );
				messageSock.closeTotal();
				return;
				}
			catch( IOException ioe ) {
				AppLog.error( "Connection broken while receive thread "
					+ "receiving message." );
				ErrorHandler.report( ioe );
				return;
				}
			}	// end run()
		}	// end inner class ReceiveThread		
				
	}	// end class client

		
