// Jason Rohrer
// DownloadThread.java

/**
*
*	Thread for file download w/ status display GUI hooks
*
*	Created 5-11-2000
*	Mods:
*       Jason Rohrer    6-30-2000
*		Jason Rohrer	8-27-2000	Added JLog logging support.
*									Fixed bug that left behind 0B files when
*									initial download connections failed.
*		Jason Rohrer	8-28-2000	Fixed a bug in code that determines
*									whether IOException was caused by an abort.
*		Jason Rohrer	9-23-2000	Fixed bug in transfer abort code.  Now
*									closes socket as well as setting abort
*									flag to make sure thread will get abort
*									signal even if blocked on socket i/o.
*/

package konspire.client;

import konspire.common.Message;
import konspire.common.log.AppLog;
import konspire.common.ErrorHandler;

import konspire.common.FileDescriptor;
import konspire.common.TwoWayMessageSocket;

import java.io.File;
import java.io.*;

/**
 * TransferThread for downloads.
 *
 * @author Jason Rohrer
 */
public class DownloadThread extends TransferThread {

    File filesDir;
	
	
	/**
	 * Construct an DownloadThread.
	 *
	 * @param fDes 			descriptor of file to download
	 * @param inFilesDir 	directory where files are located
	 *
	 */	
	public DownloadThread( FileDescriptor fDes, File inFilesDir ) {
		super( "DownloadThread", fDes );
		filesDir = inFilesDir;
		}
	
	boolean aborted = false;
	
	TwoWayMessageSocket messageSock = null;
	
	// the following functions implement the TransferThread abstract functions
	
	protected void startTransfer() {
		
		
		try {
			AppLog.info( "Trying to download: " + f );
			
			// decide if partial file exists locally
			
			long fileSize = f.getSize();
			
			long bytesRemain = fileSize;
			
			// create path to where file will be put when downloaded
			String filePath = filesDir.getPath() + File.separator + f.getName();
			
			AppLog.info( "Local downloaded file location: " + filePath );
			
			// for file log
			String logFilePath = filePath + ".partial";
			
			// setup streams
			RandomAccessFile fileOutput;
			File outFile;
			
			long bytesToSkip = 0;	// how many bytes are already in partial file (if any)?
			
			// look for log file (means partial download exists)
			File logFile = new File( logFilePath );
			if( logFile.exists() ) {	// log exists
				// resume download instead of starting new
				
				// stream for log file
				DataInputStream logFileInput = new DataInputStream (
						new FileInputStream( logFilePath ) );
				
				bytesToSkip = logFileInput.readLong();
				
				// make sure partial file is valid (matches file described by
				// log)
				outFile = new File( filePath );
				if( outFile.exists() && bytesToSkip == outFile.length() ) {
					// partial file exists and matches log description
					fileOutput = new RandomAccessFile( filePath, "rw" );						
					// skip to end of partial file
					fileOutput.skipBytes( (int)bytesToSkip );
					bytesRemain -= bytesToSkip;
					}
				else {
					if( outFile.exists() ) {
						outFile.delete();
						}
					fileOutput = new RandomAccessFile( filePath, "rw" );
					}
				logFileInput.close();
				}
			else {	// no log file exists
				// delete any existing file
				outFile = new File( filePath );
				if( outFile.exists() ) {
					outFile.delete();
					}
				fileOutput = new RandomAccessFile( filePath, "rw" );
				}
			
			
			// done initing partial output file
						
			Message m = null;
			
			try {
				messageSock = new TwoWayMessageSocket( f.getHost() );
			
				// include # bytes to skip with file request
				messageSock.send( new ClientFileRequest( f, bytesToSkip ) );
			
				m = messageSock.receive();
				}
			catch( IOException inIOE ) {
				AppLog.error( "FileTransferInit message receive failed in "
					+ "download thread." );
				ErrorHandler.report( inIOE );
				// catch exceptions from these net operation separately
				// since we need to delete the file if these operations fail
				
				fileOutput.close();
				transferFailed( "failed to connect to host" );
				File thisFile = new File( filePath );
				if( thisFile.length() == 0 ) {
					thisFile.delete();
					}
				return;
				}
				
			
			if( m instanceof FileTransferInit ) {
				FileTransferInit init = (FileTransferInit) m;
				if( ! init.getAccepted() ) {
					AppLog.info( "file download rejected by host: " +
						init.getStatus(), true );
					messageSock.close();
					fileOutput.close();
					transferFailed( "file download rejected by host: " +
						init.getStatus() );
					File thisFile = new File( filePath );
					if( thisFile.length() == 0 ) {
						thisFile.delete();
						}
					return;
					}
				else {
					// transfer accepted by host
					AppLog.info( "file download accepted by host." );
					
					int blockSize = init.getBlockSize();
					byte buff[] = new byte[ blockSize ];
					
					// start receiving file
					
					DataInputStream input = messageSock.getInputStream();
					
					
					// stream for log file
					DataOutputStream logFileOutput = new DataOutputStream (
								new FileOutputStream( logFilePath ) );
					
					try {
						while( bytesRemain > blockSize ) { // full blocks remain
							if( aborted ) {
								messageSock.close();
								throw new IOException( "Download aborted by user" );
								}
							input.readFully( buff );
							fileOutput.write( buff );
							bytesRemain -= blockSize;
							setProgress( fileSize - bytesRemain );
							}
						
						// last block may be a partial block
						if( bytesRemain > 0 ) {
							if( aborted ) {
								messageSock.close();
								throw new IOException( "Download aborted by user" );
								}
							input.readFully( buff, 0, (int)bytesRemain );
							fileOutput.write( buff, 0, (int)bytesRemain );
							bytesRemain = 0;
							setProgress( fileSize );
							}
						// done downloading file
						AppLog.info( "Done downloading file: " + filePath );
						}
					catch( IOException ioe ) {
						File thisFile = new File( filePath );
						long bytesSoFar = thisFile.length();
						
						if( bytesSoFar > 0 ) {
							AppLog.warning( "partial file download: " + f.getName(),
								true );
							}
							
						ErrorHandler.report( ioe );
						
						String failedMessage = 
							"connection broken during download";
						if( aborted ) {
							failedMessage = "aborted by local user";
							}
						
						transferFailed( failedMessage );
						
						// make sure socket is closed to free up connection on sending client
						messageSock.close();
						
						// file may have been downloaded partially
						fileOutput.close();
						
						logFileOutput.writeLong( bytesSoFar );
						
						logFileOutput.close();
						
						// delete file and .partial if file size 0 ( and actual file not
						// of size 0 )
						if( bytesSoFar == 0 && f.getSize() != 0 ) {
							thisFile.delete();
							logFile.delete();
							}
						return;
						}

					fileOutput.close();
					
					// file dl-ed successfully, no need for log file
					logFileOutput.close();
					logFile.delete();
					
					messageSock.close();
					return;	
					}
				
				}
			else {
				AppLog.error( "Unexpected message type received from file host client: " 
					+ m.getClass().toString() );
				AppLog.info( "perhaps a new version of konspire is available.", true );
				
				fileOutput.close();
				transferFailed( "received an unexpected message from host" );
				File thisFile = new File( filePath );
				if( thisFile.length() == 0 ) {
					thisFile.delete();
					}
				return;
				}
			
		}
		catch( IOException e ) {
			AppLog.error( "IO exception in download thread." );
			ErrorHandler.report( e );
			transferFailed( "connection broken during download" );
			}

		return;
		}



	public void abortTransfer() {
		AppLog.info( "File download aborted." );
		aborted = true;
		if( messageSock != null ) {
			try {
				// break the socket to for thread in mid receive/send
				// to unblock and get an IOException
				messageSock.close();
				}
			catch( IOException inIOE ) {
				// do nothing but log this error, 
				// since it probably means the socket is broken anyway.
				ErrorHandler.report( inIOE );
				}
			}
		}

	}
