// Jason Rohrer
// ListTablePanel.java

/**
*
*	Abstract class that supports displaying a scrolling, clickable
*   list table with an action button.
*
*	Created 8-13-2000
*	Mods:
*		Jason Rohrer	8-16-2000	Changed add() to addElement() to support 
*									java 1.1
*		Jason Rohrer	8-27-2000	Changed gui updates so they are swing thread
*									safe.
*		Jason Rohrer	9-30-2000	Added an auxillary button panel to the
*									left of the action button and added a
*									public method for adding buttons to it.
*/

package konspire.client;


import konspire.common.Vector;
import konspire.common.log.AppLog;
import konspire.common.ErrorHandler;
import java.lang.reflect.InvocationTargetException;

import javax.swing.*;
import java.awt.FlowLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Dimension;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

/**
 * Abstract class that supports displaying a scrolling, clickable
 * list table with an action button.  Action button removes selected
 * items from the table and calls an action method.
 *
 * @author Jason Rohrer
 */
public abstract class ListTablePanel extends EtchedTitledJPanel {

	private String mTableTitle = "Abstract List Table Panel";
	
	private DownloadController parent;
	
	
 	private JScrollPane mTableListScrollPane;
 	
 	private ClickableJList mTableList = new ClickableJList(); 
 	private Vector mTableListData = new Vector();
	
	/**
	 * Button used to perform an action on selected table items.
	 */
 	private JButton mActionButton = new JButton( "Action" );
	
	// panel for action button
	private JPanel mActionButtonPanel;
	
	// panel for any auxillary action buttons, to left of mActionButton
	private JPanel mAuxillaryActionButtonPanel;
	
	/**
	 * Construct a <code>ListTablePanel</code>.
	 * 
	 * @param inRenderer renderer used to render items in <code>List</code>.
	 */
	protected ListTablePanel( ListCellRenderer inRenderer ) {
		setTitle( mTableTitle );
		
		
		mTableListScrollPane = new JScrollPane( mTableList );
		
		setLayout( new GridBagLayout() );
		GridBagConstraints gbc = new GridBagConstraints();

		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.gridwidth = 2;
		gbc.gridheight = 1;

		gbc.fill = GridBagConstraints.BOTH;
		gbc.gridx = 0;
		gbc.gridy = 0;
		gbc.weightx = 1.0;
		gbc.weighty = 1.0;
		add( mTableListScrollPane, gbc );
		
		
		mTableList.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
		
		mTableList.setCellRenderer( inRenderer );
		
		mActionButton.addActionListener( new ButtonListener() );
		mTableList.addListChoiceListener( new ListChooseListener() );
		mTableList.addListSelectionListener( new ListSelectListener() );
		
		mActionButtonPanel = new JPanel( new FlowLayout( FlowLayout.RIGHT ) );
		mActionButtonPanel.add( mActionButton );
		// button starts out disabled.  This will change
		// when some list items are selected.
		mActionButton.setEnabled( false );
		
		mAuxillaryActionButtonPanel = new JPanel( 
			new FlowLayout( FlowLayout.LEFT ) );
		
		gbc.gridwidth = 1;
		gbc.gridx = 0;
		gbc.gridy = 1;
		gbc.weightx = 0;
		gbc.weighty = 0;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		
		add( mAuxillaryActionButtonPanel, gbc );
		
		gbc.gridx = 1;
		add( mActionButtonPanel, gbc );
		}
	
	
	
	/**
	 * Adds an auxillary button to the left of the main action button.
	 *
	 * @param inButton the button to add.
	 */
	public void addAuxillaryButton( JButton inButton ) {
		mAuxillaryActionButtonPanel.add( inButton );
		}
	
	/**
	 * Sets the items displayed in the table.
	 *
	 * @param inItems a <code>Vector</code> of items to display in the table.
	 */
	protected void setTableItems( Vector inItems ) {
		mTableListData = inItems;
		mTableList.setListData( mTableListData );
		}
	
	
	
	/**
	 * Adds an item to the table.
	 *
	 * @param inItem item to add.
	 */
	protected void addTableItem( Object inItem ) {
		final Object finalItem = inItem;
		// anonymous class that wraps change to mTableList
		// so that it can be run by event dispatch thread
		Runnable setListDataWrapper = new Runnable() {
			public void run() {
				mTableListData.addElement( finalItem );
				mTableList.setListData( mTableListData );
				}
			};
		if( !SwingUtilities.isEventDispatchThread() ) {
			try {
				SwingUtilities.invokeAndWait( setListDataWrapper );
				}
			catch( InterruptedException inIE ) {
				AppLog.error( "Interrupted while waiting for event dispatch"
					+ " thread to add an element to the table." );
				ErrorHandler.report( inIE );
				}
			catch( InvocationTargetException inITE ) {
				ErrorHandler.report( inITE );
				}
			}
		else {
			setListDataWrapper.run();
			}
		}
		
	
	/**
	 * Sets the title of this table.
	 * 
	 * @param inTableTitle the title of this table.
	 */
	protected void setTableTitle( String inTableTitle ) {
		final String finalTableTitle = inTableTitle;
		// anonymous class that wraps change to title
		// so that it can be run by event dispatch thread
		Runnable setTitleWrapper = new Runnable() {
			public void run() {
				mTableTitle = finalTableTitle;
				setTitle( mTableTitle );
				}
			};
			
		if( !SwingUtilities.isEventDispatchThread() ) {
			try {
				SwingUtilities.invokeAndWait( setTitleWrapper );
				}
			catch( InterruptedException inIE ) {
				AppLog.error( "Interrupted while waiting for event dispatch"
					+ " thread set the title of this table." );
				ErrorHandler.report( inIE );
				}
			catch( InvocationTargetException inITE ) {
				ErrorHandler.report( inITE );
				}
			}
		else {
			setTitleWrapper.run();
			}
		}
	


	/**
	 * Sets the <code>String</code> displayed on the action button.
	 *
	 * @param inButtonString the text displayed on the action button.
	 */
	protected void setActionButtonText( String inButtonString ) {
		mActionButton.setText( inButtonString );
		}
	
	
	
	/**
	 * Sets the virtual key code for the key used to initiate
	 * an action (as an alternate to double clicking).
	 *
	 * @param inVirtualKeyCode virtual key code for the action key.
	 *  Valid values are constants defined in {@link #KeyEvent}.
	 */
	public void setActionKey( int inVirtualKeyCode ) {
		mTableList.setChoiceKey( inVirtualKeyCode );
		}
	
	
	/**
	 * Method called when action button is pushed (or when another
	 * user action chooses the selected items).  
	 * <p>
	 * This method must be implemented by subclasses and should perform
	 * subclass-specific operations using <code>inSelectedItems</code>. 
	 *
	 * @param inSelectedItems the items selected when the user choice
	 *  action occurred.
	 */
	protected abstract void action( java.util.Vector inSelectedItems );
	
	
	
	// implements ActionListener interface
	private class ButtonListener implements ActionListener {
		
		public void actionPerformed( ActionEvent inAE ) {
			if( inAE.getSource() == mActionButton ) {
				Object[] selectedObjects = mTableList.getSelectedValues();
				int numItems = selectedObjects.length;
				
				Vector selectedItemsVector = new Vector( numItems );
				
				for( int i=0; i<numItems; i++ ) {
					selectedItemsVector.addElement( selectedObjects[i] );
					mTableListData.removeElement( selectedObjects[i] );
					}
				if( numItems > 0 ) {
					//make sure resetting list data is worth it
					mTableList.setListData( mTableListData );
					}
				// call subclass callback
				action( selectedItemsVector );
				}
			}
		}
		
	// implements ListChoiceListener interface
	private class ListChooseListener implements ListChoiceListener {
		
		public void listItemChosen( ListChoiceEvent inEvent ) {
				
			java.util.Vector selectedItemsVector = inEvent.getChosenItems();
			int numItems = selectedItemsVector.size();
			
			for( int i=0; i<numItems; i++ ) {
				mTableListData.removeElement( selectedItemsVector.elementAt(i) );
				}
			if( numItems > 0 ) {
				//make sure resetting list data is worth it
				mTableList.setListData( mTableListData );
				}
			// call subclass callback
			action( selectedItemsVector );
			}
		}
	
	
	
	// implements ListSelectionListener interface
	private class ListSelectListener implements ListSelectionListener {
		
		public void valueChanged( ListSelectionEvent inLSE ) {
			// enable action button only if some items are selected
			boolean buttonEnabled = 
				!( mTableList.getSelectedIndex() == -1 );

			mActionButton.setEnabled( buttonEnabled );
			}
		}
	
	}
