// Jason Rohrer
// ScrollTablePanel.java

/**
*
*	GUI component that contains a scrolling multi-column table of GUI components
*
*	Created 5-9-2000
*	Mods:
*		Jason Rohrer	6-23-2000	Fixed action listener with a hack so that Java 1.3 will 
*									forward click events to buttons in the table.
*									The problem seems to be that the buttons are returning
*									(0,0) for their dimensions when they're in the table.  Thus
*									any mouse event is out of bounds.  The hack involves triggering
*									the very general "doClick()" on the button directly.
*									This works, but it doesn't look very nice from the user standpoint,
*									because the button doesn't look depressed if the user clicks
*									and holds down on the button....  More work needed here.
*/

package konspire.client;

import javax.swing.*;

import javax.swing.table.*;

import javax.swing.table.DefaultTableModel;

import java.awt.Component;
import java.awt.FlowLayout;	

import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;

import java.awt.event.ActionEvent;

import konspire.common.Vector;

/**
 * A panel containing a scrolling JTable with support for containing JButtons
 * and other components.
 *
 * @author Jason Rohrer
 */
public class ScrollTablePanel extends EtchedTitledJPanel {

	private String title;
	private String columnTitles[];
	private Class columnClasses[];

	Vector rows = new Vector();		// vector of TableRow wrappers in same order as table rows

	int numCols;

	protected JTable table;
	private GUITableModel tableModel;
	
	private JScrollPane tableScrollPane;
	
	private boolean beenSetup = false;
	
	/**
	 * Constructs a <code>ScrollTablePanel</code> with default values.
	 *
	 */
	protected ScrollTablePanel() {
		// constructor called by subclasses when custom setup needed
		}

	/**
	 * Constructs a <code>ScrollTablePanel</code> specifying values.
	 *
	 * @param titl title to be displayed on panel
	 * @param colTitles array of titles for columns in the table
	 * @param colClasses array of types that will be contained in each column
	 */	
	public ScrollTablePanel( String titl, String[] colTitles, Class[] colClasses ) {
		setup( titl, colTitles, colClasses );
		}
	
	/**
	 * Sets up the panel.  Shared by multiple constructors.
	 *
	 * @param titl title to be displayed on panel
	 * @param colTitles array of titles for columns in the table
	 * @param colClasses array of types that will be contained in each column
	 */
	protected void setup( String titl, String[] colTitles, Class[] colClasses ) {
		
		title = titl;
		setTitle( title );
		
		columnTitles = colTitles;
		columnClasses = colClasses;
		
		numCols = columnTitles.length;
		
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	
		
		tableModel = new GUITableModel( columnTitles, 0); 
		table = new JTable( tableModel );
		
		TableCellRenderer defaultRenderer = table.getDefaultRenderer( Component.class );
		JTableGUIRenderer guiRenderer = new JTableGUIRenderer(defaultRenderer);
		
		table.setDefaultRenderer( Component.class, guiRenderer );
		
		table.setCellSelectionEnabled( false );
		table.setRowSelectionAllowed( false );
		table.setColumnSelectionAllowed( false );
		
		JTableGUIMouseListener tableListener = new JTableGUIMouseListener();
		
		table.addMouseListener( tableListener );
		
		tableScrollPane = new JScrollPane( table );
		
		add( tableScrollPane );
		
		beenSetup = true;
		}
	
	
	/** 
	 * Adds a row of objects to table.
	 * Object types must match with fixed classes defined for each table column.
	 * 
	 * @param oArray array of objects to add to table
	 *
	 * @return a TableRow wrapper corresponding to this row (for row removal)
	 */
	public TableRow addRow( Object [] oArray ) {
		Object b = oArray[0];
		// make sure class order matches fixed class order
		int maxInd = oArray.length;
		if( maxInd > numCols ) {
			maxInd = numCols;
			}
		for( int i=0; i<maxInd; i++ ) {
			if( (oArray[i]).getClass() != columnClasses[i] ) {
				System.out.println("Object array can't be added to table:  Type mismatch with fixed column types.");
				return null;
				}
			}
		TableRow rowWrapper = new TableRow( oArray );
		rows.addElement( rowWrapper );
			
		// else all classes match
		tableModel.addRow( oArray );
		//return tableModel.getRowCount() - 1;
		return rowWrapper;
		}
	
	/**
	 * Removes a row from table.
	 *
	 * @param rowWrapper row to remove
	 *
	 * @return true iff row existed in table and was removed successfully
	 */
	public boolean removeRow( TableRow rowWrapper ) {
		if( rows.contains( rowWrapper ) ) {
			int index = rows.indexOf( rowWrapper ); 
			
			tableModel.removeRow( index );
			rows.removeElementAt( index );
			
			return true;
			}
		return false;
		}
	
	
	/**
	 * Removes all rows from table.
	 */
	public void removeAllRows() {
		int lastRow = tableModel.getRowCount() -1;
		for( int i=lastRow; i>=0; i-- ) {
			tableModel.removeRow( i );
			}
		rows.removeAllElements();
		}
	
	
	/**
	 * Table model for tables with embedded gui components
	 */
	private class GUITableModel extends DefaultTableModel {
		 
		 public GUITableModel(Object[] columnNames, int numRows) {
			super(columnNames, numRows);
			} 
		 
		 public Class getColumnClass(int c) {
		 	// FIXED 6-13-2000
		 //	return getValueAt(0, c).getClass();
			if( beenSetup ) {
				return columnClasses[ c ];
				}
			else {
				return Object.class;
				}
		
			}
		
		public boolean isCellEditable(int row, int column) {
			return false;
			}	
		}
		
	/**
	 * Cell renderer for tables with embedded gui components
	 */
	private class JTableGUIRenderer implements TableCellRenderer {
		private TableCellRenderer defaultRenderer;

		public JTableGUIRenderer(TableCellRenderer renderer) {
			defaultRenderer = renderer;
			}

		public Component getTableCellRendererComponent(JTable table, Object value,
							 boolean isSelected,
							 boolean hasFocus,
							 int row, int column) {
			if(value instanceof Component) {
				return (Component)value;
				}
			return defaultRenderer.getTableCellRendererComponent( 
					table, value, isSelected, hasFocus, row, column);
			}
		}
		
	/**
	 * Mouse listener for tables with embedded gui components
	 */	
	class JTableGUIMouseListener implements MouseListener {
		
		private void passEventToComponent( MouseEvent e ) {
						
			TableColumnModel columnModel = table.getColumnModel();
			int column = columnModel.getColumnIndexAtX(e.getX());
			int row    = e.getY() / table.getRowHeight();
			Object value;
			Component comp;
			MouseEvent compEvent;

			if(row >= table.getRowCount() || row < 0 ||
			   column >= table.getColumnCount() || column < 0) {
				return;
				}
			
			value = table.getValueAt(row, column);
			
			
			if( !(value instanceof Component) ) {
			  return;
			  }
		  	
		  	comp = (Component)value;
		  	
		  	if( comp instanceof JButton ) {
				// special for a JButton, since not working in Java 1.3

				JButton thisButton = (JButton) comp;
				
				compEvent =
			  		(MouseEvent)SwingUtilities.convertMouseEvent( table, e, thisButton );
				if( compEvent.getX() < 0 || compEvent.getX() > thisButton.getWidth() ||
					compEvent.getY() < 0 || compEvent.getY() > thisButton.getHeight() ) {
					// We must be in Java 1.3, where convertMouseEvent is broken!
					// actually, it seems to be related to the fact that the button 
					// is returning (0,0) as its dimensions, which breaks convertMouseEvent
					if( mouseClicked ) {
						thisButton.doClick();
						}
					}
				else {
					// probably not in java 1.3, since it seems to be working normally!
					thisButton.dispatchEvent(compEvent);
					}
				}
			else {
				// other components
				compEvent =
			  		(MouseEvent)SwingUtilities.convertMouseEvent( table, e, comp );
				comp.dispatchEvent(compEvent);
				}

			table.repaint();
			}
		
		boolean mouseClicked = false;
		
		public void mouseClicked(MouseEvent e) {
			mouseClicked = true;
			passEventToComponent(e);
			}

		public void mouseEntered(MouseEvent e) {
			mouseClicked = false;
			passEventToComponent(e);
			}

		public void mouseExited(MouseEvent e) {
			mouseClicked = false;
			passEventToComponent(e);
			}

		public void mousePressed(MouseEvent e) {
			mouseClicked = false;
			passEventToComponent(e);
			}

		public void mouseReleased(MouseEvent e) {
			mouseClicked = false;
			passEventToComponent(e);
			}

		}
	
	}