// Jason Rohrer
// UploadThread.java

/**
*
*	Thread for file upload w/ status display GUI hooks
*
*	Created 5-11-2000
*	Mods:
*		Jason Rohrer	6-2-2000	Added support for sending files from subdirectories.
*		Jason Rohrer	8-27-2000	Added JLog logging support.
*		Jason Rohrer	9-23-2000	Updated abort transfer code to mirror
*									behavior of DownloadThread (even though
*									not entirely necessary).  Now, socket
*									is closed immediately on abort to force
*									a thread blocked on socket IO to
*									return from the blocking call.
*									Thus, the abort operation should
*									happen immediately instead of at the
*									end of the current block of the transfer
*									data.
*		Jason Rohrer	9-30-2000	Removed member filesDir because it is no
*									longer being used.  This causes constructor
*									to change.
*/

package konspire.client;

import konspire.common.Message;
import konspire.common.log.AppLog;
import konspire.common.ErrorHandler;

import konspire.common.FileDescriptor;
import konspire.common.TwoWayMessageSocket;

import java.io.File;
import java.io.*;


/**
 * TransferThread for uploads.
 *
 * @author Jason Rohrer
 */
public class UploadThread extends TransferThread {
	
	int blockSize = 20000;	// 20 KB blocks
	
	long bytesToSkip = 0;
	
	boolean accept;
	
	TwoWayMessageSocket messageSock;
	
	Client parent;

	/**
	 * Construct an UploadThread.
	 *
	 * @param req 			file request thread will handle
	 * @param acc 			should upload request be accepted
	 * @param sock 			socket for connection
	 * @param par			parent client of upload thread
	 *
	 */	
	public UploadThread( ClientFileRequest req, boolean acc, TwoWayMessageSocket sock, Client par ) {
		super( "UploadThread", req.getFileDescriptor() );
		
		bytesToSkip = req.getBytesToSkip();
		
		accept = acc;
		
		messageSock = sock;
		
		parent = par;
		}
	
	boolean aborted = false;
	
	
	// the following functions implement the TransferThread abstract functions
	
	
	// includes sending acceptance message.	
	protected void startTransfer() {
		
		
		// default status to send if download rejected
		String status = "no free download slots.";
		
		
		// first, make sure file still exists
		String filePath = f.getPath();
		AppLog.info( "received request for file upload: " + filePath );
		
		File fileToSend = new File( filePath );
		
		if( !fileToSend.exists() || fileToSend.length() != f.getSize() ) {
			// file has changed or no longer exists
			status = "file no longer exists";
			accept = false;
			}
		
		if( !accept ) {	
			AppLog.info( "rejecting upload request: " + status );
			try {
				// reject this request
				FileTransferInit init = new FileTransferInit( accept, status, blockSize );
				
				messageSock.send( init );
				messageSock.close();
				}
			catch( IOException e ) {
				AppLog.error( "Connection broken while sending reject message "
				 	+ "to client." );
				ErrorHandler.report( e );
				}
			transferFailed( "rejected: " + status );
			return;
			}
			
		
		try {
			AppLog.info( "accepting upload request" );
			
			long fileSize = f.getSize();
			
			FileTransferInit init = new FileTransferInit( accept, "", blockSize );
				
			messageSock.send( init );
			
			
			// now start sending file
			
			byte buff[] = new byte[ blockSize ];
			
			// setup streams
			RandomAccessFile fileInput = new RandomAccessFile( filePath, "r" );
			fileInput.skipBytes( (int)bytesToSkip );
			long bytesRemain = fileSize - bytesToSkip;

			DataOutputStream output = messageSock.getOutputStream();			
			
						
			while( bytesRemain > blockSize ) { // full blocks remain
				if( aborted ) {
					messageSock.close();
					fileInput.close();
					parent.clientDisconnected();
					transferFailed( "aborted by local user" );
					return;
					}
				fileInput.readFully( buff );
				output.write( buff );

				bytesRemain -= blockSize;
				setProgress( fileSize - bytesRemain );
				}
			
			// last block may be a partial block
			if( bytesRemain > 0 ) {
				if( aborted ) {
					messageSock.close();
					fileInput.close();
					parent.clientDisconnected();
					return;
					}
				fileInput.readFully( buff, 0, (int)bytesRemain );
				output.write( buff, 0, (int)bytesRemain );
				bytesRemain = 0;
				setProgress( fileSize );
				}
			
			// done uploading
			messageSock.close();
			fileInput.close();
			parent.clientDisconnected();
			AppLog.info( "Done uploading file: " + filePath );
			return;
			
			}
		catch( IOException e ) {
			AppLog.error( "Connection broken while uploading file." );
			ErrorHandler.report( e );
			transferFailed( "connection broken during upload" );
			}
			
		parent.clientDisconnected();
		return;
		}


	public void abortTransfer() {
		AppLog.info( "File upload aborted." );
		aborted = true;
		if( messageSock != null ) {
			try {
				// break the socket to for thread in mid receive/send
				// to unblock and get an IOException
				messageSock.close();
				}
			catch( IOException inIOE ) {
				// do nothing but log this error, 
				// since it probably means the socket is broken anyway.
				ErrorHandler.report( inIOE );
				}
			}
		}

	}
