// Jason Rohrer
// BaseTwoUnitsPrefixFormat.java

/**
*
*	Scales values for units that use base two for orders of magnitude.
*	E.g., 2048 B would be converted to 2 KiB.
*
*	Created 7-3-2000
*	Mods:
*		Jason Rohrer	7-15-2000	Changed to IEEE proposal for base-2 order of magnitude prefixes
*		Jason Rohrer	7-31-2000	Fixed an error in the prefixes
*/

package konspire.common;

import java.text.NumberFormat;

/**
 * Converts values and units to values in units with order of
 * magnitude prefixes (for units that use base-2 for orders of magnitude).
 * Conforms to the IEEE proposal for base-2 unit prefixes.
 *
 */
public class BaseTwoUnitsPrefixFormat implements UnitsPrefixFormat {

	/**
	 * Convert a value and unit to a scaled value and prefixed unit.
	 *
	 * @param inValue the value to convert, in base units
	 * @param inUnits string representing the base units
	 *
	 * @returns String representing scaled value with units
	 */
	public String convertUnitsPrefix( double inValue, String inUnits ) {
		
		// only go up to next OOM if value has bypassed 2000 of next smalles OOM units
		//  E.g., 1500 B stays as 1500 B
		//        2048 B is converted to 2 KiB
		
		String unitsString = inUnits;
		
		if( inValue >= 2048 ) {
			unitsString = "Ki" + inUnits;
			inValue = inValue / 1024;
			} 
			
		if( inValue >= 2048 ) {
			unitsString = "Mi" + inUnits;
			inValue = inValue / 1024;
			}
			
		if( inValue >= 2048 ) {
			unitsString = "Gi" + inUnits;
			inValue = inValue / 1024;
			}
			
		if( inValue >= 2048 ) {
			unitsString = "Ti" + inUnits;
			inValue = inValue / 1024;
			}
	
		if( inValue >= 2048 ) {
			unitsString = "Pi" + inUnits;
			inValue = inValue / 1024;
			}
		
		if( inValue >= 2048 ) {
			unitsString = "Ei" + inUnits;
			inValue = inValue / 1024;
			}
		
		if( inValue >= 2048 ) {
			unitsString = "Zi" + inUnits;
			inValue = inValue / 1024;
			}
			
		if( inValue >= 2048 ) {
			unitsString = "Yi" + inUnits;
			inValue = inValue / 1024;
			}
		
		NumberFormat numFormat = NumberFormat.getNumberInstance();
		
		numFormat.setMaximumFractionDigits( 2 );
		
		String scaledValueWithUnits = numFormat.format( inValue ) + 
								" " + unitsString;
		
		return scaledValueWithUnits;
		}
		
		
	/**
	 *	This class conforms to the IEEE proposal for base-2 prefixes:
	 *
	 *	Factor   	Name	Symbol	Origin 					Derivation  
	 *	2^10 		kibi 	Ki		kilobinary: (2^10)^1 	kilo: 	(10^3)^1 
		2^20 		mebi	Mi		megabinary: (2^10)^2    mega: 	(10^3)^2  
		2^30 		gibi	Gi		gigabinary: (2^10)^3 	giga: 	(10^3)^3 
		2^40 		tebi	Ti		terabinary: (2^10)^4 	tera: 	(10^3)^4 
		2^50 		pebi	Pi		petabinary: (2^10)^5 	peta: 	(10^3)^5 
		2^60 		exbi	Ei		exabinary: 	(2^10)^6 	exa: 	(10^3)^6
		2^70 		zebi	Zi		zettabinary: (2^10)^7 	zeta: 	(10^3)^7
		2^80 		yobi	Yi		yottabinary: (2^10)^8 	iota: 	(10^3)^8
	
	 */	
	
	}