// Jason Rohrer
// CheckedMessageSocket.java

/**
*
*	Message socket that sends pings to make sure other host is alive.
*
*	Created 9-22-2000
*	Mods:
*		Jason Rohrer	10-1-2000	Fixed a bug in the ping detection mechanism
*									in the receive method.
*/

package konspire.common;


import java.io.*;
import java.net.*;

import konspire.common.log.AppLog;

/**
 * TwoWayMessage socket that periodically sends and receives pings to make 
 * sure other host is alive.
 * Also supports timeouts on socket initialization and sending.
 *
 * NOTE:  The <code>Socket</code> inside a <code>CheckedMessageSocket</code>
 * should only be used via <code>CheckedMessageSocket</code> methods.  The
 * socket should not be retrieved and manipulated directly.  The synchronization
 * provided by the <code>CheckedMessageSocket</code> methods ensures that 
 * sends and receives of messages are atomic operations (thus, pings get
 * interleaved with message sends and receives).  If the <code>Socket</code>
 * is accessed directly to send data bytes, behavior is undefined.
 *
 * Note that this is different from the usage guidelines for the superclass
 * TwoWayMessageSocket, where retrieving the underlying <code>Socket</code>
 * and writing to it is permissable.
 *
 * Thus, <code>CheckedMessageSocket</code>s can only be used for <code>
 * Message</code>s and can't be used to send byte arrays of binary data.
 *
 * @author Jason Rohrer
 */
public class CheckedMessageSocket extends TwoWayMessageSocket {
	
	/**
	 * How long to wait between sending pings to the other host
	 */
	private int mPingTime = 60 * 1000;
	
	
	/**
	 * How long to wait between receiving pings before deducing that 
	 * other host is dead.
	 */
	private int mPingWaitTimeout = 2 * mPingTime;
	
	
	/**
	 * Thread that sends pings to other host.
	 */
	private PingSendingThread mPingThread;
	
	
	
	/**
	 * Constructs a <code>CheckedMessageSocket</code>.
	 *
	 * @param inDestination <code>Host</code> to establish this socket with.
	 *
	 * @throws IOException if socket fails to initialize (or timeout reached).
	 */
	public CheckedMessageSocket( Host inDestination ) 
		throws IOException {
		
		super( inDestination );
		
		restartPingThread();
		
		//this( inDestination, -1 );
		}
	
	
	/**
	 * Constructs a <code>CheckedMessageSocket</code>.
	 *
	 * @param inDestination <code>Host</code> to establish this socket with.
	 *
	 * @param inTimeout timeout for initialization in milliseconds.  Passing in
	 *   -1 specifies no timeout.
	 *
	 * @throws IOException if socket fails to initialize (or timeout reached).
	 */
/*	public CheckedMessageSocket( Host inDestination, long inTimeout ) 
		throws IOException {
		
		// thread for setup timeout
		if( inTimeout >= 0 ) {
			TimerThread timeThread = new TimerThread( Thread.currentThread() );
			timeThread.setTime( inTimeout );
			timeThread.start();
			}
		
		try {
			setupSocket( inDestination );
			}
		catch( InterruptedException inIE ) {
			// timer thread interrupted us because of timeout
			throw new IOException( "Socket broken during initialization due " +
				"to timeout." );
			}
		
		restartPingThread();
		}
*/
		
	
	/**
	 * Constructs a <code>CheckedMessageSocket</code> from an existing
	 * established socket connection.
	 * <p>
	 * Note that socket connection WILL NOT be closed by <@link close>.<BR>
	 * Use <@link closeTotal> to close the passed in socket along with the <code>TwoWayMessageSocket</code>.
	 *
	 * @param s an already established <code>Socket</code> connection to attach
	 *  this <code>CheckedMessageSocket</code> to.
	 *
	 * @throws IOException if attaching to socket fails
	 */
	public CheckedMessageSocket( Socket s )
		throws IOException {
		
		super( s );
		
		restartPingThread();
		}
	
	
	/**
	 * Creates a new ping thread and starts it.
	 */
	protected void restartPingThread() {
		mPingThread = new PingSendingThread( mPingTime );
		mPingThread.start();
		}
	
	/**
	 * Stops the current ping thread permanently.
	 */
	protected void stopPingThread() {
		mPingThread.stopSending();
		} 
	 
	 
	// Overrides TwoWayMessageSocket send(Message) method
	public void send( Message m ) throws IOException { 
		synchedSend( m );
		}
		
	
	// Overrides TwoWayMessageSocket receive() method	
	public Message receive() throws IOException {
		
		// should not be sending pings while waiting to receive
		stopPingThread();
		
		try {
		
			Message m = super.receive( mPingWaitTimeout );
			
			// filter out ping messages and reset timer thread
			while( m instanceof Ping ) {
				//System.out.println( "Ping Received." );
				
				m = super.receive( mPingWaitTimeout );
				}
				
			// done receiving, so restart ping thread
			restartPingThread();
				
			// a non-ping message was received.  
			// this is the message we want to return.	
			return m;	
			}
		catch( InterruptedIOException inIE ) {
			// One of our receives timed out.
			// This means other host didn't send ping in time.
			// Assume the other host is dead.
			
			closeTotal();
			throw new IOException( "Socket broken during receive due " +
				"to timeout while wating for ping from other host." );
			}
		
		}
	
	
	// Overrides TwoWayMessageSocket close() method
	public void close() throws IOException {
		stopPingThread();
		super.close();
		}
	
	
	// Overrides TwoWayMessageSocket closeTotal() method
	public void closeTotal() throws IOException {
		stopPingThread();
		super.closeTotal();
		}
	
	 
	/**
	 * Send used internally to actually send a message through
	 * the <code>Socket</code>.  Synchronized to make sure that this 
	 * operation is safe for the multiple threads associated with this socket.
	 *
	 * @param m the <code>Message</code> to send
	 *
	 * @throws IOException if send fails
	 */
	synchronized private void synchedSend( Message m ) throws IOException {
		super.send( m );
	}
	 
	 
	
	/**
	 * Thread that sleeps for a specified time and then interrupts its
	 * parent thread.
	 */
	private class TimerThread extends Thread {
		
		private Thread mParentThread;
		
		// time in milliseconds to wait before interrupting parent
		private long mTime;
		
		// true iff new time has been set and thread hasn't retrieved it yet
		private boolean mNewTimeSet = false;
		
		/**
		 * Constructs a <code>TimerThread</code>.
		 *
		 * @param inParent the parent to interrupt if time runs out.
		 */
		public TimerThread( Thread inParentThread ) {
			mParentThread = inParentThread;
			}
		
		/**
		 * Resets timeout for this thread.
		 *
		 * @param inNewTime thread will now wait <code>inNewTime</code>
		 *   milliseconds before interrupting parent thread.
		 */
		synchronized public void setTime( long inNewTime ) {
			mTime = inNewTime;
			mNewTimeSet = true;
			this.interrupt();
			}
		
		/**
		 * Gets time last set for this thread to wait before interrupting parent.
		 *
		 * @return last milliseconds set for thread to wait before interrupting 
		 *   parent thread.
		 */
		synchronized public long getTime() {
			mNewTimeSet = false;
			return mTime;
			}
		
		/**
		 * Gets whether a new time has been set that this thread hasn't
		 * retrieved yet.
		 *
		 * @return true if there is a new time set that this thread hasn't
		 *   retrieved.
		 */
		synchronized public boolean isNewTime() {
			return mNewTimeSet;
			}
		
		
		// Overrides Thread run() method
		public void run() {
			while( isNewTime() ) {
				try {
					sleep( getTime() );
					}
				catch( InterruptedException inIE ) {
					// interrupted, probably by a setTime() call.
					// while loop condition will handle verifying this
					}
				}
			
			// time ran out before a new time was set
			// interrupt the parent thread
			mParentThread.interrupt();
			}
		}
	
	
	/**
	 * <code>Thread</code> that sends pings periodically on this
	 * socket to let the other host know we're alive.
	 */
	private class PingSendingThread extends Thread {
	
		long mTimeBetweenPings;
		
		// true if thread should stop sending pings
		// and return from run method.
		boolean mShouldStop = false;
		
		/**
		 * Constructs a <code>PingSendingThread</code>.
		 *
		 * @param inTimeBetweenPings the time to wait between sending pings.
		 */
		public PingSendingThread( long inTimeBetweenPings ) {
			mTimeBetweenPings = inTimeBetweenPings;
			}
		
		/**
 		 * Tells this thread that it should stop sending pings
		 * and return from run method.
		 */
		synchronized public void stopSending() {
			mShouldStop = true;
			}
		
		/**
 		 * Gets whether this thread has been stopped.
		 *
		 * @return true iff this thread has been stopped.
		 */
		synchronized public boolean isSendingStopped() {
			return mShouldStop;
			}
			
		
		public void run() {
			try {
				sleep( mTimeBetweenPings );
				while( !isSendingStopped() ) {
					send( new Ping() );
					sleep( mTimeBetweenPings );
					}
				}
			catch( InterruptedException inIE ) {
				// this ping thread interrupted while sleeping
				}
			catch( IOException inIOE ) {
				// socket broken during send of ping
				// ignore this for now, since
				// other parent class method calls will 
				// notice this broken socket as well.
				}
			}
		}
	
		
	}
