// Jason Rohrer
// FileHost.java

/**
*
*	Description of a network host that has files
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	7-23-2000	Made scanDir( File, Vector ) method static.
*		Jason Rohrer	8-14-2000	Added a server member variable to allow
*									clients to be removed from database when
*									their server goes down.
*		Jason Rohrer	8-20-2000	Added a toString() method.
*		Jason Rohrer	9-30-2000	Changed so that files are scanned from a
*									FileDirectoryPath instead of just from 
*									a single directory File.
*/

package konspire.common;

import konspire.common.Vector;

import konspire.common.FileDirectoryPath;

import java.io.File;

/**
 * Message containing a description of a network host that is sharing files.
 * <p>
 * Contains the host address and a collection of <code>FileDescriptors</code>,
 * one for each file that the host is sharing.
 * <p>
 * Includes functionality for building the collection of <code>FileDescriptors</code>
 * from a shared files directory with subdirectories.
 *
 * @author Jason Rohrer
 */
public class FileHost implements Message {
	
	
	
	/**
	 * Address of this <code>FileHost</code>.
	 */
	protected Host address;
	
	
	
	/**
	 * Server that this <code>FileHost</code> is connected to.
	 */
	protected Host mServer = null;
	
	
	
	/**
	 * Vector of <code>FileDescriptors</code> for files on this host.
	 */
	protected Vector fileVector;
	
	
	// true iff the download directory exists
	// Doesn't depend on whether the upload directories exist or not.
	private boolean dirExists = false;
	
	
	
	/**
	 * Constructs a <code>FileHost</code from host address and <code>Vector</code> 
	 * of <code>FileDescriptor</code>s.
	 *
	 * @param addr the address of this host
	 * @param fVector a <code>Vector</code> of <code>FileDescriptor</code>s 
	 * describing files shared by this host.
	 */
	public FileHost( Host addr, Vector fVector ) {
		address = addr;
		fileVector = fVector;
		}
	
	
	
	/**
	 * Constructs a <code>FileHost</code> from a host address and a shared file directory.
	 *
	 * @param addr the address of this host
	 * @param inFilesPath the directory path (with subdirectories) 
	 *   on this host where shared files reside.
	 */
	public FileHost( Host addr, FileDirectoryPath inFilesPath ) {
		address = addr;
		rescanFileList( inFilesPath );
		}
	
	
	
	/**
	 * Rescans the file list on this host given a shared file directory.
	 * <p>
	 * Note that this method only works if it's called on this host.
	 *
	 * @param inFilesPath the directory path (with subdirectories) 
	 *   on this host where shared files reside.
	 */
	public void rescanFileList( FileDirectoryPath inFilesPath ) {
		fileVector = new Vector();

		// vector of directories to scan
		Vector directoryVector = new Vector();
		directoryVector.addElement( inFilesPath.getDownloadDirectory() );
		directoryVector.addAll( inFilesPath.getUploadDirectories() );
		
		// make sure download directory exists
		dirExists = inFilesPath.getDownloadDirectory().exists();
		
		int numDirectories = directoryVector.size();
		
		for( int i=0; i<numDirectories; i++ ) {
			File filesDir = (File)( directoryVector.elementAt( i ) );
		
			if( filesDir.exists() ) {
				// add descriptors for all files in directory and subdirectories to fileVector
				scanDir( address, filesDir, fileVector );
				}
			} 
		}
	
	
	
	/**
	 * Gets whether the download directory specified by the file
	 * directory path exists on this host.
	 *
	 * @return true iff the download directory on this host exists
	 */
	public boolean downloadDirectoryExists() {
		return dirExists;
		}
	
	
	
	/**
	 * Gets the address of this host.
	 *
	 * @return the address of this host
	 */
	public Host getHost() {
		return address;
		}
	
	
	
	/**
	 * Replaces the host address for all <code>FileDescriptors</code> in this message.
	 *
	 * @param addr the new address to be attached to all <code>FileDescriptors</code> in this message
	 */
	public void setHost( Host addr ) {
		address = addr;
		
		int numFiles = fileVector.size();
		for( int i=0; i<numFiles; i++ ) {
			FileDescriptor fDes = (FileDescriptor)( fileVector.elementAt(i) );
			fDes.setHost( address ); 
			}
		}
	
	
	
	/**
	 * Gets the server that this <code>FileHost</code> is connnected to.
	 *
	 * @return the server that this host is connected to.
	 */
	public Host getServer() {
		return mServer;
		}
	
	
	
	/**
	 * Sets the server that this <code>FileHost</code> is connnected to.
	 *
	 * @param inServer the server that this host is connected to.
	 */
	public void setServer( Host inServer ) {
		mServer = inServer;
		}
	
	
	
	/**
	 * Gets the <code>Vector</code> of <code>FileDescriptors</code> contained in this message.
	 *
	 * @return the <code>Vector</code> of <code>FileDescriptors</code> contained in this message
	 */
	public Vector getFileVector() {
		return fileVector;
		}
	
	
	
	/**
	 * Sets the <code>Vector</code> of <code>FileDescriptors</code> contained in this message.
	 *
	 * @param fVector the <code>Vector</code> of <code>FileDescriptors</code> to be 
	 * contained in this message
	 */	
	public void setFileVector( Vector fVector ) {
		fileVector = fVector;
		}



	// overrides the Object equals( Object ) method
	public boolean equals( Object o ) {
		if( o instanceof FileHost ) {
			FileHost other = (FileHost) o;
			return address.equals( other.getHost() );
			}
		return false;
		}
	
	
	
	/**
	 * Recursively scans a directory, adding all files to a <code>Vector</code> 
	 * as <code>FileDescriptor</code>s.
	 *
	 * @param inAddress address of host that is serving these files
	 * @param dirFile the directory to scan for files
	 * @param destVector the <code>Vector</code> to add <code>FileDescriptor</code>s to.
	 */
	public static void scanDir( Host inAddress, File dirFile, Vector destVector ) {	
		String [] fileList;
		fileList = dirFile.list();
		for( int i=0; i<fileList.length; i++ ) {
			// create file descriptor for file
			File f = new File( dirFile.getPath() + File.separator + fileList[i] );
			
			if( f.isDirectory() ) {
				scanDir( inAddress, f, destVector );
				}
			else {
				FileDescriptor fDes = new FileDescriptor( inAddress, fileList[i], f.getPath(), f.length() );
				destVector.addElement( fDes );
				}
			}
		}
		
		
	// overrides the Object toString() method
	public String toString() {
		int vectorSize = 0;
		if( fileVector != null ) {
			vectorSize = fileVector.size(); 
			}
		return "FileHost: " + address
				+ "\n\tServer: " + mServer
				+ "\n\tNumFiles: " + vectorSize;
		}
			
	}	
