// Jason Rohrer
// NamedErrorHandlingThread.java

/**
*
*	Thread class that catches errors thrown by the run method and passes
*	them to the error handling framework.  Thread names are also enforced.
*
*	Created 8-23-2000
*	Mods:
*/

package konspire.common;


/**
 * Thread class that catches <code>Throwable</code>s thrown by the run
 * methods of all subclasses.  <code>Throwables</code> are reported to the
 * error handling framework.  All subclassing <code>Thread</code>s are 
 * required to pass in a name for themselves.
 * <p>
 * Note that all constructors require a non-null name to be passed in.
 * Thus, other <code>Thread</code> constructors are not overridden.
 *
 * @author Jason Rohrer
 */
public abstract class NamedErrorHandlingThread extends ErrorHandlingThread {

	
	
	/**
	 * Constructs a <code>NamedErrorHandlingThread</code>.
	 *
	 * @param inName name of this <code>Thread</code>. Must
	 *  be non-null.
	 */
	protected NamedErrorHandlingThread( String inName ) {
		super( checkName( inName ) );
		}
	
	
	
	/**
	 * Constructs a <code>NamedErrorHandlingThread</code>, specifying
	 * a <code>Runnable</code>.
	 *
	 * @param inRunnable the runnable to be run by this <code>Thread</code>.
	 * @param inName name of this <code>Thread</code>. Must
	 *  be non-null.
	 */
	protected NamedErrorHandlingThread( Runnable inRunnable, String inName ) {
		super( inRunnable, checkName( inName ) );
		}
	
	
	
	/**
	 * Constructs a <code>NamedErrorHandlingThread</code>, specifying
	 * a <code>ThreadGroup</code>.
	 *
	 * @param inGroup the <code>ThreadGroup</code> for this <code>Thread</code>.
	 * @param inRunnable the <code>Runnable</code> to be run by this <code>Thread</code>.
	 * @param inName name of this <code>Thread</code>. Must
	 *  be non-null.
	 */
	protected NamedErrorHandlingThread( ThreadGroup inGroup, String inName ) {
		super( inGroup, checkName( inName ) );
		}
	
	
	
	/**
	 * Constructs a <code>NamedErrorHandlingThread</code>, specifying
	 * a <code>Runnable</code> and a <code>ThreadGroup</code>.
	 *
	 * @param inGroup the <code>ThreadGroup</code> for this <code>Thread</code>.
	 * @param inRunnable the <code>Runnable</code> to be run by this <code>Thread</code>.
	 * @param inName name of this <code>Thread</code>. Must
	 *  be non-null.
	 */
	protected NamedErrorHandlingThread( ThreadGroup inGroup,
		Runnable inRunnable, String inName ) {
		
		super( inGroup, inRunnable, checkName( inName ) );
		}
	
	
	
	/**
	 * Checks that a <code>Thread</code> name is non-<code>null</code>.  
	 * Throws an <code>IllegalArgumentException</code> if the name is 
	 * <code>null</code>.
	 * Note that the exception is not declared by this method.
	 *
	 * @param inName name to check for non-<code>null</code> property.
	 *
	 * @return the passed in name (for convienience).
	 */
	private	static String checkName( String inName ) {
		if( inName == null ) {
			throw new IllegalArgumentException( 
				"Thread name must not be null" );
			}
		return inName;
		}	
		
	}
