// Jason Rohrer
// SearchQuery.java

/**
*
*	Message from client with search query
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	8-23-2000	Added JLog logging.
*/

package konspire.common;


import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import konspire.common.log.AppLog;

/**
 * Message from <code>Client</code> to <code>Server</code> containing a search query.
 *
 * @author Jason Rohrer
 */
public class SearchQuery implements Message {


	private String[] searchTerms;
	
	// note that the server database of file names is in all
	// lower case.  Upper case strings won't find matches in the database
	
	private int maxNumResults;
	
	
	/**
	 * Constructs a <code>SearchQuery</code>, specifying separate search terms and the maximum 
	 * number of results to return
	 *
	 * @param terms an array of separate search terms
	 * @param numResults the maximum number of results to return
	 */ 
	public SearchQuery( String[] terms, int numResults ) {
		searchTerms = terms;
		
		maxNumResults = numResults;
		
		// convert all terms to lower case
		for( int i=0; i<searchTerms.length; i++ ) {
			searchTerms[i] = searchTerms[i].toLowerCase();
			}
		}
	
	/**
	 * Constructs a <code>SearchQuery</code>, specifying a single term string and the maximum 
	 * number of results to return
	 *
	 * @param termString a single <code>String</code> containing all search terms
	 *   (usually the search string as the user typed it)
	 * @param numResults the maximum number of results to return
	 */
	public SearchQuery( String termString, int numResults  ) {
		
		maxNumResults = numResults;
		
		termString = termString.toLowerCase();
		
		StringTokenizer tokenMaker = new StringTokenizer( termString );
		
		int numTokens = tokenMaker.countTokens();
		
		searchTerms = new String[numTokens];
		
		try {
			for( int i=0; i<numTokens; i++ ) {
				searchTerms[i] = tokenMaker.nextToken();
				}
			}
		catch( NoSuchElementException e ) {
			AppLog.error( "Search string didn't tokenize correctly: " + termString );
			ErrorHandler.report( e );
			}
		
		}	
	
	/**
	 * Gets the array of search terms contained in this query.
	 *
	 * @return the array of search terms contained in this query
	 */
	public String[] getSearchTerms() {
		return searchTerms;
		}	
	
	/**
	 * Gets the maximum number of results to be returned for this query.
	 *
	 * @return the maximum number of results to be returned for this query
	 */
	public int getNumResults() {
		return maxNumResults;
		}
	
	}
