// Jason Rohrer
// ServerList.java

/**
*
*	Message from server containing list of all known servers
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	7-4-2000	Added receiverCorrectAddress
*									so that clients are assured of
*									knowing their outside-accessible
*									local addresses.
*/

package konspire.common;

import konspire.common.Vector;

/**
 * Message sent from a <code>Server</code> containing a list of all known <code>Server</code>s.
 * <p>
 * This message is sent to both <code>Client</code>s and <code>Server</code>s.
 * <p>
 * This message also contain a correct address for the receiver.  This is used in the
 * case when the host requesting the live server list did not retrieve its local address
 * correctly before sending the request (in the case of a host behind a firewall, for
 * example).  The server sending back the live server list will detect a discrepency
 * between the address contained in the request (inserted by the requester) and the
 * address attached to the socket and send the requester the correct address (the address
 * attached to the socket).  
 *
 * @author Jason Rohrer
 */
public class ServerList extends ToClientMessage {


	private Vector hostVector;
	
	// address that receiver should replace its localAddress with
	// This is used if the receiver was unable to retrieve its local
	// address correctly.
	private Host receiverCorrectAddress;
	
	/**
	 * Constructs a <code>ServerList</code>.
	 *
	 * @param hosts a <code>Vector</code> of <code>Host</code>s containing addresses of all
	 *  known live <code>Server</code>s
	 * @param inReceiverCorrectAddress correct local address for receiver (in case
	 *  the receiver has not retrieved its local address correctly
	 */
	public ServerList( Vector hosts, Host inReceiverCorrectAddress ) {
		hostVector = hosts;
		receiverCorrectAddress = inReceiverCorrectAddress;
		}
	
	/** 
	 * Gets a <code>Vector</code> of live <code>Server</code> <code>Host</code>s contained
	 *  in this message.
	 *
	 * @return a <code>Vector</code> of live <code>Server</code> <code>Host</code>s contained
	 *  in this message
	 */
	public Vector getHostVector() {
		return hostVector;
		}	
	
	
	/**
	 * Gets the corrected receiver address contained in this message.
	 *
	 * @return the corrected receiver address contained in this message
	 */
	public Host getReceiverCorrectAddress() {
		return receiverCorrectAddress;
		}
		

	}