// Jason Rohrer
// ServerListFile.java

/**
*
*	Object that handles writing/reading server list to an from file
*
*	Created 5-19-2000
*	Mods:
*		Jason Rohrer	8-23-2000	Added JLog logging.
*		Jason Rohrer	8-27-2000	Fixed some output message type-o's.
*/

package konspire.common;

import java.io.*;

import konspire.common.log.AppLog;

/**
 * Handles writing/reading a <code>Server</code> list to and from file.
 *
 * @author Jason Rohrer 
 */
public class ServerListFile {

	String hostsFileName;

	
	/**
	 * Constructs a <code>ServerListFile</code>.
	 *
	 * @param fileName the name of the file where the list should be written/read to or from
	 */
	public ServerListFile( String fileName ) {
		hostsFileName = fileName;
		}

	/**
	 * Reads a <code>Vector</code> of <code>Host</code>s from file.
	 *
	 * @return a <code>Vector</code> of <code>Host</code>s read from file
	 */
	public synchronized Vector readHostsVectorFromFile() {
		Host[] hArray = readHostsFromFile();
		
		Vector v = new Vector();
		
		if( hArray == null ) {
			return v;
			}
		
		for( int i=0; i< hArray.length; i++ ) {
			v.addElement( hArray[i] );
			} 
		return v;
		}
	
	/**
	 * Writes a <code>Vector</code> of <code>Host</code>s to file
	 *
	 * @param hostVector the <code>Host</code>s to write to file
	 */
	public synchronized void writeHostsVectorToFile( Vector hostVector ) {
	
		if( hostVector.size() == 0 ) {
			writeHostsToFile( null );
			}
		else {
			Host[] hArray = new Host[ hostVector.size() ];
			
			for( int i=0; i< hArray.length; i++ ) {
				hArray[i] = (Host)( hostVector.elementAt(i) );
				}
			writeHostsToFile( hArray );	 
			}
		}
	
	
	/**
	 * Reads an array of <code>Host</code>s from file.
	 *
	 * @return an array of <code>Host</code>s read from file
	 */
	public synchronized Host[] readHostsFromFile() {
		Host[] hArray;
		
		// use serialization to write hosts to file
		
		try {
			DataInputStream fileInput = new DataInputStream (
							new FileInputStream( hostsFileName ) );
			
			ObjectInputStream objInStream 
					= new ObjectInputStream( fileInput );
			
			try {
				Object inObject = objInStream.readObject();

				if( inObject instanceof Integer ) {
					Integer wrappedNumHosts = (Integer)inObject;
					
					int numHosts = wrappedNumHosts.intValue();
					
					if( numHosts >= 0 ) {
						
						hArray = new Host[numHosts];
						
						for( int i=0; i<numHosts; i++ ) {
							hArray[i] = (Host) objInStream.readObject();
							}
						
						objInStream.close();
						fileInput.close();
						return hArray;
						}
					else {
						// no hosts in file
						return null;
						}
					}
				return null;	
				}
			catch( ClassNotFoundException ex ) {
				AppLog.error( "unknown object read from host file: " +
					hostsFileName, true );
				ErrorHandler.report( ex );
				fileInput.close();
				return null;
				}	

			}
		catch( IOException e ) {
			AppLog.error( "reading host list from file " + hostsFileName +
				" failed.", true );
			ErrorHandler.report( e );
			}
		return null;
		}
	
	
	/**
	 * Writes an array of <code>Host</code>s to file
	 *
	 * @param hArray the <code>Host</code>s to write to file
	 */
	public synchronized void writeHostsToFile( Host[] hArray ) {
		
		// use serialization to write hosts to file

		try {
			DataOutputStream fileOutput = new DataOutputStream (
							new FileOutputStream( hostsFileName ) );
			
			ObjectOutputStream objOutStream 
					= new ObjectOutputStream( fileOutput );
			
			int numHosts = 0;
			if( hArray != null ) {
				numHosts = hArray.length;
				}
				
			Integer wrappedNumHosts = new Integer( numHosts );
			
			objOutStream.writeObject( wrappedNumHosts );
			
			for( int i=0; i<numHosts; i++ ) {
				objOutStream.writeObject( hArray[i] );
				}
				
			objOutStream.flush();
			objOutStream.close();
			fileOutput.close();
			}
		catch( IOException e ) {
			AppLog.error( "writing host list to file " + hostsFileName +
				" failed.", true );
			ErrorHandler.report( e );
			}
			
		}
	
	/**
	 * Writes a <code>Vector</code> of <code>Host</code>s to file as human-readable text.
	 *
	 * @param hostVector the <code>Host</code>s to write to file
	 * @param textFile the <code>File</code> where the text <code>Server</code> list should be written
	 */
	public synchronized void writeHostsVectorToTextFile( Vector hostVector, File textFile ) {
		if( hostVector.size() == 0 ) {
			writeHostsToTextFile( null, textFile );
			}
		else {
			Host[] hArray = new Host[ hostVector.size() ];
			
			for( int i=0; i< hArray.length; i++ ) {
				hArray[i] = (Host)( hostVector.elementAt(i) );
				}
			writeHostsToTextFile( hArray, textFile );
			}
		}
		
	
	/**
	 * Writes an array of <code>Host</code>s to file as human-readable text.
	 *
	 * @param hArray the <code>Host</code>s to write to file
	 * @param textFile the <code>File</code> where the text <code>Server</code> list should be written
	 */	
	public synchronized void writeHostsToTextFile( Host[] hArray, File textFile ) {
		try {
			PrintWriter fileWriter = new PrintWriter (
							new FileOutputStream( textFile ) );
			
			fileWriter.println( "live server list:" );
			
			if( hArray == null ) {
				fileWriter.println( "no servers in list." );
				}
			else if( hArray.length == 0 ) {
				fileWriter.println( "no servers in list." );
				}
			else {
				for( int i=0; i< hArray.length; i++ ) {
					fileWriter.println( hArray[i].getAddress() + " : " + 	hArray[i].getPort() );
					}
				fileWriter.println( "end of list." );
				}
			
			fileWriter.flush();
			fileWriter.close();
			}
		catch( IOException e ) {
			AppLog.error( "writing host list to text file " + textFile.getName() + 
				" failed.", true );
			ErrorHandler.report( e );
			}
		}
		
	}
