// Jason Rohrer
// TwoWayMessageSocket.java

/**
*
*	Class for sending and receiving message between hosts
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	7-3-2000	Added support for getting remote
*									host address directly from socket.
*		Jason Rohrer	8-23-2000	Added JLog logging.
*		Jason Rohrer	9-22-2000	Modified some of the constructors
*									to make this class more subclass friendly.
*									Added a receive method that supports
*									timeouts.
*		Jason Rohrer	9-23-2000	Added locking capabilities so that 
*									threads can force their send and receive
*									pairs to be atomic.
*									Note that this is a cooperative
*									monitor scheme:  All participating
*									threads must attempt to obtain
*									the lock before executing send
*									and receive paired operations, and
*									then must release the lock when done.
*									Added code to each method in which
*									socket could break so that all
*									waiting threads are notified when
*									socket is broken.
*									Synchronized all methods that access
*									the socket to ensure that
*									current thread owns monitor and can
*									therefore notify all waiting threads
*									if socket breaks.  Synchronization
*									is wise anyway since on the client side,
*									multiple threads can be sending and
*									receiving on the same socket.
*/

package konspire.common;


import java.io.*;
import java.net.*;

import konspire.common.log.AppLog;

/**
 * Socket for sending and receiving serialized object messages between <code>Host</code>s.
 *
 * @author Jason Rohrer
 */
public class TwoWayMessageSocket {
	
	private Socket sock;
					
	private DataOutputStream output;
	private DataInputStream input;

	private ObjectOutputStream objOutStream;
	private ObjectInputStream objInStream;

	private boolean socketCreated;	// was socket created internally (or passed in)?
	
	// host connected to other end of socket
	private Host remoteHost;
	
	
	/**
	 * Constructs a <code>TwoWayMessageSocket</code> specifying the <code>Host</code> to connect to.
	 *
	 * @param dest the <code>Host</code> to connect to
	 *
	 * @throws IOException if socket fails to initialize
	 */
	public TwoWayMessageSocket( Host dest ) 
		throws IOException {		
		setupSocket( dest );
		}
	
	/**
	 * Constructs a <code>TwoWayMessageSocket</code> specifying
	 * no configuration.
	 */
	protected TwoWayMessageSocket() {
		}
	
	
	/**
	 * Sets up the socket.  Called by constructors.
	 *
	 * @param dest the <code>Host</code> to connect to
	 *
	 * @throws IOException if socket fails to initialize
	 */
	synchronized protected void setupSocket( Host dest ) 
		throws IOException {
		
		remoteHost = dest;
		try {
			sock = new Socket( dest.getAddress(), dest.getPort() );
			socketCreated = true;

			output = new DataOutputStream( sock.getOutputStream() );
			input = new DataInputStream( sock.getInputStream() );

			objOutStream = new ObjectOutputStream( output );
			objInStream = new ObjectInputStream( input );
			}
		catch( IOException inIOE ) {
			// notify all threads waiting on this socket's locks
			setBroken();
			notifyAll();
			throw inIOE;
			}
		}
	
	
	
	/**
	 * Constructs a <code>TwoWayMessageSocket</code> from an existing
	 * established socket connection.
	 * <p>
	 * Note that socket connection WILL NOT be closed by <@link close>.<BR>
	 * Use <@link closeTotal> to close the passed in socket along with the <code>TwoWayMessageSocket</code>.
	 *
	 * @param s an already established <code>Socket</code> connection to attach
	 *  this <code>TwoWayMessageSocket</code> to.
	 *
	 * @throws IOException if attaching to socket fails
	 */
	public TwoWayMessageSocket( Socket s )
		throws IOException {
		try {
			sock = s;
			socketCreated = false;

			output = new DataOutputStream( sock.getOutputStream() );
			input = new DataInputStream( sock.getInputStream() );

			objOutStream = new ObjectOutputStream( output );
			objInStream = new ObjectInputStream( input );


			InetAddress remoteHostAddress = sock.getInetAddress();
			int remoteHostPort = sock.getPort();

			remoteHost = new Host( remoteHostAddress.getHostAddress(), 
				remoteHostPort );
			}
		catch( IOException inIOE ) {
			// notify all threads waiting on this socket's locks
			setBroken();
			notifyAll();
			throw inIOE;
			}
		}
	
	/**
	 * Gets the <code>Host</code> that this socket is connected to.
	 *
	 * @return the <code>Host</code> that this socket is connected to
	 */
	public Host getRemoteHost() {
		return remoteHost;
		}
	
	/**
	 * Sends a <code>Message</code> through this socket.
	 *
	 * @param m the <code>Message</code> to send
	 *
	 * @throws IOException if send fails
	 */	
	synchronized public void send( Message m ) throws IOException {
		try {
			objOutStream.writeObject( m );
			objOutStream.flush();
			}
		catch( IOException inIOE ) {
			// notify all threads waiting on this socket's locks
			setBroken();
			notifyAll();
			throw inIOE;
			}
		} 	
	
	/**
	 * Gets the <code>DataInputStream</code> attached to this socket.
	 *
	 * @return the <code>DataInputStream</code> attached to this socket
	 */
	public DataInputStream getInputStream() {
		return input;
		}
	
	/**
	 * Gets the <code>DataOutputStream</code> attached to this socket.
	 *
	 * @return the <code>DataOutputStream</code> attached to this socket
	 */
	public DataOutputStream getOutputStream() {
		return output;
		}
	
	
	/**
	 * Receives a <code>Message</code> through this socket.
	 *
	 * @return the received <code>Message</code>
	 *
	 * @throws IOException if receive fails
	 */	
	public Message receive() throws IOException {
		return receive( 0 );	
		}
	
	
	/**
	 * Receives a <code>Message</code> through this socket, with a timeout
	 * specified.
	 *
	 * @param inTimeout the socket timeout in milliseconds.  Specifying 0
	 *   sets an infinite timeout ( equivalent to calling receive() ).
	 *
	 * @return the received <code>Message</code>
	 *
	 * @throws IOException if receive fails
	 * @throws InterruptedIOException if receive times out
	 */	
	synchronized public Message receive( int inTimeout ) 
		throws IOException, InterruptedIOException {
		
		try {
			int oldTimeout = sock.getSoTimeout();

			sock.setSoTimeout( inTimeout );

			try {
				Object received = objInStream.readObject();		

				sock.setSoTimeout( oldTimeout );

				if( received instanceof Message ) {
					return (Message) received;
					}
				return null;
				}
			catch( ClassNotFoundException e ) {
				AppLog.error( "Unknown class received on object input stream." );
				ErrorHandler.report( e );
				return null;
				}
			}
		catch( IOException inIOE ) {
			// notify all threads waiting on this socket's locks
			setBroken();
			notifyAll();
			throw inIOE;
			}	
		}
	
	
	private boolean beenClosed = false;
	private boolean broken = false;
	
	/**
	 * Marks this socket as broken.
	 */
	synchronized private void setBroken() {
		broken = true;
		}
	
	
	/**
	 * Closes this <code>TwoWayMessageSocket</code>.  If the corresponding
	 * <code>Socket</code> was created externally, the <code>Socket</code> is
	 * not closed, just the streams attached to it.
	 *
	 * @throws IOException if close operation fails
	 */
	public synchronized void close() throws IOException {
		try {
			if( !beenClosed ) {
				objOutStream.close();
				objInStream.close();
				output.close();
				input.close();

				// only close socket if created internally
				if( socketCreated ) {
					sock.close();
					}

				beenClosed = true;

				// Notify any threads that are waiting on locks for
				// this socket
				notifyAll();
				}
			}
		catch( IOException inIOE ) {
			// notify all threads waiting on this socket's locks
			setBroken();
			notifyAll();
			throw inIOE;
			}
		}
	
	/**
	 * Closes this <code>TwoWayMessageSocket</code> completely, including closing 
	 * any externally created <code>Socket</code> (e.g., if <code>Socket</code>
	 * was passed in to constructor.
	 *
	 * @throws IOException if close operation fails
	 */
	public void closeTotal() throws IOException {
		socketCreated = true;
		close();
		}	
	
	
	
	// Locks allowing multiple threads to execute sets of socket operations 
	// atomically.  All threads manipulating the socket must be lock-aware
	// for this scheme to work.
	
	private boolean locked;
	
	/**
	 * Obtain a lock on this socket.
	 *
	 * @throws InterruptedException if thread interupted while waiting for lock.
	 */
	synchronized public void obtainLock() 
		throws InterruptedException {	
		// let thread go right through lock if socket has already been closed
		if( !beenClosed && !broken ) {
			if( locked ) {
				wait();
				}
			locked = true;
			}	
		}
	
	/**
	 * Obtain a lock on this socket.
	 */
	synchronized public void releaseLock() {
		locked = false;
		notify();
		}
		
	}
