package konspire.common.log;

import java.io.*;
import java.util.*;
import java.text.*;

/**
 * AbstractFileLoggingMechanism is an abstract class which
 * defines the format and location of log files,
 * so they all work consistently for the same application.  The default
 * file name where logs are created is:
 * <pre>
 *		[Path][App][LogClass][date].[extn]
 *
 *		where:  	[Path] 		= "c:\temp\"
 *					[App]  		= "App"
 *					[LogClass]	= The Java Class name of the Log Class
 *                                (only used for unique logs)
 *					[date]		= today's date (only used for dated logs)
 *					[extn]		= "log"
 *
 * </pre>
 * The [App] default can be overridden by the application, it is taken
 * from the LogManager class.  The [Path] and [extn] defaults can be
 * overridden by settings in a properties file, see the LogProperties
 * class for more information.
 *
 * @author Todd Lauinger
 * @version $Revision: 1.3 $
 */

public abstract class AbstractFileLoggingMechanism extends AbstractLoggingMechanism {


	// Default class variables for where logs are put
	protected static String logPath = LogProperties.DEFAULT_LOG_FILE_PATH;

	// Instance variable to hold the output file stream
	protected FileOutputStream fileOutputStream;
	
	/**
	 * Instance variable to hold the current output file name
	 */
	protected String currentFileName = "";

	// Returns the filename prefix the log should write to
	protected String logFileNamePrefix(Log log) {

        StringBuffer sb = new StringBuffer();
        
        // All files have a path and application name
        sb.append(LogProperties.getInstance().getLoggingPath(log));
        sb.append(LogManager.getInstance().getApplicationName());
        
        // Unique logs also have the log name
        if (! isSharedLog() ) {
            sb.append(log.getName());
        }
        
        // Dated logs also have the current date
        if (isDatedLog()) {
            sb.append(compactCurrentDate());
        }
        
        return sb.toString();
	}

 	// Returns the filename suffix the log should write to
	protected String logFileNameSuffix(Log log) {

		return "." + LogProperties.getInstance().getLoggingExtension(log);
	}
	
	public String getFileName() {
	    return currentFileName;
	}

    // Open a file log
	synchronized protected void open(Log log) {

        super.open(log);

		String fileName = logFileNamePrefix(log) + logFileNameSuffix(log);
		
		try {
			fileOutputStream = new FileOutputStream(fileName, true);  // append
			printStream = new PrintWriter(fileOutputStream, true);  // autoflush
			currentFileName = fileName;
		} catch (IOException e) {
			System.err.println("Caught IOException: " +
					e.getMessage() + " trying to open file named " + fileName);
		    currentFileName = "";
        }
        	
	}

    // Close a file log
	synchronized protected void close() {

		try {
			fileOutputStream.close();
		} catch (IOException e) {
			System.err.println("Caught IOException: " +
					e.getMessage() + " trying to close log file");
      }

      super.close();

	}

	/**
	 * Return a nice debug string.
	 */
	public String toString() {
	    return super.toString() + " is logging to file " + currentFileName;
	}

}
