package konspire.common.log;

import java.io.*;
import java.util.*;
import java.text.*;
import java.lang.reflect.*;

/**
 * AbstractLoggingMechanism is an abstract class which serves to define a
 * hierarchy of different logging mechanisms.  Subclasses of
 * this class will implement the different mechanisms defined herein.
 *
 * The main responsibility of this class hierarchy is to offload the
 * implementation work for differences in logging streams for different
 * mechanisms.  For example, the standard error stream is a very different
 * implementation than a shared local output file.
 *
 * @author Todd Lauinger
 * @version $Revision: 1.3 $
 */

public abstract class AbstractLoggingMechanism {

	// Keep a character print stream to write out the log to.  The default is
	// System.err, in case anything goes wrong.
	PrintWriter printStream = new PrintWriter(System.err);
	
	/**
	 * The date the mechanism was first opened.
	 */
	Date dateOpened;

	// Keep a list of logs that use the logging mechanism, so we know when
	// to open/close the log properly.
	Vector logs = new Vector(10);

	// Set up dates and times that are short to take minimal space in the log
	protected static SimpleDateFormat timeFormat =
		new SimpleDateFormat ("HH:mm:ss:SSS");

	protected static SimpleDateFormat compactDateFormat =
		new SimpleDateFormat ("MMddyy");

	protected static DateFormat dateFormat =
		DateFormat.getDateInstance(DateFormat.SHORT);

	// Quick methods to return current date and time
	protected String currentDate() {

		return (dateFormat.format(new Date()));
	}

	protected String compactCurrentDate() {

		return (compactDateFormat.format(new Date()));
	}

	protected String currentTime() {

		return (timeFormat.format(new Date()));
	}


    /**
     * Mechanisms can be dated or non-dated.  If dated, the log 
     * needs to have a date appended to a
     * file name (if it has one) and it must be rolled over to a new
     * one when the current date changes.  The default setting is
     * non-dated, this method must be overridden when dated.
     */
    protected boolean isDatedLog() {
        return false;
    }

    /**
     * Mechanisms can be shared or unique.  Shared mechanisms
     * require the log name to be in every log entry (so you
     * can distinguish between the log entries).  Unique
     * mechanisms name themselves uniquely (for example, have
     * the log name in their file name) so they do not
     * have to put the name of the log in their log entry.
     * The default is shared (true).
     */
    protected boolean isSharedLog() {
        return true;
    }

	/**
	 * Returns a string which is to be concatenated before the
	 * log message, so that the log message can be properly understood
	 * in the context of other messages and logs (if a shared log).
	 *
	 * <pre>
	 * The format is [loggingLevel]|[time]|
	 * where  [loggingLevel] is the integer logging level the msg was logged at
	 *        [time] is the local time in SHORT DateFormat
	 * </pre>
	 */
	protected String loggingPrefix(Log log, int loggingLevel) {

        StringBuffer sb = new StringBuffer();
        
        // If a shared log must prepend with log name
        if (isSharedLog() ) {
            sb.append(log.getName());
            sb.append(log.getLogFieldSeparator());
        }
        
        sb.append("L");
        sb.append(loggingLevel);
        sb.append(log.getLogFieldSeparator());
        sb.append(currentTime());
        sb.append(log.getLogFieldSeparator());
        
        return sb.toString();
	}

	/**
	 * This method opens the particular logging mechanism so that messages
	 * can be output to the printSteam instance variable.
	 */
	protected void open(Log log) {
	    dateOpened = new Date();
	}

	/**
	 * This method closes the particular logging mechanism so that messages
	 * no longer get logged to the mechanism.  Overrides should call the superclass
	 * method so that printStream gets reset to StdErr in case a subsequent open
	 * has problems and the error can be logged somewhere!
	 */
	synchronized protected void close() {

		printStream = new PrintWriter(System.err);
	}

	// Put out a nice message that a log was opened on this mechanism
	synchronized protected void notifyLogOpened(Log log, 
	    int loggingLevel) {

		printStream.println("Log " + log.getName() + " opened on " +
			currentDate() + " at " + currentTime());
		notifyLoggingLevelChange(log, loggingLevel);
	}

	// Put out a nice message that a log was closed on this mechanism
	synchronized protected void notifyLogClosed(Log log) {

		printStream.println("Log " + log.getName() + " closed on " +
			currentDate() + " at " + currentTime());
	}

	// Put out a nice message that a log changed its logging level
	synchronized void notifyLoggingLevelChange(Log log, int loggingLevel) {

		printStream.println("Log " + log.getName() + " is logging at level " +
			loggingLevel + " at " + currentTime());
	}

	/**
	 * Logs call this method to register themselves with the mechanism.
	 * It will automatically open the mechanism if needed and log the fact
	 * that a log started using this mechanism.
	 */
	synchronized void openForLog(Log log, int loggingLevel) {

		// If no logs registered, this must be the first.  Open the mechanism.
		if (logs.size() <= 0)
			open(log);

		// If the log isn't already in the set of logs, add it and put out
		// a notification message
		if (! logs.contains(log)) {
			logs.addElement(log);
			notifyLogOpened(log, loggingLevel);
		}
	}

	/**
	 * Logs call this method to deregister themselves with the mechanism.
	 * It will automatically close the mechanism if needed and log the fact
	 * that a log stopped using this mechanism.
	 */
	synchronized void closeForLog(Log log) {

		// If the log is in the set of logs, remove it and put out
		// a notification message.  Then close the logging mechanism if the last
		// log registered.
		if (logs.contains(log)) {
			logs.removeElement(log);
			notifyLogClosed(log);

			if (logs.size() <= 0)
				close();
		}
	}
	
    /**
     * Checks the current date verses the date the log was opened.
     * If the current date is one or more days greater (that is,
     * have we passed midnight), return true.  Else, return false.
     * To make this method reasonably fast, it does not compare years, only
     * days within a year.
     */
    protected boolean hasDateRolledOver() {
        int todaysDayOfWeek = Calendar.getInstance().get(Calendar.DAY_OF_YEAR);
        Calendar logOpenedCalendar = Calendar.getInstance();
        logOpenedCalendar.setTime(dateOpened);
        int logOpenedDayOfWeek = logOpenedCalendar.get(Calendar.DAY_OF_YEAR);
        return (todaysDayOfWeek != logOpenedDayOfWeek);
    }

	/**
	 * Rolls a log over, typically on a date change, by closing the log
	 * and re-opening it.
	 */
	protected void rollLogOver() {
	    
	    Enumeration e = logs.elements();
	    Log log = null;
	    while (e.hasMoreElements()) {
	        log = (Log) e.nextElement();
	        this.notifyLogClosed(log);
	    }
	    
	    close();
	    open(log);
	    
	    e = logs.elements();
	    while (e.hasMoreElements()) {
	        log = (Log) e.nextElement();
	        this.notifyLogOpened(log, log.getLoggingLevel());
	    }
	}
	    

	/**
	 * Logs a string to printStream, complete with a descriptive prefix so viewers
	 * of the log can interpret who logged which strings and how important they were.
	 */
	synchronized protected void logString(Log log, int loggingLevel, String logString) {
		
		// If the log is dated and we rolled over a date, must close and
		// reopen the log to switch dates!
		if (isDatedLog() && hasDateRolledOver()) {
		    rollLogOver();
		}
				
		printStream.println(loggingPrefix(log, loggingLevel) + logString);
		printStream.flush();

	}
	
	//
	// Private helper method to get a good logging mechanism instance from an
	// input logging mechanism class string.
	//
	static AbstractLoggingMechanism getLoggingMechanismForString(String mechanismString)
		throws Exception {

		Class loggingMechanismClass, abstractLoggingMechanismClass;
		Method getInstanceMethod;

		// Translate the input class string into a class, see if a legal subclass of
		// logging mechanisms.
		loggingMechanismClass = Class.forName(mechanismString);
		abstractLoggingMechanismClass = AbstractLoggingMechanism.class;
		
		if (! abstractLoggingMechanismClass.isAssignableFrom(loggingMechanismClass))
			throw new ClassNotFoundException("Error: Logging Mechanism Class "
				+ mechanismString +
				" is not a legal subclass of AbstractLoggingMechanism!");
				
		// Return the current logging mechanism for the particular class found
		getInstanceMethod = loggingMechanismClass.getMethod("getInstance", new Class[0]);
		return ((AbstractLoggingMechanism) getInstanceMethod.invoke(loggingMechanismClass, new Object[0]));

	}

	/**
	 * Return the name of the mechanism.  Today this is the full class name.
	 */
	public String getName() {

		return this.getClass().getName();
	}
	
	/**
	 * Return a nice debug string.
	 */
	public String toString() {
	    return getName();
	}

}
