// Jason Rohrer
// AppLog.java

/**
*
*	Modified from original AppLog by Todd Lauinger for use by konspire.
*
*	Created 8-21-2000
*	Mods:
*		Jason Rohrer	8-23-2000	Changed so that all message reporting
*									methods are now overridden with methods
*									that take booleans specifying whether
*									message should also be printed directly
*									to System.out (without log tags).
*		Jason Rohrer	8-23-2000	Changed to use UserMessageHandlers for
*									"print also" messages.
*/

package konspire.common.log;

import konspire.common.UserMessageHandler;
import konspire.common.StdUserMessageHandler;

/**
 * <code>AppLog</code> wraps the Log class to do application-general
 * logging.  It provides a singleton accessor, along with lots of 
 * handy class method {static} accessors. <p>
 *
 * <code>AppLog</code> should be your general application log to write out events
 * of a non-specific nature.  You can create additional special-purpose
 * logs (for example, DbLog to log database events) by copying
 * this class and creating your own handy class method accessors. <p>
 *
 * @author Todd Lauinger
 * @version $Revision: 1.7 $
 *
 * @see Log
 *
 */
public class AppLog {

    /**
     * Provides the class variable for the Singleton pattern,
     * to keep track of the one and only instance of this class.
     */
    protected static Log singleton = new MultiThreadLog("AppLog");


	/**
     * Message handler to send "print also" message to if message
	 * handler printing is turned on.
	 */
	private static UserMessageHandler mMessageHandler = new StdUserMessageHandler();


	/**
	 * Sets the <code>UserMessageHandler</code> used by this <code>Log</code>
	 * for "print also" messages.
	 * <p>
	 * The handler defaults to a <code>StdUserMessageHandler</code> (prints
	 * to System.out) if none is specified.
	 *
	 * @param inHandler the <code>UserMessageHandler</code> this
	 *  <code>Log</code> will use for "print also" messages.
	 */
	public static void setMessageHandler( UserMessageHandler inHandler ) {
		mMessageHandler = inHandler;
		}	


    /**
     * Returns the singleton instance of this class.
     */
    public static Log getInstance() {
        return singleton;
    }
	
    /**
     * Logs a text string at the logging level specified.
     *
     * This method is used when your code needs to parameterize
     * the logging level.  Otherwise you probably should use the
     * err(), warning(), info(), detail(), or trace() methods below,
     * they will be less verbose in your code.
     */
    public static void logString(String logString, int errorLevel, 
		boolean inPrintAlso) {

        // Dispatch the call to the instance side
        singleton.logString(logString, errorLevel);
		
		if( inPrintAlso ) {
			mMessageHandler.showMessage( logString );
		}
		
    }

    /**
     * Logs a text string at CRITICAL_ERROR_LEVEL
     *
     * This method provides the implementation for a quick accessor
     * to log a critical error,
     * replacing typical calls to System.err.println().
     */
    public static void criticalError(String logString) {

        logString(logString, Log.CRITICAL_ERROR_LEVEL, false);
    }

	/**
     * Logs a text string at ERROR_LEVEL
     *
     * This method provides the implementation for a quick accessor
     * to log a critical error,
     * replacing typical calls to System.err.println().
     */
    public static void error(String logString) {

        logString(logString, Log.ERROR_LEVEL, false);
    }

    /**
     * Logs a text string at WARNING_LEVEL
     */
    public static void warning(String logString) {

        logString(logString, Log.WARNING_LEVEL, false);
    }


    /**
     * Logs a text string at INFO_LEVEL
     */
    public static void info(String logString) {

        logString(logString, Log.INFO_LEVEL, false);
    }

    /**
     * Logs a text string at DETAIL_LEVEL
     */
    public static void detail(String logString) {

        logString(logString, Log.DETAIL_LEVEL, false);	
    }

    /**
     * Logs a text string at TRACE_LEVEL
     */
    public static void trace(String logString) {

        logString(logString, Log.TRACE_LEVEL, false);
    }


	
	/**
     * Logs a text string at CRITICAL_ERROR_LEVEL
     *
     * This method provides the implementation for a quick accessor
     * to log a critical error,
     * replacing typical calls to System.err.println().
     */
    public static void criticalError(String logString, boolean inPrintAlso) {

        logString(logString, Log.CRITICAL_ERROR_LEVEL, inPrintAlso);
    }

	/**
     * Logs a text string at ERROR_LEVEL
     *
     * This method provides the implementation for a quick accessor
     * to log a critical error,
     * replacing typical calls to System.err.println().
     */
    public static void error(String logString, boolean inPrintAlso) {

        logString(logString, Log.ERROR_LEVEL, inPrintAlso);
    }

    /**
     * Logs a text string at WARNING_LEVEL
     */
    public static void warning(String logString, boolean inPrintAlso) {

        logString(logString, Log.WARNING_LEVEL, inPrintAlso);
    }


    /**
     * Logs a text string at INFO_LEVEL
     */
    public static void info(String logString, boolean inPrintAlso) {

        logString(logString, Log.INFO_LEVEL, inPrintAlso);
    }

    /**
     * Logs a text string at DETAIL_LEVEL
     */
    public static void detail(String logString, boolean inPrintAlso) {

        logString(logString, Log.DETAIL_LEVEL, inPrintAlso);	
    }

    /**
     * Logs a text string at TRACE_LEVEL
     */
    public static void trace(String logString, boolean inPrintAlso) {

        logString(logString, Log.TRACE_LEVEL, inPrintAlso);
    }


}
