// Jason Rohrer
// LogProperties.java

/**
*
*	Modified from original LogProperties by Todd Lauinger for use by konspire.
*
*	Created 8-21-2000
*	Mods:
*		Jason Rohrer	8-21-2000	Removed error message that is displayed when
*									the properties file is not found.
*									Now default properties are used quietly when
*									no properties file exists.  
*									Default properties have been changed to
*									be more konspire-friendly.
*		Jason Rohrer	8-23-2000	Fixed a bug in the default log path.
*									Changed to detail logging level for default.
*/

package konspire.common.log;

import java.io.*;
import java.util.*;
import java.text.*;

/**
 * LogProperties is a concrete class which manages all of the default
 * start-up properties for logs.  There is only one singleton
 * instance of this class.  The first time the application uses this
 * class, it will attempt to read its properties from the file named
 * [App].properties in the current directory, where [App] is the application
 * name taken from the LogManager class instance.
 *
 * The following property tags and defaults will be used when
 * reading the log property file:
 *
 * <pre>
 * LogFileExtension = log
 * LogFilePath = ./
 * LoggingLevel = 5
 * LoggingMechanism = UniqueFileLoggingMechanism
 * LogFieldSeparator = |
 * </pre>
 *
 * Any of the above property tags can be overridden for a particular
 * log, by prefixing the property key by the log name.  Also note
 * that because you may have more than one logging mechanism open
 * for the same log, append a consecutive integer to the
 * property starting at 2.  For example:
 *
 * <pre>
 * LogFilePath = /tmp
 * AppLog.LoggingLevel = 3
 * AppLog.LoggingMechanism2 = log.SharedFileLoggingMechanism
 * AppLog.LogFieldSeparator = ~
 * </pre>
 *
 * @author Todd Lauinger
 * @version $Revision: 1.6 $
 */
public class LogProperties extends Properties {

	// Property file extension
	protected static final String PROPERTY_FILE_EXTENSION = "properties";

	// Keys and default values	
	protected static final String LOG_FILE_PATH_PROPERTY = "LogFilePath";
	public static final String DEFAULT_LOG_FILE_PATH = "./";

	protected static final String LOG_FILE_EXTENSION_PROPERTY = "LogFileExtension";
	public static final String DEFAULT_LOG_FILE_EXTENSION = "log";

	protected static final String LOGGING_MECHANISM_PROPERTY = "LoggingMechanism";
	public static final String DEFAULT_LOGGING_MECHANISM =
		Log.UNIQUE_FILE_MECHANISM;
	
	protected static final String LOGGING_LEVEL_PROPERTY = "LoggingLevel";
	public static final int DEFAULT_LOGGING_LEVEL = Log.DETAIL_LEVEL;

	protected static final String LOG_FIELD_SEPARATOR_PROPERTY = "LogFieldSeparator";
	public static final String DEFAULT_LOG_FIELD_SEPARATOR = "|";

	protected static boolean initialized = false;
	
	protected static LogProperties singleton = new LogProperties();

	/**
	 * Class accessor (Factory Method) which answers the 
	 * Singleton instance of the LogProperties for this virtual machine.
	 */
	public static synchronized LogProperties getInstance() {

		FileInputStream fileInputStream;
		String fileName;
		
		// If not initialized, read in the properties from the default file name
		if (!initialized) {

			fileName = LogManager.getInstance().getApplicationName() + "." +
					PROPERTY_FILE_EXTENSION;
			try {
				fileInputStream = new FileInputStream(fileName);
				singleton.load(fileInputStream);
				fileInputStream.close();
			} catch (IOException e) {
				//System.err.println("Caught IOException: " +
				//		e + " trying to open file named " + fileName);
				// do nothing, just use defaults
	      }
			
			initialized = true;
		}
		
		return singleton;
	}

	/**
	 * Checks the property file to see if a property is in the file.
	 * A property can either be defined as a property global to all
	 * logs or specific to the log given.
	 */
	protected boolean containsLogProperty(Log log, String key) {

		String value;

		value = getProperty(log.getName() + "." + key, "");

        // If both the log property and the global property are not
        // found, return false
		if (value.equals("")) {
			value = getProperty(key, "");
			if (value.equals("")) {
			    return false;
			}
		}

        // One or the other found, return true!
		return true;
	}

	/**
	 * Returns the default property for the key given by first looking
	 * for a property prefixed by the log name.  If none found, then
	 * try a generic property for all logs, and if none found there,
	 * return the defaultValue given.
	 */
	protected String getLogProperty(Log log, String key, String defaultValue) {

		String value;

		value = getProperty(log.getName() + "." + key, "");

		if (value.equals("")) {
			value = getProperty(key, defaultValue);
		}

		return value;
	}

	/**
	 * Return the file path the log file should be written out under.
	 */
	public String getLoggingPath(Log log) {

		return getLogProperty(log,
			LOG_FILE_PATH_PROPERTY, DEFAULT_LOG_FILE_PATH);
	}
	
	/**
	 * Return the file name extension the log file should be written out under.
	 */
	public String getLoggingExtension(Log log) {

		return getLogProperty(log,
			LOG_FILE_EXTENSION_PROPERTY, DEFAULT_LOG_FILE_EXTENSION);
	}

	/**
	 * Return the logging mechanisms for the log given.
	 */
	public Vector getLoggingMechanismNames(Log log) {

        Vector loggingMechanismNames = new Vector(10);
        
        // Get the first default mechanism, then see if any
        // others defined.
		loggingMechanismNames.addElement(
		    getLogProperty(log,	LOGGING_MECHANISM_PROPERTY, 
		        DEFAULT_LOGGING_MECHANISM                   ) );
		        
		// Look for other logging mechanisms starting at 2, stop
		// looking when we get the default mechanism!
		int mechanismNumber = 2;
		String nextLoggingMechanismName = null;
		while (containsLogProperty(log,	
		        LOGGING_MECHANISM_PROPERTY + mechanismNumber)) {
		    nextLoggingMechanismName = getLogProperty(log,	
		        LOGGING_MECHANISM_PROPERTY + mechanismNumber, 
		        DEFAULT_LOGGING_MECHANISM);
		    loggingMechanismNames.addElement(nextLoggingMechanismName);
		    mechanismNumber++;
		}
		
		return loggingMechanismNames;

	}

	/**
	 * Return the logging level for the log given.
	 */
	public int getLoggingLevel(Log log) {

		return Integer.parseInt( 
		    getLogProperty(log, LOGGING_LEVEL_PROPERTY, 
		        Integer.toString(DEFAULT_LOGGING_LEVEL) ) );
	}

	/**
	 * Return the logging level for the log given.
	 */
	public String getLogFieldSeparator(Log log) {

		return getLogProperty(log,
			LOG_FIELD_SEPARATOR_PROPERTY, DEFAULT_LOG_FIELD_SEPARATOR);
	}

}
