package konspire.common.log;

import java.util.*;
import java.text.*;


/**
 * <code>MultiThreadLog</code> is an extension of the 
 * <code>Log</code> class to add information to each log
 * entry specifying the thread that generated the entry.  In a
 * multi-threaded application, a particular log file may be
 * intermixed with a number of entries from a number of
 * threads.  To make sense of what is going on in a particular
 * thread, this type of log will prepend all entries with the name
 * of the thread, so you can "grep" out a particular thread of
 * interest.<p>
 *
 * Like the <code>Log</code> class, this class could be extended to do 
 * application-specific behavior,
 * or better yet you could create a class "wrapper" like <code>AppLog</code>
 * to make instances well-known Singletons with handy static/class methods.
 *
 * @author Todd Lauinger
 * @version $Revision: 1.3 $
 *
 * @see Log
 * @see AppLog
 */
public class MultiThreadLog extends Log {

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Constructor section

	/**
	 * Public constructor to register a new <code>MultiThreadLog</code>
	 * with a particular name.
	 * The name of the log is very important.  It is used as a key into
	 * the log properties file to find settings for how the log is to
	 * be opened.
	 */
	public MultiThreadLog(String name) {
	    super(name);
	}

	/**
	 * Adds a message string to the log if the log's level is
	 * currently set at or above the loggingLevel passed in.
	 * Prepends the current thread name before the log entry.
	 *
	 * @param logString The string to write to all logging mechanisms
	 *
	 * @param loggingLevel The level at which to log the string.
	 */
	synchronized public void logString(String logString, int loggingLevel) {

		// Log the message if higher or equal to current logging level.
		// Note, the subclass does this check as well, but for performance
		// we will catch it here so we don't have to look up the
		// current thread information if we don't need to.
		if (loggingLevel <= currentLoggingLevel) {
		    super.logString(Thread.currentThread().getName() + 
		            logFieldSeparator + logString, loggingLevel) ;
		}
	}
}
