package konspire.common.log;

import java.util.*;
import java.rmi.*;
import java.rmi.server.*;

/**
 * RemoteLogManagerImpl is a concrete class which provides an RMI interface to
 * the LogManager class so that RMI clients can remote control the application's logs.
 *
 * This class houses the default RMI registration in its main() method,
 * binding with a name of (ApplicationName)RemoteLogManager, where
 * (ApplicationName) is the name of the application as recorded in
 * the property file or overridden in the LogManager class.
 *
 * @author Todd Lauinger
 * @version $Revision: 1.3 $
 *
 * @see LogManager
 * @see RemoteLogManager
 */
public class RemoteLogManagerImpl extends UnicastRemoteObject
    implements RemoteLogManager {


    /**
     * Default constructor method.
     */
    public RemoteLogManagerImpl() throws RemoteException {
        super();
    }

    /**
     * Helper method to find a log in the LogManager given a String
     * name.
     */
	protected Log logForName(String logName) {

		for (Enumeration e = LogManager.getInstance().elements(); e.hasMoreElements(); ) {

			Log log = (Log) e.nextElement();
			if (log.getName().equals(logName))
				return log;
		}

		// Didn't find a log for the input name, return null
		return null;
	}

    /**
     * Retrieves all of the names of the logs registered in
     * the virtual machine.
     */
	public synchronized String[] getLogNames() throws RemoteException {

		String[] logNames = new String[LogManager.getInstance().size()];

		int position = 0;
		for (Enumeration e = LogManager.getInstance().elements(); e.hasMoreElements(); position++) {
			Log log = (Log) e.nextElement();
			logNames[position] = log.getName();
		}

		return logNames;
	}

    /**
     * Gets the current log field separator of a particular log.
     */
	public synchronized String getLogFieldSeparator(String logName) throws RemoteException {

		Log log;

		log = logForName(logName);

		if (log != null)
			return log.getLogFieldSeparator();

		// Can't find the log, return DEACTIVATE_LEVEL as a default.
		return LogProperties.DEFAULT_LOG_FIELD_SEPARATOR;
	}

    /**
     * Resets the current log field separator of a particular log.
     */
	public synchronized void setLogFieldSeparator(String logName, String logFieldSeparator) throws RemoteException {

		Log log;

		log = logForName(logName);

		if (log != null)
			log.setLogFieldSeparator(logFieldSeparator);
	}

    /**
     * Gets the current logging level of a particular log.
     */
	public synchronized int getLoggingLevel(String logName) throws RemoteException {

		Log log;

		log = logForName(logName);

		if (log != null)
			return log.getLoggingLevel();

		// Can't find the log, return DEACTIVATE_LEVEL as a default.
		return Log.DEACTIVATE_LEVEL;
	}

    /**
     * Resets the current logging level of a particular log.
     */
	public synchronized void setLoggingLevel(String logName, int loggingLevel) throws RemoteException {

		Log log;

		log = logForName(logName);

		if (log != null)
			log.setLoggingLevel(loggingLevel);
	}

    /**
     * Returns all of the names of the logging mechanisms currently
     * active for the log given as input.
     */
    public synchronized String[] getLoggingMechanismNames(String logName) throws RemoteException {
        
		Log log;
		log = logForName(logName);

		if (log == null)
		    return new String[0];
		    
		String[] mechanismNames = new String[log.getLoggingMechanisms().size()];

		int position = 0;
		for (Enumeration e = log.getLoggingMechanisms().elements(); e.hasMoreElements(); position++) {
			AbstractLoggingMechanism mechanism = (AbstractLoggingMechanism) e.nextElement();
			mechanismNames[position] = mechanism.getName();
		}

		return mechanismNames;
	}
	
    /**
     * Adds a logging mechanism to the log given.
     */
    public synchronized void addLoggingMechanism(String logName, String mechanismName) 
        throws RemoteException {

        Log log;
		log = logForName(logName);

        // If couldn't find the log requested, do nothing and return
		if (log == null)
		    return;
		
		log.addLoggingMechanism(mechanismName);
        
    }

    /**
     * Removes a logging mechanism from the log given.
     */
    public synchronized void removeLoggingMechanism(String logName, String mechanismName) 
        throws RemoteException {

        Log log;
		log = logForName(logName);

        // If couldn't find the log requested, do nothing and return
		if (log == null)
		    return;
		
		log.removeLoggingMechanism(mechanismName);
        
    }

	/**
	 * This method binds the remote log manager so RMI clients can access
	 * it.  The naming convention is <ApplicationName>.RemoteLogManager
	 */
    public static void main(String[] args) {

		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());
		}

		String name = LogManager.getInstance().getApplicationName() +
			"RemoteLogManager";
		try {
			RemoteLogManagerImpl remoteLogManager = new RemoteLogManagerImpl();
			Naming.rebind(name, remoteLogManager);
			/* System.out.println(name + " bound"); */
			AppLog.info(name + " bound");
		} catch (Exception e) {
			System.err.println("Exception trying to bind RMI Log Manager named " +
			name + ": " + e.getMessage());
		}
	}

	/**
	 * Adds a message string to the log if the log's level is
	 * currently set at or above the loggingLevel passed in.
	 *
	 * @param logName The name of the log as a string
	 *
	 * @param logString The string to write to all logging mechanisms
	 *
	 * @param loggingLevel The level at which to log the string.
	 */
	public synchronized void logString(String logName, String logString, int loggingLevel) 
	    throws RemoteException {
	
		Log log;

		log = logForName(logName);

		if (log != null)
			log.logString(logString, loggingLevel);
	    
	}

}
