// Jason Rohrer
// Server.java

/**
*
*	Main server class for konspire
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	6-14-2000	Added support for passing client host name into 
*									database for file search. (used to filter results)
*		Jason Rohrer	7-4-2000	Fixed bug involving clients/servers that don't retrieve
*									their own IP Addresses correctly.  Previously, their FileHost
*									and NewServerUp messages did not contain a correct address 
*									(the localhost 127.0.0.1 address has even been witnessed).
*									Now, the server checks the included address against the address
*									attached to the other end of the socket.  The correct address is
*									passed back to the client/server in the ServerList message.
*		Jason Rohrer	8-14-2000	Changed so that this server's address is
*									added to a FileHost for a connected client
*									before propagating the FileHost.
*		Jason Rohrer	8-21-2000	Added JLog loggin support.
*		Jason Rohrer	8-23-2000	Replaced all occurences of System.out.print
*									with appropriate JLog messaging. Changed so
*									that all threads override
*									NamedErrorHandlingThread instead of Thread.
*		Jason Rohrer	8-23-2000	Moved setting of log application name out of
*									constructor and into main method.
*		Jason Rohrer	9-22-2000	Changed all occurences of
*									TwoWayMessageSocket to CheckedMessageSocket 
*									to ensure that other host is alive.
*		Jason Rohrer	9-23-2000	Added logging of incoming connection IP 
*									address as soon as new receive thread
*									is launched.
*									Client handler now spawns a thread to
*									propagate the client's FileHostMessage 
*									so that the client can begin searching 
*									immediately after sending up it's file list.
*									Added support for sending clients
*									a refreshed server list.
*		Jason Rohrer	9-30-2000	Added support for a 1-simultaneous-fault
*									tolerant message propagation scheme.  
*									Changed so that ClientDisconnect messages
*									use this fault-tolerant scheme.
*									
*/

package konspire.server;

import konspire.server.database.ClientDatabase;
import konspire.common.*;

import java.net.*;
import java.io.*;

import java.util.Date;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import konspire.common.Vector;

import konspire.common.log.LogManager;
import konspire.common.log.AppLog;

/**
 * Main server class for konspire.
 *
 * @author Jason Rohrer
 */
public class Server extends NamedErrorHandlingThread {
	
	// database of all connected clients and files
	// network-wide, not just for clients connected to this server
	ClientDatabase database;
	
	// used to propagate messages to all known servers
	ServerPropagator propagator;
	
	// handles writing/reading server list on disk
	ServerListFile sListFile;
	
	Host localAddress;

	int maxClients;	// max clients to allow
	int maxConnections;	// max connections to backlog total
	int numClientsConnected = 0;
	
	static String hostsFileName = "servers.dat";
	
	/**
	 * Gets the network address of this <code>Server</code>.
	 *
	 * @return the network address of this <code>Server</code>
	 */
	public Host getLocalAddress() {
		return localAddress;
		}
	
	/**
	 * Constructs a <code>Server</code>.
	 *
	 * @param me local host 
	 * @param maxClie maximum number of simultaneous client connections
	 * @param newChain true iff this server is first in group of servers
	 *  (if true, this server won't try to connect to others to get the live server list)
	 */
	public Server( Host me, int maxClie, boolean newChain ) {
		super( "Server" );
		
		database = new ClientDatabase();
		
		localAddress = me;
		
		maxClients = maxClie;
		maxConnections = 2 * maxClients;
		
		sListFile = new ServerListFile( hostsFileName );
		
		if( ! newChain ) {
			AppLog.info( "Server will connect to an existing server chain.",
				true );
			AppLog.info( "Reading server seed list from file: " + hostsFileName,
				true );
			Host[] suspectedOthers = sListFile.readHostsFromFile();
			
			
			
			boolean haveHostList = false;
			
			while( ! haveHostList ) {	// keep trying to get seed host from user until valid connection made
				if( suspectedOthers == null ) {
					System.out.println();
					System.out.println( "Please enter information about a known server." );
					suspectedOthers = getServerFromUser();
					}
				// get a list of actual servers from one of the seed servers
				AppLog.info( "Trying to get a list of live servers.", true );
				Host[] knownOthers = lookForLiveServer( suspectedOthers );
				
				if( knownOthers != null ) {
					propagator = new ServerPropagator( knownOthers, sListFile, localAddress, this );
					AppLog.info("Writing live server list to file (" +
						hostsFileName + ") for future use.", true );
					sListFile.writeHostsToFile( knownOthers );
					haveHostList = true;
					}
				else {
					AppLog.info("No servers from provided host list found! " +
						"(We're alone in this cold world.)", true);
					suspectedOthers = null;
					haveHostList = false;
					}
				}
			}
		else {		// new chain, don't look for server to receive server list from
			AppLog.info( "Server will be first node in a new server chain.",
				true );
			propagator = new ServerPropagator( sListFile, localAddress, this );
			}
			
		}
	

	// overrides the ErrorHandlingThread runMethod() method

	public void runMethod() {
		ServerSocket servSock;
		try {
			servSock = new ServerSocket( localAddress.getPort(), maxConnections );
			}
		catch( IOException e ) {
			AppLog.error( "Server socket failed to initialize.", true );
			ErrorHandler.report( e );
			return;
			}
		while( true ) {
			try {
				AppLog.info( getTimeStamp() + ": Listening for connections.",
					true );
				Socket sock = servSock.accept();
				AppLog.info( getTimeStamp() + ": Connection received.",
					true );
				
				// connection received
				// pass it to receiver thread
				ReceiveThread rThread = new ReceiveThread( sock );
				
				rThread.start();
				}
			catch( IOException e ) {
				AppLog.error( "Connection broken on server accept.", true );
				ErrorHandler.report( e );
				}
			}
		}	// end runMethod()
		

		
	/**
	 * Processes command line arguments and constructs a <code>Server</code> at launch.
	 */
	public static void main( String[] args ) {
		// retrieve user arguments and start the server thread
		
		Thread.currentThread().setName( "Server.main" );
		
		LogManager.setApplicationName( "konspireServer" );
		
		Host serverLocalHost;
		
		String sAddress = "";
		int sPort = 5443;
		int sMaxConnect = 100;
		
		boolean newChain = false;	// should this server start up alone (w/o contacting others) ?
		
		
		if( args.length == 0 ) {
			// no defaults
			usage();
			}
		else if( args.length < 2 ) usage();
		else {
			// number of args <= 2
			try{
				sPort = Integer.parseInt(args[0]);
				sMaxConnect = Integer.parseInt(args[1]);
				if( args.length >= 3 ) {
					int newChainInt = Integer.parseInt( args[2] );
					if( newChainInt == 1 ) {
						newChain = true;
						}
					}
				}
			catch( NumberFormatException e ) {
				usage();
				}
			
			try {
				InetAddress iAddress = InetAddress.getLocalHost();
				sAddress = iAddress.getHostAddress();
				}
			catch( UnknownHostException e ) {
				AppLog.error( "Failed to retrieve local host IP address.", true );
				ErrorHandler.report( e );
				System.exit(1);
				}
				
			AppLog.info( "Server at " + sAddress + 
						":" + sPort + 
						", MaxConnect: "+ sMaxConnect + "\n", true );
			}
		
		serverLocalHost = new Host( sAddress, sPort );
		
		AppLog.info( "Initializing server.", true );
		Server s = new Server( serverLocalHost, sMaxConnect, newChain );
		
		
		AppLog.info( "Starting server.", true );	
		s.start();
		
		System.out.println( "Enter q to shutdown server nicely." );
		System.out.flush();
		
		try {
			while( System.in.read() != 'q' ) {
				// wait for user quit signal	
				}
			}
		catch( IOException e ) {
			AppLog.error( "Exception occurred while reading user keyboard input: " + 
				e.toString(), true );
			ErrorHandler.report( e );
			}
		
		s.shutdown();
		}	// end main()
	
	
	/**
	 * Gets a <code>String</code> representation of the current date and time.
	 *
	 * @return <code>String</code> representing current date and time
	 */
	public String getTimeStamp() {
		Date current = new Date();
		return current.toString();
		}
	
	
	
	/**
	 * Kills the server.
	 */
	public void shutdown() {
		AppLog.info( "Shutting down server.", true );
		
		// propagate a going down message
		
		ServerGoingDown m = new ServerGoingDown( localAddress );
		
		AppLog.info( "Sending notification to server chain.", true );
		propagator.propagateMessage( m );
		
		AppLog.info( "The server is halted.", true );
		try {
			sleep( 2000 );
			}
		catch( InterruptedException ie ) {
			}
		// exiting here seems to cause problems because some threads are still running...
		//System.exit(0);
		}
	
	/**
	 * Prompts the user to enter a seed server.
	 *
	 * @return an array of <code>Host</code> entered by the user, or <code>null</code>
	 * if retrieving input from user fails
	 */
	protected Host[] getServerFromUser() {
		try {
			BufferedReader userInput = new BufferedReader( new InputStreamReader( System.in ) );
			
			boolean hostNameRead = false;
			String hostName = "";
			
			while( ! hostNameRead ) {
				System.out.print( "Enter host name (or IP address): " );
				System.out.flush();
				
				String tempLine = null;
				
				tempLine = userInput.readLine();
				if( tempLine == null ) {
					AppLog.error( "Keyboard input isn't working.", true );
					return null;
					}

				try {
					StringTokenizer hostTokenizer = new StringTokenizer( tempLine );
					hostName = hostTokenizer.nextToken();
					
					hostNameRead = true;
					}
				catch(  NoSuchElementException nsee ) {
					System.out.println("Invalid input string: " + tempLine );
					}
				}			

			int port = 0;
			boolean portRead = false;
			while( !portRead ) {
				System.out.print("Enter port number: " );
				System.out.flush();
				
				String tempLine = userInput.readLine();
				if( tempLine == null ) {
					AppLog.error( "Keyboard input isn't working.", true );
					return null;
					}

				try{
					port = Integer.valueOf( tempLine ).intValue();
					portRead = true;
					}
				catch( NumberFormatException e ) {
					System.out.println("Port number must be an integer.");
					}
				}		
				
	    	Host [] returnHostArray = new Host[1];
	    	returnHostArray[0] = new Host( hostName, port );
			
			return returnHostArray;
			}
		catch( IOException e ) {
			AppLog.error( "Exception occurred while reading user keyboard input: " + 
				e.toString(), true );
			ErrorHandler.report( e );
			}
		return null;
		}
	
	
	
	/**
	 * Queries a list of suspected servers until a live one is found.<BR>
	 * Live server should send current list of live servers.
	 *
	 * @param suspectedOthers array of <code>Host</code>s that are believed to be alive.<BR>
	 * 
	 * @return array of live <code>Host</code>s that are returned by a live server, or 
	 * <code>null</code> if no live server connection is established
	 */
	protected Host[] lookForLiveServer( Host[] suspectedOthers ) {
		
		boolean found = false;
		NewServerUp m = new NewServerUp( localAddress );
		int i=0;
		
		ServerList sList = null;
		
		while( !found && i< suspectedOthers.length ) {
			
			Host h = suspectedOthers[i];
			if( ! h.equals( localAddress ) ) {	// make sure we don't try ourselves
				try {
					AppLog.info( "Trying to connect to " + h.toString(), true );
					CheckedMessageSocket sock = new CheckedMessageSocket( h );
					
					AppLog.info( "Asking server " + h.toString() + " for live "
						+ "server list.", true );
					sock.send( m );
					
					Message received = sock.receive();
					
					if( received instanceof ServerList ) {
						sList = (ServerList) received;
						found = true;
						AppLog.info( "Live server list received.", true );
						}
					else {
						AppLog.error( "Bad message type (expecting ServerList) "
							+ "received from host " + h.toString(), true );
						AppLog.info( "Perhaps a new version of konspire is "
							+ "available.", true );
						}	
					
					sock.close();
					}
				catch( IOException e ) {
					AppLog.error( "Connection to host failed: " + h.toString(),
						true );
					ErrorHandler.report( e );
					}
				}
			i++;
			}
		
		if( found ) {
			Vector hostVector = sList.getHostVector();
			
			localAddress = sList.getReceiverCorrectAddress();
			
			int size = hostVector.size();
			if( size != 0 ) {
				Host [] returnArray = new Host[ size ];
				for( int j=0; j<size; j++ ) {
					returnArray[j] = (Host)( hostVector.elementAt(j) );
					}
				return returnArray;
				}
			
			return null;
			}
		return null;
		}	// end lookForLiveServer()	
		
	
	
	
	
	
	
	
	/**
	 * Prints usage help message and exits.
	 */
	private static void usage() {				// print usage help message and exit

		System.out.print("usage:\n");
		System.out.print("Server port_number max_connect [new_chain]\n");
		
		System.out.print("examples:\n\t");
		System.out.print("Server 5433 15\n\t");
		System.out.print("Server 5433 15 0\n\t");
		System.out.print("Server 5433 15 1\n\t");
		
		System.exit(1);
		}
	private static long nextReceiveThreadID = 0;

	/**
	 * Increments <code>nextReceiveThreadID</code> in a thread-safe manner.
	 *
	 * @return the value of the ID before it was incremented.
	 */
	private synchronized static long incrementReceiveThreadID() {
		long oldID = nextReceiveThreadID;
		nextReceiveThreadID++;
		return oldID;
		}

	private static long nextClientThreadID = 0;
	
	private static long nextFileHostHandlerID = 0;

	/**
	 * Increments <code>nextClientThreadID</code> in a thread-safe manner.
	 *
	 * @return the value of the ID before it was incremented.
	 */
	private synchronized static long incrementClientThreadID() {
		long oldID = nextClientThreadID;
		nextClientThreadID++;
		return oldID;
		}

	/**
	 * Increments <code>nextFileHostHandlerID</code> in a thread-safe manner.
	 *
	 * @return the value of the ID before it was incremented.
	 */
	private synchronized static long incrementFileHostHandlerID() {
		long oldID = nextFileHostHandlerID;
		nextFileHostHandlerID++;
		return oldID;
		}
	
	
	/**
	 * <code>Thread</code> for receiving and processing initial messages from a 
	 * newly received socket connection.
 	 * <p>
 	 * This thread spawns other threads as appropriate.
 	 */
	class ReceiveThread extends NamedErrorHandlingThread {

		Socket sock;

		CheckedMessageSocket messageSock;


		
		/**
		 * Constructs a <code>ReceiveThread</code>.
		 *
		 * @param s <code>Socket</code> that this thread should receive messages on
		 */
		public ReceiveThread( Socket s ) {
			super( "ReceiveThread" + incrementReceiveThreadID() );
			sock = s;
			}
		
		
		// overrides the ErrorHandlingThread runMethod() method
		
		public void runMethod() {
			
			try {
				messageSock = new CheckedMessageSocket( sock );
				}
			catch( IOException e ) {
				ErrorHandler.report( e );
				AppLog.error("Connection broken while receive thread setting " 
					+ "up messsage socket.", true);
				return;
				}
			
			try {
				AppLog.info( "Waiting to receive message from new connection: "
					+ messageSock.getRemoteHost(), false );
					
				Message m = messageSock.receive();
				
				
				if( m instanceof ClientRequestConnect ) {
					AppLog.info( "Client connection request message received.",
						true );
					
					// let ClientHandler decide what to do
					ClientHandlerThread cThread = new ClientHandlerThread( messageSock, (ClientRequestConnect)m );
					cThread.start();
					return;	// return without closing socket
					}
				else {
					// If we're the first server receiving the NewServerUp, i.e., directly
					// from the new server...
					if( m instanceof NewServerUp ) {
						// need to make sure correct address is in this message.
						// sometimes (if behind a firewall, e.g.), a server
						// won't retrieve it's local address correctly
						
						//NOTE:  Only fix address, not port, since socket port won't
						// be same port that server is listening on
						
						NewServerUp newUp = (NewServerUp)m;
						
						Host hostInMessage = newUp.getServerAddress();
						Host hostInSocket = messageSock.getRemoteHost();

						if( !hostInMessage.getAddress().equals( hostInSocket.getAddress() ) ) {
							hostInMessage.setAddress( hostInSocket.getAddress() );
							// need to fix host in message
							m = new NewServerUp( hostInMessage );
							}
							
						

						// now send new server the server list
						// include the possibly corrected local address for the server
						ServerList sList = new ServerList( propagator.getHostVector(), hostInMessage );
						messageSock.send( sList );
						// don't propagate the new server up message here...
						// that is handled below
						
						// however, if server is already in the server list,
						// then we should remove it now before doing anything
						// else with the NewServerUp (to avoid having
						// the NewServerUp propagated to the server itself).
						AppLog.info("Removing server from propagation chain "
							+ "before handling a NewServerUp from this server." );
						propagator.removeServer( newUp.getServerAddress() );
						}	
					
					boolean needsPropagated = true;
					if( m instanceof PropMessage ) {
						PropMessage pMessage;
						if( m instanceof OneFaultPropMessage ) {
							AppLog.info( "Fault-tolerant propagated " +
								"message received.", true );
						
							pMessage = 
								(OneFaultPropMessage)m;
							// propagate to other servers
							propagator.faultTolerantPropagateMessage( 
								(OneFaultPropMessage)pMessage,
								messageSock );
							}
						else {	// non-fault-tolerant propagation message
							AppLog.info( "Propagated message received.",
								true );
						
							pMessage = (PropMessage)m;
							// propagate to other servers
							propagator.propagateMessage( pMessage );
							}
						// extract wrapped message and continue acting on it
						m = pMessage.getMessage();
						needsPropagated = false;
						}
					
					if( needsPropagated ) {
						// only non-fault tolerant propagate here...
						// server origninating message must initiate
						// a fault-tolerant propagation
						propagator.propagateMessage( m );
						}
					
					if( m instanceof NewServerUp ) {
						
						NewServerUp newUp = (NewServerUp) m;
						
						AppLog.info( "New server up message received:  " +
							newUp.getServerAddress(), true );
						
						// now add server to propagation list
						propagator.addServer( newUp.getServerAddress() );
						}
					
					else if( m instanceof ServerGoingDown ) {
						ServerGoingDown goDown = (ServerGoingDown) m;
						AppLog.info( "Server going down message received:  " +
							goDown.getServerAddress(), true );

						propagator.removeServer( goDown.getServerAddress() );
						database.removeFileHostsForServer(
											goDown.getServerAddress() );
						}
						
					else if( m instanceof FileHost ) {
						AppLog.info( "File host file list message received.",
							true );
						
						FileHost fHost = (FileHost) m;
						database.addFileHost( fHost );
						}
						
					else if( m instanceof ClientDisconnect ) {
						ClientDisconnect cDis = (ClientDisconnect) m;
						AppLog.info( "Client disconnect message received:  " +
							cDis, true );
						FileHost fHost = new FileHost( cDis.getClientAddress(), new Vector() );
						database.removeFileHost( fHost );
						}
					}
				AppLog.info( "ReceiveThread closing message socket.", true );
				messageSock.closeTotal();
				}
			catch( IOException e ) {
				AppLog.error( "Message receive failed in receive thread: " +
					e.toString(), true );
				ErrorHandler.report( e );
				}
			}	// end runMethod()

		}	// end ReceiveThread inner class	
	
	
	
	/**
	 * <code>Thread</code> for dealing with a client connection.
	 */
	class ClientHandlerThread extends NamedErrorHandlingThread {
	
		CheckedMessageSocket messageSock;
		
		Host connectedClient;
		
		ClientRequestConnect requestMessage;
		
		
		
		/**
		 * Constructs a <code>ClientHandlerThread</code>.
		 *
		 * @param mSock socket with connection to <code>Client</code>
		 * @param inRequestMessage message received from client requesting a connection
		 */
		public ClientHandlerThread( CheckedMessageSocket mSock, ClientRequestConnect inRequestMessage ) {
			super( "ClientHandler" + incrementClientThreadID() );
			messageSock = mSock;
			requestMessage = inRequestMessage;
			}
		
		
		// overrides ErrorHandlingThread runMethod() function
		
		public void runMethod() {
			Host hostInMessage;
			try {
				
				hostInMessage = requestMessage.getClientAddress();
						
				Host hostInSocket = messageSock.getRemoteHost();
				
				if( !hostInMessage.getAddress().equals( hostInSocket.getAddress() ) ) {
					// client didn't send its local address correctly
							
					//NOTE:  Only fix address, not port, since socket port won't
					// be same port that server is listening on
							
					hostInMessage.setAddress( hostInSocket.getAddress() );
					}
				
				// first, send server list
				ServerList sList = new ServerList( propagator.getHostVector(), hostInMessage );
				sList.setStats( database.getNumClients(), database.getNumFiles(), database.getNumBytes() );	// add stats to message
				
				messageSock.send( sList );
				}
			catch( IOException e ) {
				AppLog.error( "Client connection failed while sending server " +
					"list.", true );
				ErrorHandler.report( e );
				return;	
				}
			
			// send accept or reject
			if( numClientsConnected >= maxClients ) {
				try {
					// reject client
					ConnectResponse response = new ConnectResponse( false );
					
					messageSock.send( response );
					
					messageSock.closeTotal();
					}
				catch( IOException e ) {
					AppLog.error( "Client connection failed while sending " +
						"connection respones", true );
					ErrorHandler.report( e );
					return;	
					}
				}
			
			else {
				// there's room for client
				numClientsConnected++;
				
				// so we can remove client from DB on disconnect
				FileHost savedFileHost;
				
				ConnectResponse response = new ConnectResponse( true );
				
				try {
					messageSock.send( response );
					
					
					// now receive client file list
					
					Message m = messageSock.receive();
					
					if( m instanceof FileHost ) {
						FileHost fHost = (FileHost) m;
						
						connectedClient = fHost.getHost();
						
						Host hostInSocket = messageSock.getRemoteHost();
					
						// before propagating, add the address of this
						// server to the FileHost
						fHost.setServer( localAddress );
						
						savedFileHost = fHost;
						
						// spawn a thread to propagate file
						// host message and add it to the DB
						FileHostHandler fHostHandler = 
							new FileHostHandler( fHost );
						
						fHostHandler.start();
						}
					else {
						AppLog.error( "Client sent an out of order message.",
							true);
						AppLog.error( "Message: " + m.toString(), false );
						messageSock.closeTotal();
						AppLog.info( "Perhaps a new version of konspire is "
							+ "available.", true );
						return;
						}
					}
				catch( IOException e ) {
					AppLog.error( "Client connection failed while receiving "
						+ " FileHost message.", true );
					ErrorHandler.report( e );
					numClientsConnected--;
					return;	
					}
				
				
				// now go into main client loop
			
				boolean connected = true;
				try {
					while( connected ) {
						Message m = messageSock.receive();
						
						if( m instanceof FileHost ) {
							// a new copy of client file list arrived
							FileHost fHost = (FileHost) m;
							
							// before propagating, add the address of this
							// server to the FileHost
							fHost.setServer( localAddress );
							
							// spawn a thread to propagate file
							// host message and add it to the DB
							FileHostHandler fHostHandler = 
								new FileHostHandler( fHost );
							fHostHandler.start();
							}
						else if( m instanceof ClientDisconnect ) {
							ClientDisconnect cDis = (ClientDisconnect) m;
							database.removeFileHost( savedFileHost );
							
							// ClientDisconnects must be propagated
							// completely to maintain file index correctness
							// Use fault tolerant propagate.
							
							// specify receive socket as null since
							// we're originating the message
							propagator.faultTolerantPropagateMessage( cDis,
								null );
								
							messageSock.closeTotal();
							
							connected = false;
							}
						else if( m instanceof SearchQuery ) {
							SearchQuery sQ = (SearchQuery) m;
							Vector results = database.searchForStrings( sQ.getSearchTerms(), sQ.getNumResults(), connectedClient );
							SearchResults sR = new SearchResults( results, sQ );
							sR.setStats( database.getNumClients(), database.getNumFiles(), database.getNumBytes() );	// add stats to message
							messageSock.send( sR );
							}
						else if( m instanceof ServerListRequest ) {
							// client is requesting an updated server list
							// need to clone vector to force it to 
							// actually be sent across the wire again.
							Vector clonedServerVector =
								(Vector)( propagator.getHostVector().clone() );
							ServerList sList = new ServerList(
								clonedServerVector, 
								hostInMessage );
							// add stats to message
							sList.setStats( database.getNumClients(), 
								database.getNumFiles(), 
								database.getNumBytes() );	
							messageSock.send( sList );
							}
						else {
							AppLog.error( "Unknown message type received from "
								+ "client.", true );
							AppLog.error( "Message: " + m.toString(), false );
							
							AppLog.info( "Perhaps a new version of konspire is "
								+ "available.", true );
							}
						}
					}
				catch( IOException e ) {
					AppLog.error( "Client connection failed.", true );
					database.removeFileHost( savedFileHost );
					Host clientAddress = savedFileHost.getHost();
					// propagate using fault-tolerant scheme
					propagator.faultTolerantPropagateMessage(
						new ClientDisconnect( clientAddress ), null );
					}
				
				// client has disconnected if we get to this point
				numClientsConnected--;
				}	
			}	// end runMethod()

		
		/**
		 * Inner class <code>Thread</code>of <code>ClientHandlerThread</code> 
		 * that is spawned whenever a <code>FileHost</code> message comes
		 * in that needs to be propagated and added to the DB.
		 */
		private class FileHostHandler extends NamedErrorHandlingThread {
			
			private FileHost mFileHost;
			
			public FileHostHandler( FileHost inFileHost ) {
				super( "FileHostHandler" + incrementFileHostHandlerID() );
				mFileHost = inFileHost;
				}
				
			public void runMethod() {
				// propagate first, then add to DB
				// since DB add changes fhost object
				propagator.propagateMessage( mFileHost );
				database.addFileHost( mFileHost );
				}
			}	// end FileHostHandler inner-inner class


		}	// end ClientHandlerThread inner class
	
	
		
	}	// end Server class
