// Jason Rohrer
// ServerPropagator.java

/**
*
*	Class for propagating messages between servers
*
*	Created 5-2-2000
*	Mods:
*		Jason Rohrer	8-20-2000	Fixed bug in way downed servers are
*									detected.
*									Fixed a bug in addServer( Vector ) that
*									could potentially allow server to be listed
*									twice.
*		Jason Rohrer	8-23-2000	Added JLog logging.
*		Jason Rohrer	9-30-2000	Added support for a 1-simultaneous-fault
*									tolerant scheme.
*									
*/

package konspire.server;

import konspire.common.Message;
import konspire.common.Host;
import konspire.common.MessageSender;
import konspire.common.ServerListFile;

import konspire.common.Ack;
import konspire.common.CheckedMessageSocket;

import konspire.common.ServerGoingDown;

import java.io.IOException;

import konspire.common.Vector;
import java.util.Enumeration;

import konspire.common.log.AppLog;
import konspire.common.ErrorHandler;

/**
 * Handles propagating messages to other servers.
 *
 * @author Jason Rohrer
 */
public class ServerPropagator extends MessageSender {


	Vector hostVector = new Vector();
	
	Host self;
	
	Server parent;
	
	// for writing updated server lists out to file
	ServerListFile sListFile;
	
	/** 
	 * Constructs a propagator, specifying a list of other live hosts.
	 *
	 * @param sList list of other live <code>Host</code>s
	 * @param slf file where server list will be updated
	 * @param me the <code>Host</code> this propagator is running on (this <code>Host</code>)
	 * @param par the parent <code>Server</code> that this propagator is running on
	 */
	public ServerPropagator( Host sList[], ServerListFile slf, Host me, Server par ) {
		// deep copy
		for( int i=0; i<sList.length; i++ ) {
			// not *that* deep, just a separate list of Host pointers
			hostVector.addElement( sList[i] );
			}
		
		sListFile = slf;
		
		self = me;
		parent = par;
		hostVector.addElement( self );
		}
	
	/** 
	 * Constructs a propagator without specifying a list of known hosts.
	 *
	 * @param slf file where server list will be updated
	 * @param me the <code>Host</code> this propagator is running on (this <code>Host</code>)
	 * @param par the parent <code>Server</code> that this propagator is running on
	 */
	public ServerPropagator( ServerListFile slf, Host me, Server par ) {		
		sListFile = slf;
		
		self = me;
		parent = par;
		hostVector.addElement( self );
		}
	
	/**
	 * Adds a server to the list that this propagator is propagating message to.
	 *
	 * @param addition <code>Host</code> to add to the list
	 */
	public synchronized void addServer( Host addition ) {
		// first, make sure host vector doesn't already have this host
		
		boolean alreadyAdded = false;
		
		Enumeration e = hostVector.elements();
		while( e.hasMoreElements() && ! alreadyAdded ) {
			Object o = e.nextElement();
			if( addition.equals( o ) ) {
				alreadyAdded = true;
				}
			}
		if( ! alreadyAdded ) {	
			hostVector.addElement( addition );
			sListFile.writeHostsToFile( getHostArray() );
			AppLog.info("Adding server to prop chain: " + addition.toString() );
			}
		else {
			AppLog.info("Server already in prop chain: " + addition.toString() );
			}
		}
	
	
	/**
	 * Adds a vector of servers to the list that this propagator is propagating message to.
	 *
	 * @param serverVector vector of <code>Host</code>s to add to the list
	 */
	public synchronized void addServer( Vector serverVector ) {
		Enumeration e = serverVector.elements();
		while( e.hasMoreElements() ) {
			Host newHost = (Host)( e.nextElement() );
			addServer( newHost );
			}
		}
	
	
	/**
	 * Removes a server from the list that this propagator is propagating message to.
	 *
	 * @param removal <code>Host</code> to remove from the list
	 */
	public synchronized void removeServer( Host removal ) {
		hostVector.removeElement( removal );
		sListFile.writeHostsToFile( getHostArray() );
		AppLog.info("Removing server from prop chain " + removal.toString() );
		}
	
	
	/**
	 * Gets the vector of <code>Host</code>s that this propagator is propagating message to.
	 *
	 * @return vector of <code>Host</code>s that this propagator is propagating message to
	 */
	public synchronized Vector getHostVector() {
		return hostVector;
		}
	
	

	/**
	 * Passes on a message that is already in the middle of being propagated
	 *
	 * @param p the propagation message to pass on
	 */
	public void propagateMessage( PropMessage p ) {
		
		AppLog.info("Trying to propagate: " + p.getMessage().toString() );
		
		// mark ourself in this message, since we've received it
		p.markServer( self );
		
		boolean sent = false;
		
		while( !sent ) {
			Host hostToSend = p.getUnmarked();
			if( hostToSend != null ) {
				// Before attempting to propagate to a host listed in this message,
				// we should check that the host is still alive (actually in our
				// live server list).  If the host is not in our live server list,
				// this may be an older propagation message that was created before
				// the now-dead host went down (or at least before it was detected).
				if( ! hostVector.contains( hostToSend ) ) {
					// this host isn't alive in our book
					AppLog.warning( "Host listed in a prop message list that " +
								"isn't in our live server list: " + hostToSend );
					// mark this host in the prop message
					p.markServer( hostToSend );
					}
				else {
					// server in our live server list
					// try propagating message to server
					
					sent = sendMessage( p, hostToSend );

					if( !sent  ) {	// send failed
						// server dead??

						AppLog.warning( "Message WAS NOT propagated successfully to " +
							hostToSend.toString(), true );

						// First, remove server from the propagation chain
						removeServer( hostToSend );
					
						// send message to self about dead server
						// this will cause receive thread to start propagating
						// a server dead message
						AppLog.info( "Sending message to self about dead " 
												+ "server." );
						Host localAddress = parent.getLocalAddress();
						sendMessage( new ServerGoingDown( hostToSend ), localAddress );
						// this should avoid synchronization problems
						// plus, we may eventually want the server to tell clients about the
						// dead server....

						// mark the dead server in the current message
						// so that other servers won't try to propagate to it
						// ( They will discover that the server is dead when they get
						// our ServerGoingDown message )
						p.markServer( hostToSend );
						}
					else {	// sent successful
						AppLog.info("Propagated successfully to " +
							hostToSend.toString(), true );
						}	
					}
				}
			else {	// no hosts left to try propagating to
				AppLog.info("No hosts left to propagate message to: " + 
					p.getMessage().getClass().toString(), true );
				sent = true;
				}
			}
		}
	
	
	
	
	/**
	 * Propagates a message using a 1-simultaneous-fault tolerant message
	 * passing scheme.
	 *
	 * @param inMessage the propagation message to pass on.
	 * @param inReceiveSocket the socket that the message was received on
	 *   used to send acks as part of the scheme.  Set to <code>null</code>
	 *   if we're the server originating the message. 
	 */
	public void faultTolerantPropagateMessage( OneFaultPropMessage inMessage,
		CheckedMessageSocket inReceiveSocket ) {
		
		AppLog.info("Trying to one-fault-tolerant propagate: " + 
			inMessage.getMessage().toString() );
		
		// mark ourself in this message, since we've received it
		inMessage.markServer( self );
		
		boolean sent = false;
		
		// first, try to at least send message to another server
		// using a two-way message socket
		// after this succeeds, or if no servers are left, 
		// then we send an ack back through inReceiveSocet.
		
		// hold onto the socke that we use to propagate the message,
		// we will wait for an ack on this socket
		
		// use a checked message socket to guard against dead sockets
		// while waiting for the Ack.
		CheckedMessageSocket sendSocket = null;
		// hold onto the host that we're sending to also
		Host hostToSend = null;
		
		// should we wait for an Ack from the server we send to?
		boolean waitForAck = true;
		
		while( !sent ) {
			hostToSend = inMessage.getUnmarked();
			if( hostToSend != null ) {
				// Before attempting to propagate to a host listed in this message,
				// we should check that the host is still alive (actually in our
				// live server list).  If the host is not in our live server list,
				// this may be an older propagation message that was created before
				// the now-dead host went down (or at least before it was detected).
				if( ! hostVector.contains( hostToSend ) ) {
					// this host isn't alive in our book
					AppLog.warning( "Host listed in a prop message list that " +
								"isn't in our live server list: " + hostToSend );
					// mark this host in the prop message
					inMessage.markServer( hostToSend );
					}
				else {
					// server in our live server list
					// try propagating message to server
					
					try {
						sendSocket = new CheckedMessageSocket( hostToSend );
						sendSocket.send( inMessage );
						sent = true;
						}
					catch( IOException inIOE ) {
						sent = false;
						}

					if( !sent  ) {	// send failed
						// server dead??

						AppLog.warning( "Message WAS NOT propagated successfully to " +
							hostToSend.toString(), true );

						// First, remove server from the propagation chain
						removeServer( hostToSend );
					
						// send message to self about dead server
						// this will cause receive thread to start propagating
						// a server dead message
						AppLog.info( "Sending message to self about dead " 
												+ "server." );
						Host localAddress = parent.getLocalAddress();
						sendMessage( new ServerGoingDown( hostToSend ), localAddress );
						// this should avoid synchronization problems
						// plus, we may eventually want the server to tell clients about the
						// dead server....

						// mark the dead server in the current message
						// so that other servers won't try to propagate to it
						// ( They will discover that the server is dead when they get
						// our ServerGoingDown message )
						inMessage.markServer( hostToSend );
						}
					else {	// sent successful
						AppLog.info( "Propagated successfully to " +
							hostToSend.toString(), true );
						}	
					}
				}
			else {	// no hosts left to try propagating to
				AppLog.info( "No hosts left to propagate message to: " + 
					inMessage.getMessage().getClass().toString(), true );
				sent = true;
				waitForAck = false;
				}
			}	// end while loop waiting for first send to succeed
			
		// sent==true at this point.

		// send an Ack to the server that sent us the message
		// (if null, then we're the originating server, so skip sending Ack)
		if( inReceiveSocket != null ) {
			try {
				AppLog.info( "Sending Ack to the server that sent us this "
					+ "fault-tolerant propagation message." );
				inReceiveSocket.send( new Ack() );
				}
			catch( IOException inIOE ) {
				// log the exeption, but don't worry about it.
				// if socket was broken, we don't need to send an ack.
				AppLog.error( "Socket broken while trying to send Ack." );
				ErrorHandler.report( inIOE );
				}
			}

		// now wait for an ack from this server we sent to
		// if server goes down before ack arrives, 
		// send to another and wait for an ack
		boolean ackReceived = false;
		while( !ackReceived && waitForAck ) {

			try {
				AppLog.info( "Waiting to receive Ack from server we "
					+ "propagated to." );
				Message received = sendSocket.receive();
				if( received instanceof Ack ) {
					ackReceived = true;
					AppLog.info( "Received Ack.  Done with propagation." );
					}
				else {
					AppLog.error( "Received a bad message type when " +
						"expecting Ack:  " + received.getClass().toString() );
					throw new IOException( "Bad message type received." );
					}
				}
			catch( IOException inIOE ) {
				AppLog.error( "Socket broken while " +
					"trying to receive Ack." );
				ErrorHandler.report( inIOE );

				// mark host on list since Ack receive failed
				inMessage.markServer( hostToSend );

				AppLog.info( "Trying to send 1-fault-tolerant " +
					"message to another server on our list." );

				// try sending mess to another server on list
				sent = false;
				while( !sent ) {
					hostToSend = inMessage.getUnmarked();
					if( hostToSend != null ) {
						// Before attempting to propagate to a host listed in this message,
						// we should check that the host is still alive (actually in our
						// live server list).  If the host is not in our live server list,
						// this may be an older propagation message that was created before
						// the now-dead host went down (or at least before it was detected).
						if( ! hostVector.contains( hostToSend ) ) {
							// this host isn't alive in our book
							AppLog.warning( "Host listed in a prop message list that " +
										"isn't in our live server list: " + hostToSend );
							// mark this host in the prop message
							inMessage.markServer( hostToSend );
							}
						else {
							// server in our live server list
							// try propagating message to server

							try {
								sendSocket = new CheckedMessageSocket( hostToSend );
								sendSocket.send( inMessage );
								sent = true;
								}
							catch( IOException inIOEb ) {
								sent = false;
								}

							if( !sent  ) {	// send failed
								// server dead??

								AppLog.warning( "Message WAS NOT propagated successfully to " +
									hostToSend.toString(), true );

								// First, remove server from the propagation chain
								removeServer( hostToSend );

								// send message to self about dead server
								// this will cause receive thread to start propagating
								// a server dead message
								AppLog.info( "Sending message to self about dead " 
														+ "server." );
								Host localAddress = parent.getLocalAddress();
								sendMessage( new ServerGoingDown( hostToSend ), localAddress );
								// this should avoid synchronization problems
								// plus, we may eventually want the server to tell clients about the
								// dead server....

								// mark the dead server in the current message
								// so that other servers won't try to propagate to it
								// ( They will discover that the server is dead when they get
								// our ServerGoingDown message )
								inMessage.markServer( hostToSend );
								}
							else {	// sent successful
								AppLog.info( "Propagated successfully to " +
									hostToSend.toString(), true );
								}	
							}
						}
					else {	// no hosts left to try propagating to
						AppLog.info( "No hosts left to propagate message to: " + 
							inMessage.getMessage().getClass().toString(), true );
						sent = true;
						// stop waiting for an ack if we've exhausted
						// the propagation list
						waitForAck = false;
						}
					} // end while loop waiting for message to be sent

				} // end catch(IOException) for receiving Ack
			} // end while loop waiting for Ack received

		// ack received at this point, or we've run out of servers
		try {
			if( sendSocket != null ) {
				sendSocket.close();
				}
			}
		catch( IOException inIOE ) {
			AppLog.info( "1-fault tolerant send socket " +
				"broken during close." );
			ErrorHandler.report( inIOE );
			}
		} // end faultTolerantMessagePropagate
	
	
	// Note:  synchonized methods prevent a thread from
	// adding or removing a host from the host vector while
	// another thread is trying to start a new message
	
	
	/**
	 * Starts a propagation from a new message that hasn't been propagated yet.
	 *
	 * @param m the message to start propagating
	 */
	public void propagateMessage( Message m ) {
		// wrap message and propagate it
		PropMessage p = new PropMessage( getHostArray(), m );
		propagateMessage( p );
		}
	
	
	/**
	 * Starts a fault tolerant propagation from a new message that 
	 * hasn't been propagated yet.
	 *
	 * @param m the message to start propagating
	 */
	public void faultTolerantPropagateMessage( Message inMessage, 
		CheckedMessageSocket inSocket ) {
		
		// wrap message and propagate it
		OneFaultPropMessage p = new OneFaultPropMessage( getHostArray(), 
			inMessage );
		faultTolerantPropagateMessage( p, inSocket );
		}
	
	/**
	 * Gets array of <code>Host</code>s from host vector.
	 *
	 * @return array of <code>Host</code>s from host vector
	 */ 
	private synchronized Host [] getHostArray() {
		int size = hostVector.size();
		if( size != 0 ) {
			Host [] returnArray = new Host[ size ];
			for( int i=0; i<size; i++ ) {
				returnArray[i] = (Host)( hostVector.elementAt(i) );
				}
			return returnArray;
			}
		else {
			return null;
			}
		}
	}
