// Jason Rohrer
// StringTree.java

/**
*
*	Trinary search tree used to search for 
*	substring matches in a collection of strings
*
*	Note:  Only file names containing ASCII characters 
*	can be inserted.
*
*	(This constraint is to reduce size of the substring
*	pointer array.)
*
*	Created 5-4-2000
*	Mods:
*		Jason Rohrer	6-14-2000	Added support for filtering file results so
*									that no files on searching client are returned.
*		Jason Rohrer	8-23-2000	Added JLog logging.
*/


package konspire.server.database;

import konspire.common.FileDescriptor;
import konspire.common.Host;

import konspire.common.log.AppLog;
import konspire.common.ErrorHandler;

import konspire.common.Vector;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * Trinary search tree used to search for 
 * substring matches in a collection of strings.
 * <p?
 * Note:  Only file names containing ASCII characters 
 * can be inserted.
 *
 * @author Jason Rohrer
 */
public class StringTree {

	StringTreeNode root;
	
	int minTermLength = 3;	// min length of each search term
	
	
	// an array of 128 vectors, one for each ASCII character
	// Vectors contain pointers to StringTreeNodes that match each character
	int numChars = 128;
	Vector charPointer[];	
	
	
	// Thus, when trying to do substring matches (i.e., returning all file names
	// that contain a given string), you can follow each node pointer in the
	// first character's vector to find all spots in tree where the character
	// occurs.
	
	
	/**
	 * Constructs a <code>StringTree</code>.
	 */
	public StringTree() {
		charPointer = new Vector[numChars];
		for( int i=0; i<numChars; i++ ) {
			charPointer[i] = new Vector();
			}
		}
	
	
	
	/**
	 * Tells parent of tree that a new node has been added.
	 *
	 * @param node the node that has been added
	 */
	public void nodeAdded( StringTreeNode node ) {
		
		// parent indexes nodes by character for fast mid-tree access
		
		char nodeChar = node.getChar();
		
		try {
			// add pointer to this tree node to char array
			charPointer[(int)nodeChar].addElement( node );
			}
		catch( ArrayIndexOutOfBoundsException e ) {
			AppLog.detail( "File name contained a non-ASCII character: " + nodeChar + ", " + (int)nodeChar );
			ErrorHandler.report( e );
			}
		
		}
	
	
	/**
	 * Inserts FileDescriptor into this tree.
	 *
	 * @param f the descriptor to insert
	 *
	 * @return a StringTreeFileDescriptor describing where the descriptor was inserted
	 */
	public StringTreeFileDescriptor insertFile( FileDescriptor f ) {
		String name = f.getName();
		
		// convert string to lower case before adding to tree
		String lowerName = name.toLowerCase();
		
		
		if( root == null ) {
			// create new node at root
			root = new StringTreeNode( lowerName.charAt( 0 ), null, this );
			}
		return root.insertFile( lowerName, f );	
		}
	
	
	
	/**
	 * Removes all terms that are less than minLength long.
	 * Puts most infrequent search term first (thus, "zip" won't be first, most likely).
	 *
	 * @param searchTerms array of terms to strip
	 * @param minLength the shortest allowable search term
	 *
	 * @return new array of stripped search terms
	 */
	private String[] stripTerms( String [] searchTerms, int minLength ) {
		
		// first, count terms
		int numGoodTerms = 0;
		for( int i=0; i<searchTerms.length; i++ ) {
			if( searchTerms[i].length() >= minLength ) {
				numGoodTerms++;
				}
			}
		
		String[] newArray = new String[numGoodTerms];
		
		// now strip out good strings	
		int goodIndex = 0;
		for( int i=0; i<searchTerms.length; i++ ) {
			if( searchTerms[i].length() >= minLength ) {
				newArray[goodIndex] = searchTerms[i];
				goodIndex++;
				}
			}
		
		// next, make sure terms are unique
		Vector termVector = new Vector();
		
		for( int i=0; i<numGoodTerms; i++ ) {
			if( !termVector.contains( newArray[i] ) ) {
				termVector.addElement( newArray[i] );
				}
			}
		numGoodTerms = termVector.size();
		
		newArray = new String[ numGoodTerms ];
		for( int i=0; i<numGoodTerms; i++ ) {
			newArray[i] = (String)( termVector.elementAt(i) );
			}
			
		// now look for search term that seems the most infrequent in database
		// (i.e., terms likely to have fewest matches)
		
		// do this by looking at charPointer for first char of each term
		// term with fewest charPointers is our winner
		
		if( numGoodTerms > 1 ) {
			int numPointers = Integer.MAX_VALUE;
			
			int bestIndex = 0;
			for( int i=0; i<numGoodTerms; i++ ) {
				
				char firstChar = newArray[i].charAt( 0 );
				
				int thisNumPointers = charPointer[firstChar].size();
				
				if( thisNumPointers < numPointers ) {
					numPointers = thisNumPointers;
					bestIndex = i;
					}
				}
			if( bestIndex != 0 ) {	// swap needs to happen
				String temp = newArray[0];
				newArray[0] = newArray[bestIndex];
				newArray[bestIndex] = temp;
				}		
			}
			
		return newArray;		
		}
	
	
	
	/**
	 * Searches for terms in tree (find entries that match all terms).
	 * Results are filtered so that no files on searching client are returned.
	 * 
	 * @param searchTerms an array of terms to match in search
	 * @param numResults maximum number of results to return
	 * @param searchingClient the client performing this search
	 *
	 * @return a <code>Vector</code> of results
	 */
	public Vector searchForStrings( String[] searchTerms, int numResults, Host searchingClient ) {
		
		Vector returnVector = new Vector();
		boolean returnVecSeeded = false;	// has first set been added?
		
		String [] strippedTerms = stripTerms( searchTerms, minTermLength );
		
		int numSearchTerms = strippedTerms.length;
		
		
		boolean limitNumResults = false;	// limit results throughout each step of search
				
		// if only one term in list
		if( numSearchTerms == 1 ) {
			limitNumResults = true;
			}
		
		// Thus, we're first ignoring any terms shorter than minTermLength
		// If we only have one term remaining, we limit the number of results 
		// througout the search.
		// Thus, if someone searches for "zip", we won't go through the entire
		// database, but only through approximately the first numResult zip files
		
		// If search terms contain two or more minTermLength terms, the the
		// filtering by numResults will be more automatic
		
		// for each search term
		for( int i=0; i<numSearchTerms; i++ ) {
			
			// if this is the last term
			if( i == numSearchTerms - 1 ) {
				limitNumResults = true;
				// thus, we can stop searching immediately after we've found numResults matches
				}
			
			
			String s = strippedTerms[i];
					
			char firstChar = s.charAt( 0 );
			
			// iterate through all nodes in tree labeled firstChar
			
			Enumeration e = charPointer[firstChar].elements();
			
			Vector matches = new Vector();
			int numMatches = 0;
			
			try {
				// if in limited mode, make sure we haven't passed search results limit
				while( ( !limitNumResults || numMatches < numResults ) && e.hasMoreElements() )  {
					StringTreeNode node = (StringTreeNode)( e.nextElement() );
					Vector m = node.searchForString( s );
					
					if( m != null ) {
						Enumeration em = m.elements();
						while( em.hasMoreElements() ) {
							// Object o = em.nextElement();
							FileDescriptor o = (FileDescriptor)( em.nextElement() );
							
							
							// don't consider this match at all if it's a file that's on the
							//   client doing the searching
							
							if( ! o.getHost().equals( searchingClient ) ) {
							
								if( !returnVecSeeded ) {
									// only add to matches if not already in matches
									// (don't want multiple results for the same file)
									if( !matches.contains(o) ) {
										matches.addElement( o );
										numMatches ++;
										}
									}
								else {
									// only keep match if it also matches results from previous terms
									if( returnVector.contains( o ) ) {
										// only add to matches if not already in matches
										// (don't want multiple results for the same file)
										if( !matches.contains(o) ) {
											matches.addElement( o );
											numMatches ++;
											}
										}
									// thus, matches will be a subset of returnVector	
									}
								}	// end check against files on the searching host
							}	// end iteration over all elements in vector
						}	// end check against a null vector returned from search
					}	// end iteration over all search terms	
				}
			catch( NoSuchElementException ex ) {
				AppLog.error( "Iterating through search results vector failed." );
				ErrorHandler.report( ex );
				}
			
			// now have vector of matches to this search term
			
			// also, throughout this part of search, we've made sure matches is a subset of the latest returnVector
			// thus, intersection( matches, returnVector ) = matches 
			
			returnVector = matches;
			
			if( !returnVecSeeded ) {
				returnVecSeeded = true;
				}
			}	// end loop over search terms
		
		// more results found than desired	
		if( returnVector.size() > numResults ) {
	/*		System.out.println( "Natural num results = " + returnVector.size() 
								+ ", desired results = " + numResults );
			System.out.println( "Search String = ");
			for( int i=0; i< searchTerms.length; i++ ) {
				System.out.print( searchTerms[i] + ", " );
				}
			System.out.println( "Stripped Search String = ");
			for( int i=0; i< strippedTerms.length; i++ ) {
				System.out.print( strippedTerms[i] + ", " );
				}
	*/		
			// strip results down to size
			
			Vector newReturnVector = new Vector();
			for( int i=0; i<numResults; i++ ) {
				newReturnVector.addElement( returnVector.elementAt(i) );
				}
			return newReturnVector;
			}		
				
		return returnVector;
		}
		
	}
