// Jason Rohrer
// StringTreeNode.java

/**
*
*	Node in a trinary search tree used to search for 
*	substring matches in a collection of strings
*
*	Created 4-8-2000
*	Mods:
*		Jason Rohrer	5-4-2000	Code first completed
*/


package konspire.server.database;


import konspire.common.Vector;
import konspire.common.FileDescriptor;


/**
 * Node in a trinary search tree used to search for 
 * substring matches in a collection of strings. 
 *
 * @author Jason Rohrer
 */
public class StringTreeNode {

	char nodeChar;
	
	// for removing unused notes after a file is removed
	StringTreeNode parentNode;
	
	// for callbacks to parent of tree
	StringTree parentTree;
	
	StringTreeNode left;
	StringTreeNode right;
	StringTreeNode down;
	
	Vector fileVector = new Vector();		// vector of file descriptors

	/**
	 * Constructs a <code>StringTreeNode</code>.
	 *
	 * @param c <code>String</code> character this node will be marked with
	 * @param parNode node above this node in the tree
	 * @param parTree tree that contains this node
	 */
	public StringTreeNode( char c, StringTreeNode parNode, StringTree parTree ) {		
		nodeChar = c;
		
		left = null;
		right = null;
		down = null;
		
		parentNode = parNode;
		parentTree = parTree;
		
		// tell parent about this node
		parentTree.nodeAdded( this );
		}
	
	/**
	 * Gets character marking this node.
	 *
	 * @return character marking this node.
	 */
	public char getChar() {
		return nodeChar;
		}
	
	
	/**
	 * Gets file matches that must include this node's character (and anything after it).
	 * This query is achieved by moving down one position in the tree and then
	 * gathering all files below that point.
	 *
	 * @return <code>Vector</code> of <code>FileDescriptors</code> that match this node in the tree
	 */
	public Vector getFileMatchThis() {
		// recursively return vector of files that match including this char
		Vector returnFiles = new Vector( fileVector );	// add our files
		
		// add files below
		if( down != null ) {
			returnFiles.addAll( down.getFileMatchAny() );
			}
		
		return returnFiles;
		}
	
	/**
	 * Gets file matches below this point that don't have to match the character at this node.
	 * This query is achieved by gathering all files to the left, to the right, or down from
	 * this node in the tree.
	 * 
	 * @return <code>Vector</code> of <code>FileDescriptors</code> below this node in the tree
	 */
	public Vector getFileMatchAny() {	
		// recursively return vector of all files pointed to by this node
		Vector returnFiles = new Vector( fileVector );	// add our files
		
		// add files below
		if( down != null ) {
			returnFiles.addAll( down.getFileMatchAny() );
			}
		
		// add files to left and right
		if( left != null ) {
			returnFiles.addAll( left.getFileMatchAny() );
			}
		if( right != null ) {
			returnFiles.addAll( right.getFileMatchAny() );
			}
		
		return returnFiles;	
		}

	
	/**
	 * Gets a vector of <code>StringTreeFileDescriptors</code> from this 
	 * point in tree that start-match this partial string.<BR>
	 * ( "full" matches "full" and "fuller" ).
	 *
	 * @param s string that must be a prefixed of names for all returned files
	 *
	 * @return <code>Vector</code> of <code>StringTreeFileDescriptors</code> that
	 * start-match the specified string
	 */
	public Vector searchForString( String s ) {

		char nextChar = s.charAt( 0 );
			
		// if our char doesn't match next char in string
		if( nextChar != nodeChar ) {
			// pass search on to left or right
			if( nextChar < nodeChar ) {
				if( left != null ) {
					return left.searchForString( s );
					}
				else {
					// no chars left in tree that may match
					return null;
					}
				}
			else {
				if( right != null ) {
					return right.searchForString( s );
					}
				else {
					// no chars left in tree that may match
					return null;
					}
				}
			}
		else {
			// our char matches
			
			// if this last char
			if( s.length() == 1 ) {
				return getFileMatchThis();
				}
			// else still more chars remain to be matched
			else {
				// chop off this char, and pass down
				if( down != null ) {
					return down.searchForString( s.substring( 1 ) );
					}
				else {
					// no chars left in tree that may match
					return null;
					}
				}
			}
		}	// end searchForString()
	
	
	
	/**
	 * Inserts a <code>FileDescriptor</code> into the tree starting from this node.
	 * 
	 * @param name the name to use as descriptor is inserted into the tree
	 * @param f the descriptor to insert 
	 *
	 * @return a <code>StringTreeFileDescriptor</code> with a pointer to node where 
	 * the descriptor resides in the tree
	 */
	public StringTreeFileDescriptor insertFile( String name, FileDescriptor f ) {
		
		// keep chipping away at file name and passing it down recursively
		
		char nextChar = name.charAt( 0 );
			
		// if our char doesn't match next char in string
		if( nextChar != nodeChar ) {
			// pass insertion on to left or right
			if( nextChar < nodeChar ) {
				if( left == null ) {
					// add a new node to left
					left = new StringTreeNode( nextChar, this, parentTree );
					}
				return left.insertFile( name, f );	
				}
			else {
				if( right == null ) {
					// add a new node to right
					right = new StringTreeNode( nextChar, this, parentTree  );
					}
				return right.insertFile( name, f );	
				}
			}
		else {
			// our char matches
			
			// if this last char
			if( name.length() == 1 ) {
				// add FileDescriptor to vector
				fileVector.addElement( f );
				// return FileDescriptor wrapped as a StringTreeFileDescriptor
				return new StringTreeFileDescriptor( f, this );
				}
			// else still more chars remain to be matched
			else {
				// chop off this char, and pass down
				if( down == null ) {
					// add a new node down
					down = new StringTreeNode( nextChar, this, parentTree );
					}
				return down.insertFile( name.substring( 1 ), f );
				}
			}
		}	// end insertFile()
	
		
	/**
	 * Removes a <code>FileDescriptor</code> from this node.  If this node contains no other
	 * descriptors and is not pointing to other nodes, it will be removed from the tree.
	 * 
	 * @param f the descriptor to remove 
	 */
	public void removeFile( FileDescriptor f ) {
		fileVector.removeElement( f );
		checkAndRemoveNode();
		}	// end removeFile()
	
	
	/**
	 * Removes this node if it is no longer useful.<BR>
	 * Node is removed if it contains no file descriptors and points to
	 * no other nodes.
	 */
	protected void checkAndRemoveNode() {
		// if no files remain in this node
		if( fileVector.size() == 0 ) {
			// if this node points to nothing else
			if( left==null && right==null && down == null && parentNode != null ) {
				// remove this node from the tree	
				parentNode.removeNode( this );
				}
			}
		}
	
	
	/**
	 * Removes a node from set of nodes this node points to.
	 *
	 * @param node node to remove
	 */
	public void removeNode( StringTreeNode node ) {
		if( left == node ) {
			left = null;
			}
		else if( right == node ) {
			right = null;
			}
		else if( down == node ) {
			down = null;
			}
			
		// check if this node needs to be removed now.	
		checkAndRemoveNode();	
		}	// end removeNode()


	}	// end class StringTreeNode
