/*
 * A Python module interface to the block splitting code.
 *
 * This is not really C++ code, however it requires a C++ compiler on windows due to
 * C not allowing non-constant initializers in the FOO_type = {...} definitions.  [gcc
 * didn't have a problem with these]
 *
 * $Id: blocksharemod.c,v 1.6 2001/05/04 00:20:35 burris Exp $
 */

#ifdef __cplusplus
extern "C" {
#endif
#include "Python.h"
#include "share.h"
#include "share_local.h"
#ifdef __cplusplus
}
#endif

/*
 * Undefine this to exclude all embeded documentation (__doc__ fields).
 */
#define ONLINE_DOCS

/* Share types */

/**********************************************************************/
typedef struct {
  PyObject_HEAD
  SHARE share;
  SIZE_PARAMS sp;
  unsigned int msize, esize;  /* message and extra data size in bytes */
} share_obj;



/*
 * Initialize this share objects parameters, recompute msize and
 * esize, and reallocate the memory.  Returns self on success or
 * NULL on an error.
 *
 * NOTE: This will DESTROY any data currently in the share message or
 * extra packet fields as it will be reallocated to the new size.
 */
static share_obj * share_init_params(share_obj *self, int msg_pkts,
    int extra_pkts, int field_size, int num_segs)
{
  unsigned int msize, esize;
  u_int32 *new_m, *new_s;

  /* parameter errors are reported as a python error from within
   * init_params when it is compiled with PYTHON_MODULE defined */
  if (init_params(msg_pkts, extra_pkts, field_size, num_segs, &self->sp))
    return NULL;

  /* calculate the total message size in bytes */
  msize = msg_pkts * field_size * num_segs * sizeof(u_int32);

  /* calculate the total extra paket size in bytes */
  esize = extra_pkts * field_size * num_segs * sizeof(u_int32);

  /* allocate data for message and extra packets */
  new_m = (u_int32 *) malloc(msize);
  new_s = (u_int32 *) malloc(esize);
  if (!new_m || !new_s)
  {
    PyErr_SetString(PyExc_MemoryError,
        "Can't allocate memory for share data.");
    return NULL;
  }

  memset(new_m, 0, msize);
  memset(new_s, 0, esize);

  /* store the new values in the share */
  if (self->share.message)
    free(self->share.message);
  if (self->share.shares)
    free(self->share.shares);

  self->share.message = new_m;
  self->share.shares = new_s;
  self->msize = msize;
  self->esize = esize;

  return self;
}





/* share_obj methods */

static void share_obj_dealloc(share_obj *self)
{

  if (self->share.message)
    free(self->share.message);
  if (self->share.shares)
    free(self->share.shares);

  /* XXX these were annoyingly allocated within init_params() */
  if (self->sp.FieldEltToExp)
    free(self->sp.FieldEltToExp);
  if (self->sp.ExpToFieldElt)
    free(self->sp.ExpToFieldElt);

  PyMem_DEL(self);
}


static PyObject *share_obj_make_shares(share_obj *self, PyObject *args)
{
  if (!PyArg_NoArgs(args))
    return NULL;


  /* This could be compute intensive (and doesn't make any calls to
   * Py_ functions itself) so we allow other threads to continue.  */
  Py_BEGIN_ALLOW_THREADS;
  make_shares(&self->share);
  Py_END_ALLOW_THREADS;

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *share_obj_set_message(share_obj *self, PyObject *args)
{
  char *data;
  unsigned int datalen;


  if (!PyArg_ParseTuple(args, "s#", &data, &datalen))
  {
    return NULL;
  }

  /* the message -must- be the right size */
  if (datalen > self->msize)
  {
    PyErr_SetString(PyExc_ValueError,
        "Message data must be <= length reported by get_msize()");
    return NULL;
  }
  memcpy(self->share.message, data, datalen);

  /* zero-fill any extra space in the buffer */
  if (datalen < self->msize)
  {
    memset(((char*)self->share.message) + datalen, 0, self->msize - datalen);
  }


  Py_INCREF(Py_None);
  return Py_None;
}


/*
 * Return the data from the specified share number.  make_shares is
 * assumed to have already been called.
 */
static PyObject *share_obj_get_part(share_obj *self, PyObject *args)
{
  unsigned int sharenum, offset;
  PyObject *result_str;

  if (!PyArg_ParseTuple(args, "i", &sharenum))
  {
    return NULL;
  }

  offset = self->sp.pkt_len * sharenum;

  /* be good and do some bounds checking */
  if (offset + self->sp.pkt_len > self->esize)
  {
    PyErr_SetString(PyExc_ValueError,
        "Offset out of range for the given part number");
    return NULL;
  }

  result_str = PyString_FromStringAndSize(
      (char*) &(self->share.shares[offset/sizeof(self->share.shares[0])]),
      self->sp.pkt_len);

  if (!result_str)
  {
    PyErr_SetString(PyExc_MemoryError,
        "Can't allocate memory to return share part");
    return NULL;
  }

  return result_str;
}

static PyObject *share_obj_get_msize(share_obj *self, PyObject *args)
{
  if (!PyArg_NoArgs(args))
    return NULL;
  return Py_BuildValue("i", self->msize);
}


static PyMethodDef share_obj_methods[] = {
  {"make_shares",       (PyCFunction)share_obj_make_shares, 0,
#ifdef ONLINE_DOCS
   "make_shares() - Computes the various share parts.  Use get_part to\n"
   "retreive them.\n"
#endif
  },
  {"set_message",       (PyCFunction)share_obj_set_message, METH_VARARGS,
#ifdef ONLINE_DOCS
   "set_message() - Supply the message text for this share.  It must be\n"
   "exactly get_msize() bytes.\n"
#endif
  },
  {"get_part",          (PyCFunction)share_obj_get_part, METH_VARARGS,
#ifdef ONLINE_DOCS
   "get_part(part_no) - Return the data for the specified part\n\n"
   "This method assumes that make_shares has already been called.\n"
#endif
  },
  {"get_msize",         (PyCFunction)share_obj_get_msize, 0,
#ifdef ONLINE_DOCS
   "get_msize() - Returns the expected message size.\n"
#endif
  },
  {NULL,                NULL}   /* sentinel */
};

static PyObject * share_obj_getattr(share_obj *self, char* name)
{
  return Py_FindMethod(share_obj_methods, (PyObject *)self, name);
}

/**********************************************************************/

statichere PyTypeObject Share_type = {
#if ((PY_MAJOR_VERSION == 1) && (PY_MINOR_VERSION < 6))
        PyObject_HEAD_INIT(&PyType_Type)  /* python 1.5.2 */
#else
        PyObject_HEAD_INIT(NULL)
#endif
        0,                        /*ob_size*/
        "Share",                  /*tp_name*/
        sizeof(share_obj),        /*tp_size*/
        0,                        /*tp_itemsize*/
        /* methods */
        (destructor)share_obj_dealloc,  /*tp_dealloc*/
        0,                        /*tp_print*/
        (getattrfunc)share_obj_getattr, /*tp_getattr*/
        0,                        /*tp_setattr*/
        0,                        /*tp_compare*/
        0,                        /*tp_repr*/
        0,                        /*tp_as_number*/
        0,                        /*tp_as_sequence*/
        0,                        /*tp_as_mapping*/
        0,                        /*tp_hash*/
        0,                        /*tp_call*/
        0,                        /*tp_str*/
        0,                        /*tp_getattro*/
        0,                        /*tp_setattro*/
        0,                        /*tp_as_buffer*/
        0,                        /*tp_xxx4*/
        0,                        /*tp_doc*/
};


/*
 * Allocate a new python Share object, initializing its size
 * parameters and raising an exception on errors.
 */
static share_obj * new_share_object(int msg_pkts, int extra_pkts,
    int field_size, int num_segs, char* data, unsigned int datalen)
{
  share_obj *self;


  if (!(self = PyObject_NEW(share_obj, &Share_type)))
    return NULL;

  /* initialize self contents */
  memset(&self->share, 0, sizeof(self->share));
  memset(&self->sp, 0, sizeof(self->sp));
  self->share.sp = &self->sp;


  if (!share_init_params(self, msg_pkts, extra_pkts, field_size, num_segs))
  {
    goto err;
  }


  if (data)
  {
    /* the message -must- be the right size */
    if (datalen > self->msize)
    {
      PyErr_SetString(PyExc_ValueError,
          "Message data must be <= length reported by get_msize()");
      goto err;
    }
    /* If data is less than self->msize, the remainder of the share
     * will remain filled with nulls from share_init_params() */
    memcpy(self->share.message, data, datalen);
  }


  return self;

err:
  Py_DECREF((PyObject *)self);
  return NULL;
}


/**********************************************************************/
/* Share functions */

static PyObject * Share_new(PyObject *self, PyObject *args)
{
  share_obj *newself;
  char* data = NULL;
  unsigned int datalen = 0;
  int msg_pkts, extra_pkts, field_size, num_segs;
  
  if (!PyArg_ParseTuple(args, "iiii|s#", &msg_pkts, &extra_pkts,
        &field_size, &num_segs, &data, &datalen))
    return NULL;

  newself = new_share_object(msg_pkts, extra_pkts, field_size, num_segs,
      data, datalen);
  if (!newself)
    return NULL;

  return (PyObject *)newself;
}


/**********************************************************************/
typedef struct {
  PyObject_HEAD
  REBUILD rb;
  SIZE_PARAMS sp;
  unsigned int msize, esize;  /* message and extra data size in bytes */
} rebuild_obj;




/* rebuild_obj methods */

static void bs_rebuild_dealloc(rebuild_obj *self)
{
  /* free resources within self->rb */
  if (self->rb.pre_block)
    free(self->rb.pre_block);
  if (self->rb.post_block)
    free(self->rb.post_block);
  if (self->rb.rec_table)
    free(self->rb.rec_table);
  if (self->rb.rec_mark)
    free(self->rb.rec_mark);

  /* XXX these were annoyingly allocated within init_params() */
  if (self->sp.FieldEltToExp)
    free(self->sp.FieldEltToExp);
  if (self->sp.ExpToFieldElt)
    free(self->sp.ExpToFieldElt);

  PyMem_DEL(self);
}


static PyObject *bs_rebuild_share_in(rebuild_obj *self, PyObject *args)
{
  unsigned int pkt_index, packet_sz;
  char* packet;

  if (!PyArg_ParseTuple(args, "is#", &pkt_index, &packet, &packet_sz))
    return NULL;
  
  /* share_in raises Python errors itself when compiled with
   * PYTHON_MODULE defined */
  if (share_in(pkt_index, packet, packet_sz, &self->rb))
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


static PyObject * bs_rebuild_get_block(rebuild_obj *self, PyObject *args)
{
  if (!PyArg_NoArgs(args))
    return NULL;

  /* only return data if the rebuild is complete */
  if (self->rb.done)
  {
    PyObject *result_str;

    /* return msize bytes from rebuild_struct.post_block */
    result_str = PyString_FromStringAndSize(
        (char*) self->rb.post_block, self->msize);

    if (!result_str)
    {
      PyErr_SetString(PyExc_MemoryError,
          "Can't allocate memory to return reassembled block");
      return NULL;
    }

    return result_str;
  }

  /* alternative: we could raise an exception instead of returning None... */
  Py_INCREF(Py_None);
  return Py_None;
}


/*
 * Return a list of outstanding [not share_in'd] share packet numbers.
 * (note that not all packets are needed to rebuild the block)
 */
static PyObject * bs_rebuild_share_list(rebuild_obj *self, PyObject *args)
{
  int idx;
  unsigned int list_type = 0;  /* 0 = outstanding, 1 = received packets */
  PyObject *list = PyList_New(0);

  if (list == NULL)
    return NULL;

  if (!PyArg_ParseTuple(args, "|i", &list_type))
    return NULL;
  if (list_type > 1)
    list_type = 1;
  
  /* create the list of desirable share numbers */
  for (idx=0; idx < self->sp.rpkt_count; idx++)
  {
    /* if this share is needed, add the share # to the list */
    if (self->rb.rec_mark[idx] == list_type)
    {
      PyObject *pktnum = Py_BuildValue("i", idx);

      if ((pktnum == NULL) || (PyList_Append(list, pktnum) < 0))
      {
        Py_DECREF(list);
        return NULL;
      }
    }
  } /* for */

  return list;
}


static PyMethodDef rebuild_obj_methods[] = {
  {"share_in",          (PyCFunction)bs_rebuild_share_in, METH_VARARGS,
#ifdef ONLINE_DOCS
   "share_in(pkt_num, packet_data) - Inputs share number pkt_num\n"
#endif
  },
  {"get_block",         (PyCFunction)bs_rebuild_get_block, 0,
#ifdef ONLINE_DOCS
   "get_block() - Returns the rebuilt data string.\n"
   "Returns None if enough shares have not been input.\n"
#endif
  },
  {"share_list",        (PyCFunction)bs_rebuild_share_list, METH_VARARGS,
#ifdef ONLINE_DOCS
    "share_list([type]) - Returns a list of share numbers that have\n"
    "or have not been input using share_in.  The default is to list\n"
    "shares which have NOT been input.  Passing it a true parameter\n"
    "will cause it to list shares that have been input.\n"
#endif
  },
  {NULL,                NULL}   /* sentinel */
};

static PyObject * bs_rebuild_getattr(rebuild_obj *self, char* name)
{
  return Py_FindMethod(rebuild_obj_methods, (PyObject *)self, name);
}


/********* python "buffer" type access to the rebuilt data **********/
/** (buffer is an undocumented memory access type in python 1.5.2) **/

static int bs_rebuild_buffer_getreadbuf(rebuild_obj *self,
    int index, const void **ptr)
{
  if (index != 0)
  {
    PyErr_SetString(PyExc_SystemError,
        "Accessing non-existent rebuild segment");
    return -1;
  }
  if (!self->rb.done)
  {
    PyErr_SetString(PyExc_RuntimeError,
        "Not enough shares have been input for a rebuild");
    return -1;
  }
  if (!ptr)
  {
    PyErr_SetString(PyExc_SystemError,
        "NULL pointer passed to bs_rebuild_buffer_getreadbuf");
    return -1;
  }
  *ptr = self->rb.post_block;  /* pointer to rebuilt data */
  return self->msize;  /* bytes of rebuilt data */
}

static int bs_rebuild_buffer_getwritebuf(rebuild_obj *self,
    int index, const void *ptr)
{
  PyErr_SetString(PyExc_SystemError, "Rebuild buffers are read-only");
  return -1;
}

static int bs_rebuild_buffer_getlen(rebuild_obj *self, int *lenp)
{
  if (lenp)
  {
    *lenp = self->msize; /* bytes of rebuilt data */
  }
  return 1;
}

static PyBufferProcs bs_rebuild_as_buffer = {
  (getreadbufferproc)bs_rebuild_buffer_getreadbuf,
  (getwritebufferproc)bs_rebuild_buffer_getwritebuf,
  (getsegcountproc)bs_rebuild_buffer_getlen,
};


/**********************************************************************/

statichere PyTypeObject Rebuild_type = {
#if ((PY_MAJOR_VERSION == 1) && (PY_MINOR_VERSION < 6))
        PyObject_HEAD_INIT(&PyType_Type)  /* python 1.5.2 */
#else
        PyObject_HEAD_INIT(NULL)
#endif
        0,                        /*ob_size*/
        "Rebuild",                /*tp_name*/
        sizeof(rebuild_obj),   /*tp_size*/
        0,                        /*tp_itemsize*/
        /* methods */
        (destructor)bs_rebuild_dealloc,  /*tp_dealloc*/
        0,                        /*tp_print*/
        (getattrfunc)bs_rebuild_getattr, /*tp_getattr*/
        0,                        /*tp_setattr*/
        0,                        /*tp_compare*/
        0,                        /*tp_repr*/
        0,                        /*tp_as_number*/
        0,                        /*tp_as_sequence*/
        0,                        /*tp_as_mapping*/
        0,                        /*tp_hash*/
        0,                        /*tp_call*/
        0,                        /*tp_str*/
        0,                        /*tp_getattro*/
        0,                        /*tp_setattro*/
        &bs_rebuild_as_buffer,    /*tp_as_buffer*/
        0,                        /*tp_xxx4*/
        0,                        /*tp_doc*/
};

static rebuild_obj * new_rebuild_object(int msg_pkts, int extra_pkts,
    int field_size, int num_segs)
{
  rebuild_obj *self;

  if (!(self = PyObject_NEW(rebuild_obj, &Rebuild_type)))
    return NULL;

  /* initialize self contents */
  memset(&self->rb, 0, sizeof(self->rb));
  memset(&self->sp, 0, sizeof(self->sp));
  self->rb.sp = &self->sp;

  /* parameter errors are reported as a python error from within
   * init_params when it is compiled with PYTHON_MODULE defined */
  if (init_params(msg_pkts, extra_pkts, field_size, num_segs, &self->sp))
    return NULL;

  /* calculate the total message size in bytes */
  self->msize = msg_pkts * field_size * num_segs * sizeof(u_int32);

  /* calculate the total extra paket size in bytes */
  self->esize = extra_pkts * field_size * num_segs * sizeof(u_int32);

  /* allocate memory necessary for reassembly */
  self->rb.rec_table = (u_int32 *) calloc(extra_pkts, sizeof(u_int32));
  self->rb.rec_mark = (u_int8 *) calloc(extra_pkts, sizeof(u_int8));
  self->rb.pre_block = (u_int32 *) malloc(self->esize);
  self->rb.post_block = (u_int32 *) malloc(self->msize);
  if (!self->rb.rec_table || !self->rb.rec_mark ||
      !self->rb.pre_block || !self->rb.post_block)
  {
    PyErr_SetString(PyExc_MemoryError,
        "Can't allocate memory for rebuild data.");
    return NULL;
  }
  memset(self->rb.pre_block, 0, self->esize);
  memset(self->rb.post_block, 0, self->msize);

  return self;

err:
  Py_DECREF((PyObject *)self);
  return NULL;
}

/**********************************************************************/
/* Rebuild functions */

static PyObject * BS_Rebuild_new(PyObject *self, PyObject *args)
{
  rebuild_obj *newself;
  int msg_pkts, extra_pkts, field_size, num_segs;
  
  if (!PyArg_ParseTuple(args, "iiii", &msg_pkts, &extra_pkts,
        &field_size, &num_segs))
    return NULL;

  newself = new_rebuild_object(msg_pkts, extra_pkts, field_size, num_segs);
  if (!newself)
    return NULL;

  return (PyObject *)newself;
}


/**********************************************************************/

static PyMethodDef bs_module_functions[] = {
  {"Share",   (PyCFunction)Share_new, METH_VARARGS,
#ifdef ONLINE_DOCS
   "Share(msg_pkts, extra_pkts, field_size, num_segments, [data string])\n"
   " - Create a new share object\n\n"
   "Creates a new share object initialized to the given packet, field size\n"
   "and segmentation parameters.  The message data may optionally be\n"
   "given on initialization as well.\n"
#endif
  },
  {"Rebuild", (PyCFunction)BS_Rebuild_new, METH_VARARGS,
#ifdef ONLINE_DOCS
   "Rebuild(msg_pkts, extra_pkts, field_size, num_segments)\n"
   " - Create a new rebuild object\n\n"
   "Creates a new rebuild object initialized to the given packet, field size\n"
   "and segmentation parameters.\n"
#endif
  },
  {NULL,        NULL}   /* Sentinel */
};


/**********************************************************************/
#ifndef __APPLE__
#ifdef ONLINE_DOCS
char* module_doc =
"blockshare module quick reference:\n"
"(Object instances have their own __doc__ fields with a full description)\n\n"
"A Share object has the following methods:\n"
"  __init__(msg_pkts, extra_pkts, field_size, num_segs) - see Share.__doc__\n"
"  make_shares()\n"
"  get_msize()\n"
"  set_message()\n"
"  get_part(part_no)\n"
"\n"
"A Rebuild object has the following methods:\n"
"  __init__(msg_pkts, extra_pkts, field_size, num_segs) - see Rebuild.__doc__\n"
"  share_in(pkt_num, packet_data)\n"
"  share_list([type])\n"
"  get_block()\n"
"\n";
#else
char* module_doc = NULL;
#endif
#endif // __AAPLE__

/* Initialize this module. */

DL_EXPORT(void)
initblocksharemodule()
{
  PyObject *m;
  #ifdef __APPLE__
  m = Py_InitModule3("blockshare", bs_module_functions, "");
  #else
  m = Py_InitModule3("blockshare", bs_module_functions, module_doc);
  #endif
  /* no need to check the error here, the caller will do that */
}

DL_EXPORT(void)
initblockshare()
{
  initblocksharemodule();
}


/* vim:sw=2
 */
