#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#


### standard modules
import exceptions
import os
import sys
import tempfile
import types
import whrandom  # for testing

### our modules
import MojoConstants
import MojoErrors
import blockshare
import cryptutil
import debug
from human_readable import hr
import idlib
import mencode
import mojoutil
from confutils import confman


# This is a wrapper around blocksharemodule.so, which attempts to decrypt the shares before
# using them to reconstitute the block.  And symmetrically, it encrypts the shares when it
# generates them.


# `Share' and `Rebuild' are just wrappers around the C code of the same name.
# You really want to use the more abstract `split()' and `unsplit()' methods, below, which reduce the number of calls you have to make to disperse or regenerate chunks.
# The `Share' interface looks like this:
# init(); set_message(); make_shares(); get_part(); get_part(); get_part(); get_part(); get_part(); get_part(); get_part(); get_part()
# The `split' interface looks like this:
# split() # takes a message, returns a tuple of shares

class Rebuild:
    def __init__(self, *params):
        self._rebuild = apply(blockshare.Rebuild, params)

    def share_in(self, sharenum, blobdata):
        return self._rebuild.share_in(sharenum, blobdata)

    def share_list(self):
        return self._rebuild.share_list()

    def get_block(self):
        return self._rebuild.get_block()

class Share:
    def __init__(self, *params):
        self._sh = apply(blockshare.Share, params)

    def set_message(self, smap):
        return self._sh.set_message(smap)

    def make_shares(self):
        return self._sh.make_shares()

    def get_part(self, part):
        return self._sh.get_part(part)

def split(message, totalsharesneeded, totalsharesinpiece, keyseed=None, encrypt=cryptutil.desx_encrypt):
    """
    @param keyseed keying material;  If `None', keying material will be generated from the
        pieceid.  If empty string '', then '' will be used as keying material.

    @returns a sequence of length totalsharesinpiece of tuples of (shareId, share), where `shareId' is the mojosixbit encoded id;  (The Nth share in the sequence is the share with sharenum N.)

    XXX Why the constraint on K+M<=128?

    @precondition K + M must be less than 128.: totalsharesneeded + totalsharesinpiece <= 128: "totalsharesneeded: %s, totalsharesinpiece: %s" % (hr(totalsharesneeded), hr(totalsharesinpiece))
    """
    assert totalsharesneeded + totalsharesinpiece <= 128, "precondition: K + M must be less than 128." + " -- " + "totalsharesneeded: %s, totalsharesinpiece: %s" % (hr(totalsharesneeded), hr(totalsharesinpiece))

##    MINMSGSIZE = int(confman['PUBLICATION']['MIN_SHARE_SIZE']) * totalsharesneeded
##    if len(message) < MINMSGSIZE:
##        debug.mojolog.write("WARNING: MINMSGSIZE: %s, len(message) %s\n", args=(MINMSGSIZE, len(message),))

    # This function does two things: information dispersal (a.k.a. "splitting") on the message and encryption of the resulting shares.
    # It then takes the secure hash of each encrypted share to serve as that share's unique id.
    # The keys used to encrypt the shares are derived from the `keyseed' which is usually `None' and hence is derived from the pieceid.
    # The pieceid, is, of course, the secure hash of the piece.

    params = getshareparams(len(message), totalsharesneeded=totalsharesneeded, totalsharesinpiece=totalsharesinpiece)

    id = idlib.make_id(message, 'msg')

    if keyseed is None:
        keyseed = _gen_piece_keyseed(id)

    # debug.mojolog.write("blocksharez.split(%s, %s, %s, %s): len(message): %s\n", args=(idlib.make_id(message), totalsharesneeded, totalsharesinpiece, keyseed, len(message),), v=5, vs="debug")

##    tmpfname = os.path.join(tempfile.gettempdir(), "smsg_" + idlib.to_ascii(id))
##    f = open(tmpfname, "w")
##    f.write(message)
##    f.flush()
##    f.close()
##    f = None
##    del f

    s = apply(Share, params)
    s.set_message(message)
    s.make_shares()
    shares = []
    for sharenum in range(params[1]):
        share = s.get_part(sharenum)

        # Here we generate a unique keyseed for each share by hash-expanding the keyseed that
        # was passed in and the sharenum.
        thisshareskeyseed = _gen_share_keyseed(sharenum, keyseed)

        encshare = encrypt(share, thisshareskeyseed)
        shareid = idlib.to_mojosixbit(idlib.make_id(encshare, 'block'))

##        tmpfname = os.path.join(tempfile.gettempdir(), "sencshare_" + `sharenum` + "_" + shareid)
##        debug.mojolog.write("XXX just opened %s to write %s\n", args=(tmpfname, encshare,))
##        g = open(tmpfname, "w")
##        g.write(encshare)
##        g.flush()
##        g.close()
##        g = None
##        del g

        shares.append((shareid, encshare))

    return shares

def _gen_share_keyseed(sharenum, keyseed):
    """
    @precondition `keyseed' must be a string.: type(keyseed) == types.StringType: "keyseed: %s" % hr(keyseed)
    """
    assert type(keyseed) == types.StringType, "precondition: `keyseed' must be a string." + " -- " + "keyseed: %s" % hr(keyseed)

    canonicalseed = mencode.mencode({'sharenum': sharenum, 'keyseed': keyseed})

    return cryptutil.hashexpand(canonicalseed, MojoConstants.SIZE_OF_UNIQS)

def _gen_piece_keyseed(pieceid):
    """
    @precondition `pieceid' must be a string.: type(pieceid) == types.StringType: "pieceid: %s" % hr(pieceid)
    """
    assert type(pieceid) == types.StringType, "precondition: `pieceid' must be a string." + " -- " + "pieceid: %s" % hr(pieceid)

    canonicalseed = mencode.mencode({'pieceid': idlib.to_mojosixbit(pieceid)})

    return cryptutil.hashexpand(canonicalseed, MojoConstants.SIZE_OF_UNIQS)

class UnsplitError(MojoErrors.ReassemblyError): pass

def _internal_unsplit(sharez, decrypt, keyseed, totalsharesneeded, totalsharesinpiece):
    """
    @param sharez a sequence of tuples of (sharenum, sharedata)
    @param keyseed keying material;  It must not be `None'.  If empty string '', then '' will be
        used as keying material.

    @returns the original block

    @precondition `totalsharesneeded' must be positive.: totalsharesneeded > 0: "totalsharesneeded: %s" % hr(totalsharesneeded)
    @precondition The `sharez' list must contain enough shares.: len(sharez) >= totalsharesneeded: "sharez: %s" % hr(sharez)
    @precondition `keyseed' must be a string.: type(keyseed) == types.StringType: "keyseed: %s :: %s" % (hr(keyseed), hr(type(keyseed)))
    """
    assert totalsharesneeded > 0, "precondition: `totalsharesneeded' must be positive." + " -- " + "totalsharesneeded: %s" % hr(totalsharesneeded)
    assert len(sharez) >= totalsharesneeded, "precondition: The `sharez' list must contain enough shares." + " -- " + "sharez: %s" % hr(sharez)
    assert type(keyseed) == types.StringType, "precondition: `keyseed' must be a string." + " -- " + "keyseed: %s :: %s" % (hr(keyseed), hr(type(keyseed)))

    # Note when adding encryption that you should use the length of a cleartext share not an encrypted share in this calculation.  --Zooko 2000-10-03
    # This assumes that piecelen is always equal to the length of one of the shares times totalsharesneeded;
    # it should be as long as we're using a normal information dispersal algorithm.
    piecelen=len(sharez[0][1])*totalsharesneeded

    # Get parameters needed to pass to the C blockshare module interface
    params = makerebuildparams(piecelen, totalsharesneeded=totalsharesneeded, totalsharesinpiece=totalsharesinpiece)
    # debug.mojolog.write("params: %s\n", args=(params,)) # XXX

    r = apply(Rebuild, params)
    for (sharenum, sharedata) in sharez[:totalsharesneeded]:
        # Here we generate a unique keyseed for each share by hash-expanding the keyseed that
        # was passed in and the sharenum.
        thisshareskeyseed = _gen_share_keyseed(sharenum, keyseed)

        sharedatadec = decrypt(sharedata, thisshareskeyseed)
        r.share_in(sharenum, sharedatadec)

        # debug.mojolog.write("Decrypted a share: sharedatadec == sharedata: %s\n", args=(sharedatadec == sharedata,), v=3, vs="debug")
    return r.get_block()

def unsplit(sharez, id, totalsharesneeded, totalsharesinpiece, length=None, keyseed=None):
    """
    @param sharez a sequence of tuples of (sharenum, sharedata)
    @param id the id of the original piece (that will be returned);  This is used as a checksum
        to be sure we got the original data back, and as a key if keyseed is None.
    @param totalsharesneeded N of for N of M info dispersal
    @param totalsharesinpiece M of for N of M info dispersal
    @param length the length of the data to check the id of and return; it must be <= the length of
        the reassembled data and greater than zero.  If None, the entire piece of data is returned.
    @param keyseed keying material;  If `None', keying material will be generated from the
        pieceid.  If empty string '', then '' will be used as keying material.

    @returns the original block

    @precondition The `sharez' list must contain enough shares.: len(sharez) >= totalsharesneeded: "sharez: %s" % hr(sharez)
    """
    assert totalsharesneeded > 0
    assert len(sharez) >= totalsharesneeded, "precondition: The `sharez' list must contain enough shares." + " -- " + "sharez: %s, total needed %s" % (hr(sharez), totalsharesneeded)

    if keyseed is None: # note: changing this to "if not keyseed:" breaks the interface.  None != '' here.  --Zooko 2000-10-27
        keyseed = _gen_piece_keyseed(id)

    # debug.mojolog.write("blocksharez.unsplit(%s, %s, %s, %s, %s, %s)\n", args=(map(lambda x: (hr(x[0]), hr(idlib.make_id(x[1]))), sharez), id, totalsharesneeded, totalsharesinpiece, length, keyseed,), v=5, vs="debug")

    for share in sharez:
        shareid = idlib.to_mojosixbit(idlib.make_id(share[1], 'block'))
##        tmpfname = os.path.join(tempfile.gettempdir(), "uencshare_" + `share[0]` + "_" + shareid)
##        f = open(tmpfname, "w")
##        # debug.mojolog.write("XXX just opened %s to write %s\n", args=(tmpfname, share[1],))
##        f.write(share[1])
##        f.flush()
##        f.close()
##        f = None
##        del f

    # Try unsplitting with the `decrypt()' provided in the param, and if that fails try again
    # with `cryptutil.dummy_decrypt()'.  (We try decryption first because everything has been
    # encrypted since 0.927, circa 2000/12/05 22:12:29.)
    try:
        piecedata = _internal_unsplit(sharez, decrypt=cryptutil.desx_decrypt, keyseed=keyseed, totalsharesneeded=totalsharesneeded, totalsharesinpiece=totalsharesinpiece)
        if length:
            piecedata = piecedata[:length]
##        tmpfname = os.path.join(tempfile.gettempdir(), "umsg" + idlib.to_ascii(idlib.make_id(piecedata, 'msg')))
##        f = open(tmpfname, "w")
##        f.write(piecedata)
##        f.flush()
##        f.close()
##        f = None
##        del f

        if idlib.equal(idlib.make_id(piecedata), id):
            return piecedata

        # Doesn't match hash?  Wrong data!  Try again below, without decryption...
    except:
        (t, v, tb) = sys.exc_info()
        debug.mojolog.write("Exception from `_internal_unsplit()', which we are ignoring since it might be caused by failing to decrypt shares, and we are about to try again with decryption: %s\n", args=(v,), v=3, vs="debug")
        pass

    piecedata = _internal_unsplit(sharez, decrypt=cryptutil.dummy_decrypt, keyseed=keyseed, totalsharesneeded=totalsharesneeded, totalsharesinpiece=totalsharesinpiece)
    if length:
        piecedata = piecedata[:length]

    # debug.mojolog.write("piecedata id: %s, id: %s, length: %s, keyseed: %s, sharez: %s\n" % (hr(idlib.make_id(piecedata, 'block')), hr(id), hr(length), hr(keyseed), hr(map(lambda x: idlib.make_id(x[1]), sharez)), ), v=0, vs="debug")
    if not idlib.equal(idlib.make_id(piecedata, 'block'), id):
        # Doesn't match hash?  Wrong data!  exception!
        raise UnsplitError, "piecedata doesn't match id: %s, length: %s, keyseed: %s, sharez: %s" % (hr(id), hr(length), hr(keyseed), (map(lambda x: (hr(x[0]), hr(idlib.make_id(x[1]))), sharez),),)

    return piecedata


#########################################################################
# Create parameters for Rebuild() given the length of data stored in a
# piece.  The data length must be a multiple of 128 bytes.
#########################################################################
def makerebuildparams(len, totalsharesneeded, totalsharesinpiece):
    assert type(totalsharesneeded) is types.IntType, "totalsharesneeded must be an integer"
    assert type(totalsharesinpiece) is types.IntType, "totalsharesinpiece must be an integer"
    if (len % 128) != 0:
        raise ValueError, "len %s must be a multiple of 128 bytes" % len
    bytes_per_fieldunit = 4    # 32-bit words are used in the C blocksharemodule
    fieldunits_per_field = 8   # 8*bytes_per_fieldunit length info-dispersal fields
    return (totalsharesneeded, totalsharesinpiece, 8, len/(bytes_per_fieldunit*fieldunits_per_field*totalsharesneeded))


#########################################################################
# Determines the minimum supported piece size to use for storing the
# given number of bytes and returns parameters appropriate for
# creating a Share() or Rebuild() object in a tuple.  Use apply to
# pass them as parameters when creating blockshare module objects.
#########################################################################
def getshareparams(len, totalsharesneeded, totalsharesinpiece, poweroftwo=1) :
    """
    @param poweroftwo causes len to be rounded up to the next higher power of two.  (for backwards compatibility)
    """
    if poweroftwo:
        len = get_poweroftwo_chunk_len(len, totalsharesneeded)
    elif len % 128 != 0:
        len = len + (128 - (len % 128))
    return makerebuildparams(len, totalsharesneeded=totalsharesneeded, totalsharesinpiece=totalsharesinpiece)


def get_poweroftwo_chunk_len(len, totalsharesneeded):
    """
    Return the rounded up length of a data chunk/piece that will be used to store data of length len.
    """
    MinPieceSize = int(confman['PUBLICATION']['MIN_SHARE_SIZE']) * totalsharesneeded
    power = MinPieceSize
    while 1:
        if len <= power:
            len = power
            break
        power = power * 2
    return len


#########################################################################
def test_split_and_unsplit_no_encryption():
    msg = str(whrandom.randint(0,3063096))
    msg = "How do you know she's a witch?  She turned me into a newt!  (I got bettah)" * 10
    msg = msg + ((4096 - len(msg)) * '~')
    msg_id = idlib.make_id(msg)
    shares = split(msg, encrypt=cryptutil.dummy_encrypt, totalsharesneeded=4, totalsharesinpiece=8)
    sharenum = 0
    sharemap = {}
    for shareid, sharedata in shares:
        sharemap[sharenum] = sharedata
        sharenum = sharenum + 1
    sharetuples = sharemap.items()
    output = unsplit(sharetuples[:4], msg_id, totalsharesneeded=4, totalsharesinpiece=8)
    assert msg == output

def test_split_and_unsplit_desx_encryption_default_keyseed():
    msg = str(whrandom.randint(0,3063096))
    msg = msg + "How do you know she's a witch?  She turned me into a newt!  (I got bettah)" * 10
    msg = msg + ((4096 - len(msg)) * '#')
    msg_id = idlib.make_id(msg)
    shares = split(msg, encrypt=cryptutil.desx_encrypt, totalsharesneeded=4, totalsharesinpiece=8)
    sharenum = 0
    sharemap = {}
    for shareid, sharedata in shares:
        sharemap[sharenum] = sharedata
        sharenum = sharenum + 1
    sharetuples = sharemap.items()
    output = unsplit(sharetuples[:4], msg_id, totalsharesneeded=4, totalsharesinpiece=8)
    assert msg == output

def test_split_and_unsplit_desx_encryption_default_keyseed_16_32():
    SHARES_NEEDED=16
    SHARES_TOTAL=32
    MINMSGSIZE = int(confman['PUBLICATION']['MIN_SHARE_SIZE']) * SHARES_NEEDED
    MINMSGSIZE=3000

    msg = str(whrandom.randint(0,3063096))
    msg = msg + "How do you know she's a witch?  She turned me into a newt!  (I got bettah)" * 10
    msg = msg + ((MINMSGSIZE - len(msg)) * '#')
    msg_id = idlib.make_id(msg)
    shares = split(msg, encrypt=cryptutil.desx_encrypt, totalsharesneeded=SHARES_NEEDED, totalsharesinpiece=SHARES_TOTAL)
    sharenum = 0
    sharemap = {}
    for shareid, sharedata in shares:
        sharemap[sharenum] = sharedata
        sharenum = sharenum + 1
    sharetuples = sharemap.items()
    rot = sharetuples
    rot = mojoutil.shuffleList(sharetuples)
    rot=rot[:SHARES_NEEDED]
    # rot.sort()
    output = unsplit(rot, msg_id, length=len(msg), totalsharesneeded=SHARES_NEEDED, totalsharesinpiece=SHARES_TOTAL)
    assert msg == output

def test_split_and_unsplit_short_data():
    msg = str(whrandom.randint(0,3063096))
    msg = msg + "How do you know she's a witch?  She turned me into a newt!  (I got bettah)" * 3
    msg_id = idlib.make_id(msg)
    shares = split(msg, encrypt=cryptutil.desx_encrypt, totalsharesneeded=4, totalsharesinpiece=8)
    sharenum = 0
    sharemap = {}
    for shareid, sharedata in shares:
        sharemap[sharenum] = sharedata
        sharenum = sharenum + 1
    sharetuples = sharemap.items()
    output = unsplit(sharetuples[:4], msg_id, length=len(msg), totalsharesneeded=4, totalsharesinpiece=8)
    assert msg == output

def test_split_and_unsplit_desx_encryption_with_keyseed():
    msg = str(whrandom.randint(0,3063096))
    msg = msg + "How do you know she's a witch?  She turned me into a newt!  (I got bettah)" * 10
    msg = msg + ((4096 - len(msg)) * '\0')
    msg_id = idlib.make_id(msg)
    shares = split(msg, encrypt=cryptutil.desx_encrypt, keyseed='eggs taste better!', totalsharesneeded=4, totalsharesinpiece=8)
    sharenum = 0
    sharemap = {}
    for shareid, sharedata in shares:
        sharemap[sharenum] = sharedata
        sharenum = sharenum + 1
    sharetuples = sharemap.items()
    try:
        unsplit(sharetuples[:4], msg_id, keyseed='eggs', totalsharesneeded=4, totalsharesinpiece=8)
        raise RuntimeError, "test should've failed to decrypt pieces"
    except:
        pass
    output = unsplit(sharetuples[:4], msg_id, keyseed='eggs taste better!', totalsharesneeded=4, totalsharesinpiece=8)
    assert msg == output

def test_makerebuildparams_are_backwards_compatible():
    # this test ensures that the new makerebuildparams is returning results compatible with the old hard coded defaults
    SharesPerPiece = 8
    assert makerebuildparams(4096, totalsharesneeded=4, totalsharesinpiece=8) ==   (4, SharesPerPiece, 8,    32)
    assert makerebuildparams(8192, totalsharesneeded=4, totalsharesinpiece=8) ==   (4, SharesPerPiece, 8,    64)
    assert makerebuildparams(16384, totalsharesneeded=4, totalsharesinpiece=8) ==  (4, SharesPerPiece, 8,   128)
    assert makerebuildparams(32768, totalsharesneeded=4, totalsharesinpiece=8) ==  (4, SharesPerPiece, 8,   256)
    assert makerebuildparams(65536, totalsharesneeded=4, totalsharesinpiece=8) ==  (4, SharesPerPiece, 8,   512)
    assert makerebuildparams(131072, totalsharesneeded=4, totalsharesinpiece=8) == (4, SharesPerPiece, 8,  1024)
    assert makerebuildparams(262144, totalsharesneeded=4, totalsharesinpiece=8) == (4, SharesPerPiece, 8,  2048)
    assert makerebuildparams(524288, totalsharesneeded=4, totalsharesinpiece=8) == (4, SharesPerPiece, 8,  4096)

def test_getshareparams():
    # this test guarantees that getshareparams default is backwards compatible with the old hardcoded table
    SharesPerPiece = 8
    assert getshareparams(20000, totalsharesneeded=4, totalsharesinpiece=8) == (4, SharesPerPiece, 8,  256)
    assert getshareparams(32768, totalsharesneeded=4, totalsharesinpiece=8) == (4, SharesPerPiece, 8,  256)
    assert getshareparams(32769, totalsharesneeded=4, totalsharesinpiece=8) == (4, SharesPerPiece, 8,  512)
    # this tests new non-power of two length functionality
    assert getshareparams(12799, totalsharesneeded=4, totalsharesinpiece=8, poweroftwo=0) == (4, SharesPerPiece, 8,  100)
    assert getshareparams(12800, totalsharesneeded=4, totalsharesinpiece=8, poweroftwo=0) == (4, SharesPerPiece, 8,  100)
    assert getshareparams(12801, totalsharesneeded=4, totalsharesinpiece=8, poweroftwo=0) == (4, SharesPerPiece, 8,  101)

def _inner_test_getshareparams(testlen, sharesneeded, sharestotal):
    MinPieceSize = int(confman['PUBLICATION']['MIN_SHARE_SIZE']) * sharesneeded
    assert testlen >= MinPieceSize, "this test will fail if given testlen < int(confman['PUBLICATION']['MIN_SHARE_SIZE']) * totalsharesneeded"
    params = getshareparams(testlen, totalsharesneeded=sharesneeded, totalsharesinpiece=sharestotal)
    #print '%d byte %d of %d params -> %s' % (testlen, sharesneeded, sharestotal, params)
    s = apply(Share, params)
    s.set_message('z'*testlen)
    s.make_shares()
    for i in range(sharestotal):
        part = s.get_part(i)
        if len(part) != testlen/sharesneeded:
            #print `part`
            raise RuntimeError, 'incorrect %d of %d sharing; generated %d byte shares for %d bytes data' % (sharesneeded, sharestotal, len(part), testlen)

def test_getshareparams_is_correct():
    _inner_test_getshareparams(4096, 4, 8)
    _inner_test_getshareparams(4096, 2, 6)
    _inner_test_getshareparams(8192, 8, 24)
    _inner_test_getshareparams(131072, 4, 12)
    _inner_test_getshareparams(1024*1024, 8, 24)
    _inner_test_getshareparams(2**20, 8, 16)
    _inner_test_getshareparams(2**20, 16, 32)


mojo_test_flag = 1
def run():
    import RunTests
    RunTests.runTests(["blocksharez"])

if __name__ == '__main__':
    run()
