
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "share.h"
#include "share_local.h"

#ifdef PYTHON_MODULE
#include "Python.h"
#endif

// Turn the shares we've received back into the original
// message block
//
int rebuild_block(REBUILD *rb) {
  int block_size;
  int row, col;
  u_int32 *C, *D, *E, *F;
  u_int32 ExpFE;
  u_int32 row_off, msg_ind, seg_ind, source_ind;
  u_int32 col_eqn, row_eqn;

  // de-reference for clarity, fingers, speed
  u_int32 *FieldEltToExp = rb->sp->FieldEltToExp;
  u_int32 *ExpToFieldElt = rb->sp->ExpToFieldElt;
  u_int32 *rec_table = rb->rec_table;
  u_int32 mpkt_count = rb->sp->mpkt_count;
  u_int32 topbit = rb->sp->topbit;
  u_int32 MultFieldMax = rb->sp->MultFieldMax;
  u_int32 field_size = rb->sp->gfexp;
  u_int32 seg_count = rb->sp->seg_count;
  u_int32 *pre_block = rb->pre_block;
  u_int32 *post_block = rb->post_block;
  
  if(rb->rec_packets < rb->sp->mpkt_count) {
#ifdef PYTHON_MODULE
    PyErr_SetString(PyExc_ValueError, "not enough packets to rebuild block");
#else
    do_error("not enough packets to rebuild block\n", __FILE__, __LINE__);
#endif
    return -1;
  }

  // make sure message block is cleared 
  block_size = mpkt_count * rb->sp->pkt_len;
  memset(rb->post_block, 0, block_size);

  // allocate storage for arrays used to build the determinant 
  C = (u_int32 *) calloc(mpkt_count, sizeof(u_int32));
  D = (u_int32 *) calloc(mpkt_count, sizeof(u_int32));
  E = (u_int32 *) calloc(mpkt_count, sizeof(u_int32));
  F = (u_int32 *) calloc(mpkt_count, sizeof(u_int32));

  if((C==0) || (D==0) || (E==0) || (F==0)) {
#ifdef PYTHON_MODULE
    PyErr_SetString(PyExc_MemoryError,
            "couldn't allocate memory for determinants");
#else
    do_error("out of memory", __FILE__, __LINE__);
#endif
    if (C) free(C);
    if (D) free(D);
    if (E) free(E);
    if (F) free(F);
    return -1;
  }

#ifdef PYTHON_MODULE
  // allow threads; we're in a compute intensive part which will be
  // hogging one CPU for a little while.
  Py_BEGIN_ALLOW_THREADS;
#endif

  // Compute the determinant of the matrix in the finite field 
  for (row = 0; row < mpkt_count; row++) {
    for (col = 0; col < mpkt_count; col++) {

      if (col != row) {
        C[row] += FieldEltToExp[rec_table[row]^rec_table[col]];
        D[col] += FieldEltToExp[row^col];
      };

      E[row] += FieldEltToExp[rec_table[row]^col^topbit];
      F[col] += FieldEltToExp[rec_table[row]^col^topbit];

    };
  };

  // Rebuild the message based on the matrix determinant and
  // the available shares
  // 
  for (row = 0; row < mpkt_count; row++) {
    row_off = row * field_size * seg_count;

    for (col = 0; col < mpkt_count; col++) {

      ExpFE = E[col] + F[row] - C[col] - D[row]
        - FieldEltToExp[rec_table[col]^row^topbit];

      if (ExpFE < 0)
        ExpFE = MultFieldMax-((-ExpFE) % MultFieldMax);

      ExpFE = ExpFE % MultFieldMax;

      for (row_eqn=0; row_eqn < field_size; row_eqn++) {
        // where the recovered information is going
        msg_ind = row_eqn*seg_count + row_off;
        for (col_eqn=0; col_eqn < field_size; col_eqn++) {

          // matrix determines where to get recovery info 
          if (ExpToFieldElt[ExpFE+row_eqn] & BIT[col_eqn]) {
            // where the recovered information is coming from
            source_ind = seg_count * (col_eqn + (col*field_size));

            // all of the segment follows the same pattern 
            for (seg_ind=0; seg_ind < seg_count; seg_ind++) {
               // printf("doing XOR\n");
               // fflush(stdout);
               post_block[seg_ind+msg_ind] ^= pre_block[source_ind];
               source_ind++;
            };
          };

        };
      };

    };
  };

#ifdef PYTHON_MODULE
  Py_END_ALLOW_THREADS;
#endif

  free(C);
  free(D);
  free(E);
  free(F);

  rb->done = 1;
  return 0;
}
