#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# The Broker is the single application that should be run in order to
# launch the client web server, and any other Mojo Nation (TM) servers that
# you wish to run on a machine.
#
# $Id: Broker.py,v 1.143 2001/06/09 16:39:10 zooko Exp $

# standard python modules
import os
import string
import whrandom
import time
import sys
import traceback
import signal
import code
import threading
import httplib
from cStringIO import StringIO

# The signal handling must be done before we import any Mojo modules
# because some start threads or do other stuff.  This still seems ineffectual.
# Greg and Neju's current theory is that if the main thread is waiting on a
# thread lock, signal handlers aren't called.
#
# This translates the SIGTERM and SIGINT signal sent by a normal "kill" into
# something we can catch to try and shutdown more cleanly.
#
class SignalError(StandardError): pass

def raise_exit_sig_handler(signum, frame):
    raise SignalError

# Try to install signal handlers if we are the main thread:
if threading.currentThread().getName() == 'MainThread': # Is there a better test than this?
    signal.signal(signal.SIGTERM, raise_exit_sig_handler)
    signal.signal(signal.SIGINT, raise_exit_sig_handler)

# our python modules
try:
    from loggedthreading import LoggedThread
except ImportError, e:
    print 'I could not import the Mojo Nation Python modules.'
    print 'Is the "Broker" executable still located in the "evil" directory?'
    print 'If not, then is does your EVILDIR environment variable contain the full pathname of the "evil" directory?'
    print 'If you continue to get this error after checking those settings, then perhaps the Mojo Nation Python modules have been compiled incorrectly.'
    print 'The error was: '
    traceback.print_exc()
    sys.exit(1)
import BlockServerEGTP
import CommStrat
import ContentTrackerEGTP
import DoQ
import IdTrackerEGTP
import MetaTrackerLib
import MojoConstants
from MojoConstants import true, false, DEBUG_MODE
import MojoToken
import MojoTransaction
import Paytool
import PublicationTrackerHandlers
from RelayServerHandlers import RelayServerHandlers
import SEXP
import TCPCommsHandler
from confutils import confman, BROKER_VERSION_STR, CRYPTOPP_VERSION_STR
import confutils
from contentlib import ContentTypeDefRepository
import debug
import ipaddresslib
import mojobootpage
import mojohttpd
import mojosixbit
import mojoutil
import randinit
import Passphrase
import unsplit
import FindBeforehandBlockFetcher
import BuddyList
import MojoMessage
import Presentation
import localblockstore
import archive
import unsplit
import MojoMod
import UserCommand
import ClientEGTPlib
from AssertApply import assert_apply_will_work


# Some constants:
DELAY_BETWEEN_WITHDRAWALS=10
DELAY_AFTER_FAILED_WITHDRAWAL=whrandom.randint(1000, 10000)
BASIC_DENOM = MojoToken.MAP_DENOM_TO_KEYID.keys()[0] ### !!! XXX Zooko: add denoms  --Zooko 2000/06/13
# AMOUNT_OF_WITHDRAWAL=10000
AMOUNT_OF_WITHDRAWAL=BASIC_DENOM
# AMOUNT_OF_WITHDRAWAL=BASIC_DENOM * 100

# Module-scoped variables:
# The BROKER reference is so that anyone can get a reference to the process' Broker object by doing "import Broker; b = Broker.BROKER".
BROKER = None # If a Broker is instantiated this refers to that instance.  There should be only one instance per process.

_startedflag = threading.Event()

def main(argv):
    """
    This is the main procedure to launch the MojoNation Broker.
    This function is called and passed sys.argv if this module is run as an application.

    Third party software may choose to:
    A) Call main with their own arguments.
    B) Instantiate a Broker object and use it directly.
    C) Launch the Broker as a standalone app.   The third party software is implemented as a MojoMod.
    D) Launch the Broker as a standalone app and use and IPC mechanism such as SOAP or COLA to access the Broker.
    """
    if confman.is_true_bool(['DUPLICATE_PROCESS_SANITY_CHECK']) and other_broker_exists():
        print "There is another Broker running with the same config directory."
        if confman.is_true_bool(['DUPLICATE_UI_SPAWN']):
            kludgey_show_other_broker()
        print "Stopped."
        return
    
    # there is one big try/except around this code for catching most
    # errors and nicely logging them as well as sleeping for a few seconds
    # to allow the user to see the error message on systems where the window
    # closes after it exits.
    errorcode = 0
    broker = None
    try:
        debug.mojolog.write('Cleaning up temporary directory.\n', vs='Broker', v=2)
        cleanup_mojo_tmp_dir()
    
        # Welcome Message:
        debug.stderr.write("Welcome to Mojo Nation!\n")
        debug.stderr.write("protocol version: %s\n" % MojoMessage.CURRENT_MOJO_VER)
        debug.stderr.write("broker version: %s\n" % BROKER_VERSION_STR)
        debug.stderr.write("python version: %s\n" % sys.version)
        debug.stderr.write("Crypto++ version: %s\n" % CRYPTOPP_VERSION_STR)

        # Show info for unit testing:
        check_testres()

        # Parse the commandline args:
        debug.mojolog.write('Parsing commandline arguments.\n', vs='Broker', v=0)
        (recoverdb, interact) = parse_argv(argv)
    
        # Initialize randomness:
        randinit.randinit()

        # Currently you must run the localhost gateway in order to accept the terms of service.
        if not confman.is_true_bool(("AGREED_TO_TERMS_OF_SERVICE",)) and not confman.is_true_bool(("YES_NO", "RUN_LOCALHOST_GATEWAY",)):
            # TODO tell non-web gateway users how to deal with terms of service properly
            debug.stderr.write("You must agree to the terms of service at least once.\n")
            debug.stderr.write("for now: enable the localhost web gateway and acknowledge it that way.\n")
            debug.stderr.write("in the future we'll make this easier even w/o the web gateway.\n")
            raise SystemExit, "user without localhost web gateway needs to accept the terms of service."

        # Instantiate a broker:
        debug.mojolog.write('Instantiating Broker.\n', vs='Broker', v=0)
        broker = Broker(recoverdb=recoverdb)

        debug.mojolog.write('Launching Components.\n', vs='Broker', v=0)
        broker.launch_mtms()
        broker.launch_all_components()
        broker.initialize_doq_tasks()
        broker.init_buddylist()

        debug.stderr.write("\n\n)))))))))))) Your Broker Is Up And Running ((((((((((((\n\n")

        _startedflag.set()
        if interact:
            broker.start_interact()

        broker.wait_for_exit()
        broker = None
    except KeyboardInterrupt:
        debug.mojolog.write("Broker caught a keyboard interrupt.\n")
        errorcode = 0
    except SignalError:
        debug.mojolog.write("Broker caught a SIGINT or SIGTERM.\n")
        errorcode = 0
    except:
        debug.mojolog.write("Broker caught a FATAL ERROR:\n")
        traceback.print_exc(file=debug.stderr)
        errorcode = 1

    if broker is not None:
        # If the broker reference still exists an exception must have been raised:
        broker.exit()

    debug.mojolog.write("deleteing our reference to broker object\n", v=3)
    del broker
    try:
        # force a garbage collector run here
        import gc
        debug.mojolog.write("calling gc.collect()\n", v=3)
        if DEBUG_MODE:
            gc.set_debug(gc.DEBUG_STATS)
        gc.collect()
    except ImportError:
        pass

    debug.mojolog.write('Exiting with errorcode %d\n' % errorcode)
    sys.exit(errorcode)



class Broker(ClientEGTPlib.BlobQueryInterface):
    """
    A Broker object represents a singular interface to the Mojo Nation Broker.
    Currently it is unwise to instantiate more than one per process.

    A quick summary of the Broker interface.
    
    Here is an example of the most straight forward use of the Broker:
    # Begin code example:
    broker = Broker()
    broker.launch_all_components()
    broker.wait_for_exit() # Blocks until Broker exits.
    # End code example.

    Notice that the Broker runs on a separate thread from that which instantiates
    and uses the Broker.  All of the Mojo Nation Broker software strives to use
    an asynchronous program flow model, as opposed to a synchronous (threaded)
    one.  It is suggested that third party software does the same.

    To interact with the Broker asynchronously, use the broker objects "doq" attribute.
    Call the doq's "add_task" method and pass it a python callable.  Here's an example
    of a task which prints the amount of Mojo in the Broker's stash roughly every 20
    seconds:

    # Begin code example:
    broker = Broker()
    broker.launch_all_components()

    def print_stash(broker):
        print broker.ServicesPT.get_amount_in_purse()
        broker.doq.add_task(func=print_stash, args=(broker,), delay=20)

    broker.doq.add_task(func=print_stash, args=(broker,), delay=20)
    broker.wait_for_exit()
    # End code example.

    A final caveat to the DoQ interface is that a DoQ task must not block or
    make blocking calls.  XXX


    NOTE: A higher-level synchronous model 
    As the SOAP interface is developed, an analogous python interface will develop.
    This interface will follow a synchronous model, ie: SOAP procedure calls will
    block until a response is available.

    NOTE: This interface should mirror the SOAP interface for simlicity of implementation, documentation, and use.
    (The SOAP interface is not yet complete.)
    """
    def __init__(self, recoverdb=None):
        """
        Construct a Broker instance.  It's most likely a bad idea to construct two in the same process.

        recoverdb is a flag passed to components which use BerkeleyDB 3.
        handlesigs if true installs our own signal handlers for a cleaner shutdown.
        """
        # Handle the module-scope broker reference:
        global BROKER
        if BROKER is not None:
            debug.mojolog.write("ACK!  More than one Broker instance created in this process.\n", vs='Broker', v=-1)
        BROKER = self

        self.recoverdb = recoverdb # Is None an acceptable default?

        ### Public Attributes:
        # "Tool" attributes (those which are general to all of the Mojo Nation code):
        self.mojolog = debug.mojolog
        self.doq = DoQ.DoQ()
        assert DoQ.doq is not None, "There should be a DoQ referenced by the global doq reference."
        self.confman = confman # Part of the attempt to unify all global stuff into this object.
        self.ServicesMTM = None
        self.TokenServerMTM = None
        self.ServicesPT = None

        # when a windows package without stdin/stdout, free the dummy one on
        # occasion to avoid leaks incase something manages to use print.
        if os.environ.has_key('__RUN_MOJO_BROKER_SILENT'):
            self.doq.add_task(self._flush_dummy_stdout, delay=30)

        # "Component" attributes (those that can be en/dis-abled from config and which can run independently of eachother:
        # These are None until the given component is "launched" with one of the launch_* methods: (except for module references)
        self.mojohttpd = mojohttpd # The web-gateway (later it will have a class-based interface rather than module-based).
        self.mojohttpd_is_running = false
        self.metatracker = None
        self.blobserver = None
        self.relay_server_handlers = None
        self.publicationagent = None
        self.content_tracker = None
        self.multiblockstore = None
        self.bblsubscriber = None
        self.downloadcache = localblockstore.BlockStore(
            path=os.path.expandvars(confman.dict["PATH"]["DOWNLOAD_CACHE_DIR"]),
            maxspace=(2**20)*mojoutil.longpopL(confman.dict["DOWNLOADED_BLOCK_DISK_CACHE_MEGABYTES"]),
            recoverdb=self.confman.is_true_bool(["BLOCKSERVER", "AUTO_RECOVER_DB"]),
            name="download cache", depth=2)

        # "Miscellaneous" attributes:
        self.__exitcondition = threading.Condition()
        self.__exitcondition.acquire() # Released when self.wait_for_exit is called.
        self.__interact_thread = None
        self.__is_shutting_down = false
        self.__is_shutdown = false
        self.modman = MojoMod.MojoModManager(self) # The Broker interface to MojoMods.
        self.starttime = time.time() # Useful for uptime statistics.

        ### Miscellaneous Initialization:
        self.load_bootpage()

        # rotate the mojolog once a day (only on win32 or if '--rotated-log' flag given; read the function yourself)
        secs_between_rotate = int(60*60*float(confman['MOJOLOG']['LOG_ROTATION_FREQUENCY_IN_HOURS']))
        if secs_between_rotate < 60:
            self.mojolog.write('ultrafast log rotation not allowed, forcing a 60 second minimum\n', vs='Broker')
            secs_between_rotate = 60
        self.doq.add_task(debug.rotate_mojolog, delay=secs_between_rotate, kwargs={'doq': self.doq, 'delay': secs_between_rotate})
        self.doq.add_task(debug.cleanup_old_mojologs, args=( max(2, mojoutil.intorlongpopL(confman['MOJOLOG']['MAX_ROTATED_LOG_AGE_IN_HOURS'])), '') )


    def _flush_dummy_stdout(self):
        """ used in the windows package where stdout doesn't exist """
        assert os.environ.has_key('__RUN_MOJO_BROKER_SILENT')
        self.doq.add_task(self._flush_dummy_stdout, delay=30)
        # create a new dummy (causing the old one to be freed)
        newdummy = StringIO()
        sys.stdout = newdummy
        sys.stderr = newdummy

    # General use methods:
    def log_error(self, etype=None, evalue=None, stacktrace=None):
        """
        Log an exception to mojolog.  If all three parameters are None (the default),
        the current exception is printed (as per traceback.print_exc).
        """
        if etype or evalue or stacktrace:
            traceback.print_exception(etype, evalue, stacktrace, file=self.mojolog)
        else:
            traceback.print_exc(file=self.mojolog)

    def wait_for_exit(self):
        self.__exitcondition.acquire()
        while not self.__is_shutdown:
            # Only wait if we aren't already shutdown:
            # (don't wait forever here as that seems to keep us from getting signals)
            self.__exitcondition.wait(2)
        self.__exitcondition.release()

    def exit(self):
        """
        This doesn't return until the Broker is completely shutdown.

        If the Broker is already shutdown, it returns quickly (after some
        lock twiddling).

        This notifies all threads which are blocking on the Broker
        "wait_for_exit" method.
        """
        self.__is_shutting_down = true
        try:
            if confman.is_true_bool(['YES_NO', 'RUN_LOCALHOST_GATEWAY']):
                mojohttpd.write_shutdown_intropage()
                mojohttpd.globaldict['shutdown'] = 1
        except:
            self.mojolog.write("Error writing shutdown intropage\n", vs='Broker')
            pass
        self.__exitcondition.acquire()
        if not self.__is_shutdown:
            # be kind (and help our own reputation), tell metatracker(s) that we're signing off
            self.ServicesMTM.send_goodbye_to_metatrackers()
            time.sleep(1)  # give the goodbye's a little time to be sent
            # Only shutdown if we haven't already done so.
            self.doq.add_task(self.__shutdown)
            # Only let others know if we haven't already shutdown:  (If we're already shutdown other's shouldn't be waiting.)
            self.doq.wait_until_done_shutting_down()
            self.__is_shutdown = true
            self.doq = None
            self.__exitcondition.notifyAll()
        self.__exitcondition.release()
        self.mojolog.write('Broker shutdown at %s.\n\n', args=(mojoutil.iso_utc_time(),), vs='Broker', v=0)
 
    def is_shutdown(self):
        """returns true if this broker has already been shutdown"""
        return self.__is_shutdown
    
    def __del__(self):
        debug.mojolog.write("Broker.__del__ called.\n")
        
    def __finish_shutting_down(self):
        """
        @precondition This method must be called on the DoQ.: self.doq.is_currently_doq()
        """
        self.mojolog.write("Broker finishing shutting down.\n")

        try:
            # This inner func (shutdown) is kinda silly, but it's a fun toy and it works.
            def shutdown(thing, attr='shutdown', self=self):
                debug.mojolog.write("Attempting to shutdown %s...\n", args=(thing,), v=3, vs="debug")
                def default_shutdown(thing=thing, attr=attr, self=self):
                    self.mojolog.write('Could not shutdown %s, no attribute %s found.\n', args=(thing, attr), vs='Broker', v=3)
                getattr(getattr(self, thing, thing), attr, default_shutdown)() # Call the shutdown attribute.
                setattr(self, thing, None) # Get rid of our reference
                debug.mojolog.write("Done attempting to shutdown %s.\n", args=(thing,), v=3, vs="debug")

            # MojoMods:
            shutdown('modman')

            shutdown('content_tracker')

            shutdown('bblsubscriber')

            shutdown('multiblockstore')

            # close our paytool to close the Purse db properly
            shutdown('ServicesPT')

            # stop serving
            shutdown('TokenServerMTM')
            shutdown('ServicesMTM')

            # let things finish up in other threads (XXX we should wait on
            # something here with a timeout instead of a blind sleep)

            # I propose we add a "register_shutdown_callback" method and
            # all interested components can call it from their Broker object
            # reference.  --Nejucomo 2000-11-16
            # time.sleep(2) # ? I think waiting til the DoQ is done does it!  --Zooko 2001-04-17

            # stop the DoQ
            self.doq.shutdown_slowly_and_gracefully()

        except:
            debug.stderr.write('exception while shutting down:')
            traceback.print_exc(file=debug.stderr)

    def __shutdown(self):
        """
        @precondition This method must be called on the DoQ.: self.doq.is_currently_doq()
        """
        assert self.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        self.mojolog.write("Broker shutting down.\n")

        try:
            self.TokenServerMTM.stop_listening()
            self.ServicesMTM.stop_listening()

            MetaTrackerLib.save_partyinfo_to_disk()

            self.doq.add_idle_task(self.__finish_shutting_down)
        except:
            debug.stderr.write('exception while shutting down:')
            traceback.print_exc(file=debug.stderr)

    def get_version_string(self):
        return BROKER_VERSION_STR
   
    def start_interact(self, banner='', readfunc=raw_input, local=None):
        """
        Starts the interactive Broker on a new thread.

        The arguments are the same as used in the standard code.interact function.
        """
        if local is None:
            local = {}
        self.__interact_thread = LoggedThread(target=self.__interact, args=(banner, readfunc, local))
        self.__interact_thread.start()

    def __interact(self, banner, readfunc, local):
        print "Welcome to the Mojo Nation Broker interactive mode."
        print "Warning: This is for experts only; use this feature at your own risk."
        local['broker'] = self
        code.interact(banner, readfunc, local)

    # The component launching methods:
    def launch_mtms(self):
        """
        This initializes the mtms (it also launches mojohttpd).
        """
        ## mojohttpd:
        # mojohttpd  must be done first, so that users who launch our
        # software and then open "http://mojonation.net/broker"
        # won't see our web-site wrongfully.  Also, if the user is using
        # pass-phrases this must be done before the MTM's
        # can be instantiated.

        # XXX I think that both of these reasons are obsolete, and I keep getting exceptions thrown from trying to use a servicemtm while initializing, so I'm going to try reversing this order of initialization here.  --Zooko 2001-04-23

        # Give mojohttpd a reference to this Broker object:
        self.mojohttpd.globaldict['broker'] = self

        self.kick_mojohttpd()

        # Now instantiate the two MTM's.  (The following function blocks if a user pass-phrase is required.)
        apply(self.initialize_mtms, self.mojohttpd.get_mtm_ids())

        assert self.ServicesMTM is not None
        assert self.TokenServerMTM is not None

        self.mojohttpd.set_servicemtm(self.ServicesMTM, self.TokenServerMTM)
        debug.stderr.write("Switching localhost web gateway to normal operation mode.\n", vs='Broker', v=3)
        self.mojohttpd.switch_builtin_request_handler_dict(self.mojohttpd.main_requestbuiltins)

    def launch_all_components(self):
        """
        Launches Mojo Nation Broker components (except mojohttpd).  This method
        launches components based on config file settings.  To launch a component regardless of
        config file settings, use its specific launch_* method  (ie: launch_mojohttpd()).
        """
        # Initialize the self.multiblockstore:
        self.__initialize_multiblockstore()

        # we're initialized; start loading modules
        self.doq.add_task(self.modman.autoload)
        
        MetaTrackerLib.load_partyinfo_from_disk()
        self.doq.add_task(self.periodic_save_metainfocache, delay=mojoutil.intorlongpopL(confman["METATRACKERLIB"]["SECONDS_BETWEEN_SAVE"]))

        self.start_or_stop_configured_servers()

    def periodic_save_metainfocache(self):
        self.doq.add_task(self.periodic_save_metainfocache, delay=mojoutil.intorlongpopL(confman["METATRACKERLIB"]["SECONDS_BETWEEN_SAVE"]))
        MetaTrackerLib.save_partyinfo_to_disk()

    def __initialize_multiblockstore(self):
        # convert from old single store config values into the always using the multiple stores list
        if not confman.is_true_bool(['BLOCKSERVER', 'USE_MULTIPLE_STORES'], default="true"):
            debug.mojolog.write("Converting of config file to use BLOCKSTORE.STORAGE_SPACE_LIST\n", vs='Broker')
            storedict = {
                'READONLY': 'false',
                'PATH': confman["PATH"]["BLOCK_SERVER_DIR"],
                'MAX_MEGABYTES': mojoutil.longpopL(confman["BLOCKSERVER"]["MAX_MEGABYTES_OF_DISK_TO_USE"]),
                'SUBDIRDEPTH': 3,
            }
            confman.setlist(['BLOCKSERVER', 'STORAGE_SPACE_LIST'], [storedict])
            del confman['PATH']['BLOCK_SERVER_DIR']
            del confman['BLOCKSERVER']['MAX_MEGABYTES_OF_DISK_TO_USE']
            if confman['BLOCKSERVER'].has_key('USE_MULTIPLE_STORES'):
                del confman['BLOCKSERVER']['USE_MULTIPLE_STORES']
            confman.save()
        if confman['BLOCKSERVER'].has_key('USE_MULTIPLE_STORES'):
            del confman['BLOCKSERVER']['USE_MULTIPLE_STORES']

        sclist = confman.getlist(("BLOCKSERVER", "STORAGE_SPACE_LIST"))
        # all blockstores must be at least 50 megs
        for storedict in sclist:
            if mojoutil.longpopL(storedict['MAX_MEGABYTES']) < 50:
                storedict['MAX_MEGABYTES'] = 50
                debug.mojolog.write("NOTE: raising blockstore %s's allocation to a minimum of 50 megs\n", args=(storedict['PATH'],), vs='Broker')

        if (len(sclist) > 1):
            debug.mojolog.write("Using Multiple BlockStores.\n", vs='Broker')

        blockstorelist = []
        for index, sc in mojoutil.irange(sclist):
            if string.lower(sc['READONLY']) in ("yes", "true"):
                bsclass = localblockstore.ReadOnlyBLBlockStore
            else:
                bsclass = localblockstore.BLBlockStore

            blockstorelist.append(
                bsclass(
                    path=os.path.expandvars(sc['PATH']),
                    maxspace=(2**20)*mojoutil.longpopL(sc['MAX_MEGABYTES']),
                    depth=mojoutil.intorlongpopL(sc['SUBDIRDEPTH']),
                    name="block storage #%02d %s" % (index, sc['PATH']))
                )

        debug.mojolog.write('Using %s block stores in the multiblockstore.\n', args=(len(blockstorelist),), vs='Broker', v=4)
        if len(blockstorelist) > 20:
            debug.mojolog.write("WARNING:  Using %s separate BlockStores, that's a lot!\n", vs="Broker")

        self.multiblockstore = localblockstore.MultiBlockStore(blockstorelist)
        self.bblsubscriber = localblockstore.BadBlockListSubscriber(broker=self, blblockstore=self.multiblockstore)
        
    def start_or_stop_configured_servers(self):
        self.kick_metatracker()
        self.kick_relayserver()
        self.kick_blockserver()
        self.kick_contenttracker()
        self.kick_publicationagent()

        # NOTE: this save() is -very- important as some of the servers above make additional changes to the config
        # as necessary during startup (to add idmask defaults, etc..)
        # this confman.save() is also important because mojohttpd relies on it to save the config file.
        confman.save()

    # launching methods for specific components:
    def launch_mojohttpd(self):
        """
        Starts the mojohttpd webserver and handles the various UI-related startup modes.
        Does not check config values to see if the user wanted the web-server launched.

        Note: This function blocks if the user has not accepted Terms-Of-Service.
        """
        # XXX ToDo: Improve mojohttpd interface;  As part of this improvement, move all the complicated
        # pass-phrase/terms-of-service/start-up logic into mojohttpd.  Provide a simple interface to
        # launching mojohttpd.

        # Next, immediately start the web-browser:
        self.__mojohttpd_thread = LoggedThread(target=self.mojohttpd.run, name='MojoHTTPD', kwargs={'doq': self.doq, 'port': mojoutil.intorlongpopL(confman.dict["CLIENT_PORT"]), 'broker': self})
        self.__mojohttpd_thread.setDaemon(1)
        self.__mojohttpd_thread.start()

        # If the config file requests this feature, launch the web browser:
        if confman.is_true_bool(('AUTO_LAUNCH_BROWSER',)):
            if (confutils.platform == 'win32' or os.environ.has_key('DISPLAY')) and (not os.environ.has_key('MOJO_DONT_LAUNCH_BROWSER')):
                self.mojohttpd.launch_browser()
            if(sys.platform == 'darwin1'):
                import pyobjc
                ws = pyobjc.lookup_class("NSWorkspace").sharedWorkspace()
                ws.openFile_(os.path.expandvars(confutils.confman.dict['PATH']['INTRO_PAGE_v2']))

        # Check that the user has agreed to the terms of service.  (Blocking call!)
        if not mojohttpd.wait_for_terms_of_service():
            # User rejected terms of service:
            raise SystemExit, "Terms of Service rejected, exiting..."

    def kick_mojohttpd(self):
        if confman.is_true_bool(('YES_NO', 'RUN_LOCALHOST_GATEWAY',)):
            self.launch_mojohttpd() # Blocks if Terms-Of-Service has not been accepted.
            self.mojohttpd_is_running = true
        else:
            if self.is_mojohttpd_required():
                self.mojolog.write('The web-gateway (mojohttpd) is required for required user input.  Please enable it and restart the Broker.\n')
                raise SystemExit, 'Web Gateway required for necessary user input.'

            # XXX this is a gross hack to init commandmanager even without calling mojohttpd.run()
            if not mojohttpd.globaldict.get('commandmanager'):
                debug.mojolog.write("Creating mojohttpd.globaldict['commandmanager'].\n", vs='Broker')
                mojohttpd.globaldict['commandmanager'] = UserCommand.MultiCommandManager(self.doq)


    def launch_metatracker(self):
        """
        Launches a meta tracker.
        """
        debug.stderr.write("Starting Meta/Id Tracker\n", vs='Broker')
        self.metatracker = IdTrackerEGTP.IdTrackerEGTP(MTM=self.ServicesMTM)

        # The bootpage path:
        bpp = self.confman.getpath(('PATH','BOOT_PAGE'), '/bootpage.txt')
        self.mojolog.write('Writing MetaTracker bootpage to %s.\n', args=(bpp,), vs='Broker')
        bpf = open(bpp, 'w')
        bpf.write(self.metatracker.get_bootpage_string())
        bpf.close()

        # If we are a central root metatracker, eat our own bootpage dog food
        if self.confman.is_true_bool(['METATRACKER', 'CENTRAL']):
            self.mojolog.write('We are a central MetaTracker: eating our own bootpage dog food\n', vs='Broker')
            bpd = mojobootpage.BootPage(self.metatracker.get_bootpage_string())
            bpd.use(self.confman)
            del bpd

        # run a /bootpage.txt http server if possible
        try:
            import bootpaged
            bps = bootpaged.BPServer(bppath=bpp, server_address=('', 7777))
            bpServerThread = LoggedThread(target=bps.serve_forever)
            bpServerThread.setDaemon(1)
            bpServerThread.start()
            debug.mojolog.write("/bootpage.txt HTTP server started on port 7777\n", vs='Broker')
        except ImportError:
            debug.mojolog.write("bootpaged module not found, not running /bootpage.txt HTTP server\n", vs='Broker')
        
        time.sleep(2) # Is this necessary?  If it is, should it be inside this method, or should the caller do it?

    def stop_metatracker(self):
        if hasattr(self, 'metatracker') and self.metatracker:
            debug.stderr.write("Stopping Meta/Id Tracker\n", vs='Broker')
            self.metatracker.shutdown()
            self.metatracker = None
    
    def kick_metatracker(self):
        ## metatracker:
        if confman.is_true_bool(('YES_NO', 'RUN_META_TRACKER',)):
            if self.metatracker is None:
                self.launch_metatracker()
        else:
            self.stop_metatracker()

    def launch_relayserver(self):
        """
        Launches a relay server.
        """
        debug.stderr.write("Starting Relay Server\n", vs='Broker')
        self.relay_server_handlers = RelayServerHandlers(self.ServicesMTM)
        self.ServicesMTM.add_handler_funcs(self.relay_server_handlers.get_handlers_dict())
        self.ServicesMTM.add_announced_services([self.relay_server_handlers.get_services_dict()])

    def stop_relayserver(self):
        if hasattr(self, 'relay_server_handlers') and self.relay_server_handlers:
            debug.stderr.write("Stopping Relay Server\n", vs='Broker')
            self.ServicesMTM.remove_announced_services(self.relay_server_handlers.get_services_dict().keys())
            self.ServicesMTM.remove_handler_funcs(self.relay_server_handlers.get_handlers_dict().keys())
            self.relay_server_handlers = None
    
    def kick_relayserver(self):
        ## relay server:
        if confman.is_true_bool(("YES_NO", "RUN_RELAY_SERVER",)) and not confman.is_true_bool(("YES_NO", "SERVE_USING_A_RELAY",)):
            if self.relay_server_handlers is None:
                self.launch_relayserver()
        else:
            self.stop_relayserver()

    def launch_blockserver(self):
        """
        Launches a block server.
        """
        debug.stderr.write("Starting Block Server\n", vs='Broker')
        self.blobserver = BlockServerEGTP.BlobServerEGTP(
            broker=self,
            throttler = self.FindBeforehandBlockFetcher.throttler)
    
    def stop_blockserver(self):
        return # XXX disabled for now until crashing bug can be fixed
        # (we don't need this anymore: all brokers are always blockservers)
        if hasattr(self, 'blobserver') and self.blobserver:
            debug.stderr.write("Stopping Block Server\n", vs='Broker')
            self.blobserver.shutdown()
            self.blobserver = None

    def kick_blockserver(self):
        # All brokers are block servers no matter what
        confman['YES_NO']['RUN_BLOCK_SERVER'] = "yes"

        if self.blobserver is None:
            self.launch_blockserver()

    def launch_contenttracker(self):
        """
        Launches a content tracker.
        """
        debug.stderr.write("Starting Content Tracker\n", vs='Broker')
        self.content_typedef_repository = mojohttpd.globaldict['typedefrep']
        self.content_tracker = ContentTrackerEGTP.ContentTrackerEGTP(
            broker=self, typedef_repository=self.content_typedef_repository)
 
    def stop_contenttracker(self):
        if hasattr(self, 'content_tracker') and self.content_tracker:
            debug.stderr.write("Stopping Content Tracker\n", vs='Broker')
            self.content_tracker.shutdown()
            self.content_tracker = None
 
    def kick_contenttracker(self):
        # - if we're setup to run a content tracker, start it.
        if confman.is_true_bool(("YES_NO", "RUN_CONTENT_TRACKER",)):
            if self.content_tracker is None:
                self.launch_contenttracker()
        else:
            self.stop_contenttracker()

    def launch_publicationagent(self):
        debug.stderr.write("Starting Publishing Agent\n", vs='Broker')
        self.publicationagent = PublicationTrackerHandlers.run(self.ServicesMTM)

    def stop_publicationagent(self):
        if hasattr(self, 'publicationagent') and self.publicationagent:
            debug.stderr.write("Stopping Publishing Agent\n", vs='Broker')
            self.publicationagent.shutdown()
            self.publicationagent = None

    def kick_publicationagent(self):
        # Publication Agent:
        if confman.is_true_bool(("YES_NO", "RUN_PUBLISH_AGENT",)):
            if self.publicationagent is None:
                self.launch_publicationagent()
        else:
            self.stop_publicationagent()

    # Miscellaneous initialization methods:
    def initialize_doq_tasks(self):
        self.doq.add_task(self.task_periodic_hello)
        if not confman.is_true_bool(["COUNTERPARTY", "DISABLE_PRICE_CHECKING_AND_HANDICAPPING"]):
            self.doq.add_task(self.task_periodic_withdrawal)
        self.doq.add_task(self.task_bootpage_update, delay=whrandom.randint(87600/2, 87600))

    def initialize_mtms(self, TokenMTM_ID=None, ServicesMTM_ID=None):
        self.TokenServerMTM = MojoTransaction.MojoTransactionManager(doQ = self.doq,
                                                                     serialized = None,
                                                                     pt = None,
                                                                     recoverdb=self.recoverdb,
                                                                     dontbind=1,
                                                                     keyID = TokenMTM_ID,
                                                                     allow_send_metainfo=false)

        # - create our Paytool, used by the ServicesMTM for doing TokenPayments:
        self.ServicesPT = Paytool.Paytool(self.TokenServerMTM)
        
        # - create our MojoTransactionManager used for running services and speaking to Mojo Nation (TM) servers.

        # If the user specifies automatic port searching use it, if she
        # specifies a specific port, use that.
        port_string = confman.get("TRANSACTION_MANAGER_LISTEN_PORT")
        if not port_string:
            listenport = None # This signals MojoTransactionManager to select a random port.
        else:
            listenport = mojoutil.intorlongpopL(port_string)
        pickyport = confman.is_true_bool(("TRANSACTION_MANAGER_PICKY_PORT",))

        self.ServicesMTM = MojoTransaction.MojoTransactionManager(
            doQ=self.doq,
            serialized=None,
            pt=self.ServicesPT,
            listenport=listenport,
            pickyport=pickyport,
            handler_funcs={},
            announced_service_dicts=[],
            recoverdb=self.recoverdb,
            keyID = ServicesMTM_ID)

        # check our comm strat, if it is non-routable, set SERVE_USING_A_RELAY and make a new MTM
        # XXX this is an ugly hack, we shouldn't have to create a MTM solely to see if it's routable so that we can shut it down.
        if not CommStrat.is_reachable(self.ServicesMTM.get_comm_strategy()) and not confman.is_true_bool(("YES_NO", "SERVE_USING_A_RELAY",)):
            self.ServicesMTM.shutdown()
            self.ServicesMTM = None
            self.mojolog.write("You have a non-routable IP address.\n", v=0, vs="broker")
            self.mojolog.write("SERVE_USING_A_RELAY has been enabled for you.\n", v=0, vs="broker")

            confman.dict["YES_NO"]["SERVE_USING_A_RELAY"] = "yes"
            confman.save()

            self.ServicesMTM = MojoTransaction.MojoTransactionManager(
                doQ=self.doq,
                serialized=None,
                pt=self.ServicesPT,
                listenport=listenport,
                handler_funcs={},
                announced_service_dicts=[] ,
                recoverdb=self.recoverdb,
                keyID=ServicesMTM_ID)

        # if we generated a new key, then save the ID in the config file
        if TokenMTM_ID is None:
            confman.dict['TMTM_ID'] = mojosixbit.b2a(self.TokenServerMTM.get_id())
            confman.save()
        if ServicesMTM_ID is None:
            confman.dict['PMTM_ID'] = mojosixbit.b2a(self.ServicesMTM.get_id())
            confman.save()
            
        self.__init_block_fetching()

        debug.stderr.write("This Broker's ID is: %s\n" % mojosixbit.b2a(self.ServicesMTM.get_id()), vs='Broker', v=-1)

        # tell the TokenServerMTM to use our main comms MTM for MetaTrackerLib calls
        self.TokenServerMTM.set_metalookup_mtm(self.ServicesMTM)
    
        if confman.is_true_bool(("YES_NO", "SERVE_USING_A_RELAY",)):
            self.ServicesMTM.use_relay_server()

    def _mojobootpage_callback(self, text):
        """ A private callback for the bootpage loader which sets self.bootpage. """
        if self.ServicesMTM:
            self.task_periodic_hello(reschedule=false)
        
    def load_bootpage(self):
        # load the boot page
        if self.confman.get("IGNORE_BAD_IP_ADDRS") == "testing" :  # XXX total hack for testing!
            debug.stderr.write("Not loading bootpage since IGNORE_BAD_IP_ADDRS was set to testing\n")
        else :
            mojobootpage.load_and_use(self.doq, block=1, timeout=15, textcallback=self._mojobootpage_callback)

    def is_mojohttpd_required(self):
        """
        This method should be called if the web gateway (mojohttpd) is disabled in config.
        This method checks the state of the Broker to determine if mojohttpd is required.
        It returns whether or not mojohttpd is required to run.

        It checks whether or not a passphrase is required and whether TOS has been accepted.

        @return a true value indicates mojohttpd must be enabled to run the broker properly;
        a false value indicates that mojohttpd is not required in order to continue.
        """
        # check for TOS:
        if confman.dict.get("AGREED_TO_TERMS_OF_SERVICE") != 'yes':
            self.mojolog.write('The Broker cannot run until the Terms of Service are accepted.\n', v=-1000)
            self.mojolog.write('The Terms of Service must be accepted via the web gateway.\n', v=-1000)
            return true

        # Check for pass phrase:
        if confman.dict['ENCRYPT_LOCAL_KEYS'] == 'yes':
            self.mojolog.write('The Broker is configured to use a pass-phrase which requires the web gateway.\n', v=-1000)
            return true

        # Check for encrypted keys:
        if confman.dict.has_key('ENCRYPTED_TOKEN_SERVER_KEY') or confman.dict.has_key('ENCRYPTED_BROKER_KEY'):
            self.mojolog.write('The Broker detected encrypted keys; the web-gateway is required to enter the decryption passphrase.\n', v=-1000)
            return true

        # The web-gateway is not necessary:
        return false

    def __init_block_fetching(self):
        """
        Initializes the Block Fetcher.
        
        @precondition self.initialize_mtms has been called.
        @postcondition unsplit.theGlobalFindBeforehandBlockFetcher is set.
        """
        # initialize our single FindBeforehandBlockFetcher
        self.mojolog.write("Creating our single FindBeforehandBlockFetcher\n")
        self.FindBeforehandBlockFetcher = FindBeforehandBlockFetcher.FindBeforehandBlockFetcher(self)
        unsplit.theGlobalFindBeforehandBlockFetcher = self.FindBeforehandBlockFetcher

    def init_buddylist(self):
        """
        Initializes self.BuddyList.

        @precondition self.initialize_mtms has been called.
        @postcondition self.buddylist is set to an initialized BuddyList object.
        """
        if not confman.is_true_bool(['BUDDY_LIST_ENABLED']):
            return
        self.buddylist = BuddyList.BuddyList(self.doq, self.ServicesMTM)
        self.ServicesMTM.get_handicapper().add_handicapper(self.buddylist.handicapper)
        self.doq.add_task(self.buddylist.periodic_reload_and_save_task)
        self.doq.add_task(self.buddylist.periodic_refresh_contact_info_task, delay=241)

    # Useful common methods:  (The basic functionality of MN- publish, retrieve, search, etc...)
    def get_archive_by_dinode(self, dinode):
        """
        Downloads the data represented by dinode from the MN network.

        dinode - A string dinode.  It may either be mojosixbit encoded or bare.
        """
        return archive.Reader(unsplit.ReassemblerSplitFile(TM=self.ServicesMTM, dinode=dinode))

    # DoQ tasks:
    def task_periodic_hello(self, reschedule=true):
        if self.__is_shutting_down:
            return
        # schedule another announcement for later first so that even if there is an exception
        # below we still keep trying (approx 15 minutes)
        if reschedule:
            self.doq.add_task(self.task_periodic_hello, delay=(15*60 - whrandom.randint(20,80)))

        # make sure we advertise ourselves fully if we're running our own meta/id tracker.
        if self.metatracker :
            self.metatracker.reload_own_hello_info()

        self.mojolog.write("Sending hello to meta tracker(s)\n", v=3)
        self.ServicesMTM.send_hello_to_meta_trackers()

    def task_periodic_withdrawal(self):
        if self.__is_shutting_down:
            return
        def handle_withdraw_outcome(widget, outcome, failure_reason=None, self=self):
            if self.__is_shutting_down:
                return
            if (not failure_reason) and (outcome[0] == "success"):
                # If this withdraw succeeded, then schedule another one for 10 minutes from now.
                self.doq.add_task(self.task_periodic_withdrawal, delay=DELAY_BETWEEN_WITHDRAWALS)
            else:
                # Else, schedule another one for a day or so from now.
                self.doq.add_task(self.task_periodic_withdrawal, delay=DELAY_AFTER_FAILED_WITHDRAWAL)

        self.ServicesMTM._pt.withdraw(AMOUNT_OF_WITHDRAWAL, outcome_func=handle_withdraw_outcome)

    def task_bootpage_update(self):
        if self.__is_shutting_down:
            return
        # check the mojonation.net bootpage 1-2 times a day to make sure a new version
        # or new root metatracker hasn't appeared recently.  (NOTE this won't be nice to
        # our web server when we take over the world, oh darn. ;-)

        mojobootpage.load_and_use(doQ=self.doq, textcallback=self._mojobootpage_callback)
        self.doq.add_task(self.task_bootpage_update, delay=whrandom.randint(87600/2, 87600))

def check_testres():
    try:
        import testres
        debug.stderr.write("unit tested CVS version: %s\n" % testres.cvsId)
    except:
        pass

def parse_argv(argv):
    """
    Takes argv (which follows the semantics of) sys.argv and returns a tuple:
    @return (recoverdb, interact)
    """
    recoverdb = true
    interact = false
    for arg in argv[1:]:
        if arg == "--recoverdb":
            recoverdb = true
        elif arg == "--norecoverdb":
            recoverdb = false
        elif arg == "--interact":
            interact = true
        elif arg in ("--no-log-file", "--no-log-stderr", "--rotated-log", "--log-file"):
            pass  # ignore; it is processed in mojostd.py
        elif arg[:len("--log-file=")] == "--log-file=":
            pass  # ignore; it is processed in mojostd.py
        else:
            debug.mojolog.write('Warning: Unknown command-line argument "%s".\n' % arg)
    return (recoverdb, interact)

def cleanup_mojo_tmp_dir():
    """Be kind; attempt to remove all files in our temporary directory."""
    tmpdir = os.path.expandvars(confman.dict['PATH']['TMP_DIR'])
    try:
        tmpfiles = map(lambda fn, tmpdir=tmpdir: os.path.join(tmpdir, fn), os.listdir(tmpdir))
    except OSError:
        return
    debug.mojolog.write("Cleaning up %s temporary files in %s\n", args=(len(tmpfiles), tmpdir), v=2, vs='Broker')
    for file in tmpfiles:
        try:
            os.unlink(file)
        except OSError:
            continue

# The following are quite hackish UI kludges:
def _get_base_url():
    """For use by the kludgey "talk to other brokers" functions below."""
    path =confman.getpath(('PATH', 'BASE_URL_FILE'))
    f = open(path, 'r')
    url = f.read()[len('http://'):]
    f.close()
    host, path = string.split(url, '/', 1)
    return host, '/' + path + '/'
    
def other_broker_exists():
    """
    Check the base.url file and try to connect to see if another broker is running.
    """
    try:
        hostpart, reqpath = _get_base_url()
    except:
        # Malformed...  This could happen when the Broker config dir is first created (empty base.url file)
        # or if a silly human mangles the file.
        return false
    try:
        debug.mojolog.write('Checking for other Brokers running with the same config file.\n', vs='Broker', v=0)
        h = httplib.HTTP(hostpart)
        h.putrequest('GET', reqpath)
        h.endheaders()
        rcode, rmsg, headers = h.getreply()
        if headers and headers.has_key('Server') and headers['Server'][:9] == 'MojoHTTPD':
            return true
        else:
            return false
    except httplib.socket.error, e:
        return false
    
def kludgey_show_other_broker():
    """
    This is a kludgey hack:  If another broker is running,
    this attempts to launch a browser to the main page.

    This is intended to make the user experience a tad nicer,
    especially to users who might not diligently refrain from launching
    two Brokers at once.

    Precondition: other_broker_exists has been called and returned true.
    """
    host, pcode = _get_base_url()
    import webbrowser
    webbrowser.open('http://' + host + pcode)
    
if __name__ == '__main__':
    main(sys.argv)
