# This module is used to automatically determine content types and descriptions for a given file.
# Note, this module is not as flexible as contentlib...  Maybe the criteria for determining the
# content type of a file, as well as ways to fill in content description fields automatically should
# be part of the content types, or in some way under the control of contentlib.  This would make
# it easier to maintain the relationship between content types and how the UI handles them.

# Standard Modules:
import os
import types
import string
import mimetypes
import imghdr
import sndhdr
import debug
import traceback

# Mojo Nation Modules:
from human_readable import hr
import mp3
import mojoutil

def describe(path, ctype="guess", cdesc={}, typedefrep=None):
    """
    This function takes the path to a file and returns its guessed content type and description values.

    @param path the path to the file
    @param ctype a string representing the ctype or "guess" (recommended), or "private" or
        None (deprecated) or "NONE" (deprecated);  If `ctype' is `None' or "guess", try to guess
        which type of content the files is.  If `ctype' is "private" or "NONE", return
        `("private", {}), indicating that no content tracker submission should happen.  Else,
        `ctype' specifies the content type of the file.
    @param cdesc - A dict of content description values.  Any values passed in over-ride guessed values.

    @returns (ctype, cdesc) - A tuple containing a string which is the content type, and a dict, which is the content description.

    @precondition `ctype' must be None or a string.: (type(ctype) == types.StringType) or (ctype is None): "ctype: %s :: %s" % (hr(ctype), hr(type(ctype)))
    @precondition `cdesc' must be a dict.: type(cdesc) == types.DictType: "cdesc: %s :: %s" % (hr(cdesc), hr(type(cdesc)))
    """
    assert (type(ctype) == types.StringType) or (ctype is None), "precondition: `ctype' must be None or a string." + " -- " + "ctype: %s :: %s" % (hr(ctype), hr(type(ctype)))
    assert type(cdesc) == types.DictType, "precondition: `cdesc' must be a dict." + " -- " + "cdesc: %s :: %s" % (hr(cdesc), hr(type(cdesc)))

    if ctype is None:
        ctype = guess_ctype(path)

    if (ctype) and ((string.lower(ctype) == "private") or (string.lower(ctype) == "none")):
        return ("Private", {})

    try:
        description = get_description(path, ctype, typedefrep)
    except:
        debug.mojolog.write("Couldn't describe file with path: %s \n\tException Follows:\n" % path)
        traceback.print_exc(file=debug.mojolog)
        description = {}

    description.update(cdesc)

    return (ctype, description)


def guess_ctype(path):
    """
    This function takes a path and returns a ctype.
    It defaults to "Generic Description".

    @param path the path to the file
    @returns ctype the content type
    """

    mtype, encoding = mimetypes.guess_type(path)

    if not mtype:
        # A mime type could not be determined.
        if string.lower(os.path.splitext(path)[1]) == '.mp3':
            return 'Audio'
        return "Generic Description"
        
    # Now check in out MimeToCType dict to guess the content type.
    # First, check the entire mimetype string:
    if MimeToCType.has_key(mtype):
        return MimeToCType[mtype]
    
    # Next check only against the major part:
    major, minor = string.split(mtype, '/')
    if MimeToCType.has_key(major):
        return MimeToCType[major]

    # A mime-type we don't know about.
    return "Generic Description"

def get_description(path, ctype, typedefrep=None):
    description = {}

    if (typedefrep is None) or (typedefrep and typedefrep.get_type_desc(ctype).has_key('Size in KB')):
        size = mojoutil.get_path_size(path)
        if size is not None:
            description['Size in KB'] = size / (2 ** 10) # Reduce from Bytes to KBytes.

    descriptor = DescriptionMap.get(ctype)
    if type(descriptor) == types.FunctionType:
        description = descriptor(path)
    elif type(descriptor) == types.DictType:
        for field, func in descriptor.items():
            description[field] = func(path)
    return description

def striptrailingnulls(s):
    """
    @precondition `s' must be a string.: type(s) == types.StringType: "s: %s" % `s`
    """
    assert type(s) == types.StringType, "precondition: `s' must be a string." + " -- " + "s: %s" % `s`

    if len(s) <= 0:
        return s

    i = len(s) - 1
    while (i >= 0) and (s[i] == '\000'):
        i = i - 1

    return s[:(i+1)]

def describe_mp3(f):
    """
    @returns metainfodict
    """
    md = mp3.get_mp3tag(f)

    mapnames = {
        'TITLE': 'Title',
        'ARTIST': 'Performer',
        'ALBUM': None,
        'YEAR': None,
        'COMMENT': 'Description',
        }

    mapstats = {
        'VERSION': None, 
        'MM': None, 
        'SS': None, 
        'STEREO': None, 
        'LAYER': None, 
        'MODE': None, 
        'COPYRIGHT': None, 
        'BITRATE': 'Bitrate in kbit/sec', 
        'FREQUENCY': None, 
        }

    metainfodict = {}
    for k, v in md.items():
        mnk = mapnames.get(k)

        if mnk:
            metainfodict[mnk] = striptrailingnulls(v)
    od = {}
    try:            
        od = mp3.mp3info(f)
    except IndexError:
        # this is a variable bitrate MP3 and mp3info chokes on that
        pass
    
    for k, v in od.items():
        mnk = mapstats.get(k)

        if mnk:
            metainfodict[mnk] = `v`

    mp3v = "mp3"
    if od.has_key('VERSION'):
        mp3v = mp3v + ", version " + `od.get('VERSION')`
    if od.has_key('LAYER'):
        mp3v = mp3v + ", layer " + `od.get('LAYER')`
    metainfodict['Encoding'] = mp3v

    sizeinbytes = os.path.getsize(f)
    sizeinkb = sizeinbytes / (2 ** 10)
    metainfodict['Size in KB'] = sizeinkb

    return metainfodict
    
def describe_audio(path):
    # XXX, content type wish list item:
    # It would be nice if the following fields were in the audio content type:
    # sample rate, channels, frames, bits-per-sample.
    if os.path.isdir(path):
        # XXX this should probably at least figure out how big the directory is
        return {}
    info = sndhdr.what(path)
    if info:
        return {
            'Title' : get_title_from_path(path),
            'Encoding' : info[0],
            'Bitrate in kbit/sec' : info[1],
            'Size in KB' : os.path.getsize(path) / (2 ** 10),
            }
    elif string.lower(os.path.splitext(path)[1]) == '.mp3':
        return describe_mp3(path)
    else:
        return {}

def image_file_format(path):
    if os.path.isdir(path):
        # XXX this should probably compute the size of the directory...
        return {}
    return imghdr.what(path)

def get_title_from_path(path):
    return os.path.splitext(os.path.basename(path))[0]

# This maps mime type strings to Mojo Nation content type strings.  The keys can either be the
# full mimetype string or they can be just the major part.  Full mimetypes are checked before
# the less specific major mimetypes.
MimeToCType = {
    'image' : 'Image',
    'audio' : 'Audio',
    'text/html' : 'Web Site',
    }

# This maps MojoNation content types to either functions or dicts.  If it's a function, that function
# when given a path returns a dict which is a content description.  If it is a dict, it maps content
# description fieldsnames to functions, which when passed a path, return that fieldnames value.
DescriptionMap = {
    'Generic Description' : {
        'Description' : get_title_from_path,
        },
    'Audio' : describe_audio,
    'Image' : {
        'Title' : get_title_from_path,
        'File Format' : image_file_format,
        },
    }

def interactive_test_describe(path):
    """
    This function doesn't work with testem, because you need to point it to a directory full of
    files to test!  :(
    """
    import pprint
    results = map(lambda f: (f, describe(f)), map(lambda f, b=path: os.path.join(b, f), os.listdir(path)))
    pprint.pprint(results)
    return results
    
