#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#  a cluster of classes for publishing a file object
#
## the main object to use is PublishSession

### standard modules
from Queue import Queue, Empty
from time import time
import string

### our modules
import BidStrategy
from ClientEGTPlib import publish_blob
import debug
import idlib
import idmasklib
import makesplit
import MetaTrackerLib
import mojoutil
import sharemap
from DataTypes import *
import pricelib
from confutils import confman



class PublishQueue:
    def __init__(self):
        self.q = []
        self.state = 'stopped'
    def addJob(self, session):
        self.q.append(session)
        self.do_something()
    def do_something(self):
        if self.state == 'stopped':
            if len(self.q) > 0:
                session = self.q[0]
                self.q = self.q[1:]
                self.state = 'started'
                self.current = session
                session.startPublishing()
    def publish_finished(self, session):
        self.state = 'stopped'
        self.do_something()

    def publish_cancelled(self, session):
        if self.current == session:
            self.state = 'stopped'
        else:
            try:
                self.q.remove(session)
            except ValueError:
                pass
        self.do_something()

        
pubQ = PublishQueue()


def find_publish_blob_servers_and_call_back(MTM, callback, blob_id) :
    """
    Returns a list of blob servers that should be willing to publish
    a blob with the given Id.
    """
    # XXX fine tune how many bits we specify in this search mask (it
    # should probably be semi-dynamic based on some info about how fine
    # of a mask range servers are speaking these days)  That way we'll
    # get the most out of our meta tracker response without giving
    # away too much info about what we're looking for.
    search_id_mask = idmasklib.IdMask(mask=idlib.to_binary(blob_id), bits=14)

    retval = MetaTrackerLib.get_publishing_blob_servers_and_call_back(MTM, callback, blob_id_mask=search_id_mask, min=5)

    return retval



# this class publishes a single blob, it holds state and
# it handles the callback and reports back to it's Piece the
# results of the publishing action publishing
class PutBlob:
    def __init__(self, shareid, sharedata, sharedatalen, piece, MTM, timeout):
        self.MTM = MTM
        self.piece = piece
        self.shareid = shareid
        self.sharedata = sharedata
        self.sharedatalen = sharedatalen
        self.timeout = timeout
    
    def handle_result_of_publish(self, widget, outcome, failure_reason=None):
        if failure_reason:
            # our share failed to publish
            self.piece.put_blob_failed(self, failure_reason)
        else:
            self.piece.put_blob_succeeded(self, widget.get_counterparty_id())

    def find_servers_callback(self, serverlist):
        if not serverlist:
            self.piece.put_blob_failed(self, "No publication blob servers available.")
        else:
            publish_blob(MTM=self.MTM,
                         serverlist=serverlist.items(),
                         blobid=self.shareid, blobdata=self.sharedata,
                         outcome_func=self.handle_result_of_publish,
                         timeout=self.timeout, passalong=1)

    def put_blob(self):
        find_publish_blob_servers_and_call_back(self.MTM, self.find_servers_callback, self.shareid)



# this class encapsulates the progress of publishing a complete piece, 
class Piece:
    def __init__(self, pieceid, sharelist, session, MTM, totalsharesneeded, totalsharesinpiece, is_sharemap=None):
        self.MTM = MTM
        self.pieceid = pieceid
        self.session = session
        self.num_shares = len(sharelist)
        self.totalsharesneeded = totalsharesneeded
        self.totalsharesinpiece = totalsharesinpiece
        assert self.totalsharesneeded <= self.totalsharesinpiece
        assert self.num_shares >= self.totalsharesneeded, "won't publish Piece without enough shares"

        # number of shares that published successfully
        self.passed = mojoutil.Counter()

        # number of shares that failed to publish
        self.failed = mojoutil.Counter()

        # if I'm a sharemap piece (distinction only used for accurate user feedback)
        self.is_sharemap = is_sharemap

        # create PutBlob instances for all my shares
        for shareid, share in sharelist:
            blob = PutBlob(shareid, share, len(share), self, self.MTM, perfStrat.put_timeout())
            self.session.q_blob(blob)
        self.length = self.totalsharesneeded * len(share)

    # called by PutBlobs when they fail to publish
    def put_blob_failed(self, blob, reason):
        debug.mojolog.write('Failed to put blob for PieceID: %s.\n', args=(self.pieceid,), vs='PublishSession', v=4)
        self.failed.inc()
        if self.failed.get() > self.totalsharesneeded:
            # couldn't publish at least the necessary number of shares of this piece
            # fail completely XXX
            myreason = "The publication failed. [%s]  (Piece id %s (block %s) could not be published.)\n" % (reason, idlib.id_to_abbrev(self.pieceid), idlib.id_to_abbrev(blob.shareid))
            self.session.publication_failed(myreason)
        else:
            # is this piece complete?
            if (self.passed.get() + self.failed.get()) >= self.num_shares:
                self.session.piece_succeeded(self)
            self.session.dispatch_next_share()

    # called by PutBlobs when they succeed in publishing
    def put_blob_succeeded(self, blob, serverid):
        self.passed.inc()

        # store location information
        self.session.store_location_info(blob, serverid)

        # is this piece complete?
        if (self.passed.get() + self.failed.get()) >= self.num_shares:
            self.session.piece_succeeded(self)

        self.session.dispatch_next_share()

# this is the main controller object
# it just splits up the data and sends it out
# arcfile: a file object or StringIO pointing to a mojoarc
# length: length of arcfile contents
# MTM: ...
# callback(failure_reason, compressed_dinode) : failure_reason is None in successful publishing, otherwise a reason
#           compressed_dinode raw compressed dinode of the published file
# statusfunc: optional function that takes a string updating the user on progress
class PublishSession:
    # take a file object pointing to a file object and a size
    def __init__(self, arcfile, length, MTM, callback, totalsharesneeded, totalsharesinpiece, statusfunc=None, errorfunc=None):
        assert totalsharesneeded in (1,2,4,8,16,32,64,128), "totalsharesneeded must be a reasonable power of two"
        assert totalsharesinpiece >= totalsharesneeded, "totalsharesinpiece must be greater than totalsharesneeded"

        self.MTM = MTM

        self.callback = callback
        self.totalsharesneeded = totalsharesneeded
        self.totalsharesinpiece = totalsharesinpiece

        # size
        self.length = length
        
        # time we start publishing
        self.starttime = None
        
        # function that takes a string, to keep the user updated on progress
        self.statusfunc = statusfunc

        # An error-handling function for doq tasks:
        self.errorfunc = errorfunc
        
        # location information for the pubtracker
        self.locations = []

        # shares to be published
        self.shareQ = Queue(0)
        
        # how many bytes published so far
        self.bytes_published = 0L

        # how many bytes of the sharemap have been published
        self.smap_bytes_published = 0L

        # current operation state!   one of:
        # "running"  publishing main pieces
        # "failed" show stopper occurred...  self.reason should have the reason
        # "success" all done, everything went a okay
        self.status = 'running' 

        # counter of pieces that have been published successfully
        self.completed_pieces = mojoutil.Counter()

        # total pieces queued up so far, doesn't represent all of the
        # pieces in the file until self.splits is exhausted!
        # (at that point will also include the sharemap pieces)
        self.total_pieces = 0L

        # flag indicating whether the sharemap pieces have been queued up yet
        # zero means, not yet, 1 means yes, -1 means we determined there won't be a sharemap
        self.queued_sharemap = 0

        # max number of piece descriptions that will be included in the dinode
        if confman.is_true_bool(('PUBLISH_USING_BIG_DINODES',)):
            self.pieces_included_in_dinode = 3  # include info on the first few pieces in the dinode
        else:
            self.pieces_included_in_dinode = 0  # don't publish using new (not supported by all clients) dinode format

        self.arcfile = arcfile

        self.MTM._doQ.add_task(pubQ.addJob, args=(self,))
        self.statusfunc("Queued...\n")

    def startPublishing(self):
        if self.status == 'failed':
            return
        # split 'em
        self.splits = makesplit.Splitter(self.arcfile, self.length, totalsharesneeded=self.totalsharesneeded, totalsharesinpiece=self.totalsharesinpiece)
        del(self.arcfile)

        # get the first piece ready to go
        self.q_next_piece()

        # go!  don't change this string, the publishcommand uses it to recognize when the publish starts
        self.statusfunc('Publication starting...\n')
        for i in range(perfStrat.max_puts()):
            self.MTM._doQ.add_task(self.dispatch_next_share, errorfunc=self.errorfunc)
        self.MTM._doQ.add_task(self._success_monitor, delay=5)
        self.MTM._doQ.add_task(self.update_status, delay=2)
            
    # pulls the next piece from makesplit
    # the Piece instance is responsible for putting the shares on the queue with q_blob
    # returns 0 if no more pieces, 1 if a new piece was queued
    def q_next_piece(self):
        if self.status == 'failed':
            return
        pieceid, sharelist = self.splits.getnextsplit()
        if pieceid:
            self.total_pieces = self.total_pieces + 1
            Piece(pieceid, sharelist, self, self.MTM, totalsharesneeded=self.totalsharesneeded, totalsharesinpiece=self.totalsharesinpiece)
            return 1
        else:
            # no more pieces in the main file
            if self.queued_sharemap == 0:
                self.piecelist = self.splits.get_piece_list()
                if len(self.piecelist) > self.pieces_included_in_dinode:
                    debug.mojolog.write('Starting to publish sharemap.\n', v=4, vs='PublishSession')
                    self.queued_sharemap = 1
                    # split the sharemap
                    topsharemap, piecelist, self.sharemap_len = self.splits.getsharemap()
                    sharemapid, sharemapshares, topsharemaplen = topsharemap

                    self.make_dinode(sharemapid, sharemapshares, topsharemaplen)
                    # queue up the sharemap pieces and go!
                    for pieceid, sharelist in piecelist:
                        self.total_pieces = self.total_pieces + 1
                        Piece(pieceid, sharelist, self, self.MTM, totalsharesneeded=self.totalsharesneeded, totalsharesinpiece=self.totalsharesinpiece, is_sharemap=1)
                    return 1
                else:
                    debug.mojolog.write('%s or less pieces, no sharemap needed.\n', args=(self.pieces_included_in_dinode,), v=4, vs='PublishSession')
                    self.make_dinode()
                    self.queued_sharemap = -1
                    return 0
            else:
                return 0


    # puts a blob on the queue
    def q_blob(self, blob):
        self.shareQ.put(blob)

    # status function
    def is_publication_failed(self):
        if self.status == 'failed' or self.status == 'cancelled':
            return 1
        return 0
        

    # called by a Piece when it fails to publish, shuts down the
    # publication process
    def publication_failed(self, reason):
        debug.mojolog.write('Publication fails.\n', vs='PublishSession', v=2)
        self.status = 'failed'
        self.reason = reason
        self.shareQ = None
        self.MTM._doQ.add_task(self.callback, (reason, None), errorfunc=self.errorfunc)
        pubQ.publish_finished(self)

    # called by UserCommand whet the publish is cancelled.
    # publication process
    def publication_cancelled(self, reason):
        debug.mojolog.write('Publication cancelled.\n', vs='PublishSession', v=2)
        self.status = 'cancelled'
        self.reason = reason
        self.shareQ = None
        pubQ.publish_cancelled(self)

    # called by a Piece when it succeeds
    def piece_succeeded(self, piece):
        if self.is_publication_failed():
            # forget it, we're already failed
            return
        self.completed_pieces.inc()
        if piece.is_sharemap:
            self.smap_bytes_published = self.smap_bytes_published + piece.length
        else:
            self.bytes_published = self.bytes_published + piece.length

        if self.statusfunc:
            self.update_status()

    def update_status(self):
        # calculate our throughput so far
        if self.starttime is None:
            self.statusfunc("Uh...  I quit before I began...  self.starttime: None", briefstatus="silly")
            return

        delta = time() - self.starttime
        smap_ksofar = self.smap_bytes_published/1024
        ksofar = self.bytes_published/1024
        # ksofar is size of the original pieces, but extra data is actually published due to info dispersal
        kps = (ksofar / delta) * (self.totalsharesinpiece/float(self.totalsharesneeded))
        smap_kps = (smap_ksofar / delta) * (self.totalsharesinpiece/float(self.totalsharesneeded))
        main_percent = int(float(self.bytes_published)*100/self.length)

        if main_percent > 100:
            main_percent = 100
        if self.queued_sharemap == 1:
            smap_percent = int(float(self.smap_bytes_published)*100/self.sharemap_len)
            if smap_percent > 100:
                smap_percent = 100
            self.statusfunc("%d of %d kb published [%.1f Kb/sec - %d%% complete]  Sharemap: %d of %.1f kb %.1fkb/s %d%%\n" %
                            (ksofar, self.length/1024, kps, main_percent,
                             smap_ksofar, self.sharemap_len/1024.0, smap_kps, smap_percent),
                             briefstatus="finalizing")
        else:
            self.statusfunc("%d of %d kb published [%.1f Kb/sec - %d%% complete]\n" %
                            (ksofar, self.length/1024, kps, main_percent),
                            briefstatus='running [%d%%]' % main_percent)

    # pulls a blob off the queue and publishes it
    # figures out if we are done and if so, call's out to LocationBroadcaster
    def dispatch_next_share(self):
        if self.is_publication_failed():
            # don't do any more...
            return
        if not self.starttime:
            self.starttime = time()
        blob = None
        try:
            blob = self.shareQ.get(block=0)
        except Empty:
            if self.q_next_piece():
                blob = self.shareQ.get(block=0)
            
        if blob:
            ## dispatch blob here
            self.MTM._doQ.add_task(blob.put_blob, errorfunc=self.errorfunc)
        # no more shares, check to see if we have completed all of the pieces
        elif self.completed_pieces.get() >= self.total_pieces:
            # XXX HACK self._success_monitor() stuff used to happen in here, now its asynchronous
            pass
                                   
    def _success_monitor(self):
        # XXX this asynch monitoring function was done as a hack during testing but we left it rather than twiddle something that works
        debug.mojolog.write('%s completed pieces %s, total pieces %s\n', args=(self, self.completed_pieces.get(), self.total_pieces), vs='PublishSession', v=4)
        if self.completed_pieces.get() >= self.total_pieces:
            # we're done
            self.status = 'success'
            self.shareQ = None
            # now broadcast location information
            LocationBroadcaster(self.MTM, self.finished_location_broadcasting, self.locations, self.statusfunc)
            # callback now!  since broadcasting is flaky
            self.MTM._doQ.add_task(self.callback, (None, self.dinode), errorfunc=self.errorfunc)
            pubQ.publish_finished(self)
        elif self.status != 'failed':
            self.MTM._doQ.add_task(self._success_monitor, delay=5)
            return
        debug.mojolog.write('%s success_monitor done\n', args=(self,), v=2, vs='PublishSession')
            
    # store location information (for the publication brodcaster)
    def store_location_info(self, blob, serverid):
        # this causes the Broker to appear to leak slowly while publishing
        # a smarter Broker would periodically publish this information instead of waiting till the end
        # the publication broadcaster is partially obsoleted by the BlockWholesaler
        self.locations.append({'blob id' : blob.shareid, 'server id' : serverid, 'bytes': str(int(blob.sharedatalen))})


    def finished_location_broadcasting(self):
        # we used to callback here, now we have already called back
        pass


    def make_dinode(self, sharemapid='', sharemapshares=[], sharemaplen=0):
        self.sharemapid = sharemapid
        sharemapdict = {}
        sharenum = 0
        for shareid, sharedata in sharemapshares:
            sharemapdict[sharenum] = shareid
            sharenum = sharenum + 1
        if sharemapshares:
            totalshares = len(sharemapshares)
        else:
            totalshares = len(self.piecelist[0]['shares'])
        self.dinode = sharemap.make_dinode(smapid=sharemapid,
                                               smaplength=sharemaplen,
                                               smapsharedict=sharemapdict,
                                               totalshares=totalshares,
                                               sharesneeded=self.totalsharesneeded,
                                               piecelist=self.piecelist[:self.pieces_included_in_dinode],
                                               compress=1)


def find_publishing_agents_and_call_back(MTM, callback, get_broadcasters_only=None) :
    """
    Returns a list of tuples containing publishing agent ids and their
    metainfo dicts.
    """
    return MetaTrackerLib.get_publishing_agents_and_call_back(MTM, callback=callback, get_broadcasters_only=get_broadcasters_only)


locations_template = ListMarker({'blob id' : UNIQUE_ID, 'server id' : BINARY_SHA1, 'bytes' : INTEGER})


class LocationBroadcaster:
    def __init__(self, mtm, callback, locations, statusfunc = None):
        self.callback = callback
        self.mtm = mtm
        checkTemplate(locations, locations_template)
        self.locations = locations
        if len(locations) == 0:
            # do nothing!
            return1
        self.total_broadcasters = 0
        self.passed = mojoutil.Counter()
        self.failed = mojoutil.Counter()
        self.statusfunc = statusfunc
        if statusfunc:
            statusfunc("Searching for publication tracker broadcasters.\n")
        find_publishing_agents_and_call_back(mtm, get_broadcasters_only=true, callback=self.publish_locations)

    def publish_locations(self, broadcasters):
        pubagent_list = []
        # find -broadcasters- who accept store location information messages
        for pub_tracker_id, pub_tracker_info in broadcasters.items():
            for i in pub_tracker_info['services']:
                if i['type'] == 'publication agent' and i.get('allows store location information') == 'yes':
                    pubagent_list.append((pub_tracker_id, pub_tracker_info))
                    break
        if not pubagent_list:
            debug.stderr.write('WARNING no publication tracker broadcasters found\n')
            self.handle_publication(None, None, "No publication tracker broadcasters found.")

        msgtype = 'store location information'
        msgbody = {'locations' : self.locations}
        if len(pubagent_list) > 1:
            new_pubagent_list = self.mtm.get_handicapper().sort_by_preference(pubagent_list, msgtype, msgbody)
            if len(new_pubagent_list) > 0:
                pubagent_list = new_pubagent_list

        debug.mojolog.write("Storing location information in %s publication broadcasters.\n" % len(pubagent_list), vs="LocationBroadcaster", v=2)
        preferenceno=0
        for pubagent_id, pubagent_metainfo in pubagent_list:
            self.total_broadcasters = self.total_broadcasters + 1
            msgbody = {'locations': self.locations}
            fixedprice = pricelib.publication_agent_store_location_info_price(pubagent_metainfo, msgbody)
            if fixedprice is not None:
                # Remove this once we've grandfathered-out most of the fixed-price servers...  --Zooko 2001-06-01
                mojooffer = fixedprice
            else:
                mojooffer = BidStrategy.choose_bid("MO_STORE_LOCATION_INFORMATION", mtm, counterparty_id=pubagent_id, preferenceno=preferenceno)
            preferenceno=preferenceno+1

            self.mtm.initiate(pubagent_id, msgtype, msgbody, mojooffer, timeout=120, outcome_func=self.handle_publication)

    def handle_publication(self, widget, outcome, failure_reason=None):
        if not failure_reason:
            self.failed.inc()
        else:
            self.passed.inc()
            if self.statusfunc :
                self.statusfunc("Successfully sent location information to %s out of %s publication broadcasters.\n" % (self.passed.get(), self.total_broadcasters))

        if self.failed.get() + self.passed.get() >= self.total_broadcasters:
            # done
            if self.callback:
                self.callback()
        




#--------------------------------------------
# base class for publish performance strategies
class PerfStrategy:
    _max_puts = 0
    _put_timeout = 0

    def set_max_puts(self, n):
        self._max_puts = n
    def max_puts(self):
        return self._max_puts

    def set_put_timeout(self, n):
        self._put_timeout = n
    def put_timeout(self):
        return self._put_timeout


# strategy for medium bandwidth connections (tested on 384kbps link)
class Symmetric384(PerfStrategy):
    _max_puts = 6
    _put_timeout = 120

class T1(PerfStrategy):
    _max_puts = 12
    _put_timeout = 120

# modems  cable modem should also get this, since their upload is slow
class Modem(PerfStrategy):
    _max_puts = 2
    _put_timeout = 145


speed = int(confman.dict['CONN_SPEED_OUT'])
if speed > 1000:
    perfStrat = T1()
elif speed > 250:
    perfStrat = Symmetric384()
else:
    perfStrat = Modem()
