#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# This module provides a back-end to handling UI "commands".
#
# The highest-level API provided here should allow UI programmers
# to simply call functions, each of which corresponds to a "command"
# as the user understands it.  For instance to publish a file with a content
# description, you simply call a function called "publish_file" and pass
# it the necessary info.

# Standard Modules:
import os
import sys
import traceback
import string
import time
import types
import stat


# Mojo Modules:
from MojoConstants import true, false, DEBUG_MODE
from PublishSession import PublishSession
import archive
from confutils import confman
import contentlib
import debug
import htmlspider
from human_readable import hr
import mojoutil
import mojosixbit
import publish
from loggedthreading import LoggedThread
import idlib
import MetaTrackerLib

MAX_CONTENT_TRACKERS = 30 # maximum number of content trackers to submit to

class Error(StandardError):
    pass

class MultiCommandManager:
    """
    This class is responsible for:
    - Keeping a reference to all commands.  (This is its main use.)

    These features would be nice (aka Wishlist):
    - Cancelling, Pausing, Resuming commands. (Not implemented.)
    - Keeping track of completed commands until they are no
      longer needed. (Not implemented.)
    - Managing the persistence of commands. (Not implemented.)

    Note: The semantics of this class have been greatly simplified since Broker version Beta 0.9
    """
    def __init__(self, doq):
        self.doq = doq
        self.idmap = {} # Maps command id's to the commands.
        # These lists aren't necessary but allow the UI to access the two categories quickly (rather than sorting self.idmap values)...
        self.current = [] # Contains un-completed commands.
        self.completed = [] # Contains completed commands, temporarily.
        self.active = [] # contains actively running commands
        debug.stderr.write('MultiCommandManager initialized.\n', vs='UserCommand', v=4)

    def add_command(self, usercommand):
        self.idmap[usercommand.id] = usercommand
        self.current.append(usercommand)
        usercommand.set_make_active(self.make_command_active)
        debug.stderr.write('Command of type "%s" and serial number %s initiated.\n', args=(usercommand.type, usercommand.id), v=4, vs='UserCommand')
        self.doq.add_task(usercommand.start, args=(self.doq, self.command_callback))
        return usercommand.id

    def command_callback(self, command):
        """
        This is called by a command when it is finished.
        """
        try:
            self.current.remove(command)
            self.active.remove(command)
        except ValueError:
            pass
        self.completed.append(command)
        debug.mojolog.write('Command id: %s, type: %s completed, with status %s.\n', args=(command.id, command.type, command.status), vs='UserCommand', v=1)
        
    def get_all_commands(self):
        return self.idmap.values()

    def get_all_current_commands(self):
        """
        The caller mustn'y mutate the return value, unless she knows what she's doing!
        """
        return self.current

    def get_all_active_commands(self):
        """
        The caller mustn'y mutate the return value, unless she knows what she's doing!
        """
        return self.active

    def get_all_completed_commands(self):
        """
        The caller mustn'y mutate the return value, unless she knows what she's doing!
        """
        return self.completed
    
    def get_command(self, id):
        return self.idmap.get(id)
    
    def get_commands(self, idlist):
        results = []
        for id in idlist:
            try:
                results.append(self.get_command(int(id)))
            except ValueError:
                # It wasn't intable, pass:
                pass
        return filter(None, results)

    def make_command_active(self, command):
        self.active.append(command)
    
class UserCommand:
    """
    This class is the superclass of actual specific user commands.

    It's main use it for categorizing commands (via class hierarchy) and for storing
    the command-specific data (which can be altered with callback functions).

    Note: Unless you can recover from an error, you don't need to catch it, instead, pass
    self.default_error_callback as the "errorfunc" parameter to self.doq.add_task.
    """
    _public_methods_ = []  # Allow access via COM
    _public_attrs_ = ['type', 'id', 'starttime', 'stoptime',
                      'status', 'progress', 'name',
                     ]

    nextid = mojoutil.Counter() # This is a class-scoped variable; each UserCommand should have a unique ID.
    
    def __init__(self, servicemtm):
        self.name = '*unset*'
        self._validobject = false
        self.servicemtm = servicemtm

        self.id = UserCommand.nextid.inc() - 1 # This seems roundabout, but it's threadsafe.  Kind of like a threadsafe "a = b++;" in C.

        # self.type is a string representing the type of command.
        self.type = string.lower(self.__class__.__name__)

        # self.status is a string and must equal "starting", "running", "done", or "error"
        self.status = 'starting'
        
        # This dict should only be assigned to inside the sub-classes __init__ method.
        self.input = {}

        # self.state stores state of the command such that given this data, one may continue
        # this command where it left off.  (ie: This state helps one resume a command.)
        self.state = {}

        # self.results stores resultant state which the user is interested in.  No key in this dict should be
        # written to more than once.
        self.results = {}

        # A time-stamp detailing when the command was created:
        # XXX - This should move to a more organized location.  (self.results, perhaps?)
        self.starttime = time.time()
        self.stoptime = None   # if this is not None, this command is finished.

        # self.steps is a tuple of functions to perform when self._perform_discrete_step is called.
        self.steps = ()
        self.stepindex = mojoutil.Counter(-1) # The next step to perform

        # this string can be updated periodically by the command to a meaningful one line description of the commands overall progress
        self.progress = ''

        # self.doq is the DoQ and is defined in self.start:
        self.doq = None
        # self.finish_func is called when this command is completed (either successfully or with an error).  It is defined in self.start:
        self.finish_func = None

        self.excinfo = None # This is set in case of an exception to values similar to the return values of sys.exc_info.

        self._debugfinishstack = None

        self._validobject = true

    def start(self, doq, callback):
        """
        Starts the command (by placing a command-specific function on the doq).
        """
        self.doq = doq
        self.finish_func = callback
        self.status = 'running'
        self.starttime = time.time()
        # Sub-class must override this and place the correct method on the doq.
        
    def set_progress(self, value, briefstatus=None):
        """
        Set's the progress string, as well as logging the new value.
        """
        if value[-1:] == "\n":
            debug.mojolog.write(value, vs='UserCommand', v=1)
        else:
            debug.mojolog.write(value + "\n", vs='UserCommand', v=1)
        self.progress = value
        if briefstatus:
            self.status = briefstatus
        
    def finish(self, status='done'):
        """
        @precondition This instance must not already have been finish()ed.: self._validobject: "self: %s, self._debugfinishstack: %s" % (hr(self), hr(self._debugfinishstack))
        """
        assert self._validobject, "precondition: This instance must not already have been finish()ed." + " -- " + "self: %s, self._debugfinishstack: %s" % (hr(self), hr(self._debugfinishstack))

        self._validobject = false
        if DEBUG_MODE:
            self._debugfinishstack = traceback.extract_stack()

        self.status = status
        self.stoptime = time.time()
        if hasattr(self, 'state'):
            del self.state
        self.finish_func(self)  # the finish_func takes the usercommand object as its only parameter

    _public_methods_.append('is_finished')
    def is_finished(self):
        return self.stoptime is not None

    def default_error_callback(self, etype, evalue, stacktrace):
        """
        Useful as the 'errorfunc' parameter to self.doq.add_task.
        """
        self.excinfo = (etype, evalue, stacktrace)

        # Present the error nicely to the user:
        if type(etype) == types.StringType:
            estr = etype
        elif hasattr(etype, '__name__'):
            estr = etype.__name__
        else:
            estr = repr(etype)
        debug.mojolog.write('Command (id: %s, type: %s) encountered error of type %s.\n', args=(self.id, self.type, estr), vs='UserCommand', v=1)
        if int(confman.dict['MAX_VERBOSITY']) >= 2:
            debug.mojolog.write('Exception details:\n', vs='UserCommand')
            traceback.print_exception(etype, evalue, stacktrace, None, debug.mojolog)
        # self.finish(status='error')
        # XXX I *think* that `self.finish()' always gets called from somewhere else when there is an error.  In any case I get assertion failures due to finish() getting called twice... --Zooko 2001-06-03

    # func to call when we are actually running
    def set_make_active(self, make_active):
        self.make_active = make_active


class Download(UserCommand):
    """do nothing class providing a data interface for save to disk downloads to be on the activity list"""
    _public_attrs_ = []
    _public_attrs_.extend(UserCommand._public_attrs_)   # for COM access
    def __init__(self, servicemtm, name, mojoid):
        UserCommand.__init__(self, servicemtm)
        self.name = name[:64]
        self.mojoid = mojoid
        self.status = 'starting'

        # unsplit.ReassemblerSplitFile is passed self.set_progress for updating the status display (fyi)
    
    def set_finished(self, status='done'):
        """called directly by the thread that instantiated this object as appropriate"""
        self.status = status
        # these are set by UserCommand.start if it was ever called
        if self.doq and callable(self.finish_func):
            debug.mojolog.write('Download of %s finished\n', args=(self.mojoid,), v=3, vs='Download')
            self.doq.add_task(self.finish_func, args=(self,))



class Publish(UserCommand):
    _public_methods_ = []
    _public_methods_.extend(UserCommand._public_methods_)   # for COM access
    _public_attrs_ = []
    _public_attrs_.extend(UserCommand._public_attrs_)   # for COM access
    def __init__(self,
                 broker,
                 file,
                 ctype=None,
                 cdesc=None,
                 ctracker_id_list=None,
                 totalsharesneeded=None, totalsharesinpiece=None,
                 compression=archive.CTYPE_ZLIB, allowhtmlspider=true,
                 pub_date_cdesc_fieldname=None
                ):
        """
        @param ctracker_id_list contains a list of content tracker ids
            to submit the content description to.  If none is given, we'll
            get our list from MetaTrackerLib.
        @param totalsharesneeded the N of the N of M info dispersal splitting.  Must be a power of 2.
        @param totalsharesinpiece the M of the N or M info dispersal splitting.  Must be >= totalsharesneeded.
        @param compression specifies the file compression type to use when creating the archive.
               Must be one of archive.CTYPE_* constants.
        @param allowhtmlspider is a boolean that can be used to disable the default behavior of
               spidering html files for relative images and links.
        @param pub_date_cdesc_fieldname if specified is the name of the field in self.cdesc that will
               be filled in with the UTC ISO8601 timestamp immediately after archive creation as publication begins.
        """
        if totalsharesneeded is None:
            totalsharesneeded = int(confman['PUBLICATION']['SHARES_NEEDED'])
        if totalsharesinpiece is None:
            totalsharesinpiece = int(confman['PUBLICATION']['SHARES_GENERATED'])

        if totalsharesneeded not in (1,2,4,8,16,32,64,128):
            raise ValueError, ("totalsharesneeded must be a reasonable power of two", totalsharesneeded)
        if totalsharesinpiece < totalsharesneeded:
            raise ValueError, ("totalsharesinpiece must be >= totalsharesneeded", totalsharesinpiece)
        if (totalsharesneeded * int(confman['PUBLICATION']['MAX_SHARE_SIZE'])) > 1024*1024:
            raise ValueError, "resulting piece/chunk size from PUBLICATION configuration is too large, totalsharesneeded %s, totalsharesinpiece %s, MAX_SHARE_SIZE %s" % (totalsharesneeded, totalsharesinpiece, confman['PUBLICATION']['MAX_SHARE_SIZE'])

        self.failure_reason = None

        UserCommand.__init__(self, broker.ServicesMTM)
        self.broker = broker

        self.ctracker_id_list = ctracker_id_list
        self.totalsharesneeded = totalsharesneeded
        self.totalsharesinpiece = totalsharesinpiece
        self.compression = compression
        self.allowhtmlspider = allowhtmlspider
        self.pub_date_cdesc_fieldname = pub_date_cdesc_fieldname

        self.set_progress('Initializing\n')

        self.name = file
        self.input = {
            'file' : file,
            'ctype' : ctype,
            'cdesc' : cdesc,
            } # This dict shouldn't be modified.
        # A hack to make all the input parameters attributes of self:
        self.__dict__.update(self.input)

        self.results = {
            'Content Publication' : None, # Should be a string description.
            'Description Submission' : {'Content Tracker Results' : {}},
            }
        self.publish_ses = None

        self.set_progress('Initialized.\n')

    def start(self, doq, callback):
        UserCommand.start(self, doq, callback)
        self.status = 'initializing'

        self.doq.add_task(self.publish_file, errorfunc=self.default_error_callback)

    _public_methods_.append('cancel_publish')
    def cancel_publish(self, reason):
        """
        @precondition This instance must not already have been finish()ed.: self._validobject: "self: %s, self._debugfinishstack: %s" % (hr(self), hr(self._debugfinishstack))

        note, if a PublishSession never got created then this will return 0
        returns 1 if it sent the cancel message to the publish session
        the publish session may have already finished or shut down on it's own...
        """
        assert self._validobject, "precondition: This instance must not already have been finish()ed." + " -- " + "self: %s, self._debugfinishstack: %s" % (hr(self), hr(self._debugfinishstack))

        if self.publish_ses:
            self.publish_ses.publication_cancelled(reason)
            self.publish_ses = None
            return 1
        self.status = 'cancelled'
        self.set_progress("Cancelled by user request.")
        if self.finish_func:
            self.finish_func(self)
        return 0


    def publish_file(self):
        """
        Create the actual file tree if a directory tree or html archive is being published.
        Creates the publication archive in a new thread and begins the actual publication once the archive is ready.
        """
        startingdir = os.getcwd()
        self.tempfilename = None
        arcfile = None  # set here due to its use in the "finally:"
        try:
            # expand directories into their complete list of files
            debug.stderr.write('Publishing: file "%s" using %s of %s dispersal.\n', args=(self.file, self.totalsharesneeded, self.totalsharesinpiece), vs='UserCommand')
            filelist = []
            # strip any trailing / or \ (windows doesn't seem to like it)
            if self.file[-1] == os.sep and len(self.file) > 1:
                self.file = self.file[:-1]
    
            if not os.path.exists(self.file):
                raise Error, "File '%s' does not exist." % self.file
            
            if os.path.isdir(self.file) :
                debug.stderr.write('Publishing: expanding directory "%s"\n', args=(self.file,), v=2, vs='UserCommand')
                filelist.extend( archive.recursive_listdir(self.file) )
                for f in filelist:
                    debug.stderr.write('\t%s\n', args=(f,), v=3, vs='UserCommand')
            elif self.allowhtmlspider and htmlspider.html_file_re.match(self.file):
                debug.stderr.write("Publishing: html spider parsing '%s' for relative src's and href's\n", args=(self.file,), v=2, vs='UserCommand')
                filelist.extend( htmlspider.get_spider_files(self.file, maxdepth=10) )
                for f in filelist:
                    debug.stderr.write('\t%s\n', args=(f,), v=3, vs='UserCommand')
            else :
                # this really just appends self.file to filelist,
                filelist.extend( archive.recursive_listdir(self.file) )
            
            prefixdir = os.path.dirname(os.path.commonprefix(filelist))
            debug.stderr.write('Publishing: Using prefix dir "%s".\n' % prefixdir, vs='UserCommand')
    
            self.state['common path prefix'] = prefixdir

            # strip prefixdir from all files in filelist
            if prefixdir and os.path.isdir(prefixdir):
                # strip prefixdir of all files in filelist
                newfilelist = []
                for name in filelist :
                    afilename = name[len(prefixdir):]
                    if afilename[0] == os.sep:
                        afilename = afilename[1:]
                    newfilelist.append(afilename)
                filelist = newfilelist
    
            if not self.allowhtmlspider or not htmlspider.html_file_re.match(self.file):
                # if we did -not- generate the file list using a web spider, sort the files as best we can
                def html_tree_cmp(a, b):
                    """Always put index.html first in any archive, DOS DOS DOS!"""
                    a = string.lower(a)
                    b = string.lower(b)
                    if a == 'index.html' or a == 'index.htm':
                        return -1
                    elif b == 'index.html' or b == 'index.htm':
                        return 1
                    else:
                        return cmp(a,b)
                filelist.sort(html_tree_cmp)
    
            self.state['sorted file list'] = filelist

            # pick a temporary file name
            self.tempfilename = apply(os.path.join, mojoutil.get_tempfile_name(tmpdir=os.path.expandvars(confman.dict["PATH"]["TMP_DIR"]), pre="mojohttpd.publish.", post=".mojoarc"))
    
            self.set_progress('creating archive for publication\n')

            debug.stderr.write('Publishing: Creating archive in temporary file %s.\n' % self.tempfilename, vs='UserCommand')
            self.status = 'initializing'
            rt = LoggedThread(target=self.create_archive, args=(filelist, prefixdir), name="%s create archive" % `self`)
            rt.setDaemon(1)
            rt.start()
            
        except:
            debug.mojolog.write("Publishing Failed.\n")
            traceback.print_exc(file=debug.mojolog)
            if self.tempfilename and os.path.isfile(self.tempfilename) :
                try:
                    os.unlink(self.tempfilename)
                except OSError, e:
                    debug.stderr.write("Ignoring error %s removing %s\n", args=(e, self.tempfilename), vs="UserCommand")
            raise # The error callback will handle this.


    def create_archive(self, filelist, prefixdir=''):
        if not prefixdir:
            prefixdir=0
        try:
            arc = archive.Creator(compression=self.compression)
            if not prefixdir:
                prefixdir = ''
            arc.queuefiles(filelist, pathprefix=prefixdir)
            arc.writeto(self.tempfilename)
            del arc

            self.state['archive path'] = self.tempfilename

            arcfile = open(self.tempfilename, "rb")
            arcsize = os.path.getsize(self.tempfilename)
            self.servicemtm._doQ.add_task(self.start_publish_session, args=(arcfile, arcsize))
        except:
            debug.stderr.write("Publishing Failed.")
            traceback.print_exc(file=debug.mojolog)
            if self.tempfilename and os.path.isfile(self.tempfilename) :
                try:
                    os.unlink(self.tempfilename)
                except OSError, e:
                    debug.stderr.write("ERROR %s removing %s", args=(e, self.tempfilename), vs="UserCommand")
            self.finish('failed')
                                         

    def start_publish_session(self, arcfile, arcsize):
        if self.status == 'initializing':
            if self.pub_date_cdesc_fieldname:
                self.cdesc[self.pub_date_cdesc_fieldname] = mojoutil.iso_utc_time()
            self.status="queued"
            self.publish_ses = PublishSession(MTM=self.servicemtm,
                                              arcfile=arcfile, length=arcsize,
                                              statusfunc=self.set_progress,
                                              callback=self.handle_results_of_publish,
                                              errorfunc=self.default_error_callback,
                                              totalsharesneeded=self.totalsharesneeded,
                                              totalsharesinpiece=self.totalsharesinpiece)


    def handle_results_of_publish(self, failure_reason, compressed_dinode):
        if failure_reason:
            self.failure_reason = failure_reason
            debug.stderr.write('Publishing failed: %s\n' % (failure_reason,), vs='UserCommand')
            self.set_progress(str(failure_reason))
            if self.tempfilename and os.path.isfile(self.tempfilename) :
                try:
                    os.unlink(self.tempfilename)
                except OSError, e:
                    debug.stderr.write("ERROR %s removing %s", args=(e, self.tempfilename), vs="UserCommand")
            if failure_reason == 'cancelled by user':
                self.finish('cancelled')
            else:
                self.finish('error')
            return

        self.set_progress( 'Block distribution complete.\n')

        self.state['ascii dinode'] = mojosixbit.b2a(compressed_dinode)
        self.results['dinode'] = self.state['ascii dinode']
        self.results['Content Publication'] = 'Successfully Completed'
        if self.tempfilename and os.path.isfile(self.tempfilename) :
            try:
                os.unlink(self.tempfilename)
            except OSError, e:
                debug.stderr.write("ERROR %s removing %s", args=(e, self.tempfilename), vs="UserCommand")
        self.doq.add_task(self.publish_content_description, errorfunc=self.default_error_callback) # This used to be a straight-forward call, but a finer granularity on the DoQ seems like a Good Thing.

    def publish_content_description(self):
        if (not self.ctype) or self.ctype == "NONE":
            self.set_progress('Skipping content description publication, no content type supplied.\n')
            self.set_progress('Publishing completed successfully (no content description submitted)')
            self.finish()
            return

        self.set_progress( 'Publishing content description.\n')

        descvalues = self.cdesc.copy()
        descvalues['__DINODE__'] = self.state['ascii dinode']

        self.submitxml = contentlib.make_contentdesc_xml(self.ctype, descvalues)
        self.state['submission XML'] = self.submitxml

        if self.ctracker_id_list:
            # if we were given a list to submit to, do it directly
            ctracker_infotuple_list = map(lambda a, self=self: (a, MetaTrackerLib.get_contact_info(self.servicemtm, a, timeout=60)), self.ctracker_id_list)
            self.doq.add_task(self.handle_find_content_trackers, args=(ctracker_infotuple_list,))
        else:
            contentlib.find_content_trackers_and_call_back(self.servicemtm, callback=self.handle_find_content_trackers, min=10, max=MAX_CONTENT_TRACKERS)

    def handle_find_content_trackers(self, trackers):
        if type(trackers) == types.DictType:
            content_tracker_tuple_list = trackers.items()
        elif type(trackers) == types.ListType:
            content_tracker_tuple_list = trackers
        # XXX contentlib.FOO calls are weird, this should
        # happen in there and they should do the find call
        # for us...  (except that we're using this handle_find
        # callback as a hack for now to publish to a pre-specified
        # set of trackers)
        content_tracker_tuple_list = self.servicemtm.get_handicapper().sort_by_preference(content_tracker_tuple_list, 'content tracker submit', {'XML data': self.submitxml})

        self.tracker_submissions = mojoutil.Counter()
        self.content_tracker_successes = mojoutil.Counter()
        self.content_tracker_fails = mojoutil.Counter()
        total = len(content_tracker_tuple_list)
        
        self.results['Description Submission']['Number of Submissions'] = self.tracker_submissions.get()
        self.results['Description Submission']['Number of Successes'] = self.content_tracker_successes.get()
        debug.stderr.write('Submitting content tracking association to these Content Trackers:\n', vs='UserCommand', v=6)

        preferenceno = 0
        for content_tracker_tuple in content_tracker_tuple_list:
            idascii = idlib.to_ascii(content_tracker_tuple[0])
            cb = lambda widget, outcome, failure_reason, self=self, idascii=idascii: \
                       self.handle_content_tracker_submit(widget, outcome, failure_reason, idascii)

            debug.stderr.write('Trying Content Tracker whose ascii encoded ID is "%s"...\n', args=(idascii,), vs='UserCommand', v=6)
            self.tracker_submissions.inc()
            self.results['Description Submission']['Number of Submissions'] = self.tracker_submissions.get()
            contentlib.content_tracker_submit(content_tracker_tuple, self.servicemtm, self.submitxml, callback=cb, timeout=90, preferenceno=preferenceno)
            preferenceno=preferenceno+1

    def handle_content_tracker_submit(self, widget, outcome, failure_reason, idascii):
        if failure_reason or outcome.get('result') == 'failure' :
            self.results['Description Submission']['Content Tracker Results'][idascii] = 'failed'
            debug.stderr.write('Submission to "%s" failed.\n', args=(idascii,), vs='UserCommand', v=6)
            self.content_tracker_fails.inc()
        else:
            if not self.results.has_key('mojoid'):
                self.results['mojoid'] = contentlib.get_mojoid_from_asciidinode(self.results['dinode'])
            self.results['Description Submission']['Content Tracker Results'][idascii] = 'success'
            debug.stderr.write('Submission to "%s" was successful.\n', args=(idascii,), vs='UserCommand', v=6)
            self.content_tracker_successes.inc()
            self.results['Description Submission']['Number of Successes'] = self.content_tracker_successes.get()

        self.set_progress("Content tracker submission: %s pass, %s fail out of %s total trackers.\n" %
                          (self.content_tracker_successes.get(), self.content_tracker_fails.get(), self.tracker_submissions.get()))
        # XXX confirm the content tracker really accepted it!
        if self.content_tracker_successes.get() + self.content_tracker_fails.get() >= self.tracker_submissions.get():
            debug.stderr.write('%s tries with %s successes...  Done for now.  (XXX Note: Make the number of tries configurable.)\n',
                               args=(self.tracker_submissions.get(), self.content_tracker_successes.get(),), vs='UserCommand', v=7)
            if self.content_tracker_successes.get() == 0:
                self.set_progress( 'Description was not accepted by any content trackers.  Searches will not show this publication.')
                self.finish(status = 'error')
            # XXX more cleanup could probably be done here...
            self.set_progress( 'Publishing completed successfully.\n')
            self.finish()

            
    def set_progress(self, value, briefstatus=None):
        if hasattr(self, 'failure_reason') and self.failure_reason:
            return  # we already failed, don't bother
        if value == 'Publication starting...\n':
            self.make_active(self)
            self.status = "running"
        UserCommand.set_progress(self, value, briefstatus)


