#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# Objects and functions for handling archives.
#
# $Id: archive.py,v 1.16 2001/05/03 23:09:37 greg Exp $

######## libraries ########

import os
import stat
import string
import sys
import time
import xdrlib    # We use RFC-1014 xdr data formats
import re
import zlib


import debug
from human_readable import hr
import mojoutil

######## module configuration globals ########
true, false = 1, None

# the buffer size to use while reading or writing archive data
# NOTE: this is only used on files that don't implement ReassemblerSplitFile's extended optimal read interface.
IOBufferSize = 128*1024

######## constants ########

# These two are for the file header; they must total 8 bytes
HeaderID = "\032EGBT"   # for magic file identification
FormatVersion = "003"   # the file format version

# Do not change these without changing the FormatVersion, they define the archive format.
MAX_PATH_LENGTH = 128   # the maximum length of a path+filename (must be multiple of four)
DATE_FORMAT_STR = '%Y%m%d%H%M%S'
DATE_STR_LENGTH = len( time.strftime( DATE_FORMAT_STR, time.gmtime( time.time() ) ) )
# round up to the next multiple of four
if (DATE_STR_LENGTH % 4) != 0 :
    DATE_STR_LENGTH = DATE_STR_LENGTH + (4 - (DATE_STR_LENGTH % 4))

# value of ctype field indicating the zlib compression format
CTYPE_STR_LENGTH = 4
CTYPE_NONE = 'none'
CTYPE_ZLIB = 'zlib'

# Maximum length of an index entry
MaxIndexEntryLength = 4 + MAX_PATH_LENGTH + DATE_STR_LENGTH + CTYPE_STR_LENGTH + 8 + 8 + 8

# a translation table for using /s in stored pathnames
PathXlate = string.maketrans('\\', '/')

######## exceptions ########

class ArchiveError(StandardError) :
    pass

class NotReady(ArchiveError) :
    args = "The archive was not ready to read, seek to a part first"
class NameTooLong(ArchiveError) :
    args = "Pathname too long for archive"
class NotAnArchive(ArchiveError) :
    args = "File is not an archive"
class WrongVersion(ArchiveError) :
    args = "Unsupported archive version"
class IndexFormatError(ArchiveError) :
    args = "Bad length, csize or offset found in archive index"
class BadPathname(ArchiveError) :
    args = "Pathnames in archive may not be absolute or reference .."
class BadArchivePart(ArchiveError) :
    args = "Part does not exist in the archive"
class UnknownCtypeError(ArchiveError) :
    args = "Unknown compression type specified"
class CompressionError(ArchiveError) :
    args = "Compression library raised an error"
class CompressionMemoryError(ArchiveError, MemoryError):
    args = "decompression tried to use too much memory"


#########################################################################
# This object is more of a struct or record containing the fields of a
# file header in the archive index.
#########################################################################
class IndexEntry : 
    # these allow access via COM
    _public_methods_ = ['get_pathname', 'get_local_pathname']
    _public_attrs_ = ['date', 'ctype', 'size', 'csize', 'offset']

    def __init__(self, name=None, pathprefix='', normalizepaths=None) :
        self.__pathname = ''  # full path + filename within the archive
        self.date = ''      # a string containing YYYYMMDDHHMMSS
        self.ctype = CTYPE_NONE    # compression method
        self.size = 0L      # length of the original file
        self.csize = 0L     # size of the compressed file data
        self.offset = 0L    # absolute offset within the archive
        self.__normalizepaths=normalizepaths

        # pathprefix is not stored in the index, it is here solely for
        # use when creating or extracting archives to a given directory
        # without actually changing to that directory.
        self.pathprefix = pathprefix

        if name is not None :
            self.fillin(name)

    #####################################################################
    # Use this to access the pathname since it must not be changed
    # from outside this object (unlike the other member variables).
    #####################################################################
    def get_pathname(self) :
        """get_pathname() - return the pathname in this entry"""
        return self.__pathname

    #####################################################################
    def get_local_pathname(self) :
        """get_local_pathname() - return the local pathname of the file for this entry"""
        return os.path.normpath(os.path.join(self.pathprefix, self.__pathname))

    #####################################################################
    # Load an entry from a file object
    #####################################################################
    def read(self, file) :
        # read the length of this index entry
        hdrdata = file.read(4)
        entrylen = xdrlib.Unpacker(hdrdata).unpack_uint()
        if entrylen > MaxIndexEntryLength or entrylen <= 0 :
            raise IndexFormatError

        # read the index entry
        hdrdata = file.read(entrylen)
        if len(hdrdata) != entrylen :
            raise IOError, "too few bytes read when reading archive index"
        hdr = xdrlib.Unpacker(hdrdata)

        # the path is packed as a string in write(), however we unpack
        # it manually so that we can sanity check the length to avoid
        # insane values.
        fullpath_len = hdr.unpack_uint()
        if fullpath_len > MAX_PATH_LENGTH or fullpath_len <= 0 :
            raise NameTooLong
        fullpath = hdr.unpack_fstring(fullpath_len)

        if self.__normalizepaths:
            # fixup multiple /s, convert to \s on windows, etc..
            fullpath = os.path.normpath(fullpath)

        # All this just to remove the stupid padding NULLs..
        self.__pathname = string.translate(fullpath,
            string.maketrans("", ""), "\000")

        self.date = hdr.unpack_fstring(DATE_STR_LENGTH)
        self.ctype = hdr.unpack_fstring(CTYPE_STR_LENGTH)
        self.size = hdr.unpack_uhyper()
        self.csize = hdr.unpack_uhyper()
        self.offset = hdr.unpack_uhyper()
        hdr.done()

    #####################################################################
    def get_binary(self) :
        """get_binary() - Return a binary string representation of this entry"""
        hdr = xdrlib.Packer()
        # standardize the path seperator in the name we store
        hdr.pack_string(string.translate(self.__pathname, PathXlate))
        hdr.pack_fstring(DATE_STR_LENGTH, self.date)
        hdr.pack_fstring(CTYPE_STR_LENGTH, self.ctype)
        hdr.pack_uhyper(self.size)
        hdr.pack_uhyper(self.csize)
        hdr.pack_uhyper(self.offset)
        entrystr = hdr.get_buffer()
        # return it encoded with its length preceeding it
        entrylen = xdrlib.Packer()
        entrylen.pack_uint(len(entrystr))
        return entrylen.get_buffer() + entrystr

    #####################################################################
    def write(self, file) :
        """write(file) - write this entry to the given file object"""
        file.write(self.get_binary())

    #####################################################################
    def len(self) :
        """len() - return the length in bytes of this entry"""
        return len(self.get_binary())

    #####################################################################
    # Fill this index entry in using the status of the given filename
    #####################################################################
    def fillin(self, name) :
        if len(name) > MAX_PATH_LENGTH :
            raise NameTooLong, name
        self.__pathname = name
        localname = os.path.join(self.pathprefix, name)
        modtime = os.path.getmtime(localname)

        # NOTE: we are no longer storing the date in the archive in
        # order to promote canonical archives since file dates are often
        # different for the same data on different computers.  If two
        # agents publish an archive containing an identical file it
        # should generate the same archive and therefore the same blocks.
        #old#self.date = time.strftime(DATE_FORMAT_STR, time.gmtime(modtime))
        self.date = "date not stored"[:DATE_STR_LENGTH]

        self.ctype = CTYPE_NONE    # default to uncompressed
        # XXX Python 1.5.2 on doesn't support > 32-bit sizes..
        self.size = os.path.getsize(localname)
        self.csize = 0L   # *unknown* until written
        self.offset = 0L  # *unknown* until written



#########################################################################
# This object can be used to read parts (files) out of an archive.  It
# handles the archive index lookups and decompression of compressed
# data.
#########################################################################
class Reader :
    # these allow access via COM
    _public_methods_ = ['extract_all', 'extract_part',
        'get_parts_list', 'seekpart', 'seekfirstpart', 'seeknextpart',
        'getpartname', 'readcomp', 'extract', 'com_get_index']
    _public_attrs_ = []

    #####################################################################
    # Archive readers needs an open file like object on
    # initialization.
    #####################################################################
    def __init__(self, file, normalizepaths=None) :
        self.index = []          # a list of IndexEntry objects
        self.indexhash = {}      # maps names to indices of self.index
        self.currentpart = None  # the current parts index number
        self.partpos = 0L        # position w/in the current part
        self.decompressor = zlib.decompressobj()  # used to decompress parts
        self.file = file
        self.__normalizepaths = normalizepaths

        # check the file header
        self.file.seek(0)
        hdrstr = self.file.read(len(HeaderID))
        if (hdrstr != HeaderID) :
            raise NotAnArchive, hdrstr
        hdrstr = self.file.read(len(FormatVersion))
        if (hdrstr != FormatVersion) :
            raise WrongVersion, hdrstr

        # load the archive index
        self.loadindex()
    
    def com_get_index(self):
        """Return the index as a list of COMified IndexEntry objects"""
        import win32com.server.util
        return map(win32com.server.util.wrap, self.index)

    def extract_all(self, destdir, allowexceptions=false, exceptionstoignore=(), download_monitor=None):
        """
        This extracts the archive to the local file-system.  The structure of the archive is mirrored in the directory structure below destdir.

        If allowexceptions is true, ArchiveError and any exceptions listed in exceptionstoignore will be caught and ignored.  This allows
        the extraction of an archive from a "file" that has unreadable/unretrievable parts.
        """
        if allowexceptions:
            es_tocatch = (ArchiveError,) + tuple(exceptionstoignore)
        else:
            class Dummy: pass
            es_tocatch = (Dummy,)

        for part in self.get_parts_list():
            try:
                self.extract_part(part, destdir)
            except es_tocatch, e:
                debug.mojolog.write("Ignoring error extracting part %s: %s\n", args=(part, e), v=2, vs='archive')
                if download_monitor:
                    download_monitor.increment_error_count(partname=part)

    def extract_part(self, part, destpath, usepartpath=true, sanepaths=true, overwrite=true, makedirs=true, mode=0700):
        """
        Writes the named part to a local file.
        
        part - the archive part to extract.
        destpath - a path which is either an absolute path or a base path, based on usepartpath.
        usepartpath - if true the absolute path of the file will be a combination of the base "destpath" and the archive-relative path for this part.
        sanepaths - if true, basic sanity checks and modifications are done to the archive-relative path before using it.  (It's a bad idea for this to be false!)
        overwrite - specifies if existing files should be overwritten
        makedirs - create subdirectories as needed when extracting
        mode - the unix file mode to create directories and files with
        """
        relpath = ''
        if usepartpath:
            # remove relative path references (eggs/bar/../spam.gif -> eggs/spam.gif)
            # This has the "feature" that excess /..s will simply stay in the top directory
            if sanepaths and string.count(part, '/..') :
                def replace_last_pathbit(x, y) :
                    try:
                        return x[:string.rindex(x, '/')] + y
                    except ValueError:
                        return x+y
                relpath = reduce(replace_last_pathbit, string.split('/'+part, '/..'))[1:]
            else:
                relpath = part
        abspath = os.path.join(destpath, relpath)

        dirname = os.path.dirname(abspath)
        if makedirs and dirname and not os.path.isdir(dirname):
            os.makedirs(dirname, mode=mode)

        if os.path.exists(abspath) and not overwrite:
            raise ArchiveError, "file %s already exists" % abspath

        self.seekpart(part)
        f = open(abspath, 'wb')
        self.extract(f, cbytes=-1)
        f.close()
        os.chmod(abspath, mode & 0666)

    def get_parts_list(self, subpath=''):
        """
        Returns a list of part strings.
        @param subpath restricts the list to entries beginning with subpath and removes subpath from the returned strings.
        """
        partslist = []
        for entry in self.index:
            partslist.append(entry.get_pathname())
        if subpath:
            # only return parts starting with subpath
            partslist = filter(lambda p, subpath=subpath: p[:len(subpath)], partslist)
            # strip the subpath prefix off of the returned parts
            partslist = map(lambda p, strip=strip: p[strip:], partslist)
        return partslist
    
    def get_parts_tree(self, partslist=None):
        """
        Returns a recursive dictionary representing the directory tree stored in this archive as follows:
          * keys are filenames at the current level.
          * keys map to:
            + (filesize,)
            + dicts of the same format (meaning they are subdirectories)
        
        @param partslist is used during recursion, do not supply one
        """
        if partslist is None:
            partslist = []
            for entry in self.index:
                partslist.append( (entry.get_pathname(), entry.size) )
            partslist.sort()
        treedict = {}
        pl_idx = 0
        while pl_idx < len(partslist):
            part, size = partslist[pl_idx]
            pathsepidx = string.find(part, '/')
            if pathsepidx >= 0:
                path = part[:pathsepidx]
                if treedict.has_key(path):
                    assert 0, "encountered part %s when path %s was already in treedict.  partslist = %s" % (part, path, partslist)
                else:
                    treedict[path] = {}
                # take care adding the whole subdirectory at once via recursion
                sublist=[]
                while (pl_idx < len(partslist)) and (part[:pathsepidx] == path):
                    # pass the recursive function the subdirectory with this directory name stripped off
                    sublist.append( (part[pathsepidx+1:], size) )
                    pl_idx = pl_idx + 1
                    if pl_idx < len(partslist):
                        part, size = partslist[pl_idx]
                treedict[path].update( self.get_parts_tree(partslist=sublist) )
            else:
                treedict[part] = (size,)
                pl_idx = pl_idx + 1
        return treedict

    #####################################################################
    # Return a string containing an HTML formatted archive index.
    # The string is also valid XHTML [http://www.w3.org/TR/xhtml1/]
    #####################################################################
    def htmlindex(self) :
        buf = ""
        buf = buf + "<h1>Archive Index</h1>\n"
        buf = buf + "<hr /><p>\n"
        buf = buf + '<table width="66%"><tr><th align="left">Pathname</th><th align="right">Bytes</th><th align="right">Date</th></tr>\n'
        # html escape (&#xx;) any special characters in pathname
        for entry in self.index :
            path = entry.get_pathname()
            htmlpath = string.replace(path, '<', '&lt;')
            htmlpath = string.replace(htmlpath, '>', '&gt;')
            htmlpath = string.replace(htmlpath, '&', '&#38;')
            # make a printable number out of the size
            sizestr = str(entry.size)
            if sizestr[-1] == 'L' :  # stupid L from python 1.5.2 longs..
                sizestr = sizestr[:-1]
            datestr = entry.date
            # remove trailing nulls
            while datestr and datestr[-1] == '\0' :
                datestr = datestr[:-1]
            buf = buf + ( '<tr><td align="left"><a href="%s">%s</a></td><td align="right">%s</td><td align="right">%s</td></tr>\n' % (htmlpath, htmlpath, sizestr, datestr) )
        buf = buf + "</table>\n</p><hr />\n"
        return buf


    #####################################################################
    # Read and parse the archive index; replacing any currently loaded
    # index.
    #####################################################################
    def loadindex(self) :
        # get the number of index entries and their total length
        rawnum = self.file.read(8)
        up = xdrlib.Unpacker(rawnum)
        entries = up.unpack_uint()
        indexbytes = up.unpack_uint()

        prevcsize = 0
        prevoffset = len(HeaderID) + len(FormatVersion) + 4 + 4

        # read the index
        while entries > 0 :
            entry = IndexEntry(normalizepaths=self.__normalizepaths)
            entry.read(self.file)

            # Sanity check the index.  All index entries must occur in
            # the order that the data is stored in the file.
            if entry.offset < (prevoffset + prevcsize) :
                raise IndexFormatError, "entry.offset was: "+str(entry.offset)

            # TODO do more syntax checking here to avoid finding problems
            # later while reading and extracting data and to avoid
            # security problems with trojan archive headers.

            if entry.ctype != CTYPE_NONE and entry.ctype != CTYPE_ZLIB :
                raise UnknownCtypeError, entry.ctype

            prevcsize = entry.csize
            prevoffset = entry.offset

            self.indexhash[entry.get_pathname()] = len(self.index)
            self.index.append(entry)
            entries = entries - 1


    #####################################################################
    # Seek to the location of the specified name in the archive, an
    # exception is raised if it is not found in the index.  Return its
    # IndexEntry or None if name wasn't in the index.
    #####################################################################
    def seekpart(self, name) :
        """@param name can be found by IndexEntry.get_pathname()"""
        if not self.indexhash.has_key(name) :
            raise BadArchivePart

        self.currentpart = self.indexhash[name]
        self.partpos = 0L
        self.decompressor = zlib.decompressobj()  # used to decompress parts
        self.file.seek(self.index[self.currentpart].offset)

        return self.index[self.currentpart]


    #####################################################################
    # Seek to first part and return its IndexEntry
    #####################################################################
    def seekfirstpart(self) :
        self.currentpart = 0
        self.partpos = 0L
        self.decompressor = zlib.decompressobj()  # used to decompress parts
        self.file.seek(self.index[self.currentpart].offset)

        return self.index[self.currentpart]


    #####################################################################
    # Seek to the next part and return its IndexEntry or None if there
    # are no more parts.
    #####################################################################
    def seeknextpart(self) :
        if self.currentpart is None :
            return self.seekfirstpart()

        # don't go past the end
        if self.currentpart+1 >= len(self.index) :
            return None

        self.currentpart = self.currentpart + 1
        self.partpos = 0L
        self.decompressor = zlib.decompressobj()  # used to decompress parts

        self.file.seek(self.index[self.currentpart].offset)
        return self.index[self.currentpart]
    

    #####################################################################
    def getpartname(self) :
        """getpartname() - return the pathname of the current part"""
        return self.index[self.currentpart].get_pathname()


    #####################################################################
    # Read up to the specified number of bytes from the current
    # position in the archive.  Seekpart must have been called prior
    # to this.  If bytes is negative, the entire part is read and
    # returned no matter how huge it may be (Caller be warned!).
    # [An easy denial-of-service on a machine would be to
    # insert data w/ a compression ratio of 1000:1]
    #
    # Note: bytes specifies the size of the compressed data, the size
    # of the data actually returned could be much larger or smaller.
    # An empty string can be returned if enough data to seed the
    # decompressor was not read.
    #####################################################################
    def readcomp(self, cbytes=-1) :
        if self.currentpart is None :
            raise NotReady

        cbytesleft = self.index[self.currentpart].csize - self.partpos

        # keep within the part boundaries
        if (cbytes < 0) or (cbytes > cbytesleft) :
            cbytes = cbytesleft

        #print "DEBUG: bytesleft %d, partpos %d, csize %d" % (bytesleft, self.partpos, self.index[self.currentpart].csize)

        buf = self.file.read(cbytes)
        self.partpos = self.partpos + len(buf)

        # decompress data in buf if needed
        if self.index[self.currentpart].ctype == CTYPE_ZLIB :
            try:
                buf = self.decompressor.decompress(buf)
                # flush the decompressor at the end of a part
                if cbytes == cbytesleft :
                    flushed = self.decompressor.flush()
                    if flushed :
                        buf = buf + flushed
                    # reset the part decompressor object
                    self.decompressor = zlib.decompressobj()
            except zlib.error, e:
                raise CompressionError, e
            except MemoryError, e:
                # zlib tried to allocate too much memory during decompression due to a super high compression ratio
                raise CompressionMemoryError, e

        return buf
        

    #####################################################################
    # Extract the current seeked to part into the dest file object;
    # stopping after the specified number of compressed data bytes,
    # not going past the end of the part.  If cbytes < 0 the entire
    # part is extracted.
    #
    # Note: cbytes specifies the size of the compressed data, the size
    # of the data actually extracted could be much larger or smaller.
    #####################################################################
    def extract(self, dest, cbytes=-1) :
        if self.currentpart is None :
            raise NotReady, "One must seekpart first to reach enlightenment, grasshopper!"

        # keep within the part boundaries
        cbytesleft = self.index[self.currentpart].csize - self.partpos
        if (cbytes < 0) or (cbytes > cbytesleft) :
            cbytes = cbytesleft

        #print "DEBUG: bytesleft %d, partpos %d, csize %d" % (bytesleft, self.partpos, self.index[self.currentpart].csize)

        if hasattr(self.file, 'get_optimal_read_length') and callable(self.file.get_optimal_read_length):
            optimal_read = 1
        else:
            optimal_read = 0

        buf = "oink"
        while cbytes > 0 and len(buf) > 0 :
            if optimal_read:
                bytestoread = self.file.get_optimal_read_length()
                bytestoread = min(cbytes, bytestoread)
            else:
                bytestoread = min(cbytes, IOBufferSize)

            buf = self.file.read(bytestoread)
            assert buf is not None, "buf: %s" % hr(buf)

            self.partpos = self.partpos + len(buf)

            cbytes = cbytes - len(buf)

            # decompress data in buf if needed
            if self.index[self.currentpart].ctype == CTYPE_ZLIB :
                try:
                    increment = min(8192, IOBufferSize)
                    idx = 0
                    while idx < len(buf):
                        dest.write( self.decompressor.decompress(buf[idx:idx+increment]) )
                        idx = idx + increment
                    # flush the decompressor at the end of a part
                    if cbytes <= 0 :
                        dest.write( self.decompressor.flush() )
                        # reset the part decompressor object
                        self.decompressor = zlib.decompressobj()
                except zlib.error, e:
                    raise CompressionError, e
                except MemoryError, e:
                    # zlib tried to allocate too much memory during decompression due to a super high compression ratio
                    raise CompressionMemoryError, e
            else:
                dest.write(buf)
    


#########################################################################
# This object is used to create an archive.
#########################################################################
class Creator :
    """archive.Creator([compression=CTYPE_NONE]) - Return an archive Creator object
    Compression specifies the compression method to use when storing files:
      archive.CTYPE_NONE = none
      archive.CTYPE_ZLIB = zlib
    """
    # these allow access via COM
    _public_methods_ = ['queuefiles', 'writeto']
    _public_attrs_ = []

    def __init__(self, compression=CTYPE_NONE) :
        if compression != CTYPE_NONE and compression != CTYPE_ZLIB :
            raise UnknownCtypeError, compression
        self.compression = compression
        self.indexhash = {}
        self.index = []
        self.indexbytes = 0  # number of bytes the index occupies


    #####################################################################
    def queuefiles(self, namelist, pathprefix='') :
        """
        Queue a list of files to be added to the archive when the write
        method is called.  Files will be queued using the current
        archive setting for compression type.

        Absolute and pathnames containing .. are not allowed.

        @param pathprefix - specifies a common path prefix to prepend
        to filenames before trying to open them.  The prefix will not
        be stored in the archive.
        """
        num = len(self.index)

        for name in namelist :
            name = str(name)   # XXX fix because string.translate below barfs on unicode strings in python 2.0

            # prevent drive letters, absolute or relative pathnames.
            # This will need updating on the mac.
            if re.search(r"(^.:|^(/|\\)|(/|\\)\.\.)", name) :
                raise BadPathname, "Absolute or relative pathnames are not allowed " + name
               
            # Don't add duplicate files and only use forward slashes
            name = string.translate(name, PathXlate)
            if self.indexhash.has_key(name) :
                continue

            if len(name) > MAX_PATH_LENGTH :
                raise NameTooLong

            # Create an index entry for this file
            entry = IndexEntry(name, pathprefix=pathprefix)
            self.index.append(entry)
            self.indexbytes = self.indexbytes + entry.len()

            # save the index in the hash
            self.indexhash[name] = num
            num = num + 1

    #####################################################################
    # Write the archive header + index as well as each files
    # data (compressing as needed) to the given file object, followed
    # by rewriting the archive header with the correct offsets and
    # compressed data lengths.
    #
    # Returns the total length of the archive
    #####################################################################
    def writeto(self, filename) :
        f = open(filename, "wb")

        # write archive header, version number
        f.write(HeaderID)
        f.write(FormatVersion)

        # write the number of index entries and the index length in bytes
        numentries = xdrlib.Packer()
        numentries.pack_uint(len(self.index))
        numentries.pack_uint(self.indexbytes)
        f.write(numentries.get_buffer())
        del numentries

        # write incomplete headers to reserve space
        for entry in self.index :
            entry.write(f)

        # write out each files data, compressing as desired
        for entry in self.index :
            entry.offset = f.tell()

            compressor = None
            if self.compression == CTYPE_ZLIB :
                compressor = zlib.compressobj()
                entry.ctype = self.compression

            # copy the file into the archive
            buf = "moo"
            infile = open(entry.get_local_pathname(), "rb")
            while len(buf) > 0 :
                buf = infile.read(IOBufferSize)

                if compressor :
                    try:
                        zbuf = compressor.compress(buf)
                        entry.csize = entry.csize + len(zbuf)
                        f.write( zbuf )
                        if len(buf) <= 0 :
                            zbuf = compressor.flush()
                            entry.csize = entry.csize + len(zbuf)
                            f.write( zbuf )
                    except zlib.error, e:
                        raise CompressionError, e
                else :
                    f.write(buf)

            infile.close()

            if not compressor :
                entry.csize = entry.size

        totalbytes = f.tell()

        # go back to the beginning and fill in the headers
        f.seek(len(HeaderID)+len(FormatVersion)+4+4)
        for entry in self.index :
            entry.write(f)

        f.close()

        return totalbytes


########################################################################
########################################################################
def recursive_listdir(path, followlinks=1, visited=None) :
    """
    In short: A recursive version of os.listdir().
    Return a list of all files within the given directory
    tree in a breadth first search order.  The followlinks
    parameter specifies whether symbolic links should be followed or
    ignored.  The visited parameter is for internal (recursive) use
    only to prevent loops due to circular links or bad filesystems.
    """
    #print 'recursive_listdir', path, followlinks, visited
    if not visited:
        visited = {}

    if not os.path.isdir(path) :
        # if path is a file itself, fake a call using its directory with a
        # directory listing containing only that file.  The return value
        # will be [path] as desired.
        files = [os.path.basename(path)]
        path = os.path.dirname(path)
    else:
        files = []

    # avoid infinite loops, never visit anyplace twice
    if visited.has_key(os.path.abspath(path)) :
        return []

    visited[os.path.abspath(path)] = None  # mark this path as visited

    listing = []  # used to build our resulting list

    dirs = []     # a list of subdirectories we need to visit
    if not files:
        files = os.listdir(path)

    for name in files :
        fullname = os.path.join(path, name)
        if not followlinks and os.path.islink(fullname) :
            continue  # skip symbolic links?

        if os.path.isdir(fullname) :
            dirs.append( fullname )
        else :
            listing.append(fullname)
    
    for fullname in dirs :
        subdir = recursive_listdir(fullname, followlinks=followlinks, visited=visited)
        listing.extend(subdir)

    return listing



mojo_test_flag = 1
import RunTests

def test_recursive_listdir_empty():
    dirname = RunTests.get_next_dir_name()
    assert recursive_listdir(dirname) == []

def test_recursive_listdir_single():
    dirname = RunTests.get_next_dir_name()
    filename = os.path.join(dirname, 'spamfile')
    open(filename, 'w').close()  # create an empty file
    assert recursive_listdir(filename) == [filename], "should have returned a single file"
    assert recursive_listdir(dirname) == [filename], "should have returned a single file"

def test_recursive_listdir_onelevel():
    dirname = RunTests.get_next_dir_name()
    filename = os.path.join(dirname, '0spamfile')
    open(filename, 'w').close()  # create an empty file
    dirname2 = os.path.join(dirname, '1subdir')
    os.mkdir(dirname2)
    filename2 = os.path.join(dirname2, 'file2')
    assert (recursive_listdir(filename) == [filename])
    assert (recursive_listdir(dirname) == [filename]), "empty subdirectory shouldn't have been included in returned list"
    open(filename2, 'w').close()  # create empty file number 2
    assert (recursive_listdir(dirname) == [filename, filename2])
    assert (recursive_listdir(dirname2) == [filename2])



########################################################################
########################################################################
def _flat_unarchive(filename) :
    """
    Given a filename, extract the archive into the current
    directory, not making any subdirectories.  (this was written for
    manual archive testing; it does not support subdirectories).

    The extension .extracted will be appended to all filenames extracted
    from the archive.
    """
    file = open(filename, "rb")
    arc = Reader(file, normalizepaths=1)
    while arc.seeknextpart() :
        print 'Extracting', arc.getpartname(),
        f = open(arc.getpartname() + '.extracted', 'wb')
        firstbytes = arc.readcomp(100)
        print '[readcomp(100) got', len(firstbytes), 'bytes]...',
        f.write(firstbytes)
        arc.extract(f)
        print 'done.'


########################################################################
# This is the main program when not used as a module.
########################################################################
def run():
    # expect the archive name as the first argument and a list of
    # files for the remaining arguments.

    # check for valid command line arguments
    if len(sys.argv) <= 2 and sys.argv[1] != 'TEST':
        print "Usage:\n  "+sys.argv[0]+" archive_name [options] filename[s]"
        print "  -z   enables zlib data compression"
        print "  -l   lists the contents of the archive"
        print "  -lt  lists the contents as a tree"
        print "  -x   option causes the archive to be extracted; this will"
        print "       not work on archives with subdirectories!  All extracted"
        print "       files will have .extracted appended to their name."
        print "       *** FOR TESTING ONLY ***"
        print ""
        print "  If archive name is TEST, RunTests will be run on the module instead"
        sys.exit()

    files = []
    arcname = sys.argv[1]

    if arcname == 'TEST':
        RunTests.runTests('archive')
        return

    # Add each argument, recursing if it is a directory
    compressiontype = CTYPE_NONE
    for filename in sys.argv[2:] :
        if filename == '-z' :
            compressiontype = CTYPE_ZLIB
            continue
        if filename == '-x' :
            _flat_unarchive(arcname)
            sys.exit()
        if filename == '-l' :
            file = open(arcname, "rb")
            arc = Reader(file)
            for entry in arc.index:
                print "%s\t\t%s\t%s" % (entry.get_pathname(), entry.size, entry.date)
            sys.exit()
        if filename == '-lt':
            import pprint
            file = open(arcname, "rb")
            arc = Reader(file)
            pprint.pprint(arc.get_parts_tree())
            sys.exit()
        if os.path.isdir(filename) :
            print 'DIR:', filename
            files = files + recursive_listdir(filename)
        else :
            files.append(filename)

    # don't overwrite an existing file
    # (this is to protect myself after a bad experience... ;)
    if os.path.isfile(arcname) :
        raise arcname+" already exists, aborting."

    # remove all occurances of the archive from the file list
    # XXX this could miss it on case insensitive filesystems...
    try:
        files.remove(arcname)
    except ValueError: pass

    print "Creating archive", arcname, "using these files:"
    print files

    arc = Creator(compression=compressiontype)
    arc.queuefiles(files)
    arc.writeto(arcname)

    print "done."

if __name__ == '__main__' :
    run()
