#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

# standard modules
import htmllib
import formatter
import os
import re

# our modules
import debug
from human_readable import hr
from HashQueue import HashQueue


__doc__ = \
"""
This module provides a web spider useful for generating a list of files all
related to a single web page (html file).  The primary interface is:

get_spider_files(htmlfile)
"""



html_file_re = re.compile(r".*\.html?$", re.I)


def get_spider_files(htmlfile, rootdir=None, maxdepth=20, file_hq=None):
    """return a list of absolute path names to files sourced or linked from htmlfile.

    @param htmlfile the file to start with.
    @param rootdir is the directory to never go above if relative links contain ..s
        by default it is os.path.dirname(htmlfile).
    @param maxdepth is the maximum recursion depth.

    @param file_hq is a HashQueue of files already found; for internal use only.
    """
    assert os.path.exists(htmlfile), "htmlfile %s doesn't exist" % htmlfile
    htmlfile = os.path.abspath(htmlfile)
    if rootdir is None:
        rootdir = os.path.dirname(htmlfile)
    assert os.path.isdir(rootdir), "rootdir %s is not a directory" % rootdir
    rootdir = os.path.abspath(rootdir)
    if rootdir[-1] != os.sep:
        rootdir = rootdir + os.sep

    # we use a HashQueue to store filenames for fast duplicate avoidance while
    # maintaining their order of occurrance.
    if file_hq is None:
        file_hq = HashQueue()
        # make the first file in the list be the root htmlfile
        file_hq[htmlfile] = 1

    # find files linked to from this one (if this is the lowest recursion level, don't
    # include hrefs)
    linkedfilelist = get_relative_linked_files(htmlfile, rootdir=rootdir, includehrefs=maxdepth)
    for linkedfile in linkedfilelist:
        if not file_hq.has_key(linkedfile):
            file_hq[linkedfile] = None
        else:
            file_hq[linkedfile] = 1

    if maxdepth > 1:
        # don't recurse on non-html files or those we've already visited
        nextfiles = filter(lambda f, file_hq=file_hq: html_file_re.match(f) and not file_hq.get(f), linkedfilelist)
        # spider down through the links
        for file in nextfiles:
            linkedfilelist = get_spider_files(file, rootdir=rootdir, maxdepth=maxdepth-1, file_hq=file_hq)
            for linkedfile in linkedfilelist:
                if not file_hq.has_key(linkedfile):
                    file_hq[linkedfile] = None
                else:
                    file_hq[linkedfile] = 1

    return file_hq.get_ordered_keys()
    


class SourceHTMLParser(htmllib.HTMLParser):
    """HTMLParser derivitive that also makes a self.sourcelist of img srcs"""
    def __init__(self, *args, **kwargs):
        apply(htmllib.HTMLParser.__init__, (self,) + args, kwargs)
        # used to list all img src tags
        # in the future could contain other autoloaded URIs
        self.sourcelist = []

    def handle_image(self, source, alt, ismap, align, width, height):
        self.sourcelist.append(source)
        htmllib.HTMLParser.handle_image(self, source, alt, ismap, align, width, height)


def get_relative_linked_files(htmlfile, rootdir=None, includehrefs=1):
    """return a list of relative links in htmlfile pointing to existing files.

    (The output of this function should be ideal for passing to archive.py)

    They will be listed in the order that a web browser is likely to
    load them (ie: hrefs come last).
    
    If relative links include ..'s, no links will be followed above
    rootdir.  By default rootdir is the directory of htmlfile.

    includehrefs is a boolean indicating if href references should be included in
    the result.
    """
    assert os.path.exists(htmlfile), "htmlfile %s doesn't exist" % htmlfile
    htmlfile = os.path.abspath(htmlfile)
    if rootdir is None:
        rootdir = os.path.dirname(htmlfile)
    assert os.path.isdir(rootdir), "rootdir %s is not a directory" % rootdir
    rootdir = os.path.abspath(rootdir)
    if rootdir[-1] != os.sep:
        rootdir = rootdir + os.sep

    # parse the html
    p = SourceHTMLParser(formatter.NullFormatter())
    p.feed(open(htmlfile).read())
    p.close()
    # TODO handle exceptions that a HTMLParser could raise

    # debug.mojolog.write("p.anchorlist: %s\n", args=(p.anchorlist,), v=1, vs="debug")

    sources = p.sourcelist
    if includehrefs:
        hrefs = p.anchorlist
    else:
        hrefs = []

    # only keep non absolute hrefs
    relative_hrefs = filter(lambda href: ':' not in href and href[0] != '/', hrefs)
    # remove duplicates hrefs (XXX isn't there an easier way to write this?)
    relative_hrefs.sort()
    hrefs = []
    for idx in xrange(len(relative_hrefs)):
        if idx > 0 and relative_hrefs[idx-1] == relative_hrefs[idx]:
            continue
        hrefs.append(relative_hrefs[idx])

    # only keep non absolute sources
    relative_sources = filter(lambda source: ':' not in source and source[0] != '/', sources)
    # remove duplicate sources (XXX easier way?)
    relative_sources.sort()
    sources = []
    for idx in xrange(len(relative_sources)):
        if idx > 0 and relative_sources[idx-1] == relative_sources[idx]:
            continue
        sources.append(relative_sources[idx])

    links = sources + hrefs   # sources come before hrefs
    # only keep links that are underneath rootdir
    basedir = os.path.dirname(htmlfile)
    # debug.mojolog.write("rootdir: %s\n", args=(rootdir,), v=1, vs="debug")
    # for f in links:
    #     debug.mojolog.write("absjoin(b, f): %s\n", args=(os.path.abspath(os.path.join(basedir, f)),), v=1, vs="debug")

    links = filter(lambda link, basedir=basedir, rootdir=rootdir: os.path.abspath(os.path.join(basedir, link))[:len(rootdir)] == rootdir, links)
    # convert all links to absolute filenames
    files = map(lambda link, basedir=basedir: os.path.abspath(os.path.join(basedir, link)), links)
    # only keep links that point to existing files
    # debug.mojolog.write("check for existing: %s\n", args=(files,), v=1, vs="debug")
    files = filter(lambda file, basedir=basedir: os.path.exists(file), files)
    return files
