#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# A module for creating the necessary splits and shares needed for
# publishing an archive.
#
__cvsid = '$Id: makesplit.py,v 1.26.2.4 2001/06/13 00:05:38 zooko Exp $'

# standard modules
import sha
import math
from cStringIO import StringIO
import sys
import os.path
from time import time


# custom modules
import const
import sharemap
import blocksharez
import human_readable
import idlib
import mojosixbit
from MojoConstants import true, false
import HashRandom
import debug
from confutils import confman

BSIZE_GROW = 0
BSIZE_EQUAL = 1


#########################################################################
# Get the first piece size that can fit the specified amount of data,
# otherwise return the largest piecesize if none will fit.
#########################################################################
def getpsize(len, num_shares_needed) :
    """return the power of two piece size that should hold data of length len; returns the max allowed piece size of len is too large"""
    MinPieceSize = int(confman['PUBLICATION']['MIN_SHARE_SIZE']) * num_shares_needed
    MaxPieceSize = int(confman['PUBLICATION']['MAX_SHARE_SIZE']) * num_shares_needed
    power = MinPieceSize   # this is the minimum size 
    len = max(power, len)
    while power < len:
        power = power * 2
        if power > MaxPieceSize :
            return MaxPieceSize
    return power



#########################################################################
# This class is used to create the necessary splits for publishing a
# file/archive as well as its sharemap.
#########################################################################
class Splitter :
    def __init__(self, archive, length, totalsharesneeded, totalsharesinpiece, maxpiecesize=0, flag=BSIZE_EQUAL):
        """
        @param archive a file object to read archive data from
        @param length the length of the archive
        @param totalsharesneeded the N in the N of M info dispersal we will perform
        @param totalsharesinpiece the M in the N of M info dispersal we will perform
        @param maxpiecesize the maximum size of an individual piece share
        @param flag flags indicating how the piece share sizes should be chosen:
            BSIZE_GROW   - use continually increasing piece share sizes
            BSIZE_EQUAL  - use a set of same sized pieces (maxpiecesize) and a best fit for the last piece
        """
        if totalsharesneeded not in (1,2,4,8,16,32,64):
            raise ValueError, "totalsharesneeded must be a reasonable power of two"
        if totalsharesinpiece not in (1,2,4,8,16,32,64,128):
            raise ValueError, "totalsharesinpiece must be a reasonable power of two"
        if totalsharesinpiece < totalsharesneeded:
            raise ValueError, "totalsharesneeded must be greater than totalsharesinpiece"
        if maxpiecesize > 2048*1024:
            raise ValueError, ("maxpiecesize is unreasonably large", maxpiecesize)
        if maxpiecesize != 0 and maxpiecesize not in const.PieceSizes:
            raise ValueError, ("maxpiecesize must be in const.PieceSizes", maxpiecesize)

        self.arcfile = archive
        self.arclen = length
        self.maxpiecesize = maxpiecesize
        self.flag = flag
        self.donesplitting = 0  # boolean: have all splits have been created?

        # used to cache the first maxpiecesize bytes of data read in from arcfile for
        # use in padding out the sharemap chunk with the first bytes of the publication.
        self.start_of_file = StringIO()
        self.start_of_file_limit = maxpiecesize

        self.num_shares_needed = totalsharesneeded
        self.num_shares_total = totalsharesinpiece

        MinPieceSize = int(confman['PUBLICATION']['MIN_SHARE_SIZE']) * self.num_shares_needed
        MaxPieceSize = int(confman['PUBLICATION']['MAX_SHARE_SIZE']) * self.num_shares_needed
        if self.maxpiecesize == 0:
            self.maxpiecesize = MaxPieceSize
        assert MinPieceSize >= 4096, "unreasonable minimum piece size"
        assert MaxPieceSize <= 2048*1024, ("unreasonable maximum piece size (MAX_SHARE_SIZE*totalsharesneeded)", MaxPieceSize)
        self.maxpiecesize = min(min(self.maxpiecesize, const.PieceSizes[-1]), MaxPieceSize)

        # The default maxpiecesize is no more than the size
        # of the entire data archive and no smaller than the smallest
        # allowed piece.
        if self.maxpiecesize <= 0 :
            self.maxpiecesize = max(length, MinPieceSize)

        self.maxpiecesize = min(self.maxpiecesize, MaxPieceSize)

        self.compute_split_locations()


    #####################################################################
    # Used internally.  This function computes how many splits are
    # needed and where they will occur.  It initializes the internal
    # self.pieces list which can later be used to create the sharemap
    # once all of the share splits have been created (therefore
    # computing their pieceids).
    #####################################################################
    def compute_split_locations(self) :
        self.pieces = []
        self.currentpiece = 0
        bytesleft = self.arclen
        curpos = 0

        # Pick a good starting piece size
        if self.flag == BSIZE_GROW :
            # make "small" content fit within one or two pieces/chunks rather
            # than using the start smallish and grow algorithm below
            if self.arclen <= 1024*1024:
                for sz in const.PieceSizes:
                    if self.arclen < sz:
                        bsize = sz
                        break
                else:
                    bsize = self.maxpiecesize
            else:
                # A "nifty" way to compute the starting piece size
                # XXX this is black magic used to determine what piece/chunk size to start
                # breaking the file up into -greg
                idx = int(math.log(self.arclen)) - 10
                if idx < 0 : idx = 0
                elif idx > len(const.PieceSizes) - 1 :
                    idx = len(const.PieceSizes) - 1
                bsize = const.PieceSizes[idx]

            if bsize > self.maxpiecesize :
                bsize = self.maxpiecesize
        elif self.flag == BSIZE_EQUAL :
            bsize = self.maxpiecesize
        else :
            raise ValueError, "Unknown piece size flag"

        # XXX much better piece size allocation can be done!
        #
        # For instance tapering the piece size off to fit the end of
        # the file more closely would be good.
        #
        # The maximum piece size should be chosen based on the
        # expected access patterns of the underlying archive.
        # Archives will also work better when created with care by
        # inserting files in the expected order of access.  A local
        # web spider algorithm could be used to help this for archives
        # arranged internally using HTML.

        while bytesleft > 0 :
            entry = {}
            if bytesleft > bsize :
                entry['length'] = bsize
                entry['offset'] = curpos
                self.pieces.append(entry)
                curpos = curpos + bsize
                bytesleft = bytesleft - bsize

                if self.flag == BSIZE_GROW :
                    # Step up to a larger piece size if there are more
                    # than 4 of the current-size pieces left
                    if bytesleft > 4*bsize :
                        if bsize*4 <= self.maxpiecesize :
                            bsize = getpsize(bsize*4, self.num_shares_needed)
                        elif bsize*2 <= self.maxpiecesize :
                            bsize = getpsize(bsize*2, self.num_shares_needed)
            else :
                entry['length'] = bytesleft
                entry['offset'] = curpos
                self.pieces.append(entry)
                curpos = curpos + bytesleft
                bytesleft = 0
                break


    def get_piece_list(self):
        """
        return the list of pieces used to make the sharemap
        """
        return self.pieces

    #####################################################################
    # Returns a Sharemap or Supersharemap, depending on the length of
    # the data. the return values are as follows:
    # ((<sharemap tuple>), <shares>, <length>)
    #
    # <sharemap tuple>: contains the sharemap/supersharemap piece ID, a list of
    # (shareid, share) tuples comprising the single
    # sharemap/supersharemap piece, and the sharemap/supersharemap
    # piece length.  This tuple contains data for generating the Dinode
    #
    # <shares> is a list of (piecid, [(shareid, share), (shareid, share), (shareid, share), ...])
    # tuples containing all shares of the entire sharemap and supersharemap, if applicable.
    # This list contains shares that are part of the main sharemap
    # when a supersharemap is in use.
    #
    # <length> is the length of all pieces including the sharemap and
    # superesharemap
    #
    # Raises a RuntimeError exception if called before all piece
    # shares have been created with getnextshare().
    #####################################################################
    def getsharemap(self) :
        if not self.donesplitting :
            raise RuntimeError, "create all splits before making the sharemap"

        smap = sharemap.make_sharemap(self.pieces)
        # omle = human_readable.mrepr.maxlevel
        # human_readable.mrepr.maxlevel=240
        # omd = human_readable.mrepr.maxdict
        # human_readable.mrepr.maxdict=240
        # omli = human_readable.mrepr.maxlist
        # human_readable.mrepr.maxlist=240
        # omt = human_readable.mrepr.maxtuple
        # human_readable.mrepr.maxtuple=240
        # debug.mojolog.write("XXX self.pieces: %s\n", args=(self.pieces,))
        # human_readable.mrepr.maxlevel = omle
        # human_readable.mrepr.maxdict = omd
        # human_readable.mrepr.maxlist = omli
        # human_readable.mrepr.maxtuple = omt

        if len(smap) > self.maxpiecesize:
            raise RuntimeError, ("sharemap length %s exceeds maxpiecesize %s" % (len(smap), self.maxpiecesize))

        # if there is extra padding room in the sharemap chunk/piece, fill it
        # with a copy of the first data bytes for faster download startup.
        # (based on zooko's idea in an irc discussion - 2001-05-23)
        bytes_of_padding = blocksharez.get_poweroftwo_chunk_len(len(smap), self.num_shares_needed) - len(smap)
        if bytes_of_padding > 0:
            self.start_of_file.seek(0)
            smap = smap + self.start_of_file.read(bytes_of_padding)
            self.start_of_file = None   # free this, we're done with it.

        # Generate the piece shares
        try: 
            sharemapShares = blocksharez.split(smap, totalsharesneeded=self.num_shares_needed, totalsharesinpiece=self.num_shares_total)
        except ValueError:
            # super sharemaps are currently disabled!
            raise RuntimeError, "The file is too large.  Please upgrade your Broker."

            debug.stderr.write("Sharemap larger than single piece!  Creating SuperShareMap!", v=2)
            ## The sharemap was larger than a single piece, make a supersharemap
            allShares = []
            # first, split the sharemap
            smapSplits = Splitter(StringIO(smap), len(smap), totalsharesneeded=4, totalsharesinpiece=8)
            pieceid, shares = smapSplits.getnextsplit()
            while pieceid:
                allShares.append((pieceid, shares))
                pieceid, shares = smapSplits.getnextsplit()
            # create supersharemap from sharemap pieces
            superSmap = sharemap.make_super_sharemap(smapSplits.pieces)
            superSmapLen = len(superSmap)
            # split supersharemap
            superSmapShares = blocksharez.split(superSmap, totalsharesneeded=self.num_shares_needed, totalsharesinpiece=self.num_shares_total)
            superSmapPieceID = mojosixbit.b2a(idlib.string_to_id(superSmap))
            allshares.append((superSmapPieceID, superSmapShares))
            return ((superSmapPieceID, superSmapShares, superSmapLen), allShares, len(smap) + superSmapLen)
        else:
            smappieceid = mojosixbit.b2a(idlib.string_to_id(smap))
            return ((smappieceid, sharemapShares, len(smap)), [(smappieceid, sharemapShares)], len(smap))


    def getnextsplit(self) :
        """
        Get the next piece (as split into M shares).

        @return in a tuple of (pieceid, shares), where `pieceid' is the id of this piece and
            `shares' is list of length M of tuples of (shareid, sharedata), where `sharedata' is
            the encrypted share and `shareid' is the id of the encrypted share; returns (None,
            None) when there are no more share splits.
        """
        if self.donesplitting :
            return (None, None)

        # Compute the pieceid and save it in the self.pieces
        piecedata = self.arcfile.read(self.pieces[self.currentpiece]['length'])
        pieceid = mojosixbit.b2a(idlib.string_to_id(piecedata))
        self.pieces[self.currentpiece]['pieceid'] = pieceid

        # Cache the first self.start_of_file_limit bytes in self.start_of_file for use in padding
        # out the sharemap's chunk with the first bytes of file data rather the zeroes.
        if self.start_of_file and self.start_of_file.tell() < self.start_of_file_limit:
            sof_bytes = self.start_of_file_limit - self.start_of_file.tell()
            self.start_of_file.write(piecedata[:sof_bytes])

        # Generate the piece shares.
        shares = blocksharez.split(piecedata, totalsharesneeded=self.num_shares_needed, totalsharesinpiece=self.num_shares_total)

        # Save a map of sharenum -> shareid in self.pieces (for use in the sharemap).
        self.pieces[self.currentpiece]['shares'] = {}
        sharenum = 0
        for (shId, sh) in shares:
            self.pieces[self.currentpiece]['shares'][sharenum] = shId
            sharenum = sharenum + 1

        self.currentpiece = self.currentpiece + 1
        if self.currentpiece >= len(self.pieces) :
            self.arcfile.close()
            self.donesplitting = 1

        return (pieceid, shares)



########################################################################
def run():
    print "Test pieceshare split creator."
    print "All blobs will be created in ./teststore"

    if len(sys.argv) < 2 :
        print "\nUsage:\n  "+sys.argv[0]+" archivename"
        sys.exit()

    arcname = sys.argv[1]
    arcsize = os.path.getsize(arcname)
    print arcname, arcsize

    storedir='./teststore'
    store = localpiecestore.BlockStore(path='./teststore', depth=1)
    store.init_space(storedir)

    infile = open(arcname, "rb")
    s = Splitter(infile, arcsize, totalsharesneeded=4, totalsharesinpiece=8)

    pieceid, shares = s.getnextsplit()
    while pieceid :
        # Write each of the shares to an appropriately named file
        print "Writing Piece: ", pieceid, "(%d bytes)" % (4*len(shares[0][1]))
        for shareid, sharedata in shares :
            sharefile = store.open_tempfile(shareid, len(sharedata))
            sharefile.write(sharedata)
            sharefile.close()
            store.store_tempfile(shareid, sharefile)
            print "  stored share blob", shareid

        pieceid, shares = s.getnextsplit()
        print `pieceid`

    infile.close()

    sharemap, piecedict, smaplen = s.getsharemap()
    
    # Write each of the sharemap shares to an appropriately named file
    print "Writing Sharemap:", pieceid, "(%d bytes)" % (4*len(shares[0][1]))
    for shareid, sharedata in shares :
        tf = store.open_tempfile(shareid, len(sharedata))
        tf.write(sharedata)
        tf.close()
        store.store_tempfile(shareid, tf)
        print "  stored Sharemap share blob", shareid


if __name__ == '__main__':
    run()
