#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# A simple Mojo Nation (TM) client interface web server.
#
# A newer, newer, better, better mojohttpd.  This one can serve arbitrary HTML files and call
# CGI scripts, as well as use built-in python handler funcs.  Unfortunately, CGI scripts cannot
# have access to the MojoTransactionManager, so all Mojonation (tm) service-level interfacing
# must be within a built-in handler.
#
#
__cvsid = '$Id: mojohttpd.py,v 1.242.2.1 2001/06/12 07:00:10 zooko Exp $'

# Notes on terminology used in comments within this file:

# backend - This refers to any code that requires a MojoTransactionManager object to function.
#    Often the term "servicemtm" is used as variable names or an abbreviation for this
#    MojoTransactionManager instance.  This is because there should be two MojoTransactionManagers
#    for every running broker.  The user interface is only concerned with the "service-level"
#    MojoTransactionManager because it provides an adequate API to accomplish any necessary
#    task on the Mojo Nation network that the UI is responsible for.

# Standard Modules:
import types
import os
import string
import re
import threading
import socket
import SimpleHTTPServer
import cgi
import asyncore
import Queue
import shutil
import urllib
import mimetypes
from cStringIO import StringIO
import traceback
import time
import sys
import copy

# Mojo Modules:
from loggedthreading import LoggedThread
import UserCommand
import Cache
from MojoConstants import true, false, DEBUG_MODE
import MetaTrackerLib
import mojobootpage
import MojoNationHTTPServer
import Presentation
import confutils
from confutils import confman, BROKER_VERSION_STR 
import contentlib
import debug
from human_readable import hr
import idlib
import id_lock
import mojobootpage
import mojosixbit
import mojoutil
from FindBeforehandBlockFetcher import NUM_FINISHED_REASSEMBLIES
import archive
import randsource
import unsplit
import Passphrase
import Describe
import MojoErrors
import mencode
import modval
import MojoKey
import sharemap
import js_list

import DoQ # this contains the global reference to the DoQ  --Zooko 2001-04-14

try:
    from BitmapDownloadMonitor import BitmapDownloadMonitor
except (ImportError, SyntaxError), e:
    BitmapDownloadMonitor = None
    debug.mojolog.write('BitmapDownloadMonitor could not be imported, not using one.\n%s\n', args=(e,), v=3, vs='MojoHTTPd')
try:
    import xml.dom.minidom
except ImportError:
    pass


# Restrict CGI files to the directory specified in the config file:
#CGIHTTPServer.cgi_directories=[os.path.expandvars(confman.dict['PATH']['CGI-BIN_DIR'])]

class MojoHTTPDError(StandardError):
    pass

# These are global variables that should be initiated before MojoRequestHandler is initiated:
globaldict = {
        'regexes' : {
                # This is used to reduce multiple slashes in a path to a single one:
                'multislash' : re.compile(r"//+"),
                # This is used for extracting dinodes from the path:
                'extractdinode' : re.compile(r"^/Dinode/(?P<dinode>[-_A-Za-z0-9]+)(|/(?P<part>.*))$", re.I),
                # This is used for extracting mojoids from the path:
                'extractmojoid' : re.compile(r"^/(Id|mojoid)/(?P<mojoid>"+mojosixbit._asciihash_re_str+")(|/(?P<part>.*))$", re.I),
                # Used to extract the privacy code from the http GET request:
                'extractpcode' : re.compile(r"^/([^/]*)(/.*)$"), # Group 1 is the privacy code, Group 2 is the actual path.
                # Used in URL's to indentify the broker:
                'brokername' : re.compile(r"^/(nexus|broker|station)(.*)$", re.I),  # group 2 is the real path
                # To identify the root URL:
                'rootpath' : re.compile(r"^/$"),
            },
        # note: this typedefrep is also accessed from Broker (is it still? why? -2001-05-17)
        'typedefrep' : contentlib.ContentTypeDefRepository(confman.getpath(["PATH", "CONTENT_TYPEDEF_DIR"])),
        'httpd' : None, # This holds the actual server instance, once created.
        'port': 0, # This holds the actual localhost server port once bind has been called
    }

# load our mime.types file
mimetypes.init([os.path.join(os.path.expandvars(confman.dict['PATH']['WEBROOT_DIR']), 'mime.types')])

def wait_for_terms_of_service():
    """
    This sets the web server into terms-of-service mode, if necessary.
    It checks the config to see if this is necessary (if not it returns true
    immediately).

    Note: This function blocks in the case that TOS has not been accepted.

    @precondition The mojohttpd web server is already running.
    @return true value if user accepted, false value if they declined.
    """
    if confman.dict.get("AGREED_TO_TERMS_OF_SERVICE") == 'yes':
        return true
    switch_builtin_request_handler_dict(tos_requestbuiltins) # Set the web gateway to "terms of service" mode.
    # Let the user know what must be done in the debug log:  (UI should improve to where this is not necessary.)
    debug.stderr.write("""

))))))                                                      ((((((
))))))  Waiting for user to acknowledge terms of service.   ((((((
))))))  Open up the "intropage" located in:                ((((((
        %s
))))))  And agree to the Terms of Service.                ((((((

))))))  You must do this before you can use Mojo Nation     ((((((
))))))                                                      ((((((
    """ % (os.path.expandvars(confman.dict['PATH']['INTRO_PAGE_v2']),) , vs='Broker')
    user_acknowledged_tos_event.wait()
    debug.stderr.write("Done waiting on Terms of Service.\n", v=3, vs='Broker')

    if confman.dict.get("AGREED_TO_TERMS_OF_SERVICE") != 'yes':
        # They still haven't accepted.
        debug.stderr.write("User declined the terms of service.\n", v=-100000, vs='Broker')
        return false
    else:
        debug.stderr.write("terms of service agreement accepted.\n", v=3, vs='Broker')
        return true

def fetch_key(dbparentdir, keyid):
    serialized = None
    try:
        f = open(os.path.join(dbparentdir, keyid + '/key'), 'r')
        serialized = f.read()
        f.close()
    except:
        debug.mojolog.write("WARNING: exception trying to read key file %s\n" % os.path.join(dbparentdir, keyid+ '/key'))
    return serialized

def get_key_id(serialized):
    skdict = mencode.mdecode(serialized)
    skdict = mencode.mdecode(skdict['session keeper'])
    keyMV = modval.new_serialized(mojosixbit.a2b(skdict['private key serialized']))
    pkey = MojoKey.makePublicRSAKeyForCommunicating(keyMV)
    return mojosixbit.b2a(idlib.make_id(pkey, 'broker'))

def get_mtm_ids():
    """
    This sets the web server into pass-phrase mode, if necessary.
    It handles all four cases involving pass-phrases and encrypted keys:
    A) The user wants encrypted keys and passphrases and has both.
    B) The user wants encrypted keys but has not yet supplied a passphrase (the keys are plaintext).
    C) The user does not want encrypted keys, but his keys are currently encrypted (and must be converted to plaintext).
    D) The user does not want encrypted keys, and they are not encrypted.
    
    Note: This function blocks for cases A, B, and C.

    @precondition The mojohttpd web server is already running.
    @precondition The config contains either a 'mtm for communication with token server' (exclusive) or 'ENCRYPTED_TOKEN_SERVER_KEY' key.
    @precondition The config contains either a 'TRANSACTION_MANAGER_STRING' (exclusive) or 'ENCRYPTED_BROKER_KEY' key.
    @postcondition The mojohttpd mode is set to "start-up"..
    @return a tuple (tokenmtm, peermtm) which are each (clear-text) serialized mtm strings.

# XXX key encryption is currently broken/disabled

This function can take a while if it is generating new RSA keys, but it doesn't really "block" in the sense of waiting indefinitely pending another event.
    """
    PLAIN_TMTM_CKEY = 'mtm for communication with token server' # Config key for plain-text token-server MTM.
    PLAIN_PMTM_CKEY = 'TRANSACTION_MANAGER_STRING' # Config key for plain-text broker-peer MTM.
    TMTM_ID = 'TMTM_ID' # Config key for encrypted token-server MTM.
    PMTM_ID = 'PMTM_ID' # Config key for encrypted broker-peer MTM.


    tokenmtm = confman.dict.get(PLAIN_TMTM_CKEY)
    peermtm = confman.dict.get(PLAIN_PMTM_CKEY)

    tokenmtm_id = confman.dict.get(TMTM_ID)
    peermtm_id = confman.dict.get(PMTM_ID)
    
    dbparentdir=os.path.expandvars(confman.dict["PATH"]["MOJO_TRANSACTION_MANAGER_DB_DIR"])

    changed = None
    if not tokenmtm_id and tokenmtm:
        tokenmtm_id = get_key_id(tokenmtm)
        confman.dict['TMTM_ID'] = tokenmtm_id
        del confman.dict['mtm for communication with token server']
        changed = 1
    if not peermtm_id and peermtm:
        peermtm_id = get_key_id(peermtm)
        confman.dict['PMTM_ID'] = peermtm_id
        del confman.dict['TRANSACTION_MANAGER_STRING']
    if changed:
        confman.save()

    switch_builtin_request_handler_dict(startup_requestbuiltins)
    return (tokenmtm_id, peermtm_id)
    """
    # XXX encrypted keys currently disabled
    # Switch to pass-phrase mode:
    switch_builtin_request_handler_dict(phrase_requestbuiltins)
    debug.mojolog.write('\n)))))))  Waiting for user to enter passphrase...  (((((((\n\n', vs='MojoHTTPD', v=-1)
    user_entered_passphrase.wait()
    # confutils.passphrase should now be set appropriately.

    # Switch to startup mode:
    switch_builtin_request_handler_dict(startup_requestbuiltins)

     #check if key/stash encryption has been disabled
    ENCRYPT_LOCAL_KEYS = (confman.dict['ENCRYPT_LOCAL_KEYS'] == 'yes')
    # check for encrypted keysENCRYPTED_BROKER_KEY
    ENCRYPTED_KEYS = (confman.dict.has_key(ENC_TMTM_CKEY) or confman.dict.has_key(ENC_PMTM_CKEY))

    if ENCRYPTED_KEYS:
        # Decrypt the keys:
       tokenmtm = Passphrase.decrypt(confutils.passphrase, confman.dict[ENC_TMTM_CKEY])
       peermtm = Passphrase.decrypt(confutils.passphrase, confman.dict[ENC_PMTM_CKEY])

    if ENCRYPT_LOCAL_KEYS:
        if not ENCRYPTED_KEYS:
            # Write the encrypted keys to the config file, delete the plain-text keys:
            confman.dict[ENC_TMTM_CKEY] = Passphrase.encrypt(confutils.passphrase, tokenmtm)
            confman.dict[ENC_PMTM_CKEY] = Passphrase.encrypt(confutils.passphrase, peermtm)
            if confman.dict.has_key(PLAIN_TMTM_CKEY):
                del confman.dict[PLAIN_TMTM_CKEY]
            if confman.dict.has_key(PLAIN_PMTM_CKEY):
                del confman.dict[PLAIN_PMTM_CKEY]
    else:
        # Write the plain text keys to the config file, delete the encrypted keys:
        confman.dict[PLAIN_TMTM_CKEY] = tokenmtm
        confman.dict[PLAIN_PMTM_CKEY] = peermtm
        if confman.dict.has_key(ENC_TMTM_CKEY):
            del confman.dict[ENC_TMTM_CKEY]
        if confman.dict.has_key(ENC_PMTM_CKEY):
            del confman.dict[ENC_PMTM_CKEY]

    return (tokenmtm, peermtm)
    """

def launch_browser(url="file:%s" % urllib.pathname2url(os.path.expandvars(confman.dict['PATH']['INTRO_PAGE_v2'])), newthread=true, newwindow=true):
    """
    If the webbrowser module is available, this launches the browser to the given URL.

    Note: This function launches a new thread (unless webbrowser could not be imported),
    and is non-blocking.
    
    @param url is a url string which the browser will open.
    @return true value if the webbrowser was launched; false value if it could not.
    """
    try:
        import webbrowser
    except ImportError:
        return false
    debug.mojolog.write("Launching web browser.\n", v=1, vs='Broker')
    def open_web_browser(webbrowser=webbrowser, url=url, newwindow=newwindow):
        localgateway_ready_event.wait()
        debug.mojolog.write("Opening web browser\n", v=1, vs='Broker')
        webbrowser.open(url, new=newwindow)
    if newthread:
        LoggedThread(target=open_web_browser).start()
    else:
        open_web_browser(newwindow=newwindow)
    return true

def debuglog(output, args=(), v=2):
    """
    A thin wrapper around debug.mojolog.write which uses a consistent vs parameter.
    """
    return debug.mojolog.write(output, args=args, vs='MojoHTTPD', v=v)
    

def initialize_intro_page(port=None):
    """
    Initializes the intro page.

    @precondition `globaldict' must have key 'privacycode'.: globaldict.has_key('privacycode'): "globaldict: %s" % hr(globaldict)
    """
    assert globaldict.has_key('privacycode'), "`globaldict' must have key 'privacycode'." + " -- " + "globaldict: %s" % hr(globaldict)

    if port is None:
        port = globaldict['port']

    introfile = open(os.path.expandvars(confman.dict['PATH']['INTRO_PAGE_v2']), 'w')
    # NOTE: this needs to be the first line of the file; the MojoMunchy proxy
    # reads this and snarfs the current port and privacy code out of it.
    # NOTE: my utilscript/snarfprivacycode does, too.  --Zooko 2000-08-11
    introfile.write('<!-- Port: %d PrivacyCode: %s -->\r\n' % (port, globaldict['privacycode']))

    oldbaseurl = Presentation.variables.get('baseurl')
    if confman.is_true_bool(('PUT_PRIVACY_CODE_IN_URLS_v2',)):
        Presentation.variables.set('baseurl', 'http://localhost:%s/%s' % (port, globaldict['privacycode']))
    else:
        Presentation.variables.set('baseurl', 'http://localhost:%s/' % (port,))

    introfile.write(Presentation.show('introtemplate', variables=Presentation.variables))

    Presentation.variables.set('baseurl', oldbaseurl)
    introfile.close()

    # This is so people don't have to use snarfprivacycode:
    bufile = open(os.path.expandvars(confman.dict['PATH']['BASE_URL_FILE']), 'w')
    if confman.is_true_bool(('PUT_PRIVACY_CODE_IN_URLS_v2',)):
        bufile.write('http://localhost:%s/%s' % (port, globaldict['privacycode']))
    else:
        bufile.write('http://localhost:%s/' % (port,))
    bufile.close()


def write_shutdown_intropage():
    """
    Overwrite the intropage with one that indicates the broker is not running.
    """
    introfile = open(os.path.expandvars(confman.dict['PATH']['INTRO_PAGE_v2']), 'w')
    introfile.write(Presentation.show('introtemplate_down', variables=Presentation.variables))
    introfile.close()


# XXX TODO how can we easily determine which ", . _ : or whatever" number seperator the user wants to use?
# XXX TODO this function should be obsoleted:
#  DocumentTemplate's DTML has the ability to do this to numbers when displaying them.
def integer_to_pretty_string(num, spacer=','):
    """
    Convers an int or long to a pretty printable string.
    Inserts spacer between every three digits from the right hand side of numstr.
      1000000 -> 1_000_000
      276029760, spacer=',' -> 276,029,760
    """
    if not (type(num) == type(0) or type(num) == type(0L)) :
        raise ValueError, ("int or long value required", num)
    numstr = str(num)
    if numstr[-1] == 'L' :
        numstr = numstr[:-1]
    if numstr[0] == '-' :
        prefix = '-'
        numstr = numstr[1:]
    else :
        prefix = ''
    numstr_triples = []
    while len(numstr) > 0 :
        numstr_triples.append(numstr[-3:])
        numstr = numstr[:-3]
    numstr_triples.reverse()
    return prefix + string.join(numstr_triples, spacer)

class MojoRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
    # The following methods over-ride super-class methods:
    def __init__(self, *_args, **_kwargs):
        """
        @precondition The global MojoHTTPD requestbuiltins mapping must be a dict of regex (string or object) to request-handler function.: type(requestbuiltins) == types.DictType: "requestbuiltins: %s" % (hr(requestbuiltins), `type(requestbuiltins)`)
        """
        assert type(requestbuiltins) == types.DictType, "The global MojoHTTPD requestbuiltins mapping must be a dict of regex (string or object) to request-handler function." + " -- " + "requestbuiltins: %s" % (hr(requestbuiltins), `type(requestbuiltins)`)

        
        self.builtins = requestbuiltins
        self.no_message_body = false  # set to true for HEAD requests (or could be on response codes that don't allow a message body)

        apply(SimpleHTTPServer.SimpleHTTPRequestHandler.__init__, (self,) + _args, _kwargs)

    def do_POST(self):
        debuglog('received POST request.\n', v=5) # verbose
        self.variables = Presentation.variables.clone()  # each request needs its own Presentation.variables namespace
        self.variables.set('userinput', {})

        try:
            entitytype = self.headers['Content-Type']
        except KeyError:
            debug.mojolog.write('No entity type header provided.\n', vs='MojoHTTPD', v=4)
            entitytype = None

        entityhandler = post_type_handlers.get(entitytype)

        if entityhandler is None:
            if entitytype is not None:
                debug.mojolog.write('No handler for entity type specified in POST headers: %s\n' % entitytype, vs='MojoHTTPD', v=2)
        else:
            entityhandler(self)
        self._handle_request()

    def do_GET(self):
        debuglog('received GET request.\n', v=5) # verbose
        self.variables = Presentation.variables.clone()  # each request needs its own Presentation.variables namespace
        self.variables.set('userinput', {})
        self.no_message_body = false
        try:
            self._handle_request()
        except:
            debug.mojolog.write("ERROR: do_GET caught this exception:\n", vs='MojoHTTPD')
            traceback.print_exc(file=debug.mojolog)
            if hasattr(self.wfile, 'closed') and not self.wfile.closed:
                stackstrio = StringIO()
                traceback.print_exc(file=stackstrio)
                self.variables.set('longerror', stackstrio.getvalue())
                self.show_error('Exception while processing request.', httpcode=500)
    
    def do_HEAD(self):
        debuglog('received HEAD request.\n', v=5) # verbose
        self.variables = Presentation.variables.clone()  # each request needs its own Presentation.variables namespace
        self.variables.set('userinput', {})
        self.no_message_body = true
        self._handle_request()

    def _handle_request(self):
        """
        This method is called for GET, POST, and HEAD requests.
        """
        debuglog('Request path: "%s".\n' % self.path, v=9) # verbose
        # include the privacycode in baseurl only if we're not being accessed via the mojo proxy
        if self.headers.getheader('mojoproxy') or not confman.is_true_bool(('PUT_PRIVACY_CODE_IN_URLS_v2',)):
            self.variables.set('baseurl', '')
        else:
            self.variables.set('baseurl', '/'+globaldict['privacycode']+'/')
        if self.headers.getheader('mojoproxy'):
            self.variables.set('url_host', 'mojonation.net')
        else:
            self.variables.set('url_host', 'localhost:%d' % globaldict['port'])
        
        # try to determine what web browser is being used
        useragent = self.headers.getheader('User-Agent', '')
        if string.find(useragent, 'MSIE') != -1:
            self.variables.set('browser', 'msie')
        else:
            useragent = string.lower(useragent)
            if string.find(useragent, 'mozilla/4') != -1:
                self.variables.set('browser', 'netscape')
        # add more possible 'browser' values here as we find things that we should -only- do on certain browsers

        # Reduce all duplicated slashes to a single slash in all of the path before the '?' designator:
        # Note: Either GET or POST methods may contain a query string.
        spliturl = string.split(self.path, '?', 1)
        spliturl[0] = globaldict['regexes']['multislash'].sub("/", spliturl[0])
        self.path = string.join(spliturl, '?')

        if confman.is_true_bool(('PUT_PRIVACY_CODE_IN_URLS_v2',)):
            match = globaldict['regexes']['extractpcode'].match(self.path)
            if globaldict['regexes']['extractdinode'].match(self.path) or \
               globaldict['regexes']['extractmojoid'].match(self.path) or \
               self.path[:len('/images/')] == '/images/':
                # fake the privacy code for /Dinode/ and /Id/ so that we can retrieve things from mojo nation without it.
                # XXX this is needed for network-effect generating URLs but is a potential security problem.
                # XXX Specifically the attacker may be able to trick the victim's mojohttpd into grabbing a Dinode,
                # XXX thereby depleting Mojo and denying service.
                privacycode = globaldict['privacycode']
                # -and self.path remains untouched.
            elif match is None :
                self.cachable = false
                self.variables.set('privmsg', 'Privacy Code malformed.')
                self.send_entity(self.show('privacycode'), rcode=400, rmsg="Privacy Code Error")
                return
            else :
                debuglog('validating privacy code.\n', v=4) # verbose
                privacycode, self.path = match.groups()

            # verify that they had a valid privacy code
            if privacycode != globaldict['privacycode'] :
                self.cachable = false
                self.variables.set('privmsg', 'Invalid Privacy Code.')
                self.send_entity(self.show('privacycode'), rcode=400, rmsg="Privacy Code Error")
                return

        debuglog('current path: %s\n' % self.path, v=6) # verbose
        # strip the word nexus/broker/stash for this application
        m = globaldict['regexes']['brokername'].match(self.path)
        if m :
            self.path = m.group(2)
            if self.path and self.path[0] != '/' :
                self.path = '/' + self.path
            debuglog('new path: %s\n' % self.path, v=6) # verbose

        debuglog('handling request for path "%s".\n' % self.path, v=4)
        for regex in self.builtins.keys():
            if regex.match(self.path):
                debuglog('parsing URL query.\n', v=5)
                try:
                    userinput = cgi.parse_qs(string.split(self.path, '?', 1)[1])
                except IndexError:
                    userinput = {}
        
                self.variables.get('userinput').update(userinput)

                self.cachable = false
                debuglog('invoking builtin with regex "%s".\n', args=(regex.pattern, ), v=7) # verbose
                try:
                    # (we use get so that the startup and termsofservice handlers can get by with None)
                    # assert globaldict.get('broker').ServicesMTM is not None # oops -- not when agreeing to ToC.  argh...  --Zooko 2001-04-27
                    self.builtins[regex](self, globaldict.get('broker').ServicesMTM) # XXX: The handlers should become methods of this class eventually.
                except IOError:
                    debug.mojolog.write("client closed connection - GET %s\n", args=(self.path,), v=2, vs='MojoHTTPD')
                break
        else:
            # if none of the builtin regex's matched, serve this up as a regular file
            self.cachable = true
            if globaldict['regexes']['rootpath'].match(self.path):
                self.path = '/index.html'
            debuglog('serving file "%s".\n', args=(self.path, ), v=6)
            # give all static files a 7 minute expiration time so that the browser hopefully won't reload them so often
            self.expiresin = 7*60
            SimpleHTTPServer.SimpleHTTPRequestHandler.do_GET(self)

    def show(self, filename, **kwargs):
        """
        Calls Presentation.show() with variables=self.variables

        other keyword arguments are passed directly to the dtml parser for use as dtml variables
        """
        return apply(Presentation.show, (filename, self.variables), kwargs)

    def send_entity(self, entity, rcode=200, rmsg=None, headers=None):
        """
        Use this always if the response entity is complete
        by the time this call is sent.

        entity - a string, the http response entity.
        rcode - an int, the http response code.
        rmsg - a string, the http response message.
        headers - None or a dict of string mapped to string, headers.

        This method does the following:
        A) This sends a response line, according to rcode and rmsg.
        B) If one is not present, adds a "content-type: text/html" header.
        C) If one is not present, adds a "content-length" header,
            according to len(entity).
        D) Sends the headers.
        E) Sends the entity.

        The only cases when this method should not be used are
        when the complete entity can't be sent at once (ie:
        incremental sends), or when the response line and/or
        headers must be sent before this can be called, for some
        reason.
        """
        if headers is None:
            headers = {}

        mojoutil.setdefault(headers, 'content-type', 'text/html')
        mojoutil.setdefault(headers, 'content-length', str(len(entity)))

        self.send_response(rcode, rmsg)
        for key, value in headers.items():
            self.send_header(key, value)

        self.end_headers()
        if not self.no_message_body:   # allows HEAD requests to work properly
            self.wfile.write(entity)
        
    def send_response(self, code, message=None):
        SimpleHTTPServer.SimpleHTTPRequestHandler.send_response(self, code, message=message)
        if code < 400 and hasattr(self, 'expiresin'):
            # send an expires header for this page expiresin seconds in the future
            expiretime = time.time() + self.expiresin
            year, month, day, hh, mm, ss, wd, y, z = time.gmtime(expiretime)
            s = "%s, %02d %3s %4d %02d:%02d:%02d GMT" % (
                    self.weekdayname[wd],
                    day, self.monthname[month], year,
                    hh, mm, ss)
            self.send_header('Expires', s)
        elif code >= 400 or (hasattr(self, 'cachable') and not self.cachable):
            # send an Expires header of the the same
            # thing that BaseHTTPServer put in the Date header
            # to indicate that this data is not cachable.
            self.send_header('Expires', self.date_time_string())

    def send_html_response(self, errcode, message=None):
        self.send_response(errcode, message)
        self.send_header('content-type', 'text/html')
        self.end_headers()

    def log_message(self, output, *_formargs):
        if len(_formargs) > 0 :
            output = output % _formargs
        debuglog(output + "\n", v=1)

    def show_error(self, errorstr, httpcode=200, logfile=debug.mojolog):
        """
        Display an error message to the client.  This sends a the HTTP
        response code specified in httpcode (200 by default) unless httpcode
        is None.
        """
        try:
            logfile.write('Error: %s\n' % errorstr, vs='MojoHTTPD')
            self.variables.set('errorstring', errorstr)
            self.send_entity(self.show('error'), rcode=httpcode, headers={'content-type': 'text/html'})
        except socket.error:
            pass  # client closed connection, ignore it

    def send_rewrite_response(self, url, prependhost=true):
        """
        Send a 301 "Moved Permanently" HTTP response redirecting the browser
        to the given URL.

        prependhost determines if we should prepend our own http://foo:bar
        string to url before including it in location.  If so, we'll do it
        intelligently depending on if mojoproxy was used to access us or not.
        """
        # NOTE don't argsify the % in this debug.stderr.write, it calls hr() on the URL which makes it useless in many cases
        debuglog('Sending "location moved" response with url: %s\n' % url, v=7)
        host = ''
        if prependhost:
            assert url[:1] == '/', 'rewrite_response requires absolute urls'
            if self.headers.getheader('mojoproxy'):
                host = 'http://mojonation.net'
            else:
                if self.headers.has_key('host'):
                    # XXX don't need port number here, too?  --Zooko 2001-04-22
                    host = 'http://%s' % self.headers['host']
                else:
                    host = 'http://localhost:%d' % globaldict['port']
        else:
            assert url[:4] == 'http', 'rewrite_response requires fully specified urls'
        self.send_entity(
            entity='The file you want <code>%s</code> is <a href="%s">located here</a>' % (url, url),
            rcode=301,
            rmsg='URL Redirect.',
            headers={'location': host + url})

def _once_around_the_loop(continue_func, httpd):
    if continue_func(httpd):
        debuglog('ready to handle new request.\n', v=4)
        httpd.handle_request()

localgateway_ready_event = threading.Event()  # gets set once the server is up
def request_handler_loop(httpd, continue_func = lambda httpd : true):
    localgateway_ready_event.set()
    while continue_func(httpd):
        _once_around_the_loop(continue_func, httpd)

# Built-in request handlers:
user_acknowledged_tos_event = threading.Event()
def handle_terms_of_service(reqhandler, servicemtm):
    try:
        queryvalues = cgi.parse_qs(string.split(reqhandler.path, '?', 1)[1])
    except IndexError:
        queryvalues = {}
    
    if queryvalues.has_key('agreement'):
        if queryvalues['agreement'][0] == "Accept":
            confman.dict["AGREED_TO_TERMS_OF_SERVICE"] = "yes"
            # save the fact that the user has agreed!
            confman.save()
            # signal
            # XXX this is a hack used to let the outside world (Broker.py) know that we're done asking them to sign the TOS
            switch_builtin_request_handler_dict(startup_requestbuiltins) # switch to startup mode.
            try:
                # Redirect new users directly to the preferences page in hopes
                # that they set their connection speed properly.
                reqhandler.send_rewrite_response(reqhandler.variables.get('baseurl')+'/preferences')
            finally:
                user_acknowledged_tos_event.set()
        else:
            try:
                reqhandler.send_entity(entity=reqhandler.show('termsofservice_decline'))
            finally:
                user_acknowledged_tos_event.set()
        return
    else:
        reqhandler.send_entity(reqhandler.show('termsofservice'))

user_entered_passphrase = threading.Event()
def handle_passphrase(reqhandler, servicemtm):
    # XXX quick hack to make in progress in HEAD work for now
    ENC_TMTM_CKEY = 'ENCRYPTED_TOKEN_SERVER_KEY' # Config key for encrypted token-server MTM.
    ENC_PMTM_CKEY = 'ENCRYPTED_BROKER_KEY' # Config key for encrypted broker-peer MTM.
    try:
        queryvalues = cgi.parse_qs(string.split(reqhandler.path, '?', 1)[1])
    except IndexError:
        queryvalues = {}

    def redirect_startup(reqhandler=reqhandler):
        switch_builtin_request_handler_dict(startup_requestbuiltins) # switch to startup mode.
        try:
            reqhandler.send_rewrite_response('/')
        finally:
            user_entered_passphrase.set()

    if queryvalues.has_key('NoPhrase'):
        try:
            confirm = queryvalues['NoPhraseConfirm'][0]
        except KeyError:
            confirm = ''
        if string.upper(confirm) == 'NO PASSPHRASE':
                confutils.confman.dict['ENCRYPT_LOCAL_KEYS'] = 'no'
                confutils.confman.save()
                redirect_startup()
        else:
            reqhandler.variables.set('PhraseError', "If you do not want to encrypt your Broker's identity you must spell 'No Passphrase' correctly (case insensitive).")
            reqhandler.send_entity(reqhandler.show('EnterNewPassphrase'))

    elif queryvalues.has_key('PhraseConfirm') or queryvalues.has_key('NewPhrase'):
        ## entering a new passphrase
        newPhrase = ''
        phraseConfirm = ''
        try:
            newPhrase = queryvalues['NewPhrase'][0]
            phraseConfirm = queryvalues['PhraseConfirm'][0]
        except KeyError:
            ## failure
            newPhrase = 0
            phraseConfirm = 1
        if newPhrase == phraseConfirm:
            # passphrases match, save and go
            confutils.passphrase = queryvalues['NewPhrase'][0]
            redirect_startup()
        else:
            reqhandler.variables.set('PhraseError', "Your passphrases didn't match.  Please try again.")
            reqhandler.send_entity(reqhandler.show('EnterNewPassphrase'))
    elif queryvalues.has_key('Phrase'):
        # check passphrase here
        passphrase = queryvalues['Phrase'][0]
        tokenKey = None
        brokerKey = None
        try:
            tokenKey = confutils.confman.dict[ENC_TMTM_CKEY]
            brokerKey = confutils.confman.dict[ENC_PMTM_CKEY]
        except KeyError:
            pass
        try:
            if tokenKey:
                Passphrase.decrypt(passphrase, tokenKey)
            if brokerKey:
                Passphrase.decrypt(passphrase, brokerKey)
        except Passphrase.InvalidPassphrase:
            reqhandler.variables.set('PhraseError', 'The passphrase you entered was invalid, please try again.')
            reqhandler.send_entity(reqhandler.show('EnterPassphrase'))
        else:
            confutils.passphrase = passphrase
            redirect_startup()
    else:
        if confutils.confman.dict['ENCRYPT_LOCAL_KEYS'] == 'no':
            reqhandler.variables.set('PhraseError', 'Local key encryption turned off, please enter your passphrase one last time so we can store your keys unencrypted.')
        else:
            reqhandler.variables.set('PhraseError', '')
        if confutils.confman.dict.has_key(ENC_TMTM_CKEY) or confutils.confman.dict.has_key(ENC_PMTM_CKEY):
            reqhandler.send_entity(reqhandler.show('EnterPassphrase'))
        else:
            # unencrypted keys
            reqhandler.variables.set('phrase_error', '')
            reqhandler.send_entity(reqhandler.show('EnterNewPassphrase'))

        

def handle_startup(reqhandler, servicemtm):
    debuglog( "entered handle_startup\n", v=9)
    reqhandler.send_entity(reqhandler.show('startup'))

def handle_root(reqhandler, servicemtm):
    debuglog( "entered handle_root\n", v=9)
    reqhandler.send_entity(reqhandler.show('mainmenu'))

def handle_direct_dtml_show(reqhandler, servicemtm):
    # XXX HACK this regex needs to sync-up with the one specified in the builtins dict
    dtmlfilename = re.search(r"^/(?P<dtmlfilename>\w+).dtml$", reqhandler.path).group('dtmlfilename')
    if dtmlfilename:
        reqhandler.send_entity(reqhandler.show(dtmlfilename))
    else:
        reqhandler.send_error('bad DTML file name given\n', httpcode=404)

def handle_search(reqhandler, servicemtm):
    """
    There are two stages to this handler depending on whether arguments are given.
    """
    userinput = reqhandler.variables.get('userinput')
    debuglog( "type of reqhandler.variables is %s\n" % `type(reqhandler.variables)`, v=9)
    debuglog( "reqhandler.variables is %s\n" % `reqhandler.variables`, v=9)

    debuglog( "entered handle_search\n", v=9)
    debuglog('userinput is %s\n' % `userinput`)
    if userinput.has_key('activity_start'):
        del userinput['activity_start']
        for key, value in userinput.items():
            debuglog( "userinput item: %s : %s\n" % (`key`, `value`) , v=9)
            if value:
                userinput[key] = value[0]
            else:
                userinput[key] = ''

        querystr = urllib.urlencode(userinput)
        reqhandler.send_rewrite_response(reqhandler.variables.get('baseurl') + '/search_query?' + querystr)
        return

    typedefrep = globaldict['typedefrep']
    typedefrep.reload_directory()
    typedefrep.reload_typedefs()
    reqhandler.send_entity(reqhandler.show('search', typedefrep=typedefrep))
    return

# maps mojoid -> dinode, this gets populated by all content tracker searches and mojoid searches.  Using this
# it is generally safe to use mojoid links from the search results page instead of the much longer dinode links.
MojoidToBinaryDinodeCache=Cache.StatsCache(maxitems=1000, autoexpireinterval=15*60, autoexpireparams={'maxage': 6*60*60, 'staletime': 2*60*60})

def handle_search_query(reqhandler, servicemtm, start_results_template='search_results_start', results_unit_template='search_results_unit', end_results_template='search_results_end', results_gap_template='search_results_gap', expiresin=2*60*60):
    """
    @precondition `servicemtm' must not be None.: servicemtm is not None
    """
    assert servicemtm is not None, "precondition: `servicemtm' must not be None."

    debuglog( "entered handle_search_query\n", v=9)
    userinput = reqhandler.variables.get('userinput')
    queryvalues = {}
    # if the search form's selection was to search based on the input fields instead of searching for all content:
    if int(userinput.get('all', [0])[0]) == 0:
        # First, transform queryvalues into a dict suitable to pass to multi_content_tracker_lookup:
        for key in userinput.keys():
            if key not in ('activity_start', 'ctype', 'all', 'public', 'tracker_ids'):
                value = userinput[key][0]
                if value:  # don't allow empty fields to even show up in the query, thier wildcard is implied
                    queryvalues[key] = value

    ctype = userinput['ctype'][0]

    # Search for public content == 'yes', search private content == 'no':
    # Searching with public == 'no' when you don't have permission has the effect of searching with public == 'yes'. 
    public = userinput.get('public', ['yes'])[0]

    # If this field is present it is a list of ascii (mojosixbit) content tracker ids:
    tracker_ids = userinput.get('tracker_ids')
    # XXX Should we delete the above entries from userinput?  Check for DTML bugs.
    
    exactmatch = queryvalues.get('matchexactly')
    if exactmatch:
        del queryvalues['matchexactly']

    # TODO verify value type or other info?

    # if no fields were filled in, make a query with a % wildcard in
    # the dinode to return all fields associated with this content type.
    if not queryvalues:
        queryvalues['__DINODE__'] = ""  # wildcard search

    # If we were given an explicit list of content tracker ids, use it:
    if tracker_ids:
        tracker_ids = map(idlib.to_binary, tracker_ids)
        contenttrackers = []
        # this serial loop of blocking single lookups is gross.  however, most of the time
        # the info should already be in the cache.
        for tid in tracker_ids:
            infodict = MetaTrackerLib.get_contact_info(servicemtm, tid, timeout=60)
            if infodict:
                contenttrackers.append((tid, infodict))
    else: # -else, get a list from our Broker and prioritize it.
        # TODO get the authoritative list of content trackers for this type if it is a branded/moderated content type.
        # find content trackers for public/unmoderated types:
        contenttrackers = contentlib.find_content_trackers(servicemtm,
            min=int(confman.dict['CONTENTTRACKER_CLIENT']['SEARCH_QUERY_LIMITS']['MIN']),
            max=int(confman.dict['CONTENTTRACKER_CLIENT']['SEARCH_QUERY_LIMITS']['MAX']) ) # XXX this won't scale w/ lots of trackers...
        if not contenttrackers:
            reqhandler.show_error('No Content Trackers found.  Please try again later.')
            return

        # sort by preference
        # XXX this is a total hack to put in msgtype and fake msgbody here
        contenttrackers = servicemtm.get_handicapper().sort_by_preference(contenttrackers, 'content tracker lookup', {'XML data': ""})
        if len(contenttrackers) == 0:
            debuglog("WARNING: no content trackers found after handicapping.\n")
            reqhandler.show_error('No Content Trackers found.  Please try again later.')
            return

        # be -sure- the list is not longer than max (I think find_content_trackers does this for us but it can't hurt)
        contenttrackers = contenttrackers[:int(confman.dict['CONTENTTRACKER_CLIENT']['SEARCH_QUERY_LIMITS']['MAX'])]

    reqhandler.variables.set('download_links_use_mojoid', confman.is_true_bool(['USE_MOJOID_IN_DOWNLOAD_LINKS'], default="no"))
    # XXXX For Incremental Display:
    knowndinodes = {} # Keeps track of duplicate results.
    # 
    descdicts_to_show_queue = Queue.Queue(0)
    # A call-back which displays a single content trackers results to the user:
    def show_result(widget, resultlist, reqhandler=reqhandler, knowndinodes=knowndinodes, descdicts_to_show_queue=descdicts_to_show_queue):
        #debug.mojolog.write("Received the following results from %s:\n", args=(widget.get_counterparty_id(),), v=4, vs='DEBUG')
        for ctype, descdict in map(contentlib.content_xml_to_dict, resultlist):
            if ctype is None:
                debug.mojolog.write("content tracker %s returned unparsable search result\n", args=(widget.get_counterparty_id(),), v=4, vs='MojoHTTPd')
                continue
            #debug.mojolog.write("   Result: %s\n" % (descdict,), v=4, vs='DEBUG')
            dinode = descdict.get('__DINODE__')
            if not dinode:
                # This search result is almost completely useless, since we can't retrieve the data.
                # Also, content trackers shouldn't do this, so someone has tweaked there content tracker,
                # or it's a third-party implementation.
                debug.mojolog.write("content tracker %s returned search result without a dinode!\n", args=(widget.get_counterparty_id(),), v=4, vs='MojoHTTPd')
                #D#debug.mojolog.write('descdict: %s\n', args=(descdict,), vs='MojoHTTPd')
                continue # This entry is useless, it can't be merged.
            # XXX we should verify that the dinode is parsable here
            if knowndinodes.has_key(dinode):
                # Duplicate result.
                continue
            knowndinodes[dinode] = None # Record this dinode.
            computedmojoid = contentlib.get_mojoid_from_asciidinode(dinode)
            if not descdict.has_key('__MOJOID__'):
                # Fix any results that don't have a mojoid
                descdict['__MOJOID__'] = computedmojoid
            else:
                if computedmojoid != descdict['__MOJOID__']:
                    debuglog("Warning: skipping invalid content description due to incorrect mojoid\n")
                    continue
            try:
                MojoidToBinaryDinodeCache[computedmojoid] = mojoutil.a2b(descdict['__DINODE__'])
            except:
                debuglog("Warning: skipping invalid content description due to unparsable dinode\n")
                continue
            # hand this result to the rendering thread
            descdicts_to_show_queue.put(descdict)
        
        # putting the counterparty id on the queue indicates the end
        # of results from the given counterparty id.
        descdicts_to_show_queue.put(widget.get_counterparty_id())
    
    done_event = threading.Event()
    def search_result_rendering_thread(reqhandler=reqhandler, descdicts_to_show_queue=descdicts_to_show_queue, done_event=done_event, public=public, results_gap_template=results_gap_template, results_unit_template=results_unit_template):
        try:
            to_show = descdicts_to_show_queue.get()
            while to_show is not None:
                if hasattr(reqhandler.wfile, 'closed') and reqhandler.wfile.closed:
                    break  # if the browser has closed the connection, exit now instead of rendering into never never land
                try:
                    if type(to_show) is types.StringType:  # indicates the end of results from one server
                        reqhandler.variables.set('contenttracker_id', idlib.to_ascii(to_show))
                        reqhandler.wfile.write(reqhandler.show(results_gap_template))
                    else:
                        reqhandler.variables.set('descdict', to_show)
                        reqhandler.variables.set('public', public)
                        reqhandler.wfile.write(reqhandler.show(results_unit_template))
                    reqhandler.wfile.flush()
                except (socket.error, IOError):
                    debuglog("couldn't display content tracker result, web browser already closed connection\n", v=4)
                    return
                to_show = descdicts_to_show_queue.get()
        finally:
            done_event.set()

    # rendering search results is -extreemly- slow using DTML according to the profiler.  We push it off to be done on its
    # own thread here so that it doesn't monopolize the doq and prevent other broker activities.
    displaythread = LoggedThread(name="search result rendering thread", target=search_result_rendering_thread)
    displaythread.setDaemon(1)
    displaythread.start()

    # this page can be cached by the browser for two hours (this will
    # only matter so that hitting the back button from a download
    # doesn't re-issue a search, it will not affect searches with different
    # criteria.)
    reqhandler.expiresin = expiresin
    reqhandler.send_html_response(200)

    reqhandler.variables.set('contenttrackers', contenttrackers)

    reqhandler.wfile.write(reqhandler.show(start_results_template))
    reqhandler.wfile.flush()

    searchresult = contentlib.multi_content_tracker_lookup(contenttrackers, servicemtm, typename=ctype, querydict=queryvalues, individualcallback=show_result, exactmatch=exactmatch)
    descdicts_to_show_queue.put(None)  # tell the display thread the search has completed

    # unnecessary idea: we could loop on shorter waits here here checking if the http socket was still
    # open and if not, tell the display thread to end early and exit.
    while not done_event.isSet():
        done_event.wait(2)  # waiting in a loop like this prevents the thread from hanging when it needs to shutdown
        if hasattr(reqhandler.wfile, 'closed') and reqhandler.wfile.closed:
            return  # the user hit their 'stop' button, no sense in waiting around
        if globaldict.get('shutdown'):
            return
    try:
        reqhandler.wfile.write(reqhandler.show(end_results_template))
        reqhandler.wfile.flush()
    except (socket.error, IOError):
        pass  # so what, we were done, the user hit 'stop' early


def handle_publish(reqhandler, servicemtm):
    debuglog( "entered handle_publish\n", v=9)
    # This depends on the reqhandler setting up userinput inside reqhandler.variables!
    userinput = reqhandler.variables.get('userinput')

    # First extract the activity_start variable and remove it from userinput:
    activity_start = userinput.get('activity_start', (None,))[0]
    if activity_start:
        del userinput['activity_start']

    # Extract file:
    file = userinput.get('file', (None,))[0]
    if file:
        del userinput['file']
        
    # Extract ctype:
    ctype = userinput.get('ctype', (None,))[0]
    if ctype:
        del userinput['ctype']
        
    # Now create a content description dict from the userinput (these values override auto-described values):
    cdesc={}
    for key in userinput.keys():
        cdesc[key] = userinput[key][0]
        del userinput[key]

    # userinput should be empty now.

    if file and (ctype != 'NONE'):
        # Do a file sanity check:
        if not os.path.exists(file):
            userinput['badfile'] = ['yes']

        # If there is a specified file, we are going to attempt to auto-describe it.  We must make sure user-provided values override
        # the automatically generated values!
        ctype, cdesc = Describe.describe(file, ctype, cdesc, typedefrep=globaldict['typedefrep'])

    if activity_start:
        # If the user clicked the Publish button, go at it:
        cmdnum = globaldict['commandmanager'].add_command(UserCommand.Publish(globaldict['broker'], file, ctype, cdesc))
        reqhandler.send_rewrite_response(reqhandler.variables.get('baseurl') + '/activity?id=%d' % cmdnum)
        return

    # Else re-insert all the (possibly changed) values, except activity_start:
    userinput['file'] = [file]
    userinput['ctype'] = [ctype]
    # Unwrap our content description dict into user-input:
    if cdesc:
        for key, value in cdesc.items():
            userinput[key] = [value]

    # Else, replace the userinput with our tamered version:
    reqhandler.variables.set('userinput', userinput)
    reqhandler.variables.set('urllib', urllib)

    typedefrep = globaldict['typedefrep']
    typedefrep.reload_directory()
    typedefrep.reload_typedefs()

    # And show the publish input page:
    reqhandler.send_entity(reqhandler.show('publish', typedefrep=typedefrep))
    return

def handle_activity(reqhandler, servicemtm):
    # This depends on the reqhandler setting up userinput inside reqhandler.variables!
    userinput = reqhandler.variables.get('userinput')

    cancel = userinput.get('cancel', [None])[0]
    id = userinput.get('id', [None])[0]
    if (cancel is not None) and (id is not None):
        globaldict['commandmanager'].get_command(int(id)).cancel_publish('cancelled by user')

    reqhandler.send_entity(reqhandler.show('activity'))
    return
    
def handle_status(reqhandler, servicemtm):
    debuglog( "entered handle_status\n", v=9)
    reqhandler.send_entity(reqhandler.show('status'))
    return    
    
def handle_configure(reqhandler, servicemtm):
    reqhandler.cachable = false
    userinput = reqhandler.variables.get('userinput')

    message = ''
    if userinput.has_key('Save'):
        del userinput['Save']
        try:
            set_config_values(userinput)
            message = 'New settings saved. Restart Broker to use.'
        except:
            message = 'Error while saving settings, please try again.'
    elif not userinput.has_key('configureoption') or userinput.has_key('Reset'):
        # we just reset the form or clicked into the configure page from the menu bar
        if userinput.has_key('Reset'):
            del userinput['Reset']

    # XXX this means that if you forget to put the hidden INPUT in advanced/general in the dtml it will silently work in the other page but not carry the value through.
    # initialize userinput with a flat dict of config values from confman
    userinput.update(Presentation.get_flat_confman_dict())

    userinput['message'] = [message]   # unit_menu.dtml expects it in this weird place for some reason

    # convert userinput value-lists into the first value in the list
    for key, value in userinput.items():
        if key == 'message':
            continue
        if type(value) is types.ListType:
            userinput[key] = value[0]

    dtmlvars = {}
    dtmlvars.update(userinput)
    # include the configuration defaults with _DEFAULT appended to their name
    dtmlvars.update(Presentation.get_flat_confman_dict(confutils.confdefaults, appendstr='_DEFAULT'))
    dtmlvars.update({
        'userinput': userinput,
        'message': [message],
        'configureoption': userinput.get('configureoption', 'general'),
    })

    reqhandler.send_entity(
        apply(reqhandler.show, ('configure',), dtmlvars)
    )


def handle_stash(reqhandler, servicemtm):
    reqhandler.cachable = false
    reqhandler.send_entity(reqhandler.show('stash'))

def handle_help(reqhandler, servicemtm):
    reqhandler.cachable = false
    reqhandler.send_entity(reqhandler.show('help'))

def handle_broker_shutdown(reqhandler, servicemtm):
    reqhandler.cachable = false
    reqhandler.send_entity(reqhandler.show('shutdown'))
    reqhandler.wfile.flush()
    globaldict['broker'].exit()

def handle_reshow_termsofservice(reqhandler, servicemtm):
    reqhandler.cachable = true
    reqhandler.send_entity(reqhandler.show('termsofservice'))

def handle_mojoid_get(reqhandler, servicemtm):
    """
    Handle nice short urls that only contain the sha1 of the raw uncompressed dinode
    """
    # parse mojoid and part out of the URL
    m = globaldict['regexes']['extractmojoid'].match(reqhandler.path)
    if not m :
        reqhandler.show_error('No MojoID supplied or incorrect format used.')
        return
    
    mojoid = m.group('mojoid')
    part = m.group('part')
    # If the URL has a trailing slash then part will be "/", else it will be None.
    if part is None:
        # return 301 to the same URL with a trailing slash.
        reqhandler.send_rewrite_response(reqhandler.path + "/")
        return

    part = urllib.unquote(part)

    dinodedata = get_dinodedata_for_mojoid(reqhandler, servicemtm, mojoid)
    if dinodedata is None:
        return  # an error occurred, show_error should have been called already

    # perform a regular dinode query now that we have it
    handle_dinode_stream(reqhandler, servicemtm, dinodedata=dinodedata, part=part)


def handle_dinode_stream(reqhandler, servicemtm, dinodedata=None, part=None):
    """
    If dinodedata is None, the dinode is extracted from the path.  If this happens
    the value of part is assigned according to the path, and the value passed in is
    ignored.
    """
    if dinodedata is None :
        m = globaldict['regexes']['extractdinode'].match(reqhandler.path)
        if not m :
            reqhandler.show_error('No Dinode supplied, or incorrect format used.', httpcode=400)
            return
        try:
            dinodedata = mojosixbit.a2b(m.group('dinode'))
        except mojosixbit.Error:
            reqhandler.show_error('Bad dinode.', httpcode=400)
            return
        part = m.group('part')
        # If the URL has no trailing slash, then part will be None.
        if part is None:
            # return 301 to the same URL with a trailing slash.
            reqhandler.send_rewrite_response(reqhandler.path + "/")
            return

        part = urllib.unquote(part)

    try:
        download_dinode(reqhandler, servicemtm, dinodedata, part)
    except socket.error:
        pass  # http client closed connection, ignore it


def get_dinodedata_for_mojoid(reqhandler, servicemtm, mojoid):
    """
    Given a mojoid, get the dinodedata for it (either from our cache, or do
    an actual lookup if we have to).
    """
    dinodedata = MojoidToBinaryDinodeCache.get(mojoid)
    if not dinodedata:
        # lookup the id using a __MOJOID__ content tracker lookup
        querydict = {'__MOJOID__': mojoid}
        # find content trackers:
        contenttrackers = contentlib.find_content_trackers(servicemtm,
            min=int(confman.dict['CONTENTTRACKER_CLIENT']['MOJOID_QUERY_LIMITS']['MIN']),
            max=int(confman.dict['CONTENTTRACKER_CLIENT']['MOJOID_QUERY_LIMITS']['MAX']) ) # XXX this won't scale w/ lots of trackers...
        if not contenttrackers:
            reqhandler.show_error('No Content Trackers found to perform the mojoid lookup.  Please try again later.')
            return

        # sort by preference
        # XXX this is a total hack to put in msgtype and fake msgbody here
        contenttrackers = servicemtm.get_handicapper().sort_by_preference(contenttrackers, 'content tracker lookup', {'XML data': ""})

        # TODO implement a verify_success_func for multi_content_tracker_lookup that ensures that the dinode returned
        # was indeed the droid we were looking for so that we don't stop after the first result comes in only to find
        # out that its a wookie.  -greg 22/aug/2000
        lookupresult = contentlib.multi_content_tracker_lookup(contenttrackers, servicemtm, typename='MOJOID', querydict=querydict, exactmatch=1, minsuccesses=1)

        if not lookupresult['list']:
            reqhandler.show_error('Mojo ID <code>%s</code> could not be found at this time.' % mojoid)
            return
        
        # parse the XML in lookupresult to get the dinode
        for matchxml in lookupresult['list']:
            descname, descdict = contentlib.content_xml_to_dict(matchxml)
        
            if descname:
                asciidinode = descdict.get('__DINODE__')
                if not asciidinode:
                    debuglog("NOTE: skipping content description w/o a DINODE\n")
                    continue
                del descdict['__DINODE__']

                returnedmojoid = descdict.get('__MOJOID__')
                if returnedmojoid and returnedmojoid != mojoid:
                    debuglog("NOTE: skipping content description with undesired mojoid\n")
                    continue

                # compute the mojoid from the dinode ourself
                computedmojoid = contentlib.get_mojoid_from_asciidinode(asciidinode)
                if computedmojoid != mojoid:
                    debuglog("NOTE: skipping invalid content description due to incorrect mojoid\n")
                    continue

                if descdict.has_key('__PATH__'):
                    del descdict['__PATH__']
                
                dinodedata = mojoutil.a2b(asciidinode)
            else:
                debuglog('MojoHTTPD Warning: skipping unparsable XML content description - %s\n' % `xmlstr`)
        
        # cache this lookup so that we don't have to do it later
        MojoidToBinaryDinodeCache[mojoid] = dinodedata

    return dinodedata


def handle_mojoid_save_to_disk(reqhandler, servicemtm, part=''):
    getmojoid_re = re.compile(r"^/save_id/(?P<mojoid>[-_A-Za-z0-9]+)(|/(?P<part>.*))$", re.I)
    m = getmojoid_re.match(reqhandler.path)
    mojoid = m.group('mojoid')
    part = m.group('part')

    dinodedata = get_dinodedata_for_mojoid(reqhandler, servicemtm, mojoid)
    if dinodedata is None:
        return  # an error occurred, show_error should have been called already

    return handle_dinode_save_to_disk(reqhandler, servicemtm, dinodedata=dinodedata, part=part)


def handle_dinode_save_to_disk(reqhandler, servicemtm, dinodedata=None, part=''):
    """
    If dinodedata is None, the dinode is extracted from the path.  If this happens
    the value of part is assigned according to the path, and the value passed in is
    ignored.
    """
    if dinodedata is None :
        getdinode_re = re.compile(r"^/save_dinode/(?P<dinode>[-_A-Za-z0-9]+)(|/(?P<part>.*))$", re.I)
        m = getdinode_re.match(reqhandler.path)
        if not m :
            reqhandler.show_error('No Dinode supplied, or incorrect format used.', httpcode=400)
            return
        try:
            dinodedata = mojosixbit.a2b(m.group('dinode'))
        except mojosixbit.Error:
            reqhandler.show_error('Bad dinode.', httpcode=400)
            return
        part = m.group('part')
        # If the URL has no trailing slash, then part will be None.
        if part is None:
            # return 301 to the same URL with a trailing slash.
            reqhandler.send_rewrite_response(reqhandler.path + "/")
            return

    part = urllib.unquote(part)

    try:
        save_dinode_to_disk(reqhandler, servicemtm, dinodedata, displayname=part[:64])
    except socket.error:
        pass  # http client closed connection, ignore it

def handle_download_status_graph(reqhandler, servicemtm):
    m = re.search(r"/(id|mojoid)/(?P<mojoid>"+mojosixbit._asciihash_re_str+")(|/(?P<part>.*))$", reqhandler.path, re.I)
    if m:
        mojoid = m.group('mojoid')
        if not mojoid:
            # TODO have unsplit cache based on mojoid rather than binary dinode so that this won't matter
            reqhandler.show_error('Bad mojoid in URL.' % mojoid, httpcode=400)
            return
    else:
        m = re.search(r"/dinode/(?P<dinode>[a-z0-9-_]+)(|/(?P<part>.*))$", reqhandler.path, re.I)
        if not m:
            reqhandler.show_error('Bad dinode in URL', httpcode=400)
            return
        asciidinode = m.group('dinode')
        mojoid = contentlib.get_mojoid_from_asciidinode(asciidinode)
        dinode = mojoutil.a2b(asciidinode)

    download_monitor = unsplit.get_download_monitor(mojoid)
    debug.mojolog.write('download_monitor was %s\n', args=(download_monitor,), v=5, vs='MojoHTTPd')
    if not download_monitor or not hasattr(download_monitor, 'get_image_size') or not hasattr(download_monitor, 'write_image'):
        reqhandler.show_error('no graphical download status monitor exists for '+mojoid, httpcode=404)
        return

    part = m.group('part')
    if part:
        imagefile = StringIO()
        download_monitor.write_image(imagefile, 'gif')
        imagedata = imagefile.getvalue()
        del imagefile
        reqhandler.expiresin = 2  # expire this dynamic image semi-quickly
        reqhandler.send_entity(imagedata, headers={'content-type': 'image/gif'})
        del imagedata
    else:
        reqhandler.variables.set('toomuchdetail', download_monitor.toomuchdetail)
        reqhandler.variables.set('mojoid', mojoid)
        x, y = download_monitor.get_image_size()
        reqhandler.variables.set('img_width', x)
        reqhandler.variables.set('img_height', y)
        reqhandler.variables.set('save_to_disk_path', download_monitor.save_to_disk_path)
        reqhandler.variables.set('urllib', urllib)
        reqhandler.send_entity(reqhandler.show('download_status_graph'))


def handle_mods(reqhandler, servicemtm):
    # This method uses reqhandler.broker correctly, rather than relying on global variables.
    # In other words, it is safe to change the parameter interface, or change this into a RequestHandler method.

    reqhandler.server.broker.modman.scan() # In case they add stuff to the MojoMod directory at runtime.

    # Handle the checkboxes which signfy which modules to start or stop:
    startmods = reqhandler.variables.get('userinput').get('start_mod')

    # If there were no checkboxes, then don't process (which would stop all MojoMods!):
    if startmods is not None:
        startmods.remove('NULL') # A hack, because cgi.parse_qs('start_mod=') yields {}, instead of {'start_mod': []}
        for mod in reqhandler.server.broker.modman.mojomods.keys():
            if mod in startmods:
                reqhandler.server.broker.modman.start(mod)
            else:
                reqhandler.server.broker.modman.stop(mod)

    reqhandler.send_entity(reqhandler.show('mods_main'))

# Built-In request handlers:  (Broker.py sets these as appropriate)
#(Use "switch_builtin_request_handler_dict" to switch between sets of handlers.)

# used when the user still neexundods to sign the terms of service
tos_requestbuiltins = {
    re.compile(r"(?!^\/images\/)") : handle_terms_of_service,
}

# used when the broker is waiting for the user to enter their passphrase
phrase_requestbuiltins = {
    re.compile(r"(?!^\/images\/)") : handle_passphrase,
}

# used when the broker is still in the process of starting up, before
# the real mojohttpd is ready.
startup_requestbuiltins = {
    re.compile(r"(?!^\/images\/)") : handle_startup,
}

main_requestbuiltins = {
    re.compile(r"^$|^/$|^/index.htm[l]?$", re.I) : handle_root,
    re.compile(r"^/(?P<dtmlfilename>\w+).dtml$") : handle_direct_dtml_show,
    re.compile(r"^/search(\?.*)?$") : handle_search,
    re.compile(r"^/search_query") : handle_search_query,
    re.compile(r"^/publish") : handle_publish,
    re.compile(r"^/stash") : handle_stash,
    re.compile(r"^/(configure|preferences)") : handle_configure,
    re.compile(r"^/help") : handle_help,
    re.compile(r"^/activity") : handle_activity,
    re.compile(r"^/(status|log)") : handle_status,
    re.compile(r"^/Dinode", re.I) : handle_dinode_stream,
    re.compile(r"^/save_dinode", re.I) : handle_dinode_save_to_disk,
    re.compile(r"^/save_id", re.I) : handle_mojoid_save_to_disk,
    re.compile(r"^/(id|mojoid)", re.I) : handle_mojoid_get,
    re.compile(r"^/(shutdown|exit)", re.I) : handle_broker_shutdown,
    re.compile(r"^/termsofservice", re.I) : handle_reshow_termsofservice,
    re.compile(r"^/mods", re.I) : handle_mods,
    re.compile(r"^/download_status/(id|mojoid|dinode)/[a-z0-9-_]+/", re.I): handle_download_status_graph,
}

# The default set of handlers (Broker.py will change these):
requestbuiltins = tos_requestbuiltins

def switch_builtin_request_handler_dict(dict):
    """
    This function ensures you can alter the way the web server handles
    incoming get requests without disrupting its serving.
    """
    global requestbuiltins
    requestbuiltins = dict

def set_servicemtm(servicemtm, tokenservermtm):
    """
    This should be called only once.  It passes mojohttpd a reference to the servicemtm.
    """
    if hasattr(Presentation, 'variables'):
        Presentation.variables.prepare_servicemtm_values(servicemtm,tokenservermtm) # For Presentation variables.

# The following code is for handling HTTP POST entity content types:
def post_type_handler_form_urlencoded(reqhandler):
    try:
        entity = reqhandler.rfile.read(int(reqhandler.headers['Content-Length']))
    except KeyError:
        debug.mojolog.write('No Content-Length specified for form input.', vs='MojoHTTPD', v=2)
        entity = reqhandler.rfile.read()
        
    try:
        userinput = cgi.parse_qs(entity)
    except IndexError:
        userinput = {}
        
    reqhandler.variables.get('userinput').update(userinput)
    return

post_type_handlers = {
    'application/x-www-form-urlencoded': post_type_handler_form_urlencoded,
    }


# The main procedure:
def run(doq, port=None, continue_func=None, broker=None): # Raises ValueError when confman's 'CLIENT_PORT' is not convertible to an int.
    """
    This starts the web-server.
    
    Precondition: you must assign Privacy Code globaldict before calling this function, unless you'd like to operate without a privacy code..

    Note:  Before the back-end code can function, you must call "set_servicemtm".
    """
    debuglog('starting.\n')

    if port is None:
        port = int(confman["CLIENT_PORT"])
    if string.lower(confman.dict['YES_NO']['ALLOW_ACCESS_FROM_OUTSIDE_LOCALHOST']) == 'yes' or sys.platform=='darwin1':
        address = ('', port)
    else:
        address = ('127.0.0.1', port)

    if not globaldict.has_key('privacycode'):
        if confman.is_true_bool(('PUT_PRIVACY_CODE_IN_URLS_v2',)):
            debuglog('creating a new privacy code.\n')
            globaldict['privacycode'] = mojosixbit.b2a(randsource.get(12))
        else:
            debuglog('NOT creating a new privacy code (broker is configured not to use privacy code).\n')
            globaldict['privacycode'] = ""
       
    debuglog('changing directory to %s.\n' % os.path.expandvars(confman.dict['PATH']['WEBROOT_DIR']))
    os.chdir(os.path.expandvars(confman.dict['PATH']['WEBROOT_DIR']))

    # We assume this is already set.  (I hate these globaldict semantics...)
    if confman.is_true_bool(('PUT_PRIVACY_CODE_IN_URLS_v2',)):
        Presentation.variables = Presentation.VariableManager(globaldict['typedefrep'], globaldict['privacycode'])
    else:
        Presentation.variables = Presentation.VariableManager(globaldict['typedefrep'])
    Presentation.variables.set('broker', broker)
    debuglog('Presentation Variables initialized.\n')

    if not globaldict.get('commandmanager'):
        debuglog('Initializing UserCommand.\n')
        globaldict['commandmanager'] = UserCommand.MultiCommandManager(doq)
        Presentation.variables.set('cmdmanager', globaldict['commandmanager'])
    try:
        debuglog('binding port.\n')
        remainingporttries = 100
        debuglog('starting.\n')
        while 1:
            try:
                # The following is a hackish assignment which makes the "Server" HTTP response
                # header more descriptive.  If it breaks something, it can be commented out
                # without hurting anything.
                MojoRequestHandler.server_version = 'MojoHTTPD/' + confutils.BROKER_VERSION_STR
                httpd = MojoNationHTTPServer.ReuseThreadingHTTPServer(broker, address, MojoRequestHandler)
                globaldict['httpd'] = httpd
                break
            except socket.error:
                # XXX this is a lazy hack here to avoid looping forever if bind continually barfs.
                if remainingporttries <= 0:
                    debuglog('could not bind to a port for local web interface:\n')
                    traceback.print_exc(file=debug.mojolog)
                    raise MojoHTTPDError, "couldn't bind to port"
                newaddr = (address[0], address[1] + 1)  # increment the port number
                debuglog("couldn't bind to %s, trying %s instead.\n", args=(address, newaddr))
                address = newaddr
                remainingporttries = remainingporttries - 1
        
        globaldict['port'] = address[1]    # note the actual port we bound to

        debuglog('Creating Intro Page.\n')
        initialize_intro_page(address[1])

        debuglog('beginning service.\n')
        if continue_func is not None:
            request_handler_loop(httpd, continue_func=continue_func)
        else:
            request_handler_loop(httpd)
    finally:
        globaldict['httpd'] = None # Clear the module-scoped reference.
        try:
            del httpd  # unbind the server  (clear our local reference)
        except UnboundLocalError, e:
            # httpd wasn't assigned, we don't care:
            pass
        time.sleep(0.5) # this hopefully lets the unbind take effect within the OS
        debuglog('exiting.\n')


# Back-end:
def set_config_values(values):
    for key, value in values.items():
        Presentation.set_config_value(key, value[0])
    # this call causes confman.save() to be called and save the new config
    globaldict['broker'].start_or_stop_configured_servers()
    # force a new hello to be sent incase we changed any server configs
    globaldict['broker'].task_periodic_hello(reschedule=false)


def download_dinode(reqhandler, servicemtm, dinode, part=None):
    try:
        archivefile = unsplit.ReassemblerSplitFile(TM=servicemtm, dinode=dinode, download_monitor_class=BitmapDownloadMonitor, freeonclose=0)
    except sharemap.SharemapError, le:
        debug.mojolog.write('caught exception. le: %s\n', args=(le,), v=3, vs='MojoHTTPd')
        reqhandler.show_error("the map to the file was garbled", httpcode=500)
        return
    except (IOError, archive.ArchiveError, MojoErrors.ReassemblyError), description:
        reqhandler.show_error('Could not download file.  Some of the block servers that have the pieces of this file may be off-line and may come back on-line later, so please try again later.  (part of the file couldn\'t be retrieved)', httpcode=500)
        return

    # prepare to read from the archive (the reader will load its index)
    try:
        arc = archive.Reader(archivefile)
    except MojoErrors.ReassemblyError, description:
        reqhandler.show_error('Could not download file.  Some of the block servers that have the pieces of this file may be off-line and may come back on-line later, so please try again later.  (publication index couldn\'t be retrieved)', httpcode=500)
        archivefile.close()
        return
    except archive.ArchiveError, description:
        reqhandler.show_error("Error: bad publication: %s." % cgi.escape(hr(description), true), httpcode=500)
        archivefile.close()
        return

    # Before checking for a part or an index, check to see if this archive
    # contains only one file.  If it does, retrieve it without showing an
    # index.
    if not part and len(arc.index) == 1:
        part = arc.index[0].get_pathname()
        if reqhandler.path[-1] == '/':
            reqhandler.send_rewrite_response(reqhandler.path + urllib.quote(part))
        else:
            reqhandler.send_rewrite_response(reqhandler.path + '/' + urllib.quote(part))
        archivefile.close()
        return

    # If no part was specified look for an index.html otherwise display
    # a raw archive index.  Or, if the first file in the archive is
    # a html file, jump right to it instead of showing a file index.
    if not part :
        firstname = arc.index[0].get_pathname()
        if string.lower(firstname[-5:]) == '.html' or string.lower(firstname[-4:]) == '.htm':
            arc.seekpart(firstname)
            part = firstname
        else:
            try:
                arc.seekpart('index.html')
                part = 'index.html'
            except archive.ArchiveError:
                # try DOS naming convention...
                try:
                    arc.seekpart('index.htm')
                    part = 'index.htm'
                except archive.ArchiveError:
                    part = '__INDEX__'

    # Generate a "pretty" flat archive index in HTML
    if part == '__INDEX__' :
        reqhandler.expiresin = 60*60*24*365  # dinode based data never changes, let the browser cache it well
        htmlindex = []
        for partentry in arc.index:
            htmlindex.append('<tr><td align="left"><a href="%s">%s</a>&nbsp;</td><td align="right">%s&nbsp;</td></tr>' % (
                    urllib.quote(partentry.get_pathname()),
                    partentry.get_pathname(),
                    integer_to_pretty_string(partentry.size)
                ))
        reqhandler.variables.set('htmlindex', string.join(htmlindex, '\n'))
        reqhandler.send_entity(reqhandler.show('download'))
        archivefile.close()
        return

    # Generate a heirarchical tree archive index using the tree_index.dtml template
    if part and part[-len('__TREE_INDEX__'):] == '__TREE_INDEX__':
        subpath = part[:- len('__TREE_INDEX__') - 1]
        reqhandler.expiresin = 60*60*24*365  # dinode based data never changes, let the browser cache it well

        # export archive, archivetree and archiveindex to the DTML
        reqhandler.variables.set('arc', arc)
        if subpath:
            # pass in a partial subdirectory index
            tree = arc.get_parts_tree()
            subpathlist = string.split(subpath, '/')
            partstree = tree.get(subpathlist[0])
            if partstree is None:
                reqhandler.show_error('directory %s not found in archive.' % (subpathlist[0]), httpcode=404)
                return
            for subpathunit in subpathlist[1:]:
                partstree = partstree.get(subpathunit)
                if partstree is None:
                    reqhandler.show_error('subdirectory %s not found in archive.' % (subpathunit), httpcode=404)
                    return
            assert type(partstree) == type({}), "partstree %s wasn't a subdirectory for subpath %s in %s" % (partstree, subpath, tree)
            reqhandler.variables.set('arcpath', subpath)
        else:
            # pass in the entire archive index
            partstree = arc.get_parts_tree()
        dirs = filter(lambda k, st=partstree: type(st[k]) == type({}), partstree.keys())
        files = filter(lambda k, st=partstree: type(st[k]) != type({}), partstree.keys())
        reqhandler.variables.set('arctree', partstree)
        reqhandler.variables.set('arcdirs', dirs)
        reqhandler.variables.set('arcfiles', files)
        reqhandler.variables.set('dinode', dinode)
        reqhandler.variables.set('urlquote', urllib.quote)
        reqhandler.variables.set('build', js_list.build)
        reqhandler.variables.set('gettotal', js_list.gettotal)
        reqhandler.variables.set('range', range)
        reqhandler.send_entity(reqhandler.show('tree_index'))
        archivefile.close()
        return

    # remove relative path references (eggs/bar/../spam.gif -> eggs/spam.gif)
    # This has the "feature" that excess /..s will simply stay in the top directory
    if string.count(part, '/..') :
        def replace_last_pathbit(x, y) :
            try:
                return x[:string.rindex(x, '/')] + y
            except ValueError:
                return x+y
        debuglog('Compressing "/.." directories within %s\n' % part)
        path = reduce(replace_last_pathbit, string.split('/'+part, '/..'))[1:]
        debuglog('Resulting path: %s\n' % part)

    # try and open the part.  If that doesn't work, try and load part + "index.html" incase it was a
    # subdirectory containing an index.html.  If that still doesn't work, display an error.
    try:
        debuglog('seeking part "%s".\n' % part, v=4)
        partentry = arc.seekpart(part)
    except archive.ArchiveError, description:
        origpart = part
        # Append index.html, including a seperating / if needed to see
        # if this was a directory reference and if it contained an
        # index.html
        part = os.path.join(origpart + 'index.html')
        try:
            partentry = arc.seekpart(part)
        except archive.ArchiveError:
            ## Last try, could be DOSsified
            part = os.path.join(origpart + 'index.htm')
            try:
                partentry = arc.seekpart(part)
            except archive.ArchiveError:
                reqhandler.show_error("Error: Could not find part %s: %s" % (cgi.escape(hr(origpart), true), cgi.escape(hr(description), true),), httpcode=404)
                return

    # NOTE: we could use python's mime-magic library instead to try and guess
    # the filetype from the first few bytes (below when reading tinybit)
    mtype, encoding = mimetypes.guess_type(part)
    if mtype is None :
        mtype = confman.dict.get("DEFAULT_MIME_TYPE")
    debug.mojolog.write('part %s mimetype %s encoding %s\n', args=(part, mtype, encoding), v=4, vs='MojoHTTPD')

    # read a tiny bit of data before printing the response header
    # to verify that we can actually read this archive part without
    # an exception.
    debuglog('Reading tinybit.\n', v=5)
    try:
        # don't try and read too large of a "tiny bit" (but it is nice
        # if tinybit is enough to give the first few lines of an
        # index.html file)
        tinybit_size = min(min(partentry.csize, partentry.size), 300)
        tinybit = arc.readcomp(tinybit_size)
    except MojoErrors.ReassemblyError:
        reqhandler.show_error('Error: failed to find enough blocks to reassemble part.\n', httpcode=500)
        return
    except archive.ArchiveError, description:
        reqhandler.show_error("Error: publication format error: %s." % cgi.escape(hr(description), true), httpcode=500)
        return
    except:
        reqhandler.show_error('Error: Exception starting extraction of part "%s":\n' % part, httpcode=500)
        traceback.print_exc(file=debug.mojolog)
        return

    # if we got here, we were able to start reading the archive
    reqhandler.expiresin = 60*60*24*365  # dinode based data never changes, let the browser cache it well
    reqhandler.send_response(200)
    # Send headers:
    content_len_str = str(partentry.size)
    if content_len_str[-1] == 'L' :
        content_len_str = content_len_str[:-1]
    reqhandler.send_header('content-length', content_len_str)
    if mtype:
        reqhandler.send_header('content-type', mtype)
    if encoding != None :
        reqhandler.send_header('content-encoding', encoding)
    reqhandler.end_headers()

    if reqhandler.no_message_body:
        return  # HEAD request, we're not allowed to send actual data

    # Send the actual data
    try:
        reqhandler.wfile.write(tinybit)
    except IOError:
        # ignore this, client closed connection, we're done
        debuglog('Tried to write to client which had a closed connection.\n', v=3)
        archivefile.close()
        return

    debuglog('Extracting part to http client.\n', v=5)
    try:
        arc.extract(reqhandler.wfile)
        debuglog('Done writing part to http client.\n', v=3)
    except MojoErrors.ReassemblyError:
        debuglog('Error getting blocks to reassemble part.\n')
        traceback.print_exc(file=debug.mojolog)
    except IOError, description:
        debuglog('WARNING: IOError (client closed connection?) extracting part "%s": %s\n', args=(part, description.args), v=0)
    except:
        debuglog('ERROR: Exception extracting part "%s":\n', args=(part,), v=0)
        traceback.print_exc(file=debug.mojolog)

    archivefile.close()
    return

global _idlock
_idlock = id_lock.Locker()

def save_dinode_to_disk(reqhandler, servicemtm, dinode, basepath=None, displayname='', on_activity_list=1):
    """
    @param reqhandler is an http request handler object or something faking a similar interface
    @param servicemtm is the broker.ServicesMTM
    @param dinode is the dinode of the archive to download (prefferably in ascii zlib compressed form? or does it matter?)
    @param basepath is the path to create a subdirectory in to extract files into.
    @param displayname is the first part of the subdirectory name under basepath to extract to.  It will be
        sanitized not to include any special characters.

    Starts downloading the publication described in the dinode, extracting all files into a
    subdirectory under basepath using displayname as the first part of the subdirectory name
    and the mojoid as the second part.

    Sends a http rewrite response to point the browser to the download status graph if all
    is well, otherwise it calls reqhandler.show_error with the appropriate error message.

    Exports the following via reqhandler.variables for use by FakeRequestHandlers:
      mojoid   - the mojoid for this dinode
      watcher  - the UserCommand.Download activity list object for this download
      download_monitor - the DownloadMonitor instance for this download

    @returns true if everything succeeded, None otherwise.
    """
    mojoid = contentlib.get_mojoid_from_dinode(dinode)
    global _idlock
    _idlock.lock_and_do(mojoid, _save_dinode_to_disk, args=(reqhandler, servicemtm, dinode,), kwargs={'basepath': basepath, 'displayname': displayname, 'on_activity_list': on_activity_list})

def _save_dinode_to_disk(reqhandler, servicemtm, dinode, basepath=None, displayname='', on_activity_list=1):
    if not basepath:
        basepath = confman.getpath( ("PATH", "SAVE_TO_DISK_DIR") )
    mojoid = contentlib.get_mojoid_from_dinode(dinode)
    if displayname:
        # sanitize displayname for safety and to keep it from being an absolute path
        lst = []
        for char in tuple(displayname):
            if string.lower(char) in 'abcdefghijklmnopqrstuvwxyz0123456789-+=_()[]/:\\ ':
                if char in ('\\', '/', ':'):
                    char = '_'
                lst.append(char)
        displayname = string.join(lst, '')
        extractpath = os.path.join(basepath, displayname+'='+mojoid)
    else:
        extractpath = os.path.join(basepath, mojoid)
    debug.mojolog.write('Downloading archive contents to %s\n', args=(extractpath,), v=3, vs='MojoHTTPd')

    # create a watcher to put on the activity list if desired
    watcher = UserCommand.Download(globaldict['broker'].ServicesMTM, displayname, mojoid)

    try:
        archivefile = unsplit.ReassemblerSplitFile(TM=servicemtm, dinode=dinode, download_monitor_class=BitmapDownloadMonitor, pieces_to_start_readahead_after=max(1, int(round(NUM_FINISHED_REASSEMBLIES*0.30))), pieces_to_readahead=max(1, int(round(NUM_FINISHED_REASSEMBLIES*0.65))), progressfunc=watcher.set_progress)
    except sharemap.SharemapError, le:
        debug.mojolog.write('caught exception.  le: %s\n', args=(le,), v=3, vs='MojoHTTPd')
        reqhandler.show_error("the map to the file was garbled", httpcode=500)
        return None
    except (IOError, archive.ArchiveError, MojoErrors.ReassemblyError), description:
        reqhandler.show_error("Could not download file.  Some of the block servers that have the pieces of this file may be off-line and may come back on-line later, so please try again later.  (Error message: %s)" % cgi.escape(hr(description), true), httpcode=500)
        download_monitor = unsplit.get_download_monitor(mojoid)
        if download_monitor:
            download_monitor.increment_error_count(partname="*** sharemap ***")
            download_monitor.record_stoptime()
        return None

    # store the path we're extracting to for use in the status display
    download_monitor = unsplit.get_download_monitor(mojoid)
    if download_monitor:
        download_monitor.save_to_disk_path = extractpath
        if displayname:
            download_monitor.name = displayname
        else:
            displayname = mojoid

    if on_activity_list and download_monitor:
        commandmanager = globaldict['commandmanager']
        cmdnum = commandmanager.add_command(watcher)
        commandmanager.make_command_active(watcher)

    # these are set to give the reqhandler access to them (useful when using a FakeRequestHandler)
    reqhandler.variables.set('mojoid', mojoid)
    reqhandler.variables.set('watcher', watcher)
    reqhandler.variables.set('download_monitor', download_monitor)

    # send a HTTP redirect to the download_status_graph for this mojoid
    reqhandler.send_rewrite_response(reqhandler.variables.get('baseurl', '') + '/download_status/id/' + mojoid + '/')
    reqhandler.wfile.flush()

    # prepare to read from the archive (the reader will load its index)
    try:
        arc = archive.Reader(archivefile)
    except MojoErrors.ReassemblyError, description:
        # XXX this reqhandler.show_error call should fail since we already sent the rewrite response above!
        reqhandler.show_error("Error: could not rebuild publication index: %s." % cgi.escape(hr(description), true), httpcode=500)
        archivefile.close()
        watcher.set_finished('error')
        if download_monitor:
            download_monitor.increment_error_count(partname="*** archive index ***")
            download_monitor.record_stoptime()
        return None
    except archive.ArchiveError, description:
        # XXX this reqhandler.show_error call should fail since we already sent the rewrite response above!
        reqhandler.show_error("Error: bad publication: %s." % cgi.escape(hr(description)), httpcode=500)
        archivefile.close()
        watcher.set_finished('error')
        if download_monitor:
            download_monitor.increment_error_count(partname="*** archive index ***")
            download_monitor.record_stoptime()
        return None

    # read a tiny bit of data before printing the response header
    # to verify that we can actually read this archive part without
    # an exception.
    debuglog('Extracting..\n', v=5)
    try:
        # extract the entire archive to the destination directory
        arc.extract_all(extractpath, allowexceptions=true, exceptionstoignore=(MojoErrors.ReassemblyError, IOError), download_monitor=download_monitor)
    except MojoErrors.ReassemblyError, description:
        # XXX archive.extract_all will never raise this error
        debug.mojolog.write('Error downloading %s: failed to find enough blocks to reassemble one of the archive pieces/chunks:\n%s' % (mojoid, description.args,), vs='MojoHTTPd')
        watcher.set_finished('error')
        if download_monitor:
            download_monitor.increment_error_count()
            download_monitor.record_stoptime()
        return None
    except archive.ArchiveError, description:
        # XXX archive.extract_all will never raise this error
        debug.mojolog.write('Error downloading %s: publication format error: %s.\n' % (mojoid, description.args,), vs='MojoHTTPd')
        watcher.set_finished('error')
        if download_monitor:
            download_monitor.increment_error_count()
            download_monitor.record_stoptime()
        return None
    except:
        debug.mojolog.write('Error downloading %s: Unexpected exception during extraction:\n' % (mojoid,), vs='MojoHTTPd')
        watcher.set_finished('error')
        if download_monitor:
            download_monitor.increment_error_count()
            download_monitor.record_stoptime()
        traceback.print_exc(file=debug.mojolog)
        return None

    watcher.set_finished()
    debuglog('DONE Extracting..\n', v=5)
    archivefile.close()
    if download_monitor:
        download_monitor.record_stoptime()
    return 1   # success


    return

