#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# Implementation of the functions necessary for creating and parsing
# sharemaps and dinodes.
#
__cvsid = '$Id: sharemap.py,v 1.20 2001/05/15 20:30:08 greg Exp $'

# standard modules
import os
import string
from cStringIO import StringIO
import re
import zlib

# our modules
import MojoXML
from human_readable import hr
import idlib
import mojoutil
import mojosixbit
from VersionNumber import VersionNumber

# Available in PyXML at http://www.python.org/sigs/xml-sig/
from xml.sax import saxlib, saxexts
import xml.sax


# Change this version number for new inode XML formats (keeping support
# for any older version that has been deployed)
old_XML_Dinode_Version =   VersionNumber("0.9.0")
old_XML_Sharemap_Version = VersionNumber("0.9.1")
XML_Dinode_Version =   VersionNumber("0.10.0")
XML_Sharemap_Version = VersionNumber("0.10.0")
XML_SuperSharemap_Version = VersionNumber("0.10.0")

MAX_RAW_DINODE_LEN = 64 * 1024
MAX_RAW_SHAREMAP_LEN = 4 * 1024 * 1024


class SharemapError(MojoXML.XMLParseError) : pass
class BadSharemap(SharemapError) : pass
class ParseError(BadSharemap) : pass


class __OurParserInterfaceDocHandler(saxlib.HandlerBase) :
    """
    This provides our parser interface methods.  All derived classes
    MUST call setup_chardata_stuff() in their constructor.
    """
    def setup_chardata_stuff(self) :
        self.__char_stack = [None]
        self.__char_f = None

    ####### parser interface: #######
    ####################################################################
    # This method is called by the XML parser on all non-xml-tag
    # character data.  All character data at the end of any tag will
    # be available in self.chardata.
    ####################################################################
    def characters(self, ch, start, length) :
        self.__char_f.write(ch[start:start+length])

    ####################################################################
    # This is called by the parser when parsing errors occur
    ####################################################################
    def fatalError(self, exception) :
        diags = [exception]
        def diagem(io, name="io", diags=diags):
            diags.append(name)
            diags.append(io.tell())
            io.seek(0)
            diags.append(io.read())
          
        if self.__char_f:
            diagem(self.__char_f, "char_f")
        for sio in self.__char_stack:
            if sio:
                diags.append(sio.read())
                diagem(sio)

        raise ParseError, tuple(diags)

    error = fatalError
    warning = fatalError

    ####################################################################
    # This is called by the parser on properly matched start elements
    ####################################################################
    def startElement(self, name, attrs) :
        #D#print 'start:', name
        # raise an exception if we encounter an unknown element
        self.chardata = None
        self.__char_stack.append( StringIO() )
        self.__char_f = self.__char_stack[-1]
        if hasattr(self, 'start_'+name) :
            method = getattr(self, 'start_'+name)
            method(attrs)
        else:
            raise KeyError, 'unknown element '+name

    ####################################################################
    # This is called by the parser on properly matched end elements
    ####################################################################
    def endElement(self, name) :
        #D#print 'end:', name
        self.chardata = self.__char_f.getvalue()
        del self.__char_stack[-1]
        self.__char_f = self.__char_stack[-1]
        # ignore unsupported end elements
        if hasattr(self, 'end_'+name) :
            method = getattr(self, 'end_'+name)
            method()
    ####### end parser interface #######



class EncryptionKeySeedElementHandler:
    """
    To be used as a mix-in via inheritance.  This element handler
    provides extraction of a single encryption key into the
    encryptiontype and encryptionkeyseed variables.  Values of None should
    indicate to the application to use the defaults.
    """
    def __init__(self):
        self.encryptiontype = None
        self.encryptionkeyseed  = None

    def start_EncryptionKeySeed(self, attrs) :
        if self.encryptiontype:
            raise SharemapError, "only one EncryptionKeySeed element is allowed"
        cipher = attrs.get('cipher', '3DES')
        if cipher != '3DES':
            raise SharemapError, "unknown cipher"
        self.encryptiontype = cipher

    def end_EncryptionKeySeed(self):
        try:
            ekey = mojosixbit.a2b(self.chardata)
        except mojosixbit.Error:
            raise SharemapError, "invalid ascii encoded data in EncryptionKeySeed"
        # XXX put a check for key sanity here
        self.encryptionkeyseed = ekey


class pieceElementHandler:
    """
    A MixIn class for handling <piece><share></piece> style XML as
    used in sharemaps and dinodes for listing pieces.
    """
    def __init__(self):
        self.piecelist = []

    # Record this piece.  pieces must be listed in order and have
    # consistent length and offset parameters.
    def start_piece(self, attrs) :
        if not idlib.is_ascii_id(attrs['id']) :
            raise BadSharemap, "invalid pieceid"
        length = int(attrs['length'])
        offset = long(attrs['offset'])
        # verify correctness of the length and offset parameters
        if length <= 0 :
            raise BadSharemap, "piece length must be positive"
        if len(self.piecelist) > 0 :
            prev = self.piecelist[len(self.piecelist)-1]
            this_offset = prev['offset'] + prev['length']
            if this_offset != offset :
                raise BadSharemap, "inconsistent sharemap offset"
        else:
            if offset != 0 :
                raise BadSharemap, "first piece offset was not zero"
        
        self.sharenums_seen = {}
        self.piece = {
            'pieceid': str(attrs['id']),  # str() makes sure its not unicode
            'length' : length,
            'offset' : offset,
            'shares' : {},
        }

    def start_share(self, attrs) :
        if not idlib.is_sloppy_id(attrs['blobid']) :
            raise BadSharemap, "invalid blobid"
        sharenum = int(attrs['num'])
        if self.sharenums_seen.has_key(sharenum) :
            raise BadSharemap, "multiply defined share"
        self.sharenums_seen[sharenum] = sharenum

        self.piece['shares'][sharenum] = str(attrs['blobid'])   # str() makes sure its not unicode

    # end of piece, add it to the piecelist
    def end_piece(self) :
        del self.sharenums_seen
        self.piecelist.append( self.piece )
        self.piece = None



class DinodeDocHandler(__OurParserInterfaceDocHandler, EncryptionKeySeedElementHandler, pieceElementHandler) :
    """
    All dinode data will be in the self.dinode dict after parsing is complete.
    valid dict keys:
      smapid, length, shares, totalshares, sharesneeded,
      encryptiontype, encryptionkeyseed
    """

    ####################################################################
    def __init__(self) :
        EncryptionKeySeedElementHandler.__init__(self)
        pieceElementHandler.__init__(self)
        self.setup_chardata_stuff()
        self.version = None       # Dinode version string

        # this will be accessed publicly after parsing a dinode
        self.numtotalshares = 8   # use the defaults if not specified
        self.numsharesneeded = 4  # ""

        # this will be accessed publicly after parsing a dinode
        self.dinode = {}          # dict keyed on 'smapid', 'length', 'shares', 'totalshares', 'sharesneeded'

        self.__seen_elements = {}

    ####################################################################
    # The following methods are element handlers for the XML parser
    # called by the startElement and endElement methods above.
    ####################################################################
    def start_Dinode(self, attrs) :
        self.version = VersionNumber(attrs['version'])
        if self.version > XML_Dinode_Version :
            raise SharemapError, "Unsupported Dinode version"
        smapid = attrs.get('smapid', None)  # XXX leave the None argument here for compatibility with a bug in PyXML 0.5.3 -greg 2001-01-08
        if smapid and not idlib.is_mojosixbitencoded_id(smapid) :
            raise SharemapError, "invalid Dinode smapid attribute"
        try:
            length = int(attrs.get('length', 0))
            if length < 0 :
                raise ValueError
        except ValueError:
            raise SharemapError, "invalid Dinode length "+str(length)

        self.dinode = {
            'smapid' : str(smapid),  # str() makes sure its not unicode
            'length' : length,
            'shares' : {},
        }
    
    def end_Dinode(self) :
        self.dinode['totalshares'] = self.numtotalshares
        self.dinode['sharesneeded'] = self.numsharesneeded
        self.dinode['encryptiontype'] = self.encryptiontype
        self.dinode['encryptionkeyseed'] = self.encryptionkeyseed
        if self.piecelist is None:
            self.piecelist = []
    
    def start_Share(self, attrs) :
        if not idlib.is_sloppy_id(attrs['blobid']) :
            raise SharemapError, ("invalid blobid", attrs['blobid'])
        blobid = str(attrs['blobid'])  # str() makes sure its not unicode
        sharenum = int(attrs['num'])
        if self.dinode['shares'].has_key(sharenum) :
            raise SharemapError, "multiply defined Share "+len(sharenum)
        self.dinode['shares'][sharenum] = blobid
    
    def start_TotalShares(self, attrs) :
        if self.__seen_elements.has_key('TotalShares') :
            raise SharemapError, "only one TotalShares element allowed in a Dinode"
        self.__seen_elements['TotalShares'] = None
    
    def end_TotalShares(self) :
        num = int(self.chardata)
        if num <= 0 :
            raise SharemapError, "invalid value within TotalShares element"
        self.numtotalshares = num

    def start_SharesNeeded(self, attrs) :
        if self.__seen_elements.has_key('SharesNeeded') :
            raise SharemapError, "only one SharesNeeded element allowed in a Dinode"
        self.__seen_elements['SharesNeeded'] = None

    def end_SharesNeeded(self) :
        num = int(self.chardata)
        if num <= 0 :
            raise SharemapError, "invalid value within SharesNeeded element"
        self.numsharesneeded = num
    
    def start_pieceList(self, attrs) :
        if hasattr(self, '_inpieceList'):
            raise SharemapError, "only one pieceList element is allowed"
        self._inpieceList = 1
    
    def end_pieceList(self):
        self._inpieceList = 0
    
    def start_piece(self, attrs):
        if not hasattr(self, '_inpieceList') or not self._inpieceList:
            raise SharemapError, "pieces must occur within pieceList"
        pieceElementHandler.start_piece(self, attrs)



class SharemapDocHandler(__OurParserInterfaceDocHandler, EncryptionKeySeedElementHandler, pieceElementHandler) :
    """
    Document handler for Sharemap XML.  After parsing, the following
    member variables will be useful:
      version, piecelist, encryptiontype, encryptionkeyseed
    """
    def __init__(self) :
        EncryptionKeySeedElementHandler.__init__(self)
        pieceElementHandler.__init__(self)
        self.setup_chardata_stuff()
        self.version = None  # Sharemap version
        self.piece = None    # (piece, len, offset) currently being processed

        # this will be accessed publicly after parsing a sharemap
        self.piecelist = [] # list of dicts keyed on: pieceid, len, & offset

    ####################################################################
    # The following methods are element handlers for the XML parser
    # called by the startElement and endElement methods.
    ####################################################################
    def start_Sharemap(self, attrs) :
        self.version = VersionNumber(attrs['version'])
        if self.version > XML_Sharemap_Version :
            raise BadSharemap, "Unsupported Sharemap version"
    
    def start_piece(self, attrs) :
        if not hasattr(self, 'version'):
            raise BadSharemap, "piece may only occur within a Sharemap element"
        pieceElementHandler.start_piece(self, attrs)


class SuperSharemapDocHandler(__OurParserInterfaceDocHandler, EncryptionKeySeedElementHandler, pieceElementHandler) :
    """
    Document handler for SuperSharemap XML.  After parsing, the following
    member variables will be useful:
      version, piecelist, encryptiontype, encryptionkeyseed
    """
    def __init__(self) :
        EncryptionKeySeedElementHandler.__init__(self)
        pieceElementHandler.__init__(self)
        self.setup_chardata_stuff()
        self.version = None  # Sharemap version
        self.piece = None    # (piece, len, offset) currently being processed

        # this will be accessed publicly after parsing a sharemap
        self.piecelist = [] # list of dicts keyed on: pieceid, len, & offset

    ####################################################################
    # The following methods are element handlers for the XML parser
    # called by the startElement and endElement methods.
    ####################################################################
    def start_SuperSharemap(self, attrs) :
        self.version = VersionNumber(attrs['version'])
        if self.version > XML_SuperSharemap_Version :
            raise BadSharemap, "Unsupported SuperSharemap version"

    def start_piece(self, attrs) :
        if not hasattr(self, 'version'):
            raise BadSharemap, "piece may only occur within a SuperSharemap element"
        pieceElementHandler.start_piece(self, attrs)


def is_sharemap_piece(piecedict):
    """
    @returns boolean indicating if the piece described is part of the sharemap or the actual data.
    """
    assert piecedict.has_key('pieceid'), "piecedict must be a piece dictionary as returned by parse_dinode or parse_sharemap"
    return not piecedict.has_key('offset')

#########################################################################
# Parse the data returned in a Dinode into a dictionary representation.
#
# Raises a sharemap.SharemapError exception on errors.
#########################################################################
def parse_dinode(dinode) :
    """
    dinode - A string that can either be:
    A) a zlib compressed XML string, or
    B) an XML string.
    
    Returns a dict keyed on:
      piecelist
      encryptiontype
      encryptionkeyseed
      totalshares - total number of shares this piece was split into
      sharesneeded - number of those shares needed in order to reassemble

    piecelist is a list of dicts keyed on:
      pieceid - the id (SHA-1 hash) of this piece
      length - the length of this piece
      offset - if present: the offset within the entire publication that this piece makes up
               if not present: this piece is one of the sharemap pieces.
               [see the is_sharemap_piece() function]
      shares - a dict keyed on sharenums mapping to blockids
    """
    # attempt to decompress it if needed, otherwise assume it is not compressed
    dinodefile = StringIO()
    mojoutil.safe_zlib_decompress_to_file(dinode, maxlen=MAX_RAW_DINODE_LEN, fileobj=dinodefile)
    dinodefile.seek(0)

    dinodeHandler = DinodeDocHandler()
    dinodeParser = MojoXML.create_mojo_xml_parser(dinodeHandler)
    try:
        dinodeParser.parseFile(dinodefile)
    finally:
        dinodeParser.close()

    retval = {
        'encryptiontype': dinodeHandler.encryptiontype,
        'encryptionkeyseed': dinodeHandler.encryptionkeyseed,
        'totalshares': dinodeHandler.dinode['totalshares'],
        'sharesneeded': dinodeHandler.dinode['sharesneeded'],
        'piecelist': [],
    }
    # put the sharemap piece description first in the list
    dd = dinodeHandler.dinode
    if dd['smapid']:
        retval['piecelist'].append( {'pieceid': dd['smapid'], 'length': dd['length'], 'shares': dd['shares']} )
    # append any info on the first pieces of the archive to the list
    retval['piecelist'].extend( dinodeHandler.piecelist )

    return retval

def parse_sharemap(sharemap) :
    """
    Parse the data returned in a sharemap piece.
    Returns a dict keyed on:
      piecelist
      encryptiontype
      encryptionkeyseed

    piecelist is a list of dicts keyed on:
      pieceid - the id (SHA-1 hash) of this piece
      length - the length of this piece
      offset - the offset within the entire publication that this piece makes up
      shares - a dict keyed on sharenums mapping to blockids

    Raises a sharemap.BadSharemap exception on errors.
    """
    # attempt to decompress it if needed, otherwise assume it is not compressed
    sharemap = mojoutil.safe_zlib_decompress_to_retval(sharemap, maxlen=MAX_RAW_SHAREMAP_LEN, exception=BadSharemap)

    sharemapHandler = SharemapDocHandler()
    xml.sax.parseString(sharemap, sharemapHandler)

    return {'piecelist': sharemapHandler.piecelist,
            'encryptiontype': sharemapHandler.encryptiontype,
            'encryptionkeyseed': sharemapHandler.encryptionkeyseed}


#########################################################################
# Return a tuple of the entry containing (piecedict, piecelistindex)
# and its index within piecelist.  The entry corresponds to the given
# position within the file described in piecelist.
#
# piecelist is assumed to be contiguous as returned by parse_sharemap.
# (ie: it starts at offset 0 and there are no gaps between pieces)
#
# piecedict is keyed on 'pieceid', 'length', 'offset' and 'shares'
#########################################################################
def searchpiecelist(piecelist, pos) :
    prev = piecelist[0]
    previdx = 0

    for entry in piecelist[1:] :
        # is this too far? then it must be in the previous piece
        if pos < entry['offset'] :
            break
        prev = entry
        previdx = previdx + 1

    return (prev, previdx)


def make_sharemap(piecesharelist, compress=1, encryptionkeyseed=None) :
    """
    Create a sharemap, which is a compressed XML thingie telling you how to construct a
    file. The sharemap contains the pieceids for the file as well as the shareid, sharenum, and
    length for each pieceid.

    @param piecesharelist a list of dicts, where each dict has keys 'pieceid', 'length',
        'offset', and 'shares';  'shares' is a dict keyed on sharenum integers mapping to
        shareids for the given pieceid.
    @param encryptionkeyseed is a random number which should be 160 bits long;  It is used to
        generate the symmetric keys and initialization vectors which encrypt each share.
    """
    sharemapstrs = ['<?xml version="1.0"?>']
    if encryptionkeyseed:
        vers = XML_Sharemap_Version
    else:
        vers = old_XML_Sharemap_Version
    sharemapstrs.append('<Sharemap version="%s">' % vers)

    if encryptionkeyseed:
        sharemapstrs.append('<EncryptionKeySeed>%s</EncryptionKeySeed>' % mojosixbit.b2a(encryptionkeyseed))

    for piece in piecesharelist :
        sharemapstrs.append( '<piece id="%s" length="%d" offset="%d">' % (piece['pieceid'], piece['length'], piece['offset']) )
        for sharenum, shareid in piece['shares'].items() :
            if not idlib.is_sloppy_id(shareid) :
                raise SharemapError, 'invalid share blobid: '+shareid
            sharemapstrs.append( '<share blobid="%s" num="%d"/>' % (shareid, sharenum) )
        sharemapstrs.append( '</piece>' )

    sharemapstrs.append( '</Sharemap>' )

    if compress :
        tmpstr = zlib.compress(string.join(sharemapstrs, ''), 9)
        # prevent the compressed string from ending with a null
        # character because unsplit.py strips all trailing nulls when
        # reading this from a Mojo Nation (TM) piece.  (zlib ignores extra
        # data when decompressing)
        if tmpstr[-1] == '\0' :
            tmpstr = tmpstr + '\001'
        return tmpstr
    else :
        return string.join(sharemapstrs, '')


def make_super_sharemap(piecesharelist, compress=1, encryptionkeyseed=None) :
    """
    Create a super sharemap, which is a compressed XML thingie telling you how to construct a
    full sharemap. The super sharemap contains the pieceids for the sharemap as well as the shareid, sharenum, and
    length for each pieceid.

    @param piecesharelist a list of dicts, where each dict has keys 'pieceid', 'length',
        'offset', and 'shares';  'shares' is a dict keyed on sharenum integers mapping to
        shareids for the given pieceid.
    @param encryptionkeyseed is a random number which should be 160 bits long;  It is used to
        generate the symmetric keys and initialization vectors which encrypt each share.
    """
    sharemapstrs = ['<?xml version="1.0"?>']
    vers = XML_SuperSharemap_Version

    sharemapstrs.append('<SuperSharemap version="%s">' % vers)

    if encryptionkeyseed:
        sharemapstrs.append('<EncryptionKeySeed>%s</EncryptionKeySeed>' % mojosixbit.b2a(encryptionkeyseed))

    for piece in piecesharelist :
        sharemapstrs.append( '<piece id="%s" length="%d" offset="%d">' % (piece['pieceid'], piece['length'], piece['offset']) )
        for sharenum, shareid in piece['shares'].items() :
            if not idlib.is_sloppy_id(shareid) :
                raise SharemapError, 'invalid share blobid: '+shareid
            sharemapstrs.append( '<share blobid="%s" num="%d"/>' % (shareid, sharenum) )
        sharemapstrs.append( '</piece>' )

    sharemapstrs.append( '</SuperSharemap>' )

    if compress :
        tmpstr = zlib.compress(string.join(sharemapstrs, ''), 9)
        # prevent the compressed string from ending with a null
        # character because unsplit.py strips all trailing nulls when
        # reading this from a Mojo Nation (TM) piece.  (zlib ignores extra
        # data when decompressing)
        if tmpstr[-1] == '\0' :
            tmpstr = tmpstr + '\001'
        return tmpstr
    else :
        return string.join(sharemapstrs, '')


#########################################################################
def make_dinode(totalshares, sharesneeded, smapid='', smaplength=0, smapsharedict={}, piecelist=None, compress=1, encryptionkeyseed=None) :
    """
    @param smapid - the piece id of the reassembled Sharemap
    @param smapsharedict - a dict mapping sharenums to share blobids for the Sharemap
    @param encryptionkeyseed is a random number which should be 160 bits long;  It is used to
        generate the symmetric keys and initialization vectors which encrypt each share.

    @returns a string containing an XML Dinode.
    """
    dinode = '<?xml version="1.0"?>'  # XXX not needed
    if encryptionkeyseed:
        vers = XML_Dinode_Version
    else:
        vers = old_XML_Dinode_Version

    # only include sharemap id and length if sharemap info was given
    if smapsharedict:
        dinode = dinode + ('<Dinode version="%s" smapid="%s" length="%d">' % (vers, smapid, smaplength))
    else:
        dinode = dinode + ('<Dinode version="%s">' % (vers,))

    if encryptionkeyseed:
        dinode = dinode + ('<EncryptionKeySeed>%s</EncryptionKeySeed>' % mojosixbit.b2a(encryptionkeyseed))

    # this only includes sharemap shares if some were given
    for sharenum, blobid in smapsharedict.items() :
        dinode = dinode + '<Share blobid="%s" num="%d"/>' % (idlib.to_ascii(blobid), sharenum)

    dinode = dinode + ('<TotalShares>%d</TotalShares>' % totalshares)
    dinode = dinode + ('<SharesNeeded>%d</SharesNeeded>' % sharesneeded)
    if piecelist:
        dinode = dinode + '<pieceList>'
        for piece in piecelist:
            dinode = dinode + ('<piece id="%s" length="%d" offset="%d">' % (idlib.to_ascii(piece['pieceid']), piece['length'], piece['offset']))
            for sharenum, shareid in piece['shares'].items() :
                dinode = dinode + ('<share blobid="%s" num="%d"/>' % (idlib.to_ascii(shareid), sharenum))
            dinode = dinode + '</piece>'
        dinode = dinode + '</pieceList>'

    dinode = dinode + '</Dinode>'

    if compress :
        return zlib.compress(dinode, 9)
    else :
        return dinode



def test_Dinode_creation_and_parsing() :
    sdict = {
        0 : 'jAAAAAAAAAAAspamAAAAAAAAAAk',
        1 : 'WzAAAAAspamAAAAAABAAAAAAAAE',
        2 : 'qAAzAAAAAAAAAAAAAAAspamAAAI',
        3 : 'sihlakhghlahgSpamlakghlahgA',
        4 : 'ialhglakhglkahh_alkjgh-aklI',
        5 : '__AzAAA_AAA_dogfoodeggsAAAI',
        6 : 'SpamEggsSpamBeanSpamAndSpAA',
        7 : 'GrossStuffGoesInHereOhBoy_0',
    }
    # a VERY fake dinode (unreasonable totalshares values, etc..)
    dinode = make_dinode(smapid='AAAAAAAAAAAAAAAAAAAAAAAAAAA', smaplength=10000, smapsharedict=sdict, totalshares=84, sharesneeded=42, compress=0)

    pdict = parse_dinode(dinode)
    assert pdict['totalshares'] == 84
    assert pdict['sharesneeded'] == 42
    assert pdict['piecelist'][0]['pieceid'] == 'A' * 27
    t = pdict['piecelist'][0]['shares'].keys()
    t.sort()
    s = sdict.keys()
    s.sort()
    assert t == s

def test_Dinode_with_no_sharemap_and_pieces():
    piece_of_spam = {
        'pieceid': 'AAAAAAAAAAAAAAAAAAAAAAAAAAk',
        'length': 12345,
        'offset': 0,
        'shares': {
            0 : 'spamspamspamspamspamspamspA',
            3 : 'mSpamsPaMspaMspamityspamspA',
            7 : 'MmmSspamspamspamspam__spam0',
            2 : 'spamspamspamspamspamYspam-8',
        }
    }
    dinode = make_dinode(totalshares=8, sharesneeded=4, piecelist=[piece_of_spam], compress=0)
    pdict = parse_dinode(dinode)
    assert pdict['piecelist'][0] == piece_of_spam, 'parsed dinode returned different data'

def test_dinode_encryption_keys():
    sdict = {
        0 : 'jAAAAAAAAAAAspamAAAAAAAAAAk',
        1 : 'WzAAAAAspamAAAAAABAAAAAAAAE',
        2 : 'qAAzAAAAAAAAAAAAAAAspamAAAI',
        3 : 'sihlakhghlahgSpamlakghlahgA',
        4 : 'ialhglakhglkahh_alkjgh-aklI',
        5 : '__AzAAA_AAA_dogfoodeggsAAAI',
        6 : 'SpamEggsSpamBeanSpamAndSpAA',
        7 : 'GrossStuffGoesInHereOhBoy_0',
    }
    # a VERY fake dinode (unreasonable totalshares values, etc..)
    dinode = make_dinode(smapid='AAAAAAAAAAAAAAAAAAAAAAAAAAA', smaplength=10000, smapsharedict=sdict, totalshares=84, sharesneeded=42, compress=0)

    pdict = parse_dinode(dinode)
    assert not pdict['encryptiontype']
    assert not pdict['encryptionkeyseed']

    dinode = make_dinode(smapid='AAAAAAAAAAAAAAAAAAAAAAAAAAA', smaplength=10000, smapsharedict=sdict, totalshares=84, sharesneeded=42, encryptionkeyseed='Pioneer')

    pdict = parse_dinode(dinode)
    assert pdict['encryptiontype'] == '3DES'
    assert pdict['encryptionkeyseed'] == 'Pioneer'
    

def test_Sharemap_creation_and_parsing() :
    """Create a simple sharemap and try parsing it"""
    import copy
    piece_of_spam = {
        'pieceid': 'AAAAAAAAAAAAAAAAAAAAAAAAAAk',
        'length': 12345,
        'offset': 0,
        'shares': {
            0 : 'spamspamspamspamspamspamspA',
            3 : 'mSpamsPaMspaMspamityspamspA',
            7 : 'MmmSspamspamspamspam__spam0',
            2 : 'spamspamspamspamspamYspam-8',
        }
    }
    piece_of_spam_keys = piece_of_spam.keys()
    piece_of_spam_keys.sort()
    piece_of_spam2 = copy.copy(piece_of_spam)
    piece_of_spam2['pieceid'] = 'BBBBBBBBBBBBBBBBBBBBBBBBBBk'
    piece_of_spam2['offset'] = piece_of_spam['length']

    warm_spam = make_sharemap([piece_of_spam, piece_of_spam2])
    cold_spam = make_sharemap([piece_of_spam, piece_of_spam2], compress=0)

    breakfast_platter = parse_sharemap(warm_spam) 
    assert not breakfast_platter['encryptiontype']
    assert not breakfast_platter['encryptionkeyseed']

    breakfast = breakfast_platter['piecelist']
    lunch = parse_sharemap(cold_spam)['piecelist']

    assert len(breakfast) == 2
    assert len(lunch) == 2

    breakfast_keys = breakfast[0].keys()
    lunch_keys = lunch[0].keys()
    breakfast_keys.sort()
    lunch_keys.sort()
    assert breakfast_keys == piece_of_spam_keys
    assert lunch_keys == piece_of_spam_keys


def test_Sharemap_encryption_key_support():
    piece_of_spam = {
        'pieceid': 'AAAAAAAAAAAAAAAAAAAAAAAAAAk',
        'length': 12345,
        'offset': 0,
        'shares': {
            0 : 'spamspamspamspamspamspamspA',
            3 : 'mSpamsPaMspaMspamityspamspA',
            7 : 'MmmSspamspamspamspam__spam0',
            2 : 'spamspamspamspamspamYspam-8',
        }
    }
    piece_of_spam_keys = piece_of_spam.keys()
    piece_of_spam_keys.sort()

    scrambled_spam = make_sharemap([piece_of_spam], encryptionkeyseed='beans')
    scrambled = parse_sharemap(scrambled_spam)
    assert scrambled['encryptiontype'] == '3DES'
    assert scrambled['encryptionkeyseed'] == 'beans'



mojo_test_flag = 1

def run() :
    import RunTests
    RunTests.runTests(["sharemap"])

if __name__ == '__main__' :
    run()
