#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# This module provides a read only object that implements archive lookup,
# retrieval and reassembly transparently through open/close/read/seek
# file-ish interface.
#
# The ReassemblerSplitFile object is the useful class.  It should be
# used to open and read dinodes.  It will read their sharemap and
# deal with retrieving the piece data for parts of the file as needed
# by interfacing to the block share fetcher / reassembler to collect
# block shares and assemble them into the original data.
# (the FindBeforehandBlockFetcher does this)
#
# ReadOnlySplitFile objects support files on the local disk.  This
# object was created solely for testing.
#
__cvsid = '$Id: unsplit.py,v 1.59 2001/05/16 07:12:59 zooko Exp $'

# standard python libraries
import copy
import sys
import string
import tempfile
import threading
import traceback
import types

# custom modules
import Buf
from MojoConstants import true, false, DEBUG_MODE
from confutils import confman
import debug
from human_readable import hr
import mojoutil
import sharemap
from FindBeforehandBlockFetcher import FindBeforehandBlockFetcher
from LastAccessedCache import LastAccessedCache
from contentlib import get_mojoid_from_dinode

import DoQ # this contains the global reference to the DoQ  --Zooko 2001-04-14

# This specifies the maximum read-ahead that ReassemblerSplitFile will
# perform when requesting information from the reassembler.
# Fine tuning would be useful...
# (Note: 32k worked "well" with the OldSchoolBlockFetcher -greg)
# a lower value here is nice for more streaming-like performance and
# partial loading of partially downloaded things.
DefaultReadBufferSize = 16384


theGlobalFindBeforehandBlockFetcher = None

# this global cache is shared across all instances of ReassemblerSplitFile
# objects it maps mojoid -> {...} which is used to save/update internal state to
# prevent refetching and parsing dinodes+sharemaps many times in a row
_MojoIDCache = LastAccessedCache(10)

# tune this after we figure out how big something needs to be
MAX_SHAREMAP_SIZE = 512*1024 - 1


# This is derived from IOError as unsplit is a file-like object so its
# exceptions should always at least be caught by IOError handlers.
class UnsplitError(IOError): pass


def get_download_monitor(mojoid=None, dinode=None):
    """
    @returns a DownloadMonitor instance for the given mojoid or dinode if one exists or None
    """
    assert mojoid or dinode
    if dinode:
        mojoid = get_mojoid_from_dinode(dinode)
    ddict = _MojoIDCache.get(mojoid)
    if ddict:
        return ddict.get('download_monitor')
    return None


#########################################################################
# This splitfile interface reads the dinode for the publication and
# talks to the block fetching and reassembling code.  It provides a
# limited file-like interface for files stored as splitup piece shares
# (blocks) scattered around the world.  The caller never needs to know
# that its data was split up and scattered around the universe much
# like being passed through a star trek transporter..
#########################################################################
class ReassemblerSplitFile :
    """
    A file-like object used to hide the fact that the file is
    really split up into smitherenes all over the net.  It needs a
    MojoTransactionManager object to use for talking to block
    servers to get the appropriate data from.
    """
    # these allow access via COM
    _public_methods_ = ['close', 'seek', 'tell', 'get_optimal_read_length', 'read']
    _public_attrs_ = []

    def __init__(self, TM, dinode, buffersize=DefaultReadBufferSize, download_monitor_class=None,
            pieces_to_start_readahead_after=None, pieces_to_readahead=None, progressfunc=None, freeonclose=1 ) :
        """
        @param download_monitor_class can be a DownloadMonitor class to instantiate for this dinode if none already exists
        @param progressfunc can be a function that takes a value and optional briefstatus parameter that would be useful
            to display to the user to show progress in downloading a publication.
        @param freeonclose is a boolean indicating whether or not we should explicitly call the block fetcher's free()
            method on any known pieces when we are closed (via close() or our destructor).  Set this to false if you
            expect there to be multiple ReassemblerSplitFile instances accessing the same dinode.
        """
        self.pieces_to_start_readahead_after = pieces_to_start_readahead_after
        self.pieces_to_readahead = pieces_to_readahead
        self.progressfunc = progressfunc
        self.freeonclose = freeonclose

        # tune readahead based on our connection speed setting if no values were passed in
        # definitions:
        #  1) every time we advance past a piece divisible by this number we call
        #     start_readahead on the next self.pieces_to_readahead pieces.
        #  2) every time the above condition is met as well as when seeking to a
        #     new location within the file we call start_readahead on this many
        #     following pieces.
        __speed = int(confman.dict['CONN_SPEED_IN'])
        # XXX note that this is our publication max piece/chunk size, published pieces/chunks
        # may be very different but we don't yet have enough information to tell what size this
        # publication uses.  This is known after parsing the dinode so we could adjust the readahead
        # values there.
        # TODO tune self.*_readahead_* values based on the publications piece/chunk size to restrict memory usage.
        __MaxPieceSize = int(confman['PUBLICATION']['MAX_SHARE_SIZE']) * int(confman['PUBLICATION']['SHARES_NEEDED'])
        if __speed > 500:
            # every time we advance past a piece divisible by this number we call
            # start_readahead on the next self.pieces_to_readahead pieces.
            if not self.pieces_to_start_readahead_after:
                self.pieces_to_start_readahead_after = 7
            # every time the above condition is met as well as when seeking to a
            # new location within the file we call start_readahead on this many
            # following pieces.
            if not self.pieces_to_readahead:
                self.pieces_to_readahead = 12
        elif __speed > 180:
            # medium speed line
            if not self.pieces_to_start_readahead_after:
                self.pieces_to_start_readahead_after = 4
            if not self.pieces_to_readahead:
                self.pieces_to_readahead = 7
        elif __speed >= 112:
            # dual channel isdn, better than a modem and all over Europe (damn those telco monopolies)
            if not self.pieces_to_start_readahead_after:
                self.pieces_to_start_readahead_after = 3
            if not self.pieces_to_readahead:
                self.pieces_to_readahead = 5
        else:
            # better on a modem, don't try and readahead so much since it takes forever
            if not self.pieces_to_start_readahead_after:
                self.pieces_to_start_readahead_after = 2
            if not self.pieces_to_readahead:
                self.pieces_to_readahead = 3

        assert self.pieces_to_readahead >= self.pieces_to_start_readahead_after, "bad start readahead value"

        name_of_bf_strategy = confman.get('BLOCK_FETCHING_STRATEGY_v2', "FindBeforehandBlockFetcher")

        if name_of_bf_strategy == 'FindBeforehandBlockFetcher':
            self._blockfetcher = theGlobalFindBeforehandBlockFetcher
        else:
            try:
                stratClass = eval(name_of_bf_strategy)
                self._blockfetcher = stratClass(TM)
            except:
                # Maybe it was a "module.class" string...
                dotidx = string.find(name_of_bf_strategy, ".")
                if dotidx != -1:
                    modname = name_of_bf_strategy[:dotidx]
                    classname = name_of_bf_strategy[dotidx+1:]
                else:
                    # Or it is inside a module of the same name...
                    modname = name_of_bf_strategy
                    classname = name_of_bf_strategy 

                try:
                    m = __import__(modname)
                    stratClass = m.__dict__.get(classname)
                    self._blockfetcher = stratClass(TM)
                except:
                    (t,v, tb) = sys.exc_info()
                    debug.mojolog.write("Got exception instantiating block fetching strategy: %s\n", args=(v,), v=2, vs="debug")

                    self._blockfetcher = theGlobalFindBeforehandBlockFetcher

        assert self._blockfetcher, "A valid blockfetcher was not set."

        if isinstance(self._blockfetcher, FindBeforehandBlockFetcher):
            assert self._blockfetcher == theGlobalFindBeforehandBlockFetcher, "don't instantiate new ones of these in each unsplit"

        debug.mojolog.write("Using `%s' strategy to fetch blocks in order to reassemble the file.\n", args=(self._blockfetcher.__class__.__name__,), v=3, vs="unsplit")

        self.TM = TM
        # An internal read buffer and the absolute file location it
        # starts from.
        self.__maxbuf = buffersize
        self.__buf = ''
        self.__bufpos = 0
        self.closed = 0

        self.pieceidx = None # the index in the piecelist of which piece we're sucking on right now;  starts at `None', to indicate that we haven't started sucking on any piece, not even the 0-indexed piece.
        self.pos_in_piece = None # the index in the piece of which byte we want to read next.  `self.pos_in_piece' should be `None' iff `self.pieceidx' is `None'.
        self.pos = None # the index in the file of which byte we want to read next.  `self.pos' should be `None' iff `self.pieceidx' is `None'.

        self.download_monitor_class = download_monitor_class

        self.open(dinode=dinode)
        self.seek(0)

    #####################################################################
    #####################################################################
    def open(self, dinode) :
        self.mojoid = get_mojoid_from_dinode(dinode)
        preparsed_dict = _MojoIDCache.get(self.mojoid)
        if preparsed_dict is not None:
            dinodedict = preparsed_dict['dinodedict']
            self.piecelist = preparsed_dict['piecelist']
            self.download_monitor = preparsed_dict['download_monitor']
            self.num_blocks_needed = dinodedict['sharesneeded']
            self.total_block_shares = dinodedict['totalshares']
            self._preparsed_dict = preparsed_dict
        else:
            try:
                dinodedict = sharemap.parse_dinode(dinode)
            except sharemap.SharemapError, le:
                # debug.mojolog.write("ZZZ: le: %s, mojoutil.safe_zlib_decompress_to_retval(dinode): %s\n" % (hr(le), mojoutil.safe_zlib_decompress_to_retval(dinode))) # very very verbose -- can be 100 KB of ascii text.
                raise sharemap.SharemapError, (mojoutil.safe_zlib_decompress_to_retval(dinode), le,)

            self.num_blocks_needed = dinodedict['sharesneeded']
            self.total_block_shares = dinodedict['totalshares']

            if self.download_monitor_class:
                self.download_monitor = self.download_monitor_class(name='put mojoid here', piecelist=dinodedict['piecelist'], num_shares_needed=self.num_blocks_needed)
            else:
                self.download_monitor = None

            # load the sharemap

            # start finding and prefetching the stuff needed for the sharemap (and the first few pieces of the archive)
            if hasattr(self._blockfetcher, 'seed_finder'):
                self._blockfetcher.seed_finder(dinodedict['piecelist'])
            self.start_readahead(dinodedict['piecelist'][:self.pieces_to_readahead], priority=2)  # XXX fix for multi/super sharemap

            # if there is a sharemap for this publication, load it
            if not dinodedict['piecelist'][0].has_key('offset'):
                # TODO add multi/super sharemap support here
                sharemap1data = self._get_piece_data(piece=dinodedict['piecelist'][0])  # XXX fix for multi/super sharemap
                # strip off trailing NULL padding that comes with a generic piece retrieval
                sharemap1data = sharemap1data[:dinodedict['piecelist'][0]['length']]  # XXX fix for multi/super sharemap
                if len(sharemap1data) >= MAX_SHAREMAP_SIZE :
                    raise SharemapError, "Sharemap too large, its at least " + MAX_SHAREMAP_SIZE + " bytes!"

                try:
                    parsed_sharemap_dict = sharemap.parse_sharemap(sharemap1data)
                except sharemap.SharemapError, le:
                    # debug.mojolog.write("ZZZ: le: %s, mojoutil.safe_zlib_decompress_to_retval(sharemap1data): %s\n" % (hr(le), mojoutil.safe_zlib_decompress_to_retval(sharemap1data))) # very very verbose -- can be 100 KB of ascii text.
                    raise sharemap.SharemapError, (mojoutil.safe_zlib_decompress_to_retval(sharemap1data), le,)
                self.piecelist = parsed_sharemap_dict['piecelist']
                if self.download_monitor:
                    self.download_monitor.extend_piecelist(self.piecelist)
            else:
                self.piecelist = copy.copy(dinodedict['piecelist'])
                # remove all sharemap pieces from our saved copy of piecelist
                while not self.piecelist[0].has_key('offset'):
                    self.piecelist.pop(0)
            self.update_progress()

            # Tell the fetcher everything we might possibly want to read
            if hasattr(self._blockfetcher, 'seed_finder'):
                self._blockfetcher.seed_finder(self.piecelist)
            
            # save this parsed data for future generations

            cachedict = {
                'dinodedict': dinodedict,
                'piecelist': self.piecelist,
                'download_monitor': self.download_monitor,
            }
            _MojoIDCache[self.mojoid] = cachedict
        # ... end if dinode+sharemap already parsed ...

        # make sure the fetcher has started working on the first several pieces
        self.start_readahead(self.piecelist[:self.pieces_to_readahead], priority=0)

    #####################################################################
    #####################################################################
    def update_progress(self):
        if callable(self.progressfunc):
            # if we've gotten the entire piecelist, report the current
            # file position as the percentage complete.
            if self.closed:
                self.progressfunc('publication closed', briefstatus='done')
            elif hasattr(self, 'piecelist') and self.piecelist and self.piecelist[-1].has_key('offset') and self.pos is not None:
                totallen = self.piecelist[-1]['offset'] + self.piecelist[-1]['length']
                if self.pos == totallen:
                    briefstatus = 'done'
                else:
                    briefstatus = 'running [%d%%]' % (int(100*float(self.pos)/totallen))
                self.progressfunc('read %d of %d bytes' % (self.pos, totallen),
                            briefstatus=briefstatus )
            else:
                self.progressfunc('retrieving the sharemap', briefstatus='finding')

    #####################################################################
    #####################################################################
    def start_readahead(self, piecelist, priority):
        self._blockfetcher.start_readahead(piecelist, priority=priority, num_blocks_needed=self.num_blocks_needed, total_block_shares=self.total_block_shares, download_monitor=self.download_monitor)

    #####################################################################
    #####################################################################
    def _get_piece_data(self, piece, pos_in_piece=0, bytes=0):
        """
        @param pos_in_piece the index within the piece from which to start reading

        @precondition `piece' must be the normal "piece" dict.: piece.has_key('pieceid'): "piece: %s" % hr(piece)
        """
        assert piece.has_key('pieceid'), "precondition: `piece' must be the normal \"piece\" dict." + " -- " + "piece: %s" % hr(piece)

        piecedata = self._blockfetcher.get_piece(piece, num_blocks_needed=self.num_blocks_needed, total_block_shares=self.total_block_shares)
        assert piecedata, "piecedata: %s" % hr(piecedata)

        self.update_progress()
        if not bytes :
            bytes = len(piecedata)
        pos_in_piece = int(pos_in_piece)  # make sure they're ints, not longs
        bytes = int(bytes)                # "
        if pos_in_piece == 0 and bytes == len(piecedata) :
            return str(piecedata)
        else :
            return piecedata[pos_in_piece:pos_in_piece+bytes]

    #####################################################################
    #####################################################################
    def close(self) :
        self.closed = 1
        self.update_progress()
        # TODO get rid of the freeonclose thing and do reference counting on the number
        # of instances currently accessing a given mojoid
        if self.freeonclose and hasattr(self, 'piecelist'):
            for piecedict in self.piecelist:
                self._blockfetcher.free(piecedict)

    #####################################################################
    #####################################################################
    def __del__(self):
        if self.freeonclose and hasattr(self, 'piecelist'):
            for piecedict in self.piecelist:
                self._blockfetcher.free(piecedict)

    #####################################################################
    #####################################################################
    def get_optimal_read_length(self):
        """
        returns bytes to ask read for that will result in the least bulk data copies; returns 0 at EOF.

        The whole point of this function to -avoid- using any of our internal
        buffering at all when possible to make things much faster by not doing
        data copies from underlying singlereassembly objects.
        """
        # offset within our internal buffer of any data ready to be read
        offset = self.pos - self.__bufpos
        if offset < len(self.__buf):
            # There's no getting around a slice data copy in the next read as we're mid-buffer.
            # Just let it read to the end of our current buffer so that future reads can
            # be done optimally.
            debug.mojolog.write("optimal read returning remaining buffer length %s\n", args=(len(self.__buf) - offset,), v=10, vs='ReassemblerSplitFile')
            return len(self.__buf) - offset
        elif (offset == 0) and len(self.__buf) > 0:
            debug.mojolog.write("optimal read returning full buffer length %s\n", args=(len(self.__buf),), v=10, vs='ReassemblerSplitFile')
            return len(self.__buf)
        else:
            # Our internal buffer is empty.
            assert offset >= len(self.__buf)
            # this is the best case (and is what two or more calls and reads in a row using this function will lead to)
            if self.pos_in_piece >= self.piece['length']:
                assert self.pieceidx is not None
                if self.pieceidx+1 >= len(self.piecelist):
                    self.wait_for_piecelist()
                if self.pieceidx+1 < len(self.piecelist):
                    current_piece_length = self.piecelist[self.pieceidx+1]['length']
                else:
                    # this was the last piece/end of file, nothing more to read
                    debug.mojolog.write("optimal read returning 0\n", v=4, vs='ReassemblerSplitFile')
                    return 0
            else:
                current_piece_length = self.piece['length'] - self.pos_in_piece
            debug.mojolog.write("optimal read returning piece length %s\n", args=(current_piece_length,), v=10, vs='ReassemblerSplitFile')
            return current_piece_length 

    #####################################################################
    # NOTE: Unbounded reads are not supported; size must be positive.
    #####################################################################
    def read(self, size) :
        if size <= 0 :
            raise UnsplitError, "infinite read not supported in splitfile (%s specified)" % size
        size = int(size)
   
        data = ''

        assert (self.pos is None) == (self.pos_in_piece is None) == (self.pieceidx is None), "These three should all be `None' or else none of them should be `None'. -- self.pos: %s, self.pos_in_piece: %s, self.pieceidx: %s" % (hr(self.pos), hr(self.pos_in_piece), hr(self.pieceidx))

        # If we are already sucking on a piece, there might be data in our internal buffer.
        if self.pos is not None:
            # offset within our internal buffer of any data we want
            offset = self.pos - self.__bufpos

            # If any of the data we want is in our buffer, use it
            if offset >= 0 and offset < len(self.__buf) :
                usefulbytes = len(self.__buf) - offset

                # All of the data is in our buffer, return it
                if size <= usefulbytes :
                    self.pos_in_piece = self.pos_in_piece + size
                    self.pos = self.pos + size
                    return self.__buf[offset:offset+size]

                data = self.__buf[offset:offset+usefulbytes]
                self.pos_in_piece = self.pos_in_piece + usefulbytes
                self.pos = self.pos + usefulbytes
        
        # We need more data
        sizeleft = size - len(data)

        # Advance to the next piece in the archive if needed
        if self.pos_in_piece >= self.piece['length'] :
            try:
                self.piece = self.piecelist[self.pieceidx+1]
            except IndexError:
                # this was the last piece/end of file, we're done
                return data

            self.pieceidx = self.pieceidx + 1
            # Tell the fetcher that we are probably going to read the next few blocks next.
            if (self.pieceidx % self.pieces_to_start_readahead_after) == 0:
                self.start_readahead(self.piecelist[self.pieceidx:self.pieceidx+self.pieces_to_readahead], priority=1)
            # Tell the fetcher that we are done with the previous piece.
            debug.mojolog.write("Done with this piece because we just read past it. self.pieceidx-1: %s, piece: %s\n", args=(self.pieceidx-1, self.piecelist[self.pieceidx-1],), v=10, vs="debug")
            self._blockfetcher.free(self.piecelist[self.pieceidx - 1])
            self.pos_in_piece = 0
        
        # Only fill our buffer up to piece boundaries if they're
        # smaller than our buffer size.
        if self.__maxbuf > (self.piece['length'] - self.pos_in_piece) :
            tobuffer = self.piece['length'] - self.pos_in_piece
        else :
            tobuffer = self.__maxbuf
        
        # If this read would overflow our next buffer, don't buffer it
        if sizeleft > tobuffer :
            output = data + self.__unbuffered_read(sizeleft)
            return output
        else :
            # fill the new buffer
            self.__buf = self._get_piece_data(piece=self.piece, pos_in_piece=self.pos_in_piece, bytes=tobuffer)
            self.__bufpos = self.pos
        
        self.pos = self.pos + sizeleft
        self.pos_in_piece = self.pos_in_piece + sizeleft
        output = data + self.__buf[:sizeleft]
        return output


    #####################################################################
    # An internal unbuffered read implementation.  Reads -exactly- size
    # bytes.  (this was the original read() until I saw the abysmal
    # performance when doing small reads -greg; however it is best when
    # reading up to a piece boundary as get_optimal_read_length would
    # have you do)
    #####################################################################
    def __unbuffered_read(self, size) :
        if size <= 0 :
            raise UnsplitError, "infinite read not supported in splitfile"

        assert (self.pos is None) == (self.pos_in_piece is None) == (self.pieceidx is None), "These three should all be `None' or else none of them should be `None'. -- self.pos: %s, self.pos_in_piece: %s, self.pieceidx: %s" % (hr(self.pos), hr(self.pos_in_piece), hr(self.pieceidx))

        sizeleft = size     # bytes left in order to read size bytes
        readdata = []

        # loop to read data from successive pieces
        while sizeleft > 0 :
            # don't try to reassemble past the end of the current piece
            bytesinpiece = self.piece['length'] - self.pos_in_piece
            if sizeleft > bytesinpiece :
                toget = bytesinpiece
            else :
                toget = sizeleft

            data = self._get_piece_data(piece=self.piece, pos_in_piece=self.pos_in_piece, bytes=toget)

            # This should only happen if the piece was not the expected
            # size or if the server screws up.
            if len(data) < toget :
                raise UnsplitError, "Only read %d; expecting %d" % (len(data), toget)
            assert len(data) == toget

            readdata.append(data)
            del data   # help garbage collection

            self.pos_in_piece = self.pos_in_piece + toget
            self.pos = self.pos + toget
            sizeleft = sizeleft - toget

            # Advance to the next piece in the archive if needed
            assert self.pos_in_piece <= self.piece['length']
            if self.pos_in_piece >= self.piece['length'] :
                try:
                    self.piece = self.piecelist[self.pieceidx+1]
                except IndexError:
                    # this was the last piece/end of file, we're done
                    debug.mojolog.write("attempt to read past last piece.", v=4, vs='ReassemblerSplitFile')
                    break
                self.pieceidx = self.pieceidx + 1
                # Tell the fetcher that we are probably going to read the next few blocks next.
                # (only tell it this once every several blocks for efficiency)
                # Note: we already issue all possible 'do you have blobs' messages to find block location for an entire archive up front as soon as we get the sharemap (we've done it that way since soon after the FBFH download code was integrated)  --greg 2001-04-23.
                if (self.pieceidx % self.pieces_to_start_readahead_after) == 0:
                    self.start_readahead(self.piecelist[self.pieceidx:self.pieceidx+self.pieces_to_readahead], priority=1)
                # Tell the fetcher that we are done with the previous piece.
                debug.mojolog.write("Done with this piece because we just read past it unbufferedly. self.pieceidx-1: %s, piece: %s\n", args=(self.pieceidx-1, self.piecelist[self.pieceidx-1],), v=10, vs="debug")
                self._blockfetcher.free(self.piecelist[self.pieceidx - 1])
                self.pos_in_piece = 0

        return string.join(readdata, '')

    #####################################################################
    # Only forward seeks need be allowed as this "file" will only be
    # used by a sequential archive reader extracting a single file.
    #####################################################################
    def seek(self, pos, whence=0) :
        debug.mojolog.write("seek(%s, %s, %s)\n", args=(self.mojoid, pos, whence,), v=10, vs="debug")
        if pos < 0 :
            raise ValueError, "pos must be positive"
        if type(pos) != type(42) :
            _pos = int(pos)
            if _pos != pos :
                raise ValueErrror, "pos to large"
            pos = _pos

        if whence == 0 :    # absolute seek
            newpos = pos
        elif whence == 1 :  # relative seek
            newpos = self.pos + pos
        else :
            raise ValueError, "invalid seek type"

        newpiece, newpieceidx = sharemap.searchpiecelist(self.piecelist, newpos)
        # Tell the fetcher that we are probably going to read the next few blocks next.
        self.start_readahead(self.piecelist[newpieceidx:newpieceidx+self.pieces_to_readahead], priority=1)
        # Tell the fetcher that we are done with the old piece.
        if (self.pieceidx is not None) and (self.pieceidx != newpieceidx):
            assert self.pieceidx < newpieceidx
            debug.mojolog.write("Done with this piece because we just seeked past it. self.pieceidx: %s, piece: %s\n", args=(self.pieceidx, self.piecelist[self.pieceidx],), v=10, vs="debug")
            self._blockfetcher.free(self.piecelist[self.pieceidx])
        newpos_in_piece = newpos - newpiece['offset']

        # this shouldn't happen unless an invalid sharemap snuck by or
        # if we try and position past the end of the archive.
        if newpos_in_piece > newpiece['length'] :
            raise UnsplitError, "position outside of piece!"

        self.piece = newpiece
        self.pieceidx = newpieceidx
        self.pos_in_piece = newpos_in_piece
        self.pos = newpos

    #####################################################################
    # Return our current location within the "file"
    #####################################################################
    def tell(self) :
        return self.pos



#########################################################################
# A read-only split file object designed for reading archives hiding
# where the parts physically reside from the user.
#########################################################################
class ReadOnlySplitFile :  # XXX test object
    #####################################################################
    #####################################################################
    def open(self, dinode) :
        self.dinode = dinode
        self.file = open(dinode, "rb")  # XXX test code

    #####################################################################
    #####################################################################
    def close(self) :
        self.file.close()  # XXX test code
        self.file = None

    #####################################################################
    #####################################################################
    def read(self, size=-1) :
        return self.file.read(size)  # XXX test code

    #####################################################################
    #####################################################################
    def seek(self, pos, whence=0) :
        return self.file.seek(pos, whence)  # XXX test code

    #####################################################################
    #####################################################################
    def tell(self) :
        return self.file.tell()  # XXX test code

    #####################################################################
    #####################################################################
    def __init__(self, dinode) :
        self.open(dinode)

    #####################################################################
    #####################################################################
    def __del__(self) :
        if not self.file is None :
            self.close()

