#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

import types
try:
    from human_readable import hr
except ImportError:
    hr = repr

true = 1
false = 0

def _real_assert_apply_will_work(func, args = [], kwargs = {}):
    """
    asserts if apply(func, args, kwargs) will fail based on parameters
    
    throws up it's hands and doesn't assert for native code

    XXX doesn't yet support functions which use *args and **kwargs
    """
    assert callable(func)
    assert type(args) == types.ListType or type(args) == types.TupleType
    assert type(kwargs) == types.DictType
    if hasattr(func, '__call__'):
        func = func.__call__
    if type(func) == types.BuiltinFunctionType or type(func) == types.BuiltinMethodType:
        return
    if type(func) == types.MethodType:
        args = [func.im_self] + list(args)
        func = func.im_func

    defaults = func.func_defaults
    if defaults is None:
        defaults = []
    varnames = func.func_code.co_varnames[:func.func_code.co_argcount]

    params = {}
    for i in varnames:
        params[i] = false
    assert len(args) <= len(varnames), 'bad args for func %s: args %s, kwargs %s' % (func, hr(args), hr(kwargs))
    for i in xrange(len(args)):
        params[varnames[i]] = true
    for key, value in kwargs.items():
        assert params.has_key(key), 'bad args for func %s: args %s, kwargs %s' % (func, hr(args), hr(kwargs))
        assert params.get(key) == false, 'bad args for func %s: args %s, kwargs %s' % (func, hr(args), hr(kwargs))
        params[key] = true
    for i in varnames[:len(varnames) - len(defaults)]:
        assert params[i], 'bad args for func %s: args %s, kwargs %s' % (func, hr(args), hr(kwargs))


if __debug__:   # if we're not in -O or -OO mode
    assert_apply_will_work = _real_assert_apply_will_work
else:
    assert_apply_will_work = lambda func, args=None, kwargs=None: None


def test_non_callable_fails():
    try:
        assert_apply_will_work(2)
        return false
    except AssertionError:
        return true

def test_builtin():
    assert_apply_will_work(apply)

def test_function_pass():
    assert_apply_will_work(lambda x, y, z = 3: None, args = [1], kwargs = {'y': 2})

def test_with_varname():
    def spam(x):
        y = 3
    assert_apply_will_work(spam, args = [2])

def test_funcion_pass2():
    def spam(x, y = 2, z = 3):
        pass
    assert_apply_will_work(spam, args = [4])

def test_too_many_params():
    try:
        assert_apply_will_work(lambda: None, args = [1])
        return false
    except AssertionError:
        return true

def test_param_unassigned():
    try:
        assert_apply_will_work(lambda x: None)
        return false
    except AssertionError:
        return true

def test_unexpected_parameter():
    try:
        assert_apply_will_work(lambda: None, kwargs = {'a': 2})
        return false
    except AssertionError:
        return true

def test_param_redefined():
    try:
        assert_apply_will_work(lambda x: None, args = [3], kwargs = {'x': 2})
        return false
    except AssertionError:
        return true

def test_callable_object_pass():
    class spam:
        def __call__(self):
            pass
    assert_apply_will_work(spam())

def test_callable_object_fail():
    class spam:
        def __call__():
            pass
    try:
        assert_apply_will_work(spam(), args = [2])
        return false
    except AssertionError:
        return true

def test_method_pass():
    class spam:
        def eggs(self, x, y = 1):
            pass
    assert_apply_will_work(spam().eggs, args = [0])

def test_method_fail():
    class spam:
        def eggs(self):
            pass
    try:
        assert_apply_will_work(spam().eggs, args = [2])
        return false
    except AssertionError:
        return true

mojo_test_flag = 1
if __name__ == '__main__':
    import RunTests
    RunTests.runTests('AssertApply')
