#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# Classes used to manage inbound and outbound network bandwidth (by
# making calls into the object around every metered send and receive).
#
# Unless otherwise stated, all time values in this module are
# represented in seconds.
#
# Copyright (c) 2000 - Autonomous Zone Industries
# Original Author:  Gregory P. Smith <greg@mad-scientist.com>
#
# TODO provide a side interface into the bandwidth manager allowing
# access to the current bandwidth usage state.  Also keep some simple
# statistics (such as the number of times we've hit the limit).
#
__cvsid = '$Id: Bandwidth.py,v 1.2 2000/09/05 07:31:59 zooko Exp $'

from human_readable import hr
from time import time, sleep
from sched import scheduler
import thread
from threading import Lock, Event, Thread
import debug


class BandwidthLowError(StandardError) : pass


class BandwidthManager:
    """BandwidthManager(maxinbytes=None, maxoutbytes=None, granularity=30)

    An object to be called surrounding send/recv calls on a socket
    to perform bandwidth allocation.  This object starts its own
    event scheduler thread when instantiated.

      allow_send(numbytes, block=1) - account for sending numbytes data.
      Blocking until we're ready/allowed to send the data if necessary.
      Call this before calling send().

      allow_recv(block=1) - Block until we're allowed to receive more
      data off the wire.  Call this before calling recv().

      account_recv(numbytes) - Account for data that we just received.
      Call this after calling recv().

    All 'block' arguments specify the blocking behavior.  If set to 0,
    a BandwidthLowError exception will be raised instead of blocking.
    """

    ####################################################################
    def __init__(self, maxinbytes=None, maxoutbytes=None, granularity=30):
        assert (maxinbytes is None) or (maxinbytes > 0)
        assert (maxoutbytes is None) or (maxoutbytes > 0)
        assert granularity > 1, "granularity way too small"
        # small granularities might interfere with TCP algorithms in a bad way.  -greg 17-may-2000
        if granularity <= 11 :
            debug.stderr.write('WARNING: low granularity used in bandwidth manager\n')

        self.granularity = granularity
        self.maxinbytes = maxinbytes
        self.maxoutbytes = maxoutbytes

        self.outbytes_lock = Lock()
        self.outbytes = 0
        self.outbytes_reset_time = time()
        self.inbytes_lock = Lock()
        self.inbytes = 0
        self.inbytes_reset_time = time()

        # start our event scheduler (used as an alarm clock to wake up
        # sleepers awaiting more bandwidth in our account_SPAM methods)
        self.sched = scheduler(time, sleep)
        def run_sched_loop(g2=self.granularity/2, s=self.sched) :
            while 1 :
                s.run()
                sleep(g2)
        self.thread = Thread(target=run_sched_loop, name="BandwidthManager Scheduler")
        # NO, NO, DON'T DO IT!
        self.thread.setDaemon(1)
        self.thread.start()
    
    ####################################################################
    def __del__(self) :
        # cause the thread to exit when we go away
        self.sched.enter(0, 9999999, thread.exit, ())

    ####################################################################
    def wait_until_abstime(self, atime) :
        """wait_until_abstime(atime) - Block until atime"""
        ev = Event()
        self.sched.enterabs(atime, 1, ev.set, ()) # XXX replace 1 w/ random()?
        ev.wait()

    ####################################################################
    def allow_send(self, block=1):
        """allow_send(block=1) - call this before calling send()"""
        if self.maxoutbytes is None :
            return
        self.outbytes_lock.acquire()
        try:
            # reset our counter if enough time has elapsed
            if time() - self.outbytes_reset_time >= self.granularity :
                if self.outbytes < self.maxoutbytes :
                    self.outbytes = 0
                else :
                    self.outbytes = self.outbytes - self.maxoutbytes
                self.outbytes_reset_time = time()

            # if we're over our limit, block or signal an error
            if self.outbytes >= self.maxoutbytes :
                if not block :
                    raise BandwidthLowError

                # block until we're back within range of our limit.
                reset_delta = self.granularity * (self.outbytes / self.maxoutbytes)
                self.wait_until_abstime(self.outbytes_reset_time + reset_delta)
                self.outbytes = self.outbytes % self.maxoutbytes
                self.outbytes_reset_time = time()

        finally:
            self.outbytes_lock.release()

    ####################################################################
    def account_send(self, numbytes):
        """account_send(numbytes) - call this after send() returns"""
        if self.maxoutbytes is None or numbytes <= 0 :
            return
        self.outbytes_lock.acquire()
        try:
            try:
                self.outbytes = self.outbytes + numbytes
            except OverflowError:
                # Wow, better switch to large numbers!
                self.outbytes = long(self.outbytes) + numbytes
        finally:
            self.outbytes_lock.release()
    

    ####################################################################
    def allow_recv(self, block=1):
        """allow_recv(block=1) - call this before calling recv()"""
        if self.maxinbytes is None :
            return
        self.inbytes_lock.acquire()
        try:
            # reset our counter if enough time has elapsed
            if time() - self.inbytes_reset_time >= self.granularity :
                if self.inbytes < self.maxinbytes :
                    self.inbytes = 0
                else :
                    self.inbytes = self.inbytes - self.maxinbytes
                self.inbytes_reset_time = time()

            # if we're over our limit, block or signal an error
            if self.inbytes >= self.maxinbytes :
                if not block :
                    raise BandwidthLowError

                # block until we're back within range of our limit.
                reset_delta = self.granularity * (self.inbytes / self.maxinbytes)
                self.wait_until_abstime(self.inbytes_reset_time + reset_delta)
                self.inbytes = self.inbytes % self.maxinbytes
                self.inbytes_reset_time = time()

        finally:
            self.inbytes_lock.release()

    ####################################################################
    def account_recv(self, numbytes):
        """account_recv(numbytes) - call this after recv() returns"""
        if self.maxinbytes is None or numbytes <= 0:
            return
        self.inbytes_lock.acquire()
        try:
            try:
                self.inbytes = self.inbytes + numbytes
            except OverflowError:
                # Wow, better switch to large numbers!
                self.inbytes = long(self.inbytes) + numbytes
        finally:
            self.inbytes_lock.release()


########################################################################
# System Tests
########################################################################

# be sure unrestricted bandwidth managers work
def test_01() :
    mybw = BandwidthManager()
    mybw.allow_send(block=0)
    mybw.account_send(1234567890)
    mybw.allow_recv(block=0)
    mybw.allow_recv(block=0)
    mybw.account_recv(1234567890)
    mybw.account_recv(1234567890)
    mybw.allow_recv(block=0)
    mybw.account_recv(1234567890)
    mybw.allow_send(block=0)
    mybw.account_send(1234567890)
    mybw.allow_send(block=0)
    mybw.account_send(1234567890)


def test_02() :
    mybw = BandwidthManager(maxinbytes=1000000, maxoutbytes=1000000, granularity=60)
    mybw.allow_send(block=0)
    mybw.account_send(1234567890)
    try:
        mybw.allow_send(block=0)
        assert "failed to raise BandwidthLowError in allow_send"
    except BandwidthLowError:
        pass

    mybw.allow_recv(block=0)
    mybw.account_recv(1234567890)
    try:
        mybw.allow_recv(block=0)
        assert "failed to raise BandwidthLowError in allow_recv"
    except BandwidthLowError:
        pass


def test_03() :
    mybw = BandwidthManager(maxinbytes=400, maxoutbytes=3000, granularity=60)

    mybw.allow_send(block=0)
    mybw.account_send(999)
    mybw.allow_send(block=0)
    mybw.account_send(1001)
    try:
        mybw.allow_send(block=0)
        assert "failed to raise BandwidthLowError in send"
    except BandwidthLowError:
        pass

    mybw.allow_recv(block=0)
    mybw.account_recv(250)
    mybw.allow_recv(block=0)
    mybw.account_recv(500)
    try:
        mybw.allow_recv(block=0)
        assert "failed to raise BandwidthLowError in recv"
    except BandwidthLowError:
        pass

def test_04() :
    mygran = 2.1
    mybw = BandwidthManager(maxoutbytes=10000, granularity=mygran)
    mybw.allow_send(block=0)
    mybw.account_send(25000)
    start_timer = time()
    mybw.allow_send()
    mybw.account_send(5001)
    end_timer1 = time()
    mybw.allow_send()
    end_timer2 = time()
    assert end_timer1 - start_timer >= (mygran * 1.9)
    assert end_timer1 - start_timer < (mygran * 3)
    assert end_timer2 - end_timer1 >= mygran
    assert end_timer2 - end_timer1 < (mygran * 2)


mojo_test_flag = 1



def run():
    import RunTests 
    RunTests.runTests(["Bandwidth"])

if __name__ == '__main__' :
    run()
