#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: BlockFinder.py,v 1.49 2001/06/07 23:00:11 greg Exp $'

### standard modules
import traceback

### our modules
import debug
import bisect
from human_readable import hr
import idmasklib
from idlib import make_ascii_id
import idlib
from HashQueue import HashQueue
from CounterHash import CounterHash
import RunTests
import DoQ
from LastAccessedCache import LastAccessedCache
import MetaTrackerLib
true = 1
false = 0

class BlockFinder:
    """
    A BlockFinder takes care of searching block servers to find
    which ones have blocks that we're interested in downloading.  It
    caches these results and intelligently manages the number of
    queries with various block servers all happening at once.

    The find_blocks_and_callback method is the primary interface.
    """
    def __init__(self, broker, find_function, max_queries_at_once=20, max_cached_results=100000):
        """
        @param broker - the Broker instance
        @param find_function - takes (server_id, block_ids, callback, preferencenumber)
        @param max_queries_at_once - maximum queries to have running at once
        @param max_cached_results - the number of blockid locations we'll keep cached in memory

        @precondition `find_function' must be callable.: callable(find_function): "find_function: %s" % hr(find_function)
        """
        assert callable(find_function), "precondition: `find_function' must be callable." + " -- " + "find_function: %s" % hr(find_function)

        self.broker = broker
        self.find_function = find_function
        self.data = BlockFinderData(max_queries_at_once, max_cached_results)
    
    def find_blocks(self, block_ids, allow_download_cache=1):
        """
        Add blockids to this blockfinder's internal list of ids to find.  Starts finding them immediately.
        """
        assert self.broker.doq.is_currently_doq()
        # debug.mojolog.write("find_blocks called; calling hold on block_ids %s\n", args=(block_ids,), v=4, vs='BlockFinder')
        block_ids = map(idlib.to_ascii, block_ids)
        blocks2 = []
        for i in block_ids:
            if not allow_download_cache or self.broker.downloadcache.get_block_size(i) == 0:
                blocks2.append(i)
            else:
                debug.mojolog.write("not searching for block %s -- it's in our download cache\n", args=(i,), v=5, vs='BlockFinder')
        block_ids = blocks2
        for i in block_ids:
            self.data.hold(i)
        def callback(outcome, block_ids = block_ids, self = self):
            # debug.mojolog.write("get_blob_servers returned %s server_infos; unholding blocks\n", args=(len(outcome),), v=3, vs="BlockFinder")
            map(self.data.unhold, block_ids)
            server_tuple_list = outcome.items()
            self.data.start_find(block_ids, server_tuple_list, handicapper=self.broker.ServicesMTM.get_handicapper())
            # we've already been on the DoQ thread a while here (start_find can take a while to process
            # stuff) so put the final call on the DoQ to let other things continue.
            self.broker.doq.add_batch_task('download', self._check)
        MetaTrackerLib.get_blob_servers_and_call_back(self.broker.ServicesMTM, callback, min = 120) # XXX un-hardcode min!

    def cancel(self, block_id):
        """
        Remove block_id from those being searched for.  This is useful when SingleReassembly objects get enough
        blocks and don't need the others as well as when cancelling a download.

        @precondition This method must be called from the DoQ.: self.broker.doq.is_currently_doq()
        """
        assert self.broker.doq.is_currently_doq(), "precondition: This method must be called from the DoQ."

        # debug.mojolog.write('cancel called on %s. traceback: %s\n', args=(block_id, traceback.extract_stack(),), v=4, vs='BlockFinder') # XXX traceback is expensive -- don't commi thtis.  --Zooko 2001-04-16
        self.data.cancel(idlib.to_ascii(block_id))
        self._check()

    def get_block_locations_from_cache(self, block_id):
        """
        Returns [(server id, server info)] for servers having this blockid
        """
        assert self.broker.doq.is_currently_doq()
        return self.data.get_cached(idlib.to_ascii(block_id))

    def callback_when_found(self, block_id, found_callback, finished_callback):
        """
        Calls callback asynchronously on the DoQ.
        
        callback takes block id and an optional failure reason.
        
        If there is anything in the cache at all, calls back immediately.
        
        Otherwise, if there is no current attempt to find the block_id 
        currently, assumes it failed and calls back immediately.
        
        Otherwise, calls back when either the first possible place comes in, 
        or the current attempt to find a location is completely given up.
        """
        assert self.broker.doq.is_currently_doq()
        block_id = idlib.to_ascii(block_id)
        self.data.add_callback(block_id, found_callback, finished_callback)
        self._check()

    def _query_callback(self, server_id, server_info, block_ids):
        assert self.broker.doq.is_currently_doq()
        self.data.data_came_in(server_id, server_info, block_ids)
        self._check()

    def _check(self):
        """
        loop through the finds that need to be issued and start them.  Also loop through any
        callbacks that need to happen and toss them onto the DoQ.
        """
        # note: do them in order of preference for -all- the servers with listed pending requests. -greg 2000-10-28
        # (this is pretty much taken care now of by sorting the list of servers before queueing up the requests)
        calls = []
        preferenceno=0
        for server_id, server_info, block_ids in self.data.get_request():
            calls.append({'server_id': server_id, 'server_info': server_info, 'block_ids' : block_ids, 'callback': self._query_callback, 'preferenceno': preferenceno})
            preferenceno=preferenceno+1
        if len(calls) > 0:
            # TODO should be converted to use doq.add_batch_task()
            FindBatcher(self.broker.doq, self.find_function, calls)
        for block_id, callbacks in self.data.get_found_callback():
            for i in callbacks:
                self.broker.doq.add_batch_task('download', i, kwargs = {'block_id': block_id})
        for block_id, callbacks in self.data.get_finished_callback():
            for i in callbacks:
                self.broker.doq.add_batch_task('download', i, kwargs = {'block_id': block_id})

class FindBatcher:
    """
    take a list of find_function calls and send them out in chunks, on the DoQ
    the list contains kwargs dicts of arguments for the BlockFinder's find function
    """
    def __init__(self, doq, find_function, l):
        self.doq = doq
        self.find_function = find_function
        self.list = l
        self.doq.add_task(self.do_some)
    def do_some(self):
        chunk = self.list[:5]
        self.list = self.list[5:]
        for kwargs in chunk:
            self.doq.add_task(self.find_function, kwargs = kwargs)
        if len(self.list) > 0:
            self.doq.add_task(self.do_some)

def test_standard_operation():
    class FakeBroker: pass
    broker = FakeBroker()
    broker.doq = DoQ.DoQ()
    find_log = []
    def find_function(server_id, block_ids, callback, find_log = find_log):
        find_log.append(server_id)
        find_log.append(block_ids)
        find_log.append(callback)
    bf = BlockFinder(broker, find_function)
    server_id = make_ascii_id('server id')
    bf.set_server_infos([(server_id, idmasklib.generate_random_mask(0))])
    block_id = make_ascii_id('block')
    broker.doq.add_task(bf.find_blocks, kwargs = {'block_ids': [block_id]})
    broker.doq.flush()
    assert find_log == [server_id, [block_id], bf._query_callback]

    callback_log = []
    def callback_function(block_id, callback_log = callback_log):
        callback_log.append(block_id)
    broker.doq.add_task(bf.callback_when_found, kwargs = {'block_id': block_id, 'callback': callback_function})
    broker.doq.flush()
    assert find_log == [server_id, [block_id], bf._query_callback]
    assert callback_log == []

    broker.doq.add_task(find_log[2], kwargs = {'server_id': server_id, 'block_ids': [block_id]})
    broker.doq.flush()
    broker.doq.flush()
    assert find_log == [server_id, [block_id], bf._query_callback]
    assert callback_log == [block_id]

    cache_result = []
    def cache_func(cache_result = cache_result, bf = bf, block_id = block_id):
        cache_result.append(bf.get_block_locations_from_cache(block_id))
    broker.doq.add_task(cache_func)
    broker.doq.flush()
    assert cache_result == [[server_id]]

def setdefault(thing, key, default):
    if not thing.has_key(key):
        thing[key] = default
        return default
    else:
        return thing[key]

class BlockFinderData:
    def __init__(self, max_downloads, max_cached):
        self.max_downloads = max_downloads
        # maps server id to list of blocks want to request
        self.waiting_requests = HashQueue()
        # maps server id to list of blocks last asked
        self.active_requests = {}
        # maps block id to number of pending requests for it
        self.block_counts = CounterHash()
        # maps block id to [callbacks]
        self.callbacks = {}
        # maps block id to server ids
        self.cache = LastAccessedCache(max_cached)
        # list of recently found blocks which might trigger callbacks
        self.found_blocks = []
        # blocks which are being processed externally
        self.held_blocks = CounterHash()
        self.hold = self.held_blocks.start_counting_and_increment
        self.unhold = self.held_blocks.decrement_count_and_delete_if_zero

    def cancel(self, block_id):
        if self.held_blocks.is_counting(block_id) or self.block_counts.get_count(block_id) > 0:
            self.held_blocks.delete(block_id)
            self.block_counts.set_count_to_zero_if_present(block_id)
            if self.cache.has_key(block_id):
                del self.cache[block_id]

    def start_find(self, block_ids, server_infos, handicapper=None):
        assert type(block_ids) == type([])
        assert type(server_infos) == type([])
        debug.mojolog.write("start_find of %s block_ids %s\n", args=(len(block_ids), block_ids,), v=3, vs='BlockFinder')
        # filter out blocks that we've already started finding or have info on
        block_ids = filter(lambda x, self=self: not self.cache.has_key(x), block_ids)
        debug.mojolog.write("after removing those with locations already in our cache: %s block_ids %s\n", args=(len(block_ids), block_ids,), v=3, vs='BlockFinder')
        #block_ids = filter(lambda x, self=self: self.block_counts.is_counting(x), block_ids)
        #debug.mojolog.write("after filtering out ones we're not counting: %s block_ids %s\n", args=(len(block_ids), block_ids,), v=3, vs='BlockFinder')
        if len(block_ids) <= 0:
            return
        # mark these block_ids as being processed.  This is necessary so that the block_finished callback will still happen
        # even if there were no suitable servers to ask about some of the block ids.
        map(self.block_counts.start_counting, block_ids)
        # prepare for binary searching using bisect
        # (as bram later noted; this needs to be done in binary format for proper sorting/searching)
        binary_block_ids = map(idlib.to_binary, block_ids)
        binary_block_ids.sort()
        if handicapper:
            server_infos = handicapper.sort_by_preference(server_infos, 'do you have blobs', [])
        # find which block_ids to ask each server about
        for server_id, info in server_infos:
            server_id = idlib.to_ascii(server_id)
            masklist = MetaTrackerLib.get_blobserver_mask_list(info)
            #debug.mojolog.write('server %s masklist: %s\n', args=(server_id, masklist,), v=9, vs='BlockFinder')
            block_ids_for_this_server = []
            for mask in masklist:
                startid, stopid = map(idlib.to_binary, mask.get_first_and_last_match())
                startpos = bisect.bisect(binary_block_ids, startid)
                if startpos > 0 and binary_block_ids[startpos-1] >= startid:
                    startpos = startpos - 1  # backup if it skipped an exact match
                stoppos = bisect.bisect(binary_block_ids, stopid)
                #debug.mojolog.write('startid %s, stopid %s, startpos %s, stoppos %s\n', args=(startid, stopid, startpos, stoppos), v=9, vs='BlockFinder')
                block_ids_for_this_server.extend( map(idlib.to_ascii, binary_block_ids[startpos:stoppos]) )
            # TODO remove duplicates from block_ids_for_this_server (or disallow overlapping mask ranges)
            map(self.block_counts.start_counting_and_increment, block_ids_for_this_server)
            if len(block_ids_for_this_server) > 0:
                debug.mojolog.write('asking server %s about %s blocks %s\n', args=(server_id, len(block_ids_for_this_server), block_ids_for_this_server), v=4, vs='BlockFinder')
                # XXX store block ids for each server in a hash as well as a list to prevent it from having
                # duplicates yet preserving order (a HashQueue perhaps?).
                setdefault(self.waiting_requests, server_id, (info, []))[1].extend(block_ids_for_this_server)

    def add_callback(self, block_id, found_callback, finished_callback):
        assert callable(found_callback)
        assert callable(finished_callback)
        block_id = idlib.to_ascii(block_id)
        self.block_counts.start_counting(block_id)
        setdefault(self.callbacks, block_id, []).append((found_callback, finished_callback))
        if self.cache.has_key(block_id):
            self.found_blocks.append(block_id)

    def data_came_in(self, server_id, server_info, block_ids):
        """
        failure should be indicated by block_ids having length 0
        """
        debug.mojolog.write("data_came_in %s %s\n", args=(server_id, block_ids), v=4, vs='BlockFinder')
        server_id = idlib.to_ascii(server_id)
        block_ids = map(idlib.to_ascii, block_ids)
        if not self.active_requests.has_key(server_id):
            # this should not happen currently, but is generally safe programming practice
            debug.mojolog.write("returning from data_came_in because %s is not active\n", args=(server_id,), v=3, vs='BlockFinder')
            return
        for i in block_ids:
            setdefault(self.cache, i, []).append((server_id, server_info))
            self.found_blocks.append(i)
        for i in self.active_requests.get(server_id):
            self.block_counts.decrement_count(i)
        del self.active_requests[server_id]

    def get_request(self):
        """
        returns [(server_id, [block_ids])]
        
        returns [] if just called
        """
        if len(self.active_requests) >= self.max_downloads:
            return []
        r = []
        server_id = self.waiting_requests.get_first_key()
        while server_id is not None:
            next_server_id = self.waiting_requests.after(server_id)
            # we have a policy of only having one request going to a particular server at once
            if not self.active_requests.has_key(server_id):
                info, block_ids = self.waiting_requests[server_id]
                del self.waiting_requests[server_id]
                debug.mojolog.write("adding server %s to active requests\n", args=(server_id,), v=4, vs='BlockFinder')
                self.active_requests[server_id] = block_ids
                r.append((server_id, info, block_ids))
            server_id = next_server_id
        return r

    def get_found_callback(self):
        """
        returns [(block id, [callback])]
        
        returns [] if just called
        """
        r = []
        while len(self.found_blocks) > 0:
            block_id = self.found_blocks.pop()
            block_id = idlib.to_ascii(block_id)
            result = []
            if not self.callbacks.has_key(block_id):
                continue
            for find_func, finished_func in self.callbacks[block_id]:
                result.append(find_func)
            # debug.mojolog.write("get_found_callback returning %s, %s\n", args=(block_id, result), v=20, vs='BlockFinder')
            r.append((block_id, result))
        return r

    def get_finished_callback(self):
        """
        returns [(block id, [callback])]

        returns [] if just called
        """
        r = []
        block_ids = self.block_counts.get_all_zero_counts()
        for block_id in block_ids:
            if self.held_blocks.is_counting(block_id):
                continue
            if not self.cache.has_key(block_id):
                # debug.mojolog.write("block %s could not be found or was cancelled\n", args=(block_id,), v=3, vs='BlockFinder')
                # it would make sense to put a neg here, but cancel() hoses the info anyway -
                # we aren't really using the cache as a cache, it should probably be replaced 
                # with something which contains a bunch of one-use objects
                pass
            self.block_counts.delete(block_id)
            if not self.callbacks.has_key(block_id):
                continue
            result = []
            for find_func, finished_func in self.callbacks[block_id]:
                result.append(finished_func)
            del self.callbacks[block_id]
            r.append((block_id, result))
        return r

    def get_cached(self, block_id):
        """
        returns [(server id, server info)]
        """
        # debug.mojolog.write("get_cached %s\n", args=(block_id,), v=20, vs='BlockFinder')
        return self.cache.get(idlib.to_ascii(block_id), [])

def test_returns_immediately_on_unstarted():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    callback = lambda: None
    ida = make_ascii_id('a')
    d.add_callback(ida, callback)
    assert d.get_request() is None
    assert d.get_finished_callback() == (ida, [callback])
    assert d.get_finished_callback() is None

def test_does_not_start_two_requests_to_same_server_at_once():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server id')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0))])
    assert d.get_request() == (server_id, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None

    idb = make_ascii_id('b')
    d.start_find([idb], [(server_id, idmasklib.generate_random_mask(0))])
    assert d.get_request() is None
    assert d.get_finished_callback() is None

    d.data_came_in(server_id, [])
    assert d.get_request() == (server_id, [idb])
    assert d.get_request() is None
    assert d.get_finished_callback() is None

def test_standard():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server id')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0))])
    assert d.get_request() == (server_id, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    callback = lambda: None
    d.add_callback(ida, callback)
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() == (ida, [callback])
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == [server_id]

def test_request_multiple_at_once():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server id')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0))])
    assert d.get_request() == (server_id, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    callback = lambda: None
    d.add_callback(ida, callback)
    callback2 = lambda: None
    d.add_callback(ida, callback2)
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id, [ida])
    assert d.get_request() is None
    dg = d.get_finished_callback()
    assert dg == (ida, [callback, callback2]) or dg == (ida, [callback2, callback])
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == [server_id]

def test_succeed_on_second_server():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server 1')
    server_id2 = make_ascii_id('server 2')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0)),(server_id2, idmasklib.generate_random_mask(0))])
    dr = d.get_request()
    assert dr == (server_id, [ida]) or dr == (server_id2, [ida])
    dr2 = d.get_request()
    assert dr2 == (server_id, [ida]) or dr2 == (server_id2, [ida])
    assert dr != dr2
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    callback = lambda: None
    d.add_callback(ida, callback)
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id, [])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id2, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() == (ida, [callback])
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == [server_id2]

def test_finish_on_second_server_fail():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server 1')
    server_id2 = make_ascii_id('server 2')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0)),(server_id2, idmasklib.generate_random_mask(0))])
    dr = d.get_request()
    assert dr == (server_id, [ida]) or dr == (server_id2, [ida])
    dr = d.get_request()
    assert dr == (server_id, [ida]) or dr == (server_id2, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    callback = lambda: None
    d.add_callback(ida, callback)
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id, [])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id2, [])
    assert d.get_request() is None
    assert d.get_finished_callback() == (ida, [callback])
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

def test_callback_after_first_find():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server 1')
    server_id2 = make_ascii_id('server 2')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0)),(server_id2, idmasklib.generate_random_mask(0))])
    dr = d.get_request()
    assert dr == (server_id, [ida]) or dr == (server_id2, [ida])
    dr = d.get_request()
    assert dr == (server_id, [ida]) or dr == (server_id2, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    callback = lambda: None
    d.add_callback(ida, callback)
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() == (ida, [callback])
    assert d.get_cached(ida) == [server_id]

    d.data_came_in(server_id2, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == [server_id, server_id2] or d.get_cached(ida) == [server_id2, server_id]

def test_uses_mask():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server 1')
    server_id2 = make_ascii_id('server 2')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(50)),(server_id2, idmasklib.generate_random_mask(0))])
    assert d.get_request() == (server_id2, [ida])

def test_handles_unrequested_cleanly():
    d = BlockFinderData(max_downloads = 10, max_cached = 10)
    server_id = make_ascii_id('server id')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0))])
    assert d.get_request() == (server_id, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    callback = lambda: None
    d.add_callback(ida, callback)
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    gid = make_ascii_id('garbage')
    d.data_came_in(server_id, [gid, ida])
    assert d.get_request() is None
    assert d.get_finished_callback() == (ida, [callback])
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == [server_id]
    assert d.get_cached(gid) == [server_id]

def test_throttles_on_max_downloads():
    d = BlockFinderData(max_downloads = 1, max_cached = 10)
    server_id = make_ascii_id('server 1')
    server_id2 = make_ascii_id('server 2')

    ida = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0)), (server_id2, idmasklib.generate_random_mask(0))])
    dr = d.get_request()
    assert dr == (server_id, [ida]) or dr == (server_id2, [ida])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    idb = make_ascii_id('a')
    d.start_find([ida], [(server_id, idmasklib.generate_random_mask(0)), (server_id2, idmasklib.generate_random_mask(0))])
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

    d.data_came_in(server_id, [])
    dr2 = d.get_request()
    assert dr2 == (server_id, [ida]) or dr2 == (server_id2, [ida])
    assert dr != dr2
    assert d.get_request() is None
    assert d.get_finished_callback() is None
    assert d.get_cached(ida) == []

mojo_test_flag = 1

if __name__ == '__main__':
    RunTests.runTests('BlockFinder')




