#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: BlockWholesaler.py,v 1.21 2001/06/04 20:34:45 zooko Exp $'


### standard modules
import math
import string

### our modules
import BidStrategy
from BlockFinder import BlockFinder
import MojoConstants
import Cache
from confutils import confman
import debug
from human_readable import hr
import idlib
import mojoutil
import pricelib

true = 1
false = 0

WHOLESALER_FLAG = idlib.make_id('remember to make your message bodies dicts - do not pass this request on')


class BlockWholesaler:
    def __init__(self, broker, throttler):
        """
        @precondition `broker' must be an instance of Broker.: broker.__class__.__name__ == "Broker": "self: %s, broker: %s" % (hr(self), hr(broker))
        """
        assert broker.__class__.__name__ == "Broker", "precondition: `broker' must be an instance of Broker." + " -- " + "self: %s, broker: %s" % (hr(self), hr(broker))

        _oldval = confman['BLOCKSERVER'].get('MAX_BLOCKS_PER_WHOLESALE_RUN')
        if  _oldval:
            # 100 was the old 0.938/0.940 default setting, if its still there, replace it.
            # Otherwise use the users customized setting.
            del confman['BLOCKSERVER']['MAX_BLOCKS_PER_WHOLESALE_RUN']
            if int(_oldval) != 100:
                confman['BLOCKSERVER']['MAX_BLOCKS_PER_WHOLESALE_RUN_v2'] = _oldval

        # This contains the ids of blocks that are either currently unfindable or already in our blockstore.  (It is faster to store their ids here once we have noticed them in our blockstore than to look up *all* blocks that people have requested in our blockstore, which can be slow.)  Actually maybe this could be replaced by a fast "get list of blobids" method in BlockStore, but for now we'll use this little cache trick.)  --Zooko 2001-03-13
        EXPTIME = 5 * (60*60)   # 5 hours
        self.unwanted_blocks_map = Cache.StatsCache(autoexpireinterval=EXPTIME, autoexpireparams={'maxage': max(EXPTIME, 10 * int(confman.dict['BLOCKSERVER']['TIME_BETWEEN_WHOLESALE_QUERIES'])) })

        self.broker = broker
        self.finder = BlockFinder(self.broker, self._find_function)
        self.throttler = throttler
        self.queries = {}
        self.to_download = []
        self.downloads = []
        self.broker.doq.add_task(self._do_wholesale, delay = int(confman.dict['BLOCKSERVER']['TIME_BETWEEN_WHOLESALE_QUERIES'])/10 )
        
    def add_queries(self, block_ids):
        assert self.broker.doq.is_currently_doq()
        if len(block_ids) > 0 and idlib.equal(block_ids[0], WHOLESALER_FLAG):
            return
        # sort the list in canonical form so we skip duplicates reported in a single do you have blobs list
        block_ids = map(idlib.to_binary, block_ids)
        block_ids.sort()
        prev_id = None
        for block_id in block_ids:
            if block_id == prev_id:
                continue   # skip duplicates in the list
            if not self.queries.has_key(block_id):
                self.queries[block_id] = 0
            self.queries[block_id] = self.queries[block_id] + 1
            prev_id = block_id
    
    def _download_done(self, download):
        assert self.broker.doq.is_currently_doq()
        if download not in self.downloads:
            return
        self.downloads.remove(download)
        self._download_more()

    def _download_more(self):
        """
        @precondition This method must be called on the DoQ.: self.broker.doq.is_currently_doq()
        """
        assert self.broker.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        # the BlockThrottler keeps too many actual requests from going out at once; this just prevents
        # us from bothering to even create and queue up too many SingleDownloads at once.
        max_to_queue_at_once = 10

        while self.to_download and len(self.downloads) < max_to_queue_at_once:
            block_id = self.to_download.pop(0)
            self.downloads.append(SingleDownload(broker=self.broker,
                finder = self.finder, throttler = self.throttler, 
                callback = self._download_done, block_id = block_id))

    def _do_wholesale(self):
        """
        @precondition `self.broker' must be present.: hasattr(self, 'broker'): "self: %s" % hr(self)
        @precondition `self.broker' must be not None.: self.broker is not None: "self: %s, self.broker: %s" % (hr(self), hr(self.broker))
        @precondition `self.broker' must be an instance of Broker.: self.broker.__class__.__name__ == "Broker": "self: %s, self.broker: %s" % (hr(self), hr(self.broker))
        @precondition `self.broker' must have a multiblockstore.: hasattr(self.broker, 'multiblockstore'): "self: %s, self.broker: %s" % (hr(self), hr(self.broker))
        """
        assert hasattr(self, 'broker'), "precondition: `self.broker' must be present." + " -- " + "self: %s" % hr(self)
        assert self.broker is not None, "precondition: `self.broker' must be not None." + " -- " + "self: %s, self.broker: %s" % (hr(self), hr(self.broker))
        assert self.broker.__class__.__name__ == "Broker", "precondition: `self.broker' must be an instance of Broker." + " -- " + "self: %s, self.broker: %s" % (hr(self), hr(self.broker))
        assert hasattr(self.broker, 'multiblockstore'), "precondition: `self.broker' must have a multiblockstore." + " -- " + "self: %s, self.broker: %s" % (hr(self), hr(self.broker))
        assert self.broker.doq.is_currently_doq()

        # schedule the next run
        self.broker.doq.add_task(self._do_wholesale, delay = int(confman.dict['BLOCKSERVER']['TIME_BETWEEN_WHOLESALE_QUERIES']))

        numwanted = int(confman['BLOCKSERVER']['MAX_BLOCKS_PER_WHOLESALE_RUN_v2'])

        # local names are faster for use inside loops...
        queries = self.queries
        unwanted_blocks_map = self.unwanted_blocks_map

        # Now filter out ones that were recently unfindable, or which were memoized as being already in our block store.  (See below for where the "already gotem" memos are added, and see `_count_unfound_blocks()' for where the "couldn't find 'em" memos were added.)
        # (Do it whichever way will be faster...)
        if len(unwanted_blocks_map) > len(queries):
            for k in queries.keys():
                if unwanted_blocks_map.has_key(k):
                    del queries[k]
        else:
            for k in unwanted_blocks_map.keys():
                if queries.has_key(k):
                    del queries[k]

        # Throw out all but the top `numwanted * 3' most-requested blocks.
        mojoutil.skim(queries, numwanted * 3)

        # We only want blocks we don't have.
        # Note that it is very rare that there are any entries in `queries' for blocks that we already have, since we don't add to the `queries' dict if someone requests a block and we have it.
        assert self.broker.multiblockstore is not None
        store = self.broker.multiblockstore
        for k in queries.keys():
            if store.get_block_size(k) > 0:
                # Already got that one.
                unwanted_blocks_map[k] = None
                del queries[k]
        
        # pick a random selection 
        blockids_to_find = mojoutil.rotatelist(queries.keys())
        self.queries = {}

        # replace+cancel any old but still running wholesale tasks
        self.to_download = blockids_to_find[:numwanted]
        for i in self.downloads:
            i.cancel()
        self.downloads = []

        if len(self.to_download) < 1:
            debug.mojolog.write('no blocks to search for from others\n', v=3, vs='BlockWholesaler')
            return

        # start finding the suckers
        debug.mojolog.write("searching for %s blocks that others asked about but we did not have\n", args=(len(self.to_download),), v=2, vs='BlockWholesaler')
        self.broker.doq.add_task(self.finder.find_blocks, kwargs={'block_ids': self.to_download, 'allow_download_cache': 0})
        self.broker.doq.add_task(self._count_unfound_blocks)

        # start the actual work for this wholesale run
        self.broker.doq.add_task(self._download_more)

    def _count_unfound_blocks(self):
        """Used to keep track of blocks that just aren't findable"""
        for block_id in self.to_download:
            self.unwanted_blocks_map[block_id] = None
        def _found_callback(block_id, failure_reason=None, self=self):
            try:
                del self.unwanted_blocks_map[block_id]
            except KeyError:
                pass
        for block_id in self.to_download:
            self.broker.doq.add_task(self.finder.callback_when_found, args=(block_id,), kwargs={'found_callback': _found_callback, 'finished_callback': lambda block_id, failure_reason=None: None})

    def _find_function(self, server_id, server_info, block_ids, callback, preferenceno=0):
        """
        called by BlockFinder to do actual message generation and spewing over mtm
        """
        assert len(block_ids) > 0, "_find_function called without any block_ids to find!"
        self.broker.do_you_have_blobs((server_id, server_info), block_ids, callback, allow_recursion=0, mojooffer=BidStrategy.choose_bid("MO_DO_YOU_HAVE_BLOBS_BLOCK_WHOLESALING", self.broker.ServicesMTM, preferenceno=preferenceno))

class SingleDownload:
    def __init__(self, broker, finder, throttler, callback, block_id):
        assert broker.doq.is_currently_doq()
        self.broker = broker
        self.throttler = throttler
        self.finder = finder
        self.callback = callback
        self.block_id = block_id
        self.block_length = 65536
        self.finished = false
        self.done_finding = false
        self.currently_downloading = false
        self.tried_servers = []

        speed = int(confman.dict['CONN_SPEED_IN'])
        # these values probably need tweaking
        if speed > 1000:
            self.max_concurrent_request_blobs = 14
        elif speed > 500:
            self.max_concurrent_request_blobs = 7
        elif speed > 180:
            self.max_concurrent_request_blobs = 5
        elif speed >= 112:
            self.max_concurrent_request_blobs = 4
        else:
            self.max_concurrent_request_blobs = 2

        finder.callback_when_found(block_id = block_id, found_callback = self._found_callback, 
            finished_callback = self._finished_callback)

    def _found_callback(self, block_id):
        assert self.broker.doq.is_currently_doq()
        if self.finished:
            return
        self.broker.doq.add_task(self.throttler.do_download, args = [self])

    def _finished_callback(self, block_id):
        assert self.broker.doq.is_currently_doq()
        if self.finished:
            return
        self.done_finding = true
        self.broker.doq.add_task(self.throttler.do_download, args = [self])

    def is_finished(self):
        assert self.broker.doq.is_currently_doq()
        return self.finished

    def cancel(self):
        assert self.broker.doq.is_currently_doq()
        if self.finished:
            return
        self.finished = true
        self.broker.doq.add_batch_task('download', self.finder.cancel, kwargs = {'block_id': self.block_id})
        self.broker.doq.add_batch_task('download', self.throttler.do_download, args = [self])
        self.broker.doq.add_batch_task('download', self.callback, args = [self])

    def get_priority(self):
        assert self.broker.doq.is_currently_doq()
        return -2

    def start_next(self, current_servers):
        """
        current_servers is [server id] - it is modified by this function (appended to).  it is intended to be a list of
        server_ids that we're currently downloading blocks from so that we don't send multiple request
        blobs to the same server at once (as that doesn't help streaming and causes timeouts on the later requests).
        """
        assert self.broker.doq.is_currently_doq()
        if self.finished or self.currently_downloading:
            return false

        completely_failed = self.done_finding

        message_type = 'request blob'
        message_body = {'blob id': self.block_id, 'len': self.block_length}
        infolist = self.finder.get_block_locations_from_cache(self.block_id)
        infolist = self.broker.ServicesMTM.get_handicapper().sort_by_preference(infolist, message_type, message_body)
        infolist2 = []
        for infotuple in infolist:
            if infotuple[0] in self.tried_servers:
                continue
            if infotuple[0] in current_servers:
                completely_failed = false
            else:
                infolist2.append(infotuple)
        infolist = infolist2
        if len(infolist) == 0:
            if completely_failed:
                self.cancel()
            return false

        self.currently_downloading = true
        server_id = infolist[0][0]
        self.tried_servers.append(server_id)
        self.doing_download = true
        current_servers.append(server_id)

        server_id, server_info = infolist[0]
        topay = pricelib.blob_server_purchase_price(server_info, message_body)

        connection_speed_bytes_per_second = 2400  # 2400 should be true of even a non-great 28.8k modem connection
        # timeouts are always at least 45 just due to general software load and send/receive via relay issues
        maxtime = min(max(self.block_length / (connection_speed_bytes_per_second / 
            self.max_concurrent_request_blobs / 2), 45), MojoConstants.DEFAULT_TIMEOUT)

        timer = mojoutil.Timer()
        def process_callback(widget, outcome = None, failure_reason = None, 
                timer = timer, self = self, current_servers = current_servers, 
                server_id = server_id):
            assert self.broker.doq.is_currently_doq()
            block_id = idlib.to_ascii(self.block_id)
            timer.stop()
            self.broker.doq.add_task(self.throttler.download_done)
            current_servers.remove(server_id)
            self.currently_downloading = false

            if self.finished:
                return

            self.broker.doq.add_task(self.finder.cancel, args = (block_id,))
            self.broker.doq.add_task(self.cancel)  # this sets self.finished=true and performs our callbacks

            if failure_reason is not None:
                return
            if outcome.get('result') == 'failure':
                return
            block_data = outcome['data']
            if not idlib.equal(block_id, idlib.make_id(block_data)):
                return

            # store the latency overall and on a per-block size basis
            cp_obj = self.broker.ServicesMTM._keeper.get_counterparty_object(server_id)
            cp_obj.update_custom_stat_weighted_sample('request blob bytes/sec', float(len(block_data))/timer.get())

            self.broker.ServicesMTM._blobserver.try_to_store_blob(block_id, block_data)

        timer.start()
        self.broker.ServicesMTM.initiate(server_id, message_type, message_body, mojooffer=0, outcome_func = process_callback, timeout=maxtime)
        return true


class GivenLocationBlockFinder:
    """
    This fake BlockFinder interface is only for use by a BlockWholesaler.SingleDownload.
    It is designed to be used during publication tracker polling when we are told the
    location of a block in advance.
    """
    def __init__(self, broker, block_id, block_len, location_tuple):
        self.broker = broker
        self.block_id = block_id
        self.block_len = block_len
        self.location_tuple = location_tuple

    def get_block_locations_from_cache(self, block_id):
        if idlib.equal(block_id, self.block_id):
            return [self.location_tuple]
        else:
            return []

    def cancel(self, block_id):
        if idlib.equal(block_id, self.block_id):
            self.block_id = None

    def callback_when_found(self, block_id, found_callback, finished_callback):
        if idlib.equal(block_id, self.block_id):
            self.broker.doq.add_task(found_callback, args=(block_id,))
        self.broker.doq.add_task(finished_callback, args=(block_id,))

