# Manages a list of broker ID's this Broker prefers.

# Standard Modules:
import time

# Mojo Modules:
import MojoHandicapper
import debug
import SEXP
import confutils
import idlib
import MetaTrackerLib

BUDDY = 1 # Buddies are given a 0 penalty.
ENEMY = 0 # Enemies are always disqualified.

BUDDY_BONUS = 10000 # How much to penalize non-Buddy, non-Enemy counterparties.
REFRESH_CONTACT_INFO_DELAY = (75 * 60) # Every 75 minutes.
RELOAD_AND_SAVE_DELAY = 111


def debug_out(output, args=(), v=3):
    debug.mojolog.write(output, args=args, v=v, vs='BuddyList')
    
class BuddyList:
    def __init__(self, doq, mtm):
        """
        @pre doq is a DoQ.
        @pre mtm is a MetaTransactionManager.
        """
        self.doq = doq
        self.mtm = mtm
        # Maps id's to dicts with keys 'type' and 'timestamp'.
        # type is either BUDDY or ENEMY.
        # timestamp represents the last time contact info was successfully refreshed.  (It is initially absent.)
        self.buddyids = {} 
        debug_out('Initializing.\n', v=4)
    def refresh_list(self):
        """Refresh the list from the config file."""
        debug_out('(Re)Loading ID list from config file.\n')
        buddycount, enemycount = 0, 0
        confutils.confman.load()
        # Convert the base64-encoded config-file values to binary then add them to self.buddyids:
        for id in confutils.confman.dict.get('BUDDY_LIST_IDS', {}).values():
            id = idlib.to_binary(id)
            self.__get_with_default(id)['type'] = BUDDY
            buddycount = buddycount + 1
        for id in confutils.confman.dict.get('ENEMY_LIST_IDS', {}).values():
            id = idlib.to_binary(id)
            self.__get_with_default(id)['type'] = ENEMY
            enemycount = enemycount + 1
        debug_out('Found %s buddies and %s enemies.\n', args=(buddycount, enemycount,), v=3)
    def __get_with_default(self, id):
        if not self.buddyids.has_key(id):
            self.buddyids[id] = {}
        return self.buddyids[id]
    def save_list(self):
        debug_out('Saving ID list to config file.\n')
        # Convert the keys of self.buddyids to base64-encoded strings then convert that list into an integer-keyed (list-like) dict, then assign it to the configuration manager:
        confutils.confman.dict['BUDDY_LIST_IDS'] = SEXP.listToDict(map(idlib.to_ascii, self.buddyids.keys()))
        # Save the configuration data to the config file:
        confutils.confman.save()
    def periodic_reload_and_save_task(self, delay=RELOAD_AND_SAVE_DELAY):
        """The 'task' in the name refers to the fact that this is meant to be operated with a DoQ."""
        self.refresh_list()
        #self.save_list()
        # Append ourself to the DoQ and preserve the delay we were passed:
        self.doq.add_task(self.periodic_reload_and_save_task, args=(delay,), delay=delay)
    def refresh_contact_info(self):
        """
        Calls MetaTrackerLib.get_contact_info on all ID's with a timestamp older than 1 hour.
        This is important to ensure all buddy's are considered for business transactions (by
        ensuring their contact info is cached).

        XXX Is this necessary?  Does it work as described in the previous paragraph?  --Neju  2000-11-03
        """
        debug_out('Refreshing contact info of Buddies.\n')
        def callback_factory_func(id, buddyids=self.buddyids):
            """
            The lookup contact info doesn't have access to the ID?!
            We use this factory function to make a separate callback for each id!  Kinda icky (but fun)!
            """
            def contact_callback(widget, outcome, failure_reason=None, id=id, buddyids=buddyids):
                if outcome and outcome['result'] == 'success':
                    buddyids[id]['timestamp'] = time.time()
            return contact_callback
        for id, valdict in self.buddyids.items():
            if (valdict.get('type') == BUDDY) and (time.time() - valdict.get('timestamp', 0) > (60 * 60)):
                # Only if it is a buddy and the contact info is over 1 hour old:
                MetaTrackerLib.get_contact_info(self.mtm, id, callback=callback_factory_func(id))
    def periodic_refresh_contact_info_task(self, delay=REFRESH_CONTACT_INFO_DELAY ):
        """The 'task' in the name refers to the fact that this is meant to be operated with a DoQ."""
        self.refresh_contact_info()
        self.doq.add_task(self.periodic_refresh_contact_info_task, args=(delay,), delay=delay)
    def handicapper(self, counterparty_id, metainfo, message_type, message_body):
        """A handicapper that strongly prefers members of this buddylist."""
        if self.buddyids.has_key(counterparty_id):
            if self.buddyids[counterparty_id]['type'] == BUDDY:
                debug_out('Handicapper for %s, result: %s\n', args=(counterparty_id, 'BUDDY'), v=4)
                return 0 # We like!
            else:
                debug_out('Handicapper for %s, result: %s\n', args=(counterparty_id, 'ENEMY'), v=4)
                return MojoHandicapper.DISQUALIFIED # We NO like!
        else:
            debug_out('Handicapper for %s, result: %s\n', args=(counterparty_id, 'STRANGER'), v=8)
            return BUDDY_BONUS # We are ambivalent, but we prefer this counterparty much less than our buddys.
