#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# This module implements a basic dictionary-like Cache object as well
# as an example more advanced StatsCache which can remove items based
# on their access patterns.
#
# $Id: Cache.py,v 1.17 2001/06/04 06:44:25 greg Exp $

### standard modules
import threading
import time


### our modules
from MojoConstants import true, false
from human_readable import hr


class SimpleCache:
    """
    A very simple cache object.

    This is really just a wrapper around dicts so that you can subclass them, plus a convenient
    `remove()' method which returns the removed item.

    If you override `insert()' then you can be sure you'll catch any insertion of any items into
    this dict.  If you override `remove()' then you can be sure you'll catch any removal of any
    items from this dict.  Remember to actually add/remove them from self._dict if you do this,
    or else call `SimpleCache.insert()' or `SimpleCache.remove()' respectively.
    """
    def __init__(self):
        self._dict = {}
        self._lock = threading.RLock()

    def remove(self, key, x=None, strictkey=true):
        """
        remove an item from the cache and return it

        @param strictkey `true' if you want a KeyError in the case that `key' is not there,
            `false' if you want a reference to `x' in the case that `key' is not there
        @param x the object to return if `key' is not there; This is ignore if `strictkey'.
        """
        self._lock.acquire()
        try:
            item = self.get(key, x)
            if self.has_key(key):
                del self._dict[key]
            else:
                if strictkey:
                    raise KeyError, key
            return item
        finally:
            self._lock.release()

    def insert(self, key, item=None):
        """
        @param item if `None', then the key itself is the item
        """
        if item:
            self._dict[key] = item
        else:
            self._dict[key] = key

    def __nonzero__(self):
        return self._numitems != 0

    def __len__(self):
        return len(self._dict)

    def __setitem__(self, key, item):
        return self.insert(key, item)

    def keys(self):
        return self._dict.keys()

    def items(self):
        return self._dict.items()

    def values(self):
        return self._dict.values()

    def has_key(self, key):
        return self._dict.has_key(key)

    def get(self, key, x=None):
        return self._dict.get(key, x)

    def __getitem__(self, key):
        return self._dict[key]

    def __delitem__(self, key):
        return self.remove(key)

   
class SimpleStatsCache(SimpleCache):
    """
    A SimpleCache which has most of the guts copied out of StatsCache, except for the part about
    removing things automatically according to a parameterized "throw things out that don't make
    muster" cleaning strategy.  Subclass this in order to implement your favorite cache cleaning
    strategy.

    I left out the size-accounting feature since it is implemented in Cache instead of
    StatsCache and since I don't need it right now.  --Zooko 2000-08-15

    One more difference between this and StatsCache: if you call `values()' or `items()' on
    this, it touches every item, incrementing its usage number.
    """
    def __init__(self):
        SimpleCache.__init__(self)

        self._stats = {}

    def insert(self, key, item=None):
        self._enter()
        try:
            result = SimpleCache.insert(self, key, item)
            now = time.time()
            self.stats[key] = [now, now, 0]
            return result
        finally:
            self._exit()

    def get(self, key, x=None):
        self._enter()
        try:
            self._touch(key)
            return SimpleCache.get(self, key, x)
        finally:
            self._exit()

    def remove(self, key, strictkey=true):
        """
        @see SimpleCache for interface documentation
        """
        self._enter()
        try:
            item = SimpleCache.remove(self, key, strictkey=strictkey)
            try:
                del self.stats[key]
            except KeyError:
                # We do not really care if there are no stats.  If `strictkey', then it is entirely consistent for their to be no stats, and no actual item either!
                pass
            return item
        finally:
            self._exit()

    def items(self):
        """
        This touches (increases the `numaccesses' by 1) all items.
        """
        self._enter()
        try:
            for key in self._dict.keys():
                self._touch(key)
            return SimpleCache.items()
        finally:
            self._exit()

    def values(self):
        """
        This touches (increases the `numaccesses' by 1) all items.
        """
        self._enter()
        try:
            for key in self._dict.keys():
                self._touch(key)
            return SimpleCache.values()
        finally:
            self._exit()

    def _touch(self, key):
        self._enter()
        try:
            now = time.time()
            statslist = self._stats.get(key)
            if statslist is None:
                debug.stderr.write("SimpleStatsCache._touch(): no stats for key %s.  Fixing by adding stats now.\n" % `key`)
                statslist = [now, now, 0]

            statslist[key][1] = now # update access time
            try:
                statslist[key][2] = statslist[key][2] + 1 # update numaccesses
            except OverflowError:
                statslist[key][2] = statslist[key][2] + 1L # switch to a long

            self._stats[key] = statslist
        finally:
            self._exit()


class Cache :
    """Cache([maxsize], [maxitems], [autolen], [initialdict]) - Resource limited object cache
    A Caching dictionary-like object.  A maximum number of items or
    maximum total size of all items may be specified.  Items will be
    removed in FIFO order to make room for new items.

    initialdict contains the items to initialize this Cache object
    with.  They will be inserted in the order that initialdict.items()
    returns them; subject to the cache size limits and removal rules.
    initialdict values must have a valid __len__ method.

    If autolen is true, len() will be called on all inserted items; if
    a TypeError exception is raised itemsize will be used instead.
    """
    def __init__(self, maxsize=None, maxitems=None, autolen=None, initialdict={}) :
        """
        @param maxsize the maximum amount of ram this cache can hold, or `None' if there is no limit;  This cannot be 0.
        @param maxitems the maximum number of items this cache can hold, or `None' if there is no limit;  This cannot be 0.

        @precondition `maxsize' must be positive.: maxsize is None or maxsize > 0: "maxsize: %s" % hr(maxsize)
        @precondition `maxitems' must be positive.: maxitems is None or maxitems > 0: "maxitems: %s" % hr(maxitems)
        """
        assert maxsize is None or maxsize > 0, "precondition: `maxsize' must be positive." + " -- " + "maxsize: %s" % hr(maxsize)
        assert maxitems is None or maxitems > 0, "precondition: `maxitems' must be positive." + " -- " + "maxitems: %s" % hr(maxitems)

        assert type(initialdict) == type({})

        self.lock = threading.RLock()

        # dict of tuples of the cached item and its size: (item, itemsize)
        self.cache = {}
        self.fifo = []    # insertion order list of keys
        self.numitems = 0
        self.totalsize = 0

        self.autolen = autolen

        self.maxsize = maxsize
        self.maxitems = maxitems

        # initialize the cache
        for key, value in initialdict.items() :
            self.insert(key, value, len(value))
  
    def __nonzero__(self) :
        return self.numitems != 0
  
    def __len__(self) :
        return len(self.cache)
  
    ####################################################################
    def insert(self, key, item, itemsize=0) :
        """
        insert(key, item, itemsize=0) - insert item into the cache,
        to be retrieved using key.  itemsize should be the size in bytes
        you wish this item to be counted as.

        @precondition Item size must be non-negative.: itemsize >= 0: "key: %s, item: %s, itemsize: %s" % (hr(key), hr(item), hr(itemsize))
        """
        assert itemsize >= 0, "precondition: Item size must be non-negative." + " -- " + "key: %s, item: %s, itemsize: %s" % (hr(key), hr(item), hr(itemsize))

        self.lock.acquire()
        try:
            # if the object already exists in the cache, replace it
            if self.cache.has_key(key) :
                self.remove(key)
            
            if self.autolen :
                try:
                    itemsize = len(item)
                except TypeError:
                    # len() failed, use the original itemsize
                    pass

            # make sure there is space in the cache
            if (self.maxsize is not None) and (self.totalsize + itemsize) > self.maxsize:
                self.makeroomfor(size=itemsize)
            if (self.numitems is not None) and (self.numitems + 1) > self.maxitems:
                self.makeroomfor(numitems=1)

            # insert the item into the cache and update our stats
            self.cache[key] = (item, itemsize)
            self.fifo.append(key)
            self.totalsize = self.totalsize + itemsize
            self.numitems = self.numitems + 1
        finally:
            self.lock.release()

    def __setitem__(self, key, item, itemsize=0) :
        self.insert(key, item, itemsize)


    ####################################################################
    def insert_and_return_unique(self, key, item, itemsize=0) :
        """
        Performs this atomic operation:
          - If key does not exist in the cache, insert it.
          - Return the value of self[key]

        Compare the return value to your item parameter using "is" to
        determine if the object already existed in the cache or if
        your new one was inserted.
        """
        self.lock.acquire()
        try:
            if self.cache.has_key(key) :   # if self.has_key(key) :
                return self.cache[key][0]  #     return self[key]

            # Calling insert here is important for proper
            # functionality in derived classes.  -greg
            self.insert(key, item, itemsize)

            return self.cache[key][0]      # return self[key]
        finally:
            self.lock.release()
    
    setdefault = insert_and_return_unique
    

    ####################################################################
    def keys(self) :
        self.lock.acquire()
        try:
            k = self.cache.keys()
        finally:
            self.lock.release()
        return k

    ####################################################################
    def items(self) :
        self.lock.acquire()
        try:
            i = self.cache.items()
            i = map(lambda a: (a[0], a[1][0]), i)  # remove size info from values
        finally:
            self.lock.release()
        return i

    ####################################################################
    def values(self) :
        self.lock.acquire()
        try:
            v = self.cache.values()
            v = map(lambda a: a[0], v)   # remove size info
        finally:
            self.lock.release()
        return v

    ####################################################################
    def has_key(self, key) :
        """has_key(key) - See if an item is in this cache.  This is not
        deterministic if the cache is being accessed by multiple threads
        at once. """
        self.lock.acquire()
        try:
            return self.cache.has_key(key)
        finally:
            self.lock.release()

    ####################################################################
    def get(self, key, default=None) :
        """get(key) - Lookup an item in the cache, return None if it doesn't exist"""
        self.lock.acquire()
        try:
            itemtuple = self.cache.get(key)
            if itemtuple is None :
                return default
            return itemtuple[0]
        finally:
            self.lock.release()

    def __getitem__(self, key) :
        self.lock.acquire()
        try:
            if self.has_key(key) :
                return self.get(key)
            else :
                raise KeyError, key
        finally:
            self.lock.release()

    ####################################################################
    def remove(self, key) :
        """ remove(key) - Remove an item from the cache (and return it) """
        self.lock.acquire()
        try:
            item = self.cache[key][0]
            self.totalsize = self.totalsize - self.cache[key][1]
            self.numitems = self.numitems - 1
            del self.cache[key]
            self.fifo.remove(key)
        finally:
            self.lock.release()
        return item

    def __delitem__(self, key) :
        self.remove(key)

    ####################################################################
    def makeroomfor(self, size=0, numitems=0) :
        """
        makeroomfor(size=0, numitems=0) -
        Make room in the cache for numitems objects and size bytes.  Raises
        a ValueError exception if either numitems or size is greater than
        their respective maximums for this Cache object.

        @postcondition not self.maxsize or self.totalsize + size < self.maxsize, "There are enough free bytes (if we care about how much memory the cache takes up)"
        @postcondition not self.maxitems or self.numitems + numitems <= self.maxitems, "There are enough free slots (if we care about how many items there are)"
        """
        old_totalsize = self.totalsize
        old_numitems = self.numitems

        if self.maxsize and size > self.maxsize :
            # raise ValueError, "can't make room for more than maxsize"
            size = self.maxsize
        if self.maxitems and numitems > self.maxitems :
            raise ValueError, "can't make room for more than maxitems"

        assert len(self.fifo) == len(self.cache), "cache len != fifo len, impossible"
        self.lock.acquire()
        try:
            # Free up space in the cache
            if self.maxsize :
                # XXX This greedy approach may behave poorly if the
                # size of the items in the cache varies greatly
                # relative to the magnitude of maxsize!
                while self.maxsize - self.totalsize < size:
                    assert (self.totalsize == 0) or (len(self.fifo) > 0), "len(self.fifo): %s,  self.totalsize: %s,  self.numitems: %s, self.maxsize: %s, size: %s" % (hr(len(self.fifo)), hr(self.totalsize), hr(self.numitems), hr(self.maxsize), hr(size))
                    assert (self.numitems == 0) or (len(self.fifo) > 0), "len(self.fifo): %s,  self.totalsize: %s,  self.numitems: %s, self.maxsize: %s, size: %s" % (hr(len(self.fifo)), hr(self.totalsize), hr(self.numitems), hr(self.maxsize), hr(size))
                    self.remove(self.fifo[0])
            if self.maxitems :
                while self.maxitems - self.numitems <= numitems:
                    assert (self.totalsize == 0) or (len(self.fifo) > 0), "len(self.fifo): %s,  self.totalsize: %s,  self.numitems: %s, self.maxsize: %s, size: %s" % (hr(len(self.fifo)), hr(self.totalsize), hr(self.numitems), hr(self.maxsize), hr(size))
                    assert (self.numitems == 0) or (len(self.fifo) > 0), "len(self.fifo): %s,  self.totalsize: %s,  self.numitems: %s, self.maxsize: %s, size: %s" % (hr(len(self.fifo)), hr(self.totalsize), hr(self.numitems), hr(self.maxsize), hr(size))
                    self.remove(self.fifo[0])
        finally:
            self.lock.release()

        assert not self.maxsize or self.totalsize + size <= self.maxsize
        assert not self.maxitems or self.numitems + numitems <= self.maxitems

    ####################################################################
    # Empty the cache
    ####################################################################
    def empty(self) :
        """ empty() - empty this Cache object """
        self.lock.acquire()
        try:
            self.cache = {}
            self.fifo = []
            self.totalsize = 0
            self.numitems = 0
        finally:
            self.lock.release()
    

########################################################################
# This Cache keeps statistics for each item in the cache including the
# insert time, last access time, and total number of accesses.  It
# also adds the expire() method [see below] for removing items from
# the cache based on the kept stats.
########################################################################
class StatsCache(Cache) :
    """A Cache object that keeps insertion and access time statistics
    on each item.  It adds the expire method for removing items based
    on their stats."""

    ####################################################################
    def __init__(self, maxsize=None, maxitems=None, autolen=0, initialdict={}, autoexpireinterval=None, autoexpireparams=None) :
        """
        @param autoexpireinterval is the number of seconds between automatically calling our expire method.
        @param autoexpireparams is a dict of keyword parameters for self.expire() that will be called every autoexpireinterval seconds.
        """
        Cache.__init__(self, maxsize=maxsize, maxitems=maxitems, autolen=autolen, initialdict=initialdict)

        # stats is a dictionary of lists containing:
        # [insert time, last access time, number of accesses]
        self.stats = {}

        self.autoexpireinterval = autoexpireinterval
        self.autoexpireparams = autoexpireparams
        self.last_autoexpire_run = time.time()
        self.__in_autoexpire = 0

        if self.autoexpireinterval :
            assert (type(self.autoexpireparams) == type({}))
            # verify that autoexpireparams are acceptable to self.expire() now rather than later
            apply(self.expire, (), self.autoexpireparams)


    ####################################################################
    def __auto_expire(self) :
        """Run expire if it is time, this must be called with self.lock acquired"""
        if self.__in_autoexpire :
            # don't let autoexpire trigger itself
            return
        now = time.time()
        time_since_last_autoexpire = now - self.last_autoexpire_run
        if self.autoexpireinterval <= time_since_last_autoexpire :
            self.last_autoexpire_run = now
            self.__in_autoexpire = 1
            try:
                apply(self.expire, (), self.autoexpireparams)
            finally:
                self.__in_autoexpire = 0

    ####################################################################
    def insert(self, key, item, itemsize=0) :
        """
        @precondition Item size must be non-negative.: itemsize >= 0: "key: %s, item: %s, itemsize: %s" % (hr(key), hr(item), hr(itemsize))
        """
        assert itemsize >= 0, "precondition: Item size must be non-negative." + " -- " + "key: %s, item: %s, itemsize: %s" % (hr(key), hr(item), hr(itemsize))

        self.lock.acquire()
        try:
            if self.autoexpireinterval :
                self.__auto_expire()
            Cache.insert(self, key, item, itemsize)
            if self.stats.has_key(key) :
                del self.stats[key]
            now = time.time()
            # (insertion time, accesstime, number of accesses)
            self.stats[key] = [now, now, 0]
        finally:
            self.lock.release()

    ####################################################################
    def get(self, key, default=None) :
        self.lock.acquire()
        try:
            if self.autoexpireinterval :
                self.__auto_expire()
            item = Cache.get(self, key, default)
            if self.stats.has_key(key) :
                self.stats[key][1] = time.time()             # update access time
                try:
                    self.stats[key][2] = self.stats[key][2] + 1  # update numaccesses
                except OverflowError:
                    self.stats[key][2] = self.stats[key][2] + 1L # switch to a long
        finally:
            self.lock.release()
        return item

    ####################################################################
    def remove(self, key) :
        self.lock.acquire()
        try:
            item = Cache.remove(self, key)
            del self.stats[key]
            if self.autoexpireinterval :
                self.__auto_expire()
        finally:
            self.lock.release()
        return item

    ####################################################################
    def keys(self) :
        if self.autoexpireinterval :
            self.__auto_expire()
        return Cache.keys(self)

    ####################################################################
    def items(self) :
        if self.autoexpireinterval :
            self.__auto_expire()
        return Cache.items(self)

    ####################################################################
    def values(self) :
        if self.autoexpireinterval :
            self.__auto_expire()
        return Cache.values(self)

    ####################################################################
    def expire(self, minaccesses=0, staletime=0, maxage=None, cleanerfunc=None) :
        """expire(minaccesses=0, staletime=0, maxage=None) :
        Remove all items from the Cache that:
           a) have been accessed less than minaccesses times,
           b) have not been accessed in the past staletime seconds
        or c) are older than maxage seconds

        cleanerfunc (if given) should take a key and value as parameters and return
        a boolean indicating if they should remain in the cache.  It is only called
        on expired key/value pairs. This can be used to perform extra cleanup of
        objects as they are removed, or to veto an object's removal.  If a removal
        is vetoed that object's access time and number of accesses is updated.
        """
        self.lock.acquire()
        now = time.time()
        try:
            for key in self.stats.keys() :
                if maxage and now - self.stats[key][0] > maxage :        # (c)
                    if cleanerfunc and cleanerfunc(key, self.get(key)) :
                        continue
                    self.remove(key)
                    continue
                if staletime and now - self.stats[key][1] > staletime :  # (b)
                    if cleanerfunc and cleanerfunc(key, self.get(key)) :
                        continue
                    self.remove(key)
                    continue
                if self.stats[key][2] < minaccesses :                    # (a)
                    if cleanerfunc and cleanerfunc(key, self.get(key)) :
                        continue
                    self.remove(key)
        finally:
            self.lock.release()


    ####################################################################
    def empty(self) :
        self.lock.acquire()
        try:
            Cache.empty(self)
            self.stats = {}
        finally:
            self.lock.release()
    

########################################################################

def _help_test_empty_lookup(d) :
    assert d.get('spam') is None

def _help_test_key_error(d) :
    try :
        d['spam']
        assert 0
    except KeyError :
        pass

def _help_test_insert_and_get(d) :
    d.insert("spam", "eggs")
    d["spam2"] = "eggs2"
    assert d.get("spam") == "eggs", str(d)
    assert d.get("spam2") == "eggs2"
    assert d["spam"] == "eggs"
    assert d["spam2"] == "eggs2"

def _help_test_insert_and_remove(d):
    d.insert('spam', "eggs")
    x = d.remove('spam')
    assert x == "eggs", "x: %s" % `x`
    
def test_Cache_empty_lookup():
    _help_test_empty_lookup(Cache())

def test_StatsCache_empty_lookup():
    _help_test_empty_lookup(StatsCache())

def test_SimpleCache_empty_lookup():
    _help_test_empty_lookup(SimpleCache())

def test_Cache_key_error():
    _help_test_key_error(Cache())
    
def test_StatsCache_key_error():
    _help_test_key_error(StatsCache())
    
def test_SimpleCache_key_error():
    _help_test_key_error(SimpleCache())
    
def test_Cache_insert_and_get():
    _help_test_insert_and_get(Cache())

def test_StatsCache_insert_and_get():
    _help_test_insert_and_get(StatsCache())

def test_SimpleCache_insert_and_get():
    _help_test_insert_and_get(SimpleCache())

def test_Cache_insert_and_remove():
    _help_test_insert_and_remove(Cache())

def test_StatsCache_insert_and_remove():
    _help_test_insert_and_remove(StatsCache())

def test_SimpleCache_insert_and_remove():
    _help_test_insert_and_remove(SimpleCache())

def test_StatsCache_autoexpire() :
    success = 0
    try:
        c = StatsCache(autoexpireinterval=0.5, autoexpireparams={'foo': "bar"})
    except:
        success = 1
    if not success :
        assert 0, "StatsCache didn't reject bad autoexpireparams"
    c = StatsCache(autoexpireinterval=0.5, autoexpireparams={'minaccesses':3})
    c["spam"] = "eggs"
    c["swallow"] = "african"
    w = c["spam"]  # spam 1
    w = c["spam"]  # spam 2
    w = c["spam"]  # spam 3
    x = c.get("swallow") # swallow 1
    assert c.get("swallow") == "african" # swallow 2
    time.sleep(1)
    c.insert("foo", "bar")  # use the cache, trigger an autoexpire run
    assert c.get("spam") == "eggs"  # spam 4
    assert c.get("swallow") == None # swallow 3

def test_StatsCache_expire_cleanerfunc() :
    # basic test of cleanerfunc
    cleaned_items = []
    c = StatsCache()
    c['spam'] = "eggs"
    c['mog'] = "barf"
    time.sleep(0.6)
    assert c['mog'] == "barf"
    c.expire(staletime=0.5, cleanerfunc=lambda k, v, c=cleaned_items: c.append((k,v)) )
    assert not c.has_key('spam')
    assert len(cleaned_items) == 1 and cleaned_items[0] == ('spam', 'eggs')

    # test that cleanerfunc veto's removal
    c['spam'] = "eggs"
    time.sleep(0.5)
    c.expire(staletime=0.1, cleanerfunc=lambda k, v: k == 'mog')
    assert c.has_key('mog')
    assert not c.has_key('spam')


def test_generic() :
    # print "Testing Cache(maxitems=23, maxsize=2600):"
    C = Cache(maxsize=2600, maxitems=23)
    for i in xrange(0,50) :
        C.insert(i, str(i*i), (i%10)*10)
    # print "  after inserting 50 items:",
    # print len(C), "cache entries (", C.totalsize, "`bytes' )"
    for i in xrange(50,100) :
        C.insert(i, str(i*i), i*5)
    # print "  after inserting 50 more `larger' items:",
    # print len(C), "cache entries (", C.totalsize, "`bytes' )"
    # print
    del C

    ### Test StatsCache
    import time
    # print "Testing StatsCache():"
    S = StatsCache()
    # print "  populating cache with items 1..100 & 310:",
    for i in xrange(1,101) :
        S[i] = i*i
    S[310] = 69
    # print "now contains", len(S), "items."

    # print "  Accessing item 11"
    a = S[11]

    # print "  sleeping 2 seconds..."
    time.sleep(2)
    # print "  adding items 101..200:",
    for i in xrange(101,201) :
        S[i] = i*i
    # print "now contains", len(S), "items."

    # print "  Accessing items 22, 105 twice, and 113"
    a = S[22]
    a = S[113]
    a = S[105]
    a = S[105]

    # print "  sleeping 1 more seconds..."
    time.sleep(2)
    # print "  adding items 201..300:",
    for i in xrange(201,301) :
        S[i] = i*i
    # print "now contains", len(S), "items."

    # print "  Accessing items 11 and 224"
    a = S[11]
    a = S[224]

    # print "  sleeping 1 more second..."
    time.sleep(1)
    # print "  adding items 301..400:",
    for i in xrange(301,401) :
        S[i] = i*i
    # print "now contains", len(S), "items."

    # print "  Accessing item 33, 334 twice, 224, and 205"
    a = S[33]
    a = S[334]
    a = S[224]
    a = S[334]
    a = S[205]

    # print " Expiration tests:"
    # print "  items older than 4 seconds...",
    S.expire(maxage=4)
    # print "now contains", len(S), "items."

    # print "  items accessed less than 2 times...",
    S.expire(minaccesses=2)
    # print "now contains", len(S), "items."
    # print S.keys()

    # print "  items not accessed in the last 1 second...",
    S.expire(staletime=1)
    # print "now contains", len(S), "items."
    # print S.keys()

    try:
        worked = 0
        # print "Testing a cache miss on item 11...",
        d = S[11]
    except KeyError:
        # print "Success."
        worked = 1
    if not worked :
        pass
        # print " TEST FAILED!"
    
    # print "\nTests complete"


def test_get_default_on_Cache():
    s = StatsCache()
    s['spam'] = 'eggs'
    assert s.get('ham', 69) == 69
    assert s.get('spam', 69) == 'eggs'

def test_Cache_values():
    c = Cache()
    c[3] = 'three'
    c[4] = 'four'
    assert c.values() == ['three', 'four'] or c.values() == ['four', 'three']

mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests('Cache')

if __name__ == '__main__' :
    run()
