#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# Implements functions for purchasing and publishing blobs
# this should probably have a different name.
#

# standard modules
import traceback
import threading
import time
import types
import copy

# our modules
import BidStrategy
import MojoTransaction
import MojoMessage
import mojoutil
import pricelib
import debug
from human_readable import hr
import idlib
import MetaTrackerLib
from AssertApply import assert_apply_will_work
from confutils import confman


class ClientEGTPlibError(StandardError): pass
class NoServersError(ClientEGTPlibError): pass

# XXX this timeout is -way- too short for a modem (or dsl line that's in heavy use).
# (a 2 transaction protocol of "hello, will you accept a blob" followed by the
# actual publish would be a good way to solve this for the most common cases?)
def publish_blob(MTM, serverlist, blobid, blobdata, outcome_func, timeout=200, passalong=0) :
    """
    Sends the given blob to one of the servers in serverlist.  The outcome goes to "outcome_func".
    
    @param serverlist a list of (server id, server meta info dict)

    @return no return value

    @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
    @precondition `serverlist' must contain some servers.: len(serverlist) > 0: "serverlist: %s" % hr(serverlist)
    """
    assert idlib.is_sloppy_id(blobid), "`blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
    assert len(serverlist) > 0, "`serverlist' must contain some servers." + " -- " + "serverlist: %s" % hr(serverlist)

    # blobid is mojosixbit-encoded for this protocol:
    blobid = idlib.to_mojosixbit(blobid)

    msgtype = 'put blob'
    msgbody = {
        'blob id': str(blobid),
        'data': blobdata,
    }
    if passalong:
        msgbody['passalong'] = 1

    serverlist = MTM.get_handicapper().sort_by_preference(serverlist, msgtype, msgbody)

    if len(serverlist) <= 0 :
        debug.stderr.write("WARNING: publish_blob() says no servers in serverlist (after handicapping). blobid: %s\n", args=(blobid,))
        MTM._doQ.add_task(outcome_func, kwargs = {'widget': None, 'outcome': None, 'failure_reason': 'No acceptable block servers'})
        return

    outcome_func_refs = []  # XXX this is a hack! [inner_outcome_func doesn't seem to be able to refer to itself] -greg
    latency_timer_map = {}  # maps serverid's to the mojoutil.Timer() started when the `initiate()' to them was started
    
    # this callback will initiate a message to the next server in serverlist if the current outcome was a failure.
    # it calls the outer_outcome_func on success if there are no more servers in serverlist
    # XXX if it looks like an object, and it quacks like an object, it should be made into an object - bram 2001-01-19
    def inner_outcome_func(widget = None, outcome = None, failure_reason=None, MTM=MTM, outer_outcome_func=outcome_func, serverlist=serverlist, msgtype=msgtype, msgbody=msgbody, timeout=timeout, outcome_func_refs=outcome_func_refs, latency_timer_map=latency_timer_map, bloblen=len(blobdata)):
        if widget is not None:
            global_current_publication_servers.remove(widget.get_counterparty_id())
            for i in global_current_stalled_publications:
                MTM._doQ.add_task(i)
            del global_current_stalled_publications[:]
        if outcome and outcome.get('result') == 'success' :
            # break a circular reference to ourselves!  this was possibly causing a major memory leak on publishing -g/n
            del outcome_func_refs[0]
            # success! record the latency of this put blob for future reputation judgements
            server_id = widget.get_counterparty_id()
            timer = latency_timer_map[server_id]
            timer.stop()
            # store the thruput
            cp_obj = MTM._keeper.get_counterparty_object(server_id)
            cp_obj.update_custom_stat_weighted_sample('put blob bytes/sec', float(len(msgbody['data']))/timer.get())

            outer_outcome_func(widget=widget, outcome=outcome, failure_reason=failure_reason)
        else:
            if serverlist:
                preferenceno=0
                for i in xrange(len(serverlist)):
                    server_id, server_meta_info = serverlist[i]
                    if server_id not in global_current_publication_servers:
                        del serverlist[i]
                        break
                else:
                    global_current_stalled_publications.append(outcome_func_refs[0])
                    return
                fixedprice = pricelib.blob_server_publish_price(server_meta_info, msgbody)
                if fixedprice is not None:
                    # Remove this once we've grandfathered-out most of the fixed-price servers, since it allows servers to "steal" mojo...  --Zooko 2001-06-01
                    if msgbody.get('passalong', 0) > 0:  # it costs double if we ask the other end to pass it along
                        mojooffer = fixedprice * 2
                    else:
                        mojooffer = fixedprice
                else:
                    # Modern dynamic-pricing servers don't have an announced price at all, so it'll be `None' and we'll do this:  --Zooko 2001-06-01
                    mojooffer = BidStrategy.choose_bid('MO_PUT_BLOB', MTM, preferenceno=preferenceno)


                preferenceno=preferenceno+1
                global_current_publication_servers.append(server_id)
                # ... initiate_and_callback on the next server since the previous one didn't accept the block
                latency_timer_map[server_id] = mojoutil.Timer(start=1)
                MTM.initiate(server_id, msgtype, msgbody, mojooffer, outcome_func=outcome_func_refs[0], timeout=timeout)
            else:
                # no block server worked, pass the failure on up
                del outcome_func_refs[0]
                outer_outcome_func(widget=widget, outcome=outcome, failure_reason="couldn't publish this message to any known block server: %s" % hr(msgbody))

    outcome_func_refs.append(inner_outcome_func)
    inner_outcome_func()

global_current_publication_servers = []
global_current_stalled_publications = []


class BlobQueryInterface:
    """
    Broker should inherit from this class.  It provides a method for issuing block finding queries.
    """
    def do_you_have_blobs(self, counterparty_info_tuple, block_id_list, callback, allow_post_timeout_callback=None, allow_recursion=0, mojooffer=50):
        """
        Issue a 'do you have blobs' query to the given counterparty for the blocks in blob_id_list, the result will come via a callback.

        @param callback a callable that takes (server_id, server_info, block_ids) keyword arguments.
            It is called no matter whan with the response or after a timeout.  If allow_post_timeout_callback is true,
            it may be called one additional time after the message timeout incase a response comes in late.
        @param allow_post_timeout_callback is a boolean indicating if late responses should be handled.
        @param allow_recursion tells the counterparty if this request should be considered when doing block wholesaling

        TODO this should really just be a method of the Broker object.
        """
        assert (len(counterparty_info_tuple) == 2) and idlib.is_sloppy_id(counterparty_info_tuple[0]), ("bad counterparty_info_tuple", counterparty_info_tuple)
        assert type(block_id_list) in (types.ListType, types.TupleType), "block_id_list must be a list"
        assert len(block_id_list) > 0, "refusing to issue an empty find"
        assert_apply_will_work(callback, kwargs={'server_id':None, 'server_info':None, 'block_ids':None}), "callback must accept proper keyword parameters"

        server_id, server_info = counterparty_info_tuple
        timer = mojoutil.Timer()
        def do_you_have_blobs_callback(widget, outcome = None, failure_reason = None, callback = callback, server_info = server_info, timer=timer, self=self):
            server_id = widget.get_counterparty_id()
            if failure_reason is not None:
                outcome = []
            else:
                cp_obj = self.ServicesMTM._keeper.get_counterparty_object(server_id)
                cp_obj.update_custom_stat_weighted_sample('do you have blobs latency', timer.get())
            if type(outcome) is not types.ListType:  # i don't know how this happens but it does on rare occasions
                debug.mojolog.write("got strange 'do you have blobs response' from %s where outcome = %s\n", args=(server_id, outcome), v=2, vs='FindBeforehandBlockFetcher')
                outcome = []
            callback(server_id = server_id, server_info = server_info, block_ids = outcome)

        if allow_post_timeout_callback:
            post_timeout_callback = callback
        else:
            post_timeout_callback = None

        fixedprice = pricelib.blob_server_have_blobs_price(server_info, block_id_list)
        if fixedprice is not None:
            # This is an old fixed-price server, so we'll pay whatever it wants.
            # Remove this once we've grandfathered-out most of the fixed-price servers, since it allows servers to "steal" mojo...  --Zooko 2001-06-01
            mojooffer = fixedprice
        else:
            mojooffer = mojooffer

        # TODO better timeout tuning could be done (ie: an average 95%ile timeout with this
        # counterparty should be better, that should be true for most all message types)
        min_speed = min(int(confman.dict['CONN_SPEED_IN']), int(confman.dict['CONN_SPEED_OUT']))
        if min_speed <= 64:
            timeout_scaling_factor = 200
            maxtime = 360
        elif min_speed <= 128:
            timeout_scaling_factor = 400
            maxtime = 300
        else:
            # not scaling things too drastically here as we don't know about the other side of the connection
            timeout_scaling_factor = 600
            maxtime = 240
        timeout = min( max(90, (len(block_id_list) / timeout_scaling_factor) * 60), maxtime )

        msgbody = {'block id list': block_id_list}
        if not allow_recursion:
            msgbody['do not recurse'] = "true"

        timer.start()
        self.ServicesMTM.initiate(server_id, 'do you have blobs', msgbody, mojooffer, outcome_func=do_you_have_blobs_callback, post_timeout_outcome_func=post_timeout_callback, timeout=timeout)



def run():
    import RunTests
    RunTests.runTests(["ClientEGTPlib"])
    pass


#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
