2#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

# standard modules
import exceptions
import time # actually just for debugging
import traceback # actually just for debugging
import types

# our modules
import BidStrategy
import DataTypes
import OurMessages
import MetaTrackerLib
import debug
import idlib
import pricelib

from CommHints import HINT_EXPECT_RESPONSE, HINT_EXPECT_MORE_TRANSACTIONS, HINT_EXPECT_NO_MORE_COMMS, HINT_NO_HINT
from DataTypes import UNIQUE_ID, ANY, ASCII_ARMORED_DATA, NON_NEGATIVE_INTEGER, MOD_VAL, INTEGER, NON_NEGATIVE_FLOAT, ListMarker, OptionMarker
from MojoConstants import true, false, DEBUG_MODE
import MojoConstants
import MojoKey
import MojoMessage
import OurMessages
from OurMessagesCommStrat import *
import RunTests
import SEXP
import TCPConnection
from human_readable import hr
import ipaddresslib
import mencode
import modval
import mojoutil
import string



class Error(exceptions.StandardError): pass
class UnsupportedTypeError(exceptions.StandardError): pass


"""
TCP(host, port, asyncsock) - a strategy for opening tcp connections or else using a currently open tcp socket

Relay(relay_server) - makes a strategy which routes through the given relay server

Crypto(pubkey, lowerstrategy) - a combination of a pub key and another strategy for transmitting ciphertext

Pickup() - if you have a message for this person, store it and she will contact you and pick it up;  Relay servers use a Pickup strategy to give messages to their clients currently.

Dummy() - for testing
"""

def is_reachable(cs):
    # Find the bottommost strategy...
    while hasattr(cs, "_lowerstrategy"):
        cs = cs._lowerstrategy

    if (cs) and ((isinstance(cs, TCP) and (ipaddresslib.is_routable(cs.host))) or (isinstance(cs, Relay))):
        return true
    else:
        return false

def choose_best_strategy(cs1, cs2):
    """
    You should pass the newer, or most recently "suggested" cs as `cs2'.  If there is a tie we will prefer that one.
 
    @param cs1 an instance of CommStrat; Can be `None'.
    @param cs2 an instance of CommStrat; Can be `None'.

    @return a reference to whichever of `cs1' or `cs2' is most preferable
    """
    if cs1 is None:
        return cs2
    if cs2 is None:
        return cs1

    # We always prefer TCP over non-TCP...
    if isinstance(cs1, TCP) and not isinstance(cs2, TCP):
        return cs1
    if isinstance(cs2, TCP) and not isinstance(cs1, TCP):
        return cs2

    # Among TCPs, we prefer open sockets over IP addresses:
    if isinstance(cs1, TCP) and isinstance(cs2, TCP):
        if ((cs1.asyncsock) and not cs1.asyncsock.is_closing()) and ((not cs2.asyncsock) or (cs2.asyncsock.is_closing())):
            return cs1
        if ((cs2.asyncsock) and not cs2.asyncsock.is_closing()) and ((not cs1.asyncsock) or (cs1.asyncsock.is_closing())):
            return cs2

        # Among open sockets, we prefer the newer one if it is a different socket (sort of arbitrary).
        if (cs1.asyncsock) and (not cs1.asyncsock.is_closing()) and (cs2.asyncsock) and (not cs2.asyncsock.is_closing()):
            if not cs2.same(cs1):
                return cs2
            else:
                return cs1

        # Among IP addresses, we definitely prefer routeable ones...
        if ipaddresslib.is_routable(cs1.host) and not ipaddresslib.is_routable(cs2.host):
            return cs1
        if ipaddresslib.is_routable(cs2.host) and not ipaddresslib.is_routable(cs1.host):
            return cs2

        # Among routeable IP addresses, we prefer the newer one if it is a different address (sort of arbitrary).
        if not cs2.same(cs1):
            return cs2
        else:
            return cs1

    # Note: if we reach here then neither strat is a TCP.
    # We always prefer Relay over "none of the above":
    if isinstance(cs1, Relay) and not isinstance(cs2, Relay):
        return cs1
    if isinstance(cs2, Relay) and not isinstance(cs1, Relay):
        return cs2

    # Among Relays, we prefer the newer one if it is different (sort of arbitrary):
    # Note, it would be nice to check if we have an open TCP connection to one of the relay servers, and
    # if so prefer that Relay strat.  Also it might be nice to actually handicap the Relay servers here.
    if not cs2.same(cs1):
        return cs2
    else:
        return cs1

class CommStrat:
    def __init__(self, broker_id=None):
        """
        @precondition `broker_id' must be `None' or an id.: (broker_id is None) or (idlib.is_sloppy_id(broker_id)): "broker_id: %s" % hr(broker_id)
        """
        assert (broker_id is None) or (idlib.is_sloppy_id(broker_id)), "precondition: `broker_id' must be `None' or an id." + " -- " + "broker_id: %s" % hr(broker_id)

        self.hint = HINT_NO_HINT
        self.hintnumexpectedresponses = 0
        self.hintnumexpectedsends = 0
        self._broker_id = broker_id 

    def toDict(self):
        return { 'broker id': self._broker_id }

#### XXXX I need to split OpenConnection strategy from TCP strategy.  For example, if you are selling metatracker services, you need to remember your counterparty's TCP strategies and sell those, but filter out your open connections to them.  --Zooko 2000-09-08

class TCP(CommStrat):
    def __init__(self, host=None, port=None, asyncsock=None, broker_id=None):
        """
        @param asyncsock an instance of TCPConnection

        @precondition `asyncsock' must be an instance of TCPConnection or nothing.: (asyncsock is None) or isinstance(asyncsock, TCPConnection.TCPConnection): "asyncsock: %s :: %s" % (hr(asyncsock), hr(type(asyncsock)))
        """
        assert (asyncsock is None) or isinstance(asyncsock, TCPConnection.TCPConnection), "`asyncsock' must be an instance of TCPConnection or nothing." + " -- " + "asyncsock: %s :: %s" % (hr(asyncsock), hr(type(asyncsock)))

        CommStrat.__init__(self, broker_id)

        self.host = host
        self.port = port
        self.asyncsock = asyncsock
    
    def __repr__(self):
        return '<%s to %s:%s via %s at %x>' % (self.__class__.__name__, self.host, self.port, self.asyncsock, id(self))

    def same(self, other):
        """
        Two TCP's are same iff they both have the same non-None socket object or if they both have None socket objects and the same host and port

        @return `true' iff `self' and `other' are actually the same strategy
        """
        if not hasattr(other, 'asyncsock'):
            assert hasattr(self, 'asyncsock')
            return false
        if self.asyncsock is not None:
            return self.asyncsock is other.asyncsock
        else:
            if other.asyncsock is not None:
                return false
        if not hasattr(other, 'host'):
            assert hasattr(self, 'host')
            return false
        if not hasattr(other, 'port'):
            assert hasattr(self, 'port')
            return false
        return (self.host == other.host) and (self.port == other.port)

    def toDict(self):
        d = CommStrat.toDict(self)

        d['comm strategy type'] = "TCP"

        if self.host:
            d['IP address'] = self.host

        if self.port:
            d['port number'] = `self.port`

        if self.asyncsock:
            d['open connection'] = 'true'
            
            try:
                peername = self.asyncsock.getpeername()
                d['open connection peername'] = `peername`
            except:
                pass

        return d

class Relay(CommStrat):
    def __init__(self, relayer_id, broker_id, mtm):
        """
        @precondition `relayer_id' must be an id.: idlib.is_sloppy_id(relayer_id): "relayer_id: %s" % hr(relayer_id)
        """
        assert idlib.is_sloppy_id(relayer_id), "precondition: `relayer_id' must be an id." + " -- " + "relayer_id: %s" % hr(relayer_id)

        CommStrat.__init__(self, broker_id)

        self._relayer_id = relayer_id
        self._mtm = mtm

        if DEBUG_MODE:
            self.debcr = traceback.extract_stack()

    def __repr__(self):
        return '<%s to %s via %s at %x>' % (self.__class__.__name__, hr(self._broker_id), hr(self._relayer_id), id(self))

    def same(self, other):
        """
        Two Relay's are same iff they have the same `_relayer_id'.

        @return `true' iff `self' and `other' are actually the same strategy
        """
        if not hasattr(other, '_relayer_id'):
            assert hasattr(self, '_relayer_id')
            return false
        return idlib.equal(self._relayer_id, other._relayer_id)

    def toDict(self):
        d = CommStrat.toDict(self)
        d['comm strategy type'] = "relay" # XXXX This should be changed to "Relay" to match the name of the class.  --Zooko 2000-08-02
        d['relayer id'] = self._relayer_id
        return d

    def send(self, msg, mojooffer, hint=HINT_NO_HINT, fast_fail_handler=None, timeout=None):
        timeout = long(timeout)

        if idlib.equal(self._relayer_id, self._mtm.get_id()):
            fast_fail_handler(failure_reason="aborting a recursive `pass this along' to myself")
            return

        if idlib.equal(self._relayer_id, self._broker_id):
            debug.mojolog.write("Weird: we have gotten the idea that the comm strat for sending messages to %s is to relay through herself!  CommStrat.Relay: %s\n", args=(self._broker_id, self,), v=2, vs="debug")
            fast_fail_handler(failure_reason="aborting a recursive `pass this along' to %s via herself" % hr(self._relayer_id,))
            return

        if DEBUG_MODE:
            assert self._mtm is not None, "self: %s, self.debcr: %s" % (hr(self), hr(self.debcr))
        assert self._mtm is not None, "self: %s" % hr(self)

        # Why do we do this `MetaTrackerLib.lookup()' first instead of letting MojoTransaction handle that?  Because we need to know the price.
        # Perhaps the price should be managed separately and passed in (the way it is done for all other `initiate()' invocations).  This would also eliminate the aforementioned fraud issue.  --Zooko 2001-05-14
        # This is solved by dynamic pricing.  Once we don't care to interoperate with old fixed-price relay servers, we can take out this extra visit to MTL.lookup.  --Zooko 2001-06-01
        def after_lookup(widget, outcome, failure_reason=None, self=self, msg=msg, hint=hint, fast_fail_handler=fast_fail_handler, mojooffer=mojooffer, timeout=timeout):
            if failure_reason or (outcome.get('result') != "success"):
                fast_fail_handler(failure_reason="couldn't find contact info for relay server, ll failure_reason: %s, ll outcome: %s" % (hr(failure_reason), hr(outcome)))
                return

            if not DataTypes.is_template_matching(outcome, OurMessages.PHONEBOOK_ENTRY):
                fast_fail_handler(failure_reason="couldn't get contact info for relay server")

            encappedmsgbody = {'recipient': idlib.to_mojosixbit(self._broker_id), 'message': msg}
            relayfixedprice = pricelib.relay_server_pass_along_price(outcome, encappedmsgbody)
            if relayfixedprice is not None:
                # Remove this once we've grandfathered-out most of the fixed-price servers, since it allows servers to "steal" mojo...  --Zooko 2001-06-01
                mojooffer = relayfixedprice
            else:
                mojooffer = BidStrategy.choose_bid("MO_PASS_THIS_ALONG", self._mtm)

            def outcome_func_from_pass_this_along(widget, outcome, failure_reason=None, self=self, msg=msg):
                assert idlib.equal(widget.get_counterparty_id(), self._relayer_id)
                debug.mojolog.write("CommStrat.Relay: Got result of `pass this along'.  self._relayer_id: %s, widget: %s, outcome: %s, failure_reason: %s\n", args=(self._relayer_id, widget, outcome, failure_reason,), v=3, vs="commstrats")
                if (not failure_reason) and (outcome.get('result') != "ok") and (outcome.get('result') != "success"):
                    self._mtm.forget_comm_strategy(self._broker_id, idlib.make_id(msg, 'msg'), outcome=outcome, failure_reason="got failure from relay server: %s" % hr(outcome))
                if failure_reason:
                    self._mtm.forget_comm_strategy(self._broker_id, idlib.make_id(msg, 'msg'), outcome=outcome, failure_reason="couldn't contact relay server: %s" % hr(outcome))

            debug.mojolog.write("CommStrat.Relay: Initiating `pass this along'...  self._relayer_id: %s, relayfixedprice: %s\n", args=(self._relayer_id, relayfixedprice,), v=3, vs="commstrats")

            self._mtm.initiate(self._relayer_id, 'pass this along v2', encappedmsgbody, mojooffer, outcome_func=outcome_func_from_pass_this_along, timeout=timeout * 3)

        debug.mojolog.write("CommStrat.Relay: Looking up relayer contact info...  self._relayer_id: %s\n", args=(self._relayer_id,), v=3, vs="commstrats")
        MetaTrackerLib.nonblocking_get_contact_info(self._mtm, self._relayer_id, callback=after_lookup)

class Crypto(CommStrat):
    def __init__(self, pubkey, lowerstrategy, broker_id=None):
        """
        @param lowerstrategy the lower-level comms strategy, either given by meta-tracking or
            suggested by the way that the last message arrived (e.g. For TCP, the suggested
            strategy is to send a message back down the connection over which the last message
            arrived.)

        @precondition `pubkey' must be a well-formed MojoKey.: MojoKey.publicKeyForCommunicationSecurityIsWellFormed(pubkey)
        @precondition `lowerstrategy' must be a CommStrat.: isinstance(lowerstrategy, CommStrat): "lowerstrategy: %s" % hr(lowerstrategy)
        """
        assert MojoKey.publicKeyForCommunicationSecurityIsWellFormed(pubkey), "`pubkey' must be a well-formed MojoKey."
        assert isinstance(lowerstrategy, CommStrat), "`lowerstrategy' must be a CommStrat." + " -- " + "lowerstrategy: %s" % hr(lowerstrategy)

        CommStrat.__init__(self, broker_id)

        self._pubkey = pubkey
        self._lowerstrategy = lowerstrategy

    def __repr__(self):
        return '<%s pubkey_id %s at %x, lowerstrategy: %s>' % (self.__class__.__name__, idlib.to_ascii(idlib.make_id(self._pubkey)), id(self), self._lowerstrategy)

    def same(self, other):
        """
        Two Crypto's are same iff they have the same pub key and their lowerstrategies are same.

        @return `true' iff `self' and `other' are actually the same strategy
        """
        if not hasattr(other, '_pubkey'):
            return false
        if self._pubkey != other._pubkey:
            return false
        if not hasattr(other, '_lowerstrategy'):
            return self._lowerstrategy.same(other._lowerstrategy)

    def toDict(self):
        d = CommStrat.toDict(self)
        d['comm strategy type'] = "crypto" # XXXX This should be changed to "Crypto" to match the name of the class.  --Zooko 2000-08-02
        d['pubkey'] = mencode.mdecode(self._pubkey)
        d['lowerstrategy'] = self._lowerstrategy.toDict()
        return d

class Pickup(CommStrat):
    def __init__(self, broker_id=None):
        CommStrat.__init__(self, broker_id)
        return

    def toDict(self):
        d = CommStrat.toDict(self)
        d['comm strategy type'] = "pickup" # XXXX This should be changed to "Pickup" to match the name of the class.  --Zooko 2000-08-02
        return d

    def same(self, other):
        """
        Two Pickup's are always the same.

        @return `true' iff `self' and `other' are actually the same strategy
        """
        return isinstance(other, Pickup)

class Dummy(CommStrat):
    def __init__(self):
        CommStrat.__init__(self)

def dict_to_strategy(dict, mtm=None, broker_id=None):
    return dictToStrategy(dict, mtm=mtm, broker_id=broker_id)

def dictToStrategy(dict, mtm=None, broker_id=None):
    """
    @raises an UnsupportedTypeError if `dict' is not either a TCP, Relay, Crypto, or Pickup
    
    @precondition `broker_id' must be an id or None.: (broker_id is None) or (idlib.is_sloppy_id(broker_id)): "broker_id: %s :: %s" % (hr(broker_id), hr(type(broker_id)))

    @deprecated in favor of `dict_to_strategy()' for naming reasons
    """
    assert (broker_id is None) or (idlib.is_sloppy_id(broker_id)), "precondition: `broker_id' must be an id or None." + " -- " + "broker_id: %s :: %s" % (hr(broker_id), hr(type(broker_id)))

    if not ((dict.get('comm strategy type') == "TCP") or (dict.get('comm strategy type') == "relay") or (dict.get('comm strategy type') == "Relay") or (dict.get('comm strategy type') == "crypto") or (dict.get('comm strategy type') == "Crypto") or (dict.get('comm strategy type') == "pickup")) or (dict.get('comm strategy type') == "Pickup"):
        raise UnsupportedTypeError, "`dict' must be either a TCP, Relay, Crypto or Pickup." + " -- " + "dict: [%s]" % hr(dict)

    MojoMessage.checkTemplate(dict, COMM_STRAT_TEMPL)

    dictbroker_id = dict.get('broker id')
    if (broker_id is not None) and (dictbroker_id is not None):
        assert idlib.equal(broker_id, dictbroker_id)
    if broker_id is None:
        broker_id = dictbroker_id

    cst = dict['comm strategy type']
    if cst == "TCP":
        MojoMessage.checkTemplate(dict, TCP_COMM_STRAT_TEMPL)
        return TCP(dict['IP address'], int(dict['port number']), broker_id=broker_id)

    if (cst == "Relay") or (cst == "relay"):
        MojoMessage.checkTemplate(dict, RELAY_COMM_STRAT_TEMPL)
        return Relay(relayer_id=dict['relayer id'], mtm=mtm, broker_id=broker_id)

    if (cst == "Crypto") or (cst == "crypto"):
        MojoMessage.checkTemplate(dict, CRYPTO_COMM_STRAT_TEMPL)
        return Crypto(mencode.mencode(dict['pubkey']), dictToStrategy(dict['lowerstrategy'], broker_id=broker_id, mtm=mtm), broker_id=broker_id)

    if (cst == "Pickup") or (cst == "pickup"):
        MojoMessage.checkTemplate(dict, PICKUP_COMM_STRAT_TEMPL)
        return Pickup(broker_id=broker_id)

    raise Error, "Unknown strategy %s." % `cst`

def test_relay_reserialize():
    mv = modval.new_random(MojoConstants.SIZE_OF_MODULAR_VALUES, 3)
    key = MojoKey.makePublicRSAKeyForCommunicating(mv)
    rs = Relay(idlib.string_to_id(key))
    d1 = rs.toDict()
    rs2 = dictToStrategy(d1)
    d2 = rs2.toDict()
    assert str(d1) == str(d2)

def test_tcp_reserialize():
    rs = TCP(host = '10.5.222.9', port = 53)
    d1 = rs.toDict()
    rs2 = dictToStrategy(d1)
    d2 = rs2.toDict()
    assert str(d1) == str(d2)

if __name__ == '__main__':
    RunTests.runTests('CommStrat')

mojo_test_flag = 1

