#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import exceptions
import types
import threading
import traceback

### our modules
import CommStrat
from CommHints import HINT_EXPECT_RESPONSE, HINT_EXPECT_MORE_TRANSACTIONS, HINT_EXPECT_NO_MORE_COMMS, HINT_NO_HINT
from MojoConstants import true, false
import SingleThreader
import debug
from human_readable import hr
import idlib
import mojoutil


class Error(exceptions.StandardError): pass

class CannotSendError(Error): pass

class CannotListenError(Error): pass # for TCPCommsHandler, this means "couldn't bind to a port".


def example_inmsg_handler(counterparty_id, msg, commstrat=None):
    """
    @param commstrat a comms strategy that is suggested by the fact of the incoming message (for example, if someone opens a TCP connection to you then this will be a CurrentTCPConnection strat, enabling you to use that open connection for future comms, or if someone sends a message to you through a relay server then this will be a Relay strat).  This enables higher layers to decide whether or not to use the new strategy based on things like crypto (see CryptoCommsHandler) or economics (see MojoTransaction).  You can also just ignore it.
    """
    assert 0, "this class is only an interface -- not to be invoked itself"


class CommsHandler:
    """
    You can subclass CommsHandler and implement `send_msg()', `start_listening()', `get_id()',
    `use_comm_strategy()', and `get_comm_strategy()'.

    Or you can subclass BaseCommsHandler and implement `_internal_send_msg()', optionally
    `_activate_commstrat()', and optionally `get_comm_strategy()' and `get_id()'.

    The point of the `_internal_spam()' methods is that the BaseCommsHandler object will lazily
    invoke `_activate_commstrat()' only when it is needed (i.e. just before it tries to
    send).
    """
    def shutdown(self):
        """
        This gets called after stop_listening() when someone wants to
        shut this comms handler down completely never to be used again
        (in preparation for deleting the object; this should perform
        nice cleanup on any references held and break all circular ones)
        """
        assert 0, "this class is only an interface -- not to be invoked itself"

    def send_msg(self, counterparty_id, msg, hint=HINT_NO_HINT):
        """
        Non-blocking.

        @param hint one of the HINT_* constants indicating how to optimize the lower level communication system for this situation

        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id)
        @precondition `msg' must be a string.: type(msg) == types.StringType
        """
        assert 0, "this class is only an interface -- not to be invoked itself"
      

    def stop_listening(self):
        assert 0, "this class is only an interface -- not to be invoked itself"

        
    def start_listening(self, inmsg_handler_func):
        """
        Start trying to get incoming messages which you will then pass to `inmsg_handler_func()'.

        @param inmsg_handler_func the function to be called whenever a message for us comes in
        """
        assert 0, "this class is only an interface -- not to be invoked itself"


    def get_id(self):
        """
        Return your id.
        """
        assert 0, "this class is only an interface -- not to be invoked itself"


    def use_comm_strategy(self, counterparty_id, strategy, orig_cs=None) :
        """
        @param orig_cs may optionally be the original metainfo representation of the comm strat.
                       it will be used in fast fail handlers for reporting with comm strat failed.
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id)
        @precondition `strategy' must be an instance of CommStrat.Crypto.: isinstance(strategy, CommStrat.Crypto)
        """
        assert 0, "this class is only an interface -- not to be invoked itself"


    def get_comm_strategy(self):
        """
        Return a comm strategy that can be used to talk to you.
        """
        assert 0, "this class is only an interface -- not to be invoked itself"

# class BaseCommsHandler(CommsHandler, SingleThreader.BombingSingleThreader):
class BaseCommsHandler(CommsHandler, SingleThreader.SingleThreader):
    """
    This is a base class providing common functionality for comms handlers.  Historical note: this was originally created by factoring common functionality out of CryptoCommsHandler and RelayCommsHandler, and to a lesser extent TCPCommsHandler and DummyCommsHandler.)
    """
    def __init__(self, llch=None):
        # SingleThreader.BombingSingleThreader.__init__(self)
        SingleThreader.SingleThreader.__init__(self)

        # lower-level comms handler
        # primitive comms handlers like TCPCommsHandler and DummyCommsHandler don't have a llch, and need to override `use_comm_strategy()' and `start_listening()'.
        self._llch = llch

        # map cid to instance of CommStrat
        self._cid_to_cs = {}
        # map cid to "not yet"/"yes"
        self._cids_to_activation = {}
    
    def shutdown(self):
        if hasattr(self, '_llch') and self._llch is not None:
            self._llch.shutdown()
        if hasattr(self, '_cid_to_cs'):
            del self._cid_to_cs

    def send_msg(self, counterparty_id, msg, hint=HINT_NO_HINT, fast_fail_handler=None, timeout=None):
        """
        Non-blocking.

        Subclassers of BaseCommsHandler want to override `_internal_send_msg()', not `_send_msg()'.

        @precondition `msg' must be a string.: type(msg) == types.StringType: "msg: %s" % hr(msg)
        """
        assert type(msg) == types.StringType, "`msg' must be a string." + " -- " + "msg: %s" % hr(msg)
        counterparty_id = idlib.to_binary(counterparty_id)  # this also checks the precondition of counterparty_id being an id

        self._enter()
        try:
            # debug.stderr.write("BaseCommsHandler.send_msg()\n")
            self._ensure_comm_strategy_has_activated(counterparty_id)

            self._internal_send_msg(counterparty_id, msg, hint=hint, fast_fail_handler=fast_fail_handler, timeout=timeout)
        finally:
            self._exit()

    def stop_listening(self):
        self._enter()
        try:
            self._llch.stop_listening()
        finally:
            self._exit()

    def is_listening(self):
        self._enter()
        try:
            self._llch.is_listening()
        finally:
            self._exit()

    def start_listening(self, inmsg_handler_func):
        """
        Start trying to get incoming messages which you will then pass to `inmsg_handler_func()'.

        @param inmsg_handler_func the function to be called whenever a message for us comes in
        """
        self._enter()
        try:
            self._upward_inmsg_handler = inmsg_handler_func
            # start listening
            self._llch.start_listening(inmsg_handler_func = self._inmsg_handler)
        finally:
            self._exit()

    def _ensure_comm_strategy_has_activated(self, counterparty_id):
        """
        idempotent
        """
        counterparty_id = idlib.to_binary(counterparty_id)
        self._enter()
        try:
            if self._cids_to_activation.get(counterparty_id) == "not yet":
                try:
                    cs = self._cid_to_cs[counterparty_id]
                except KeyError:
                    debug.mojolog.write("%s._cis_to_cs did not contain %s (%s)\nkeys: %s\n" % (self, `counterparty_id`, hr(counterparty_id), `self._cid_to_cs.keys()`))
                    raise
                self._activate_commstrat(counterparty_id, cs)
                # QUERY: but what if _activate_commstrat failed? (such as a TCP connection that couldn't
                # be established)  This would mean that we think we have an active connection to them when we don't.

                # Zooko responds:  We don't have any notion of an active connection in the
                # higher-layer code.  There is, for example, no "connection problem" error, only
                # a "could not send" error.  You just tell the comms handler to use a connection
                # strategy and then you try to send and the earliest that you can find out that
                # the strategy is bad is when you can't send.  This is actually considered to be
                # a feature-not-a-bug -- our code for finding new connection strategies is
                # executed only lazily (when we try to send and fail), rather than eagerly
                # (whenever we get a connection strategy).  So for example you can remember for
                # weeks and weeks that the connection strategy to talk to <EZR> is a certain TCP
                # address, and every day or two you send <EZR> a message, and fortunately <EZR>
                # happens to be on-line each time, so the message gets through.  Now the fact is
                # that <EZR> is off-line most of the rest of the time and trying to activate the
                # connection strategy would fail, but you don't care.  --Zooko 2000-07-22

                self._cids_to_activation[counterparty_id] = "yes"
        finally:
            self._exit()

    def _activate_commstrat(counterparty_id, cs):
        """
        The base implementation calls down to the llch, if it is there.

        @precondition `cs' must be an instance of CommStrat.CommStrat.: isinstance(cs, CommStrat.CommStrat): "cs: %s" % hr(c)
        """
        assert isinstance(cs, CommStrat.CommStrat), "`cs' must be an instance of CommStrat.CommStrat." + " -- " + "cs: %s" % hr(c)

        self._enter()
        try:
            if self._llch:
                return self._llch.use_comm_strategy(counterparty_id, cs, orig_cs=cs.__dict__.get('_orig_top_cs'))
        finally:
            self._exit()

    def use_comm_strategy(self, counterparty_id, strategy, orig_cs=None) :
        """
        You probably don't want to override this -- instead you want to override `_activate_commstrat()' to actually do any work necessary (e.g. open a TCP connection)  `_activate_commstrat()' will be called lazily, and no more than once per call to `use_comm_strategy()'.

        Note: we're going to use naming magic here:  if `strategy' is an instance of a class named CommStrat.Foo, and the self object is an instance of a class named FooCommsHandler, then we store this strategy object here for later activation.  Else, we pass it down to the llch.

        @param orig_cs (optional) the original metainfo representation of the comm strat;  It
            will be used in fast fail handlers for reporting with comm strat failed.
        """
        counterparty_id = idlib.to_binary(counterparty_id)  # this also checks the precondition of counterparty_id being an id

        debug.stderr.write("CommsHandler(%s).use_comm_strategy(%s, %s -- %s)\n", args=(self, counterparty_id, strategy, strategy.toDict()), v=21, vs="commstrats")

        self._enter()
        try:
            # XXX TODO this is utterly incomprehensible code!
            # (and looks like a "very bad idea" to me to depend on the module and class naming for functionality linking!)

            # Zooko says: <shrug>  If you want you want to replace this with "if isinstance(self, CryptoCommsHandler) and isinstance(strategy, CommStrat.Crypto): self.use_comm_strategy(strategy)" and so forth for all types of CommStrat, I don't mind.  Maybe in the future there will be a good way to put the "handler" functionality right into the CommStrat object so it will be a simple method invocation, using the Python interpreter to do the "if it is type X then invoke method X.use_comm_strategy()" logic.  But I already tried to write it like that (with the handler functionality in the Strat class) and failed once.  --Zooko 2000-07-23
            # P.S.:  There is also the __doc__ string above.  --Zooko 2000-07-26

            # There is much simpler way; it's called a dict of function pointers: use_comm_strat_func_dict[X]()  -greg

            # Oh yeah, good idea...  Would it be easier to read...  ?  --Zooko 2000-08-16

            stratCName = strategy.__class__.__name__
            stratMName = strategy.__class__.__module__
            csCName = self.__class__.__name__

            x = stratCName + "CommsHandler"

            if orig_cs is not None:
                strategy._orig_top_cs = orig_cs  # XXX HACK - adds a member to the object for use in MojoTransaction._ffh()

            if (stratMName == "CommStrat") and (x == csCName):
                self._cid_to_cs[counterparty_id] = strategy
                self._cids_to_activation[counterparty_id] = "not yet"

                if hasattr(strategy, '_lowerstrategy'):
                    self._llch.use_comm_strategy(counterparty_id, strategy._lowerstrategy, orig_cs=orig_cs)
            else:
                #### XXXX !!!! special exception temporary until RelayCommsHandler is implemented.  --Zooko 2000-07-24
                if isinstance(strategy, CommStrat.Relay):
                    return
                if not hasattr(self._llch, 'use_comm_strategy'):
                    debug.mojolog.write('GGG not making bad call to use_comm_strategy\n', v=3, vs='commstrats')
                    debug.mojolog.write('GGG the callstack that got here was: %s\n' % [traceback.extract_stack()], v=3, vs='commstrats')
                    debug.mojolog.write('GGG self is %s, strategy is %s, counterparty_id %s\n', args=[self, strategy, counterparty_id], v=3, vs='commstrats')
                    return
                self._llch.use_comm_strategy(counterparty_id, strategy, orig_cs=orig_cs)
        finally:
            self._exit()

    def get_comm_strategy(self):
        """
        Return a comm strategy that can be used to talk to you.

        The base implementation calls down to the llch, if it is there.

        @precondition This CH must have already started listening.: self._upward_inmsg_handler
        """
        assert self._upward_inmsg_handler, "This CH must have already started listening."

        self._enter()
        try:
            if self._llch:
                return self._llch.get_comm_strategy()
        finally:
            self._exit()

    def get_id(self):
        """
        The base implementation calls down to the llch, if it is there.
        """
        self._enter()
        try:
            if self._llch:
                return self._llch.get_id()
        finally:
            self._exit()
        

def _msgcb_warper(icb):
    import threading

    reswrapper={
        'msg': None,
        'cid': None,
        'cs': None,
        }
    doneflag = threading.Event()
    def _wcb(counterparty_id, msg, commstrat, doneflag=doneflag, reswrapper=reswrapper, icb=icb):
        icb(counterparty_id, msg, commstrat)
        reswrapper['msg'] = msg
        reswrapper['cid'] = counterparty_id
        reswrapper['cs'] = commstrat
        doneflag.set()

    return reswrapper, doneflag, _wcb
  
    
def _msgcb_maker():
    return _msgcb_warper(lambda counterparty_id, msg, commstrat: counterparty_id)

  

