#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import traceback
from string import atof
import time
from whrandom import randint

### our modules
import BidStrategy
from BlockFinder import BlockFinder
from BlockThrottler import BlockThrottler
from BlockWholesaler import WHOLESALER_FLAG
from bsddb3 import dbtables, db
import contentlib
import debug
import idlib
import mojosixbit
from human_readable import hr

import mojoutil
from mojoutil import Counter
from SingleReassembly import SingleReassembly, ReassemblyError
from sharemap import parse_sharemap, parse_dinode

import pricelib


# constant, determines the weight of new entries 
RELIABILITY_WEIGHT = 0.25

# how unreliable before we delete the entry...
RELIABILITY_THRESHOLD = 0.15

# check the complete database at most every X days
# may take longer...
CHECK_DURATION = (24 * 60 * 60) * 2  # two days

# number of dinodes to check at once
CHUNK_SIZE = 10

# min delay - minimum delay between verifier sessions
MIN_DELAY = 60 * 15  # 15 minutes



# This is the main class.  It handles the scheduling of VerifierSessions
# it also maintains the retrieability database and deletes broken mojoids
class ContentVerifier:
    def __init__(self, broker, contentdb):
        # (XXX hack) this allows the verification db to be within the same directory.
        # but remember to always close it before calling contentdb.close()!
        self.contentdb = contentdb.db 
        # method to delete a mojoid
        self.DELETE_MOJOID = contentdb.DeleteMojoid
        self.MARK_MOJOID_MISSING = contentdb.MarkMojoidMissing
        self.UNMARK_MOJOID_MISSING = contentdb.UnMarkMojoidMissing
        self.broker = broker
        self.MTM = broker.ServicesMTM

        # average amount of time for each VerifySession to finish
        self.avg_spent = 0

        # open our availability history DB
        try:
            self.avail_db = db.DB(self.contentdb.env)
            self.avail_db.open('avail.db', db.DB_HASH, db.DB_CREATE | db.DB_THREAD)
        except (db.DBError, IOError), e:
            debug.mojolog.write("WARNING: couldn't start the content tracker entry verifier, couldn't open DB:\n", vs='ContentVerifier')
            self.avail_db = None
            traceback.print_exc(file=debug.mojolog)
    
    def __del__(self):
        if hasattr(self, 'avail_db') and self.avail_db is not None:
            self.avail_db.close()
            del self.avail_db

    def shutdown(self):
        if self.avail_db is not None:
            self.avail_db.close()
            self.avail_db = None
        
    # gets a snapshot of the content tracker database
    def _fetch_snapshot(self):
        debug.mojolog.write('_fetch_snapshot\n', v=3, vs='ContentVerifier')
        l = self.contentdb.Select('MOJOID', ['__DINODE__', '__MOJOID__'], {'__DINODE__':dbtables.Cond()})
        return mojoutil.rotatelist(l)

    
    # starts a complete search through the database
    def begin_cycle(self):
        debug.mojolog.write('begin_cycle self.avail_db is %s\n', args=(self.avail_db,), v=3, vs='ContentVerifier')
        if self.avail_db is not None:
            self.start_time = time.time()
            self.dinodes = self._fetch_snapshot()
            self.MTM._doQ.add_task(self.check_some_dinodes, delay=MIN_DELAY)

    # returns the next chunk of dinodes
    def get_more_dinodes(self):
        dinodes, self.dinodes = self.dinodes[:CHUNK_SIZE], self.dinodes[CHUNK_SIZE:]
        debug.mojolog.write('get_more_dinodes returning %s\n', args=(len(dinodes),), v=3, vs='ContentVerifier')
        return dinodes

    # starts up a new VerifierSession and feeds it the next chunk of dinodes
    def check_some_dinodes(self):
        debug.mojolog.write('check_some_dinodes\n', v=3, vs='ContentVerifier')
        self.session_start = time.time()
        dinodes = self.get_more_dinodes()
        if dinodes:
            session = VerifierSession(self.broker, self.session_finished,
                                      self.update_mojoid_unavail, self.update_mojoid_avail)
            session.find_dinodes(dinodes)
        else:
            self.session_finished()
                              
    # VerifierSession calls this when it is done
    def session_finished(self):
        # schedule the next session
        spent = time.time() - self.session_start
        if len(self.dinodes):
            delay = self._delay_for_next_session(spent)
            self.MTM._doQ.add_task(self.check_some_dinodes, delay=delay)
        else:
            debug.mojolog.write("Finished checking content tracker database availability. it took: %s days.\n" %
                                ((time.time() - self.start_time) / 60.0 / 60.0 / 24.0))
            # start up the next cycle!
            self.MTM._doQ.add_task(self.begin_cycle, delay=self._time_left())

    def _time_left(self):
        return CHECK_DURATION - (time.time() - self.start_time)
    
    # retuns how long to wait before starting the next session
    # spent is the length in seconds of the duration of the previous session
    def _delay_for_next_session(self, spent):
        time_left = self._time_left()
        if time_left < 0:
            delay =  MIN_DELAY
        else:
            avg_spent = self._avg_time_spent(spent)
            chunks_left = max((len(self.dinodes) / 10) * 1.0, 1.0)
            delay = (time_left - (avg_spent * chunks_left)) / chunks_left
            if delay < MIN_DELAY:
                delay = MIN_DELAY
        debug.mojolog.write("DELAY = %s\n" % delay, v=3, vs='ContentVerifier')
        return delay

    # returns the average amount of time each VerifierSession has taken
    def _avg_time_spent(self, spent):
        if self.avg_spent == 0:
            self.avg_spent = spent
        self.avg_spent = (0.25 * self.avg_spent) + ((1.0 - 0.25) * spent)
        return self.avg_spent
    
    # used by reliability stat
    def _increment_filter(self, x):
        return RELIABILITY_WEIGHT + (1 - RELIABILITY_WEIGHT) * x

    def _decrement_filter(self, x):
        return (1 - RELIABILITY_WEIGHT) * x

    # returns the availability statistic for a mojoid
    def availability_for_mojoid(self, mojoid):
        return atof(self.avail_db.get(mojoid, '0.5'))

    # called by VerifierSession when it successfully locates all the pieces in a mojoid
    # updates the mojoid's availability stat
    def update_mojoid_avail(self, mojoid):
        avail = self.availability_for_mojoid(mojoid)
        self.avail_db.put(mojoid, `self._increment_filter(avail)`)
        self.UNMARK_MOJOID_MISSING(mojoid)
        return self.availability_for_mojoid(mojoid)

    # called by VerifierSession when it fails to locate any of the pieces in a mojoid
    # updates the stat and deletes bunk mojoids
    def update_mojoid_unavail(self, mojoid):
        avail = self.availability_for_mojoid(mojoid)
        self.avail_db.put(mojoid, `self._decrement_filter(avail)`)
        self.MARK_MOJOID_MISSING(mojoid)
        if self.availability_for_mojoid(mojoid) < RELIABILITY_THRESHOLD:
            # unretrievable, delete it
            debug.mojolog.write("Deleting unretrievable MojoID: %s with retrievability of %s.\n" %
                                (idlib.to_ascii(mojoid), self.availability_for_mojoid(mojoid)), v=2, vs='ContentVerifier')
            self.DELETE_MOJOID(idlib.to_ascii(mojoid))
        return self.availability_for_mojoid(mojoid)


# this verifies a set of dinodes
class VerifierSession:
    ## db = the contenttracker db (broker.content_tracker.contentdb.db)
    def __init__(self, broker, callback, mojoid_failed, mojoid_succeeded):
        """
        mojoid_failed, mojoid_succeeded : funciotns taking a mojoid
        called when finding the mojoids's pieces fail or succeed
        """
        self.broker = broker
        self.MTM = broker.ServicesMTM

        self.callback = callback
        self.mojoid_failed = mojoid_failed
        self.mojoid_succeeded = mojoid_succeeded

        self.max_concurrent = 12 # unhardcode me  XXX

        self.throttler = broker.FindBeforehandBlockFetcher.throttler
        # we use our own BlockFinder so that it and any cached info
        # (bad and good) is discarded when this session is deleted
        # and so that active user downloads don't interfere with us.
        # (maybe it'd be better to use the broker.FBFH.finder one?)
        self.finder = BlockFinder(broker, self._find_function)

        # list of block ids to find
        self.blocks_to_find = []

        # you can't register for a block id call back with BlockFinder until
        # after you have seeded it with block ids... so we hold the callbacks here
        self.callbacks_to_register = []

        # total number of dinodes we are searching for
        self.num_dinodes = Counter()

        # number of sharemaps that have been fetched and have their pieces queued
        self.queued = Counter()

        # number of dinodes that had 4/8 shares located for every piece
        self.finished = Counter()

        # number of dinodes that had one or more pieces with less than 4/8 shares located
        self.failed = Counter()

        # state indicator, set true after we have issued the finds for the main blocks
        self.in_second_phase = None

        # used for debugging, tracking how long this session took
        self.lifetimer = mojoutil.Timer()
        self.lifetimer.start()


    def __del__(self):
        debug.mojolog.write('%s being garbage collected after %s seconds\n', args=(self, self.lifetimer.get()), v=3, vs='ContentVerifier')


    def find_dinodes(self, dinodes):
        """
        This is the main function.  Call this with a list of dicts
        containing __MOJOID__ and __DINODE__ as you would get back
        from the content tracker DB.  It will eventually callback when
        it is done.

        basically it gets all the pieces and creates SingleReassemblies for them
        registers all the blocks with BlockFinder
        not necessarily in that order (can't create SingleReassemblies till BlockFinder is seeded)
        """
        if len(dinodes) == 0:
            self.callback()
        assemblies = []
        debug.mojolog.write("Determining availability of content tracker entries. \n", v=2, vs="ContentVerifier")

        for i in dinodes:
            dinodedict = parse_dinode(mojosixbit.a2b(i['__DINODE__']))
            self.num_dinodes.inc()

            if dinodedict['piecelist'][0].has_key('offset'):
                # this is a sharemapless dinode, start finding its few pieces immediately
                dinodeObj = Dinode(i['__MOJOID__'], 0, self, sharesneeded=dinodedict['sharesneeded'], totalshares=dinodedict['totalshares'])
                dinodeObj.init_piecelist(dinodedict['piecelist'])
                continue
            dinodeObj = Dinode(i['__MOJOID__'], dinodedict['piecelist'][0]['length'], self, sharesneeded=dinodedict['sharesneeded'], totalshares=dinodedict['totalshares'])
            dinodeObj.smapid = dinodedict['piecelist'][0]['pieceid']  # XXX fix for multi/super sharemap
            dinodeObj.smap_piece_length = dinodedict['piecelist'][0]['length']  # XXX fix for multi/super sharemap
            num_blocks_needed = dinodedict['sharesneeded']
            block_positions = [None] * dinodedict['totalshares']
            for index, block_id in dinodedict['piecelist'][0]['shares'].items():  # XXX fix for multi/super sharemap
                block_positions[index] = block_id

            # can't create a SingleReassembly until the BlockFinder is seeded
            # because you can't register for a callback until the BlockFinder knows about that blockid
            assemblies.append((dinodeObj, block_positions))
            for block in block_positions:
                self.blocks_to_find.append(block)

        # sharemap blocks are queued up, now tell the finder to find them
        self.finder.find_blocks(self.blocks_to_find, allow_download_cache=0)
        self.blocks_to_find = []
        # this should be split up and done on the DoQ for better scheduling XXX
        for dinodeObj, block_positions in assemblies:
            SingleReassembly(do_function = self.throttler.do_download, 
                             find_function = self.finder.callback_when_found, finder = self.finder,
                             cancel_find_function = self.finder.cancel, download_done_function = self.throttler.download_done, broker = self.broker,
                             block_positions = block_positions, num_blocks_needed = dinodeObj.sharesneeded,
                             piece_id = dinodeObj.smapid, piece_length = dinodeObj.smap_piece_length, max_concurrent_blobs =self.max_concurrent,
                             callback = dinodeObj._sharemap_reassembled, mtm=self.MTM, allow_download_cache=0)


    def _find_function(self, server_id, server_info, block_ids, callback, preferenceno=0):
        """
        called by BlockFinder to do actual message generation and spewing over mtm
        this was lifted from FindBeforehandBlockFetcher
        """
        assert len(block_ids) > 0, "_find_function called without any block_ids to find!"
        self.broker.do_you_have_blobs((server_id, server_info), block_ids, callback, allow_recursion=0, mojooffer=BidStrategy.choose_bid("MO_DO_YOU_HAVE_BLOBS_CONTENT_VERIFYING", self.broker.ServicesMTM, preferenceno=preferenceno))

    def _find_block(self, blockid, found_callback, finished_callback):
        self.blocks_to_find.append(blockid)
        self.callbacks_to_register.append((blockid, found_callback, finished_callback))
        
    def _finished_queueing_blocks(self):
        self.queued.inc()
        self._find_main_blocks()
        
    def _find_main_blocks(self):
        if not self.in_second_phase:
            if self.queued.get() + self.failed.get() >= self.num_dinodes.get():
                self.in_second_phase = 1
                self.queued = Counter()
                self.finder.find_blocks(self.blocks_to_find, allow_download_cache=0)
                self.blocks_to_find = []
                # TODO replace with a for loop calling doq.add_batch_task()
                CallbackRegistrar(self.finder, self.callbacks_to_register, self.broker)
                self.callbacks_to_register = []


    # record that it has failed
    def _dinode_failed(self, dinode):
        self.failed.inc()
        val = self.mojoid_failed(dinode.mojoid)
        self._find_main_blocks()
        if self.finished.get() + self.failed.get() >= self.num_dinodes.get():
            debug.mojolog.write("Found %s, failed %s\n" % (self.finished.get(), self.failed.get()), vs="ContentVerifier")
            self.callback()
    # record that it has succeded
    def _dinode_succeeded(self, dinode):
        self.finished.inc()
        val = self.mojoid_succeeded(dinode.mojoid)
        if self.finished.get() + self.failed.get() >= self.num_dinodes.get():
            debug.mojolog.write("Found %s, failed %s\n" % (self.finished.get(), self.failed.get()), vs="ContentVerifier")
            self.callback()

class CallbackRegistrar:
    """
    Internal class that registers a buncha callbacks with a block
    finder in a slow, asynchronous manner
    """
    def __init__(self, finder, callbacks, broker):
        self.callbacks = callbacks
        self.finder = finder
        self.broker = broker
        self.broker.doq.add_task(self.reg_some_callbacks)
    def reg_some_callbacks(self):
        l = self.callbacks[:10]
        self.callbacks = self.callbacks[10:]
        if len(l) > 0:
            for blockid, found, finished in l:
                self.finder.callback_when_found(blockid, found, finished)
            self.broker.doq.add_task(self.reg_some_callbacks)
            
# internal class used by VerifierSession
# handles the callback from SingleReassembly, queues up main blocks for that dinode
# creates Pieces and registers them with the BlockFinder
# handles callbacks from each piece
# holds state about the progress of finding all the pieces
# reports back if it succeeds or fails
class Dinode:
    def __init__(self, mojoid, sharemap_length, verifier, sharesneeded, totalshares):
        self.mojoid = mojoid
        self.sharemap_length = sharemap_length
        self.verifier = verifier
        self.pieces = Counter()
        self.succeeded = Counter()
        self.failed = 0
	self.done = 0
        self.sharesneeded = sharesneeded
        self.totalshares = totalshares
        
    def _sharemap_reassembled(self, reassembly):
        """
        this is the callback given to SingleReassembly...
        """
        try:
            sharemap = reassembly.nonblocking_reassemble()
        except ReassemblyError:
            self.verifier._dinode_failed(self)
            return
        sharemap = sharemap[:self.sharemap_length]
        # parse sharemap
        sharemapdict = parse_sharemap(sharemap)
        # list of piecedicts
        piecelist = sharemapdict['piecelist']
        self.init_piecelist(piecelist)
    
    def init_piecelist(self, piecelist):
        """
        call this once we have the full list of piece dicts making up this publication
        """
        # register blocks with Verifier
        for dict in piecelist:
            shares = dict['shares'].items()
            p = Piece(len(shares), self.piece_failed, self.piece_succeeded, sharesneeded=self.sharesneeded)
            self.pieces.inc()
            for i, blockid in shares:
                self.verifier._find_block(blockid, p.block_found, p.block_finished)
        self.verifier._finished_queueing_blocks()

    def piece_failed(self):
        if not self.failed and not self.done:
            self.failed = 1
            # alert verifier we are not retrieveable
            self.verifier._dinode_failed(self)
            # break external references (helps the GC)
            del self.verifier

    def piece_succeeded(self):
        if self.failed or self.done:
            return
        self.succeeded.inc()
        if self.succeeded.get() >= self.pieces.get():
            self.done = 1
            # alert verifier we are retrievable
            self.verifier._dinode_succeeded(self)
            # break external references (helps the GC)
            del self.verifier


# just something to encapsulate the state of finding a piece
# handles callbacks from block finder
# reports back to it's Dinode if it fails or succeeds (finds 4/8 blocks)
class Piece:
    def __init__(self, num_shares, piece_failed_callback, piece_succeeded_callback, sharesneeded):
        self.piece_failed_callback = piece_failed_callback
        self.piece_succeeded_callback = piece_succeeded_callback
        self.num_shares = num_shares
        self.passed_block_map = {}
        self.finished_block_map = {}
        self.called_callback = 0  # boolean, set to true when we call a callback
        self.sharesneeded = sharesneeded
        
    def block_finished(self, block_id):
        debug.mojolog.write('Piece %s block_finished %s\n', args=(self, block_id), v=4, vs='ContentVerifier')
        if not self.finished_block_map.has_key(block_id):
            self.finished_block_map[block_id] = block_id
        if not self.called_callback and len(self.finished_block_map) >= self.num_shares:
            debug.mojolog.write('Piece %s got %s of %s blockshares\n', args=(self, len(self.passed_block_map), len(self.finished_block_map)), v=4, vs='ContentVerifier')
            self.called_callback = 1
            if len(self.passed_block_map) >= self.sharesneeded:
                self.piece_succeeded_callback()
            else:
                self.piece_failed_callback()
            # break external references (helps the GC)
            del self.piece_succeeded_callback
            del self.piece_failed_callback

        # TODO Bram says: it is much better to totally ignore the
        # found callbacks completely and only use the finished
        # callbacks, asking the BlockFinder for the list of locations
        # for each block after enough or all of the finds have
        # finished.  -greg  2001-01-09

    def block_found(self, block_id):
        if not self.passed_block_map.has_key(block_id):
            debug.mojolog.write('Piece %s block_found %s\n', args=(self, block_id), v=4, vs='ContentVerifier')
            self.passed_block_map[block_id] = block_id
        if not self.called_callback and len(self.passed_block_map) >= self.sharesneeded:
            # call the piece succeeded callback as soon as we have enough blockshares even if finds for the others haven't finished yet
            self.called_callback = 1
            self.piece_succeeded_callback()
            # break external references (helps the GC)
            del self.piece_succeeded_callback
            del self.piece_failed_callback

