#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: Conversation.py,v 1.76 2001/06/04 06:48:59 greg Exp $'

### standard modules
import threading
import types
import traceback
import sys
from sha import sha
from time import time
import exceptions
import socket
import os
import types
from traceback import print_exc

### our modules
import Cache
from CommHints import HINT_EXPECT_RESPONSE, HINT_EXPECT_MORE_TRANSACTIONS, HINT_EXPECT_NO_MORE_COMMS, HINT_EXPECT_TO_RESPOND, HINT_THIS_IS_A_RESPONSE, HINT_NO_HINT
import CommStrat
import MojoConstants
from MojoConstants import true, false
import MojoKey
import MojoMessage
import RunTests
import SEXP
import SingleThreader
from confutils import confman
import debug
import idlib
from human_readable import human_readable, hr
import modval
import mojosixbit
import mojoutil
import randsource
import MetaTrackerLib
from MojoHandicapper import DISQUALIFIED


debugwrite = debug.mojolog.write


HANDLED = 0
EXPECTING_RESPONSE = 1

class ConversationManager(SingleThreader.DummySingleThreader):
    """
    It is asserted that this class is only called by one thread at a time.  If you experience a
    multithread exception (due the the BombingSingleThreader mixin class), then you should
    probably take whatever function call is calling the ConversationManager and put it on the
    DoQ instead of calling it from your thread.
    """
    def __init__(self, MTM):
        SingleThreader.DummySingleThreader.__init__(self)

        self._MTM = MTM
        # maps first message id's to (counterparty_id, callback function, notes for callback, conversation type)
        self.__callback_functions = {}
        # maps message id's of messages who's responses have timed out to (recipient_id, conversation type, time of timeout, post timeout callback function, notes [only if post timeout callback function])
        self.__failed_conversations = Cache.StatsCache(autoexpireinterval=120, maxitems=500, autoexpireparams={'maxage': 900})

        # maps binary counterparty_id to number of response messages that have not been received
        # (used for the pending_responses_handicapper)
        self.__outstanding_messages = Cache.StatsCache(autoexpireinterval=45, autoexpireparams={'maxage': 600})

        # maps message id to (binary counterparty_id, message type, response status)
        # where status is HANDLED or EXPECTING_RESPONSE
        # This data structure is used between receiving a specific initiating message and
        # sending the response to it.  It is not used for incoming response messages.
        self._map_inmsgid_to_info = {}

        # This is a map from counterparty_id to the binary msgId of the last message you
        # received from that counterparty.  We use the msgId in outgoing messages, thus proving
        # to the counterparty that our messages are fresh (i.e., they are not replay attacks or
        # other sneaky trickery).  Actually our current software does not check the incoming
        # freshness proofs, but this code ensures that when we ship a new, replay-attack-proof
        # version which _does_ verify freshness proofs, then older apps which are running _this_
        # version of our software will be able to interoperate with it.
        self._map_cid_to_freshness_proof = Cache.Cache(maxitems=20000)

        self._in_message_num = 0L   # used only in debugging

    def shutdown(self):
        debug.mojolog.write("self._map_inmsgid_to_info: %s\n", args=(self._map_inmsgid_to_info,), v=6, vs="debug")
        self.__callback_functions = {}
        self.__failed_conversations.empty()
        if hasattr(self, '_MTM'):
            del self._MTM  # break our circular reference
    
    def __del__(self):
        debug.mojolog.write("ConversationManager %s __del__ called\n", args=(self,))
        self.shutdown()
    
    def pending_responses_handicapper(self, counterparty_id, metainfo, message_type, message_body):
        """
        Severely handicaps any counterparty based on the number of response messages it has outstanding.

        The idea behind this is to prevent sending a counterparty another query until it responds to the
        earlier one we sent it.

        This should prevent queries that are failing from building up and failing due to timeouts
        piling on top of eachother with responses to earlier messages still being delivered but
        ignored due to their transaction already timing out.

        XXX This is hopefully fixed by late-response processing and more clever timeout values.  This handicapper can cause serious harm by disqualifying all of the block servers that you most want to use, if it is the case that you can generate queries faster than the queries can travel to the BS, get processed, and travel back.  Anyway, I'm just adjusting it to be a handicap instead of DISQ.  --Zooko 2001-04-29

        timed out messages count for a while after they were first initiated.  The exact time is controlled
        by the expire time for the __outstanding_messages StatsCache in the constructor.

        If more than 4 messages are outstanding this counterparty is temporarily disqualified from further business.
        If it is still unreliable after that, the persistent unreliability handicapper will eventually take care of it.
        """
        HANDICAP_FOR_PENDING=100000 # This should be dramatically reduced when we have smart stats-based timeouts.  --Zooko 2001-05-09
        num = self.__outstanding_messages.get(counterparty_id, 0)
        return num * HANDICAP_FOR_PENDING

    def initiate_and_return_first_message(self, counterparty_id, conversationtype, firstmsgbody, outcome_func, timeout = MojoConstants.DEFAULT_TIMEOUT, notes = None, mymetainfo=None, post_timeout_outcome_func=None):
        """
        @precondition `counterparty_id' must be  an id.: idlib.is_sloppy_id(counterparty_id): "id: %s" % hr(id)

        returns a tuple of (message_id, binary_message_string)
        """
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be  an id." + " -- " + "id: %s" % hr(id)

        counterparty_id = idlib.canonicalize(counterparty_id, "broker")

        self._enter()
        try:
            # Our `freshness proof' is the hash of the last message that we saw from this counterparty.

            # If we don't have such a last message, we leave it blank.

            # Currently, Mojo Nation brokers accept any freshnessproof and don't check to be
            # sure that it is really a proof of freshness, so currently either of these cases
            # will work.  In the future Mojo Nation brokers will start rejecting messages with
            # blank freshnessproofs and insisting on a hash of a recent message.  This case is
            # handled (even in the current version: v0.920) with a special handler inside
            # MojoTransaction.  --Zooko 2000-09-27

            # XXXX Zooko: add the freshness challenge handler inside MojoTransaction!  --Zooko 2000-09-29

            message = MojoMessage.makeInitialMessage(msgtype=conversationtype, msgbody=firstmsgbody, recipient_id=idlib.to_binary(counterparty_id), nonce=idlib.new_random_uniq(), freshnessproof=self._map_cid_to_freshness_proof.get(counterparty_id), mymetainfo=mymetainfo)

            msgId = idlib.make_id(message, 'msg')

            # only keep track of the message details if we have a response message handler
            if outcome_func:
                self.__callback_functions[msgId] = (counterparty_id, outcome_func, notes, conversationtype, post_timeout_outcome_func)
                num = self.__outstanding_messages.get(counterparty_id, 0)
                self.__outstanding_messages[counterparty_id] = num + 1

            return msgId, message
        finally:
            self._exit()

    def fail_conversation(self, msgId, failure_reason='generic failure', istimeout=0):
        self._enter()
        try:
            initial = self.__callback_functions.get(msgId)
            if initial is None :
                # This means that a response has been received, so it shouldn't fail.
                return
            del self.__callback_functions[msgId]
            recipient_id, callback_function, notes, conversationtype, post_timeout_callback_function = initial
            if not istimeout:
                # only include these in the failed_conversation map for later calling if it was a timeout
                post_timeout_callback_function = None
            if post_timeout_callback_function is None:
                # only needed for post timeout callback functions
                post_timeout_notes = None
            else:
                post_timeout_notes = notes
            self.__failed_conversations[msgId] = recipient_id, conversationtype, time(), post_timeout_callback_function, post_timeout_notes
            if self._MTM._doQ.is_currently_doq():
                callback_function(failure_reason=failure_reason, notes=notes)
            else:
                self._MTM._doQ.add_task(callback_function, kwargs = {'failure_reason': failure_reason, 'notes': notes})
        finally:
            self._exit()

    def is_unsatisfied_message(self, msgId):
        """
        @returns `true' if and only if the msg identified by `msgId' has been sent out to a
            counterparty, and it is a message to which we expect a response, and we have not
            yet received a response, and we have not yet timed out and given up hope of
            getting a response
        """
        self._enter()
        try:
            return self.__callback_functions.has_key(msgId)
        finally:
            self._exit()

    def drop_request_state(self, firstmsgId):
        """cleanup all internal state about received request firstmsgId"""
        del self._map_inmsgid_to_info[firstmsgId]

    def send_response(self, prevmsgId, msgbody, mymetainfo=None):
        """
        @param msgbody the message body to be sent back
        
        @precondition `prevmsgId' must be of the right form for an id.: idlib.is_sloppy_id(prevmsgId): "prevmsgId: %s" % hr(prevmsgId)
        @precondition `msgbody' must be either None or else the full msg body dict, containing either a "mojo header" subdict or a "mojo message" subdict or both.: (not msgbody) or msgbody.has_key('mojo header') or msgbody.has_key('mojo message'): "msgbody: %s" % hr(msgbody)
        @precondition internal1: self._map_inmsgid_to_info.get(prevmsgId) is not None: "prevmsgId: %s" % hr(prevmsgId)
        @precondition internal2: (type(self._map_inmsgid_to_info.get(prevmsgId)) in (types.TupleType, types.ListType)): "self._map_inmsgid_to_info.get(prevmsgId): %s :: %s" % (hr(self._map_inmsgid_to_info.get(prevmsgId)), hr(type(self._map_inmsgid_to_info.get(prevmsgId))))
        @precondition internal3: self._map_inmsgid_to_info.get(prevmsgId)[2] == EXPECTING_RESPONSE: "self._map_inmsgid_to_info.get(prevmsgId): %s" % hr(self._map_inmsgid_to_info.get(prevmsgId))
        @precondition internal4: idlib.is_binary_id(self._map_inmsgid_to_info.get(prevmsgId)[0]): "self._map_inmsgid_to_info.get(prevmsgId)[0]: %s :: %s" % (hr(self._map_inmsgid_to_info.get(prevmsgId)[0]), hr(type(self._map_inmsgid_to_info.get(prevmsgId)[0])))
        """
        assert idlib.is_sloppy_id(prevmsgId), "precondition: `prevmsgId' must be of the right form for an id." + " -- " + "prevmsgId: %s" % hr(prevmsgId)
        assert (not msgbody) or msgbody.has_key('mojo header') or msgbody.has_key('mojo message'), "precondition: `msgbody' must be either None or else the full msg body dict, containing either a \"mojo header\" subdict or a \"mojo message\" subdict or both." + " -- " + "msgbody: %s" % hr(msgbody)
        assert self._map_inmsgid_to_info.get(prevmsgId) is not None, "precondition: internal1" + " -- " + "prevmsgId: %s" % hr(prevmsgId)
        assert (type(self._map_inmsgid_to_info.get(prevmsgId)) in (types.TupleType, types.ListType)), "precondition: internal2" + " -- " + "self._map_inmsgid_to_info.get(prevmsgId): %s :: %s" % (hr(self._map_inmsgid_to_info.get(prevmsgId)), hr(type(self._map_inmsgid_to_info.get(prevmsgId))))
        assert self._map_inmsgid_to_info.get(prevmsgId)[2] == EXPECTING_RESPONSE, "precondition: internal3" + " -- " + "self._map_inmsgid_to_info.get(prevmsgId): %s" % hr(self._map_inmsgid_to_info.get(prevmsgId))
        assert idlib.is_binary_id(self._map_inmsgid_to_info.get(prevmsgId)[0]), "precondition: internal4" + " -- " + "self._map_inmsgid_to_info.get(prevmsgId)[0]: %s :: %s" % (hr(self._map_inmsgid_to_info.get(prevmsgId)[0]), hr(type(self._map_inmsgid_to_info.get(prevmsgId)[0])))

        counterparty_id, inmsgtype, status = self._map_inmsgid_to_info.get(prevmsgId)
        assert idlib.is_binary_id(counterparty_id), "`counterparty_id' must be a binary id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)
        self.drop_request_state(prevmsgId)

        if idlib.equal(counterparty_id, self._MTM.get_id()):
            extrametainfo = None
        else:
            extrametainfo = MetaTrackerLib.get_preencoded_extra_metainfo_for(counterparty_id)
        msgstr = MojoMessage.makeResponseMessage(inmsgtype + ' response', msgbody, prevmsgId,
                    freshnessproof=self._map_cid_to_freshness_proof.get(counterparty_id),
                    mymetainfo=mymetainfo,
                    extrametainfo=extrametainfo)
        self._MTM.send_message_with_lookup(counterparty_id, msgstr, mojooffer=0, hint=HINT_EXPECT_MORE_TRANSACTIONS | HINT_THIS_IS_A_RESPONSE)

    def _process(self, msg, msgId, counterparty_id, commstrat=None):
        """
        This gets called for all incoming messages, by `handle_raw_message()'.  It verifies the
        message's integrity and either send it to a handler func to generate a response or send
        it to the appropriate callback func.

        @returns the full msg body (containing a 'mojo header' and/or 'mojo message' subdict) in dict form, or None

        @throws MojoMessage.BadFormatError if the message isn't properly formatted in
            MojoMessage format

        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)

        @postcondition Result must not be `None'.: result is not None: "result: %s" % hr(result)
        @postcondition Result must be either MojoConstants.NO_RESPONSE or MojoConstants.ASYNC_RESPONSE or else the full msg body dict, containing either a "mojo header" subdict or a "mojo message" subdict or both.: (result is MojoConstants.NO_RESPONSE) or (result is MojoConstants.ASYNC_RESPONSE) or result.has_key('mojo header') or result.has_key('mojo message'): "result: %s" % hr(result)
        @postcondition Result must not be a body with an `ASYNC_RESPONSE' or a `NO_RESPONSE' buried in it.: not (responsebody and (type(responsebody) is types.DictType) and ((responsebody.get('mojo message') is MojoConstants.ASYNC_RESPONSE) or (responsebody.get('mojo message') is MojoConstants.NO_RESPONSE))): "responsebody: %s" % hr(responsebody)
        """ 
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

        counterparty_id = idlib.canonicalize(counterparty_id, "broker")

        # begin DEBUG do not uncomment this in normal code
        # save all uncompressed incoming messages to unique files to be used for post analysis
        # such as real world mdecode() performance tweaking, etc.
        # (note: this writes them in the current directory, normally localweb/webroot)
        ##_dbg_fname = 'message.%05d' % self._in_message_num
        ##_dbg_f = open(_dbg_fname, 'wb')
        ##_dbg_f.write(msg)
        ##_dbg_f.close()
        ##self._in_message_num += 1  # note: python 2.0 syntax
        # end DEBUG

        self._enter()
        try:
            reference = MojoMessage.getReference(msg)
            nonce = MojoMessage.getNonce(msg)
            recipient_id = MojoMessage.getRecipient(msg)
            senders_metainfo = MojoMessage.getSendersMetaInfo(msg)
            extra_metainfo = MojoMessage.getExtraMetaInfo(msg)

            if senders_metainfo and self._MTM._allow_send_metainfo:  # (the allow_send also means allow_receive, bad name..)
                try:
                    metainfo_id = MetaTrackerLib.metainfo_dict_to_id(senders_metainfo)
                    # verify that counterparty_id is the id of the info described in senders_metainfo
                    if idlib.equal(counterparty_id, metainfo_id):
                        # if it is, we'll accept it as authentic
                        debug.mojolog.write("received metainfo for counterparty %s in a message from them.\n", args=(counterparty_id,), v=4, vs="conversation")
                        MetaTrackerLib.cache_meta_tracker_responses([senders_metainfo])
                    else:
                        debug.mojolog.write("counterparty %s sent us unauthenticated metainfo for %s, ignoring.\n", args=(counterparty_id, metainfo_id), v=3, vs="conversation")
                except:
                    debug.mojolog.write("non-fatal error parsing metainfo in message from %s: %s\n", args=(counterparty_id, senders_metainfo,), v=3, vs="conversation")
                    if int(confman.dict.get('MAX_VERBOSITY', 0)) >= 3:
                        traceback.print_exc(file=debug.mojolog)

            if extra_metainfo and self._MTM._allow_send_metainfo:  # (the allow_send also means allow_receive, bad name..)
                try:
                    MetaTrackerLib.cache_meta_tracker_responses(extra_metainfo)
                    debug.mojolog.write("received %s extra metainfo entries from counterparty %s.\n", args=(len(extra_metainfo), counterparty_id), v=4, vs="conversation")
                except:
                    debug.mojolog.write("non-fatal error caching extra metainfo in message from %s: %s\n", args=(counterparty_id, extra_metainfo,), v=3, vs="conversation")
                    if int(confman.dict.get('MAX_VERBOSITY', 0)) >= 3:
                        traceback.print_exc(file=debug.mojolog)

            if nonce is not None :
                # this is a first message
                if (reference is not None) or (recipient_id is None):
                    debug.stderr.write("WARNING: a Mojo Message arrived with inconsistent conversation markers: nonce: %s, reference: %s, recipient_id: %s\n", args=(nonce, reference, recipient_id), v=1, vs="conversation")
                    return MojoConstants.NO_RESPONSE
                if not idlib.is_sloppy_id(nonce) :
                    debug.stderr.write("WARNING: a Mojo Message arrived with badly formed nonce: %s\n", args=(nonce,), v=1, vs="conversation")
                    return MojoConstants.NO_RESPONSE
                conversationtype = MojoMessage.getType(msg)

                # We now have a hint -- we're expecting to respond to this.
                if commstrat:
                    commstrat.hint = commstrat.hint | HINT_EXPECT_MORE_TRANSACTIONS | HINT_EXPECT_TO_RESPOND
                    commstrat.hintnumexpectedsends = commstrat.hintnumexpectedsends + 1

                if self._map_inmsgid_to_info.has_key(msgId):
                    # This can only happen if we have already started processing this unique message.
                    return MojoConstants.ASYNC_RESPONSE
                self._map_inmsgid_to_info[msgId] = (counterparty_id, MojoMessage.getType(msg), EXPECTING_RESPONSE)
                # Reminder: do not somehow change this handle_initiating_message call to be on the DoQ in the future without changing
                # the MTM.__in_message_for_you logic used in fast relay to prevent nested 'message for you' messages.  -greg
                result = self._MTM.handle_initiating_message(counterparty_id, conversationtype, MojoMessage.getBody(msg), firstmsgId=msgId) 
                if result is None:
                    result = MojoConstants.NO_RESPONSE
                if result is MojoConstants.NO_RESPONSE:
                    self.drop_request_state(msgId)
                assert (result is MojoConstants.NO_RESPONSE) or (result is MojoConstants.ASYNC_RESPONSE) or result.has_key('mojo header') or result.has_key('mojo message'), 'Result must be either None or else the full msg body dict, containing either a "mojo header" subdict or a "mojo message" subdict or both.' + ' -- ' + "result: %s" % hr(result)
                return result
            else :
                # this is a response message
                if (reference is None) or (recipient_id is not None):
                    debug.stderr.write("WARNING: a Mojo Message arrived with inconsistent conversation markers: nonce: %s, reference: %s, recipient_id: %s\n", args=(nonce, reference, recipient_id), v=1, vs="conversation")
                    return MojoConstants.NO_RESPONSE

                responsetype = MojoMessage.getType(msg)

                # Make sure that this is a response to a message that we sent, from the person to whom we sent it.
                initial = self.__callback_functions.get(reference)
                if initial is None:
                    recipient_id, conversationtype, msgtimeout, post_timeout_callback_function, post_timeout_notes = self.__failed_conversations.get(reference, (None, None, None, None, None))
                    if recipient_id is not None:
                        del self.__failed_conversations[reference]  # proactively clean up this cache
                    if recipient_id != counterparty_id:
                        return MojoConstants.NO_RESPONSE
                    # log the late response
                    if conversationtype is not None:
                        debug.mojolog.write("received %s to msgId %s %s from %s %s seconds after the timeout\n", args=(responsetype, reference, conversationtype, counterparty_id, time() - msgtimeout), v=2, vs='Conversation')

                    if post_timeout_callback_function:
                        # recipient_id and conversationtype were set above
                        callback_function = post_timeout_callback_function
                        notes = post_timeout_notes
                        debug.mojolog.write("post timeout callback for response to msgId %s %s\n", args=(reference, responsetype), vs='Conversation')
                    else:
                        # keep track of the number of outstanding response messages for handicapping purposes
                        num = self.__outstanding_messages.get(recipient_id, 0)
                        if num > 0:  # make sure it stays >= 0 just in case threaded access happened somehow and messed it up
                            self.__outstanding_messages[recipient_id] = num - 1
                        return MojoConstants.NO_RESPONSE
                else:
                    recipient_id, callback_function, notes, conversationtype, post_timeout_callback_function = initial

                if counterparty_id != recipient_id:
                    return MojoConstants.NO_RESPONSE

                try:
                    del self.__callback_functions[reference]
                except KeyError:
                    # try..except needed here so that del works even if this is a post timeout callback
                    pass

                # keep track of the number of outstanding response messages for handicapping purposes
                num = self.__outstanding_messages.get(recipient_id, 0)
                if num > 0:  # make sure it stays >= 0 just in case threaded access happened somehow and messed it up
                    self.__outstanding_messages[recipient_id] = num - 1
                
                if conversationtype + ' response' != responsetype:
                    debug.mojolog.write("message of unexpected type %s from %s in response to a %s\n", args=(responsetype, counterparty_id, conversationtype), v=3, vs='Conversation')
                    return MojoConstants.NO_RESPONSE

                # Do hints about expecting responses...
                if commstrat:
                    if (commstrat.hint & HINT_EXPECT_RESPONSE) and (commstrat.hintnumexpectedresponses > 0):
                        commstrat.hintnumexpectedresponses = commstrat.hintnumexpectedresponses - 1
                        if commstrat.hintnumexpectedresponses == 0:
                            # No longer expecting response!
                            commstrat.hint = commstrat.hint & (~ HINT_EXPECT_RESPONSE)
                        else:
                            pass

                self._MTM._doQ.add_task(callback_function, kwargs = {'outcome': MojoMessage.getBody(msg), 'notes': notes})
                return MojoConstants.NO_RESPONSE
        finally:
            self._exit()

    def handle_raw_message(self, counterparty_id, inmsg, commstrat=None):
        """
        This gets called for all incoming messages.  It calls `self._process()' and sends the
        response returned, if any, back to the counterparty.
        """
        counterparty_id = idlib.to_binary(counterparty_id)   # also asserts is_sloppy_id(counterparty_id)

        if commstrat:
            assert isinstance(commstrat, CommStrat.Crypto)
        elif not hasattr(self, '_complained'):
            self._complained = true
            debug.mojolog.write("complaint: I wish that I had a reference to a persistent CommStrat object that represents this particular way of talking to this counterparty, so that I could use it to store hints to optimize behaviour.  Oh well.  counterparty_id: %s\n", args=(counterparty_id,), v=5, vs="debug")

        self._enter()
        try: # for the `self._enter()' ... `self._exit()' shtick...
            try: # for BadFormatErrors
                msgId = idlib.make_id(inmsg, 'msg')
                def debugprintreceive(counterparty_id=counterparty_id, inmsg=inmsg, msgId=msgId):
                    printstr="<=== %s: %s\n"
                    theseargs=[counterparty_id, inmsg]

                    maxverb = int(confman.dict.get("MAX_VERBOSITY", 0))
                    if maxverb >= 5:
                        debug.stderr.write("<=== %s: receiving: %s, %s bytes uncomp\n", args=(counterparty_id, inmsg, len(inmsg)), v=5, vs="conversation")
                    elif maxverb >= 4:
                        # XXX note, this slows sending down A LOT! calling MojoMessage.getType() calls mdecode() & template check on the message.  The mdecode part is -extremely- slow (esp on big messages).  -greg 11-oct-2000
                        # if we want to display this quickly, we'll need to pass the type in from before we called mencode()
                        debug.stderr.write("<=== %s: receiving: (id: %s, type: %s, %s bytes uncomp, ...)\n", args=(counterparty_id, msgId, MojoMessage.getType(inmsg), len(inmsg)), v=4, vs="conversation") # semi-verbose
                    elif maxverb >= 3:
                        debug.stderr.write("<=== %s: receiving: (id: %s, %s bytes uncomp, ...)\n", args=(counterparty_id, msgId, len(inmsg)), v=3, vs="conversation") # semi-verbose

                debugprintreceive()

                self._map_cid_to_freshness_proof[counterparty_id] = msgId

                responsebody = self._process(inmsg, msgId, counterparty_id, commstrat)
                result = responsebody # just an alternate name for postcondition assertions
                assert result is not None, "Result must not be `None'." + " -- " + "result: %s" % hr(result)

                assert (result is MojoConstants.NO_RESPONSE) or (result is MojoConstants.ASYNC_RESPONSE) or result.has_key('mojo header') or result.has_key('mojo message'), "postcondition: Result must be either MojoConstants.NO_RESPONSE or MojoConstants.ASYNC_RESPONSE or else the full msg body dict, containing either a \"mojo header\" subdict or a \"mojo message\" subdict or both." + "--" + "result: %s" % hr(result)
                assert not (responsebody and (type(responsebody) is types.DictType) and ((responsebody.get('mojo message') is MojoConstants.ASYNC_RESPONSE) or (responsebody.get('mojo message') is MojoConstants.NO_RESPONSE))), "Result must not be a body with an `ASYNC_RESPONSE' or a `NO_RESPONSE' buried in it." + "--" + "responsebody: %s" % hr(responsebody)

                if (responsebody is not MojoConstants.ASYNC_RESPONSE) and (responsebody is not MojoConstants.NO_RESPONSE):
                    self.send_response(msgId, responsebody)

            except MojoMessage.BadFormatError, le:
                debug.stderr.write("_process(): BadFormatError in message from %s, msg: %s, error: %s\n", args=(counterparty_id, inmsg, le,), v=2, vs="conversation")
        finally:
            self._exit()

class ConversationTest :
    def __init__(self) :
        self.fail_reason = None
        self.stored = []
        self.expected_id = None
        self.expected_type = None
        self.expected_message_flag = None
        self.next_response = None
        self.id1 = idlib.string_to_id('one')
        self.id2 = idlib.string_to_id('two')
        self.id3 = idlib.string_to_id('three')
        self.c1 = ConversationManager(self)
        self.c2 = ConversationManager(self)
        self.c3 = ConversationManager(self)
        self.expected_response_flag = None

    def handle_response(self, outcome, failure_reason=None) :
        if self.expected_response_flag is None :
            self.fail_reason = 'no response expected'
        else :
            if outcome['mojo message'].get('flag') != self.expected_response_flag :
                print "outcome: %s" % hr(outcome)
                self.fail_reason = 'mismatched response'
        self.expected_response_flag = None

    def expect(self, expected_id, expected_type, expected_message_flag) :
        self.expected_id = expected_id
        self.expected_type = expected_type
        self.expected_message_flag = expected_message_flag
        
    def send_message_with_lookup(self, counterparty_id, respMsg, hint = None) :
        self.stored.append((counterparty_id, respMsg))

    def handle_initiating_message(self, counterparty_id, message_type, message, firstmsgId) :
        if not idlib.equal(self.expected_id, counterparty_id):
            self.fail_reason = 'unmatched id'
        if not self.expected_type == message_type :
            self.fail_reason = 'unmatched type'
        if not message.get('flag') == self.expected_message_flag :
            self.fail_reason = 'unmatched flag'
        self.expect(None, None, None)
        return self.next_response

    def test_message_and_response_association(self) :
        m1_id, m1 = self.c1.initiate_and_return_first_message(self.id2, 't1', {'flag' : 'spam1'}, outcome_func=self.handle_response)
        self.next_response = {'mojo message' : {'flag' : 'f1'}}
        self.expect(self.id1, 't1', 'spam1')
        self.c2.handle_raw_message(self.id1, m1)
        assert len(self.stored) == 1
        cid, msg = self.stored[0]
        assert idlib.equal(cid, self.id1), "cid: %s, self.id1: %s" % (hr(cid), hr(self.id1))
        self.expected_response_flag = 'f1'
        self.c1.handle_raw_message(self.id2, msg)
        assert len(self.stored) == 1
        assert self.fail_reason is None, "self.fail_reason: %s" % hr(self.fail_reason)

# Generic stuff
NAME_OF_THIS_MODULE="Conversation"

mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests(NAME_OF_THIS_MODULE)

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()


