#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: CryptoCommsHandler.py,v 1.55 2001/06/08 00:50:42 greg Exp $'

# standard modules
import traceback
import types
import sha
import string
import zlib

# our modules
import Cache
import CommsHandler
import CommStrat
from CommHints import HINT_EXPECT_RESPONSE, HINT_EXPECT_MORE_TRANSACTIONS, HINT_EXPECT_NO_MORE_COMMS, HINT_EXPECT_TO_RESPOND, HINT_THIS_IS_A_RESPONSE, HINT_NO_HINT
import MojoConstants
import MojoKey
import SEXP
import confutils
import debug
from human_readable import human_readable, hr
import idlib
import mesgen
import mojosixbit
import mojoutil


# only for testing
import TCPCommsHandler
import DoQ


# Things to test:
# * Opens TCP connection, sends message, keeps TCP connection, receives reply.
# * Sends via relay.
# * Receives via relay.
# * Someone opens TCP connection to you, sends message, you keep TCP connection, send reply.
# * Someone opens TCP connection, never sends message, you keep it for awhile, but clean it when
#     the total conns > TCP_MAX_CONNECTIONS.
# * You are operating behind relay.  Someone sends you a message, you have phonebook entry from
#     them (either from them or from MT), and you know their comm strat is TCP, you open a TCP
#     connection to them and send reply.
# * They are operating behind relay.  You send them a message via relay.  They open a TCP
#     connection to you and send reply.  Next time you want to talk to them, you use the open
#     TCP connection.
# * You have an open TCP connection to a counterparty, which is working.  You get a new comm
#     strat from somewhere that says to use relay.  You ignore it and keep using the working TCP
#     connection.
# * You have an IP addr and port number for a counterparty.  You get a new comm strat from
#     somewhere that says to use relay.  You ignore it and try to open a connection using the IP
#     addr.
# * You are talking via TCP, and the connection dies.  You then learn from a MT that the
#     counterparty is available via relay.  You start using relay.



# messages who's cleartext is larger than this will be dropped
MAX_CLEARTEXT_LEN = 4*1024*1024  # 4megs is #)^*!)^& huge for a single message!


class CryptoCommsHandler:
    """
    """
    def __init__(self, mesgen, tcpch, doQ):
        """
        @param doQ It's the DO QUEUE!
        @param mesgen the crypto object
        @param tcpch the TCP Comms Handler to send messages to after they have been encrypted, and from which to receive messages that need to be decrypted and then passed on to `inmsg_handler_func()'
        """
        self._tcpch = tcpch
        self._mesgen = mesgen
        self._doQ = doQ
        # map cid to instance of CommStrat
        self._cid_to_cs = {}
    
    def shutdown(self):
        if hasattr(self, '_tcpch') and self._tcpch is not None:
            self._tcpch.shutdown()
        if hasattr(self, '_cid_to_cs'):
            del self._cid_to_cs
        if hasattr(self, '_mesgen'):
            del self._mesgen
        if hasattr(self, '_doQ'):
            del self._doQ

    def stop_listening(self):
        self._tcpch.stop_listening()

    def is_listening(self):
        self._tcpch.is_listening()

    def start_listening(self, inmsg_handler_func):
        """
        Start trying to get incoming messages which you will then pass to `inmsg_handler_func()'.

        @param inmsg_handler_func the function to be called whenever a message for us comes in
        """
        self._upward_inmsg_handler = inmsg_handler_func
        # start listening
        self._tcpch.start_listening(inmsg_handler_func = self._inmsg_handler)

    def send_msg(self, counterparty_id, msg, mojooffer, fast_fail_handler, hint=HINT_NO_HINT, timeout=None):
        """
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        @precondition `msg' must be a string.: type(msg) == types.StringType
        """
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)
        assert type(msg) == types.StringType, "precondition: `msg' must be a string."
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)
        assert type(msg) == types.StringType, "precondition: `msg' must be a string."

        # debug.mojolog.write("CryptoCommsHandler.send_msg()\n")
        if idlib.equal(counterparty_id, self.get_id()):
            self._doQ.add_task(self._push_msg_up, args=(counterparty_id, msg, None))
        else:
            try:
                cs = self._cid_to_cs.get(counterparty_id)
                if not cs:
                    fast_fail_handler(msgId=idlib.make_id(msg, 'msg'), failure_reason="no CommStrat.Crypto for this counterparty")
                    return

                llstrat = cs._lowerstrategy
                if not llstrat:
                    fast_fail_handler(msgId=idlib.make_id(msg, 'msg'), failure_reason="no llstrat for this counterparty")
                    return

                if len(msg) > MAX_CLEARTEXT_LEN :
                    raise CommsHandler.CannotSendError, "message too long: %s bytes" % len(msg)

                origmsglen = len(msg)
                try:
                    msg = zlib.compress(msg, 3)
                    debug.mojolog.write("CryptoCommsHandler.send_msg(): sending compressed msg (%s bytes, orig %s)\n", args=(len(msg), origmsglen), v=14, vs="crypto")
                except zlib.error, e:
                    debug.mojolog.write("CryptoCommsHandler.send_msg(): got `%s' during compression (%s bytes)\n", args=(e, origmsglen), v=5, vs="crypto") # verbose
                cyphertext = self._mesgen.generate_message(recipient_id = counterparty_id, message = msg)
                del msg  # we're done with it, encourage faster gc
            except mesgen.NoCounterpartyInfo, le:
                raise CommsHandler.CannotSendError, ("no counterparty public key", le)

            # Now call down to the next lower comms handler.
            # In the future, each lower level comm strat will have a `send()' method and we'll just call it.  This will simplify several things in the comms code.
            # Here we check to see whether we have a new style lower level comm strat with "hasattr()".  If so use it, if not pass this call down to the
            # tcpch (TCPCommsHandler).
            # debug.mojolog.write("CryptoCommsHandler: llstrat: %s, hasattr(llstrat, 'send'): %s\n", args=(llstrat, hasattr(llstrat, 'send')), v=4, vs="debug")
            if hasattr(llstrat, 'send'):
                # debug.mojolog.write("CryptoCommsHandler: doing llstrat.send(): %s\n", args=(llstrat,), v=4, vs="debug")
                llstrat.send(msg=cyphertext, mojooffer=mojooffer, hint=hint, fast_fail_handler=fast_fail_handler, timeout=timeout)
            else:
                self._tcpch.send_msg(counterparty_id, msg=cyphertext, hint=hint, fast_fail_handler=fast_fail_handler, timeout=timeout)

    def _inmsg_handler(self, msg, lowerstrategy, strategy_id_for_debug):
        """
        Processes an incoming message.

        @param lowerstrategy a comm strategy which is suggested by the caller;  Currently if the caller is a TCP comms handler, then this is a strategy for using the extant TCP connection, upon which `msg' arrived, to send further messages to the counterparty.  If the caller is a Relay comms handler, then this is `None'.  TODO: make it so that you consider using the same relay server to send back
        @param strategy_id_for_debug is a unique identifier which the *lower-level* comms handler uses to identify `lowerstrategy';  This is only used for diagnostic output.  Currently if the lower-level comms is TCP, and we don't yet know a public key which has signed messages that were sent over this TCP connection, then it is a random id, else it is the id of the public key which has signed messages that were sent over this TCP connection.  If the lower-level comms is Relay, then this is `None'.

        @precondition `msg' must be a string.: type(msg) == types.StringType: "msg: %s :: %s" % (hr(msg), hr(type(msg)))
        @precondition `strategy_id_for_debug' must be None if and only if `lowerstrategy' is None.: (strategy_id_for_debug is None) == (lowerstrategy is None): "strategy_id_for_debug: %s, lowerstrategy: %s" % (hr(strategy_id_for_debug), hr(lowerstrategy))
        """
        assert type(msg) == types.StringType, "precondition: `msg' must be a string." + " -- " + "msg: %s :: %s" % (hr(msg), hr(type(msg)))
        assert (strategy_id_for_debug is None) == (lowerstrategy is None), "precondition: `strategy_id_for_debug' must be None if and only if `lowerstrategy' is None." + " -- " + "strategy_id_for_debug: %s, lowerstrategy: %s" % (hr(strategy_id_for_debug), hr(lowerstrategy))

        try:
            counterparty_pub_key, cleartext = self._mesgen.parse(msg)
        except mesgen.Error, e:
            if isinstance(e, mesgen.SessionInvalidated):
                debug.mojolog.write("%s\n", args=(e.args[0],), v=3, vs="crypto")
                return

            if lowerstrategy:
                if hasattr(lowerstrategy, 'asyncsock'):
                    debug.mojolog.write("WARNING: a message arrived with suggested lowerstrategy %s, asyncsock: %s, strategy_id_for_debug: %s, that couldn't be decrypted.  Perhaps it was cleartext, or garbled.  The message was %s.  The error was: %s\n", args= (lowerstrategy.toDict(), lowerstrategy.asyncsock, strategy_id_for_debug, msg, e), v=1, vs="crypto")
                else:
                    debug.mojolog.write("WARNING: a message arrived with suggested lowerstrategy %s, strategy_id_for_debug: %s, that couldn't be decrypted.  Perhaps it was cleartext, or garbled.  The message was %s.  The error was: %s\n", args=(lowerstrategy.toDict(), strategy_id_for_debug, msg, e), v=1, vs="crypto")
            else:
                debug.mojolog.write("WARNING: a message arrived with suggested strategy_id_for_debug %s that couldn't be decrypted.  Perhaps it was cleartext, or garbled.  The message was %s.  The error was: %s\n", args=(strategy_id_for_debug, msg, e), v=1, vs="crypto")

            # if it is an UnknownSession error, attempt to send a note back down the lowerstrategy if it happens to be a two way connection
            if isinstance(e, mesgen.UnknownSession) and lowerstrategy:
                # XXX if all CommStrats are updated in the future to have a send() method, just use it if the CommStrat has a "twowaycomms" flag?
                if isinstance(lowerstrategy, CommStrat.TCP) and lowerstrategy.asyncsock:
                    debug.mojolog.write("sending an 'invalidate session' note back on the TCP connection\n", v=1, vs="CryptoCommsHandler")
                    # send e.invalidate_session_msg back out this TCP Connection
                    lowerstrategy.asyncsock.send(e.invalidate_session_msg)
                else:
                    debug.mojolog.write("message did not come via a two way CommStrat, cannot send an 'invalidate session'\n", v=3, vs="CryptoCommsHandler")

            return # drop it on the floor

        counterparty_id = idlib.string_to_id(counterparty_pub_key)

        assert MojoKey.publicRSAKeyForCommunicationSecurityIsWellFormed(counterparty_pub_key), "postcondition of `mesgen.parse()': `counterparty_pub_key' is a public key." + " -- " + "counterparty_pub_key: %s" % hr(counterparty_pub_key)

        if len(cleartext) > MAX_CLEARTEXT_LEN :
            debug.mojolog.write("msg from %s dropped due to length: %s\n", args=(counterparty_id, len(cleartext)), v=0, vs="ERROR")
            return
        try:
            decomptext = mojoutil.safe_zlib_decompress_to_retval(cleartext, maxlen=MAX_CLEARTEXT_LEN)
        except (mojoutil.UnsafeDecompressError, mojoutil.TooBigError), le:
            debug.mojolog.write("msg from %s dropped. le: %s\n", args=(counterparty_id, le), v=0, vs="ERROR")
            return
        except mojoutil.ZlibError, le:
            debug.mojolog.write("msg from %s not zlib encoded. processing it anyway for backwards compatibility. le: %s\n", args=(counterparty_id, le,), v=3, vs="debug")
            decomptext = cleartext

        if lowerstrategy is not None:
            self._doQ.add_task(self.use_comm_strategy, args=(counterparty_id, CommStrat.Crypto(counterparty_pub_key, lowerstrategy)))
        self._doQ.add_task(self._push_msg_up, args=(counterparty_id, decomptext, self._cid_to_cs.get(counterparty_id)))

    def _forget_comm_strategy(self, counterparty_id):
        """
        @precondition This thread must be the DoQ thread.: self._doQ.is_currently_doq()
        """
        assert self._doQ.is_currently_doq(), "precondition: This thread must be the DoQ thread."

        curcs = self._cid_to_cs.get(counterparty_id)
        if curcs:
            debug.mojolog.write("CryptoCommsHandler: Forgetting comm strategy.  counterparty_id: %s, curcs: %s\n", args=(counterparty_id, curcs,), v=4, vs="commstrats")
            ls = curcs._lowerstrategy
            if ls:
                if isinstance(ls, CommStrat.TCP):
                    self._tcpch.forget_comm_strategy(counterparty_id, ls)

            curcs._lowerstrategy = None
            del self._cid_to_cs[counterparty_id]

    def forget_comm_strategy(self, counterparty_id):
        """
        If the higher-level code realizes that a comm strategy is failing, they can call
        `forget_comm_strategy()' to get rid of it.

        TODO: have a sorted list instead of just one CS.  Maybe change this to "demote
        commstrat"?  --Zooko 2001-02-04

        XXX XYZ need to pass which CS to forget, here.  --Zooko 2001-02-05
        """
        if self._doQ.is_currently_doq():
            self._forget_comm_strategy(counterparty_id)
        else:
            self._doQ.add_task(self._forget_comm_strategy, args=(counterparty_id,))

    def hold_incoming_connections(self):
        self._tcpch.hold_incoming_connections()

    def unhold_incoming_connections(self):
        self._tcpch.unhold_incoming_connections()

    def use_comm_strategy(self, counterparty_id, commstrat, orig_cs=None):
        """
        There are several things to do to with this new suggested the comm strat.

        First, make sure that CryptoCommsHandler has a CommStrat.Crypto for this counterparty.

        Second, choose your favourite among available lowerstrategies.

        Third, if you are replacing the current one with the new one then forget the current one.

        Fourth, start using the new one.

        @param commstrat a CommStrat.Crypto which is suggested to use;  The `_lowerstrategy' may
            or may not be adopted depending on whether it is better than the current one.  If
            you want to force adoption of a new strategy, call `forget_comm_strategy()' first.

        @precondition This thread must be the DoQ thread.: self._doQ.is_currently_doq()
        """
        assert self._doQ.is_currently_doq(), "precondition: This thread must be the DoQ thread."

        assert idlib.equals(idlib.make_id(commstrat._pubkey, 'broker'), counterparty_id)

        # If we already have a CommStrat.Crypto object then we'll continue to use it (it might
        # have useful hint information in it) instead of the new one, but we'll examine the new
        # one in order to consider taking the newly suggested lowerstrategy.

        curcs = self._cid_to_cs.get(counterparty_id)

        # See if there are already any lowerstrategies available...

        if curcs:
            assert curcs._pubkey == commstrat._pubkey

            preferredlcs = CommStrat.choose_best_strategy(curcs._lowerstrategy, commstrat._lowerstrategy)

            # Now if we are replacing the curcs with the new one, then forget the old one:
            if (preferredlcs is not curcs._lowerstrategy) and (isinstance(curcs._lowerstrategy, CommStrat.TCP)):
                debug.mojolog.write("forgetting comm strategy before using new one.  counterparty_id: %s, preferredlcs: %s, curcs: %s\n", args=(counterparty_id, preferredlcs, curcs,), v=5, vs="debug")
                self._tcpch.forget_comm_strategy(counterparty_id, curcs._lowerstrategy)
                # XXX TODO save the old one in case the new one fails...  --Zooko 2001-05-15

            # Now start using the new one:
            # XXX why the dependency on the type?
            # answer: in the future, this will go away, when the CommStrat.TCP itself manages itself and we don't need
            # to inform the _tcpch about it.
            if isinstance(preferredlcs, CommStrat.TCP):
                self._tcpch.use_comm_strategy(counterparty_id, preferredlcs)

        else:
            # New crypto comm strat!
            self._cid_to_cs[counterparty_id] = commstrat
            self._mesgen.store_key(commstrat._pubkey)
            # XXX why the dependency on the type?
            if isinstance(commstrat._lowerstrategy, CommStrat.TCP):
                self._tcpch.use_comm_strategy(counterparty_id, commstrat._lowerstrategy)

    def _push_msg_up(self, counterparty_id, msg, commstrat):
        """
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        """
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

        return self._upward_inmsg_handler(counterparty_id, msg, commstrat)

    def get_id(self):
        return idlib.canonicalize(self._mesgen.get_id(), "broker")

    def get_comm_strategy(self):
        return CommStrat.Crypto(pubkey=self.get_public_key(), lowerstrategy=self._tcpch.get_comm_strategy())

    def get_public_key(self):
        return self._mesgen.get_public_key()

def _help_test_make_mesgen():
    testdir="/var/tmp/RunTests"
    return mesgen.create_MessageMaker(testdir) 

def test_send_msg_a_to_b():
    doQ = DoQ.DoQ() # test
    try:
        cA = CryptoCommsHandler(_help_test_make_mesgen(), tcpch=TCPCommsHandler.TCPCommsHandler(), doQ=doQ)
        cB = CryptoCommsHandler(_help_test_make_mesgen(), tcpch=TCPCommsHandler.TCPCommsHandler(), doQ=doQ)
        cBid = cB.get_id()

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper()

        cB.start_listening(_wcb1)

        cA.use_comm_strategy(cBid, cB.get_comm_strategy())
        # cA.start_listening(lambda counterparty_id, msg, commstrat: counterparty_id) # This is just to satisfy a silly assertion which is in there to prevent people from doing something they don't mean.
        cA.send_msg(cBid, "ping")
        doneflag1.wait(2) # If it takes longer than 2 seconds you fail the test!
    finally:
        try:
            assert (reskw1.get('msg') == "ping") or ((len(res1) > 1) and (res1[1] == "ping")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))
        finally:
            doQ.shutdown()

def test_send_msg_a_to_b_then_back_on_same_connection():
    doQ = DoQ.DoQ() # test
    try:
        cA = CryptoCommsHandler(_help_test_make_mesgen(), tcpch=TCPCommsHandler.TCPCommsHandler(), doQ=doQ)
        cB = CryptoCommsHandler(_help_test_make_mesgen(), tcpch=TCPCommsHandler.TCPCommsHandler(), doQ=doQ)
        cAid = cA.get_id()
        cBid = cB.get_id()

        def _handle_first_msg(counterparty_id, msg, commstrat, cB=cB, *args, **kwargs):
            if args or kwargs:
                print "------ unhandled args: %s, kwargs: %s" % (hr(args), hr(kwargs))

            cB.use_comm_strategy(counterparty_id, commstrat)
            cB.send_msg(counterparty_id, "pong")

        def _handle_second_msg(counterparty_id, msg, commstrat, *args, **kwargs):
            if args or kwargs:
                print "------ unhandled args: %s, kwargs: %s" % (hr(args), hr(kwargs))

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper(_handle_first_msg)
        res2, reskw2, doneflag2, _wcb2 = mojoutil._cb_warper(_handle_second_msg)

        cB.start_listening(_wcb1)

        cA.use_comm_strategy(cBid, cB.get_comm_strategy())
        cA.start_listening(_wcb2)
        cA.send_msg(cBid, "ping")

        doneflag1.wait(2) # If it takes longer than 2 seconds you fail the test!
        doneflag2.wait(2) # If it takes longer than 2 seconds you fail the test!
    finally:
        try:
            assert (reskw1.get('msg') == "ping") or ((len(res1) > 1) and (res1[1] == "ping")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))
            assert (reskw2.get('msg') == "pong") or ((len(res1) > 1) and (res2[1] == "pong")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))
        finally:
            doQ.shutdown()

def test_send_msg_a_to_b_then_back_on_different_connection():
    doQ = DoQ.DoQ() # test
    try:
        cA = CryptoCommsHandler(_help_test_make_mesgen(), tcpch=TCPCommsHandler.TCPCommsHandler(), doQ=doQ)
        cB = CryptoCommsHandler(_help_test_make_mesgen(), tcpch=TCPCommsHandler.TCPCommsHandler(), doQ=doQ)
        cAid = cA.get_id()
        cBid = cB.get_id()

        def _handle_first_msg(counterparty_id, msg, commstrat, cB=cB, *args, **kwargs):
            if args or kwargs:
                print "------ unhandled args: %s, kwargs: %s" % (hr(args), hr(kwargs))

            assert msg == "ping", "msg: %s" % hr(msg)

            cB.send_msg(counterparty_id, "pong")

        def _handle_second_msg(counterparty_id, msg, commstrat, *args, **kwargs):
            if args or kwargs:
                print "------ unhandled args: %s, kwargs: %s" % (hr(args), hr(kwargs))

            assert msg == "pong", "msg: %s" % hr(msg)

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper(_handle_first_msg)
        res2, reskw2, doneflag2, _wcb2 = mojoutil._cb_warper(_handle_second_msg)

        cB.start_listening(_wcb1)
        cA.start_listening(_wcb2)

        cA.use_comm_strategy(cBid, cB.get_comm_strategy())
        cB.use_comm_strategy(cAid, cA.get_comm_strategy())
        cA.send_msg(cBid, "ping")

        doneflag1.wait(2) # If it takes longer than 2 seconds you fail the test!
        doneflag2.wait(2) # If it takes longer than 2 seconds you fail the test!
    finally:
        try:
            assert (reskw1.get('msg') == "ping") or ((len(res1) > 1) and (res1[1] == "ping")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))
            assert (reskw2.get('msg') == "pong") or ((len(res2) > 1) and (res2[1] == "pong")), "reskw2: %s, res2: %s" % (hr(reskw2), hr(res2))
        finally:
            doQ.shutdown()


# Generic stuff
NAME_OF_THIS_MODULE="CryptoCommsHandler"

mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests(NAME_OF_THIS_MODULE)

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
