#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: DoQ.py,v 1.81 2001/06/07 23:00:11 greg Exp $'


# standard modules
import bisect
import profile
import sys
import tempfile
import threading
import time
import traceback
import types


# our modules
import MojoMessage
import MojoConstants
from MojoConstants import true, false
import SingleThreader
import mojoasyncore
from confutils import confman
import debug
from human_readable import hr
import idlib
import mojoutil
import mencode
from AssertApply import assert_apply_will_work

if confman.is_true_bool(('DEBUG_GC',)):
    import gc
    gc.set_threshold(1000000,1000000,1000000)
    gc.set_debug(gc.DEBUG_STATS|gc.DEBUG_UNCOLLECTABLE|gc.DEBUG_INSTANCES|gc.DEBUG_OBJECTS)

   
# Here is a global reference to the doq.  --Zooko 2001-04-14
global doq
doq = None

import exceptions

class ShutdownException(exceptions.StandardError): pass

# How many seconds the DoQ can attempt to invoke a function before that function is considered "blocking".
# This is just for debugging.
BLOCKTIME=720

IDLEDELAY=4.3

class DoQ:
    THE_NAME="The DoQ Thread "

    def __init__(self, maxqsize=0, blocktime=BLOCKTIME):
        debug.stderr.write("DoQ: Creating the DoQ.\n", v=4, vs="DoQ")

        self.blocktime = blocktime

        self.if_doQ_thread_i_am_i_am_then_start_new_doQ_thread_with_spam = self.dissociate
        
        self._lock = threading.Condition()

        self._doneflag = threading.Event() # This goes to `set' when the DoQ has stopped.

        self._flag = threading.Event()  # This is used to wake up the watcher thread
        self._lastttdtime = time.time() # The time that the last thingtodo was started. this seems obsolete

        self._last_wait_time = time.time()  # last time we waited around twiddling our thumbs
       
        # In the `evQ' below, `stacktrace' is an optional value for debugging.  It contains
        # the entire stack trace of where you were when you put this task onto the DoQ.  In
        # case of exception, this will be printed.  This is resource-consuming so it is done
        # only when MojoConstants.DEBUG is true.
        self._evQ = [] # Things to do: (when, func, args, kwargs, callback, stacktrace).  If func is `None' then shut down now.
        self._idleevQ = [] # Things to do when not doing anything from evQ: (when, func, args, kwargs, callback, stacktrace).
        # DoQTaskBatchers can be used to make the DoQ two dimensional so that scheduling tons
        # of tasks at once doesn't delay other more important tasks from running.
        self._taskbatchers = {}

        self._asyncorelooper = mojoasyncore.Asyncore(doq=self)

        global doq
        assert doq is None, "there should only be one doq at a time"
        doq = self
        assert doq is not None, "there should be a doq referenced by the global reference"

        self.start()

    def __del__(self):
        global doq
        assert doq is not None, "there should be a doq referenced by the global reference"
        doq = None

        
    def _start_a_new_doq_thread(self):
        # we -must- set this in here because dissociate() calls this
        # and we don't want the watcher_thread to notice an old hung event in
        # the case where there isn't any new event for this new DoQ
        # thread to start doing immediately.  (thanks to zooko for
        # noticing this! -greg)
        self._nowdoingsomething = false
        self._flag.clear()
        # _the_thread is the "Do Queue baton"
        if MojoConstants.DEBUG_MODE:
            self._the_thread = threading.Thread(target=self._do_the_doQ, name=self.THE_NAME + hr(traceback.extract_stack())) # verbose -- for debugging
        else:
            self._the_thread = threading.Thread(target=self._do_the_doQ, name=self.THE_NAME) # non-verbose
        self._the_thread.setDaemon(1)
        self._the_thread.start()

    def start(self):
        # flags that the watcher_thread uses
        self._nowdoingsomething = false
        self._timesomethingstarted = None
        self._timetoquit = false
        self._flag.clear()

        # start the monitoring safety net thread
        self.watcher_thread_running = false
        self.watcher_thread = threading.Thread(target=self._watch, name='DoQ Watcher')
        self.watcher_thread.setDaemon(1)
        self.watcher_thread.start()

        # start the actual DoQ thread (and kick asyncore off and running - HACK)
        self._start_a_new_doq_thread()
        self._asyncorelooper.start_asyncore()

    def restart(self):
        self.start()

    def flush(self):
        """
        Waits till everything currently on the DoQ is complete. Useful for testing.

        @precondition This method must not be called from the DoQ.: not self.is_currently_doq()
        """
        assert not self.is_currently_doq(), "precondition: This method must not be called from the DoQ."

        flag = threading.Event()
        self.add_task(flag.set)
        flag.wait()

    def dissociate(self):
        """
        This spawns a new thread and passes the Do Queue baton to it
        if the calling thread is the Do Queue thread.  Otherwise it
        does nothing.

        Call this before you make any blocking call if there is any
        possibility of yourself being the Do Queue thread.
        """
        self._lock.acquire()
        try:
            if self._the_thread is threading.currentThread():
                debug.stderr.write("DoQ: Now doing Greg and Zooko's ultra-klooooge.\n", v=8, vs="DoQ")

                self._start_a_new_doq_thread()
        finally:
            self._lock.release()

    spam_dance = dissociate

    def _shutdown(self):
        self._lock.acquire()
        try:
            self._timetoquit = true
            self._flag.set()  # tell the watcher_thread to stop
            debug.stderr.write("DoQ: Shutting down the DoQ.\n", v=0, vs="DoQ")

            if MojoConstants.DEBUG_MODE:
                tb = traceback.extract_stack()
            else:
                tb = None
            newevent = (time.time(), None, None, None, None, tb, None)
            index = bisect.bisect(self._evQ, newevent)
            self._evQ.insert(index, newevent)
            self._asyncorelooper.stop_asyncore()
            if index == 0:
                self._lock.notify()
        finally:
            self._lock.release()

    def add_idle_task(self, func, args=(), kwargs={}, delay=0, callback=None, errorfunc=None):
        """
        @param callback if not `None', this callback will be called and passed one argument
            containing the return value from the invokation of `func'.
        @param errorfunc if not `None' this callable is called and passed three _keyword_ arguments:
            (`etype', `evalue', `stacktrace')  The values correspond to the return values of sys.exc_info,
            except `stacktrace' is the stracktrace of this function call, rather than the actual stacktrace
            of the DoQ call.

        @precondition `func' must be callable.: callable(func): "func: %s" % hr(func)
        @precondition `callback' must be `None' or callable.: (not callback) or callable(callback)
        @precondition `errorfunc' must be `None' or callable.: (not errorfunc) or callable(errorfunc)
        @precondition `delay' must be a number.: (type(delay) == types.IntType) or (type(delay) == types.LongType) or (type(delay) == types.FloatType): "delay: %s" % hr(delay)
        @precondition `args' must be None or a sequence.: (args is None) or (type(args) == types.TupleType) or (type(args) == types.ListType): "args: %s" % hr(args)
        """
        assert callable(func), "precondition: `func' must be callable." + " -- " + "func: %s" % hr(func)
        assert (not callback) or callable(callback), "precondition: `callback' must be `None' or callable."
        assert (not errorfunc) or callable(errorfunc), "precondition: `errorfunc' must be `None' or callable."
        assert (type(delay) == types.IntType) or (type(delay) == types.LongType) or (type(delay) == types.FloatType), "precondition: `delay' must be a number." + " -- " + "delay: %s" % hr(delay)
        assert (args is None) or (type(args) == types.TupleType) or (type(args) == types.ListType), "precondition: `args' must be None or a sequence." + " -- " + "args: %s" % hr(args)

        debug.mojolog.write("DoQ: adding when-idle task to DoQ with delay %s: %s(%s, %s)\n", args=(delay, func, args, kwargs,), v=21, vs="DoQ")

        self._lock.acquire()
        try:
            if MojoConstants.DEBUG_MODE:
                tb = traceback.extract_stack()
            else:
                tb = None
            newevent = (time.time() + delay, func, args, kwargs, callback, tb, errorfunc)
            index = bisect.bisect(self._idleevQ, newevent)
            self._idleevQ.insert(index, newevent)
            if index == 0:
                self._lock.notify()
        finally:
            self._lock.release()

    def shutdown_slowly_and_gracefully(self):
        self.add_idle_task(func=self._shutdown)

    def wait_until_done_shutting_down(self):
        self._doneflag.wait()

    def suspend(self):
        return self.shutdown_slowly_and_gracefully()

    def is_currently_doq(self):
        return self._the_thread is threading.currentThread()

    def _watch(self):
        self._flag.wait(2)  # do this to delay our start incase an old _watch thread is still shutting down
        assert not self.watcher_thread_running, "More than one DoQ watcher may not be started"
        while 1:
            debug.mojolog.write('DoQ watcher thread is watching\n', v=21, vs="DoQ")
            self._flag.wait(self.blocktime/2.0)  # wake up every now and then and check the status
            if self._timetoquit:
                self.watcher_thread_running = None
                return
            # if the DoQ is in the middle of doing something, and it
            # has been for over the self.blocktime, consider it hung.
            self._lock.acquire()
            try:
                if self._nowdoingsomething and (self._timesomethingstarted + self.blocktime < time.time()) :
                    debug.stderr.write("ERROR: The DoQ thread appears to be blocked.  NOT starting a new one.  The method that apparently blocked was: %s(%s, %s), which was put onto the DoQ from this stack: %s\n", args=(self._dbfunc, self._dbargs, self._dbkwargs, self._dbtb), v=0, vs="ERROR")
                    # self._start_a_new_doq_thread()
                else:
                    self._flag.clear()
            finally:
                self._lock.release()

    def _do_the_doQ(self):
        return mojoutil.profit(self._internal__do_the_doQ)

    def _do_thing(self, func, args=(), kwargs={}, callback=None, stacktrace=None, errorfunc=None):
        self._dbfunc = func # for debugging only
        self._dbargs = args # for debugging only
        self._dbkwargs = kwargs # for debugging only
        self._dbtb = stacktrace # for debugging only

        # exit if someone is shutting down the DoQ/RealScheduler
        if func is None:
            raise ShutdownException

        # Never hold a lock while making an external call!
        self._lock.release()
        try:
            self._timesomethingstarted = time.time()  # always set the time before setting _nowdoingsomething
            self._nowdoingsomething = true
            self._flag.clear()
            debug.mojolog.write("DoQ: now doing: %s(%s, %s)\n", args=(func, args, kwargs), v=19, vs="DoQ")

            # Do the thing off the queue
            result = mojoutil.callback_wrapper(func, args, kwargs)
            if confman.is_true_bool(('DEBUG_GC',)):
                collected = gc.collect()
                debug.mojolog.write("DoQ: DEBUG_GC: collected: %s\n", args=(collected,), v=5, vs="DEBUG_GC")

            # queue up its 'finished' callback if there was one
            if callback:
                self.add_task(callback, args=(result,))

            debug.mojolog.write("DoQ: done doing: %s(%s, %s)\n", args=(func, args, kwargs), v=20, vs="DoQ")
        except:
            self._lock.acquire()
            if errorfunc:
                debug.mojolog.write("_do_the_doQ caught an exception.  Calling the events errorfunc (%s)\n", args=(errorfunc,), v=2, vs="DoQ")
                (etype, evalue, realtb) = sys.exc_info()
                try:
                    errorfunc(etype=etype, evalue=evalue, stacktrace=realtb)
                    del realtb # The documentation for sys.exc_info recommends this.  XXX because of circular garbage issues, which are no longer a concern in python >= 2.0.  --Zooko 2001-06-03
                    realtb = None # This is for the del statement below.  --Neju 2001-06-06
                except:
                    debug.mojolog.write("_do_the_doQ caught an exception when calling errorfunc %s with parameters (%s, %s, %s):\n", args=(errorfunc, etype, evalue, stacktrace), vs="ERROR")
                    traceback.print_exc(file=debug.mojolog)
                del realtb # The documentation for sys.exc_info recommends this.

            else:
                (etype, evalue, realtb) = sys.exc_info()
                debug.mojolog.write("_do_the_doQ caught an exception.  The stack trace from when this task was added to the DoQ is: %s.  exception: %s :: %s, the stack trace follows:\n", args=(stacktrace, evalue, etype,), v=0, vs="ERROR")
                del realtb # The documentation for sys.exc_info recommends this.  XXX because of circular garbage issues, which are no longer a concern in python >= 2.0.  --Zooko 2001-06-03
                traceback.print_exc(file=debug.mojolog)

        self._nowdoingsomething = false
        self._lock.acquire()

    def _internal__do_the_doQ(self):
        # exit if we no longer have the Do Queue baton
        if self._the_thread is not threading.currentThread():
            return

        debug.stderr.write("new DoQ thread started: %s\n", args=(threading.currentThread(),), v=3, vs="DoQ")
        while true:
            self._lock.acquire() #...  XXXX Hm.  Is this lock necessary?  --Zooko 2000-08-31
            try:
                # exit if we no longer have the Do Queue baton
                if self._the_thread is not threading.currentThread():
                    return

                self._doneflag.clear()
                errorfunc = None
                while (len(self._evQ) > 0) and (self._evQ[0][0] <= time.time()):
                    # exit if we no longer have the Do Queue baton
                    if self._the_thread is not threading.currentThread():
                        return

                    debug.stderr.write("~~~ DoQ doing 1 of %s (%s) things\n", args=(self.get_length_of_immediate_queue(), len(self._evQ),), v=7, vs='DoQ')
                    (when, func, args, kwargs, callback, stacktrace, errorfunc) = self._evQ.pop(0)
                    try:
                        self._do_thing(func=func, args=args, kwargs=kwargs, callback=callback, stacktrace=stacktrace, errorfunc=errorfunc)
                    except ShutdownException:
                        self._doneflag.set()
                        return

                    # exit if we no longer have the Do Queue baton
                    if self._the_thread is not threading.currentThread():
                        return

                if len(self._evQ) == 0:
                    self._last_wait_time = time.time()
                    # execute something from the idle task queue here
                    if (len(self._idleevQ) > 0) and (self._idleevQ[0][0] <= time.time()):
                        debug.stderr.write("~~~ DoQ doing 1 of %s (%s) idle queue things\n", args=(self.get_length_of_immediate_idle_queue(), len(self._idleevQ),), v=7, vs='DoQ')
                        (when, func, args, kwargs, callback, stacktrace, errorfunc) = self._idleevQ.pop(0)
                        try:
                            self._do_thing(func=func, args=args, kwargs=kwargs, callback=callback, stacktrace=stacktrace, errorfunc=errorfunc)
                        except ShutdownException:
                            self._doneflag.set()
                            return

                    self._lock.wait()
                    self._last_wait_time = time.time()
                else:
                    delay = self._evQ[0][0] - time.time()
                    self._last_wait_time = time.time()
                    # if delay > idle_delay:
                    #    execute something from the idle task queue here
                    if (delay > IDLEDELAY) and (len(self._idleevQ) > 0) and (self._idleevQ[0][0] <= time.time()):
                        debug.stderr.write("~~~ DoQ doing 1 of %s (%s) idle queue things\n", args=(self.get_length_of_immediate_idle_queue(), len(self._idleevQ),), v=7, vs='DoQ')
                        (when, func, args, kwargs, callback, stacktrace, errorfunc) = self._idleevQ.pop(0)
                        try:
                            self._do_thing(func=func, args=args, kwargs=kwargs, callback=callback, stacktrace=stacktrace, errorfunc=errorfunc)
                        except ShutdownException:
                            self._doneflag.set()
                            return

                    self._lock.wait(delay)
                    self._last_wait_time = time.time()
            finally:
                self._lock.release()

        debug.stderr.write("DoQ: Former DoQ thread %s, is now entering a well-earned retirement.\n", args=(threading.currentThread().getName(),), v=14, vs="DoQ")

    def add_batch_task(self, batch, func, args=(), kwargs={}):
        """
        @param batch if given specifies a DoQTaskBatcher name to add this task to instead of directly on the DoQ.
            if no DoQTaskBatcher of the given name exists, a new one will be created.
        """
        ##return self.add_task(func, args=args, kwargs=kwargs) # uncomment this if you want to disable batch support for testing
        assert_apply_will_work(func, args = args, kwargs = kwargs)
        assert (args is None) or (type(args) == types.TupleType) or (type(args) == types.ListType), "precondition: `args' must be None or a sequence." + " -- " + "args: %s" % hr(args)
        assert (type(batch) == types.StringType), "batch must be a string"

        debug.mojolog.write("adding to DoQ batch %s: %s(%s, %s)\n", args=(batch, func, args, kwargs), v=21, vs="DoQ")

        if batch is not None:
            batcher = None
            self._lock.acquire()
            try:
                if self._taskbatchers.has_key(batch):
                    batcher = self._taskbatchers[batch]
                else:
                    batcher = DoQTaskBatcher(doq=self, name=batch)
                    self._taskbatchers[batch] = batcher
            finally:
                self._lock.release()
            # this must be done outside of the lock as it may call doq.add_task
            batcher.add_task(func, args, kwargs)


    def add_task(self, func, args=(), kwargs={}, delay=0, callback=None, errorfunc=None):
        """
        @param callback if not `None', this callback will be called and passed one argument
            containing the return value from the invokation of `func'.
        @param errorfunc if not `None' this callable is called and passed three _keyword_ arguments:
            (`etype', `evalue', `stacktrace')  The values correspond to the return values of sys.exc_info,
            except `stacktrace' is the stracktrace of this function call, rather than the actual stacktrace
            of the DoQ call.

        @precondition `func' must be callable.: callable(func): "func: %s" % hr(func)
        @precondition `callback' must be `None' or callable.: (not callback) or callable(callback)
        @precondition `errorfunc' must be `None' or callable.: (not errorfunc) or callable(errorfunc)
        @precondition `delay' must be a number.: (type(delay) == types.IntType) or (type(delay) == types.LongType) or (type(delay) == types.FloatType): "delay: %s" % hr(delay)
        @precondition `args' must be None or a sequence.: (args is None) or (type(args) == types.TupleType) or (type(args) == types.ListType): "args: %s" % hr(args)
        """
        assert_apply_will_work(func, args = args, kwargs = kwargs)
        assert (not callback) or callable(callback), "precondition: `callback' must be `None' or callable."
        assert (not errorfunc) or callable(errorfunc), "precondition: `errorfunc' must be `None' or callable."
        assert (type(delay) == types.IntType) or (type(delay) == types.LongType) or (type(delay) == types.FloatType), "precondition: `delay' must be a number." + " -- " + "delay: %s" % hr(delay)
        assert (args is None) or (type(args) == types.TupleType) or (type(args) == types.ListType), "precondition: `args' must be None or a sequence." + " -- " + "args: %s" % hr(args)

        debug.mojolog.write("DoQ: adding to DoQ with delay %s: %s(%s, %s)\n", args=(delay, func, args, kwargs,), v=21, vs="DoQ")

        self._lock.acquire()
        try:
            if MojoConstants.DEBUG_MODE:
                tb = traceback.extract_stack()
            else:
                tb = None
            newevent = (time.time() + delay, func, args, kwargs, callback, tb, errorfunc)
            index = bisect.bisect(self._evQ, newevent)
            self._evQ.insert(index, newevent)
            if index == 0:
                self._lock.notify()
        finally:
            self._lock.release()

    def get_length_of_immediate_queue(self, delay=IDLEDELAY):
        """returns the number of things currently queued to be done in the next second"""
        index = bisect.bisect(self._evQ, (time.time()+delay,))
        return index

    def get_length_of_immediate_idle_queue(self, delay=IDLEDELAY):
        """returns the number of things currently queued to be done in the next second"""
        index = bisect.bisect(self._idleevQ, (time.time()+delay,))
        return index

class DoQTaskBatcher:
    """
    This queue feeds into the DoQ, batching tasks up into small groups
    and reinserting itself on the DoQ so that the tasks don't occupy space
    on the DoQ preventing other immediate tasks from running.

    Good for doing tasks which are numerous and can take a while
    but aren't urgent so they should not block others from running.
    """
    def __init__(self, doq, some=4, name=''):
        self.doq = doq
        self.some = some
        self.name = name
        self._list = []
        self._list_lock = threading.Lock()

    def add_task(self, task, args=(), kwargs={}):
        if MojoConstants.DEBUG_MODE:
            stacktrace = traceback.extract_stack()
        else:
            stacktrace = None
        self._list_lock.acquire()
        try:
            self._list.append((task, args, kwargs, stacktrace))
            # if the list was empty and we just added an item, start running again
            # (written like this so that self.doq.add_task is done after we've released our lock)
            if len(self._list) != 1:
                # self._do_some must already be on the queue, return without adding the task
                return
        finally:
            self._list_lock.release()
        self.doq.add_task(self._do_some)

    def _do_some(self):
        self._list_lock.acquire()
        try:
            chunk = self._list[:self.some]
            self._list = self._list[self.some:]
            need_to_do_more = (len(self._list) > 0)
        finally:
            self._list_lock.release()
        # reinsert _do_some on the DoQ if needed (outside of the list lock)
        if need_to_do_more:
            self.doq.add_task(self._do_some)
        debug.mojolog.write("DoQTaskBatcher(%s)._do_some doing %s tasks\n", args=(self.name, len(chunk)), v=7, vs="DoQ")
        for task, args, kwargs, stacktrace in chunk:
            try:
                apply(task, args, kwargs)
            except:
                debug.mojolog.write("DoQTaskBatcher(%s)._do_some caught an exception.  The stack trace from when this task was added is: %s.  The exception itself follows:\n", args=(self.name, stacktrace,), v=0, vs="ERROR")
                traceback.print_exc(file=debug.mojolog)


def test_is_currently_doq():
    q = DoQ()
    assert not q.is_currently_doq()
    spot = []
    f = threading.Event()
    def on_q(spot = spot, f = f, q = q):
        spot.append(q.is_currently_doq())
        f.set()
    q.add_task(on_q)
    f.wait()
    assert spot[0]

def test_do_something():
    q = DoQ() # test
    try:
        l = []
        def something(l):
            l.append(8)
        q.add_task(something, args=(l,))
        time.sleep(1)
        assert l == [8], "something never happened"
    finally:
        q.shutdown_slowly_and_gracefully()

def test_do_something_with_delay():
    q = DoQ() # test
    try:
        l = []
        def something(l):
            l.append(8)
        q.add_task(something, args=(l,), delay=1)
        assert l == [], "something happened too soon"
        time.sleep(2)
        assert l == [8], "something never happened"
    finally:
        q.shutdown_slowly_and_gracefully()
    

def test_do_something_that_stops_being_the_DoQ():
    q = DoQ() # test
    try:
        l = []
        def something(l, a, q=None):
            if q:
                q.if_doQ_thread_i_am_i_am_then_start_new_doQ_thread_with_spam()
                time.sleep(0.1)
            l.append(a)
        q.add_task(something, args=(l, 2), kwargs={'q':q}, delay=0.5)
        q.add_task(something, args=(l, 3), delay=1)
        time.sleep(2)
        assert l == [2, 3], "both somethings never happened %s" % l
        q.add_task(something, args=(l, 4))
        time.sleep(0.2)
        assert l == [2, 3, 4]
    finally:
        q.shutdown_slowly_and_gracefully()

def test_watcher_thread():
    q = DoQ(blocktime=1)
    try:
        original_thread = q._the_thread
        time.sleep(2)
        assert q._the_thread == original_thread, "watcher restarted DoQ when it shouldn't (1)"
        q.add_task(time.sleep, args=(3,))
        time.sleep(1)
        assert q._the_thread == original_thread, "watcher restarted DoQ when it shouldn't (2)"
        time.sleep(2)
        assert q._the_thread != original_thread, "watcher didn't restart DoQ"
    finally:
        q.shutdown_slowly_and_gracefully()

def test_watcher_dissociate_interaction():
    q = DoQ(blocktime=1.1)
    try:
        original_thread = q._the_thread
        time.sleep(2)
        def thing(q=q):
            time.sleep(0.5)
            q.dissociate()
            time.sleep(3)
        q.add_task(thing)
        assert q._the_thread == original_thread, "watcher restarted DoQ when it shouldn't (1)"
        time.sleep(2)
        second_thread = q._the_thread
        assert original_thread != second_thread, "watcher restarted DoQ when it shouldn't (2)"
        time.sleep(2)
        assert q._the_thread == second_thread, "watcher restarted DoQ when it shouldn't (3)"
    finally:
        q.shutdown_slowly_and_gracefully()

def test_flush():
    q = DoQ()
    l = threading.Event()
    def func(l = l):
        time.sleep(.5)
        l.set()
    q.add_task(func)
    q.flush()
    assert l.isSet()

# Generic stuff
NAME_OF_THIS_MODULE="DoQ"

mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests(NAME_OF_THIS_MODULE)

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
