#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: DownloadMonitor.py,v 1.6 2001/04/30 19:01:47 greg Exp $'

# standard modules
import types
from time import time

# our modules
import sharemap
from idlib import to_binary

import mojoutil

true = 1
false = None
class Unknown: pass


class PieceMonitor:
    def __init__(self, num_shares_needed, pieceid, length, shares, offset=None):
        """
        Most of the parameters of this contructor are the same as the members
        of a piecedict returned by sharemap.parse_sharemap so that you
        can merely apply them as kwargs.

        These objects should only be constructed by DownloadManager objects.
        """
        self.num_shares_needed = num_shares_needed
        self.pieceid = pieceid
        self.length = length
        self._sharenum_to_blockid_map = shares
        self.total_shares = len(shares)
        self.offset = offset

        # these are set as info comes in
        self.possible = Unknown        # indicates if this reassembly is possible or not
        self.completed = false         # true if this piece was successfully completed
        self.discarded = 0             # > 0 if the SingleReassembly object for this piece has been discarded
        self.found_sharenums = {}      # maps sharenum -> number of places we found it
        self.started_sharenums = {}    # maps sharenum -> times we attempted to download it
        self.completed_sharenums = {}  # if has_key(sharenum) it has been downloaded


    ##### Interface for a DownloadMonitor class to use to report progress: #####
    def ev_found_block(self, sharenum):
        count = self.found_sharenums.get(sharenum, 0)
        self.found_sharenums[sharenum] = count + 1

    def ev_reassembly_possible(self):
        self.possible = true

    def ev_failed_reassembly(self, reasonstr=""):
        self.possible = false

    def ev_started_block(self, sharenum):
        count = self.started_sharenums.get(sharenum, 0)
        self.started_sharenums[sharenum] = count + 1

    def ev_finished_block(self, sharenum):
        self.completed_sharenums[sharenum] = None

    def ev_finished_reassembly(self):
        self.completed = true

    def ev_discarded(self):
        self.discarded = self.discarded + 1


class DownloadMonitor:
    """
    SingleReassembly should seperately signal this class as each piece:
      * receives a location for one of its blockshares
      * fails to find any location for one of its blockshares
      * starts retrieving a block
      * successfully retrieves a block
      * successfully reassembles the piece
      * determines that not enough blocks are to be found
    """
    _public_methods_ = []   # allow COM access
    _public_attrs_ = ['starttime',  'iso_starttime',
                      'stoptime',   'iso_stoptime',
                      'num_errors', 'error_partnames',
                      'name', 'save_to_disk_path',
                     ]
    def __init__(self, name, piecelist, num_shares_needed, piecemonitorclass=PieceMonitor, save_to_disk_path=None):
        """
        @param name is text name useful for display for this download
        @param piecelist is a list of dicts keyed on:
          pieceid - the id (SHA-1 hash) of this piece
          length - the length of this piece
          offset - if present: the offset within the entire publication that this piece makes up
                   if not present: this piece is one of the sharemap pieces.
                   [see the sharemap.is_sharemap_piece() function]
          shares - a dict keyed on sharenums mapping to blockids
        @param piecemonitorclass is a class that implements the PieceMonitor interface
        @param save_to_disk_path may be used to indicate to others where on disk this download is being saved to
        """
        # public variables
        self.name = name
        self.save_to_disk_path = save_to_disk_path

        # for internal use
        assert num_shares_needed > 0
        self.piecemonitorclass = piecemonitorclass
        self.num_shares_needed = num_shares_needed

        self._piece_order = []
        self._pm_map = {}
        for piecedict in piecelist:
            piecemonitor = apply(self.piecemonitorclass, (self.num_shares_needed,), piecedict)
            self._piece_order.append(piecemonitor)
            self._pm_map[to_binary(piecedict['pieceid'])] = piecemonitor
        
        # record our start and stop times for statistical reasons
        self.starttime = time()
        self.iso_starttime = mojoutil.iso_utc_time(self.starttime)
        self.stoptime = 0
        self.iso_stoptime = ''

        # to be used by the creator to call increment_error_count on
        # each error encountered extracting desired data from this
        # publication.
        self.num_errors = 0
        # a list of strings associated with the errors, they can be specified when
        # increment_error_count is called.
        self.error_partnames = []
    
    def record_stoptime(self):
        self.stoptime = time()
        self.iso_stoptime = mojoutil.iso_utc_time(self.stoptime)

    def increment_error_count(self, partname=None):
        self.num_errors = self.num_errors + 1
        if partname is not None:
            self.error_partnames.append(partname)
    
    def extend_piecelist(self, piecelist):
        """
        Useful on new format dinode downloads where the sharemap and first few piece infos are in the dinode
        but more will become available after the sharemap has been retrieved.
        """
        for piecedict in piecelist:
            pieceid = to_binary(piecedict['pieceid'])
            if not self._pm_map.has_key(pieceid):
                piecemonitor = apply(self.piecemonitorclass, (self.num_shares_needed,), piecedict)
                self._piece_order.append(piecemonitor)
                self._pm_map[pieceid] = piecemonitor

    ##### Interface that SingleReassembly objects use to report their progress: #####
    def ev_found_block(self, pieceid, sharenum):
        """called only when a block could not be located or failed to download"""
        self._pm_map[to_binary(pieceid)].ev_found_block(sharenum)

    def ev_started_block(self, pieceid, sharenum):
        """called when a block download is started"""
        self._pm_map[to_binary(pieceid)].ev_started_block(sharenum)

    def ev_finished_block(self, pieceid, sharenum):
        """called when a block download completes successfully"""
        self._pm_map[to_binary(pieceid)].ev_finished_block(sharenum)

    def ev_reassembly_possible(self, pieceid):
        """called after enough blocks have known locations, always called before finished_reassembly"""
        self._pm_map[to_binary(pieceid)].ev_reassembly_possible()

    def ev_failed_reassembly(self, pieceid, reasonstr=""):
        """called when not enough blocks could be found"""
        self._pm_map[to_binary(pieceid)].ev_failed_reassembly(reasonstr=reasonstr)

    def ev_finished_reassembly(self, pieceid):
        """called when enough blocks have been retrieved and reassembled successfully"""
        self._pm_map[to_binary(pieceid)].ev_finished_reassembly()

    def ev_discarded(self, pieceid):
        """called when a SingleReassembly object is destroyed"""
        self._pm_map[to_binary(pieceid)].ev_discarded()



def test_instantiation():
    DownloadMonitor(name="test", num_shares_needed=4, piecelist=[{'pieceid': 'A'*27, 'length': 42, 'shares': {0:'A'*27}}], piecemonitorclass=PieceMonitor)


mojo_test_flag = 1
def run():
    import RunTests
    RunTests.runTests("DownloadMonitor")

if __name__ == '__main__':
    run()
