#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

import RunTests
import time
import idlib
import debug

true = 1
false = 0

class FailBasedSelecter:
    def __init__(self, interval_function, numfail_function, initial = None):
        assert not initial or idlib.is_sloppy_id(initial), "initial must be None or an id"
        if initial is not None:
            initial = idlib.to_binary(initial)
        self.interval_function = interval_function
        self.numfail_function = numfail_function
        # maps binary counterparty id -> list of fail times
        self.failures = {}
        self.changed = false
        self.current = initial

    def _is_acceptable(self, counterparty):
        cutoff = self.numfail_function()
        fs = 0
        list_of_fail_times = self.failures.get(idlib.canonicalize(counterparty))
        if list_of_fail_times is not None:
            fs = len(list_of_fail_times)
        #debug.mojolog.write("%s _is_acceptable returning %s\n", args=(counterparty, fs < cutoff), vs='FBS')
        return fs < cutoff

    def fail_came_in(self, counterparty):
        return self._fail_came_in(counterparty, time.time())

    def _fail_came_in(self, counterparty, current_time):
        counterparty = idlib.to_binary(counterparty)
        if self.failures.has_key(counterparty):
            self.failures[counterparty].append(current_time)
        else:
            self.failures[counterparty] = [current_time]
        self.changed = true

    def get_preferred(self, clist = []):
        return self._get_preferred(time.time(), clist)

    def _get_preferred(self, current_time, clist):
        """
        If last one being used is still acceptable, return it.
        
        Otherwise, pick the first acceptable one from clist 
        (clist is assumed to be sorted in descending order of preference)
        make that the new preferred thing and return it
        
        If neither of those techniques come up with anything, return None
        """
        if not self.changed:
            if self.current is not None:
                return self.current
        else:
            cutoff = current_time - self.interval_function()
            for counterparty, list_of_times in self.failures.items():
                new_times = []
                for t in list_of_times:
                    if t > cutoff:
                        new_times.append(t)
                # replace the list in self.failures with the shortened one
                list_of_times[:] = new_times
                #debug.mojolog.write("(cutoff %s) list of times for %s: %s\n", args=(cutoff, counterparty, new_times), vs='FBS')

            if self.current is not None and not self._is_acceptable(self.current):
                self.current = None
            self.changed = false
        if self.current is None:
            for i in clist:
                i = idlib.to_binary(i)
                if self._is_acceptable(i):
                    self.current = i
                    break
        return self.current

def test_normal_operation():
    a = idlib.make_new_random_id()
    f = FailBasedSelecter(lambda: 2, lambda: 2, a)
    assert f._get_preferred(0, []) == a

def test_no_initial():
    a = idlib.make_new_random_id()
    f = FailBasedSelecter(lambda: 2, lambda: 2)
    assert f._get_preferred(0, []) == None
    assert f._get_preferred(0, [a]) == a

def test_fail_thing():
    a = idlib.make_new_random_id()
    f = FailBasedSelecter(lambda: 2, lambda: 2, a)
    assert f._get_preferred(0, []) == a
    f._fail_came_in(a, 0)
    assert f._get_preferred(0, []) == a
    f._fail_came_in(a, 0)
    assert f._get_preferred(0, []) == None

def test_fail_other():
    a = idlib.make_new_random_id()
    b = idlib.make_new_random_id()
    f = FailBasedSelecter(lambda: 2, lambda: 2, a)
    f._fail_came_in(b, 0)
    assert f._get_preferred(0, []) == a
    f._fail_came_in(b, 0)
    assert f._get_preferred(0, []) == a
    f._fail_came_in(a, 0)
    assert f._get_preferred(0, []) == a
    f._fail_came_in(a, 0)
    assert f._get_preferred(0, []) == None
    assert f._get_preferred(0, [b]) == None

def test_failure_timeout():
    a = idlib.make_new_random_id()
    f = FailBasedSelecter(lambda: 2, lambda: 2, a)
    assert f._get_preferred(0, []) == a
    f._fail_came_in(a, 0)
    assert f._get_preferred(0, []) == a
    f._fail_came_in(a, 3)
    assert f._get_preferred(3, []) == a

mojo_test_flag = 1

if __name__ == '__main__':
    RunTests.runTests("FailBasedSelecter")

