#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import string
import time
import traceback
import types

### our modules
import BidStrategy
from Cache import StatsCache
from LastAccessedCache import LastAccessedCache
from MojoErrors import ReassemblyError
from BlockFinder import BlockFinder
from SingleReassembly import SingleReassembly
from BlockThrottler import BlockThrottler
import idlib
import debug
import mojoutil
from MojoConstants import true, false
import MojoConstants
from confutils import confman
from human_readable import hr


# this controls how many SingleReassembly objects are kept in ram
# after completing.  Do not make this too low or you will likely
# thrash the download cache and possibly have to refetch some blocks.
# Also, reading ahead more than this many SingleReassemblies at once
# is a bad idea.
# NOTE: this is used from mojohttpd to control save-to-disk readahead
_MaxPieceSize = int(confman['PUBLICATION']['MAX_SHARE_SIZE']) * int(confman['PUBLICATION']['SHARES_NEEDED'])
if _MaxPieceSize > 512*1024:
    NUM_FINISHED_REASSEMBLIES = 12
if _MaxPieceSize > 256*1024:
    NUM_FINISHED_REASSEMBLIES = 18
else:
    NUM_FINISHED_REASSEMBLIES = 24


class FindBeforehandBlockFetcher:
    def __init__(self, broker):
        self.broker = broker

        speed = int(confman.dict['CONN_SPEED_IN'])
        # these values probably need tweaking
        if speed > 2000:
            self.max_concurrent_request_blobs = 30
        if speed > 1000:
            self.max_concurrent_request_blobs = 20
        elif speed > 500:
            self.max_concurrent_request_blobs = 12
        elif speed >= 320:
            self.max_concurrent_request_blobs = 8
        elif speed > 180:
            self.max_concurrent_request_blobs = 5            
        elif speed >= 112:
            self.max_concurrent_request_blobs = 4            
        else:
            self.max_concurrent_request_blobs = 2
        
        # maps piece id's to SingleReassembly objects
        self.reassemblies = {}
        self.finished_reassemblies = LastAccessedCache(NUM_FINISHED_REASSEMBLIES)
        # maps piece id's to successfully completed SingleReassembly objects that self.free() has been called on.
        _cache_timeout = int(confman['PIECE_CACHE_TIMEOUT'])
        self.freed_successful_reassembly_cache = StatsCache(maxsize=int(confman['PIECE_CACHE_SIZE_v2']), autoexpireinterval=max(20, _cache_timeout/8), autoexpireparams={'maxage':_cache_timeout})

        # used internally to try and prevent start_readahead calls from unsplit.py
        # (in particular from seek()) when loading web archives from taking so long
        if int(confman['PUBLICATION']['MAX_SHARE_SIZE']) * int(confman['PUBLICATION']['SHARES_NEEDED']) > (256*1024):
            self._start_readahead_piece_id_cache = LastAccessedCache(30)
        else:
            self._start_readahead_piece_id_cache = LastAccessedCache(40)

        self.finder = BlockFinder(self.broker, self._find_function)
        self.throttler = BlockThrottler(doq = self.broker.doq, max_concurrent = self.max_concurrent_request_blobs)

    def get_piece(self, piecedict, num_blocks_needed, total_block_shares, priority = 2):
        """return the data for the piece described by piecedict.
        This defaults to priority 2 as it is a blocking call.
        """
        # to ensure that a prior call to start_readahead gets finished
        self.broker.doq.flush()
        piece_id = piecedict['pieceid']
        debug.mojolog.write("get_piece(%s)\n", args=(piecedict,), v=3, vs="FindBeforehandBlockFetcher")
        s = self.freed_successful_reassembly_cache.get(piece_id)
        if s is None:
            s = self.finished_reassemblies.get(piece_id)
        if s is not None:
            return s.reassemble()
        s = self.reassemblies.get(piece_id)
        if s is not None:
            if priority is not None:
                s.set_priority(priority)
            return s.reassemble()

        # no SingleReassembly for piece_id existed, cause one to be created and wait for it.
        self.start_readahead([piecedict], priority=priority, _checkcache=0, num_blocks_needed=num_blocks_needed, total_block_shares=total_block_shares)
        self.broker.doq.flush()

        s = self.finished_reassemblies.get(piece_id)
        if s is not None:
            return s.reassemble()
        s = self.reassemblies.get(piece_id)
        if s is not None:
            if priority is not None:
                s.set_priority(priority)
            return s.reassemble()
        raise ReassemblyError, "SingleReassembly object dropped too fast: %s" % hr(piecedict)


    def seed_finder(self, piece_list):
        """
        Tells the block finder about all of the blocks we need for all pieces in piece_list.
        """
        self.broker.doq.add_task(self._seed_finder, args=(piece_list,))
    
    def _seed_finder(self, piece_list):
        # tell the finder to start looking for all the blocks
        block_ids = []
        for i in piece_list:
            for j in i['shares'].values():
                block_ids.append(j)
        debug.mojolog.write('calling find_blocks for %s ids: %s\n', args=(len(block_ids), block_ids), v=3, vs="FindBeforehandBlockFetcher")
        self.finder.find_blocks(block_ids)

    def start_readahead(self, piece_list, priority = None, _checkcache=1, num_blocks_needed=None, total_block_shares=None, download_monitor=None):
        if _checkcache:
            def need_to_call_readahead(piece_dict, priority=priority, srcache=self._start_readahead_piece_id_cache):
                prev_priority = srcache.get(piece_dict['pieceid'], 'notfound')
                if prev_priority == 'notfound' or priority > prev_priority:
                    return 1   # we need to actually call start_readahead on this piece dict
                return 0   # start_readahead called recently on this piece dict
            piece_list = filter(need_to_call_readahead, piece_list)
        self.broker.doq.add_task(self._start_readahead, kwargs = {'piece_list': piece_list, 'priority': priority, 'num_blocks_needed': num_blocks_needed, 'total_block_shares': total_block_shares, 'download_monitor': download_monitor})

    def _start_readahead(self, piece_list, priority = None, num_blocks_needed=None, total_block_shares=None, download_monitor=None):
        """
        piecelist is a [{'pieceid': hash of reassembled thing, 'length': length in bytes, 'shares': {share number: block id}}]
        TODO unsplit has num_blocks_needed and num_blocks_total from the dinode & sharemap; it needs to pass them in.

        @precondition This method must be called from the DoQ.: self.broker.doq.is_currently_doq()
        """
        assert self.broker.doq.is_currently_doq(), "precondition: This method must be called from the DoQ."

        for i in piece_list:
            debug.mojolog.write("piece info for reassembly is %s\n", args=(i,), v=4, vs="FindBeforehandBlockFetcher")
            piece_id = i['pieceid']
            piece_length = i['length']

            # mark that we recently called start_readahead on this piece
            self._start_readahead_piece_id_cache[piece_id] = priority

            if not self.reassemblies.has_key(piece_id) and not self.finished_reassemblies.has_key(piece_id):
                block_positions = [None] * total_block_shares
                for index, block_id in i['shares'].items():
                    block_positions[index] = block_id
                debug.mojolog.write('putting key %s in reassemblies\n', args=(piece_id,), v=4, vs="FindBeforehandBlockFetcher")
                self.reassemblies[piece_id] = SingleReassembly(
                    broker=self.broker,
                    do_function = self.throttler.do_download, 
                    find_function = self.finder.callback_when_found, finder = self.finder,
                    cancel_find_function = self.finder.cancel, download_done_function = self.throttler.download_done,
                    block_positions = block_positions, num_blocks_needed = num_blocks_needed, 
                    piece_id = piece_id, piece_length = piece_length, max_concurrent_blobs=self.max_concurrent_request_blobs,
                    callback = self._reassembly_completed_callback, download_monitor = download_monitor)
            if priority is not None:
                if self.reassemblies.has_key(piece_id):
                    self.reassemblies[piece_id].set_priority(priority)
                elif self.finished_reassemblies.has_key(piece_id):
                    self.finished_reassemblies[piece_id].set_priority(priority)

    def _reassembly_completed_callback(self, reassembly):
        piece_id = reassembly.get_piece_id()
        if self.reassemblies.has_key(piece_id) and self.reassemblies[piece_id].get_piece_id() == piece_id:
            self.finished_reassemblies[piece_id] = reassembly
            del self.reassemblies[piece_id]

    def _find_function(self, server_id, server_info, block_ids, callback, preferenceno):
        """
        called by BlockFinder to do actual message generation and spewing over mtm
        """
        assert len(block_ids) > 0, "_find_function called without any block_ids to find!"
        # TODO check if its safe for us to pass allow_post_timeout_callback in here
        self.broker.do_you_have_blobs((server_id, server_info), block_ids, callback, allow_recursion=1, mojooffer=BidStrategy.choose_bid("MO_DO_YOU_HAVE_BLOBS_DOWNLOADING", self.broker.ServicesMTM, counterparty_id=server_id, preferenceno=preferenceno))

    def free(self, piecedict):
        """
        Removes a completed piece from our active reassembly object tables and into the
        successful reassembly (downloaded piece) cache if appropriate.

        If the reassembly is not finished, it is cancelled.
        """
        # debug.mojolog.write("free() called -- piecedict: %s, traceback: %s\n", args=(piecedict, traceback.extract_stack(),), v=1, vs="debug") # don't commit this when commented in -- traceback.extract_stack() can be slow.  --Zooko 2001-04-16
        piece_id = piecedict['pieceid']
        reassembly = self.reassemblies.get(piece_id)
        if reassembly is not None:
            try:
                del self.reassemblies[piece_id]
            except KeyError:
                # this could happen if free() is not called from DoQ thread, no harm done
                pass
            self.broker.doq.add_task(reassembly.cancel, kwargs={'reason': "cancelled"})
        # XXX ? I don't understand what this is supposed to do.  If you turn MojoConstants.DEBUG_MODE on you'll see that we create a second SingleReassembly for the same pieceid without ever deleting the original SingleReassembly.  Although I suppose that this could be because the garbage collector hasn't collected the original SingleReassembly yet.  Hm.  This happens when you try to download the same file that you already downloaded in this process.  What *should* we do in that case?  --Zooko 2001-05-06
        reassembly = self.finished_reassemblies.get(piece_id)
        if reassembly is None:
            return
        try:
            del self.finished_reassemblies[piece_id]
        except KeyError:
            # this could happen if free() is not called from DoQ thread, no harm done
            pass
        try:
            size = len(reassembly.nonblocking_reassemble())
            self.freed_successful_reassembly_cache.insert(piece_id, reassembly, itemsize=size)
        except ReassemblyError:
            # this SingleReassembly was unsuccessful, don't save it
            pass

