import RunTests
from threading import Lock

class HashQueue:
    def __init__(self):
        self.first_key = None
        self.last_key = None
        # maps keys to {'value': value, 'previous key': previous key, 'next key': next key}
        self.hash = {}
        self._lock = Lock()

    def keys(self):
        return self.hash.keys()

    def values(self):
        self._lock.acquire()
        try:
            return map(lambda i: i['value'], self.hash.values())
        finally:
            self._lock.release()

    def items(self):
        self._lock.acquire()
        try:
            return map(lambda i: (i[0], i[1]['value']), self.hash.items())
        finally:
            self._lock.release()

    def get_first_key(self):
        return self.first_key

    def has_key(self, key):
        assert key is not None
        return self.hash.has_key(key)

    def get(self, key, default = None):
        assert key is not None
        self._lock.acquire()
        try:
            val = self.hash.get(key)
            if val is None:
                return default
            return val['value']
        finally:
            self._lock.release()
    
    def after(self, key):
        assert key is not None
        return self.hash[key]['next key']

    def get_ordered_keys(self):
        self._lock.acquire()
        try:
            result = []
            key = self.get_first_key()
            while key is not None:
                result.append(key)
                key = self.after(key)
            return result
        finally:
            self._lock.release()

    def __len__(self):
        return len(self.hash)
    
    def __getitem__(self, key):
        assert key is not None
        val = self.hash[key]
        return val['value']

    def __setitem__(self, key, value):
        assert key is not None
        self._lock.acquire()
        try:
            if self.hash.has_key(key):
                self.hash[key]['value'] = value
            else:
                if self.first_key is None:
                    self.first_key = key
                    self.last_key = key
                    self.hash[key] = {'value': value, 'previous key': None, 'next key': None}
                else:
                    old_last = self.last_key
                    self.hash[old_last]['next key'] = key
                    self.last_key = key
                    self.hash[key] = {'value': value, 'previous key': old_last, 'next key': None}
        finally:
            self._lock.release()

    def __delitem__(self, key):
        assert key is not None
        self._lock.acquire()
        try:
            if not self.hash.has_key(key):
                raise KeyError, key
            previous = self.hash[key]['previous key']
            next = self.hash[key]['next key']
            if previous is None:
                self.first_key = next
            else:
                self.hash[previous]['next key'] = next
            if next is None:
                self.last_key = previous
            else:
                self.hash[next]['previous key'] = previous
            del self.hash[key]
        finally:
            self._lock.release()

def test_keys():
    h = HashQueue()
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    r = h.keys()
    r.sort()
    assert r == [0, 1, 2]

def test_values():
    h = HashQueue()
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    r = h.values()
    r.sort()
    assert r == ['a', 'b', 'c']

def test_items():
    h = HashQueue()
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    r = h.items()
    r.sort()
    assert r == [(0, 'c'), (1, 'b'), (2, 'a')]

def test_get_ordered_keys():
    h = HashQueue()
    h[2] = 'a'
    h[0] = 'b'
    h[1] = 'c'
    r = h.get_ordered_keys()
    assert r == [2, 0, 1], r

def test_delete_first():
    h = HashQueue()
    assert h.get_first_key() == None
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    assert h.get_first_key() == 2
    del h[2]
    assert h.get_first_key() == 1
    del h[1]
    assert h.get_first_key() == 0
    del h[0]
    assert h.get_first_key() == None

def test_delete_last():
    h = HashQueue()
    assert h.get_first_key() == None
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    assert h.get_first_key() == 2
    del h[0]
    assert h.get_first_key() == 2
    del h[2]
    assert h.get_first_key() == 1
    del h[1]
    assert h.get_first_key() == None

def test_delete_middle():
    h = HashQueue()
    assert h.get_first_key() == None
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    assert h.get_first_key() == 2
    del h[1]
    assert h.get_first_key() == 2
    del h[2]
    assert h.get_first_key() == 0
    del h[0]
    assert h.get_first_key() == None

def test_assignment():
    h = HashQueue()
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    assert h[0] == 'c'
    assert h.get(0) == 'c'
    assert h[1] == 'b'
    assert h.get(1) == 'b'
    assert h[2] == 'a'
    assert h.get(2) == 'a'

def test_get_missing():
    h = HashQueue()
    assert h.get(0) == None
    assert h.get(0, 'a') == 'a'

def test_overwrite_existing():
    h = HashQueue()
    h[2] = 'a'
    h[1] = 'b'
    h[0] = 'c'
    h[1] = 'd'
    assert h[2] == 'a'
    assert h[1] == 'd'
    assert h[0] == 'c'
    assert h.get_first_key() == 2
    del h[2]
    assert h.get_first_key() == 1
    del h[1]
    assert h.get_first_key() == 0
    del h[0]
    assert h.get_first_key() == None

def test_deplete_and_add():
    h = HashQueue()
    assert not h.has_key(0)
    h[0] = 'a'
    assert h.has_key(0)
    del h[0]
    assert not h.has_key(0)
    assert h.get(0) is None
    h[0] = 'b'
    assert h.get(0) == 'b'
    assert h.get_first_key() == 0
    h[1] = 'c'
    del h[0]
    assert h.get_first_key() == 1

def test_KeyError_on_del():
    h = HashQueue()
    try:
        h[0]
        assert false
    except KeyError:
        pass

def test_KeyError_on_lookup():
    h = HashQueue()
    try:
        del h[0]
        assert false
    except KeyError:
        pass

def test_after():
    h = HashQueue()
    h[0] = 'a'
    h[1] = 'b'
    assert h.after(0) == 1
    assert h.after(1) is None

def test_deplete_and_assign():
    h = HashQueue()
    h[0] = 'a'
    del h[0]
    h[1] = 'b'

def test_deplete_and_assign_same():
    h = HashQueue()
    h[0] = 'a'
    del h[0]
    h[0] = 'b'

mojo_test_flag = 1

if __name__ == '__main__':
    RunTests.runTests('HashQueue')

