#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

from HashQueue import HashQueue
import RunTests
from threading import Lock

false = None
true = 1

class LastAccessedCache:
    def __init__(self, max_size):
        self.max_size = max_size
        self.dict = HashQueue()
        self._lock = Lock()

    def get(self, key, default = None):
        self._lock.acquire()
        try:
            if not self.dict.has_key(key):
                return default
            val = self.dict[key]
            del self.dict[key]
            self.dict[key] = val
            return val
        finally:
            self._lock.release()

    def __getitem__(self, key):
        self._lock.acquire()
        try:
            val = self.dict[key]
            del self.dict[key]
            self.dict[key] = val
            return val
        finally:
            self._lock.release()

    def has_key(self, key):
        self._lock.acquire()
        try:
            if not self.dict.has_key(key):
                return false
            val = self.dict[key]
            del self.dict[key]
            self.dict[key] = val
            return 1
        finally:
            self._lock.release()

    def __delitem__(self, key):
        self._lock.acquire()
        try:
            del self.dict[key]
        finally:
            self._lock.release()

    def __setitem__(self, key, value):
        self._lock.acquire()
        try:
            if self.dict.has_key(key):
                del self.dict[key]
            self.dict[key] = value
            while len(self.dict) > self.max_size:
                del self.dict[self.dict.get_first_key()]
        finally:
            self._lock.release()

    def keys(self):
        return self.dict.keys()

    def values(self):
        return self.dict.values()

    def items(self):
        return self.dict.items()

    def __len__(self):
        return len(self.dict)

def test_dropping():
    c = LastAccessedCache(2)
    c['a'] = 0
    c['b'] = 1
    c['c'] = 2
    ks = c.keys()
    ks.sort()
    assert ks == ['b', 'c']

def test_access_changes_order():
    c = LastAccessedCache(2)
    c['a'] = 0
    c['b'] = 1
    c['a']
    c['c'] = 2
    ks = c.keys()
    ks.sort()
    assert ks == ['a', 'c']

def test_get_changes_order():
    c = LastAccessedCache(2)
    c['a'] = 0
    c['b'] = 1
    c.get('a')
    c['c'] = 2
    ks = c.keys()
    ks.sort()
    assert ks == ['a', 'c']

def test_has_key_changes_order():
    c = LastAccessedCache(2)
    c['a'] = 0
    c['b'] = 1
    c.has_key('a')
    c['c'] = 2
    ks = c.keys()
    ks.sort()
    assert ks == ['a', 'c']
    
def test_overwrite_changes_order():
    c = LastAccessedCache(2)
    c['a'] = 0
    c['b'] = 1
    c['a'] = 3
    c['c'] = 2
    ks = c.keys()
    ks.sort()
    assert ks == ['a', 'c']

def test_del():
    c = LastAccessedCache(2)
    c['a'] = 0
    del c['a']
    assert c.keys() == []

def test_items():
    c = LastAccessedCache(2)
    c['a'] = 0
    assert c.items() == [('a', 0)]

def test_values():
    c = LastAccessedCache(2)
    c['a'] = 0
    assert c.values() == [0]

def test_len():
    c = LastAccessedCache(2)
    assert len(c) == 0
    c['a'] = 0
    assert len(c) == 1

mojo_test_flag = 1

if __name__ == '__main__':
    RunTests.runTests('LastAccessedCache')
