#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import traceback
import types

### our modules
import mojoutil
import RunTests
import debug
from human_readable import hr
import idlib

DISQUALIFIED = ()

def getnameofhandicapper(thingie):
    if hasattr(thingie, '__name__'):
        if thingie.__name__ != "compute_handicap":
            return thingie.__name__

    if hasattr(thingie, 'im_class'):
        return getnameofhandicapper(thingie.im_class)

    if hasattr(thingie, '__class__'):
        return getnameofhandicapper(thingie.__class__)

    return hr(thingie)

class MojoHandicapper :
    """
    A very simple business logic utility - for each message to be sent out, compute the 
    sum of all estimated real costs of it, and sort based on those.
    
    Available methods are add_handicapper and sort_by_preference
    """
    def __init__(self, doq) :
        self.__handicappers_map = {None: []}   # maps from message type to a list of handicappers (None = all messages)
        self.doq = doq
        
    def add_handicapper(self, handicapper, message_types=None) :
        """
        Adds a handicap computing function to the list of all of them.

        A handicap function must take the parameters counterparty_id, message_type, and
        message_body and return a numerical value, DISQUALIFIED or a dict containing:
        'multiplier': a float to multiply the inprogress handicap computation by 'additive': a
        float to add to the handicap (same as returning a float)

        @param message_types if specified is a list of message types this handicapper
        should be applied to (None indicates all message types).
        [message type specific handicappers will always be evaulated before generic handicappers]

        Sum of squares addition is used: All additives are squared before being used.

        Numerical values will be added to produce a total handicap.  Multipliers will be applied
        after all additions have taken place.  DISQUALIFIED will completely disqualify the given
        counterparty.

        Examples of handicappers are one which returns the advertised price, one 
        which returns the cost of the relay server, and one which returns DISQUALIFIED for 
        any counterparty which has cheated you in the past, and 0 otherwise.

        A sophisticated reputation system could be plugged in here -- if you had information
        from friends of friends indicating that a given counterparty is unreliable, you could
        make a handicapper which returns your friend's friend's opinion, and it will be factored
        in along with the other considerations.

        @precondition `handicapper' must be a callable.: callable(handicapper): "handicapper: %s :: %s" % (hr(handicapper), `type(handicapper)`)
        """
        assert callable(handicapper), "`handicapper' must be a callable." + " -- " + "handicapper: %s :: %s" % (hr(handicapper), `type(handicapper)`)
        assert message_types is None or type(message_types) in (type([]), type(())), "message_types must be None or a sequence"

        if message_types is None:
            self.__handicappers_map[None].append(handicapper)
        else:
            for msgtype in message_types:
                mojoutil.setdefault(self.__handicappers_map, msgtype, []).append(handicapper)
        
    def _compute_handicap(self, counterparty_id, metainfo, message_type, message_body):
        """
        Computes the 'handicap', or estimated real cost including risk of failure, 
        of sending the message to the counterparty.
        """
        counterparty_id = idlib.canonicalize(counterparty_id, "broker")  # guarantee its in good form for the handicapper functions
        handicap = 0
        _hmap = self.__handicappers_map
        for h in (_hmap.get(message_type, []) + _hmap[None]) :
            try:
                amount = h(counterparty_id = counterparty_id, metainfo = metainfo, message_type = message_type, message_body = message_body)
            except:
                debug.stderr.write("WARNING: counterparty %s disqualified due to exception:\n Handicapper Params (message_type=%s, message_body=%s, metainfo=%s)\n:", args=(counterparty_id, message_type, message_body, metainfo), v=2, vs="business logic")
                traceback.print_exc(file=debug.stderr)
                return DISQUALIFIED

            assert (type(amount) is types.IntType) or (type(amount) is types.LongType) or (type(amount) is types.FloatType) or (type(amount) is type(DISQUALIFIED)) or (type(amount) is types.DictType), "Handicapper function must return certain type. -- func: %s, result: %s, counterparty_id: %s, metainfo: %s\n" % (hr(h), hr(amount), hr(counterparty_id), hr(metainfo)) # postcondition

            if amount == DISQUALIFIED :
                debug.mojolog.write("Handicap %s DISQUALIFIED for %s on %s\n", args=(getnameofhandicapper(h), counterparty_id, message_type), v=5, vs="business logic")
                return DISQUALIFIED
            elif type(amount) == type({}) :
                if amount.has_key('multiplier') :
                    assert amount['multiplier'] >= 0, "amount %s" % amount
                    debug.mojolog.write("Handicap %s multiplier %s for %s on %s\n", args=(getnameofhandicapper(h), "%0.2f" % amount['multiplier'], counterparty_id, message_type), v=8, vs="business logic")
                    handicap = handicap * (amount['multiplier'])
                if amount.has_key('additive') :
                    assert amount['additive'] >= 0, "amount %s" % amount
                    debug.mojolog.write("Handicap %s additive %s for %s on %s\n", args=(getnameofhandicapper(h), "%0.2f" % amount['additive'], counterparty_id, message_type), v=8, vs="business logic")
                    handicap = handicap + (float(amount['additive'])**2)
            else:
                assert amount >= 0
                debug.mojolog.write("Handicap %s additive %s for %s on %s\n", args=(getnameofhandicapper(h), "%0.2f" % amount, counterparty_id, message_type), v=8, vs="business logic")
                handicap = handicap + (float(amount)**2)
        debug.mojolog.write("Handicap %s for %s on %s\n", args=("%0.0f" % handicap, counterparty_id, message_type), v=8, vs="business logic")
        return handicap

    def sort_by_preference(self, counterparties, message_type, message_body):
        """
        @param counterparties a sequence of (counterparty_id, service_info_dict);  It is okay to be an empty list.

        Cogitates on the possibility of sending each counterparty in counterparties 
        the given message and returns a list sorted by order of preference of which 
        counterparties to try to use. Disqualified counterparties are not included 
        in the list at all.

        @precondition `counterparties' must be a list of (id, infodict) tuples.: len(filter(lambda x: (len(x) == 2) and idlib.is_sloppy_id(x[0]), counterparties)) == len(counterparties)
        @precondition `message_type' must be a string.: type(message_type) == types.StringType
        """
        assert len(filter(lambda x: (len(x) == 2) and idlib.is_sloppy_id(x[0]), counterparties)) == len(counterparties), "precondition: `counterparties' must be a list of (id, infodict) tuples."
        assert type(message_type) == types.StringType, "precondition: `message_type' must be a string."

        # contains (counterparty_id, cost)
        unsorted = []
        for i in counterparties :
            cost = self._compute_handicap(i[0], i[1], message_type=message_type, message_body=message_body)
            if cost is not DISQUALIFIED :
                unsorted.append((i, cost))
        unsorted.sort(lambda a, b : cmp(a[1], b[1]))
        result = []
        for i in unsorted :
            result.append(i[0])
        return result

mojo_test_flag = 1

def test_disqualified_not_returned() :
    id_a = idlib.string_to_id('a')
    id_b = idlib.string_to_id('b')
    h = MojoHandicapper()
    def cfunc(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        if counterparty_id == id_a :
            return DISQUALIFIED
        else :
            return 0
    h.add_handicapper(cfunc)
    result = h.sort_by_preference(counterparty_ids = [(id_a, None), (id_b, None)], message_type = 'a message', message_body = {})
    assert len(result) == 1
    assert result[0][0] == id_b

def test_returns_nothing_for_nothing() :
    assert len(MojoHandicapper(None).sort_by_preference([], 'm', {})) == 0

def test_asserts_for_negative_value() :
    id_a = idlib.string_to_id('a')
    h = MojoHandicapper()
    def cfunc(counterparty_id, metainfo, message_type, message_body) :
        return -2
    h.add_handicapper(cfunc)
    try :
        h.sort_by_preference(counterparty_ids = [(id_a, None)], message_type = 'a message', message_body = {})
        return 0
    except AssertionError :
        return 1
    
def test_normal_operation() :
    id_a = idlib.string_to_id('a')
    id_b = idlib.string_to_id('b')
    h = MojoHandicapper()
    def cfunc(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        if counterparty_id == id_a :
            return 1
        else :
            return 0
    h.add_handicapper(cfunc)
    result = h.sort_by_preference(counterparty_ids = [(id_a, None), (id_b, None)], message_type = 'a message', message_body = {})
    assert len(result) == 2
    assert result[0][0] == id_b
    assert result[1][0] == id_a
    
def test_normal_operation_opposite_order() :
    id_a = idlib.string_to_id('a')
    id_b = idlib.string_to_id('b')
    h = MojoHandicapper()
    def cfunc(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        if counterparty_id == id_a :
            return 1
        else :
            return 0
    h.add_handicapper(cfunc)
    result = h.sort_by_preference(counterparty_ids = [(id_b, None), (id_a, None)], message_type = 'a message', message_body = {})
    assert len(result) == 2
    assert result[0][0] == id_b
    assert result[1][0] == id_a
    
def test_equal_handicap_returns_something() :
    id_a = idlib.string_to_id('a')
    id_b = idlib.string_to_id('b')
    h = MojoHandicapper()
    def cfunc(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        return 1
    h.add_handicapper(cfunc)
    result = h.sort_by_preference(counterparty_ids = [(id_a, None), (id_b, None)], message_type = 'a message', message_body = {})
    assert len(result) == 2
    assert result[0][0] != result[1]
    
def test_uses_first_handicapper_stored() :
    id_a = idlib.string_to_id('a')
    id_b = idlib.string_to_id('b')
    h = MojoHandicapper()
    def cfunc(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        if counterparty_id == id_a :
            return 1
        else :
            return 0
    h.add_handicapper(cfunc)
    def cfunc2(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        if counterparty_id == id_a :
            return 0
        else :
            return 2
    h.add_handicapper(cfunc2)
    result = h.sort_by_preference(counterparty_ids = [(id_b, None), (id_a, None)], message_type = 'a message', message_body = {})
    assert len(result) == 2
    assert result[0][0] == id_a
    assert result[1][0] == id_b
    
def test_uses_second_handicapper_stored() :
    id_a = idlib.string_to_id('a')
    id_b = idlib.string_to_id('b')
    h = MojoHandicapper()
    def cfunc(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        if counterparty_id == id_a :
            return 0
        else :
            return 2
    h.add_handicapper(cfunc)
    def cfunc2(counterparty_id, metainfo, message_type, message_body, id_a = id_a, id_b = id_b) :
        if counterparty_id == id_a :
            return 1
        else :
            return 0
    h.add_handicapper(cfunc2)
    result = h.sort_by_preference(counterparty_ids = [(id_b, None), (id_a, None)], message_type = 'a message', message_body = {})
    assert len(result) == 2
    assert result[0][0] == id_a
    assert result[1][0] == id_b

if __name__ == '__main__':
    RunTests.runTests("MojoHandicapper")
