### Module imports:
# Standard Modules:
import os
import sys
import traceback

# Mojo Nation Modules:
from MojoConstants import true, false

### Constants:
# The "OLD_SCHOOL_*" stuff is for backwards compatibility with previous MojoMod APIs:
OLD_SCHOOL_INIT_FUNC_NAMES = ('initialize_dynamic_broker_mojomod', 'initialize_dynamic_broker_extension')
OLD_SCHOOL_UN_INIT_FUNC_NAMES = ('uninitialize_dynamic_broker_mojomod',)

### MojoMod API:
class MojoModHook:
    """
    Handles MojoMod hooks in an OO fashion.
    The Broker uses sub-classes of MojoModHook
    as interfaces to MojoMods.

    If you need to control module load order, add a member
    to the python module called 'MojoModLoadOrder' that
    contains an integer.  The default value is 1000, lower
    numbers will be loaded first and unloaded last.

    To write a MojoMod, write a normal python
    module which includes a subclass of MojoModHook.
    The subclass must also be named "MojoModHook".

    This allows the MojoMod author to write more than
    one subclass of MojoModHook and designate which
    class is to be instantiated, ie:

    import MojoMod
    
    class NormalMojoModHook (MojoMod.MojoModHook):
        ...

    class DebugMojoModHook (NormalMojoModHook):
        ...
    
    MojoModHook = NormalMojoModHook # Change this to "DebugMojoModHook" to turn on debugging.
    """
    def __init__(self, broker):
        self.broker = broker

    def close(self):
        """
        MojoMods are responsible for stopping any threads they start, closing databases, etc... in this method.
        """
        self.log('Closing.\n')

    def log(self, output, args=(), vs=None, v=2):
        if vs is None:
            vs = self.__module__
        self.broker.mojolog.write(output, args=args, vs=vs, v=v)

    def set_autoload(self, autoload=true):
        self.broker.modman.set_autoload(__name__, autoload)

### Broker API:
class MojoModManager:
    """
    The Broker uses an instance of this class to manage MojoMods.

    Terminology:
    module - A typical python module.  In the code this refers to an actual module object, not a string name or path.
    MojoMod - A dynamically-loadable extension to the Mojo Nation Broker; currently these must be implemented
        as python modules which subclass the "MojoModHook" class found in this module.
    imported - When referring to a MojoMod, this means the python module containing this MojoMod has been imported.
    running - When referring to a MojoMod, this means a single instance of the MojoModHook subclass found in the
        MojoMod's python module has been created.  If a MojoMod is not "running" no instance of this class should exist.
        A MojoMod which is running necessarily is imported (see above).
    """
    def __init__(self, broker):
        self.broker = broker

        # This hack ensures we can import the python modules in the MojoMod directory:
        if self.broker.confman.getpath(('PATH', 'MOJOMOD_DIR')) not in sys.path:
            sys.path.append(self.broker.confman.getpath(('PATH', 'MOJOMOD_DIR')))

        # self.mojomods maps extension mods names to a dictionary with these keys:
        # 'autoload': a boolean of whether or not to load this module at broker start-up.
        # 'modref': None if the module is not imported, else the imported module instance.
        # 'modhook': None if the module is not "running", else an instance of the module's hook class "MojoModHook".
        # 'loadorder': An integer used to determine the order modules load/unload.  Lower numbers loads first and unloads last.  Default: 1000
        self.mojomods = {}

        self.scan()

        # NOTE: the broker must call autoload() from the DoQ once it has
        # initialized itself to a usable state.

    def _create_MojoMod(self, module):
        """Given an imported module, create a MojoModHook interface to it."""
        if hasattr(module, 'MojoModHook') and issubclass(module.MojoModHook, MojoModHook):
            return module.MojoModHook(self.broker)
        else:
            return OldSchoolMojoModWrapper(self.broker, module)

    def _log(self, output, args=(), vs='MojoMod', v=2):
        self.broker.mojolog.write(output, args=args, vs=vs, v=v)

    def shutdown(self):
        for mod, moddata in self._sorted_modtuple_list(reverse=1):
            self.remove(mod)
            
    def get_known_mods(self):
        return self.mojomods.keys()

    def get_running_mods(self):
        running = []
        for name, data in self.mojomods.items():
            if data['modhook'] is not None:
                running.append(name)
        return running
    
    def add(self, mod, autoload=false):
        """
        Parameters:
        mod - A string which is the MojoMod python module name.
        autoload - Whether or not to start this module at initialization.

        If the module is already added, the builtin "reload" function is called on it.
        """
        self._log('Adding MojoMod %s to manager.\n', args=(mod,), v=3)
        try:
            if self.mojomods.has_key(mod):
                reload(self.mojomods[mod]['modref'])
            else:
                modref = __import__(mod)
                if hasattr(modref, 'MojoModLoadOrder'):
                    loadorder = modref.MojoModLoadOrder
                else:
                    loadorder = 1000
                self.mojomods[mod] = {
                        'autoload': autoload, 'modref': modref, 'modhook': None,
                        'loadorder': 1000,
                    }
        except:
            traceback.print_exc(file=self.broker.mojolog)

    def remove(self, mod):
        self._log('Removing MojoMod %s from manager.\n', args=(mod,), v=3)
        if self.mojomods[mod]['modhook'] is not None:
            self.stop(mod)
        del self.mojomods[mod]
        
    def start(self, mod):
        """
        Initializes a MojoMod and puts it in a "running" state.
        """
        self._log('Starting MojoMod %s.\n', args=(mod,))
        if self.mojomods[mod]['modhook'] is None:
            try:
                self.mojomods[mod]['modhook'] = self._create_MojoMod(self.mojomods[mod]['modref'])
            except:
                self._log('Could not load mojomod %s:\n', args=(mod,), v=1)
                traceback.print_exc(file=self.broker.mojolog)
        else:
            self._log('An instance of %s is already running.\n', args=(mod,), v=3)

    def stop(self, mod):
        self._log('Stopping MojoMod %s.\n', args=(mod,))
        if self.mojomods[mod]['modhook'] is None:
            self._log('There is no instance of mod %s running.\n', args=(mod,), v=3)
            return
        self.mojomods[mod]['modhook'].close()
        self.mojomods[mod]['modhook'] = None

    def reload(self, mod):
        """
        Stops, adds, and starts a MojoMod.
        If the MojoMod is already stopped or removed, it silently skips these steps.

        Note:  This has the wonderful but possibly confusing feature of incorporating
        new changes in the MojoMod source at runtime!  In other words, a developer
        can change the source to a MojoMod and then call this method to see the
        new changes without (in many cases) restarting the Broker!
        """
        if self.mojomods.has_key(mod) and self.mojomods[mod]['modhook'] is not None:
                self.stop(mod)
        self.add(mod) # Reloads the python module if it is already present.
        self.start(mod)
        
    def set_autoload(self, mod, autoload=true):
        self.mojomods[mod]['autoload'] = autoload
        modlist = self.broker.confman.getlist(['EXTENSION_MODS'])
        if autoload and (mod not in modlist):
            modlist.append(mod)
        elif (not autoload) and (mod in modlist):
            modlist.remove(mod)
        self.broker.confman.setlist(['EXTENSION_MODS'], modlist)
        
    def scan(self):
        """
        This searches self.broker.confman['PATH']['MOJOMOD_DIR'] and EVILDIR/mnmods for things that look like python modules,
        and assumes they are Broker mojomods (aka MojoMods).

        It also uses the union of those search results and the self.broker.confman['EXTENSION_MODS'] list to
        compile a list of potential mojomods which it stores in self.mojomods.
        """
        self._log("Scanning for MojoMods.\n")

        def endswith(s, e):  # python 2.0 strings have this as a method
            return s[-len(e):] == e

        # Add MojoMods from MojoDir:
        moddirs = []
        moddirs.append( self.broker.confman.getpath(('PATH', 'MOJOMOD_DIR')) )
        if self.broker.confman.is_true_bool(['SCAN_EVILDIR_FOR_MODS']):
            moddirs.append( os.path.join(os.environ['EVILDIR'], 'mnmods') )
        for moddir in moddirs:
            for name in os.listdir(moddir):
                fullpath = os.path.join(moddir, name)
                if os.path.isfile(fullpath) and endswith(name, '.py'):
                    # It's a normal python module:
                    self.add(name[:-3], autoload=false)
                elif os.path.isdir(fullpath) and os.path.exists(os.path.join(fullpath, '__init__.py')):
                    # We think its a python package:
                    self.add(name, autoload=false)
                else:
                    # TODO add support for precompiled-only or binary modules in here:
                    # (*.so on unix, *.pyd on windows, *.pyc and *.pyo for bytecode depending on __debug__)
                    self._log('Unknown item in the MojoMod directory: %s\n', args=(name,), v=3)

        # Add MojoMods from the config file:
        for mod in self.broker.confman.getlist(['EXTENSION_MODS']):
            if self.mojomods.has_key(mod):
                # we already found the mod and it was enabled in the config file
                self.start(mod)
            else:
                # we didn't find the mod but the config file listed it, try loading it anyways
                self.add(mod, autoload=true)

    def autoload(self):
        """
        This loads all of those which have "autoload" set to true
        (which includes any specified in the config file).
        """
        for name, moddata in self._sorted_modtuple_list():
            if moddata['autoload']:
                # do not put self.start calls on the DoQ, for some reason it
                # causes them to be loaded out of order (how?!?) -greg
                self.start(name)
    
    def _sorted_modtuple_list(self, reverse=0):
        """return a sorted list of (name, moddata) tuples"""
        l = self.mojomods.items()
        l.sort(lambda a,b: cmp(a[1]['loadorder'], b[1]['loadorder']))
        if reverse:
            l.reverse()
        return l


class OldSchoolMojoModWrapper (MojoModHook):
    """This encapsulates MojoMod's written with older APIs."""
    def __init__(self, broker, module):
        """
        Parameters:
        broker - A Broker instance.
        module - An imported module (which should not have a "MojoModHook" attribute).
        """
        MojoModHook.__init__(self, broker)
        self.module = module
        self.log('Creating MojoMod API wrapper for module %s.\n', args=(self.module.__name__,), vs='MojoMod', v=1)

        for old_api_hook in OLD_SCHOOL_INIT_FUNC_NAMES :
            initfunc = getattr(self.module, old_api_hook, None)
            if initfunc is not None:
                self.log('Using old school MojoMod initialization hook %s.\n', args=(old_api_hook,), vs='MojoMod', v=1)
                initfunc(self.broker)
                break

    def __del__(self):
        for old_api_hook in OLD_SCHOOL_UN_INIT_FUNC_NAMES:
            uninitfunc = getattr(self.module, old_api_hook, None)
            if uninitfunc is not None:
                self.log('Using old school MojoMod uninitialization hook %s.\n', args=(old_api_hook,), vs='MojoMod', v=1)
                uninitfunc()
                break
