#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# This file contains our derived classes necessary for implementing
# Mojo Nation HTTP servers properly.
#
# $Id: MojoNationHTTPServer.py,v 1.8 2001/05/01 08:57:21 greg Exp $

# standard modules
import BaseHTTPServer
import SocketServer
import socket
import string
import sys
import cgi
import re
import traceback

# our modules
from loggedthreading import LoggedThread
import debug


Version = "0.3"


class LoggingErrorHandler:
    def handle_error(self, request, client_address):
        debug.stderr.write("Exception during request from %s\n", args=(client_address,), vs='httpd')
        traceback.print_exc(file=debug.stderr)

########################################################################
# (derived from code in SocketServer.py)
# SocketServer.py's Threading* classes use the thread module rather
# than the threading module.  The threading module prefers threads
# created using it when doing locking, events, etc.
#
# TODO limit the maximum number of currently spawned off threads, etc.
########################################################################
class ThreadingMixIn:
    """Mix-in class to handle each request in a new LoggedThread() class"""
    def process_request(self, request, client_address):
        """Start a new thread to process the request."""
        t = LoggedThread(target=self.finish_request, args=(request, client_address))
        t.start()
    
    if sys.version >= "2.1":
        # XXX work around an annoying -BUG- in python 2.1:
        # http://sourceforge.net/tracker/index.php?func=detail&aid=417845&group_id=5470&atid=105470
        def close_request(self, request):
            pass

class ThreadingUDPServer(ThreadingMixIn, SocketServer.UDPServer): pass
class ThreadingTCPServer(ThreadingMixIn, SocketServer.TCPServer): pass


########################################################################
# A HTTP server that spawns a new thread for each incoming
# request (in the process_request() call)
########################################################################
class ThreadingHTTPServer(LoggingErrorHandler, ThreadingTCPServer,
                          BaseHTTPServer.HTTPServer):
    request_queue_size = 15    # the size of the listen queue for this server (used in SocketServer.TCPServer)
    pass


########################################################################
# This class simply adds the SO_REUSEADDR option to the socket so that
# We can restart a test server shortly after it exits without it
# whining about "address already in use" due to the old socket.
########################################################################
class ReuseHTTPServer(LoggingErrorHandler, BaseHTTPServer.HTTPServer) :
    def server_bind(self) :
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        BaseHTTPServer.HTTPServer.server_bind(self)


########################################################################
# A ReuseHTTPServer that spawns a new thread for each incoming request
# (in the process_request() call from SocketServer)
########################################################################
class ReuseThreadingHTTPServer(LoggingErrorHandler, ThreadingTCPServer,
                               BaseHTTPServer.HTTPServer) :
    def __init__(self, broker, *_args, **_kwargs):
        """
        broker is a Broker instance.
        """
        apply(BaseHTTPServer.HTTPServer.__init__, (self,) + _args, _kwargs)
        self.broker = broker
    
    def server_bind(self) :
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        BaseHTTPServer.HTTPServer.server_bind(self)

########################################################################
# This request handler implements query string parsing and URL
# rewriting.  The descendant classes need to implement them by calling
# the following methods in this order from their do_GET and other
# request handlers:
#    rewrite_path()
#    parse_query_string()
#
# Rewrite rules are specified in the rewrite_rules list of tuples.
# Each tuple should contain a compiled regular expression followed by
# its substitution string or function (the second argument to re.subn).
#
# Query string contents will be placed in the form field as simple
# key/value pairs.
########################################################################
class RewriteHTTPRequestHandler(LoggingErrorHandler, BaseHTTPServer.BaseHTTPRequestHandler):
    server_version = 'MojoNation'+Version

    form = {}  # form data from the query string

    # a list of tuples containing a compiled regular expression and a
    # substitution string or function to generate a new path.
    rewrite_rules = []


    #####################################################################
    # Run the rewrite rules on this requests URL
    ####################################################################
    def rewrite_path(self):
        for rule in self.rewrite_rules :
            (self.path, num) = rule[0].subn(rule[1], self.path)
            if num > 0 :
                return


    #####################################################################
    # Parse the query string and fill the data into the form dictionary
    #####################################################################
    def parse_query_string(self):
        parts = string.split(self.path, '?', 1)
        if parts and len(parts) == 2 :
            (self.path, qs) = parts
            self.form = cgi.parse_qs(qs, 1)

