#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

# standard modules
import math
import sha
import exceptions

# our modules
from DataTypes import UNIQUE_ID, MOD_VAL
import HashRandom
import MojoConstants
from MojoConstants import true, false
import MojoKey
import OurMessagesTokens
import SEXP
import canon
import cryptutil
import debug
from human_readable import hr
import idlib
import mencode
import modval
import mojosixbit
import mojoutil
import randsource
import tokendists


class Error(exceptions.StandardError): pass
class BadSignatureError(Error): pass
class UntrustedPublicKeyError(BadSignatureError): pass


# Hard-coded TokenServer constants:
HARDCODED_TOKEN_SERVER_COMMS_ID=idlib.string_to_id('(4:dict(6:string10:key header)(4:dict(6:string12:cryptosystem)(6:string3:RSA)(6:string4:type)(6:string6:public)(6:string5:usage)(6:string31:only for communication security))(6:string10:key values)(4:dict(6:string15:public exponent)(6:string1:3)(6:string14:public modulus)(6:string171:uxDLNYPqDkuJetlgcU1zrD4M1I5gys5SDf2pRNRhYRrlM6D3A4v0Msv13pTU_aY0NJldr9frXuB7TTx5QWuFPey_CEiX03bRATua93WZG9TwgvpWpLWjlzph-L7X_HXImMbI9Fj-KaLGitJYrZh_hZgmPDZH3i1S1ZVyUJSVOPE)))')

PUB_KEY_OF_TOKEN_SIGNER_SEXP='(4:dict(6:string10:key header)(4:dict(6:string12:cryptosystem)(6:string3:RSA)(6:string4:type)(6:string6:public)(6:string5:usage)(6:string35:only for verifying token signatures)(6:string21:value of signed token)(4:dict(6:string6:amount)(6:string2:16)(6:string8:currency)(6:string16:Mojo Test Points)))(6:string10:key values)(4:dict(6:string15:public exponent)(6:string1:3)(6:string14:public modulus)(6:string171:uiI2e3V8-zoprTlZMoYcE_P1h5e2KbY5NmXpnIqWjOPkh-7OGfjF-6-DpT0aDskhbjT8kP5IZNozHf7oqEUpQJnm1wOxgwrXHPNnfJSl-8RahiN7ntMC85sX_FJRxWy5r1XqY3ohczEvS2VjfjMjspxrxH637jskf0SQgvByEOU)))'

DEMO_ID_OF_TOKEN_SIGNER=MojoKey.pubkey_to_id(PUB_KEY_OF_TOKEN_SIGNER_SEXP)


MAP_KEYID_TO_PUBLIC_KEY={DEMO_ID_OF_TOKEN_SIGNER: PUB_KEY_OF_TOKEN_SIGNER_SEXP }

# This maps from public key ids to denomination values (long integers).


TEST_DENOM=10000

MAP_KEYID_TO_DENOM = { DEMO_ID_OF_TOKEN_SIGNER : TEST_DENOM }

MAP_DENOM_TO_KEYID = { TEST_DENOM : DEMO_ID_OF_TOKEN_SIGNER }


def make_tokenId(seed, pubkeysexp):
    """
    @param seed the token id seed
    @param pubkeysexp the sexp string of the public key for verifying this token

    @precondition `seed' is of correct form.: idlib.is_sloppy_id(seed): "seed: %s" % hr(seed)
    @precondition `pubkeysexp' is well-formed RSA public key.: MojoKey.publicRSAKeyIsWellFormed(pubkeysexp): "pubkeysexp: %s" % hr(pubkeysexp)
    @precondition `pubkeysexp' is sane.: MojoKey.publicRSAKeyIsSane(pubkeysexp): "pubkeysexp: %s" % hr(pubkeysexp)
    """
    assert idlib.is_sloppy_id(seed), "`seed' is of correct form." + " -- " + "seed: %s" % hr(seed)
    assert MojoKey.publicRSAKeyIsWellFormed(pubkeysexp), "`pubkeysexp' is well-formed RSA public key." + " -- " + "pubkeysexp: %s" % hr(pubkeysexp)
    assert MojoKey.publicRSAKeyIsSane(pubkeysexp), "`pubkeysexp' is sane." + " -- " + "pubkeysexp: %s" % hr(pubkeysexp)

    # ids are generated by taking the seed and expanding it to be the size of the modval using a deterministic pseudo-random number generator.  Also we do an iterative try-and-check loop to ensure that the id is both really randomly distributed and also less than the public modulus.  It will be quicker for you to read the source code (which is pretty straightforward) than for me to try to explain in English.  --Zooko 2000-07-03

    pkid = idlib.make_id(pubkeysexp, 'token verifying public key')
    seed = idlib.canonicalize(seed, "token id seed")

    rseed = idlib.make_id(mencode.mencode({'input values for generating token id': {'token id seed': seed, 'id of public key for verifying this token': pkid}}), 'rseed')

    kmv = MojoKey.makeRSAPublicKeyMVFromSexpString(pubkeysexp)

    res = cryptutil.get_rand_lt_n(rseed, kmv)

    debug.mojolog.write("------------------------------------------ make_tokenId(): seed: %s, pubkeysexp: %s, res: %s\n", args=(seed, pubkeysexp, res), v=11, vs="payment")

    return res

def make_tokenId_OLD(seed, pubkeysexp):
    """
    @param seed the token id seed
    @param pubkeysexp the sexp string of the public key for verifying this token

    @precondition `seed' is of correct form.: idlib.is_sloppy_id(seed): "seed: %s" % hr(seed)
    @precondition `pubkeysexp' is well-formed RSA public key.: MojoKey.publicRSAKeyIsWellFormed(pubkeysexp): "pubkeysexp: %s" % hr(pubkeysexp)
    @precondition `pubkeysexp' is sane.: MojoKey.publicRSAKeyIsSane(pubkeysexp): "pubkeysexp: %s" % hr(pubkeysexp)

    @deprecated in favor of `make_tokenId()'; See note in `is_token_sig_valid()'.  --Zooko 2000-08-28
    """
    assert idlib.is_sloppy_id(seed), "`seed' is of correct form." + " -- " + "seed: %s" % hr(seed)
    assert MojoKey.publicRSAKeyIsWellFormed(pubkeysexp), "`pubkeysexp' is well-formed RSA public key." + " -- " + "pubkeysexp: %s" % hr(pubkeysexp)
    assert MojoKey.publicRSAKeyIsSane(pubkeysexp), "`pubkeysexp' is sane." + " -- " + "pubkeysexp: %s" % hr(pubkeysexp)

    # ids are generated by taking the seed and expanding it to be the size of the modval using a deterministic pseudo-random number generator.  Also we do an iterative try-and-check loop to ensure that the id is both really randomly distributed and also less than the public modulus.  It will be quicker for you to read the source code (which is pretty straightforward) than for me to try to explain in English.  --Zooko 2000-07-03

    pkid = idlib.string_to_id(pubkeysexp)

    rseed = sha.new(SEXP.dictToString({'input values for generating token id': {'token id seed': mojosixbit.b2a(seed), 'id of public key for verifying this token': mojosixbit.b2a(pkid)}})).digest()

    kmv = MojoKey.makeRSAPublicKeyMVFromSexpString(pubkeysexp)

    res = cryptutil.get_rand_lt_n_with_prepended_0(rseed, kmv)

    debug.mojolog.write("------------------------------------------ make_tokenId_OLD(): seed: %s, pubkeysexp: %s, res: %s\n", args=(seed, pubkeysexp, res), v=11, vs="payment")

    return res

def is_token_sig_valid(tokidsig, tokidseed, idofpubkey4ving=MAP_DENOM_TO_KEYID[TEST_DENOM]):
    """
    @param tokidsig signature on the token id; Preferred form is mojosixbitencoded, but the straight binary form is also accepted.
    @param tokidseed the token id seed; Preferred form is mojosixbitencoded, but the straight binary form is also accepted.
    @param idofpubkey4ving id of public key for verifying this token

    @returns `true' if and only if `tokidsig' actually represents a signature on the token
        generated from `tokidseed'

    @precondition `tokidseed' must be an id.: idlib.is_sloppy_id(tokidseed): "tokidseed: %s" % hr(tokidseed)
    """
    assert idlib.is_sloppy_id(tokidseed), "precondition: `tokidseed' must be an id." + " -- " + "tokidseed: %s" % hr(tokidseed)

    try:
        if len(tokidsig) == 128:
            bintokidsig = tokidsig
        else:
            bintokidsig = mojosixbit.a2b(tokidsig)
            if len(bintokidsig) != 128:
                raise mojosixbit.Error, "sigs must be 1024 bits long"
    except mojosixbit.Error:
        # Oh well, nevermind.  Hopefully it was already binary.
        bintokidsig = tokidsig

    assert len(bintokidsig) == 128, "sigs must be exactly 1024 bits long, bintokidsig: %s, tokidsig: %s, tokidseed: %s" % (hr(bintokidsig), hr(tokidsig), hr(tokidseed))

    bintokidseed = idlib.canonicalize(tokidseed, 'tokenidseed')

    # canonicalize the id (allowing for backwards compatibility):
    idofpubkey4ving = idlib.canonicalize(idofpubkey4ving, thingtype='pubkey')

    sigkeysexp = MAP_KEYID_TO_PUBLIC_KEY.get(idofpubkey4ving)
    if not sigkeysexp:
        return false

    keymv = MojoKey.makeRSAPublicKeyMVFromSexpString(sigkeysexp)
    debug.stderr.write("keymv.get_modulus(): %s, keymv.get_exponent(): %s\n", args=(keymv.get_modulus(), keymv.get_exponent()), v=11, vs="token")
    try:
        keymv.set_value_string(bintokidsig)
    except modval.Error, le:
        debug.mojolog.write("MojoToken.is_token_sig_valid(tokidsig=%s, tokidseed=%s) returning false due to mod val error: %s.\n", args=(tokidsig, tokidseed, le), v=1, vs="payment")
        return false

    keymv.undo_signature()
    recoveredtid = canon._canon(keymv.get_value(), len(keymv.get_modulus()))

    gennedid = make_tokenId(bintokidseed, sigkeysexp)

    if recoveredtid != gennedid:
        # XXXXXX  Now here is a special exception for backwards compatibility.  Brokers of
        # version <= 0.882 used a hash-expansion function that was a little faster but didn't
        # fully cover the set of "Zed-star-N".  For now we'll accept tokens of that format, but
        # as soon as few enough people are using the old format we'll remove this exception
        # here.  --Zooko 2000-08-28
        # Also, the older version used "SEXP" encoding format and the new one uses "mencode".  --Zooko 2000-09-03
        debug.mojolog.write("MojoToken.is_token_sig_valid(tokidsig=%s, tokidseed=%s) trying old token format, because recoveredtid: %s != gennedid: %s.\n", args=(tokidsig, tokidseed, recoveredtid, gennedid), v=1, vs="payment")

        gennedid = make_tokenId_OLD(bintokidseed, sigkeysexp)

    if recoveredtid != gennedid:
        debug.mojolog.write("MojoToken.is_token_sig_valid(tokidsig=%s, tokidseed=%s) returning false, because recoveredtid: %s != gennedid: %s.\n", args=(tokidsig, tokidseed, recoveredtid, gennedid), v=1, vs="payment")

    return recoveredtid == gennedid

def verify_token_sigs(toks):
    """
    @param toks a list of MojoDicts satisfying the PAYMENT_TOKEN_TEMPL template

    @throws BadSignatureError if any one of the tokens contained within `set' does not come with
        a valid signature
    @throws UntrustedPublicKeyError if any one of the tokens contained within `set' is not
        signed by a trusted public key
    """
    for tok in toks:
        if not is_token_sig_valid(tok['token id signature'], tok['token id seed'], tok['id of public key for verifying this token']):
            raise BadSignatureError

def msg_of_tokens_to_cdist(msg):
    cDist = {}
    
    for t in msg:
        tk = idlib.canonicalize(t['id of public key for verifying this token'], thingtype='pubkey')

        denom = MAP_KEYID_TO_DENOM.get(tk)
        if not denom:
            raise UntrustedPublicKeyError, tk
        mojoutil.add_long_to_dict(cDist, denom, 1)

    return cDist
   
def tally_tokens(toks):
    """
    @return the value that is allegedly contained in the message

    @throws UntrustedPublicKeyError if any one of the tokens contained within `set' is not
        signed by a trusted public key
    """
    aggval = 0
    for tok in toks:
        idopubkey = idlib.canonicalize(tok['id of public key for verifying this token'], 'tokensigner')
        den = MAP_KEYID_TO_DENOM.get(idopubkey)
        if not den:
            raise UntrustedPublicKeyError, idopubkey

        aggval = aggval + den

    assert tokendists.amount_of_token_dist(msg_of_tokens_to_cdist(toks)) == aggval, "Two different functions that tally tokens must yield the same answer. -- aggval: %s, tokendists.amount_of_token_dist(msg_of_tokens_to_cdist(toks)): %s" % (hr(aggval), hr(tokendists.amount_of_token_dist(msg_of_tokens_to_cdist(toks))))
    return aggval

class Token:
    def __init__(self, seed, serverPubKeySexpStr):
        """
        Make a new token with a given seed for a given server pubkey.

        @param seed a unique, unguessable number of 160-bit length
        @param serverPubKeySexpStr the public key (in an s-expression string with accompanying
            meta-data) which can be used to verify this token

        @precondition `seed' must be the right form.: idlib.is_sloppy_id(seed): "seed: %s" % hr(seed)
        @precondition `serverPubKeySexpStr' must be well-formed.: MojoKey.publicRSAKeyIsWellFormed(serverPubKeySexpStr): "serverPubKeySexpStr: %s" % hr(serverPubKeySexpStr)
        @precondition `serverPubKeySexpStr' must be sane.: MojoKey.publicRSAKeyIsSane(serverPubKeySexpStr): "serverPubKeySexpStr: %s" % hr(serverPubKeySexpStr)
        """
        assert idlib.is_sloppy_id(seed), "`seed' must be the right form." + " -- " + "seed: %s" % hr(seed)
        assert MojoKey.publicRSAKeyIsWellFormed(serverPubKeySexpStr), "`serverPubKeySexpStr' must be well-formed." + " -- " + "serverPubKeySexpStr: %s" % hr(serverPubKeySexpStr)
        assert MojoKey.publicRSAKeyIsSane(serverPubKeySexpStr), "`serverPubKeySexpStr' must be sane." + " -- " + "serverPubKeySexpStr: %s" % hr(serverPubKeySexpStr)

        self.__serverPubKeySexpStr = serverPubKeySexpStr

        self.__seed = seed

        self.__idForSigning = None

        self.__signature = None


    def get_server_public_key_id(self):
        """
        A universally unique number that can be used to identify this Token Server public key
        among all public keys.

        @postcondition Result is of correct form.: idlib.is_sloppy_id(result)
        """
        return MojoKey.pubkey_to_id(self._get_server_public_key_sexp())


    def _get_server_public_key_sexp(self):
        """
        `serverPubKeySexpStr' is the public key used by the token server for this token, in
        human-readable canonical-sexp-string form.
        """
        return self.__serverPubKeySexpStr


    def __getServerPubKeyMV(self):
        """
        `serverPubKeyMV' is the public key used by the token server for this token, in a modval
        instance.
        """
        return MojoKey.makeRSAPublicKeyMVFromSexpString(self._get_server_public_key_sexp())

    def get_server_pub_key_len(self):
        """
        Returns the length in bytes of the server pub key.  This should be at least 128,
        although smaller values may be seen in testing (because smaller pub keys make for faster
        tests.)
        """
        return len(self.__getServerPubKeyMV().get_modulus())

    def get_id(self):
        """
        `id' is a unique number which distinguishes this token from every other token that will
        ever exist.  It is also unguessable.

        @postcondition Result must be of correct form.: canon.is_canonical(result, self.get_server_pub_key_len())
        """
        return make_tokenId(self.__getSeed(), self._get_server_public_key_sexp())


    def __getSeed(self):
        """
        `seed' is used to generate id, and is later used to prove that id was correctly
        generated.  It must be a universally unique, unguessable number.

        @postcondition Result is of correct form.: idlib.is_sloppy_id(result)
        """
        return self.__seed


    def __get_id_for_signing(self):
        """
        `idForSigning' is the representation of the id which the token-signer should sign.

        @postcondition Result is of correct form.: canon.is_canonical(result, self.get_server_pub_key_len())
        """
        self.__preprocessForSigning()
        return self.__idForSigning


    def __getSignature(self):
        """
        `signature' is the signature by the token-signer upon the token id.

        @postcondition Result is of correct form.: canon.is_canonical(result, self.get_server_pub_key_len())
        """
        return self.__signature

    def _get_signature(self):
        return self.__getSignature()

    def is_signed(self):
        """
        @return `true' if and only if a good signature has been applied to this token with the
            `set_signature()' method.
        """
        return self.isSigned()

    def isSigned(self):
        """
        @return `true' if and only if a good signature has been applied to this token with the
            `set_signature()' method.

        @deprecated use `is_signed()' for modern naming convention
        """
        return self.__getSignature() != None

    def getFormattedFullToken(self):
        """
        `formattedFullToken' is the id, the seed, the signature and the verifying public key id 
        with some meta-data.  This is the representation of the token which a recipient of
        payment needs in order to claim the payment.

        @precondition Token must have been signed by calling `set_signature()'.: self.is_signed()

        @postcondition Result must match OurMessagesToken.PAYMENT_TOKEN_TEMPL.: MojoMessage.checkTemplate(result, OurMessagesToken.PAYMENT_TOKEN_TEMPL): "result: %s" % hr(result)
        """
        assert self.is_signed(), "Token must have been signed by calling `set_signature()'."

        # un-canonicalize the id (to allow for backwards compatibility):
        pubkeybinaaid = mojosixbit.b2a(idlib.sloppy_id_to_bare_binary_id(self.get_server_public_key_id())) # XXXX Zooko: upgrade this...  --Zooko 2000-08-29

        return { 'token id seed': mojosixbit.b2a(self.__getSeed()), 'token id signature': mojosixbit.b2a(self._get_signature()), 'id of public key for verifying this token': pubkeybinaaid }

    def get_formatted_full_token(self):
        return self.getFormattedFullToken()

    def get_id_for_signing(self):
        """
        `id_for_signing' is the id and the verifying public key id with some meta-data
        """
        # un-canonicalize the id (to allow for backwards compatibility):
        pubkeybinaaid = mojosixbit.b2a(idlib.sloppy_id_to_bare_binary_id(self.get_server_public_key_id())) # XXXX Zooko: upgrade this...  --Zooko 2000-08-29

        return { 'token id for signing': mojosixbit.b2a(self.__get_id_for_signing()), 'id of public key for verifying this token': pubkeybinaaid }

    def __preprocessForSigning(self):
        """
        Do any preparations before you will request that the token id be signed.  Also, put the id, formatted for signing, into `idForSigning'.
        """
        id = self.get_id()

        assert canon.is_canonical(id, self.get_server_pub_key_len()), "postcondition: Result must be of correct form. -- id: %s" % hr(id)

        self.__idForSigning = canon._canon(id, self.get_server_pub_key_len())

    def set_signature(self, signature):
        """
        @param signature a signature on this token by the token server's private key

        @throws BadSignatureError if `signature' is not a signature on this token signed by
            `self.__serverPubKey'

        @precondition `signature' is of correct form.: canon.is_canonical(signature, self.get_server_pub_key_len()): "signature: %s" % hr(signature)
        """
        assert canon.is_canonical(signature, self.get_server_pub_key_len()), "`signature' is of correct form." + " -- " + "signature: %s" % hr(signature)

        self.__postprocessAfterSigning(signature)
    
    def __postprocessAfterSigning(self, signature):
        """
        Do any post-processing after receiving the signature back from the token-signer.

        This verifies that the signature is correct, raising an exception if it is not.

        @param signature a signature on this token by the token server's private key

        @throws BadSignatureError if `signature' is not a signature on this token signed by
            `self.__serverPubKey'

        @precondition `signature' is of correct form.: canon.is_canonical(signature, self.get_server_pub_key_len()): "signature: %s" % hr(signature)
        """
        assert canon.is_canonical(signature, self.get_server_pub_key_len()), "`signature' is of correct form." + " -- " + "signature: %s" % hr(signature)

        if modval.verify_key_and_value(self.__getServerPubKeyMV().get_modulus(), signature) != None:
            raise BadSignatureError

        kMV = self.__getServerPubKeyMV()

        kMV.set_value_string(signature)

        kMV.undo_signature()

        backAgain = kMV.get_value()

        assert canon.is_canonical(backAgain, self.get_server_pub_key_len())

        assert canon.is_canonical(self.__get_id_for_signing(), self.get_server_pub_key_len())

        if backAgain != self.__get_id_for_signing():
            raise BadSignatureError

        self.__signature = signature
