#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: MojoTransaction.py,v 1.264.2.1 2001/06/12 03:37:30 zooko Exp $'


# standard modules
import copy
import exceptions
import math
import os
from sha import sha
import string
import sys
import threading
import time
import traceback
from traceback import print_stack, print_exc
import types


# our modules
import BidStrategy
import BlockServerHandicapper
import Cache
from CommHints import HINT_EXPECT_RESPONSE, HINT_EXPECT_MORE_TRANSACTIONS, HINT_EXPECT_NO_MORE_COMMS, HINT_EXPECT_TO_RESPOND, HINT_THIS_IS_A_RESPONSE, HINT_NO_HINT
import CommStrat
import CommsHandler
import Conversation
import CryptoCommsHandler
import DoQ
from FailBasedSelecter import FailBasedSelecter
import IdTrackerEGTP
import MetaTrackerLib
import MojoConstants
from MojoConstants import true, false, DEBUG_MODE
from MojoHandicapper import MojoHandicapper
import MojoKey
import MojoMessage
import MojoTransaction
import NewbieLovingHandicapper
import PerformanceHandicapper
import Purse
import RelayServerHandlers
import RequestHandler
import RunTests
import SEXP
import TCPCommsHandler
from UnreliableHandicapper import UnreliableHandicapper
import confutils
from confutils import confman
import counterparties
import debug
from human_readable import hr
import idlib
import ipaddresslib
import loggedthreading
import mencode
import mesgen
import mojosixbit
import mojoutil
import pricelib
import randsource
import tokendists
import whrandom
import contentlib
import MojoToken

debugwrite = debug.mojolog.write


# this is the number of times we'll retry getting contact information (every time with a different metatracker presumably)
DEFAULT_RETRIES = 3

# make users throttle a lot less than the metatracker
if confman.is_true_bool(('YES_NO', 'RUN_META_TRACKER',)):
    MAX_TIME_BETWEEN_IDLE = 90  # 90 seconds
else:
    MAX_TIME_BETWEEN_IDLE = 180  # three minutes

    
"""
When any message arrives that initiates a new mojotransaction then the appropriate "receive
func" that you registered will be called, but only after the counterparty has been charged the
appropriate amount.  How much is the appropriate amount?  That is determined by a dymanic 2nd
best price auctioning system.
"""

class Widget:
    def __init__(self, counterparty_id, firstmsgId=None):
        """
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        """
        assert idlib.is_sloppy_id(counterparty_id), "`counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

        counterparty_id = idlib.canonicalize(counterparty_id, "broker")

        self._counterparty_id=counterparty_id
        self._firstmsgId=firstmsgId

    def __repr__(self):
        return "<MojoTransaction.Widget instance (id: %x), counterparty_id: %s, conversation_id: %s>" % \
               (id(self), hr(self._counterparty_id), hr(self._firstmsgId))

    def get_counterparty_id(self):
        return self._counterparty_id

    def get_conversation_id(self):
        return idlib.make_id_from_uniq(uniq=self._firstmsgId, thingtype='conversation')
        

class Error(exceptions.StandardError): pass
class FailureError(Error): pass # FailureError is for failures in the conversation/transaction layer
class PricerError(Error): pass

POLLING = 0
NOT_POLLING = 1
GOING_TO_STOP_POLLING = 2

BASIC_DENOM = MojoToken.MAP_DENOM_TO_KEYID.keys()[0] ### !!! XXX Zooko: add denoms  --Zooko 2000/06/13



STAY_PAID_UP_TOKEN_SIZE = BASIC_DENOM



class ResponseMarker: pass

# The following symbols really just belong here in MojoTransaction.py, but Python's inability to do mutually recursive modules means that I can't access ConversationManager from here and also have ConversationManager access these symbols from Conversation.py.  So I'll just shove them into MojoConstants.  Bleah.  --Zooko 2000-09-28

ASYNC_RESPONSE = ResponseMarker()
NO_RESPONSE = ResponseMarker()

# XXX this is very bad to be adding member to another module at run time!  please use this module's references to the above instead -g
MojoConstants.ASYNC_RESPONSE = ASYNC_RESPONSE # possible return value from an incoming message handler. # `ASYNC_RESPONSE' means I'll provide the response later.
MojoConstants.NO_RESPONSE = NO_RESPONSE # possible return value from an incoming message handler. # `NO_RESPONSE' means that the correct, permanent, immutable response to the incoming message is to send no response.  Mappings from incoming messages to responses are immutable: you MUST NOT change your mind and send a response later after you have returned `NO_RESPONSE'.  If you mean "I have no response right now, but later I'll come up with the correct, permanent, immutable response.", then return `ASYNC_RESPONSE'.

# Currently, handlers that handles _responding_ messages are assumed to have no response (that is: every transaction is a two-move protocol), so if you return `None' (the default return value in Python, same as what happens if you have no return statement at all), then it translates to `NO_RESPONSE' for you.  It wouldn't hurt to go ahead and return `NO_RESPONSE' from those handlers, if you wish to emphasize that they are there are no more moves in the protocol.

# Currently, every handler that handles _initiating_ messages returns a response, with the exception of the handler that handles payment messages.  That handler has two modes of operation: 1. quick payments, in which a response is immediately sent to the payer that indicates that the payer can go ahead with their transactions, and 2. safe payments, in which no response is sent back to the payer until after the payee has contacted the token server to verify the tokens.  In the former case, the payments handler returns a response (which is sent back to the payer), and in the latter case, the payments handler returns `ASYNC_RESPONSE', no response is sent to the payer until later.


class MojoTransactionManager:
    """
    There are basically two interfaces to MojoTransactionManager, one for handling transactions
    initiated by you and satisfied by a remote broker acting as a server, and one for handling
    transactions initiated by a remote broker acting as client and satisfied by you.

    If you want to handle incoming request messages (for example, you are a block server and you
    want to handle incoming 'get block' messages), then you define a handler func, which is a
    function that takes two arguments named 'widget' and 'msgbody', and returns a response.  The
    response will be sent back to the client, unless it is an instance of ResponseMarker.  See
    the documentation for the ResponseMarkers, above.

    If you want to initiate a transaction with a remote broker acting as server, call
    `MojoTransactionManager.initiate()', passing the id of the remote broker, the contents of
    the query message, and optionally a callback function that will be called when the
    transaction completes or fails.

    For an example of server behavior, see "server/merchant/BlockServerEGTP.py".  For an example
    of client behavior, see "common/Paytool.py".
    """
    def __init__(self, doQ, pt=None, announced_service_dicts=[], handler_funcs={}, serialized=None, relay_server=None, listenport=None, recoverdb=true, pickyport=false, dontbind=false, keyID=None, allow_send_metainfo=true):
        """
        @param dontbind `true' if and only if you don't want to bind and listen to a port

        @precondition `doQ' must be a DoQ.: isinstance(doQ, DoQ.DoQ): "doQ: %s" % hr(doQ)
        @precondition `announced_service_dicts' must be a list.: type(announced_service_dicts) == types.ListType: "announced_service_dicts: %s" % hr(announced_service_dicts)
        @precondition `handler_funcs' must be a dict.: type(handler_funcs) == types.DictType: "handler_funcs: %s" % hr(handler_funcs)
        @precondition `listenport' must be a non-negative integer or `None'.: (listenport is None) or ((type(listenport) == types.IntType) and (listenport > 0)): "listenport: %s" % hr(listenport)
        """
        assert isinstance(doQ, DoQ.DoQ), "`doQ' must be a DoQ." + " -- " + "doQ: %s" % hr(doQ)
        assert type(announced_service_dicts) == types.ListType, "`announced_service_dicts' must be a list." + " -- " + "announced_service_dicts: %s" % hr(announced_service_dicts)
        assert type(handler_funcs) == types.DictType, "`handler_funcs' must be a dict." + " -- " + "handler_funcs: %s" % hr(handler_funcs)
        assert (listenport is None) or ((type(listenport) == types.IntType) and (listenport > 0)), "`listenport' must be a non-negative integer or `None'." + " -- " + "listenport: %s" % hr(listenport)

        self._shuttingdownflag = false

        self._tcpch = None  # will later contain our TCPCommsHandler instance if any
        self._reqhandler = RequestHandler.RequestHandler(self)

        if handler_funcs:
            for i in handler_funcs.keys():
                assert type(i) == types.StringType
            for i in handler_funcs.values():
                assert type(i) == types.TupleType
                assert len(i) == 2
                assert type(i[0]) == types.FunctionType or type(i[0]) == types.MethodType
                assert type(i[1]) == types.FunctionType or type(i[1]) == types.MethodType

        self._doQ=doQ

        self.__announced_service_dicts=copy.copy(announced_service_dicts)
        self._handicapper=MojoHandicapper(doq=self._doQ)
        self._relay_server_failer = FailBasedSelecter(lambda confman=confman: int(confman.dict.get("RELAY_SERVER_POLL_FAILURE_WINDOW_SECS_v2", 900)), lambda confman=confman: int(confman.dict.get("RELAY_SERVER_POLL_FAILURE_LIMIT", 4)), relay_server)
        self._most_recent_relay_poll_time = 0
        self.__polling = NOT_POLLING

        self._allow_send_metainfo = allow_send_metainfo  # controls if we allow adding our metainfo to outgoing messages on occasion
        self.__counterparties_metainfo_sent_to_map = Cache.StatsCache(maxitems=10000, autoexpireinterval=600, autoexpireparams={'maxage': 1800})
        self.__need_sequence_update = true  # determines if we update our sequence number when generating a hello

        self._metatrackers_sent_hello = []  # list of (id, info) tuples of metatrackers we most recently said hello to

        # a map of counterpartyids -> the last time we tried sending a token to them in order to stay paid up
        # (used as a quick 'works in most all cases' hack to prevent sending several tokens to someone at once
        # to stay paid up)
        self._stay_paid_up_time_map = Cache.StatsCache(maxitems=5000, autoexpireinterval=300, autoexpireparams={'maxage': 60})

        self.calls_to_use_relay_server = 0
        self.__retried_list_relay_servers = false
        # this is set during use_relay_server() processing while its still waiting for contact info so that it can pick a relay server.
        # (it is used in __announce_self_to_id_trackers and IdTrackerEGTP.reload_own_metainfo())
        self.allow_hello_to_be_sent = true
        # used to determine to set_relay_server and start_listening should send a hello messsage.  This prevents
        # the "problem" of sending two hellos when serving via relay.
        self.initial_hello_has_been_sent = None

        if handler_funcs:
            self._handler_funcs=copy.copy(handler_funcs)
        else:
            self._handler_funcs={}

        dbparentdir=os.path.expandvars(confman.dict["PATH"]["MOJO_TRANSACTION_MANAGER_DB_DIR"])
        if (serialized is None) and (keyID is None):
            if RunTests.is_testing():
                # if testing, use a precomputed value
                try:
                    # try to read an id out of our file of test ids...
                    # make sure that the file is open...
                    global testidfile
                    if not testidfile:
                        mtmdbdir=os.path.expandvars(confman.dict["PATH"]["MOJO_TRANSACTION_MANAGER_DB_DIR"])
                        testidfname=os.path.join(mtmdbdir, "test-keys.txt")
                        testidfile=open(testidfname, "r+")

                    myid_aa=mojoutil.popTrailingNewlines(testidfile.readline())
                    
                    dir=os.path.normpath(os.path.join(dbparentdir, myid_aa))
                    self._mesgen=mesgen.load_MessageMaker(dir, recoverdb=recoverdb)
                    debug.stderr.write("Using testing key generated during a previous run.\n")
                except:
                    # Whatever.  Generate a new one, then.
                    self._mesgen=mesgen.create_MessageMaker(dbparentdir, recoverdb=recoverdb)

                    debug.stderr.write("Couldn't find precomputed key, generating new key.\n")

                    # Now try to store its id in our file of test ids.
                    try:
                        # make sure that the file is open...
                        global testidfile
                        if not testidfile:
                            mtmdbdir=os.path.expandvars(confman.dict["PATH"]["MOJO_TRANSACTION_MANAGER_DB_DIR"])
                            testidfname=os.path.join(mtmdbdir, "test-keys.txt")
                            testidfile=open(testidfname, "a+")

                        testidfile.write(mojosixbit.b2a(self._mesgen.get_id()) + "\n")

                        debug.stderr.write("Putting new key in %s.\n", args=(testidfname,))
                    except:
                        # Whatever.  So we couldn't store it.  Oh well.
                        pass
            else:
                # if not testing, generate new, secret, secure one
                self._mesgen=mesgen.create_MessageMaker(dbparentdir=dbparentdir, recoverdb=recoverdb)
        else:
            if keyID and dbparentdir:
                self._mesgen=mesgen.MessageMaker(dir=os.path.join(dbparentdir, keyID), serialized=serialized, recoverdb=recoverdb)
            else:
                self._mesgen=mesgen.MessageMaker(dbparentdir=dbparentdir, serialized=serialized, recoverdb=recoverdb)

        self._cm=Conversation.ConversationManager(self)
        # If `listenport' is None, we can choose any port we like.  We'll choose one determined by the first few bits
        # of our pubkeyid, and make sure it is higher than 1025 and lower than 32767.
        if listenport is None:
            FLOOR=1026
            CEIL=32766
            listenport = ((ord(self._mesgen.get_id()[0]) * 256 + ord(self._mesgen.get_id()[1])) % (CEIL - FLOOR)) + FLOOR

        TCPch = TCPCommsHandler.TCPCommsHandler(listenport=listenport, pickyport=pickyport, dontbind=dontbind)
        self._tcpch = TCPch   # used by RequestHandler to watch bandwidth usage
        self._doQ._asyncorelooper._tcpcommshandler_list.append(TCPch)   # XXX HACK see mojoasyncore.py
        self._ch=CryptoCommsHandler.CryptoCommsHandler(mesgen=self._mesgen, tcpch=TCPch, doQ=self._doQ)

        self._keeper=counterparties.CounterpartyObjectKeeper(dbparentdir, local_id=self.get_id(), recoverdb=true)
        # IMPORTANT NOTE:  handicappers are ordered as some of them are in-progress multipliers (such as
        # the performance handicappers which should come after price so that they can scale independently
        # of later handicappers values)
        # When reading the following list, remember that all scalar additives are squared before
        # being added.  The "@returns" doco below is _before_ squaring.  All "@returns" numbers
        # below are additives unless specified to be multipliers.

        # for message_type[:len('content tracker')] == 'content tracker'
        #  @returns a complex function of content tracking history, I think its magnitude is between 0 and 1700.
        self.get_handicapper().add_handicapper(contentlib.ContentTrackerHandicapper(self._keeper), message_types=['content tracker lookup', 'content tracker submit'])

        # for all msgtypes 
        #  @returns counterparty's costliness * TUNING_FACTOR, where TUNING_FACTOR == 1000
        self.get_handicapper().add_handicapper(pricelib.PriceHandicapper(self._keeper).compute_handicap)

        # for msgtype == 'put blob' 
        #  @returns a function of put blob bytes/sec history scaled based on connection speed settings and a configured scaling factor
        self.get_handicapper().add_handicapper(PerformanceHandicapper.PutBlobThruputHandicapper(self._keeper), message_types=['put blob'])

        #  for msgtype == 'request blob' 
        #  @returns a function of put blob bytes/sec history scaled based on connection speed settings and a configured scaling factor
        self.get_handicapper().add_handicapper(PerformanceHandicapper.RequestBlobThruputHandicapper(self._keeper), message_types=['request blob'])

        # for msgtype == 'do you have blobs'
        #  @returns average *recent* latency in seconds times 500, or 500 if no recent information is available
        self.get_handicapper().add_handicapper(PerformanceHandicapper.DoYouHaveBlobsLatencyHandicapper(self._keeper), message_types=['do you have blobs'])

        # for msgtype in ('do you have blobs', 'request blob', 'put blob')
        # @returns (20 - length of bitmask) * 10 or 100, or 10000 if mask excludes blob; Note that default bitmask is length 7.
        self.get_handicapper().add_handicapper(BlockServerHandicapper.compute_handicap, message_types=('do you have blobs', 'request blob', 'put blob'))

        # for all msgtypes
        #  @returns 1000 if counterparty is "unreliable";  With default settings, you are unreliable if you have dropped more than 30% of queries or more than 2000 queries throughout all history.
        self.get_handicapper().add_handicapper(UnreliableHandicapper(self._keeper, self.get_id()))

        # for all msgtypes
        #  @returns the value of all MojoTokens I have paid her minus the value of all MojoTokens she has paid me
        # self.get_handicapper().add_handicapper(NewbieLovingHandicapper.NewbieLovingHandicapper(self._keeper).compute_handicap) # don't comment this in!  it is for experimental purposes -- not good business.  --Zooko 2000-11-02
        if confman.is_true_bool(['KLOOGE_TEST_NEWBIE_LOVER']):
            self.get_handicapper().add_handicapper(NewbieLovingHandicapper.NewbieLovingHandicapper(self._keeper).compute_handicap)

        # significantly handicap counterparties that haven't been responding to their messages recently at all or fast enough
        self.get_handicapper().add_handicapper(self._cm.pending_responses_handicapper)

        # this handicapper must go last, it multiplies our handicap by 0
        # so that we're our own lowest handicap.
        def always_prefer_myself(counterparty_id, metainfo, message_type, message_body, our_id=self.get_id(),
            # we don't want to prefer ourselves for some things...
            types_we_dont_want = [
                'pass this along v2',
                'are there messages v2',
                'retrieve messages v2',
            ],
        ) :
            if idlib.equal(counterparty_id, our_id) and message_type not in types_we_dont_want :
                return {'multiplier': 0}
            return 0
        self.get_handicapper().add_handicapper(always_prefer_myself)

        # this is used to prevent >1 update from occurring at the same time
        self.__handler_funcs_and_services_dicts_update_lock=threading.Lock()

        # this is used to prevent deep or infinite nesting when decoding 'message for you' messages
        self.__in_message_for_you = false
        # allow us to receive 'message for you' messages that relay servers will use if we request fast relay from them
        self.__init_message_for_you_handler()

        self.set_pt(pt)

        self._lastpolltime=time.time()

        self.set_metalookup_mtm(self)

        self._doQ.add_task(self._reqhandler._stale_check_doq_event)   # toss stale events off the incoming request queue

        self.stopped_due_to_load = false
        self.start_listening()

    def get_total_mojo(self):
        """
        Total tokens in purse, plus total mojo others owe you, minus total mojo you owe others.

        This could be misleading, especially as there might be lots of counterparties who owe
        you but who are never ever going to pay up...
        """
        return self._keeper.get_total_mojo_offer_balance_adjustment() + self._pt.get_amount_in_purse()
    
    def start_listening(self):
        debug.mojolog.write("start_listening()\n", v=1, vs='MojoTransaction')
        self._ch.start_listening(inmsg_handler_func=self._cm.handle_raw_message)
        self.stopped_due_to_load = false
        if self.initial_hello_has_been_sent:
            self.send_hello_to_meta_trackers()

    def stop_listening(self):
        self._ch.stop_listening()

    def is_listening(self):
        return self._ch.is_listening()

    def set_pt(self, pt):
        self._pt = pt
        if self._pt:
            self._pt._register_omtm(self)
    
    def set_metalookup_mtm(self, m):
        self._metamtm = m
     
    def server_thread_join(self, t=1000000000):
        time.sleep(31000000L)

    def get_handicapper(self):
        return self._handicapper


    def clear_all_handler_funcs(self):
        self.__handler_funcs_and_services_dicts_update_lock.acquire()
        try:
            self._handler_funcs={}
        finally:
            self.__handler_funcs_and_services_dicts_update_lock.release()
            pass
    
    def clear_all_announced_services(self):
        self.__handler_funcs_and_services_dicts_update_lock.acquire()
        try:
            self._announced_service_dicts = {}
        finally:
            self.__handler_funcs_and_services_dicts_update_lock.release()
            pass

    def clear_references(self):
        if hasattr(self, '_blobserver'):  # XXX yuck. works with the kloodge in BlockServerEGTP
            del self._blobserver
        if hasattr(self, '_handicapper'):
            del self._handicapper
        if hasattr(self, '_keeper'):
            del self._keeper
        if hasattr(self, '_mesgen'):
            del self._mesgen
        if hasattr(self, '_metamtm'):
            del self._metamtm
        if hasattr(self, '_cm'):
            self._cm.shutdown()
            del self._cm
        if hasattr(self, '_tcpch'):
            del self._tcpch
        if hasattr(self, '_ch'):
            self._ch.stop_listening()
            self._ch.shutdown()
            del self._ch

    def shutdown(self):
        self._shuttingdownflag = true
        our_id = self.get_id()
        debug.mojolog.write("entering MTM.shutdown for %s\n", args=(our_id,))
        self.stop_polling()
        self.clear_all_announced_services()
        self.clear_all_handler_funcs()
        self.clear_references()
        debug.mojolog.write("exiting MTM.shutdown for %s, %s\n", args=(our_id, repr(self)))


    def __del__(self):
        debug.mojolog.write("%s __del__ called\n", args=(self,))
        self.clear_references()

    
    def add_handler_funcs(self, updated_handler_funcs_dict):
        """Add more handler_funcs to this MTM or redefine current ones."""
        self.__handler_funcs_and_services_dicts_update_lock.acquire()
        try:
            new_handler_funcs_dict=copy.copy(self._handler_funcs)
            new_handler_funcs_dict.update(updated_handler_funcs_dict)
            # The update of self._handler_funcs is done this way so that
            # it is atomic and can't possibly interfere with other threads
            # currently accessing the old self._handler_funcs dict.
            self._handler_funcs=new_handler_funcs_dict
        finally:
            self.__handler_funcs_and_services_dicts_update_lock.release()
            pass
        self._hello_sequence_num_needs_increasing()

    def add_announced_services(self, service_dict_list):
        """
        Add more announced services to our list of servers we run.

        @precondition `service_dict_list' must be a list.: type(service_dict_list) == types.ListType: "service_dict_list: %s" % hr(service_dict_list)
        """
        assert type(service_dict_list) == types.ListType, "`service_dict_list' must be a list." + " -- " + "service_dict_list: %s" % hr(service_dict_list)

        self.__handler_funcs_and_services_dicts_update_lock.acquire()
        try:
            assert type(self.__announced_service_dicts) == types.ListType, "self.__announced_service_dicts: %s" % hr(self.__announced_service_dicts)
            self.__announced_service_dicts.extend(service_dict_list)
        finally:
            self.__handler_funcs_and_services_dicts_update_lock.release()
            pass
        self._hello_sequence_num_needs_increasing()
    
    def remove_handler_funcs(self, message_type_list):
        """
        remove handler functions for the specified message types.
        """
        assert type(message_type_list) == types.ListType, "%s is not a list." % hr(message_type_list)
        self.__handler_funcs_and_services_dicts_update_lock.acquire()
        try:
            for msgtype in message_type_list:
                if self._handler_funcs.has_key(msgtype):
                    del self._handler_funcs[msgtype]
                    self._hello_sequence_num_needs_increasing()
        finally:
            self.__handler_funcs_and_services_dicts_update_lock.release()

    def remove_announced_services(self, service_name_list, sendhello=true):
        """
        remove service announcements for the given service types
        """
        assert type(service_name_list) == types.ListType, "%s is not a list." % hr(service_name_list)
        self.__handler_funcs_and_services_dicts_update_lock.acquire()
        try:
            new_service_dicts = []
            for servicedict in self.__announced_service_dicts:
                for servicename in service_name_list:
                    if servicedict.get('type') == servicename:
                        self._hello_sequence_num_needs_increasing()
                        break
                else:
                    new_service_dicts.append(servicedict)
            self.__announced_service_dicts[:] = new_service_dicts
        finally:
            self.__handler_funcs_and_services_dicts_update_lock.release()
        if sendhello and self.initial_hello_has_been_sent:
            self.send_hello_to_meta_trackers()

    def send_hello_to_meta_trackers(self):
        self.__announce_self_to_id_trackers()
        self.initial_hello_has_been_sent = 1

    def _hello_sequence_num_needs_increasing(self):
        if not self.__need_sequence_update:
            self.__need_sequence_update = true
            # wipe the cache of who we've sent metainfo to since it is now false as our metainfo has been updated
            self.__counterparties_metainfo_sent_to_map.expire(maxage=0)

    def _get_our_hello_msgbody(self):
        """
        This is for use solely by a Meta/Id tracker that we are running to avoid the special
        cases with sending ourself a hello message.
        """
        my_cs=self.get_comm_strategy()

        hello_body={
            'connection strategies': [my_cs.toDict()],
            }
        if len(self.__announced_service_dicts) > 0:
            hello_body['services']=self.__announced_service_dicts
        hello_body['broker version'] = confman.dict.get("BROKER_VERSION_STR", "I'm not telling!")
        hello_body['platform'] = confutils.platform # XXX Jim says: "This is to be removed after the beta period."  --Zooko 2000-08-22

        hello_body['sequence num'] = self.get_hello_sequence_num()
        hello_body['dynamic pricing'] = "true"

        return hello_body

    def get_hello_sequence_num(self):
        """
        Get our current hello sequence number, incrementing it and
        saving it to the config file if it needs updating.
        """
        seqconfkey = "MTM_HELLO_SEQUENCE_NUMS"
        asciiid = idlib.to_ascii(self.get_id())
        needtosave = false
        if not confman.dict.has_key(seqconfkey):
            confman.dict[seqconfkey] = {}
            needtosave = true
        if not confman.dict[seqconfkey].has_key(asciiid):
            confman.dict[seqconfkey][asciiid] = str(0) #
            needtosave = true
        if self.__need_sequence_update:
            self.__need_sequence_update = false
            strstr= confman.dict[seqconfkey][asciiid]
            if strstr[-1:] == "L": # remove annoying L's -- python 2.0 doesn't do this
                strstr = strstr[:-1]
            s = str(long(strstr) + 1)
            if s[-1] == 'L':
                s = s[:-1]
            confman.dict[seqconfkey][asciiid] = s
            needtosave = true
        if needtosave:
            # save the current sequence number in the config file so that we never lower it
            # (XXX storing it somewhere other than the cfg file such as the mtmdb/<id>/ directory
            # would be a good idea before we implement persistent metainfo... -greg)
            confman.save()
        strstr = confman.dict[seqconfkey][asciiid]
        if strstr[-1:] == "L": # remove annoying L's -- python 2.0 doesn't do this
            strstr = strstr[:-1]
        seqnum = long(strstr)
        return seqnum

    def get_metainfo_dict(self):
        """
        Return an info dict about ourselves suitable for use with
        MetaTrackerLib (basically this is what a meta tracker lookup
        of us would return)
        """
        hello_dict=self._get_our_hello_msgbody()
        metainfo_dict = copy.copy(hello_dict)
        return metainfo_dict

    def __announce_self_to_id_trackers(self):
        if self._shuttingdownflag:
            return

        if not self.allow_hello_to_be_sent:
            # wait a few seconds and try again, we're still waiting on new relay server contact info
            self._doQ.add_task(self.__announce_self_to_id_trackers, delay=5)
            return
        hello_body = self._get_our_hello_msgbody()
        MetaTrackerLib.stuff_hello_into_cache_for_testing(hello_body)

        # Announce yourself to a few of the MT's, randomly chosen.
        mts = mojoutil.rotatelist(MetaTrackerLib.find_root_meta_trackers(self._metamtm))

        # How many?  log_base_2 of the number of MTs, plus 1.  That means the following map from num MTs to num to hello: { 1: 1, 2: 2, 3: 2, 4: 3, 5: 3, 6: 3, 7: 3, 8: 4, 16: 5, 32: 6, 64: 7, 128: 8, 256: 9, 512: 10 }
        numtohello = mojoutil.int_log_base_2(len(mts)) + 1

        self._metatrackers_sent_hello = mts[:numtohello]
        for (metatracker_id, metatracker_info) in mts[:numtohello]:
            if self.get_id() == metatracker_id:
                # don't announce self to self, that gets weird. (???)
                continue

            fixedprice = pricelib.meta_tracker_hello_price(metatracker_info, hello_body)
            if fixedprice is not None:
                # Remove this once we've grandfathered-out most of the fixed-price MTs...  --Zooko 2001-06-01
                mojooffer = fixedprice
            else:
                mojooffer = BidStrategy.choose_bid("MO_HELLO", self, counterparty_id=metatracker_id)

            self.initiate(counterparty_id=metatracker_id, conversationtype='hello', firstmsgbody=hello_body, mojooffer=mojooffer)

    def send_goodbye_to_metatrackers(self):
        """Send sign off message to meta/id trackers we most recently said hello to; this does not wait for the send to succeed!"""
        for (metatracker_id, metatracker_info) in self._metatrackers_sent_hello:
            if self.get_id() == metatracker_id:
                continue
            fixedprice = pricelib.meta_tracker_hello_price(metatracker_info, {})
            if fixedprice is not None:
                # Remove this once we've grandfathered-out most of the fixed-price MTs...  --Zooko 2001-06-01
                mojooffer = fixedprice
            else:
                mojooffer = BidStrategy.choose_bid("MO_GOODBYE", self, counterparty_id=metatracker_id)

            self.initiate(counterparty_id=metatracker_id, conversationtype='goodbye', firstmsgbody={}, mojooffer=mojooffer, timeout=1)

    def get_public_key(self):
        """Return value is a SEXP string."""
        return self._mesgen.get_public_key()

    def get_id(self):
        return self._ch.get_id()

    def get_comm_strategy(self):
        """
        @returns the comm strategy in the form of an instance of CommStrat.CommStrat
        """
        rs = self._relay_server_failer.get_preferred()
        if rs:
            return CommStrat.Crypto(pubkey=self.get_public_key(), lowerstrategy=CommStrat.Relay(relayer_id=rs, broker_id=self.get_id(), mtm=self._metamtm))
        else:
            return self._ch.get_comm_strategy()

    def serialize(self):
        return mencode.mencode({'session keeper': self._mesgen._session_keeper.serialize()})

    def respond_with(self, prevmsgId, msgbody, mojoheader=None, _extras=None):
        """
        You must ensure that you call `respond_with()' using a given `prevmsgId', no more than
        one time in the history of the universe.  Also you must not call `response_with()' using
        a given `prevmsgId' if you returned a "response body" return value from the handler
        which originally processed the previous message.  In practice, this is pretty easy to
        ensure by relying upon MojoTransaction's guarantee that your handler will be invoked
        exactly one time in the history of the universe for one message.

        @param prevmsgId the id of the message to which this is a response
        @param msgbody the body of the response
        @param mojoheader is an optional 'mojo header' to send with this response.  It should not be used
            by normal callers as it is parsed and interpreted by MojoTransaction internally on the receiving end.

        @precondition `msgbody' is just the innermost message body, not having a "mojo header" or "mojo message" subdict.: (type(msgbody) != types.DictType) or ((not msgbody.has_key('mojo header')) and (not msgbody.has_key('mojo message'))): "msgbody: %s" % hr(msgbody)

        @precondition internal1: self._cm._map_inmsgid_to_info.get(prevmsgId) is not None: "prevmsgId: %s, msgbody: %s" % (hr(prevmsgId), hr(msgbody))
        @precondition internal2: (type(self._cm._map_inmsgid_to_info.get(prevmsgId)) == types.TupleType) or (type(self._cm._map_inmsgid_to_info.get(prevmsgId)) == types.ListType): "self._cm._map_inmsgid_to_info.get(prevmsgId): %s :: %s" % (hr(self._cm._map_inmsgid_to_info.get(prevmsgId)), hr(type(self._cm._map_inmsgid_to_info.get(prevmsgId))))
        @precondition internal3: self._cm._map_inmsgid_to_info.get(prevmsgId)[2] == Conversation.EXPECTING_RESPONSE: "self._cm._map_inmsgid_to_info.get(prevmsgId): %s" % hr(self._cm._map_inmsgid_to_info.get(prevmsgId))
        @precondition internal4: idlib.is_sloppy_id(self._cm._map_inmsgid_to_info.get(prevmsgId)[0]): "self._cm._map_inmsgid_to_info.get(prevmsgId)[0]: %s :: %s" % (hr(self._cm._map_inmsgid_to_info.get(prevmsgId)[0]), hr(type(self._cm._map_inmsgid_to_info.get(prevmsgId)[0])))
        """
        assert (type(msgbody) != types.DictType) or ((not msgbody.has_key('mojo header')) and (not msgbody.has_key('mojo message'))), "precondition: `msgbody' is just the innermost message body, not having a \"mojo header\" or \"mojo message\" subdict." + " -- " + "msgbody: %s" % hr(msgbody)
        assert self._cm._map_inmsgid_to_info.get(prevmsgId) is not None, "precondition: internal1" + " -- " + "prevmsgId: %s, msgbody: %s" % (hr(prevmsgId), hr(msgbody))
        assert (type(self._cm._map_inmsgid_to_info.get(prevmsgId)) == types.TupleType) or (type(self._cm._map_inmsgid_to_info.get(prevmsgId)) == types.ListType), "precondition: internal2" + " -- " + "self._cm._map_inmsgid_to_info.get(prevmsgId): %s :: %s" % (hr(self._cm._map_inmsgid_to_info.get(prevmsgId)), hr(type(self._cm._map_inmsgid_to_info.get(prevmsgId))))
        assert self._cm._map_inmsgid_to_info.get(prevmsgId)[2] == Conversation.EXPECTING_RESPONSE, "precondition: internal3" + " -- " + "self._cm._map_inmsgid_to_info.get(prevmsgId): %s" % hr(self._cm._map_inmsgid_to_info.get(prevmsgId))
        assert idlib.is_sloppy_id(self._cm._map_inmsgid_to_info.get(prevmsgId)[0]), "precondition: internal4" + " -- " + "self._cm._map_inmsgid_to_info.get(prevmsgId)[0]: %s :: %s" % (hr(self._cm._map_inmsgid_to_info.get(prevmsgId)[0]), hr(type(self._cm._map_inmsgid_to_info.get(prevmsgId)[0])))

        counterparty_id = self._cm._map_inmsgid_to_info.get(prevmsgId, [''])[0]  # XXX HACK (ideally this would be in Conversation or Conversation and MojoTransaction would be merged)

        # include our metainfo in messages the first time we send a message to a counterparty
        # (or again when it has been updated; this map is emptied when our metainfo changes)
        mymetainfo = None
        if self.allow_hello_to_be_sent and self._allow_send_metainfo and not self.__counterparties_metainfo_sent_to_map.has_key(counterparty_id):
            mymetainfo = self._get_our_hello_msgbody()
            self.__counterparties_metainfo_sent_to_map[counterparty_id] = None

        body = {}
        if msgbody is not None:
            body['mojo message'] = msgbody
        if mojoheader:
            body['mojo header'] = mojoheader
        if _extras:   # XXX only used as a hack to implement old style (pre 0.986.2) 'mojo refund' compatibility
            body.update(_extras)

        self._cm.send_response(prevmsgId, body, mymetainfo=mymetainfo)

    def handle_initiating_message(self, counterparty_id, msgtype, msgbody, firstmsgId):
        """
        Place the incoming request in the RequestHandler, or return a rejection letter if there is something wrong with it.

        @returns the full msg body to be sent back in response (containing a 'mojo header' and/or 'mojo message' subdict) in dict form, or None, or MojoConstants.ASYNC_RESPONSE

        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)

        @postcondition Result must be either None or MojoConstants.ASYNC_REPONSE or else the full msg body dict, containing either a "mojo header" subdict or a "mojo message" subdict or both.: (not result) or (result is MojoConstants.ASYNC_RESPONSE) or result.has_key('mojo header') or result.has_key('mojo message'): "result: %s" % hr(result)
        """ 
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

        ### if it's been a while since you've been idle, stop listening

        if time.time() - self._doQ._last_wait_time > MAX_TIME_BETWEEN_IDLE and not self.stopped_due_to_load:
            debug.mojolog.write("DoQ overloaded, %s holding incoming connections\n", args=(self.get_id(),), v=3, vs='MojoTransaction')
            self.stopped_due_to_load = true
            self._ch.hold_incoming_connections()
            # Put a call to start_listening() on the DoQ so that we start listening again once our load has gone down
            # (a delay of 5 seconds would be really unnecessary as even without the delay this event will be put at the end
            # of the things to do...)
            def unhold_incoming_connections(self=self):
                self._ch.unhold_incoming_connections()
                self.stopped_due_to_load = false
            self._doQ.add_task(unhold_incoming_connections) 

        vals=self._handler_funcs.get(msgtype)
        if not vals:
            debug.stderr.write("DEBUG: received a message of unhandled type `%s': `%s'.\n", args=(msgtype, msgbody), v=1, vs="debug")
            return None

        serverfunc, pricer=vals
        widget=Widget(counterparty_id, firstmsgId=firstmsgId)

        mojoheaderdict=msgbody['mojo header']
        mojomessagedict=msgbody['mojo message']
        mojooffer=long(mojoheaderdict['mojo offer'])

        self._reqhandler.add(firstmsgId, counterparty_id, msgtype, msgbody, mojooffer, serverfunc)

        return ASYNC_RESPONSE

    def initiate_and_wait(self, counterparty_id, conversationtype, firstmsgbody, mojooffer, timeout=MojoConstants.DEFAULT_TIMEOUT, retry=false):
        """
        Do NOT use this function in production code, it exists for easy interactive testing and simple test clients.
        This function is a simple blocking version of initiate().  It can not be called from the DoQ thread.
        """
        if self._doq.is_currently_doq():
            raise RuntimeError, "initiate_and_wait is for testing only, it may not be called from the DoQ"

        res1, reskw1, doneflag1, _wcb1=mojoutil._cb_warper()
        self.initiate(counterparty_id, conversationtype, firstmsgbody, mojooffer, outcome_func=_wcb1, timeout=timeout, retry=retry)
        doneflag1.wait()

        # The failure_reason can be by keyword or as the 3rd argument.
        if reskw1.get('failure_reason'):
            raise FailureError, reskw1.get('failure_reason')

        if (len(res1) > 3) and (res1[2]):
            raise FailureError, res1[2]

        # The outcome can be by keyword or as the 2nd argument.
        if reskw1.get('outcome'):
            return reskw1.get('outcome')

        if (len(res1) > 2) and (res1[1]):
            return res1[1]

    def initiate(self, counterparty_id, conversationtype, firstmsgbody, mojooffer, outcome_func=None, timeout=MojoConstants.DEFAULT_TIMEOUT, retry=false, notes = None, post_timeout_outcome_func=None, use_dynamic_timeout=None):
        """
        Initiates a transaction with the specified counterparty.

        @param counterparty_id the id of the counterparty with whom to transact
        @param conversationtype human readable string "message type" or "conversation type";
            See OurMessages*.py for examples.
        @param firstmsgbody the contents of the initiating message of the transaction;  This
            can be any of the Pythonic types supported by the mencode module, currently
            (in v0.913): integers (ints or longs), strings, None, sequences (lists or tuples)
            and dicts, where the items in the sequences and the keys and values in the dicts
            may be any of these same types.
        @param mojooffer the amount of the mojo offer that will be made;  This is the maximum
            that you will be charged for this service.
        @param outcome_func a callback that will be called exactly once with "widget",
            "outcome", "failure_reason", and "notes" keyword arguments;  <p>This func will be
            called with a `failure_reason' argument other than `None' if and only if there is a
            failure below the level of the transaction layer.  Typical causes of failure below
            this layer include communications error, timeout, or failure to pay Mojo.  (Note
            that sometimes your broker _chose_ incur a failure to pay Mojo, for example if the
            counterparty was charging more than you were willing to spend.)  <p>This func will
            be called with an `outcome' if and only if a satisfactory response message is
            received from the counterparty.  The `outcome' can be any of the Pythonic types
            handled by mencode (see above in doc of `firstmsgbody'), and typically is a dict
            with strings as names. <p>The "notes" argument is the same as was passed to
            this initiate call.  This allows data to be passed from initiation code to response
            handling code.
        @param timeout the maximum amount of time before `outcome_func' is called with a
            `failure_reason' indicating timeout if no response message has been received
        @param retry `true' if and only if the initiating message should be resent at
            increasing intervals until either a response is received or the timeout period is
            reached; Currently we use retrying for messages involving tokens (which are
            unforgeable and cannot be freely replaced if lost), but not for other messages.
        @param post_timeout_outcome_func is a callback with the same interface as outcome_func.
            It will only be called if outcome_func has already been called with a failure_reason
            of 'timeout' and the response messages arrives (known as a "post timeout response").  It
            is not guaranteed to be called at all.  The idea of this function is that response
            messages still contain valid data that your application could use even if they took longer
            than expected.

        @param use_dynamic_timeout is a flag to indiciate whether a dynamic timeout
            based on previous timings with this kind of message to this counterparty
            would be a good idea.  If no data is available, fall back to using timeout.
            Set this to 0 to indicate never using a dynamic timeout, 1 to indicate always,
            and None to indicate 'use one iff post_timeout_outcome_func is not None'

        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        @precondition `mojooffer' must be an integer.: (type(mojooffer) == types.IntType) or (type(mojooffer) == types.LongType): "mojooffer: %s :: %s" % (hr(mojooffer), str(type(mojooffer)))
        @precondition This MTM must not be shutting down.: not self._shuttingdownflag
        """
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)
        assert (type(mojooffer) == types.IntType) or (type(mojooffer) == types.LongType), "precondition: `mojooffer' must be an integer." + " -- " + "mojooffer: %s :: %s" % (hr(mojooffer), str(type(mojooffer)))
        assert not self._shuttingdownflag, "precondition: This MTM must not be shutting down."
        
        counterparty_id = idlib.to_binary(counterparty_id)
        if notes is not None and outcome_func:
            def wrapped_outcome_func(widget, outcome, failure_reason, 
                    outcome_func = outcome_func, notes = notes):
                outcome_func(widget, outcome, failure_reason, notes = notes)
            outcome_func = wrapped_outcome_func
        self._doQ.add_task(self._initiate, args=(counterparty_id, conversationtype, firstmsgbody, mojooffer, outcome_func,), kwargs={'timeout': timeout, 'retry': retry, 'post_timeout_outcome_func': post_timeout_outcome_func, 'use_dynamic_timeout':use_dynamic_timeout})

    def _initiate(self, counterparty_id, conversationtype, firstmsgbody, mojooffer, outcome_func, timeout=MojoConstants.DEFAULT_TIMEOUT, retry=false, post_timeout_outcome_func=None, use_dynamic_timeout=None):
        """
        @precondition mojooffer must be an integer.: (type(mojooffer) is types.IntType) or (type(mojooffer) is types.LongType): "mojooffer: %s :: %s" % (hr(mojooffer), hr(type(mojooffer)))
        @precondition This method must be called on the DoQ.: self._doQ.is_currently_doq()
        """
        assert (type(mojooffer) is types.IntType) or (type(mojooffer) is types.LongType), "precondition: mojooffer must be an integer." + " -- " + "mojooffer: %s :: %s" % (hr(mojooffer), hr(type(mojooffer)))
        assert self._doQ.is_currently_doq(), "precondition: This method must be called on the DoQ."

        counterparty_id = idlib.canonicalize(counterparty_id, "broker")
            
        ### BEGIN DYNAMIC TIMERS code
        # XXX Maybe collect stats only if we'd use them?
        if confman.is_true_bool(['COUNTERPARTY', 'USE_DYNAMIC_TIMING_testing']):
            if (use_dynamic_timeout
                or (use_dynamic_timeout is None
                    and post_timeout_outcome_func is not None)):
                counterparty = self._keeper.get_counterparty_object(counterparty_id)
                stat = 'roundtrip_time['+conversationtype+']'
                if counterparty.get_custom_stat(stat) is None:
                    # No dynamic estimate is available.  Don't mess with timeout. 
                    pass 
                else:
                    (mu,sigma,ignore) = counterparty.get_custom_stat(stat)
                    timeout = mu + 2*sigma
                    debug.mojolog.write("using dynamic timeout %s with for %s to %s\n", args=(timeout,stat,counterparty_id), v=3, vs='counterparty')
        ### END DYNAMIC TIMERS code



        ### BEGIN DYNAMIC TIMERS code
        def collect_timings(counterparty_id=counterparty_id,
                            start_time=time.time(),
                            conversationtype=conversationtype,
                            self=self):
            """Update timing statistics then call outcome_func

            The elapsed time is used to update estimates of mean and
            standard deviation of arrival times.  As a statistical
            tool, a warning is output every time the estimate is under
            by two standard deviations.  The intention is that one
            could use mu + 2*sigma as a timeout.

            The exponential averaging uses a time constant of 10, so
            it takes about 10 packets to learn aout the timeouts.
            This may be too high; it should be a configuration option.

            """
            elapsed_time = time.time() - start_time
            time_constant = float(confman["COUNTERPARTY"]["AVERAGING_TIMESCALE"])
            # DEFAULT_TIMEOUT is too high.
            default_sigma = elapsed_time

            stat = 'roundtrip_time['+conversationtype+']'
            
            counterparty = self._keeper.get_counterparty_object(idlib.canonicalize(counterparty_id, "broker"))
            if counterparty.get_custom_stat(stat) is not None:
                # We already have an estimate of how long it should have taken.
                # How good was our guess?
                try:
                    # defaults get returned only if it's you
                    (mu, sigma, ignore) = counterparty.get_custom_stat(stat,(0,0,0))
                    debug.mojolog.write("dynamic timing: message %s to %s took %s seconds (mu: %s, sigma: %s)\n", args=(stat, counterparty_id, elapsed_time, mu, sigma), v=3, vs="MojoTransaction")
                    if elapsed_time > (mu+2*sigma):
                        # Assuming a normal distribution, 99.7% caught
                        debug.mojolog.write("dynamic timing: UNUSUALLY LONG DELAY (normal distribution) on message %s to %s (took: %s mu: %s, sigma: %s)\n", args=(stat, counterparty_id, elapsed_time, mu, sigma), v=4, vs="MojoTransaction")
                    if elapsed_time > 6*mu:
                        # Assuming an exponential distribution, 99.7% caught
                        debug.mojolog.write("dynamic timing: UNUSUALLY LONG DELAY (exponential distribution) on message %s to %s (took: %s mu: %s)\n", args=(stat, counterparty_id, elapsed_time, mu), v=4, vs="MojoTransaction")
                except TypeError:
                    # Statistics are damaged.  Ignore.
                    debug.mojolog.write("dynamic timing: deleting damaged statistic %s for %s\n",
                                        args=(stat,counterparty_id),
                                        v=2, vs="MojoTransaction")
                    # this shouldn't happen, but we should nuke the damaged stat in the event that it does
                    counterparty.delete_custom_stat(stat)

            counterparty.update_custom_stat_weighted_sample_with_deviation(stat,
                                                                           elapsed_time,
                                                                           math.exp(-1./time_constant),
                                                                           default_sigma)


        def wrapped_outcome_func_collect_timings(widget,outcome,failure_reason,
                                                 outcome_func=outcome_func,
                                                 collect_timings=collect_timings):
            if failure_reason != 'timeout':
                collect_timings()
            return outcome_func(widget,outcome,failure_reason)


        def wrapped_post_timeout_outcome_func_collect_timings(widget,outcome,failure_reason,
                                              post_timeout_outcome_func=post_timeout_outcome_func,
                                              collect_timings=collect_timings):
            debug.mojolog.write("dynamic timing: late return from %s\n", args=(widget.get_counterparty_id(),), v=3, vs='MojoTransaction')
            if failure_reason != 'timeout':
                collect_timings()
            if post_timeout_outcome_func is not None:
                return post_timeout_outcome_func(widget,outcome,failure_reason)
            else:
                return None

        if confman.is_true_bool(['COUNTERPARTY', 'COLLECT_DYNAMIC_TIMING']) and outcome_func:
            outcome_func = wrapped_outcome_func_collect_timings
            post_timeout_outcome_func = wrapped_post_timeout_outcome_func_collect_timings
        ### END DYNAMIC TIMERS code

        if mojooffer > 0:
            self._keeper.get_counterparty_object(counterparty_id).spend(mojooffer, reasonstr=conversationtype)
        else:
            # this log message is similar to the one that the above spend() call prints in counterparties.py
            debug.mojolog.write("accounting with %s: requesting free transaction [reason: %s]\n", args=(counterparty_id, conversationtype), v=1, vs="counterparty")

        mojoheaderdict={'mojo offer': mojooffer}

        bodydict={
            'mojo header': mojoheaderdict,
            'mojo message': firstmsgbody,
            }

        notes = {'outer_outcome_func': outcome_func, 
            'mojooffer': mojooffer, 
            'conversationtype': conversationtype, 'firstmsgbody': firstmsgbody, 
            'timeout': timeout, 'retry': retry, 'timer': mojoutil.Timer(true), 
            'counterparty_id': counterparty_id,
            'outer_post_timeout_outcome_func': post_timeout_outcome_func}
            
        # include our metainfo in messages the first time we send a message to a counterparty
        # (or again when it has been updated; this map is emptied when our metainfo changes)
        mymetainfo = None
        if self._allow_send_metainfo and not self.__counterparties_metainfo_sent_to_map.has_key(counterparty_id):
            mymetainfo = self._get_our_hello_msgbody()
            self.__counterparties_metainfo_sent_to_map[counterparty_id] = None

        first_message_id, msg = self._cm.initiate_and_return_first_message(counterparty_id, conversationtype, bodydict, outcome_func=self._outcome_func_to_do_mojo_header, timeout=timeout, notes=notes, mymetainfo=mymetainfo, post_timeout_outcome_func=post_timeout_outcome_func)

        notes['first_message_id'] = first_message_id

        self.send_message_with_lookup(counterparty_id, msg, mojooffer, timeout=timeout, hint=HINT_EXPECT_RESPONSE | HINT_EXPECT_MORE_TRANSACTIONS, retry=retry)

        # Add resender.
        if retry:
            # It would be clever to make the initial delay equal to the average latency of this counterparty times 2.1 or something like that.
            # Currently we'll make the initial delay equal to the log of timeout times 4.  This number is 11 seconds for a 20-second timeout, 13 seconds for a 30-second timeout, 16 seconds for a 60-second timeout, 21 seconds for a 200-second timeout, and 60 seconds for a 600000-second timeout (which is what we currently (v0.913) have for token messages).
            # ...and I'm adding a min() with 1/2 of the timeout, because in unit tests, with a timeout < 10.0, the initial delay is > 9.0!
            # Similarly, the cleverest value for initial backoff would probably be the average latency of this counterparty times 1.1, or something (maybe a min of that and some fraction of time remaining until timeout...).
            # Current initial backoff is just log of timeout.  (That's 2.9 for 20 seconds, 3.4 for 30 seconds, 4.0 for 60 seconds, and 13.3 for 600000 seconds.)  (and a min() with that and timeout/10 for the same reason as the min described above.)
            self._doQ.add_task(self._retry_doq_loop, args=(counterparty_id, msg, mojooffer,), kwargs={'timeout': timeout, 'hint': HINT_EXPECT_RESPONSE | HINT_EXPECT_MORE_TRANSACTIONS, 'backoff': min(timeout/10.0, math.log(timeout))}, delay=min(timeout/2.0, math.log(timeout)*4.0))

        # Add timeout checker (only if we have an outcome func capable of handling it).
        if outcome_func:
            self._doQ.add_task(self._cm.fail_conversation, kwargs={'msgId': first_message_id, 'failure_reason': 'timeout'}, delay=timeout)

    def _outcome_func_to_do_mojo_header(self, outcome = None, failure_reason = None, notes = None):
        """
        XXX adding yet another feature to this handler (it already has several separate feature not adequately described by "to do mojo header").  The new one is forget a comm strategy if a conversation fails.  --Zooko 2001-05-04
        XXX In the future, we need to keep track of which actual specific comm strategy the conversation was trying to use, and only forget that one, rather than forgetting whatever one we are using *now*.  --Zooko 2001-05-04

        @precondition `outcome' is none or else the full message body, having a "mojo header" or "mojo message" subdict.: (not outcome) or outcome.has_key('mojo header') or outcome.has_key('mojo message'): "outcome: %s" % hr(outcome)
        """
        assert (not outcome) or outcome.has_key('mojo header') or outcome.has_key('mojo message'), "`outcome' is none or else the full message body, having a \"mojo header\" or \"mojo message\" subdict." + " -- " + "outcome: %s" % hr(outcome)

        counterparty_id = notes['counterparty_id']
        outer_outcome_func = notes['outer_outcome_func']
        mojooffer = notes['mojooffer']
        conversationtype = notes['conversationtype']
        firstmsgbody = notes['firstmsgbody']
        timeout = notes['timeout']
        retry = notes['retry']
        timer = notes['timer']
        timer.stop()

        widget=Widget(counterparty_id)
        widget._firstmsgId = notes['first_message_id']

        counterparty_obj = self._keeper.get_counterparty_object(counterparty_id)

        if failure_reason is not None:
            debug.mojolog.write("MTM: msgId: %s :: %s with %s, failed after %s seconds, failure_reason: %s\n", args=(widget._firstmsgId, conversationtype, widget.get_counterparty_id(), "%1.2f" % timer.get(), failure_reason), v=0, vs="Conversation")
            self.forget_comm_strategy(counterparty_id, widget._firstmsgId, outcome=outcome, failure_reason=failure_reason)

            counterparty_obj.decrement_reliability()
            return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': failure_reason})
        else:
            counterparty_obj.increment_reliability()

        mojoheader=outcome.get('mojo header', {})
        mojomessage=outcome.get('mojo message')
        mojorefund=0
        if outcome.has_key('mojo refund'):                   # old method of sending refunds
            mojorefund = min(outcome.get('mojo refund'), mojooffer)
        if mojoheader.has_key('amount of offer claimed'):    # new method of sending refunds
            mojorefund = max(0, mojooffer - mojoheader.get('amount of offer claimed'))
        if mojorefund > 0:
            counterparty_obj.unspend(mojorefund, reasonstr=conversationtype)

        amountspent = mojooffer - mojorefund

        debug.mojolog.write("MTM: msgId: %s :: %s with %s, completed after %s seconds for %s mojo\n", args=(widget._firstmsgId, conversationtype, widget.get_counterparty_id(), "%1.2f" % timer.get(), int(amountspent)), v=0, vs="Conversation")

        if mojooffer > 0:
            counterparty_obj.update_custom_stat_weighted_sample("costliness", amountspent/mojooffer, historyweight=0.99) # randomly guessed 0.99  --Zooko 2001-05-29
            # Note that this is number, `costliness' isn't simply the amount claimed, but a fraction of how much you
            # he claimed versus how much you were *willing* to spend...  If you were willing to spend 0, then we don't
            # count this data point.

        if (mojoheader.get('result') == "failure") and (mojoheader.get('reason') == "credit limit reached"):
            # Consider whether we are willing to front this counterparty tokens.
            ### !!! XXX TODO: make this transactional (with the processing of the message?)  --Zooko 2000/07/02
            mycreditlimitwithher=long(mojoheader['your credit limit'])
            amountshethinksioweher=long(mojoheader['amount you owe me'])
            amountioweher=counterparty_obj.get_amount_owe()
            amountiwillfronther=counterparty_obj.get_amount_will_front()
            priceofthisservice=long(mojoheader['price of this service'])

            if amountshethinksioweher != amountioweher:
                debug.stderr.write("NOTE: I am " + idlib.id_to_abbrev(self.get_id()) + ", talking to " + idlib.id_to_abbrev(widget.get_counterparty_id()) + ", she thinks I owe her " + hr(amountshethinksioweher) + " and I think I owe her " + hr(amountioweher) + ".\n", v=1, vs="accounting")

            # Now the amount of mojo she wants before she'll satisfy my request is...
            shewantsatleast=(amountshethinksioweher + priceofthisservice) - mycreditlimitwithher

            # Note!  I've decided that instead of paying her the minimum she wants (thus running a constant credit with her), I'll pay either enough to bring my balance into the positive or else as much as I'm willing to loan her, whichever is the lesser payment.  This means fewer payment transactions with more coins in each one.  --Zooko 2000-07-29
            # Also: not paying her more than I think I owe her...  --Zooko 2000-08-02
            shewantsMORE=max(amountshethinksioweher, amountioweher)

            # The least valuable batch of Mojo tokens that is worth greater than or equal to this amount is...
            if shewantsMORE > 0:
                minam=tokendists.amount_of_smallest_dist_greater_than(shewantsMORE)
            else:
                minam=tokendists.amount_of_smallest_dist_greater_than(1)

            assert (type(minam) == types.IntType) or (type(minam) == types.LongType), "`minam' must be integer. -- minam: %s" % hr(minam)
            assert minam > 0, "`minam' must be positive. -- minam: %s" % hr(minam)

            # If I gave her that much then she would owe me...
            shewouldoweme=(amountioweher * -1) + minam
            # Is it okay with me if she owes me that much?
            if shewouldoweme <= amountiwillfronther:
                debug.stderr.write("MojoTransaction.outcome_func_to_do_mojo_header(): I'm willing to pay her as much as she wants.  shewouldoweme: %s, amountiwillfronther: %s\n", args=(shewouldoweme, amountiwillfronther), v=1, vs="accounting")

                # If yes, pay her, and once she acknowledges a successful payment then do a new instantiation of the original request.
                def handle_payment_outcome(self=self, widget=widget, outcome=None, failure_reason=None, counterparty_id=widget.get_counterparty_id(), conversationtype=conversationtype, firstmsgbody=firstmsgbody, mojooffer=mojooffer, outer_outcome_func=outer_outcome_func, timeout=timeout):
                    debug.stderr.write("MojoTransaction.handle_payment_outcome(outcome=%s)\n", args=(outcome,), v=3, vs="accounting")
                    if failure_reason:
                        return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': failure_reason})

                    if outcome.get('result') == "success":
                        self.initiate(counterparty_id, conversationtype, firstmsgbody, mojooffer, outcome_func=outer_outcome_func, timeout=timeout)
                    else:
                        # Okay the payment didn't work.  Whatever.
                        debug.stderr.write("WARNING: a token payment apparently failed.  %s\n", args=(outcome,), v=0, vs="ERROR")
                        return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': 'token payment failed'})

                try:
                    assert (type(minam) == types.IntType) or (type(minam) == types.LongType), "`minam' must be integer. -- minam: %s" % hr(minam)
                    assert minam > 0, "`minam' must be positive. -- minam: %s" % hr(minam)
                    self._pt.pay(widget.get_counterparty_id(), minam, outcome_func=handle_payment_outcome)
                except Purse.NotEnoughMojoError, le:
                    return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': "failed due to not enough Mojo in local Purse: %s" % str(le)})

            else:
                # If no, (no it is not okay with me to give her as much as she wants)...
                # Would she still do business if given less?
                # (Note: should probably combine the redundant logic here and above...)  --Zooko 2000-08-02
                # (and below...)    --Zooko 2000-08-22


                # The least valuable batch of Mojo tokens that is worth greater than or equal to this amount is...
                if shewantsatleast > 0:
                    minam=tokendists.amount_of_smallest_dist_greater_than(shewantsatleast)
                else:
                    minam=tokendists.amount_of_smallest_dist_greater_than(1)

                assert (type(minam) == types.IntType) or (type(minam) == types.LongType), "`minam' must be integer. -- minam: %s" % hr(minam)
                assert minam > 0, "`minam' must be positive. -- minam: %s" % hr(minam)

                # If I gave her that much then she would owe me...
                shewouldoweme=(amountioweher * -1) + minam
                # Is it okay with me if she owes me that much?
                if shewouldoweme <= amountiwillfronther:
                    debug.stderr.write("MojoTransaction.outcome_func_to_do_mojo_header(): I'm willing to pay her as much as her minimal demand.  shewouldoweme: %s, amountiwillfronther: %s\n", args=(shewouldoweme, amountiwillfronther), v=1, vs="accounting")

                    # If yes, pay her, and once she acknowledges a successful payment then do a new instantiation of the original request.
                    def handle_payment_outcome(self=self, widget=widget, outcome=None, failure_reason=None, counterparty_id=widget.get_counterparty_id(), conversationtype=conversationtype, firstmsgbody=firstmsgbody, mojooffer=mojooffer, outer_outcome_func=outer_outcome_func, timeout=timeout):
                        debug.stderr.write("MojoTransaction.handle_payment_outcome(outcome=%s)\n", args=(outcome,), v=3, vs="accounting")
                        if failure_reason:
                            return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': failure_reason})

                        if outcome.get('result') == "success":
                            self.initiate(counterparty_id, conversationtype, firstmsgbody, mojooffer, outcome_func=outer_outcome_func, timeout=timeout)
                        else:
                            # Okay the payment didn't work.  Whatever.
                            debug.stderr.write("WARNING: a token payment apparently failed.  %s\n", args=(outcome,), v=0, vs="ERROR")
                            return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': "token payment failed"})

                    try:
                        assert (type(minam) == types.IntType) or (type(minam) == types.LongType), "`minam' must be integer. -- minam: %s" % hr(minam)
                        assert minam > 0, "`minam' must be positive. -- minam: %s" % hr(minam)
                        self._pt.pay(widget.get_counterparty_id(), minam, outcome_func=handle_payment_outcome)
                    except Purse.NotEnoughMojoError, le:
                        return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': "failed due to not enough Mojo in local Purse: %s" % str(le)})

                else:
                    # Okay we give up, she isn't willing to loan us the credit, and we aren't willing to give her an actual token.  We can't do business.  Someday really cool technology like ripped certificates will solve this problem, but for now we'll just find someone else to deal with.  --Zooko 2000-08-02
                    # Or maybe before we have really cool tech like ripped certs, we'll have a reconciliation mechanism that is protected from systemic abuse by either local or distributed reputations.  --Zooko 2000-09-30
                    debug.stderr.write("MojoTransaction.outcome_func_to_do_mojo_header(): WARNING: I'm not willing to pay her even the minimal amount so that she will do business with me, so she and I can do no business.  If I paid her, she would then owe me %s.  The amount I am willing to lend to her is: %s.\n", args=(shewouldoweme, amountiwillfronther), v=0, vs="accounting")
                    return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': "not willing to front mojo"})
        elif mojoheader.get('result') != "failure":
            # This is a successful happy response message

            # keep an average price for this service
            #paid = long(mojoheader.get('price of this service'))
            #Pricer.update_price(conversationtype, paid)
            # record that the mojo offer we sent for this message actually bought us a response
            counterparty_obj.record_spend_success()
            # stay paid up.  send a token as soon as we think we owe one.
            # (this can't be -relied- upon for payment as anyone can remove this in their own client; brokers
            # not using this code will end up with worse performance and eventually given worse prices due to
            # a lower credibility if they do remove this)
            def handle_stay_paid_up_outcome(widget, outcome=None, failure_reason=None):
                debug.mojolog.write("handle_stay_paid_up_outcome(outcome=%s)\n", args=(outcome,), v=3, vs="accounting")
                if failure_reason or outcome.get('result') != "success":
                    debug.mojolog.write("token payment failed while trying to stay paid up. [%s]\n", args=(outcome,), v=0, vs="Warning")
            # XXX the _stay_paid_up_time_map is used to prevent most cases where we might end up trying to send
            # a token to stay paid up with someone several times in a row due to the asynchronous nature of this
            # call to pay().  We won't try to 'stay paid up' more than once every 30 seconds.
            counterparty_id = widget.get_counterparty_id()
            if (counterparty_obj.get_amount_owe() > STAY_PAID_UP_TOKEN_SIZE) and \
               (self._stay_paid_up_time_map.get(counterparty_id, 0) < (time.time() - 30)):
                try:
                    tokenamount = tokendists.amount_of_smallest_dist_greater_than(STAY_PAID_UP_TOKEN_SIZE/2)
                    debug.mojolog.write("sending token(s) worth %s to %s to stay paid up\n", args=(tokenamount, counterparty_id,), v=2, vs="accounting")
                    self._stay_paid_up_time_map[counterparty_id] = time.time()
                    self._pt.pay(counterparty_id, tokenamount, outcome_func=handle_stay_paid_up_outcome)
                except Purse.NotEnoughMojoError, e:
                    debug.mojolog.write("not enough mojo in stash while trying to stay paid up with %s\n", args=(counterparty_id,), v=2, vs='Warning')
                except:
                    debug.mojolog.write("very strange error in stay paid up:\n", v=0, vs='ERROR')
                    traceback.print_exc(file=debug.mojolog)
                    debug.mojolog.write("DEBUG: self = %s, self._pt = %s, hello = %s\n", args=(self, self._pt, self._get_our_hello_msgbody()), vs='ERROR')

            # call the callback with the response message
            return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': mojomessage, 'failure_reason': None})
        else:
            # this is some other undefined failure response.
            return mojoutil.callback_wrapper(func=outer_outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': mojoheader.get('failure_reason', "failure reported in mojoheader")})

    def _retry_doq_loop(self, counterparty_id, msg, mojooffer, timeout, hint, backoff):
        """
        This is a "doq loop", which does something and then puts itself back on the doq to be
        done again later.  The loop ends when the message in question is no longer
        "unsatisfied", either because we received a response to it or timed out.
        """
        msgId = idlib.make_id(msg, 'msg')
        if self._shuttingdownflag:
            return
        if self._cm.is_unsatisfied_message(msgId):
            self.send_message_with_lookup(counterparty_id, msg, mojooffer, timeout=timeout, hint=HINT_EXPECT_RESPONSE | HINT_EXPECT_MORE_TRANSACTIONS, retry=true)

            # Each time through we exponentially increase the backoff.  (The `+0.1' is just to be sure we don't stick at 0.0!)
            self._doQ.add_task(self._retry_doq_loop, args=(counterparty_id, msg, mojooffer,), kwargs={'timeout': timeout, 'hint': HINT_EXPECT_RESPONSE | HINT_EXPECT_MORE_TRANSACTIONS, 'backoff': (backoff + 0.1) * 2.0}, delay=backoff)

    def send_message_with_lookup(self, counterparty_id, msg, mojooffer, timeout=MojoConstants.DEFAULT_TIMEOUT, hint=HINT_NO_HINT, retry=false):
        """
        This just calls `send_message_with_lookup_internal()' from the DoQ thread.
       
        But if this _is_ the DoQ thread, then it calls `send_message_with_lookup_internal()' directly.
        """
        if self._doQ.is_currently_doq():
            self.send_message_with_lookup_internal(counterparty_id, msg, mojooffer, timeout=timeout, hint=hint, retry=retry)
        else:
            self._doQ.add_task(self.send_message_with_lookup_internal, args=(counterparty_id, msg, mojooffer), kwargs={'timeout': timeout, 'hint': hint, 'retry': retry})

    def _ffh(self, msgId, retry=false, failure_reason=None, bad_commstrat=None, counterparty_id=None):
        """
        An internal use only fast fail handler that calls the conversation properly.

        @param counterparty_id is the id of the counterparty we were sending this message to.
        @param bad_commstrat (if not None) must be the particular comm strategy that the lower layers think is bad.
        """
        if retry:
            debug.mojolog.write("MojoTransactionManager: Attempt to send %s failed.  Ignoring: will try again later.  failure_reason: %s\n", args=(msgId, failure_reason,), v=5, vs="Conversation")
            return

        # XXX If we were clever we would be able to tell which specific comm strat we used to
        # send the failed message, and we would neg that particular comm strat but not any
        # others that we might know.  But we're not that smart so we'll just brute force it and
        # destroy all contact info for this counterparty and start fresh.
        # On a similar note, greg wrote most of the necessary code here to identify the
        # difference between failure due to bad comm strat (i.e. TCP connection refused) and
        # failure due to some other unknown reason (i.e. timeout), but we can't really discern
        # the difference -- for example if Alice's relay server, Roger, has gone down, then we
        # will think that our message to Alice failed for "some other reason", instead of for
        # having a bad comm strat, but it really does have a bad commstrat.
        # So for now let's brute-force that issue too -- any failure leads to total deletion of
        # all contact info and comm strats.
        if not idlib.is_sloppy_id(counterparty_id):
            debug.mojolog.write("WARNING: I want a counterparty_id here.  counterparty_id: %s :: %s, msgId: %s, failure_reason: %s, stack trace: %s\n", args=(counterparty_id, type(counterparty_id), msgId, failure_reason, traceback.extract_stack(),), v=1, vs="debug")
        else:
            MetaTrackerLib.neg_contact_info(counterparty_id)
            self.forget_comm_strategy(counterparty_id, firstmsgId=msgId, failure_reason=failure_reason)

        if bad_commstrat is not None and counterparty_id is not None:
            debug.mojolog.write("bad comm strategy for %s: %s = %s?\n", args=(counterparty_id, bad_commstrat, bad_commstrat.__dict__.get('_orig_top_cs')), v=3, vs="FastFail")

        debug.mojolog.write("MojoTransactionManager: Attempt to send %s failed.  Aborting transaction.  failure_reason: %s\n", args=(msgId, failure_reason,), v=5, vs="Conversation")
        if self._doQ.is_currently_doq():
            self._cm.fail_conversation(msgId, failure_reason=failure_reason)
        else:
            self._doQ.add_task(self._cm.fail_conversation, args=(msgId,), kwargs={'failure_reason': failure_reason})

    def forget_comm_strategy(self, counterparty_id, firstmsgId=None, outcome=None, failure_reason=None):
        # XXX would like to make this have same arguments as normal callbacks.  --Zooko 2001-05-14
        debug.mojolog.write("MojoTransaction: Forgetting comm strategy.  firstmsgId: %s, failure_reason: %s, counterparty_id: %s\n", args=(firstmsgId, failure_reason, counterparty_id,), v=4, vs="commstrats")
        # MetaTrackerLib.neg_contact_info(counterparty_id) # XXX I think maybe this causes you to starve yourself of known counterparties and suffer a very poor Mojo Net.  --Zooko 2001-05-15
        self._ch.forget_comm_strategy(counterparty_id)

    def send_message_with_lookup_internal(self, counterparty_id, msg, mojooffer, timeout=MojoConstants.DEFAULT_TIMEOUT, hint=HINT_NO_HINT, retry=false):
        """
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        @precondition `msg' must be a string, and non-empty.: (type(msg) == types.StringType) and (len(msg) > 0): "msg: %s" % hr(msg)
        """
        assert idlib.is_sloppy_id(counterparty_id), "`counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)
        assert (type(msg) == types.StringType) and (len(msg) > 0), "`msg' must be a string, and non-empty." + " -- " + "msg: %s" % hr(msg)

        counterparty_id = idlib.canonicalize(counterparty_id, "broker")

        msgId = idlib.string_to_id(msg)
        def debugprint(self=self, counterparty_id=counterparty_id, diagstr="", args=(), v=0, vs="conversation"):
            printstr="---> %s: " + diagstr + "\n"
            theseargs=[counterparty_id]

            theseargs.extend(list(args))
            debug.stderr.write(printstr, args=theseargs, v=v, vs=vs)

        def outer_fast_fail_handler(msgId=msgId, failure_reason="cannot send message", bad_commstrat=None, counterparty_id=counterparty_id, self=self, retry=retry):
            self._ffh(msgId, retry=retry, failure_reason=failure_reason, bad_commstrat=bad_commstrat, counterparty_id=counterparty_id)

        maxverb = int(confman.dict.get("MAX_VERBOSITY", 0))
        if maxverb >= 5:
            debugprint(diagstr="sending: %s, %s bytes uncomp", args=(msg, len(msg)), v=5) # super verbose
        elif maxverb >= 4:
            # XXX note, this slows sending down A LOT! calling MojoMessage.getType() calls mdecode() & template check on the message.  The mdecode part is -extremely- slow (esp on big messages).  -greg 11-oct-2000
            # if we want to display this quickly, we'll need to pass the type in from before we called mencode()
            debugprint(diagstr="sending: (id: %s, type: %s, %s bytes uncomp, ...)", args=(msgId, MojoMessage.getType(msg), len(msg)), v=4) # semi-verbose
        elif maxverb >= 3:
            debugprint(diagstr="sending: (id %s, %s bytes uncomp, ...)", args=(msgId, len(msg)), v=3)

        def fast_fail_handler(msgId=msgId, failure_reason="cannot send message", bad_commstrat=None, counterparty_id=counterparty_id, debugprint=debugprint, self=self, msg=msg, timeout=timeout, outer_fast_fail_handler=outer_fast_fail_handler, retry=retry, mojooffer=mojooffer):
            debugprint(diagstr="failed to send, forgetting comm strat and trying to look up new one...", v=5)
            # XXX someday, pass `bad_commstrat' to `forget_comm_strategy()' so that we can forget only the bad one.
            # XXX if we kept a list of comm strategies instead of just one, then we could manage remembering comm strats in just one place.
            # Currently we basically have a list of 2 comm strats, one kept in comms handlers and one kept in MetaTrackerLib.
            # Anyway, here we forget the comms handler cs, but preserve the MTL contact info.  --Zooko 2001-05-15
            # self.forget_comm_strategy(counterparty_id, msgId, failure_reason=failure_reason)
            self.forget_comm_strategy(counterparty_id, firstmsgId=msgId, failure_reason=failure_reason)
            self.__query_for_counterparty_and_send(counterparty_id, msg, mojooffer, timeout=timeout, fast_fail_handler=outer_fast_fail_handler)

        debugprint(diagstr="sending directly [b] ...", v=15, vs="conversation")
        try:
            self._ch.send_msg(counterparty_id, msg, mojooffer, hint=HINT_EXPECT_RESPONSE | HINT_EXPECT_MORE_TRANSACTIONS, fast_fail_handler=fast_fail_handler, timeout=timeout) # hold open connections as many connections as you can
            debugprint(diagstr="sent.", v=15, vs="conversation")
        except CommsHandler.CannotSendError, le:
            debug.stderr.write("MojoTransaction.send_message_with_lookup(): called `_ch.send_msg(%s, msgId:%s)' and got CannotSendError: %s\n", args=(counterparty_id, msgId, le), v=10, vs="commstrats")
            fast_fail_handler(failure_reason=hr(le))
        debugprint(diagstr="sent", v=15)

    def __query_for_counterparty_and_send(self, counterparty_id, msg, mojooffer, timeout=MojoConstants.DEFAULT_TIMEOUT, fast_fail_handler=None):
        def debugprint(self=self, counterparty_id=counterparty_id, msg=msg, diagstr="", args=(), v=0, vs="conversation"):
            printstr="---> %s: " + diagstr + "\n"
            theseargs=[counterparty_id]

            theseargs.extend(list(args))
            debug.stderr.write(printstr, args=theseargs, v=v, vs=vs)

        if not MetaTrackerLib.find_root_meta_trackers(self._metamtm)[0][0]:
            debugprint(diagstr="WARNING: don't know any metatracker; I give up.", v=0, vs="metatracking")
            return

        def send_orig_msg(widget, outcome, failure_reason=None, counterparty_id=counterparty_id, msg=msg, self=self, debugprint=debugprint, timeout=timeout, fast_fail_handler=fast_fail_handler, mojooffer=mojooffer):
            """
            @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
            """
            assert idlib.is_sloppy_id(counterparty_id), "`counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

            counterparty_id = idlib.canonicalize(counterparty_id, "broker")

            msgId = idlib.string_to_id(msg)
            if failure_reason:
                debug.stderr.write("MojoTransaction: lookup failed; failure_reason: %s\n", args=(failure_reason,), v=1, vs="metatracking")
                fast_fail_handler(failure_reason=failure_reason)
                return

            if type(outcome) != types.DictType or outcome.get('result') != 'success':
                debugprint(diagstr="Couldn't get new comm strat.", v=5, vs="metatracking")
                fast_fail_handler(failure_reason="couldn't get contact info.  outcome: %s" % hr(outcome))
                return

            # The info better be in the MetaTrackerLib cache by now, otherwise we couldn't find it
            # (we do this to avoid having to parse & verify the hello response ourselves)
            counterparty_info = MetaTrackerLib.get_contact_info(self._metamtm, counterparty_id, fromcacheonly=true)

            if not counterparty_info:
                debug.stderr.write("WARNING: no valid contact info for %s found\n", args=(counterparty_id,), v=0, vs="metatracking")
                fast_fail_handler(failure_reason="invalid contact info from metatracker")
                return

            # XXX TODO sort the connection strategies in order of preference; (relay vs direct, etc..)
            if len(counterparty_info['connection strategies']) > 1:
                debug.stderr.write("WARNING: multiple connection strategies for %s, blindly using the first one.\n", args=(counterparty_id,), v=2, vs="metatracking")
            newcs = CommStrat.dict_to_strategy(counterparty_info['connection strategies'][0], mtm=self._metamtm, broker_id=counterparty_id)
            assert idlib.is_sloppy_id(newcs._broker_id), "newcs: %s" % hr(newcs)
            self._ch.use_comm_strategy(counterparty_id, newcs, orig_cs=counterparty_info['connection strategies'][0])
            try:
                debugprint(diagstr="sending directly [a] ...", v=5, vs="conversation")
                self._ch.send_msg(counterparty_id, msg, mojooffer, hint=HINT_EXPECT_RESPONSE | HINT_EXPECT_MORE_TRANSACTIONS, fast_fail_handler=fast_fail_handler, timeout=timeout) # hold open connections as many connections as you can
                debugprint(diagstr="sent.", v=15, vs="conversation")
            except CommsHandler.CannotSendError, le:
                debug.stderr.write("MojoTransaction.send_message_with_lookup(): called `_ch.send_msg(%s, msgId:%s)' after `use_comm_strategy(%s, strategy: %s)' and got CannotSendError: %s\n", args=(counterparty_id, idlib.make_id(msg, 'msg'), counterparty_id, counterparty_info['connection strategies'][0], le), v=10, vs="commstrats")
                fast_fail_handler(failure_reason="couldn't send either directly or via relay")

        debugprint(diagstr="looking up contact info...", v=15, vs="metatracking")
        MetaTrackerLib.nonblocking_get_contact_info(self._metamtm, counterparty_id, callback=send_orig_msg, timeout=timeout)
        debugprint(diagstr="done calling nonblocking_get_contact_info...", v=15, vs="metatracking")

    def stop_polling(self):
        if self.__polling == POLLING:
            self.__polling = GOING_TO_STOP_POLLING

    def use_relay_server(self):
        self.allow_hello_to_be_sent = false
        self.preferred_rs_id_list = confman.getlist(["PREFERRED_RELAY_SERVER_ID_LIST"])[:2]  # don't pay attention to more than 2
        # Looking these up individually merely makes sure that we have them
        # in our cache for the following get_relay_servers call.  We'll sort
        # these to the front later
        if len(self.preferred_rs_id_list) <= 0:
            MetaTrackerLib.get_relay_servers_and_call_back(self._metamtm, callback=self._gotrelayservers_use_relay_server)
            return
        preferred_id_info_map = {}  # maps rsid to contact info for that id as we find it
        for rsid in self.preferred_rs_id_list :
            MetaTrackerLib.nonblocking_get_contact_info(self._metamtm, rsid, callback=self.__gotpreferred_use_relay_server, notes={'rsid': rsid, 'preferred_id_info_map': preferred_id_info_map}, timeout=30)
    
    def __gotpreferred_use_relay_server(self, widget, outcome, failure_reason=None, notes=None):
        rsid = notes['rsid']
        preferred_id_info_map = notes['preferred_id_info_map']  # maps rsid to contact info for that id as we find it

        if failure_reason or (not outcome) or (not type(outcome) == types.DictType) or (len(outcome.get('connection strategies', [])) < 1):
            preferred_id_info_map[rsid] = None
        else:
            preferred_id_info_map[rsid] = outcome

        if len(preferred_id_info_map) < len(self.preferred_rs_id_list):
            debug.mojolog.write('still awaiting more preferred relayer id contact info\n', v=3, vs="relay")
            # we're waiting on more nonblocking_get_contact_info responses
            return
        debug.mojolog.write('done getting preferred relayer id contact info\n', v=3, vs="relay")

        # this will callback immediately without a metatracker query if any of our preferred relay
        # servers were successfully looked up and we at least know of a few others as their contact
        # info will be in our cache.
        MetaTrackerLib.get_relay_servers_and_call_back(self._metamtm, callback=self._gotrelayservers_use_relay_server, min=6)

    def _gotrelayservers_use_relay_server(self, relay_server_id_info_map):
        relay_server_list = relay_server_id_info_map.items()
        num_found_relay_servers = len(relay_server_list)
        self.calls_to_use_relay_server = self.calls_to_use_relay_server + 1

        def retry_list_relay_servers(callback, self=self, num_found_relay_servers=len(relay_server_list)):
            debug.mojolog.write("no acceptable relay servers found, trying to find more\n", v=2, vs="relay")
            # hmm, maybe our preferred ones changed to hiding behind relay to confuse us, try to more if possible (once)
            MetaTrackerLib.get_relay_servers_and_call_back(self._metamtm, callback=callback, doremotequery="forceremotequery")

        # reject any relay server that does not have a direct TCP
        # connection strategy (until we have time to debug why relay
        # chaining isn't working and add loop detection)
        for i in copy.copy(relay_server_list):
            has_tcp = 0
            for j in i[1]['connection strategies']:
                try:
                    if j['lowerstrategy']['comm strategy type'] == 'TCP':
                        has_tcp = 1
                        break
                except KeyError:
                    pass
            if not has_tcp:
                relay_server_list.remove(i)
                debug.stderr.write('rejected ' + hr(i[0]) + " because it didn't have a TCP strategy\n", v=2, vs="relay")
        for i in copy.copy(relay_server_list):
            if not self._relay_server_failer._is_acceptable(i[0]):
                relay_server_list.remove(i)
                debug.stderr.write('rejected ' + hr(i[0]) + " due to previous failures\n", v=1, vs="relay")

        if len(relay_server_list) == 0:
            retry_list_relay_servers(self._gotrelayservers_use_relay_server)
            return

        # see if any of our preferred ones are available
        relay_server = None
        relay_server_ids = map(lambda a: a[0], relay_server_list)
        for rsid in self.preferred_rs_id_list :
            try:
                idx = relay_server_ids.index(idlib.to_binary(rsid))
                debug.stderr.write("Using preferred relay server %s\n", args=(rsid,), v=1, vs="relay")
                relay_server, our_relay_server_info = relay_server_list[idx]
                break
            except ValueError:
                pass

        # no preferred one available?  pick one.
        if not relay_server :
            # sort the list of relay servers by preference
            relay_server_list = self.get_handicapper().sort_by_preference(relay_server_list, 'are there messages v2', {})

            if len(relay_server_list) == 0:
                retry_list_relay_servers(self._gotrelayservers_use_relay_server)
                return

            # pick a random relay server out of the perferred 2/3 of the ones available
            # (to avoid everyone using the best one and overloading it)
            relay_server, our_relay_server_info = whrandom.choice(relay_server_list[:(2*len(relay_server_list)/3)+1])
            ascii_relay_server_id = idlib.to_mojosixbit(relay_server)
            try:
                oldidx = self.preferred_rs_id_list.index(ascii_relay_server_id)
                del self.preferred_rs_id_list[oldidx]
            except ValueError:
                pass
            # put this relay server at the front of our preferred relayer id list so that we try
            # and use it first the next time we come up.
            self.preferred_rs_id_list.insert(0, ascii_relay_server_id)
            self.preferred_rs_id_list = self.preferred_rs_id_list[:2]  # don't keep more than 2 in the list
            confman.dict["PREFERRED_RELAY_SERVER_ID_LIST"] = SEXP.listToDict(self.preferred_rs_id_list)
            # save the config file to store our new preferred relay server id
            confman.save()

        self.__retried_list_relay_servers = false

        debug.stderr.write("now using relay server %s.\n", args=(relay_server,), v=0, vs="relay")
        self.set_relay_server(relay_server)
        self.allow_hello_to_be_sent = true

    def set_relay_server(self, relay_server):
        assert idlib.is_sloppy_id(relay_server)
        if not idlib.equal(self._relay_server_failer.get_preferred(), relay_server):
            self._relay_server_failer.get_preferred([relay_server])
            if self.initial_hello_has_been_sent:
                self.send_hello_to_meta_trackers()
        if self.__polling == GOING_TO_STOP_POLLING:
            self.__polling = POLLING
        if self.__polling == NOT_POLLING:
            self.__polling = POLLING
            self._most_recent_relay_poll_time = 0
            self._schedule_next_relay_server_poll()

    def _poll_relay_server_start(self, rs):
        if self._shuttingdownflag or confman.is_true_bool(('KLOOGE_NO_RELAY',)):
            return

        MetaTrackerLib.nonblocking_get_contact_info(self._metamtm, rs, callback=self._poll_relay_server_metainfo_callback, notes={'rs_id': rs})

    def _poll_relay_server_metainfo_callback(self, widget, outcome=None, failure_reason=None, notes=None):
        if self._shuttingdownflag:
            return

        rs_id = notes['rs_id']

        if (not outcome) or (not type(outcome) == types.DictType) or (len(outcome.get('connection strategies', [])) < 1):
            debug.mojolog.write("WARNING: bad or non-existent contact info for relay server, outcome: %s\n", args=(outcome,), v=4, vs="debug")
            self._process_relay_server_failure(rs_id, 'bad or non-existent contact info for relay server')
            self._schedule_next_relay_server_poll()
            return
        rs_info = outcome
        assert idlib.equal(MetaTrackerLib.metainfo_dict_to_id(rs_info), rs_id), "meta tracker returned contact info for a different relay server"
        msgbody = {'response version': 2, 'enable fast relay': 1}
        fixedprice = pricelib.relay_server_are_there_messages_price(rs_info, msgbody)
        if fixedprice is not None:
            # Remove this once we've grandfathered-out most of the fixed-price servers...  --Zooko 2001-06-01
            mojooffer=fixedprice
        else:
            mojooffer = BidStrategy.choose_bid("MO_ARE_THERE_MESSAGES", self, counterparty_id=rs_id)

        latency_start = time.time()

        def _are_there_messages_callback(widget, outcome=None, failure_reason=None, notes=None, self=self, latency_start=latency_start, rs_id=rs_id, rs_info=rs_info):
            if self._shuttingdownflag:
                return

            try:
                now = time.time()
                latency = now - latency_start
                def update_recent_latency_list(current_stat, latency=latency):
                    current_stat['last update'] = time.time()
                    current_stat['list'].append(latency)
                    current_stat['list'] = current_stat['list'][-30:]  # XXX make this constant configurable
                    return current_stat
                self._keeper.get_counterparty_object(rs_id).update_custom_stat('are there messages v2 latency', update_recent_latency_list, {'last update': now, 'list': []})
                if (not outcome) or failure_reason:
                    self._process_relay_server_failure(rs_id, failure_reason)
                    return
                if outcome.get('result') == 'no':
                    return  # no messages, how boring
                msgbody = {'messages': outcome['messages info'], 'bytes': outcome['total bytes']}
                fixedprice = pricelib.relay_server_get_message_price(rs_info, msgbody)
                if fixedprice is not None:
                    # Remove this once we've grandfathered-out most of the fixed-price servers...  --Zooko 2001-06-01
                    mojooffer=fixedprice
                else:
                    mojooffer = BidStrategy.choose_bid("MO_RETRIEVE_MESSAGES", self, counterparty_id=rs_id)

                self.initiate(rs_id, 'retrieve messages v2', msgbody, mojooffer, outcome_func=self._relay_retrieve_messages_callback)
            finally:
                # try and schedule another poll no matter what happens
                self._schedule_next_relay_server_poll()
            # end _are_there_messages_callback

        self._most_recent_relay_poll_time = time.time()
        self.initiate(rs_id, "are there messages v2", msgbody, mojooffer, outcome_func=_are_there_messages_callback, timeout=60)

    def _relay_retrieve_messages_callback(self, widget, outcome=None, failure_reason=None, notes=None):
        if self._shuttingdownflag:
            return

        if (not outcome) or failure_reason:
            self._process_relay_server_failure(widget.get_counterparty_id(), failure_reason)
            return
        if outcome['status'] != 'messages attached v2':
            raise Error, "couldn't get message which was supposedly there"
        # XXX check whether the messages sent back were the ones requested
        for i in outcome['messages']:
            try:
                # TODO tell _ch._inmsg_handler to verify that the message really came from i['sender id']?
                # I don't suppose we really care since the true sender is encoded inside the message. ?  --Zooko 2001-05-07
                # In fact, we could just optimize out the 'sender id' field entirely.  --Zooko 2001-05-07
                self._ch._inmsg_handler(i['message'], None, None)

            except mesgen.Error, e:
                debug.mojolog.write("WARNING: message retrieved from relay server caused mesgen.Error %s\n", args=(e,), vs="relay")
  
    def _process_relay_server_failure(self, rs_id, failure_reason):
        assert idlib.is_sloppy_id(rs_id), "rs_id must be an id"
        assert not idlib.equal(rs_id, self.get_id()), "rs_id must not be our own id"
        debug.mojolog.write("relay server %s failed; failure_reason: %s\n", args=(rs_id, failure_reason,), v=0, vs="relay")
        self._relay_server_failer.fail_came_in(rs_id)
        preferred_rs_id = self._relay_server_failer.get_preferred()
        debug.mojolog.write("preferred_rs_id: %s\n", args=(preferred_rs_id,), vs="relay")
        if preferred_rs_id is None:
            debug.mojolog.write("too many failures for %s, choosing a new relay server\n", args=(rs_id,), vs="relay")
            self.use_relay_server()
        elif idlib.equal(preferred_rs_id, rs_id):
            debug.mojolog.write("not enough failures, %s is still our preferred relay server\n", args=(rs_id,), vs="relay")

    def _schedule_next_relay_server_poll(self):
        if self._shuttingdownflag:
            return

        if (self.__polling == GOING_TO_STOP_POLLING) or (self.__polling == NOT_POLLING):
            self.__polling = NOT_POLLING
            return
        poll_period = min(600, max(1, float(confman.dict.get("RELAY_SERVER_POLL_PERIOD_v2", 10))))
        # always leave at least 5 seconds between a call to this function and the actual poll
        next_poll_delay = max(5, poll_period - (time.time() - self._most_recent_relay_poll_time))
        assert next_poll_delay >= 5 and next_poll_delay <= 600
        # schedule the next poll
        rs_id = self._relay_server_failer.get_preferred()
        if rs_id is not None:
            self._doQ.add_task(self._poll_relay_server_start, args=(rs_id,), delay=next_poll_delay)
        else:
            debug.mojolog.write("WARNING: no relay server set (perhaps we're in the middle of picking a new one)\n", vs="relay")
            self._doQ.add_task(self._schedule_next_relay_server_poll, delay=next_poll_delay)

    def __init_message_for_you_handler(self):
        self.add_handler_funcs({'message for you': (self._message_for_you_handler, self._message_for_you_pricer)})

    def _message_for_you_pricer(self, widget, message_type, message_body):
        # sending us an encapsulated messasge is free, we like to get our messages from our relay servers
        return 0

    def _message_for_you_handler(self, widget, message_body):
        debug.mojolog.write("counterparty %s fast-relayed us a %s byte message\n", args=(widget.get_counterparty_id(), len(message_body['message'])), v=3, vs="relay")
        if self.__in_message_for_you:
            debug.mojolog.write("WARNING: nested 'message for you' from %s not allowed, dropping.\n", args=(widget.get_counterparty_id(),), v=2, vs="relay")
            return NO_RESPONSE
        def process_message_for_us(self=self, encmsg=message_body['message']):
            self.__in_message_for_you = 1
            try:
                # XXX _ch._inmsg_handler does things on the DoQ which make this self.__in_message_for_you nesting avoidance flag useless.
                self._ch._inmsg_handler(encmsg, None, None)
            finally:
                self.__in_message_for_you = None
        self._doQ.add_task(process_message_for_us)
        return {'result': 'success'}


def make_mtm_for_testing(doQ, receivers={}, listenport=None, relay_server=None, pt=None):
    def return_one_pricer(widget, message_type, message):
        return 1

    hs={}
    for i in receivers.keys():
        hs[i]=(receivers[i], return_one_pricer)
    mtm=MojoTransactionManager(doQ=doQ, pt=pt, handler_funcs=hs, listenport=listenport, relay_server=relay_server)
    return mtm

def test_serialize_and_reserialize():
    doQ = DoQ.DoQ() # test
    try:
        mtm1=MojoTransactionManager(doQ)
        mtm2=MojoTransactionManager(doQ, serialized=mtm1.serialize())
        assert mtm1.get_id() == mtm2.get_id()
    finally:
        doQ.shutdown()

def test_retry_on_no_resp():
    doQ = DoQ.DoQ() # test
    try:
        doQ.suspend() # Don't want things happening multithreadedly while we set up...
        
        counterholder=[0]
        def handle_ping(widget, msgbody, counterholder=counterholder):
            debug.stderr.write("MojoTransaction.test_retry(): received: %s\n", args=(msgbody,), v=0, vs="testem")
            if counterholder[0] < 1:
                debug.stderr.write("MojoTransaction.test_retry(): ignoring it to see if it will retry.\n", v=0, vs="testem")
                response = MojoConstants.ASYNC_RESPONSE
            else:
                debug.stderr.write("MojoTransaction.test_retry(): answering\n", v=0, vs="testem")
                response = "PONG!"
            counterholder[0] = counterholder[0] + 1
            return response

        c1=make_mtm_for_testing(doQ)
        c2=make_mtm_for_testing(doQ, receivers={'test': handle_ping})

        # Tell c1 how to talk to c2.
        c1.use_comm_strategy(c2.get_id(), c2.get_comm_strategy())

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper()

        # Initiate a transaction with retrying.
        c1.initiate(c2.get_id(), 'test', "ping!", 1, outcome_func=_wcb1, timeout=20.0, retry=true)

        doQ.restart() # Okay now the engine can run...

        doneflag1.wait(16.0)

        # Okay by now we should have gotten a PONG back.
        assert (reskw1.get('outcome') == 'PONG!') or ((len(res1) > 1) and (res1[1] == 'PONG!')), "res1: %s, reskw1: %s\n" % (hr(res1), hr(reskw1))
    finally:
        doQ.shutdown()


def test_retry_on_rejected_conn():
    doQ = DoQ.DoQ() # test
    try:
        doQ.suspend() # Don't want things happening multithreadedly while we set up...
        
        debug.stderr.write("hi there.  I'm done calling `DoQ.suspend()'\n", v=0, vs="debug")
        
        counterholder=[0]
        def handle_ping(widget, msgbody, counterholder=counterholder):
            debug.stderr.write("MojoTransaction.test_retry(): received: %s\n", args=(msgbody,), v=0, vs="testem")
            if counterholder[0] < 1:
                debug.stderr.write("MojoTransaction.test_retry(): ignoring it to see if it will retry.\n", v=0, vs="testem")
                response = MojoConstants.ASYNC_RESPONSE
            else:
                debug.stderr.write("MojoTransaction.test_retry(): answering\n", v=0, vs="testem")
                response = "PONG!"
            counterholder[0] = counterholder[0] + 1
            return response

        c1=make_mtm_for_testing(doQ)
        c2=make_mtm_for_testing(doQ, receivers={'test': handle_ping})

        # Tell c1 how to talk to c2.
        c1.use_comm_strategy(c2.get_id(), c2.get_comm_strategy())

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper()

        # ... but have c2 stop accepting connections
        c2.stop_listening()

        # Initiate a transaction with retrying.
        c1.initiate(c2.get_id(), 'test', "ping!", 1, outcome_func=_wcb1, timeout=20.0, retry=true)

        doQ.restart() # Okay now the engine can run...

        WAITBEFORELISTEN=1.0
        TOTALWAIT=16.0
        WAITAFTERLISTEN=(TOTALWAIT - WAITBEFORELISTEN)

        time.sleep(WAITBEFORELISTEN)

        # we should _not_ have gotten a PONG back yet -- c2 isn't listening.
        assert (reskw1.get('outcome') != 'PONG!') or ((len(res1) < 2) or (res1[1] != 'PONG!')), "res1: %s, reskw1: %s\n" % (hr(res1), hr(reskw1))

        doneflag1.wait(WAITAFTERLISTEN)

        # Okay by now we should have gotten a PONG back.
        assert (reskw1.get('outcome') == 'PONG!') or ((len(res1) > 1) and (res1[1] == 'PONG!')), "res1: %s, reskw1: %s\n" % (hr(res1), hr(reskw1))
    finally:
        doQ.shutdown()


global testidfile
testidfile=None

# Generic stuff
NAME_OF_THIS_MODULE="MojoTransaction"

mojo_test_flag=1

def run():
    RunTests.runTests(NAME_OF_THIS_MODULE)

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()


