#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# This is a simple class derived from saxlib.HandlerBase that
# implements the common parser-interface routines that we want.
# (which are derived from a saxlib demo program)
#
# This module is safe to import using 'from mojoxml import *' if you
# just want this library and the necessary xml.sax modules.
#
__cvsid = '$Id: MojoXML.py,v 1.7 2001/06/08 21:17:36 zooko Exp $'

# standard modules
import re
import string
import types
import base64
from cStringIO import StringIO

# Available in PyXML at http://www.python.org/sigs/xml-sig/
from xml.sax import saxlib, saxexts
import xml
import xml.sax
import xml.sax.drivers.drv_xmlproc
import xml.parsers
import xml.parsers.xmlproc

# our modules
import idlib
import mojoutil
import SEXP
import debug

# generic XML parsing error (error baseclass for this module)
class XMLParseError(StandardError) : pass

# raised when an unknown element is encountered
class XMLBadElementError(XMLParseError) : pass


def normalize_whitespace(text):
    "Remove redundant whitespace from a string"
    return string.join( string.split(text), ' ')

__valid_xml_keystr_element_name_re = re.compile('^[A-Za-z:][A-Za-z0-9 .:-]*$')
def keystr_to_xml_element(keystr) :
    """
    Convert an mojodict key string to a string usable as an xml element.
    Spaces are replaced with underscores (_).  Underscores may not
    appear in the original keystr.
    """
    keystr = 'X-' + keystr 
    if not __valid_xml_keystr_element_name_re.match(keystr) :
        raise XMLBadElementError, "keystr contains invalid xml name characters"
    return string.replace(keystr, ' ', '_')

def xml_element_to_keystr(elementname) :
    """Convert an xml element name into an mojo dict key string"""
    if elementname[:2] != 'X-' :
        raise XMLBadElementError, "keystr element must start with 'X-'"
    # convert _ back to ' ' and remove the leading 'X-'
    return string.replace(elementname, '_', ' ')[2:]

# matches characters which are not legal within text in XML [excluding
# the \000 character which the re module doesn't like :( ]
__non_xml_text_char_re = re.compile('[\001-\010\013\014\016-\037]')

__all_digits_re = re.compile('^[0-9]*$')
def __mojodict_to_mojoxml_list(mojodict) :
    """
    Returns an ordered list of strings to be concatenated to become the xml.
    """
    olist = []

    def encode_value_into_olist(value, olist=olist) :
        """insert the value into olist, properly encoded"""
        # XXX we could be smarter about encoding.  base64
        # isn't needed in many cases where a couple '&#0;'
        # style encodings would do.  -greg 8/jun/2000
        if __non_xml_text_char_re.search(value) or ('\000' in value) :
            olist.append(' encoding="RFC1421">')  # use RFC1421 base64 encoding
            olist.append(base64.encodestring(value))
        else :
            olist.append('>')
            if '&' in value or '>' in value or '<' in value :
                value = string.replace(value, '&', '&amp;')
                value = string.replace(value, '<', '&lt;')
                value = string.replace(value, '>', '&gt;')
            olist.append(value)

    if type(mojodict) == types.IntType or type(mojodict) == types.LongType:
        olist.append('<int>')
        olist.append(str(mojodict))
        olist.append('</int>')
        return olist
    if mojodict is None:
        olist.append('<null/>')
        return olist
    if type(mojodict) == types.ListType or type(mojodict) == types.TupleType:        
        olist.append('<list>')
        for value in mojodict:
            if type(value) is types.StringType :
                if len(value) > 0 :
                    olist.append('<li')  # don't close, encode_value does it for us after adding an encoding type
                    encode_value_into_olist(value)
                    olist.append('</li>')
                else :
                    olist.append('<li/>')
            else:
                olist.append('<li>')
                olist.extend(__mojodict_to_mojoxml_list(value))
                olist.append('</li>')
        olist.append('</list>')
        return olist

    assert type(mojodict) == types.DictType
    olist.append('<dict>')
    for key, value in mojodict.items() :
        assert type(key) == types.StringType, "mojodict keys must be strings"

        keyelement = keystr_to_xml_element(key)
        olist.append('<')
        olist.append(keyelement)

        if type(value) is types.StringType:
            if len(value) > 0 :
                encode_value_into_olist(value)
                # close the element
                olist.append('</')
                olist.append(keyelement)
                olist.append('>')
            else :
                # close the element using shorthand when there is no data
                olist.append('/>')
        else:
            # recursively convert sub mojo-dicts
            olist.append('>')
            olist.extend(__mojodict_to_mojoxml_list(value))
            # close the element
            olist.append('</')
            olist.append(keyelement)
            olist.append('>')
    olist.append('</dict>')
    return olist


########################################################################
def mojodict_to_mojomsg_xml_str(mojodict) :
    """
    Convert an Mojo Dict as returned by SEXP.stringToDict or as
    passed to SEXP.dictToString into a MojoMsg XML string.
    """
    strlist = __mojodict_to_mojoxml_list(mojodict)
    strlist.insert(0, '<?xml version="1.0"?><MojoMsg>')
    strlist.append('</MojoMsg>')
    return string.join(strlist, '')


########################################################################
def mojodict_to_mojomsg_xml_file(mojodict, ofile) :
    """
    Convert an Mojo Dict as returned by SEXP.stringToDict or as
    passed to SEXP.dictToString into MojoMsg XML written to ofile.
    """
    # XXX TODO this function needs work, it was written for old alpha style SEXP messages
    strlist = __mojodict_to_mojoxml_list(mojodict)
    strlist.insert(0, '<?xml version="1.0"?><MojoMsg>')
    strlist.append('</MojoMsg>')
    for s in strlist :
        ofile.write(s)


########################################################################
def mojomsg_xml_file_to_mojodict(xml_file) :
    """Read MojoMsg XML from xml_file and return the Mojo Dict"""
    # XXX TODO this function needs work, it was written for old alpha style SEXP messages
    mojomsg_dh = MojoMsgDocHandler()
    mojomsgparser = create_mojo_xml_parser(mojomsg_dh)
    try:
        mojomsgparser.parseFile(xml_file)
    finally:
        mojomsgparser.close()
    
    return mojomsg_dh.MojoDict


########################################################################
def mojomsg_xml_str_to_mojodict(xml_str) :
    """Parse MojoMsg XML from xml_str and return the MojoDict"""
    return mojomsg_xml_file_to_mojodict(StringIO(xml_str))


########################################################################
def create_mojo_xml_parser(dochandler) :
    """
    Given a dochandler object (presumable MojoNationXMLDocHandler
    derived) this function returns an XML parser setup to use it.
    Reminder: ALWAYS call the close method of a parser (don't rely on garbage collection)
    """
    # XXX this parser is specified explicitly because it gives much
    # more useful error messages than the default (pyexpat).  pyexpat
    # was also causing coredumps after dealing with enough incorrect
    # XML (using pyxml 0.5.3).  This python based one can't coredump.

    # XXX we don't do this:
    #parser = saxexts.make_parser("xml.sax.drivers.drv_xmlproc")
    # XXX because it doesn't work well for building windows packages
    # using gordons installer (the package can't find the parser modules)

    parser = xml.sax.drivers.drv_xmlproc.create_parser()
    parser.setDocumentHandler(dochandler)
    parser.setErrorHandler(dochandler)
    return parser



########################################################################
class MojoMsgDocHandler(saxlib.HandlerBase) :
    """Used in the conversion of MojoMsg XML back into our python MojoDicts."""
    def __init__(self) :
        # used to collect character data within an element
        self.__chars = None
        self.__need_base64_decode = 0

        self.__in_MojoMsg = 0

        self.__current = None
        self.__currentstack = []
        self.__value = None

        self.__dict_key_names = []

        # after parsing, this contains our mojodict
        self.MojoDict = None

    # This is called by the parser when parsing errors occur
    def fatalError(self, exception) :
        # XXX log this error for debugging?
        raise XMLParseError, exception.msg
    error = fatalError
    warning = fatalError

    # This method is called by the XML parser on all non-xml-tag
    # character data.
    def characters(self, ch, start, length) :
        if self.__chars is not None :
            self.__chars.append(ch[start:start+length])

    # This is called by the parser on properly matched start elements
    def startElement(self, name, attrs) :
        #D#print '..... start name:', name
        if not self.__in_MojoMsg :
            if name == 'MojoMsg' :
                self.__in_MojoMsg = 1
            else:
                raise XMLParseError, 'missing MojoMsg element'
        else:
            if name == 'list' or name == 'dict' :
                if name == 'list' :
                    newcurrent = []
                else :
                    newcurrent = {}
                if self.__current is None :
                    self.__current = newcurrent
                elif type(self.__current) == types.ListType :
                    self.__current.append(newcurrent)
                else :
                    self.__current[self.__dict_key_names[len(self.__dict_key_names)-1]] = newcurrent
                self.__currentstack.append(self.__current)
                self.__current = newcurrent
                # stop collecting character data
                self.__chars = None
            elif name == 'li' :
                if type(self.__current) != types.ListType :
                    raise XMLParseError, 'li element can only occur within a list'
                # start collecting character data
                self.__chars = []
            else :
                #D#print '... pushing:', name
                self.__dict_key_names.append(xml_element_to_keystr(name))
                # start collecting character data
                self.__chars = []

            encoding = attrs.get('encoding', None)
            if encoding == 'RFC1421' :
                self.__need_base64_decode = 1
            elif encoding is None or encoding == 'binary' :
                self.__need_base64_decode = 0
            else:
                raise XMLParseError, 'unknown encoding attribute: '+encoding

    # This is called by the parser on properly matched end elements
    def endElement(self, name) :
        #D#print '..... end name:', name
        if name == 'dict' :
            self.__value = self.__current
            self.__current = self.__currentstack.pop()
        elif name == 'list' :
            self.__value = SEXP.listToDict(self.__current)
            self.__current = self.__currentstack.pop()
        elif name == 'MojoMsg' :
            self.__in_MojoMsg = 0
            self.MojoDict = self.__value
            #D#print 'self.__value:', self.__value
            #D#print 'self.__current:', self.__current
            #D#print 'self.__currentstack:', self.__currentstack
        else :
            # if we recorded character data, finish processing it
            if self.__chars is not None :
                self.__value = string.join(self.__chars, '')
                if self.__need_base64_decode :
                    self.__value = base64.decodestring(self.__value)

            # insert the value as appropriate
            if name == 'li' :
                assert type(self.__current) == types.ListType, 'xml parser out of sync!'
                self.__current.append(self.__value)
            else:
                assert type(self.__current) == types.DictType, 'xml parser out of sync!'
                #D#print '... popping:', self.__dict_key_names
                self.__current[self.__dict_key_names.pop()] = self.__value
            
            self.__value = None
            self.__chars = None


def test_mojomsg() :
    """Test the MojoMsgDocHandler class"""
    spam = {
        'msgtype': "feed me",
        'spamlist' : SEXP.listToDict(['green', 'eggs', 'and', 'ham', 'is', 'not', 'as', 'good', 'as', 'spam']),
        'msgbody': {
           'breakfast': 'poached spam',
           'lunch': 'spam beans and spam',
           'dinner': 'spam spam spam eggs and spam',
        },
    }

    mojomsg_xml = mojodict_to_mojomsg_xml_str(spam)
    twiceparsedspam = mojomsg_xml_str_to_mojodict(mojomsg_xml)
    assert spam == twiceparsedspam



#########################################################################
# All that's needed to create your own simple XML parser from this is
# an __init__ method to define any local data that you need as well as
# functioned named start_FOO and end_FOO where FOOs are the names of
# XML tags this document handler is expecting.  start_FOO methods
# take an "args" argument which is a dictionary of arguments to that
# XML tag.
#########################################################################
class MojoNationXMLDocHandler(saxlib.HandlerBase) :
    ####### parser interface: #######
    ####################################################################
    # This is called by the parser when parsing errors occur
    ####################################################################
    def fatalError(self, exception) :
        raise XMLParseError, exception
    error = fatalError
    warning = fatalError

    ####################################################################
    # This is called by the parser on properly matched start elements
    ####################################################################
    def startElement(self, name, attrs) :
        # raise an exception if we encounter an unknown element
        if hasattr(self, 'start_'+name) :
            method = getattr(self, 'start_'+name)
            method(attrs)
        else:
            raise XMLBadElementError, "unknown element "+name

    ####################################################################
    # This is called by the parser on properly matched end elements
    ####################################################################
    def endElement(self, name) :
        # ignore unsupported end elements
        if hasattr(self, 'end_'+name) :
            method = getattr(self, 'end_'+name)
            method()
    ####### end parser interface #######



mojo_test_flag = 1

#### this runs if you import this module by itself
def run():
    import RunTests
    RunTests.runTests(["MojoXML"])
    ##print mojodict_to_mojoxml_str({'spam': 'eggs', 'Food like substance': {'001': 'Spam & Eggs were', '002': 'and', '003': 'Binary \007 beeping \007 spam'}})

if __name__ == '__main__':
    run()

# EOF

