#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

# standard modules
import time
import types
import operator
import os

# our modules
import Purse
import canon
import debug
import fileutil
from human_readable import hr
import tokendists
import randsource
import mojosixbit
import mojoutil
import MojoToken
import SEXP
import IdTrackerEGTP
import MojoTransaction
import id_lock
import MojoConstants
from MojoConstants import true, false
import confutils
import DoQ


debugwrite = debug.mojolog.write


BASIC_DENOM = MojoToken.MAP_DENOM_TO_KEYID.keys()[0]



def _make_prototokens(amount):
    """
    @returns a list of prototokens in Token instance form
    """

    # Choose what kind (what denoms) of tokens to withdraw.
    # !! XX TODO Zooko: even smarter (/even more complicated) dist chooser which knows what dist you already have in your purse...  --Zooko 2000/06/06
    dist = tokendists.default_choose_token_dist(amount)

    listoprototokens = []

    for (denom, num) in dist.items():
        for x in range(num):
            tok = MojoToken.Token(randsource.get(20), MojoToken.MAP_KEYID_TO_PUBLIC_KEY[MojoToken.MAP_DENOM_TO_KEYID[denom]])
            assert tok.get_id_for_signing(), "Token must be fully formed now.  tok: %s" % hr(tok.get_id_for_signing())

            listoprototokens.append(tok)
            # !! XX TODO Zooko: some day we'll want to have more control over random seeds...  --Zooko 2000/06/07
    return listoprototokens


class Paytool:
    def __init__(self, tsmtm):
        """
        @param tsmtm the MTM to use when talking to the token server
        """
        self._tsmtm = tsmtm
        self._omtm = None
        self._purse = Purse.DbPurse(dbparentdir=os.path.expandvars(confutils.confman.dict["PATH"]["MOJOSTASH_DIR"]))
    
    def __del__(self):
        if hasattr(self, '_purse'):
            self._purse.shutdown()

    def shutdown(self):
        if self._purse:
            self._purse.shutdown()
            self._purse = None

    def _register_omtm(self, omtm):
        """
        @param omtm the MTM to use when talking to the rest of the world

        Registers itself with the MojoTransactionManager `omtm' as being the handler for
        incoming payments.  If any payments arrive then it tries to deposit them with the
        TokenServer, and if the deposit is successful then it updates the appropriate
        counterparty object to indicate that that person has paid us.  So currently there is no
        way for you to get notified of incoming payments.  We can add such a feature if it is
        needed.
        """
        self._omtm = omtm
   
        def free_pricer(widget, incoming_type, incoming_body):
            return 0

        my_funcs = {
            'payment': (self._handle_payment, free_pricer)
            }
        self._omtm.add_handler_funcs(my_funcs)

    def get_amount_in_purse(self):
        return self._purse.get_aggregate_amount()

    def withdraw(self, amount, outcome_func = None, timeout=(10000L*MojoConstants.DEFAULT_TIMEOUT), retry=true):
        """
        outcome_func will be called with two arguments: widget, outcome
            `outcome' will be one of these three tuples:
                ("success", amountadded)
                ("failure withdrawal response", response from token server)
                ("bad withdrawal response", response from token server)
                #### XXXX !!!! add time-out failure.  in fact, change this whole interface by adding failure_reason optional parameter, mirroring MojoTransaction
                # Actually "failure_reason" is reserved for transaction-level failures like timeout, no connection, stuff like that.  If the token server rejects you, it has to be put into "outcome".  Sorry about that.    --Zooko 2000-08-28
        """
        lopts = _make_prototokens(amount)

        wmsgbody = []
        for tok in lopts:
            wmsgbody.append(tok.get_id_for_signing())

        def handle_response(widget, outcome, failure_reason=None, outcome_func=outcome_func, self=self, lopts=lopts):
            if failure_reason:
                debug.stderr.write('Warning:  Withdraw failed due to transaction failure, this might result in the loss of a token.  We should handle this better some day.\n', v=9, vs="payment")
                return mojoutil.callback_wrapper(func=outcome_func, kwargs={'widget': widget, 'outcome': outcome, 'failure_reason': failure_reason}, defaultreturnval=MojoConstants.NO_RESPONSE)

            if outcome.get('result') != "success":
                # Hm.  Notify the app that this failed.
                debug.stderr.write('Withdraw failed due to rejection from TokenServer.  This is normal, in order to withdraw all tokens.\n', v=9, vs="payment")
                return mojoutil.callback_wrapper(func=outcome_func, kwargs={'widget': widget, 'outcome': ("failure withdrawal response", outcome)}, defaultreturnval=MojoConstants.NO_RESPONSE)

            return self._verify_and_store_withdrawn_tokens(widget, outcome['signed token ids'], lopts, outcome_func=outcome_func)

        self._tsmtm.initiate(MojoToken.HARDCODED_TOKEN_SERVER_COMMS_ID, conversationtype='withdraw', firstmsgbody=wmsgbody, mojooffer=0, outcome_func=handle_response, timeout=timeout, retry=retry)

    def pay(self, recipient_id, amount, outcome_func = None, timeout=(10000*MojoConstants.DEFAULT_TIMEOUT), retry=true):
        """
        @param recipient_id the id of the recipient (use the "omtm" id, which that person uses
            to talk to others, instead of the "tsmtm" id, which that person uses to talk to
            the TokenServer.  Actually you do not know their tsmtm id unless they are you
            yourself. :-))
        @param retry `true' if and only if you want to resend the payment message in the case that
            you don't hear an acknowledgement;  This is the right thing to do.  Leave it set
            to `true'.  If you set it to `false', and there is a comms failure, you _don't_
            get your token back, but the other person doesn't get your token either, which is
            a problem.  If you set it to `true' then you lose your token, but you keep
            resending it over and over until the other person gets it, which is the goal of
            making a payment.

        @throws Purse.NotEnoughMojoError if there isn't enough Mojo in this paytool
        @throws Purse.NoExactChangeError if there isn't correct change in this paytool

        @precondition `amount' must be an integer.: (type(amount) == types.IntType) or (type(amount) == types.LongType): "amount: %s" % hr(amount)
        @precondition `amount' must be positive.: amount > 0: "amount: %s" % hr(amount)
        """
        assert (type(amount) == types.IntType) or (type(amount) == types.LongType), "precondition: `amount' must be an integer." + " -- " + "amount: %s" % hr(amount)

        if recipient_id == self._omtm.get_id():
            debug.stderr.write("WARNING: depositing tokens into the TokenServer / paying yourself is a deprecated function that doesn't actually do anything.\n", v=2, vs="debug")

        setoftoks = self._purse.get_tokens(amount)

        listof_tok_info_for_paying = []
        for tok in setoftoks.values():
            listof_tok_info_for_paying.append(tok.get_formatted_full_token())

        pmsgbody = listof_tok_info_for_paying

        co = self._omtm._keeper.get_counterparty_object(recipient_id)

        #### !!! XXX TODO Zooko: make these next two lines robust against arbitrary interruption (which is to say, make it transactional but do it the easy way by taking advantage of its idempotence instead of the hard way by doing arbitrary transactions).  --Zooko 2000/06/19
        co.token_payment_went_out(amount)
        self._omtm.initiate(recipient_id, 'payment', pmsgbody, mojooffer=0, outcome_func=outcome_func, timeout=timeout, retry=retry) 

    def _verify_and_store_withdrawn_tokens(self, widget, listofsiggedtokid4s, lopts, outcome_func):
        amountadded = 0
        index = 0

        # Apply all of the sigs to our tokens (this will raise an Error if the sig isn't good).
        # Note that the order of tokens in `lopts' must match the order of tokens returned in the withdrawal response.
        for tok in lopts:
            sig = mojosixbit.a2b(listofsiggedtokid4s[index])

            if not canon.is_canonical(sig, tok.get_server_pub_key_len()):
                # Hm.  Notify the app that this failed.
                expl="bad withdrawal response: signature is not a modular value of the correct modulus"
                debug.stderr.write('WARNING:  Withdraw failed due to: %s.\n', args=(expl,), v=7, vs="WARNING")
                return mojoutil.callback_wrapper(func=outcome_func, kwargs={'widget': widget, 'outcome': (expl, outcome)}, defaultreturnval=MojoConstants.NO_RESPONSE)

            try:
                tok.set_signature(sig)
            except MojoToken.BadSignatureError, le:
                expl="bad withdrawal response: bad signature: %s" % hr(le)
                debug.stderr.write('WARNING:  Withdraw failed due to: %s.\n', args=(expl,), v=7, vs="WARNING")
                return mojoutil.callback_wrapper(func=outcome_func, kwargs={'widget': widget, 'outcome': (expl, outcome)}, defaultreturnval=MojoConstants.NO_RESPONSE)

            index = index + 1
            self._purse.add_token(tok)
            amountadded = amountadded + MojoToken.MAP_KEYID_TO_DENOM[tok.get_server_public_key_id()]

        # Notify the app that this succeeded.
        return mojoutil.callback_wrapper(func=outcome_func, kwargs={'widget': widget, 'outcome': ("success", amountadded)}, defaultreturnval=MojoConstants.NO_RESPONSE)

    def _handle_payment(self, widget, msgbody):
        """
        The (NEW! IMPROVED!) way to handle a payment is to extend some credit temporarily to the
        counterparty while depositing the token at the token server.  When the deposit response
        arrives you remove the temporarily extended credit and mark down that she paid you in
        real tokens.

        Now, what happens if someone pays only in double-spent (pre-spent) tokens, so that the
        deposit response always says "failed"?  You don't want to continually extend credit to
        this person, give them service while you are awaiting the deposit response, and then
        revoke their credit, because they can just do the same thing again and still get their
        service.  So what we do is, if they already owe us more than their credit limit says
        they are allowed (a condition that they can get into only by giving us a bad coin and
        then spending on their temporary credit until it is revoked), then we fall back to the
        old 'make them wait until the token is proved to be good' behavior.
        """
        debugwrite("Paytool._handle_payment();  msgbody: %s\n", args=(msgbody,), v=7, vs="payment")
        # Check the signatures by the token server on these tokens.
        try:
            MojoToken.verify_token_sigs(msgbody)
        except MojoToken.BadSignatureError, le:
            # Whoa.  This is a rotten payment.
            debugwrite("Paytool._handle_payment(): bad signature;  le: %s, msgbody: %s\n", args=(le, msgbody), v=2, vs="payment")
            return {'result': "failure", 'reason': "bad signature"}

        co = self._omtm._keeper.get_counterparty_object(widget.get_counterparty_id())
        amount = MojoToken.tally_tokens(msgbody)

        # If she already owes _more_ than we're willing to front...
        ioweher = co.get_amount_owe()
        sheowesme = 0L - ioweher
        amwillfront = co.get_amount_will_front()
        if sheowesme > amwillfront:
            debugwrite("Paytool._handle_payment();  slow payment!  sheowesme: %s, amwillfront: %s\n", args=(sheowesme, amwillfront,), v=7, vs="payment")
            # The slow deposit protocol -- make her wait until we are sure her token is valid.
            def outcome_func(widget, outcome, failure_reason=None, payment_msgId=widget._firstmsgId, amount=amount, co=co, self=self):
                debugwrite("Paytool._handle_payment() -- slowp outcome_func();  outcome: %s, failure_reason: %s\n", args=(outcome, failure_reason,), v=7, vs="payment")
                if failure_reason:
                    # The conversation with the token server failed.
                    # For now, we consider this to be a failed payment, then.
                    self._omtm.respond_with(payment_msgId, {'result': "failure", 'reason': "failure to contact token server"})
                    debug.stderr.write("Paytool._handle_deposit_response(): WARNING: transaction failure in attempt to deposit.\n", v=0, vs="payment")
                    return 

                if outcome.get('result') == "success":
                    # We got MOJO from them!  Whoo-hoo!
                    co.token_payment_came_in(amount)

                    # Now let's go ahead and withdraw this amount of tokens into our local stash.
                    self.withdraw(amount)

                self._omtm.respond_with(payment_msgId, outcome)

            self._tsmtm.initiate(MojoToken.HARDCODED_TOKEN_SERVER_COMMS_ID, conversationtype='deposit', firstmsgbody=msgbody, mojooffer=0, outcome_func=outcome_func, retry=true, timeout=(10000*MojoConstants.DEFAULT_TIMEOUT))
            return MojoTransaction.ASYNC_RESPONSE # -1 # `ASYNC_RESPONSE' means I'll provide the response later. (by calling `respond_with()')
        else:
            debugwrite("Paytool._handle_payment();  fast payment!  sheowesme: %s, amwillfront: %s\n", args=(sheowesme, amwillfront,), v=7, vs="payment")
            # The fast deposit protocol -- give her a temporary credit extension while we check to see if her token is valid.
            co.raise_temp_awf(amount)

            def outcome_func(widget, outcome, failure_reason=None, co=co, amount=amount, self=self):
                debugwrite("Paytool._handle_payment() -- fastp outcome_func();  outcome: %s, failure_reason: %s\n", args=(outcome, failure_reason,), v=7, vs="payment")
                co.lower_temp_awf(amount)

                if failure_reason:
                    # The conversation with the token server failed.
                    # For now, we consider this to be a failed payment, then.
                    debug.stderr.write("Paytool._handle_deposit_response(): WARNING: transaction failure in attempt to deposit.\n", v=0, vs="payment")
                    return 

                if outcome.get('result') == "success":
                    # We got MOJO from them!  Whoo-hoo!
                    co.token_payment_came_in(amount)

                    # Now let's go ahead and withdraw this amount of tokens into our local stash.
                    self.withdraw(amount)

            self._tsmtm.initiate(MojoToken.HARDCODED_TOKEN_SERVER_COMMS_ID, conversationtype='deposit', firstmsgbody=msgbody, mojooffer=0, outcome_func=outcome_func, retry=true, timeout=(10000*MojoConstants.DEFAULT_TIMEOUT))

            return {'result': "success", 'disposition': "awaiting deposit response"}


mojo_test_flag = 1

def do_all_tests():
    import RunTests
    RunTests.runTests(["Paytool"])


#### this runs if you import this module by itself
if __name__ == '__main__':
    # tsPoe = TokenServer()
    do_all_tests()

