#!/usr/bin/env python

import time
import types
import threading
from sha import sha


import RunTests
import debug
import idlib
import mojosixbit
from human_readable import hr

class PopularityMeter:
    def __init__(self, interval_to_consider_popular, min_time_between_reports, poll_interval, publication_function, doQ):
        assert type(publication_function) is types.FunctionType or type(publication_function) is types.MethodType
        self.interval_to_consider_popular = interval_to_consider_popular
        self.min_time_between_reports = min_time_between_reports
        self.poll_interval = poll_interval
        self.publication_function = publication_function
        self._doQ = doQ
        # maps blob id to {'status' : 'requested', 'time' : time, 'counterparty' : counterparty_id}
        # or {'status' : 'published', 'time' : time reported publication}
        self.statuses = {}
        self.lock = threading.Lock()
        self._polling = 1
        # list of blob ids
        self._to_publish = []

##        t = threading.Thread(target = self._poll)
##        t.setDaemon(1)
##        t.start()

        self._poll_doqish()

    def stop_polling(self):
        self._polling = 0

    def _poll_doqish(self):
        # copied from `_poll()' and altered to be a doq pseudothread instead of a Python thread.  --Zooko 2000-10-14
        to_publish = []
        for blob_id in self.statuses.keys():
            self._check_expire(blob_id)
        to_publish = self._to_publish
        self._to_publish = []
        if len(to_publish) > 0:
            debug.stderr.write("PopularityMeter._poll(): calling publication_function().\n")
            self._doQ.add_task(self.publication_function, args=(to_publish,))

        debug.mojolog.write("YYY PopularityMeter._poll_doqish(), poll_interval: %s\n", args=(self.poll_interval,), v=7, vs="test")

        # Now reregister on the DoQ for later iteration.
        if self._polling:
            self._doQ.add_task(self._poll_doqish, delay=self.poll_interval)

    def _poll(self):
        while self._polling:
            to_publish = []
            time.sleep(self.poll_interval)
            self.lock.acquire()
            try:
                for blob_id in self.statuses.keys():
                    self._check_expire(blob_id)
                to_publish = self._to_publish
                self._to_publish = []
            finally:
                self.lock.release()
            if len(to_publish) > 0:
                debug.stderr.write("PopularityMeter._poll(): calling publication_function().\n")
                self.publication_function(to_publish)

    def _check_expire(self, blob_id):
        status = self.statuses.get(blob_id)
        if status is None:
            return
        if status['status'] == 'requested':
            if time.time() - status['time'] > self.interval_to_consider_popular:
                del self.statuses[blob_id]
        elif status['status'] == 'published':
            if time.time() - status['time'] > self.min_time_between_reports:
                del self.statuses[blob_id]
        else:
            assert 0

    
    def _internal_report_access(self, blob_id, counterparty_id, blob_size):
        # copied from `report_access()' and made to use the DoQ.  --Zooko 2000-10-14
        """
        @precondition `blob_id' must be an id.: idlib.is_sloppy_id(blob_id): "blob_id: %s" % hr(blob_id)
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        """
        assert idlib.is_sloppy_id(blob_id), "precondition: `blob_id' must be an id." + " -- " + "blob_id: %s" % hr(blob_id)
        assert idlib.is_sloppy_id(counterparty_id), "precondition: `counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

        blob_id = idlib.to_mojosixbit(blob_id)
        counterparty_id = idlib.canonicalize(counterparty_id, "broker")

        self._check_expire(blob_id)
        status = self.statuses.get(blob_id)
        if status is None:
            self.statuses[blob_id] = {'status' : 'requested', 'time' : time.time(), 'counterparty' : counterparty_id}
        elif status['status'] == 'requested':
            if status['counterparty'] != counterparty_id:
                self._to_publish.append( (blob_id, blob_size) )
                self.statuses[blob_id] = {'status' : 'published', 'time' : time.time()}
        else:
            assert status['status'] == 'published'

        debug.mojolog.write("YYY finished _internal_report_access\n", v=7, vs="test")

    def report_access_doqish(self, blob_id, counterparty_id, blob_size):
        # copied from `report_access()' and made to use the DoQ.  --Zooko 2000-10-14
        """
        @precondition `blob_id' must be an id.: idlib.is_sloppy_id(blob_id): "blob_id: %s" % hr(blob_id)
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        """
        assert idlib.is_sloppy_id(blob_id), "`blob_id' must be an id." + " -- " + "blob_id: %s" % hr(blob_id)
        assert idlib.is_sloppy_id(counterparty_id), "`counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

        blob_id = idlib.sloppy_id_to_bare_mojosixbitencoded_id(blob_id)
        counterparty_id = idlib.canonicalize(counterparty_id, "broker")

        self._doQ.add_task(self._internal_report_access, args=(blob_id, counterparty_id, blob_size,))

    def report_access(self, blob_id, counterparty_id, blob_size):
        """
        @precondition `blob_id' must be an id.: idlib.is_sloppy_id(blob_id): "blob_id: %s" % hr(blob_id)
        @precondition `counterparty_id' must be an id.: idlib.is_sloppy_id(counterparty_id): "counterparty_id: %s" % hr(counterparty_id)
        """
        assert idlib.is_sloppy_id(blob_id), "`blob_id' must be an id." + " -- " + "blob_id: %s" % hr(blob_id)
        assert idlib.is_sloppy_id(counterparty_id), "`counterparty_id' must be an id." + " -- " + "counterparty_id: %s" % hr(counterparty_id)

        blob_id = idlib.sloppy_id_to_bare_mojosixbitencoded_id(blob_id)
        counterparty_id = idlib.canonicalize(counterparty_id, "broker")

        self.lock.acquire()
        try:
            self._internal_report_access(blob_id, counterparty_id, blob_size)
        finally:
            self.lock.release()

mojo_test_flag = 1

def test_does_not_publish_on_single():
    id1 = mojosixbit.b2a(sha('first id').digest())
    c1 = mojosixbit.b2a(sha('first counterparty').digest())
    l = []
    def pub_func(blob_list, l = l):
        l.append(blob_list)
    pm = PopularityMeter(interval_to_consider_popular = 5, min_time_between_reports = 5, poll_interval = 1, publication_function = pub_func)
    pm.report_access(id1, c1, 2)
    time.sleep(2)
    assert l == []
    pm.stop_polling()

def test_does_not_pubish_on_multiple_access_from_same_counterparty():
    id1 = mojosixbit.b2a(sha('first id').digest())
    c1 = mojosixbit.b2a(sha('first counterparty').digest())
    c2 = mojosixbit.b2a(sha('second counterparty').digest())
    l = []
    def pub_func(blob_list, l = l):
        l.append(blob_list)
    pm = PopularityMeter(interval_to_consider_popular = 5, min_time_between_reports = 5, poll_interval = 1, publication_function = pub_func)
    pm.report_access(id1, c1, 2)
    pm.report_access(id1, c1, 2)
    time.sleep(2)
    assert l == []
    pm.stop_polling()

def test_publishes_on_two():
    id1 = mojosixbit.b2a(sha('first id').digest())
    c1 = mojosixbit.b2a(sha('first counterparty').digest())
    c2 = mojosixbit.b2a(sha('second counterparty').digest())
    l = []
    def pub_func(blob_list, l = l):
        l.append(blob_list)
    pm = PopularityMeter(interval_to_consider_popular = 5, min_time_between_reports = 5, poll_interval = 1, publication_function = pub_func)
    pm.report_access(id1, c1, 2)
    pm.report_access(id1, c2, 2)
    time.sleep(2)
    assert (l == [[(id1, 2)]]), l
    pm.stop_polling()

def test_expires_if_too_long_between_accesses():
    id1 = mojosixbit.b2a(sha('first id').digest())
    c1 = mojosixbit.b2a(sha('first counterparty').digest())
    c2 = mojosixbit.b2a(sha('second counterparty').digest())
    l = []
    def pub_func(blob_list, l = l):
        l.append(blob_list)
    pm = PopularityMeter(interval_to_consider_popular = 1, min_time_between_reports = 5, poll_interval = .5, publication_function = pub_func)
    pm.report_access(id1, c1, 2)
    time.sleep(2)
    pm.report_access(id1, c2, 2)
    time.sleep(2)
    assert l == []
    pm.stop_polling()

def test_does_not_publish_second_time():
    id1 = mojosixbit.b2a(sha('first id').digest())
    c1 = mojosixbit.b2a(sha('first counterparty').digest())
    c2 = mojosixbit.b2a(sha('second counterparty').digest())
    l = []
    def pub_func(blob_list, l = l):
        l.append(blob_list)
    pm = PopularityMeter(interval_to_consider_popular = 5, min_time_between_reports = 5, poll_interval = 1, publication_function = pub_func)
    pm.report_access(id1, c1, 2)
    pm.report_access(id1, c2, 2)
    time.sleep(2)
    assert (l == [[(id1, 2)]]), l
    pm.report_access(id1, c1, 2)
    pm.report_access(id1, c2, 2)
    time.sleep(2)
    assert (l == [[(id1, 2)]]), l
    pm.stop_polling()

def test_lets_publish_after_time_has_passed():
    id1 = mojosixbit.b2a(sha('first id').digest())
    c1 = mojosixbit.b2a(sha('first counterparty').digest())
    c2 = mojosixbit.b2a(sha('second counterparty').digest())
    l = []
    def pub_func(blob_list, l = l):
        l.append(blob_list)
    pm = PopularityMeter(interval_to_consider_popular = 5, min_time_between_reports = 1, poll_interval = 1, publication_function = pub_func)
    pm.report_access(id1, c1, 2)
    pm.report_access(id1, c2, 2)
    time.sleep(2)
    assert (l == [[(id1, 2)]]), l
    pm.report_access(id1, c1, 2)
    pm.report_access(id1, c2, 2)
    time.sleep(2)
    assert (l == [[(id1, 2)], [(id1, 2)]]), l
    pm.stop_polling()

def test_publish_two_at_once():
    id1 = mojosixbit.b2a(sha('first id').digest())
    id2 = mojosixbit.b2a(sha('second id').digest())
    c1 = mojosixbit.b2a(sha('first counterparty').digest())
    c2 = mojosixbit.b2a(sha('second counterparty').digest())
    l = []
    def pub_func(blob_list, l = l):
        l.append(blob_list)
    pm = PopularityMeter(interval_to_consider_popular = 5, min_time_between_reports = 5, poll_interval = 1, publication_function = pub_func)
    pm.report_access(id1, c1, 2)
    pm.report_access(id2, c1, 1)
    pm.report_access(id1, c2, 2)
    pm.report_access(id2, c2, 1)
    time.sleep(2)
    assert (l == [[(id1, 2), (id2, 1)]] or l == [[(id2, 1), (id1, 2)]]), l
    pm.stop_polling()

if __name__ == '__main__':
    RunTests.runTests('PopularityMeter')








