#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

import RunTests


class PriorityList:
    def __init__(self):
        # contains {thing: {'priority': priority, 'previous': previous, 'after': after}}
        self.things = {}
        # contains {priority: {'highest': highest, 'last': last}}
        self.priorities = {}
        
    def set(self, thing, priority):
        assert thing is not None
        if self.things.has_key(thing):
            if self.things[thing]['priority'] == priority:
                return
            self.forget(thing)
        if self.priorities.has_key(priority):
            previous = self.priorities[priority]['last']
            after = self.things[previous]['after']
            self.things[previous]['after'] = thing
            if after is not None:
                self.things[after]['previous'] = thing
            self.things[thing] = {'priority': priority, 'previous': previous, 'after': after}
            self.priorities[priority]['last'] = thing
        else:
            prev_priority = None
            after_priority = None
            for i in self.priorities.keys():
                if i > priority and (prev_priority is None or i < prev_priority):
                    prev_priority = i
                if i < priority and (after_priority is None or i > after_priority):
                    after_priority = i
            if prev_priority is None:
                previous = None
            else:
                previous = self.priorities[prev_priority]['last']
                self.things[previous]['after'] = thing
            if after_priority is None:
                after = None
            else:
                after = self.priorities[after_priority]['highest']
                self.things[after]['previous'] = thing
            self.things[thing] = {'priority': priority, 'previous': previous, 'after': after}
            self.priorities[priority] = {'highest': thing, 'last': thing}
        
    def forget(self, thing):
        if not self.things.has_key(thing):
            return 0
        old_entry = self.things[thing]
        del self.things[thing]
        previous = old_entry['previous']
        after = old_entry['after']
        if previous is not None:
            self.things[previous]['after'] = after
        if after is not None:
            self.things[after]['previous'] = previous
        p_entry = self.priorities[old_entry['priority']]
        if p_entry['highest'] == thing and p_entry['last'] == thing:
            del self.priorities[old_entry['priority']]
        else:
            if p_entry['highest'] == thing:
                p_entry['highest'] = after
            if p_entry['last'] == thing:
                p_entry['last'] = previous
        return 1
        
    def highest(self):
        highest_priority = None
        for i in self.priorities.keys():
            if highest_priority is None or i > highest_priority:
                highest_priority = i
        if highest_priority is None:
            return None
        else:
            return self.priorities[highest_priority]['highest']
        
    def after(self, thing):
        return self.things[thing]['after']

def assert_order(p, l):
    if len(l) == 0:
        assert p.highest() is None
        return
    assert p.highest() == l[0]
    for i in range(len(l) - 1):
        assert p.after(l[i]) == l[i+1]
    assert p.after(l[-1]) is None

def test_absolute_unitary():
    p = PriorityList()
    assert_order(p, [])
    p.set('a', 0)
    assert_order(p, ['a'])
    p.forget('a')
    assert_order(p, [])
    p.forget('a')
    assert_order(p, [])
    p.set('a', 0)
    assert_order(p, ['a'])

def test_absolute_beginning():
    p = PriorityList()
    p.set('a', 1)
    p.set('b', 2)
    assert_order(p, ['b', 'a'])
    p.forget('b')
    assert_order(p, ['a'])
    p.set('b', 2)
    assert_order(p, ['b', 'a'])

def test_absolute_end():
    p = PriorityList()
    p.set('a', 3)
    p.set('b', 2)
    assert_order(p, ['a', 'b'])
    p.forget('b')
    assert_order(p, ['a'])
    p.set('b', 2)
    assert_order(p, ['a', 'b'])

def test_middle_beggining():
    p = PriorityList()
    p.set('a', 0)
    p.set('b', 0)
    p.set('c', 1)
    p.forget('a')
    assert_order(p, ['c', 'b'])
    p.set('a', 0)
    assert_order(p, ['c', 'b', 'a'])

def test_middle_unitary():
    p = PriorityList()
    p.set('a', 0)
    p.set('b', 1)
    p.set('c', 2)
    p.forget('b')
    assert_order(p, ['c', 'a'])
    p.set('b', 1)
    assert_order(p, ['c', 'b', 'a'])

def test_double_delete():
    p = PriorityList()
    p.set('a', 0)
    p.set('b', 0)
    p.set('c', 0)
    p.forget('b')
    p.forget('c')
    assert_order(p, ['a'])
    p.set('d', 0)
    assert_order(p, ['a', 'd'])

def test_beginning_unitary():
    p = PriorityList()
    p.set('a', 1)
    p.set('b', 0)
    p.forget('a')
    assert_order(p, ['b'])
    p.set('a', 1)
    assert_order(p, ['a', 'b'])

def test_end_unitary():
    p = PriorityList()
    p.set('a', 1)
    p.set('b', 2)
    p.forget('a')
    assert_order(p, ['b'])
    p.set('a', 1)
    assert_order(p, ['b', 'a'])

def test_middle_middle():
    p = PriorityList()
    p.set('a', 0)
    p.set('b', 0)
    p.set('c', 0)
    p.forget('b')
    assert_order(p, ['a', 'c'])
    p.forget('b')
    assert_order(p, ['a', 'c'])
    p.set('b', 0)
    assert_order(p, ['a', 'c', 'b'])

def test_move():
    p = PriorityList()
    p.set('a', 1)
    p.set('b', 0)
    p.set('a', 0)
    assert_order(p, ['b', 'a'])

def test_does_not_move_on_same_set():
    p = PriorityList()
    p.set('a', 0)
    p.set('b', 0)
    p.set('a', 0)
    assert_order(p, ['a', 'b'])

def test_general_operation():
    p = PriorityList()
    p.set('a', 0)
    p.set('b', 1)
    p.set('c', 0)
    p.set('d', 2)
    p.set('e', 1)
    p.set('f', 2)
    assert_order(p, ['d', 'f', 'b', 'e', 'a', 'c'])

def test_insert_high_low_mid():
    p = PriorityList()
    p.set('a', 2)
    p.set('c', 0)
    p.set('b', 1)
    assert_order(p, ['a', 'b', 'c'])

def test_reprioritize_works():
    p = PriorityList()
    p.set('a', 2)
    p.set('f', 0)
    p.set('g', 0)
    p.set('h', 0)
    assert_order(p, ['a', 'f', 'g', 'h'])
    p.set('f', 1)
    p.set('g', 1)
    assert_order(p, ['a', 'f', 'g', 'h'])
    p.forget('a')
    assert_order(p, ['f', 'g', 'h'])
    p.forget('f')
    assert_order(p, ['g', 'h'])
    p.forget('h')
    assert_order(p, ['g'])
    p.forget('g')
    assert p.highest() is None

def test_forget_internals():
    p = PriorityList()
    p.set('a', 0)
    p.set('aa', 0)
    p.set('b', 1)
    p.forget('aa')
    p.forget('b')
    p.forget('a')
    assert p.highest() is None
    assert p.things == {}
    assert p.priorities == {}


mojo_test_flag = 1

if __name__ == '__main__':
    RunTests.runTests('PriorityList')


