#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# Collection of Mojo Tokens
#

### standard modules
import copy
import exceptions
import threading
import os
import sys
import traceback
import types
import zlib
try:
    import cPickle
    pickle=cPickle
    del cPickle
except ImportError:
    import pickle

### our modules
import MojoToken
from bsddb3 import db, dbobj
import canon
import debug
# HACK: this is needed to allow unpickling of Tokens created using the old "Token" modulename instead of "MojoToken"
sys.modules['Token'] = sys.modules['MojoToken']
import fileutil
from human_readable import hr
import tokendists
import mojoutil
from confutils import confman

from CleanLogDb import CleanLogDbEnv

from MojoConstants import true, false


SHOW_GET_MORE_MOJO_THRESH = 500000

#######################################################################

class PurseError(exceptions.StandardError):
    pass

class NotEnoughMojoError(PurseError):
    """
    An operation cannot be completed because of a lack of Mojo in a purse.
    """
    pass

class NoExactChangeError(PurseError):
    """
    An operation cannot be completed because of a lack of the correct change distribution in a
    purse.  (E.g. a call to `makePayment(10, p)' will raise this error if the Purse `p' contains
    only an 8-Mojo token and a 4-Mojo token, since it is impossible to make a payment of exactly
    10 Mojo with that token distribution.)
    """
    pass

class TokenAlreadyInPurse(PurseError):
    """
    This error is raised if you call add_token with a token that
    already exists in a DbPurse.  This should never happen.
    """
    pass

class BadEncryptionError(PurseError):
    """
    The encryption/decryption functions given to a Purse do not work.
    """
    pass

#######################################################################


class DbPurse:
    """A token purse that makes good use of a BerkeleyDB database"""
    def __init__(self, dbparentdir="", recoverdb=true, encryptionfunc=None, decryptionfunc=None):
        """
        @param encryptionfunc/decryptionfunc must be callable, take one string parameter and return the encrypted string.
        Encryption is done on both the token ids and serialized tokens.  The goal is only to make unauthorized access
        difficult, not to prevent knowing how much the purse contains.
        """
        assert (encryptionfunc is None and decryptionfunc is None) or (callable(encryptionfunc) and callable(decryptionfunc))
        self._encrypt = encryptionfunc
        if self._encrypt is None:
            self._encrypt = lambda thing: thing  # no encryption
        self._decrypt = decryptionfunc
        if self._decrypt is None:
            self._decrypt = lambda thing: thing  # no encryption
        
        assert self._decrypt(self._encrypt("king phillip came over for great shoes\n")) == "king phillip came over for great shoes\n"

        self._dbdir = os.path.join(dbparentdir, "tokens")
        fileutil.make_dirs(self._dbdir)
        self._db_env = CleanLogDbEnv()
        self._db_env.set_lk_detect(db.DB_LOCK_DEFAULT)
        extraflags = 0
        if recoverdb:
            extraflags = extraflags | db.DB_RECOVER
        extraflags = extraflags | db.DB_PRIVATE   # disable multiprocess access, it seems to be causing problems

        try:
            self._db_env.open(self._dbdir, db.DB_CREATE | db.DB_INIT_MPOOL | db.DB_INIT_LOCK | db.DB_INIT_TXN | db.DB_THREAD | extraflags)
        except db.DBError, dbe:
            debug.mojolog.write('Failed to open the database environment the first time, reason: %s\nTrying again...\n', args=(dbe,), vs='DbPurse', v=2)
            # XXX HACK sometimes trying again after one open with DB_RECOVER works.
            try:
                self._db_env.open(self._dbdir, db.DB_CREATE | db.DB_INIT_MPOOL | db.DB_INIT_LOCK | db.DB_INIT_TXN | db.DB_THREAD | extraflags | db.DB_RECOVER)
            except db.DBError, dbe:
                debug.mojolog.write('Failed to open the database environment the second time, reason: %s\nTrying again...\n', args=(dbe,), vs='DbPurse', v=2)
                # XXX DOUBLE CHOCOLATEY HACK sometimes trying *again* after one open *without* DB_RECOVER works.
                self._db_env.open(self._dbdir, db.DB_CREATE | db.DB_INIT_MPOOL | db.DB_INIT_LOCK | db.DB_INIT_TXN | db.DB_THREAD | extraflags & (~db.DB_RECOVER))


        # Key -> Value(s) mapping
        #  denom string -> ids of tokens of this denomination (multiple values, use a database cursor)
        self._denom_db = dbobj.DB(self._db_env)
        self._denom_db.set_flags(db.DB_DUP)
        self._denom_db.open("denom.db", db.DB_BTREE, db.DB_CREATE)

        # Key:                     Value:
        #  encrypted token ids  ->  encrypted pickled MojoToken object
        self._token_db = dbobj.DB(self._db_env)
        self._token_db.open("token.db", db.DB_BTREE, db.DB_CREATE)

        # for safety, test that the given encryption/decryption functions are working and that
        # we can successfully load a token from the db.
        tokencursor = self._token_db.cursor()
        try:
            keyandvalue = tokencursor.first()
            if keyandvalue:
                enctoken_id, enctoken = keyandvalue
                tmptoken = pickle.loads(zlib.decompress(self._decrypt(enctoken)))
                if self._decrypt(enctoken_id) != tmptoken.get_id():  # this should not be possible
                    raise BadEncryptionError, "decrypted token id didn't match id returned from decrypted token"
                assert type(enctoken_id) == types.StringType, "assertion: enctoken_id must be a string. -- enctoken_id: %s :: %s" % (hr(enctoken_id), hr(type(enctoken_id)))
                assert canon.is_canonical_modval(enctoken_id), "assertion: enctoken_id must be the right length. -- enctoken_id: %s, self._decrypt(enctoken_id): %s" % (hr(enctoken_id), hr(self._decrypt(enctoken_id)))
                assert type(enctoken) == types.StringType, "assertion: enctoken must be a string. -- enctoken: %s :: %s" % (hr(enctoken), hr(type(enctoken)))
        except (zlib.error, pickle.UnpicklingError):
            tokencursor.close()
            raise BadEncryptionError, "couldn't decrypt a stored token using the given decryption function"
        except db.DBError:
            pass  # assume that this was an empty purse and continue
        tokencursor.close()  # needed due to current Db cursor bugs
        del tokencursor
        
        self.__open = 1
        # automatically upgrade any existing old-style Purse object
        self._import_old_purse()

        # search for lost tokens due to pre v0.938 bugs at least once
        if confman.get('Recovered_Lost_DbPurse_Mojo_Tokens', 'no') != 'yes':
            debug.mojolog.write("Checking DbPurse for pre-0.938 lost tokens\n", v=2, vs='Purse')
            self.__recover_lost_tokens()
            confman.dict['Recovered_Lost_DbPurse_Mojo_Tokens'] = 'yes'
            confman.save()

    def __del__(self):
        if hasattr(self, '_denom_db'):
            self._denom_db.close()
        if hasattr(self, '_token_db'):
            self._token_db.close()
    
    def shutdown(self):
        if not self.__open:
            return
        self.__open = None
        try:
            self._db_env.txn_checkpoint(0)
        except db.DBError:
            self._db_env.nosyncerror_txn_checkpoint(0)
        self._denom_db.close()
        self._token_db.close()

    def _import_old_purse(self, oldpursename='purse'):
        """
        Move all tokens from an old style Purse "database" existing in the
        directory we were created in into a DbPurse.
        """
        assert self.__open, "DbPurse object already shutdown"
        if not os.path.isfile(os.path.join(self._dbdir, oldpursename)):
            return  # no old purse to upgrade, we're done
        debug.mojolog.write("Migrating old-style token Purse %s to DbPurse\n", args=(oldpursename,), vs="Purse")
        debug.mojolog.write("*** WARNING: do not kill your Broker while it is upgrading the purse ***\n", vs="Purse")
        old_db = dbobj.DB(self._db_env)
        try:
            old_db.open(oldpursename, db.DB_BTREE, 0)
            if old_db.has_key('10000'):
                old_purse_tokens = pickle.loads(old_db['10000'])
                iter = 0
                for tokenid, token in old_purse_tokens.items():
                    self.add_token(token, sync=None)
                    del old_purse_tokens[tokenid]
                    iter = iter + 1
                    if iter % 50 == 0:
                        self.sync()
                        debug.mojolog.write("  50 tokens migrated\n", vs="Purse")
                        debug.mojolog.write("*** WARNING: do not kill your Broker while it is upgrading the purse ***\n", vs="Purse")
                del old_db['10000']
                self.sync()
                debug.mojolog.write("%s total tokens migrated\n", args=(iter,), vs="Purse")
            old_db.close()
            debug.mojolog.write("Purse migration complete, deleting old %s.\n", args=(oldpursename,), vs="Purse")
            os.unlink(os.path.join(self._dbdir, oldpursename))
        except:
            debug.mojolog.write("ERROR importing old style purse %s:\n", args=(oldpursename,), vs='Purse')
            traceback.print_exc(file=debug.mojolog)
    
    def __recover_lost_tokens(self):
        """
        A bug existed in DbPurse from its creation through versions
        just before MojoNation 0.938 was released.  It would cause an
        extra token to be `lost' from the purse index every time any
        tokens were removed from the purse.  This function collects
        that lost change from between the couch cushions.

        Only call this from the constructor before this object could
        ever be in use by other threads.
        """
        txn = None  # NOTE: don't use a transaction here!  2001-01-22
        # this does much work for a single txn which results in a db.DBError (12, "couldn't allocate memory")
        # (dying after processing about 1800 tokens in my tests -greg)
        # not using a transaction means the db needs no internal state so it shouldn't have a problem
        denomcursor = None
        tokencursor = None
        try:
            # build a hash table of all indexed token ids
            denomcursor = self._denom_db.cursor(txn)
            indexed_enctoken_ids = {}
            keyandvalue = denomcursor.get(db.DB_FIRST)
            while keyandvalue:
                indexed_enctoken_ids[keyandvalue[1]] = 1  # mark this token id as being indexed
                keyandvalue = denomcursor.get(db.DB_NEXT)
            denomcursor.close()
            denomcursor = None

            # find all tokens in token_db that were not indexed
            reclaimed = 0L
            num = 0
            tokencursor = self._token_db.cursor(txn)
            keyandvalue = tokencursor.get(db.DB_FIRST)
            while keyandvalue:
                debug.mojolog.write("checking tokens, reclaimed %s, token num %s\n", args=(reclaimed, num), v=4, vs='Purse')
                num = num + 1
                enctoken_id, enctoken = keyandvalue
                if not indexed_enctoken_ids.has_key(enctoken_id):
                    losttoken_id, losttoken = self._decrypt(enctoken_id), pickle.loads(zlib.decompress(self._decrypt(enctoken)))
                    losttokenvalue = MojoToken.MAP_KEYID_TO_DENOM[losttoken.get_server_public_key_id()]
                    denomstr = str(losttokenvalue)
                    # re-index this token in the appropriate token denomination index.
                    self._denom_db.put(denomstr, enctoken_id, txn=txn)
                    reclaimed = reclaimed + losttokenvalue
                keyandvalue = tokencursor.get(db.DB_NEXT)
            tokencursor.close()
            tokencursor = None

            debug.mojolog.write("Recovered tokens worth %s from between the cushions\n", args=[reclaimed], v=2, vs='Purse')  # due to an old implementation with holes in its pockets
            if txn:
                txn.commit()
                txn = None
        finally:
            if denomcursor:
                denomcursor.close()
                del denomcursor
            if tokencursor:
                tokencursor.close()
                del tokencursor
            if txn:
                txn.abort()
            self.sync()

    def sync(self):
        assert self.__open, "DbPurse object already shutdown"
        try:
            self._token_db.sync()
        except db.DBIncompleteError:
            pass  # ignore DB_INCOMPLETE sync didn't finish
        try:
            self._denom_db.sync()
        except db.DBIncompleteError:
            pass  # ignore DB_INCOMPLETE sync didn't finish

    def get_token_dist(self, _txn=None):
        """
        Return a dictionary mapping token denominations to the number of that size token in the purse.
        (tokendists.py needs dictionaries like this)
        """
        assert self.__open, "DbPurse object already shutdown"

        # XXX Could optimize by storing this integer directly and updating it every time you add or subtract tokens.  Currently we call `get_token_dist()' very frequently (in every call to `get_aggregate_amount()' or `get_tokens()', in fact.)

        #debug.mojolog.write(">entering get_token_dist\n", v=4, vs='Purse') # v=15 or comment out the line...  It's all the same to me.
        try:
            dist = {}
            # cursors always need to be transaction protected in a transactional database
            denomcursor = None
            txn = self._db_env.txn_begin(_txn)
            try:
                denomcursor = self._denom_db.cursor(txn)
                keyandvalue = denomcursor.get(db.DB_FIRST)
                while keyandvalue:
                    dist[long(keyandvalue[0])] = denomcursor.count()
                    keyandvalue = denomcursor.get(db.DB_NEXT_NODUP)
                denomcursor.close()
                denomcursor = None
            finally:
                if denomcursor:
                    denomcursor.close()
                    del denomcursor
                # we didn't actually write anything, always abort this txn
                if txn:
                    txn.abort()

            return dist
        finally:
            #debug.mojolog.write("<exiting get_token_dist\n", v=4, vs='Purse')
            pass

    def get_aggregate_amount(self, dist=None):
        """
        Returns the aggregate amount of Mojo in the purse.  If
        specified, dist is the result of get_token_dist; otherwise
        we'll call it ourselves.
        """
        assert self.__open, "DbPurse object already shutdown"
        debug.mojolog.write(">entering get_aggregate_amount\n", v=4, vs='Purse')
        try:
            am = 0
            if not dist:
                dist = self.get_token_dist()
            for denom, numberwehave in dist.items():
                am = am + (denom * numberwehave)
            if confman.dict['GOT_MOJO'] == 'no':
                if long(am) >= SHOW_GET_MORE_MOJO_THRESH:
                    confman.dict['GOT_MOJO'] = 'yes'
                    confman.save()
            return long(am)
        finally:
            debug.mojolog.write("<exiting get_aggregate_amount\n", v=4, vs='Purse')

    def add_tokens(self, newtokenlist):
        debug.mojolog.write("add_tokens(%s)\n", args=(newtokenlist,), v=3, vs="debug")
        for newtoken in newtokenlist:
            self.add_token(newtoken, sync=None)
        self.sync()

    def add_token(self, newtoken, sync=1):
        """
        @precondition newtoken must be a signed token.: newtoken.isSigned(): "newtoken: %s" % hr(newtoken)
        @precondition DbPurse object must not be already shutdown.: self.__open: "self: %s" % hr(self)
        """
        assert newtoken.isSigned(), "precondition: newtoken must be a signed token." + " -- " + "newtoken: %s" % hr(newtoken)
        assert self.__open, "precondition: DbPurse object must not be already shutdown." + " -- " + "self: %s" % hr(self)

        debug.mojolog.write(">entering add_token\n", v=7, vs='Purse')
        try:
            tskeyid = newtoken.get_server_public_key_id()
            denomstr = str(MojoToken.MAP_KEYID_TO_DENOM[tskeyid])

            enctoken_id = self._encrypt(newtoken.get_id())
            assert type(enctoken_id) == types.StringType, "assertion: enctoken_id must be a string. -- enctoken_id: %s :: %s" % (hr(enctoken_id), hr(type(enctoken_id)))
            assert canon.is_canonical_modval(enctoken_id), "assertion: enctoken_id must be the right length. -- enctoken_id: %s, self._decrypt(enctoken_id): %s" % (hr(enctoken_id), hr(self._decrypt(enctoken_id)))

            # no transaction is needed here as we store the token in _token_db before we add
            # the reference to its id in _denom_db.
            try:
                enctoken = self._encrypt(zlib.compress(pickle.dumps(newtoken)))
                assert type(enctoken) == types.StringType, "assertion: enctoken must be a string. -- enctoken: %s :: %s" % (hr(enctoken), hr(type(enctoken)))
                self._token_db.put(enctoken_id, enctoken, None, flags=db.DB_NOOVERWRITE)
                # ??? self._token_db.put(enctoken_id, enctoken, txn=None, flags=db.DB_NOOVERWRITE) ???  --Zooko 2001-03-12
            except db.DBKeyExistError, e:
                debug.mojolog.write("token already exists in the DbPurse: %s\n", args=(newtoken,), v=2, vs="Purse")
                return

            # add enctoken_id to the list stored under denomstr (this db was opened with DB_DUP)
            self._denom_db.put(denomstr, enctoken_id)
        finally:
            if sync:
                self.sync()
            debug.mojolog.write("<exiting add_token\n", v=7, vs='Purse')

    def get_tokens(self, amount):
        """
        Removes a set of tokens from the purse worth amount and returns them in a
        dictionary mapping token id to token.

        @param amount the aggregate amount desired
        
        @return dict with key: token-id and value: Token instance

        @throws NotEnoughMojoError if there is not enough Mojo in purse to make a payment of size
            `amount'
        @throws NoExactChangeError if there are not correctly sized tokens in purse to make a 

        @precondition `amount' must be positive.: amount > 0: "amount: %s" % hr(amount)
        """
        assert amount > 0, "`amount' must be positive." + " -- " + "amount: %s" % hr(amount)
        assert self.__open, "DbPurse object already shutdown"
        debug.mojolog.write(">entering get_tokens\n", v=4, vs='Purse')
        denomcursor = None
        txn = self._db_env.txn_begin()
        try:
            dist = self.get_token_dist(txn)
            if amount > self.get_aggregate_amount(dist=dist):
                raise NotEnoughMojoError
            if not tokendists.is_amount_possible(amount, dist):
                raise NoExactChangeError

            cdist = tokendists.choose_token_dist_from_set(amount, dist)

            collection = {}

            # Now we go through the denoms in descending order of size...
            denoms = cdist.items()
            denoms.sort()
            denoms.reverse()

            denomcursor = self._denom_db.cursor(txn)
            for denom, amountofthisdenom in denoms:
                tmpdenom, enctoken_id = denomcursor.set(str(denom))
                assert type(enctoken_id) == types.StringType, "assertion: enctoken_id must be a string. -- enctoken_id: %s :: %s" % (hr(enctoken_id), hr(type(enctoken_id)))
                assert canon.is_canonical_modval(enctoken_id), "assertion: enctoken_id must be the right length. -- enctoken_id: %s, self._decrypt(enctoken_id): %s" % (hr(enctoken_id), hr(self._decrypt(enctoken_id)))

                while amountofthisdenom > 0:
                    debug.mojolog.write("beginning to seek token, denom: %s, amountofthisdenom: %s\n", args=(denom, amountofthisdenom,), v=3, vs="debug")
                    denomcursor.delete()  # we're taking this token, remove it from the list of available tokens
                    try:
                        token_id, token = self.__remove_token(enctoken_id, txn)
                        collection[token_id] = token
                        del token_id, token  # good habit: never keep more than one reference to a token
                        amountofthisdenom = amountofthisdenom - 1
                    except BadEncryptionError, le:
                        debug.mojolog.write("Found bogus token.  Ignoring.  le: %s\n", args=(le,), v=3, vs="debug")
                        # go on to the next token then.

                    next_token = denomcursor.next()
                    if next_token is None:
                        if amountofthisdenom > 0:
                            # this should never happen due to the transaction
                            # ... except if some of our tokens in the db are bogus...
                            # assert amountofthisdenom == 0, "internal error: unexpectedly ran out of %d sized tokens, wanted %d more" % (denom, amountofthisdenom)
                            raise NotEnoughMojoError, "internal error: unexpectedly ran out of %d sized tokens, wanted %d more" % (denom, amountofthisdenom)
                        break
                    tmpdenom, enctoken_id = next_token
                    assert type(enctoken_id) == types.StringType, "assertion: enctoken_id must be a string. -- enctoken_id: %s :: %s" % (hr(enctoken_id), hr(type(enctoken_id)))
                    assert canon.is_canonical_modval(enctoken_id), "assertion: enctoken_id must be the right length. -- enctoken_id: %s, self._decrypt(enctoken_id): %s" % (hr(enctoken_id), hr(self._decrypt(enctoken_id)))
                    del next_token  # good habit: never keep more than one reference to a token
                    if (long(tmpdenom) != denom) and (amountofthisdenom > 0):
                        # this should never happen due to the transaction
                        # ... except if some of our tokens in the db are bogus...
                        # assert amountofthisdenom == 0, "internal error: unexpectedly ran out of %d sized tokens, wanted %d more" % (denom, amountofthisdenom)
                        raise NotEnoughMojoError, "internal error: unexpectedly ran out of %d sized tokens, wanted %d more" % (denom, amountofthisdenom)

            denomcursor.close()
            denomcursor = None
            txn.commit()
            txn = None
            self._db_env.nosyncerror_txn_checkpoint(1)

            return collection
        finally:
            if denomcursor:
                denomcursor.close()
                del denomcursor
            debug.mojolog.write("<exiting get_tokens\n", v=4, vs='Purse')
            if txn:
                txn.abort()
    
    def __remove_token(self, enctoken_id, txn):
        """
        Remove a specific token from our _token_db database, decrypt it and return it.

        @precondition `enctoken_id' must be a string.: type(enctoken_id) == types.StringType: "enctoken_id: %s :: %s" % (hr(enctoken_id), hr(type(enctoken_id)))
        @precondition `enctoken_id' must be the right length.: canon.is_canonical_modval(enctoken_id): "enctoken_id: %s" % hr(enctoken_id)
        """
        assert type(enctoken_id) == types.StringType, "precondition: `enctoken_id' must be a string." + " -- " + "enctoken_id: %s :: %s" % (hr(enctoken_id), hr(type(enctoken_id)))
        assert canon.is_canonical_modval(enctoken_id), "precondition: `enctoken_id' must be the right length." + " -- " + "enctoken_id: %s" % hr(enctoken_id)
        assert type(enctoken_id) == types.StringType, "precondition: `enctoken_id' must be a string." + " -- " + "enctoken_id: %s :: %s" % (hr(enctoken_id), hr(type(enctoken_id)))
        assert canon.is_canonical_modval(enctoken_id), "precondition: `enctoken_id' must be the right length." + " -- " + "enctoken_id: %s" % hr(enctoken_id)

        debug.mojolog.write(">entering __remove_token\n", v=4, vs='Purse')
        try:
            try:
                enctoken = self._token_db.get(enctoken_id, txn=txn, flags=db.DB_RMW)
                self._token_db.delete(enctoken_id, txn=txn) # remove it, whether it was good or not.
                token_id, token = self._decrypt(enctoken_id), pickle.loads(zlib.decompress(self._decrypt(enctoken)))
            except (zlib.error, pickle.UnpicklingError), le:
                raise BadEncryptionError, "decryption of a stored token failed -- enctoken_id: %s, enctoken: %s, le: %s" % (hr(enctoken_id), hr(enctoken), hr(le))
            return (token_id, token)
        finally:
            debug.mojolog.write("<exiting __remove_token\n", v=4, vs='Purse')


def test_DbPurse_init():
    p = DbPurse(RunTests.get_next_dir_name())
    assert p.get_aggregate_amount() == 0, "new purse wasn't empty"

mojo_test_flag = 1
import RunTests
#### generic stuff
def run():
    RunTests.runTests(["Purse"])
    
#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
