#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import threading
import whrandom

### our modules
import Cache
import MetaTrackerLib
import MojoConstants
import RunTests
import idlib
from confutils import confman, integerizePrice
import pricelib
import debug


MAX_COUNTERPARTIES = 5000

class RelayServerHandlers :
    def __init__(self, mtm) :
        # maps recipient ids to {sender_id: {message_id: message}}
        self.maxsize = int( max(3.5, float(confman["RELAYSERVER"]["MAX_MEMORY_MEGABYTES"])) *1000*1024)  # the 1000 accounts for overhead
        self.__counterparties_to_messages = Cache.StatsCache(maxsize=self.maxsize, maxitems=MAX_COUNTERPARTIES, autoexpireinterval=150, autoexpireparams={'staletime': 600})

        self.__lock = threading.Lock()
        self.mtm = mtm

        # a table of ascii counterparty ids that have recently (in the last 10 minutes) told us that we should try and
        # fast-relay any messages we receive for them (assuming we have an open connection, otherwise they still need
        # to pick them up).
        self._cids_checked_in_map = Cache.StatsCache(maxitems=MAX_COUNTERPARTIES, autoexpireinterval=150, autoexpireparams={'maxage': 600})

        self.announced_services = {
            'type': "relay server v2",
            'are there messages price v2': integerizePrice(confman.dict["RELAYSERVER"]["PRICES"]["POLL"]),
            # NOTE: do not integerize "per kilobyte" prices, they can be floats as they're used in a price computation.
            'pass this along per kilobyte price v2': confman.dict["RELAYSERVER"]["PRICES"]["FORWARD_PER_KB"],
            'retrieve messages per kilobyte price v2': confman.dict["RELAYSERVER"]["PRICES"]["RETRIEVE_PER_KB"],
        }

    def compute_cost_are_there_messages(self, widget, message_type, message_body):
        return int(self.announced_services['are there messages price v2'])

    def compute_cost_pass_this_along_v2(self, widget, message_type, message_body):
        price_per_kb = float(self.announced_services['pass this along per kilobyte price v2'])
        bytes = len(message_body['message'])
        return pricelib.get_price_using_per_kilobyte(price_per_kb, bytes)

    def compute_cost_retrieve_messages_v2(self, widget, message_type, message_body):
        recipient_id = idlib.to_binary(widget.get_counterparty_id())
        price_per_kb = float(self.announced_services['retrieve messages per kilobyte price v2'])
        if not message_body.has_key('messages'):
            bytes = self.compute_bytes_for_counterparty(recipient_id)
            return pricelib.get_price_using_per_kilobyte(price_per_kb, bytes)
        else:
            bytes = 0
            for i in message_body['messages']:
                bytes = bytes + i['length']
            return pricelib.get_price_using_per_kilobyte(price_per_kb, bytes)

    def get_handlers_dict(self) :
        dict = {}
        dict['are there messages v2'] = (self.has_messages, self.compute_cost_are_there_messages)
        dict['pass this along v2'] = (self.relay_message, self.compute_cost_pass_this_along_v2)
        dict['retrieve messages v2'] = (self.get_message, self.compute_cost_retrieve_messages_v2)
        return dict

    def get_services_dict(self) :
        """
        Returns an announced services dict suitable for including in a
        hello message when you're running a RelayServer.
        """
        return self.announced_services
    
    def compute_bytes_for_counterparty(self, counterparty_id):
        counterparty_id = idlib.to_binary(counterparty_id)
        h = self.__counterparties_to_messages.get(counterparty_id, {})
        bytes = 0
        for message_cache in h.values():
            bytes = bytes + reduce(lambda a,b: a+b, map(len, message_cache.values()), 0)  # compute bytes used by message_cache
        return bytes

    def relay_message(self, widget, message_body) :
        recipient_id = idlib.canonicalize(message_body.get('recipient'))
        sender_id = idlib.canonicalize(widget.get_counterparty_id())
        message = message_body.get('message')
        firstmsgId = widget._firstmsgId

        cid_registration = self._cids_checked_in_map.get(recipient_id)

        def hold_message_for_pickup(self=self, recipient_id=recipient_id, sender_id=sender_id, message=message):
            debug.mojolog.write("holding %s byte message from %s to %s for pickup\n", args=(len(message), sender_id, recipient_id,), v=3, vs="RelayServer")
            h = self.__counterparties_to_messages.get(recipient_id)
            if h is None :
                h = Cache.StatsCache(maxsize=self.maxsize/3, autoexpireinterval=70, autoexpireparams={'maxage': 600})
                self.__counterparties_to_messages.insert(recipient_id, h, itemsize=len(recipient_id)+80)
            mdict = h.get(sender_id, {})
            mdict[idlib.make_id(message, 'msg')] = message
            mdict_bytes = reduce(lambda a,b: a+b, map(len, mdict.values()), 0)  # compute bytes of memory used by mdict
            # reinsert it in the StatsCache to keep the its memory totals updated
            h.insert(sender_id, mdict, itemsize=mdict_bytes)
            counterparty_bytes = self.compute_bytes_for_counterparty(recipient_id)
            self.__counterparties_to_messages.insert(recipient_id, h, itemsize=counterparty_bytes+len(recipient_id)+80)

        debug.mojolog.write("attempting fast-relay of message from %s to %s\n", args=(sender_id, recipient_id), v=3, vs="RelayServer")
        def outcome_func(widget, outcome, failure_reason=None, hold_message_for_pickup=hold_message_for_pickup, self=self, recipient_id=recipient_id, firstmsgId=firstmsgId):
            if failure_reason or (outcome.get('result') == "failure"):
                if self._cids_checked_in_map.has_key(recipient_id):
                    hold_message_for_pickup()
                    self.mtm.respond_with(firstmsgId, {'result': 'ok', 'note': "slow delivery"})
                else:
                    self.mtm.respond_with(firstmsgId, {'result': 'failure', 'reason': "delivery failure and recipient does not pickup messages here", 'underlying failure_reason': failure_reason})
            else:
                self.mtm.respond_with(firstmsgId, {'result': 'ok'})
        self.mtm.initiate(recipient_id, 'message for you', { 'message': message }, 0, outcome_func=outcome_func)

        return MojoConstants.ASYNC_RESPONSE

    def has_messages(self, widget, message_body) :
        recipient_id = idlib.to_binary(widget.get_counterparty_id())
        self._cids_checked_in_map[recipient_id] = 'fast'
        h = self.__counterparties_to_messages.get(recipient_id)
        if h is None or len(h) == 0:
            return {'result' : 'no'}
        mlist = []
        totalmessagebytes = 0
        totalmessages = 0
        for sender_id, message_dict in h.items():
            totalmessages = totalmessages + len(message_dict)
            for message_id, message in message_dict.items():
                totalmessagebytes = totalmessagebytes + len(message)
                mlist.append({'sender id': sender_id, 'length': len(message), 
                    'message id': message_id})
        return {'result': 'yes', 'number of messages': str(totalmessages),
            'total bytes': str(totalmessagebytes), 'messages info': mlist}

    def get_message(self, widget, message_body) :
        recipient_id = idlib.to_binary(widget.get_counterparty_id())
        self.__lock.acquire()
        try :
            if not message_body.has_key('messages'):
                h = self.__counterparties_to_messages.get(recipient_id)
                if h is None or len(h) == 0 :
                    return {'status': 'no messages'}
                sender_id, message_dict = h.items()[0]
                key, message = message_dict.items()[0]
                del message_dict[key]
                if len(message_dict) == 0:
                    del h[sender_id]
                if len(h) == 0:
                    del self.__counterparties_to_messages[recipient_id]
                return {'status' : "messages attached", 'sender id' : sender_id, 
                    'message' : message}
            else:
                rlist = []
                for i in message_body['messages']:
                    h = self.__counterparties_to_messages.get(recipient_id)
                    if h is None:
                        continue
                    i['sender id'] = idlib.to_binary(i['sender id'])
                    message_dict = h.get(i['sender id'])
                    if message_dict is None:
                        continue
                    i['message id'] = idlib.to_binary(i['message id'])
                    message = message_dict.get(i['message id'])
                    if message is not None and len(message) == i['length']:
                        del message_dict[i['message id']]
                        rlist.append({'sender id': i['sender id'], 'message': message})
                        if len(message_dict) == 0:
                            del h[i['sender id']]
                        if len(h) == 0:
                            del self.__counterparties_to_messages[recipient_id]
                return {'status': 'messages attached v2', 'messages': rlist}
        finally :
            self.__lock.release()

mojo_test_flag = 1

if __name__ == '__main__' :
    RunTests.runTests('RelayServerHandlers')

class MyWidget :
    def __init__(self, idstring) :
        self.counterparty_id = idlib.make_id(idstring)
        
    def get_counterparty_id(self) :
        return self.counterparty_id

def test_get_message_returns_no_messages_initially() :
    rsh = RelayServerHandlers(None)
    result = rsh.get_message(MyWidget('a'), {})
    assert result['status'] == 'no messages'
    
def test_has_message_returns_no_initially() :
    rsh = RelayServerHandlers(None)
    result = rsh.has_messages(MyWidget('a'), {})
    assert result['result'] == 'no'

def test_store_and_retrieve_two_messages() :
    rsh = RelayServerHandlers(None)
    wa = MyWidget('a')
    ida = idlib.to_binary(wa.get_counterparty_id())
    wb = MyWidget('b')
    idb = idlib.to_binary(wb.get_counterparty_id())
    val0 = rsh.has_messages(wb, {})
    assert val0['result'] == 'no'
    result1 = rsh.relay_message(MyWidget('a'), {'recipient' : idb, 'message' : 'foo'})
    assert result1['result'] == 'ok'
    val1 = rsh.has_messages(wb, {})
    assert val1['result'] == 'yes'
    assert val1['number of messages'] == '1'
    assert val1['total bytes'] == '3'
    result2 = rsh.relay_message(MyWidget('a'), {'recipient' : idb, 'message' : 'fo'})
    assert result2['result'] == 'ok'
    val2 = rsh.has_messages(wb, {})
    assert val2['result'] == 'yes'
    assert val2['number of messages'] == '2'
    assert val2['total bytes'] == '5'
    m1 = rsh.get_message(wb, {})
    assert m1['status'] == 'messages attached'
    assert m1['sender id'] == ida
    if m1['message'] == 'fo':
        assert m1['message'] == 'fo'
        val1 = rsh.has_messages(wb, {})
        assert val1['result'] == 'yes'
        assert val1['number of messages'] == '1'
        assert val1['total bytes'] == '3'
        m2 = rsh.get_message(wb, {})
        assert m2['status'] == 'messages attached'
        assert m2['sender id'] == ida
        assert m2['message'] == 'foo'
        val0 = rsh.has_messages(wb, {})
        assert val0['result'] == 'no'
        m3 = rsh.get_message(wb, {})
        assert m3['status'] == 'no messages'
    else:
        assert m1['message'] == 'foo'
        val1 = rsh.has_messages(wb, {})
        assert val1['result'] == 'yes'
        assert val1['number of messages'] == '1'
        assert val1['total bytes'] == '2'
        m2 = rsh.get_message(wb, {})
        assert m2['status'] == 'messages attached'
        assert m2['sender id'] == ida
        assert m2['message'] == 'fo'
        val0 = rsh.has_messages(wb, {})
        assert val0['result'] == 'no'
        m3 = rsh.get_message(wb, {})
        assert m3['status'] == 'no messages'

def test_store_two_identical_messages_works() :
    rsh = RelayServerHandlers(None)
    wa = MyWidget('a')
    ida = idlib.to_binary(wa.get_counterparty_id())
    wb = MyWidget('b')
    idb = idlib.to_binary(wb.get_counterparty_id())
    val0 = rsh.has_messages(wb, {})
    assert val0['result'] == 'no'
    result1 = rsh.relay_message(MyWidget('a'), {'recipient' : idb, 'message' : 'foo'})
    assert result1['result'] == 'ok'
    val1 = rsh.has_messages(wb, {})
    assert val1['result'] == 'yes'
    assert val1['number of messages'] == '1'
    assert val1['total bytes'] == '3'
    result2 = rsh.relay_message(MyWidget('a'), {'recipient' : idb, 'message' : 'foo'})
    assert result2['result'] == 'ok'
    val2 = rsh.has_messages(wb, {})
    assert val2['result'] == 'yes'
    assert val2['number of messages'] == '1'
    assert val2['total bytes'] == '3'
    m1 = rsh.get_message(wb, {})
    assert m1['status'] == 'messages attached'
    assert m1['sender id'] == ida
    assert m1['message'] == 'foo'
    val0 = rsh.has_messages(wb, {})
    assert val0['result'] == 'no'
    m3 = rsh.get_message(wb, {})
    assert m3['status'] == 'no messages'

def test_retrieve_multi_message_returns_nothing_initially():
    rsh = RelayServerHandlers(None)
    wa = MyWidget('a')
    ida = idlib.to_binary(wa.get_counterparty_id())
    wb = MyWidget('b')
    
    result1 = rsh.has_messages(wa, {'response version': 2})
    assert result1 == {'result': 'no'}

    result2 = rsh.get_message(wa, {'bytes': '0', 'messages': []})
    assert result2 == {'status': 'messages attached v2', 'messages': []}

    result2 = rsh.get_message(wb, {'bytes': '0', 'messages': []})
    assert result2 == {'status': 'messages attached v2', 'messages': []}
    
def test_retrieve_multi_message_returns_one_then_nothing():
    rsh = RelayServerHandlers(None)
    wa = MyWidget('a')
    ida = idlib.to_binary(wa.get_counterparty_id())
    wb = MyWidget('b')
    idb = idlib.to_binary(wb.get_counterparty_id())

    result0 = rsh.has_messages(wb, {})
    assert result0 == {'result': 'no'}
    
    result1 = rsh.relay_message(wa, {'recipient': idb, 'message': 'spam'})
    assert result1 == {'result': 'ok'}

    result2 = rsh.has_messages(wb, {'response version': 2})
    assert result2 == {'result': 'yes', 'number of messages': '1', 'total bytes': '4', 
        'messages info': [{'sender id': ida, 'message id': idlib.make_id('spam'), 
        'length': 4}]}

    result3 = rsh.get_message(wb, {'bytes': '0', 'messages': result2['messages info']})
    assert result3 == {'status': 'messages attached v2', 'messages': [{'sender id': ida, 
        'message': 'spam'}]}

    result4 = rsh.get_message(wb, {'bytes': '0', 'messages': result2['messages info']})
    assert result4 == {'status': 'messages attached v2', 'messages': []}

def test_retrieve_multi_message_two_messages_from_different_counterparties():
    rsh = RelayServerHandlers(None)
    wa = MyWidget('a')
    ida = idlib.to_binary(wa.get_counterparty_id())
    wb = MyWidget('b')
    idb = idlib.to_binary(wb.get_counterparty_id())
    wc = MyWidget('c')
    idc = idlib.to_binary(wc.get_counterparty_id())

    result0 = rsh.has_messages(wb, {})
    assert result0 == {'result': 'no'}
    
    result1 = rsh.relay_message(wa, {'recipient': idb, 'message': 'spam'})
    assert result1 == {'result': 'ok'}

    result1a = rsh.relay_message(wc, {'recipient': idb, 'message': 'spamc'})
    assert result1 == {'result': 'ok'}

    result2 = rsh.has_messages(wb, {'response version': 2})
    assert result2 == {'result': 'yes', 'number of messages': '2', 'total bytes': '9', 
        'messages info': [
            {'sender id': ida, 'message id': idlib.make_id('spam'), 'length': 4}, 
            {'sender id': idc, 'message id': idlib.make_id('spamc'), 'length': 5}]} or \
        result2 == {'result': 'yes', 'number of messages': '2', 'total bytes': '9', 
        'messages info': [
            {'sender id': idc, 'message id': idlib.make_id('spamc'), 'length': 5},
            {'sender id': ida, 'message id': idlib.make_id('spam'), 'length': 4}]}

    result3 = rsh.get_message(wb, {'bytes': '0', 'messages': result2['messages info']})
    assert result3 == {'status': 'messages attached v2', 'messages': [
        {'sender id': ida, 'message': 'spam'},
        {'sender id': idc, 'message': 'spamc'}]} or \
        result3 == {'status': 'messages attached v2', 'messages': [
        {'sender id': idc, 'message': 'spamc'},
        {'sender id': ida, 'message': 'spam'}]}

    result4 = rsh.get_message(wb, {'bytes': '0', 'messages': result2['messages info']})
    assert result4 == {'status': 'messages attached v2', 'messages': []}

def test_retrieve_multi_message_two_messages_from_the_same_counterparty():
    rsh = RelayServerHandlers(None)
    wa = MyWidget('a')
    ida = idlib.to_binary(wa.get_counterparty_id())
    wb = MyWidget('b')
    idb = idlib.to_binary(wb.get_counterparty_id())
    
    result0 = rsh.has_messages(wb, {})
    assert result0 == {'result': 'no'}

    result1 = rsh.relay_message(wa, {'recipient': idb, 'message': 'spam'})
    assert result1 == {'result': 'ok'}

    result1a = rsh.relay_message(wa, {'recipient': idb, 'message': 'spamc'})
    assert result1 == {'result': 'ok'}

    result2 = rsh.has_messages(wb, {'response version': 2})
    assert result2 == {'result': 'yes', 'number of messages': '2', 'total bytes': '9', 
        'messages info': [
            {'sender id': ida, 'message id': idlib.make_id('spam'), 'length': 4}, 
            {'sender id': ida, 'message id': idlib.make_id('spamc'), 'length': 5}]} or \
        result2 == {'result': 'yes', 'number of messages': '2', 'total bytes': '9', 
        'messages info': [
            {'sender id': ida, 'message id': idlib.make_id('spamc'), 'length': 5},
            {'sender id': ida, 'message id': idlib.make_id('spam'), 'length': 4}]}

    result3 = rsh.get_message(wb, {'bytes': '0', 'messages': result2['messages info']})
    assert result3 == {'status': 'messages attached v2', 'messages': [
        {'sender id': ida, 'message': 'spam'},
        {'sender id': ida, 'message': 'spamc'}]} or \
        result3 == {'status': 'messages attached v2', 'messages': [
        {'sender id': ida, 'message': 'spamc'},
        {'sender id': ida, 'message': 'spam'}]}

    result4 = rsh.get_message(wb, {'bytes': '0', 'messages': result2['messages info']})
    assert result4 == {'status': 'messages attached v2', 'messages': []}
