#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: RequestHandler.py,v 1.18 2001/06/07 22:35:34 zooko Exp $'

# standard modules
from bisect import bisect
import types


# our modules
import DoQ                 # XXX this should not be here, TODO remove all DoQ references in favor of self._mtm._doQ
import MojoTransaction
import debug
from human_readable import hr
import idlib
import MetaTrackerLib
from confutils import confman


class RequestHandler:
    """
    This object stores a set of outstanding requests for service and runs a DoQ loop which selects a
    request to satisfy and then processes it.  You use this by calling `RequestHandler.add()' to add a
    request into the pool. It operates a "doq loop" which repeatedly selects an request to process, removes
    it from the pool, and invokes the appropriate `serverfunc' to process the request.

    RequestHandler implements a "just in time second-price auction".  The "just in time" part means that
    competition between requests *only* kicks in when there are more than one outstanding requests for
    service.  That is, when requests are getting added faster than they are getting processed.  If there is
    only one outstanding request for service, then we do it for free.

    The "second-price auction" part means that the request that came with the highest offer gets to go
    first, but that the requester gets charged as much as the *second-highest* offer.  This strange
    behaviour is actually a good idea based on a well-researched issue in game theory (search for
    "second-price auction" or "Vickrey auction").  The salient fact is that if you charge a requester as
    much as his own offer, then he is motivated to try to figure out or to guess how much the other people
    are offering and then to offer that +1.  If you charge a requester as much as the second-highest offer,
    then he is motivated to simply offer you as much as the service is worth to him -- no more and no less.

    Note that the game theoretic research was in the context of a single-event auction, not a repeated,
    rolling, just-in-time auction like this one, which can be seen as a series of auctions in which
    everyone who lost the Nth auction automatically submits the exact same bid in the N+1th auction.  It
    may well be that this behaviour is suboptimal for this context...

    Note that the desired behavior of "we do it for free as long as nobody else is buying anyway" is
    simply the trivial case of the 2nd-price behavior: if you are the only bidder, then the second-price
    is 0.
    """
    def __init__(self, mtm):
        self._mtm = mtm
        self._reqs = [] # Items of this list are (-1 * mojooffer, firstmsgId, counterparty_id, msgtype, msgbody, serverfunc, staleness_seqnum)
        self._curprice = 0 # the amount to charge for service.  This value gets updated every time self._reqs gets changed.  It is *always* equal to the second price (the price of the index=1 element of self._reqs) or 0 if there are less than 2 requests.
        self._current_staleness_num = 0   # incremented once every self._stale_check_doq_event() run

    def _server_doq_event(self):
        """
        Pick one request from the pool and call `self._serve_request()'.
        Then add a new `_server_doq_event' repeatedly in a doq pseudo-loop.
      
        @precondition This method must be called on the DoQ.: DoQ.doq.is_currently_doq()
        """
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        assert ((len(self._reqs) <= 1) and (self._curprice == 0)) or (self._curprice == (-1 * self._reqs[1][0])), "self._reqs: %s, self._curprice: %s, self._reqs[1][0]: %s" % (hr(self._reqs), hr(self._curprice), hr(self._reqs[1][0]))

        # The following check is necessitated by the time-bounded request pruning strategy.  Without that, we could instead assert that len(self._reqs) > 0 here, since `_server_doq_event()' never gets called at all when there are not outstanding requests.  Not complaining, just a detail for evolvers of this codebase to know...  --Zooko 2001-06-07
        if len(self._reqs) == 0:
            return

        newev = self._reqs.pop(0)
        assert type(newev) is types.TupleType, "newev: %s :: %s" % (hr(newev), hr(type(newev)))
        (negmojooffer, firstmsgId, counterparty_id, msgtype, msgbody, serverfunc, staleness_seqnum) = newev
        mojooffer = -1 * negmojooffer

        if (confman.is_true_bool('MIN_BID_1',)) and (mojooffer > 0) and (self._curprice == 0):
            mojocost = 1
        else:
            # assert mojooffer >= self._curprice # See comment on next line.
            mojocost = min(self._curprice, mojooffer) # the min() would be unnecessary except for the current trick of inserting "important" msgtypes at the front.  Without that trick, you could instead assert that mojoofer >= self._curprice.  --Zooko 2001-06-07
        self._serve_request(firstmsgId, counterparty_id, msgtype, msgbody, mojooffer, serverfunc, mojocost=mojocost)

        lenreqs = len(self._reqs)
        if lenreqs == 0:
            assert(self._curprice == 0)
        elif lenreqs == 1:
            self._curprice = 0
            DoQ.doq.add_task(self._server_doq_event)
        else:
            assert len(self._reqs) >= 2, "self._reqs: %s" % hr(self._reqs)
            assert len(self._reqs[1]) >= 1, "self._reqs[1]: %s" % hr(self._reqs[1])
            self._curprice = -1 * self._reqs[1][0]
            DoQ.doq.add_task(self._server_doq_event)
    
    def _stale_check_doq_event(self):
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."
        if int(confman['MAX_REQUEST_HANDLER_DELAY']) <= 0:
            return
        if self._mtm._shuttingdownflag:
            return

        # TODO rather than toss out stale requests based -solely- on our own setting of how long we'll allow a
        # request to "age" we could accept a suggested timeout from the requestor and toss things out after
        # min(suggested_timeout, max_request_handler_delay) time has elapsed without processing.

        starting_reqs_len = len(self._reqs)
        # remove requests from self._reqs that are too old (marked with an old staleness value)
        def keep_this_request(reqtuple, min_staleness=self._current_staleness_num-1, _cm=self._mtm._cm):
            keep = (reqtuple[6] >= min_staleness)
            if not keep:
                # TODO potentially send a tiny "mojo offer too low" response to some requests (not all,
                # remember that we are doing this to decrease resource usage by not processing requests
                # that are too old).  [*never* send a "mojo offer too low" response on a 0 mojooffer
                # request, that would make denial of service free]

                # clean up internal state about this dropped request
                _cm.drop_request_state(reqtuple[1])
            return keep
        self._reqs = filter(keep_this_request, self._reqs)

        num_dropped = starting_reqs_len - len(self._reqs)
        if num_dropped:
            debug.mojolog.write("%s dropped %s stale messages\n", args=(self._mtm.get_id(), num_dropped,), v=3, vs="RequestHandler")
            # Reset self._curprice
            if len(self._reqs) > 1:
                self._curprice = -1 * self._reqs[1][0]
            else:
                self._curprice = 0

        self._current_staleness_num = self._current_staleness_num + 1
        self._mtm._doQ.add_task(self._stale_check_doq_event, delay=max(10, int(confman['MAX_REQUEST_HANDLER_DELAY']))/2)

    def add(self, firstmsgId, counterparty_id, msgtype, msgbody, mojooffer, serverfunc):
        """
        @precondition This method must be called on the DoQ.: DoQ.doq.is_currently_doq()
        """
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        # We use the negative value of `mojooffer' in order to get the old python < 2.1 `bisect' to do the right thing
        # (that is: when there are two requests with the same mojooffer, the earliest-added one remains at the front
        # of the queue instead of the most recently added one going to the front of the queue).
        newev = (-1 * mojooffer, firstmsgId, counterparty_id, msgtype, msgbody, serverfunc, self._current_staleness_num)

        if len(self._reqs) == 0:
            self._reqs.append(newev)
            self._curprice = 0 # If there is nobody else waiting in line, then our price is 0.
            DoQ.doq.add_task(self._server_doq_event)
            return

        if msgtype in [ "payment", "message for you", "hello", "lookup contact info", ]:
            # This needs to be evolved a bit more, but for now putting these four messages at the very front of the queue seems like the safe thing to do for today's stable release.  --Zooko 2001-06-07
            index = 0
        else:
            index = bisect(self._reqs, newev)

        if index == 0:
            # If this is the highest offer, then whatever is now the 2nd highest offer is our new self._curprice.
            self._curprice = -1 * self._reqs[0][0]
        elif index == 1:
            # If this is the second-highest offer, then it is our new self._curprice.  See doco of "2nd price auction" in RequestHandler.RequestHandler.__doc__.
            self._curprice = mojooffer

        self._reqs.insert(index, newev)

    def _serve_request(self, firstmsgId, counterparty_id, msgtype, msgbody, mojooffer, serverfunc, mojocost):
        """
        Perform the given service and send the result back to the requesting counterparty.

        @precondition This method must be called on the DoQ.: DoQ.doq.is_currently_doq()
        """
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        counterparty_id = idlib.canonicalize(counterparty_id, "broker")
        cp=self._mtm._keeper.get_counterparty_object(counterparty_id)
        # `balance' is the amount that you owe me.  (Hint: I will reject you if it is larger than your credit limit.)
        if mojocost > 0:
            (ok, limit, balance)=cp.charge(mojocost, reasonstr=msgtype)
            if not ok:
                self._mtm.respond_with(firstmsgId, msgbody=None, mojoheader={ 'result': "failure", 'reason': "credit limit reached", "your credit limit": limit, "amount you owe me": balance, "price of this service": mojocost, "amount of offer claimed": 0 }, _extras={'mojo refund': mojooffer })
                # XXX Note: In the future, we will use `amount of offer claimed', in the mojo header, instead of `mojo refund', in the mojo message.  The former is absolute, the latter depends on what they offered.  Note that currently we pass `mojooffer' in here solely to include it in this rejection letter for backwards compatibility.  --Zooko 2001/05/23; version in my sandbox calls itself v0.985.
                return

        else:
            # this log message is similar to the one that the above charge() call prints in counterparties.py
            debug.mojolog.write("accounting with %s: performing free transaction [reason: %s]\n", args=(counterparty_id, msgtype), v=1, vs="counterparty")

        widget = MojoTransaction.Widget(counterparty_id, firstmsgId)
        # Okay, now actually invoke the server func:
        result = serverfunc(widget, msgbody['mojo message'])

        if result is MojoTransaction.NO_RESPONSE:
            self._mtm._cm.drop_request_state(firstmsgId)
            return

        if result is MojoTransaction.ASYNC_RESPONSE:
            return
           
        mojoheader = { 'mojo offer': 0, 'amount of offer claimed': mojocost }
        # "This 'mojo offer': 0 is just for backwards compatibility with old clients that template check that every mojo header has either a mojo offer or else "result": 'failure'.  --Zooko 2001-06-05
        if mojooffer > mojocost:   # this old (pre 0.986.2) style refund can go away once enough brokers have upgraded
            old_style_refund = { 'mojo refund': mojooffer - mojocost }
        else:
            old_style_refund = None

        self._mtm.respond_with(firstmsgId, result, mojoheader=mojoheader, _extras=old_style_refund)

