#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

import traceback
import sys
import threading
import os
import types


__doc__ = """
A testing tool, inspired by JUnit and utilizing a lot of 
introspection

Contains the following functions, which have separate doc strings
runTests()
runAll()
get_next_port()
get_next_dir_name()
assert_testing()
"""

def runTests(modules = None, filename = None, flag = 'mojo_test_flag', directory =
             '/var/tmp/RunTests/', verboselevel="spew everything please", testalldeps=None) :
    """
    Runs a bunch of tests. Prints a dot to stdout for every test passed. 
    Prints stack traces thrown by tests, and messages when they fail. 
    When done, prints either 'everything passed' or 'the following tests 
    failed:' followed by a list of tests which failed.

    A module is tested by testing all the appropriate functions and classes 
    in it.

    Functions (as opposed to methods) which start with 'test' are tested by 
    running them passing no arguments. If the return value is None or 
    something which means 'true' it passed. If it raised an exception or 
    returned something other than None which means 'false', it failed.

    Classes whose name ends in 'Test' have each method whose name starts with 
    'test' tested. Methods are tested by first creating an instance 
    (including calling the constructor,) then calling the setUp() method if 
    there is one, then calling the test method with no parameters, then 
    calling the tearDown() method if there is one. Method return values are 
    interpreted the same way as function return values.

    Named parameters - 
    
    modules - a list strings which are names of modules to be tested, or a 
    single module name. The default is to test everything currently imported
    
    filename - a file to put output in, for each test executed either 
    'passed spam' or 'failed spam' is appended to the file, with possibly 
    some messages of 'failed eggs' for modules named eggs which couldn't 
    be imported at all
    
    flag - the name of a variable to check for in all modules. Modules not 
    containing this variable name won't be tested (this is to avoid hitting 
    incompatible test code with a similar naming scheme.) Defaults to 
    'mojo_test_flag'
    
    directory - the directory which temporary subdirectories (created with 
    get_next_dir_name()) are made in. Anything currently in the directory 
    is assumed to be junk from previous test runs which wasn't cleaned up 
    properly and deleted. Everything is removed after tests run as well.

    verboselevel - options include "spew everything please" and
    "just dots to stdout but results to file please"
    """
    if type(modules) == types.StringType :
        modules = [modules]
    if modules is None :
        modules = sys.modules.keys()
    assert type(modules) == types.ListType
    assert type(flag) == types.StringType
    if filename is not None :
        assert type(filename) == types.StringType
    # commented this out because RunTests doesn't need randsource
    #try :
    #    import randsource
    #except ImportError, e :
    #    pass
    __DoTests(modules, filename, flag, directory, verboselevel, testalldeps)

class __DoTests :
    def __init__(self, modules, filename, flag, directory, verboselevel, testalldeps) :
        self._verboselevel = verboselevel
        self._testalldeps = testalldeps

        _testing[0] = 1
        _dir[0] = directory
        try :
            os.makedirs(directory)
        except OSError :
            pass
        self.flag = flag
        self.file = None
        if filename is not None :
            self.file = open(filename,'w')
        self.failed_tests = []
        self.modules = modules
        _clear_dirs()
        self.test_modules()
        self.clean_up()

    def clean_up(self) :
        if len(self.failed_tests) == 0 :
            sys.stdout.write('TEST: everything passed\n')
            sys.stdout.flush()
        else :
            sys.stdout.write('TEST: the following tests failed:\n')
            sys.stdout.flush()
            for i in self.failed_tests :
                sys.stdout.write('TEST: failed ' + i + '\n')
                sys.stdout.flush()

        if self.file is not None :
            self.file.close()
        try :
            _clear_dirs()
        except OSError :
            traceback.print_exc
        _testing[0] = 0

    def test_modules(self) :
        modnames = list(self.modules)
        modnames.sort()
        oldmods = []
        for m in modnames :
            try :
                oldmod = __import__(m)
                oldmods.append(oldmod)
            except :
                self.add_failure(m, type='module')

        if self._testalldeps:
            newmodnames = sys.modules.keys()
            newmodnames.sort()
            newmods = []
            for newmodname in newmodnames:
                newmod = sys.modules.get(newmodname)
                if newmod:
                    newmods.append(newmod)
          
            for newmod in newmods :
                self.test_module(newmod)
        else:
            for oldmod in oldmods :
                self.test_module(oldmod)

    def _find_all_teardowns(self):
        tds = []
        for mod in sys.modules.values():
            if mod:
                td = mod.__dict__.get("_test_tear_down")
                if td:
                    tds.append(td)

        return tds

    def test_function(self, module, function):
        try:
            if self._verboselevel != "just dots to stdout but results to file please":
                print 'TEST: trying ' + module.__name__ + '.' + function.__name__ + '()'

            result = function()

            if result is not None :
                assert result
            self.add_success(module.__name__ + '.' + function.__name__)
        finally:
            for td_func in self._find_all_teardowns():
                td_func()
        
    def test_module(self, module) :
        # to avoid invoking other incompatible test code which uses a similar naming convention
        if not hasattr(module, self.flag) :
            return
        possible_tests = module.__dict__.keys()
        possible_tests.sort()
        for pt in possible_tests :
            ptname = module.__name__ + '.' + pt
            next = module.__dict__[pt]
            try :
                if pt[-4:] == 'Test' :
                    if type(next) == types.ClassType :
                        self.test_class(next, ptname)
                elif pt[:4] == 'test' :
                    if type(next) == types.FunctionType :
                        self.test_function(module, next)
            except :
                self.add_failure(ptname)

    def test_class(self, classobject, classname) :
        ms = classobject.__dict__.keys()
        ms.sort()
        for m in ms :
            if m[:4] == 'test' and type(classobject.__dict__[m]) == types.FunctionType:
                methodname = classname + '.' + m
                if self._verboselevel != "just dots to stdout but results to file please":
                    print 'TEST: trying ' + methodname + '()'
                try :
                    instance = classobject()
                    if hasattr(instance, 'setUp') :
                        instance.setUp()
                    # call the particular test method
                    result = classobject.__dict__[m](instance)
                    if result is not None :
                        assert result
                    self.add_success(methodname)
                    if hasattr(instance, 'tearDown'):
                        instance.tearDown()
                except :
                    self.add_failure(methodname)

    def add_failure(self, current_test, type='function') :
        if type == 'function' :
            outline = 'TEST: ' + current_test + '() -> failed\n'
        elif type == 'module' :
            outline = 'TEST: module ' + current_test + ' -> failed\n'
        else:
            raise Exception, 'Test failure type "%s" unknown.' % type

        if self._verboselevel == "just dots to stdout but results to file please":
            sys.stdout.write(".")
            sys.stdout.flush()
        else:
            sys.stdout.write(outline)
            sys.stdout.flush()
        if self.file is not None :
            self.file.write(outline)
        traceback.print_exc()
        self.failed_tests.append(current_test)

    def add_success(self, current_test) :
        outline = 'TEST: ' + current_test + '() -> passed\n'
        if self._verboselevel == "just dots to stdout but results to file please":
            sys.stdout.write(".")
            sys.stdout.flush()
        else:
            sys.stdout.write(outline)
            sys.stdout.flush()
        if self.file is not None :
            self.file.write(outline)

# yes, we drink our own Kool-Aid
mojo_test_flag = 1

_dir = ['/var/tmp/RunTests/', 0]

def _clear_dirs(dirname = _dir[0]) :
    for f in os.listdir(dirname) :
        fullname = os.path.join(dirname, f)
        if os.path.isdir(fullname) :
            _clear_dirs(fullname)
            os.rmdir(fullname)
        else :
            os.remove(fullname)

__port = [20000]

__lock = threading.Lock()

_testing = [0]

def is_testing() :
    return _testing[0]

def assert_testing() :
    """
    A nice method to call from functionality which is for testing only - 
    asserts false if not in the middle of a call to runTests()
    """
    assert _testing[0]

def get_next_dir_name() :
    """
    Creates and returns the name of a temporary directory which can be 
    used by databases and the like
    """
    __lock.acquire()
    try :
        name = os.path.join(_dir[0], str(_dir[1]))
        os.makedirs(name)
        return name
    finally :
        _dir[1] = _dir[1] + 1
        __lock.release()

def get_next_port() :
    """
    Returns a port number which wasn't returned to anyone else - useful for 
    tests of servers
    """
    __lock.acquire()
    try :
        return __port[0]
    finally :
        __port[0] = __port[0] + 1
        __lock.release()

class PassTest :
    def test(self) :
        pass
        
    def test2(self) :
        return 1

def test_one() :
    pass

def test_return() :
    return 1

def test_different_ports() :
    assert type(get_next_port()) == type(0)
    assert get_next_port() != get_next_port

def test_different_dirs() :
    assert get_next_dir_name() != get_next_dir_name()

def test_assert_testing() :
    try :
        assert_testing()
        return 1
    except AssertionError :
        traceback.print_exc()
        return 0



