#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import threading
import time

### our modules
import BidStrategy

from blocksharez import split, unsplit
from MojoErrors import ReassemblyError
from idlib import make_ascii_id
import RunTests
from DoQ import DoQ
from copy import copy
from human_readable import hr
import idlib
import debug
import const
import localblockstore
import mojoutil
import pricelib
import MojoConstants
import MojoTransaction

true = 1
false = 0

# maps from pieceid to SingleReassembly instance
if MojoConstants.DEBUG_MODE:
    global debugmap
    debugmap = {}

class SingleReassembly:
    def __init__(self, broker, do_function, find_function, finder, 
        cancel_find_function, download_done_function, block_positions, 
        num_blocks_needed, piece_id, piece_length, priority = 0, max_concurrent_blobs=2,
        callback = (lambda reassembly: None), mtm=None, allow_download_cache=1, download_monitor=None):
        """
        Creating a SingleReassembly starts a single download. The callback is called 
        once enough stuff has been downloaded or it's no longer possible to download 
        the number of necessary blocks.

        The callback might happen almost immediately! To keep it from happening while 
        in the constructor you must either get a lock while call the constructor and processing 
        the callback, or call the constructor from inside the doq.

        You should *not* construct multiple SingleReassemblys for the same piece!
        They might decide on different sets of blocks to download and do redundant work. Instead, 
        create a single SingleReassembly and later on call it's reassemble() method.

        block_positions contains ids to be reconstructed in order, or None for ones which are unavailable
        
        num_blocks_needed is the number of blocks necessary to reconstruct the piece
        
        priority is a number, the higher it is the greater the priority. Default is zero.

        find_function takes a block_id and a callback as parameters (see BlockFinder.callback_when_found)

        cancel_find_function takes a block_id as a parameter (see BlockFinder.cancel)

        download_monitor is an object implementing the DownloadMonitor interface or None
        """
        debug.mojolog.write("To reconstruct chunk: %s, trying to find %s of these shares: %s\n", args=(piece_id, num_blocks_needed, block_positions,), v=6, vs="SingleReassembly")

        if MojoConstants.DEBUG_MODE:
            global debugmap
            if debugmap.has_key(piece_id):
                debug.mojolog.write("Debug Warning: Creating a second SingleReassembly for the same pieceid: %s\n" % hr(piece_id), vs='SingleReassembly')
            debugmap[piece_id] = self

        self.broker = broker
        assert self.broker.doq.is_currently_doq(), "SingleReassembly objects must be instantiated on the DoQ"
        self.finder = finder
        self.do_function = do_function
        self.find_function = find_function
        self.cancel_find_function = cancel_find_function
        self.num_blocks_needed = num_blocks_needed
        self.num_blocks_total = len(block_positions)
        self.block_positions = block_positions
        # this map is used to prevent linear searches of block_positions
        self.blockid_pos_map = {} # blockid -> integer block num (idx of block_positions, blocks, done_finding, & currently_downloading)
        for x in xrange(len(self.block_positions)):
            self.blockid_pos_map[idlib.to_binary(self.block_positions[x])] = x
            self.blockid_pos_map[idlib.to_ascii(self.block_positions[x])] = x

        self.piece_id = piece_id
        self.piece_length = piece_length
        self.priority = priority
        self.callback = callback
        self.download_done_function = download_done_function
        self.download_monitor = download_monitor
        
        self.block_length = self._compute_share_block_size(self.piece_length)
        self._finflag = threading.Event()
        self.piece = None

        self.done_finding = [false] * len(block_positions)
        self.currently_downloading = [false] * len(block_positions)
        # either None or the block corresponding to the same index of block_positions
        self.blocks = [None] * len(block_positions)
        # {block id: [server ids]}
        self.tried_servers = {}
        self.max_concurrent_request_blobs = max_concurrent_blobs
        self.block_store = {}
        if self.download_monitor:
            # block ids are keyed in here when a find says they exist (to determine which pieces are possible to download)
            self.block_has_location = {}
        
        if allow_download_cache:
            num_still_need = self.num_blocks_needed
            for block_id in self.block_positions:
                if block_id is None:
                    continue
                try:
                    if self.broker.downloadcache.get_block_size(block_id) > 0:
                        debug.mojolog.write('pieceid %s Getting block %s from download cache\n', args=(self.piece_id, block_id,), v=4, vs='SingleReassembly')
                        blobfile = self.broker.downloadcache.get_block_file(block_id)
                        if blobfile is not None:
                            blob = blobfile.read()
                            blobfile.close()
                            self.blocks[self.blockid_pos_map[block_id]] = blob
                            num_still_need = num_still_need - 1
                            if num_still_need <= 0:
                                self.broker.doq.add_batch_task('download', self.cancel)
                                return
                except(localblockstore.Hell, IOError), e:
                    debug.mojolog.write('WARNING get_block_file (blob file) exception: %s\n' % e.args)
        self.broker.doq.add_batch_task('download', self._start)
   
    def __del__(self):
        if MojoConstants.DEBUG_MODE:
            global debugmap
            if debugmap:
                if not debugmap.has_key(self.piece_id):
                    debug.mojolog.write("Debug warning: deleting SingleReassembly that was never created, piece_id: %s\n", args=(self.piece_id,), v=0, vs="debug")
                    return
                if not debugmap[self.piece_id] == self:
                    debug.mojolog.write("Debug warning: deleting SingleReassembly that wasn't the one in the debugmap, piece_id: %s\n", args=(self.piece_id,), v=0, vs="debug")
                del debugmap[self.piece_id]

        if hasattr(self, 'download_monitor') and self.download_monitor:
            self.download_monitor.ev_discarded(self.piece_id)

    def _start(self):
        assert self.broker.doq.is_currently_doq()
        if self._finflag.isSet():
            return
        for i in xrange(len(self.block_positions)):
            if self.block_positions[i] is not None and self.blocks[i] is None:
                self.find_function(block_id = self.block_positions[i], 
                    found_callback = self._find_found, 
                    finished_callback = self._find_finished)
            else:
                self.done_finding[i] = true
        self.broker.doq.add_batch_task('download', self.do_function, args = (self,))

    def _find_found(self, block_id):
        # callback for find_function
        assert self.broker.doq.is_currently_doq()
        if self._finflag.isSet():
            return
        debug.mojolog.write("Find Found %s\n" % block_id, v=12, vs="SingleReassembly")
        if self.download_monitor:
            self.download_monitor.ev_found_block(self.piece_id, self.blockid_pos_map[block_id])
            self.block_has_location[block_id] = None
            if len(self.block_has_location) == self.num_blocks_needed:
                self.download_monitor.ev_reassembly_possible(self.piece_id)
        self.broker.doq.add_batch_task('download', self.do_function, args = (self,))

    def _find_finished(self, block_id):
        debug.mojolog.write("Find Finished %s\n" % block_id, v=12, vs="SingleReassembly")
        assert self.broker.doq.is_currently_doq()
        if self._finflag.isSet():
            return
        self.done_finding[self.blockid_pos_map[block_id]] = true
        self.broker.doq.add_batch_task('download', self.do_function, args = (self,))

    def set_priority(self, priority):
        self.broker.doq.add_batch_task('download', self._set_priority, args=(priority,))

    def _set_priority(self, priority):
        assert self.broker.doq.is_currently_doq()
        if self._finflag.isSet():
            return
        self.priority = priority
        self.do_function(self)

    def reassemble(self):
        """
        this method will block until enough shares are available or if it can't be reassembled.

        @precondition This method must not be called on the DoQ (as it might block).: not self.broker.doq.is_currently_doq()
        """
        assert not self.broker.doq.is_currently_doq(), "precondition: This method must not be called on the DoQ (as it might block)."

        debug.mojolog.write("reassemble(): %s._finflag = %s\n", args=(self, self._finflag.isSet()), v=4, vs='SingleReassembly')
        self._finflag.wait()
        return self.nonblocking_reassemble()

    def nonblocking_reassemble(self):
        if self.piece is None:
            # if finflag was set and piece wasn't assigned then failure_reason must have been assigned
            if not hasattr(self, 'failure_reason'):
                return ReassemblyError, "Find not finished."
            else:
                raise ReassemblyError, self.failure_reason
        return self.piece

    def consolidate(self):
        assert self.broker.doq.is_currently_doq()
        if self._finflag.isSet():
            return
        blocks = []
        for i in xrange(len(self.blocks)):
            if self.blocks[i] is not None:
                blocks.append((i, self.blocks[i]))
        if len(blocks) < self.num_blocks_needed:
            raise ReassemblyError, 'not enough blocks found - need '+str(self.num_blocks_needed)+', have ' + str(len(blocks))
        result = unsplit(blocks, self.piece_id, length = self.piece_length, totalsharesneeded=self.num_blocks_needed, totalsharesinpiece=self.num_blocks_total)
        if self.download_monitor:
            self.download_monitor.ev_finished_reassembly(self.piece_id)
        
        # check the list of shares, if any of the shares in this
        # piece match our block servers mask but we did -not- download
        # them, use split to regenerate them here and store the
        # matching ones on our block server.  Thanks goes to Donwulff
        # on irc for this great idea!  (we'll call this Donwulff block storing)
        if self.broker.blobserver is not None:
            debug.mojolog.write("checking to see if any blocks can be reassembled to fit on our block server\n", v=4, vs='SingleReassembly')
            blocks_to_recreate_and_store = []  # list of tuples (sharenum, block_id)
            for idx in xrange(len(self.block_positions)):
                idx_and_id = idx, self.block_positions[idx]
                # see if we could use it on our block server
                block_id = idx_and_id[1]
                debug.mojolog.write("  checking block %s\n", args=(block_id,), v=6, vs='SingleReassembly')
                if self.broker.multiblockstore.get_block_size(block_id) > 0:
                    continue   # we already have this block, next please!
                for mask in self.broker.blobserver.masklist:
                    if mask.match(idlib.to_binary(block_id)):
                        debug.mojolog.write("  going to try and get %s via reassembly\n", args=(block_id,), v=5, vs='SingleReassembly')
                        blocks_to_recreate_and_store.append(idx_and_id)
                        break
                    pass
            if len(blocks_to_recreate_and_store) > 0:
                splitblocks = split(result, totalsharesneeded=self.num_blocks_needed, totalsharesinpiece=self.num_blocks_total)
                for sharenum, supposed_shareid in blocks_to_recreate_and_store:
                    if idlib.equal(splitblocks[sharenum][0], supposed_shareid):
                        debug.mojolog.write("  storing %s after reassembly\n", args=(splitblocks[sharenum][0],), v=4, vs='SingleReassembly')
                        self.broker.blobserver.try_to_store_blob(splitblocks[sharenum][0], splitblocks[sharenum][1])

        self.piece = result
        self.blocks = [None] * len(self.block_positions)

    def is_finished(self):
        return self._finflag.isSet()

    def get_priority(self):
        return self.priority

    def cancel(self, reason=None):
        """
        @precondition This method must be called on the DoQ.: self.broker.doq.is_currently_doq()
        """
        assert self.broker.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        if self._finflag.isSet():
            return
        try:
            self.consolidate()
        except ReassemblyError, e:
            debug.mojolog.write("pieceid %s failed: %s\n", args=(self.piece_id,e), v=3, vs="SingleReassembly")
            if not reason:
                reason = str(e)
            if self.download_monitor:
                self.download_monitor.ev_failed_reassembly(self.piece_id, reasonstr=reason)
            self.failure_reason = reason
        self._finflag.set()
        for block_id in self.block_positions:
            if block_id is not None:
                self.broker.doq.add_batch_task('download', self.cancel_find_function, args = (block_id,))
        self.broker.doq.add_batch_task('download', self.do_function, args = (self,))
        self.broker.doq.add_batch_task('download', self.callback, kwargs = {'reassembly': self})

    def get_piece_id(self):
        return self.piece_id

    # XXX this is a kludge for now
    def _compute_share_block_size(self, piece_len):
        # return the next greater size
        # note: PieceSizes is a sorted list
        for size in const.PieceSizes :
            if size >= piece_len:
                return size / self.num_blocks_needed

    def start_next(self, current_servers):
        """
        current_servers is [server id] - it is modified by this function (appended to).  it is intended to be a list of
        server_ids that we're currently downloading blocks from so that we don't send multiple request
        blobs to the same server at once (as that doesn't help streaming and causes timeouts on the later requests).
        """
        if self._finflag.isSet():
            return false
        num_currently_downloading = 0
        for i in self.currently_downloading:
            if i:
                num_currently_downloading = num_currently_downloading + 1
        num_still_need = self.num_blocks_needed
        for i in self.blocks:
            if i is not None:
                num_still_need = num_still_need - 1
        if num_currently_downloading >= num_still_need:
            return false

        completely_failed = [true] * len(self.block_positions)
        for i in xrange(len(self.done_finding)):
            if not self.done_finding[i]:
                completely_failed[i] = false
            elif self.blocks[i] is not None:
                completely_failed[i] = false

        for block_index in xrange(len(self.block_positions)):
            if self.currently_downloading[block_index]:
                completely_failed[block_index] = false
                continue
            if self.blocks[block_index] is not None:
                continue
            block_id = self.block_positions[block_index]
            if not self.tried_servers.has_key(block_id):
                self.tried_servers[block_id] = []

            message_type = 'request blob'
            message_body = {'blob id': block_id, 'len': self.block_length}
            infolist = self.finder.get_block_locations_from_cache(block_id)
            infolist = self.broker.ServicesMTM.get_handicapper().sort_by_preference(infolist, message_type, message_body)
            infolist2 = []
            for infotuple in infolist:
                # if infotuple[0] in self.tried_servers[block_id]:   # [ XXX the old statement that tries a given broker only once ]
                if self.tried_servers[block_id].count(infotuple[0]) >= 2:
                    # skip brokers that we've already tried to get this block from more than once
                    continue
                if infotuple[0] in current_servers:
                    completely_failed[block_index] = false
                else:
                    infolist2.append(infotuple)
            infolist = infolist2
            if len(infolist) == 0:
                # no new places to try to get this block from
                continue

            server_id = infolist[0][0]
            self.tried_servers[block_id].append(server_id)
            current_servers.append(server_id)
            self.currently_downloading[self.blockid_pos_map[block_id]] = true

            server_id, server_info = infolist[0]
            fixedprice = pricelib.blob_server_purchase_price(server_info, message_body)
            if fixedprice is not None:
                # Remove this once we've grandfathered-out most of the fixed-price MTs...  --Zooko 2001-06-01
                mojooffer = fixedprice
            else:
                mojooffer = BidStrategy.choose_bid("MO_REQUEST_BLOB_DOWNLOADING", self.broker.ServicesMTM)

            connection_speed_bytes_per_second = 1800  # 1800 should be true even if the counterparty is on a poor 28.8k connection
            # timeouts are always at least 60 just due to general software load and send/receive via relay issues
            maxtime = min(max(self.block_length / (connection_speed_bytes_per_second / 
                self.max_concurrent_request_blobs / 2), 60), MojoConstants.DEFAULT_TIMEOUT)

            timer = mojoutil.Timer()
            def process_callback(widget, outcome = None, failure_reason = None, block_id = block_id, block_index = block_index,
                    timer = timer, self = self, current_servers = current_servers, server_id = server_id):
                """
                this 'request blob response' callback is idempotent; calling it more than once per initiated transaction is okay.

                @precondition This method must be called on the DoQ.: self.broker.doq.is_currently_doq()
                """
                assert self.broker.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

                binary_block_id = idlib.to_binary(block_id)
                ascii_block_id = idlib.to_ascii(block_id)
                del block_id
                timervalue = timer.get()  # don't use timer.stop() so that this is idempotent incase the first callback is a timeout failure
                self.broker.doq.add_batch_task('download', self.download_done_function)
                current_servers.remove(server_id)
                self.currently_downloading[self.blockid_pos_map[binary_block_id]] = false

                if self._finflag.isSet():
                    return
                if failure_reason is not None:
                    return
                if (outcome is None) or (outcome.get('result') == 'failure'):
                    return
                block_data = outcome['data']
                if not idlib.equal(binary_block_id, idlib.make_id(block_data)):
                    return

                if self.download_monitor:
                    self.download_monitor.ev_finished_block(self.piece_id, block_index)
                debug.mojolog.write("block_finished keeping block_id %s in memory\n", args=(ascii_block_id,), v=3, vs="SingleReassembly")
                self.broker.doq.add_batch_task('download', self.cancel_find_function, args = (ascii_block_id,))
                self.blocks[self.blockid_pos_map[binary_block_id]] = block_data
                num_still_need = self.num_blocks_needed
                for i in self.blocks:
                    if i is not None:
                        num_still_need = num_still_need - 1
                if num_still_need <= 0:
                    self.cancel()

                # store the latency overall and on a per-block size basis
                cp_obj = self.broker.ServicesMTM._keeper.get_counterparty_object(server_id)
                cp_obj.update_custom_stat_weighted_sample('request blob bytes/sec', float(len(block_data))/timervalue)

                # store the blob on our block server's localblockstore if were running one and it matches our mask
                if self.broker.blobserver is not None:
                    if self.broker.multiblockstore.get_block_size(ascii_block_id) <= 0 and self.broker.blobserver.try_to_store_blob(ascii_block_id, block_data):
                        debug.mojolog.write("stored block %s for our server.\n", args=(ascii_block_id,), v=5, vs="SingleReassembly")
                # store the block in our download cache localblockstore if we have one
                try:
                    if self.broker.downloadcache:
                        tempF = self.broker.downloadcache.open_tempfile(ascii_block_id, len(block_data))
                        tempF.write(block_data)
                        tempF.close()
                        self.broker.downloadcache.store_tempfile(ascii_block_id, tempF)
                except localblockstore.NoSpaceError:
                    pass

            if self.download_monitor:
                self.download_monitor.ev_started_block(self.piece_id, block_index)
            timer.start()
            # if we're sending the request to ourself, cheat and avoid initiate's overhead.  (XXX initiate should really do this for us)
            if idlib.equal(server_id, self.broker.ServicesMTM.get_id()) and self.broker.blobserver:
                widget = MojoTransaction.Widget(server_id)
                try:
                    outcome = self.broker.blobserver.handle_request_blob(widget, message_body)
                    failure_reason = None
                except:
                    debug.mojolog.write("Caught exception trying to get blob from ourself:\n", vs='SingleReassembly')
                    traceback.print_exc(file=debug.mojolog)
                    outcome = None
                    failure_reason = 'exception'
                self.broker.doq.add_task(process_callback, kwargs={
                        'widget' : widget,  'outcome': outcome, 'failure_reason': failure_reason,
                    })
                return true
            # send the request
            self.broker.ServicesMTM.initiate(server_id, message_type, message_body, mojooffer, outcome_func = process_callback, timeout=maxtime, post_timeout_outcome_func=process_callback)
            return true
        num_completely_failed = 0
        for i in completely_failed:
            if i:
                num_completely_failed = num_completely_failed + 1
        if num_completely_failed + self.num_blocks_needed > len(self.block_positions):
            debug.mojolog.write("NARF: cancelling, completely_failed = %s, self.piece_id: %s, \n", args=(completely_failed, self.piece_id,), v=3, vs='SingleReassembly')
            self.cancel()
        return false            



def test_standard_operation():
    q = DoQ()
    block_store = {}
    
    do_results = []
    finds = []
    

    def do_function(reassembly, do_results = do_results):
        if reassembly.is_finished():
            do_results.append('finished')
            return
        while 1:
            next = reassembly.next_block_id()
            do_results.append(next)
            if next is None:
                return
    
    def find_function(block_id, callback, finds = finds):
        finds.append((block_id, callback))
    
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    block_positions = map(lambda x: x[0], l)
    block_positions[1] = None
    
    done_flag = threading.Event()
    
    block_store[l[0][0]] = l[0][1]

    callback = (lambda reassembly, done_flag = done_flag: done_flag.set())

    s = SingleReassembly(do_function = do_function, find_function = find_function, cancel_find_function = lambda x: None, doq = q,
        block_positions = block_positions, num_blocks_needed = 4, piece_id = make_ascii_id(crap), 
        piece_length = len(crap), callback = callback, block_store = block_store)

    def flush(q = q):
        for i in xrange(3):
            f = threading.Event()
            q.add_task(f.set)
            f.wait()

    flush()
    
    assert do_results == [None]
    assert len(finds) == 6
    ffs = map(lambda x: x[0], finds)
    ffs.sort()
    bcopy = copy(block_positions)
    bcopy = bcopy[2:]
    bcopy.sort()
    assert ffs == bcopy
    assert not done_flag.isSet()

    q.add_task(finds[0][1], kwargs = {'block_id': finds[0][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[0][0], None]
    assert not done_flag.isSet()
    
    q.add_task(finds[1][1], kwargs = {'block_id': finds[1][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[0][0], None, finds[1][0], None]
    assert not done_flag.isSet()
    
    q.add_task(finds[2][1], kwargs = {'block_id': finds[2][0], 'failure_reason': "it didn't work and stuff"})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[0][0], None, finds[1][0], None]
    assert not done_flag.isSet()
    
    q.add_task(finds[3][1], kwargs = {'block_id': finds[3][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[0][0], None, finds[1][0], None, finds[3][0], None]
    assert not done_flag.isSet()

    q.add_task(s.block_finished, kwargs = {'block_id': finds[0][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[0][0], None, finds[1][0], None, finds[3][0], None]
    assert not done_flag.isSet()

    q.add_task(s.block_finished, kwargs = {'block_id': finds[1][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[0][0], None, finds[1][0], None, finds[3][0], None]
    assert not done_flag.isSet()

    q.add_task(s.block_finished, kwargs = {'block_id': finds[3][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[0][0], None, finds[1][0], None, finds[3][0], None, 'finished']
    assert done_flag.isSet()

def test_finds_fail():
    q = DoQ()
    block_store = {}
    
    do_results = []
    finds = []
    
    def do_function(reassembly, do_results = do_results):
        if reassembly.is_finished():
            do_results.append('finished')
            return
        while 1:
            next = reassembly.next_block_id()
            do_results.append(next)
            if next is None:
                return
    
    def find_function(block_id, callback, finds = finds):
        finds.append((block_id, callback))
    
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    block_positions = map(lambda x: x[0], l)
    block_positions[1] = None
    
    done_flag = threading.Event()
    
    block_store[l[0][0]] = l[0][1]

    callback = (lambda reassembly, done_flag = done_flag: done_flag.set())

    s = SingleReassembly(do_function = do_function, find_function = find_function, cancel_find_function = lambda x: None, doq = q,
        block_positions = block_positions, num_blocks_needed = 4, piece_id = make_ascii_id(crap), 
        piece_length = len(crap), callback = callback, block_store = block_store)

    def flush(q = q):
        for i in xrange(3):
            f = threading.Event()
            q.add_task(f.set)
            f.wait()

    flush()
    
    assert do_results == [None]
    assert len(finds) == 6
    ffs = map(lambda x: x[0], finds)
    ffs.sort()
    bcopy = copy(block_positions)
    bcopy = bcopy[2:]
    bcopy.sort()
    assert ffs == bcopy
    assert not done_flag.isSet()

    q.add_task(finds[0][1], kwargs = {'block_id': finds[0][0], 'failure_reason': "He's dead, Jim"})
    flush()

    assert len(finds) == 6
    assert do_results == [None]
    assert not done_flag.isSet()
    
    q.add_task(finds[1][1], kwargs = {'block_id': finds[1][0], 'failure_reason': "He's dead, Jim"})
    flush()

    assert len(finds) == 6
    assert do_results == [None]
    assert not done_flag.isSet()
    
    q.add_task(finds[2][1], kwargs = {'block_id': finds[2][0], 'failure_reason': "He's dead, Jim"})
    flush()

    assert len(finds) == 6
    assert do_results == [None]
    assert not done_flag.isSet()
    
    q.add_task(finds[3][1], kwargs = {'block_id': finds[3][0], 'failure_reason': "He's dead, Jim"})
    flush()

    assert len(finds) == 6
    assert do_results == [None, 'finished']    
    assert done_flag.isSet()

def test_retrieves_fail():
    q = DoQ()
    block_store = {}
    
    do_results = []
    finds = []
    
    def do_function(reassembly, do_results = do_results):
        if reassembly.is_finished():
            do_results.append('finished')
            return
        while 1:
            next = reassembly.next_block_id()
            do_results.append(next)
            if next is None:
                return
    
    def find_function(block_id, callback, finds = finds):
        finds.append((block_id, callback))
    
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    block_positions = map(lambda x: x[0], l)
    block_positions[1] = None
    
    done_flag = threading.Event()
    
    block_store[l[0][0]] = l[0][1]

    callback = (lambda reassembly, done_flag = done_flag: done_flag.set())

    s = SingleReassembly(do_function = do_function, find_function = find_function, cancel_find_function = lambda x: None, doq = q,
        block_positions = block_positions, num_blocks_needed = 4, piece_id = make_ascii_id(crap), 
        piece_length = len(crap), callback = callback, block_store = block_store)

    def flush(q = q):
        for i in xrange(3):
            f = threading.Event()
            q.add_task(f.set)
            f.wait()

    flush()
    
    assert do_results == [None]
    assert len(finds) == 6
    ffs = map(lambda x: x[0], finds)
    ffs.sort()
    bcopy = copy(block_positions)
    bcopy = bcopy[2:]
    bcopy.sort()
    assert ffs == bcopy
    assert not done_flag.isSet()

    q.add_task(finds[0][1], kwargs = {'block_id': finds[0][0], 'failure_reason': 'it just croaked'})
    flush()

    assert len(finds) == 6
    assert do_results == [None]
    assert not done_flag.isSet()
    
    q.add_task(finds[1][1], kwargs = {'block_id': finds[1][0], 'failure_reason': 'it just croaked'})
    flush()

    assert len(finds) == 6
    assert do_results == [None]
    assert not done_flag.isSet()
    
    q.add_task(finds[2][1], kwargs = {'block_id': finds[2][0], 'failure_reason': 'it just croaked'})
    flush()

    assert len(finds) == 6
    assert do_results == [None]
    assert not done_flag.isSet()
    
    q.add_task(finds[3][1], kwargs = {'block_id': finds[3][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[3][0], None]
    assert not done_flag.isSet()

    q.add_task(s.block_finished, kwargs = {'block_id': finds[3][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[3][0], None]
    assert not done_flag.isSet()

    q.add_task(finds[4][1], kwargs = {'block_id': finds[4][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[3][0], None, finds[4][0], None]
    assert not done_flag.isSet()
    
    q.add_task(finds[5][1], kwargs = {'block_id': finds[5][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[3][0], None, finds[4][0], None, finds[5][0], None]
    assert not done_flag.isSet()
    
    q.add_task(s.block_finished, kwargs = {'block_id': finds[3][0], 'failure_reason': 'it went boom'})
    flush()

    assert len(finds) == 6
    assert do_results == [None, finds[3][0], None, finds[4][0], None, finds[5][0], None, 'finished']
    assert done_flag.isSet()

def test_cancel():
    q = DoQ()
    block_store = {}
    
    do_results = []
    finds = []
    
    def do_function(reassembly, do_results = do_results):
        if reassembly.is_finished():
            do_results.append('finished')
            return
        while 1:
            next = reassembly.next_block_id()
            do_results.append(next)
            if next is None:
                return
    
    def find_function(block_id, callback, finds = finds):
        finds.append((block_id, callback))
    
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    block_positions = map(lambda x: x[0], l)
    block_positions[1] = None
    
    done_flag = threading.Event()
    
    block_store[l[0][0]] = l[0][1]

    callback = (lambda reassembly, done_flag = done_flag: done_flag.set())

    s = SingleReassembly(do_function = do_function, find_function = find_function, cancel_find_function = lambda x: None, doq = q,
        block_positions = block_positions, num_blocks_needed = 4, piece_id = make_ascii_id(crap), 
        piece_length = len(crap), callback = callback, block_store = block_store)

    def flush(q = q):
        for i in xrange(3):
            f = threading.Event()
            q.add_task(f.set)
            f.wait()

    flush()

    q.add_task(s.cancel)
    
    flush()
    
    q.add_task(finds[0][1], kwargs = {'block_id': finds[0][0]})
    flush()

    assert len(finds) == 6
    assert do_results == [None, 'finished']
    assert done_flag.isSet()

def test_get_priority():
    q = DoQ()
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    l = map(lambda x: x[0], l)
    failflag = threading.Event()

    sr = SingleReassembly(do_function = lambda reassembly: None, 
        find_function = lambda block_id, callback: None ,
        cancel_find_function = lambda x: None, doq = q, 
        block_store = {}, block_positions = l, num_blocks_needed = 4, 
        piece_id = make_ascii_id(crap), piece_length = len(crap), callback = lambda reassembly, f = failflag: f.set(),
        priority = 3)

    def c(f = failflag, sr = sr):
        if sr.get_priority != 3:
            print 'AUGH!'
            f.set()
    
    def flush(q = q):
        fl = threading.Event()
        q.add_task(fl.set)
        fl.wait()

    q.add_task(sr.set_priority, kwargs = {'priority': 2})
    flush()
    
    def c2(f = failflag, sr = sr):
        if sr.get_priority != 2:
            print 'AUGH2!'
            f.set()
    flush()
    flush()
    
    assert not failflag.isSet()

def test_not_enough_ids():
    q = DoQ()
    block_store = {}
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    l = map(lambda x: x[0], l)
    l[0] = None
    for i in xrange(4, 8):
        l[i] = None
    failflag = threading.Event()
    f = threading.Event()
    sr = SingleReassembly(do_function = lambda reassembly: None, 
        find_function = lambda block_id, callback, f = failflag: f.set() , cancel_find_function = lambda x: None, doq = q, 
        block_store = block_store, block_positions = l, num_blocks_needed = 4, 
        piece_id = make_ascii_id('something else'), piece_length = len(crap), callback = lambda reassembly, f = f: f.set())
    f.wait()
    assert not failflag.isSet()
    try:
        sr.reassemble()
        return 0
    except ReassemblyError:
        pass

def test_wrong_piece_id():
    q = DoQ()
    block_store = {}
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    for i in l:
        block_store[i[0]] = i[1]
    failflag = threading.Event()
    f = threading.Event()
    sr = SingleReassembly(do_function = lambda reassembly: None, 
        find_function = lambda block_id, callback, f = failflag: f.set() , cancel_find_function = lambda x: None, doq = q, 
        block_store = block_store, block_positions = map(lambda x: x[0], l), num_blocks_needed = 4, 
        piece_id = make_ascii_id('something else'), piece_length = len(crap), callback = lambda reassembly, f = f: f.set())
    f.wait()
    assert not failflag.isSet()
    try:
        sr.reassemble()
        return 0
    except ReassemblyError:
        pass

def test_simple_case():
    q = DoQ()
    block_store = {}
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    for i in l[1:5]:
        block_store[i[0]] = i[1]
    failflag = threading.Event()
    f = threading.Event()
    piece_id = make_ascii_id(crap)
    sr = SingleReassembly(do_function = lambda reassembly: None, 
        find_function = lambda block_id, callback, f = failflag: f.set() , cancel_find_function = lambda x: None, doq = q, 
        block_store = block_store, block_positions = map(lambda x: x[0], l), num_blocks_needed = 4, 
        piece_id = piece_id, piece_length = len(crap), callback = lambda reassembly, f = f: f.set())
    f.wait()
    assert not failflag.isSet()
    assert sr.reassemble() == crap

def test_fall_out_of_cache():
    q = DoQ()
    block_store = {}
    crap = 'stnaoeusnathoeusnthaysnuhaosenuhtsadu'
    crap = crap + ('\000' * (4096 - len(crap)))
    l = split(crap, 4, 8)
    for i in l[1:5]:
        block_store[i[0]] = i[1]
    failflag = threading.Event()
    f = threading.Event()
    sr = SingleReassembly(do_function = lambda reassembly: None, 
        find_function = lambda block_id, callback, f = failflag: f.set() , cancel_find_function = lambda x: None, doq = q, 
        block_store = block_store, block_positions = map(lambda x: x[0], l), num_blocks_needed = 4, 
        piece_id = make_ascii_id(crap), piece_length = len(crap), callback = lambda reassembly, f = f: f.set())
    f.wait()
    assert not failflag.isSet()
    del block_store[l[1][0]]
    try:
        sr.reassemble()
        return 0
    except ReassemblyError:
        pass

mojo_test_flag = 1

if __name__ == '__main__':
    RunTests.runTests('SingleReassembly')




