#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import string
import threading
import traceback
import time

### our modules
from MojoConstants import true, false, DEBUG_MODE
from confutils import confman
import debug
from wrappedmethod import wrappedmethod


# just for testing
import RunTests


# Normally, if `DEBUG_MODE', then use Bombing mode, else use Blocking mode.  But when
# debugging you might want to see what happens in Blocking mode, but with verbose diagnostics.
# So turn on DEBUG_MODE _and_ BLOCKING_MODE for that purpose.  Sorry if this is confusing --
# basically we need to be able to test and debug "non-debug mode", which we then ship to users.
# --Zooko 2000-10-15

# BLOCKING_MODE = false
BLOCKING_MODE = true


DEADLOCK_INITIAL_BACKOFF=0.07 # Initial backoff.
DEADLOCK_TOLERANCE=66.6 # number of seconds that you'll wait before declaring deadlock


# if REALLY_REALLY_FAST_MODE, then do nothing.  This overrides settings above.
# REALLY_REALLY_FAST_MODE = false
REALLY_REALLY_FAST_MODE = true

_Dummy = lambda: None

class SingleThreader:
    """
    If `DEBUG_MODE', then this acts like BombingSingleThreader, else it acts like
    BlockingDeadlockDetectingSingleThreader.

    There is also a fast-and-sloppy mode (assert on a flag without locking) that isn't enabled.

    There's also really _really_ fast mode, which does nothing.

    @deprecated You really don't want to use any "*SingleThreader" class, which requires you to put a try: ... finally block around every method body and add a "_enter()" and "_exit()" call to each.  Instead you want the CoolLocker class, which does all that for you (with magical Python introspection weirdness) when you subclass CoolLocker.
    """
    def __init__(self):
        if REALLY_REALLY_FAST_MODE:
            self._enter = _Dummy
            self._exit = _Dummy
            return

        self._tlastthread = threading.currentThread()
        self._tlastthread2 = None
        self._tlaststack = traceback.extract_stack()
        self._tlaststack2 = None

        FAST = false
        if FAST:
            self._busy = false
        else: 
            self._tlock = threading.RLock()

            if (not DEBUG_MODE) or (BLOCKING_MODE):
                self._locklock = threading.Lock() # For protecting deadlock detector.

                # XXX this causes a tight circular reference that prevents any SingleThreader derived object from being garbage collected
                # (self._enter holds a reference to self via the instance method type assigned to it from self._enter_foo)
                self._enter = self._enter_block
                self._exit = self._exit_block
            else:
                # XXX this causes a tight circular reference that prevents any SingleThreader derived object from being garbage collected
                # (self._enter holds a reference to self via the instance method type assigned to it from self._enter_foo)
                self._enter = self._enter_bomb
                self._exit = self._exit_bomb

    def _enter_bomb(self):
        res = self._tlock.acquire(blocking=false)
        assert res, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread.  previous thread: %s, previous stack trace: %s, current thread: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._tlastthread`, `self._tlaststack`, `threading.currentThread()`, `self._tlastthread2`, `self._tlaststack2`)
        self._tlaststack2 = self._tlaststack
        self._tlaststack = traceback.extract_stack()
        self._tlastthread2 = self._tlastthread
        self._tlastthread = threading.currentThread()
        return

    def _exit_bomb(self):
        self._tlock.release()
        return

    def _enter_fast_bomb(self):
        assert not self._busy, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread.  previous thread: %s, previous stack trace: %s, current thread: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._tlastthread`, `self._tlaststack`, `threading.currentThread()`, `self._tlastthread2`, `self._tlaststack2`)
        self._busy = true
        self._tlaststack2 = self._tlaststack
        self._tlaststack = traceback.extract_stack()
        self._tlastthread2 = self._tlastthread
        self._tlastthread = threading.currentThread()
        
    def _exit_fast_bomb(self):
        assert self._busy, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread.  previous thread: %s, previous stack trace: %s, current thread: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._tlastthread`, `self._tlaststack`, `threading.currentThread()`, `self._tlastthread2`, `self._tlaststack2`)
        self._busy = false
        self._tlaststack2 = self._tlaststack
        self._tlaststack = traceback.extract_stack()
        self._tlastthread2 = self._tlastthread
        self._tlastthread = threading.currentThread()
        
    def _enter_block(self):
        debug.stderr.write("SingleThreader %s >-> ...  %s num=%s\n", args=(self, threading.currentThread().getName(), self._tlock._RLock__count), v=31, vs="debug")
        debug.stderr.write("SingleThreader %s: tb: %s\n", args=(self, traceback.extract_stack()), v=31, vs="debug")

        self._locklock.acquire()
        try:
            satime = time.time()
            gotit = self._tlock.acquire(blocking=false)
            backoff = DEADLOCK_INITIAL_BACKOFF
            while ((time.time() - satime)<DEADLOCK_TOLERANCE) and (not gotit):
                debug.stderr.write("SingleThreader %s --- retrying! ...  %s - num=%s\n", args=(self, threading.currentThread().getName(), self._tlock._RLock__count), v=5, vs="debug")

                time.sleep(backoff)
                backoff = (backoff * 2) + 0.001 # `+ 0.001' to avoid sticking at "0"
                gotit = self._tlock.acquire(blocking=false)
            assert gotit, "error -- Deadlock.  previous thread: %s, previous stack trace: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._tlastthread`, `self._tlaststack`, `self._tlastthread2`, `self._tlaststack2`)
            self._tlaststack2 = self._tlaststack
            self._tlaststack = traceback.extract_stack()
            self._tlastthread2 = self._tlastthread
            self._tlastthread = threading.currentThread()
        finally:
            self._locklock.release()

    def _exit_block(self):
        self._tlock.release()
        debug.stderr.write("SingleThreader %s <-< ...  %s - num=%s\n", args=(self, threading.currentThread().getName(), self._tlock._RLock__count), v=31, vs="debug")




### !!! XXX TODO: use metaprogramming trick instead of requiring subclasses to do "self._enter(); try: ... finally: self._exit()" in each method.

class BombingSingleThreader:
    def __init__(self):
        self._tlock = threading.RLock() # This is just for detecting someone calling this class multithreadedly...  ("Top LOCK")
        self._tlaststack = traceback.extract_stack() # This is just for detecting someone calling this class multithreadedly...
        self._tlastthread = threading.currentThread() # This is just for detecting someone calling this class multithreadedly...

    def _enter(self):
        res = self._tlock.acquire(blocking=false)
        assert res, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread.  previous thread: %s, previous stack trace: %s, current thread: %s" % (`self._tlastthread`, `self._tlaststack`, `threading.currentThread()`)
        self._tlaststack = traceback.extract_stack()
        self._tlastthread = threading.currentThread()
        return

    def _exit(self):
        self._tlock.release()
        return

class BlockingSingleThreader:
    def __init__(self):
        self._tlock = threading.RLock()
        self._enter = self._tlock.acquire
        self._exit = self._tlock.release


class BlockingDeadlockDetectingSingleThreader:
    def __init__(self):
        self._tlock = threading.RLock()
        self._tlaststack = traceback.extract_stack()
        self._locklock = threading.Lock() # For protecting deadlock detector.

    def _enter(self):
        self._locklock.acquire()
        satime = time.time()
        gotit = self._tlock.acquire(blocking=false)
        backoff = DEADLOCK_INITIAL_BACKOFF
        while ((time.time() - satime)<DEADLOCK_TOLERANCE) and (not gotit):
            time.sleep(backoff)
            backoff = backoff * 2
            gotit = self._tlock.acquire(blocking=false)
        assert gotit, "error -- Deadlock.  previous stack trace = %s" % `self._tlaststack`
        self._tlaststack = traceback.extract_stack()
        self._locklock.release()

    def _exit(self):
        self._tlock.release()

class VerboseBlockingDeadlockDetectingSingleThreader:
    def __init__(self):
        self._tlock = threading.RLock()
        self._tlaststack = traceback.extract_stack()
        self._locklock = threading.Lock() # For protecting deadlock detector.

    def _enter(self):
        debug.stderr.write("VerboseBlockingSingleThreader %s >-> ...  %s - %s  num=%s\n", args=(self, threading.currentThread().getName(), self, self._tlock._RLock__count), v=31, vs="debug")

        self._locklock.acquire()
        satime = time.time()
        gotit = self._tlock.acquire(blocking=false)
        backoff = DEADLOCK_INITIAL_BACKOFF
        while ((time.time() - satime)<DEADLOCK_TOLERANCE) and (not gotit):
            time.sleep(backoff)
            backoff = backoff * 2
            gotit = self._tlock.acquire(blocking=false)
        assert gotit, "error -- Deadlock.  previous stack trace = %s" % `self._tlaststack`
        self._tlaststack = traceback.extract_stack()
        self._locklock.release()

    def _exit(self):
        self._tlock.release()
        debug.stderr.write("VerboseBlockingSingleThreader %s <-< ...  %s - %s  num=%s\n", args=(self, threading.currentThread().getName(), self, self._tlock._RLock__count), v=31, vs="debug")

class DummySingleThreader:
    def __init__(self):
        pass
    def _enter(self):
        pass
    def _exit(self):
        pass

class VerboseDummySingleThreader:
    def __init__(self):
        self.num = 0
        pass

    def _enter(self):
        self.num = self.num + 1
        debug.stderr.write("VerboseDummySingleThreader %s >-> ...  %s  num=%s\n" % (`self`, `threading.currentThread().getName()`, `self.num`))
        pass

    def _exit(self):
        self.num = self.num - 1
        debug.stderr.write("VerboseDummySingleThreader %s <-< ...  %s  num=%s\n" % (`self`, `threading.currentThread().getName()`, `self.num`))
        pass

class CoolLocker:
    """
    This ensures that objects are accessed by only one thread at a time.  It has five possible
    modes of operation, only one of which can be used per object, and which is chosen in the
    constructor.

    The five modes of operation are "blocking", "blocking-deadlock-detecting", "bombing",
    "sloppy" and "dummy".

    "blocking" means that the second thread that invokes a method of this object will be blocked
    until the first thread exits this object.  This mode can result in deadlocks, and is not
    recommended, although it is faster than "blocking-deadlock-detecting".

    "blocking-deadlock-detecting" mode is like "blocking" mode except that deadlock detection is
    also enabled;  An exception is raised if deadlock is detected, which will be very verbose if
    DEBUG_MODE is set.  This mode is recommended if you need to use the object in
    a multi-threaded way, but it is the slowest mode.

    "bombing" means that if more than one thread enters the object simultaneously, an exception
    will immediately be raised.  This mode is recommended if you intend to use the object in a
    single-threaded way, but bugs in your program are causing it to be used multi-threadedly,
    and you want to find those bugs.

    "sloppy" mode is like "bombing" mode, but it uses a fast and sloppy mechanism which can
    sometimes fail to detect multi-threading.

    "dummy" means don't do anything.  This is the fastest mode.
    """
    ### XXX TODO: in "dummy" mode, don't do the whole wrapping-of-functions thing, which is probably by far the biggest slowdown here.
    def __init__(self, mode="bombing", include=None, exclude=None):
        """
        @param mode one of {"blocking", "blocking-deadlock-detecting", "bombing", "sloppy", "dummy"}
        @param include if `None', then include all methods of the object under threading
            protection; if not `None', then it is a sequence of method names -- include only
            those methods under threading protection
        @param exclude if not `None', then it is a sequence of method names -- exclude those
            methods from threading protection (regardless of whether they were listed in
            `include' or not)

        @precondition `mode' must be one of {"blocking", "blocking-deadlock-detecting", "bombing", "sloppy", "dummy"}.: mode in ("blocking", "blocking-deadlock-detecting", "bombing", "sloppy", "dummy"): "mode: %s" % hr(mode)
        """
        assert mode in ("blocking", "blocking-deadlock-detecting", "bombing", "sloppy", "dummy"), "precondition: `mode' must be one of {\"blocking\", \"blocking-deadlock-detecting\", \"bombing\", \"sloppy\", \"dummy\"}." + " -- " + "mode: %s" % hr(mode)

        # rename each method by prepending "_SingleThreader_real_", and then make a new method that does:
        # def foo():
        #     self._enter()
        #     try:
        #         _SingleThreader_real_foo()
        #     finally:
        #         self._exit()

        if include is not None:
            for incname in include:
                if not self.__class__.__dict__.has_key(incname):
                    debug.mojolog.write("WARNING: self: %s doesn't have method: %s which was supposed to be included in threading protection.\n", args= (self, incname,), v=4, vs="debug")

        if exclude is not None:
            for excname in exclude:
                if not self.__class__.__dict__.has_key(excname):
                    debug.mojolog.write("WARNING: self: %s doesn't have method: %s which was supposed to be excluded from threading protection.\n", args= (self, excname,), v=4, vs="debug")

        for (fname, func) in self.__class__.__dict__.items():
            if (len(fname) >= 2) and (fname[:2] == "__"):
                continue
            if (include is not None) and not (fname in include):
                continue
            if (exclude is not None) and (fname in exclude):
                continue

            # The only actual use for the newfnames is as a clue for programmers who don't know
            # what SingleThreader does.  While inspecting their objects/classes in a debugger,
            # the presence of "_SingleThreader_real_*" will hopefully tip them off...
            newfname = "__SingleThreader_real_" + fname
            self.__class__.__dict__[newfname] = func

            meth = eval("self." + fname)

            if isinstance(meth, wrappedmethod):
                # This one is already wrapped!  Probably caused by repeated inheritance.
                continue

            # debug.mojolog.write("fname: %s, meth: %s\n", args=(fname, meth,), vs="debug")
            # Generate a callable object to take the place of this func.

            wm = wrappedmethod(meth)

            # Put this wrapper callable obj where it will be used.
            self.__class__.__dict__[fname] = wm

            if mode == "blocking":
                self._enter = self._enter_blocking
                self._exit = self._exit_blocking
                self._lock = threading.RLock()
            elif mode == "blocking-deadlock-detecting":
                self._enter = self._enter_blocking_deadlock_detecting
                self._exit = self._exit_blocking_deadlock_detecting
                self._lock = threading.RLock() # deadlock detector
                self._locklock = threading.Lock() # For protecting deadlock detector.
                if DEBUG_MODE:
                    self._lastthread = threading.currentThread()
                    self._lastthread2 = None
                    self._laststack = traceback.extract_stack()
                    self._laststack2 = None
            elif mode == "bombing":
                self._enter = self._enter_bombing
                self._exit = self._exit_bombing
                self._lock = threading.RLock()
                if DEBUG_MODE:
                    self._lastthread = threading.currentThread()
                    self._lastthread2 = None
                    self._laststack = traceback.extract_stack()
                    self._laststack2 = None
            elif mode == "sloppy":
                self._enter = self._enter_sloppy
                self._exit = self._exit_sloppy
                self._busy = false
                if DEBUG_MODE:
                    self._lastthread = threading.currentThread()
                    self._lastthread2 = None
                    self._laststack = traceback.extract_stack()
                    self._laststack2 = None
            elif mode == "dummy":
                self._enter = self._enter_dummy
                self._exit = self._exit_dummy

    def _enter_dummy(self):
        pass
    def _exit_dummy(self):
        pass

    def _enter_bombing(self):
        res = self._lock.acquire(blocking=false)
        if DEBUG_MODE:
            assert res, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread.  previous thread: %s, previous stack trace: %s, current thread: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._lastthread`, `self._laststack`, `threading.currentThread()`, `self._lastthread2`, `self._laststack2`)

            self._laststack2 = self._laststack
            self._laststack = traceback.extract_stack()
            self._lastthread2 = self._lastthread
            self._lastthread = threading.currentThread()
        else:
            assert res, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread."
        return

    def _exit_bombing(self):
        self._lock.release()
        return

    def _enter_sloppy(self):
        if DEBUG_MODE:
            assert not self._busy, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread.  previous thread: %s, previous stack trace: %s, current thread: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._lastthread`, `self._laststack`, `threading.currentThread()`, `self._lastthread2`, `self._laststack2`)

            self._laststack2 = self._laststack
            self._laststack = traceback.extract_stack()
            self._lastthread2 = self._lastthread
            self._lastthread = threading.currentThread()
        else:
            assert not self._busy, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread."

        self._busy = true

    def _exit_sloppy(self):
        if DEBUG_MODE:
            assert self._busy, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread.  previous thread: %s, previous stack trace: %s, current thread: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._lastthread`, `self._laststack`, `threading.currentThread()`, `self._lastthread2`, `self._laststack2`)

            self._laststack2 = self._laststack
            self._laststack = traceback.extract_stack()
            self._lastthread2 = self._lastthread
            self._lastthread = threading.currentThread()
        else:
            assert self._busy, "error -- This class is not multi-thread safe and it has been called simultaneously from more than one thread."
        self._busy = false

    def _enter_blocking(self):
        self._lock.acquire()
    def _exit_blocking(self):
        self._lock.release()

    def _enter_blocking_deadlock_detecting(self):
        debug.stderr.write("SingleThreader %s >-> ...  %s num=%s\n", args=(self, threading.currentThread().getName(), self._lock._RLock__count), v=31, vs="debug")

        self._locklock.acquire()
        try:
            satime = time.time()
            gotit = self._lock.acquire(blocking=false)
            backoff = DEADLOCK_INITIAL_BACKOFF
            while ((time.time() - satime)<DEADLOCK_TOLERANCE) and (not gotit):
                debug.stderr.write("SingleThreader %s --- retrying! ...  %s - num=%s\n", args=(self, threading.currentThread().getName(), self._lock._RLock__count), v=5, vs="debug")

                time.sleep(backoff)
                backoff = (backoff * 2) + 0.001 # `+ 0.001' to avoid sticking at "0"
                gotit = self._lock.acquire(blocking=false)
            if DEBUG_MODE:
                assert gotit, "error -- Deadlock.  previous thread: %s, previous stack trace: %s, previous^2 thread: %s, previous^2 stack trace: %s" % (`self._lastthread`, `self._laststack`, `self._lastthread2`, `self._laststack2`)

                self._laststack2 = self._laststack
                self._laststack = traceback.extract_stack()
                self._lastthread2 = self._lastthread
                self._lastthread = threading.currentThread()
            else:
                assert gotit, "error -- Deadlock."
        finally:
            self._locklock.release()

    def _exit_blocking_deadlock_detecting(self):
        self._lock.release()
        debug.stderr.write("SingleThreader %s <-< ...  %s - num=%s\n", args=(self, threading.currentThread().getName(), self._lock._RLock__count), v=31, vs="debug")


def test():
    class Spam(SingleThreader):
        def __init__(self):
            SingleThreader.__init__(self)
            print "after SingleThreader: self.__dict__: %s\nself.__class__.__dict__: %s\n\n" % (self.__dict__, self.__class__.__dict__)
        def setx(self, newx):
            self.x = newx
        def printx(self, msg):
            print "msg: %s, self.x: %s\n" % (msg, self.x)

    s=Spam()

    s.setx(3)
    s.printx("hi there!")


# Generic stuff
NAME_OF_THIS_MODULE="SingleThreader"

mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests(NAME_OF_THIS_MODULE)

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()

