#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import UserDict
import asyncore
import threading
import time
import socket
import types
import struct
import traceback
import string

### our modules
import mojoasyncore
import BandwidthThrottler
import Cache
import CommsHandler
from CommHints import HINT_EXPECT_RESPONSE, HINT_EXPECT_MORE_TRANSACTIONS, HINT_EXPECT_NO_MORE_COMMS, HINT_EXPECT_TO_RESPOND, HINT_THIS_IS_A_RESPONSE, HINT_NO_HINT
import CommStrat
import DoQ
from MojoConstants import true, false, DEBUG_MODE
import MojoConstants
import SEXP
import SingleThreader
import TCPConnection
import confutils
import debug
import idlib
import mojoutil
from confutils import confman
import ipaddresslib
from human_readable import human_readable, hr


SECS_BETWEEN_IPADDR_RECHECK = 180   # recheck our IP address every 3 minutes incase it has changed


class TCPCommsHandler(CommsHandler.BaseCommsHandler, asyncore.dispatcher):
    """
    This accepts incoming connections and spawns off a TCPConnection for each one.
    """
    def __init__(self, listenport=None, pickyport=false, dontbind=false):
        """
        @param listenport the preferred port to listen on
        @param pickyport `true' if you want to fail in the case that `listenport' is unavailable, false if you want to get another arbitrary port

        @precondition `listenport' must be a non-negative integer or `None'.: (listenport is None) or ((type(listenport) is types.IntType) and (listenport > 0)): "listenport: %s :: %s" % (hr(listenport), `type(listenport)`)
        @precondition `pickyport' is false or listenport is an integer.: (not pickyport) or (type(listenport) is types.IntType)
        """
        assert (listenport is None) or ((type(listenport) is types.IntType) and (listenport > 0)), "precondition: `listenport' must be a non-negative integer or `None'." + " -- " + "listenport: %s :: %s" % (hr(listenport), `type(listenport)`)
        assert (not pickyport) or (type(listenport) is types.IntType), "precondition: `pickyport' is false or listenport is an integer."

        asyncore.dispatcher.__init__(self)
        CommsHandler.BaseCommsHandler.__init__(self)

        self._requested_listenport = listenport
        self._pickyport = pickyport

        self._conncache = TCPConnCache(cid_to_cs=self._cid_to_cs)

        # this boolean determines if we actually bind to a port
        self._dontbind = dontbind

        self._throttlerout = BandwidthThrottler.BandwidthThrottler(throttle=confman.is_true_bool(('TCP_THROTTLE_OUT',)), Kbps=mojoutil.longpopL(confman.get('TCP_MAX_KILOBITS_PER_SECOND_OUT', "56")))
        self._throttlerin = BandwidthThrottler.BandwidthThrottler(throttle=confman.is_true_bool(('TCP_THROTTLE_IN',)), Kbps=mojoutil.longpopL(confman.get('TCP_MAX_KILOBITS_PER_SECOND_IN', "56")))

        # My "id" is just a random number.  Nobody really uses this except for testing.
        self._id = idlib.make_new_random_id(thingtype='TCPCommsHandler')

        self._upward_inmsg_handler = None

        # If you are bound to a port and listening, then `self._ip' and `self._listenport' will be non-None.
        # If you want to see if your address is routable or not, then call `ipaddresslib.is_routable(self._ip)', but you should
        # only do this from within TCPCommsHandler.  If you are writing higher-level code, and you want to see if
        # you have a good comm strat, then just call `get_comm_strategy()' and see if you like the strategy offered.
        # TCPCommsHandler will give you back a "CommStrat.TCP" (or, in the worst case, a "CommStrat.Pickup",
        # which you almost certainly do not like).  Call `ipaddresslib.is_routable(cs.host)'.  The reason to go through the
        # CommStrat abstraction and not just peek at `self._ip' is because people can set up tunnels and announce
        # themselves as being available on a different host/port.
        self._ip = None
        self._last_ip_check_time = 0
        self._listenport = None
        self._accept_incoming_connections = 1


    def hold_incoming_connections(self):
        self._accept_incoming_connections = None
    
    def unhold_incoming_connections(self):
        self._accept_incoming_connections = 1


    def stop_listening(self):
        self._enter()
        try:
            # prevent an AttributeError in asyncore when this TCPCommsHandler is a dummy for outgoing only without a socket
            if hasattr(self, 'socket'):
                self.close()
            self._ip = None
            self._listenport = None

            # Clean up all outstanding connections.
            for k in self._conncache.keys():
                self._conncache.remove(k)
        finally:
            self._exit()
   
    def is_listening(self):
        self._enter()
        try:
            debug.mojolog.write("TCPCommsHandler.is_listening(): %s, self._ip: %s\n", args=(self._ip != None, self._ip,), v=0, vs="debug")

            return self._ip != None
        finally:
            self._exit()

    def start_listening(self, inmsg_handler_func):
        self._enter()
        try:
            self._upward_inmsg_handler = inmsg_handler_func

            # if dontbind is set, do nothing here, we don't want to listen
            if self._dontbind :
                return

            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            self.set_reuse_addr()

            self._listenport = None # Will be set to the port number only after we've bound to it.
            if self._pickyport:
                try:
                    self.bind(("", self._requested_listenport))
                except socket.error:
                    raise CommsHandler.CannotListenError, "couldn't bind to port, and I'm picky about which port I listen on"

                self._listenport = self._requested_listenport
            else:
                MAX_PORT = 32767 # standard?  But you can feel free to change it without changing the rest of this code.
                MIN_PORT = 1025 # Arbitrarily chosen.  Change at your whim and you don't need to change the rest of this code.
                if self._requested_listenport:
                    try_listenport = self._requested_listenport
                else:
                    try_listenport = 20001 # Arbitrarily chosen.  Change at your whim and you don't need to change the rest of this code.
                firstporttried = try_listenport

                if string.lower(confman.dict["YES_NO"]["SERVE_USING_A_RELAY"]) == 'yes':
                    debug.mojolog.write("SERVE_USING_A_RELAY is set, not listening on a socket.\n", v=1, vs='TCPCommsHandler')
                    self._ip = None
                    return

                while not self._listenport:
                    try:
                        self.bind(("", try_listenport))
                        debug.mojolog.write("successfully bound to port %s.\n", args=(try_listenport,), v=1, vs="TCPCommsHandler")
                        self._listenport = try_listenport

                    except socket.error:
                        debug.mojolog.write("couldn't bind to port %s; trying next port.\n", args=(try_listenport,), v=1, vs="TCPCommsHandler")

                        try_listenport = ((try_listenport + 1 - MIN_PORT) % (MAX_PORT + 1 - MIN_PORT)) + MIN_PORT

                        if try_listenport == firstporttried:
                            # Well, we've tried them all!  Give up.
                            raise CommsHandler.CannotListenError, "couldn't bind to any port"

            self.listen(int(confman['TCP_MAX_CONNECTIONS']))

            self._ip = ipaddresslib.get_primary_ip_address(nonroutableok=true)
            self._last_ip_check_time = time.time()

            assert self._listenport > 0
            debug.mojolog.write("I am listening on port %s and telling %s.\n", args=(self._listenport, inmsg_handler_func), v=4, vs="TCPCommsHandler")
        finally:
            self._exit()

    def get_comm_strategy(self):
        self._enter()
        try:
            assert (self._ip is not None) == (self._listenport is not None), "internal error: either both or none of {_listenport, _ip} should be defined -- self._ip: %s, self._listenport: %s" % (`self._ip`, `self._listenport`)

            if self._listenport:
                # Now IP_ADDRESS_OVERRIDE overrides the actual bound IP address, and TRANSACTION_MANAGER_ANNOUNCED_PORT overrides the actual bound port.
                announceip = confman.get('IP_ADDRESS_OVERRIDE')
                if not announceip:
                    # this causes us to recheck our IP address every so often incase it has changed
                    if self._last_ip_check_time + SECS_BETWEEN_IPADDR_RECHECK < time.time():
                        self._ip = ipaddresslib.get_primary_ip_address(nonroutableok=true)
                        self._last_ip_check_time = time.time()
                    announceip = self._ip
                if confman.get('TRANSACTION_MANAGER_ANNOUNCED_PORT'):
                    announceport = int(confman['TRANSACTION_MANAGER_ANNOUNCED_PORT'])
                else:
                    announceport = self._listenport

                return CommStrat.TCP(host=announceip, port=announceport)
            else:
                if string.lower(confman.dict["YES_NO"]["SERVE_USING_A_RELAY"]) != 'yes':
                    debug.mojolog.write("TCPCommsHandler.get_comm_strategy() is returning CommStrat.Pickup() and you're not behind a relay server!\n", v=0, vs="ERROR")
                return CommStrat.Pickup()
        finally:
            self._exit()

    def get_id(self):
        """
        """
        self._enter()
        try:
            return self._id
        finally:
            self._exit()
       
    def _internal_send_msg(self, counterparty_id, msg, hint=HINT_NO_HINT, fast_fail_handler=None, timeout=None):
        counterparty_id = idlib.to_binary(counterparty_id)  # this also checks the precondition of counterparty_id being an id

        self._enter()
        try:
            tcpc = self._conncache.get(counterparty_id)
            if not tcpc :
                raise CommsHandler.CannotSendError, "no TCP connection"

            assert isinstance(tcpc, TCPConnection.TCPConnection)

            cs = self._cid_to_cs.get(counterparty_id)
            if cs:
                # If we were expecting to send a response and this is it...
                if (hint & HINT_THIS_IS_A_RESPONSE) and (cs.hint & HINT_EXPECT_TO_RESPOND) and (cs.hintnumexpectedsends > 0):
                    cs.hintnumexpectedsends = cs.hintnumexpectedsends - 1
                    if cs.hintnumexpectedsends == 0:
                        cs.hint = cs.hint & (~ HINT_EXPECT_TO_RESPOND)

                if hint & HINT_EXPECT_RESPONSE:
                    cs.hintnumexpectedresponses = cs.hintnumexpectedresponses + 1

                cs.hint = cs.hint | hint & (~ HINT_THIS_IS_A_RESPONSE)

            tcpc.send(msg, fast_fail_handler)
            del msg  # we're done with it, encourage faster gc

            # Now if we have a hint that says no response is expected, close the connection.
            if cs and (cs.hint & HINT_EXPECT_NO_MORE_COMMS):
                pn = None
                try:
                    pn = tcpc.getpeername()
                except:
                    pass
                   
                debug.mojolog.write("TCPCommsHandler._internal_send_msg(): removing connection %s _cid_for_debugging: %s, .getpeername(): %s due to HINT_EXPECT_NO_MORE_COMMS.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                self._conncache.remove(counterparty_id)

            self._conncache._cleanup_nice()
        finally:
            self._exit()

    def forget_comm_strategy(self, counterparty_id, commstrat):
        curcs = self._cid_to_cs.get(counterparty_id)
        if (curcs is None) or (not curcs.same(commstrat)):
            # We aren't using `commstrat', so there is nothing to forget.
            debug.mojolog.write("TCPCommsHandler was asked to forget unknown commstrat: %s, curcs: %s\n", args=(commstrat, curcs,), v=5, vs="debug")
            return

        debug.mojolog.write("TCPCommsHandler: Forgetting comm strategy.  counterparty_id: %s, curcs: %s\n", args=(counterparty_id, curcs,), v=5, vs="commstrats")

        remmed = self._conncache.remove_if_present(counterparty_id)
        # assert (remmed is None) or (remmed is curcs.asyncsock), "curcs: %s, remmed: %s" % (hr(curcs.asyncsock), hr(remmed))

        del self._cid_to_cs[counterparty_id]
        del self._cids_to_activation[counterparty_id]

    def use_comm_strategy(self, counterparty_id, commstrat):
        """
        @precondition `commstrat' must be a CommStrat.TCP.: isinstance(commstrat, CommStrat.TCP): "commstrat: %s :: %s" % (hr(commstrat), hr(type(commstrat)))
        """
        assert isinstance(commstrat, CommStrat.TCP), "precondition: `commstrat' must be a CommStrat.TCP." + " -- " + "counterparty_id: %s, commstrat: %s :: %s" % (hr(counterparty_id), hr(commstrat), hr(type(commstrat)))

        self._enter()
        try:
            # If the new one is an open connection, or the old one is just an IP address, then
            # we'll use the new one.  (Which is the same as saying we use the new one unless the
            # new one is just an IP address and we already have an open connection.)

            curcs = self._cid_to_cs.get(counterparty_id)

            if (not curcs) or (not curcs.asyncsock) or (curcs.asyncsock.is_closing()) or (commstrat.asyncsock and (not commstrat.asyncsock.is_closing())):
                # Use the new one!
                # ... but if the old one had IP address and port number, and the new one
                # doesn't, then copy them over.
                if (not commstrat.host) and (curcs) and (curcs.host):
                    commstrat.host = curcs.host
                if (not commstrat.port) and (curcs) and (curcs.port):
                    commstrat.port = curcs.port

                self._cid_to_cs[counterparty_id] = commstrat
                self._cids_to_activation[counterparty_id] = "not yet"

                if commstrat.asyncsock:
                    commstrat.asyncsock._cid_for_debugging = counterparty_id
                    self._conncache.insert(counterparty_id, commstrat.asyncsock)
            else:
                # Continue using the old one.  No change.
                # ... but if the new one has IP address and port number, and the old one
                # doesn't, then copy them over.
                if not curcs.host:
                    curcs.host = commstrat.host
                if not curcs.port:
                    curcs.port = commstrat.port

                pass
        finally:
            self._exit()

    def _activate_commstrat(self, counterparty_id, cs):
        debug.mojolog.write("TCPCommsHandler._activate_commstrat(%s, %s)\n", args=(counterparty_id, cs,), v=21, vs="commstrats") # verbose -- commstrats
        counterparty_id = idlib.canonicalize(counterparty_id, 'broker')  # this also checks the precondition of counterparty_id being an id

        self._enter()
        try:
            if cs.asyncsock and (not cs.asyncsock.is_closing()):
                assert isinstance(cs.asyncsock, TCPConnection.TCPConnection), "cs.asyncsock: %s :: %s" % (hr(cs.asyncsock), hr(type(cs.asyncsock)))
                tcpc = cs.asyncsock
                tcpc._upward_inmsg_handler=self._inmsg_handler
            elif cs.host and cs.port:
                tcpc = TCPConnection.TCPConnection(inmsg_handler_func=self._inmsg_handler, key=counterparty_id, host=cs.host, port=cs.port, commstratobj=cs, throttlerin=self._throttlerin, throttlerout=self._throttlerout, cid_for_debugging=counterparty_id)
                cs.asyncsock = tcpc
            else:
                return # can't use this comm strat -- this problem will be discovered momentarily when someone tries to send a message  --Zooko 2000-09-26

            assert isinstance(tcpc, TCPConnection.TCPConnection)

            debug.mojolog.write("TCPCommsHandler._activate_commstrat(%s, %s): after activation\n", args=(counterparty_id, cs,), v=21, vs="commstrats")

            self._conncache.insert(counterparty_id, tcpc)
            assert len(self._conncache) > 0
        finally:
            self._exit()

    def _inmsg_handler(self, asyncsock, msg):
        """
        # XXX this is now called from the DoQ.  --Zooko 2001-02-07
        @precondition This method must be called on the DoQ.: DoQ.doq.is_currently_doq()
        """
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        # XXX How come this is sometimes `None'?  I think it should have been put in the cache
        # by a DoQ tasks launched from `handle_accept()', and that it can't be removed until after
        # `inmsg_handler()' has already run.  --Zooko 2001-05-18
        # Okay the answer is that in at least once case it was removed due to timeout, and then moments later
        # a message came in on it.  Currently removal takes quite a while to finalize -- as it has to go to the
        # asyncore thread and then back, so instead of trying to leave it in _conncache until it is finalized, I'm
        # just going to comment out this assertion.  --Zooko 2001-05-29
        # assert self._conncache.get(asyncsock._key) is asyncsock, "asyncsock: %s: asyncsock._key: %s, self._conncache.get(asyncsock._key): %s" % (hr(asyncsock), hr(asyncsock._key), hr(self._conncache.get(asyncsock._key)),)

        # Testing whether this is the only case that this happens.
        assert (self._conncache.get(asyncsock._key) is asyncsock) or (asyncsock.is_closing()), "asyncsock: %s: asyncsock._key: %s, self._conncache.get(asyncsock._key): %s" % (hr(asyncsock), hr(asyncsock._key), hr(self._conncache.get(asyncsock._key)),)

        cs = self._cid_to_cs.get(asyncsock._key)
        if (cs is not None) and (cs.asyncsock is asyncsock) and (self._conncache.get(asyncsock._key) is asyncsock) and (not asyncsock.is_closing()):
            # If the current `cs' points to a useful (i.e. registered and non-closing) socket, then just recommend to keep the current cs.
            self._upward_inmsg_handler(msg, cs, asyncsock._key)
        elif not asyncsock.is_closing():
            # If `asyncsock' is a useful socket, recommend to use it.
            self._upward_inmsg_handler(msg, CommStrat.TCP(asyncsock=asyncsock), asyncsock._key)
        else:
            # There is no known useful socket.  Recommend nothing.
            self._upward_inmsg_handler(msg, None, None)

    # The `handle_spam()' methods and the `_inmsg_handler()' are the "bottom" interface, to be called by the asyncore thread.  There can be (one) thread touching the bottom interface and (one) thread touching the top interface at the same time.
    def handle_accept(self) :
        # self._enter()
        try:
            try:
                thingie = self.accept()

                if not (((type(thingie) is types.TupleType) or (type(thingie) is types.ListType))):
                    debug.mojolog.write("got a non-sequence from `socket.accept()'.  Ignoring.  thingie: %s :: %s\n", args=(thingie, type(thingie),), v=0, vs="debug")
                    return
            except socket.error, le:
                debug.mojolog.write("got an exception from `socket.accept()'.  Ignoring.  le: %s\n", args=(le,), v=0, vs="debug")
                return

            (sock, addr) = thingie
            key = idlib.make_new_random_id(thingtype='TCPConnection')
            tcpc = TCPConnection.TCPConnection(inmsg_handler_func=self._inmsg_handler, key=key, sock=sock, throttlerin=self._throttlerin, throttlerout=self._throttlerout)
            pn = None
            try:
                pn = tcpc.getpeername()
            except:
                pass

            debug.mojolog.write("TCPCommsHandler: handle_accept(): the new socket %s has peername: %s and addr: %s\n", args=(sock, pn, addr), v=21, vs="commstrats")

            DoQ.doq.add_task(self._conncache.insert, args=(key, tcpc))
        finally:
            pass # self._exit()

    def handle_write(self):
        debug.mojolog.write("TCPCommsHandler: handle_write() called.  self: %s\n", args=(self,), v=0, vs="debug")
        return

    def handle_connect(self):
        debug.mojolog.write("TCPCommsHandler: handle_connect() called.  self: %s\n", args=(self,), v=0, vs="debug")
        return

    def handle_read(self):
        # `handle_read()' gets called on Linux 2.4.0.  Oh well.  --Zooko 2001-01-20
        # debug.mojolog.write("TCPCommsHandler: handle_read() called.  self: %s\n", args=(self,), v=0, vs="debug")
        return

    def handle_close(self):
        debug.mojolog.write("TCPCommsHandler: handle_close() called.  self: %s\n", args=(self,), v=0, vs="debug")
        return

    def readable(self) :
        # in this listening socket, this controls whether handle_accept gets called on incoming
        # connections or not (if not, they wait in the OS listen() queue until it is called).
        return self._listenport and self._accept_incoming_connections and (not self._throttlerin.is_maxed())

    def writable(self) :
        return 0   # this is a listening socket only, its never writable

    def log(self, message):
        # for faster operation, comment this whole method out and replace it with "def log(): return".  --Zooko 2000-12-11
        #self._enter()
        try:
            if message[-1:] == "\n":
                debug.mojolog.write("TCPCommsHandler: asyncore log: " + message, v=21, vs="commstrats")
            else:
                debug.mojolog.write("TCPCommsHandler: asyncore log: " + message + "\n", v=21, vs="commstrats")
        finally:
            pass #self._exit()


class TCPConnCache(Cache.SimpleCache):
    """
    A cache object, which closes TCP connections when their time is up, which
    preserves TCP connections that are in active use, and which keeps a few connections around
    if it has been hinted that they will be useful in the future (and which throws out the least
    frequently used when it needs to throw some out).

    Note that this inherits from SimpleCache which inherits from CoolLocker, so all access to a
    TCPConnCache object will be multithreaded according the CoolLocker mode specified by
    SimpleCache.
    """
    def __init__(self, cid_to_cs={}):
        """
        The following "virtual parameters" are pulled out of confman whenever they are referenced:
        @param TCP_MAINTAINED_CONNECTIONS the number of connections which have "probably will be
            used again someday" hints to keep open, or `-1' if all of them should be kept open
        @param TCP_MAX_CONNECTIONS the maximum number of conns to keep open (this is the
            absolute max, a `-1' in `TCP_MAINTAINED_CONNECTIONS' notwithstanding)
        @param TCP_TIMEOUT the number of seconds to give a connection even if we have no hint to
            keep it
        @param cid_to_cs a dict mapping cids to CommStrats; We look in the CommStrats to get
            hints.  We never change this dict.
        """
        Cache.SimpleCache.__init__(self)
        self._timelastcleaned = 0
        self._nice_cleanup_result = true
        self._cid_to_cs = cid_to_cs

    def remove_if_present(self, key, default=None):
        self._lock.acquire()
        try:
            if not self.has_key(key):
                return default

            tcpc = Cache.SimpleCache.remove(self, key)
            pn = None
            try:
                pn = tcpc.getpeername()
            except:
                pass

            debug.mojolog.write("TCPCommsHandler.TCPConnCache.remove(): key: %s, val: %s, val._cid_for_debugging: %s, val.getpeername(): %s\n", args=(key, tcpc, tcpc._cid_for_debugging, pn), v=8, vs="commstrats")
            if tcpc:
                tcpc.close_in_mojoasyncore()
            else:
                debug.mojolog.write("TCPCommsHandler.TCPConnCache.remove(): WARNING: didn't find any conn for cid %s\n", args=(key,), v=1, vs="ERROR")
            return tcpc
        finally:
            self._lock.release()

    def remove(self, key):
        """
        Over-riding SimpleCache's `remove()' to catch all things being removed from this cache.

        Remove the connection and call `close()' on it.
        """
        self._lock.acquire()
        try:
            tcpc = Cache.SimpleCache.remove(self, key)
            pn = None
            try:
                pn = tcpc.getpeername()
            except:
                pass

            debug.mojolog.write("TCPCommsHandler.TCPConnCache.remove(): key: %s, val: %s, val._cid_for_debugging: %s, val.getpeername(): %s\n", args=(key, tcpc, tcpc._cid_for_debugging, pn), v=8, vs="commstrats")
            if tcpc:
                tcpc.close_in_mojoasyncore()
            else:
                debug.mojolog.write("TCPCommsHandler.TCPConnCache.remove(): WARNING: didn't find any conn for cid %s\n", args=(key,), v=1, vs="ERROR")
            return tcpc
        finally:
            self._lock.release()

    def insert(self, key, item):
        """
        Over-riding SimpleCache's `insert()' to catch all things being inserted into this cache.

        This calls `_cleanup()' before inserting.

        @precondition `item' must be a TCPConnection.: isinstance(item, TCPConnection.TCPConnection): "item: %s :: %s" % (hr(item), hr(type(item)))
        """
        assert isinstance(item, TCPConnection.TCPConnection), "precondition: `item' must be a TCPConnection." + " -- " + "item: %s :: %s" % (hr(item), hr(type(item)))

        self._lock.acquire()
        try:
            # If this item is already in the cache under this key, we're done.
            if self.get(key) is item:
                return

            # If this item is already in the cache under an old key, then remove it from there...
            if item._key:
                if self.get(item._key) is item:
                    Cache.SimpleCache.remove(self, item._key)

            # Set the new key.
            item._key = key

            # If there is a different item under this key already, then remove it:
            if self.has_key(key):
                self.remove(key)

            if not self._cleanup(int(confman.get('TCP_MAX_CONNECTIONS', 50)) - 1):
                raise CommsHandler.CannotSendError, "too many busy connections"

            return Cache.SimpleCache.insert(self, key, item)
        finally:
            self._lock.release()

    def _cleanup(self, maxconns):
        """
        @returns "success" -- `true' if and only if it pared down to less than or equal to
            `maxconns' connections;  The only time it can fail is if there are more than `max'
            connections that are currently busy (they all return `true' from
            `TCPConnection.is_busy()').

        XXX Note: this might be a slow method when there are many (?? how many?) connections
        because it does a lot of sorting in order to choose the very best connections to keep.
        This could probably be sped-up but it probably wouldn't be a win since N is normally so
        low. In fact, one solution if this turns out to be slow when there are many connections
        is just to turn down TCP_MAX_CONNECTIONS to be small enough that this is fast!  :-)
        --Zooko 2000-08-15

        @precondition This method must be called on the DoQ.: DoQ.doq.is_currently_doq()
        """
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        def cmpmaybs(tupa, tupb, self=self):
            (ka, tcpa) = tupa
            (kb, tcpb) = tupb
            return -1 * cmp(tcpa._inmsgs, tcpb._inmsgs)

        def cmpprobs(tupa, tupb, self=self):
            tcp_timeout = int(confman.get('TCP_TIMEOUT', 60))
            (ka, tcpa) = tupa
            (kb, tcpb) = tupb
            if tcpa.is_busy(tcp_timeout):
                if tcpb.is_busy(tcp_timeout):
                    # Both are in the middle of something.
                    return -1 * cmp(tcpa._last_io_time, tcpb._last_io_time) # highest (most recent) io_time first
                else:
                    # tcpa is in the middle of something so it goes first.
                    return -1
            else:
                if tcpb.is_busy(tcp_timeout):
                    # tcpb is in the middle of something so it goes first.
                    return 1

            # Neither are in the middle of something.
            return -1 * cmp(tcpa._last_io_time, tcpb._last_io_time) # highest (most recent) io_time first

        self._lock.acquire()
        try:
            debug.mojolog.write("TCPConnCache._cleanup(): starting... self: %s, MAINT: %s, MAX: %s, len(self): %s\n", args=(self, confman.get('TCP_MAINTAINED_CONNECTIONS', '5'), confman.get('TCP_MAX_CONNECTIONS', '50'), len(self)), v=6, vs="comm hints") ### for faster operation, comment this line out.  --Zooko 2000-12-11

            now = time.time()

            # "maintain" connections, sorted with most frequently used first;  This list never
            # exceeds nummaintainedconns, and it is the first to go if we approach maxconns.
            listomaybes = []
            # busy and "expect response" connections, sorted so that the best bets are first;
            # This list never exceeds maxconns.
            listoprobablies = []

            def addtomaybes(k, tcpc, self=self, listomaybes=listomaybes, listoprobablies=listoprobablies, cmpmaybs=cmpmaybs, maxconns=maxconns):
                listomaybes.append((k, tcpc))
                # Sort the maybes with the most-used at the front of the list.
                listomaybes.sort(cmpmaybs)
                # If adding this put us over the limit...
                # We can have only this many maybes:
                cutoff = min(long(confman.get('TCP_MAINTAINED_CONNECTIONS', '5')), maxconns - len(listoprobablies))
                if (cutoff >= 0) and (len(listomaybes) > cutoff):
                    assert (cutoff + 1) == len(listomaybes), "internal error -- the listomaybes should be checked at each insert and never allowed to grow over limit."
                    (k, tcpc) = listomaybes.pop(cutoff)
                    assert cutoff == len(listomaybes)
                    pn = None
                    try:
                        pn = tcpc.getpeername()
                    except:
                        pass

                    if cutoff == (maxconns - len(listoprobablies)):
                        debug.mojolog.write("TCPConnCache._cleanup(): removing connection %s _cid_for_debugging: %s, .getpeername(): %s, which I 'maybe' wanted due to too many total connections.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                    else:
                        debug.mojolog.write("TCPConnCache._cleanup(): removing connection %s._cid_for_debugging: %s, .getpeername(): %s, which I 'maybe' wanted due to too many maintained connections.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                    self.remove(k)

            def addtoprobablies(k, tcpc, self=self, listomaybes=listomaybes, listoprobablies=listoprobablies, cmpprobs=cmpprobs, maxconns=maxconns):
                listoprobablies.append((k, tcpc))
                # Sort the probablies with the best at the front of the list.
                # First: active connections with half-finished messages and recent activity,
                # sorted with most recent network activity first.
                # Second: "expect response" connections sorted with most-frequently-used first.
                listoprobablies.sort(cmpprobs)
                # We can only have this many probablies:
                cutoff = maxconns
                if (cutoff >= 0) and (len(listoprobablies) > cutoff):
                    assert (cutoff + 1) == len(listoprobablies), "internal error -- the listoprobablies should be checked at each insert and never allowed to grow over limit."
                    (k, tcpc) = listoprobablies.pop(cutoff)
                    assert cutoff == len(listoprobablies)
                    # If we are about to kill a busy connection then don't do it and instead return `false'.
                    if tcpc.is_busy(long(confman.get('TCP_TIMEOUT', '60'))):
                        return false # failure
                    pn = None
                    try:
                        pn = tcpc.getpeername()
                    except:
                        pass

                    debug.mojolog.write("TCPConnCache._cleanup(): removing connection %s._cid_for_debugging: %s, .getpeername(): %s, which I 'probably' wanted due to too many total connections.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                    self.remove(k)

                # We can only have this many maybes:
                cutoff = min(long(confman.get('TCP_MAINTAINED_CONNECTIONS', '5')), maxconns - len(listoprobablies))
                if (cutoff >= 0) and (len(listomaybes) > cutoff):
                    assert (cutoff + 1) == len(listomaybes), "internal error -- the listomaybes should be checked at each insert and never allowed to grow over limit."
                    (k, tcpc) = listomaybes.pop(cutoff)
                    assert cutoff == len(listomaybes)
                    pn = None
                    try:
                        pn = tcpc.getpeername()
                    except:
                        pass

                    debug.mojolog.write("TCPConnCache._cleanup(): removing connection %s._cid_for_debugging: %s, .getpeername(): %s, due to too many maintained connections.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                    self.remove(k)


            for (k, tcpc) in self._dict.items():
                pn = None
                try:
                    pn = tcpc.getpeername()
                except:
                    pass

                cs = self._cid_to_cs.get(k)
                hint = 0 # XYZ XXX ???  why do we have no cs here sometimes?--Zooko 2001-02-07
                if cs:
                    hint = cs.hint

                    debug.mojolog.write("TCPConnCache._cleanup(): examining connection %s._cid_for_debugging: %s, .getpeername(): %s -- hint: %s, hintnumexpectedresponses: %s, hintnumexpectedsends: %s.\n", args=(tcpc, tcpc._cid_for_debugging, pn, hint, cs.hintnumexpectedresponses, cs.hintnumexpectedsends), v=21, vs="comm hints") ### for faster operation, comment this line out.  --Zooko 2000-12-11

                # Remove it if it is closed.
                if tcpc.is_closed():
                    debug.mojolog.write("TCPConnCache._cleanup(): removing connection %s._cid_for_debugging: %s, .getpeername(): %s, because it has been closed.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                    self.remove(k)
                    continue

                # Leave it alone if it is busy.
                if tcpc.is_busy(long(confman.get('TCP_TIMEOUT', '60'))):
                    debug.mojolog.write("TCPConnCache._cleanup(): keeping connection %s._cid_for_debugging: %s, .getpeername(): %s probably: it is busy.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=7, vs="comm hints") ### for faster operation, comment this line out.  --Zooko 2000-12-11

                    addtoprobablies(k, tcpc)
                    continue

                if (hint & HINT_EXPECT_RESPONSE) or (hint & HINT_EXPECT_TO_RESPOND):
                    # We're waiting for a response.  This one is a definite keeper.
                    debug.mojolog.write("TCPConnCache._cleanup(): keeping connection %s._cid_for_debugging: %s, .getpeername(): %s probably: waiting for response.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=21, vs="comm hints") ### for faster operation, comment this line out.  --Zooko 2000-12-11

                    addtoprobablies(k, tcpc)
                    continue
                if hint & HINT_EXPECT_NO_MORE_COMMS:
                    # loser
                    debug.mojolog.write("TCPConnCache._cleanup(): removing connection %s._cid_for_debugging: %s, .getpeername(): %s, due to HINT_EXPECT_NO_MORE_COMMS.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                    self.remove(k)
                    continue
                if hint & HINT_EXPECT_MORE_TRANSACTIONS:
                    debug.mojolog.write("TCPConnCache._cleanup(): keeping connection %s._cid_for_debugging: %s, .getpeername(): %s maybe: expect more transactions.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=21, vs="comm hints") ### for faster operation, comment this line out.  --Zooko 2000-12-11
                    addtomaybes(k, tcpc)
                    continue

                # If we have no better clue to go on, then just time it out if it is old.
                if tcpc.is_idle(long(confman.get('TCP_TIMEOUT', '60'))):
                    debug.mojolog.write("TCPConnCache._cleanup(): removing connection %s._cid_for_debugging: %s, .getpeername(): %s, due to idleness.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=5, vs="comm hints")
                    self.remove(k)
                else:
                    debug.mojolog.write("TCPConnCache._cleanup(): keeping connection %s._cid_for_debugging: %s, .getpeername(): %s maybe: it isn't idle.\n", args=(tcpc, tcpc._cid_for_debugging, pn), v=8, vs="comm hints") ### for faster operation, comment this line out.  --Zooko 2000-12-11
                    pass

                    addtomaybes(k, tcpc)

            return true # success
        finally:
            self._timelastcleaned = time.time()
            debug.mojolog.write("TCPConnCache._cleanup(): finishing.  len(self): %s\n", args=(len(self),), v=21, vs="comm hints") ## verbose connection caching diags ### for faster operation, comment this line out.  --Zooko 2000-12-11
            self._lock.release()

    def _cleanup_nice(self):
        """
        This is "nice" because it just returns if less than 5 seconds have elapsed since the
        last cleanup.
        """
        self._lock.acquire()
        try:
            if (time.time() - self._timelastcleaned) < 5:
                return self._nice_cleanup_result

            self._nice_cleanup_result = self._cleanup(long(confman.get('TCP_MAX_CONNECTIONS', '50')))
            return self._nice_cleanup_result
        finally:
            self._lock.release()

def test_a_to_b_and_back():
    as = mojoasyncore.Asyncore()
    as.start_asyncore()
    try:
        ler = TCPCommsHandler()

        def _handle_first(counterparty_id, msg, commstrat):
            commstrat.asyncsock.send("gretings, oh unlightened one!")

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper(_handle_first)

        res2, reskw2, doneflag2, _wcb2 = mojoutil._cb_warper()

        ler.start_listening(_wcb1)
        lercs = ler.get_comm_strategy()
        ser = TCPCommsHandler()
        ser.start_listening(_wcb2)
        ser.use_comm_strategy(ler.get_id(), ler.get_comm_strategy())
        ser.send_msg(ler.get_id(), "hi there silly face!")

        doneflag1.wait(3) # If it takes longer than 3 seconds you fail the test!
        assert (reskw1.get('msg') == "hi there silly face!") or ((len(res1) > 1) and (res1[1] == "hi there silly face!")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))
        doneflag2.wait(3) # If it takes longer than 3 seconds you fail the test!
        assert (reskw2.get('msg') == "gretings, oh unlightened one!") or ((len(res1) > 1) and (res2[1] == "gretings, oh unlightened one!")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))
    finally:
        as.stop_asyncore()


def test_drop_connection_and_detect_dropped_connection():
    as = mojoasyncore.Asyncore()
    as.start_asyncore()
    try:
        ler = TCPCommsHandler()

        def _handle_first(counterparty_id, msg, commstrat):
            commstrat.asyncsock.send("gretings, oh unlightened one!")

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper(_handle_first)

        res2, reskw2, doneflag2, _wcb2 = mojoutil._cb_warper()

        ler.start_listening(_wcb1)
        lercs = ler.get_comm_strategy()
        ser = TCPCommsHandler()
        ser.start_listening(_wcb2)
        ser.use_comm_strategy(ler.get_id(), ler.get_comm_strategy())
        ser.send_msg(ler.get_id(), "hi there silly face!")

        doneflag1.wait(3) # If it takes longer than 3 seconds you fail the test!
        assert (reskw1.get('msg') == "hi there silly face!") or ((len(res1) > 1) and (res1[1] == "hi there silly face!")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))
        doneflag2.wait(3) # If it takes longer than 3 seconds you fail the test!
        assert (reskw2.get('msg') == "gretings, oh unlightened one!") or ((len(res2) > 1) and (res2[1] == "gretings, oh unlightened one!")), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))

        # Now close one side of it.
        cs = (reskw1.get('commstrat') or res1[2])
        cs.asyncsock.close()

        try:
            time.sleep(1) # necessitated by the pause during select in asyncore,  --Zooko 2000-07-25
            ser.send_msg(ler.get_id(), "hi there silly face!")
            raise "should have gotten CannotSendError"
        except CommsHandler.CannotSendError: # testing
            return # success
    finally:
        as.stop_asyncore()


def _help_test_open_connections_and_ping_pong(NUM, MAINT, MAX):
    """
    This should exercise some "clean up of old connections" behavior...
    """
    as = mojoasyncore.Asyncore()
    as.start_asyncore()
    try:
        confman['TCP_MAINTAINED_CONNECTIONS'] = MAINT
        confman['TCP_MAX_CONNECTIONS'] = MAX

        ler = TCPCommsHandler()

        def _handle_first(counterparty_id, msg, commstrat, ler=ler):
            ler.use_comm_strategy(counterparty_id, commstrat)
            ler.send_msg(counterparty_id, "gretings, oh unlightened one! msg: %s" % hr(msg), hint=HINT_EXPECT_MORE_TRANSACTIONS | HINT_THIS_IS_A_RESPONSE)

        res1, reskw1, doneflag1, _wcb1 = mojoutil._cb_warper(_handle_first)

        ler.start_listening(_wcb1)
        lercs = ler.get_comm_strategy()

        reskws = []
        ress = []
        doneflags = []
        sers = []

        for i in range(NUM):
            ser = TCPCommsHandler(listenport=(9000 + i))
            res2, reskw2, doneflag2, _wcb2 = mojoutil._cb_warper()
            reskws.append(reskw2)
            ress.append(res2)
            doneflags.append(doneflag2)

            sers.append(ser)
            ser.start_listening(_wcb2)
            ser.use_comm_strategy(ler.get_id(), ler.get_comm_strategy())

        for i in range(len(sers)):
            sers[i].send_msg(ler.get_id(), "hi there silly face! i: %s" % hr(i), hint=HINT_EXPECT_RESPONSE)

        doneflag1.wait(1 * NUM) # If it takes longer than 1 second per pingpong you fail the test!
        assert (reskw1.get('msg')) or ((len(res1) > 1) and (res1[1])), "reskw1: %s, res1: %s" % (hr(reskw1), hr(res1))

        doneflags[0].wait(2) # If it takes longer than 2 seconds you fail the test!
        for i in range(len(doneflags)):
            doneflags[i].wait(1 * NUM) # If it takes longer than 1 second per pingpong you fail the test!
            assert (reskws[i].get('msg')) or ((len(ress[i]) > 1) and (ress[i][1])), "i: %s, reskws[i]: %s, ress[i]: %s" % (hr(i), `reskws[i]`, `ress[i]`)
    finally:
        as.stop_asyncore()

def test_open_15_connections_maint_5_max_20():
    _help_test_open_connections_and_ping_pong(15, 5, 20)

def test_open_15_connections_maint_5_max_10():
    _help_test_open_connections_and_ping_pong(15, 5, 10)

def test_open_15_connections_maint_20_max_20():
    _help_test_open_connections_and_ping_pong(15, 20, 20)

def test_open_15_connections_maint_20_max_10():
    _help_test_open_connections_and_ping_pong(15, 20, 10)


# Generic stuff
NAME_OF_THIS_MODULE="TCPCommsHandler"

mojo_test_flag = 1

def run():
    confman['MAX_VERBOSITY'] = "9"
    import RunTests
    RunTests.runTests(NAME_OF_THIS_MODULE)

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
