#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import asyncore
import threading
import time
import socket
import types
import struct
import traceback
import sys

### our modules
import Buf
import CommsHandler
import DoQ
from MojoConstants import true, false
import SingleThreader
import debug
from human_readable import hr
import idlib
import mojoutil
import mojoasyncore
from confutils import confman

# only for testing
import RunTests


# This is the maximum lowlevel EGTP message size; attempting to
# receive a message longer than this will cause the EGTP connection
# to be aborted.  Attempting to send a message longer than this
# will raise a ValueError.
MAXIMUM_MSG_SIZE = 640*1024   # 640k



# class TCPConnection(asyncore.dispatcher, SingleThreader.BombingSingleThreader):
class TCPConnection(asyncore.dispatcher, SingleThreader.SingleThreader):
    """
    Sends and receives buffers on TCP connections.  Prepends lengths for each message.
    """
    def __init__(self, inmsg_handler_func, key, host=None, port=None, sock=None,
                 commstratobj=None, throttlerin=None, throttlerout=None,
                 cid_for_debugging=None):
        """
        @param key a key for identifying this connection;  (Hint: if you know the counterparty
            id use that, else use `idlib.make_new_random_id(thingtype='TCPConnection')')

        @precondition `key' is an id.: idlib.is_sloppy_id(key): "key: %s :: %s" % (hr(key), hr(type(key)))
        """
        assert idlib.is_sloppy_id(key), "precondition: `key' is an id." + " -- " + "key: %s :: %s" % (hr(key), hr(type(key)))

        key = idlib.canonicalize(key)

        # SingleThreader.BombingSingleThreader.__init__(self)
        SingleThreader.SingleThreader.__init__(self)

        # `_cid_for_debugging' is for debugging.
        self._cid_for_debugging = cid_for_debugging

        # `_key' is what we use to index this object in the TCPConnCache.  If we know the
        # counterparty id, then `_key' gets set to the counterparty id by the higher-level code
        # (by direct access to the `_key' member), else `_key' is set to a random unique id by
        # the higher-level code (in the constructor).
        self._key = key
        # XXX Hey -- could this be causing unexpected behaviour at some point if there were two different comm strats using the same key because they were with the same counterparty?  It shouldn't, because currently we supposedly always forget an old comm strat for a given counterparty before trying a new one.  (Although I don't know off the top of my head if this is accomplished with an if: condition or just by replacing the entry in the _conncache.)  But in the future we might keep multiple comm strats for one counterparty, so perhaps we shouldn't use the counterparty id as the key...  --Zooko 2001-05-07

        # for passing through to the fast fail handler
        self._commstratobj = commstratobj

        self._upward_inmsg_handler = inmsg_handler_func

        self._throttlerread = throttlerin
        self._throttlerwrite = throttlerout

        self._outbuf = Buf.DumbBuf(MAXIMUM_MSG_SIZE)
        self._inbuf = Buf.DumbBuf(MAXIMUM_MSG_SIZE)
        self._current_fast_fail_handler = None

        self._outmsgq = [] # contains (msg, fast_fail_handler_func)

        if sock:
            self._everconnected = true  # we're already connected
        else:
            self._everconnected = false # `handle_connect()' sets this to true
        self._closing = false
        self._closed = false # This gets set to `true' in `close()'.
        self._last_io_time = time.time() # the last time an IO event happened on this connection
        self._inmsgs = 0 # The total number of incoming messages that have come through this connection.
        self._nummsgs = 0 # The sum of outgoing and incoming messages that have come through this connection.
        self._outbytes = 0L # The total bytes ever sent over this connection.
        self._inbytes = 0L # The total bytes ever received over this connection.

        # this lock is used to prevent more than one thread from being
        # inside of handle_write (manipulating _outbuf) at once.  (both the
        # asyncore polling thread and external threads [DoQ or other] can
        # trigger a handle_write [external threads trigger it from send()])
        self._write_lock = threading.Lock()

        asyncore.dispatcher.__init__(self, sock=sock)
        if not sock:
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                self.connect((host, port))
            except socket.error, le:
                # This is not guaranteed to detect a failed connection, since connect() is a non-blocking call.  But if we _do_ get a socket error here then we have definitely failed.
                debug.stderr.write("TCPConnection: couldn't connect to (%s, %s): %s\n", args=(host, port, le), v=1, vs="commstrats")
                self.close_in_mojoasyncore()
                raise CommsHandler.CannotSendError, ("socket.error on connect", le)
            
            DoQ.doq.add_task(self._fail_if_not_connected, delay=int(confman['TCP_CONNECT_TIMEOUT']))

        pn = None
        try:
            pn = self.getpeername()
        except:
            pass

        debug.stderr.write("TCPConnection(): %s, cid: %s, host: %s, port: %s, pn: %s\n", args=(self, self._cid_for_debugging, host, port, pn), v=5, vs="debug")

    def _fail_if_not_connected(self):
        if self._everconnected:
            return
        # this causes it to be cleaned up and the fast fail handlers to be called appropriately
        self._fast_fail_reason = "connect() timeout"
        self.close_in_mojoasyncore()

    def close_in_mojoasyncore(self):
        # Put a new DoQ task so that any processing that is still queued on the DoQ will finish before the socket closes.
        if not self._closing:
            DoQ.doq.add_task(mojoasyncore.TCPConnections_to_close.append, args=(self,))
        self._closing = true

    def is_closing(self):
        return self._closing

    def is_closed(self):
        return self._closed

    def send(self, msg, fast_fail_handler=None):
        self._enter()
        try:
            if self.is_closing():
                raise CommsHandler.CannotSendError, "connection closed"

            # TODO this is gross, it does a full data copy of the message just to prepend 4 bytes
            # (send should ideally accept buffer chains instead of a single string messages)
            # (this is not such a big deal as we're already spending many more cycles to encrypt every message)
            str = struct.pack('>L', len(msg)) + msg

            self._outmsgq.append((str, fast_fail_handler))

            self._nummsgs = self._nummsgs + 1

            # try to send some data immediately if possible
            self.handle_write(called_from_send=true)

            DoQ.doq._asyncorelooper.wake_select()
        finally:
            self._exit()

    def is_idle(self, idletimeout=30):
        """
        @returns `true' if and only if there have been no I/O events have occured on this socket in >= idletimeout seconds or if it is closed
        """
        if self.is_closing():
            return true
        return (time.time() - self._last_io_time) >= idletimeout

    def is_talking(self):
        """
        @returns `true' if and only if there is a message actually half-sent or half-received
        """
        return (self._outbuf.len() > 0) or (self._inbuf.len() > 0) or (len(self._outmsgq) > 0)

    def is_busy(self, idletimeout):
        """
        @returns `true' if and only if (there is a message actually (half-sent or half-received)
            and not `is_idle(idletimeout)')
        """
        return self.is_talking() and not self.is_idle(idletimeout)

    #### methods below here are for internal use
    # The `handle_spam()' methods are the "bottom" interface, to be called by the asyncore thread.  There can be (one) thread touching the bottom interface and (one) thread touching the top interface at the same time.  Also `_chunkify()', which get called from `handle_spam()' methods.

    def close(self):
        """
        This should only be called from the asyncore polling thread!!!

        It calls `asyncore.dispatcher.close(self)' to clean up the socket object.

        It also does a fast-fail on any messages that were queued to be sent but haven't been
        sent yet.
        """
        self._closing = true
        asyncore.dispatcher.close(self)

        if (not self._everconnected) or (self._outbytes == 0 and self._inbytes == 0):
            debug.mojolog.write("connection refused to %s: commstrat=%s, _everconnected=%s, _outbytes=%s, _inbytes=%s\n", args=(self._cid_for_debugging, self._commstratobj, self._everconnected, self._outbytes, self._inbytes), v=5, vs="TCPConnection")
            connection_refused = 1
        else:
            connection_refused = None

        # Fail any partially sent message:
        if (self._outbuf.len() > 0) and (self._current_fast_fail_handler):
            if hasattr(self, '_fast_fail_reason'):
                self._current_fast_fail_handler(failure_reason="TCPConnection: "+self._fast_fail_reason, bad_commstrat=self._commstratobj)
            elif connection_refused:
                self._current_fast_fail_handler(failure_reason="TCPConnection: connection refused", bad_commstrat=self._commstratobj)
            else:
                self._current_fast_fail_handler(failure_reason="TCPConnection: closed before message was sent")

        # Now fail any queued messages.
        while len(self._outmsgq) > 0:
            (str, ffh) = self._outmsgq.pop(0)
            if ffh:
                if connection_refused:
                    ffh(failure_reason="TCPConnection: connection refused", bad_commstrat=self._commstratobj)
                else:
                    ffh(failure_reason="TCPConnection: cannot send message")

        self._closed = true

    def handle_write(self, called_from_send=false):
        self._write_lock.acquire()
        try:
            if self.is_closing():
                return

            # if send() called us we only want to proceed if out _outbuf does not
            # already have data in it.  We should only empty an existing _outbuf
            # when asyncore.poll calls us.
            if called_from_send and (not self._everconnected or (self._outbuf.len() > 0)):
                return

            self._last_io_time = time.time()

            # load up the next message if any.
            if (self._outbuf.len() == 0) and (len(self._outmsgq) > 0):
                (str, ffh) = self._outmsgq.pop(0)
                self._current_fast_fail_handler = ffh
                self._outbuf.append(str)

            if self._outbuf.len() > 0:
                try:
                    num_sent = asyncore.dispatcher.send(self, self._outbuf.read())
                    debug.stderr.write("TCPConnection(%s).handle_write(): sent [%s] bytes\n", args=(self, num_sent), v=9, vs="commstrats") ### for faster operation, comment this line out.  --Zooko 2000-12-11
                except socket.error, le:
                    debug.stderr.write("TCPConnection.handle_write(): got exception: %s\n", args=(le.args,), v=6, vs="commstrats")
                    self._write_lock.release()  # release our lock before making a call out
                    self.close_in_mojoasyncore()
                    return

                self._outbytes = self._outbytes + num_sent
                self._outbuf.chop(num_sent)
                self._throttlerwrite.used(num_sent) # notify throttler we just used up some bandwidth

        finally:
            if self._write_lock.locked(): self._write_lock.release()

    def writable(self):
        """returns true if we're willing to service handle_write events for sending data"""
        #self._enter()
        try:
            if (self.is_closing()) or (self._throttlerwrite.is_maxed()):
                ###debug.mojolog.write("TCPConnection%s.writable(): `false', because self.is_closing(): %s, self._throttlerwrite.is_maxed(): %s\n", args=(self, self.is_closing(), self._throttlerwrite.is_maxed(),), v=20, vs='debug') ### for faster operation, comment this line out.  --Zooko 2000-12-11
                return 0

            ###debug.mojolog.write("TCPConnection%s.writable(): %s, because self._outbuf.len(): %s, len(self._outmsgq): %s\n", args=(self, ((self._outbuf.len() > 0) or (len(self._outmsgq) > 0)), self._outbuf.len(), len(self._outmsgq),), v=20, vs='debug') ### for faster operation, comment this line out.  --Zooko 2000-12-11
            return (self._outbuf.len() > 0) or (len(self._outmsgq) > 0)
        finally:
            pass #self._exit()

    def _chunkify(self, nextstream):
        #self._enter()
        try:
            self._inbuf.append(nextstream)

            # Now try to read length-prefixed messages out of here until we don't have any more.
            while self._inbuf.len() >= 4:
                str = self._inbuf.read()

                binsize = str[:4]
                amount = struct.unpack('>L', binsize)[0]

                assert type(amount) == types.LongType

                if amount > MAXIMUM_MSG_SIZE:
                    # Too big.
                    self.close_in_mojoasyncore()

                if len(str) < (4 + amount):
                    # The message isn't here (yet).
                    return

                intamt = int(amount)

                msg = str[4:(4 + intamt)]
                self._inbuf.chop(4 + intamt)

                DoQ.doq.add_task(self._upward_inmsg_handler, args=(self, msg))

                self._inmsgs = self._inmsgs + 1
                self._nummsgs = self._nummsgs + 1
        finally:
            pass #self._exit()

    def handle_read(self):
        #self._enter()
        try:
            if self.is_closing():
                return

            self._last_io_time = time.time()
            try:
                data = self.recv(65536)
                debug.stderr.write("TCPConnection(%s).handle_read(): received [%s] bytes\n", args=(self, len(data)), v=9, vs="commstrats") ### for faster operation, comment this line out.  --Zooko 2000-12-11
            except socket.error, le:
                # This is the socket's way of telling us that we are _closed_.
                debug.stderr.write("TCPConnection(%s): closed socket detected. le:%s\n", args=(self, le), v=5, vs="commstrats")
                self.close_in_mojoasyncore()
                return

            if len(data) == 0:
                # This is the socket's way of telling us that we are _closed_.
                debug.stderr.write("TCPConnection(%s): closed socket detected.  (read of length 0)\n", args=(self,), v=5, vs="commstrats")
                self.close_in_mojoasyncore()
                return
            self._inbytes = self._inbytes + len(data)
            self._throttlerread.used(len(data)) # notify throttler we just used up some bandwidth
            self._chunkify(data)
        finally:
            pass #self._exit()

    def handle_accept(self) :
        debug.stderr.write("TCPConnection.handle_accept() checking to see if this gets called...\n", v=0, vs="debug")

    def handle_connect(self):
        self._last_io_time = time.time()
        self._everconnected = true
        pn = None
        try:
            pn = self.getpeername()
        except:
            pass
        debug.stderr.write("TCPConnection(%s).handle_connect(), cid: %s, pn: %s\n", args=(self, self._cid_for_debugging, pn), v=5, vs="commstrats")

    def handle_close(self):
        debug.stderr.write("TCPConnection(%s).handle_close()\n", args=(self,), v=5, vs="commstrats")
        self.close_in_mojoasyncore()

    def readable(self) :
        """returns true if we're willing to service handle_read events for receiving data"""
        #self._enter()
        try:
            ###debug.mojolog.write("TCPConnection%s.readable(): %s, self.is_closing(): %s, self._throttlerread.is_maxed(): %s\n", args=(self, ((not self.is_closing()) and (not self._throttlerread.is_maxed())), self.is_closing(), self._throttlerread.is_maxed(),), v=20, vs="debug") ### for faster operation, comment this line out.  --Zooko 2000-12-11
            return (not self.is_closing()) and (not self._throttlerread.is_maxed())
        finally:
            pass #self._exit()

    def log(self, message):
        # for faster operation, comment this whole method out and replace it with "def log(): return".  --Zooko 2000-12-11
        # return
        #self._enter()
        try:
            if message[-1:] == "\n":
                debug.stderr.write("TCPConnection: asyncore log: %s", args=(message,), v=15, vs="commstrats")
            else:
                debug.stderr.write("TCPConnection: asyncore log: %s\n", args=(message,), v=15, vs="commstrats")
        finally:
            pass #self._exit()


# Generic stuff
NAME_OF_THIS_MODULE="TCPConnection"

mojo_test_flag = 1

def run():
    RunTests.runTests(NAME_OF_THIS_MODULE)

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
