#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
from threading import Thread, Condition, currentThread

class ThreadManager:
    def __init__(self, min_threads = 4, max_threads = 8):
        assert max_threads > 0
        self._min_threads = min_threads
        self._max_threads = max_threads
        self._lock = Condition()
        self._running = 1
        self._num_threads_waiting = 0
        self._tasks = []
        self._non_blocking_threads = {}
        for i in range(max_threads):
            self._start_new_thread()

    def add_task(self, task):
        """
        cause task to get executed by a managed thread
        """
        assert callable(task)
        self._lock.acquire()
        try:
            if not self._running:
                return
            self._tasks.append(task)
            self._start_new_if_needed()
            self._lock.notify()
        finally:
            self._lock.release()

    def about_to_block(self):
        """
        called before making a blocking call, to let new threads be 
        created if necessary
        """
        self._lock.acquire()
        try:
            if not self._running:
                return
            try:
                del self._non_blocking_threads[currentThread()]
            except KeyError:
                return
            self._start_new_if_needed()
        finally:
            self._lock.release()

    def shut_down(self):
        """
        stop doing anything - get rid of all threads
        """
        self._lock.acquire()
        try:
            self._running = 0
            del self._tasks[:]
            self._lock.notifyAll()
        finally:
            self._lock.release()

    def _start_new_if_needed(self):
        if not self._running:
            return
        if len(self.tasks) == 0 or self._num_threads_waiting > 0:
            return
        diff = self._min_threads - len(self._non_blocking_threads)
        if diff > 0:
            for i in range(diff):
                self.start_new_thread()

    def _start_new_thread(self):
        thread = Thread(target = self._poll)
        thread.setDaemon(1)
        thread.start()
        
    def _poll(self):
        try:
            while 1:
                self._lock.acquire()
                try:
                    self._num_threads_waiting = self._num_threads_waiting + 1
                    if self._num_threads_waiting > self._max_threads:
                        return
                    self._non_blocking_threads[currentThread()] = 1
                    while not self._tasks:
                        if not self._running:
                            return
                        self._lock.wait()
                    task = self._tasks.pop(0)
                finally:
                    self._num_threads_waiting = self._num_threads_waiting - 1
                    self._start_new_if_needed()
                    self._lock.release()
                task()
        finally:
            self.about_to_block()





