#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

# standard modules
import time
import types

# our modules
from MojoConstants import DEFAULT_TIMEOUT, true, false
import pricelib
import idlib
import mojoutil
import debug


def purchase_block_and_callback(mtm, block_id, block_size, serverinfotuple, callback, checkhandicap=true):
    """
    Issues a request blob message to the server described in serverinfotuple and calls the
    callback after processing the response and storing the block on our own block server
    if it fits our mask.

    @param callback - must be callable and accept arguments as follows:
            (server_id, block_id, seconds_elapsed=0, mojo_spent=0, block_data=None, failure_reason=None)
    @param checkhandicap - controls if this ever denies the query due to the handicapper rejecting the given server.
    """
    assert idlib.is_sloppy_id(block_id), "bad block id"
    assert block_size > 0, "bad block size"
    assert len(serverinfotuple) == 2 and idlib.is_sloppy_id(serverinfotuple[0]), "bad serverinfotuple"
    assert callable(callback), 'callback must be callable'

    message_type = 'request blob'
    message_body = {'blob id': idlib.to_binary(block_id), 'len': block_size}

    server_id, server_info = serverinfotuple
    topay = pricelib.blob_server_purchase_price(server_info, message_body)

    if checkhandicap:
        _tmp_infolist = mtm.get_handicapper().sort_by_preference([serverinfotuple], message_type, message_body)
        if not _tmp_infolist:
            mtm._doQ.add_task(callback, kwargs = {'server_id': server_id, 'block_id': block_id, 'failure_reason': 'server denied by handicapping'})
            return

    timer = mojoutil.Timer()
    def process_callback(widget, outcome=None, failure_reason=None, notes=None,
                            block_id=block_id, block_size=block_size, callback=callback,
                            timer=timer, mtm=mtm, topay=topay):
        timer.stop()
        server_id = widget.get_counterparty_id()

        if failure_reason is not None:
            # TODO when we have refund logic in place, tell the callback the actual amount paid so it can be passed up properly here.
            callback(server_id=server_id, block_id=block_id, seconds_elapsed=timer.get(), failure_reason=failure_reason)
            return
        if not outcome or outcome.get('result') == 'failure':
            # TODO when we have refund logic in place, tell the callback the actual amount paid so it can be passed up properly here.
            callback(server_id=server_id, block_id=block_id, seconds_elapsed=timer.get(), mojo_spent=topay,
                        failure_reason='server returned failure')
            return

        block_data = outcome['data']
        if not idlib.equal(block_id, idlib.make_id(block_data)):
            # TODO when we have refund logic in place, tell the callback the actual amount paid so it can be passed up properly here.
            callback(server_id=server_id, block_id=block_id, seconds_elapsed=timer.get(), mojo_spent=topay,
                        failure_reason='server returned block with wrong id')
            return

        # store the latency overall and on a per-block size basis
        cp_obj = mtm._keeper.get_counterparty_object(server_id)
        cp_obj.update_custom_stat_weighted_sample('request blob bytes/sec', float(len(block_data))/timer.get())

        # store the block on our own block server if were running one and it matches our mask
        if hasattr(mtm, '_blobserver'):
            if mtm._blobserver.try_to_store_blob(block_id, block_data):
                debug.mojolog.write("stored block %s on our server.\n", args=(block_id,), v=6, vs="blockserverclientlib")

        # TODO when we have refund logic in place, tell the callback the actual amount paid so it can be passed up properly here.
        callback(server_id=server_id, block_id=block_id, seconds_elapsed=timer.get(), mojo_spent=topay, block_data=block_data)

    # choose a timeout less than the default if it makes sense (most common on small blocks).
    # reminder: this connection speed applies to -both ends- of the connection, when buying from a
    # broker on a modem you will have a painfully slow transaction no matter what.
    connection_speed_bytes_per_second = 2400  # 2400 should be true of even a non-great 28.8k modem connection
    # timeouts are always at least 45 just due to general software load and send/receive via relay issues
    maxtime = min( max(block_size / connection_speed_bytes_per_second, 45), DEFAULT_TIMEOUT)

    timer.start()
    mtm.initiate(server_id, message_type, message_body, topay, outcome_func=process_callback, timeout=maxtime)

