/*
 * mencode.py helper functions written in C for speed
 * 
 * $Id: _c_mencode_help.c,v 1.4 2001/06/04 19:27:46 greg Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h> /* debugging */

#ifdef __cplusplus
extern "C" {
#endif
#include <Python.h>
#ifdef __cplusplus
}
#endif


#define RETURN_IF_EXC(ret)    if (!ret) { if (owrite_method) { Py_DECREF(owrite_method); } return NULL;  }


static PyObject* MencodeError;

static PyObject *colon_tup_const, *close_paren_tup_const, *str_tup_const, *dict_tup_const;

static PyObject *encoder_dict;

// need this prototype
static PyObject *real_encode_io(PyObject *data, PyObject *output);


static PyObject *real_encode_string(PyObject *data, PyObject *output)
{
    PyObject *ret, *str, *length = NULL;
    PyObject *owrite_method;

    owrite_method = PyObject_GetAttrString(output, "write");
    if (!owrite_method || !PyCallable_Check(owrite_method)) {
        PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "output object must have a callable write method"));
        if (owrite_method) { Py_DECREF(owrite_method); }
        return NULL;
    }

    //    result.write('(6:string')
    ret = PyObject_CallObject(owrite_method, str_tup_const);
    RETURN_IF_EXC(ret);
    Py_DECREF(ret);
 
    //    result.write(str(len(data)))
    length = PyInt_FromLong((long)PyObject_Length(data));
    str = PyObject_Str(length);
    ret = PyObject_CallFunction(owrite_method, "S", str);
    Py_DECREF(length);
    Py_DECREF(str);
    RETURN_IF_EXC(ret);
    Py_DECREF(ret);
    
    // result.write(':')
    ret = PyObject_CallObject(owrite_method, colon_tup_const);
    RETURN_IF_EXC(ret);
    Py_DECREF(ret);
    
    // result.write(str(data))
    ret = PyObject_CallFunction(owrite_method, "S", data);
    RETURN_IF_EXC(ret);
    Py_DECREF(ret);
    
    // result.write(')')
    ret = PyObject_CallObject(owrite_method, close_paren_tup_const);
    RETURN_IF_EXC(ret);
    Py_DECREF(ret);
    
    if (owrite_method) { Py_DECREF(owrite_method); }

    Py_INCREF(Py_None);
    return Py_None;
} 

static PyObject *encode_string(PyObject *self, PyObject *args)
{
    PyObject *data, *output = NULL;
    
    if (!PyArg_ParseTuple(args, "OO:encode_string", &data, &output)) {
        PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "expected two python objects as parameters"));
        return NULL;
    }
    return real_encode_string(data, output);
}

static PyObject *real_encode_dict(PyObject *data, PyObject *output)
{
    PyObject *keys, *key, *ret;
    int i, keylen;
    PyObject *owrite_method;

    owrite_method = PyObject_GetAttrString(output, "write");
    if (!owrite_method || !PyCallable_Check(owrite_method)) {
        PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "output object must have a callable write method"));
        if (owrite_method) { Py_DECREF(owrite_method); }
        return NULL;
    }

    // result.write('(4:dict')
    ret = PyObject_CallObject(owrite_method, dict_tup_const);
    RETURN_IF_EXC(ret);
    Py_DECREF(ret);

    // keys = data.keys()
    keys = PyDict_Keys(data);

    // keys.sort()
    // XXX FIXME bad!  this depends on the python implementations cross type comparison (string > integer or long)
    if (PyList_Sort(keys) != 0) {
        PyErr_SetObject(MencodeError, Py_BuildValue("(is)", 0, "PyList_Sort failed [returned non zero]"));
        Py_DECREF(keys);
        RETURN_IF_EXC(0);
    }
    
    keylen = PyList_Size(keys);
    // for key in keys:
    for (i = 0; i < keylen; i++) {
        key = PyList_GetItem(keys, i);
        //     if type(key) not in (types.StringType, types.IntType, types.LongType):
        // TODO support BufferType in the future
        if (!PyString_Check(key) && !PyDict_Check(key) && !PyInt_Check(key) && !PyLong_Check(key)) {
                Py_DECREF(keys);
                // it would be nice if this exception included the key we tried to encode...
                PyErr_SetObject(MencodeError, Py_BuildValue("(is)", 0, "mencoded dictionary keys must be strings or numbers"));
                RETURN_IF_EXC(0);
        }
        //     encode_io(key, result)
        real_encode_io(key, output);
        //     encode_io(data[key], result)
        real_encode_io(PyDict_GetItem(data, key), output);
    }
    
    Py_DECREF(keys);
    
    // result.write(')')
    ret = PyObject_CallObject(owrite_method, close_paren_tup_const);
    RETURN_IF_EXC(ret);
    Py_DECREF(ret);

    if (owrite_method) { Py_DECREF(owrite_method); }

    Py_INCREF(Py_None);
    return Py_None;
} 


static PyObject *encode_dict(PyObject *self, PyObject *args)
{
    PyObject *data, *output;
    
    if (!PyArg_ParseTuple(args, "OO:encode_dict", &data, &output)) {
        PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "expected two python objects as parameters"));
        return NULL;
    }
    
    return real_encode_dict(data, output);
}


static PyObject *real_encode_io(PyObject *data, PyObject *output)
{
    PyObject *type, *func, *tup, *ret = NULL;
    if (PyString_Check(data)) {
            /* TODO accept buffer objects */
            return real_encode_string(data, output);
        }
    else if (PyDict_Check(data)) {
            return real_encode_dict(data, output);
        }
    else {
            type = PyObject_Type(data);
            func = PyDict_GetItem(encoder_dict, type);   /* borrowed reference to func */
            if (!func) {
                /* TODO include a reference to data in this error */
                PyErr_SetObject(MencodeError, Py_BuildValue("(is)", 0, "encoder_dict did not contain an encoding function for data"));
                return NULL;
            }
            Py_DECREF(type);
            tup = PyTuple_New(2);
            Py_INCREF(data);
            Py_INCREF(output);
            PyTuple_SET_ITEM(tup, 0, data);
            PyTuple_SET_ITEM(tup, 1, output);
            ret =  PyObject_CallObject(func, tup);
            Py_DECREF(tup);
            return ret;
        }
}

static PyObject *encode_io(PyObject *self, PyObject *args)
{
    PyObject *data, *output;
    if (!PyArg_ParseTuple(args, "OO:encode_io", &data, &output)) {
        PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "expected two python objects as parameters"));
        return NULL;
    }
    return real_encode_io(data, output);
}


/* @param a python string object and an integer index as python parameters */
/* @returns tuple of a python string and an ending index into its string parameter */
static PyObject *decode_raw_string(PyObject *self, PyObject *args)
{
    PyObject *dataobj = NULL;
    int dataptrlen;
    int index, strlength, dataend;
    char *dataptr, *digitptr, *colonptr;

    /*printf("args = %p\n", args);*/
    if (!PyArg_ParseTuple(args, "Oi:decode_raw_string", &dataobj, &index)) {
        PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "expected a string and an integer as parameters"));
        return NULL;
    }
    /*printf("dataobj = %p, index = %d\n", dataobj, index);*/

    if (index < 0) {
        PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "index must be a non-negative integer"));
        return NULL;
    }
    if (PyString_Check(dataobj)) {
        PyString_AsStringAndSize(dataobj, &dataptr, &dataptrlen);
        if (!dataptr) {
            PyErr_SetObject(PyExc_TypeError, Py_BuildValue("(is)", 0, "first parameter was not a python object?"));
            return NULL;
        }
        if (index > dataptrlen) {
            PyErr_SetObject(PyExc_ValueError, Py_BuildValue("(is)", 0, "index is greater than the string length"));
            return NULL;
        }
        dataptr = dataptr + index;
        dataptrlen = dataptrlen - index;
    } else if (PyBuffer_Check(dataobj)) {
        /* TODO accept buffer objects */
        PyErr_SetObject(PyExc_TypeError, Py_BuildValue("(is)", 0, "Buffer objects are not yet supported"));
        return NULL;
    } else {
        PyErr_SetObject(PyExc_TypeError, Py_BuildValue("(is)", 0, "parameter must be a python String object"));
        return NULL;
    }
    /*printf("dataptr = %p, dataptrlen = %d\n", dataptr, dataptrlen);*/
    /*printf("*dataptr = %s\n", dataptr);*/

    colonptr = strchr(dataptr, ':');
    if (!colonptr || (colonptr > dataptr + dataptrlen)) {
        PyErr_SetObject(MencodeError, Py_BuildValue("(is)", 0, "bad string length"));
        return NULL;
    }

    /* we don't use atoi() because it doesn't check for bad characters.  (does strtol?) */
    strlength = 0;
    digitptr = dataptr;
    while (digitptr < colonptr) {
        char digit = *digitptr;
        if ((digit < '0') || (digit > '9')) {
            PyErr_SetObject(MencodeError, Py_BuildValue("(is)", 0, "length contained non-digit character"));
            return NULL;
        }
        strlength = (10*strlength) + (digit - '0');
        digitptr++;
    }

    if ((strlength != 0) && (*dataptr == '0')) {
        PyErr_SetObject(MencodeError, Py_BuildValue("(is)", 0, "positive string length must not begin with `0'"));
        return NULL;
    }
    dataend = colonptr + 1 + strlength - dataptr;
    if (dataend > dataptrlen) {
        PyErr_SetObject(MencodeError, Py_BuildValue("(is)", 0, "unexpected end of string"));
        return NULL;
    }
    /* TODO return a buffer object instead of a string if strlength is over a determined threshold (40 chars or so?) */
    return Py_BuildValue("s#i", colonptr+1, strlength, dataend + index);
}

static PyMethodDef _c_mencode_helper_methods[] = {
    {"_c_decode_raw_string",  (PyCFunction)decode_raw_string,     METH_VARARGS},
    {"_c_encode_string",  (PyCFunction)encode_string,     METH_VARARGS},
    {"_c_encode_dict",  (PyCFunction)encode_dict,     METH_VARARGS},
    {"_c_encode_io",  (PyCFunction)encode_io,     METH_VARARGS},
    {NULL,      NULL}       /* sentinel */
};


DL_EXPORT(void) init_c_mencode_help(void)
{
    PyObject* m;
    PyObject* d;
    
    m = Py_InitModule("_c_mencode_help", _c_mencode_helper_methods);
    d = PyModule_GetDict(m);

    /* add our base exception class */
    MencodeError = PyErr_NewException("mencode._c_MencodeError", PyExc_StandardError, NULL);
    PyDict_SetItemString(d, "_c_MencodeError", MencodeError);

    encoder_dict = PyDict_New();
    PyDict_SetItemString(d, "_c_encoder_dict", encoder_dict);

    colon_tup_const = Py_BuildValue("(s)", ":");
    close_paren_tup_const = Py_BuildValue("(s)", ")");
    str_tup_const = Py_BuildValue("(s)", "(6:string");
    dict_tup_const = Py_BuildValue("(s)", "(4:dict");
    
    /* Check for errors */
    if (PyErr_Occurred()) {
        PyErr_Print();
        Py_FatalError("can't initialize module _c_mencode_help");
    }

}

