#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# This module abstracts the creation of XML required for talking to a
# content tracker.
#
__cvsid = '$Id: contentlib.py,v 1.58.2.1 2001/06/12 05:09:26 zooko Exp $'

# standard modules
from cStringIO import StringIO
import os
import shutil
from stat import ST_MTIME
import htmlentitydefs
import string
import bisect
import copy
import threading
import zlib
import sha
import math
try:
    import xml.dom.minidom
    have_minidom = 1
except ImportError:
    have_minidom = 0
import whrandom

# our modules
import BidStrategy
from MojoXML import *  # this also imports the saxlib+saxexts for us
import MojoTransaction
import MetaTrackerLib
from confutils import confman
from human_readable import human_readable, hr
import pricelib
import debug
import mojosixbit
import mojoutil
import fileutil



MAX_COMPRESSED_BINARY_DINODE_LEN = 6144


# the content tracking system's base error class
class ContentTrackerError(StandardError) : pass
# there was an error using the database
class ContentDBError(ContentTrackerError) : pass
# there was an error parsing the XML
class ContentXMLError(ContentTrackerError, XMLParseError) : pass
# the request was not for a supported content type
class ContentTypeError(ContentTrackerError) : pass



########################################################################
def find_content_trackers(MTM, min=5, max=10) :
    contenttrackers_id_map = MetaTrackerLib.get_content_trackers(MTM, min=min, max=max)

    # XXX call business logic to sort/filter the list in order of preference
    return contenttrackers_id_map.items()

def find_content_trackers_and_call_back(MTM, callback, min=5, max=10):
    MetaTrackerLib.get_content_trackers_and_call_back(MTM, callback=callback, min=min, max=max)

########################################################################
def content_tracker_submit(contenttrackertuple, MTM, submission, callback, timeout=None, preferenceno=0):
    """
    Make a submission to a content tracker

    @param contenttrackertuple a tuple containing the id and meta_info about the content tracker we wish to use.
    @param callback the result will be "returned" by means of a standard MTM callback

    @dontblock this function does not block
    """
    if not timeout:
        timeout = int(confman.dict['CONTENTTRACKER_CLIENT']['INDIVIDUAL_SUBMIT_TIMEOUT'])
    contenttracker_id, contenttracker_info = contenttrackertuple
    msgtype = 'content tracker submit'
    msgbody = {'XML data': submission}
    fixedprice = pricelib.content_tracker_submit_price(contenttracker_info, msgbody)
    if fixedprice is not None:
        # Remove this once we've grandfathered-out most of the fixed-price servers...  --Zooko 2001-06-01
        mojooffer = fixedprice
    else:
        mojooffer = BidStrategy.choose_bid("MO_CONTENT_TRACKER_SUBMIT", MTM, preferenceno=preferenceno)

    def inner_outcome_func(widget, outcome, failure_reason=None, callback=callback):
        if failure_reason:
            debug.stderr.write("content tracker submit failed from %s\n", args=(widget.get_counterparty_id(),), v=2)
        # TODO record latency here
        # TODO record success/fail stats here
        callback(widget=widget, outcome=outcome, failure_reason=failure_reason)

    MTM.initiate(contenttracker_id, msgtype, msgbody, mojooffer, outcome_func=inner_outcome_func, timeout=timeout)

########################################################################
def get_content_tracker_service_dict(metainfo):
    """
    Returns a reference to the 'content tracker' service announcement
    found in the metainfo dictionary.  raises KeyError if metainfo
    does not describe a content tracker.
    """
    for service in metainfo['services']:
        if (service is not None) and (service['type'] == 'content tracker'):
            return service
    raise KeyError, (metainfo, "no 'content tracker' service")

########################################################################
def multi_content_tracker_lookup(servertuplelist, MTM, typename, querydict, verify_success_func=None, minsuccesses=99999999, individualtimeout=None, individualcallback=None, exactmatch=None, publicity_criterion="public"):
    """
    Query content trackers in servertuplelist for information.

    @param querydict - is passed to make_contentsearch_xml with the
    appropriate parameters for the content tracker in question.

    @param exactmatch - controls whether the search looks for exact
    matches of the values in querydict or parses them into a more
    user friendly query string.

    minsuccesses - the minimum success responses before returning.  The
    default of 'None' means infinite (a response or timeout will be
    required from each content tracker before returning)

    verify_success_func - if one is given, this function should take a
    content tracker lookup response list of XML data as its parameter and
    return a boolean indicating whether or not to count this as a
    successful response.  It will be called on each success response
    received before counting it as a success.
    IDEAS: in the future, use this to verify the xml or do our own
    filtering on it before considering and counting it as a success

    individualtimeout - the number of seconds to wait for a response
    from any given content tracker.  (if None, it'll use the conf file
    default)

    individualcallback - a callback function that will be called once
    for each content tracker response as they come in.  It must have
    a "widget" and "resultlist" parameters.  resultlist will be
    a list of XML responses from one server (described in widget).

    publicity_criterion - Must be either "public", "private" or "both".  (see OurMessages.py)

    Returns: a dict keyed on 'list' and 'totalresponses' containing a
    list of xml content tracker entries and the total number of servers
    that we heard back from (or that timed out).
    """
    assert type(servertuplelist) == type([])
    assert minsuccesses > 0

    if not individualtimeout:
        individualtimeout = int(confman.dict['CONTENTTRACKER_CLIENT']['INDIVIDUAL_QUERY_TIMEOUT_v3'])

    queryxml_smart = make_contentsearch_xml(typename, querydict, exactmatch=exactmatch, use_smart=1)

    # Credit: thanks to Andy Sloane for "fixing" the old incarnation of this to
    # search all content trackers until our UI handles preferred content trackers better!

    searchresult = {'list' : [], 'result': 'no match', 'totalsuccesses': 0, 'totalresponses' : 0}

    # use these to send all queries at once and wait for the responses
    searchresult_condition = threading.Condition()
    def got_content_lookup_result_callback(widget, outcome, failure_reason=None, verify_success_func=verify_success_func, searchresult=searchresult, searchresult_condition=searchresult_condition, individualcallback=individualcallback) :
        searchresult_condition.acquire()
        try:
            searchresult['totalresponses'] = searchresult['totalresponses'] + 1
            if outcome and outcome.get('result') == 'success' :
                if verify_success_func and not verify_success_func(outcome['list']):
                    return  # verification failed, don't count as a success
                searchresult['totalsuccesses'] = searchresult['totalsuccesses'] + 1
                searchresult['list'] = list(searchresult['list'])
                searchresult['list'].extend(list(outcome['list']))
                if individualcallback:
                    individualcallback(widget=widget, resultlist=outcome['list'])
            else:
                if individualcallback:
                    individualcallback(widget=widget, resultlist=[])
        finally:
            searchresult_condition.notify()
            searchresult_condition.release()

    # we're about to block, make sure we're not the DoQ
    MTM._doQ.dissociate()

    max_queries_at_once = int(confman.dict["CONTENTTRACKER_CLIENT"]["MAX_QUERIES_AT_ONCE_v5"])
    num_queries_sent = 0
    # send the queries, up to max at a time
    for tracker in servertuplelist :
        queryxml = queryxml_smart
        nonblocking_content_tracker_lookup(tracker, MTM, queryxml, callback=got_content_lookup_result_callback, timeout=individualtimeout, publicity_criterion=publicity_criterion, preferenceno=num_queries_sent)
        num_queries_sent = num_queries_sent + 1

        # don't allow more than num_queries_at_once outstanding queries
        searchresult_condition.acquire()
        try:
            while searchresult['totalsuccesses'] < minsuccesses and (num_queries_sent - searchresult['totalresponses']) >= max_queries_at_once :
                searchresult_condition.wait()
        finally:
            searchresult_condition.release()
        
        if searchresult['totalsuccesses'] >= minsuccesses:
            break  # stop sending queries, we have enough responses
    
    # block until all content trackers have responded, timed out or we have enough
    searchresult_condition.acquire()
    try:
        while searchresult['totalresponses'] < num_queries_sent and searchresult['totalsuccesses'] < minsuccesses :
            searchresult_condition.wait()
    finally:
        searchresult_condition.release()
    
    return {'totalresponses': searchresult['totalresponses'], 'list': copy.copy(searchresult['list'])}


########################################################################
def nonblocking_content_tracker_lookup(contenttrackertuple, MTM, query, callback, timeout=None, publicity_criterion="public", preferenceno=0):
    """
    Query a content tracker for information.  This function will not block and the result will
    be "returned" by means of a standard MTM callback.

    @param contenttrackertuple - is a tuple containing the id and meta_info about the content tracker we wish to use.

    @param publicity_criterion - (see OurMessages.py) Should be "public", "private", or "both".
    
    Returns: None, use the callback silly
    """
    if not timeout:
        timeout = int(confman.dict['CONTENTTRACKER_CLIENT']['INDIVIDUAL_QUERY_TIMEOUT_v3'])

    contenttracker_id, contenttracker_info = contenttrackertuple
    msgtype = 'content tracker lookup'
    msgbody = {'XML data': query, 'publicity criterion' : publicity_criterion}
    fixedprice = pricelib.content_tracker_lookup_price(contenttracker_info, msgbody)
    if fixedprice is not None:
        # Remove this once we've grandfathered-out most of the fixed-price servers...  --Zooko 2001-06-01
        mojooffer = fixedprice
    else:
        mojooffer = BidStrategy.choose_bid("MO_CONTENT_TRACKER_LOOKUP", MTM, preferenceno=preferenceno)
        # We want the earlier lookups to be higher-priority than later ones, as users are presumably waiting
        # on the first couple, but happily browsing or downloading on subsequent ones...  --Zooko 2001-06-01
        mojooffer = long(mojooffer * (0.94 ** preferenceno))

    def inner_outcome_func(widget, outcome, failure_reason=None, callback=callback, sessionkeeper=MTM._keeper) :
        if failure_reason :
            debug.stderr.write("content tracker lookup failed from %s, reason: %s\n" % (human_readable(widget.get_counterparty_id()), failure_reason), v=3, vs='contentlib')
        # record success/fail stats here
        server_id = widget.get_counterparty_id()
        cp_obj = sessionkeeper.get_counterparty_object(server_id)
        if outcome is None:
            results_total = 0
        else:
            results_total = len(outcome.get('list', []))
        cp_obj.update_custom_stat_weighted_sample('content tracker results', results_total)
        if failure_reason or not outcome or (outcome.get('result', 'failure') != 'success'):
            success_value = 0
        else:
            success_value = 1
        cp_obj.update_custom_stat_weighted_sample('content tracker success', success_value)
        callback(widget=widget, outcome=outcome, failure_reason=failure_reason)
    MTM.initiate(contenttracker_id, msgtype, msgbody, mojooffer, outcome_func=inner_outcome_func, timeout=timeout)


########################################################################
class ContentTrackerHandicapper:
    """
    This handicapper is designed to handicap content tracker selection
    based on past performance of the content tracker in question

    for message_type[:len('content tracker')] == 'content tracker'

    @returns a complex function of content tracking history, I think its magnitude is between 0 and 1700.  --Zooko 2000-11-20
    """
    def __init__(self, sessionkeeper):
        if confman.is_true_bool(['CONTENTTRACKER_CLIENT', 'HANDICAPPER', 'DISABLED']):
            raise RuntimeError, "Content Tracker Client handicapping disabled in config file, this shouldn't have been called"
        self.sessionkeeper = sessionkeeper

    def __call__(self, counterparty_id, metainfo, message_type, message_body):
        # only handicap content tracker related messages
        if message_type[:len('content tracker')] != 'content tracker':
            return 0

        cp_obj = self.sessionkeeper.get_counterparty_object(counterparty_id)
        avg_results = cp_obj.get_custom_stat('content tracker results')
        if avg_results is not None:
            query_failures_ratio = 1.0 - cp_obj.get_custom_stat('content tracker success', 0)
            handicap = (query_failures_ratio * float(confman.dict['CONTENTTRACKER_CLIENT']['HANDICAPPER']['FAILURE_RATIO_WEIGHT']))
            if avg_results > 0:
                inverse_log_numresults_ratio = math.log(float(confman.dict['CONTENTTRACKER_CLIENT']['HANDICAPPER']['NUMRESULTS_LOG_BASE'])) / math.log(avg_results + 2)
                # `inverse_log_numresults_ratio' will be between 2.32 (if avg_results is ~0) to 0.29 (if avg_results is 256).
                handicap = handicap + (inverse_log_numresults_ratio * float(confman.dict['CONTENTTRACKER_CLIENT']['HANDICAPPER']['NUMRESULTS_WEIGHT_v2']))
            return handicap
        else:
            return float(confman.dict['CONTENTTRACKER_CLIENT']['HANDICAPPER']['DEFAULT'])


########################################################################
def get_mojoid_from_asciidinode(asciidinode):
    """Return the ascii mojoid for the given dinode."""
    # Note: XML dinodes aren't strictly canonical so an archive can
    # technically have more than one possible mojoid, but so what.
    assert type(asciidinode) == type('')
    if mojosixbit._mojosixbit_re.match(asciidinode):
        try:
            zbindinode = mojosixbit.a2b(asciidinode)
        except:
            # this should never happen
            debug.mojolog.write("dinode %s matched regexp but failed ascii decoding!\n" % `asciidinode`, vs='contentlib')
            zbindinode = asciidinode  # not ascii encoded
    else:
        zbindinode = asciidinode  # not ascii encoded
    if len(zbindinode) > MAX_COMPRESSED_BINARY_DINODE_LEN:
        raise ContentTrackerError, "compressed binary __DINODE__ too large"
    try:
        rawdinode = zlib.decompress(zbindinode)
    except zlib.error:
        rawdinode = zbindinode
    mojoid = mojosixbit.b2a(sha.new(rawdinode).digest())
    return mojoid

get_mojoid_from_dinode = get_mojoid_from_asciidinode

########################################################################
char_to_xmlentity = {}
for entity, char in htmlentitydefs.entitydefs.items() :
    char_to_xmlentity[char] = entity

def make_xml_entity_safe(s) :
    """convert chars such as &, <, > into their safely encoded form for use within xml"""
    s = str(s)
    if have_minidom:
        # if we've got late enough xml library, it'll do it for us
        d = xml.dom.minidom.Document()
        tn = d.createTextNode(s)
        result = tn.toxml()
        d.unlink()
        return result
    def entity_mapper(c) :
        if char_to_xmlentity.has_key(c) :
            return '&' + char_to_xmlentity[c] + ';'
        else :
            return c
    return string.join(map(entity_mapper, s), '')

########################################################################
def make_contentdesc_xml(typename, datadict, query=0) :
    """Create XML describing a record in the database.
    typename is the data template type of this record.
    datadict maps fields to their values.
    """
    myxml = StringIO()
    if query :
        myxml.write('<ContentDesc query="yes">')
    else :
        myxml.write('<ContentDesc>')
    myxml.write('<TypeName>')
    myxml.write(make_xml_entity_safe(typename))
    myxml.write('</TypeName>')
    for field, value in datadict.items() :
        if value is None :
            value = ""
            continue
        myxml.write('<Data><Field>')
        myxml.write(make_xml_entity_safe(field))
        myxml.write('</Field><Value>')
        myxml.write(make_xml_entity_safe(value))
        myxml.write('</Value></Data>')
    myxml.write('</ContentDesc>')
    return myxml.getvalue()


########################################################################
def make_contentsearch_xml(typename, matchdict, exactmatch=0, use_smart=0) :
    """Create XML describing a content search.
       The values in matchdict will be parsed and turned into more useful query strings the content tracker.

       (this is called by multi_content_tracker_lookup, it should not be needed outside of this module.)
    """
    myxml = StringIO()
    myxml.write('<ContentDesc query="yes"')
    if use_smart:
        myxml.write(' smart="yes"')
    myxml.write('>')
    myxml.write('<TypeName>')
    myxml.write(make_xml_entity_safe(typename))
    myxml.write('</TypeName>')
    for field, value in matchdict.items() :
        if value is None:  # do not check for the empty string, that indicates a global special cases
            continue

        # Convert the value strings into useful search strings for the content tracker
        #   - the more recent use_smart style will match if any of the words in value occur in that field.
        #   - the old style matches unintuitively: if all of the words in value occur in that field in the given order.
        if exactmatch:
            if use_smart:
                if field != '__MOJOID__':   # __MOJOID__ lookups are always exact non-regexp 
                    value = '^' + value + '$'
        else:
            if not use_smart:
                value = '%' + string.replace(value, ' ', '%') + '%'

        myxml.write('<Data><Field>')
        myxml.write(make_xml_entity_safe(field))
        myxml.write('</Field><Value>')
        myxml.write(make_xml_entity_safe(value))
        myxml.write('</Value></Data>')
    myxml.write('</ContentDesc>')
    return myxml.getvalue()


########################################################################
def make_contenttypedef_xml(typename, fieldtypedict) :
    """
    Create XML describing a defined content type.  The types listed
    in fieldtypedict must be 'number' or 'string'

    @precondition Must specift at least one field.: len(fieldtypedict) > 0: "fieldtypedict: %s" % hr(fieldtypedict)
    """
    assert len(fieldtypedict) > 0, "Must specify at least one field." + " -- " + "fieldtypedict: %s" % hr(fieldtypedict)

    myxml = StringIO()
    myxml.write('<ContentTypeDef>\n')
    myxml.write('<TypeName>')
    myxml.write(make_xml_entity_safe(typename))
    myxml.write('</TypeName>\n')
    for field, fieldtype in fieldtypedict.items() :
        if fieldtype != 'number' and fieldtype != 'string' :
            raise ValueError, "Field types must be 'number' or 'string'"
        myxml.write('<Field type="%s">' % fieldtype)
        myxml.write(make_xml_entity_safe(field))
        myxml.write('</Field>\n')
    myxml.write('</ContentTypeDef>')
    return myxml.getvalue()


class ContentTypeDef:
    """
    A container object for the DocHandler to create a list of.  It
    will always have the following public attributes:

    * name is the name of the type
    * fielddict maps fieldnames to their data type ['string' or 'number'].
    * sortedfieldnamelist is a sorted list of field names to use when
      presenting them to a user.
    * multifile is a boolean indicating if multifile publications should
      be allowed under this type.
    * automaticfieldmap is a map of field names that are theoretically filled in automaticlly when publishing.
    * moderated indicates if new submissions to this content type require approval before being returned in a search
    * version is the floating point version number of this type
    * url is the url used to get the most recent version of this type description
    * contenttrackeridlist is the list of mojosixbit encoded content tracker ids officially supporting this type
    * searchedfieldmap is a map of field names that should be allowed to be searched on
    * submittedfieldmap is a map of field names that should be allowed to be filled in and submitted during publication
    """
    _public_methods_ = []  # enable COM access
    _public_attrs_ = []

    def __init__(self, name, fielddict, sortedfieldnamelist, multifile, automaticfieldmap, moderated, version, url, contenttrackeridlist, searchedfieldmap, submittedfieldmap):
        self.name = name
        self.fielddict = fielddict
        self.sortedfieldnamelist = sortedfieldnamelist
        self.multifile = multifile
        self.automaticfieldmap = automaticfieldmap
        self.moderated = moderated
        self.version = version
        self.url = url
        self.contenttrackeridlist = contenttrackeridlist
        self.searchedfieldmap = searchedfieldmap
        self.submittedfieldmap = submittedfieldmap

        # for access via COM (via mojocom.smart_wrap)
        self._public_attrs_.extend( self.__dict__.keys() )

    def legacy_tuple(self):
        return (self.name, self.fielddict, self.sortedfieldnamelist, self.multifile, self.automaticfieldmap)

    _public_methods_ = ['get_name']
    def get_name(self):
        return self.name

########################################################################
class ContentTypeDefDocHandler(MojoNationXMLDocHandler) :
    """After this document handler has been used to parse one or more
    documents its member variable called 'types' will be a list of
    ContentTypeDef instances describing each type defined in the XML.

    """
    def __init__(self) :
        # this is an array of tuples containing:
        # (typename, fieldsdict, sortedfieldslist, multifile flag,
        # automatic fields list) for each content type definition
        # parsed.
        self.types = []

        # this is used to represent the current XML parsing state
        self.__state = [0]

        # used during reading; contains a sorted list of tuples of (pos, fieldname)
        # that can be used to determine the order to display the fields on a form
        self.__sortedfieldnametuples = []

    # This method is called by the XML parser on all non-xml-tag
    # character data.
    def characters(self, ch, start, length) :
        state = self.__state[0]
        if state == self.inTypeName :
            self.name = self.name + ch[start:start+length]
        elif state == self.inField :
            self.fieldname = self.fieldname + ch[start:start+length]
        elif state == self.inContentTrackerId:
            self.id = self.id + ch[start:start+length]

    inContentType = 801
    def start_ContentTypeDef(self, attrs) :
        self.multifile = None
        if attrs.has_key('multifile'):
            if attrs['multifile'] == 'yes':
                self.multifile = 1
            elif attrs['multifile'] != 'no':
                raise ContentXMLError, "'multifile' attribute must be (no|yes)"
        if attrs.has_key('url'):
            self.url = attrs['url']
        else:
            self.url = None
        self.moderated = 0
        if attrs.has_key('moderated'):
            if attrs['moderated'] not in ('yes', 'no'):
                raise ContentXMLError, "'moderated' attribute must be (no|yes)"
            if attrs['moderated'] == 'yes':
                self.moderated = 1
        if not (attrs.has_key('version') and attrs.has_key('url')):
            raise ContentXMLError, "'version' and 'url' attributes required"
        self.version = attrs['version']
        self.url = attrs['url']
        # TODO url attribute to ContentTypeDef tag indicating the URL of where to get the latest version of this content type
        # must have a name attr
        self.name = None
        self.fields = {}
        self.automaticfields = {}
        self.searchedfields = {}
        self.submittedfields = {}
        self.contenttrackeridlist = []
        self.__state.insert(0, self.inContentType)
    def end_ContentTypeDef(self) :
        if not self.name :
            raise ContentXMLError, "No name specified"
        if not self.fields :
            raise ContentXMLError, "No fields specified"
        self.types.append( ContentTypeDef(self.name, self.fields, map(lambda tup: tup[1], self.__sortedfieldnametuples), self.multifile, self.automaticfields, self.moderated, self.version, self.url, self.contenttrackeridlist, self.searchedfields, self.submittedfields) )
        del self.name
        del self.fields
        del self.multifile
        del self.automaticfields
        del self.moderated
        del self.version
        del self.url
        del self.contenttrackeridlist
        self.__state.pop(0)

    # these must occur within ContentTypeDef elements, the name of this content type
    inTypeName = 803
    def start_TypeName(self, attrs) :
        if self.__state[0] != self.inContentType :
            raise ContentXMLError, "TypeName must be within ContentTypeDef"
        self.name = ''
        self.__state.insert(0, self.inTypeName)
    def end_TypeName(self) :
        if not self.name :
            raise ContentXMLError, "empty name"
        self.__state.pop(0)

    # these must occur within ContentTypeDef elements, the name of this content type
    inContentTrackerId = 804
    def start_ContentTrackerId(self, attrs) :
        if self.__state[0] != self.inContentType :
            raise ContentXMLError, "ContentTrackerId must be within ContentTypeDef"
        self.id = ''
        self.__state.insert(0, self.inContentTrackerId)
    def end_ContentTrackerId(self) :
        if not self.id :
            raise ContentXMLError, "empty ContentTrackerId"
        if not idlib.is_ascii_id(self.id):
            raise ContentXMLError, "invalid mojosixbit encoded ContentTrackerId: %s" % `self.id`
        self.contenttrackeridlist.append(self.id)
        del self.id
        self.__state.pop(0)

    # these must occur within ContentTypeDef elements
    inField = 802
    def start_Field(self, attrs) :
        if self.__state[0] != self.inContentType :
            raise ContentXMLError, "Field must be within ContentTypeDef"
        # can have a type attr specifying number/string
        self.fieldname = ''
        if attrs.has_key('type') :
            self.fieldtype = attrs['type']
            if self.fieldtype != 'string' and \
               self.fieldtype != 'number' :
                raise ContentXMLError, "Field type must be 'string' or 'number'"
        else :
            self.fieldtype = 'string'
        if attrs.has_key('position') :
            self.fieldposition = int(attrs['position'])
        else :
            self.fieldposition = 0x7fffffff   # last
        self.fieldautomatic = (attrs.get('automatic', "no") == "yes")
        self.fieldsearched = (attrs.get('searched', "yes") == "yes")
        self.fieldsubmitted = (attrs.get('submitted', "yes") == "yes")
        self.__state.insert(0, self.inField)
    def end_Field(self) :
        if not self.fieldname :
            raise ContentXMLError, "empty field"

        self.fields[self.fieldname] = self.fieldtype
        if self.fieldautomatic:
            # mark which fields are (in theory) automatically filled in
            self.automaticfields[self.fieldname] = self.fieldtype
        # maintain searchedfields and submittedfields dictionaries
        if self.fieldsearched:
            self.searchedfields[self.fieldname] = self.fieldtype
        if self.fieldsubmitted:
            self.submittedfields[self.fieldname] = self.fieldtype

        bisect.insort(self.__sortedfieldnametuples, (self.fieldposition, self.fieldname))
        self.__state.pop(0)


########################################################################
class ContentDescDocHandler(MojoNationXMLDocHandler) :
    """Handles content description documents which represent a single record in the database"""
    def __init__(self) :
        self.query = 0        # boolean: indicates if this document describes a query
        self.smart = 0        # boolean: indicated if this query should be done smartly by the tracker
        self.field = None
        self.value = None
        self.__state = [0]
        self.__chars = ''
        self.typename = None  # this will hold the type name
        self.fieldvalues = {} # this will hold the field->value map

    # This method is called by the XML parser on all non-xml-tag
    # character data.
    def characters(self, ch, start, length) :
        self.__chars = self.__chars + ch[start:start+length]

    inContentDesc = 901
    def start_ContentDesc(self, attrs) :
        if attrs.has_key('query') :
            if attrs['query'] == 'yes' :
                self.query = 1
        else :
            self.query = 0
        if attrs.has_key('smart') and attrs.get('smart') == 'yes':
            self.smart = 1
        self.__state.insert(0, self.inContentDesc)
    def end_ContentDesc(self) :
        self.__state.pop(0)

    inTypeName = 902
    def start_TypeName(self, attrs) :
        if self.__state[0] != self.inContentDesc :
            raise ContentXMLError, "TypeName must be within ContentDesc element"
        if self.typename :
            raise ContentXMLError, "Only one TypeName is allowed in a ContentDesc element"
        self.__chars = ''
        self.__state.insert(0, self.inTypeName)
    def end_TypeName(self) :
        self.typename = self.__chars
        self.__chars = ''
        self.__state.pop(0)

    inData = 903
    def start_Data(self, attrs) :
        if self.__state[0] != self.inContentDesc :
            raise ContentXMLError, "Data must be within ContentDesc element"
        self.__state.insert(0, self.inData)
    def end_Data(self) :
        if not self.field or self.value is None :  # field can't be empty, value can be empty but must always be specified
            raise ContentXMLError, "Field or Value missing within Data element"
        if self.fieldvalues.has_key(self.field) :
            raise ContentXMLError, "Duplicate field definition"
        self.fieldvalues[self.field] = self.value
        self.field = None
        self.data = None
        self.__state.pop(0)

    inField = 904
    def start_Field(self, attrs) :
        if self.__state[0] != self.inData :
            raise ContentXMLError, "Field must occur within Data element"
        self.__chars = ''
        self.__state.insert(0, self.inField)
    def end_Field(self) :
        self.field = self.__chars
        self.__chars = ''
        self.__state.pop(0)

    inValue = 905
    def start_Value(self, attrs) :
        if self.__state[0] != self.inData :
            raise ContentXMLError, "Value must occur within Data element"
        self.__chars = ''
        self.__state.insert(0, self.inValue)
    def end_Value(self) :
        self.value = self.__chars
        self.__chars = ''
        self.__state.pop(0)


def content_xml_to_dict(xmlstring) :
    """
    Return a tuple containing (content type string, fieldvaluedict)
    The type string will be None if there was an error parsing the XML.
    """
    xmlfileobj = StringIO(xmlstring)
    desc_dochandler = ContentDescDocHandler()
    parser = create_mojo_xml_parser(desc_dochandler)
    try:
        try:
            parser.parseFile(xmlfileobj)
        finally:
            parser.close()
    except (XMLParseError, ContentTrackerError), e:
        debug.mojolog.write("Got error %s parsing content xml: " + `xmlstring` + "\n", args=(e,), v=4, vs='contentlib')
        return (None, {})
    
    # A catch-all HTML filter for content description fields:
    for name, value in desc_dochandler.fieldvalues.items():
        # Remove the unsafe entry:
        del desc_dochandler.fieldvalues[name]
        # Filter:
        name = string.replace(name, '&', '&amp;')
        name = string.replace(name, '<', '&lt;')
        name = string.replace(name, '>', '&gt;')
        value = string.replace(value, '&', '&amp;')
        value = string.replace(value, '<', '&lt;')
        value = string.replace(value, '>', '&gt;')
        # Re-insert:
        desc_dochandler.fieldvalues[name] = value
        
    return (desc_dochandler.typename, desc_dochandler.fieldvalues)


def test_content_xml_to_dict() :
    spamxml = make_contentdesc_xml('spam', {'opinion': 'i love it'})
    xtype, xdict = content_xml_to_dict(spamxml)
    assert xtype == 'spam'
    assert xdict.get('opinion') == 'i love it'


TYPEDEF_FILE_EXTENSION = '.mct'  # Mojo content type


class ContentTypeDefRepository :
    """
    Maintain a subdirectory containing a bunch of content type definition
    templates in individual files. This object is not yet thread safe.
    """
    _public_methods_ = []  # enable COM access
    _public_attrs_ = []

    def __init__(self, dirname) :
        debug.mojolog.write('Content Type directory: "%s"\n' % dirname, vs='contentlib')
        self.dirname = dirname

        # Make sure the directory exists:
        if not os.path.isdir(dirname):
            fileutil.make_dirs(dirname)
            # Temporary hack to upgrade:
            # Move content types from EVILDIR to dirname.
            olddir = confman.getpath(['PATH', 'CONTENT_TYPEDEF_DIR'])
            if olddir and os.path.isdir(olddir):
                del confman['PATH']['CONTENT_TYPEDEF_DIR']
                debug.mojolog.write('Copying Content Type Definitions from "%s" to "%s".\n' % (olddir, dirname), vs='contentlib', v=4)
                for file in os.listdir(olddir):
                    if file[-4:] not in ['.xml', '.mct']:
                        continue
                    shutil.copyfile(
                        os.path.join(olddir, file),
                        os.path.join(dirname, file))

        self.filestats = {}       # results of os.stat() call on each typedef file
        self.file_to_typenames_map = {}  # maps filenames to typedefs contained within
        self.name_to_typedef_map = {}  # maps typenames to ContentTypeDef objects
        self.typedef_map = {}  # contains type name -> typedef fields dict
        self.typedef_ordered_map = {}  # contains type name -> ordered list of field names
        self.typedef_multifile = {}  # contains type name -> boolean multifile allowed indicator

        # just like self.typedef_map but only contains automatically or manually filled in fields
        self.typedef_automatic_map = {}
        self.typedef_manual_map = {}

        # load typedefmap with the list of types currently found
        self.__load_directory()
        self.__load_all_typedefs()

    def __load_directory(self) :
        """
        Load the applicable filenames from the directory into self.filestats
        """
        self.dirstats = os.stat(self.dirname)
        files = os.listdir(self.dirname)
        def pick_typedef_files(f, dirname=self.dirname) :
            return f[-len(TYPEDEF_FILE_EXTENSION):] == TYPEDEF_FILE_EXTENSION and os.path.isfile(os.path.join(dirname, f))
        files = filter(pick_typedef_files, files)
        self.filestats.clear()
        for f in files :
            self.filestats[f] = os.stat(os.path.join(self.dirname, f))

    def __load_typedef(self, filename) :
        """Read content type definition(s) from the given file in self.dirname"""
        debug.mojolog.write('Loading Type Definition file: "%s"\n' % filename, vs='contentlib', v=0)
        if self.file_to_typenames_map.has_key(filename) :
            del self.file_to_typenames_map[filename]

        fullname = os.path.join(self.dirname, filename)
        fileobj = open(fullname, 'r')

        try:
            typedh = ContentTypeDefDocHandler()
            parser = create_mojo_xml_parser(typedh)
            try:
                parser.parseFile(fileobj)
            finally:
                parser.close()
        except XMLParseError, le:
            debug.stderr.write("XMLParseError in content type def file %s: %s\n" % (fullname, `le.args`))
            return
        if len(typedh.types) < 1 :
            debug.stderr.write("WARNING: "+fullname+" did not contain any valid content typedefs\n")
            return

        self.file_to_typenames_map[filename] = map(lambda x: x.name, typedh.types)
        for defobj in typedh.types:
            name, fields, sortedfieldnamelist, multifile, automaticfields = defobj.legacy_tuple()
            if self.name_to_typedef_map.has_key(name):
                debug.mojolog.write('NOTE: replacing typedef %s with one from %s\n', args=(name, filename), vs='contentlib')
            self.name_to_typedef_map[name] = defobj
            self.typedef_map[name] = fields
            self.typedef_ordered_map[name] = sortedfieldnamelist
            self.typedef_multifile[name] = multifile
            # a map of which fields should only be filled in automatically for this type
            self.typedef_automatic_map[name] = automaticfields
            # build the opposite of automatic_map
            self.typedef_manual_map[name] = {}
            for key, value in fields.items():
                if not automaticfields.has_key(key):
                    self.typedef_manual_map[name][key] = value
        
        #debug.mojolog.write("typedef_automatic_map %s\n" % self.typedef_automatic_map, vs="contentlib")
        #debug.mojolog.write("typedef_manual_map %s\n" % self.typedef_manual_map, vs="contentlib")

    def __load_all_typedefs(self) :
        for f in self.filestats.keys() :
            self.__load_typedef(f)

    _public_methods_.append('add_typedef_by_url')
    def add_typedef_by_url(self, typedefurl) :
        """add_typedef_by_url(typedefxml)
        Adds a content type definition to our repository, storing it in the repository directory on disk.
        @returns a tuple of a boolean indicating whether it was accepted and added or not
         and a string suitable for showing to a user.
        """
        # This function needs to make use of an HTTP proxy if the user has
        # defined one!  It also desperately needs to check its input, because
        # its input comes from a web form and could be absolutely anything.

        # Grab the content from the URL.  This is normally done synchronously,
        # while the user waits, so the user must know that this method is
        # looking up a URL (so he or she will allow time before getting impatient).
        #debug.mojolog.write("type of typedefxml is %s\n" % type(typedefxml), vs="contentlib")
        assert (type(typedefxml) == type(''))
        import urllib
        try:
            url = urllib.urlopen(typedefurl)
            infomsg = url.info()
            clength = infomsg.getheader('Content-Length')
            # The urllib module, for reasons I don't understand, does not return
            # an IOError for error 404 as it does for other errors.  So make
            # sure we have a Content-Length header.
            if not clength:
                return(0, 'Could not retrieve URL %s\n' % typedefurl)
            try:
                contentlength = int(clength)
            except ValueError:
                debug.mojolog.write('Cannot convert received Content-Length %s to int\n' % `clength`, vs="contentlib")
                return(0, 'Unspecified error retrieving URL %s\n' % typedefurl)
            xmltype = url.read(contentlength)
            debug.mojolog.write('Downloaded URL %s and found contents: %s\n' % (typedefurl, xmltype))
        except IOError:
            return (0, 'Could not retrieve URL %s\n' % typedefurl)
        rtn = self.add_typedef(xmltype)
        if rtn == 0:
            return (0, 'Could not add XML file %s to content type repository; it had no fields or I could not parse the XML.\n' % xmltype)
        else:
            return (1, 'Successfully added the content type at %s\n', typedefurl)


    _public_methods_.append('add_typedef')
    def add_typedef(self, typedefxml) :
        """add_typedef(typedefxml)
        Adds a content type definition to our repository, storing it in the repository directory on disk.
        @returns a boolean indicating whether it was accepted and added or not.
        """
        # try parsing this xml
        debug.mojolog.write("add_typedef(): entered add_typedef\n")
        fileobj = StringIO(typedefxml)
        debug.mojolog.write("add_typedef(): got fileobj\n")
        try:
            typedh = ContentTypeDefDocHandler()
            parser = create_mojo_xml_parser(typedh)
            try:
                parser.parseFile(fileobj)
            finally:
                parser.close()
        except XMLParseError, le:
            debug.mojolog.write("XMLParseError in added content type, skipping: %s\n" % `le.args`)
            return 0
        debug.mojolog.write("add_typedef(): finished XML parsing\n")
        if len(typedh.types) < 1 :
            debug.mojolog.write("WARNING: added content type xml did not contain any valid content typedefs\n")
            return 0

        # save it to a unique filename in our repository (uniqueness linked to the url of the file if possible)
        debug.mojolog.write("add_typedef(): about to save\n")
        if typedh.types[0].url:
            filename = idlib.to_ascii(idlib.make_id(typedh.types[0].url)) + TYPEDEF_FILE_EXTENSION 
        else:
            filename = idlib.to_ascii(idlib.make_id(typedefxml)) + TYPEDEF_FILE_EXTENSION 
        fullname = os.path.join(self.dirname, filename)
        debug.mojolog.write("add_typedef(): filename is %s\n" % fullname)
        f = open(fullname, "wb")
        f.write(typedefxml)
        f.close()
        debug.mojolog.write("add_typedef(): closed file\n")

        # This needs to update self.filestats -- it currently doesn't, which is a bug.  -jill

        self.file_to_typenames_map[filename] = map(lambda x: x.name, typedh.types)
        for defobj in typedh.types:
            name, fields, sortedfieldnamelist, multifile, automaticfields = defobj.legacy_tuple()
            if self.name_to_typedef_map.has_key(name):
                debug.mojolog.write('NOTE: replacing typedef %s with one from %s', args=(name, filename), vs='contentlib')
            self.name_to_typedef_map[name] = defobj
            self.typedef_map[name] = fields
            self.typedef_ordered_map[name] = sortedfieldnamelist
            self.typedef_multifile[name] = multifile
            # a map of which fields should only be filled in automatically for this type
            self.typedef_automatic_map[name] = automaticfields
            # build the opposite of automatic_map
            self.typedef_manual_map[name] = {}
            for key, value in fields.items():
                if not automaticfields.has_key(key):
                    self.typedef_manual_map[name][key] = value
        
        debug.mojolog.write("add_typedef(): about to return: self is: %s\n" % `self`)
        return 1

    _public_methods_.append('remove_typedef')
    def remove_typedef(self, typename):
        """
        Remove this typedef and its file on disk.

        Return a boolean saying whether this was successful or not.
        """
        # Make sure typename actually exists.  
        if typename not in self.name_to_typedef_map.keys():
            debuglog('did not find typename: name_to_typedef_map.keys() is %s' % `name_to_typedef_map.keys()`)
            return 0

        # Check basic sanity.
        assert(typename in self.filestats.values())
        assert(typename in self.file_to_typenames_map.values())
        assert(self.name_to_typedef_map.has_key(typename))
        assert(self.typedef_ordered_map.has_key(typename))
        assert(self.typedef_multifile.has_key(typename))
        assert(self.typedef_automatic_map.has_key(typename))
        assert(self.typedef_manual_map.has_key(typename))

        # Find the name of the file and remove it.
        found = 0
        debug.mojolog.write('self is %s.\n' % `self`)
        debug.mojolog.write('self.file_to_typenames_map is %s.\n' % `self.file_to_typenames_map`)
        for filename in self.file_to_typenames_map.keys():
            typelist = [ typename ]
            if self.file_to_typenames_map[filename] == typelist:
                found = 1
                break
        assert(found == 1)
        fullname = os.path.join(self.dirname, filename)
        debug.mojolog.write('remove_typedef(): attempting to remove file %s.\n' % `fullname`)
        try:
            rtn = os.remove(fullname)
        except OSError, e:
            debug.mojolog.write('NOTE: unsuccessful attempt to remove file %s.\n' % fullname)
            debug.mojolog.write('NOTE: e is %s.\n' % `e`)
            return 0
        del self.filestats[filename]
        del self.file_to_typenames_map[filename]
        del self.name_to_typedef_map[typename]
        del self.typedef_ordered_map[typename]
        del self.typedef_multifile[typename]
        del self.typedef_automatic_map[typename]
        del self.typedef_manual_map[typename]
        self.reload_typedefs()
        return 1

    _public_methods_.append('reload_typedefs')
    def reload_typedefs(self) :
        """Check all typedef files, reload any that have changed"""
        for f, stat in self.filestats.items() :
            fullname = os.path.join(self.dirname, f)
            if os.stat(fullname)[ST_MTIME] > stat[ST_MTIME] :
                self.__load_typedef(f)

    _public_methods_.append('reload_directory')
    def reload_directory(self) :
        """If the directory or any of the typedefs have been modified, read them again"""
        if os.stat(self.dirname)[ST_MTIME] > self.dirstats[ST_MTIME] :
            self.__load_all_typedefs()

    _public_methods_.append('has_type')
    def has_type(self, name):
        return self.typedef_map.has_key(name)

    _public_methods_.append('known_type_list')
    def known_type_list(self) :
        """known_type_list() - Return a sorted list of currently known types"""
        self.reload_directory()
        self.reload_typedefs()
        resultlist = self.typedef_map.keys()

        def _content_type_name_cmp(a, b):
            A = string.lower(a)
            B = string.lower(b)

            if A[:len('generic')] == 'generic':
                if B[:len('generic')] == 'generic':
                    return cmp(B, A)
                else :
                    return -1  # put generic entries first
            if B[:len('generic')] == 'generic':
                return 1  # put generic entries first

            if A[:len('adult')] == 'adult':
                if B[:len('adult')] == 'adult':
                    return cmp(A, B)
                else :
                    return 1  # put Adult entries last
            if B[:len('adult')] == 'adult':
                return -1  # put Adult entries last
            return cmp(A, B)

        resultlist.sort(_content_type_name_cmp)
        return resultlist
    
    _public_methods_.append('known_typeobj_list')
    def known_typeobj_list(self):
        """known_typeobj_list returns a sorted list of ContentTypeDef objects"""
        namelist = self.known_type_list()
        return map(self.name_to_typedef_map.get, namelist)
    
    _public_methods_.append('is_type_multifile')
    def is_type_multifile(self, typename):
        """is_type_multifile(typename) - Boolean indicator if this type allows multi-file publications (or raise KeyError)"""
        return self.typedef_multifile[typename]

    _public_methods_.append('get_type_desc')
    def get_type_desc(self, typename) :
        """get_type_desc(typename) - Return the typedef dict for typename or raise KeyError"""
        try:
            return self.typedef_map[typename]
        except KeyError, le:
            raise KeyError, "le: %s, self.typedef_map.keys(): %s" % (hr(le), hr(self.typedef_map.keys()),)
    
    _public_methods_.append('get_type_obj')
    def get_type_obj(self, typename):
        """get_type_obj(typename) - Return the typedef object for typename or raise KeyError"""
        return self.name_to_typedef_map[typename]
    
    _public_methods_.append('get_ordered_type_fieldname_list')
    def get_ordered_type_fieldname_list(self, typename):
        """get_ordered_type_fieldname_list(typename) - Return fieldnames for typename in the desired order or raise KeyError"""
        return self.typedef_ordered_map[typename]
    
    _public_methods_.append('is_automatic_type_fieldname')
    def is_automatic_type_fieldname(self, typename, fieldname):
        return self.typedef_automatic_map[typename].has_key(fieldname)
    
    _public_methods_.append('is_moderated_type')
    def is_moderated_type(self, typename):
        return self.name_to_typedef_map[typename].moderated

    _public_methods_.append('is_branded_type')
    def is_branded_type(self, typename):
        return len(self.name_to_typedef_map[typename].contenttrackeridlist) > 0

    _public_methods_.append('is_public_type')
    def is_public_type(self, typename):
        return self.name_to_typedef_map[typename].public

    def get_typedefxml_list(self) :
        """get_typedefxml_list() - Return a list of xml content type descriptions for everything we support"""
        self.reload_directory()
        self.reload_typedefs()
        typedefxml_list = []
        for typename, typedefdict in self.typedef_map.items() :
            typedefxml =  make_contenttypedef_xml(typename, typedefdict)
            typedefxml_list.append(typedefxml)
        return typedefxml_list

#    def __repr__(self):
#        """ Just show what directory and files this instance is using. """
#        return "<%s.ContentTypeDefRepository instance: { dirname: '%s', file_to_typenames_map: '%s' }" % ( self.__module__, self.dirname, `self.file_to_typenames_map` )

    def __repr__(self):
        """  If you need to see what's going on in huge and tremendous detail, 
          this function is for you.
        """
        datastring = """
length of self.filestats is %d
length of self.file_to_typenames_map is %d
length of self.name_to_typedef_map is %d
length of self.typedef_ordered_map is %d
length of self.typedef_multifile is %d
length of self.typedef_automatic_map is %d
length of self.typedef_manual_map is %d
        """ % ( len(self.filestats), len(self.file_to_typenames_map), len(self.name_to_typedef_map), len(self.typedef_ordered_map), len(self.typedef_multifile), len(self.typedef_automatic_map), len(self.typedef_manual_map))
        datastring = datastring + """
\nself.filestats is %s
self.file_to_typenames_map is %s
self.name_to_typedef_map is %s
self.typedef_ordered_map is %s
self.typedef_multifile is %s
self.typedef_automatic_map is %s
self.typedef_manual_map is %s
        """ % ( `self.filestats`, `self.file_to_typenames_map`, `self.name_to_typedef_map`, `self.typedef_ordered_map`, `self.typedef_multifile`, `self.typedef_automatic_map`, `self.typedef_manual_map`)

        return "<%s.ContentTypeDefRepository instance: { dirname: '%s', file_to_typenames_map: '%s' }\ndata is %s" % ( self.__module__, self.dirname, `self.file_to_typenames_map`, datastring )
        self.filestats
        self.file_to_typenames_map
        self.name_to_typedef_map
        self.typedef_ordered_map
        self.typedef_multifile
        self.typedef_automatic_map
        self.typedef_manual_map

# DEBUG code:
def dump_processed_typedefs(srcdir, dstdir):
    import os

    tdr = ContentTypeDefRepository(srcdir)
    for typename in tdr.known_type_list():
        outpath = os.path.join(dstdir, typename.lower().replace('/', '_').replace(' ', '_') + '.mct')
        print 'Writing "%s"...' % outpath
        outfile = open(outpath, 'w')
        outfile.write(make_contenttypedef_xml(typename, tdr.get_type_desc(typename)))
        outfile.close()

mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests('contentlib')

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
