#!/usr/bin/env python

#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

### standard modules
import exceptions
import os
from repr import aRepr, Repr
import sha
import string
from string import translate
import sys
import types

### our modules
from MojoConstants import true, false
import MojoConstants
import MojoErrors
from idlib import _asciihash_re
import mencode
from mojosixbit import b2a


printableascii = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_=+!@#$%^&*()`~[]\{}|;':\",./<>? \t" # I just typed this in by looking at my keyboard.  It probably doesn't matter much if I missed some, because I only use it to guess whether a 20-byte string should be represented as a string or as an ID.  If all of the characters in the string are found `printableascii', then we guess that it is a string, not an id.
nulltrans = string.maketrans('', '')

arepr = aRepr
def builtin_human_readable(arg):
    return arepr.repr(arg)

class MojoRepr(Repr):
    def __init__(self):
        Repr.__init__(self)
        self._objidabbrev = None

    def repr_function(self, obj, level):
        return '<' + obj.func_name + '() at ' + os.path.basename(obj.func_code.co_filename) + ':' + str(obj.func_code.co_firstlineno) + '>'

    def repr_instance_method(self, obj, level):
        return '<' + obj.im_class.__name__ + '.' + obj.im_func.__name__ + '() at ' + os.path.basename(obj.im_func.func_code.co_filename) + ':' + str(obj.im_func.func_code.co_firstlineno) + '>'

    def repr_instance(self, obj, level):
        """
        If it is an instance of Exception, format it nicely (trying to emulate the format that
        you see when an exception is actually raised, plus bracketing '<''s, else defer to
        `Repr.repr_instance()'.
        """
        if isinstance(obj, exceptions.Exception):
            # Don't cut down exception strings so much.
            tms = self.maxstring
            self.maxstring = max(512, tms * 4)
            tml = self.maxlist
            self.maxlist = max(12, tml * 4)
            try:
                if hasattr(obj, 'args'):
                    if len(obj.args) == 1:
                        if type(obj.args[0]) is types.StringType:
                            return '<' + obj.__class__.__name__ + ': ' + obj.args[0] + '>'
                        else:
                            return '<' + obj.__class__.__name__ + ': ' + self.repr1(obj.args[0], level-1) + '>'
                    else:
                        return '<' + obj.__class__.__name__ + ': ' + self.repr1(obj.args, level-1) + '>'
                else:
                    return '<' + obj.__class__.__name__ + '>'
            finally:
                self.maxstring = tms
                self.maxlist = tml

        return Repr.repr_instance(self, obj, level)

    def repr_dictionary(self, obj, level, b2a=b2a):
        """
        Try to format it as a MojoMessage first.  If that fails, call normal dict formatting.
        """
        try:
            if (obj.get('header').get('protocol')[:6] == 'Mojo v') and (obj.get('header').has_key('message type')):
                if not self._objidabbrev:
                    self._objidabbrev = '<' + b2a(sha.new(mencode.mencode(obj)).digest()[:3]) + '>'

                resstr = "(id: %s, type: %s" % (self._objidabbrev, str(obj.get('header').get('message type')))

                # If it has a reference, include that:
                ref = obj.get('header').get('reference')
                if ref:
                    resstr = resstr + ", ref: <" + b2a(ref[:3]) + ">"

                # If it has a mojo offer, print the mo:
                mo = obj.get('message body', {}).get('mojo header', {}).get('mojo offer')
                if mo:
                    resstr = resstr + ", mo: %s" % str(mo)

                # If it has a `mojo message', print that instead of the whole body:
                mm = obj.get('message body').get('mojo message')
                if mm:
                    resstr = resstr + ", msg: %s)" % self.repr1(mm, level)
                else:
                    resstr = resstr + ", body: %s)" % self.repr1(obj.get('message body'), level)

                return resstr
        except:
            pass

        return Repr.repr_dictionary(self, obj, level)

    def repr_string(self, obj, level, asciihashmatch=_asciihash_re.match, b2a=b2a, translate=translate, nulltrans=nulltrans, printableascii=printableascii):
        """
        If it is a unique id, return its abbreviation inside '<' and '>'.  If it is an mencoded
        string, mdecode it and then call `self.repr1()'.  Else, ellide if it is > 70 bytes.
        """
        try:
            if len(obj) == 20:
                # But maybe it was just a 20-character human-readable string, like "credit limit reached", so this is an attempt to detect that case.
                if len(translate(obj, nulltrans, printableascii)) == 0:
                    if self.maxstring >= 22:
                        return `obj`

                    # inlining `repr.repr_string()' (with maxourstring) here...
                    s = `obj[:self.maxourstring]`
                    if len(s) > self.maxourstring:
                        i = max(0, (self.maxourstring-3)/2)
                        j = max(0, self.maxourstring-3-i)
                        s = `obj[:i] + obj[len(obj)-j:]`
                        s = s[:i] + '...' + s[len(s)-j:]
                        # ... done inlining `repr.repr_string()'
                    return s

                return '<' + b2a(obj[:3]) + '>'
            elif (len(obj) == 27) and (asciihashmatch(obj)):
                return '<' + obj[:4] + '>'

            if self.enable_mdecode:
                try:
                    objdec = mencode.mdecode(obj)
                    self._objidabbrev = '<' + b2a(sha.new(obj).digest()[:3]) + '>'
                    return self.repr1(objdec, level - 1)
                except mencode.MencodeError:
                    pass

            # inlining `repr.repr_string()' (with maxourstring) here...
            s = `obj[:self.maxourstring]`
            if len(s) > self.maxourstring:
                i = max(0, (self.maxourstring-3)/2)
                j = max(0, self.maxourstring-3-i)
                s = `obj[:i] + obj[len(obj)-j:]`
                s = s[:i] + '...' + s[len(s)-j:]
            # ... done inlining `repr.repr_string()'
            return s
        except exceptions.StandardError, le:
            return "ERROR IN PRINTING: %s: %s" % (le.__class__.__name__, str(le))
        except exceptions.Exception, le:
            (t, v, tr) = sys.exc_info()
            return "ERROR IN PRINTING: %s: %s" % (`t`, `v`)


mrepr = MojoRepr()

# Note: These levels can get adjusted dynamically!  My goal is to get more info when printing important debug stuff like exceptions and stack traces and less info when printing the contents of messages.  --Zooko 2000-10-14
if MojoConstants.DEBUG:
    mrepr.maxlevel = 24
    mrepr.maxdict = 24
    mrepr.maxlist = 24
    mrepr.maxtuple = 24
    mrepr.maxstring = 800 # note that this doesn't apply to _our_ repr_string above -- only to names of objects, etc.
    mrepr.maxother = 800
    mrepr.maxourstring = 400
    mrepr.enable_mdecode = true  # NOTE: This means dramatic performance hit!  Do not turn on MojoConstants.DEBUG on servers or release software!
else:
    mrepr.maxlevel = 6
    mrepr.maxdict = 6
    mrepr.maxlist = 6
    mrepr.maxtuple = 6
    mrepr.maxstring = 300 # note that this doesn't apply to _our_ repr_string above -- only to names of objects, etc.
    mrepr.maxother = 300
    mrepr.maxourstring = 300
    mrepr.enable_mdecode = None  # NOTE: this will be set to 1 elsewhere if the verbosity is configured high enough

def newnmproved_hr(arg):
    try:
        return mrepr.repr(arg)
    except exceptions.StandardError, le:
        return "ERROR IN PRINTING: %s" % `le.args`
    except exceptions.Exception, le:
        (t, v, tr) = sys.exc_info()
        return "ERROR IN PRINTING: %s: %s" % (`t`, `v`)

# human_readable = builtin_human_readable
human_readable = newnmproved_hr

hr = human_readable


