#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

from MojoConstants import true, false


import threading
import sys


class id_lockError(StandardError):
    pass

class ReleasedALockYouDontHaveError(id_lockError):
    pass

class Locker:
    def __init__(self):
        # You must hold this lock while adding locks to `__map_id_to_lock '.
        self.__lock_on_id_locks = threading.Lock()

        # This maps from unique ids to a lock.
        self.__map_id_to_lock = {}


    def acquire(self, id, blocking=true):
        self.__lock_on_id_locks.acquire()
        try:
            idlock = self.__map_id_to_lock.get(id)
        
            if not idlock:
                idlock = threading.Lock()
                self.__map_id_to_lock[id] = idlock
        finally:
            self.__lock_on_id_locks.release()

        # Yes, it is okay for this `idlock.acquire()' to come after
        # `self.__lock_on_id_locks.release()' (above) instead of before.  --Zooko 2000/05/14
        if blocking:
            idlock.acquire()
        else:
            return idlock.acquire(0)


    def release(self, id):
        self.__lock_on_id_locks.acquire()
        try:
            idlock = self.__map_id_to_lock.get(id)

            if not idlock:
               raise ReleaseALockYouDontHave()

            # Refcount based garbage collection of locks in the dict! :)
            #
            # if the dictionary, this function, and the getrefcount() call are the only 3 references
            # to this Lock() object, we can safely delete it because nobody else is holding a
            # reference that they're in the process of acquiring; and they can't check for it in
            # the self.__map_id_to_lock until we release self.__lock_on_id_locks
            #
            # Cleaning up the dict this way prevents it from growing out of hand over time.  Only
            # useful locks will remain in the dict.   -greg  2000/05/23
            #
            assert sys.getrefcount(idlock) >= 3, "There must be at least three references to this object."

            if sys.getrefcount(idlock) == 3:
                del self.__map_id_to_lock[id]
            
            # now that we've removed it from the dict if possible we can release it.
            idlock.release()
        finally:
            self.__lock_on_id_locks.release()


    def lock_and_do(self, id, func, args=[], kwargs={}):
        self.acquire(id)
        try:
            return apply(func, args, kwargs)
        finally:
            self.release(id)
