#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# Common code for encoding and decoding Mojo Nation (TM) object IDs
# in their two common representations (binary and ascii/mojosixbit).
#
# Everything in this file is optimized for speed, it gets called a
# -lot- throughout the program, including many hot spots.
#
# $Id: idlib.py,v 1.18.2.1 2001/06/12 01:42:46 zooko Exp $

### standard modules
import re
import sha
import types

### our modules
import mojosixbit
import randsource

_asciihash_re = mojosixbit._asciihash_re

try:
    unicode
    _strtypes = (types.StringType, types.UnicodeType)
except:
    _strtypes = (types.StringType,)

def is_canonical_uniq(thing, _strtypes=_strtypes):
    """slightly slower than is_binary_id, but more accurate due to the type check"""
    if type(thing) not in _strtypes:
        return None
    return len(thing) == 20

def identifies(id, thing, thingtype=None):
    """
    @precondition `id' must be an id.: is_sloppy_id(id): "id: %s" % repr(id)
    """
    assert is_sloppy_id(id), "precondition: `id' must be an id." + " -- " + "id: %s" % repr(id)

    return equal(id, make_id(thing, thingtype))

def make_ascii_id(data):
    """Returns a nice simple 27 char ascii encoded id (sha1) of data"""
    return mojosixbit.b2a(string_to_id(data))

def string_to_id(sexpStr):
    """
    @param sexpStr the string containing the expression in canonical form

    @return the unique id of the sexp str

    @postcondition Result is of correct form.: is_binary_id(result)

    @precondition `sexpStr' must be a string.: type(sexpStr) == types.StringType: "sexpStr: %s" % repr(sexpStr)

    @memoizable

    @deprecated in favor of new name: `make_id()'
    """
    return make_id(sexpStr)

def make_id(thing, thingtype=None):
    """
    Use this function to create a unique persistent cryptographically assured id of a string.

    @param thing the thing that you want an id of
    @param thingtype optional type of the thing (ignored)

    @precondition `thing' must be a string.: type(thing) == types.StringType: "thing: %s :: %s" % (hr(thing), hr(type(thing)))

    """
    assert type(thing) == types.StringType, "precondition: `thing' must be a string." + " -- " + "thing: %s :: %s" % (str(thing), str(type(thing)))

    return sha.new(thing).digest()

def canonicalize(id, thingtype=None):
    """
    Use this function to canonicalize an id of one of the "bare" forms (a 20-byte binary string
    or the mojosixbit encoding thereof) into the canonical form.  Useful for calling on ids
    received from other brokers over the wire.

    @param id an id, which may be a bare binary or bare mojosixbit encoded id, or a full canonical id
    @param thingtype optional type of the thing that is identified (ignored)

    @returns the full canonical id

    @precondition `id' must be an id.: is_sloppy_id(id): "id: %s" % repr(id)
    """
    assert is_sloppy_id(id), "precondition: `id' must be an id." + " -- " + "id: %s" % repr(id)

    # NOTE: this method is also known as idlib.to_binary
    # implemented for speed not maintainability:
    if len(id) == 20:
        return id
    else:
        return mojosixbit.a2b(id)

def equal(id1, id2):
    """
    @returns `true' if and only if id1 and id2 identify the same thing;  if `id1' or `id2' or both are `None', then `equals()' returns false.
    """
    if (not id1) or (not id2):
        return None

    return to_binary(id1) == to_binary(id2)

# alternate name
equals = equal

def make_new_random_id(thingtype=None):
    return new_random_uniq()

def string_to_id(sexpStr):
    """
    @param sexpStr the string containing the expression in canonical form

    @return the unique id of the sexp str

    @postcondition Result is of correct form.: is_binary_id(result)

    @precondition `sexpStr' must be a string.: type(sexpStr) == types.StringType: "sexpStr: %s" % repr(sexpStr)

    @memoizable
    """
    assert type(sexpStr) == types.StringType, "precondition: `sexpStr' must be a string." + " -- " + "sexpStr: %s" % repr(sexpStr)

    return sha.new(sexpStr).digest()

def id_to_abbrev(str):
    """
    @precondition `str' must be an id.: is_sloppy_id(str): "str: %s" % repr(str)
    """
    if (len(str) == 27) and (_asciihash_re.match(str)):
        return "<" + str[:4] + ">"
    elif len(str) == 20:
        return "<" + mojosixbit.b2a(str[:3]) + ">"
    else:
        assert is_sloppy_id(str), "precondition: `str' must be an id." + " -- " + "str: %s" % repr(str)

# this gets called a -lot-, it must be fast!
def is_sloppy_id(astr, thingtype=None, _strtypes=_strtypes, _asciihash_re=_asciihash_re):
    return (type(astr) in _strtypes) and ((len(astr) == 20) or ((len(astr) == 27) and (_asciihash_re.match(astr))))

# this gets called a -lot-, it must be fast!
def is_mojosixbitencoded_id(str, thingtype=None, _asciihash_re=_asciihash_re):
    return (len(str) == 27) and (_asciihash_re.match(str))

is_ascii_id = is_mojosixbitencoded_id

# this gets called a -lot-, it must be fast!
def is_binary_id(str, thingtype=None):
    return len(str) == 20

sloppy_id_to_bare_binary_id = canonicalize
to_binary = canonicalize

def to_mojosixbit(sid):
    """
    @precondition `sid' must be an id.: is_sloppy_id(sid): "sid: %s" % repr(sid)
    """
    if _asciihash_re.match(sid):
        return sid

    assert len(sid) == 20, "`sid' must be a mojosixbit or binary id." + " -- " + "sid: %s" % repr(sid)

    # then it must be binary, encode it
    return mojosixbit.b2a(sid)

to_ascii = to_mojosixbit

def newRandomUniq():
    """
    @return a universally unique random number

    @postcondition Result is of correct form.: is_canonical_uniq(result)

    @deprecated in favor of `new_random_uniq()'
    """
    return randsource.get(20)

new_random_uniq = newRandomUniq


#### generic stuff
mojo_test_flag = 1
def run():
    import RunTests
    RunTests.runTests(["idlib"])
    
#### this runs if you import this module by itself
if __name__ == '__main__':
    run()

