#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# This is a library for dealing with sha1 id masks (generating them,
# applying & matching them, etc).
#
__cvsid = '$Id: idmasklib.py,v 1.14 2001/04/30 19:36:03 greg Exp $'

# standard modules
import string
from struct import pack, unpack
import struct

# our modules
from MojoConstants import true, false
import debug
import mojosixbit
import mojoutil
import SEXP
import randsource


class IdMaskError(StandardError) : pass
class IdMaskFormatError(IdMaskError) : pass


def mask_list_includes(includer, includee):
    """
    @returns `true' if and only if every bit string that falls within any of includee's masks
        also falls within at least one of includer's masks

    This is used in MetaTrackerLib to say "I want to know about blob servers with masklist A.
    If we already know about blob servers with masklist B, and mask_list_includes(B, A), then
    I'll be satisfied with the information we already have."

    XXXX Unfortunately this incorrectly returns `false' if you ask it whether a mask list A
    includes mask list B, where B has one mask with 1 significant bit, equal to 0, and A has
    two masks with two significant bits each, one mask "01" and one mask "00".

    Note: A masklist of length 0 covers no strings, which means that if `includer' is a list
    of length 0, then return `false' unless the `includee' parameter is also a list of length
    0.  The mask list which covers no strings is the mask list of length 0.

    If the `includee' parameter is a list of length 0 then always return `true'.

    @precondition `includer' must not be `None' (although it may be a sequence of length 0).: includer is not None: "includer: %s" % hr(includer)
    @precondition `includee' must not be `None' (although it may be a sequence of length 0).: includee is not None: "includee: %s" % hr(includee)
    """
    assert includer is not None, "precondition: `includer' must not be `None' (although it may be a sequence of length 0)." + " -- " + "includer: %s" % hr(includer)
    assert includee is not None, "precondition: `includee' must not be `None' (although it may be a sequence of length 0)." + " -- " + "includee: %s" % hr(includee)

    # assume true, test for false, return
    for eemask in includee:
        thisoneincluded = false # assume false, test for true, break
        for ermask in includer:
            if eemask.is_subset_of(ermask):
                # Okay this particular mask from `includee' is covered.
                thisoneincluded = true
                break
        if not thisoneincluded:
            # This one is not included in any of the masks of includer.  Lose.
            return false
    return true


########################################################################
def generate_random_mask(bits, keepfull = true):
    """Returns a random IdMask of at most bits bits."""
    return IdMask(mask=randsource.get((bits+8)/8), bits=bits, keepfull = keepfull)


########################################################################
def compare_idmasks(a, b) :
    """A comparison function for comparing idmask objects (useful for sorting) [-1 if a<b, 0 of a==b, 1 if a>b]"""
    if a.significant_bits == b.significant_bits :
        # have the same number of bits? create a fake id to check if a falls within b.
        fakeid = b.mask + ('\0' * (20-len(b.mask)))
        if a.match(fakeid) :
            return 0  # the masks match
        else :
            return cmp(a.mask, b.mask)
    else :
        # 0 significant bits has the highest value of all masks.
        # Masks with fewer significant bits sort "greater than" masks which fall within them.
        if a.is_subset_of(b) :
            return -1
        elif b.is_subset_of(a) :
            return 1
        else :
            # Brainstorm: just pad out the mask to a full 20 bytes to
            # the beginning and ending value that match and use string
            # comparisons!
            arange = a.get_first_and_last_match()
            brange = b.get_first_and_last_match()
            
            if arange[1] < brange[0] :
                # if the a ends before b begins, b is greater
                return -1
            elif arange[0] > brange[1] :
                # if a begins after b ends, a is greater
                return 1

            raise RuntimeError, 'comparison failure! oops! we should never get here'



########################################################################
class IdMask :
    def __init__(self, mask='\0', bits=0, onesandzeros=None, mojodict=None, keepfull=0) :
        """
        @param mask is used to initialize our mask (if no mojodict is given)
        @param bits is used to specify the number of valid bits in our mask (if no mojodict is given)
        @param onesandzeros is a string of 1 and 0 characters (1100100010); if it is not None it will
                   be used instead of the mask & bits parameters.  This allows for easily human
                   understandable mask construction.  onesandzeros is not an efficient constructor.
        @param mojodict overrides all other initializers if specified.
        @param keepfull is a boolean, if true, a full 160 bits are always stored.  Extending or retracting
                   the length just affects how many of the 160 bits are used.
        """
        self.keepfull = keepfull
        if mojodict :
            # TODO left off here -greg 22/jun/2000
            self.mask = mojosixbit.a2b(mojodict['mask'])
            self.significant_bits = int(mojodict['bits'])
        else :
            if onesandzeros is not None:
                assert type(onesandzeros) == type('')
                assert len(onesandzeros) <= 160

                self.significant_bits = len(onesandzeros)
                if (len(onesandzeros) % 8) != 0:
                    onesandzeros = onesandzeros + ('0' * (8-(len(onesandzeros)%8)))

                # Figure out self.mask without using add_bit, in case keepfull is true.
                newmask = []
                for lowindex in range(0, len(onesandzeros), 8):
                    segment = onesandzeros[lowindex:lowindex+8]
                    byte = 0
                    for bitindex in range(len(segment)):
                        byte = byte | (int(segment[bitindex]) * 2 ** (len(segment)-bitindex-1))
                    newmask.append(chr(byte))
                self.mask = string.join(newmask, '')
            else:
                assert len(mask) <= 20, "mask %s must be a BINARY string, no more than 160 bits long" % `mask`
                # the (binary) string representation of our mask
                self.mask = mask
                # the number of significant bits in this mask (0 - 160)
                self.significant_bits = bits
                
        if self.significant_bits < 0 or self.significant_bits > 160 :
            raise IdMaskFormatError, "invalid number of significant_bits"
        if len(self.mask) < (self.significant_bits+7)/8 :
            raise IdMaskFormatError, "mask too short for given number of bits"
        if len(self.mask) > 20 :
            raise IdMaskFormatError, "mask too long: %d bytes of binary string (max is 20)" % len(self.mask)
        
        self.__trim_or_extend()
    
    def get_onesandzeros(self, fullmask=None) :
        """
        Return a string made up of the characters '1' and '0' that represents this block id mask.
        Only significant bits are returned, so the length of the string returned is equal to
        the number of significant bits in the block id mask.
        """
        onesandzeros = ''
        for byte in self.mask:
            # Only the last eight bits of each string-representation of
            # an integer are meaningful for a character.  Grab them for
            # each binary character
            onesandzeros = onesandzeros + int2string_of_bits(ord(byte))[-8:]
        significant_onesandzeros = onesandzeros[:self.significant_bits]
        if fullmask:
            suffix = '1' * (32 - self.significant_bits)
            return significant_onesandzeros + suffix
        else:
            return significant_onesandzeros

    def get_mojodict(self, fullmask=None) :
        """
        Return a mojodict describing this block id mask that we'll use for communicating it to others.

        If self.keepfull is true and fullmask is true, the dict contains the entire mask.
        If self.keepfull is true and fullmask is false, the dict appears the same as if self.keepfull is not true.
        If self.keepfull is false, fullmask has no effect.
        
        Keyed on:
          mask - a mojosixbit.b2a encoded bitmask
          bits - integer number of significant bits
        """
        assert self.significant_bits >= 0 and self.significant_bits <= 160

        self.__trim_or_extend()
        
        if self.keepfull and fullmask:
            ascii_mask = mojosixbit.b2a(str(self.mask))
        else:
            bytes_that_matter = self.significant_bits / 8
            if (self.significant_bits & 7) > 0 :
                bytes_that_matter = bytes_that_matter + 1
            if bytes_that_matter <= 0 :
                bytes_that_matter = 1
            ascii_mask = mojosixbit.b2a(str(self.mask[:bytes_that_matter]))

        dict_repr = {
            'bits': str(self.significant_bits),
            'mask': ascii_mask,
        }

        return dict_repr
    
    def get_first_and_last_match(self) :
        """Returns: tuple containing the first and last id that match this bitmask"""
        if self.significant_bits <= 0 :
            return ('\0' * 20, '\377' * 20)
        mask = self.__trim_unused_bits()
        first = mask + ('\0' * (20 - len(mask)))

        last = mask
        # add one bits to the end of the last byte if the mask is not on an even byte boundary
        if (self.significant_bits & 7) > 0 :
            last_byte_to_fix = (self.significant_bits) / 8
            ones_to_add = 8 - (self.significant_bits & 7)
            one_bits = (2**ones_to_add) - 1
            (last_byte,) = struct.unpack('B', last[-1])
            last = last[:-1] + struct.pack('B', (last_byte|one_bits))
        last = last + ('\377' * (20 - len(mask)))
        return (first, last)

    def __trim_or_extend(self):
        if self.keepfull:
            self.__extend_full_mask()
        else:
            self.__trim_unused_bits()
            
    def __trim_unused_bits(self) :
        """
        This trims any unused bits off of our self.mask string and return the result.

        If self.keepfull is true, this doesn't alter self.mask, else it sets self.mask equal
        to the return value.
        """
        bytes_that_matter = self.significant_bits / 8
        remaining_bits = (self.significant_bits & 7)
        if remaining_bits > 0 :
            bytes_that_matter = bytes_that_matter + 1
        if bytes_that_matter <= 0 :
            bytes_that_matter = 1
        mask = self.mask[:bytes_that_matter]
        if remaining_bits > 0:
            (final_byte,) = struct.unpack('B', mask[bytes_that_matter-1])
            last_byte_mask = 0xff ^ ( (1 << (8-remaining_bits)) - 1 )
            final_byte = final_byte & last_byte_mask  # remove any unused bits
            mask = mask[:bytes_that_matter-1] + struct.pack('B', final_byte)
        if not self.keepfull:
            self.mask = mask
        return mask

    def __extend_full_mask(self):
        """
        Ensures that there are a full 160 bits in the mask.

        Warning:  If len(self.mask) < 20 and self.mask isn't the minimal number of
        bytes to contain self.significant_bit number of bits, then this function will screw
        up self.mask!
        """
        if not self.keepfull:
            return
        if len(self.mask) < 20:
            filter = 2 ** ((160-self.significant_bits) % 8) - 1
            rfilter = (255 ^ filter) # rfilter is the byte-length NOT of the filter.

            if len(self.mask) > 0:
                byte = ord(self.mask[-1])
            else:
                byte = 0
            byte = byte & rfilter # Clear all bits not explicitly within the mask.
            randbits = ord(randsource.get(1))
            byte = byte | (randbits & filter) # Fill all the bits cleared in the previous step with random bits, while preserving bits within the mask.
            byte = chr(byte) # Convert back into a byte.
            
            self.mask = self.mask[:-1] + byte + randsource.get(20-len(self.mask)) # Old bytes + union byte of old and new + new bytes.

    def __repr__(self) :
        """This is nice for printing masks"""
        return '<%s.IdMask instance: %s>' % (self.__module__, repr(self.get_mojodict()), )
    
    def debug_repr(self) :
        s = self.get_mojodict()
        s['mask'] = mojosixbit.a2b(s['mask'])
        return s

    def match(self, id) :
        """
        @param id is a mojonation id (in binary or ascii, it doesn't matter)

        @returns `true' if and only if this mask matches the given id
        """
        assert len(id) == 20 or len(id) == 27

        if len(id) == 27 :
            # convert ascii ids to binary for easier use
            id = mojosixbit.a2b(id)


        full_bytes = self.significant_bits / 8
        if self.mask[:full_bytes] != id[:full_bytes] :
            return 0

        remaining_bits = (self.significant_bits & 7)
        if remaining_bits != 0 :
            last_mask = 0xff ^ ( (1 << (8-remaining_bits)) - 1 )
            (mask_last_byte, ) = unpack('B', self.mask[full_bytes])
            (id_last_byte, ) = unpack('B', id[full_bytes])

            if (mask_last_byte & last_mask) != (id_last_byte & last_mask) :
                return 0

        return 1  # the mask matches!

    def add_bit(self, value=None) :
        """
        Add a bit to our mask; it will be random unless the value parameter is specified
        If self.keepfull is true and value is not None an exception is raised.
        """
        if self.keepfull:
            assert value is None, "A bit value cannot be passed to a keepfull bitmask."
            self.__trim_or_extend()
            self.significant_bits = min(self.significant_bits + 1, 160)
            return

        if value is None :
            (randombyte,) = struct.unpack('B', randsource.get(1))
        else :
            if value :
                randombyte = 0xff
            else :
                randombyte = 0x00
        bitnum = 0x7 ^ ((self.significant_bits) & 0x7)
        bitnum_mask = 1 << bitnum
        randombit_in_byte = randombyte & bitnum_mask
        byte_in_mask_to_modify = (self.significant_bits) / 8
        #D#print self.debug_repr(), 'bitnum %02x, bitnum_mask %02x, randombit_in_byte %02x, byte_in_mask_to_modify %d' % (bitnum, bitnum_mask, randombit_in_byte, byte_in_mask_to_modify)
        if len(self.mask)-1 < byte_in_mask_to_modify :
            # we just need to append a single random bit in a new byte
            assert bitnum == 7
            if len(self.mask) != byte_in_mask_to_modify :
                # this shouldn't happen, but lets check for it anyways
                raise IdMaskError, "internal mask representation was shorter than the number of bits"
            if randombit_in_byte :
                self.mask = self.mask + '\200'
            else :
                self.mask = self.mask + '\000'
        else :
            # we need to force the new bit into the last byte of mask
            (value_of_byte_to_modify,) = struct.unpack('B', self.mask[-1])
            if randombit_in_byte :
                # the bit is one, so lets set it
                value_of_byte_to_modify = value_of_byte_to_modify | randombit_in_byte
            else :
                # the bit is zero, we need to clear it
                value_of_byte_to_modify = value_of_byte_to_modify & (0xff ^ bitnum_mask)
            # remove the old byte and append the new one
            self.mask = self.mask[:-1] + struct.pack('B', value_of_byte_to_modify)

        self.significant_bits = self.significant_bits + 1

    def remove_bit(self) :
        """Decrement the number of significant bits by one"""
        self.significant_bits = max(0, self.significant_bits-1)
        self.__trim_or_extend()

    def is_subset_of(self, othermask) :
        """
        @return `true' if and only if every bitstring that matches mask also matches othermask
        """
        if self.significant_bits >= othermask.significant_bits :
            # we have the same or more bits, create a fake id of us extended with zeros to check
            # if we fall within othermask (our fake id should match the other mask).
            fakeid = self.mask + ('\0' * (20-len(self.mask)))
            return othermask.match(fakeid)
        else:
            # If we have fewer significant bits, we can't fall within othermask.
            return false

    def set_length(self, length):
        assert length >= 0 and length < 160, "A bitmask must have a length between 0 and 160.  (Anything over 14 is questionable, currently.)"
        if self.keepfull:
            self.__extend_full_mask()
            self.significant_bits = length
        else:
            if length > self.significant_bits:
                for n in range(length - self.significant_bits):
                    self.add_bit()
            else:
                for n in range(self.significant_bits - length):
                    self.remove_bit()
        
    def overlaps_with(self, othermask) :
        """Returns true if the range covered by these masks overlap in any way"""
        return self.is_subset_of(othermask) or othermask.is_subset_of(self)
    
    def __cmp__(self, b):
        return compare_idmasks(self, b)

def int2string_of_bits(intmask):
    """
    Turn an integer into a string of zeros and ones representing
    that integer in base 2.
    e.g. _int2string_of_bits(0x64) returns '00000000000000000000000001100100'
    """
    asciistring = ''
    # Find out how many bits are in an int.
    i = 0x1
    bitcount = 0
    while i != 0:
        bitcount = bitcount + 1
        i = i << 1

    # Create a string representing the bits of the int.
    for i in range(bitcount):
        if ( 0x1 & intmask ):
            asciistring = '1' + asciistring
        else:
            asciistring = '0' + asciistring
        intmask = intmask >> 1

    return asciistring


def create_full_length_base_bitmask(precursor='', preclen=0):
    """
    Create a 160 bit string for use as a base bitmask.  This is useful because the user can adjust the bitmask
    length back and forth without worrying about excluding blocks.

    @pre precursor is a string of non-64 bit encoded bytes which represent the bitmask.
    precursor is an existing bitmask.  This function extends the precursor to 160 bits.  This is useful if there
    already exists a short (say 6-bit) mask, and the full length base bitmask should match the bits in the
    existing bitmask.  (So in the case of a 6-bit mask, this function would copy those 6 bits, then append
    154 bits.)
    """
    return precursor[:(preclen+8)/8] + randsource.get((160-preclen)/8)
    
    
##################################### Tests #####################################


def test_id_misc() :
    spam = IdMask()
    spamdict = spam.get_mojodict()
    spamrepr = repr(spam)

def test_set_length():
    import whrandom
    for keepfull in (false, true):
        idm = IdMask(keepfull=keepfull)
        for n in range(0, 10):
            length = whrandom.randint(0, 160)
            idm.set_length(length)
            assert int(idm.get_mojodict()['bits']) == length
        
def test_id_dict_init() :
    for keepfull in (false, true):
        ourmask = 'a0spamA'
        spamdict = {
            'mask': ourmask,
            'bits': '3',
            }
        spam0 = IdMask(mask=mojosixbit.a2b(ourmask), bits=3, keepfull=keepfull)
        spam1 = IdMask(mojodict=spamdict, keepfull=keepfull)
        # XXX is this a valid test (the repr could be in a different order)
        assert repr(spam0) == repr(spam1)

def test_mask_matching() :
    for keepfull in (false, true):
        mask = IdMask(mask='\377\377', bits=0, keepfull=keepfull)
        assert mask.match('\200\0'*10)
        assert mask.match(mojosixbit.b2a('\200\0'*10))
        assert mask.match('\000\0'*10)
        assert mask.match(mojosixbit.b2a('\000\0'*10))
        mask = IdMask(mask='\377\377', bits=1, keepfull=keepfull)
        assert mask.match('\300\0'*10)
        assert mask.match('\200\0'*10)
        assert not mask.match('\100\0'*10)
        assert not mask.match(mojosixbit.b2a('\100\0'*10))
        mask = IdMask(mask='\377\377', bits=9, keepfull=keepfull)
        assert mask.match('\377\371'*10)
        assert not mask.match(mojosixbit.b2a('\377\161'*10))
        assert not mask.match('\100\0'*10)
        # test even byte boundary number of bits
        mask = IdMask(mask='\315', bits=8, keepfull=keepfull)
        assert mask.match('\315abc'*5)

def test_bit_addition_and_removal() :
    for keepfull in (false, true):
        mask = IdMask(keepfull=keepfull)
        for x in range(0, 160) :
            mask.add_bit()
        assert(len(mojosixbit.a2b(mask.get_mojodict()['mask'])) == 20) # 160 random bits
        assert(mask.get_mojodict()['mask'] != 'A'*27)  # 160 random bits are rarely zero ;)
        for x in range(0, 160) :
            mask.remove_bit()
        assert(len(mask.get_mojodict()['mask']) == 2)

def test_is_subset_of() :
    # Mutables:
    m1, m2 = IdMask(keepfull=false), IdMask(keepfull=false)
    assert m1.is_subset_of(m1)
    assert m2.is_subset_of(m2)
    m1.add_bit(1)
    assert m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    m2.add_bit(0)
    assert not m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    assert m1.is_subset_of(m1)
    assert m2.is_subset_of(m2)
    m1, m2 = IdMask(mask='\315', bits=8, keepfull=false), IdMask(mask='\315', bits=8, keepfull=true)
    for x in range(0, 126) :
        m1.add_bit()
    assert m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    for x in range(0, 126) :
        m2.add_bit()
    assert not m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    # Immutable test:
    m1, m2 = IdMask(keepfull=true), IdMask(keepfull=true)
    assert m1.is_subset_of(m1)
    assert m2.is_subset_of(m2)
    m1 = IdMask(mask='\200', bits=1, keepfull=true)
    assert m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    m2 = IdMask(mask='\000', bits=1, keepfull=true)
    assert not m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    assert m1.is_subset_of(m1)
    assert m2.is_subset_of(m2)
    m1, m2 = IdMask(mask='\315', bits=8, keepfull=true), IdMask(mask='\315', bits=8, keepfull=true)
    for x in range(0, 126) :
        m1.add_bit()
    assert m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    for x in range(0, 126) :
        m2.add_bit()
    assert not m1.is_subset_of(m2)
    assert not m2.is_subset_of(m1)
    

def test_overlaps_with() :
    for keepfull in (false, true):
        m1 = IdMask(mask='\315', bits=8, keepfull=keepfull)
        m2 = IdMask(mask='\315a', bits=9, keepfull=keepfull)
        assert m1.overlaps_with(m1)
        assert m2.overlaps_with(m1)
        assert m1.overlaps_with(m2)
        assert m2.overlaps_with(m2)

def test_compare_idmasks() :
    m1 = IdMask(keepfull=false)
    m2 = IdMask(keepfull=false)
    assert compare_idmasks(m1, m2) == 0
    m1.add_bit(1)
    assert compare_idmasks(m1, m2) < 0
    m1.remove_bit()
    assert compare_idmasks(m1, m2) == 0
    m2.add_bit(1)
    assert compare_idmasks(m1, m2) > 0
    m2.remove_bit()
    assert compare_idmasks(m1, m2) == 0
    m2.add_bit(0)
    assert compare_idmasks(m1, m2) > 0
    m1.add_bit(0)
    assert compare_idmasks(m1, m2) == 0
    m1.add_bit(1)
    assert compare_idmasks(m1, m2) < 0

def test_first_and_last_match() :
    for keepfull in (false, true):
        m1 = IdMask(mask='\300', bits=2, keepfull=keepfull)
        a, b = m1.get_first_and_last_match()
        assert b == ('\377' * 20)
        assert a == ('\300' + ('\0' * 19))
        m1 = IdMask(mask='\000', bits=2, keepfull=keepfull)
        a, b = m1.get_first_and_last_match()
        assert b == ('\077' + ('\377' * 19))
        assert a == ('\0' * 20)
        m1 = IdMask(mask='spamSpamSpAm!', bits = 81, keepfull=keepfull)
        assert m1.get_first_and_last_match() == ('spamSpamSp\000\000\000\000\000\000\000\000\000\000', 'spamSpamSp\177\377\377\377\377\377\377\377\377\377')
        m1 = IdMask(mask='spamSpamSpAm!', bits = 80, keepfull=keepfull)
        assert m1.get_first_and_last_match() == ('spamSpamSp\000\000\000\000\000\000\000\000\000\000', 'spamSpamSp\377\377\377\377\377\377\377\377\377\377')


def test_int2string_of_bits(verbose=0):
    testdict = {}
    testdict[0x00000000] = '00000000000000000000000000000000'
    testdict[0x00000001] = '00000000000000000000000000000001'
    testdict[0xffffffff] = '11111111111111111111111111111111'
    testdict[0x00000064] = '00000000000000000000000001100100'
    testdict[0x0ffffff0] = '00001111111111111111111111110000'
    testdict[0x80000000] = '10000000000000000000000000000000'

    for input in testdict.keys():
        if verbose:
            print 'running test: _int2string_of_bits(0x%08x)' % input
        output = int2string_of_bits(input)
        if output != testdict[input]:
            print 'int2string_of_bits() TEST FAILED with input 0x%08x, returned %s' % (input, output)
            print 'expected', testdict[input]


def test_onesandzeros():
    for keepfull in (false, true):
        ozstring = '11111111'
        m1 = IdMask(onesandzeros=ozstring, keepfull=keepfull)
        m2 = IdMask(mask='\377', bits=8, keepfull=keepfull)
        assert m1 == m2
        assert ozstring == m1.get_onesandzeros()
        assert m2.get_onesandzeros(fullmask=1) == '1' * 32

        ozstring='11'
        m1 = IdMask(onesandzeros=ozstring, keepfull=keepfull)
        m2 = IdMask(mask='\300', bits=2, keepfull=keepfull)
        assert m1 == m2
        assert ozstring == m1.get_onesandzeros()
        assert m2.get_onesandzeros(fullmask=1) == '1' * 32

        ozstring='1100'
        m1 = IdMask(onesandzeros='1100', keepfull=keepfull)
        m2 = IdMask(mask='\300', bits=4, keepfull=keepfull)
        assert m1 == m2
        assert ozstring == m1.get_onesandzeros()
        assert m2.get_onesandzeros(fullmask=1) == '1100' + ('1' * 28)

        ozstring='1100'
        m1 = IdMask(onesandzeros='1100', keepfull=keepfull)
        m2 = IdMask(mask='\300', bits=3, keepfull=keepfull)
        assert m1 != m2
        assert '110' == m2.get_onesandzeros()
        assert m2.get_onesandzeros(fullmask=1) == '110' + ('1' * 29)

        m1 = IdMask(onesandzeros='1100', keepfull=keepfull)
        m2 = IdMask(mask='\300', bits=5, keepfull=keepfull)
        assert m1 != m2
        assert '11000' == m2.get_onesandzeros()
        assert m2.get_onesandzeros(fullmask=1) == '11000' + ('1' * 27)

        ozstring='11001011001010110101011010'
        m1 = IdMask(onesandzeros='11001011001010110101011010', keepfull=keepfull)
        m2 = IdMask(mojodict={'mask': 'yytWgA', 'bits': '26'}, keepfull=keepfull)
        assert m1 == m2
        assert ozstring == m2.get_onesandzeros()
        assert m2.get_onesandzeros(fullmask=1) == ozstring + '111111'

        # Test the longest possible mask (160 bits).
        ozstring = '11001100' * 20
        m1 = IdMask(onesandzeros=ozstring)
        longmask = mojosixbit.b2a('\314' * 20)
        m2 = IdMask(mojodict={'mask': longmask, 'bits': '160'}, keepfull=keepfull)
        assert m1 == m2
        assert ozstring == m2.get_onesandzeros()
        assert m2.get_onesandzeros(fullmask=1) == ozstring

def test_mask_list_includes():
    for keepfull in (false, true):
        m1, m2 = IdMask(keepfull=keepfull), IdMask(keepfull=keepfull)
        m3, m4 = IdMask(keepfull=keepfull), IdMask(keepfull=keepfull)
        ml1 = [m1, m2]
        ml2 = []

        assert mask_list_includes(includer=ml1, includee=ml2) # anything includes empty list
        assert not mask_list_includes(includer=ml2, includee=ml1) # empty list includes nothing

        ml2 = [m3, m4]

        assert mask_list_includes(ml1, ml2) # nulls include
        assert mask_list_includes(ml2, ml1) # nulls include

        m1.add_bit(1)

        assert mask_list_includes(ml1, ml2) # m2 still includes all and m4 still includes all.
        assert mask_list_includes(ml2, ml1) # m2 still includes all and m4 still includes all.

        m3.add_bit(0)

        assert mask_list_includes(ml1, ml2) # m2 still includes all and m4 still includes all.
        assert mask_list_includes(ml2, ml1) # m2 still includes all and m4 still includes all.

        m2.add_bit(1)

        assert not mask_list_includes(ml1, ml2) # specific doesn't include general
        assert mask_list_includes(ml2, ml1) # general includes specific
  

        m1 = IdMask(keepfull=keepfull)
        m3, m4 = IdMask(keepfull=keepfull), IdMask(keepfull=keepfull)
        ml1 = [m1]
        ml2 = [m3]

        m1.add_bit(0)
        m3.add_bit(0)
        m3.add_bit(1)
 
        assert not mask_list_includes(includer=ml2, includee=ml1) # m3 is too specific

        ml2 = [m3, m4]
        m4.add_bit(0)
        m4.add_bit(0)

        assert mask_list_includes(includer=ml1, includee=ml2), "This test is known to fail.  Fixing it would be tricky.  Bram would probably enjoy fixing this test." # the pair of m3, m4 together cover m1
        assert mask_list_includes(includer=ml2, includee=ml1), "This test is known to fail.  Fixing it would be tricky.  Bram would probably enjoy fixing this test." # the pair of m3, m4 together cover m1


__doc__ = """
  generate_random_mask(bits) function
  compare_idmasks(a, b) function
  IdMask class

  See idmasklib.IdMask.__doc__ for real details of this module.

HOWTO - for generating understandable masks by hand:

>>> import idmasklib
>>> idmasklib.IdMask(onesandzeros="110")
<idmasklib.IdMask instance: {'mask': 'wA', 'bits': '3'}>
>>>
"""


mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests('idmasklib')

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
