#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# This module implements our local disk based block storage.
#
__cvsid = '$Id: localblockstore.py,v 1.63.2.1 2001/06/11 13:54:04 zooko Exp $'

# standard modules
from cPickle import dumps, loads
import exceptions
import os
import whrandom
import time
import string
import sys
from threading import Lock
import threading
import types
from loggedthreading import LoggedThread
import math
from cStringIO import StringIO

# our modules
import DoQ
import BidStrategy
from MojoConstants import true, false
import DataTypes 
from bsddb3 import db, dbobj
from CleanLogDb import CleanLogDbEnv
import debug
from human_readable import hr
import idlib
import randsource
import mojoutil
import mojosixbit
import confutils
import MetaTrackerLib
import Cache

# so only one freaking block store grovel can happen at a time
GROVEL_LOCK = Lock()

class ShutdownException(exceptions.StandardError): pass


########################################################################
# Custom exceptions
########################################################################
class Hell(StandardError) :   # shouldn't happen == silly name :)
    msg = "BlockStore not initialized"
class PathError(Hell) :
    msg = "Path does not exist"
class NoBlock(Hell) :
    msg = "Block data not found"
class BLBlock(Hell) :
    msg = "BlobID is on black list"
class NoIndex(Hell) : pass
class NoSpaceError(Hell) : pass
class CantRunDFError(Hell) :
    msg = "Cannot run df()"
class PlatformNotSupportedError(Hell) :
    msg = "Platform not supported"


_MAX_DEPTH = 4

MINS_BETWEEN_DB_CHECKPOINTS = 1

########################################################################
########################################################################
class BlockStore:
    """class BlockStore(path=None, maxspace=None, depth=2)
    Manages a simple database using the filesystem by creating
    several wide levels of subdirectories up to a depth specified in the
    constructors depth parameter.  This minimizes the impact of a linear
    algorithm in many OSes namei implementation when looking up files in
    very large directories.  (btree and hash based filesystems such as SGIs
    XFS or linux's ReiserFS wouldn't need this; most OSes aren't so lucky)

    An index is kept in a BerkeleyDB database called index.db.  It
    contains info about what blocks exist and stats on the accesses
    of each block.
    """

    ####################################################################
    def __init__(self, path, maxspace=None, depth=2, recoverdb=true, name='*unnamed*', rebuild=true) :
        self.name = name
        # max bytes in this blockstore (`None' means infinite, 0 means files/dbs are touched and this instance is a dummy)
        self.maxspace = maxspace
        if self.maxspace == 0:
            return

        # used during cleanup when making more space
        self.__cleanup_lock = Lock()

        # set the subdirectory tree depth to create and use
        if depth < 0 or depth > _MAX_DEPTH :
            raise ValueError, "depth must be >= 0 and <= "+str(_MAX_DEPTH)
        self.subdirdepth = depth

        self.init_space(path=path)   # (this sets self.basepath)

        recover_flag = 0
        if recoverdb :
            recover_flag = db.DB_RECOVER

        db_env_open_flags = db.DB_CREATE | db.DB_INIT_MPOOL | db.DB_INIT_TXN | db.DB_TXN_NOSYNC | db.DB_INIT_LOG | db.DB_PRIVATE | recover_flag
        try:
            self.db_env = CleanLogDbEnv()
            try:
                self.db_env.open(self.basepath, db_env_open_flags)
            except db.DBError, dbe:
                debug.mojolog.write('Failed to open the database environment the first time, reason: %s\nTrying again...\n', args=(dbe,), vs='BlockStore', v=2)
                # XXX HACK sometimes trying again after one error works.
                try:
                    self.db_env.open(self.basepath, db_env_open_flags | db.DB_RECOVER)
                except db.DBError, dbe:
                    debug.mojolog.write('Failed to open the database environment the second time, reason: %s\nTrying again...\n', args=(dbe,), vs='BlockStore', v=2)
                    # XXX DOUBLE CHOCOLATEY HACK sometimes trying *again* after one open *without* DB_RECOVER works.
                    self.db_env.open(self.basepath, db_env_open_flags & (~db.DB_RECOVER))

            # keyed on blobid, stores a pickled dict keyed on 'size', 'path', 'stored on', 'last access' and 'num accesses', 'mojo earned'
            self.index_db = dbobj.DB(self.db_env)
            self.index_db.open('index.db', db.DB_BTREE, db.DB_CREATE)
        except (db.DBError, IOError), e:  # XXX
            debug.mojolog.write("WARNING: (%s) ** Exception %s opening block server index.db! rebuilding index. **\n", args=(self.name, e), vs='localblockstore')
            # clean up any half-opened things (let their destructors take care of them)
            self.index_db = None
            self.db_env = None

            # Remove all of the DB log files, for some ?%$&@E!& reason we've seen them be corrupted and
            # cause an IO Error when attempting to open the block storage database.
            dblogfiles = filter(lambda n: string.lower(n[:len('log')]) == 'log', os.listdir(self.basepath))
            map(lambda n, bp=self.basepath: os.remove(os.path.join(bp, n)), dblogfiles)

            self.db_env = CleanLogDbEnv()
            self.db_env.open(self.basepath, db_env_open_flags)
            self.index_db = dbobj.DB(self.db_env)
            self.index_db.open('index.db', db.DB_BTREE, db.DB_CREATE)

        self.recount = false
        if self.index_db.get('total bytes used') == None :
            self.index_db.put('total bytes used', dumps(0L))
            self.recount = true

        # start a thread rebuilding our index; this is done every time
        # we run now until we work out why it is even necessary to do this.
        self._shuttingdownflag = threading.Event()
        self._startedverifyingflag = false
        self._doneverifyingflag = threading.Event()
        if rebuild:
            rt = LoggedThread(target=self._rebuild_index, name="%s rebuild index" % self.name)
            rt.setDaemon(1)
            rt.start()
        else:
            self._doneverifyingflag.set()

    ####################################################################
    def __del__(self) :
        if self.maxspace == 0:
            return
        
        if hasattr(self, 'index_db') and self.index_db is not None :
            self.index_db.close()
            del self.index_db
        if hasattr(self, 'db_env') and self.db_env is not None :
            self.db_env.nosyncerror_txn_checkpoint(0)
            self.db_env.close()
            del self.db_env

    def shutdown(self):
        self._shuttingdownflag.set()
        self.wait_until_done_verifying()   # verifier threads will see the event set and exit soon

        if self.maxspace == 0:
            return

        # XXX be sure nothing else will use the blockstore before calling this!
        self.index_db.close()
        del self.index_db
        self.db_env.nosyncerror_txn_checkpoint(0)
        self.db_env.close()
        del self.db_env

    def wait_until_done_verifying(self):
        if self._startedverifyingflag:
            self._doneverifyingflag.wait()

    ####################################################################
    def init_space(self, path=None) :
        """init_space([path])
        Create the necessary filesystem structure or database in the directory
        specified by path.
        """
        if path :
            self.basepath = os.path.abspath(path)

        # Check the existence of basepath and create any necessary
        # blockstore index files there.
        if not os.path.isdir(self.basepath) :
            os.makedirs(self.basepath)

        # Check the existence of the temporary file directory
        if not os.path.isdir(os.path.join(self.basepath, 'tmp')) :
            os.makedirs(os.path.join(self.basepath, 'tmp'))


    def __get_and_make_subdir(self, blobid, base='', depth=None) :
        """
        generate a subdirectory name for storing blobid; create the directory if it doesn't exist

        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        if depth is None :
            depth = self.subdirdepth

        if depth > len(blobid) or depth > _MAX_DEPTH :
            raise ValueError, 'depth to large: '+str(depth)

        if depth < 1 :
            path = base
        else :
            path = os.path.join(base, reduce(os.path.join, list(blobid[:depth])))

        try:
            os.makedirs(path)
        except os.error:
            pass  # the directory already exists; good.
        return path

    ####################################################################
    def __generate_id_filename(self, blobid) :
        """
        figure out the full path for this blobid within our blockstore

        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        if self.subdirdepth < 1 :
            return os.path.join(self.basepath, blobid)
        else :
            return os.path.join(self.__get_and_make_subdir(blobid, self.basepath), blobid)

    ####################################################################
    def __update_space_totals(self, delta) :
        """Add delta to our total bytes used value"""
        self.db_env.nosyncerror_txn_checkpoint(MINS_BETWEEN_DB_CHECKPOINTS)
        txn = self.db_env.txn_begin()
        try:
            totalbytes = loads(self.index_db.get('total bytes used', txn=txn))
            totalbytes = totalbytes + delta
            if totalbytes < 0 :
                totalbytes = 0L
            self.index_db.put('total bytes used', dumps(totalbytes), txn=txn)
        finally:
            if txn :
                txn.commit()

    ####################################################################
    def __cleanup_old_blobs(self, minimumfree) :
        """
        make sure at least minimumfree bytes are available

        @precondition This method must be called on the DoQ.: DoQ.doq.is_currently_doq()
        """
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        # TODO XXX this needs an intelligent implementation.  For now
        # it randomly selects blobs to remove (which should work well
        # on a macro scale)

        # Part of a suggested intelligent implementation is to throw out
        # blocks that have a low 'mojo earned' flag first.  (Jim's idea)
        # [for a blockstores where most of the low earned values are the
        # same it should fall back to a more graceful behavior for
        # choosing among them]

        # we don't support cleaning up without a maximum amount of space specified
        if self.maxspace <= 0 :
            return

        # free more than specified, else others will need to call this again very soon.
        amount_to_free = minimumfree * 5
        # never make more than 1/4 of our space free at once, that's already pretty
        # devistating.  (if its less than minimumfree, oh well, you've got super tiny
        # blob store and shouldn't expect it to hold much)
        if amount_to_free > (self.maxspace/4) :
            amount_to_free = self.maxspace/4

        self.__cleanup_lock.acquire()  # prevent multiple cleanups from running at once
        try:
            # recompute the amount to free now that we've got the lock
            # (another thread could've freed some space for us)
            totalbytes = loads(self.index_db.get('total bytes used'))
            if self.maxspace - totalbytes < amount_to_free :
                amount_to_free = totalbytes - (self.maxspace - amount_to_free)
            else:
                # enough space is available, we're done
                return

            # choose which blobs to remove
            blobid_list = filter(idlib.is_sloppy_id, self.index_db.keys())
            blobs_to_remove = []
            bytes_removed = 0

            while blobid_list and bytes_removed < amount_to_free :
                blobid_idx = whrandom.randint(0, len(blobid_list)-1)
                blobid = blobid_list[blobid_idx]
                assert idlib.is_sloppy_id(blobid), "`blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
                blobid = idlib.to_mojosixbit(blobid)
                blobs_to_remove.append(blobid)
                bytes_removed = bytes_removed + self.get_block_size(blobid)
                del blobid_list[blobid_idx]

            # remove the selected blobs
            for blobid in blobs_to_remove :
                assert idlib.is_sloppy_id(blobid), "`blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
                blobid = idlib.to_mojosixbit(blobid)
                self.remove_blob(blobid)
        finally:
            self.__cleanup_lock.release()

    ####################################################################
    def get_free_disk_space(self):
        """
        Return the number of bytes free on the drive (Windows) or disk
        partition (Unix) where the blobstore resides.  The value returned
        is a long.

        Obviously, this code is all platform-specific.
        """
        if confutils.platform == 'win32':
            import win32file
            drive = os.path.splitdrive(self.basepath)[0]
            sectorsPerCluster, bytesPerSector, numFreeClusters, totalNumClusters = \
                     win32file.GetDiskFreeSpace(drive + "\\")
            return long(sectorsPerCluster) * long(numFreeClusters) * long(bytesPerSector)
        elif confutils.platform == 'linux':
            import statvfs
            fs = os.statvfs(self.basepath)
            return long(fs[statvfs.F_BAVAIL]) * long(fs[statvfs.F_BSIZE])
        elif confutils.platform == 'bsd':
            # The right way to do this is with the statfs() system call,
            # but I don't have time to write the C module, and it's not 
            # in the os module, at least under OpenBSD.  So hello to 
            # unmaintainable code.
            #
            # The df command always returns 0, so we have to parse the
            # first line to see if it failed. This is unreliable in
            # weird situations like I/O errors. 
            # Check the status in case of an OS error (cannot pfork, etc).
            import commands
            (status, output) = commands.getstatusoutput('/bin/df -P ' + self.basepath)
            if status != 0 or output.find('No such file or directory') != -1:
                raise CantRunDFError
            else:
                # Assume the df command worked.
                # Skip header line.
                output = output.split('\012')[1]
                # The third field of df output lines is 'Available'.
                numblocks = long(output.split()[3])
                # blocks are 512 bytes as was intended On High.
                return numblocks * 512
        #elif confutils.platform in 'irix': #not supported - no way to test
        else:
            debug.mojolog.write('ERROR get_free_disk_space() unsupported os: ' + sys.platform)
            raise PlatformNotSupportedError


    ####################################################################
    def __request_disk_space(self, bytes) :
        """
        Request disk space; this function returns a boolean if you have
        permission to use that amount space.  Automatic cleanup of old
        blobs may be triggered here depending upon our storage policy.
        """
        if bytes < 1 :
            raise ValueError, 'bytes must be positive'
        if self.maxspace is None:
            return 1  # there's always room in infinite blobstores

        if self.maxspace <= 0 :
            return 0  # there's never room in dummy blobstores

        try:
            totalbytes = loads(self.index_db.get('total bytes used'))
        except db.DBInvalidArgError, le:
            debug.mojolog.write("Got mysterious bsddb error: %s.  Ignoring.\n", args=(le,), vs="warning")
            # The following means we never refuse to use more disk space.  This db error needs to be fixed!
            totalbytes=self.maxspace + bytes
       
        if totalbytes + bytes <= self.maxspace :
            return 1
        else :
            # try and make room for this blob
            self.__cleanup_old_blobs(bytes)
            totalbytes = loads(self.index_db.get('total bytes used'))
            if totalbytes + bytes <= self.maxspace :
                return 1
            else:
                return 0

    ####################################################################
    def __create_index_entry(self, blobid, filepath, filesize, txn=None) :
        """
        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        # we store path as a relative pathname from self.basepath
        indexentry = {
            'path' : filepath[len(self.basepath)+len(os.sep):],
            'size' : filesize,
            'stored on' : time.strftime('%Y%m%d%H%M', time.gmtime(time.time())),
            'last access' : "Never",
            'num accesses' : 0,
        }
        self.index_db.put(blobid, dumps(indexentry), txn=txn)

    ####################################################################
    def __increment_access_count(self, blobid) :
        """
        @precondition `blobid' must be an blobid.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an blobid." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        self.db_env.nosyncerror_txn_checkpoint(MINS_BETWEEN_DB_CHECKPOINTS)
        txn = self.db_env.txn_begin()
        try:
            pickled_indexentry = self.index_db.get(blobid, txn=txn)
            if pickled_indexentry is None :
                txn.abort();  txn = None
                raise NoIndex, 'index entry not found'
            indexentry = loads(pickled_indexentry)
            indexentry['last access'] = time.strftime('%Y%m%d%H%M', time.gmtime(time.time()))  # this has a year 10,000 problem!
            indexentry['num accesses'] = indexentry['num accesses'] + 1
            self.index_db.put(blobid, dumps(indexentry), txn=txn)
        finally:
            if txn :
                txn.commit()


    ####################################################################
    def open_tempfile(self, blobid, size) :
        """
        Returns a file object for a temporary file in this block store.
        After the data has been written to the temporary file, it can
        be committed to the blockstore using store_tempfile().

        Raises a NoSpaceError if there is not enough space for the given
        expected size and we failed to make enough room.

        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        @precondition `size' must be a positive integer.: ((type(size) == types.IntType) or (type(size) == types.LongType)) and (size > 0): "size: %s" % hr(size)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        assert ((type(size) == types.IntType) or (type(size) == types.LongType)) and (size > 0), "precondition: `size' must be a positive integer." + " -- " + "size: %s" % hr(size)

        blobid = idlib.to_mojosixbit(blobid)

        if self.basepath is None :
            raise Hell

        if not self.__request_disk_space(size) :
            raise NoSpaceError, 'not enough free disk space for id'

        # create a unique filename within our temporary storage
        tempname = os.path.join(self.basepath, 'tmp')
        tempname = os.path.join(tempname, blobid+'.'+ mojosixbit.b2a(randsource.get(6)))
        
        # open the file
        return open(tempname, "wb")


    ####################################################################
    def store_tempfile(self, blobid, tempfile) :
        """store_tempfile(blobid, tempfile) - Store a file object created
        by open_tempfile.  tempfile must be closed.

        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        if self.basepath is None :
            raise Hell
        
        filepath = self.__generate_id_filename(blobid)
        filesize = os.path.getsize(tempfile.name)

        try:
            # move the tempfile into place (renames across
            # subdirectories on the same filesystem won't work
            # on many OSes)
            os.rename(tempfile.name, filepath)
        except OSError:
            # nope, rename didn't work, try doing it the hard way
            inf = open(tempfile.name, "rb")
            indata = inf.read()
            inf.close()
            os.unlink(tempfile.name)
            outf = open(filepath, "wb")
            outdata = outf.write(indata)
            outf.close()

        self.__create_index_entry(blobid, filepath, filesize)
        self.__update_space_totals(filesize)


    ####################################################################
    def get_block_size(self, blobid):
        """
        get_block_size(blobid) - return the size of the block if this
        blockstore contains block id, 0 otherwise.  This is a good
        method to test the existence of a block share id with.

        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        if self.basepath is None :
            raise Hell

        indexentry = self.index_db.get(blobid)
        if not indexentry :
            return 0
        indexentry = loads(indexentry)
        filename = os.path.join(self.basepath, indexentry['path'])

        if not os.path.isfile(filename) :
            return 0

        filesize = os.path.getsize(filename)
        if indexentry['size'] != filesize :
            debug.stderr.write("ERROR "+self.basepath+" localblockstore ("+self.name+"): "+blobid+" index entry 'size' doesn't match file\n")
            return 0

        return int(filesize)


    ####################################################################
    def get_block_file(self, blobid) :
        """
        get_block_file(blobid) - return an open file object for id

        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        if self.basepath is None :
            raise Hell

        indexentry = self.index_db.get(blobid)
        if not indexentry :
            raise NoBlock
        indexentry = loads(indexentry)

        filename = os.path.join(self.basepath, indexentry['path'])
        if not os.path.isfile(filename) :
            raise NoBlock

        return open(filename, "rb")


    ####################################################################
    def remove_blob(self, blobid) :
        """
        Remove a specific blob (does nothing if it doesn't exist)

        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        # remove the file (if it exists).  Trust the index for the
        # pathname and size computations first (as they are what was
        # used when the block was put in)
        indexentry = self.index_db.get(blobid)
        if indexentry :
            self.index_db.delete(blobid)
            indexentry = loads(indexentry)
            filename = os.path.join(self.basepath, indexentry['path'])
            sizeremoved = indexentry['size']
        else :
            filename = self.__generate_id_filename(blobid)
            sizeremoved = None
        if os.path.isfile(filename) :
            if not sizeremoved :
                sizeremoved = os.path.getsize(filename)
            os.unlink(filename)
            self.__update_space_totals(-sizeremoved)


    ####################################################################
    def space_available(self) :
        """Return some indication of how much space we have available"""
        if self.maxspace <= 0:
            return 0 # Dummy store.
        
        if self.maxspace is None:
            return 0x7fffffff  # return a large number if we have no space limit
        totalbytes = loads(self.index_db.get('total bytes used'))
        ###debug.mojolog.write("XXX self.maxspace: %s, totalbytes: %s\n", args=(self.maxspace, totalbytes,))
        return max(long(self.maxspace) - totalbytes, 0)

    ####################################################################
    def _rebuild_index(self, startdelay=20) :
        """
        This rebuilds the index database from the filesystem.  Stats on
        blobs not already in the database will be lost.
        (I wish this function weren't necessary...)
        """
        while startdelay > 0:
            startdelay = startdelay - 1
            self._shuttingdownflag.wait(1)  # non indefinite waits are easier to interrupt via signals, etc.
            if self._shuttingdownflag.isSet():
                break
        GROVEL_LOCK.acquire()
        try:
            debug.stderr.write('(%s) Verifying index.db integrity (recount=%s)\n' % (self.name, self.recount), vs='localblockstore')
            recountbytes = {'total': 0L} # this is not a hack. this is evil.
            total_ids = {'total': 0L}  # XXX hack for updating w/in subfunc
            rebuild_ids = []
            def index_directory_visit_func(basepath, dirname, names, self=self, rebuild_ids=rebuild_ids, total_ids=total_ids, recountbytes=recountbytes) :
                if self._shuttingdownflag.isSet():
                    raise ShutdownException
                for blobname in names :
                    if not idlib.is_mojosixbitencoded_id(blobname) :
                        continue
                    total_ids['total'] = total_ids['total'] + 1
                    blobsize = os.path.getsize(os.path.join(dirname, blobname))
                    if not blobsize :
                        continue
                    if not self.index_db.has_key(blobname) :
                        rebuild_ids.append(blobname)
                        self.__create_index_entry(blobid=blobname, filepath=os.path.join(dirname, blobname), filesize=blobsize)
                        if self.recount:
                            self.__update_space_totals(blobsize)
                            recountbytes['total'] = recountbytes['total'] + blobsize
                if self._shuttingdownflag.isSet():
                    raise ShutdownException
                time.sleep(0.2) # please don't hammer my system...

            try:
                if self._shuttingdownflag.isSet():
                    raise ShutdownException
                self._startedverifyingflag = true
                os.path.walk(self.basepath, index_directory_visit_func, self.basepath)
            except ShutdownException:
                # shutting down
                debug.mojolog.write("Index fix up interrupted by shutdown.\n", v=3, vs="localblockstore")
                return

            if len(rebuild_ids) > 0:
                debug.stderr.write('(%s) Rebuilt index entries for %s blocks (out of %s)\n', args=(self.name, len(rebuild_ids), total_ids['total']), v=2, vs='localblockstore')
                debug.stderr.write('(%s) blocks missing entries: %s\n', args=(self.name, rebuild_ids,), v=3, vs='localblockstore')
            else:
                debug.mojolog.write('(%s) no blocks missing\n', args=(self.name,), v=2, vs='localblockstore')
            if recountbytes['total'] > 0:
                totalbytes = loads(self.index_db.get('total bytes used'))
                debug.stderr.write('(%s) Added %s bytes to totalbytes (%s bytes) during rebuild\n', args=(self.name, recountbytes['total'], totalbytes), vs='localblockstore')
            try:
                if hasattr(self, 'index_db'):  # if we've shutdown, we don't have this anymore
                    self.index_db.sync()
            except db.DBIncompleteError:
                pass  # ignore a possible DB_INCOMPLETE
        finally:
            GROVEL_LOCK.release()
            self._doneverifyingflag.set()


class BLBlockStore(BlockStore):
    """
    A BlockStore with shitlists.  I think.
    """
    def __init__(self, path, maxspace=-1, depth=3, recoverdb=true, name='*unnamed*', rebuild=true) :
        """
        @param maxspace same as BlockStore; `-1' means this is a dummy.
        """
        BlockStore.__init__(self, path, maxspace, depth, recoverdb, name, rebuild=rebuild)
        self.blist = open(os.path.join(self.basepath, "blist.txt"), 'ac')
        try:
            self.bl_db = dbobj.DB(self.db_env)
            self.bl_db.open('bl.db', db.DB_BTREE, db.DB_CREATE)
        except (db.DBError, IOError), e:
            debug.mojolog.write("ERROR: (%s) ** Exception %s opening block server bl.db! **\n", args=(self.name, e), vs='localblockstore')
            self.bl_db = None

    def __del__(self) :
        if hasattr(self, 'bl_db') and self.bl_db is not None :
            self.bl_db.close()
            self.blist.close()
            del self.blist
            del self.bl_db
        BlockStore.__del__(self)

    def shutdown(self):
        # we -must- close this database before BlockStore.shutdown closes the underlying db_env.
        # (otherwise we can crash due to a bug in the version of pybsddb we're currently using - 2001-06-01)
        self.bl_db.close()
        del self.bl_db
        BlockStore.shutdown(self)

    def is_bl_blob_id(self, blobid):
        """
        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)

        blobid = idlib.to_mojosixbit(blobid)
        indexentry = self.bl_db.get(blobid)
        if not indexentry :
            return 0
        return 1

    def add_bl_blob_id(self, blobid, txn=None, flush=1):
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)

        # get rid of the blob if we have it
        size = self.get_block_size(blobid)
        if size:
            self.remove_blob(blobid)

        if not self.is_bl_blob_id(blobid):
            if blobid[-1] == '\n':        
                self.blist.write(blobid)
            else:
                self.blist.write(blobid + '\n')
            if flush:
                self.blist.flush()
            self.bl_db.put(blobid, '1', txn=txn)

    def add_bl_blob_list(self, bloblist):
        for blob in bloblist:
            self.add_bl_blob_id(blob, flush=0)
        self.blist.flush()
        
    def remove_bl_blob_id(self, blobid, txn=None):
        """
        If blobid is not present, this method returns silently.
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        blobid = idlib.to_mojosixbit(blobid)
        val = self.bl_db.get(blobid)
        if val:
            self.bl_db.delete(blobid)

    def get_block_size(self, blobid):
        if self.is_bl_blob_id(blobid):
            return 0
        return BlockStore.get_block_size(self, blobid)

    def get_block_file(self, blobid):
        if self.is_bl_blob_id(blobid):
            raise NoBlock
        return BlockStore.get_block_file(self, blobid)
    
    def store_tempfile(self, blobid, tempfile):
        if self.is_bl_blob_id(blobid):
            raise BLBlock
        return BlockStore.store_tempfile(self, blobid, tempfile)

    def _rebuild_bl_index(self):
        f = open(os.path.join(self.basepath, 'blist.txt'), 'r')
        l = f.readlines()
        f.close()
        for blobid in l:
            blobid = string.strip(blobid)
            assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
            blobid = idlib.to_mojosixbit(blobid)

            # get rid of the blob if we have it
            size = self.get_block_size(blobid)
            if size:
                self.remove_blob(blobid)
            # add id to the index
            self.bl_db.put(blobid, '1')
        self.bl_db.sync()

    def _rebuild_index(self):
        BlockStore._rebuild_index(self)
        if not self._shuttingdownflag.isSet():
            self._rebuild_bl_index()

########################################################################
class BadBlockListSubscriber:
    """Given a broker and a BLBlockStore, a BadBlockListSubscriber queries BadBlockListServers for bad blocks, periodically."""
    def __init__(self, broker, blblockstore, path=None, recoverdb=true):
        """
        broker - A Broker instance.
        blblockstore - Something that implements the BLBlockStore interface.
        path - Path to db directory.  (If None, it uses confman.dict['PATH']['BBLSubscriber_DIR'] setting.)
        recorverdb - If true, the DB_RECOVER flag is set for the db env.
        """
        self.broker = broker
        self.blbs = blblockstore

        self._shuttingdownflag = threading.Event()

        self.path = path
        if self.path is None:
            self.path = os.path.expandvars(self.broker.confman['PATH']['BBL_SUBSCRIBER_DIR'])

        recover_flag = 0
        if recoverdb :
            recover_flag = db.DB_RECOVER

        db_env_open_flags = db.DB_CREATE | db.DB_INIT_MPOOL | db.DB_PRIVATE | recover_flag
        try:
            self.db_env = CleanLogDbEnv()
            try:
                self.db_env.open(self.path, db_env_open_flags)
            except db.DBError, dbe:
                # XXX HACK sometimes trying again after one error works.
                debug.mojolog.write('Failed to open the database environment the first time, reason: %s\nTrying again...\n', args=(dbe,), vs='BadBlockListSubscriber', v=2)
                try:
                    self.db_env.open(self.path, db_env_open_flags | db.DB_RECOVER)
                except db.DBError, dbe:
                    debug.mojolog.write('Failed to open the database environment the second time, reason: %s\nTrying again...\n', args=(dbe,), vs='BadBlockListSubscriber', v=2)
                    # XXX DOUBLE CHOCOLATEY HACK sometimes trying *again* after one open *without* DB_RECOVER works.
                    self.db_env.open(self.path, db_env_open_flags & (~db.DB_RECOVER))

            # keyed on blobid, stores a pickled dict keyed on 'size', 'path', 'stored on', 'last access' and 'num accesses'
            self.bbls_db = dbobj.DB(self.db_env)
            self.bbls_db.open('subscriptions.db', db.DB_BTREE, db.DB_CREATE)
        except (db.DBError, IOError), e:  # XXX
            debug.mojolog.write("WARNING: (%s) ** Exception %s opening subscriptions.db! **\n", args=(self.name, e), vs='BBLSubscriber')
            # clean up any half-opened things (let their destructors take care of them)
            self.bbls_db = None
            self.db_env = None

            # Remove all of the DB log files, for some ?%$&@E!& reason we've seen them be corrupted and
            # cause an IO Error when attempting to open the database.
            # (this does nothing at the moment since this db doesn't need transactions or logging -greg)
            dblogfiles = filter(lambda n: string.lower(n[:len('log')]) == 'log', os.listdir(self.basepath))
            map(lambda n, bp=self.basepath: os.remove(os.path.join(bp, n)), dblogfiles)

            self.db_env = CleanLogDbEnv()
            self.db_env.open(self.basepath, db_env_open_flags)
            self.bbls_db = dbobj.DB(self.db_env)
            self.bbls_db.open('subscriptions.db', db.DB_BTREE, db.DB_CREATE)

        # if contact info for a particular BBL Subscription isn't available, don't try looking them up or
        # contacting them for another few hours.
        self.broken_subscription_id_cache = Cache.StatsCache(autoexpireinterval=300, autoexpireparams={'maxage': 3*60*60 })

        self.broker.doq.add_task(self.task_bbl_subscription, delay=288)

    def __del__(self):
        if hasattr(self, 'bbls_db') and self.bbls_db is not None:
            self.bbls_db.close()
            del self.bbls_db
        if hasattr(self, 'db_env') and self.db_env is not None:
            self.db_env.close()
            del self.db_env

    def shutdown(self):
        self._shuttingdownflag.set()
        self.bbls_db.close()
        self.bbls_db = None
        self.db_env.close()
        self.db_env = None
       
    def task_bbl_subscription(self, delay=1260):  # update the list roughly every twenty one minutes
        if self._shuttingdownflag.isSet():
            return

        debug.mojolog.write("task_bbl_subscription()\n", v=4, vs='BBLSubscriber')
        subscriptionlist = {}
        for bblserverid in self.broker.confman.dict.get('BadBlockSubscriptionList', {}).values():
            if not self.broken_subscription_id_cache.has_key(idlib.to_binary(bblserverid)):
                subscriptionlist[idlib.to_binary(bblserverid)] = None

        cursor = self.bbls_db.cursor()
        kvpair = cursor.first()
        while kvpair:
            bblserverid, bbldata = kvpair
            if subscriptionlist.has_key(bblserverid):
                del subscriptionlist[bblserverid]
                self.update_subscription(bblserverid, loads(bbldata)) # update existing subscriptions
            kvpair = cursor.next()
        cursor.close()  # be kind to bsddb3 cursor bugs
        del cursor

        map(self.update_subscription, subscriptionlist.keys()) # Handle new subscriptions.

        delay = delay + ((0.5 - whrandom.random())*delay/9.0)  # +/- a bit each time
        self.broker.doq.add_task(self.task_bbl_subscription, (delay, ), delay=delay) # Repeat this task.

    def update_subscription(self, bblserverid, subinfo={}):
        message = {}
        if subinfo.has_key('nonce'):
            message['nonce'] = subinfo['nonce']
        if subinfo.has_key('end'):
            # Ask for the entries beginning with where we left off (aka 'end'):
            message['start'] = subinfo['end']
        notes = {
            'message': message,
            'bblserverid': bblserverid,
        }
        MetaTrackerLib.nonblocking_get_contact_info(
            self.broker.ServicesMTM,
            bblserverid,
            callback=self.__update_subscription_metainfo_callback,
            notes=notes)
    
    def __update_subscription_metainfo_callback(self, widget, outcome, failure_reason=None, notes=None):
        message = notes['message']
        bblserverid = notes['bblserverid']
        metainfo = outcome
        if metainfo is None or metainfo.get('result') == 'failure' or failure_reason is not None:
            debug.mojolog.write("ignoring subscription to %s: unknown contact info\n" % idlib.to_ascii(bblserverid), v=3, vs='BBLSubscriber')
            # put bblserverid on a temporary shitlist so that we wait awhile before looking them up again
            self.broken_subscription_id_cache[idlib.to_binary(bblserverid)] = bblserverid
        else:
            mojooffer = BidStrategy.choose_bid("MO_GET_BAD_BLOCK_LIST", self)
            self.broker.ServicesMTM.initiate(
                bblserverid,
                'get bad block list',
                message,
                mojooffer,
                outcome_func=self.handle_response__get_bad_block_list,
                notes=message)

    def handle_response__get_bad_block_list(self, widget, outcome, failure_reason, notes):
        bblserverid = widget.get_counterparty_id()
        if failure_reason:
            self.broker.mojolog.write('Failure when receiving bad block list from %s: %s\n', args=(bblserverid, failure_reason,), vs='BBLSubscriber', v=3)
            return

        bblist = outcome['bad block list']
        if len(bblist) > 0:
            self.broker.mojolog.write('Discovered %s new bad blocks from subscription to %s.\n', args=(len(bblist), bblserverid,), vs='BBLSubscriber', v=3)
            self.blbs.add_bl_blob_list(bblist) # Add the new bad blobs.

        # Update our subscription info:
        dbentry = {
            'nonce': outcome['nonce'],
            'start': int(outcome['start']),
            }
        
        if dbentry['nonce'] == notes.get('nonce'):
            dbentry['end'] = max(int(outcome['end']), notes.get('end', -1))
        else:
            dbentry['end'] = int(outcome['end'])

        debug.mojolog.write("got bbl dbentry %s\n", args=(dbentry,), v=3, vs='BBLSubscriber')
        self.bbls_db.put(bblserverid, dumps(dbentry)) # Update our subscriber info.
        self.sync()

    def create_new_subscription(self, bblserverid):
        self.bbls_db.put(bblserverid, dumps({}))
        self.sync()
    
    def sync(self):
        try:
            self.bbls_db.sync()
        except db.DBIncompleteError:
            pass
    

########################################################################
class _ReadOnlyBSOverride:
    """
    Create a BlockStore object derived from your BlockStore and this
    (first) so that multiple inheritance causes this to override the
    open_tempfile, store_tempfile and space_available methods.
    """
    def open_tempfile(self, blobid, size):
        raise NoSpaceError, "can't add to a read-only BlockStore"
    def store_tempfile(self, blobid, tempfile):
        raise NoSpaceError, "can't add to a read-only BlockStore"
    def space_available(self):
        return 0L


class ReadOnlyBLBlockStore(_ReadOnlyBSOverride, BLBlockStore):
    """A read only BLBlockStore"""
    pass


########################################################################
class MultiBlockStore:
    """
    This implements a block store that uses multiple BlockStore
    objects underneath to store the data across multiple
    disks/directories.
    """
    def __init__(self, blockstorelist):
        assert len(blockstorelist) > 0, "no BlockStores specified"
        for bs in blockstorelist:
            assert isinstance(bs, BlockStore), ("BlockStore object(s) required, got:", bs)

        self._blockstores = []
        self._blockstores.extend(blockstorelist)

        self._tempfilemap = {}

    def __del__(self) :
        if hasattr(self, '_blockstores'):
            del self._blockstores

    #### basic BlockStore interface ####
    def shutdown(self):
        for bs in self._blockstores:
            bs.shutdown()
        self._blockstores = []
    
    def wait_until_done_verifying(self):
        for bs in self._blockstores:
            bs.wait_until_done_verifying()

    def init_space(self, path=None):
        if path is not None:
            raise Hell, "init_space() cannot be given a path in a MultiBlockStore"
        for bs in self._blockstores:
            bs.init_space()
    
    def open_tempfile(self, blobid, size):
        # if we already have this block, don't bother storing it again
        if self.get_block_size(blobid) > 0:
            self._tempfilemap[blobid] = None  # tells store_tempfile not to actually store it
            return StringIO()

        # choose a block store with size bytes of free disk space and store this file.
        # if there are non with free space, choose the store that this file fits into using
        # the first few bits of its id.
        for bs in mojoutil.rotatelist(self._blockstores):
            if bs.space_available() > size:
                tempfile = bs.open_tempfile(blobid, size)
                break
        else:
            # no BlockStores had free space, randomly pick one to store it in
            bs = whrandom.choice(self._blockstores)
            tempfile = bs.open_tempfile(blobid, size)
        # keep a reference to which bs we opened the tempfile in so we know where to store it when store_tempfile is called.
        self._tempfilemap[blobid] = bs

        return tempfile
    
    def store_tempfile(self, blobid, tempfile):
        if not self._tempfilemap.has_key(blobid):
            raise Hell, ("store_tempfile called on a blobid before open_tempfile", blobid, tempfile)
        bs = self._tempfilemap[blobid]
        if bs is not None:
            return bs.store_tempfile(blobid, tempfile)

    def get_reserved_disk_space(self):
        """
        Returns a long representing the total amount of disk space
        reserved for Mojo Nation (on all blockstores).
        """
        reslist = map(lambda bs: bs.maxspace, self._blockstores)
        return reduce(lambda x, y: long(x)+long(y), reslist)

    
    def get_free_disk_space(self):
        """
        Returns a dict mapping the paths of the blockstores 
        to the number of bytes of free space on those drives
        or disk partitions.
        """
        d = {}
        for bs in self._blockstores:
            path = bs.basepath
            d[path] = bs.get_free_disk_space()
        return d
    
    def get_block_size(self, blobid):
        for bs in self._blockstores:
            size = bs.get_block_size(blobid)
            if size:
                return size
        return 0
    
    def get_block_file(self, blobid):
        for bs in self._blockstores:
            size = bs.get_block_size(blobid)
            if size:
                return bs.get_block_file(blobid)
        raise NoBlock, ("no BlockStore contained this blob", blobid)

    def remove_blob(self, blobid):
        for bs in self._blockstores:
            bs.remove_blob(blobid)
    
    def space_available(self):
        spacelist = map(lambda bs: bs.space_available(), self._blockstores)
        return reduce(lambda a, b: long(a)+long(b), spacelist)
    
    #### BLBlockStore interface ####
    def is_bl_blob_id(self, blobid):
        for bs in self._blockstores:
            if isinstance(bs, BLBlockStore):
                if bs.is_bl_blob_id(blobid):
                    return 1
        return None

    def add_bl_blob_id(self, blobid):
        for bs in self._blockstores:
            if isinstance(bs, BLBlockStore):
                bs.add_bl_blob_id(blobid)

    def add_bl_blob_list(self, bloblist):
        for bs in self._blockstores:
            if isinstance(bs, BLBlockStore):
                bs.add_bl_blob_list(bloblist)

    def remove_bl_blob_id(self, blobid):
        for bs in self._blockstores:
            if isinstance(bs, BLBlockStore):
                bs.remove_bl_blob_id(blobid)
        


########################################################################
def test_depth_0_no_max() :
    id1 = mojosixbit.b2a(randsource.get(20))
    id2 = mojosixbit.b2a(randsource.get(20))
    bs = BlockStore(RunTests.get_next_dir_name(), depth=0, rebuild=false)

    tf = bs.open_tempfile(id1, 100)
    tf.write('x'*100)
    tf.close()
    bs.store_tempfile(id1, tf)
    assert bs.get_block_size(id1) == 100

    tf = bs.open_tempfile(id2, 200)
    tf.write('y'*200)
    tf.close()
    bs.store_tempfile(id2, tf)
    assert bs.get_block_size(id1) == 100
    assert bs.get_block_size(id2) == 200

    bs.remove_blob(id1)
    assert bs.get_block_size(id1) == 0
    assert bs.get_block_size(id2) == 200

def test_maxspace_and_depth() :
    id1 = mojosixbit.b2a(randsource.get(20))
    id2 = mojosixbit.b2a(randsource.get(20))
    id3 = mojosixbit.b2a(randsource.get(20))
    bs = BlockStore(RunTests.get_next_dir_name(), maxspace=500, depth=3, rebuild=false)
    assert bs.space_available() == 500

    idlist = []
    for x in xrange(0,50) :
        id = mojosixbit.b2a(randsource.get(20))
        tf = bs.open_tempfile(id, 10)
        tf.write('s'*10)
        tf.close()
        bs.store_tempfile(id, tf)
        idlist.append(id)
    assert bs.space_available() == 0
    
    for id in idlist :
        assert bs.get_block_size(id) == 10, 'block missing that should have existed'

    id = mojosixbit.b2a(randsource.get(20))
    tf = bs.open_tempfile(id, 100)
    tf.write('e'*100)
    tf.close()
    bs.store_tempfile(id, tf)
    assert bs.get_block_size(id) == 100

    remaining_idlist = filter(bs.get_block_size, idlist)
    assert len(remaining_idlist) < len(idlist)-10, "not enough old blobs were discarded"
    assert len(remaining_idlist) > len(idlist)*(3/4), "too many old blobs were dicarded"
    assert bs.space_available() > 0, "no extra space leftover after the cleanup"

def test_depth_limits() :
    try:
        bs=BlockStore(RunTests.get_next_dir_name(), depth=-1)
        bs=BlockStore(RunTests.get_next_dir_name(), depth=50)
        assert 0, 'BlockStore allowed bad depth'
    except ValueError:
        pass

def test_get_file() :
    id = mojosixbit.b2a(randsource.get(20))
    bs=BlockStore(RunTests.get_next_dir_name(), depth=1, rebuild=false)
    tf = bs.open_tempfile(id, 50)
    stuff = randsource.get(50)
    tf.write(stuff)
    tf.close()
    bs.store_tempfile(id, tf)
    assert bs.get_block_size(id) == 50
    bf = bs.get_block_file(id)
    assert bf.read() == stuff

def test_read_only():
    ro_bs = ReadOnlyBLBlockStore(RunTests.get_next_dir_name(), rebuild=false)
    assert ro_bs.space_available() == 0
    try:
        ro_bs.open_tempfile('a'*27, 31415)
        assert 0, "read only blockstore allowed open_tempfile call"
    except NoSpaceError:
        pass

def test_multi_blockstore():
    bslist = []
    bslist.append( BLBlockStore(RunTests.get_next_dir_name(), depth=2, maxspace=100000, rebuild=false) )
    bslist.append( ReadOnlyBLBlockStore(RunTests.get_next_dir_name(), rebuild=false) )
    bslist.append( BlockStore(RunTests.get_next_dir_name(), depth=1, maxspace=200000, rebuild=false) )
    bs = MultiBlockStore(bslist)
    assert bs.space_available() == 300000, "bad space total"
    for c in range(1,111):
        id = idlib.to_ascii(chr(c)*20)
        tmpf = bs.open_tempfile(id, c*len(id))
        tmpf.write(id*c)
        tmpf.close()
        bs.store_tempfile(id, tmpf)
    for c in range(1,111):
        id = idlib.to_ascii(chr(c)*20)
        assert bs.get_block_size(id) == (len(id)*c)
        bs.remove_blob(id)
    assert bs.space_available() == 300000, "not all space freed after deleting every inserted block"

def test_multi_blockstore_full():
    bslist = []
    bslist.append( BlockStore(RunTests.get_next_dir_name(), depth=2, maxspace=3000, rebuild=false) )
    bslist.append( BlockStore(RunTests.get_next_dir_name(), depth=1, maxspace=2800, rebuild=false) )
    bs = MultiBlockStore(bslist)
    # this will store way more than the max space available, it should not fail
    for c in range(1,80):
        id = idlib.to_ascii(chr(whrandom.randint(0,255))*20)
        tmpf = bs.open_tempfile(id, 350*(c%2+1))
        tmpf.write('\0'*350*(c%2+1))
        tmpf.close()
        bs.store_tempfile(id, tmpf)


mojo_test_flag = 1
import RunTests
def run():
    RunTests.runTests('localblockstore')

if __name__ == '__main__' :
    run()
