#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#




def is_mature(candidate) :
    """
    Returns whether this object has been declared mature
    """
    if candidate is None :
        return 1
    return hasattr(candidate, '_mature')

def is_immutable(candidate) :
    """
    Return whether this object is immutable and has been declared mature
    """
    if candidate is None :
        return 1
    return hasattr(candidate, '_immutable')

def make_mature(candidate) :
    """
    Objects are marked mature once their construction is complete, 
    usually (but not always) at the end of their constructor
    
    It's advisable to say assert make_mature(spam), much like
    it's advisable to say assert is_mature(spam)
    """
    assert hasattr(candidate, 'verify_mature'), 'to be made mature, an object must have a verify_mature() method, even if all it does is pass'
    candidate.verify_mature()
    candidate.__dict__['_mature'] = None
    if hasattr(candidate, 'notify_mature') :
        candidate.notify_mature()

class Immutable :
    """
    Immutable objects can't be altered once they are declared mature
    """
    def __setattr__(self, attr, value) :
        assert not hasattr(self, '_mature'), "Immutable objects can't be modified after maturity"
        self.__dict__[attr] = value
        
    def notify_mature(self) :
        self.__dict__['_immutable'] = None

class ForTestingBad :
    pass

class ForTesting :
    def __init__(self, proper) :
        if proper :
            self.spam = None
    
    def verify_mature(self) :
        assert hasattr(self, 'spam')

class ForTestingImmutable(Immutable) :
    def __init__(self) :
        self.spam = None
    
    def verify_mature(self) :
        assert hasattr(self, 'spam')

mojo_test_flag = 1

def test_maturity() :
    spam = ForTesting(1)
    assert not is_mature(spam)
    make_mature(spam)
    assert is_mature(spam)

def test_maturity_assertion() :
    try :
        spam = ForTesting(0)
        make_mature(spam)
    except AssertionError :
        pass
    else :
        assert 0

def test_invalid() :
    try :
        spam = ForTestingBad()
        make_mature(spam)
    except AssertionError :
        pass
    else :
        assert 0

def test_immutable_can_be_altered_before_maturity() :
    spam = ForTestingImmutable()
    assert not is_mature(spam)
    assert not is_immutable(spam)
    spam.eggs = 'hard boiled'

def test_immutable_cannot_be_altered_after_maturity() :
    spam = ForTestingImmutable()
    make_mature(spam)
    assert is_mature(spam)
    assert is_immutable(spam)
    try :
        spam.eggs = 'hard boiled'
    except AssertionError :
        pass
    else :
        assert 0

class ImmutableLink(Immutable) :
    def __init__(self, next_link) :
        """
        Normally this would make itself mature at the end of the constructor, 
        but we don't want that for testing
        """
        self.next = next_link

    def verify_mature(self) :
        assert hasattr(self, 'next')
        assert is_mature(self.next)

def test_make_immutable_chain() :
    l1 = ImmutableLink(None)
    make_mature(l1)
    l2 = ImmutableLink(l1)
    make_mature(l2)
    assert is_mature(l2)
    assert is_immutable(l2)

def test_make_immutable_chain_fail() :
    l1 = ImmutableLink(None)
    l2 = ImmutableLink(l1)
    try :
        make_mature(l2)
    except AssertionError :
        pass
    else :
        assert 0

