#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#


### standard modules
import asyncore
import threading
import time
import traceback
import socket
import select
import sys
import whrandom

### our modules
import MojoConstants
from MojoConstants import true, false
import debug
import mojoutil
from confutils import confman

# This is a list of sockets that we should close:
TCPConnections_to_close = []

class Asyncore:
    def __init__(self, doq):
        self._stopflag = false
        self._donestopping = threading.Event()
        self._waitingtostart = threading.Event()
        self._doq = doq
        self._tcpcommshandler_list = []  # XXX HACK MojoTransaction adds these to the list
        self._select_trigger = None

    def _asyncoreloop(self):
        mojoutil.profit(self._asyncoreloop_real)

    def _asyncoreloop_real(self):
        try:
            debug.mojolog.write("The Asyncore Poller thread started.\n")
            while not self._stopflag:
                if not asyncore.socket_map:
                    # poll() returns immediately when the socket map is empty
                    # rather than sleeping as we'd like so we sleep here.
                    self._waitingtostart.wait(1)
                try:
                    asyncore.poll(timeout=10)
                except select.error, e:
                    if e.args[0] == 0:
                        debug.mojolog.write("Ignoring select.error 0\n", v=3, vs='mojoasyncore')
                        time.sleep(1)
                    else:
                        debug.mojolog.write("Caught select.error while calling asyncore.poll()\n", v=0, vs="ERROR")
                        traceback.print_exc(file=debug.mojolog)
                        time.sleep(1)
                except ValueError, e:
                    # XXX don't touch asyncore.socket_map, different versions of asyncore use it differently (but len() is safe)
                    debug.mojolog.write("asyncore.poll(): %s\nasyncore.socket_map length is %s.  Forcing socket cleanup\n", args=(e, len(asyncore.socket_map)), v=2, vs='mojoasyncore')
                    # It's probably a too many file descriptors in select bug if we're on windows.
                    # force all TCPCommsHandlers to clean up their connections ASAP!
                    for TCPch in self._tcpcommshandler_list:
                        self._doq.add_task(TCPch._cid_to_tcpc._cleanup, args=[int(confman.get('TCP_MAX_CONNECTIONS', 50))-1], delay=-5)  # put this on the DoQ to be done ASAP
                    # don't pound the CPU and wait in hope that some of the above has finished
                    time.sleep(10)
                except:
                    debug.mojolog.write("Caught this exception while calling asyncore.poll()\n", v=0, vs="ERROR")
                    traceback.print_exc(file=debug.mojolog)
                    time.sleep(1) # don't want to pound the CPU and fill up the debug log on some repetitive error

                while TCPConnections_to_close:
                    try:
                        TCPConnections_to_close.pop(0).close()
                    except:
                        debug.mojolog.write("Caught this exception while calling TCPConnection.close()\n", v=0, vs="ERROR")
                        traceback.print_exc(file=debug.mojolog)

            # TODO ? After stopping, close all sockets.

        finally:
            self._donestopping.set()
            debug.stderr.write("asyncore poller thread exiting.\n")

    def start_asyncore(self):
        self._stopflag = false
        self._donestopping.clear()
        self._select_trigger_server = SelectTrigger()
        debug.mojolog.write("Asyncore.start_asyncore(): about to create and start The Asyncore Poller thread.\n", v=9, vs="debug")
        if MojoConstants.DEBUG_MODE:
            t = threading.Thread(target=self._asyncoreloop, name="The Asyncore Poller" + `traceback.extract_stack()`) # verbose -- for debugging
        else:
            t = threading.Thread(target=self._asyncoreloop, name="The Asyncore Poller")
        t.setDaemon(1)
        t.start()
        debug.mojolog.write("Asyncore.start_asyncore(): done creating and starting The Asyncore Poller thread.\n", v=9, vs="debug")
        if self._select_trigger_server.port:
            try:
                self._select_trigger = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)   # UDP socket for simplicity
                self._select_trigger.connect( ("127.0.0.1", self._select_trigger_server.port) )   # prevent the need for sendto() calls
            except socket.error, e:
                debug.mojolog.write("Got socket.error %s trying to connect SelectTrigger\n", args=(e,), vs="mojoasyncore")
                self._select_trigger = None

    def stop_asyncore(self):
        """
        Blocks briefly until asyncore is done stopping.
        """
        self._stopflag = true
        self._waitingtostart.set()
        self._donestopping.wait()
        if self._select_trigger:
            self._select_trigger.close()
        self._tcpcommshandler_list = []
        self._doq = None

    def wake_select(self):
        """
        Wake up a sleeping asyncore.poll() call if possible.  This allows acceptable client behavior
        from asyncore by not delaying data sending until asyncore.poll() wakes up and makes updated
        readable/writable calls on a socket with data freshly queued up to be sent.
        """
        if self._select_trigger:
            try:
                self._select_trigger.send('.')
            except socket.error, e:
                debug.mojolog.write("socket.error %s sending on _select_trigger, closing.\n", args=(e,), vs="mojoasyncore")
                self._select_trigger = None


class SelectTrigger(asyncore.dispatcher):
    def __init__(self):
        # do this last
        self._ready = 0
        asyncore.dispatcher.__init__(self)

        self.port = None

        self.create_socket(socket.AF_INET, socket.SOCK_DGRAM)   # UDP socket for simplicity
        self.set_reuse_addr()
        for x in xrange(20):   # try binding to up to 20 different random ports
            try:
                port = whrandom.randint(1025, 32767)
                self.bind( ("127.0.0.1", port) )
                debug.mojolog.write("SelectTrigger running on (127.0.0.1, %s) [UDP]\n", args=(port,), v=4, vs="mojoasyncore")
                break
            except socket.error:
                debug.mojolog.write("SelectTrigger couldn't bind to UDP port %s, trying another\n", args=(port,), v=5, vs="mojoasyncore")
                pass
        else:
            debug.mojolog.write("SelectTrigger giving up after failing to bind to 20 random UDP ports!\n", v=2, vs="mojoasyncore")
            return

        self.port = port
        self._ready = 1

    def handle_connect(self):
        pass

    def handle_accept(self):
        # the SelectTrigger is a special socket, we only accept a single connection which should stay open as long as we're running
        (sock, addr) = self.accept()
        self.close()

    def readable(self):
        return self._ready

    def handle_read(self):
        # eat the trigger data that someone sent us to wake asyncore.poll() up.  yum yum.
        self.recv(8192)

    def handle_close(self):
        debug.mojolog.write("mojoasyncore.SelectTrigger socket closing\n", vs="mojoasyncore")
        self.close()

    def writable(self):
        return 0

    def handle_write(self):
        debug.mojolog.write("SelectTrigger.handle_write should never get called\n", vs="mojoasyncore")

    def log(self, message):
        pass

