#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# This reads a piece of a config file from our HTTP server.  This is
# known as a "boot page" which tells the users software about the
# root meta tracker contact info and what the latest version of the
# mojo nation software is (to let them know about upgrades and in the
# future to support a "click here to auto-upgrade")
#
# In the future we also hope to have the boot page be a signed piece
# of data, signed by a hardcoded key.
#
__cvsid = '$Id: mojobootpage.py,v 1.22 2001/06/08 00:50:42 greg Exp $'

# standard modules
import copy
import traceback
import threading
import time
import asyncore
import socket
import string
from StringIO import StringIO
import re
import types
import urlparse

# our modules
import Buf
import DoQ
from MojoConstants import true, false
import SEXP
from VersionNumber import VersionNumber
import confutils
import debug
import idmasklib
import idlib
import mojoutil

MAX_BOOTPAGE_BYTES = 64001  # happy fun limit


# after a bootpage is loaded, this global will be updated to reflect the bootpage
# use the get_bootpage_version() call to really get this.
_ReportedVersion = VersionNumber(confutils.confman.dict["BROKER_VERSION_STR"])

def get_bootpage_version():
    global _ReportedVersion
    return _ReportedVersion

mojoutil.setdefault(confutils.confman, "BOOTPAGE_URL", "http://www.mojonation.net:25000/bootpage.txt")

class _simple_bootpage_loader(asyncore.dispatcher):
    def __init__(self, doq, url=None, callback=None):

        self.doq = doq
        self.callback = callback

        if not url:
            url = confutils.confman.get('BOOTPAGE_URL', 'http://www.mojonation.net:25000/bootpage.txt')

        debug.mojolog.write('bootpage url: %s\n' % url, vs='mojobootpage')
        if url[:5] == 'file:': # A local filesystem file hack.
            bpf = open(url[5:], 'r')
            self.data = bpf.read()
            bpf.close()
            if self.callback:
                self.callback(self.data)
            return
            
        _http_, host_and_port, path = urlparse.urlparse(url)[:3]
        if string.find(host_and_port, ':') == -1:
            host = host_and_port
            port = 80
        else:
            host, port = string.split(host_and_port, ':', 2)
            port = int(port)

        self._ignoring_errs = false

        self.path = path
        # this is an implicit HTTP/0.9 connection (so no headers or status will be returned)
        self.buffer = 'GET %s\r\n\r\n' % self.path
        self.data = ''
        self.connected = None

        # do this last after all other initialization
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connect( (host, port) )

    def handle_connect(self):
        self.connected = 1

    def handle_read(self):
        try:
            newdata = self.recv(8192)
        except socket.error, e:
            debug.mojolog.write('got %s in handle_read\n', args=(e,), v=3, vs='mojobootpage')
            self.handle_close()
            return
        self.data = self.data + newdata
        if len(newdata) == 0 or len(self.data) > MAX_BOOTPAGE_BYTES:
            # a 0 byte read is another way we find about closed sockets sometimes
            self.handle_close()

    def writeable(self):
        return self.connected and (len(self.buffer) > 0)

    def handle_write(self):
        if (len(self.buffer) <= 0):
            return
        try:
            sent = self.send(self.buffer)
            self._ignoring_errs = false
        except socket.error, e:
            if self._ignoring_errs:
                # don't want to fill up the log...
                pass
            else:
                debug.mojolog.write('ignoring %s in handle_write\n', args=(e,), v=3, vs='mojobootpage')
                self._ignoring_errs = true

            sent = 0
        self.buffer = self.buffer[sent:]

    def handle_close(self):
        if self.callback:
            self.doq.add_task(self.callback, args=(self.data,))
            self.callback = None  # only do the callback once
        self.close()

    def log(self, message):
        pass


def load_and_use(doQ, block=None, timeout=None, textcallback=lambda txt : None):
    """
    This is a non-blocking call.  It achieves non-blockingness using the DoQ.
    (unless you set block=1; its important to block on startup because we don't
    want to get stuck with outdated root meta tracker contact info)
    If you set block=1, timeout is the time in seconds it'll wait for it to finish.

    If the textcallback is callable, it is called with the bootpage contents once
    retrieved.
    """
    if confutils.confman.is_true_bool(['IGNORE_BOOTPAGE']):
        debug.mojolog.write(
            'Configured to ignore the bootpage; no point in fetching it...\n',
            vs='mojobootpage'
            )
        return

    # this code seems silly, all it does it use the DoQ to spawn a thread which immediately dissociates and blocks
    # but whatever, its easy to read and it works.
    wakeupevent = threading.Event()
    def got_it(bootpagedata, doQ=doQ, wakeupevent=wakeupevent, textcallback=textcallback):
        textcallback(bootpagedata)
        b = BootPage(bootpagedata)
        b.use(confmanobj=confutils.confman, doQ=doQ)
        wakeupevent.set()

    debug.mojolog.write("loading mojonation bootpage\n", vs='mojobootpage')
    try:
        _simple_bootpage_loader(doq=doQ, callback=got_it)
    except socket.error, e:
        debug.mojolog.write('socket error %s ignored\n', args=(e,), vs='mojobootpage')
        pass
    if block:
        wakeupevent.wait(timeout)



class BootPage :
    """
    BootPage(data) - processes the bootpage data given into an internal
            dictionary representation of its contents.  If an error
            occurs during processing, it will be logged and the bootpage
            will be treated as empty.
    BootPage.use(confmanobj) - overrides fields in confmanobj with the known
            safe ones from the bootpage, also checks the local vs bootpage
            software version.
    """
    def __init__(self, data) :
        self.bpdict = {}

        try:
            bootpagelines = StringIO(data).readlines()
            self.bpdict = confutils.lines_to_dict(bootpagelines)
            debug.mojolog.write('bootpage loaded.\n')
        except:
            debug.mojolog.write('WARNING: Ignoring exception while loading bootpage:\n')
            traceback.print_exc(file=debug.mojolog)

    def use(self, confmanobj, doQ=None) :
        """
        Overwrite any reasonable things in confmanobj with those from the boot page.
        If a doQ is given, the confman.save() call will be placed on it instead of done here.
        """
        try:
            changed = None
            bpversion = self.bpdict.get("BROKER_VERSION_STR")

            # Tell the user about a new spiffy cool version of our software
            if bpversion :
                global _ReportedVersion
                _ReportedVersion = VersionNumber(bpversion)
                if VersionNumber(confmanobj.dict["BROKER_VERSION_STR"]) < VersionNumber(bpversion) :
                    debug.mojolog.write("!!!! UPGRADE NOTICE: Mr. Bootpage says that the latest Mojo Nation verion\n")
                    debug.mojolog.write("!!!! is '%s' while you are running version '%s'.\n" % (bpversion, confmanobj.dict["BROKER_VERSION_STR"]))
                    debug.mojolog.write("!!!! Visit http://www.mojonation.net/ for an upgrade.\n")

            # use the bootpage's root meta tracker
            bprootinfo = self.bpdict.get("ROOT_ID_TRACKER_CONTACT_INFO")
            if bprootinfo and type(bprootinfo) == type({}) :
                debug.mojolog.write("Updating root meta tracker contact info.\n", vs='mojobootpage')
                confmanobj.dict["ROOT_ID_TRACKER_CONTACT_INFO"] = copy.copy(bprootinfo)
                changed = 1

            # use the bootpage's root meta tracker
            bprootinfo = self.bpdict.get("MULTI_ROOT_ID_TRACKER_CONTACT_INFO")
            debug.mojolog.write("MULTI_ROOT_ID_TRACKER_CONTACT_INFO = %s" % bprootinfo, v=13)
            # XXX this regex is a hack to make sure its safe to eval()
            safe_metainfo_list_re = re.compile(r'^\[[-_a-z0-9,:{}\s\.\[\]\'\"]+\n?$', re.I)
            if bprootinfo and type(bprootinfo) == type(""):
                if safe_metainfo_list_re.match(bprootinfo):
                    debug.mojolog.write("Updating root meta tracker(s) contact info.\n", vs='mojobootpage')
                    bprootlist = eval(bprootinfo)

                    try:
                        for metainfo in bprootlist:
                            def checkcs(connstrat):
                                lowerstrat = connstrat['lowerstrategy']
                                dbstr = lowerstrat['comm strategy type']
                                if dbstr == 'TCP':
                                    dbstr = dbstr + ' %s:%s' % (
                                        lowerstrat['IP address'],
                                        lowerstrat['port number']
                                        )
                                debug.mojolog.write(
                                    'Found metatracker connection info: %s\n' % dbstr,
                                    vs='mojobootpage'
                                    )
                            if type(metainfo['connection strategies']) == type({}):
                                metainfo['connection strategies'] = SEXP.dictToList(metainfo['connection strategies'])
                            map(checkcs, metainfo['connection strategies'])
                    except KeyError:
                        debug.mojolog.write('Strange bootpage format: %s\n' % `bprootlist`, vs='mojobootpage')
                        traceback.print_exc(file=debug.mojolog)

                    confmanobj.dict["MULTI_ROOT_ID_TRACKER_CONTACT_INFO"] = bprootlist
                    changed = 1
                else:
                    debug.mojolog.write("new MULTI_ROOT_ID_TRACKER_CONTACT_INFO ignored, failed the regex: %s\n" % `bprootinfo`, vs='mojobootpage')

            bpbblservers = self.bpdict.get("BadBlockSubscriptionList")
            if bpbblservers and type(bpbblservers) in [type([]), type({})]:
                if type(bpbblservers) == type({}):
                    bpbblservers = SEXP.dictToList(bpbblservers)
                conf_bbl_list = confmanobj.get('BadBlockSubscriptionList', [])
                if type(conf_bbl_list) == type({}):
                    conf_bbl_list = SEXP.dictToList(conf_bbl_list)
                bpbblservers = map(idlib.to_ascii, bpbblservers)
                conf_bbl_list = map(idlib.to_ascii, conf_bbl_list)
                for newbblserver in bpbblservers:
                    if newbblserver not in conf_bbl_list:
                        debug.mojolog.write("adding %s to BBL subscriptions\n", v=3, vs='mojobootpage')
                        conf_bbl_list.append(newbblserver)
                        changed = 1
                confmanobj['BadBlockSubscriptionList'] = SEXP.listToDict(conf_bbl_list)

            if changed:
                if doQ:
                    doQ.add_task(confmanobj.save, delay=-10)  # delay = -10 makes it happen sooner
                else:
                    confmanobj.save()

        except:
            debug.mojolog.write('WARNING: Ignoring exception while processing bootpage:\n')
            traceback.print_exc(file=debug.mojolog)




def test_it():
    q = DoQ.DoQ()
    try:
        load_and_use(q, block=1, timeout=5)
    finally:
        q.shutdown()


mojo_test_flag=1
import RunTests
def run():
    RunTests.runTests('mojobootpage')

#### this runs if you import this module by itself
if __name__ == '__main__':
    run()
