#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#  This module contains pricer functions for looking up prices from
#  meta info services dicts (as returned by MetaTrackerLib's get_SPAM
#  functions).
#
__cvsid = '$Id: pricelib.py,v 1.35 2001/06/08 14:34:06 zooko Exp $'

# standard modules
import bisect
import math
import types


# our modules
import MojoHandicapper
from confutils import confman
import counterparties
import debug
from human_readable import hr
import mojoutil


# used internally
def _get_service(info_dict, desiredtype, default={}): 
    """
    given a meta info dict, return the service dict for desiredtype

    @precondition `info_dict' must be a dict or `None'.: (info_dict is None) or (type(info_dict) == types.DictType): "info_dict: %s" % hr(info_dict)
    """
    assert (info_dict is None) or (type(info_dict) == types.DictType), "precondition: `info_dict' must be a dict or `None'." + " -- " + "info_dict: %s" % hr(info_dict)

    if info_dict is None:
        return default

    if info_dict.get('dynamic pricing'):
        return default

    try:
        servicelist = info_dict['services']
        for service in servicelist :
            if (service is not None) and (service.get('type') == desiredtype):
                return service
    except KeyError:
        # XXX quick fix - templating should prevent this!
        pass
    return default

# used internally and in our servers for figuring out their own prices
def get_price_from_pricemap(pricemap, kilobytes) :
    """
    @param pricemap can be `None'
    @param kilobytes the size of the blob in kilobytes, or `None'

    @precondition `kilobytes' must be an integer or None.: (kilobytes is None) or (type(kilobytes) == types.IntType) or (type(kilobytes) == types.LongType): "kilobytes: %s :: %s" % (hr(kilobytes), hr(type(kilobytes)))
    """
    assert (kilobytes is None) or (type(kilobytes) == types.IntType) or (type(kilobytes) == types.LongType), "precondition: `kilobytes' must be an integer or None." + " -- " + "kilobytes: %s :: %s" % (hr(kilobytes), hr(type(kilobytes)))

    if pricemap is None:
        return None

    try:
        # lookup the exact price in the map, or the next highest one if there was no exact match
        strk = str(kilobytes)
        if strk[-1:] == "L": # stupid trailing L's -- python 2.0 doesn't do this
            strk = strk[:-1]
        exactprice_str = pricemap.get(strk)
        if exactprice_str is None:
            listedindexes = map(int, pricemap.keys())
            listedindexes.sort()
            priceidx = bisect.bisect(listedindexes, kilobytes)
            # if kilobytes is greater than the highest listed, use the highest
            if priceidx > len(pricemap)-1 :
                priceidx = len(pricemap)-1
            # XXX verify that pricemap's keys are in canonical form (the smallest string representing that integer)
            strsize = str(listedindexes[priceidx])
            if strsize[-1:] == "L": # stupid trailing L's -- python 2.0 doesn't do this
                strsize = strsize[:-1]
            return int(pricemap[strsize])
        else:
            return int(exactprice_str)
    except (KeyError, IndexError, ValueError, TypeError):
        # XXX Make this mask fewer bugs! Don't catch so many random errors!
        return None

def get_price_using_per_kilobyte(price_per_kb, bytes) :
    if price_per_kb is None:
        return None
    # compute price per kilobyte
    exactprice = float(price_per_kb)/1024 * bytes
    return long(math.ceil(exactprice))

#
# Meta tracker related prices
#
def longorNone(str):
    try:
        return long(str)
    except TypeError:
        return None

def meta_tracker_hello_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "meta tracker").get("hello price", None))

def meta_tracker_lookup_contact_info_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "meta tracker").get("lookup contact info price", None))

def meta_tracker_list_servers_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "meta tracker").get("list servers price", None))

#
# Content tracker related prices
#
def content_tracker_lookup_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "content tracker").get("lookup price", None))

def content_tracker_submit_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "content tracker").get("submit price", None))

def content_tracker_download_types_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "content tracker").get("download content types price", None))

#
# Publication agent/tracker related prices
#
def publication_agent_store_location_info_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "publication agent").get("store location information price", None))

def publication_agent_get_location_info_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "publication agent").get("get location information price", None))

#
# Blob server related prices
#
def blob_server_have_blobs_price(info_dict, msgbody, qos=1):
    return longorNone(_get_service(info_dict, "blob server").get("do you have blobs price", None))

def blob_server_publish_price(info_dict, msgbody, qos=1) :
    return get_price_from_pricemap(_get_service(info_dict, "blob server").get("put blob kilobyte price map", None), len(msgbody.get('data', ""))/1024)

def blob_server_purchase_price(info_dict, msgbody, qos=1) :
    return get_price_from_pricemap(_get_service(info_dict, "blob server").get("request blob kilobyte price map", None), msgbody.get('len', 0)/1024)

#
# Relay server related prices
#
def relay_server_are_there_messages_price(info_dict, msgbody, qos=1) :
    return longorNone(_get_service(info_dict, "relay server v2").get("are there messages price v2", None))

def relay_server_pass_along_price(info_dict, msgbody, qos=1) :
    return get_price_using_per_kilobyte(_get_service(info_dict, "relay server v2").get("pass this along per kilobyte price v2", None), len(msgbody['message']))

def relay_server_get_message_price(info_dict, msgbody, qos=1) :
    service_dict = _get_service(info_dict, "relay server v2")
    if service_dict is None:
        return None
    if not service_dict.has_key("retrieve messages per kilobyte price v2"):
        # New-style server uses automatic dynamic pricing and has no fixed-price entry.
        return None
    price_per_kb = float(service_dict["retrieve messages per kilobyte price v2"])
    if msgbody.has_key('bytes'):
        return get_price_using_per_kilobyte(price_per_kb, long(msgbody['bytes']))
    else:
        total = 0
        for i in msgbody['messages']:
            total = total + i['length']
        return get_price_using_per_kilobyte(price_per_kb, total)

########################################################################
#
#  This is a table of message types to pricer_funcs.
#
msgtype_to_pricelib_func_map = {
    "request blob": blob_server_purchase_price,
    "put blob": blob_server_publish_price,
    "do you have blobs" : blob_server_have_blobs_price,

    "store location information" : publication_agent_store_location_info_price,
    "get location information" : publication_agent_get_location_info_price,

    "pass this along v2" : relay_server_pass_along_price,
    "are there messages v2" : relay_server_are_there_messages_price,
    "retrieve messages v2" : relay_server_get_message_price,

    "content tracker lookup" : content_tracker_lookup_price,
    "content tracker submit" : content_tracker_submit_price,

    "hello" : meta_tracker_hello_price,
    "lookup contact info" : meta_tracker_lookup_contact_info_price,
    "list blob servers" : meta_tracker_list_servers_price,
    "list publishing blob servers" : meta_tracker_list_servers_price,
    "list content trackers" : meta_tracker_list_servers_price,
    "list meta trackers" : meta_tracker_list_servers_price,
    "list relay servers v2" : meta_tracker_list_servers_price,
    "list publication agents" : meta_tracker_list_servers_price,
}


########################################################################

def content_tracker_submit_handicap_price(info_dict, msgbody, qos=1):
    """consider how easy it is for others to lookup this entry when submitting"""
    p = content_tracker_submit_price(info_dict, msgbody, qos=qos)
    p = 0.25 * content_tracker_lookup_price(info_dict, msgbody, qos=qos) + p
    return p

def relay_server_poll_handicap_price(info_dict, msgbody, qos=1):
    """consider overall relay server prices when picking a relay server to use"""
    p = relay_server_are_there_messages_price(info_dict, msgbody, qos=qos)
    service_dict = _get_service(info_dict, "relay server v2")
    p = 0.200 * float(service_dict["retrieve messages per kilobyte price v2"]) + p
    p = 0.100 * float(service_dict["pass this along per kilobyte price v2"]) + p
    return p

########################################################################

msgtype_to_pricehandicapper_func_map = {}
msgtype_to_pricehandicapper_func_map.update(msgtype_to_pricelib_func_map)
msgtype_to_pricehandicapper_func_map["content tracker submit"] = content_tracker_submit_handicap_price
msgtype_to_pricehandicapper_func_map["are there messages v2"] = relay_server_poll_handicap_price



########################################################################
TUNING_FACTOR=1000
class PriceHandicapper:
    def __init__(self, counterparties):
        self._counterparties = counterparties

    def compute_handicap(self, counterparty_id, metainfo, message_type, message_body):
        """
        for all msgtypes

        @returns counterparty's costliness * mojo offer for this request * TUNING_FACTOR, where TUNING_FACTOR == 1000
        """
        cpobj = self._counterparties.get_counterparty_object(counterparty_id)
        return cpobj.get_custom_stat("costliness", 0.5) * TUNING_FACTOR

mojo_test_flag = 1

def run():
    import RunTests
    RunTests.runTests("pricelib")

if __name__ == '__main__':
    run()

