#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# Collection of Mojo Tokens

### standard modules
import copy
import exceptions
import math
import types

### our modules
import MojoToken
from MojoConstants import true, false
import SEXP
from human_readable import hr
import mojoutil


def amount_of_smallest_dist_greater_than(amount):
    """
    @precondition `amount' must be non-negative.: amount >= 0: "amount: %s" % hr(amount)
    """
    assert amount >= 0, "`amount' must be non-negative." + " -- " + "amount: %s" % hr(amount)

    td = choose_greater_than_or_equal_token_dist(amount)

    return amount_of_token_dist(td)


def is_amount_possible(amount, dist):
    """
    @returns `true' if and only if it is possible to choose a subset of `dist' that has a total value of `amount'

    @precondition There must be enough aggregate value in dist.: amount_of_token_dist(dist) >= amount
    """
    assert amount_of_token_dist(dist) >= amount, "There must be enough aggregate value in dist."

    # Note: we could probably do this faster.  Oh well.
    amountneeded = amount

    ds = dist.keys()
    ds.sort()
    ds.reverse()

    for d in ds:
        amountneeded = amountneeded - (dist[d] * d)
        if amountneeded <= 0:
            return true
        
    return false


def default_choose_token_dist(amount):
    """
    For a given amount, choose a distribution of tokens that sum to that amount.

    @param amount the total amount

    @return map from denomval to numberoftokens, summing to `amount'
    """
    denoms = MojoToken.MAP_DENOM_TO_KEYID.keys()

    denoms.sort()
    denoms.reverse()

    resultdist = {}
    amountleft = amount
    while amountleft > 0:
        if len(denoms) == 0:
            raise CantMakeAmountError

        d = denoms.pop()
        numofthisdenom = amountleft / d
        if numofthisdenom > 0:
            resultdist[d] = numofthisdenom
            amountleft = amountleft - (numofthisdenom * d)

    return resultdist


def choose_greater_than_or_equal_token_dist(amount):
    """
    For a given amount, choose a distribution of tokens that sum to at least amount.

    @param amount the total amount

    @return map from denomval to numberoftokens, summing to `amount'
    """
    denoms = MojoToken.MAP_DENOM_TO_KEYID.keys()

    denoms.sort()
    denoms.reverse()

    resultdist = {}
    amountleft = amount
    while amountleft > 0:
        assert len(denoms) > 0, "We should exit this loop before this happens!"

        d = denoms.pop()
        assert (type(d) == types.IntType) or (type(d) == types.LongType), "`d' must be integer. -- d: %s" % hr(d)

        # If this is the smallest denom, then round up else round down.
        if len(denoms) == 0:
            numofthisdenom = long(math.ceil(float(amountleft)/float(d)))
        else:
            numofthisdenom = long(amountleft / d)

        if numofthisdenom > 0:
            resultdist[d] = numofthisdenom
            amountleft = amountleft - (numofthisdenom * d)

    return resultdist


def basic_choose_token_dist_from_set(amount, set):
    """
    Given a set of tokens and an amount that you want to choose out of that set, identify which
    tokens from that set you will choose.

    The "basic" token distribution choice ensures that it will choose `amount' tokens if it is
    possible to choose `amount' tokens from `set'.  Other token distribution choices do not
    guarantee this.

    @param amount the total amount
    @param set the set of tokens from which to choose as a map from denomval to numberoftokens

    @return map from denomval to numberoftokens, summing to `amount'

    @throws NoExactChangeError if there are not correctly sized tokens in `set' to make a
        collection of exactly size `amount'

    @precondition There must be enough value in this token set.: amount_of_token_dist(set) >= amount

    @postcondition There are `amount' worth of tokens in result.: amount_of_token_dist(result) == amount
    """
    assert amount_of_token_dist(set) >= amount, "There must be enough value in this token set."

    result = {}

    while amount_of_token_dist(result) < amount:
        # Pick the biggest available coin less than the amount you need.
        amountneeded = amount - amount_of_token_dist(result)

        ds = availtoksremaining.keys()
        ds.sort()
        d = ds.pop()
        while (d > amountneeded):
            if len(ds) == 0:
                # We can't make that amount out of this set.
                # Okay just return the basic distribution
                raise NoExactChangeError

            d = ds.pop()

        move_toks(availtoksremaining, result, { d: long(amountneeded / d) })

    return result


def move_toks(origset, targset, tomove):
    """
    Move a set of tokens (described by `tomove') from `origset' to `targset'.

    If `tomove' says to move more tokens of a given denomination than are present in `origset'
    then it will move as many as are present in `origset'.

    @paramorigset the `set' of tokens from which to take as a map from `denomval' to numberoftokens
    @param targset the set of tokens in which to place as a map from denomval to numberoftokens
    @param tomove how many tokens of each denom to move as a map from denomval to numberoftokens

    @returns the aggregate value of the tokens moved
    """
    totalvalmoved = 0

    for d in tomove.keys():
        tomovenum = tomove[d]
        orignum = origset[d]

        if tomovenum <= orignum:
            # Great!  You can have as many of these as you wanted.
            numtotake = tomovenum
        else:
            # Oops -- there aren't enough of those.
            # Get as many as you can:
            numtotake = orignum

        origset[d] = origset[d] - numtotake
        if origset[d] == 0:
            del origset[d]

        
        mojoutil.add_long_to_dict(targset, d, numtotake)
        totalvalmoved = totalvalmoved + (numtotake * d)

    return totalvalmoved


def choose_token_dist_from_set(amount, set, choose_func=default_choose_token_dist):
    """
    Given a set of tokens and an amount that you want to choose out of that set and a function
    defining your preferences, identify which tokens from that set you will choose.

    @param amount the total amount
    @param set the set of tokens from which to choose as a map from denomval to numberoftokens
    @param choose_func a function with the same interface as `default_choose_token_dist()' to
        indicate your preferences of token denominations

    @return map from denomval to numberoftokens, summing to `amount'

    @throws NoExactChangeError if there are not correctly sized tokens in `set' to make a
        collection of exactly size `amount'

    @precondition There must be enough value in this token set.: amount_of_token_dist(set) >= amount

    @postcondition There are `amount' worth of tokens in result.: amount_of_token_dist(result) == amount
    """
    assert amount_of_token_dist(set) >= amount, "There must be enough value in this token set."

    if not is_amount_possible(amount, set):
        raise NoExactChangeError


    availtoksremaining = set.copy()
    old_amountsofar = -1 # This is so that we will enter the first-phase loop the first time.

  
    # Now we'll allocate tokens in two phases.  In the first phase, we'll allocate all of the
    # tokens indicated by `idealdist', if they are available.  If doing so successfully
    # allocated any tokens at all (thus reducing `amountunsatisfied') then we'll loop and do it
    # again with the new amount.  If choosing `idealdist' doesn't get you any tokens then it
    # will try `default_choose_token_dist()' and see if that works.

    # If it gets stuck where it can't allocate any tokens at all then we'll go into the second
    # phase in which we try to greedily allocate any tokens that are available.

    # If _that_ doesn't work then we just call `basic_choose_token_dist_from_set()' to allocate
    # tokens out of the available set until we have enough amount.

    # If _that_ doesn't work then it isn't possible to get amount out of this set.  Actually we
    # check for that at the beginning (above).


    result = {}

    # first-phase loop
    while amount_of_token_dist(result) != old_amountsofar:
        old_amountsofar = amount_of_token_dist(result)

        # Now if you could have any kind of tokens you wanted what would your ideal distribution be?
        idealdist = choose_func(amount - amount_of_token_dist(result))
        moved = move_toks(availtoksremaining, result, idealdist)
        if moved == 0:
            otheridealdist = default_choose_token_dist(amount - amount_of_token_dist(result))
            move_toks(availtoksremaining, result, otheridealdist)


        if amount_of_token_dist(result) == amount:
            return result # done!


    # second-phase clean-up
    while amount_of_token_dist(result) < amount:
        # Pick the biggest available coin less than the amount you need.
        amountneeded = amount - amount_of_token_dist(result)

        ds = availtoksremaining.keys()
        ds.sort()
        d = ds.pop()
        while (d > amountneeded):
            if len(ds) == 0:
                # We can't make that amount out of this set.
                # Okay just return the basic distribution
                return basic_choose_token_dist_from_set(amount, set)

            d = ds.pop()

        move_toks(availtoksremaining, result, { d: long(amountneeded / d) })

    return result # done!

def amount_of_token_dist(cDist):
    """
    @param cDist a map from denom to number of tokens
    """
    sum = 0
    for denom in cDist.keys():
        sum = sum + (denom * cDist[denom])
        assert (type(sum) == types.IntType) or (type(sum) == types.LongType), "`sum' must be integer. -- sum: %s, denom: %s, cDist: %s" % (hr(sum), hr(denom), hr(cDist))

    return sum


class tokendistsError(exceptions.StandardError):
    pass

class CantMakeAmountError(tokendistsError):
    """
    An operation failed because the distribution of denominations that exist (currently stored
    in `MojoToken.MAP_DENOM_TO_KEYID.keys()') cannot be used to make a requested amount.  (You can
    avoid this condition entirely by making your denom distribution be exactly the powers of 2.)
    """
    pass

