#!/usr/bin/env python
#
# An http proxy server which takes all requests to reference.mojonation.net 
# and sends them to localhost
#
# based on munchy by Neil Schemenauer <nascheme@enme.ucalgary.ca>

# default port if none specified on command line
PORT = 8000

MOJO_NATION_SERVER_PORT = 4004

# debugging level,
#   0 = no debugging
#   1 = access debugging
#   2 = full debugging
#   3 = really full debugging
DEBUG_LEVEL = 1
ALLOW_DEBUG_ON_WIN32 = None


# the address to bind the server to
ADDR_TO_BIND_TO = '127.0.0.1'


# this is set when we should stop and exit
ShouldExitNow = None

#####################################################################
# End of user configuration
#####################################################################

import sys
import os
import time
import string
import socket
import urlparse
import re
import SocketServer
from select import select
import BaseHTTPServer
from stat import ST_MTIME
import mimetools


def log(s, level=1):
    if sys.platform == 'win32' and not ALLOW_DEBUG_ON_WIN32:
        return
    if level <= DEBUG_LEVEL:
        sys.stderr.write(s)

def try_del(dict, key):
    try:
        del dict[key]
    except KeyError:
        pass


# hostnames we wish to check for a redirect to your local broker
# (we run a real-world server on all of these so that people without
# the mojo nation software will be directed to our site for info and
# instructions on how to quickly download and install it)
hosts_to_check_re = re.compile(r'^(www\.|)(mojonation|mojo-nation)\.(com|net|org)', re.I)

# note: this will match with or without a privacy code
paths_to_redirect_re = re.compile(r'^(/[A-Za-z0-9_-]{12,27}|)/(id|mojoid|dinode|broker|nexus|station)', re.I)

# this is used to get info about the running broker from its intropage
extract_brokerinfo_re = re.compile(r'-- Port: (?P<port>\d+) PrivacyCode: (?P<privacycode>\S+) --', re.I)

# we remove all referer headers matching this
referer_to_strip_re = re.compile(r'(www\.|)mojonation\.(com|net|org)(/[A-Za-z0-9_-]{12,27}|)/(id|mojoid|dinode|broker|nexus|station)', re.I)

Intropage_Filename = None
Last_Intropage_Stat = {ST_MTIME:None}  # XXX this dict is a hack so the if in get_privacy_code_and_port works ;)
Current_Privacy_Code = ""
Current_Broker_Port = MOJO_NATION_SERVER_PORT
Current_Broker_Address = "127.0.0.1"   # the mojonation broker runs on localhost by default


def get_privacy_code_and_port() :
    """
    Checks if intropage.html has changed.  If so, it rereads the privacy code from it.

    Returns a tuple: (privacycode, broker port)
    """
    global Last_Intropage_Stat
    global Current_Privacy_Code
    global Current_Broker_Port
    try:
        newstat = os.stat(Intropage_Filename)
    except OSError:
        log('error getting status of intropage %s\n' % Intropage_Filename, 1)
        Last_Intropage_Stat = {ST_MTIME:None}
        return Current_Privacy_Code, Current_Broker_Port

    # only re-read the file if it has been modified
    if Last_Intropage_Stat[ST_MTIME] != newstat[ST_MTIME]:
        try:
            ipfile = open(Intropage_Filename, "r")
            m = extract_brokerinfo_re.search(ipfile.readline())
            if m :
                Current_Privacy_Code = '/' + m.group('privacycode')
                Current_Broker_Port = int(m.group('port'))
            Last_Intropage_Stat = newstat
        except:
            log('error reading intropage %s\n' % Intropage_Filename, 1)
    
    return Current_Privacy_Code, Current_Broker_Port


# used for replacing the path in a request if we're chaining to another proxy
http_proxy_fixup_re = re.compile(r'^(\w+) (.*) (HTTP.*)$', re.M)

class ProxyHandler(SocketServer.StreamRequestHandler):
    def handle(self):
        """handle one request from the browser"""
        http_proxy = os.environ.get('http_proxy')
        if not http_proxy :
            http_proxy = os.environ.get('HTTP_PROXY')
        try:
            host, port, request, url = self.read_request()

            # if we're chaining to another proxy, modify our request to do that
            if http_proxy and host != Current_Broker_Address :
                scheme, netloc, path, params, query, fragment = urlparse.urlparse(http_proxy)
                if string.lower(scheme) == 'http' :
                    log('using next http proxy: %s\n' % netloc, 2)
                    # set host and port to the proxy
                    if ':' in netloc:
                        host, port = string.split(netloc, ':')
                        port = string.atoi(port)
                    else:
                        host = netloc
                        port = 80
                    # replace the path within the request with the full URL for the next proxy
                    request = http_proxy_fixup_re.sub(r'\1 %s \3' % url, request, count=1)

            srfile, swfile = self.connect(host, port)
            if http_proxy :
                log('sending request to the next http proxy "%s"\n' % request, 2)
            else :
                log('sending request to server "%s"\n' % request, 2)

            self.send_request(swfile, request)
            self.handle_response(srfile)
        except IndexError:
            pass # got sent invalid data
        except IOError:
            pass # browser closed connection?
        log('finished request\n', 2)

    def read_request(self):
        """read request to find host and port and make new request"""
        request = self.rfile.readline()
        sys.stdout.write('%s - %s - %s' % (
                self.client_address[0], time.ctime(time.time()), request))
        try:
            method, url, protocol = string.split(request)
        except:
            self.error(400, "Can't parse request")
        if not url:
            self.error(400, "Empty URL")
        if method == 'MOJOSHUTDOWN':
            global ShouldExitNow
            ShouldExitNow = 1
            raise SystemExit, "Got a MOJOSHUTDOWN request"
        if method not in ['GET', 'HEAD', 'POST']:
            self.error(501, "Unknown request method (%s)" % method)
        # split url into site and path
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
        if string.lower(scheme) != 'http':
            self.error(501, "Unknown request scheme (%s)" % scheme)
        # find port number
        if ':' in netloc:
            host, port = string.split(netloc, ':')
            port = string.atoi(port)
        else:
            host = netloc
            port = 80
        path_match = paths_to_redirect_re.match(path) 

        # redirect mojonation.net urls to the local broker as appropriate
        if hosts_to_check_re.match(host) and path_match :
            self.original_host_and_port = (host, port)
            host = Current_Broker_Address
            privacycode, port = get_privacy_code_and_port()
            if len(path_match.group(1)) == 0 :   # if no privacy code was supplied, add one
                # TODO now we need some way of authenticating the user in here if we're really
                # going to support automatic adding of the privacy code like this.  <ugh..>
                path = privacycode + path
            log("redirecting to mojo broker as %s\n" % path, 2)
            # put this header in so that the broker knows we came via the proxy
            reveal_mojoproxy = 1
        else :
            self.original_host_and_port = None
            reveal_mojoproxy = 0
        path = urlparse.urlunparse(('', '', path, params, query, fragment))
        # read headers
        mimeheaders = self.read_headers(self.rfile)
        if reveal_mojoproxy:
            mimeheaders['mojoproxy'] = 'yes'
        if method == 'POST' and not mimeheaders.has_key('content-length'):
            self.error(400, "Missing Content-Length for POST method")
        length = int(mimeheaders.get('content-length', 0))
        # read content if any
        content = self.rfile.read(length)
        log('content = %s\n' % `content`, 2)
        # build new request
        # (NOTE: try_del isn't necessary anymore due to rfc822.Message semantics but it doesn't hurt)
        try_del(mimeheaders, 'accept-encoding')
        try_del(mimeheaders, 'proxy-connection')
        # strip off any referer from mojonation.net
        referer = mimeheaders.get('referer', 0)
        if referer and referer_to_strip_re.search(mimeheaders['referer']) :
            # IMHO, we should -always- do this but unfortunately some
            # stupid web sites probably depend on it.  -greg
            try_del(mimeheaders, 'referer')
        if port == 80 :
            mimeheaders['host'] = host
        else:
            # XXX hmm.. is the host header ever supposed to have the port on it?
            # Shouldn't the server know that already?  (I ask because some don't
            # like seeing a :80 on it [such as sourceforge.net])
            mimeheaders['host'] = host + ':' + str(port)

        request = '%s %s HTTP/1.0\r\n%s\r\n%s' % (method, path,
                self.join_headers(mimeheaders), content)
        return host, port, request, url

    def connect(self, host, port):
        try:
            addr = socket.gethostbyname(host)
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            log('addr = %s\n' % repr(addr), 2)
            log('port = %s\n' % repr(port), 2)
            server.connect((addr, port))
        except socket.error, err:
            log('Socket error %s connecting to %s:%d\n' % (err, host, port), 1)
            if self.original_host_and_port:
                log('Trying original host & port instead of redirecting to the mojo broker\n', 1)
                # XXX We should return a static page here rather
                # than passing the request onto the real
                # mojonation web server for the error page.
                # TODO even better, present the user with a choice to start their broker for them
                try:
                    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    server.connect(self.original_host_and_port)
                except socket.error, err:
                    self.error(500, 'Error connecting to "%s:%d" (%s)' % (self.original_host_and_port + (err,)) )
            else:
                self.error(500, 'Error connecting to "%s:%d" (%s)' % (host, port, err))
        return server.makefile('rb'), server.makefile('wb')

    def read_headers(self, input):
        return mimetools.Message(input, seekable=0)

    def join_headers(self, mimeheaders):
        return string.join(mimeheaders.headers, '')

    def send_request(self, server, request):
        try:
            server.write(request)
            server.flush()
        except socket.error, err:
            self.error(500, 'Error sending data in send_request (%s)' % (err,))

    def handle_response(self, server):
        log('reading server response\n', 2)
        response = server.readline()
        log('response = %s\n' % response, 2)
        fields = string.split(response)
        if not fields:
            self.error(500, 'No response from server.')
        elif len(fields) < 3:
            self.error(500, 'Bad response (%s)"' % response)
        version = fields[0]
        status = fields[1]
        comment = string.join(fields[2:])
        self.wfile.write('HTTP/1.0 %s %s\r\n' % (status, comment))
        log('reading response headers\n', 2)
        mimeheaders = self.read_headers(server)
        if DEBUG_LEVEL >= 3:  # do this to prevent %s expansion unless desired
            log('writing headers to client:\n%s\n' % mimeheaders, 3)
        self.wfile.write(self.join_headers(mimeheaders))
        self.wfile.write('\r\n')
        # read by blocks
        log('transfering raw data\n', 2)
        while 1:
            data = server.read(4096)
            log('data = %s\n' % `data`, 5)
            if not data:
                break
            self.wfile.write(data)
        self.wfile.flush()

    def finish(self):
        try:
            self.connection.setblocking(0)
            r, w, e = select([self.rfile], [self.wfile], [], 0)
            if r and w:
                self.wfile.write(self.rfile.read())
            SocketServer.StreamRequestHandler.finish(self)
        except IOError:
            pass

    def error(self, code, body):
        response = BaseHTTPServer.BaseHTTPRequestHandler.responses[code][0]
        self.wfile.write("HTTP/1.0 %s %s\r\n" % (code, response))
        self.wfile.write("Server: Mojo Nation Proxy\r\n")
        self.wfile.write("Content-type: text/html\r\n")
        self.wfile.write("\r\n")
        self.wfile.write('<html><head>\n<title>%d %s</title>\n</head>\n'
                '<body>\n%s\n</body>\n</html>' % (code, response, body))
        self.wfile.close()
        self.rfile.close()
        raise SystemExit

class ProxyServer(SocketServer.ThreadingTCPServer):
    def __init__(self, port):
        log('Starting proxy on %s port %d\n' % (ADDR_TO_BIND_TO, port), 1)
        SocketServer.ThreadingTCPServer.__init__(self, (ADDR_TO_BIND_TO, port), ProxyHandler)

    def server_bind(self) :
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        SocketServer.ThreadingTCPServer.server_bind(self)
    
    def handle_request(self):
        global ShouldExitNow 
        if ShouldExitNow:
            raise SystemExit, "ShouldExitNow flag was set"
        else:
            SocketServer.ThreadingTCPServer.handle_request(self)

if __name__ == '__main__':
    global Intropage_Filename
    if len(sys.argv) > 2 :  # 2rd param: the location of the intropage file containing the privacy code
        Intropage_Filename = sys.argv[2]
    else :
        Intropage_Filename = os.path.join("localweb", "intropage.html")
    log('Using intropage %s to obtain your privacy code\n' % Intropage_Filename, 1)

    if len(sys.argv) > 3 :  # 3th param: the next-step HTTP proxy can specified here (overrides the environment variable)
        os.environ['http_proxy'] = sys.argv[3]
    # display which proxy we're using
    http_proxy = os.environ.get('http_proxy')
    if not http_proxy :
        http_proxy = os.environ.get('HTTP_PROXY')
    if http_proxy :
        log("I will use the proxy at %s\n" % http_proxy, 1)

    if len(sys.argv) < 2:   # 1st param: the port for this proxy to run on
        ProxyServer(PORT).serve_forever()
    else:
        ProxyServer(int(sys.argv[1])).serve_forever()

