#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: BlockServerEGTP.py,v 1.68 2001/06/08 01:49:06 greg Exp $'


# standard modules
import exceptions
import sys
import string
import whrandom
import threading
import time

# out modules
import DoQ
import localblockstore
from confutils import confman, confdefaults, integerizePrice
import MojoTransaction
import SEXP
from PublicationTrackerHandlers import PublicationTrackerHandlers
import MetaTrackerLib
import blockserverclientlib
import TCPCommsHandler
from MojoConstants import true, false
import debug
from human_readable import hr
import mojosixbit
from PublishSession import LocationBroadcaster
import pricelib
import mojoutil
import idlib
import idmasklib
import Cache
from BlockWholesaler import BlockWholesaler, WHOLESALER_FLAG
import ClientEGTPlib

class Error(exceptions.StandardError): pass
class MaskError(Error): pass
class NoSpaceError(Error): pass
class BLBlobError(Error): pass


# tune polling based on our connection speed setting (TODO allow this to be changed at runtime)
# TODO publication tracker polling really just needs to hook into the BlockThrottler rather than use this hack
__speed = int(confman.dict['CONN_SPEED_IN'])
if __speed > 2000:
    BLOCKS_TO_PURCHASE_AT_ONCE = 5
if __speed > 500:
    BLOCKS_TO_PURCHASE_AT_ONCE = 3
elif __speed > 180:
    BLOCKS_TO_PURCHASE_AT_ONCE = 2
else:
    BLOCKS_TO_PURCHASE_AT_ONCE = 1


class BlobServerEGTP:
    def __init__(self, broker, throttler):
        self.broker = broker
        self._publication_tracker_tuple = None

        self.publishing_allowed = (string.lower(confman.dict["YES_NO"]["BLOCK_SERVER_ALLOWS_PUBLISHING"]) == "yes")

        handler_funcs = {
            'request blob': (self.handle_request_blob, self.pricer_func_for_mtm),
            'do you have blobs': (self.handle_do_you_have_blobs, self.pricer_func_for_mtm),
        }
        self.handler_func_names = handler_funcs.keys()

        if self.publishing_allowed:
            handler_funcs['put blob'] = (self.handle_put_blob, self.pricer_func_for_mtm)

        NEW_MASK_BITS = int(SEXP.dictToList(confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"])[0].get("bits", 7))
        # create a masklist (of IdMask objects rather than dicts representing the masks)
        # XXX hack: if no LIST_OF_BLOCKID_MASKS is present, then create a random one and save it.  --Zooko 2000-09-03
        if (not SEXP.dictToList(confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"])[0].get("mask")):
            debug.mojolog.write('No block server bitmask found, creating a new one.\n', vs='BlockServer')
            confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"] = SEXP.listToDict( [
                idmasklib.generate_random_mask(bits = NEW_MASK_BITS, keepfull = true).get_mojodict(fullmask=true)
                ] )
            confman.dict['BLOCKSERVER']['SELECTED_MASK_PROPERLY'] = 'yes'

        self.masklist = map(lambda d: idmasklib.IdMask(mojodict=d, keepfull=true), SEXP.dictToList(confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"]))
        debug.mojolog.write('Your block server mask: %s\n', args=(self.masklist[0],), vs='BlockServer')

        # to clean up mess resulting from generating non-random masks as prior to 0.932
        if (confman.dict['BLOCKSERVER'].get('SELECTED_MASK_PROPERLY', 'no') != 'yes') and \
            (self.masklist[0].get_mojodict(fullmask = true)['mask'][0] == 'A'):
            debug.mojolog.write('Non-random block server bitmask found, creating a new one.\n', vs='BlockServer')
            confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"] = SEXP.listToDict( [
                idmasklib.generate_random_mask(bits = NEW_MASK_BITS, keepfull = true).get_mojodict(fullmask=true)
                ] )
            self.masklist = map(lambda d: idmasklib.IdMask(mojodict=d, keepfull=true), SEXP.dictToList(confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"]))
            debug.mojolog.write('Your new block server mask: %s', args=(self.masklist[0],), vs='BlockServer')

        # always set this to yes here.
        confman.dict['BLOCKSERVER']['SELECTED_MASK_PROPERLY'] = 'yes'

        # Save these bitmasks, in case there are new ones, or old bitmasks have been extended to full bitmasks:
        confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"] = SEXP.listToDict(map(lambda idm : idm.get_mojodict(fullmask=true), self.masklist))
        confman.save()

        self.service_announcement = {
            'type': "blob server",
            'publishing allowed': confman.dict["YES_NO"]["BLOCK_SERVER_ALLOWS_PUBLISHING"],
            'request blob kilobyte price map': integerizePrice(confman.dict["BLOCKSERVER"]["PRICE_MAPS"]["RETRIEVE"]),
            'put blob kilobyte price map': integerizePrice(confman.dict["BLOCKSERVER"]["PRICE_MAPS"]["PUBLISH"]),
            'mask list': map(lambda m: m.get_mojodict(fullmask=true), self.masklist),
            'do you have blobs price': integerizePrice(confman.dict["BLOCKSERVER"]["PRICE_MAPS"]["HAVE_BLOCK_LIST"]),
        }

        self.broker.ServicesMTM.add_handler_funcs(handler_funcs)
        self.broker.ServicesMTM.add_announced_services([self.service_announcement])
        self.broker.ServicesMTM._blobserver = self  # XXX Klooge so that we can reach the block server object from several other modules

        def publish_info(blob_list, self = self):
            blobs = []
            for blob_id, blob_size in blob_list:
                blobs.append({'blob id' : idlib.canonicalize(blob_id, 'blob'), 'server id' : self.broker.ServicesMTM.get_id(), 'bytes' : str(int(blob_size))})
            LocationBroadcaster(self.broker.ServicesMTM, None, blobs)

        self.wholesaler = BlockWholesaler(broker=self.broker, throttler = throttler)

    def shutdown(self):
        self.broker.ServicesMTM.remove_announced_services(self.service_announcement.keys())
        self.broker.ServicesMTM.remove_handler_funcs(self.handler_func_names)
        self.broker.multiblockstore.shutdown()

    def change_mask(self, mojodict, index=0):
        """Change one of my masks"""
        oldmask = self.masklist[index]
        self.masklist[index] = idmasklib.IdMask(mojodict=mojodict)
        confman["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"][str(index)] = self.masklist[index].get_mojodict(fullmask=1)
        self.service_announcement['mask list'] = map(lambda m: m.get_mojodict(), self.masklist)
        self.broker.ServicesMTM._hello_sequence_num_needs_increasing()
        debug.mojolog.write("mask %s changed from %s to %s\n", args=[index, oldmask.get_mojodict(), self.masklist[index].get_mojodict()], vs='BlockServer')

    def pricer_func_for_mtm(self, widget, incoming_type, incoming_body):
        """This gets called by our MTM on incoming messages to verify their mojo offer before calling our handler func"""
        assert self.broker.doq.is_currently_doq(), "BlockServer message handlers must be called from the DoQ"
        if incoming_type == "do you have blobs":
            return int(self.service_announcement['do you have blobs price'])
        elif incoming_type == "request blob":
            blobid = incoming_body["blob id"]
            blobsize = self.broker.multiblockstore.get_block_size(blobid)
            if blobsize > 0:
                amount = pricelib.get_price_from_pricemap(self.service_announcement['request blob kilobyte price map'], blobsize/1024)
                if amount is not None:
                    return amount
            # Didn't find it!
            raise MojoTransaction.PricerError, "BlockServerEGTP.pricer_func_for_mtm(): Couldn't find block with id %s while trying to price it.\n" % idlib.to_mojosixbit(blobid)
        elif incoming_type == "put blob":
            blobid = incoming_body["blob id"]
            blobsize = len(incoming_body["data"])
            amount = pricelib.get_price_from_pricemap(self.service_announcement['put blob kilobyte price map'], blobsize/1024)
            if amount is None:
                raise MojoTransaction.PricerError
            if incoming_body.get('passalong', 0) > 0:
                amount = amount * 2
            return amount
        else:
            debug.stderr.write('ERROR: type %s unhandled in BlobServer pricer_func\n' % incoming_type)
            return 1000000000  # if someone gives us a billion for something we don't know how to do, let them. ;)

    def handle_do_you_have_blobs(self, widget, msgbody):
        assert self.broker.doq.is_currently_doq(), "BlockServer message handlers must be called from the DoQ"
        if type(msgbody) != type({}):
            blobidlist = msgbody  # v0.901 - 0.931 only support this list only message format
            allow_wholesale = (len(blobidlist) > 0 and not idlib.equal(blobidlist[0], WHOLESALER_FLAG))
        else:
            blobidlist = msgbody['block id list']
            allow_wholesale = (msgbody.get('do not recurse') != "true") and (len(blobidlist) > 0 and not idlib.equal(blobidlist[0], WHOLESALER_FLAG))

        to_find = []
        resultlist = []
        for blobid in blobidlist:
            blobsize = self.broker.multiblockstore.get_block_size(blobid)
            if blobsize > 0:
                resultlist.append(blobid)
            else:
                to_find.append(blobid)

        if len(to_find) > 0 and allow_wholesale:
            # remove unwanted blob ids from the ones we'll try and find
            to_find = filter(lambda xx, self=self: not self.broker.multiblockstore.is_bl_blob_id(xx), to_find)
            # the most concise format (needed by mask.match)
            to_find = map(idlib.to_binary, to_find)
            # make sure all blocks that we'll try and find fit within our mask(s)
            for mask in self.masklist:
                to_find = filter(mask.match, to_find)
            # tell the block wholesaler what blocks we've been asked for but don't have
            if len(to_find) > 0:
                self.broker.doq.add_task(self.wholesaler.add_queries, args = [to_find])

        return resultlist

    def handle_request_blob(self, widget, msgbody):
        assert self.broker.doq.is_currently_doq(), "BlockServer message handlers must be called from the DoQ"
        blobid = msgbody['blob id']
        allow_wholesale = (msgbody.get('do not recurse') != "true")

        # Look in our local block storage for blockshare
        blobfile = None
        blobsize = self.broker.multiblockstore.get_block_size(blobid)

        if blobsize <= 0:
            # block not found? count it as being desirable in the wholesaler
            if allow_wholesale and not self.broker.multiblockstore.is_bl_blob_id(blobid):
                # if it fits within our mask, count it as desirable to the wholesaler
                for mask in self.masklist:
                    if mask.match(idlib.to_binary(blobid)):
                        self.broker.doq.add_task(self.wholesaler.add_queries, args=([blobid],))
                        break

            return {'result': 'failure'}  # we don't have that blob

        try:
            # TODO increment this block's "mojo earned" by the mojocharge amount
            blobfile = self.broker.multiblockstore.get_block_file(blobid)
        except (localblockstore.Hell, IOError), e:
            debug.mojolog.write('WARNING get_block_file (blob file) exception: %s\n' % e.args)
            return {'result': 'failure'}  # we couldn't read it, return a failure message

        blobdata = blobfile.read()

        if len(blobdata) != blobsize:
            raise "inconsistency in local data: file " + blobfile.name + " is identified as having " + `blobsize` + " size in the database, but we just read " + `len(blobdata)` + " bytes from it."

        blobfile.close()

        return {'result': 'success', 'blob id': blobid, 'data': blobdata}

    def try_to_store_blob(self, blobid, blobdata, verifymask=1):
        """
        @returns `true' if and only if the blob was successfully stored
        """
        try:
            self.store_blob(blobid, blobdata, verifymask=verifymask)
            return true
        except Error:
            # Oh well -- not enough space or wrong mask.
            return false

    def store_blob(self, blobid, blobdata, verifymask=1):
        """
        @precondition `blobid' must be an id.: idlib.is_sloppy_id(blobid): "blobid: %s" % hr(blobid)
        @precondition `blobid' must identify `blobdata'.: idlib.equal(idlib.make_id(blobdata, 'blob'), blobid): "blobid: %s, blobdata: %s" % (hr(blobid), hr(blobdata))

        @throws MaskError if `blobid' is not within our mask
        @throws NoSpaceError if there is not enough space
        """
        assert idlib.is_sloppy_id(blobid), "precondition: `blobid' must be an id." + " -- " + "blobid: %s" % hr(blobid)
        assert idlib.equal(idlib.make_id(blobdata, 'blob'), blobid), "precondition: `blobid' must identify `blobdata'." + " -- " + "blobid: %s, blobdata: %s" % (hr(blobid), hr(blobdata))
        assert self.broker.doq.is_currently_doq(), "BlockServer message handlers must be called from the DoQ"

        blobsize = len(blobdata)
        
        # verify that this blockshare matches one of our masks
        if verifymask:
            for mask in self.masklist:
                if mask.match(idlib.sloppy_id_to_bare_binary_id(blobid)):
                    break
            else:
                raise MaskError

        if self.broker.multiblockstore.is_bl_blob_id(blobid):
            raise BLBlobError
        
        # Open a tempfile within the blockstore.
        try:
            tempF = self.broker.multiblockstore.open_tempfile(blobid, blobsize)
        except localblockstore.NoSpaceError, e:
            # this shouldn't happen as blockstores make more space by deleting things as needed
            raise NoSpaceError
        tempF.write(blobdata)
        tempF.close()

        # Commit the tempfile into the blockstore.
        self.broker.multiblockstore.store_tempfile(blobid, tempF)

    def handle_put_blob(self, widget, msgbody):
        """
        Note that the precondition should be checked by template checking on incoming messages.
        """
        # Verify that shareId matches the received data's SHA1.
        if not idlib.identifies(msgbody['blob id'], msgbody['data'], 'blob'):
            # If the blob id given didn't match, we reject this blob (saying why)
            return {'result': "failure", 'reason': "blob id did not match data"}

        try:
            # TODO increment this blob's "mojo earned" by the mojocharge amount.
            # TODO allow this handler to control the refund size so that it can
            # always keep some mojo for a 'put blob' when possible.
            self.store_blob(msgbody['blob id'], msgbody['data'], verifymask=0)
            response = {'result': "success"}
        except MaskError:
            response = {'result': "failure", 'reason': "unsupported mask"}
        except NoSpaceError:
            response = {'result': "failure", 'reason': "no space available" }
        except BLBlobError:
            response = {'result': "failure", 'reason': "internal blobstore error" }
        
        passalong = msgbody.get('passalong', 0)
        if passalong:
            # if asked, try to send it to another block server to help block distribution
            # (yes, this is a very freenet smelling feature)
            def _put_blob_passalong(servermap, self=self, ourid=self.broker.ServicesMTM.get_id(), pubid=widget.get_counterparty_id(), blobid=msgbody['blob id'], blobdata=msgbody['data']):
                for serverid in servermap.keys():
                    if idlib.equal(serverid, pubid) or idlib.equal(serverid, ourid):
                        del servermap[serverid]
                if len(servermap) > 0:
                    ClientEGTPlib.publish_blob(self.broker.ServicesMTM,
                        serverlist=servermap.items(),
                        blobid=blobid, blobdata=blobdata, passalong=0,
                        outcome_func=lambda widget, outcome=None, failure_reason=None: None)
            # lookup a few other block servers that should be willing to hold this blob
            blob_id_mask = idmasklib.IdMask(mask=idlib.to_binary(msgbody['blob id']), bits=14)
            MetaTrackerLib.get_publishing_blob_servers_and_call_back(self.broker.ServicesMTM, _put_blob_passalong, blob_id_mask=blob_id_mask, min=2)

        return response
    

    def set_publication_tracker_tuple(self, pttuple):
        assert type(pttuple) == type( () )
        assert idlib.is_sloppy_id(pttuple[0])
        assert type(pttuple[1]) == type({})
        self._publication_tracker_tuple = pttuple

import RunTests
mojo_test_flag = 1

import CommStrat


def test_do_you_have_blobs():
    # force our test servers to store everything
    confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"] = SEXP.listToDict([idmasklib.IdMask(bits=0).get_mojodict()])

    # make blob server
    blockstore = localblockstore.BlockStore(path=RunTests.get_next_dir_name(), depth = 1)
    blockstore.init_space()

    doq = DoQ.DoQ()
    doq.shutdown() # Can't run it until we're all set up.
    mtm1 = MojoTransaction.MojoTransactionManager(doq)

    blobserver = BlobServerEGTP([blockstore], mtm1)

    doq.start()
    try:
        # publish a blob to blob server
        blob = 'my blob 1'
        blobsize = len(blob)
        blob_id_1 = idlib.make_ascii_id(blob)
        blobserver.handle_put_blob(None, {'blob id': blob_id_1, 'data': blob})
        blob = 'my blob 2'
        blobsize = len(blob)
        blob_id_2 = idlib.make_ascii_id(blob)
        blobserver.handle_put_blob(None, {'blob id': blob_id_2, 'data': blob})
        blob = 'my blob 3'
        blobsize = len(blob)
        blob_id_3 = idlib.make_ascii_id(blob)
        blobserver.handle_put_blob(None, {'blob id': blob_id_3, 'data': blob})

        rogue_id = idlib.make_ascii_id('spam') 
        has = blobserver.handle_do_you_have_blobs(None, [blob_id_2, blob_id_3, rogue_id])
        assert blob_id_1 not in has
        assert blob_id_2 in has
        assert blob_id_3 in has
        assert rogue_id not in has

        has = blobserver.handle_do_you_have_blobs(None, [rogue_id])
        assert has == []

    finally:
        blobserver.time_to_stop = 1
        mtm1.shutdown()
        doq.shutdown()



def run_test_instead():
    import RunTests
    RunTests.runTests("BlockServerEGTP")

#### this runs if you import this module by itself
if __name__ == '__main__':
    run_test_instead()
