#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# This implements content tracker server EGTP message handlers.
#
__cvsid = '$Id: ContentTrackerEGTP.py,v 1.23.2.1 2001/06/12 03:21:23 nejucomo Exp $'

### standard modules
import traceback  #DEBUG#
import os
import sys

# our modules
import ContentVerifier
import MojoMessage
from MojoConstants import true, false
import SEXP
from confutils import confman, integerizePrice
import contentdb
import contentlib
import debug
import mojoutil
import MojoXML


class ContentTrackerEGTP:
    """
    Implementation of a content tracker running over EGTP.
    Instantiate the object to start running a content tracker.
    """
    def __init__(self, typedef_repository, broker):
        """
        contentdbobj - Should be a contentdb.ContentTrackerDB object
        typedef_repository - Should be a contentlib.ContentTypeDefRepository object
        """
        self.contentdb = setup_content_tracker_db(confman)
        self.typedef_repository = typedef_repository

        self.rescan_content_types()

        receive_funcs = {
            'content tracker lookup': (self.handle_content_tracker_lookup, self.price_callback_for_lookup),
            'content tracker submit': (self.handle_content_tracker_submit, self.price_callback_for_submit),
            # disabled, its never used
            ###'download content types': (self.handle_download_content_types, self.price_callback_for_downloadtypes),
            }

        self.service_announcement = {
            'type' : "content tracker",
            'content types': [],  # XXX not used, going away
            'lookup price': integerizePrice(confman.dict["CONTENTTRACKER"]["PRICES"]["LOOKUP"]),
            'submit price': integerizePrice(confman.dict["CONTENTTRACKER"]["PRICES"]["SUBMIT"]),
            'download content types price': integerizePrice(confman.dict["CONTENTTRACKER"]["PRICES"]["DOWNLOAD_TYPES"]),
            # tell clients that we do support smart keyword with logical OR ' ' and AND '+' queries
            'supports smart queries': 'yes',
        }

        broker.ServicesMTM.add_handler_funcs(receive_funcs)
        broker.ServicesMTM.add_announced_services([self.service_announcement])

        self.verifier = ContentVerifier.ContentVerifier(broker, self.contentdb)
        broker.ServicesMTM._doQ.add_task(self.verifier.begin_cycle)


    def __del__(self):
        if hasattr(self, 'verifier') and self.verifier is not None:
            self.verifier.shutdown()
            del self.verifier
        if hasattr(self, 'contentdb') and self.contentdb is not None:
            self.contentdb.close()
            del self.contentdb

    def shutdown(self):
        # XXX be sure to unregister handler functions with the MTM before this
        self.MyMTM = None
        # the content verifier must be shutdown before contentdb as it
        # has a database within contentdb's db environment.
        if self.verifier is not None:
            self.verifier.shutdown()
            self.verifier = None
        if self.contentdb is not None:
            self.contentdb.close()
            self.contentdb = None
        self.typedef_repository = None

    def rescan_content_types(self) :
        """
        Register any known content types with our Db.
        Returns boolean - true if any content types were loaded.
        """
        # (re)load all known content types
        know_some_types = 0
        for typename in self.typedef_repository.known_type_list() :
            # TODO allow typedef_repository to return the raw XML or allow contentdb to register
            # using a pre-parsed dictionary.  (to avoid this kludge)
            try:
                self.contentdb.RegisterTypeDef(self.typedef_repository.get_type_obj(typename))
                know_some_types = 1
            except contentdb.ContentDBError, e:
                debug.stderr.write('ContentDbError in RegisterType for "%s": %s\n' % (typename, `e`))
        
        if not know_some_types :
            debug.stderr.write('WARNING: no Content Type Descriptions found\n')
        
        return know_some_types


    def price_callback_for_submit(self, widget, incoming_type, incoming_body) :
        return int(self.service_announcement['submit price'])

    def price_callback_for_lookup(self, widget, incoming_type, incoming_body) :
        return int(self.service_announcement['lookup price'])
    
    def price_callback_for_downloadtypes(self, widget, incoming_type, incoming_body) :
        return int(self.service_announcement['download content types price'])


    def get_id(self):
        return self.myMTM.get_id()


    def handle_download_content_types(self, widget, msgbody):
        return self.typedef_repository.get_typedefxml_list()


    def handle_content_tracker_lookup(self, widget, msgbody):
        # content tracker searches
        try:
            # a date range to search can optionally be specified
            first_iso_date=msgbody.get('first ISO date')
            last_iso_date=msgbody.get('last ISO date')

            results = self.contentdb.Search(msgbody['XML data'], first_iso_date=first_iso_date, last_iso_date=last_iso_date)

            # return the matching items
            if results:
                resultdict = results
                respbody = {'result': "success", 'list': resultdict, 'verifier': "yes"}
            else:
                respbody = {'result': "no match", 'verifier': "yes"}

        except (contentdb.ContentXMLError, MojoXML.XMLParseError), e:
            respbody = {'result': "failure", 'reason': "xml parse error"}
            debug.mojolog.write("ERROR parsing XML: %s\n%s\n", args=(e, msgbody['XML data']), v=3, vs='ContentTracker')
        except contentdb.ContentTrackerError, description:
            respbody = {'result': "failure", 'reason': 'content tracker error: "%s"' % str(description)}
        except:
            respbody = {'result': "failure", 'reason': 'content tracker error' }
            debug.mojolog.write('Caught this exception in handle_content_tracker_lookup:\n')
            traceback.print_exc(file=debug.mojolog)

        return respbody
    # end of `def handle_content_tracker_lookup() ...'


    def handle_content_tracker_submit(self, widget, msgbody):
        # content tracker submissions
        try:
            mojoid = self.contentdb.Put(msgbody['XML data'], pub_iso_date=mojoutil.iso_utc_time())

            respbody = {'result': "success", 'mojoid': mojoid}
        except (contentdb.ContentXMLError, MojoXML.XMLParseError), e:
            respbody = {'result': "failure", 'reason': "xml parsing error"}
        except contentdb.ContentTrackerError, description:
            respbody = {'result': "failure", 'reason': 'Content Tracker Error: "%s"' % str(description)}
            debug.stderr.write('Caught this exception in handle_content_tracker_submit:\n')
            traceback.print_exc(file=debug.stderr)
        except:
            respbody = {'result': "failure", 'reason': 'Unknown Content Tracker Error'}
            debug.stderr.write('Caught this exception in handle_content_tracker_submit:\n')
            traceback.print_exc(file=debug.stderr)

        return respbody
    # end of `def handle_content_tracker_submit() ...'


########################################################################
def setup_content_tracker_db(conf) :
    """Given a config object, create/open our content tracking database and return the Db object"""
    return contentdb.ContentTrackerDB(os.path.expandvars(conf.dict["PATH"]["CONTENT_TRACKER_DB_DIR"]), name="contenttracking.db")


