#!/usr/bin/env python
#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
#
# An id->conn_strat mapper.
#

# standard modules
import copy
import string
import re
import traceback
import time
import whrandom

# our modules
from Cache import Cache, StatsCache
import CommStrat
import DoQ
import MojoConstants
from MojoConstants import true, false
import MojoMessage
import MojoTransaction
import OurMessages
import Paytool
import RunTests
import SEXP
from confutils import confman, integerizePrice
import debug
from human_readable import hr
import idlib
import idmasklib
import ipaddresslib
import mojosixbit
import mojoutil
import randinit
import MojoErrors
import mencode   # we want PreEncodedThing
import loggedthreading

_MIN_AUTOEXPIRE_TIMEOUT = 60*10+5   # 10 minutes is getting pretty low!


class IdTrackerEGTP:
    def __init__(self, MTM):

        # setup auto expiration of data
        self.__autoexpiretimeout = int(confman.dict['METATRACKER']["HELLO_INFO_TIMEOUT_SECS"])
        if self.__autoexpiretimeout < _MIN_AUTOEXPIRE_TIMEOUT:
            self.__autoexpiretimeout = _MIN_AUTOEXPIRE_TIMEOUT
            debug.stderr.write('WARNING: HELLO_INFO_TIMEOUT_SECS raised to a minimum of' + `_MIN_AUTOEXPIRE_TIMEOUT`)
            confman.dict['METATRACKER']["HELLO_INFO_TIMEOUT_SECS"] = str(_MIN_AUTOEXPIRE_TIMEOUT)
            confman.save()

        statscache_autoexpire_params = {
            'autoexpireinterval': int(confman.dict['METATRACKER']["HELLO_CACHE_CLEANUP_INTERVAL_SECS"]),
            'autoexpireparams': {'maxage': self.__autoexpiretimeout }
        }

        # our message handler setup
        receive_funcs = {
            'hello': (self.handle_hello, self.price_callback_for_hello),
            'goodbye': (self.handle_goodbye, self.price_callback_for_goodbye),
            'lookup contact info': (self.handle_look_up_contact_info, self.price_callback_for_look_up_contact_info),
            'list blob servers': (self.handle_list_blob_servers, self.price_callback_for_list_services),
            'list publishing blob servers': (self.handle_list_publishing_blob_servers, self.price_callback_for_list_services),
            'list content trackers': (self.handle_list_content_trackers, self.price_callback_for_list_services),
            'list meta trackers': (self.handle_list_meta_trackers, self.price_callback_for_list_services),
            'list relay servers v2': (self.handle_list_relay_servers, self.price_callback_for_list_services),
            'list publication agents': (self.handle_list_publication_agents, self.price_callback_for_list_services),
        }
        self.receive_func_names = receive_funcs.keys()

        self.service_announcement = {
            'type': "meta tracker",
            'hello price': integerizePrice(confman.dict["METATRACKER"]["PRICES"]["HELLO"]),
            'lookup contact info price': integerizePrice(confman.dict["METATRACKER"]["PRICES"]["LOOKUP"]),
            'list servers price': integerizePrice(confman.dict["METATRACKER"]["PRICES"]["LIST_SERVICES"]),  # XXX specify price for each bitmask length (more bits = cheaper price because we return less data)
            'seconds until expiration': str(self.__autoexpiretimeout),
        }

        self._mtm = MTM

        # Map from key id to connection strategy dicts for ids that
        # have not provided us with info about what services they run.
        self.__keyid_to_conn_map = apply(StatsCache, (), statscache_autoexpire_params)

        # If a keyid is in the following maps, it means that service
        # has been announced by that key and includes info about
        # the service in a service specific format (documented
        # in the comments in OurMessages.py).  In the future this
        # could easily be signed information for verification that the
        # meta tracker isn't lying (or that someone tried to fake it
        # out).
        self.__keyid_service_maps = {}
        self.__keyid_service_maps['blob server'] = apply(StatsCache, (), statscache_autoexpire_params)
        self.__keyid_service_maps['publishing blob server'] = apply(StatsCache, (), statscache_autoexpire_params)
        self.__keyid_service_maps['content tracker'] = apply(StatsCache, (), statscache_autoexpire_params)
        self.__keyid_service_maps['meta tracker'] = apply(StatsCache, (), statscache_autoexpire_params)
        self.__keyid_service_maps['relay server v2'] = apply(StatsCache, (), statscache_autoexpire_params)
        #self.__keyid_service_maps['relay server'] = apply(StatsCache, (), statscache_autoexpire_params)  # XXX only to support old 0.81 format hellos
        self.__keyid_service_maps['publication agent'] = apply(StatsCache, (), statscache_autoexpire_params)

        # this caches responses to 'list XXX' done by __handle_list_anyservice so that mencode.py can memoize
        # mencoding to **drastically** reduce the server load.
        # contains: tuples of (time.time() this result was generated, {dict response})
        self._list_anyservice_cache_map = {}
        self._list_anyservice_cache_map['blob server'] = (0,{'result': 'impossible meta tracker error'})
        self._list_anyservice_cache_map['publishing blob server'] = (0,{'result': 'impossible meta tracker error'})
        self._list_anyservice_cache_map['content tracker'] = (0,{'result': 'impossible meta tracker error'})
        self._list_anyservice_cache_map['meta tracker'] = (0,{'result': 'impossible meta tracker error'})
        self._list_anyservice_cache_map['relay server v2'] = (0,{'result': 'impossible meta tracker error'})
        self._list_anyservice_cache_map['publication agent'] = (0,{'result': 'impossible meta tracker error'})

        self._mtm.add_handler_funcs(receive_funcs)
        self._mtm.add_announced_services([self.service_announcement])

        # Keep track of our own contact/service info
        self.reload_own_hello_info()

        pass # end `def __init__() ...'
    
    def shutdown(self):
        self._mtm.remove_announced_services(self.service_announcement.keys())
        self._mtm.remove_handler_funcs(self.receive_func_names)

    def reload_own_hello_info(self):
        """
        Load our Meta/Id tracker data with new info about ourselves
        from the MTM.  (Use this to add more announced services)
        """
        if self._mtm.allow_hello_to_be_sent:
            self.handle_hello(widget=MojoTransaction.Widget(self._mtm.get_id()), msgbody=self._mtm._get_our_hello_msgbody() )
        else:
            # wait for the mtm to become ready and try again
            self._mtm._doQ.add_task(self.reload_own_hello_info, delay=5)


    def price_callback_for_list_services(self, widget, incoming_type, incoming_body):
        return int(self.service_announcement['list servers price'])

    def price_callback_for_look_up_contact_info(self, widget, incoming_type, incoming_body):
        return int(self.service_announcement['lookup contact info price'])
    
    def price_callback_for_hello(self, widget, incoming_type, incoming_body):
        return int(self.service_announcement['hello price'])

    price_callback_for_goodbye = price_callback_for_hello


    def __handle_list_anyservice(self, msgbody, servicename, widget):
        """
        Generic function for returning information about Ids that we
        track as running a given service (specified in servicename).
        """
        canning_delay = int(confman.dict['METATRACKER']['SECS_BEFORE_RECANNING_RESPONSES'])
        canned_response_tuple = self._list_anyservice_cache_map.get(servicename)
        if canned_response_tuple:
            born_on_date, canned_response = canned_response_tuple
            now = time.time()
            if (born_on_date + canning_delay) > now:
                return canned_response
            else:
                # NOTE: we stuff the -old- canned response back in the list_anyservice_cache_map here with
                # a timestamp not destined to expire for 100 more seconds to prevent any other threads (which
                # shouldn't ever happen thanks to the DoQ) from getting here and also starting a new canning
                # process.  We stuff the newly canned response into this tuple at the end of this function. -greg
                put_in_can = 1
                self._list_anyservice_cache_map[servicename] = ( (now-canning_delay+100), canned_response)
                debug.mojolog.write('Creating a new canned response listing "%s"s\n', args=(servicename,), v=3, vs='metatracker')
        else:
            put_in_can = None
            debug.mojolog.write("WARNING: __handle_list_service called for %s which has no canned response map\n" % `servicename`, v=3, vs="metatracker")

        service_dict = self.__keyid_service_maps[servicename]

        filtermasklist = []
# NOTE: filtermask's are ignored now that we are canning responses for speed!
#        if msgbody.has_key('filter mask list'):
#            for filtermask in filtermasklist:
#                MojoMessage.checkTemplate(filtermask, OurMessages.IDMASK_TEMPL)
#            filtermasklist.extend(msgbody['filter mask list'])
#        elif msgbody.has_key('filter mask'):
#            MojoMessage.checkTemplate(msgbody['filter mask'], OurMessages.IDMASK_TEMPL)
#            filtermasklist.append(msgbody['filter mask'])

        result_list = []

        # get our list of things, randomly rotate the list (so that we
        # expose brokers to a new set of servers in each canned response)
        service_dict_items = service_dict.items()
        if len(service_dict_items) > 0:
            rotation = whrandom.randint(0, len(service_dict_items)-1)
            service_dict_items = service_dict_items[rotation:] + service_dict_items[:rotation]

        # XXX hack needed for stability and proper block distribution; always list
        # 'broadcasting' publication agents (otherwise the list overflows so
        # the broadcaster often isn't in the list; that is "a bad thing")
        if servicename == 'publication agent':
            front_items = []
            for keyid, service_info in service_dict_items:
                if service_info['type'] == 'publication agent':
                    # note: this interacts nicely with the hack below that disallows
                    # anyone but us to advertise themselves as a publication broadcaster
                    # until we have better support for that.
                    if service_info['allows store location information'] == 'yes':
                        front_items.append((keyid, service_info))
            # XXX this can end up with duplicate entries in the returned list for the
            # broadcaster; no big deal right now as there is only one/few of these.
            service_dict_items = front_items + service_dict_items

        for keyid, service_info in service_dict_items:
            try:
                connstrats = self.__keyid_to_conn_map[keyid]
            except KeyError:
                # skip this key, we don't have its contact info
                # (which presumably means another thread is in the
                # process of deleting it from all of our id maps)
                debug.stderr.write("IdTrackerEGTP.__handle_list_anyservice - Service Name <%s> - No contact info for Key Id <%s>!\n" % (servicename, mojosixbit.b2a(keyid)[:3]))
                continue

            # if present, check service_info['mask list'] for a match on the mask
            # ranges we were asked about.
            if 0 and filtermasklist and service_info.has_key('mask list'):  # XXX note the 0 to disable mask support
                passed_mask_check = None
                for service_mask in service_info['mask list']:
                    # see if any of the filter masks fall within the service masks
                    debug.stderr.write(
                        "IdTrackerEGTP.__handle_list_anyservice - Service Name <%s> - Comparing filter_masks to our service_mask %s.\n" % (servicename, `service_mask`))
                    for filter_mask in filtermasklist :
                        filter_mask = idmasklib.IdMask(mojodict=filter_mask)
                        if filter_mask.is_subset_of(idmasklib.IdMask(mojodict=service_mask)) :
                            debug.stderr.write("IdTrackerEGTP.__handle_list_anyservice - Service Name <%s> - filter_mask %s is within our mask.\n" % (servicename, `filter_mask`))
                            passed_mask_check = 1
                            break
                        else:
                            debug.stderr.write("IdTrackerEGTP.__handle_list_anyservice - Service Name <%s> - filter_mask %s is excluded by our mask.\n" % (servicename, `filter_mask`))
                    # if at least one mask matched, stop checking
                    if passed_mask_check :
                        break
                # this server didn't match any of the requested masks,
                # skip it; the requestor doesn't want this one.
                if not passed_mask_check :
                    continue

            try:
                # return all services that this key provides, starting
                # explicitly with the one we were asked to lookup
                services_list = [service_info]

                for mapname, mapdict in self.__keyid_service_maps.items():
                    # note: all mapdicts include a 'type' field so we don't
                    # need to explicitly include the type name again here.

                    # XXX hack! XXX 'publishing blob server' is excluded here because it is always a subset of
                    # 'blob server' and was causing two blob server entries to always be returned.
                    if servicename == 'publishing blob server' :
                        service_to_exclude = 'blob server'
                    else :
                        service_to_exclude = 'publishing blob server'

                    if mapname != servicename and mapname != service_to_exclude and mapdict.has_key(keyid):
                        #debug.stderr.write("appending %s to services_list\n" % repr(mapdict[keyid])) # verbose
                        services_list.append(mapdict[keyid])

                itemdict = {
                    'connection strategies': connstrats,
                    'services': services_list,
                }
                result_list.append(itemdict)

                # don't allow us to return too many items
                if len(result_list) >= int(confman.dict['METATRACKER']['MAX_QUERY_RESULTS_RETURNED']):
                    debug.mojolog.write("NOTE: hit MAX_QUERY_RESULTS_RETURNED in __handle_list_anyservice\n", v=2, vs="metatracker")
                    break
            finally:
                pass

        if put_in_can:
            def can_new_response_and_respond(MTM=self._mtm, cache_map=self._list_anyservice_cache_map, servicename=servicename, result_list=result_list, orig_msgid=widget._firstmsgId):
                mencoded_thing = mencode.PreEncodedThing(result_list)
                cache_map[servicename] = (time.time(), mencoded_thing)
                debug.mojolog.write('New canned response stored for "%s"s (%s bytes uncomp)\n', args=(servicename,len(mencoded_thing)), v=3, vs='metatracker')
                MTM.respond_with(orig_msgid, mencoded_thing)

            # we start the canning with an async response in another thread since it can take a while (several seconds) so it shouldn't be on the DoQ
            t = loggedthreading.LoggedThread(target=can_new_response_and_respond)
            t.setDaemon(1)
            t.start()
            return MojoTransaction.ASYNC_RESPONSE # -1 # `ASYNC_RESPONSE' means I'll provide the response later. (by calling `respond_with()')
        else:
            return result_list


    def handle_list_blob_servers(self, widget, msgbody):
        return self.__handle_list_anyservice(msgbody, 'blob server', widget)

    def handle_list_publishing_blob_servers(self, widget, msgbody):
        return self.__handle_list_anyservice(msgbody, 'publishing blob server', widget)

    def handle_list_content_trackers(self, widget, msgbody):
        return self.__handle_list_anyservice(msgbody, 'content tracker', widget)

    def handle_list_meta_trackers(self, widget, msgbody):
        return self.__handle_list_anyservice(msgbody, 'meta tracker', widget)

    def handle_list_relay_servers(self, widget, msgbody):
        return self.__handle_list_anyservice(msgbody, 'relay server v2', widget)

    def handle_list_publication_agents(self, widget, msgbody):
        return self.__handle_list_anyservice(msgbody, 'publication agent', widget)


    def handle_goodbye(self, widget, msgbody):
        counterparty_id = widget.get_counterparty_id()
        if self.__keyid_to_conn_map.has_key(counterparty_id):
            del self.__keyid_to_conn_map[counterparty_id]
        for servmap in self.__keyid_service_maps.values():
            if servmap.has_key(counterparty_id):
                del servmap[counterparty_id]
        return MojoTransaction.NO_RESPONSE

    def handle_hello(self, widget, msgbody):
        counterparty_id = widget.get_counterparty_id()

        connStrats = msgbody['connection strategies']
        services = msgbody.get('services', [])

        if len(connStrats) < 1 :
            return {'result': "failure", 'reason': "no connection strategies"}

        debug.mojolog.write("id %s, platform %s, version %s, cstype: %s\n", args=(counterparty_id, msgbody.get('platform'), msgbody.get('broker version'), connStrats[0].get('lowerstrategy', {}).get('comm strategy type')), v=1, vs="HelloStats")

        # Now vet each strategy.  If it is a Crypto, its public key must
        # match the counterparty of this message.
        # If a crypto has a lower-level TCP strat, it must be a routable
        # IP address.
        for c in connStrats :
            try:
                csobj = CommStrat.dictToStrategy(c)

                if isinstance(csobj, CommStrat.Crypto) :
                    # make sure this is the pub key that signed this hello message
                    if not idlib.equal(counterparty_id, idlib.make_id(csobj._pubkey, 'broker')):
                        return {'result': "failure", 'reason': "public key in comm strat differs from public key that signed message" }

                if isinstance(csobj._lowerstrategy, CommStrat.TCP) :
                    if not RunTests.is_testing():
                        iptocheck = str(csobj._lowerstrategy.host)
                        if ipaddresslib.bad_address_re.match(iptocheck) or not ipaddresslib.valid_ipaddr_re.match(iptocheck) :
                            return {'result': "failure", 'reason': "bad or non-routable IP address (%s)" % iptocheck}
            except (MojoErrors.MojoMessageError, MojoErrors.BadFormatError, KeyError, CommStrat.UnsupportedTypeError): ## XXX ? MojoMessageError and KeyError might come from where?  --Zooko 2000-08-02
                # Okay we don't know how to vet this one.
                continue

        try:
            # ensure that all of the service descriptions given to us were valid
            for service in services:
                if service is None:
                    continue
                MojoMessage.checkTemplate(service, OurMessages.HELLO_SERVICE_TEMPL)
                MojoMessage.checkTemplate(service, OurMessages.HELLO_SERVICE_TEMPL_TYPE_MAP[service['type']])
                if service.has_key('mask list'):
                    if len(service['mask list']) > 10:
                        debug.stderr.write('IdTrackerEGTP - mask list too long in hello, hello dropped')
                        return {'result': "failure", 'reason': "mask list must contain 10 or fewer masks"}
        except (MojoErrors.MojoMessageError, MojoErrors.BadFormatError, KeyError):
            debug.stderr.write('IdTrackerEGTP - hello failed template check:' + hr(msgbody))
            traceback.print_exc(file=debug.stderr)
            return {'result': "failure", 'reason': "failed template check"}

        # create a single tuple to be referenced from the maps
        self.__keyid_to_conn_map[counterparty_id] = connStrats

        # remove all current service map entries for this server; we're registering new ones
        for servmap in self.__keyid_service_maps.values():
            if servmap.has_key(counterparty_id):
                del servmap[counterparty_id]

        # store any service specific information (only for service types that we support)
        services_to_remove = []
        for service in services:
            if service['type'] == 'blob server':
                self.__keyid_service_maps['blob server'][counterparty_id] = service
                if string.lower(service['publishing allowed']) == 'yes':
                    self.__keyid_service_maps['publishing blob server'][counterparty_id] = service
            elif service['type'] == 'publication agent':  # XXX see comment with services_to_remove loop below
                if confman.dict.get("IGNORE_BAD_IP_ADDRS") == "testing" :
                    # don't enable this check during testing
                    mn_addr_re = re.compile('.*')
                else :
                    mn_addr_re = re.compile(r'^64\.71\.128\.')
                try:
                    if connStrats[0]['lowerstrategy'].has_key('IP address') and mn_addr_re.match(connStrats[0]['lowerstrategy']['IP address']) :
                        # its one of our servers, keep it
                        debug.mojolog.write("keeping mojonation.net publication agent unaltered %s\n", args=(counterparty_id,), v=2, vs='MetaTracker')
                        self.__keyid_service_maps['publication agent'][counterparty_id] = service
                    else:
                        # XXX this is to prevent anyone else from running a top-level
                        # publication broadcasting tracker for now - it will go 
                        # away when we remove that restriction
                        if service.has_key('mask'):
                            service['allows store location information'] = 'no'
                            self.__keyid_service_maps['publication agent'][counterparty_id] = service
                        else:
                            # don't allow old (non-mask based) publication trackers
                            services_to_remove.append(service)
                except:
                    debug.stderr.write("this shouldn't be a fatal problem, but we're worried it might be a NameError or ImportError, so we're printing it out for now")
                    traceback.print_exc(file=debug.stderr)
                    services_to_remove.append(service)
            else:
                # It's kosher for us to use service['type'] as the
                # maps index here now that it passed the template
                # check above.  that means it must be one of the
                # templated services in HELLO_SERVICE_TEMPL_TYPE_MAP.
                self.__keyid_service_maps[ service['type'] ][counterparty_id] = service

        # XXX this is used solely for removing top level 'publication agent' (broadcaster) service announcements
        # until we have better support for allowing redundant ones.   -greg 
        for service in services_to_remove :
            debug.mojolog.write("NOTE: disallowing this service from %s: %s\n", args=(counterparty_id, service), v=2, vs='MetaTracker')
            services.remove(service)

        # TODO in the future, we could implement varying expiration times based on how much was paid?

        return {'result': "success", 'seconds until expiration': str(self.__autoexpiretimeout)}


    def handle_look_up_contact_info(self, widget, msgbody):
        keyId = mojosixbit.a2b(msgbody['key id'])

        conn_strat_list_from_map = self.__keyid_to_conn_map.get(keyId)
        if not conn_strat_list_from_map:
            debug.stderr.write("WARNING: failure to lookup contact info -- unknown key id %s\n" % `msgbody['key id']`)
            return { 'result': "failure" }

        # find all services that this key provides
        services_list = []
        for mapdict in self.__keyid_service_maps.values():
            # note: all mapdicts include a 'type' field so we don't
            # need to explicitly include the type name again here.
            if mapdict.has_key(keyId):
                services_list.append(mapdict[keyId])

        result_dict = {'result': "success", 'connection strategies': conn_strat_list_from_map}

        # return a list of all service descriptions for this keyid
        if services_list:
            result_dict['services'] = services_list

        # TODO return a 'seconds until expiration' value to give any requesting client
        # an idea of how long to keep it in their own cache.  (also todo with this
        # is implement an expiration time based cache on the client) -greg 27/jun/2000

        return result_dict
        # end `def handle_look_up_contact_info() ...'

################################ Tests ################################

mojo_test_flag = 1

def test_still_needs_to_be_written():
    pass

# TODO write tests to make sure we handle all hello messages properly
# as well as handling additional hello's from the same Id, etc.
#  -greg 22/jun/2000

################################ This is for running a Root Id/Meta Tracker ################################

def run():
    import sys

    ID_TRACKER_PORT=None
    if len(sys.argv) > 1:
        ID_TRACKER_PORT=sys.argv[1] 

    if len(sys.argv) > 2:
        reconst_string = sys.argv[2]
    else:
        reconst_string = confman.dict.get("ROOT_META_TRACKER_PRIVATE_KEY")

    if not reconst_string or not ID_TRACKER_PORT :
        debug.mojolog.write("\n")
        debug.mojolog.write("You must specify a port and serialized ID Tracker MTM on the command line\n")
        debug.mojolog.write("to run a Root Meta Tracker.\n")
        sys.exit(1)

    randinit.randinit()

    doQ = DoQ.DoQ() # The DoQ for this Id Tracker.
    MTM = MojoTransaction.MojoTransactionManager(doQ=doQ, serialized=reconst_string, pt=None, listenport=int(ID_TRACKER_PORT), pickyport=false, recoverdb=false, allow_send_metainfo=false)
    pt = Paytool.Paytool(tsmtm=MTM)
    MTM.set_pt(pt)
    
    myIdTracker = IdTrackerEGTP(MTM)

    debug.mojolog.write("I am a Root Id/Meta Tracker\n")

    import MojoMessage
    from confutils import BROKER_VERSION_STR
    debug.stderr.write("protocol version: %s\n" % MojoMessage.CURRENT_MOJO_VER)
    debug.stderr.write("broker version: %s\n" % BROKER_VERSION_STR)
    try:
        import testres
        debug.stderr.write("unit tested CVS version: %s\n" % testres.cvsId)
    except:
        pass

    debug.mojolog.write("Here is my public key:")
    debug.mojolog.write(myIdTracker._mtm.get_public_key())
    debug.mojolog.write("Here is my public key id:")
    pkId = idlib.string_to_id(myIdTracker._mtm.get_public_key())
    debug.mojolog.write(mojosixbit.b2a(pkId))
    debug.mojolog.write("Here is my comm strategy:\n")
    debug.mojolog.write(myIdTracker._mtm.get_comm_strategy().toDict())
    debug.mojolog.write("\n")


    debug.mojolog.write("Joining Server Thread.\n")
    myIdTracker._mtm.server_thread_join()
    debug.mojolog.write("Server Thread Joined.\n")




#### this runs if you import this module by itself
if __name__ == '__main__':
    run()

