#
#  Copyright (c) 2000 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#

from copy import copy
import threading
import types

import RunTests
import SEXP
from confutils import confman
import idlib
from idmasklib import IdMask
from mojoutil import b2a
from MojoConstants import true, false
from OurMessages import IDMASK_TEMPL
from DataTypes import ASCII_ARMORED_DATA, UNIQUE_ID

request_publication_info_template = {
    'mask' : IDMASK_TEMPL,
    'begin index' : 0,
    'nonce' : ASCII_ARMORED_DATA}

here_is_pub_info_template = {
    'matches' : ListMarker({
        'blob id' : UNIQUE_ID,
        'server id' : UNIQUE_ID}),
    'last index' : 0,
    'more left' : ['yes', 'no']}

publication_reporting_template = {
    'blob id' : UNIQUE_ID,
    'server id' : UNIQUE_ID}

publication_reporting_template_response = None

wrong_nonce_template = {
    'current nonce' : UNIQUE_ID, 
    'last index' : 0}

request_nonce_template = None

here_is_nonce_template = {
    'nonce' : ASCII_ARMORED_DATA,
    'last index' : 0}

class PublicationAgentHandlers :
    """
    Handles server side messages relating to where blobs are located
    """
    def __init__(self, number_to_store = 10000) :
        # place all info is stored in - most recent at the end, oldest at the beginning
        self.queue = []
        self.number_to_store = number_to_store
        self.number_thrown_away = 0
        self.lock = threading.Lock()

        self.service_announcement = {
            'type' : 'publication agent',
            'store location information price' : confman.dict["PublicationAgentHandlers"]["PRICES"]["STORE"],
            'get location information price' : confman.dict["PublicationAgentHandlers"]["PRICES"]["RETRIEVE"],
        }


    def compute_cost_store_location(self, widget, message_type, message_body) :
        return int(self.service_announcement['store location information price'])

    def compute_cost_get_location(self, widget, message_type, message_body) :
        return int(self.service_announcement['get location information price'])


    def get_handlers_dict(self) :
        dict = {}
        dict['store location information'] = (self.store_location_information, self.compute_cost_store_location)
        dict['get location information'] = (self.get_location_information_for_mask, self.compute_cost_get_location)
        return dict
        
    def get_services_dict(self) :
        return self.service_announcement
        
    def store_location_information(self, widget, message_body) :
        self.lock.acquire()
        try :
            links = message_body['locations']
            for link in links :
                if not idlib.is_sloppy_id(link['blob id']) :
                    return {'result': 'failure', 'reason': 'invalid blob id'}
                new_link = Link(blob_id = link['blob id'], blob_server_id = link['server id'], blob_size=link['bytes'])
                if len(self.queue) == 0 :
                    self.queue.append(new_link)
                else :
                    self.queue[-1].next = new_link
                    self.queue.append(new_link)
                    # if stored too many, drop some
                    if len(self.queue) > self.number_to_store :
                        del self.queue[0]
                        self.number_thrown_away = self.number_thrown_away + 1
            return {'result' : 'success'}
        finally :
            self.lock.release()
        
    def get_location_information_for_mask(self, widget, message_body) :
        """
        Returns everything matching message_body['mask list'], starting at index message_body['first']
        """
        limit = 0
        first_string = message_body.get('first')
        if type(first_string) is types.StringType :
            try :
                limit = int(first_string)
            except ValueError :
                return {'result': 'failure', 'minimum unsent': '0'}
        mask_list = map(lambda d : IdMask(mojodict = d), message_body['mask list'])
        self.lock.acquire()
        try :
            index = limit - self.number_thrown_away
            if index < 0 :
                index = 0
            if index >= len(self.queue) :
                link = None
            else :
                link = self.queue[index]
        finally :
            self.lock.release()
        # the purpose of Link.first is to make doing this operation outside of a lock safe
        rlist = []
        while link is not None :
            for mask in mask_list :
                if mask.match(link.blob_id) :
                    rlist.append({'blob id' : link.blob_id, 'server id' : link.blob_server_id, 'bytes': link.blob_size})
                    break
            link = link.next
            limit = limit + 1
        if len(rlist) == 0 :
            return {'result' : 'none', 'minimum unsent' : str(limit)}
        return {'result' : 'success', 'locations' : rlist, 'minimum unsent' : str(limit)}

class Link :
    def __init__(self, blob_id, blob_server_id, blob_size) :
        self.next = None
        self.blob_id = blob_id
        self.blob_server_id = blob_server_id
        self.blob_size = blob_size

mojo_test_flag = 1

if __name__ == '__main__' :
    RunTests.runTests('PublicationAgentHandlers')

def test_get_handlers_dict_coverage() :
    h = PublicationAgentHandlers()
    assert h.get_handlers_dict() is not None
    
def test_get_services_dict_coverage() :
    h = PublicationAgentHandlers()
    assert h.get_services_dict() is not None
    
def test_returns_nothing_initially() :
    h = PublicationAgentHandlers()
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()]})
    assert r['result'] == 'none'

def test_has_something_but_missed_by_mask_returns_none() :
    h = PublicationAgentHandlers()
    blob_id = b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask(bits = 160, mask = 'a' * 20).get_mojodict()]})
    assert r['result'] == 'none'
    
def test_has_something_returns_valid() :
    h = PublicationAgentHandlers()
    blob_id = b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()]})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 1
    assert l[0]['blob id'] == blob_id
    assert l[0]['server id'] == blob_server_id

def test_start_at_index_zero_explicitly_with_none_available() :
    h = PublicationAgentHandlers()
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '0'})
    assert r['result'] == 'none'
    
def test_start_at_index_zero_explicitly() :
    h = PublicationAgentHandlers()
    blob_id = b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '0'})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 1
    assert l[0]['blob id'] == blob_id
    assert l[0]['server id'] == blob_server_id
    
def test_start_at_index_one() :
    h = PublicationAgentHandlers()
    blob_id = b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    blob_id2 = b2a(idlib.string_to_id('blob2'))
    blob_server_id2 = idlib.string_to_id('blob server2')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id2, 'server id' : blob_server_id2, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '1'})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 1
    assert l[0]['blob id'] == blob_id2
    assert l[0]['server id'] == blob_server_id2
    
def test_thrown_away_data() :
    h = PublicationAgentHandlers(5)
    for i in range(0, 7) :
        blob_id = b2a(idlib.string_to_id('blob ' + str(i)))
        blob_server_id = idlib.string_to_id('blob server ' + str(i))
        h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '0'})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 5
    assert l[0]['blob id'] == b2a(idlib.string_to_id('blob 2'))
    assert l[0]['server id'] == idlib.string_to_id('blob server 2')
    
def test_returns_none_if_too_large() :
    h = PublicationAgentHandlers(5)
    for i in range(0, 7) :
        blob_id = b2a(idlib.string_to_id('blob ' + str(i)))
        blob_server_id = idlib.string_to_id('blob server ' + str(i))
        h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '10'})
    assert r['result'] == 'none'


### Start a PublicationTrackerHandler:
def run(mtm=None, sendhello=true):
    if mtm is None:
        mtm = MojoTransaction.MojoTransactionManager()
    pth = PublicationAgentHandlers()
    mtm.add_announced_services([pth.get_services_dict()])
    mtm.add_handler_funcs(pth.get_handlers_dict())
    if sendhello : # Don't send hello if you just recently sent a hello with this mtm (for example, from the Broker). -Nathan 2000-07-13
        mtm.send_hello_to_meta_trackers()
    ## mtm.server_thread_join()
    return

if __name__ == '__main__' :
    run()

