#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
__cvsid = '$Id: PublicationTrackerHandlers.py,v 1.33 2001/06/02 00:14:57 zooko Exp $'

# standard modules
from copy import copy
import threading
import types

# our modules
import idmasklib
import RunTests
from idmasklib import IdMask
import SEXP
from confutils import confman, integerizePrice
import idlib
import randsource
import mojosixbit
import mojoutil
import debug
import MetaTrackerLib
import pricelib
from MojoConstants import true, false
import MojoTransaction
from human_readable import hr


debugwrite = debug.mojolog.write


class PublicationTrackerHandlers :
    """
    Handles server side messages relating to where blobs are located
    """
    def __init__(self, mtm, doq, number_to_store = 4000):
        # place all info is stored in - most recent at the end, oldest at the beginning
        self.queue = []
        self.mtm = mtm
        self.doq = doq
        self._last_asked = '0'
        self._last_pt_nonce = None
        self.number_to_store = number_to_store
        self.number_thrown_away = 0
        self.lock = threading.Lock()

        # if we don't already have a mask, get one
        if not confman.dict["PUBLICATIONTRACKERHANDLERS"].has_key('mask'):
            if confman.dict["BLOCKSERVER"].has_key("LIST_OF_BLOCKID_MASKS"):
                # use the same mask as our own block server
                confman.dict["PUBLICATIONTRACKERHANDLERS"]['mask'] = copy(confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"]['0'])
            else:
                # generate a random new mask
                confman.dict["PUBLICATIONTRACKERHANDLERS"]['mask'] = idmasklib.generate_random_mask(bits=4).get_mojodict()
        # if we share the same mask as our blockserver (we should, that's the default); use one less bit than it
        if confman.dict["PUBLICATIONTRACKERHANDLERS"]['mask']['mask'] == confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"]['0'].get('mask'):
            confman.dict["PUBLICATIONTRACKERHANDLERS"]['mask']['bits'] = max(1, int(confman.dict["BLOCKSERVER"]["LIST_OF_BLOCKID_MASKS"]['0']['bits']) - 1)
        self.mask = idmasklib.IdMask(mojodict = confman.dict["PUBLICATIONTRACKERHANDLERS"]['mask'])

        self.service_announcement = {
            'type' : 'publication agent',
            'store location information price': integerizePrice(confman.dict["PUBLICATIONTRACKERHANDLERS"]["PRICES"]["STORE"]),
            'get location information price': integerizePrice(confman.dict["PUBLICATIONTRACKERHANDLERS"]["PRICES"]["RETRIEVE"]),
            'mask': self.mask.get_mojodict(),
            'allows store location information': confman.dict["PUBLICATIONTRACKERHANDLERS"]['allows store location information']
        }
        self.nonce = mojosixbit.b2a(randsource.get(12)) # plenty of bits (96) for a unique id indicating that we restarted
        if confman.dict["PUBLICATIONTRACKERHANDLERS"]['allows store location information'] == 'no':
            self.doq.add_task(self.poll_broadcasting_publication_tracker, delay = int(confman.dict["PUBLICATIONTRACKERHANDLERS"]['poll delay']))

    def shutdown(self):
        self.mtm.remove_announced_services(self.get_services_dict().keys())
        self.mtm.remove_handler_funcs(self.get_handlers_dict().keys())

    def poll_broadcasting_publication_tracker(self):
        # lookup a publication tracker to use, polling will continue via the callback
        MetaTrackerLib.get_publishing_agents_and_call_back(self.mtm, callback=self._continue_poll_metatrackerlib, min=1, get_broadcasters_only=true)  # XXX make min>1 once we allow more than one broadcaster

    def _continue_poll_metatrackerlib(self, pubtracker_id_to_info_map):
        # this funciton is used as a callback for MetaTrackerLib.get_publishing_agents_and_call_back
        msgtype = 'get location information'
        msgbody = {'mask list': [self.mask.get_mojodict()], 'first': self._last_asked}
        if self._last_pt_nonce :
            msgbody['nonce'] = self._last_pt_nonce
        pts = self.mtm.get_handicapper().sort_by_preference(pubtracker_id_to_info_map.items(), msgtype, msgbody)
        if len(pts) == 0:
            debug.mojolog.write("couldn't find publication tracker broadcaster\n", v=3, vs='PubTracker')
        else:
            pt = pts[0]
            pub_tracker_id, pub_tracker_info = pt
            amounttopay_pub_tracker = pricelib.publication_agent_get_location_info_price(pub_tracker_info, msgbody)
            if amounttopay_pub_tracker is not None:
                def outcome_func(widget, outcome, failure_reason=None, self=self):
                    if failure_reason:
                        # XXX should really have some logic here
                        debug.stderr.write("WARNING got a failure when trying to poll the broadcasting publication tracker: %s\n", args=(failure_reason,), v=2, vs="WARNING")
                        return

                    if outcome.get('result') == "success":
                        self._last_asked = outcome['minimum unsent']
                        self._last_pt_nonce = outcome.get('nonce')
                        self.store_location_information(None, outcome)

                    debug.mojolog.write("PublicationTrackerHandlers: I finished outcome func from polling pub trackers!\n", v=5, vs="test")

                self.mtm.initiate(pub_tracker_id, msgtype, msgbody, mojooffer=amounttopay_pub_tracker, timeout = 200, outcome_func = outcome_func)
                debug.mojolog.write("PublicationTrackerHandlers: I initiated polling pub trackers!\n", v=5, vs="test")
            else:
                # XXX punt
                debug.stderr.write("WARNING: couldn't figure out how much to pay the broadcasting publication tracker")

        poll_delay = int(confman.dict["PUBLICATIONTRACKERHANDLERS"]['poll delay'])
        if poll_delay < 10 :
            debug.mojolog.write("WARNING: extreemly low publishing broadcaster poll period! Please raise it in your configuration.  Using the minimum of 10 seconds.\n", v=0, vs='PubTracker')
            poll_delay = 10
        self.doq.add_task(self.poll_broadcasting_publication_tracker, delay = poll_delay)

    def compute_cost_store_location(self, widget, message_type, message_body) :
        return int(self.service_announcement['store location information price'])

    def compute_cost_get_location(self, widget, message_type, message_body) :
        return int(self.service_announcement['get location information price'])

    def get_handlers_dict(self) :
        dict = {}
        if confman.dict["PUBLICATIONTRACKERHANDLERS"]['allows store location information'] == 'yes':
            dict['store location information'] = (self.store_location_information, self.compute_cost_store_location)
        dict['get location information'] = (self.get_location_information_for_mask, self.compute_cost_get_location)
        return dict
        
    def get_services_dict(self) :
        return self.service_announcement
        
    def store_location_information(self, widget, message_body):
        # also called internally from the polling function
        self.lock.acquire()
        try :
            links = message_body['locations']
            for link in links :
                new_link = Link(blob_id = link['blob id'], blob_server_id = link['server id'], 
                    blob_size=link['bytes'])
                if len(self.queue) == 0 :
                    self.queue.append(new_link)
                else :
                    self.queue[-1].next = new_link
                    self.queue.append(new_link)
                    # if stored too many, drop some
                    if len(self.queue) > self.number_to_store:
                        del self.queue[0]
                        self.number_thrown_away = self.number_thrown_away + 1
            return {'result' : 'success', 'nonce': self.nonce}
        finally :
            self.lock.release()
        
    def get_location_information_for_mask(self, widget, message_body) :
        """
        Returns everything matching message_body['mask list'], starting at index message_body['first']
        """
        if not message_body.has_key('first'):
            limit = 0
        else:
            limit = int(message_body.get('first'))
        if message_body.has_key('nonce') and message_body.get('nonce') != self.nonce:
            limit = 0
        mask_list = map(lambda d : IdMask(mojodict = d), message_body['mask list'])
        self.lock.acquire()
        try :
            index = limit - self.number_thrown_away
            if index < 0:
                limit = self.number_thrown_away
                index = 0
            if index >= len(self.queue):
                limit = self.number_thrown_away + len(self.queue)
                link = None
            else :
                link = self.queue[index]
        finally :
            self.lock.release()
        rlist = []
        while link is not None :
            for mask in mask_list :
                if mask.match(link.blob_id) :
                    rlist.append({'blob id' : link.blob_id, 'server id' : link.blob_server_id, 'bytes': link.blob_size})
                    break
            link = link.next
            limit = limit + 1
        if len(rlist) == 0:
            return {'result' : 'none', 'minimum unsent' : str(limit), 'nonce': self.nonce}
        return {'result' : 'success', 'locations' : rlist, 'minimum unsent' : str(limit), 'nonce': self.nonce}

class Link :
    def __init__(self, blob_id, blob_server_id, blob_size) :
        self.next = None
        self.blob_id = blob_id
        self.blob_server_id = blob_server_id
        self.blob_size = blob_size

mojo_test_flag = 1

if __name__ == '__main__' :
    RunTests.runTests('PublicationTrackerHandlers')

def test_get_handlers_dict_coverage() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    assert h.get_handlers_dict() is not None
    
def test_get_services_dict_coverage() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    assert h.get_services_dict() is not None
    
def test_returns_nothing_initially() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()]})
    assert r['result'] == 'none'

def test_has_something_but_missed_by_mask_returns_none() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    blob_id = mojosixbit.b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask(bits = 160, mask = 'a' * 20).get_mojodict()]})
    assert r['result'] == 'none'
    
def test_has_something_returns_valid() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    blob_id = mojosixbit.b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()]})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 1
    assert l[0]['blob id'] == blob_id
    assert l[0]['server id'] == blob_server_id

def test_start_at_index_zero_explicitly_with_none_available() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '0'})
    assert r['result'] == 'none'
    
def test_start_at_index_zero_explicitly() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    blob_id = mojosixbit.b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '0'})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 1
    assert l[0]['blob id'] == blob_id
    assert l[0]['server id'] == blob_server_id
    
def test_start_at_index_one() :
    h = PublicationTrackerHandlers(mtm=None, doq=None)
    blob_id = mojosixbit.b2a(idlib.string_to_id('blob'))
    blob_server_id = idlib.string_to_id('blob server')
    blob_id2 = mojosixbit.b2a(idlib.string_to_id('blob2'))
    blob_server_id2 = idlib.string_to_id('blob server2')
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.store_location_information(None, {'locations' : [{'blob id' : blob_id2, 'server id' : blob_server_id2, 'bytes': '23'}]})
    assert r['result'] == 'success'
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '1'})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 1
    assert l[0]['blob id'] == blob_id2
    assert l[0]['server id'] == blob_server_id2
    
def test_thrown_away_data() :
    h = PublicationTrackerHandlers(mtm=None, doq=None, number_to_store=5)
    for i in range(0, 7) :
        blob_id = mojosixbit.b2a(idlib.string_to_id('blob ' + str(i)))
        blob_server_id = idlib.string_to_id('blob server ' + str(i))
        h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '0'})
    assert r['result'] == 'success'
    l = r['locations']
    assert len(l) == 5
    assert l[0]['blob id'] == mojosixbit.b2a(idlib.string_to_id('blob 2'))
    assert l[0]['server id'] == idlib.string_to_id('blob server 2')
    
def test_returns_none_if_too_large() :
    h = PublicationTrackerHandlers(mtm=None, doq=None, number_to_store=5)
    for i in range(0, 7) :
        blob_id = mojosixbit.b2a(idlib.string_to_id('blob ' + str(i)))
        blob_server_id = idlib.string_to_id('blob server ' + str(i))
        h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': '23'}]})
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '10'})
    assert r['result'] == 'none'

def test_checks_nonce() :
    h = PublicationTrackerHandlers(mtm=None, doq=None, number_to_store=10)
    for i in range(0, 7) :
        blob_id = mojosixbit.b2a(idlib.string_to_id('blob ' + str(i)))
        blob_server_id = idlib.string_to_id('blob server ' + str(i))
        h.store_location_information(None, {'locations' : [{'blob id' : blob_id, 'server id' : blob_server_id, 'bytes': str(i)}]})
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '1', 'nonce': h.nonce})
    assert r['nonce'] == h.nonce
    assert r['result'] == 'success' and r['locations'][0]['bytes'] == '1', `r`
    # test wrong nonce
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '1', 'nonce': 'scooby doobie'})
    assert r['result'] == 'success' and r['locations'][0]['bytes'] == '0', `r`
    # test no nonce keeps old behavior
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '3'})
    assert r['result'] == 'success' and r['locations'][0]['bytes'] == '3', `r`
    r = h.get_location_information_for_mask(None, {'mask list' : [IdMask().get_mojodict()], 'first' : '9'})
    assert r['result'] == 'none', `r`

def test_poll_broadcasting_publication_tracker() :
    print 'XXX write a test of poll_broadcasting_publication_tracker'

### Start a PublicationTrackerHandler:
def run(mtm=None):
    if mtm is None:
        mtm = MojoTransaction.MojoTransactionManager()
    pth = PublicationTrackerHandlers(mtm=mtm, doq=mtm._doQ)
    mtm.add_announced_services([pth.get_services_dict()])
    mtm.add_handler_funcs(pth.get_handlers_dict())
    return pth

