#!/usr/bin/env python
#
#  Copyright (c) 2001 Autonomous Zone Industries
#  This file is licensed under the
#    GNU Lesser General Public License v2.1.
#    See the file COPYING or visit http://www.gnu.org/ for details.
#
# This module implements Mojo Nation (TM) content tracking support.
#
__cvsid = '$Id: contentdb.py,v 1.31.2.1 2001/06/12 03:21:23 nejucomo Exp $'

# standard python modules
import re
import sys
import os
import os.path
import time
import whrandom
from cStringIO import StringIO
from confutils import confman

# our own modules
import DoQ
import contentlib
from contentlib import ContentTrackerError, ContentDBError, ContentXMLError, ContentTypeError
from bsddb3 import dbtables, db
from MojoXML import *  # this also imports the saxlib+saxexts for us
import mojoutil
import idlib


MAX_COMPRESSED_BINARY_DINODE_LEN = 8192


# XXX on large databases and searches this code is a problem as we do a full search to hold the result list in
# memory rather than using a cursor/iterator based search as -real- databases do.  We need to change this
# module to use a real database via the Python DB API 2.0 (works with tons of real databases) instead of
# using our poor bsddb3.dbtables implementation.


########################################################################
########################################################################

class ContentTrackerDB :
    """A Mojo Nation (TM) content tracking database.  Methods take XML
    strings as input and parse them as appropriate to store or lookup
    information.
    """
    ####################################################################
    def __init__(self, dbpath, name="contenttracking.db") :
        if not os.path.isdir(dbpath) :
            raise ContentDBError, "Database directory must exist"
        self.dbpath = dbpath
        try:
            self.db = dbtables.bsdTableDB(name, dbpath, create=1, recover=None, dbflags=db.DB_PRIVATE)
        except (dbtables.TableDBError, db.DBError), e:
            # XXX HACK try one more time.  When recover is set,
            # sometimes it raises a DB_INCOMPLETE "sync didn't finish"
            # but works the next time you try and open the DBEnv().
            try:
                self.db = dbtables.bsdTableDB(name, dbpath, create=1, recover=1, dbflags=db.DB_PRIVATE)
            except (dbtables.TableDBError, db.DBError), e:
                raise ContentDBError, e

        # setup the special MojoID -> Dinode mapping table
        mojoid_table_columns = [
            '__DINODE__',
            '__MOJOID__',
            '__LAST_VERIFICATION_RESULT__',
            '__ACCEPTED__',
            ]
        self.db.CreateOrExtendTable('MOJOID', mojoid_table_columns)

        debug.mojolog.write("self.db.ListTables() %s\n" % (self.db.ListTables(),), v=4, vs='contentdb')

    ####################################################################
    def close(self):
        if self.db is not None:
            self.db.close()
            self.db = None

    ####################################################################
    def RegisterTypeXML(self, typexml) :
        """Register a content type with this database using an XML template.
        All tables have implied '__DINODE__', '__MOJOID__', and '__PATH__'
        string fields as that is the entire purpose of the content tracking
        database.
        """
        # parse typexml
        typedochandler = contentlib.ContentTypeDefDocHandler()
        parser = create_mojo_xml_parser(typedochandler)
        try:
            parser.parseFile(StringIO(typexml))
        finally:
            parser.close()
        if len(typedochandler.types) < 1 :
            raise ContentTrackerError, "no types specified"
        for name, fields, sortedfieldnamelist, multifile, automaticfields in typedochandler.types :
            self.RegisterTypeRaw(name, fields, sortedfieldnamelist, multifile, automaticfields)
        
    def RegisterTypeDef(self, typedef):
        self.RegisterTypeRaw(
            typedef.name,
            typedef.fielddict,
            typedef.sortedfieldnamelist,
            typedef.multifile,
            typedef.automaticfieldmap)

    def RegisterTypeRaw(self, name, fields, sortedfieldnamelist, multifile, automaticfields):
        try:
            debug.mojolog.write('registering table name: %s\n', args=(name,), v=3, vs='contentdb')
            for fieldname in fields.keys():
                if fieldname[:2] == '__':
                    raise ContentTrackerError, ("Field names beginning with `__' are reserved for internal use", fieldname, name)
            columns = fields.keys()
            columns.append('__DINODE__')
            columns.append('__MOJOID__')
            columns.append('__PATH__')
            columns.append('__PUB_ISO_DATE__')
            # Invariant for __ACCEPTED__:  Always either "yes", "no", or None.  "yes" and None are synonymous.
            columns.append('__ACCEPTED__')
            self.db.CreateOrExtendTable(name, columns)
        except dbtables.TableDBError, e:
            raise ContentDBError, e.args
        self.db.checkpoint(0)

    ####################################################################
    def DeleteMojoid(self, mojoid):
        assert idlib.is_sloppy_id(mojoid)
        mojoid = idlib.to_ascii(mojoid)
        for table in self.db.ListTables():
            try:
                self.db.Delete(table, {'__MOJOID__': dbtables.ExactCond(mojoid)})
            except (dbtables.TableDBError, db.DBError), e:
                debug.mojolog.write("Got error %s removing mojoid %s from table %s\n", args=(e, mojoid, table), v=8, vs='contentdb')
                pass
    
    ####################################################################
    def AcceptMojoid(self, mojoid, ctype=None):
        """change the internal __ACCEPTED__ flag for a mojoid to 'yes'"""
        try:
            if ctype:
                list_of_tables = [ctype]
            else:
                list_of_tables = self.db.ListTables()
            for table in list_of_tables:
                try:
                    self.db.Modify(table,
                        conditions={'__MOJOID__': dbtables.ExactCond(mojoid)},
                        mappings={'__ACCEPTED__': lambda x: 'yes'})
                except (dbtables.TableDBError, db.DBError), e:
                    pass # this table didn't have __MOJOID__ or __ACCEPTED__ columns, ignore it
        except (dbtables.TableDBError, db.DBError):
            return  # its not in the mojoid table, don't do anything (this will be noticed elsewhere)

    ####################################################################
    def RejectMojoid(self, mojoid, ctype=None):
        """change the internal __ACCEPTED__ flag for a mojoid to 'no'"""
        try:
            if ctype:
                list_of_tables = [ctype]
            else:
                list_of_tables = self.db.ListTables()
            for table in list_of_tables:
                try:
                    self.db.Modify(table,
                        conditions={'__MOJOID__': dbtables.ExactCond(mojoid)},
                        mappings={'__ACCEPTED__': lambda x: 'no'})
                except (dbtables.TableDBError, db.DBError), e:
                    pass # this table didn't have __MOJOID__ or __ACCEPTED__ columns, ignore it
        except (dbtables.TableDBError, db.DBError):
            return  # its not in the mojoid table, don't do anything (this will be noticed elsewhere)

    ####################################################################
    def MarkMojoidMissing(self, mojoid):
        """
        Put a flag in the db saying that this file couldn't be found during its most recent verification run.
        """
        try:
            self.db.Modify('MOJOID',
                conditions={'__MOJOID__': dbtables.ExactCond(mojoid)},
                mappings={'__LAST_VERIFICATION_RESULT__': lambda x: 'missing'})
        except (dbtables.TableDBError, db.DBError):
            return  # its not in the mojoid table, don't do anything (this will be noticed elsewhere)

    ####################################################################
    def UnMarkMojoidMissing(self, mojoid):
        """
        Remove the flag in the db which says that this file couldn't be found during its most recent verification run.
        """
        try:
            self.db.Modify('MOJOID',
                conditions={'__MOJOID__': dbtables.ExactCond(mojoid)},
                mappings={'__LAST_VERIFICATION_RESULT__': lambda x: 'found'})
        except (dbtables.TableDBError, db.DBError):
            return  # its not in the mojoid table, don't do anything (this will be noticed elsewhere)

    ####################################################################
    def WasMojoidMissing(self, mojoid):
        """
        Returns true if the mojoid in question has been verified at least once and
        the most recent ContentVerifier run marked it as missing.
        """
        try:
            mojoid_rows = self.db.Select('MOJOID', ['__LAST_VERIFICATION_RESULT__'], {'__MOJOID__': dbtables.ExactCond(mojoid)})
            if not mojoid_rows:
                return 0
            # TODO deal with (cleanup) things that are in the database multiple times
            return mojoid_rows[0].get('__LAST_VERIFICATION_RESULT__', 'found') == 'missing'
        except (dbtables.TableDBError, db.DBError):
            return 0  # not found at all should never happen

    ####################################################################
    def Search(self, queryxml, accepted_field="yes", first_iso_date=None, last_iso_date=None, trimsubset=1) :
        """
        Search this database using an XML query template.  The
        return value is yet to be defined but will likely be a dict
        of dicts containing the search results.

        accepted_field may be None, "yes", or "no".  If None, the __ACCEPTED__
        field is ignored, else only results matching accepted_field
        (exactly) are returned.

        @param first_iso_date specifies the earliest ISO8601 date string that results may have
        in their publication date field.  (__PUB_ISO_DATE__ column internally).  None means we don't care.

        @param last_iso_date specifies the latest ISO8601 date string that results may have
        in their publication date field.  (__PUB_ISO_DATE__ column internally).  None means we don't care.

        @param trimsubset is a boolean that controls whether or not the result is trimmed due too its length
        if too many results are found.

        @precondition This method must be called on the DoQ.: DoQ.doq.is_currently_doq()
        """
        assert DoQ.doq.is_currently_doq(), "precondition: This method must be called on the DoQ."

        dochandler = contentlib.ContentDescDocHandler()
        parser = create_mojo_xml_parser(dochandler)
        try:
            parser.parseFile(StringIO(queryxml))
        finally:
            parser.close()

        if not dochandler.query :
            raise ContentTrackerError, "ContentDesc did not specify a query"

        # create the Select condition functions
        conditions = {}
        for field, matchstr in dochandler.fieldvalues.items() :
            if not matchstr:
                continue  # skip empty match strings
            if dochandler.smart:
                # exact matches are noted by being surrounded with ^...$
                if matchstr[0] == '^' and matchstr[-1] == '$':
                    if field == '__MOJOID__':
                        # only support exact matches for mojoids.
                        conditions[field] = dbtables.ExactCond(matchstr[1:-1])   # strip off the ^ and $ for use with ExactCond
                        continue
                    match_re = matchstr
                else:
                    if field == '__MOJOID__':
                        # only support exact matches for mojoids.
                        conditions[field] = dbtables.ExactCond(matchstr)
                        continue
                    # Turn the matchstr into a regular expression for a more intelligent "smart" query
                    #   example: "cats+sleeping playing" will turn into "(cats).*(sleeping|playing)"
                    # turn all spaces into logical ORs
                    match_re = string.join(string.split(matchstr), '|')
                    # turn all + signs into order preserving logical ANDs
                    match_re = '(' + string.replace(match_re, '+', ').*(') + ')'

                conditions[field] = re.compile(match_re, re.I).search
            else:
                # the old original poorly done fake sqlish "like" string way
                conditions[field] = dbtables.LikeCond(matchstr)

        # if a date range was given, check the date range.
        if first_iso_date and last_iso_date:
            conditions['__PUB_ISO_DATE__'] = lambda datestr, first=first_iso_date, last=last_iso_date: datestr is not None and (last >= datestr >= first)
        elif first_iso_date:
            conditions['__PUB_ISO_DATE__'] = lambda datestr, first=first_iso_date: datestr is not None and (datestr >= first)
        elif last_iso_date:
            conditions['__PUB_ISO_DATE__'] = lambda datestr, last=last_iso_date: datestr is not None and (datestr <= last)

        if len(conditions) == 0:
            # no conditions?  no worries, this "search" matches everything
            conditions['__MOJOID__'] = dbtables.Cond()

        try:
            matchedrows = self.db.Select(dochandler.typename, None, conditions)
        except dbtables.TableDBError, e:
            raise ContentDBError, e.args

        # don't use __ACCEPTED__ as a Select condition because select doesn't allow conditions to match columns
        # with None [not defined] values as is the case for all existing database entries.
        if accepted_field == "yes":
            # Make sure None is synonymous with "yes" in searches:
            matchedrows = filter(lambda row : row.get('__ACCEPTED__') in ['yes', None], matchedrows)
        elif accepted_field == "no":
            matchedrows = filter(lambda row : row.get('__ACCEPTED__') == 'no', matchedrows)
        # else:  Search for both.   (We suggest None, but sloppily allow any other value.)

        # If the result list is too long, we return a random subset.
        rows = len(matchedrows)
        if rows == 0:
            return []
        start=whrandom.randint(0,rows - 1)
        iterator=start
        maximumlength = int(confman.dict['CONTENTTRACKER']['MAX_RESPONSE_KILOBYTES'])*1024

        # return a list of XML descriptions for each row
        resultlist = []
        totallength = 0
        while 1 :
            # debug.mojolog.write("contentdb.Search(): iterator is %s\n", args=(iterator,), v=1, vs='contentdb')
            row=matchedrows[iterator]
            if row.has_key('__PATH__') and row['__PATH__'] is None :
                del row['__PATH__']   # don't return a path if None was stored
            if not row.has_key('__MOJOID__'):
                # this entry was submitted before we were running a version of the
                # software with mojoid support.  Compute its MOJOID.
                try:
                    asciidinode = row['__DINODE__']
                    mojoid = contentlib.get_mojoid_from_asciidinode(asciidinode)
                    row['__MOJOID__'] = mojoid
                except mojoutil.BadAsciiArmoredDataError:
                    debug.stderr.write("Skipping bad dinode `%s' in contentdb\n" % asciidinode, v=2)
                    # must've been a bad dinode, skip this entry!
                # save the mojoid to the database so that we don't need to do this again
                self.db.Insert(dochandler.typename, row)
                # store it in the mojoid lookup table for easy mojoid->dinode lookups
                self.db.Insert('MOJOID', {'__MOJOID__': mojoid, '__DINODE__': asciidinode})
            # only include it in the returned search result if it is not marked as missing
            if not self.WasMojoidMissing(row['__MOJOID__']):
                xml = contentlib.make_contentdesc_xml(dochandler.typename, row)
                totallength = totallength + len(xml)
                if totallength > maximumlength :
                    break
                resultlist.append(xml)
            iterator = iterator + 1
            if iterator == rows :
                iterator=0
            if iterator == start :
                break

        return resultlist

    ####################################################################
    def Put(self, descxml, pub_iso_date=None) :
        """
        Add the given content description xml to this database, return its MOJOID

        @param pub_iso_date - if given must contain an ISO8601 date to put in the database
        as this entries publication date (in the __PUB_ISO_DATE__ column).
        """
        dochandler = contentlib.ContentDescDocHandler()
        parser = create_mojo_xml_parser(dochandler)
        try:
            parser.parseFile(StringIO(descxml))
        finally:
            parser.close()

        if dochandler.query :
            raise ContentTrackerError, "Query specified in content description"

        if not dochandler.fieldvalues.has_key('__DINODE__') :
            raise ContentTrackerError, "__DINODE__ of content required"

        # __MOJOID__ is a generated field and can not be specified
        if dochandler.fieldvalues.has_key('__MOJOID__') :
            raise ContentTrackerError, "__MOJOID__ field not allowed in put"
        
        asciidinode = dochandler.fieldvalues['__DINODE__']

        # compute the dinode id (the ascii sha1 of the raw dinode)
        mojoid = contentlib.get_mojoid_from_asciidinode(asciidinode)
        dochandler.fieldvalues['__MOJOID__'] = mojoid

        accepted = "yes"

        # if a publication date was given, store it with one
        if pub_iso_date:
            dochandler.fieldvalues['__PUB_ISO_DATE__'] = pub_iso_date

        # we -could- parse and verify the dinode syntax here but
        # probably shouldn't in order to support newer dinode versions
        # than the version of software a particular content tracker is
        # running.

        try:
            self.db.Insert(dochandler.typename, dochandler.fieldvalues)
            # store it in the mojoid lookup table for easy mojoid->dinode lookups
            self.db.Insert('MOJOID', {
                '__MOJOID__': mojoid,
                '__DINODE__': asciidinode,
                '__ACCEPTED__': accepted})

            self.db.checkpoint(1)
        except dbtables.TableDBError, e:
            raise ContentDBError, e.args
        
        return mojoid


########################################################################
class ContentTrackerDBTest :
    def __init__(self) :
        import RunTests
        db_dir = RunTests.get_next_dir_name()
        db_name = 'TestContentTable'
        dbfilename = os.path.join(db_dir, db_name)
        if os.access(dbfilename, os.F_OK) :
            os.remove(dbfilename)
        self.db = ContentTrackerDB(db_dir, name=db_name)
    
    def __obsolete_test_reject_duplicate_typedefs(self) :
        # XXX this test is obsolete because duplicat typedefs are now
        # accepted such that they cause the database table to contain
        # the union of their set of fields.
        x = contentlib.make_contenttypedef_xml('Test-1', {'1': 'string'})
        self.db.RegisterTypeXML(x)
        try:
            # make sure duplicate table definitions are detected
            self.db.RegisterTypeXML(x)
            assert 0, "duplicate table definition was accepted"
        except ContentDBError:
            pass
    
    def test_no_reserved_fields(self) :
        # the '__DINODE__' and '__PATH__' field are not allowed
        x = contentlib.make_contenttypedef_xml('T2', {'Arnie': 'string', '__DINODE__': 'string'})
        y = contentlib.make_contenttypedef_xml('Life Of', {'Brian': 'string', '__PATH__': 'string'})
        try:
            self.db.RegisterTypeXML(x)
            assert 0, "__DINODE__ field was allowed in typedef"
            self.db.RegisterTypeXML(y)
            assert 0, "__PATH__ field was allowed in typedef"
        except ContentTrackerError:
            pass

    def test_normal_functionality(self) :
        x = contentlib.make_contenttypedef_xml('T3', {'1': 'string', '2': 'number'})
        self.db.RegisterTypeXML(x)

        x = contentlib.make_contentdesc_xml('T3', {'1': 'Fuzzy', '2': '42', '__DINODE__': 'wazzuh'})
        self.db.Put(x)
        x = contentlib.make_contentdesc_xml('T3', {'1': 'Wuzzy', '2': '23', '__DINODE__': 'wazzuh', '__PATH__': 'bear.html' })
        self.db.Put(x)
        x = contentlib.make_contentdesc_xml('T3', {'1': 'bandwidth', '2': '45', '__DINODE__': 'fake-_dIn0DE'})
        self.db.Put(x)

        x = contentlib.make_contentsearch_xml('T3', {'1': '%uzz%'})
        result = self.db.Search(x)
        assert len(result) == 2
        dh = contentlib.ContentDescDocHandler()
        parser = create_mojo_xml_parser(dh)
        try:
            parser.parseFile(StringIO(result[0]))
        finally:
            parser.close()
        assert dh.typename == 'T3'
        if len(dh.fieldvalues) == 4 :
            assert dh.fieldvalues['__PATH__'] == 'bear.html'
            bear_spotted = 1
        else :
            assert len(dh.fieldvalues) == 3
            bear_spotted = 0
        assert dh.fieldvalues['__DINODE__'] == 'wazzuh'
        zero_1 = dh.fieldvalues['1']
        dh = contentlib.ContentDescDocHandler()
        parser = create_mojo_xml_parser(dh)
        try:
            parser.parseFile(StringIO(result[1]))
        finally:
            parser.close()
        assert dh.typename == 'T3'
        if len(dh.fieldvalues) == 4 and not bear_spotted :
            assert dh.fieldvalues['__PATH__'] == 'bear.html'
        else :
            assert len(dh.fieldvalues) == 3
        assert dh.fieldvalues['__DINODE__'] == 'wazzuh'
        assert zero_1 != dh.fieldvalues['1']

        x = contentlib.make_contentsearch_xml('T3', {'2': '45'})
        result = self.db.Search(x)
        assert len(result) == 1
        dh = contentlib.ContentDescDocHandler()
        parser = create_mojo_xml_parser(dh)
        try:
            parser.parseFile(StringIO(result[0]))
        finally:
            parser.close()
        assert dh.typename == 'T3'
        assert len(dh.fieldvalues) == 3
        assert dh.fieldvalues['__DINODE__'] == 'fake-_dIn0DE'
        assert dh.fieldvalues['2'] == '45'


# putting this here allows our test suite to test this module
mojo_test_flag = 0

def run():
    import RunTests
    RunTests.runTests(['contentdb'])

if __name__ == '__main__' :
    run()
