#!/usr/bin/env/python

"""Take a broker.conf file and return the public key that corresponds
   to one or more of the secret keys in that file."""

### Imports
import sys
import os.path
import getopt
import re

### In-House modules:
import modval
import idlib
import mencode
import mojosixbit
import MojoKey

### Globals
optsdict = {}

### Functions

def _usage():
    """Print usage message and exit."""

    usagestring = """usage: %s [--commskey|-c|--tokenkey|-t] broker_config_file
The broker_config_file defaults to broker.conf.
""" % os.path.basename(sys.argv[0]) 
    sys.stderr.write(usagestring)
    sys.exit(1)

def _parseargs():
    """ Parse the command-line arguments. """
    global optsdict

    try:
        opts, args = getopt.getopt( sys.argv[1:], "ct", [ 'commskey', 'tokenkey' ] )
    except getopt.error:
        _usage()

    if len(args) < 1:
        _usage()
    optsdict['brokerfile'] = args[0]
    #print 'optsdict is ', optsdict   #DEBUG
    #print 'opts is ', opts           #DEBUG

    for opt, val in opts:
        if opt in ("-c", "--commskey"):
            optsdict['commskey'] = 1
        if opt in ("-t", "--tokenkey"):
            optsdict['tokenkey'] = 1
        if opt in ("-h", "--help"):
            _usage()

    # If the user didn't set either key option, default to both.
    if not optsdict.has_key('commskey') and not optsdict.has_key('tokenkey'):
        optsdict['commskey'] = 1
        optsdict['tokenkey'] = 1

def _get_keys_from_brokerfile( filename ):
    """Read the broker configuration file and return the long (secret) versions
       of the keys in it."""
    ckey = None
    tkey = None
    ckey_regex = re.compile( r'TRANSACTION_MANAGER_STRING: (?P<ckey>(.*)$)' )
    tkey_regex = re.compile( r'mtm for communication with token server: (?P<tkey>(.*)$)' )
    try:
        thefile = open( filename, 'r' )
    except:
        sys.stderr.write( 'Could not open file %s\n' % filename )
        sys.exit(1)
    for line in thefile.readlines():
        re_match = ckey_regex.search( line ) 
        if re_match != None:
            ckey = re_match.group( 'ckey' )
        re_match = tkey_regex.search( line ) 
        if re_match != None:
            tkey = re_match.group( 'tkey' )
    return  ( ckey, tkey )

def _process_secret_key( secretkey ):
    """Given the long text version of a secret key, return the corresponding public key."""
    skdict = mencode.mdecode(secretkey)
    skdict = mencode.mdecode(skdict['session keeper'])
    keyMV = modval.new_serialized(mojosixbit.a2b(skdict['private key serialized']))
    pub_key = MojoKey.makePublicRSAKeyForCommunicating(keyMV)
    key_id = idlib.make_id(pub_key)
    return idlib.to_ascii(key_id)
    
def run():
    _parseargs()
    comms_secret_key, token_secret_key = _get_keys_from_brokerfile( optsdict['brokerfile'] )
    if optsdict.has_key('commskey'):
        print 'comms public key is', _process_secret_key(comms_secret_key)
    if optsdict.has_key('tokenkey'):
        print 'token public key is', _process_secret_key(token_secret_key)
    return

if __name__ == '__main__':
    run()
