# this script is for brokers that are consuming too much disk space
# and are doing too much business (traffic overload)
# it will grovel through your block store and publish any blobs
# that don't fit your mask anymore (and aren't available anywhere else)
# before deleting them
#
# NOTE: this module currently causes a memory leak!  I couldn't get it
# to run successfully before it consumed all available memory and swap
# on my machine.  2001-03-17 -greg
#
# I don't think it has a leak, I think you hit the berkeleydb bug?? 20010326-drue

__cvsid = '$Id: maskutil.py,v 1.4 2001/05/03 00:50:09 greg Exp $'

import PublishSession
from confutils import confman

# standard libs
import os
from Queue import Queue, Empty
import string
import time

# mojo
import idmasklib
import SEXP
import idlib
from human_readable import hr
from ContentVerifier import CallbackRegistrar
from BlockFinder import BlockFinder
import debug
import mojoutil
import pricelib
from localblockstore import NoBlock

class DummyPiece:
    """
    gets callback from PutBlob and deletes the blob from disk
    tells blobqueue it was done so it can queue up some more blobs
    """
    def __init__(self, blockstore, bq):
        self.blockstore = blockstore
        self.bq = bq
    def put_blob_failed(self, blob, failure_reason):
        self.blockstore.remove_blob(blob.shareid)
        self.bq.fin(blob.shareid)
        
    def put_blob_succeeded(self, blob, serverid):
        # delete it
        self.blockstore.remove_blob(blob.shareid)
        self.bq.fin(blob.shareid)


class BlobQueue:
    """
    holds the list of blocks to be published/deleted
    all inter-thread comm is done in put through the Queue
    """
    def __init__(self, max_concurrent, broker):
        self.q = Queue(250)
        self.max_concurrent = max_concurrent
        self.doq = broker.doq
        self.stopped = None
        self.to_find = {}
        self.find_issued = {}
        self.broker = broker
        self.finder = BlockFinder(broker, self._find_function)
        self.publishes = []
        self.current_downloads = []
        
    def _find_function(self, server_id, server_info, block_ids, callback):
        """
        called by BlockFinder to do actual message generation and spewing over mtm
        this was lifted from FindBeforehandBlockFetcher
        """
        assert len(block_ids) > 0, "_find_function called without any block_ids to find!"
        if idlib.equal(server_id, self.broker.ServicesMTM.get_id()):
            # if we're about to ask ourselves if we have these blocks, fake a response saying we don't
            self.doq.add_task(callback, kwargs={'server_id':server_id, 'server_info':server_info, 'block_ids':[]})
            return
        self.broker.do_you_have_blobs((server_id, server_info), block_ids, callback)
        
    def start(self):
        self.doq.add_task(self.do_some, delay=10)
        self.last_time = time.time()
    def stop(self):
        self.stopped = 1
    def put(self, putblob):
        # block!  calling from user thread
        self.q.put(putblob, block=1)
    def fin(self, blobid):
        self.current_downloads.remove(blobid)
        self._start_new()


    # this is our block throttler interface
    def _start_new(self):
        if len(self.publishes) > 0:
            next = self.publishes[0]
            while next is not None and len(self.current_downloads) < self.max_concurrent:
                self._start_single(next)
                self.publishes = self.publishes[1:]

    def _start_single(self, blob):
        self.current_downloads.append(blob.shareid)
        blob.put_blob()
        
    # block finder interface
    def found(self, block_id):
        pass
    def finished(self, block_id):
        l = self.finder.get_block_locations_from_cache(block_id)
        if len(l) == 0:
            self.publishes.append(self.find_issued[idlib.to_binary(block_id)])
            self._start_new()
        else:
            self.broker.multiblockstore.remove_blob(block_id)
        self.find_issued[idlib.to_binary(block_id)] = None

    def do_some(self):
        if self.stopped and self.q.empty():
            return
        
        if self.q.qsize() >= 99 or time.time() - self.last_time > 600 or self.stopped:
            self.last_time = time.time()
            for i in range(250):
                try:
                    blob = self.q.get(block=0)
                except Empty:
                    pass
                else:
                    self.to_find[blob.shareid] = blob
                    
            self.finder.find_blocks(self.to_find.keys(), allow_download_cache=0)
            callbacks = []
            for blobid, blob in self.to_find.items():
                callbacks.append((blobid, self.found, self.finished))
                self.find_issued[blobid] = blob
            self.to_find = {}
            CallbackRegistrar(self.finder, callbacks, self.broker)
        if self.stopped and not self.q.empty():
            self.doq.add_task(self.do_some, delay = 60)
        elif not self.stopped:
            self.doq.add_task(self.do_some, delay = 60)



#################

            
def do_one_store(broker, blockstore, mask, blobq):
    """
    walk through a blockstore directory and publish/delete the blobs that don't fit our mask
    """
    dp = DummyPiece(blockstore, blobq)
    def index_directory_visit_func(basepath, dirname, names, blockstore = blockstore, broker = broker, dp=dp, bq=blobq, mask = mask) :
        for blobname in names :
            if not idlib.is_mojosixbitencoded_id(blobname) :
                continue
            bin_id = idlib.sloppy_id_to_bare_binary_id(blobname)
            if not mask.match(bin_id):
                # publish it
                size = blockstore.get_block_size(blobname)
                try:
                    f = blockstore.get_block_file(blobname)
                    # lets just delete the blob...
                    pb = PublishSession.PutBlob(bin_id, f.read(), size, dp, broker.ServicesMTM, 120)
                    bq.put(pb)
                    f.close()
                    del(pb)
                except NoBlock:
                    # block is gone forget it
                    pass
    os.path.walk(blockstore.basepath, index_directory_visit_func, blockstore.basepath)


def clean_blobs(broker, num_blobs=2):
    """
    you need to reduce your mask by hand and restart your broker before using this
    this will go through your blockstore and delete any blobs that don't match your mask
    (after publishing ones that aren't to be found elsewhere)
    num_blobs is how max number of put blobs you want happening at once lower is slower
    only blobs that can't be find elsewhere will be published
    """
    # for store in stores
    q = BlobQueue(num_blobs, broker)
    q.start()
    num = 0
    for mask in broker.blobserver.masklist:
        store = broker.multiblockstore._blockstores[num]
        do_one_store(broker, store, mask, q)
        num = num + 1
    q.stop()
