#ifndef _NAP_CMDS_H
#define _NAP_CMDS_H

/* Copyright (c) 2000  Kevin Sullivan <nite@gis.net>
 *
 * Please refer to the COPYRIGHT file for more information.
 */

#include <sys/time.h>
#include <netinet/in.h>
#include "scheck.h"

struct in_nap_cmd_s 
{
  unsigned short op;
  int (*func)(int, char *, char **, int, WINDOW *);
};
typedef struct in_nap_cmd_s in_nap_cmd_t;

struct out_nap_cmd_s 
{
  const char *nm;
  unsigned char a;
  int (*func)(int, char *, char **, int, WINDOW *);
  const char *help;
};
typedef struct out_nap_cmd_s out_nap_cmd_t;

struct user_s 
{
  char *nm, *addr;
  unsigned char conn;
  unsigned int scount;
  unsigned char flag;
  struct user_s *next;
};
typedef struct user_s user_t;

struct hotlist_s 
{
  char *nm;
  int conn;
  int on;
  struct hotlist_s *next;
};
typedef struct hotlist_s hotlist_t;

/* this is the type of channels as well as queries. A query is a
   (virtual) connection to a single user */
struct chans_s 
{
  char *nm;                   /* name of channel/query - case insensitive */
  char *topic;
  user_t *users;
  unsigned char q;            /* is it a query? */
  unsigned char flag;         /* ? */
  int l;                      /* ? */
  char *key;                  /* ? */
  unsigned char p;            /* ? */
  struct chans_s *next;
}; 
typedef struct chans_s chans_t;

struct scount_s 
{
  unsigned long libraries, songs, gigs;
};
typedef struct scount_s scount_t;

#define ICMP_ECHO 8
struct icmphdr 
{
  unsigned char type, code;
  unsigned short checksum;
  union
  {
    struct
    {
      unsigned short id, sequence;
    } echo;
    unsigned long gateway;
    struct
    {
      unsigned short blah, mtu;
    } frag;
  } un;
};

/* ---------------------------------------------------------------------- */
/* the data structure for search result items. Note: the convention is
   that the items of the search result list are numbered sequentially,
   starting from *1*. However, the list macros start numbering from
   *0*. Thus, careful when looking of item number n. */
struct ssearch_s 
{
  char *song;               /* title with pathname componentes stripped */
  char *fn;                 /* title as sent by server */
  char *cmp;                /* nickname of remote user */
  int conn;                 /* connection speed of remote client */
  unsigned int sz;          /* total size */
  int brate;                /* bitrate (e.g. 128) */
  int freq;                 /* sample frequency (e.g. 44100) */
  int time;                 /* length, in seconds */
  unsigned long nip;        /* ip address of remote client, in network
			       byte order (formerly this field was
			       called ip and was in non-host byte
			       order) */
  struct timeval s, r;      /* time ping sent, received */
  int ping;                 /* ping time */
  struct ssearch_s *next;   /* next item in list */
  struct ssearch_s *next1;  /* only used for sorting; need not initialize */
};
typedef struct ssearch_s ssearch_t;

/* ---------------------------------------------------------------------- */
/* The task lists are two of the three "backbone" data structures of
   nap (the remaining one is the connection list, type sock_t, see
   scheck.h). A "task" is an upload or a download. Tasks are kept in
   two global linked lists, "up" and "down".

   The following is my attempt at a specification of what these data
   structures represent, and how they are updated in reaction to user
   events, server events, and timeouts.

   A task can be in several different states. For download tasks,
   these are:
   
   QUEUED: requested by the user, but not yet requested from server.
   RQUEUED: "remotely queued". Try to re-request later.
   RRQUEUED: "remotely queued" and requested. 
   REQUESTED: requested from server, but not yet acknowledged by server.
   WAITING: waiting for remote client to contact us.
   CONNECTING: contacted remote client and waiting for a response.
   IN PROGRESS: reading file from the remote client.
   COMPLETE: file successfully received.
   INCOMPLETE: we were disconnected after receiving a partial file.
   FAILED: error before receiving file (e.g. it did not exist)
   TIMED OUT: waited too long for server or remote client.

   For upload tasks, the possible states are very similar:

   WAITING: waiting for remote client to contact us.
   CONNECTING: contacted remote client and waiting for a response.
   IN PROGRESS: sending file to the remote client.
   COMPLETE: file successfully sent.
   INCOMPLETE: disconnected after sending partial file.
   FAILED: error before sending file (e.g. it did not exist)
   TIMED OUT: waited too long for server or remote client.

   Note that there are no QUEUED, RQUEUED, RRQUEUED or REQUESTED
   states for uploads, since uploads are always initiated by the
   server.

   Not all of the above states need to be explicitly represented in
   the data structures. For instance, one may choose to implement
   COMPLETE, INCOMPLETE, FAILED, and TIMED OUT simply by deleting the
   task. On the other hand, a clever client might want to try to
   something smart with an INCOMPLETE, for instance, try to resume it. 

   Much of the napster protocol has to do with what messages to expect
   and send in each state, and how to change states. The following
   model explains how to react to each condition in each state.
   This could be represented as a flow graph, but that's difficult to
   do in ASCII. 

   -------------------------------------------------------------------

   1. Downloads.

   [no state]: user download request (if limit exceeded) -> create
   task -> QUEUED

   [no state]: user download request (if limit ok) -> create task ->
   send 203 (download request) to server -> REQUESTED

   QUEUED: (when a slot becomes available) -> send 203 (download
   request) to server -> REQUESTED

   RQUEUED: (when timer due and slot available) -> send 203 (download
   request) to server -> RRQUEUED

   REQUESTED or RRQUEUED: receive 206 (get error) from server -> FAILED

   REQUESTED or RRQUEUED: receive 609 (accept failure) from server -> FAILED

   REQUESTED or RRQUEUED: receive 620 (queue limit) from server -> RQUEUED (with
   timer set to retry this item).

   REQUESTED or RRQUEUED: receive 204 (download ack) from server (if remote party
   is firewalled) -> send 500 (alt download request) to server -> WAITING

   REQUESTED or RRQUEUED: receive 204 (download ack) from server (if remote party
   is not firewalled) -> contact remote client -> CONNECTING

   REQUESTED: (timed out) -> resend 203 to server -> REQUESTED

   RRQUEUED: (timed out) -> resend 203 to server -> RRQUEUED

   WAITING: remote client connects to us (if negotiation fails) -> FAILED

   WAITING: remote client connects to us (if negotiation succeeds) ->
   send 218 (downloading) to sever -> IN PROGRESS

   WAITING: (timed out) -> TIMED OUT

   CONNECTING: remote client responds (if negotiation fails) -> FAILED

   CONNECTING: remote client responds (if negotiation succeeds) -> 
   send 218 (downloading) to sever -> IN PROGRESS

   CONNECTING: (timed out) -> TIMED OUT

   IN PROGRESS: receive data -> IN PROGRESS

   IN PROGRESS: (transfer complete) -> send 219 (download complete) to
   server -> COMPLETE

   IN PROGRESS: (transfer interrupted) -> send 219 (download complete) to
   server -> INCOMPLETE

   FAILED or COMPLETE or INCOMPLETE or TIMEOUT: -> purge/delete -> [no state]

   -------------------------------------------------------------------

   2. Uploads.

   [no state]: receive 607 (upload request) from server (if we don't
   have the file, or /tquit is currently in effect) -> send 608
   (accept upload request) to server -> [no state]

   (Note: there is no "reject upload request" packet!)

   [no state]: receive 607 (upload request) from server (if upload
   limit exceeded) -> send 619 (queue limit) to server -> [no state]

   [no state]: receive 607 (upload request) from server (otherwise) ->
   create task -> send 608 (accept upload request) to server -> WAITING

   WAITING: remote client connects to us (if negotiation fails) -> FAILED

   WAITING: remote client connects to us (if negotiation succeeds) ->
   send 220 (uploading) to server and go to IN PROGRESS

   WAITING: receive 501 (alt download ack or "push request") from
   server -> contact remote client -> CONNECTING

   WAITING: (timed out) -> TIMED OUT

   (Note: apparently, 501 is only sent after 607. But we may also
   choose to accept 501 in [no state] - are not currently doing so,
   however.)

   CONNECTING: remote client responds (if negotiation fails) -> FAILED

   (Note: there might be no harm in allowing an upload of a shared
   file even if it does not seem to have been initiated by the server,
   *provided* that we have space available. Usually this happens if a
   remote client sent an upload request, we received it, then we quit
   and restarted. The remote client will now try to connect to us, but
   we have erased our upload list!)

   CONNECTING: remote client responds (if negotiation succeeds) ->
   CONNECTING1 -> (continue negotiation) -> send 220 (uploading) to
   server -> IN PROGRESS

   CONNECTING: (timed out) -> TIMED OUT

   IN PROGRESS: send data -> IN PROGRESS

   IN PROGRESS: (transfer complete) -> send 221 (upload complete) to
   server -> COMPLETE

   IN PROGRESS: (disconnected) -> send 221 (upload complete) to server
   -> INCOMPLETE

   FAILED, COMPLETE, INCOMPLETE, TIMEOUT: -> purge -> [no state]

   -------------------------------------------------------------------

   Some states may split into even more sub-states, for instance,
   CONNECTING usually involves first sending a GET and then waiting
   one more time for the remote client's response. This is implemented
   by remaining in the same state, but updating the (sock_t) "func"
   entry.

*/

#define QUEUED 0
#define RQUEUED 1
#define IN_PROGRESS 2
#define RRQUEUED 3
#define REQUESTED 4
#define WAITING 5
#define CONNECTING 6
#define CONNECTING1 7
#define FAILED -1
#define INCOMPLETE -2
#define COMPLETE -3
#define TIMED_OUT -4

#define STOPPED(x) ((x)<0)    /* FAILED, INCOMPLETE, COMPLETE, TIMED_OUT */
#define ACTIVE(x) ((x)>=2)    /* not STOPPED or QUEUED or RQUEUED */
#define TIMED(x) ((x)>=3)     /* RRQUEUED, REQUESTED, WAITING, CONNECTING(1) */

/* note: in the following, if a certain field is marked "in state
   xxx", then this means the field's value is undefined in any other
   state. In particular, such fields need not be (and are not usually)
   initialized outside their own state. All functions that update the
   state MUST define and undefine such fields as appropriate; in
   particular, if a field becomes undefined as the result of a state
   change, but contains an allocated value (such as a string), then
   that value MUST be deallocated at that time. */

/* The "new" task structure for uploads */

/* note: CONNECTING1 is identical to CONNECTING, except that the local
   file and file size have already been filled in */

struct upload_s
{
  int state;                   /* WAITING, CONNECTING, CONNECTING1,
				  IN_PROGRESS, FAILED, INCOMPLETE,
				  COMPELTE, TIMED_OUT */
  
  /* in all states: */
  char *nick;                  /* nickname of remote user */
  char *rfn;                   /* remote filename; the pair nick/xrfn
                                  must be unique among non-stopped
                                  uploads */
  char *fn;                    /* filename without directory components */
  char *lfn;                   /* local filename to read from */
  int linespeed;               /* remote client's connection speed */    
  
  /* in states WAITING, CONNECTING, and CONNECTING1 */
  time_t c_time;               /* time state entered */

  /* in states CONNECTING, CONNECTING1, and IN_PROGRESS */
  struct sock_s *sk;           /* this task's client-client connection */

  /* in states CONNECTING1 and IN_PROGRESS */
  FILE *f;                     /* local file to read from */
  size_t size;                 /* totel file size */

  /* in state IN_PROGRESS */
  size_t bsz;                  /* first byte requested */
  size_t pos;                  /* position of next byte to send */
  time_t p_time;               /* time IN_PROGRESS started */

  /* in states COMPLETE, INCOMPLETE, FAILED, and TIMED_OUT (STOPPED): */
  time_t d_time;               /* time STOPPED state was entered */

  int tmp;                     /* used locally by ddup() */
  struct upload_s *next;
};
typedef struct upload_s upload_t;

/* The "new" task structure for downloads. */

/* note: even though c_time, p_time, r_time, and d_time are mutually
   exclusive, I resisted the temptation to merge them to a single
   field.  This makes debugging easier. Also note: states REQUESTED
   and RRQUEUED are almost identical, except in the amount of
   verbosity they produce. We don't keep informing the user about
   re-requests for a remotely queued item. */

struct download_s
{
  int state;                   /* QUEUED, RQUEUED, RRQUEUED, REQUESTED, 
				  WAITING, CONNECTING, IN_PROGRESS,
                                  FAILED, INCOMPLETE, COMPLETE,
                                  TIMED_OUT */
  
  /* in all states: */
  char *nick;                  /* nickname of remote user */
  char *rfn;                   /* remote filename.  The pair
				  (nick,rfn) must not occur more than
				  once (for a non-stopped item) in the
				  download list, since this is how the
				  server refers to a particular
				  download. */
  char *fn;                    /* filename without
				  directory components */

  /* in state RQUEUED: */
  time_t r_time;               /* time this item was RQUEUED */

  /* in states REQUESTED, RRQUEUED, WAITING, CONNECTING */
  time_t c_time;               /* time REQUESTED or RRQUEUED state was
                                  entered. */

  /* in states WAITING, CONNECTING, IN PROGRESS: */
  struct sockaddr_in addr;     /* remote IP and port (see ip(4)) */
  char *check;                 /* remote client's MD5 hash (or NULL) */
  int linespeed;               /* remote client's connection speed */

  /* in states CONNECTING and IN PROGRESS */
  struct sock_s *sk;           /* connection for client-to-client transfer */

  /* in state IN PROGRESS (filled during negotiation) */
  char *lfn;                   /* local filename to write to */
  FILE *f;                     /* local file to write to */
  size_t size;                 /* file size as reported by remote client */
  size_t bsz;                  /* first byte to get */
  size_t pos;                  /* position of next byte to get
                                  (current size of local file) */
  time_t p_time;               /* time when download started */

  /* in states COMPLETE, INCOMPLETE, FAILED, and TIMED_OUT (STOPPED): */
  time_t d_time;               /* time STOPPED state was entered */

  int tmp;                     /* used locally by dddown() */
  struct download_s *next;
};
typedef struct download_s download_t;

#if 0
 /* this is the "old" task structure. It is no longer used and shown
   here just for reference. */
 struct file_s
 {
  unsigned long addr;          /* remote IP (network byte order) */
  unsigned short port;         /* remote port (host byte order) */
  time_t st;                   /* time when ONGOING started, or 0 */ 
  FILE *f;                     /* local file, open for reading or writing */
  char *lfn;                   /* local file name, downloads only */
  char *fn;                    /* uploads: full name of local file,
				  downloads: stripped filename */
  char *nm;                    /* nick of remote user */
  char *rfn;                   /* uploads: stripped filename(?),
				  downloads: remote file name */ 
  char *check;                 /* remote client's md5 hash (downloads), or 
				  NULL */
  void *timer;                 /* timer for TIMEOUT in some states */
  size_t sz, csz, bsz;         /* sz: total bytes to send/get, csz: bytes
				  already sent/received, bsz: first
			          byte requested */
  unsigned char g;             /* apparently 1 iff timer is set */
  struct sock_s *sk;           /* this task's client-client connection */
  int tmp;                     /* used locally by ddup() */
  struct file_s *next;         /* next item in list */
 };
 typedef struct file_s file_t;
#endif /* 0 */

#define I_NAP_FUNC(x) int (x)(int s, char *str, char **tok, int num, WINDOW *win)

I_NAP_FUNC(sjoin);
I_NAP_FUNC(spart);
I_NAP_FUNC(stopic);
I_NAP_FUNC(ssay);
I_NAP_FUNC(snchan);
I_NAP_FUNC(sjchan);
I_NAP_FUNC(nothing);
I_NAP_FUNC(snotice);
I_NAP_FUNC(sscount);
I_NAP_FUNC(stell);
I_NAP_FUNC(suser);
I_NAP_FUNC(swhois);
I_NAP_FUNC(soff);
I_NAP_FUNC(sget);
I_NAP_FUNC(snget);
I_NAP_FUNC(snacc);
I_NAP_FUNC(sqlimit);
I_NAP_FUNC(sfreq);
I_NAP_FUNC(sry);
I_NAP_FUNC(ssf);
I_NAP_FUNC(smpart);
I_NAP_FUNC(suon);
I_NAP_FUNC(suoff);
I_NAP_FUNC(ssret);
I_NAP_FUNC(ssend);
I_NAP_FUNC(srbrowse);
I_NAP_FUNC(sdbrowse);
I_NAP_FUNC(sop);
I_NAP_FUNC(spchange);
I_NAP_FUNC(sbport);
I_NAP_FUNC(sannounce);
I_NAP_FUNC(sbanlist);
I_NAP_FUNC(scping);
I_NAP_FUNC(scpong);
I_NAP_FUNC(ssping);
I_NAP_FUNC(sredir);
I_NAP_FUNC(scycle);
I_NAP_FUNC(sclist);
I_NAP_FUNC(sclist2);
I_NAP_FUNC(snend);
I_NAP_FUNC(sblocklist);
I_NAP_FUNC(signorelist);
I_NAP_FUNC(signoreend);
I_NAP_FUNC(signoreadd);
I_NAP_FUNC(signoreremove);
I_NAP_FUNC(signoreunknown);
I_NAP_FUNC(signoreexists);
I_NAP_FUNC(signoreclear);
I_NAP_FUNC(signorefail);
I_NAP_FUNC(scbanlist);
I_NAP_FUNC(snerr);
I_NAP_FUNC(snadd);
I_NAP_FUNC(sme);


#define O_NAP_FUNC(x) int (x)(int s, char *str, char **tok, int num, WINDOW *win)

O_NAP_FUNC(ddebug);
O_NAP_FUNC(djoin);
O_NAP_FUNC(dpart);
O_NAP_FUNC(dquit);
O_NAP_FUNC(dtell);
O_NAP_FUNC(dwhois);
O_NAP_FUNC(dpup);
O_NAP_FUNC(dpdown);
O_NAP_FUNC(dsearch);
O_NAP_FUNC(dpvars);
O_NAP_FUNC(dkill);
O_NAP_FUNC(dban);
O_NAP_FUNC(dunban);
O_NAP_FUNC(ddatap);
O_NAP_FUNC(dtopic);
O_NAP_FUNC(dlevel);
O_NAP_FUNC(dg);
O_NAP_FUNC(dforce);
O_NAP_FUNC(dretry);
O_NAP_FUNC(dretryall);
O_NAP_FUNC(dopsay);
O_NAP_FUNC(ddns);
O_NAP_FUNC(dannounce);
O_NAP_FUNC(dpchans);
O_NAP_FUNC(dbanlist);
O_NAP_FUNC(dlspeed);
O_NAP_FUNC(ddup);
O_NAP_FUNC(dddown);
O_NAP_FUNC(dpurge);
O_NAP_FUNC(dpurgedown);
O_NAP_FUNC(dpurgeup);
O_NAP_FUNC(dpsocks);
O_NAP_FUNC(dmuzzle);
O_NAP_FUNC(dunmuzzle);
O_NAP_FUNC(ddisconnect);
O_NAP_FUNC(dreload);
O_NAP_FUNC(dreconnect);
O_NAP_FUNC(dbrowse);
O_NAP_FUNC(dhelp);
O_NAP_FUNC(dping);
O_NAP_FUNC(duserpass);
O_NAP_FUNC(dreloadconf);
O_NAP_FUNC(dsver);
O_NAP_FUNC(dsetconf);
O_NAP_FUNC(dchanclear);
O_NAP_FUNC(dsetlevel);
O_NAP_FUNC(dme);
O_NAP_FUNC(dusers);
O_NAP_FUNC(dgusers);
O_NAP_FUNC(dclist);
O_NAP_FUNC(dclist2);
O_NAP_FUNC(dclear);
O_NAP_FUNC(dnews);
O_NAP_FUNC(dsraw);
O_NAP_FUNC(dquery);
O_NAP_FUNC(dcloak);
O_NAP_FUNC(dblocklist);
O_NAP_FUNC(dblock);
O_NAP_FUNC(dunblock);
O_NAP_FUNC(dwindow);
O_NAP_FUNC(dignore);
O_NAP_FUNC(dunignore);
O_NAP_FUNC(dignoreclear);
O_NAP_FUNC(dignorelist);
O_NAP_FUNC(dalias);
O_NAP_FUNC(dunalias);
O_NAP_FUNC(daliaslist);
O_NAP_FUNC(dhandler);
O_NAP_FUNC(dunhandler);
O_NAP_FUNC(dhandlerlist);
O_NAP_FUNC(dcbanlist);
O_NAP_FUNC(dcban);
O_NAP_FUNC(dcunban);
O_NAP_FUNC(dcbanclear);
O_NAP_FUNC(dkick);
O_NAP_FUNC(dloadalias);
O_NAP_FUNC(dsavealias);
O_NAP_FUNC(dclearalias);
O_NAP_FUNC(dloadhandler);
O_NAP_FUNC(dsavehandler);
O_NAP_FUNC(dclearhandler);
O_NAP_FUNC(dwstats);
O_NAP_FUNC(dnotify);
O_NAP_FUNC(dunnotify);
O_NAP_FUNC(dhotlist);
O_NAP_FUNC(dsay);
O_NAP_FUNC(dtquit);
O_NAP_FUNC(dunquit);
O_NAP_FUNC(dabout);
O_NAP_FUNC(dirc);
O_NAP_FUNC(dkickall);
O_NAP_FUNC(deval);
O_NAP_FUNC(dset);
O_NAP_FUNC(dunset);
O_NAP_FUNC(dsaveconfig);
O_NAP_FUNC(dloadconfig);
O_NAP_FUNC(dsavechannels);
O_NAP_FUNC(dloadchannels);
O_NAP_FUNC(decho);
O_NAP_FUNC(dresults);
O_NAP_FUNC(dexec);
O_NAP_FUNC(dtimer);
O_NAP_FUNC(dif);
O_NAP_FUNC(ddone);
O_NAP_FUNC(dwhile);
O_NAP_FUNC(dinc);
O_NAP_FUNC(ddec);
O_NAP_FUNC(dbreak);
O_NAP_FUNC(dlastlog);
O_NAP_FUNC(drebuild);
O_NAP_FUNC(dupdate);
O_NAP_FUNC(dchupload);
O_NAP_FUNC(dtlist);
O_NAP_FUNC(dnoprint);
O_NAP_FUNC(dstop);
O_NAP_FUNC(dserv);
O_NAP_FUNC(dserver);

char *fixquotes(char *s);
void upchan(chans_t *);
int in_cksum(u_short *, int);
void tvsub(register struct timeval *, register struct timeval *);
int icmpin(WINDOW *, struct sock_s *);
char **form_toks(char *, int *);
char **form_tokso(char *, int *);
int parsein(int, unsigned short, char *, WINDOW *);
char **fxv(char **, int, int *);
int parseout(int, char *, WINDOW *);
int parseoutn(int, char *, WINDOW *);
char *gnum(int);
char *basename(char *fn);
char *cstr(char *, int);
char *cspstr(char **, int, int);
int dupload(upload_t *);
int ddownload(WINDOW *win, download_t *task);
int interrupt_download(WINDOW *win, download_t *task);
void move_to_downloaddir(WINDOW *win, download_t *task);
void remove_turd(WINDOW *win, download_t *task, int turdsize);
void mark_incomplete(WINDOW *win, download_t *task);
void tagincompletefile(char *fn, char *str);
void adduser(chans_t *, char *, int, unsigned char);
void deluser(chans_t *, char *);
user_t *finduser(chans_t *, char *);
void timed_dup(void *data);
void timed_ddown(void *data);
void timed_command(void *data);
void timed_pingresults(void *dummy);
void showresults(WINDOW *win, int havepings);
int requestdownload(int s, ssearch_t *sitem);
int opendownloadfile(char *fn, FILE **f, char **lfn);
void qevent(void);
int parse_range(char *range);
int download_limit_reached(char *nick);
int upload_limit_reached(char *nick);
int purgedown(WINDOW *win);
int purgeup(void);
void rqueued_hook(char *nick);
int save_hotlist(WINDOW *win);

#endif /* _NAP_CMDS_H */

