#include <stddef.h>
#include <stdlib.h>
#include <string.h>


#include "litexmlparser.h"

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
    Find the DOCTYPE element in the string pointed to by 'start'. The search is
    limited to the range between 'start' and 'limit'. If 'wholeElement is non-
    zero then the range of the entire element is returned, otherwise only the
    range of the DOCTYPE value is returned.

    @param start Pointer at the begining of the range to be scanned. If start
           does not specify a C String then 'limit' must be non-NULL in order
           to establish bounds to the search area.

    @param limit Pointer at the last character which will be included in the
           search range. If NULL then 'start' is assumed to be a C String and
           'limit' is positioned at the character before the terminating NULL.
    
    @param wholeElement A boolean which, if non-zero, indicates if the range
           of the entire DOCTYPE directive should be returned or just the name
           of the root element.

    @returns A charRange descibing either the name of the root element or the
           entire DOCTYPE directive, ie. "root" or "<!DOCTYPE root>". If the
           DOCTYPE directive cannot be found then the range returned is 
           { NULL, NULL }.
*/
charRange getDocType( char const * start, char const * limit, int wholeElement )
{
charRange	result = { NULL, NULL };
tagRange	ranges = getTagRanges( "!DOCTYPE", start, limit );
char const*	end = start + (ptrdiff_t) strlen(start) - 1;

if( NULL == start )
	return result;

if( (NULL != limit) && (end > limit) )
	end = limit;

if( NULL == ranges.startTag.start )
	return result;
	
if( wholeElement ) {
	result = ranges.startTag;
	}
else {
	result.start = ranges.startTag.start + 1 + sizeof( "!DOCTYPE" ) - 1 + 1;
	
	while(	(result.start < end) &&
			(	(' ' == *(result.start)) ||							// is immediately followed by a delimiter or the end of the tag
		  		('\t' == *(result.start)) ||
		  		('\n' == *(result.start)) ) )
		result.start++;

	result.end = result.start;
	while(	(result.end < end) &&
			(' ' != *(result.end+1)) &&								// is immediately followed by a delimiter or the end of the tag
		  	('\t' != *(result.end+1)) &&
		  	('\n' != *(result.end+1)) &&
		  	('/' != *(result.end+1)) &&
		  	('>' != *(result.end+1)) )
		result.end++;
	}
	
return result;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
	Find the portion of the C string located at 'start' which forms the body of 
    the	tag identified by 'tag'. The search will extend to the end of the 
    string or to the point provided in "limit". To include the text of the 
    start and end tags in the range "includeTags" must be non-zero.
	
	Note : if 'includeTags' is zero then there is no way to differentiate between
    not finding the tag and finding an empty-element tag. (since the body is
    empty)

    NOTE : This implementation does not handle documents containing nested tags
           with the same name properly. The search for the end tag is done
           without ensuring that it is the matching end tag for this start tag.

           If you have a document of the form:

            <tag>
                stuff
                <tag>
                    other stuff
                </tag>
            </tag>

            This implementation will incorrectly return the range of the initial
            tag as ending at the first instance of </tag> and not the second.


    @param tag Pointer at a C String containing the name of the XML tag which
            will be searched for within the message specified by 'start' and
            'limit.

    @param start Pointer at the begining of the range to be scanned. 'start'
           must point at a C String or 'limit' must be non-NULL in order to 
           establish bounds to the search area. 'start' does not need to point
           to the begining of the XML document, when iterating through a
           document, 'start' is incremented to one character past the end of
           the previously found tag.

    @param limit Pointer at the last character which will be included in the
           search range. If NULL then 'start' is assumed to be a C String and
           'limit' is positioned at the character before the terminating NULL.

    @param includeTags A boolean which, if non-zero, indicates that the start
           tag and the terminating tag should be included in the range, ie.
           "body" versus "<tag>body</tag>"

    @returns A charRange descibing either the body of the first element found
           matching 'tag' or a range describing the start tag, the body and
           the end tag of the first tag matching 'tag'. If no tag matching
           'tag' can be found then { NULL, NULL } is returned. if 'includeTags'
           is zero then there is no way to differentiate between not finding
           the tag and finding an empty-element tag (since the body is empty)
           
*/
charRange getTagScope( char const * tag, char const * start, char const * limit, int includeTags )
{
charRange	result = { NULL, NULL };
tagRange	ranges = getTagRanges( tag, start, limit );

if( NULL != ranges.startTag.start ) {
	if( includeTags ) {
		result.start = ranges.startTag.start;
		result.end = ranges.endTag.end;
		}
	else {
		result = ranges.body;
		}
	}

return result;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
	Find the portion of the C string located at "start" which forms the body of
    the tag identified by "tag". The search will extend to the end of the string
    or to the point provided in "limit". 
    
    getTagRanges differs from getTagScope in that it always returns the
    location of all the components of the element specified by 'tag'. This makes
    it easier to locate just the initial tag, the terminating tag or the body.

    NOTE : This implementation does not handle documents containing nested tags
           with the same name properly. The search for the end tag is done
           without ensuring that it is the matching end tag for this start tag.

           If you have a document of the form:

            <tag>
                stuff
                <tag>
                    other stuff
                </tag>
            </tag>

            This implementation will incorrectly return the range of the initial
            tag as ending at the first instance of </tag> and not the second.

    @param tag Pointer at a C String containing the name of the XML tag which
            will be searched for within the message specified by 'start' and
            'limit.

    @param start Pointer at the begining of the range to be scanned. 'start'
           must point at a C String or 'limit' must be non-NULL in order to 
           establish bounds to the search area. 'start' does not need to point
           to the begining of the XML document, when iterating through a
           document, 'start' is incremented to one character past the end of
           the previously found tag.

    @param limit Pointer at the last character which will be included in the
           search range. If NULL then 'start' is assumed to be a C String and
           'limit' is positioned at the character before the terminating NULL.

    @returns A tagRange structure composed of three elements, the location of the
           first start tag matching 'tag', the body of the element and the end
           tag. The implementation will attempt to return partial results, but
           does require that the start tag be fully formed, ie. if the document
           terminates before the end tag is found then the result will be the
           correct location of the startTag, a body range describing the
           remainder of the document and a null 'endTag'. If tag matches against
           an empty element declaration, ie. <tag/> then startTag and endTag
           will have the same value and body will be {NULL, NULL}. If the body
           is simply empty then body.end will be equal to body.start - 1 or
           alternatly, body.start == endTag.start and body.end == startTag.end.


*/
tagRange getTagRanges( char const * tag, char const * start, char const * limit )
{
tagRange	result = { { NULL, NULL }, { NULL, NULL }, { NULL, NULL } };
ptrdiff_t	tagLen = strlen( tag );
char const*	end = start + (ptrdiff_t) strlen(start) - 1;
char const*	workPtr;
int	    	foundStartTag = 0;
int 		foundEndTag = 0;

if( NULL == start )
	return result;

if( (NULL != limit) && (end > limit) )
	end = limit;

// search for the start tag
workPtr = start;

while( !foundStartTag && (workPtr < end) ) {
	char const * foundTagText = strstr( workPtr, tag );
	char const * foundTagTerminator;
	char const * foundNextTagStart;
	char const * afterTagText = foundTagText + tagLen;
	
	// was it found or not in bounds?
	if( (NULL == foundTagText) || (afterTagText > end) )
		return result;	// it was not found
	
	// check to see if it is the start tag
	if( (foundTagText < (workPtr + 1)) || 						// it begins at a safe place in the string
		('<' != *(foundTagText - 1)) ||							// has an open tag delimiter before it.
		( (' ' != *afterTagText) &&								// is immediately followed by a delimiter or the end of the tag
		  ('\t' != *afterTagText) &&
		  ('\n' != *afterTagText) &&
		  ('/' != *afterTagText) &&
		  ('>' != *afterTagText) ) ) {								
		workPtr = afterTagText;
		continue;
		}
	
	foundTagTerminator = strchr( afterTagText, '>' );
	foundNextTagStart = strchr( afterTagText, '<' );
	
	if ( (NULL == foundTagTerminator) || 						// the tag has no terminator
		 (foundTagTerminator > end) ||							// it is past the valid range
		 (	(NULL != foundNextTagStart) &&						// there is another tag start
		 	(foundNextTagStart < foundTagTerminator) ) ) {		// and it is before the terminator we found. very bad.
		workPtr = afterTagText;
		continue;
		}

	foundStartTag = 1;
	result.startTag.start = foundTagText - 1;
	result.startTag.end = foundTagTerminator;
	}

if( !foundStartTag )
	return result;

// is it an empty element declaration?
if( '/' == *(result.startTag.end - 1) ) {
	result.endTag = result.startTag;							// end is the start and there is no body.
	return result;
	}
	
// search for the end tag

workPtr = result.startTag.end + 1;

while( !foundEndTag && (workPtr < end) ) {
	char const * foundTagText = strstr( workPtr, tag );
	
	// was it found or not in bounds?
	if( (NULL == foundTagText) || ((foundTagText + tagLen + 1) > end) )
		break;	// it was not found
	
	// check to see if it is end tag
	if(  (foundTagText < (workPtr + 2)) ||
		 ('/' != foundTagText[-1]) ||
		 ('<' != foundTagText[-2]) ||
		 ('>' != foundTagText[tagLen]) ) {
		workPtr += tagLen;
		continue;
		}
	
	foundEndTag = 1;
	result.endTag.start = foundTagText - 2;
	result.endTag.end = foundTagText + tagLen;
	}

// calculate where the body must lie.

result.body.start = result.startTag.end + 1;

if( foundEndTag ) {
	result.body.end = result.endTag.start - 1;
	}
else {
	result.body.end = end;
	}
	
return result;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/*
	Given the range location for a document, 'doc', and the name of a tag, 'tag',
    insert the data described by 'new content' into the body of the tag. 

    @param doc a charRange describing the current limits of the document into
          which the new text will be placed. 'doc.end' MUST point to the real
          end of the document for the content to be correctly inserted.

    @param docLimit a charRange describing the maximum limits by which doc may
          be expanded to accomodate the new content specified by 'newContent'

    @param tag A C String containing the name of the XML tag to which the new
          content will be added.

    @param replace A boolean which, if non-zero indicates that 'newContent' 
          should replace the current message body of 'tag' rather than being
          appended to any existing content.

    @returns A charRange describing the new bounds of the document. If the new
          content will not fit in the bounds defined by docLimit then the
          result returned will be { NULL, (required docLimit.end) }
        

*/
charRange addContentToTag( charRange doc, charRange docLimit, char const * tag, charRange newContent, int replace )
{
tagRange	intoTag = getTagRanges( tag, doc.start, doc.end );
charRange	result = { NULL, NULL };

if( (NULL != intoTag.startTag.start) && (NULL != intoTag.endTag.end) ) {
	size_t		newContentLen = (newContent.end - newContent.start + 1);
	size_t		endTagLen = (doc.end - intoTag.endTag.start + 1);
	size_t      oldContentLen = (intoTag.body.end - intoTag.body.start + 1);
	if( replace ) {
		result.end = doc.end - oldContentLen + newContentLen;
		if( result.end > docLimit.end )
			return result;
		memmove( (char*) intoTag.body.start + newContentLen, intoTag.endTag.start, endTagLen ); 
		}
	else {
		result.end = doc.end + newContentLen;
		if( result.end > docLimit.end )
			return result;
		memmove( (char*) intoTag.body.start + newContentLen, intoTag.body.start, endTagLen + oldContentLen ); 
		}
	
	result.start = doc.start;
	memmove( (char*) intoTag.body.start, newContent.start, newContentLen );
	}

return result;
}

