#ifndef LITEXMLPARSER_H_
#define LITEXMLPARSER_H_

/**
    LiteXMLParser is a small C Library which provides limited XML parsing 
    facilities suitable for use on memory limited devices. In particular, none
    of the APIs allocate memory and none of the APIs require copying of data.
    LiteXMLParser provides no validation of the XML document and makes only
    minimal checks that the document is well formed.

    Four APIs are provided : getDocType, getTagScope, getTagRanges and 
    AddContentToTag.

    getDocType is used for returning information from the DOCTYPE declaration

    getTagScope and getTagRanges are used to locate tags within an XML document.

    AddContentToTag is used to add content to the body of a tag.

*/

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
    A charRange defines a range of characters, probably within a string. The 
    range is deemed to be invalid if 'start' is NULL. If 'start' is non-NULL
    and 'end' is NULL then the range can be assumed to be a null terminated C
    string. A zero length range is, by convention, described by an 'end' value
    of 'start' - 1.
*/
typedef struct charRange {
	char const * start;
	char const * end;
	} charRange;

/**
    A tagRange is a collection of char ranges useful for describing XML 
    structures. 'startTag' is the range of the opening tag, ie. <tag>
    'endTag' is the range of the terminating tag, ie. </tag>. The 'body'
    range everything between the start and end tags. For empty-element tags
    the 'startTag' and 'endTag' will be equal and the 'body' tag will be
    NULL.
*/
typedef struct tagRange {
	charRange	startTag;
	charRange	body;
	charRange	endTag;
	} tagRange;

/**
    Contains a collection of results, all contained within the C String 
    pointed to by 'srcData'. 'resultsCount' is the number of results and
    'theResults' is an array of the charRanges with the results.
*/
typedef struct someResults {
	char const *	srcData;
	int				resultsCount;
	charRange		theResults[];
	} someResults;

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
    Find the DOCTYPE element in the string pointed to by 'start'. The search is
    limited to the range between 'start' and 'limit'. If 'wholeElement is non-
    zero then the range of the entire element is returned, otherwise only the
    range of the DOCTYPE value is returned.

    @param start Pointer at the begining of the range to be scanned. If start
           does not specify a C String then 'limit' must be non-NULL in order
           to establish bounds to the search area.

    @param limit Pointer at the last character which will be included in the
           search range. If NULL then 'start' is assumed to be a C String and
           'limit' is positioned at the character before the terminating NULL.
    
    @param wholeElement A boolean which, if non-zero, indicates if the range
           of the entire DOCTYPE directive should be returned or just the name
           of the root element.

    @returns A charRange descibing either the name of the root element or the
           entire DOCTYPE directive, ie. "root" or "<!DOCTYPE root>". If the
           DOCTYPE directive cannot be found then the range returned is 
           { NULL, NULL }.
*/
charRange getDocType( char const * start, char const * limit, int wholeElement );

/**
	Find the portion of the C string located at 'start' which forms the body of 
    the	tag identified by 'tag'. The search will extend to the end of the 
    string or to the point provided in "limit". To include the text of the 
    start and end tags in the range "includeTags" must be non-zero.
	
	Note : if 'includeTags' is zero then there is no way to differentiate between
    not finding the tag and finding an empty-element tag. (since the body is
    empty)

    NOTE : This implementation does not handle documents containing nested tags
           with the same name properly. The search for the end tag is done
           without ensuring that it is the matching end tag for this start tag.

           If you have a document of the form:

            <tag>
                stuff
                <tag>
                    other stuff
                </tag>
            </tag>

            This implementation will incorrectly return the range of the initial
            tag as ending at the first instance of </tag> and not the second.


    @param tag Pointer at a C String containing the name of the XML tag which
            will be searched for within the message specified by 'start' and
            'limit.

    @param start Pointer at the begining of the range to be scanned. 'start'
           must point at a C String or 'limit' must be non-NULL in order to 
           establish bounds to the search area. 'start' does not need to point
           to the begining of the XML document, when iterating through a
           document, 'start' is incremented to one character past the end of
           the previously found tag.

    @param limit Pointer at the last character which will be included in the
           search range. If NULL then 'start' is assumed to be a C String and
           'limit' is positioned at the character before the terminating NULL.

    @param includeTags A boolean which, if non-zero, indicates that the start
           tag and the terminating tag should be included in the range, ie.
           "body" versus "<tag>body</tag>"

    @returns A charRange descibing either the body of the first element found
           matching 'tag' or a range describing the start tag, the body and
           the end tag of the first tag matching 'tag'. If no tag matching
           'tag' can be found then { NULL, NULL } is returned. if 'includeTags'
           is zero then there is no way to differentiate between not finding
           the tag and finding an empty-element tag (since the body is empty)
           
*/
charRange getTagScope( char const * tag, char const * start, char const * limit, int includeTags );

/**
	Find the portion of the C string located at "start" which forms the body of
    the tag identified by "tag". The search will extend to the end of the string
    or to the point provided in "limit". 
    
    getTagRanges differs from getTagScope in that it always returns the
    location of all the components of the element specified by 'tag'. This makes
    it easier to locate just the initial tag, the terminating tag or the body.

    NOTE : This implementation does not handle documents containing nested tags
           with the same name properly. The search for the end tag is done
           without ensuring that it is the matching end tag for this start tag.

           If you have a document of the form:

            <tag>
                stuff
                <tag>
                    other stuff
                </tag>
            </tag>

            This implementation will incorrectly return the range of the initial
            tag as ending at the first instance of </tag> and not the second.

    @param tag Pointer at a C String containing the name of the XML tag which
            will be searched for within the message specified by 'start' and
            'limit.

    @param start Pointer at the begining of the range to be scanned. 'start'
           must point at a C String or 'limit' must be non-NULL in order to 
           establish bounds to the search area. 'start' does not need to point
           to the begining of the XML document, when iterating through a
           document, 'start' is incremented to one character past the end of
           the previously found tag.

    @param limit Pointer at the last character which will be included in the
           search range. If NULL then 'start' is assumed to be a C String and
           'limit' is positioned at the character before the terminating NULL.

    @returns A tagRange structure composed of three elements, the location of the
           first start tag matching 'tag', the body of the element and the end
           tag. The implementation will attempt to return partial results, but
           does require that the start tag be fully formed, ie. if the document
           terminates before the end tag is found then the result will be the
           correct location of the startTag, a body range describing the
           remainder of the document and a null 'endTag'. If tag matches against
           an empty element declaration, ie. <tag/> then startTag and endTag
           will have the same value and body will be {NULL, NULL}. If the body
           is simply empty then body.end will be equal to body.start - 1 or
           alternatly, body.start == endTag.start and body.end == startTag.end.


*/
tagRange getTagRanges( char const * tag, char const * start, char const * limit );

/*
	Given the range location for a document, 'doc', and the name of a tag, 'tag',
    insert the data described by 'new content' into the body of the tag. 

    @param doc a charRange describing the current limits of the document into
          which the new text will be placed. 'doc.end' MUST point to the real
          end of the document for the content to be correctly inserted.

    @param docLimit a charRange describing the maximum limits by which doc may
          be expanded to accomodate the new content specified by 'newContent'

    @param tag A C String containing the name of the XML tag to which the new
          content will be added.

    @param replace A boolean which, if non-zero indicates that 'newContent' 
          should replace the current message body of 'tag' rather than being
          appended to any existing content.

    @returns A charRange describing the new bounds of the document. If the new
          content will not fit in the bounds defined by docLimit then the
          result returned will be { NULL, (required docLimit.end) }
        

*/
charRange addContentToTag( charRange doc, charRange docLimit, char const * tag, charRange newContent, int replace );

#endif