/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Codat.java,v 1.7 2001/05/23 00:28:19 lomax Exp $
 */

package net.jxta.codat;

import net.jxta.peergroup.PeerGroupID;
import net.jxta.document.Document;

/**
 * This class represents a <b>Codat</b>. Codats are JXTA objects that can
 * hold both data or code. The core platform API is making available the
 * Codat APIs to JXTA services and applications as a placeholder for all
 * contents they are manipulating. The Core is not making any references to
 * this class.  The core only manipulates StructuredDocuments and
 * Advertisements that are subclasses of Documents.  The Codat class is
 * offered as a standard way for applications and services to exchange any
 * kinds of contents via a common API and associate a unique JXTA id to
 * these contents.
 *
 * <p>Codats are containers objects that are used to hold any kinds of
 * objects or data. A codat can represent a file, a class file, the saved
 * state of an application, a loadable C library. Codats are handled
 * transparently by the JXTA platform, and are used as placeholders for
 * any types of data. Codats hold Document that represent the data that
 * they hold.</p>
 *
 * <p>Codats are published in peer groups. A Codat can belong to
 * only one peer group. Multiple copies of a codat can be made
 * to be published in multiple peer groups.</p>
 *
 * <p>Codats are uniquely identified via a unique CodatID. This Id
 * is guaranteed to be unique within the JXTA world.
 *
 * <p>The core manipulates two main types of codats:
 *      <ul type-disc>
 *      <li><p><b>Codat</b> - hold data or code</li></p>
 *      <li><p><b>Metadata</b> - hold information about another Codat</li></p>
 *      </ul>
 *
 * The JXTA platform defines Codat as the unit of information shared and
 * exchanged within a JXTA group.  All instances of Codats reside within a
 * peer group. The PeerGroup content caching service provides storage and retrieval
 * methods for codats using codatId as index.
 *
 *
 * @see     net.jxta.codat.CodatID
 * @see     net.jxta.document.Document
 * @see     net.jxta.document.StructuredDocument
 * @see     net.jxta.document.StructuredTextDocument
 *
 * @version $Revision: 1.7 $
 * @since JXTA 1.0 */


public class Codat {

  /**
   * Codat Id of this Codat. This is the "address" which may be used to
   * refer to this Codat. @see net.jxta.codat.CodatId
   * 
   */ 

  protected CodatID id = CodatID.nullCodatID;
  
  /**
   * Codat Id of a Codat to which this Codat is related. This may be the
   * Codat Id of another codat in the same Peer Group or 
   * net.jxta.codat.CodatId.nullCodatId.
   * 
   */ 

  protected CodatID metaId = CodatID.nullCodatID;
  
  /**
   * A Jxta Document which contains the data held by this Codat.
   *
   */

  protected Document doc = null;

  /**
   *  Makes a new Codat with a new CodatId given a PeerGroupID
   *  and a document.
   *
   *  @param  groupID   PeerGroupID the group to which this
   *          codat will belong.
   *  @param about CodatId for which this Codat is metadata
   *  @param  document Document held by this codat.
   *
   *  @exception  OutOfMemoryException  not enough memory to create the codat
   *
   *  @version $Revision: 1.7 $
   *  @since JXTA 1.0
   */

  public Codat (PeerGroupID groupID, CodatID about, Document document) 
    throws OutOfMemoryError {

    try {
      this.id = new CodatID( groupID, document );
    } catch (Exception ex) {
    }

    this.metaId = about;
    this.doc = document;

  }

  /**
   *  Makes a new Codat instance from an existing Codat,
   *  with a given CodatID and a document.
   *
   *  @param id CodatId of the new codat
   *  @param about CodatID for which this Codat is metadata
   *  @param document Document hold by this codat
   *
   *  @exception  OutOfMemoryException  not enough memory to create the codat
   *
   *  @version $Revision: 1.7 $
   *  @since JXTA 1.0
   */

  public Codat (CodatID id, CodatID about, Document document)
    throws OutOfMemoryError {

    this.id = id;
    this.metaId = about;
    this.doc = document;
  }

  /**
   *  Returns the CodatId associated with this Codat.
   *
   *  @return  CodateID associated with this codat
   *
   *  @version $Revision: 1.7 $
   *  @since JXTA 1.0
   */

  public CodatID getCodatID () {
    return id;
  }

  /**
   *  Returns Codat id of related codat associated with this metadata Codat.
   *
   *  @return  CodateID associated with this codat
   *
   *  @version $Revision: 1.7 $
   *  @since JXTA 1.0
   */

  public CodatID getMetaID () {
    return metaId;
  }

  /**
   *  Returns the Document associated with this Codat.
   *
   * @returns Document associated with this Codat
   *
   *  @version $Revision: 1.7 $
   *  @since JXTA 1.0
   */
  
  public Document getDocument() {
    return doc;
  }
  
}
