/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: CodatID.java,v 1.10 2001/05/23 00:28:19 lomax Exp $
 */

package net.jxta.codat;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.security.MessageDigest;

import java.io.IOException;
import java.security.DigestException;
import java.security.NoSuchAlgorithmException;

import net.jxta.peergroup.PeerGroupID;
import net.jxta.document.Document;
import net.jxta.id.ID;
import net.jxta.id.IDFactory;

import net.jxta.impl.id.UUID;
import net.jxta.impl.id.UUIDFactory;

/**
 *
 * This class represents a CodatID. Codats are JXTA objects that can hold
 * both data or code.
 *
 * <p>Codats are containers objects that are used to hold any kinds of
 * objects or data. A codat can represent a file, a class file, the saved
 * state of an application, a loadable C library. Codats are handled
 * transparently by the JXTA platform, and are used as placeholders for
 * any type of data. Codats hold Document that represent the data that
 * they hold.</p>
 *
 * <p>A CodatID is formed by the conjuction of a PeerGroup unique ID and
 * the Codat Document hash value.The Id is the unique Id for this Codat
 * within the JXTA world. Some codats may not hold a document. In that case
 * a codatId is constructed without the document hash value.
 *
 * @see     net.jxta.codat.Codat
 * @see     net.jxta.document.Document
 * @see     net.jxta.document.StructuredDocument
 *
 * @version $Revision: 1.10 $
 * @since JXTA 1.0 */

public class CodatID extends ID implements Cloneable {

  /**
   *   The null CodatID. Useful for comparisons and cases where the CodatId
   *   is ignored.
  **/

  final static public CodatID nullCodatID = new CodatID();
  
  final static private int uuidSize = 16;
  
  /**
   * size of a SHA1 hash. I would use MessageDigest.getDigestLength, but
   * possible exceptions make it difficult to do. 
   */
  
  final static private int hashSize = 20;
  
  final static private int groupIdOffset = 0;
  final static private int idOffset = CodatID.groupIdOffset + CodatID.uuidSize;
  final static private int codatHashOffset = CodatID.idOffset
    + CodatID.uuidSize;
  final static private int padOffset = CodatID.codatHashOffset +
    CodatID.hashSize;
  final static private int padSize = ID.flagsOffset - CodatID.padOffset;
  
  /**
   *  If true then this class has been registered as the class created when
   *  JXTA URIs of type ID.flagCodatID need constructing.
   */
  
    private static boolean registered = false;
  
  static {
      try {
	registered = IDFactory.registerIDType(ID.flagCodatID,
					      CodatID.class);
      }
      catch( Exception e ) {
	registered = false;
      }
  };
    
  /**
   * Constructor used in the creation of a CodatID when no concrete Document is
   * associated. Dymamic codats.
   * 
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */
  
  public CodatID () {
    super();
    bytes[ID.flagsOffset + ID.flagsIdTypeOffset] = ID.flagCodatID;
  };
  
  /**
   * Creates a CodatID for a Codat without document. A PeerGroupID is provided
   * but no Document is provided. CodatID contains no hash value for the
   * Codat data.
   *
   * @param groupID    the group to which this content will belong.
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */

    public CodatID (PeerGroupID groupID) {
      this( );
      
      UUID    groupUUID = (UUID) groupID.getGroupUniqueId();
      UUID    id = UUIDFactory.newUUID();
      
      longIntoBytes( CodatID.groupIdOffset,
		     groupUUID.getMostSignificantBits() );
      longIntoBytes( CodatID.groupIdOffset + 8,
		     groupUUID.getLeastSignificantBits() );
      
      longIntoBytes( CodatID.idOffset, id.getMostSignificantBits() );
      longIntoBytes( CodatID.idOffset + 8, id.getLeastSignificantBits() );
    };
  
  /**
   * Constructor to create a new CodatID. CodatID contains the hash value
   * for the associated Codat document and the PeerGroup Id where this
   * codat is published.
   *
   * @param groupID    PeerGroup to which this content will belong.
   * @param document   Document which will be associcated with this CodatID
   *
   * @exception   NoSuchAlgorithmException Did not find correct encoding
   *                                       algorithm
   * @exception   IOException   I/O Error reading document
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0 
   */

  public CodatID (PeerGroupID groupID, Document document)
    throws IOException, NoSuchAlgorithmException {
    this( groupID, document.getStream() );
  };
  
  /**
   * Constructor.
   * Creates a static Codat. ID contains hash value for Codat data.
   *
   * @param groupID   the group to which this dynamic content will belong.
   * @param in        the InputStream from which the content hash is
   *                  calculated. The stream is read until EOF and then closed.
   *
   *
   * @exception   NoSuchAlgorithmException Did not find correct encoding
   *                                       algorithm
   * @exception   IOException   I/O Error reading document
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0 
   */
  
  public CodatID  (PeerGroupID groupID, InputStream in)
    throws IOException, NoSuchAlgorithmException {
    this( groupID );
        
    MessageDigest dig = MessageDigest.getInstance( "SHA" );
    
    if (dig == null) {
      throw (new NoSuchAlgorithmException("SHA1 digest algorithm found"));
    }
    
    dig.reset();
    
    do {
      int nextByte = in.read();
      if( nextByte == -1 )
	break;
      
      dig.update( (byte) nextByte);
    }
    while( true );
    
    in.close();
    
    try {
      byte [] result = dig.digest( );
      for( int eachByte = 0; eachByte < CodatID.hashSize; eachByte++ )
	bytes[eachByte + CodatID.codatHashOffset] = result[eachByte];
    }
    catch( Exception e ) {
      /* we convert this to an IO Exception to keep the number of
       * down.
       */
      throw new IOException( "Digest algorithm could not complete" );
    }
  };
  
  /**
   * Returns true if the CodatID is associated with a static Codat.
   *
   * @return boolean chcek if the codatId is for a codat without document
   * (Static codat)
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0 
   */

  public boolean isStatic () {
    for( int eachHashByte = CodatID.codatHashOffset;
	 eachHashByte < (CodatID.padOffset); eachHashByte++ )
      if( 0 != bytes[eachHashByte] )
	return true;
    
    return false;
  };
    
  /**
   * Returns the PeerGroupID of the PeerGroup associated with this Codat
   *
   * @return PeerGroupID return peer group Id associated with this codat id
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0 
   */

  public PeerGroupID getPeerGroupID() {
    UUID result = new UUID( bytesIntoLong( CodatID.groupIdOffset ),
			    bytesIntoLong( CodatID.groupIdOffset + 8 ) );
        
    return new PeerGroupID( result );
  };
  
  /**
   * Returns true if the two CodatIDs are same
   *
   *
   * @param id CodatID to be checked with 
   * @return boolean 
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0 
   */

  public boolean equals ( Object id ) {
  	if (this == id) {
	    return true;
	}
	if (id instanceof CodatID)
        return super.equals( id );
    else 
        return false;
  };
  
  /**
   * Clones this CodatID
   *
   * @return Object  clone object
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0 
   */

  public Object clone () {
    CodatID result = new CodatID();
    result.bytes = (byte [] ) this.bytes.clone();
    return result;
  };
}
