/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Discovery.java,v 1.4 2001/05/23 00:28:20 lomax Exp $
 */

package net.jxta.discovery;

import java.io.IOException;
import java.util.Enumeration;

import net.jxta.service.Service;
import net.jxta.document.Advertisement;

/**
 * The jxta DiscoveryService Service provides asynchronous mechanism in jxta for discovering
 * Peer Advertisements, Group Advertisements, and general jxta Advertisements
 * (pipe, service, etc.). The scope of discovery can controlled by specifying
 * name and attribute pair, and a threshold.  The threshold is an upper limit
 * the requesting peer specifies for responding peers not to exceed.
 * Each jxta Peer Group has an instance of a DiscoveryService Service the scope of the
 * discovery is limited to the group. for example
 *
 * <p>A peer in the soccer group invokes the soccer group's DiscoveryService Service
 * to discover pipe advertisements in the group, and is interested in a maximum of
 * 10 Advertisements from each peer:
 *
 * <pre>
 *  discovery.getRemoteAdvertisements(null, discovery.ADV,
 *                                      null, null,10);
 *
 * </pre>
 *
 * <pe>in the above example, peers that are part of the soccer group would
 *  respond.  After a getRemoteAdvertisements call is made, a call to
 *  getLocalAdvertisements can be made to retrieve results that have been found
 *  and added to the local group cache.
 *
 * <pe> DiscoveryService also provides a mechanism for publishing advertisements,
 * so that they may be discovered. The rules to follow when publishing are
 *
 * <ul type-disc>
 *   <li><pe>use the current discovery service to publish
 *    advertisements private to the group.
 *
 *
 * <pre>
 *        discovery.publish(adv,discovery.ADV);
 * </pre>
 *
 *    <li><pe>use the parent's discovery to publish advertisements that
 *    public outside of the group. e.g.  peer A would like publish the
 *    soccer group in the NetPeerGroup
 *
 *
 * <pre>
 *        parent=soccerGroup.getParent();
 *        discovery= parent.getDiscovery()
 *        discovery.publish(adv,discovery.ADV);
 * </pre>
 *</ul>
 *
 * <pe>The threshold can be utilized in peer discovery in situations
 * where a peer is only interested in other peers, and not about additional
 * peers they may know about.  to achieve this effect for peer discovery
 * set the Threshold to 0
 *
 * <pe>Advertisements are stored in a persist ant local cache. When a peers
 * boots up the same cache is referenced. this is an area where several
 * optimizations can take place, and intelligence about discovery patterns,
 * etc.
 *
 * <pe> Another feature of discovery is automatic discovery, a peer
 * initiates a discovery message by including it's own Advertisement in the
 * discovery message, which also can be viewed as a announcement.
 *
 * <pe>Another feature of discovery is a learning about other rendezvous, i.e.
 * when discovery comes across a peer advertisement of a rendezvous peer
 * it passes the information down to the endpoint router.
 *
 *
 * <pe>Message Format :
 *
 *
 * <pe>A discovery Query
 *
 * <pre>
 *      &lt;?xml version="1.0" standalone='yes'?&gt;
 *       &lt;DiscoveryQuery&gt;
 *        &lt;type&gt;int&lt;/type&gt;
 *        &lt;threshold&gt;int&lt;/threshold&gt;
 *        &lt;peeradv&gt;peeradv&lt;/peeradv&gt;
 *        &lt;attribute&gt;attribute&lt;/attribute&gt;
 *        &lt;value&gt;attribute&lt;/value&gt;
 *       &lt;/DiscoveryQuery&gt;
 * </pre>
 *
 *
 * <pe>A discovery Response
 *
 *      &lt;?xml version="1.0" standalone='yes'?&gt;
 *       &lt;DiscoveryResponse&gt;
 *        &lt;count&gt;int&lt;count&gt;
 *        &lt;type&gt;int&lt;/type&gt;
 *        &lt;peeradv&gt; adv &lt;/peeradv&gt;
 *        &lt;responses&gt;adv&lt;/responses&gt;
 *          ......
 *        &lt;responses&gt;adv&lt;/responses&gt;
 *       &lt;/DiscoveryResponse&gt;
 *
 * @see net.jxta.service.Service
 * @see net.jxta.resolver.Resolver
 * @see net.jxta.protocol.DiscoveryQueryMsg
 * @see net.jxta.protocol.DiscoveryResponseMsg
 * @see net.jxta.protocol.ResolverQueryMsg
 * @see net.jxta.protocol.ResolverResponseMsg
 *
 * @version $Revision: 1.4 $
 * @since JXTA 1.0 */

public interface Discovery extends Service {

  /**
   * Peer type
   */

    public static final int   PEER = 0;

   /**
    * Group type
    */

    public static final int  GROUP = 1;

  /**
   * DOC type
   */

    public static final int    ADV = 2;

 /**
  * This method discovers PeerAdvertisements, GroupAdvertisements
  * and jxta Advertisements. jxta Advertisements are documents that
  * describe pipes, services, etc.  The discovery scope
  * can be narrowed down firstly by Name and Value pair where the match
  * is an exact match, secondly by setting a upper limit where the
  * responding peer will not exceed.
  *
  * DiscoveryService can be performed in two ways
  *
  * 1. by specifying a null peerid, the discovery message is propagated
  *    on the local sub-net utilizing ip multicast. In addition to the
  *    multicast it is also propagated to rendezvous points.
  *
  * 2. by passing a peerid, the EndpointRouter will attempt to resolve
  *    destination peer's endpoints or route the message to other routers
  *    in attempt to reach the peer.
  *
  * @param attribute attribute name to narrow disocvery to
  *
  * @param value value of attribute to narrow disocvery to
  *
  * @param threshold the upper limit of responses from one peer
  *
  * @param peerid id of a peer, or "rendezvous" to connect to, if
  * address is null
  * @param type PEER, GROUP, ADV
  *
  * @version $Revision: 1.4 $
  * @since JXTA 1.0
  */

  public void getRemoteAdvertisements (String peerid, int type,
              String attribute, String value, int threshold);

 /**
  * Retrieve Stored Peer, Group, and General Advertisements
  *
  * @return Enumeration of stored advertisements/structured documents
  *
  * @param type PEER, GROUP, or ADV
  * @param attribute attribute name to narrow the response to
  * @param value value of the named attribute to narrow the response to
  * @exception IOException - If an I/O error occurs
  *
  * @version $Revision: 1.4 $
  * @since JXTA 1.0
  */

   public Enumeration getLocalAdvertisements (int type, String attribute,
                 String value) throws IOException;

 /**
  * Publish an advertisement
  * @param advertisement publish an adverisement within this group
  * @param type PEER, GROUP, ADV
  * @exception IOException - If an I/O error occurs
  *
  * @version $Revision: 1.4 $
  * @since JXTA 1.0
  */

    public void publish (Advertisement advertisement, int type)
        throws IOException;


 /**
  * Remote Publish an advertisement
  * @param advertisement publish an adverisement within this group
  * @param type PEER, GROUP, ADV
  *
  * @version $Revision: 1.4 $
  * @since JXTA 1.0
  */
    public void remotePublish (Advertisement adv, int type);

 /**
  * flush stored  Documents.
  *
  * @param id Document ID, Peer ID, or PeerGroup ID
  * @param type PEER, GROUP, ADV
  * @exception IOException - If an I/O error occurs
  *
  * @version $Revision: 1.4 $
  * @since JXTA 1.0
  */

    public void  flushAdvertisements(String id, int type)
        throws IOException ;

}
