/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: AdvertisementFactory.java,v 1.12 2001/06/08 23:03:22 bondolo Exp $
 */

package net.jxta.document;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.String;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ResourceBundle;
import java.util.Vector;

import java.lang.NoSuchMethodException;
import java.lang.reflect.InvocationTargetException;
import java.lang.SecurityException;
import java.util.MissingResourceException;

import net.jxta.util.ClassFactory;

/**
 * This class is a class factory for Advertisements. The class enables to hide
 * the implementation used to represent and create advertisements.
 *
 * <p>Advertisements are core objects that are used to advertise a Peer, a
 * PeerGroup, a Service, or a PipeService. The Advertisement class provides a platform
 * independent representation of core objects that can be exchanged between
 * different implementations (Java, C).
 *
 * <p>The AdvertisementFactory extends the ClassFactory to register the various
 * types of adverstisements into a static hastable. The factory is called with
 * the Advertisement type requested to create the corresponding advertisement
 * type.
 *
 * @see net.jxta.document.Advertisement
 * @see net.jxta.document.Document
 * @see net.jxta.document.MimeMediaType
 * @see net.jxta.peergroup.PeerGroup
 * @see net.jxta.protocol.PeerAdvertisement
 * @see net.jxta.protocol.PeerGroupAdvertisement
 * @see net.jxta.protocol.PipeAdvertisement
 *
 * @version $Revision: 1.12 $
 * @since JXTA 1.0
 */

public class AdvertisementFactory extends ClassFactory {
    
    /**
     *  Constructors registered with this factory need to have the following
     *  parameter sets.
     *
     *  For this factory the constructors are:
     *      constructor( void )
     *      constructor( Element )
     **/
    private static final Class requiredConstructors [] [] = {
        { },
        { net.jxta.document.Element.class }
    };
    
    /**
     *  The required set of static methods needed by classes registering with this
     *  factory.
     *
     *  For this factory the static methods are:
     *      String getAdvertisementType( )
     **/
    final static private Object [] [] requiredStaticMethods = {
        { "getAdvertisementType", new Class [] { } }
    };
    
    final static private int getAdvertisementTypeMethod = 0;
    
    /**
     *  This is the index in the constructor parameter set for the constructor
     *  used when constructing a Document from scratch.
     **/
    private static final int newInstanceConstructor = 0;
    /**
     *  This is the index in the constructor parameter set for the constructor
     *  used when constructing a portion of a StructuredDocument
     **/
    private static final int fromElementConstructor = 1;
    
    /**
     *  This class is in fact a singleton. This is the instance that backs the
     *  static methods.
     **/
    private static AdvertisementFactory factory = new AdvertisementFactory();
    
    /**
     *  This is the map of mime-types and constructors used by
     *  <CODE>newStructuredDocument</CODE>.
     **/
    private Hashtable  encodings = new Hashtable();
    
    /**
     *  If true then the pre-defined set of StructuredDocument sub-classes has
     *  been registered from the property containing them.
     **/
    private boolean loadedProperty = false;
    
    /**
     *  Registers the pre-defined set of StructuredDocument sub-classes so that
     *  this factory can construct them.
     *
     *  @return true if at least one of the StructuredDocument sub-classes could
     *  be registered otherwise false.
     **/
    private boolean doLoadProperty() {
        try {
            return registerFromResources( "net.jxta.impl.config",
            "AdvertisementInstanceTypes" );
        }
        catch ( MissingResourceException notFound ) {
            System.err.println( "Warning : Could not find net.jxta.impl properties file!" );
            return false;
        }
    }
    
    /**
     *  Used by ClassFactory methods to get the mapping of Mime Types to constructors.
     *
     *  @return the hashtable containing the mappings.
     **/
    protected Hashtable getAssocTable() {
        return encodings;
    }
    
    /**
     *  Used by ClassFactory methods to ensure that all of the instance classes
     *  which register with this factory have the correct base class
     *
     *  @return Class object of the key type.
     **/
    protected Class getBaseClassForFactory() {
        // our key is the doctype names.
        return Advertisement.class;
    }
    
    /**
     *  Used by ClassFactory methods to ensure that all keys used with the mapping are
     *  of the correct type.
     *
     *  @return Class object of the key type.
     **/
    protected Class getClassForKey() {
        // our key is the doctype names.
        return java.lang.String.class;
    }
    
    /**
     *  Used by ClassFactory methods to determine the required set of constructors
     *  needed by the instance classes generated by this factory.
     *
     *  @return Class object of the key type.
     **/
    protected Class [] []  getRequiredConstructors() {
        return requiredConstructors;
    }
    
    /**
     *  Used by ClassFactory methods to determine the required set of static
     *  methods needed by classes registering with this factory.
     *
     *  @return Object
     *
     *  @since JXTA 1.0
     */
    protected Object [] [] getRequiredStaticMethods() {
        return requiredStaticMethods;
    }
    
    /**
     *  Register a class with the factory from its class name. We override the
     *  standard implementation to get the advertisement type from the class and
     *  use that as the key to register the class with the factory.
     *
     * @param className The class name which will be regiestered.
     * @return boolean true if the class was registered otherwise false.
     * @Exception Exception   when an error occurs.
     *
     *  @since JXTA 1.0
     */
    protected boolean registerAssoc( String className ) {
        boolean registeredSomething = false;
        
        try {
            Class advClass = Class.forName( className );
            
            Method getAdvTypeMethod = advClass.getMethod( "getAdvertisementType", new Class [0] );
            
            String advType = (String) getAdvTypeMethod.invoke( null, new Object [0] );
            
            registeredSomething = registerAssoc( advType, advClass );
        }
        catch( Exception ignored ) {
            ignored.printStackTrace();
        }
        
        return registeredSomething;
    }
    
    /**
     * Register an array of constructor for an ID type to enable IDs of that type to be
     * constructed.
     *
     * @param rootType  the identifying value for this advertisement instance type
     * @param instanceClass   the constructors to use in constructing objects
     *                           of this rootTypeIns type.
     * @return boolean  true if the rootType type is registered. If there is already a
     *                   constructor for this type then false will be returned.
     *
     * @exception NoSuchMethodException invalid registration
     *
     * @exception SecurityException  security error
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */
    public static boolean registerAdvertisementInstance( String rootType, Class instanceClass )
    throws NoSuchMethodException, SecurityException {
        boolean result = factory.registerAssoc( rootType, instanceClass );
        
        return result;
    }
    
    /**
     * Constructs an instance of <CODE>Advertisement</CODE> matching the type
     * specified by the <CODE>advertisementType</CODE> parameter.
     *
     * @param advertisementType Specifies the mime media type to be associated with the
     * <CODE>StructuredDocument</CODE> to be created.
     * @return The instance of <CODE>Advertisement</CODE> or null if it
     * could not be created.
     *
     * @exception InvocationTargetException error invoking target constructor
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public static Advertisement newAdvertisement( String advertisementType )
    throws InvocationTargetException {
        Object params [] = { };
        
        if( !factory.loadedProperty )
            factory.loadedProperty = factory.doLoadProperty();
        
        return (Advertisement) factory.newInstance( advertisementType,
        requiredConstructors[AdvertisementFactory.newInstanceConstructor], params );
    }
    
    /**
     * Constructs an instance of <CODE>Advertisement</CODE> matching the type
     * specified by the <CODE>advertisementType</CODE> parameter.
     *
     * @param mimetype Specifies the mime media type to be associated with the
     * <CODE>StructuredDocument</CODE> to be created.
     * @param stream imput stream used to read data to construct the advertisement
     * @return The instance of <CODE>Advertisement</CODE> or null if it
     * could not be created.
     *
     * @exception InvocationTargetException error invoking target constructor
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public static Advertisement newAdvertisement( MimeMediaType mimetype, InputStream stream )
    throws InvocationTargetException {
        
        StructuredTextDocument doc = (StructuredTextDocument)
        StructuredDocumentFactory.newStructuredDocument( mimetype, stream );
        
        return newAdvertisement( doc );
    }
    
    /**
     * Constructs an instance of <CODE>Advertisement</CODE> matching the type
     * specified by the <CODE>advertisementType</CODE> parameter.
     *
     * @param root Specifies a portion of a StructuredDocument which will be
     * converted into an Advertisement.
     * @return The instance of <CODE>Advertisement</CODE> or null if it
     * could not be created.
     *
     * @exception InvocationTargetException error invoking target constructor
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */
    public static Advertisement newAdvertisement( TextElement root )
    throws InvocationTargetException {
        Object params [] = { root };
        
        if( !factory.loadedProperty )
            factory.loadedProperty = factory.doLoadProperty();
        
        return (Advertisement) factory.newInstance( root.getName(),
        requiredConstructors[AdvertisementFactory.fromElementConstructor], params );
    }
    
}
