/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Document.java,v 1.10 2001/06/08 17:09:36 bondolo Exp $
 */

package net.jxta.document;

import java.io.InputStream;
import java.io.OutputStream;

import java.io.IOException;

/**
 * JXTA documents are simple containers for data. JXTA makes no interpretation
 * of the data contained in a Document. Each Document is typed using a
 * MIME Media Type (as defined by 
 * {@see <a href=http://www.ietf.org/rfc/rfc2046.txt">MIME : Media Types</a>})
 *
 * Note : Document is intended to be somewhat analagous to HTTP streams
 * we don't provide for compression, encoding, etc. though and leave it all
 * up to the MIME media type and whatever parameters it may specify. Its not
 * clear if we should break this stuff up and include such attributes 
 * or leave it as is.
 *
 * @see         net.jxta.document.Document
 * @see         net.jxta.document.StructuredDocument
 * @see         net.jxta.document.StructuredTextDocument
 *
 * @version  $Revision: 1.10 $
 * @since JXTA 1.0
 */
public interface Document {

    /**
     * Returns the MIME Media type of this Document per
     * {@see <a href=http://www.ietf.org/rfc/rfc2046.txt">MIME : Media Types</a>}
     *
     * Jxta does not currently support the 'Multipart' or 'Message' media types.
     *
     * @return A MimeMediaType object containing the MIME Media Type for this 
     * Document.
     *
     * @since JXTA 1.0 
     */
    public abstract MimeMediaType getMimeType();

    /**
     * Returns a stream of bytes which represent the content of this Document.
     *
     * @return An {@see java.io.InputStream InputStream} containing the bytes of this Document.
     * @exception  IOException if an I/O error occurs.
     *
     * @since JXTA 1.0 
     */
    public abstract InputStream getStream() throws IOException;
    
    /**
     *  Rather than returning an InputStream like {@link #getStream}, send the
     *  document to the specified stream. 
     *  
     *  @param stream   The OutputStream to which the document will be written.
     *  @exception  IOException if an I/O error occurs.
     *
     *  @since JXTA 1.0 
     */
    public abstract void sendToStream( OutputStream stream ) throws IOException;

}
