/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: MimeMediaType.java,v 1.11 2001/06/22 16:42:59 bondolo Exp $
 */

package net.jxta.document;

import java.lang.String;
import java.util.Enumeration;
import java.util.Vector;

import java.lang.IllegalArgumentException;

/**
 *  This class is a utility class with mostly provides the common parsing
 *  features needed to deal with MIME media types. 
 *
 *  FIXME bondolo@jxta.org 20010219 Doesnt yet handle parameters
 *
 * @see         net.jxta.document.Document
 * @see         net.jxta.document.StructuredDocument
 * @see         net.jxta.document.StructuredTextDocument
 *
 * @version $Revision: 1.11 $ 
 * @since JXTA 1.0
 */
public class MimeMediaType {

    private String type = null;

    private String subtype = null;

    private Vector parameters = new Vector();

    /** 
     * Creates a new MimeMediaType 
     *
     * @param someString string representing a mime-type  
     *
     * @since JXTA 1.0
     */
    public MimeMediaType( final String someString ) {

        if( null == someString )
            throw new IllegalArgumentException( "input cannot be null" );

        String cleaned = someString.trim().toLowerCase();

        if( 0 == cleaned.length() )
            throw new IllegalArgumentException( "input cannot be null" );

        // determine the type
        int typeSepAt = findNextSeperator( cleaned );

        if( (-1 == typeSepAt) || (0 == typeSepAt) )
            throw new IllegalArgumentException( 
	    "expected seperator or seperator in unexpected location" );

        String itsType = cleaned.substring( 0, typeSepAt );

        // determine the sub-type
        int subtypeSepAt = findNextSeperator( cleaned, typeSepAt + 1 );

        String itsSubType = null;
        String itsParams = null;

        if(-1 == subtypeSepAt)
            itsSubType = cleaned.substring( typeSepAt + 1 );
        else {
            itsSubType = cleaned.substring( typeSepAt + 1, subtypeSepAt);
            itsParams = cleaned.substring( subtypeSepAt );  
	    // include the seperator, its significant
        }

        Vector someParams = new Vector();

        setType( itsType );
        setSubtype( itsSubType );
        setParameters( new String [0] );
    };

    /** 
     * Creates a new type/subtype MimeMediaType 
     *
     * @param type string representing a mime type 
     * @param subtype string representing a mime subtype
     *
     * @since JXTA 1.0 
     */
    
    public MimeMediaType ( final String type, final String subtype ) {
        this( type, subtype, (String) null );
    }

    /** 
     * Creates a new type/subtype MimeMediaType 
     *
     * @param type string representing a mime type 
     * @param subtype string representing a mime subtype
     * @param parameters parameters to the mime-type constructor
     *
     * @since JXTA 1.0 
     */

    public MimeMediaType ( final String type, final String subtype,
			   final String parameters ) {
        Vector someParams = new Vector();

        if( null != parameters ) {
            // FIXME bondolo@jxta.org 20010219 proccessing of parameters still needs to be done.
        }

        setType( type );
        setSubtype( subtype );
        setParameters( new String [0] );
    };

    /** 
     * Creates a new type/subtype MimeMediaType 
     *
     * @param type string representing a mime type 
     * @param subtype string representing a mime subtype
     * @param parameters[]  array of parameters to the mime-type constructor
     *
     * @since JXTA 1.0 
     */

    public MimeMediaType ( final String type, final String subtype, 
			   final String [] parameters ) {
        setType( type );
        setSubtype( subtype );
        setParameters( parameters );
    };

    /**
     * Creates a new object of the same class as this object. It then
     * initializes each of the new object's fields by assigning it the
     * same value as the corresponding field in this object. No
     * constructor is called.
     * <p>
     * The <code>clone</code> method of class <code>Object</code> will
     * only clone an object whose class indicates that it is willing for
     * its instances to be cloned. A class indicates that its instances
     * can be cloned by declaring that it implements the
     * <code>Cloneable</code> interface.
     *
     * @return  Object a clone of this instance.
     * @exception  CloneNotSupportedException  if the object's class does not
     *              support the <code>Cloneable</code> interface. Subclasses
     *              that override the <code>clone</code> method can also
     *              throw this exception to indicate that an instance cannot
     *              be cloned.
     *
     * @since JXTA 1.0 
     */

    public Object clone() throws CloneNotSupportedException {

        Object retValue = new MimeMediaType( toString() );

        return retValue;
    }

    /**
     * Compares two Objects for equality.
     * <p>
     * The <code>equals</code> method implements an equivalence relation:
     * <ul>
     * <li>It is <i>reflexive</i>: for any reference value <code>x</code>,
     *    <code>x.equals(x)</code> should return <code>true</code>.
     * <li>It is <i>symmetric</i>: for any reference values <code>x</code> and
     *    <code>y</code>, <code>x.equals(y)</code> should return
     *    <code>true</code> if and only if <code>y.equals(x)</code> returns
     *    <code>true</code>.
     * <li>It is <i>transitive</i>: for any reference values <code>x</code>,
     *    <code>y</code>, and <code>z</code>, if <code>x.equals(y)</code>
     *    returns  <code>true</code> and <code>y.equals(z)</code> returns
     *    <code>true</code>, then <code>x.equals(z)</code> should return
     *    <code>true</code>.
     * <li>It is <i>consistent</i>: for any reference values <code>x</code>
     *    and <code>y</code>, multiple invocations of <code>x.equals(y)</code>
     *    consistently return <code>true</code> or consistently return
     *    <code>false</code>.
     * <li>For any reference value <code>x</code>, <code>x.equals(null)</code>
     *    should return <code>false</code>.
     * </ul>
     * <p>
     * The equals method for class <code>Object</code> implements the most
     * discriminating possible equivalence relation on objects; that is,
     * for any reference values <code>x</code> and <code>y</code>, this
     * method returns <code>true</code> if and only if <code>x</code> and
     * <code>y</code> refer to the same object (<code>x==y</code> has the
     * value <code>true</code>).
     *
     * @param   obj   the reference object with which to compare.
     * @return  boolean <code>true</code> if this object is the same as the obj
     *         argument; <code>false</code> otherwise.
     *
     * @since JXTA 1.0 
     */

    public boolean equals(Object obj) {
        if (this == obj)
            return true;

        if( !(obj instanceof MimeMediaType) )
            return false;

        boolean retValue = (hashCode() == obj.hashCode());
        return retValue;
    }

    /**
     * Returns a hash code value for the object. This method is
     * supported for the benefit of hashtables such as those provided by
     * <code>java.util.Hashtable</code>.
     * <p>
     * The general contract of <code>hashCode</code> is:
     * <ul>
     * <li>Whenever it is invoked on the same object more than once during
     *    an execution of a Java application, the <code>hashCode</code> method
     *    must consistently return the same integer. This integer need not
     *    remain consistent from one execution of an application to another
     *    execution of the same application.
     * <li>If two objects are equal according to the <code>equals</code>
     *    method, then calling the <code>hashCode</code> method on each of the
     *    two objects must produce the same integer result.
     * </ul>
     *
     * @return int a hash code value for this object.
     *
     * @since JXTA 1.0 
     */

    public int hashCode() {
        return getMimeMediaType().hashCode();
    }

    /**
     * Returns a string representation of the object. In general, the
     * <code>toString</code> method returns a string that
     * "textually represents" this object. The result should
     * be a concise but informative representation that is easy for a
     * person to read.
     * It is recommendedthat all subclasses override this method.
     * <p>
     * The <code>toString</code> method for class <code>Object</code>
     * returns a string consisting of the name of the class of which the
     * object is an instance, the at-sign character `<code>@</code>', and
     * the unsigned hexadecimal representation of the hash code of the
     * object.
     *
     * @return String  a string representation of the object.
     *
     * @since JXTA 1.0 
     */

    public String toString() {
        if( !isValid() )
            return null;

        String retValue = type + "/" + subtype;

        for( Enumeration eachParameter = parameters.elements(); 
	     eachParameter.hasMoreElements(); )
	  retValue += ";" + (String) eachParameter.nextElement();

        return retValue;
    };

    /** 
     * Check if mime-type is valid
     *
     * @return booelan true if mime-type is well-formed
     *
     * @since JXTA 1.0
     */

    public boolean isValid() {
        boolean result = ( (null != type) && (null != subtype) );

        return result;
    };

    /** 
     * Get the full mime-type/subtype
     *
     * @return full mime-type/subtype
     *
     * @since JXTA 1.0
     */

    public String getMimeMediaType() {
        if( !isValid() )
            return null;

        String retValue = type + "/" + subtype;

        return retValue;
    }

    /** 
     * Get type of the mime-type
     *
     * @return type of the mime-type
     *
     * @since JXTA 1.0
     */

    public String getType() {
        return type;
    }

    /** 
     * Check if the mime-type is for provisional.
     *
     * @return boolean  true if it is a provisional type
     *
     * @since JXTA 1.0
     */


    public boolean isExperimentalType() {
        if( (null == type) || (type.length() < 2) )
            return false;

        return( ('x' == type.charAt(0)) && ('-' == type.charAt(1)) );
    }

    /** 
     * Set the type of MimeMediaType 
     *
     * @param type type value
     *
     * @since JXTA 1.0
     */


    public void setType( final String type ) {
        if( null == type )
            throw new IllegalArgumentException( "type cannot be null" );

        String cleaned = type.trim().toLowerCase();

        if( 0 == cleaned.length() )
            throw new IllegalArgumentException( "type cannot be null" );

        if( -1 != findNextSeperator( cleaned ) )
            throw new IllegalArgumentException( "type cannot contain a seperator" );

        this.type = cleaned;
    };

    /** 
     * Get the Subtype of the mime-type
     *
     * @retrun subtype of the mime-type
     *
     * @since JXTA 1.0
     */

    public String getSubtype() {
        return subtype;
    }

    /** 
     * Check if the mime-type is for debugging. This method will be
     * removed 
     *
     * @return boolean  true if it is a debugging type
     *
     * @since JXTA 1.0
     */

    public boolean isExperimentalSubtype() {
        if( (null == subtype) || (subtype.length() < 2) )
            return false;

        return( ('x' == subtype.charAt(0)) && ('-' == subtype.charAt(1)) );
    }

    /** 
     * Set the subtype of MimeMediaType 
     *
     * @param subtype subtype value
     *
     * @since JXTA 1.0
     */

    public void setSubtype( String subtype ) {
        if( null == subtype )
            throw new IllegalArgumentException( "subtype cannot be null" );

        String cleaned = subtype.trim().toLowerCase();

        if( 0 == cleaned.length() )
            throw new IllegalArgumentException( "subtype cannot be null" );

        if( -1 != findNextSeperator( cleaned ) )
            throw new IllegalArgumentException( "subtype cannot contain a seperator" );

        this.subtype = cleaned;
    };

    /** 
     * Get parameters value of a mime-type
     *
     * @return Enumeration return enumeration of parameters for this mime-type
     *
     * @since JXTA 1.0
     */

    public Enumeration getParameters() {
        return parameters.elements();
    };

    /** 
     * Set parameter to a mime-type
     *
     * @param parameters array of parameters
     *
     * @since JXTA 1.0
     */

    public void setParameters( String [] parameters ) {
        this.parameters = new Vector();

        for( int eachParameter = 0; eachParameter < parameters.length; eachParameter++ )
            addParameter( parameters[eachParameter] );
    }

    /** 
     * Add parameter to a mime-type
     *
     * @param newParam new parameter to add to the type
     *
     * @since JXTA 1.0
     */

    public void addParameter( String newParam ) {
        parameters.addElement( newParam );
    };

    private final static String CTL =
    "\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007" +
    "\u0008\u0009\n\u000b\u000c\r\u000e\u000f" +
    "\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017" +
    "\u0018\u0019\u001a\u001b\u001c\u001d\u001e\u001f" +
    "\u007f";

    private final static String space = "\u0020";
    private final static String tspecials = "()<>@,;:\\\"/[]?=";
    private final static String terminator = CTL + space + tspecials;

    /** 
     * Find next separator position  in mime-type 
     *
     * @param source  source location
     * @return int separator location
     *
     * @since JXTA 1.0
     */

    private int findNextSeperator( final String source ) {
        return findNextSeperator( source, 0 );
    };

    /** 
     * Find next separator position  in mime-type 
     *
     * @param source  source location
     * @param from  from location
     * @return int separator location
     *
     * @since JXTA 1.0
     */

    private static int findNextSeperator( final String source, int from ) {

        int seperator = -1;

        // find a seperator
        for( int eachChar = from; eachChar < source.length(); eachChar++ )
            if( -1 != terminator.indexOf( source.charAt( eachChar ) ) ) {
                seperator = eachChar;
                break;
            }

        return seperator;
    }
}
