/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: StructuredDocumentFactory.java,v 1.13 2001/06/12 17:40:32 bondolo Exp $
 */
package net.jxta.document;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.String;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ResourceBundle;
import java.util.Vector;

import java.lang.NoSuchMethodException;
import java.lang.reflect.InvocationTargetException;
import java.lang.SecurityException;
import java.util.MissingResourceException;

import net.jxta.util.ClassFactory;

/**
 * This class provides a static method for constructing instances of
 * various sub-classes of <CODE>StructuredDocument</CODE>. Behind the
 * scene, it also provides for the registration of the mime-types and
 * constructors needed to accomplish the construction.All supported
 * mime-types will need to register their implementation in this factory.
 *
 * <p>The configuration is done via the property "StructuredDocumentInstanceTypes"
 *
 * <p>FIXME bondolo@jxta.org 20010223 This implementation needs a way to return the list of
 * mime media types supported
 *
 * @see         net.jxta.document.Document
 * @see         net.jxta.document.StructuredTextDocument
 * @see         net.jxta.document.StructuredDocument
 * @see         net.jxta.document.MimeMediaType
 *
 * @version $Revision: 1.13 $
 * @since JXTA 1.0
 */

public class StructuredDocumentFactory extends ClassFactory {
    
    /**
     *  Constructors registered with this factory need to have the following
     *  parameter sets.
     *
     *  For this factory the constructors are:
     *      constructor( MimeMediaType, String )
     *      constructor( MimeMediaType, InputStream )
     **/
    
    private static final Class requiredConstructors [] [] = {
        { net.jxta.document.MimeMediaType.class, java.lang.String.class },
        { net.jxta.document.MimeMediaType.class, java.io.InputStream.class }
    };
    
    /**
     *  The required set of static methods needed by classes registering with this
     *  factory.
     *
     *  For this factory the static methods are:
     *      MimeMediaType []  getSupportedMimeTypes( )
     *
     *  @since JXTA 1.0
     */
    final static private Object [] [] requiredStaticMethods = {
        { "getSupportedMimeTypes", new Class [] { } }
    };
    
    final static private int getSupportedMimeTypesMethod = 0;
    
    /**
     *  This is the index in the constructor parameter set for the constructor
     *  used when constructing a Document from scratch.
     **/
    private static final int newInstanceConstructor = 0;

    /**
     *  This is the index in the constructor parameter set for the constructor
     *  used when constructing a Document from a stream.
     **/
    private static final int fromStreamConstructor = 1;
    
    /**
     *  This class is in fact a singleton. This is the instance that backs the
     *  static methods.
     **/
    private static StructuredDocumentFactory factory =
        new StructuredDocumentFactory();
    
    /**
     *  This is the map of mime-types and constructors used by
     *  <CODE>newStructuredDocument</CODE>.
     **/
    
    private Hashtable  encodings = new Hashtable();
    
    /**
     *  If true then the pre-defined set of StructuredDocument sub-classes has
     *  been registered from the property containing them.
     **/
    
    private boolean loadedProperty = false;
    
    /**
     *  Registers the pre-defined set of StructuredDocument sub-classes so that
     *  this factory can construct them.
     *
     *  @return boolean true if at least one of the StructuredDocument sub-classes could
     *  be registered otherwise false.
     *
     * @since JXTA 1.0
     */
    
    private boolean doLoadProperty() {
        try {
            return registerFromResources( "net.jxta.impl.config",
            "StructuredDocumentInstanceTypes" );
        }
        catch ( MissingResourceException notFound ) {
            System.err.println
            ( "Warning : Could not find net.jxta.impl properties file!" );
            return false;
        }
    }
    
    /**
     *  Used by ClassFactory methods to get the mapping of Mime Types to constructors.
     *
     *  @return Hastable the hashtable containing the mappings.
     *
     * @since JXTA 1.0
     *
     */
    
    protected Hashtable getAssocTable() {
        return encodings;
    }
    
    /**
     *  Used by ClassFactory methods to ensure that all keys used with the mapping are
     *  of the correct type.
     *
     *  @return Class object of the key type.
     *
     * @since JXTA 1.0
     */
    
    protected Class getClassForKey() {
        return java.lang.String.class;
    }
    
    /**
     *  Used by ClassFactory methods to ensure that all of the instance classes
     *  which register with this factory have the correct base class
     *
     *  @return Class object of the "Factory" type.
     *
     * @since JXTA 1.0
     */
    
    protected Class getBaseClassForFactory() {
        // our key is the doctype names.
        return StructuredDocument.class;
    }
    
    /**
     *  Used by ClassFactory methods to determine the required set of constructors
     *  needed by the instance classes generated by this factory.
     *
     *  @return Class[] required constructors
     *
     * @since JXTA 1.0
     */
    
    protected Class [] []  getRequiredConstructors() {
        return requiredConstructors;
    }
    
    /**
     *  Used by ClassFactory methods to determine the required set of static
     *  methods needed by classes registering with this factory.
     *
     *  @return Object
     *
     *  @since JXTA 1.0
     */
    protected Object [] [] getRequiredStaticMethods() {
        return requiredStaticMethods;
    }
    
    /**
     *  Register a class with the factory from its class name. We override the
     *  standard implementation to get the advertisement type from the class and
     *  use that as the key to register the class with the factory.
     *
     * @param className The class name which will be regiestered.
     * @return boolean true if the class was registered otherwise false.
     * @Exception Exception   when an error occurs.
     *
     *  @since JXTA 1.0
     */
    protected boolean registerAssoc( String className ) {
        boolean registeredSomething = false;
        
        //System.err.println( "Registering : " + className );

        try {
            Class docClass = Class.forName( className );
            
            Method getMimeTypesMethod = docClass.getMethod(
                (String) requiredStaticMethods[getSupportedMimeTypesMethod][0],
                (Class []) requiredStaticMethods[getSupportedMimeTypesMethod][1] );
            
            MimeMediaType [] mimeTypes = 
                (MimeMediaType []) getMimeTypesMethod.invoke( null, new Object [0] );
            
            for( int eachType = 0; eachType < mimeTypes.length; eachType++ ) {
                //System.err.println( "   Registering Type : " + mimeTypes[eachType].getMimeMediaType() );

                registeredSomething |= 
                    registerAssoc( mimeTypes[eachType].getMimeMediaType(), docClass );
            }
        }
        catch( Exception ignored ) {
            //ignored.printStackTrace();
        }
        
        return registeredSomething;
    }
    
    /**
     * Register an array of constructor for a mime-type to enable
     * that mime-type of documents to be constructed.
     *
     * @param cleanMime   the mime-type associated
     * @param instanceClass the instance class that wants to be registered..
     * @return boolean true   if the ID type is registered. If there is already a
     *                   instance class for this type then false will be returned.
     *
     * @exception NoSuchMethodException missing method
     * @exception SecurityException   security exception when registering new type
     *
     * @since JXTA 1.0
     */
    
    public static boolean registerEncoding( MimeMediaType cleanMime, Class instanceClass )
    throws NoSuchMethodException, SecurityException {
        boolean result = factory.registerAssoc( cleanMime.getMimeMediaType(),
        instanceClass );
        
        return result;
    }
    
    /**
     * Constructs an instance of <CODE>StructuredDocument</CODE> matching
     * the mime-type specified by the <CODE>mimetype</CODE> parameter. The
     * <CODE>doctype</CODE> parameter identifies the base type of the
     * <CODE>StructuredDocument</CODE>.
     *
     * @param mimetype Specifies the mime media type to be associated with
     * the <CODE>StructuredDocument</CODE> to be created.
     *
     * @param doctype Specifies the root type of the
     * <CODE>StructuredDocument</CODE> to be created.
     *
     * @return StructuredDocument The instance of
     * <CODE>StructuredDocument</CODE> or null if it could not be created.
     *
     * @exception InvocationTargetException invalid mime-media-type or constructors
     *
     * @since JXTA 1.0
     */
    
    public static StructuredDocument newStructuredDocument( MimeMediaType mimetype,
    String doctype )
    throws InvocationTargetException {
        Object params [] = { mimetype, doctype };
        
        if( !factory.loadedProperty )
            factory.loadedProperty = factory.doLoadProperty();
        
        return (StructuredDocument) factory.newInstance(
        mimetype.getMimeMediaType(),
        requiredConstructors[StructuredDocumentFactory.newInstanceConstructor], params );
    }
    
    /**
     * Constructs an instance of <CODE>StructuredDocument</CODE> matching
     * the mime-type specified by the <CODE>mimetype</CODE> parameter. The
     * <CODE>doctype</CODE> parameter identifies the base type of the
     * <CODE>StructuredDocument</CODE>.
     *
     * @param mimetype Specifies the mime media type to be associated with the
     * <CODE>StructuredDocument</CODE> to be created.
     *
     * @param stream Contains an InputStream from which the document will be constructed.
     *
     * @return StructuredDocument The instance of
     * <CODE>StructuredDocument</CODE> or null if it could not be created.
     *
     * @exception InvocationTargetException invalid mime-media-type or constructors
     *
     * @since JXTA 1.0
     */
    
    public static StructuredDocument newStructuredDocument( MimeMediaType mimetype,
    InputStream stream )
    throws InvocationTargetException {
        Object params [] = { mimetype, stream };
        
        if( !factory.loadedProperty )
            factory.loadedProperty = factory.doLoadProperty();
        
        return (StructuredDocument) factory.newInstance(
        mimetype.getMimeMediaType(),
        requiredConstructors[StructuredDocumentFactory.fromStreamConstructor], params );
    }
}
