/************************************************************************
 *
 * $Id: EndpointAddress.java,v 1.9 2001/06/09 22:46:57 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.endpoint;

import java.io.IOException;

/**
 * The EndpointAddress class defines the JXTA Endpoint address. Each peer
 * can have as many EndpointAdresses as it needs
 *
 * A JXTA endpoint address is a URI that has the following form:
 *
 * <p>protocol://network address/service name/optional parameters
 *
 * <p>For instance, a JXTA Peer on TCP, will have its PipeService Service address
 * defined as:
 *
 * <p>tcp://192.168.1.10/JxtaPipe
 *
 * <p>A particular pipe provided by the service would have the following address:
 *
 * <p>tcp://192.168.1.10/JxtaPipe/000...128 bit UUID of the pipe
 *
 * @see  net.jxta.endpoint.Message
 * @see  net.jxta.pipe.InputPipe
 * @see  net.jxta.pipe.OutputPipe
 * @see  net.jxta.pipe.Pipe
 *
 * @version     $Revision: 1.9 $
 * @since JXTA 1.0
 */

public interface EndpointAddress {

  /**
   * Return a String that contains the canonical representation of the
   * Endpoint Address.
   *
   * @return String a String containing the EndpointAddress
   * @exception IOExcetion if the EndpointAddress is malformed.
   *
   * @deprecated Please use toString() instead.
   *
   * @since JXTA 1.0
   */

  public String makeString() throws java.io.IOException;

  /**
   * Return a String that contains the name of the protocol
   * contained in the EndpointAddress
   *
   * @return a String containing the protocol name
   * @throw IOExcetion if the EndpointAddress is malformed.
   *
   * @since JXTA 1.0
   */

  public String getProtocolName() throws IOException;

  /**
   * Return a String that contains the name of the protocol
   * address contained in the EndpointAddress
   *
   * @return String a String containing the protocol address
   * @exception IOExcetion if the EndpointAddress is malformed.
   *
   * @since JXTA 1.0
   */

  public String getProtocolAddress() throws IOException;

  /**
   * Return a String that contains the name of the service name
   * contained in the EndpointAddress
   *
   * @return String a String containing the service name
   * @exception IOExcetion if the EndpointAddress is malformed.
   *
   * @since JXTA 1.0
   */

  public String getServiceName() throws IOException;

  /**
   * Return a String that contains the service
   * parameter contained in the EndpointAddress
   *
   * @return String a String containing the service parameter
   * @exception IOExcetion if the EndpointAddress is malformed, or if
   * there is no parameter associated with the service
   *
   * @since JXTA 1.0
   */

  public String getServiceParameter() throws IOException;

  /**
   * Set the protocol name.
   *
   * @param name String containing the name of the protocol
   *
   * @since JXTA 1.0
   */

  public void setProtocolName(String name);

  /**
   * Set the protocol address.
   *
   * @param address String containing the peer address.
   *
   * @since JXTA 1.0
   */

  public void setProtocolAddress(String address);

  /**
   * Set the service name.
   *
   * @param name String containing the name of the service
   *
   * @since JXTA 1.0
   */

  public void setServiceName(String name);

  /**
   * Set the service parameter
   *
   * @param name String containing the service parameter
   *
   * @since JXTA 1.0
   */

  public void setServiceParameter(String name);

}



