/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ID.java,v 1.12 2001/05/23 00:28:21 lomax Exp $
 */

package net.jxta.id;

import java.lang.reflect.Constructor;
import java.lang.String;
import java.net.URL;
import java.util.zip.Adler32;                           // used in hashCode

import java.lang.IllegalArgumentException;
import java.security.NoSuchAlgorithmException;

import net.jxta.peergroup.PeerGroupID;
import net.jxta.util.ClassFactory;

import net.jxta.impl.id.UUID;
import net.jxta.impl.id.UUIDFactory;

public abstract class ID implements Cloneable {
    
    /**
     * This class represents a JXTA ID. IDs are used to uniquely identify a
     * peer, a peer group and a pipe or any other objects manipulated by the
     * core. JXTA ID are represented as UUID.
     *
     * @see net.jxta.id.IDFactory
     * @see net.jxta.id.UUID
     * @see net.jxta.id.UUIDFactory
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    final static protected int IdByteArraySize = 64;
    // uuid + uuid + digest + 11 bytes of pad + encoding
    
    final static protected int flagsSize = 2;
    
    final static protected int flagsIdTypeOffset = ID.flagsSize - 2;
    final static protected byte flagNullID = 0x00;
    final static protected byte flagCodatID = 0x01;
    final static protected byte flagPeerGroupID = 0x02;
    final static protected byte flagPeerID = 0x03;
    final static protected byte flagPipeID = 0x04;
    
    final static protected int flagsEncodingOffset = ID.flagsSize - 1;
    final static protected byte UUIDEncoded = 0x01;
    
    final static protected int flagsOffset = ID.IdByteArraySize - ID.flagsSize;
    
    protected byte [] bytes = null;
    
    /**
     *	The null ID. Useful for comparisons and cases where the Id is ignored.
     **/
    final static public ID nullID = new ID() {
        
        public Object clone() {
            return this;  // null is only itself.
        }
        
        /**
         * Compares two IDs for equality.
         *
         * @param target  the ID to be compared against.
         * @return	boolean true if CodatIds are equal, false otherwise.
         *
         *  @since JXTA 1.0
         */
        public boolean equals( Object target ) {
            return (this == target);   // null is only itself.
        }
        
        public PeerGroupID getPeerGroupID( ) {
            PeerGroupID result = new PeerGroupID( new UUID( 0x0000000000000000L,
            0x0000000000000000L ) ); // all nulls;
            return result;
        }
    };
    
    final static public String URIEncodingName = "jxta";
    
    /**
     *  Constructor for the null case. This constructor initializes only the
     *  flag fields of the ID. This is not usually a useful
     *  operation since there are already constants. Provided for subclasses to
     *  make their own null constants.
     *
     * @since JXTA 1.0
     */
    
    public ID() {
        this.bytes = new byte[ ID.IdByteArraySize ];
        this.bytes[ID.flagsOffset + ID.flagsIdTypeOffset] = ID.flagNullID;
        this.bytes[ID.flagsOffset + ID.flagsEncodingOffset] = ID.UUIDEncoded;
    };
    
    /**
     *  Returns true if this CodatId is in the same Peer Group as the target.
     *
     *  @param  target  The CodatId which will checked for a Peer Group match.
     *  @return boolean true if the CodatIds are in the same Peer Group otherwise false.
     *
     *  @since JXTA 1.0
     */
    
    public boolean isSamePeerGroup( ID target ) {
        PeerGroupID targetGroup = target.getPeerGroupID();
        PeerGroupID myGroup = this.getPeerGroupID();
        boolean result = myGroup.equals(targetGroup);
        
        return result;
    }
    
    /**
     *  Returns PeerGroupId of the Peer Group to which this CodatId belongs.
     *
     *  @return PeerGroupId of the Peer Group which this CodatId is part of.
     *
     *  @since JXTA 1.0
     */
    
    public abstract PeerGroupID getPeerGroupID( );
    
    
    /**
     *  Returns a clone of this object.
     *
     *  @return the cloned object
     */
    public Object clone() throws CloneNotSupportedException {
        throw new CloneNotSupportedException();
    }
    
    /**
     * Compares two CodatIds for equality.
     *
     * @param target  the CodatID to be compared against.
     * @return	boolean true if CodatIds are equal, false otherwise.
     *
     *  @version     $Revision: 1.12 $
     *  @since JXTA 1.0
     */
    public boolean equals( Object target ) {
        if (this == target)
            return true;
        
        if (target instanceof ID) {
            for( int eachByte = 0; eachByte < bytes.length; eachByte++ )
                if( bytes[eachByte] != ((ID)target).bytes[eachByte] )
                    return false;
            
            return true;
        }
        else
            return false;
    }
    
    /**
     * Public member calculates a hash code for this ID. Used by Hashmaps.
     *
     * @return int Containing the hashcode of this ID.
     *
     *  @version     $Revision: 1.12 $
     *  @since JXTA 1.0
     *
     */
    
    public int hashCode() {
        Adler32 crc = new Adler32();
        
        crc.update( bytes );
        return (int) crc.getValue();
    };
    
    /**
     * Public member which returns a string representation of the ID.
     * This implementations encodes the Codat ID into a URI.
     *
     * @return	String containting the URI
     *
     *  @version     $Revision: 1.12 $
     *  @since JXTA 1.0
     */
    
    public String toString( ) {
        return getURL().toString();
    };
    
    /**
     *
     *  Private replacement for toHexString since we need the leading 0 digits.
     *  Returns a String containing byte value encoded as 2 hex characters.
     *
     *  @param  theByte a byte containing the value to be encoded.
     *  @return	String containing byte value encoded as 2 hex characters.
     */
    private String toHexDigits( byte theByte ) {
        final char [] HEXDIGITS = { '0', '1', '2', '3', '4', '5', '6', '7', 
                                    '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };
        StringBuffer result = new StringBuffer(2);
        
        result.append( HEXDIGITS[(theByte >>> 4) & 15] );
        result.append( HEXDIGITS[theByte & 15] );
        
        return result.toString();
    }
    
    /**
     * Public member which returns a URI (URL in Java nomenclature) of the ID.
     *
     * @return	URL Object containing the URI
     *
     *  @since JXTA 1.0
     */
    
    public URL getURL( ) {
        URL             result = null;
        StringBuffer    encoded = new StringBuffer();
        
        for( int eachByte = 0; eachByte < bytes.length; eachByte++ ) {
            encoded.append( toHexDigits(bytes[eachByte]) );
        }
        
        try {
            result = new URL( ID.URIEncodingName, encoded.toString().toUpperCase(), "" );
        }
        catch ( java.net.MalformedURLException ignored ) {
            System.out.println( "URL Handler for '" + ID.URIEncodingName +
            "' could not be found. Check your classpath for sun.net.www.protocol." +
            ID.URIEncodingName + ".Handler" );
        }
        
        return result;
    };
    
    /**
     * convert a long into the byte array
     *
     * @param offset offset
     * @paran value  value
     *
     *  @since JXTA 1.0
     */
    
    protected void longIntoBytes( int offset, long value ) {
        for( int eachByte = 0; eachByte < 8; eachByte ++ )
            bytes[eachByte + offset] = (byte) (value >> ((7 - eachByte) * 8L));
        
    }
    
    /**
     * return the long value
     *
     * @param offset
     * @return	long
     *
     *  @since JXTA 1.0
     */
    
    protected long bytesIntoLong( int offset ) {
        long result = 0L;
        
        for( int eachByte = 0; eachByte < 8; eachByte ++ )
            result |= ((long)(bytes[eachByte + offset] & 0xff)) << ((7 - eachByte) * 8L);
        
        return result;
    }
}
