/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: IDFactory.java,v 1.9 2001/06/08 23:03:23 bondolo Exp $
 */

package net.jxta.id;

import java.lang.reflect.Constructor;
import java.lang.String;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;

import java.lang.NoSuchMethodException;
import java.lang.reflect.InvocationTargetException;
import java.lang.SecurityException;
import java.net.MalformedURLException;
import java.net.UnknownServiceException;
import java.util.NoSuchElementException;

import net.jxta.util.ClassFactory;

/**
 * This class is a factory class for constructing sub-classes of ID from external
 * representations such as strings or URIs
 *
 * @see net.jxta.id.ID
 * @see net.jxta.impl.id.UUID
 * @see net.jxta.impl.id.UUIDFactory
 *
 * @version     $Revision: 1.9 $
 * @since       JXTA 1.0
 */

public class IDFactory extends ClassFactory {
    
    /**
     *  This class is in fact a singleton. This is the instance that backs the
     *  static methods.
     **/

    private static IDFactory factory = new IDFactory();
    
    /**
     *  Constructors registered with this factory need to have the following
     *  parameter sets.
     *
     *  For this factory the constructors are:
     *      constructor( void )
     **/

    final static private Class [] [] requiredConstructors = { {} };
    
    /**
     *  This is the index in the constructor parameter set for the constructor
     *  used when constructing the ID sub-class from a URI
     **/

    final static private int fromURIConstructor = 0;
    
    /**
     *  The required set of static methods needed by classes registering with this
     *  factory.
     *
     *  For this factory the static methods are:
     **/
    final static private Object [] [] requiredStaticMethods = {
    };

    /**
     *  This is the map of id value to the constructor rets.
     **/

    private static Hashtable IdTypes = new Hashtable();
    
    /**
     *  If true then the pre-defined set of ID sub-classes has been registered
     *  from the string containing them.
     **/

    private boolean loadedString = false;
    
    /**
     *  Registers the pre-defined set of ID sub-classes so that this factory can
     *  construct them.
     *
     *  @return booelan true if at least one of the ID sub-classes could be
     *  registered otherwise false.
     *
     *  @since JXTA 1.0 
     */

    private boolean doRegisterIDTypes()  {
        final String idInstances = "net.jxta.codat.CodatID net.jxta.peergroup.PeerGroupID net.jxta.peer.PeerID net.jxta.pipe.PipeID";
        return factory.registerFromString( idInstances );
    };
    
    /**
     *  Used by ClassFactory methods to get the mapping of ID types to constructors.
     *
     *  @return Hashtable the hashtable containing the mappings.
     *
     *  @since JXTA 1.0 
     */

    protected Hashtable getAssocTable() {
        return IdTypes;
    }
    
    /**
     *  Used by ClassFactory methods to ensure that all keys used with the mapping are
     *  of the correct type.
     *
     *  @return Class object of the key type.
     *
     *  @since JXTA 1.0 
     */

    protected Class getClassForKey() {
        return Integer.class;
    }
    
    /**
     *  Used by ClassFactory methods to ensure that all of the instance classes 
     *  which register with this factory have the correct base class
     *
     *  @return Class object of the key type.
     *
     *  @since JXTA 1.0 
     */

    protected Class getBaseClassForFactory() {
        // our key is the doctype names.
        return ID.class;
    }

    /**
     *  Used by ClassFactory methods to determine the required set of constructors
     *  needed by the instance classes generated by this factory.
     *
     *  @return Class object of the key type.
     *
     *  @since JXTA 1.0
     */
    protected Class [] []  getRequiredConstructors() {
        return requiredConstructors;
    }
    
    /**
     *  Used by ClassFactory methods to determine the required set of static
     *  methods needed by classes registering with this factory.
     *
     *  @return Object
     *
     *  @since JXTA 1.0
     */
    protected Object [] [] getRequiredStaticMethods() {
        return requiredStaticMethods;
    }

    /**
     * Construct a CodatId from a jxta URI.
     *
     * @param source  URI which will be decoded to initialize this Codat ID.
     * @exception UnknownServiceException thrown if the URI provided is not
     * a jxta URI.
     * @exception MalformedURLException bad URL
     * @exception InvocationTargetException error during invocation
     *
     *  @since JXTA 1.0 
     */
    public static ID fromURL( URL source ) throws
      MalformedURLException, UnknownServiceException, InvocationTargetException {

        ID result = null;
        final String digits = "0123456789ABCDEF";
        byte [] decoded = new byte [ID.IdByteArraySize];
        
        if ( !(ID.URIEncodingName.equals(source.getProtocol())) )
            throw new UnknownServiceException( "URI protocol type was not as expected." );
        
        String  encoded = source.getHost();
        for( int eachChar = encoded.length() - 1; eachChar >= 0; eachChar-- ) {
            int digit = digits.indexOf(encoded.charAt(eachChar));
            if ( -1 != digit )
                decoded[eachChar / 2] += ((16 - (eachChar % 2) * 15) * digit);
            else
                throw new MalformedURLException( "Invalid Character in jxta URI" );
        }
        
        if( ! factory.loadedString )
            factory.loadedString = factory.doRegisterIDTypes();
        
        if( ID.UUIDEncoded != decoded[ID.flagsOffset + ID.flagsEncodingOffset] )
            throw new MalformedURLException( "Unsupported JXTA ID encoding" );

        if( ID.flagNullID == decoded[ID.flagsOffset + ID.flagsIdTypeOffset] )
            return ID.nullID;
        
        // Find out which ID sublcass is represented by this ID and construct it.
        try {
            result = (ID) factory.newInstance( new Integer( decoded[ID.flagsOffset +
		      ID.flagsIdTypeOffset] ), requiredConstructors[fromURIConstructor],
		      new Object [0] );
        }
        catch( NoSuchElementException notFound ) {
            throw new MalformedURLException( "jxta ID type not recognized" );
        }
        
        // This step wipes whatever the constructor did, but thats ok.
        for( int eachByte = 0; eachByte < result.bytes.length; eachByte++ )
            result.bytes[eachByte] = decoded[eachByte];
        
        return result;
        
    };
    
    /**
     *    Register a constructor for an ID type to enable IDs of that type to be
     *    constructed.
     *
     *    @param  type    the identifying value for this id type
     *    @param  itsConstructor  the constructor to use in constructing objects
     *                            of this ID type. It should be the parameterless
     *                            constructor.
     *    @returns boolean true   if the ID type is registered. If there is already a
     *                    constructor for this type then false will be returned.
     *
     *    @exception NoSuchMethodException invalid method
     *    @exception SecurityException security violation
     *
     *    @since JXTA 1.0 
     */
    public static boolean registerIDType( int type, Class instanceClass ) 
      throws NoSuchMethodException, SecurityException {
      boolean result = factory.registerAssoc( new Integer( type ), instanceClass );
      
      return result;
    }
}
