/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerGroup.java,v 1.19 2001/06/19 00:34:28 jice Exp $
 */

package net.jxta.peergroup;

import net.jxta.peer.PeerID;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.protocol.PeerGroupAdvertisement;
import net.jxta.service.Service;
import net.jxta.platform.Application;

import net.jxta.resolver.Resolver;
import net.jxta.discovery.Discovery;
import net.jxta.pipe.Pipe;
import net.jxta.membership.Membership;
import net.jxta.rendezvous.RendezVous;
import net.jxta.peer.PeerInfo;

/**
 * This interface implements the peer group interface. Peer groups are formed
 * as a collection of peers that have agreed upon a common set of services.
 * Each peer group is assigned a unique peer group ID and a peer group
 * advertisement.
 * The peer group advertisement defines each of the group services (membership,
 * discovery, resolver, etc).
 *
 * <p>In order to participate in a group, a peer needs to join the group
 * using the peer group membership service.
 *
 * <p> JXTA comes with two subclasses of peer groups:
 *
 * <ul type-disc> 
 *
 * <li> <strong>Platform</strong>: this is used to represent the <em>World</em>
 * group. Every peer, when booting, becomes part of this group. This group
 * provides the minimum core services needed to find and instanciate other
 * groups on a peer. <strong>Platform</strong> has the privilege of assigning
 * a new ID to the peer, if it does not already have one. The <em>World</em>
 * group's ID is invariant.</li>
 *
 * <li> <strong>StdPeergroup</strong>: this is currently used to implement
 * all other kinds of peer groups.
 * The first such group that it is instanciated after booting is known as
 * <em>The Net Peer Group</em>. When <strong>Platform</strong> starts it may
 * optionaly search for <em>The Net Peer Group</em> on the local network and,
 * if found, instanciate it. Otherwise a default configuration of
 * <em>The Net Peer Group</em> is instanciated.
 *
 * A non-default configuration of <em>The Net Peer Group</em> may be set-up
 * by the administrator in charge of the network domain inside which the peer
 * is booting. <em>The Net Peer Group</em> is discovered via the Discovery
 * protocol. Many such groups may be configured by an administrator.<br>
 *
 * <strong>StdPeergroup</strong> may also be used to implement User-defined
 * peer groups: Users can create new peer groups which use their own set of
 * customized services.
 *</li>
 *</ul>
 *
 * @see         net.jxta.peergroup.PeerGroupID
 * @see         net.jxta.service.Service
 * @see         net.jxta.peergroup.PeerGroupFactory
 * @see         net.jxta.protocol.PeerGroupAdvertisement
 *
 * @version $Revision: 1.19 $
 * @since JXTA 1.0 */


public interface PeerGroup extends Application {

  /*
   * shortcuts to the well-known services, in order to avoid calls to lookup.
   */

    /**
     * @return Resolver an object implementing the Resolver service for this
     * group.
     */
    public Resolver getResolver();

    /**
     * @return Discovery an object implementing the Discovery service for this
     * group.
     */
    public Discovery getDiscovery();

    /**
     * @return PeerInfo an object implementing the PeerInfo service for this
     * group.
     */
    public PeerInfo getPeerInfo();

    /**
     * @return Membership an object implementing the Membership service for
     * this group.
     */
    public Membership getMembership();

    /**
     * @return Pipe an object implementing the Pipe service for this
     * group.
     */
    public Pipe getPipe();

    /**
     * @return RendezVous an object implementing the RendezVous service for this
     * group.
     */
    public RendezVous getRendezVous();

  /**
   * Check debug status
   * @return boolean true if debug is on.
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0
   */

  public boolean getDebug();
  
  /**
   * Returns the whether the group member is a Rendezvous peer for the group
   *
   * @return boolean true if the peer is a rendezvous for the group.
   * 
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0
   */

  public boolean isRendezvous();
  
  /**
   * Ask a group its group id.
   *
   * @return PeerGroupId this Group's ID
   *
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0
   */

  public PeerGroupID getID();

  /**
   * Ask a group member for its peer Id.
   * @return PeerId this peer's ID
   *
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0
   */

  public PeerID getPeerID();

  /**
   * Ask a group member for its peer Name.
   * @return String this peer's name
   *
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0
   */

  public String getPeerName();

  /**
   * Ask a group for its group advertisement
   *
   * @return PeerGroupAdvertisement this Group's advertisement
   *
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0
   */

  public PeerGroupAdvertisement getAdvertisement();
  
  /**
   * Lookup for a service by name.
   *
   * @param name the service name
   *
   * @return Service, the Service registered by that name
   *
   * @exception ServiceNotFoundException could not find the service requested
   *
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0    
   */

  public Service lookupService(String name)
    throws ServiceNotFoundException;


  /**
   * Return the default application setup for this peergroup
   * The application can be started via its init() method, or by calling
   * this group's startApp() method.
   *
   * @return Application The application object
   *
   * @version $Revision: 1.19 $
   * @since JUXTA 1.0
   */
  
    public Application getApp();
}
