/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerGroupID.java,v 1.12 2001/06/18 21:17:56 jice Exp $
 */
package net.jxta.peergroup;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.security.MessageDigest;

import net.jxta.id.ID;
import net.jxta.id.IDFactory;

import net.jxta.impl.id.UUID;
import net.jxta.impl.id.UUIDFactory;

/**
 * This class implements a PeerGroup ID. Each peer group is assigned a
 * unique peer id.UUID id are used to implement peer group id.
 *
 * @see         net.jxta.id.ID
 * @see         net.jxta.id.IDFactory
 * @see         net.jxta.id.UUID
 * @see         net.jxta.id.UUIDFactory
 * @see         net.jxta.peer.PeerID
 *
 * @version     $Revision: 1.12 $
 * @since JXTA 1.0
 */

public class PeerGroupID extends ID implements Cloneable {
    
    final static private int uuidSize = 16;
    
    final static private int groupIdOffset = 0;
    final static private int padOffset = PeerGroupID.groupIdOffset +
    PeerGroupID.uuidSize;
    
    final static private int padSize = ID.flagsOffset - PeerGroupID.padOffset;
    
    /**
     *  If true then this class has been registered as the class created when
     *  JXTA URIs
     */
    
    private static boolean registered = false;
    
    static {
        try {
            registered = IDFactory.registerIDType(
            ID.flagPeerGroupID, PeerGroupID.class );
        }
        catch( Exception e ) {
            registered = false;
        }
    };

    /**
     * The well known Unique Identifier of the world peergroup.
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    final static public PeerGroupID worldPeerGroupID = new PeerGroupID(
    new UUID( 0x5961626164616261L, 0x4A78746150325033L ) ); //Yabadaba JXTA P2P!
    
    /**
     * Constructor for creating a new PeerGroupID with a unique ID.
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public PeerGroupID( ) {
        this(UUIDFactory.newUUID());
    }
    
    /**
     * Constructor.
     * Creates a PeerGroupID. A PeerGroupID is provided
     *
     * @param groupUUID    the PeerGroupID to use to construct the new PeerGroupID
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public PeerGroupID (UUID groupUUID) {
        super( );
        bytes[ID.flagsOffset + ID.flagsIdTypeOffset] = ID.flagPeerGroupID;
        
        longIntoBytes( PeerGroupID.groupIdOffset,
        groupUUID.getMostSignificantBits() );
        longIntoBytes( PeerGroupID.groupIdOffset + 8,
        groupUUID.getLeastSignificantBits() );
    };
    
    /**
     * Returns the PeerGroupID of the PeerGroup associated with this peer group
     *
     * @return PeerGroupID return the peer group Id
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public PeerGroupID getPeerGroupID() {
        return this; // duh!
    };
    
    /**
     * Returns true if the two PeerGroup IDs are the same
     *
     * @param id Peer group id to compare with
     * @return boolean true if equals
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public boolean equals ( Object id ) {
        if (this == id)
            return true;
        
        if ( id instanceof PeerGroupID )
            return super.equals( id );
        else
            return false;
    };
    
    /**
     * Clones this Peer Group ID
     *
     * @return Object return a clone object of the peer group Id
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public Object clone () {
        PeerGroupID result = new PeerGroupID();
        result.bytes = (byte [] ) this.bytes.clone();
        return result;
    };
    
    /**
     * get the peer group unique id
     *
     * @return Object peer group unique id
     *
     * @version     $Revision: 1.12 $
     * @since JXTA 1.0
     */
    
    public Object getGroupUniqueId( ) {
        UUID result = new UUID( bytesIntoLong( PeerGroupID.groupIdOffset ),
        bytesIntoLong( PeerGroupID.groupIdOffset + 8) );
        
        return result;
    };
}
