/************************************************************************
 *
 * $Id: OutputPipe.java,v 1.7 2001/05/23 00:28:22 lomax Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.pipe;

import java.io.IOException;

import net.jxta.endpoint.Message;

/** 
 * OuputPipe defines the interface for sending messages from a
 * <b>PipeService</b>. Pipes are the core mechanism for exchanging messages
 * between two JXTA applications or services. Pipes provides a simple,
 * uni-directional and asynchronous channel of communication between two
 * peers. JXTA messages are exchanged between pipes (Input and Output
 * pipes). An application to open a receiving communication with other
 * peers will create an input pipe and bind it to a specific pipe
 * advertisement. The application will then publish the pipe advertisement
 * so other applications or services can obtain the advertisment and create
 * corresponding output pipes to send messages to that input pipe.
 *
 * <p>Pipes are uniquely identified throughout the JXTA world by a PipeId (UUID)
 * enclosed in a <b>PipeAdvertisement</b>.  The pipe advertisement has a
 * unique id that is used to create the association between Input and
 * Output pipes.
 *
 * <p>Pipes are <code>non-localized</code> communication channels not
 * binded to specific peers. This is a unique feature of JXTA pipes. The
 * mechanism to resolve the location of pipes to a physical peer is done in
 * a complete decentralized manner in JXTA via the JXTA <b>PipeService
 * ResolverService</b> protocol. The PipeService resolver protocol does not rely on a
 * centralized protocol such as DNS (bind Hostname to IP) to bind a pipe
 * advertisement (i.e.symbolic name) to an instance of a pipe on a physical
 * peer (i.e IP address). The resolver protocol uses a dynamic and adaptive
 * search mechanism that attempts at all time to find the closest peer
 * where an instance of that pipes is running leverage JXTA core DiscoveryService
 * mechanism.
 *
 * <p>The binding of output pipes to an input pipes is done dynamically at
 * runtime via the pipe resolver protocol. The pipe resolver mechanism
 * searches for a pipe advertisement and an instance of the pipe (Input or
 * output) running on the closest available peer.
 *
 * <p>The non-localization of pipes and the pipe resolver protocol make
 * JXTA pipes unique by allowing applications and services to find
 * themselves and open communication channels via the resolution of a pipe
 * advertisement. Futhermore, the pipe resolution protocol enables to have
 * pipe channels to adapt to their environment in a transparent manner, by
 * allowing applications and services to find the "closest" instance of a
 * pipe instance in their horizon of discovery.
 *
 * <p>The pipe resolver mechanism is powerful to create JXTA
 * high-availability services by allowing pipes to be dynamically connected
 * on different peers in a fully decentralized and transparent manner.
 *
 * @see net.jxta.pipe.InputPipe
 * @see net.jxta.endpoint.Message
 * @see net.jxta.protocol.PipeAdvertisement
 * @see net.jxta.pipe.Pipe
 * @see net.jxta.resolver.GenericResolver
 *
 * @version $Revision: 1.7 $
 * @since JXTA 1.0 */

public interface OutputPipe {

  /**
   * send a message throught the pipe
   *
   * @param msg is the PipeMessage to be sent.
   *
   * @exception IOException output pipe error
   *
   * @version $Revision: 1.7 $
   * @since JXTA 1.0
   */

  public void send (Message msg) throws IOException;

  /**
   * close the pipe
   *
   * @version $Revision: 1.7 $
   * @since JXTA 1.0
   */

  public void close();
  
}



