/************************************************************************
 *
 * $Id: Pipe.java,v 1.5 2001/06/07 21:54:38 raffi Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.pipe;

import java.util.Enumeration;
import java.io.IOException;

import net.jxta.service.Service;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.endpoint.Message;

/**
 * A core service of Jxta Group is the PipeService service for
 * creating and accessing pipes within a peer group.  The Pipe defines a
 * set of interfaces to create and access pipes within a peer group.
 *
 *<p>The Jxta Pipe defines an API and a Protocol.
 *
 *<p>Pipes are the core mechanism for exchanging messages
 * between two JXTA applications or services. Pipes provides a simple,
 * uni-directional and asynchronous channel of communication between two
 * peers. JXTA messages are exchanged between pipes (Input and Output
 * pipes). An application to open a receiving communication with other
 * peers will create an input pipe and bind it to a specific pipe
 * advertisement. The application will then publish the pipe advertisement
 * so other applications or services can obtain the advertisment and create
 * corresponding output pipes to send messages to that input pipe.
 *
 * <p>Pipes are uniquely identified throughout the JXTA world by a PipeId (UUID)
 * enclosed in a <b>PipeAdvertisement</b>.  The pipe advertisement has a
 * unique id that is used to create the association between Input and
 * Output pipes.
 *
 * <p>Pipes are <code>non-localized</code> communication channels not
 * binded to specific peers. This is a unique feature of JXTA pipes. The
 * mechanism to resolve the location of pipes to a physical peer is done in
 * a complete decentralized manner in JXTA via the JXTA <b>PipeService
 * ResolverService</b> protocol. The PipeService resolver protocol does not rely on a
 * centralized protocol such as DNS (bind Hostname to IP) to bind a pipe
 * advertisement (i.e.symbolic name) to an instance of a pipe on a physical
 * peer (i.e IP address). The resolver protocol uses a dynamic and adaptive
 * search mechanism that attempts at all time to find the closest peer
 * where an instance of that pipes is running leverage JXTA core DiscoveryService
 * mechanism.
 * 
 * <p>Pipes are named with their XML advertisement. The generation of
 * the PipeService Advertisements is let to the applications. A Jxta PipeService
 * advertisement is an XML document (non validated) which contains at least
 * the following structure (but can be extended).
 *
 * <?xml version="1.0" standalone='yes'?>
 * <JxtaPipe>
 *   <id> UUID </id>
 *   <name> optional symbolic name than can be used by any search engine </name>
 * <JxtaPipe>
 *
 *<p>Note that other tags can be provided by the application for its own
 * profit. Those extra tags are not interpreted by the PipeService Service.
 *
 * <p>The UUID needs to be a unique String (unique in time and space).
 * The optional tool package net.jxtax.pipe provides helper classes to
 *
 * @see net.jxta.pipe.InputPipe
 * @see net.jxta.endpoint.Message
 * @see net.jxta.protocol.PipeAdvertisement
 * @see net.jxta.pipe.Pipe
 * @see net.jxta.resolver.GenericResolver
 *
 * @version $Revision: 1.5 $
 * @since JXTA 1.0
 */

public interface Pipe extends Service {

  public final static int NonBlocking = 1;
  public final static int HandChecking = 2;
  public final static int Propagate = 3;

  /**
   * create an InputPipe from a pipe Advertisement
   *
   * @param adv is the advertisement of the PipeService.
   * @return InputPipe InputPipe object created
   *
   * @exception IOException error creating input pipe
   *
   * @version $Revision: 1.5 $
   * @since JXTA 1.0
   */

  public InputPipe createInputPipe (PipeAdvertisement adv)
    throws IOException;

  /**
   * create an OutputPipe from the pipe Advertisement
   *
   * @param adv is the advertisement of the PipeService.
   * @param type is an int and contains the name of the diffusion mode for
   * this OutputPipe (NonBlocking, HandChecking, Propagate).
   * @param timeout time createOutputPipe will block: -1 means indefinitely,
   *  otherwise, timeout in millisecond
   * @return OuputPipe Output pip object
   *
   * @exception IOException error creating output pipe
   *
   * @version $Revision: 1.5 $
   * @since JXTA 1.0
   */

  public OutputPipe createOutputPipe (PipeAdvertisement adv, int type, long timeout)
    throws IOException;

  /**
   * create an OutputPipe from the pipe Advertisement giving a PeerId(s) where
   * the corresponding InputPipe is supposed to be.
   *
   * @param adv is the advertisement of the NetPipe.
   * @param type is an int and contains the name of the diffusion mode for
   * this OutputPipe.
   * @param peers is an enumeration of the PeerId of thepeers where to look for the
   * corresponding Pipes
   * @param timeout time createOutputPipe will block: -1 means 
   *        indefinitely, otherwise, timeout in millisecond
   * @return OuputPipe
   * @exception IOException if none of the peers in the enumeration has 
   *                     the corresponding OutputPipe
   *
   * @version $Revision: 1.5 $
   * @since JXTA 1.0
   */

  public OutputPipe createOutputPipe (PipeAdvertisement adv, int type,
                                      Enumeration peers, long timeout)
    throws IOException;

  /**
   * Creates a new Message
   *
   * @return Message returns a newly allocated PipeService Message.
   *
   * @version $Revision: 1.5 $
   * @since JXTA 1.0
   */

  public Message createMessage();

  /**
   * Teach the PipeService Polcicy about the localisation of a PipeService. This
   * allows an application to help the PipeService ResolverService. Note that once
   * a set of EndpointAddresses has been set for a pipe, it cannot be
   * removed, because the PipeResolver may have already propagated the
   * addresses. This function should be used very cautiously.
   *
   * XXX: maybe some sanity check should be added here, or special
   * credential. To be revisited.
   *
   * @param adv PipeService Advertisement of the pipe
   * @param addrs Enumeration of EndpointAddress of the pipe
   *
   * @version $Revision: 1.5 $
   * @since JXTA 1.0
   */

  public void addAddresses (PipeAdvertisement adv, Enumeration addrs);

}



