/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PipeID.java,v 1.2 2001/05/23 00:28:22 lomax Exp $
 */

package net.jxta.pipe;

import java.lang.reflect.Constructor;

import net.jxta.id.ID;
import net.jxta.id.IDFactory;
import net.jxta.peergroup.PeerGroupID;

import net.jxta.impl.id.UUID;
import net.jxta.impl.id.UUIDFactory;

/**
 * This class implements a Pipe ID. Each pipe is assigned a unique id
 *
 * @see         net.jxta.id.ID
 * @see         net.jxta.id.IDFactory
 * @see         net.jxta.peergroup.PeerGroupID
 *
 * @version     $Revision: 1.2 $
 * @since       JXTA  1.0
 */

public class PipeID extends ID implements Cloneable {

  final static private int uuidSize = 16;
  
  final static private int groupIdOffset = 0;
  final static private int idOffset = PipeID.groupIdOffset + PipeID.uuidSize;
  final static private int padOffset = PipeID.idOffset + PipeID.uuidSize;
  
  final static private int padSize = ID.flagsOffset - PipeID.padOffset;

  /**
   *  If true then this class has been registered as the class created when
   *  JXTA URIs
   */

  private static boolean registered = false;
    
  static {
    try {
      registered = IDFactory.registerIDType( ID.flagPipeID, PipeID.class );
    }
    catch( Exception e ) {
      registered = false;
    }
  };

  /**
   *  Constructor used in the creation of PipeIDs from URIs. Other than that
   *  there is no reason to use this constructor.
   *
   * @version     $Revision: 1.2 $
   * @since       JXTA  1.0
   */
  
  public PipeID() {
    super();
    bytes[ID.flagsOffset + ID.flagsIdTypeOffset] = ID.flagPipeID;
    };
  
  /**
   * Constructor.
   * Creates a PipeID. A PeerGroupID is provided
   *
   * @param groupID    the group to which this will belong.
   *
   * @version     $Revision: 1.2 $
   * @since       JXTA  1.0
   */

    public PipeID(PeerGroupID groupID) {
    this( );

    UUID    groupUUID = (UUID) groupID.getGroupUniqueId();
    UUID    id = UUIDFactory.newUUID();

    longIntoBytes( PipeID.groupIdOffset, groupUUID.getMostSignificantBits() );
    longIntoBytes( PipeID.groupIdOffset + 8, groupUUID.getLeastSignificantBits() );
    
    longIntoBytes( PipeID.idOffset, id.getMostSignificantBits() );
    longIntoBytes( PipeID.idOffset + 8, id.getLeastSignificantBits() );
  };

  /**
   * Returns the PeerGroupID of the PeerGroup associated with this Peer
   *
   * @return PeerGroupID return peer group Id
   *
   * @version     $Revision: 1.2 $
   * @since       JXTA  1.0
   */

  public PeerGroupID getPeerGroupID() {
    UUID result = new UUID( bytesIntoLong( PipeID.groupIdOffset ),
			    bytesIntoLong( PipeID.groupIdOffset + 8 ) );
        
    return new PeerGroupID( result );
  };

  /**
   * Returns true if the two PipeIDs are same
   *
   * @param ip PipeID to compare
   * @return boolean true if equals
   *
   * @version     $Revision: 1.2 $
   * @since       JXTA  1.0
   */
    public boolean equals ( Object id ) {
  	if (this == id)
	    return true;

	if (id instanceof PipeID)
        return super.equals( id );
    else 
        return false;
    };

  /**
   * Clones this PipeID
   * 
   * @return Object Cloned peer id.
   *
   * @version     $Revision: 1.2 $
   * @since       JXTA  1.0
   */

  public Object clone () {
    PipeID result = new PipeID();
    result.bytes = (byte [] ) this.bytes.clone();
    return result;
  };
}
