/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: DiscoveryQueryMsg.java,v 1.11 2001/05/23 00:28:22 lomax Exp $
 */

package net.jxta.protocol;

import net.jxta.document.Document;
import net.jxta.document.MimeMediaType;

/**
 * This class defines the DiscoveryService message "Query"
 * The default behavior of this abstract class is simply  a
 * place holder for the generic resolver query fields.
 *
 * This message is part of the Peer DiscoveryService Protocol.
 *
 *
 * <pre>
 *      &lt;?xml version="1.0" standalone='yes'?&gt;
 *       &lt;DiscoveryQuery&gt;
 *        &lt;type&gt;int&lt;/type&gt;
 *        &lt;threshold&gt;int&lt;/threshold&gt;
 *        &lt;peeradv&gt;peeradv&lt;/peeradv&gt;
 *        &lt;attribute&gt;attribute&lt;/attribute&gt;
 *        &lt;value&gt;attribute&lt;/value&gt;
 *       &lt;/DiscoveryQuery&gt;
 * </pre>
 *
 *
 * @see net.jxta.protocol.DiscoveryResponseMsg
 * @version $Revision: 1.11 $
 * @since JXTA 1.0
 *
 */


public abstract class DiscoveryQueryMsg {

  /**
   * default threshold to limit the number of responses
   * from one peer
   */

    public static int threshold = 50;

    public int type=0;
    private String peeradv=null;
    private String attr=null;
    private String value=null;


  /**
   * All messages have a type (in xml this is !doctype)
   * which identifies the message
   * @return String "jxta:ResolverResponse"
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public static String getAdvertisementType() {
        return "jxta:DiscoveryQuery" ;
    }

  /**
   * Get the Threshold for number of responses
   * @return int threshold
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */
    public int getThreshold() {
      return threshold;
    }

  /**
   * Get the response type
   * @return int  PEER, or GROUP discovery type response
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public int getDiscoveryType() {
      return type;
    }

  /**
   * returns the responding's peer advertisement
   * @return String handlername name
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public String getPeerAdv() {
        return peeradv;
    }


  /**
   * returns the Attr value
   * @return String value of Attribute
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public String getAttr() {
      return attr;
    }

  /**
   * returns the value of Attr
   * @return String
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public String getValue() {
      return value;
    }

  /**
   * set the threshold
   * @param threshold value to be set
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public void setThreshold(int threshold) {
      this.threshold=threshold;
    }

  /**
   * set the Response type whether it's peer, or group dsicovery
   * @param type type of discovery
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public void setDiscoveryType(int type) {
      this.type=type;
    }

  /**
   * set the Response type whether it's peer, or group dsicovery
   * @param peer
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public void setPeerAdv(String peer) {
      this.peeradv=peer;
    }

  /**
   * set the attr
   * @param attr attribut of the query
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public void setAttr(String attr) {
      this.attr=attr;
    }

  /**
   * set the query
   * @param value value of the attribute to query
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public void setValue(String value) {
      this.value=value;
    }

  /**
   * Write advertisement into a document. asMimeType is a mime media-type
   * specification and provides the form of the document which is being
   * requested. Two standard document forms are defined. "text/text" encodes
   * the document in a form nice for printing out, and "text/xml" which
   * provides an XML representation.
   *
   * @param asMimeType mime-type format requested
   * @return Document representaion of the document as an advertisement
   *
   * @exception Exception error generating the document
   *
   * @version $Revision: 1.11 $
   * @since JXTA 1.0
   */

    public abstract Document getDocument ( MimeMediaType asMimeType )
      throws Exception;

}



