/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: EndpointAdvertisement.java,v 1.10 2001/05/23 00:28:22 lomax Exp $
 */

package net.jxta.protocol;

import java.util.*;

import net.jxta.document.Element;
import net.jxta.document.Advertisement;

/**
 * This abstract class defines an Endpoint advertisement. Enpoint
 * advertisements are used to represent each peer Platform endpoints. A
 * peer can have many available endpoints. One endpoints for each
 * configured network interfaces or protocol (IP, HTTP).  The endpoint
 * advertisement is a tag field in the peer group advertisement (see
 * PeerGroupAdvertisement) to describe the endpoints available on the
 * member peer of the peer group.
 *
 * <pre>
 * &lt;?xml version="1.0"?&gt;
 *  &lt;EndpointAdvertisement&gt;
 *   &lt;Name&gt;string&lt;/Name&gt;
 *   &lt;Keywords&gt;string&lt;/Keywords&gt;
 *   &lt;Address&gt;endpoint address&lt;/Address&gt;
 *   &lt;TransportAdvertisement&gt;transport&lt;/TransportAdvertisement&gt;
 *  &lt;/EndpointAdvertisement&gt;
 *</pre>
 *
 * @see net.jxta.protocol.PeerAdvertisement
 * @see net.jxta.protocol.PeerGroupAdvertisement
 * @see net.jxta.protocol.TransportAdvertisement
 *
 * @version $Revision: 1.10 $
 * @since JXTA 1.0 */


public abstract class EndpointAdvertisement extends Advertisement {
  
  private String name = null;
  private String keywords = null;
  private TransportAdvertisement transport = null;
  private String endpointaddress = null;

  private boolean enabled = true; // enabled is the default.
    

  /** 
   * returns the advertisement type
   * @return string type
   * 
   * @version $Revision: 1.10 $
   * @since JXTA 1.0     
   */
  
  public static String getAdvertisementType() {
    return "jxta:EndpointAdvertisement" ;
  }
  
  /** 
   * returns the name of the endpoint
   * @return string name of the endpoint
   * 
   * @version $Revision: 1.10 $
   * @since JXTA 1.0     
   */

  public String getName() {
    return name;
  }
  
  /** 
   * sets the name of the endpoint
   * @param  name of the endpoint
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0     
   */

  public void setName(String name) {
    this.name = name;
  }
  
  /** 
   * returns the keywords associated with the endpoint
   * @return String keywords
   * 
   * @version $Revision: 1.10 $
   * @since JXTA 1.0     
   * 
   */
  
  public String getKeywords() {
    return keywords;
  }

  /** 
   * sets the keywords associated with the endpoint
   * @param keywords associated with endpoint
   * 
   * @version $Revision: 1.10 $
   * @since JXTA 1.0     
   */

  public void setKeywords(String keywords) {
    this.keywords = keywords;
  }

  /** 
   * returns the transport advertisement associated with the endpoint
   * @return TransportAdvertisement  for the endpoint
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */
  
  public TransportAdvertisement getTransportAdvertisement() {
    return transport;
  }
 
  /** 
   * sets the transport advertisement
   * @param transport Transport Advertisement
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */
  
  public void setTransportAdvertisement(TransportAdvertisement transport) {
    this.transport = transport;
  }

  /** 
   * returns the endpointaddress
   * @return String return the endpoint address
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */
  
  public String getEndpointAddress() {
    return endpointaddress;
  }
 
  /** 
   * set the endpointaddress
   * @param add  endpoint address
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */

  public void setEndpointAddress(String add) {
    endpointaddress = add;
  }

  /** 
   * enable or disable this specific endpoint.
   * keeping a disabled endpoint is usefull to remember its configuration.
   * @param enabled whether it is enabled or disabled.
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */
  public void setEnabled(boolean enabled) {
      this.enabled = enabled;
  }

  /** 
   * Tells whether specific endpoint is enabled or not.
   * @return boolean whether it is enabled or disabled.
   *
   * @version $Revision: 1.10 $
   * @since JXTA 1.0
   */
  public boolean getEnabled() {
      return enabled;
  }
}
