/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerAdvertisement.java,v 1.8 2001/05/23 00:28:22 lomax Exp $
 */

package net.jxta.protocol;

import java.util.Hashtable;

import net.jxta.document.Element;
import net.jxta.document.Advertisement;

/**
 * This abstract class define a Peer advertisement. A peer Advertisement is
 * an instance of the Platform PeerGroup Advertisement. A peer is by
 * default a member of the Platform or World Peer Group.
 *
 * <pre>
 * &lt;?xml version="1.0"?&gt;
 * &lt;PeerAdvertisement&gt;
 *   &lt;Name&gt;string&lt;/Name&gt;
 *   &lt;Keywords&gt;string&lt;/Keywords&gt;
 *   &lt;Pid&gt; Peer Id &lt;/Pid&gt;
 *   &lt;Service&gt; service advertisement&lt;/Service&gt;
 *           ...........
 *   &lt;Service&gt; service advertisement&lt;/Service&gt;
 *   &lt;Endpoint&gt; endpoint Advertisement &lt;/Endpoint&gt;
 *           ...........
 *   &lt;Endpoint&gt; endpoint Advertisement &lt;/Endpoint&gt;
 *   &lt;InitialApp&gt; initial application  &lt;/InitialApp&gt;
 *   &lt;GroupImpl&gt;  group Implementation &lt;/GroupImpl&gt;
 * &lt;/PeerAdvertisement&gt;
 *</pre>
 *
 *
 * @see net.jxta.protocol.PeerGroupAdvertisement
 * @see net.jxta.protocol.ServiceAdvertisement
 * @see net.jxta.protocol.EndpointAdvertisement
 * @see net.jxta.protocol.ServiceAdvertisement
 * @see net.jxta.protocol.TransportAdvertisement
 *
 * @version $Revision: 1.8 $
 * @since JXTA 1.0
 */


public abstract class PeerAdvertisement extends PeerGroupAdvertisement {
  

    private String pid = null;
    private String name = null;
    private String keywords = null;
    private Hashtable services = new Hashtable(); 
    private Hashtable endpoints = new Hashtable();

    /**
     * Default constructor for Peer advertisement
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */

    public PeerAdvertisement() {
    }

    /** 
     * returns the advertisement type
     * @return String type of advertisement
     * 
     * @version $Revision: 1.8 $
     * @since JXTA 1.0     
     */
  
    public static String getAdvertisementType() {
        return "jxta:PeerAdvertisement" ;
    }
  
    /** 
     * returns the name of the Peer
     * @return String name of the Peer
     * 
     * @version $Revision: 1.8 $
     * @since JXTA 1.0     
     */

    public String getName() {
        return name;
    }
  
    /** 
     * sets the name of the peer
     * @param name name of the peer
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */

    public void setName(String name) {
        this.name = name;
    }

    /** 
     * returns the keywords associated with the peer
     * @return String keywords
     * 
     * @version $Revision: 1.8 $
     * @since JXTA 1.0     
     * 
     */
  
    public String getKeywords() {
        return keywords;
    }

    /** 
     * sets the keywords associated with the peer
     * @param keywords
     * 
     * @version $Revision: 1.8 $
     * @since JXTA 1.0     
     */

    public void setKeywords(String keywords) {
        this.keywords = keywords;
    }

    /** 
     *returns the peer id
     * @return String peer id
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */

    public String getPid() {
        return pid;
    }
 
    /** 
     * sets the peer id
     * @param pid peer id to be set
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */

    public void setPid(String pid) {
        this.pid = pid;
    }

    /** 
     * returns the services associated with the services available on the peer 
     * @return Hastable  hastable of service Advertisements associated with
     * the peer
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */

    public Hashtable getServiceAdvertisements() {
        return services;
    }
 
    /** 
     * sets the service advertisements for services registered on the peer
     * @param Hashtable hashtable of service Advertisements for each service
     * registered
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */
  
    public void setServiceAdvertisements(Hashtable services) {
        this.services = services;
    }


    /** 
     * returns the endpoints associated with the peer
     * @return Hashtable hashtable of endpoints for the peer
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */

    public Hashtable getEndpointAdvertisements() {
        return endpoints;
    }
 
    /** 
     * sets the endpoint advertisement
     * @param endpoints hashtable of endpoint advertisements
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */
  
    public void setEndpointAdvertisements(Hashtable endpoints) {
        this.endpoints = endpoints;
    }


    /** 
     * get the service implementation for a registered service 
     * @param service name
     *
     * @return String service implementation for the service
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0
     */
  
    public String getCode(String name) {
        ServiceAdvertisement servadv =
            (ServiceAdvertisement) services.get(name);
        return servadv.getCode();
    }


}

