/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerGroupAdvertisement.java,v 1.19 2001/06/19 03:30:39 jice Exp $
 */

package net.jxta.protocol;

import java.util.Hashtable;

import net.jxta.document.Element;
import net.jxta.document.Advertisement;

/**
 * This abstract class defines a PeerGroup advertisement A peer group
 * advertisement fully describes a PeerGroup, and uniquely identifies a
 * member of this peer group. The advertisement contains all the information
 * needed to recognize this peer as the member of the peer group and to
 * instanciate that group on another peer:
 *
 * <ul>
 * <li> The name of this peer.</li>
 * <li> The unique ID of this peer.</li>
 * <li> The name of this group.</li>
 * <li> The unique ID of this group.</li>
 * <li> The keywords that describe this group.</li>
 * <li> A reference to all the services (membership, discovery, resolver,
 * pipe, etc.) that define the peer group. With their parameters on this peer.</li>
 * <li> A reference to the group's implementation itself.</li>
 * <li> All the available peer endpoints that can be used by other peer group
 * members to contact this peer with their addresses and parameters on this peer.</li>
 * <li> The default startup application to run when the peer group instance
 * is initialized on a peer.</li>
 * </ul>
 * <br>
 * <pre>
 * &lt;?xml version="1.0"?&gt;
 * &lt;PeerAdvertisement&gt;
 *   &lt;Name&gt;string&lt;/Name&gt;
 *   &lt;Keywords&gt;string&lt;/Keywords&gt;
 *   &lt;Pid&gt; Peer Id &lt;/Pid&gt;
 *   &lt;Gid&gt;Peer Group Id &lt;/Gid&gt;
 *   &lt;isRendezVous&gt; boolean &lt;/isRendezVous&gt;
 *   &lt;Service&gt; service advertisement&lt;/Service&gt;
 *           ...........
 *   &lt;Service&gt; service advertisement&lt;/Service&gt;
 *   &lt;Endpoint&gt; endpoint Advertisement &lt;/Endpoint&gt;
 *           ...........
 *   &lt;Endpoint&gt; endpoint Advertisement &lt;/Endpoint&gt;
 *   &lt;InitialApp&gt; initial application  &lt;/InitialApplication&gt;
 *   &lt;GroupImpl&gt;  group Implementation &lt;/GroupImpl&gt;
 * &lt;PeerAdvertisement&gt;
 * </pre>
 *
 *
 * @see net.jxta.protocol.PeerGroupAdvertisement
 * @see net.jxta.protocol.ServiceAdvertisement
 * @see net.jxta.protocol.EndpointAdvertisement
 * @see net.jxta.protocol.TransportAdvertisement
 *
 * @version $Revision: 1.19 $
 * @since JXTA 1.0 */


public abstract class PeerGroupAdvertisement extends Advertisement {

  private String pid = null;
  private String gid = null;
  private String name = null;
  private String peerName = null;
  private boolean isRendezvous = false;
  private String keywords = null;
  private Hashtable services = new Hashtable();
  private Hashtable endpoints = new Hashtable();
  private ServiceAdvertisement app = null;
  private ServiceAdvertisement groupImpl = null;
  private boolean debugStatus = false;

  /**
   * returns the advertisement type
   * @return string type
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public static String getAdvertisementType() {
    return "jxta:PeerGroupAdvertisement" ;
  }

  /**
   * returns the name of the peer group
   * @return String name of the peer group
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public String getName() {
    return name;
  }

  /**
   * sets the name of the peer group
   * @param name name of the peer group to be set
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setName(String name) {
    this.name = name;
  }

  /**
   * returns the name of the peer
   * @return String name of the peer
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public String getPeerName() {
    return peerName;
  }

  /**
   * sets the name of the peer
   * @param peerName name of the peer to be set
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setPeerName(String peerName) {
    this.peerName = peerName;
  }

  /**
   * returns the keywords associated with this peer group
   * @return String keywords associated with the peer group
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   *
   */

  public String getKeywords() {
    return keywords;
  }

  /**
   * sets the keywords associated with this peer group
   * @param  keywords
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setKeywords(String keywords) {
    this.keywords = keywords;
  }

  /**
   * returns the peer id of the member of the peer group
   * @return String peer id
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public String getPid() {
    return pid;
  }

  /**
   * sets the peer id of the member of this peer group
   * @return pid peer id of the peer group member
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setPid(String pid) {
    this.pid = pid;
  }

  /**
   * returns the peer group id
   * @return String return the peer group id
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0   
   */

  public String getGid() {
    return gid;
  }

  /**
   * sets the peer group id
   * @param gid  group id to be set
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setGid(String gid) {
    this.gid = gid;
  }


  /**
   * check if the member of the peer group is a rendez vous member
   * @return boolean true if it is a rendezvous
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public boolean getIsRendezvous() {
    return isRendezvous;
  }

  /**
   * sets the peer group member as a rendezvous member
   * @param flag boolean value
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setIsRendezvous(boolean flag ) {
    this.isRendezvous = flag;
  }

  /**
   * returns the services associated with the peer group
   * @return Hashtable hashtable of service Advertisements for the peer group
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public Hashtable getServiceAdvertisements() {
    return services;
  }

  /**
   * sets the service advertisement
   * @param services hashtable of service advertisements
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setServiceAdvertisements(Hashtable services) {
    this.services = services;
  }

  /**
   * returns the endpoints associated with the peer group
   * @return Hashtable hashtable of endpoints for the peer group
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public Hashtable getEndpointAdvertisements() {
    return endpoints;
  }

  /**
   * sets the default application
   * @param ServiceAdvertisement default application service advertisement
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setApp(ServiceAdvertisement app) {
    this.app = app;
  }

  /**
   * sets the peer group service implementation. Code to implement the
   * peer group framework
   *
   * @param ServiceAdvertisement advertisement for the peer group framework service
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setGroupImpl(ServiceAdvertisement group) {
    this.groupImpl= group;
  }

  /**
   * returns the default application for the peergroup
   * @return ServiceAdvertisement default  application  for the peer group
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public ServiceAdvertisement getApp() {
    return app;
  }

  /**
   * returns the  peer group framework service
   * @return ServiceAdvertisement peer group framework implementation
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public ServiceAdvertisement getGroupImpl() {
    return groupImpl;
  }

  /**
   * sets the endpoint advertisement
   * @param endpoints hashtable of endpoint advertisements
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public void setEndpointAdvertisements(Hashtable endpoints) {
    this.endpoints = endpoints;
  }


  /**
   * get the service implementation for a specific peer group service
   * @param name name of the peer group service
   * @return String code implementation for the service
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public String getCode(String name) {
    ServiceAdvertisement servadv = (ServiceAdvertisement) services.get(name);
    return servadv.getCode();
  }

  /**
   * Convenience. Get the service implementatio for the peer group framework
   * @return String the service implementation.
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public String getGroupCode() {
    return groupImpl.getCode();
  }

  /**
   * Convenience method.
   * Returns the implementation for the default peer group application
   *
   * @return String implementation for the default peer group application
   *
   * @version $Revision: 1.19 $
   * @since JXTA 1.0
   */

  public String getAppCode() {
    return getApp().getCode();
  }

    /**
     * Tells whether the group should switch to debug mode.
     *
     * @return boolean The group's required debug status.
     */
  public boolean getDebug() {
    return debugStatus;
  }

    /**
     * Sets whether the group should be switched to debug mode.
     *
     * @param debug The desired setting.
     */
  public void setDebug(boolean debug) {
    debugStatus = debug;
  }
}
