/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerInfoAdvertisement.java,v 1.10 2001/05/24 02:15:40 akhil Exp $
 */

package net.jxta.protocol;

import net.jxta.document.Advertisement;
import java.util.Hashtable;
import java.util.Enumeration;

/*
 * This abstract class define a PeerInfo advertisement
 *
 * @since JXTA 1.0
 */
public abstract class PeerInfoAdvertisement extends Advertisement {

    /**
     * The Sender's (or Querying) Peer Id.
     */
    private String spid = null;

    /**
     * Peer Id of the peer which published this advertisement.
     */
    private String tpid = null;
  
    /**
     * Time in seconds since this peer was started.
     */
    private int uptime = 0;

    /**
     * Time in seconds when this peer was last polled since
     * "the epoch", namely January 1, 1970, 00:00:00 GMT.
     */
    private int timestamp = 0;

    /**
     * Time in seconds since midnight, January 1, 1970 UTC and when
     * this peer last recieved a message.
     */
    private int lastIncomingMessageTime = 0;

    /**
     * Time in seconds since midnight, January 1, 1970 UTC and when
     * this peer last sent a message out.
     */
    private int lastOutgoingMessageTime = 0;

    /**
     * A set of key-value pairs representing the traffic on this
     * peer. The keys represent service names and the values are
     * java.lang.Long objects representing total bytes transferred in
     * or out to/from this peer. 
     */
    private Hashtable incomingTraffic = new Hashtable();

    /**
     * A set of key-value pairs representing the traffic on this
     * peer. The keys represent service names and the values are
     * java.lang.Long objects representing total bytes transferred in
     * or out to/from this peer. 
     */
    private Hashtable outgoingTraffic = new Hashtable();

    public PeerInfoAdvertisement() {
        super();
    }

    /** 
     * returns the advertisement type
     * @return a string 
     * 
     * @since JXTA 1.0     
     */
    public static String getAdvertisementType() {
        return "jxta:PeerInfoAdvertisement";
    }

    /** 
     * returns the sender's pid
     * @return a string representing the peer's id
     *
     * @since JXTA 1.0
     */
    public String getSourcePid() {
        return spid;
    }
 
    /** 
     * sets the sender's pid
     * @param pid a string representing a peer's id
     *
     * @since JXTA 1.0
     */
    public void setSourcePid(String pid) {
        this.spid = pid;
    }

    /** 
     * returns the target pid
     * @return a string representing the peer's id
     *
     * @since JXTA 1.0
     */
    public String getTargetPid() {
        return tpid;
    }
 
    /** 
     * sets the target's pid
     * @param pid a string representing a peer's id
     *
     * @since JXTA 1.0
     */
    public void setTargetPid(String pid) {
        this.tpid = pid;
    }

    /** 
     * returns the number of seconds since this peer was started
     * @return the number of seconds
     *
     * @since JXTA 1.0
     */
    public int getUptime() {
        return uptime;
    }
 
    /** 
     * sets the number of seconds since this peer was started
     * @param seconds the number of seconds since this peer was started
     *
     * @since JXTA 1.0
     */
    public void setUptime(int seconds) {
        this.uptime = seconds;
    }

    /** 
     * returns the time when this peer was last polled 
     *
     * @return int time in seconds when this peer was last polled in
     * seconds since "the epoch", namely January 1, 1970, 00:00:00
     * GMT.
     *
     * @since JXTA 1.0 */
    public int getTimestamp() {
        return timestamp;
    }
 
    /** 
     * sets the time when this peer was last polled
     *
     * @param seconds time in seconds when this peer was last polled
     * in seconds since "the epoch", namely January 1, 1970, 00:00:00
     * GMT.
     *
     * @since JXTA 1.0 */
    public void setTimestamp(int seconds) {
        this.timestamp = seconds;
    }

    /**
     * Get the time in seconds since this peer last recieved a message
     * in seconds since "the epoch", namely January 1, 1970, 00:00:00
     * GMT.
     *
     * @return time in seconds
     */
    public int getLastIncomingMessageTime() {
        return lastIncomingMessageTime;
    }

    /**
     * Set the time in seconds since this peer last recieved a message
     * in seconds since "the epoch", namely January 1, 1970, 00:00:00
     * GMT.
     *
     * @param t time in seconds
     */
    public void setLastIncomingMessageTime(int t) {
        lastIncomingMessageTime = t;
    }

    /**
     * Get the time in seconds since this peer last sent a message
     * in seconds since "the epoch", namely January 1, 1970, 00:00:00
     * GMT.
     *
     * @return time in seconds
     */
    public int getLastOutgoingMessageTime() {
        return lastOutgoingMessageTime;
    }

    /**
     * Set the time in seconds since this peer last sent a message
     * in seconds since "the epoch", namely January 1, 1970, 00:00:00
     * GMT.
     *
     * @param t time in seconds
     */
    public void setLastOutgoingMessageTime(int t) {
        lastOutgoingMessageTime = t;
    }

    /**
     * Get an enumeration of incoming traffic channels on this peer. A
     * channel represents traffic generated by a particular service. A
     * service may open more than one data channel.
     *
     * @return an enumeration of String 
     */
    public Enumeration getIncomingTrafficChannels() {
        return incomingTraffic.keys();
    }

    /**
     * Get the number of bytes recieved on the specified channel.
     *
     * @return number of bytes recieved
     */
    public long getIncomingTrafficOnChannel(String channel) {
        Long ret = (Long) incomingTraffic.get(channel);
	if (ret != null) 
	    return ret.longValue();
	return 0;
    }

    /**
     * Set the number of bytes recieved on the specified channel.
     *
     * @param channel service name
     * @param long number of bytes recieved
     */
    public void setIncomingTrafficElement(String channel, 
                                          long bytes) {
        incomingTraffic.put(channel, new Long(bytes));
    }

    /**
     * Get an enumeration of outgoing traffic channels on this peer. A
     * channel represents traffic generated by a particular service. A
     * service may open more than one data channel.
     *
     * @return an enumeration of String 
     */
    public Enumeration getOutgoingTrafficChannels() {
        return outgoingTraffic.keys();
    }

    /**
     * Get the number of bytes sent on the specified channel.
     *
     * @return number of bytes sent
     */
    public long getOutgoingTrafficOnChannel(String channel) {
        Long ret = (Long) outgoingTraffic.get(channel);
	if (ret != null)
	    return ret.longValue();
	return 0;
    }

    /**
     * Set the number of bytes sent on the specified channel.
     *
     * @param channel service name
     * @param long number of bytes sent
     */
    public void setOutgoingTrafficElement(String channel, 
                                          long bytes) {
        outgoingTraffic.put(channel, new Long(bytes));
    }
}

