/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ServiceAdvertisement.java,v 1.12 2001/06/18 23:54:21 jice Exp $
 */

package net.jxta.protocol;

import java.util.*;
import net.jxta.document.Element;
import net.jxta.document.Advertisement;

/**
 * This abstract class defines a Service Advertisement. Service advertisement
 * describe an implementation of a peer group service or a peer service. The
 * service advertisement describes the necessary information to a peer to
 * download a version of the service on the peer and to configure the service
 * as a peer group service or as a peer service.<br>
 * Note: ServiceAdvertisement is also used to describe the main Application
 * of a PeerGroup.
 *
 * <p>A service advertisement is part of a Peer or PeerGroup Advertisement
 *
 *<pre>
 * &lt;?xml version="1.0"?&gt;
 * &lt;ServiceAdvertisement&gt;
 *   &lt;Name&gt; name &lt;/Name&gt;
 *   &lt;Version&gt; version &lt;/Version&gt;
 *   &lt;Uri&gt; Uri &lt;/Uri&gt;
 *   &lt;Provider&gt; provider &lt;/Provider&gt;
 *   &lt;Security&gt; security &lt;/Security&gt;
 *   &lt;Code&gt;  implementation code &lt;/Code&gt;
 * &lt;/ServiceAdvertisement&gt;
 *</pre>
 *
 * @see net.jxta.protocol.PeerGroupAdvertisement
 * @see net.jxta.protocol.PeerAdvertisement
 * @see net.jxta.protocol.ServiceAdvertisement
 *
 * @version $Revision: 1.12 $
 * @since JXTA 1.0
 */

public abstract class ServiceAdvertisement extends Advertisement {
  
  private String name = null;
  private String version = null;
  private String uri = null;
  private String provider = null;
  private String security = null;
  private String code = null;
  private String keywords = null;
  private PipeAdvertisement pipe = null;
  private Vector params = new Vector();

  /** 
   * returns the advertisement type
   * @return String type
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */
  
  public static String getAdvertisementType() {
    return "jxta:ServiceAdvertisement" ;
  }
  
  /** 
   * returns the name of the service
   * @return String name
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public String getName() {
    return name;
  }
  
  /** 
   * sets the name
   * @param string name
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */
  
  public void setName(String name) {
    this.name = name;
  }
  
  /** 
   * returns the service version number
   * @return String version number
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */
  
  public String getVersion() {
    return version;
  }
  
  /** 
   * sets the version of the service
   * @param version  version number
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public void setVersion(String version) {
    this.version = version;
  }
  
  /** 
   * returns the provider information for the service
   * @return String provider information
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */
  
  public String getProvider() {
    return provider;
  }
 
  /** 
   * sets the service provider information
   * @param provider information
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public void setProvider(String provider) {
    this.provider = provider;
  }

  /** 
   * returns the uri
   * @return String uri
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public String getUri() {
    return uri;
  }
 
  /** 
   * sets the uri
   * @param uri string uri
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public void setUri(String uri) {
    this.uri = uri;
  }

  /** 
   * returns the service implementation
   * @return String service implementation
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public String getCode() {
    return code;
  }
 
  /** 
   * sets the service implementation
   * @param code string service implementation
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public void setCode(String code) {
    this.code = code;
  }

  /** 
   * returns the security service advertisement
   * @return String security service
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */
  
  public String getSecurity() {
    return security;
  }
  
  /** 
   * sets the security service name for this service
   * @param security security service
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public void setSecurity(String security) {
    this.security = security;
  }   

  /** 
   * returns the pipe advertisment for this service
   * @return PipeAdvertisement pipe advertisement
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public PipeAdvertisement getPipe() {

    return pipe;
  }
  
  /** 
   * sets the pipe advertisement for the service
   * @param pipe advertisment pipe for the service
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   */

  public void setPipe(PipeAdvertisement pipe) {
    this.pipe = pipe;
  }
   

  /** 
   * returns the service search keywords
   * @return String keywords associated with the service
   * 
   * @version $Revision: 1.12 $
   * @since JXTA 1.0     
   * 
   */
  
  public String getKeywords() {
    return keywords;
  }

  /** 
   * sets the the service seach keywords
   * @param  keywords string keywords associated with the plocy
   * 
   * @since JXTA 1.0     
   */

  public void setKeywords(String keywords) {
    this.keywords = keywords;
  }

  /** 
   * returns the service params
   *
   * @return Vector params containing a collection of Strings
   * 
   * @since JXTA 1.0     
   * 
   */
  
  public Vector getParams() {
    return (Vector) params.clone();
  }

  /** 
   * sets the service params
   * @param params vector of service params
   * 
   * @since JXTA 1.0     
   */

  public void setParams(Vector params) {
    this.params = (Vector) params.clone();
  }

}
