/************************************************************************
 *
 * $Id: RendezVous.java,v 1.12 2001/05/23 00:28:22 lomax Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.rendezvous;

import java.io.IOException;
import java.util.Enumeration;

import net.jxta.peer.PeerID;
import net.jxta.protocol.PeerGroupAdvertisement; 
import net.jxta.protocol.PeerAdvertisement;
import net.jxta.endpoint.EndpointAddress;
import net.jxta.service.Service;

/** 
 * This interface provides an API to the Jxta RendezVous mechanism
 * allowing applications to add, remove, monitor RendezVous peers
 * (client side), or control the local peer, if it is a RendezVous
 * itself.
 *
 * <p>Any peer in a peer group can become a rendezvous for a peer group. A
 * rendezvous provides a long-range discovery mechanism where members of a
 * peer group can find themselves. Rendezvous are advertised as part of the
 * peer group advertisement of a peer group.
 *
 * @see net.jxta.rendezvous.RendezVousManager
 * @see net.jxta.rendezvous.RendezVousMonitor
 * @see net.jxta.discovery.Discovery
 * @see net.jxta.protocol.PeerGroupAdvertisement
 *
 * @version $Revision: 1.12 $
 * @since JXTA 1.0
 */

public interface RendezVous extends Service {

    /**
     ** This portion of the API is for the peers that are connected
     ** to a RendezVous peer.
     **/

    /**
     * Add a peer as a new RendezVous point.
     * If/When the RendezVous accepts the connection, the RendezVous
     * service will invoke the RendezVousMonitor.
     *
     * @param adv the advertisement of the RendezVous peer
     * @exception IOException when the RendezVous peer is not reachable
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public void connectRendezVous (PeerGroupAdvertisement adv) throws IOException;

    /**
     * Add a peer as a new RendezVous point.
     * If/When the RendezVous accepts the connection, the RendezVous
     * service will invoke the RendezVousMonitor.
     *
     * @param addr EndpointAddress of the rendezvous peer
     * @exception IOException when the RendezVous peer is not reachable
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public void connectRendezVous (EndpointAddress addr) throws IOException;


    /**
     * Remove a RendezVous point.
     *
     * @param rendezVous the PeerId of the RendezVous point to remove
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public void removeRendezVous (PeerID rendezVous);

    /**
     * Register a notification handler that is called each time a RendezVous
     * peer is not reachable anymore.
     *
     * @param handler a handler that is called each time a RendezVous peer is
     * is not reachable anymore.
     * @return RendezVousMonitor returns, if any, the current RendezVousMonitor
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public RendezVousMonitor setMonitor (RendezVousMonitor handler);

    /**
     * Returns an Enumeration of the PeerID all the RendezVous on which this
     * Peer is currentely connected.
     *
     * @return Enumeration enumeration of RendezVous

     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public Enumeration listRendezVous ();

    /**
     ** This portion is for peers that are RendezVous
     **/

    /**
     * Start the local peer as a RendezVous peer. The caller must provide
     * a handler to an authorization manager.
     *
     * @param handler is the RendezVousManager that will be invoked each
     * time a new Peer request to be connected.
     *
     * @exception IOException when a handler has already been connected
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public void startRendezVous (RendezVousManager handler) throws IOException;

    /**
     * Stop the RendezVous function on the local Peer. All connected Peer are
     * disconnected.
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public void stopRendezVous ();

    /**
     * Returns an Enumeration of PeerID of the peers that are currentely
     * connected.
     *
     * @return Enumeration enumeration of peers connected to that rendezvous
     * 
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public Enumeration getConnectedPeers();

    /**
     * Sends advertisement about other RendezVous to a given peer
     *
     * @param destPeer is the advertisement of the peer to which to 
     * send the RendezVous advertisement.
     * @param rendezVous is the advertisment of a RendezVous peer
     *
     * @version $Revision: 1.12 $
     * @since JXTA 1.0
     */

    public void sendRendezVousAdv (PeerAdvertisement destPeer,
				   PeerAdvertisement rendezVous);
}



