/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ClassFactory.java,v 1.11 2001/06/08 23:03:23 bondolo Exp $
 */

package net.jxta.util;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.String;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import java.util.Vector;

import java.lang.reflect.InvocationTargetException;
//import java.lang.reflect.UndeclaredThrowableException;
import java.util.MissingResourceException;
import java.util.NoSuchElementException;

/**
 * This util class provides methods needed by class construction factories.
 *
 * FIXME bondolo@jxta.org 20010223 This implementation needs a way to return the list of
 * key -> class associations
 *
 * @see net.jxta.document.StructuredDocumentFactory
 * @see net.jxta.document.AdvertisementFactory
 * @see net.jxta.id.IDFactory
 *
 *
 * @version $Revision: 1.11 $
 * @since JXTA 1.0
 */

public abstract class ClassFactory {
    
    /**
     *  Used by ClassFactory methods to get the mapping of keys to constructors.
     *
     *  @return the hashtable containing the mappings.
     *
     *  @since JXTA 1.0
     */
    protected abstract Hashtable getAssocTable();
    
    /**
     *  Used by ClassFactory methods to ensure that all keys used with the mapping are
     *  of the correct type.
     *
     *  @return Class object of the key type.
     *
     *  @since JXTA 1.0
     */
    protected abstract Class getClassForKey();
    
    /**
     *  Used by ClassFactory methods to ensure that all of the instance classes
     *  which register with this factory have the correct base class
     *
     *  @return Class object of the "Factory" type.
     *
     *  @since JXTA 1.0
     */
    protected abstract Class getBaseClassForFactory();
    
    /**
     *  Used by ClassFactory methods to determine the required set of constructors
     *  needed by the instance classes generated by this factory.
     *
     *
     *  @since JXTA 1.0
     */
    protected abstract Class [] [] getRequiredConstructors();
    
    /**
     *  Used by ClassFactory methods to determine the required set of static
     *  methods needed by classes registering with this factory.
     *
     *  @return Object
     *
     *  @since JXTA 1.0
     */
    protected abstract Object [] [] getRequiredStaticMethods();
    
    /**
     * Given a resource bundle identifier and a property name register instance
     * classes. The property must be a string containing class names which must
     * be found on the current class path. The class names are seperated by
     * spaces.
     *
     * @return boolean true if at least one instance class could be registered with
     * this factory.
     * @param resourceName name of the resource bundle
     * @param propertyName name of the property.
     *
     * @exception MissingResourceException if the resource bundle or
     * property cannot be located.
     *
     *  @since JXTA 1.0
     */
    protected boolean registerFromResources( String resourceName,
    String propertyName )
    throws MissingResourceException {
        
        ResourceBundle jxtaRsrcs = ResourceBundle.getBundle( resourceName );
        String fromProps = jxtaRsrcs.getString( propertyName ).trim();
        
        return registerFromString( fromProps );
    }
    
    /**
     *  Register instance classes given a string containing class names which 
     *  must be found on the current class path. The class names are seperated
     *  by spaces.
     *
     *  @param classNamesString The class name list
     *
     *  @return boolean true if at least one of the instance classes could be
     *  registered otherwise false.
     *
     *  @since JXTA 1.0
     */
    protected boolean registerFromString( String classNamesString ) {
        boolean registeredSomething = false;
        
        if( (null == classNamesString) || (0 == classNamesString.length()) )
            return false;
        
        // make sure the static initialisers for each instance class are called.
        for( StringTokenizer eachInstanceClass =
            new StringTokenizer(classNamesString); 
            eachInstanceClass.hasMoreTokens();  ) {
            try {
                String willInitialize = (String) eachInstanceClass.nextToken();
                registeredSomething |= registerAssoc( willInitialize );
            }
            catch( Exception allElse ) {
                //allElse.printStackTrace();
            }
        }
        
        return registeredSomething;
    }
    
    /**
     *  Register a class with the factory from its class name. Since class name
     *  doesn't tell us much, we just load the class and hope that something
     *  happens as a result of the class loading. This class is often overridden
     *  in class factories to interogate the instance class before registering
     *  the instance class.
     *
     * @param className The class name which will be regiestered.
     * @return boolean true if the class was registered otherwise false.
     * @Exception Exception   when an error occurs.
     *
     *  @since JXTA 1.0
     */
    protected boolean registerAssoc( final String className ) throws Exception {
        
        boolean registeredSomething = false;
        
        try {
            Class ignore = Class.forName( className );
            registeredSomething = true;
        }
        catch( ClassNotFoundException ignored ) {
            //System.out.println(className + "not supported" );
        }
        
        return registeredSomething;
    }
    
    /**
     * Register a key and instance class with the factory.
     * @param key The key to register.
     * @param instanceClass The class which will be registered for this key.
     *
     * @exception NoSuchMethodException If one one of the required
     * constructors or methods is not present in the instance class.
     *
     * @exception SecurityException If the instance class cannot be
     * registered because either it or one of its methods is inaccesssible
     * due to a security service.
     *
     * @return booelan true if the key was successfully registered otherwise false.
     *
     *  @since JXTA 1.0
     */
    protected boolean registerAssoc( final Object key, Class instanceClass )
    throws NoSuchMethodException, SecurityException {
        
        if( !getBaseClassForFactory().isAssignableFrom(instanceClass) )
            throw new ClassCastException( "instance class '" +
            instanceClass.getName()+"' is not a descendant of '" +
            getBaseClassForFactory().getName() + "'" );
        
        // Check the class of the key to make sure it is of the right class
        Class requiredKeyClass = getClassForKey();
        Class itsA =  key.getClass();
        if( !requiredKeyClass.isAssignableFrom(itsA ) )
            throw new IllegalArgumentException( "Incorrect Class for key type" );
        
        // Check the instance class for the correct constructors.
        Class requiredConstructors [] [] = getRequiredConstructors();
        for( int eachConstructor = 0; 
        eachConstructor < requiredConstructors.length; eachConstructor++ ) {
            // This will throw exceptions if the constructors are missing.
            try {
                Constructor aConstructor = instanceClass.getConstructor(
                requiredConstructors[eachConstructor] );
            } catch( NoSuchMethodException notFound ) {
                throw new NoSuchMethodException( "Missing constructor #"+
                eachConstructor + " for class '" + instanceClass.getName() + "'" );
            }
        }
        
        // Check the instance class for the correct static Methods.
        Object requiredStaticMethods [] [] = getRequiredStaticMethods();
        for( int eachStatic = 0; eachStatic < requiredStaticMethods.length;
        eachStatic++ ) {
            // This will throw exceptions if the methods are missing.
            try {
                String methodName = (String) requiredStaticMethods[eachStatic][0];
                Class [] methodParams = (Class []) requiredStaticMethods[eachStatic][1];
                
                Method aStatic = instanceClass.getMethod( methodName, methodParams );
                
                // must be public
                if( !Modifier.isPublic(aStatic.getModifiers()) )
                    throw new NoSuchMethodException( );
                
                // must be static
                if( !Modifier.isStatic(aStatic.getModifiers()) )
                    throw new NoSuchMethodException( );
                
            } catch( NoSuchMethodException notFound ) {
                throw new NoSuchMethodException( "Missing method '" +
                requiredStaticMethods[eachStatic][0] +
                "' for class '" + instanceClass.getName() + "'" );
            }
        }
        
        // Check the association table to make sure this key is not already present.
        if( null != getAssocTable().get( key ) )
            return false;
        
        getAssocTable().put( key, instanceClass );
        
        //System.out.println( "Factory : " + getClass().getName() +
        //" Registered key : '" + key + "' for class :" + instanceClass.getName() );
        
        return true;
    }
    
    /**
     * Create a new instance of the class associated with the provided key.
     * @param key The identifier for the instance class to be created.
     *
     * @param constructorMatching identifies which constructor is needed for
     * the parameter set provided.
     *
     * @param params the parameters which will be provided to the constructor.
     *
     * @exception InvocationTargetException If there is an error in the
     * execution of the constructor for the instance class.
     *
     * @return Object The new instance.
     *
     *  @since JXTA 1.0
     */
    protected Object newInstance( final Object key,
    final Class [] constructorMatching, Object [] params )
    throws InvocationTargetException {
        
        Class requiredKeyClass = getClassForKey();
        
        if( !requiredKeyClass.isAssignableFrom( key.getClass() ) )
            throw new IllegalArgumentException( "Incorrect Class for key type" );
        
        // Get the constructors for this key.
        Class instanceClass = (Class) getAssocTable().get( key );
        
        if( null == instanceClass )
            throw new NoSuchElementException( "key '" + key + "' not registered." );
        
        Constructor constructor = null;
        try {
            constructor = instanceClass.getConstructor( constructorMatching );
        }
        catch( Exception everthing ) {
            // we should have already caught this when the class registered.
            throw new  NoSuchElementException( "No matching method" );
        }
        
        try {
            return constructor.newInstance( params );
        } catch( InvocationTargetException fromMethod ) {
            throw fromMethod;
        } catch( Exception all  ) {
            if( all instanceof RuntimeException )
                throw (RuntimeException) all;
            else
                // XXX bondolo@jxta.org UndeclaredThrowableException would be preferable
                throw new InvocationTargetException( all );
        }
    }
    
    /**
     * Call a static method of the class associated with the provided key.
     * @param key The identifier for the instance class to be created.
     *
     * @param name identifies the method name
     *
     * @param params the parameters which will be provided to the method.
     *
     * @exception InvocationTargetException If there is an error in the
     * execution of the constructor for the instance class.
     *
     * @return Object ots result.
     *
     *  @since JXTA 1.0
     */
    protected Object callStaticMethod( final Object key, final String name,
    final Class [] paramSet, Object [] params )
    throws InvocationTargetException {
        
        Class requiredKeyClass = getClassForKey();
        
        if( !requiredKeyClass.isAssignableFrom( key.getClass() ) )
            throw new IllegalArgumentException( "Incorrect Class for key type" );
        
        // Get the constructors for this key.
        Class instanceClass = (Class) getAssocTable().get( key );
        
        if( null == instanceClass )
            throw new NoSuchElementException( "key '" + key + "' not registered." );
        
        Method method = null;
        try {
            method = instanceClass.getMethod( name, paramSet );
        }
        catch( Exception everthing ) {
            // we should have already caught this when the class registered.
            throw new  NoSuchElementException( "No matching method" );
        }
        
        try {
            return method.invoke( null, params );
        } catch( InvocationTargetException fromMethod ) {
            throw fromMethod;
        } catch( Exception all  ) {
            if( all instanceof RuntimeException )
                throw (RuntimeException) all;
            else
                // XXX bondolo@jxta.org UndeclaredThrowableException would be preferable
                throw new InvocationTargetException( all );
        }
    }
    
}
