/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Config.java,v 1.7 2001/05/23 01:00:17 lomax Exp $
 */

package net.jxta.impl.config;

import net.jxta.platform.*;
import net.jxta.document.*;
import net.jxta.endpoint.EndpointAddress;
import net.jxta.impl.endpoint.Address;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.Writer;
import java.io.InputStream;

// purely for the main() unit test
import java.io.FileWriter;
import java.io.File;
import java.io.FileInputStream;


/** Instances of Config class provides the interface for configuring a JXTA
    peer. Config provides:
     <li>an interface to write a configuration to XML (write())
     <li>an interface to read a configuration from XML (static read())
     <li>an interface to create default a configuration (static
         createDefaultConfig())
     <li>an interface to read configuration information
     <li>an interface to set configuration information
**/
public class Config {
    /** debug flag **/
    private String mDebugLevel;

    /** Hashtable of {service key (String), CodeDescriptor  } **/
    private Hashtable mServiceCodeDescriptors = new Hashtable();

    /** the descriptor of the initial class that is booted at startup **/
    private CodeDescriptor mInitialPlatform;

    /** the descriptor of the initial peergoup class that is booted at
        startup **/
    private CodeDescriptor mInitialApp;

    /** TCP Transport Configuration **/
    private TcpConfig mTcpConfig=null;

    /** Http Transport Configuration **/
    private HttpConfig mHttpConfig=null;

    /** Vector of EndpointAddress objects of rendezvous peers **/
    private Vector mRendezvousPeers = new Vector();

    /** Stores whether the peer is a router peer **/
    private boolean mIsRouter;

    /** Stores whether the peer is a rendezvous peer **/
    private boolean mIsRendezvous;

    /** Stores whether the peer should propagate messages **/
    private boolean mShouldPropagate;
    
    // Debug level constants
    public static final String DEBUG_NONE = "none";
    public static final String DEBUG_WARNING = "warning";
    public static final String DEBUG_VERBOSE = "verbose";

    // Service keys for mServiceImpls and mServiceCodebases
    public static final String SERVICE_DISCOVERY = "Discovery";
    public static final String SERVICE_MEMBERSHIP = "Membership";
    public static final String SERVICE_PIPE = "Pipe";
    public static final String SERVICE_PEERINFO = "PeerInfo";
    public static final String SERVICE_RESOLVER = "Resolver";

    /** Array of services.  This should be kept in sync with the list above--
    the config reader and writer use this array to streamline the
    reaading/writing process. **/
    private static final String[] SERVICE_KEYS = { SERVICE_DISCOVERY,
        SERVICE_MEMBERSHIP,
        SERVICE_PIPE,
        SERVICE_PEERINFO,
        SERVICE_RESOLVER};

    private static final String JXTA_CODEBASE =
        "http://www.jxta.org/download/jxta.jar";

    /** Creates a blank configuration.  Normally you will want to create a
	Config object by either reading a pre-existing XML document or
	by calling the static createDefaultConfig()
    **/
    public Config() {
    }

    
    /** Creates a default configuration instance.
    
    @throw RuntimeException A RuntimeException is thrown when there
    is a problem creating a default configuration object. This should
    never happen.
    **/
    public static Config createDefaultConfig() {
        try {
            Config conf = new Config();
            conf.setDebugLevel( DEBUG_NONE );

            conf.setServiceImplDescriptor( SERVICE_DISCOVERY, 
                new CodeDescriptor( "net.jxta.impl.discovery.Discovery",
                                    JXTA_CODEBASE ) );

            conf.setServiceImplDescriptor( SERVICE_PIPE, 
                new CodeDescriptor( "net.jxta.impl.pipe.Pipe",
                                    JXTA_CODEBASE ));

            conf.setServiceImplDescriptor( SERVICE_PEERINFO, 
                new CodeDescriptor( "net.jxta.impl.peer.PeerInfo",
                                    JXTA_CODEBASE));

            conf.setServiceImplDescriptor( SERVICE_RESOLVER, 
                new CodeDescriptor( "net.jxta.impl.resolver.Resolver",
                                    JXTA_CODEBASE));

            conf.setServiceImplDescriptor( SERVICE_MEMBERSHIP, 
                new CodeDescriptor( "net.jxta.impl.membership.NullMembership",
                                    JXTA_CODEBASE ) );

            conf.setRendezvous( false );

            // XXX: We should probably not do this in the code at all.
            // if we distribute JXTA with hardcoded rendezvous, they should
            // go into a jxtaconfig.xml that we distribute rather than
            // here.
            conf.addRendezvousPeer( new Address( "tcp://129.144.36.190:6001" ) );
            conf.addRendezvousPeer( new Address( "tcp://jxta.dioxine.net:6001" ) );
            conf.addRendezvousPeer( new Address( "http://jxta.dioxine.net:6002" ) );
            
            conf.setShouldPropagate( false );

            conf.setTcpConfig( TcpConfig.createDefaultConfig() );

            conf.setHttpConfig( HttpConfig.createDefaultConfig() );

            conf.setInitialPlatform( new CodeDescriptor(
                "net.jxta.impl.peergroup.StartNetPeerGroup", JXTA_CODEBASE ) );

            conf.setInitialApp( new CodeDescriptor (
                "net.jxta.impl.shell.bin.Shell.Shell", JXTA_CODEBASE ) );

            return conf;
        } catch ( ConfigurationException e ) {
            throw new RuntimeException( "Fatal error with default config" );
        }
    }

    /** Gets debug level. See DEBUG_ constants. **/
    public String getDebugLevel() {
        return mDebugLevel;
    }

    /** Sets debug level. See DEBUG_ constants. **/
    public void setDebugLevel( String aDebugLevel ) {
        mDebugLevel = aDebugLevel;
    }


    /** Given a service key, returns the descriptor of the implementation class.
    See the set of SERVICE_ constants above. **/
    public CodeDescriptor getServiceImplDescriptor( String aServiceKey ) {
        return(CodeDescriptor) mServiceCodeDescriptors.get( aServiceKey );
    }

    /** Correlates a service with a code descriptor. **/
    public void setServiceImplDescriptor(
        String aServiceKey,
        CodeDescriptor aDescriptor ) throws ConfigurationException {

        try {
            mServiceCodeDescriptors.put( aServiceKey, aDescriptor );
        } catch ( NullPointerException e ) {
            throw new ConfigurationException( "Null service key provided", e );
        }
    }


    /** Tells whether the peer is a rendezvous peer **/
    public boolean isRendezvous() { 
        return mIsRendezvous;
    }

    /** Sets whether the peer is a rendezvous peer **/
    public void setRendezvous( boolean aIsRendezvous ) {
        mIsRendezvous = aIsRendezvous;
    }

    /** Tells whether the peer is a router peer **/
    public boolean isRouter() {
        return mIsRouter;

    }

    /** Sets whether the peer is a rendezvous peer **/
    public void setRouter( boolean aIsRouter ) {
        mIsRouter = aIsRouter;
    }


    /** Gets the contained TCP Transport configuration **/
    public TcpConfig getTcpConfig() {
        return mTcpConfig;
    }

    /** Sets the contained TCP Transport configuration **/
    public void setTcpConfig( TcpConfig aConfig ) {
        mTcpConfig = aConfig;
    }


    /** Gets the contained Http Transport configuration **/
    public HttpConfig getHttpConfig() {
        return mHttpConfig;
    }

    /** Sets the contained Http Transport configuration **/
    public void setHttpConfig( HttpConfig aConfig ) {
        mHttpConfig = aConfig;
    }


    /** Returns the descriptor of the initial class that is booted at
        startup **/
    public CodeDescriptor getInitialPlatform() {
        return mInitialPlatform;
    }

    /** Sets the classname of the initial class that is booted at startup **/
    public void setInitialPlatform( CodeDescriptor aInitialPlatform ) {
        mInitialPlatform = aInitialPlatform;
    }


    /** Returns the classname of the initial class that is booted at startup **/
    public CodeDescriptor getInitialApp() {
        return mInitialApp;
    }

    /** Sets the classname of the initial class that is booted at startup **/
    public void setInitialApp( CodeDescriptor aInitialApp ) {
        
        mInitialApp = aInitialApp;
    }


    /** Adds a rendezvous peer **/
    public void addRendezvousPeer( EndpointAddress address ) {
        mRendezvousPeers.addElement( address );
    }

    /** Returns an enumeration of rendezvous peers **/
    public Enumeration getRendezvousPeers() {
        return mRendezvousPeers.elements();
    }

    /** Clears ou the list of rendezvous peers **/
    public void clearRendezvousPeers() {
        mRendezvousPeers = new Vector();
    }

    /** Tells whether messages should be propagated. **/
    public boolean getShouldPropagate() {
        return mShouldPropagate;
    }

    /** Sets whether messages should be propagated. **/
    public void setShouldPropagate( boolean aShouldPropagate ) {
        mShouldPropagate = aShouldPropagate;
    }


    /** Reads the configuration file in aInStream and creates a new
        Configuration. This is a static method such that the Configuration
        object can be tightly controlled (notice that there are no public
        contructors).
    **/
    public static Config read( InputStream aInStream ) throws ConfigurationException {
                
        /* XXX why does StructuredDocFctry take an inputstream
        rather than a reader?  A writer is used to writing config files. 
        Same method should be used for reading, therefore.  */

        try {
            Config config = new Config();

            StructuredTextDocument doc = (StructuredTextDocument)
                StructuredDocumentFactory.newStructuredDocument(
                    new MimeMediaType ("text/xml"), aInStream);

            config.readConfig( doc );

            return config;
        } catch ( ConfigurationException e ) {
            throw e;
        } catch ( Exception e ) {
            throw new ConfigurationException( "Fatal error reading configuration file", e );
        }
    }

    private void readConfig( StructuredTextDocument doc )
    throws ConfigurationException {
        Enumeration elements = doc.getChildren();

        if ((elements == null) || (!elements.hasMoreElements())) {
            throw new ConfigurationException( "Configuration file is empty or corrupted" );
        }

        while (elements.hasMoreElements()) {
            TextElement element = (TextElement) elements.nextElement();
            readElement( element );
        }
    }

    /** helper function for reading **/
    private void readElement( TextElement e ) throws ConfigurationException {
        String elname = e.getName();
        if (elname.equals( ConfigConstants.KEY_DEBUG_LEVEL ) )
            readDebugLevel( e );
        else if ( elname.equals( ConfigConstants.KEY_IS_ROUTER ) )
            mIsRouter = new Boolean( e.getTextValue() ).booleanValue();
        else if ( elname.equals( ConfigConstants.KEY_IS_RENDEZVOUS_SERVER ) )
            mIsRendezvous = Boolean.valueOf( e.getTextValue() ).booleanValue();
        else if ( elname.equals( ConfigConstants.KEY_RENDEZVOUS_PEERS ) )
            readRendezvousPeers( e );
        else if ( elname.equals( ConfigConstants.KEY_SERVICES ) )
            readServices( e );
        else if ( elname.equals( ConfigConstants.KEY_INITIAL_PLATFORM ) )
            readInitPlatform( e );
        else if ( elname.equals( ConfigConstants.KEY_INITIAL_APP ) )
            readInitApp( e );
        else if ( elname.equals( ConfigConstants.KEY_TCP_TRANSPORT ) )
            readTcp( e );
        else if ( elname.equals( ConfigConstants.KEY_HTTP_TRANSPORT ) )
            readHttp( e );
        else if ( elname.equals( ConfigConstants.KEY_PROPAGATE ) )
            mShouldPropagate = Boolean.valueOf( e.getTextValue() ).booleanValue();
        else
            System.err.println( "Unknown element seen during config read: " + elname );
    }

    /** helper function for reading **/
    private void readServices( TextElement e ) throws ConfigurationException {
        Enumeration services = e.getChildren();
        if ( services!=null ) {
            while ( services.hasMoreElements() ) {
                TextElement service = (TextElement) services.nextElement();
                if ( service!=null ) {
                    String servicename = service.getName();
                    CodeDescriptor cd = CodeDescriptor.newFromXML( service );
                    if ( cd != null ) {
                        setServiceImplDescriptor( servicename, cd );
                    }
                }
            }
        }
    }

    /** helper function for reading **/
    private void readDebugLevel( TextElement e ) {
        mDebugLevel = e.getTextValue().toLowerCase();
    }

    /** helper function for reading **/
    private void readRendezvousPeers( TextElement e ) throws ConfigurationException {
        Enumeration peers = e.getChildren();
        if ( peers != null ) {
            while ( peers.hasMoreElements() ) {
                TextElement peer = (TextElement) peers.nextElement();
                addRendezvousPeer( new Address( peer.getTextValue() ) );
            }
        }
    }

    /** helper function for reading **/
    private void readInitPlatform( TextElement e ) {
        mInitialPlatform = CodeDescriptor.newFromXML( e );
    }
    
    /** helper function for reading **/
    private void readInitApp( TextElement e ) {
        mInitialApp = CodeDescriptor.newFromXML( e );
    }
    
    /** helper function for reading **/
    private void readHttp( TextElement e ) {
        mHttpConfig = HttpConfig.newFromXML( e );
    }
    
    /** helper function for reading **/
    private void readTcp( TextElement e ) {
        mTcpConfig = TcpConfig.newFromXML( e );
    }

    /** Writes configuration to a writer in XML format. **/
    public void write( Writer aWriter )
    throws ConfigurationException {
        try {
            // create doc
            StructuredTextDocument doc = null;
            doc = (StructuredTextDocument) StructuredDocumentFactory.newStructuredDocument(
                new MimeMediaType("text/xml"), "jxta:config");
            
            // populate the doc
            writeConfig( doc );

            // write the doc
            doc.sendToWriter( aWriter );
        } catch ( ConfigurationException e ) {
            // same as what I'm throwing: just rethrow.
            throw e;
        } catch ( Exception e ) {
            e.printStackTrace();
            throw new ConfigurationException( "Fatal error during configuration write.", e );
        }

    }

    /** writes all the configuration **/ 
    private void writeConfig( StructuredTextDocument doc )
    throws ConfigurationException {
        if ( mHttpConfig != null )
            mHttpConfig.writeToXML( doc );

        if ( mTcpConfig != null )
            mTcpConfig.writeToXML( doc );

        Element e=null;

        e = doc.createElement( ConfigConstants.KEY_IS_ROUTER, new Boolean( mIsRouter ).toString() );
        doc.appendChild( e );

        writeRendezvousServer( doc );
        
        doc.appendChild( doc.createElement(
            ConfigConstants.KEY_PROPAGATE, String.valueOf( mShouldPropagate ) ) );
        
        writeServices( doc, mServiceCodeDescriptors );

        e = doc.createElement( ConfigConstants.KEY_INITIAL_APP );
        doc.appendChild( e );
        getInitialApp().writeToXML( doc, e );

        e = doc.createElement( ConfigConstants.KEY_INITIAL_PLATFORM );
        doc.appendChild( e );
        getInitialPlatform().writeToXML( doc, e );

        e = doc.createElement( ConfigConstants.KEY_DEBUG_LEVEL, getDebugLevel() );
        doc.appendChild( e );
    }

    /** helper function for writing **/
    private void writeRendezvousServer( StructuredTextDocument doc ) {
        writeRendezvousPeers( doc, getRendezvousPeers() );

        doc.appendChild( doc.createElement(
            ConfigConstants.KEY_IS_RENDEZVOUS_SERVER,
            new Boolean( isRendezvous() ).toString() ) );
    }

    /** writes the rendezvous peers **/
    private void writeRendezvousPeers( StructuredTextDocument doc, Enumeration peers ) {
        Element peersElement = doc.createElement( ConfigConstants.KEY_RENDEZVOUS_PEERS );
        doc.appendChild( peersElement );

        while ( peers.hasMoreElements() ) {
            EndpointAddress addr = (EndpointAddress) peers.nextElement();
            ConfigUtil.writeAddress( doc,
                                     peersElement,
                                     addr,
                                     ConfigConstants.KEY_RENDEZVOUS_PEER );
        }
    }

    /** writes the services **/
    private void writeServices( StructuredTextDocument doc, Hashtable aServiceCodeDescriptors )
    throws ConfigurationException {
        Element services = doc.createElement( ConfigConstants.KEY_SERVICES );
        doc.appendChild( services ); 
        for ( int i=0; i<SERVICE_KEYS.length; i++ ) {
            CodeDescriptor desc = (CodeDescriptor) aServiceCodeDescriptors.get( SERVICE_KEYS[i] );
            if ( desc == null )
                throw new ConfigurationException( "Service " + SERVICE_KEYS[i] + " not configured" );

            Element service = doc.createElement( SERVICE_KEYS[i] );
            services.appendChild( service );
            desc.writeToXML( doc, service );
        }
    }


    /** Parameterized test driver.  Here's the options:
        1 - create default config
        2 - read config file
        3 - write config file
        
        If no parms:
            - creates default config file
            - writes it
            - reads it
            - writes what was just read
    **/
    private static void main(String[] args) throws Exception {
        Config c = Config.createDefaultConfig();
        File firstfile = new File( "first.xml" );
        File secondfile = new File( "second.xml" );

        FileWriter firstwriter = new FileWriter( firstfile );
        FileWriter secondwriter = new FileWriter( secondfile );
        try {
            c.write( firstwriter );
            firstwriter.close();
            Config c2 = Config.read( new FileInputStream( firstfile ) );
            c2.write( secondwriter );
            secondwriter.close();
        } catch ( ConfigurationException e ) {
            System.err.println( e );
            e.getNestedException().printStackTrace();
            e.printStackTrace();
        }
    }
}
