/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: AdvertisementCompactor.java,v 1.9 2001/06/07 22:57:49 bondolo Exp $
 */


package net.jxta.impl.discovery;

import java.util.Enumeration;
import java.util.Hashtable;
import java.io.IOException;

import net.jxta.resolver.Resolver;
import net.jxta.service.Service;
import net.jxta.protocol.*;
import net.jxta.discovery.Discovery;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.impl.peergroup.*;
import net.jxta.impl.protocol.*;

public class AdvertisementCompactor  implements Runnable {

    /**
     *  The task of this thread is to Compact Peer Advertisements.
     *  The problem arises when  peers take on new identities
     *  could occur when the configuration is removed, or the
     *  the software is started/installed in a different directory.
     *  the mechanism it utilizes is a simple one it looks for
     *  duplicate endpoint for different peers. and ejects one.
     *
     *  One of the ways to enhance this is by using a service or service
     *  like pip to verify which PeerId is associated with the transport
     *
     */
     //XXX enhance hamada@jxta.org

     private PeerGroup myGroup;
     private String localPeerId;
     // lets start with every 1 hour
     // once we have pip we will utilize it to verify a peer
     // and we'll switch to a longer nap
     // this should also be a configurable in the jxtaConfig
     // volunteers who want to take on this task, are welcome
     private int nap = 1000*60*10;
     private Discovery discovery = null;
     private boolean debug=false;
     private Hashtable table ;


    public AdvertisementCompactor(PeerGroup pg) {
            myGroup = pg;
            localPeerId = myGroup.getPeerID().toString();
            debug=((RefPeerGroup) myGroup).getDebug();
    }

    /**
     *  to be called during group initialization
     *  different groups can each have a compactor
     */
    public void init (){

         debug("Start AdvertisementCompactor");
         Thread thread  = new Thread (new AdvertisementCompactor(myGroup),
                  "DiscoveryService:AdvertisementCompactor" );
         thread.start();
    }

    /**
     *  the simple mechanism, this could use improvements
     *  XXXX improve  hamada@jxta.org
     */
    public void run() {

      while (discovery == null) {
          discovery = ((GenericPeerGroup) myGroup).getDiscovery();
          doze(nap);
      }

      Enumeration peers;

      while (true) {
         table = new Hashtable(128);
         try {
             peers = discovery.getLocalAdvertisements(discovery.PEER,null,null);
             while ( (null != peers) && peers.hasMoreElements())  {

                 PeerGroupAdv padv = (PeerGroupAdv) peers.nextElement();
                 if (! padv.getPid().equals(localPeerId) ){
                     if (padv.getEndpointAdvertisements().size() == 0) {
                         // let's zap it
                      discovery.flushAdvertisements(padv.getPid(),discovery.PEER);
                     }
                     EndpointAdvertisement endp = (EndpointAdvertisement)
                         (padv.getEndpointAdvertisements().elements()).nextElement();

                     String addr = endp.getEndpointAddress();
                     if ((table.get(addr))==null){
                         debug("adding :"+addr);
                         table.put(addr, padv.getPid());
                     } else {
                         debug("duplicate found :"+addr);
                         discovery.flushAdvertisements(padv.getPid(),discovery.PEER);
                         debug("ejecting :"+padv.getPid());
                     }
                 }
             }
         } catch (IOException e ) {
             if (debug) e.printStackTrace();
         }
         debug("AdvertisementCompactor Done");
         doze(nap);
      }
    }


    private void doze (int period){
      try {
         debug("AdvertisementCompactor hibernating for:"+period/1000/60+" min");
         Thread.sleep(period);
      } catch (Exception e) { }
    }


    private void debug (String str) {
    if (debug) {
        System.out.println ("[AdvertisementCompactor] " + str);
    }
    }

}


