/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: DiscoveryInterface.java,v 1.11 2001/05/23 01:00:18 lomax Exp $
 */

package net.jxta.impl.discovery;

import java.util.*;
import java.io.*;
import java.net.*;
import java.lang.reflect.InvocationTargetException;


import net.jxta.service.*;
import net.jxta.discovery.*;
import net.jxta.resolver.*;
import net.jxta.protocol.*;
import net.jxta.document.*;
import net.jxta.endpoint.*;
import net.jxta.peergroup.*;
import net.jxta.exception.*;
import net.jxta.id.*;


/**
 * DiscoveryInterface provides a pure interface object that permits
 * interaction with the actual DiscoveryService implementation without giving
 * access to the real object.
 */

public class DiscoveryInterface implements Discovery {

    private DiscoveryService impl;

    /**
     * returns an interface object that permits to use this
     * service without having access to the real object.
     *
     * Since THIS is already such an object, it returns itself.
     * FIXME: it is kind of absurd to have this method part of the
     * interface but we do not want to define two levels of Service interface
     * just for that.
     * @return Resolver An interface object that implements
     * this service and nothing more.
     */

    public Service getInterface() {
	return this;
    }

    /**
     * Returns the advertisment for that service.
     *
     * @return Advertisement the advertisement.
     *
     * @version $Revision: 1.11 $
     * @since JXTA 1.0
     */
    public Advertisement getAdvertisement() {
        return impl.getAdvertisement();
    };

  /**
   * Only authorized constructor
   */

    public DiscoveryInterface(DiscoveryService theRealThing) {
	impl = theRealThing;
    }

    /**
     *  Initialize the application
     *  FIXME: This is meaningless for the interface object;
     * it is there only to satisfy the requirements of the
     * interface that we implement. Ultimately, the API should define
     * two levels of interfaces: one for the real service implementation
     * and one for the interface object. Right now it feels a bit heavy
     * to so that since the only different between the two would be
     * init() and may-be getName().
     * @param group PeerGroup this application is started from
     * @since   JXTA 1.0
     */

    public void init(PeerGroup pg, Advertisement a) {
    }

    /**
     * This is here for temporary class hierarchy reasons.
     * it is ALWAYS ignored. By definition, the interface object
     * protects the real object's start/stop methods from being called
     *
     * @param arg A table of strings arguments.
     * @return int status indication.
     */
    public int startApp(String[] arg) {
        return 0;
    }

    /**
     * This is here for temporary class hierarchy reasons.
     * it is ALWAYS ignored. By definition, the interface object
     * protects the real object's start/stop methods from being called
     *
     * This request is currently ignored.
     */
    public void stopApp() {
    }

  /**
   * Discover peers within a region set by the endpoint
   *
   * @param address Address of a peer, or portal to connect to, if address is null
   * then getPeerAdv "discover" in the region of the group
   * @see net.jxta.endpoint.Address
   */

    public void getRemoteAdvertisements (String peer, int type,
              String attribute, String value, int threshold) {
     impl.getRemoteAdvertisements (peer, type, attribute,
                              value, threshold);
    }


   /**
    * Retrieve Stored  Advertisements
    *
    * @return Enumeration of stored advertisements/structured documents
    * @since JXTA 1.0
    */

    public Enumeration getLocalAdvertisements (int type, String attribute,
                 String value) throws IOException {
	return impl.getLocalAdvertisements (type, attribute, value);
    }

   /**
    * flush  cached peer Advertisements.
    *
    */

    public void flushAdvertisements (String id, int type)
       throws IOException {
	 impl.flushAdvertisements(id,type);
    }

   /**
    * Publish an advertisement
    * same as above, provided as a convenience
    * @param doc StructuredDocument to publish
    */

    public void publish (Advertisement adv, int type) throws IOException {
	impl.publish(adv, type);
    }

    public void remotePublish (Advertisement adv, int type){
    impl.remotePublish(adv, type);
    }

}
