/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: DiscoveryService.java,v 1.13 2001/06/18 20:07:53 akhil Exp $
 */

package net.jxta.impl.discovery;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URL;
import java.lang.reflect.InvocationTargetException;


import net.jxta.service.Service;
import net.jxta.discovery.Discovery;
import net.jxta.resolver.Resolver;
import net.jxta.resolver.QueryHandler;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.MimeMediaType;
import net.jxta.protocol.ResolverQueryMsg;
import net.jxta.protocol.ResolverResponseMsg;
import net.jxta.protocol.PeerGroupAdvertisement;
import net.jxta.protocol.ServiceAdvertisement;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.NoResponseException;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.exception.ResendQueryException;
import net.jxta.exception.DiscardQueryException;

import net.jxta.impl.peergroup.GenericPeerGroup;
import net.jxta.impl.peergroup.RefPeerGroup;
import net.jxta.impl.endpoint.Endpoint;
import net.jxta.impl.protocol.PeerGroupAdv;
import net.jxta.impl.protocol.DiscoveryQuery;
import net.jxta.impl.protocol.DiscoveryResponse;
import net.jxta.impl.protocol.ResolverQuery;
import net.jxta.impl.protocol.ResolverResponse;
import net.jxta.impl.cm.Cm;




/**
 * DiscoveryService Service provides a mechanism to discover peers within a "
 * Network Peer Group" DiscoveryService is limited to the Network Peer Group
 * region set. jxtaDiscovery also provides a way to connect directly
 * to another peer given a Address and request other peer advertisements, this
 * method is particularly useful in the case of a portal where new relationships
 * may be established, or in the case of connecting to a predetermined peer
 * (perhaps described in address book, or through an invitation)
 *
 * @version $Revision: 1.13 $
 * Tag  $Name:  $
 * @since 1.0
 */

public class DiscoveryService implements Discovery ,QueryHandler {


    protected Cm cm;
    private String[] dirname ={"Peers","Groups","Adv"};
    private Resolver resolver;
    private RefPeerGroup myGroup=null;
    private String localPeerId = null;
    private boolean debug = false;
    private static int qid=0;
    private Endpoint endpoint;
    private long seed=System.currentTimeMillis();
    private final static long multiplier = 0x5DEECE66DL;
    private final static long addend = 0xBL;
    private final static long mask = (1L << 48) - 1;
    private Vector pendingQueries = null;

    private ServiceAdvertisement advertisement = null;

    /**
     * Default constructor
     */
    public DiscoveryService() {

    }

    /**
     * Supply arguments and starts this service if it hadn't started by itself.
     *
     * Currently this service starts by itself and does not expect
     * arguments.
     *
     * @param arg A table of strings arguments.
     * @return int status indication.
     */
    public int startApp(String[] arg) {
        return 0;
    }

    /**
     * Ask this service to stop.
     *
     * This request is currently ignored.
     */
    public void stopApp() {
    }

    /**
     * Service objects are not manipulated directly to protect usage
     * of the service. A Service interface is returned to access the service
     * methods.
     *
     * @return Service public interface of the service
     *
     * @version $Revision: 1.13 $
     * @since JXTA 1.0
     */
    public Service getInterface() {
	  return new DiscoveryInterface(this);
    }

    /**
     * Returns the advertisement for that service.
     *
     * @return Advertisement the advertisement.
     *
     * @version $Revision: 1.13 $
     * @since JXTA 1.0
     */
    public Advertisement getAdvertisement() {
        return advertisement;
    }

    /**
     * init is called by PeerGroup to init the Service with endpoint etc.
     * @param pg Peer Group
     * @throws PeerGroupException throw any exception thrown during obtaining
     * Service
     */
    public void init(PeerGroup pg, Advertisement sadv)
        throws PeerGroupException
    {

     advertisement = (ServiceAdvertisement) sadv;
     myGroup     = (RefPeerGroup) pg;
     endpoint    = myGroup.getEndpoint();
     localPeerId = myGroup.getPeerID().toString();
     pendingQueries = new Vector();

     PeerGroupAdvertisement adv = pg.getAdvertisement();
     resolver = pg.getResolver();
     resolver.registerHandler (advertisement.getName(), this);

     debug  =  pg.getDebug();
     try {
        URL url = new URL(pg.getID().toString());
        String pgdir = Integer.toHexString((url.getHost()).hashCode());
        cm = new Cm(pgdir);
	    cm.createFolder(dirname[PEER]);
	    cm.createFolder(dirname[GROUP]);
	    cm.createFolder(dirname[ADV]);
     } catch (Exception e) {
	    debug("Error during creation of local store");
	    if (debug) e.printStackTrace();
	}

	// Publish the local Peer Advertisement
	adv = ((GenericPeerGroup) pg).getPeerAdvertisement();
	try {
	    debug ("publishing local advertisement");
	    publish (adv, PEER);
	} catch (Exception e) {
	    debug ("Error when publishing local peer advertisement: " + e);
	}

     AdvertisementCompactor compactor = new AdvertisementCompactor(myGroup);
     compactor.init();


    }

    /**
     * Discover peers within a region set by the endpoint
     *
     * @param address Address of a peer, or portal to connect to, if address is null
     * then getPeerAdv "discover" in the region of the group
     * @return enumeration of peer advertisements discovered, null if none
     * @see net.jxta.endpoint.Address
     */

    public void getRemoteAdvertisements (String peer, int type,
              String attribute, String value, int threshold) {

        DiscoveryQuery dquery  = new DiscoveryQuery(type,
                          advToString(myGroup.getAdvertisement()),
                          attribute,
                          value);

        if ((attribute==null) || (value==null)){
	     dquery.setAttr("null");
	     dquery.setValue("null");
        } else {
	     dquery.setAttr(attribute);
	     dquery.setValue(value);
        }
        ResolverQuery query     = new ResolverQuery(
                      advertisement.getName(),
                      "JXTACRED",
                      "Token",
                      localPeerId,
                      dquery.toString(),qid++);
        resolver.sendQuery(peer, query);

    }

    /**
     * Retrieve Stored  Advertisements
     * based on the type it returns an enum Advertisement for PEER,
     * and GROUP
     * returns an enum of Advertisement for ADV
     *
     * @return Enumeration of stored Advertisement
     * @since JXTA 1.0
     */

    public Enumeration getLocalAdvertisements (int type,
					       String attribute,
					       String value)
	                     throws IOException {

	if ( ( type >2) || (type <0) ) {
	    throw new IOException("Unknown Advertisement type");
	}

	if (attribute == null) {
	    attribute = "null";
	}

	if (value == null) {
	    value = "null";
	}

	return( search (type,
			 attribute,
			 value,
			 Integer.MAX_VALUE,
			 false).elements());

    }

    /**
     * flush  cached peer Advertisements.
     *
     */

     public void flushAdvertisements (String id, int type)
              throws IOException {


     if ( (type < ADV) &&  (id !=null) ) {
        URL url = new URL(id);
        id= url.getHost();
        debug("flushing adv "+id+" of type "+dirname[type]);
        cm.remove (dirname[type],id);
     } else {
	    debug("flushing advertisements of type "+dirname[type]);
	    String [] fn = cm.getFileNames (dirname[type]);
	    for (int i=0; i < fn.length; i++) {
		  cm.remove (dirname[type],fn[i]);
        }
	 }
    }

    private PeerGroupAdvertisement mergeGroupAdv (PeerGroupAdvertisement adv1,
						  PeerGroupAdvertisement adv2) {

	debug ("mergeGroupAdv starts");
	Hashtable sAdvs = null;

	sAdvs = adv1.getServiceAdvertisements ();
	if (sAdvs == null) {
	    // Nothing to merge. Return adv2
	    debug ("Nothing to merge [1]");
	    return adv2;
	}
	// Get the resolver ServiceAdvertisement
	ServiceAdvertisement r1 = null;
	try {
	    r1 = (ServiceAdvertisement) sAdvs.get ("jxta.service.resolver");
	    if (r1 == null) {
		// Nothing to merge. Return adv2
		debug ("Nothing to merge [2]");
		return adv2;
	    }
	} catch (Exception e) {
	    // Return adv2
	    debug ("Nothing to merge [3]");
	    return adv2;
	}


	sAdvs = adv2.getServiceAdvertisements ();
	if (sAdvs == null) {
	    // Nothing to merge. Return adv1
	    debug ("Nothing to merge [4]");
	    return adv1;
	}
	// Get the resolver ServiceAdvertisement
	ServiceAdvertisement r2 = null;
	try {
	    r2 = (ServiceAdvertisement) sAdvs.get ("jxta.service.resolver");
	    if (r2 == null) {
		// Nothing to merge. Return adv1
		debug ("Nothing to merge [5]");
		return adv1;
	    }
	} catch (Exception e) {
	    // Return adv1
	    debug ("Nothing to merge [6]");
	    return adv1;
	}

	// Merge the param section of the resolver service advertisement
	Vector p1 = null;
	Vector p2 = null;

	p1 = r1.getParams();
	p2 = r2.getParams();

	if ((p1 == null) || (p1.size() == 0)) {
	    debug ("Nothing to merge [7]");
	    return adv2;
	}

	if ((p2 == null) || (p2.size() == 0)) {
	    debug ("Nothing to merge [8]");
	    return adv1;
	}

	String s1 = null;
	String s2 = null;
	boolean found = false;
	Vector res = (Vector) p2.clone();

	for (int i = 0; i < p1.size(); ++i) {
	    try {
		found = false;
		s1 = (String) p1.elementAt (i);
		for (int j = 0; j < p2.size(); ++j) {
		    try {
			s2 = (String) p2.elementAt (j);
			debug (" ");
			if (s1.equals(s2)) {
			    found = true;
			    break;
			}
		    } catch (Exception e2) {
			continue;
		    }
		}
		if (!found) {
		    // Add it
		    debug ("Merging: " + s1);
		    res.addElement (s1);
		}
	    } catch (Exception e) {
		continue;
	    }
	}
	// Set adv2 with the new update service advertisement
	r2.setParams (res);
	// Replace this service advertisement
	sAdvs.put ("jxta.service.resolver", r2);
	// Replace this list of service advertisement
	adv2.setServiceAdvertisements (sAdvs);

	debug ("merging done");
	return adv2;
    }

    /**
     * Publish an advertisement
     * This provides a way to advertise Peer/Group Advertisement within
     * the current group.
     * same as above, provided as a convenience
     * @param doc StructuredDocument to publish
     */

    public void publish (Advertisement adv, int type)
	throws IOException {

        URL url=null;

        if (type == GROUP) {
	    url = new URL(((PeerGroupAdv)adv).getGid() );
        } else if (type == PEER) {
	    url = new URL(((PeerGroupAdv)adv).getPid() );
        }

        StructuredDocument doc;
        try {
	    doc = (StructuredDocument) adv.getDocument(
				      new MimeMediaType("text/xml"));
        } catch( IOException error ) {
            throw error;
        } catch( Exception everything ) {
	    throw new IOException
		("Advertisement couldnt be saved because of :"
		 + everything.toString() );
        }
	switch (type) {
	case PEER: {
	    debug("publishing "+url.getHost()+ " in " + dirname[type]);
	    cm.save (dirname[type], url.getHost(), doc );
	    break;
	}
	case GROUP: {
	    debug("publishing "+url.getHost()+ " in " + dirname[type]);
	    // Check if we already have a record for this group. If we
	    // do, both advertisements will be merged.
	    try {
		StructuredDocument old = null;
		old = cm.restore (dirname[type], url.getHost());

		PeerGroupAdvertisement oldAdv = null;
		oldAdv = (PeerGroupAdvertisement)
                      AdvertisementFactory.newAdvertisement (
                                       (StructuredTextDocument) old);

		oldAdv = mergeGroupAdv ((PeerGroupAdvertisement) adv, oldAdv);
		doc = (StructuredDocument)
                          oldAdv.getDocument (new MimeMediaType("text/xml"));

	    } catch (Exception e) {
		// No advertisement was found. No merge.
	    }
	    cm.save (dirname[type], url.getHost(), doc );
	    break;
	}
	default: {
	    String str =  cm.createTmpName(doc);
	    debug("publishing "+str+ " in " + dirname[type]);
	    cm.save (dirname[type], str, doc );
	}
	}
    }

    public void remotePublish (Advertisement adv, int type){

     Vector list= new Vector();
     Enumeration peers;

     // support groups, and adv
     // to remotely publish
     if (type == PEER) return;

      Vector advert = new Vector();
      DiscoveryResponse dresponse=null;
      advert.addElement(advToString((Advertisement)adv));
      try {
        debug("Remote publishing Group");
        dresponse = new DiscoveryResponse(1,
                    type,
                    advToString((Advertisement)myGroup.getAdvertisement()),
                    "null",
                    "null",
                    advert.elements());
      } catch (Exception e) {
         if (debug) e.printStackTrace();
      }
      ResolverResponse pushRes = new ResolverResponse(advertisement.getName(),
                          "JXTACRED",
                          "Token",
                           0,dresponse.toString());
      debug("remote publishing");
      resolver.sendResponse(null,pushRes);

    }


    // methods we must implement by implementing
    // Service ,QueryHandler


   /**
    * Returns the group to which this service is attached.
    * @return PeerGroup the group
    */

    public PeerGroup getGroup () {
        return((PeerGroup) myGroup);
    }

  /**
   * deal with incoming responses for discovery
   * @param response
   */

    public void pushResponse(ResolverResponseMsg response) {

	DiscoveryResponse res;

	try {
	    res =  new DiscoveryResponse (new ByteArrayInputStream(
                        (response.getResponse()).getBytes()));
	} catch (Exception e ){
	    // we don't understand this msg, let's skip it
	    if (debug) e.printStackTrace();
	    return;
	}

	debug("Got a "+dirname[res.getDiscoveryType()]+" response");
	try {
	    String aRes = (String)res.getPeerAdv();
	    InputStream is = new ByteArrayInputStream( (aRes).getBytes() );

	    PeerGroupAdvertisement adv = (PeerGroupAdv)
            AdvertisementFactory.newAdvertisement(
            new MimeMediaType( "text/xml" ), is);

	    if (adv != null){
		publish(adv,PEER);
		if (endpoint.addPeerRoute(adv.getPid()))
		    debug("Added endpoint route"+adv.getPid());
	    }
	} catch (Exception e) {
	    if (debug) e.printStackTrace();
	    return;
	}

	Advertisement adv;
	Enumeration enum = res.getResponses();
	String str;
	if (enum != null ) {
	    while (enum.hasMoreElements()) {
		try {
		    str = (String) enum.nextElement();
		    adv =(Advertisement)
			AdvertisementFactory.newAdvertisement
                          (new MimeMediaType ("text/xml"),
                           new ByteArrayInputStream
                           (str.getBytes()));

		    if (adv != null) {
			publish(adv,res.getDiscoveryType());
		    }
		} catch (Exception e) {}

	    }
	}

        // If we are a rendezvous, check that we don't have pending queries
        // that need to be answered.

        if (!myGroup.isRendezvous()) {
	    return;
	}
	PendingQuery query = null;
	query = removePendingQuery (res.getDiscoveryType(),
				    res.getQueryAttr(),
				    res.getQueryValue());

	if (query == null) {
	    // Nope... nothing to do
	    return;
	}

	Enumeration pending = query.getRequesters();
	Vector results = null;

	while (pending.hasMoreElements()) {
	    try {
		Requester pq = (Requester) pending.nextElement();

		results = search (query.type,
				  query.requestedTag,
				  query.requestedValue,
				  pq.threshold,
				  true);

		if ((results == null) || (results.size() == 0)) {
		    // Since we just got an answer, this is highly
		    // improbable.
		    debug ("pushResponse: no response to forward");
		    break;
		}

		// Send the response
		debug("Responding  to "+ dirname[query.type]+" Query");

		// lets set the response type


		DiscoveryResponse dresponse =
		    new DiscoveryResponse(results.size(),
					  query.type,
					  advToString(myGroup.getAdvertisement()),
					  query.requestedTag,
					  query.requestedValue,
					  results.elements());


		response = new ResolverResponse(advertisement.getName(),
						"JXTACRED",
						"Token",
						pq.id,
						dresponse.toString());

		resolver.sendResponse (pq.requestingPeer, response);

	    } catch (Exception e) {
		debug ("  failed with " + e);
		continue;
	    }
	}
    }


    // bytes is set to true if the caller wants wire format of the
    // advertisement, or set to false if caller wants Advertisement
    // objects.

    private Vector search (int type,
			   String attr,
			   String value,
			   int threshold,
			   boolean bytes) {

        FileInputStream is=null;
        Enumeration enum = null;
        Vector results = new Vector();

        if ( !attr.equals("null") || !value.equals("null") ) {

	    // This is a discovery query without any searching
	    // criteria.

	    enum = cm.search (dirname[type], attr, value);

	    if (enum == null) {
                // return an empty Vector
		return results;
	    }

	    int i =0;
	    while (enum.hasMoreElements()) {
		String fn = (String) enum.nextElement();
		try {
		    if (((type <ADV) && !(fn.equals(localPeerId)))) {

			if (bytes) {
			    results.addElement(
				 new String(cm.restoreBytes(dirname[type], fn)));
			} else {
			    is = cm.getInputStream(dirname[type], fn);
			    if (is ==null) continue;

			    Advertisement adv = (Advertisement)
				AdvertisementFactory.newAdvertisement(
                                      new MimeMediaType( "text/xml" ), is );
			    results.addElement(adv);
			}
			i++;
			if ( i > threshold) break;
		    } else {

			if (bytes) {
			    results.addElement(
			       new String(cm.restoreBytes(dirname[type], fn)));
			} else {
			    is = cm.getInputStream(dirname[type], fn);
			    if (is ==null) continue;

			    Advertisement adv = (Advertisement)
				AdvertisementFactory.newAdvertisement(
                                      new MimeMediaType( "text/xml" ), is );
			    results.addElement(adv);
			}

			i++;
		    }
		} catch(Exception e) {
		   if (debug) e.printStackTrace();
		}
		try {
		    if (is != null) {
			is.close();
		    }
		} catch (Exception e) {
		    if (debug) e.printStackTrace();
		}
            }

	} else  {
	    // remote does not care about narrowing down the response
            String [] fn= cm.getFileNames(dirname[type]);
	    if (fn.length > 0) {

		if ( type <ADV) {
		    for (int i=0; i < fn.length; i++){
			try {
			    if ( ((type <ADV) && !(fn.equals(localPeerId)))) {
				if (bytes) {
				    results.addElement(new String(cm.restoreBytes(dirname[type], fn[nextInt(fn.length)])));
				} else {
				    is = cm.getInputStream(dirname[type], fn[i]);
				    if (is ==null) continue;

				    Advertisement adv = (Advertisement)
				       AdvertisementFactory.newAdvertisement(
                                          new MimeMediaType( "text/xml" ), is );
				    results.addElement(adv);
				}
				if ( i > threshold)
				    break;
			    }
            } catch(InvocationTargetException ex){
                // proceed with zpping it
                debug("got a stale advertisement, flushing it from cache");
                if (debug) ex.printStackTrace();
                try {
                     cm.remove(dirname[type], fn[i]);
                } catch (Exception e) {}
			} catch(Exception e) {
			    if (debug) e.printStackTrace();
			}
			try {
			    if (is != null) {
				is.close();
			    }
			} catch (Exception e) {
			   if(debug) e.printStackTrace();
			}
		    }

		} else {  // ADV type

		    for (int i=0; i < fn.length; i++){
			try {
			    if (bytes) {
				results.addElement(new String(cm.restoreBytes(dirname[type], fn[i])));
			    } else {
				is = cm.getInputStream(dirname[type],
						       fn[i]);
				if (is ==null) continue;

				Advertisement adv = (Advertisement)
				       AdvertisementFactory.newAdvertisement(
                                          new MimeMediaType( "text/xml" ), is );
				results.addElement(adv);
			    }

			    if ( i > threshold) break;

			} catch(Exception e) {
			    if (debug) e.printStackTrace();
			}
			try {
			    if (is != null) {
				is.close();
			    }
			} catch (Exception e) {
			    if (debug) e.printStackTrace();
			}
		    }
		}
	    }
	}
	return results;
    }


   /**
    * Handler API method
    * @return ResolverResponseMsg response
    * @NoResponseException thrown when no response is generated
    * @param query GenericResolverMsg
    * @throws NoResponseException if no Responses were received
    */

    public ResolverResponseMsg processQuery(ResolverQueryMsg query)
	throws NoResponseException,
	       DiscardQueryException,
	       IOException,
	       ResendQueryException {


	debug ("processQuery starts");

	ResolverResponse  response;
        Enumeration enum = null;
	Vector results = null;
        DiscoveryQuery dq;

	try {
	    dq = new DiscoveryQuery(new ByteArrayInputStream
		((query.getQuery()).getBytes()));
	} catch (Exception e ) {
            debug( "Error in query : " + e );
            if( debug )
                e.printStackTrace();
	    throw new IOException ();
	}

	debug("Got a "+dirname[dq.getDiscoveryType()]+" query");

	PeerGroupAdvertisement padv=null;
	try {
	    padv= (PeerGroupAdvertisement) AdvertisementFactory.newAdvertisement(
            new MimeMediaType( "text/xml" ),
            new ByteArrayInputStream(((String)dq.getPeerAdv()).getBytes()));
	    if (!(padv.getPid()).equals(localPeerId)) {
		//publish others only
		publish(padv,PEER);
	    }
	    if (endpoint.addPeerRoute(padv.getPid())){
		debug("Added endpoint route"+padv.getPid());
	    }
	} catch (Exception e) {
     	    if (debug) e.printStackTrace();
	}

	results = search (dq.getDiscoveryType(),
			  dq.getAttr(),
			  dq.getValue(),
			  dq.getThreshold(),
			  true) ;


	if ((results == null) || (results.size()) == 0) {

	    // We do not have anything to share. If this peer
	    // is a rendez vous, issue a query on our own, and
	    // forward the answers when we get them

	    if (isPendingQuery (dq.getDiscoveryType(),
				dq.getAttr(),
				dq.getValue())) {

		// We already have a pending request.
		// Discard query.
		throw new DiscardQueryException();
	    }
	    // Check if we want to get the answer.
	    if (myGroup.isRendezvous()) {

		boolean resend = processNotFound (query.getSrc(),
						  dq.getDiscoveryType(),
						  query.getQueryId(),
						  dq.getThreshold(),
						  dq.getAttr(),
						  dq.getValue());
		if (resend) {
		    debug ("   resend query");
		    throw new ResendQueryException();
		} else {
		    debug ("   forward query");
		    throw new NoResponseException ();
		}
	    }
	}

	debug("Responding  to "+dirname[dq.getDiscoveryType()]+" Query");

	// lets set the response type


    if (results == null ) throw new NoResponseException();
	DiscoveryResponse dresponse =
	    new DiscoveryResponse(results.size(),
				  dq.getDiscoveryType(),
				  advToString(myGroup.getAdvertisement()),
				  dq.getAttr(),
				  dq.getValue(),
				  results.elements());


	if  ((dq.getDiscoveryType() > PEER) && results.isEmpty()){
	    // remote peer is only interested in something other than peer
	    //(group, or adv)
	    // don't answer unless it is peer discovery
	    throw new DiscardQueryException();
	}

	response = new ResolverResponse(advertisement.getName(),
					"JXTACRED",
					"Token",
					query.getQueryId(),
					dresponse.toString());

	return(response);
    }


    private String advToString( Advertisement adv ) {

        StringWriter    out = new StringWriter();
        MimeMediaType displayAs = new MimeMediaType( "text/xml" );

        try {
	    StructuredTextDocument doc = (StructuredTextDocument) adv.getDocument( displayAs );
	    doc.sendToWriter( out );
	    return out.toString();
        }
        catch( Exception all ) {
            return null;
        }
    }


    private int nextInt(int n) {
        if (n<=0)
            throw new IllegalArgumentException("n must be positive");

        if ((n & -n) == n)  // i.e., n is a power of 2
            return (int)((n * (long)next(31)) >> 31);

        int bits, val;
        do {
            bits = next(31);
            val = bits % n;
        } while(bits - val + (n-1) < 0);
        return val;
    }

    synchronized protected int next(int bits) {
        long nextseed = (seed * multiplier + addend) & mask;
        seed = nextseed;
        return (int)(nextseed >>> (48 - bits));
    }


    // If this peer is a rendez-vous, cache the query, and ask the ResolverService
    // to resolve that request as for ourselves.

    private boolean processNotFound (String requestingPeer,
				     int    type,
				     int    id,
				     int    threshold,
                                     String requestedTag,
				     String requestedValue) {

        boolean hasToResend = true;

        debug ("processNotFound for " + requestingPeer);
	    debug ("           tag= " + requestedTag);
	    debug ("         value= " + requestedValue);

        if (myGroup.isRendezvous()) {
            if (isPendingQuery (type, requestedTag, requestedValue)) {
                // There is already a pending query for that peer.
		// No need to resend the query (that has already been done.
                hasToResend = false;
            }
            // Add the current query to the pending queries
            addPendingQuery (new String (requestingPeer),
			     type,
			     id,
			     threshold,
			     new String (requestedTag),
			     new String (requestedValue));

            return hasToResend;

        } else {
            // We are not a RendezVous, just ignore.
            return false;
        }
    }

    protected class Requester {

	protected String requestingPeer = null;
	protected int    threshold = 0;
	protected int    id = 0;

	public Requester (String peer, int threshold, int id) {

	    this.requestingPeer = peer;
	    this.threshold = threshold;
	    this.id = id;
	}

	public boolean equals (Object obj) {

	    Requester req = (Requester) obj;
	    if (req.requestingPeer.equals (requestingPeer)) {
		return true;
	    } else {
		return false;
	    }
	}
    }

    protected class PendingQuery {

	protected Vector requesters     = null;
	protected int    type           = 0;
    protected String requestedTag   = null;
	protected String requestedValue = null;
    protected long   timeOfQuery    = 0;
	protected long   delay          = 0;
	private  long    InitialDelay   = 120000; // 2 minutes

        public PendingQuery (int    type,
			     String tag,
			     String value) {

	    this.type = type;
        this.requestedTag = tag;
	    this.requestedValue = value;
        this.timeOfQuery = System.currentTimeMillis();
	    this.delay = this.timeOfQuery + InitialDelay;
	    requesters = new Vector();
        }

        public boolean equals (Object obj) {

            PendingQuery tmp =  (PendingQuery) obj;
            return (tmp.requestedTag.equals (requestedTag) &&
		    tmp.requestedValue.equals (requestedValue));
        }

	public boolean isRequester (String peer) {

	    Requester req = null;
	    for (int i = 0; i < requesters.size(); ++i) {
		try {
		    req = (Requester) requesters.elementAt (i);
		    if (req.requestingPeer.equals (peer)) {
			return true;
		    }
		} catch (Exception e) {
		    continue;
		}
	    }
	    return false;
	}

	public Enumeration getRequesters() {
	    return requesters.elements();
	}
    }

    private synchronized void addPendingQuery (String requestingPeer,
					       int type,
					       int id,
					       int threshold,
                                               String requestedTag,
					       String requestedValue) {

	PendingQuery query = null;
	Requester req = null;

        if (isPendingQuery (type, requestedTag, requestedValue)) {
	    // We already have a pending query of that type

	    query = getPendingQuery (type, requestedTag, requestedValue);

	    if (query.isRequester (requestingPeer)) {
		// This requesting peer is already recorded.
		return;
	    }
	    // add the requesting peer.
	    req = new Requester (requestingPeer, threshold, id);
	    query.requesters.addElement (req);

            return;
        }

        debug ("addPendingQuery for " + requestingPeer);

        query = new PendingQuery (type,
				  requestedTag,
				  requestedValue);

	req = new Requester (requestingPeer, threshold, id);
	query.requesters.addElement (req);
	pendingQueries.addElement (query);
	return;
    }

    private synchronized PendingQuery getPendingQuery (int type,
						       String requestedTag,
						       String requestedValue) {

    debug ("getPendingQuery for:");
	debug ("         type= " + type);
	debug ("          tag= " + requestedTag);
	debug ("        value= " + requestedValue);

        PendingQuery query = null;

        for (int i = 0; i < pendingQueries.size(); ++i) {

            try {
                query = (PendingQuery) pendingQueries.elementAt (i);
                if ((query.requestedTag.equals (requestedTag)) &&
		    (query.requestedValue.equals (requestedValue)) &&
		    (query.type == type)) {

		    return query;
                }
            } catch (Exception e) {
                continue;
            }
        }
	return null;
    }

    private synchronized PendingQuery removePendingQuery (int type,
							  String requestedTag,
							  String requestedValue) {

    debug ("removePendingQuery for:");
	debug ("         type= " + type);
	debug ("          tag= " + requestedTag);
	debug ("        value= " + requestedValue);

        PendingQuery query = null;

        for (int i = 0; i < pendingQueries.size(); ++i) {

            try {
                query = (PendingQuery) pendingQueries.elementAt (i);
                if ((query.requestedTag.equals (requestedTag)) &&
		    (query.requestedValue.equals (requestedValue)) &&
		    (query.type == type)) {
                    // Remove that query
                    try {
                        pendingQueries.removeElementAt (i);
                    } catch (Exception e) {
			break;
                    }
		    break;
                }
            } catch (Exception e) {
                continue;
            }
        }

	return query;
    }


    private synchronized boolean isPendingQuery (int type,
						 String requestedTag,
						 String requestedValue) {

    PendingQuery query = null;
	long currentTime = System.currentTimeMillis();

	debug ("isPendingQuery ");
	debug ("         type= " + type);
	debug ("          tag= " + requestedTag);
	debug ("        value= " + requestedValue);

        for (int i = 0; i < pendingQueries.size(); ++i) {

            try {
                query = (PendingQuery) pendingQueries.elementAt (i);
                if (query.requestedTag.equals (requestedTag) &&
		    query.requestedValue.equals (requestedValue) &&
		    (query.type == type)) {

		    debug ("   found pending query");
                    if ( query.delay <= currentTime) {
                        // We allow a new request to be sent
                        // Reset delay
			debug ("   allow resend");
                        query.delay = currentTime + query.InitialDelay;
                        return false;
                    }
		    debug ("   disallow resend");
                    return true;
                }
            } catch (Exception e) {
                return false;
            }
        }
        return false;
    }

    private void debug (String str) {
	if (debug) {
	    System.out.println ("[DiscoveryService] " + str);
	}
    }
}
