/*
 * $Id: LiteXMLDocument.java,v 1.17 2001/06/10 01:34:22 bondolo Exp $
 ********************
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 ********************
 */

package net.jxta.impl.document;

import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.Constructor;
import java.lang.String;
import java.util.Enumeration;
import java.util.Vector;

import java.io.IOException;

import net.jxta.document.MimeMediaType;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.TextElement;

/**
 * This class is an implementation of the StructuredDocument interface using
 * a simplified XML implementation
 *
 * @version     $Revision: 1.17 $
 * @since       JXTA 1.0
 */
public class LiteXMLDocument extends LiteXMLElement implements StructuredTextDocument {
    
    // "x-" is a mime-type convention for indicating partial or provisional compliance to a standard
    private static final MimeMediaType [] myTypes = {
        new MimeMediaType ( "Text", "Xml" ),
        new MimeMediaType ( "Text", "x-Xml" ),
        new MimeMediaType ( "Application", "Xml" ),
        new MimeMediaType ( "Application", "x-Xml" ),
    };
    
    protected String docContent = null;

    protected MimeMediaType mimeType = null;
    
    /**
     *  Returns the MIME Media types supported by this this Document per 
     *  {@link http://www.ietf.org/rfc/rfc2046.txt}. 
     *
     *  Jxta does not currently support the 'Multipart' or 'Message' media types.
     *
     *  @return An array of MimeMediaType objects containing the MIME Media Type
     *  for this Document.
     *
     */
    public static MimeMediaType [] getSupportedMimeTypes() {
        return( myTypes );
    }

    /** Creates new LiteXMLDocument */
    public LiteXMLDocument( MimeMediaType mimeType, String type) {
        // FIXME bondolo@jxta.org 20010227 This is not correct wrt charset encodings.
        this( mimeType, new ByteArrayInputStream( new String( "<?xml version=\"1.0\"?>\n\n<!DOCTYPE " + type + ">\n\n" + "<" + type + ">\n" + "</" + type + ">\n" ).getBytes()) );
    }
    
    /** Creates new LiteXMLDocument */
    public LiteXMLDocument( MimeMediaType mimeType, InputStream in ) {
        this( mimeType, new InputStreamReader( in ) );
    }
    
    /** Creates new LiteXMLDocument */
    public LiteXMLDocument( MimeMediaType mimeType, Reader in ) {
        super( null, (tagRange) null );
        loc = new tagRange();
        doc = this;
        parent = this;
        
        this.mimeType = mimeType;

        StringBuffer theDoc = new StringBuffer();
        
        char [] smallBuffer = new char[512];
        
        do {
            try {
                int readCount = in.read(smallBuffer);
                if( readCount > 0 )
                    theDoc.append( smallBuffer, 0, readCount );
                if( readCount < 0 )
                    break;
            }
            catch( IOException ignored ) {
                break;
            }
        }
        
        while( true );
        
        docContent = new String( theDoc );
        theDoc = null;                      // no longer needed.
        
        // startTag will contain the xml declaration
        loc.startTag.start = 0;
        loc.startTag.end = docContent.indexOf( '>' );
        
        // body is everything after the xml declaration
        loc.body.start = loc.startTag.end + 1;
        loc.body.end = docContent.length() - 1;
        
        // end is the end of the doc.
        loc.endTag.start = loc.body.end;
        loc.endTag.end = loc.body.end;
        
        charRange docType = getDocType( doc.docContent, false );
        
        if( docType.isValid() ) {
            loc = getTagRanges( doc.docContent, docContent.substring( docType.start, docType.end + 1 ), loc.body );
        } else{
            loc = getTagRanges( doc.docContent, null, loc.body );
        }
        
        if( !loc.isValid() )
            throw new RuntimeException( "Parsing error in source document. could not find end of document" );
        
        addChildTags( loc.body, this );         // now add the subtags
    }
    
    /**
     * get Mime Type
     */
    public MimeMediaType getMimeType() {
        return mimeType;
    }
    
    
    /**
     * create a new element without value
     */
    public net.jxta.document.Element createElement( Object key ) {
        return createElement( key, null );
    }
    
    /**
     * create a new element with value
     */
    public net.jxta.document.Element createElement( Object key, Object val) {
        if( !String.class.isAssignableFrom( key.getClass() ) )
            throw new ClassCastException( key.getClass().getName() + " not supported by createElement." );
        
        if( (null != val) && !String.class.isAssignableFrom( val.getClass() ) )
            throw new ClassCastException( val.getClass().getName() + " not supported by createElement." );
        
        return (net.jxta.document.Element) createElement( (String) key, (String) val );
    }
    
    /**
     * create a new element without value
     */
    public TextElement createElement( String name ) {
        return (TextElement) new LiteXMLElement( this, name );
    }
    
    /**
     * create a new element with value
     */
    public TextElement createElement(String name, String val) {
        return (TextElement) new LiteXMLElement( this, name, val );
    }
    
    /**
     * create a new element with value
     */
    protected TextElement createElement( tagRange loc ) {
        return (TextElement) new LiteXMLElement( this, loc );
    }
    
    /**
     *  Gets a reader for the document content.
     **/
    public Reader getReader() {
        StringWriter  stringOut = new StringWriter();
        
        stringOut.write( "<?xml version=\"1.0\"?>\n\n" );
        
        charRange result = getDocType( docContent, true );
        
        if( result.isValid() )
            stringOut.write( docContent.substring( result.start, result.end + 1 ) + "\n\n" );
        
        printNice( stringOut, 0, true );
        
        return new StringReader( stringOut.toString() );
    }
    
    /**
     *  Gets an inputStream for the document content.
     **/
    public InputStream getStream() {
        StringWriter  stringOut = new StringWriter();
        
        stringOut.write( "<?xml version=\"1.0\"?>\n\n" );
        
        charRange result = getDocType( docContent, true );
        
        if( result.isValid() )
            stringOut.write( docContent.substring( result.start, result.end + 1 ) + "\n\n" );
        
        printNice( stringOut, 0, true );
        
        return new ByteArrayInputStream( stringOut.toString().getBytes() );
    }
    
    /**
     *  Sends the document content to a writer.
     **/
    public void sendToWriter( Writer stream ) throws IOException {
        stream.write( "<?xml version=\"1.0\"?>\n\n" );
        
        charRange result = getDocType( docContent, true );
        
        if( result.isValid() )
            stream.write( docContent.substring( result.start, result.end + 1 ) + "\n\n" );
        
        printNice( stream, 0, true );
    }
    
    /**
     *  Sends the document content to a stream.
     **/
    public void sendToStream( OutputStream stream ) throws IOException {
        Writer out = new BufferedWriter( new OutputStreamWriter( stream ) );
        
        sendToWriter( out );
        out.flush();
    }
    
    protected charRange getDocType( String source, boolean wholeElement ) {
        final String xmldoctype = "!DOCTYPE";
        charRange result = new charRange();
        int start = 0;
        int end = doc.docContent.length() - 1;
        tagRange ranges = getTagRanges( source, xmldoctype, new charRange( start, end ) );
        
        if( -1 == start )
            return result;
        
        if( !ranges.startTag.isValid() )
            return result;
        
        if( wholeElement ) {
            result = ranges.startTag;
        }
        else {
            result.start = ranges.startTag.start + 1 + xmldoctype.length() - 1 + 1;
            
            while( (result.start < end) &&           // immediately followed by a delimiter or the end of the tag
            ( -1 != " \t\n\r".indexOf( source.charAt(result.start) ) ) )
                result.start++;
            
            result.end = result.start;
            
            while( ((result.end + 1) < end ) && // immediately followed by a delimiter or the end of the tag
            ( -1 == " \t\n\r/>".indexOf( source.charAt(result.end + 1) ) ) )
                result.end++;
        }
        return result;
    }
}
