/*
 * $Id: LiteXMLElement.java,v 1.19 2001/06/14 17:25:09 bondolo Exp $
 ********************
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 ********************
 */

package net.jxta.impl.document;

import java.io.Writer;
import java.lang.String;
import java.util.Enumeration;
import java.util.Vector;

import java.io.IOException;
import java.lang.ClassCastException;

import net.jxta.document.Element;
import net.jxta.document.StructuredDocument;
import net.jxta.document.TextElement;

/** An element of a <CODE>StructuredDocument</CODE>. <CODE>StructuredDocument</CODE>s
 * are made up of hierarchies of elements. LiteXMLElement is part of an implementation
 * while makes use of XML-style document conventions, but without the overhead of a
 * full parser.
 *
 * @since Jxta 0.1
 * @Version $Revision: 1.19 $
 */
public class LiteXMLElement extends TextElementCommon {
    
    /**
     * A charRange defines a range of characters, probably within a string. The
     * range is deemed to be invalid if 'start' is -1.  A zero length range is,
     * by convention, described by an 'end' value of 'start' - 1.
     */
    protected class charRange {
        
        /**
         *  Contains the start position of this range.
         **/
        public int start;
        
        /**
         * Contains the end position of this range. one weird thing: if end == start -1,
         * then the item is of zero length begining at start.
         **/
        public int end;
        
        /**
         * Constructor for a null charRange.
         **/
        public charRange() {
            start = -1;
            end = -1;
        }
        
        /**
         * Constructor for which the bounds are specified.
         *
         **/
        public charRange( int start, int end ) {
            this.start = start;
            this.end = end;
        }
        
        /**
         * Returns the length of this range.
         * @return The length of the range or -1 if the range is null.
         **/
        public int length() {
            if( (-1 == start) || (-1 == end) )
                return -1;
            
            return ( end - start + 1 );
        }
        
        /**
         * Returns true if the range is both non-null and has a length of greater
         * than or equal to zero.
         *
         * @return true if the range is a valid one, otherwise false.
         *
         **/
        public boolean isValid() {
            return length() >= 0;
        }
        
        /**
         * Compares two ranges for equality.
         *
         * @param someRange The range against which this range will be compared.
         * @return true if the two ranges are equal, otherwise false.
         *
         */
        public boolean equals( Object aRange ) {
            if (this == aRange)
                return true;
            
            if( !(aRange instanceof charRange) )
                return false;
            
            charRange someRange = (charRange) aRange;
            
            return( (start == someRange.start) &&
            (end == someRange.end) );
        }
        
        /**
         * Compares two ranges for equality.
         *
         * @param someRange The range against which this range will be compared.
         * @return true if the two ranges are equal, otherwise false.
         *
         */
        public int compareTo( Object aRange ) throws ClassCastException {
            if (this == aRange)
                return 0;
            
            if( !(aRange instanceof charRange) )
                throw new ClassCastException( "type mismatch error" );
            
            charRange someRange = (charRange) aRange;
            
            if( start < someRange.start )
                return -1;
            
            if( start > someRange.start )
                return 1;
            
            if( end < someRange.end )
                return -1;
            
            if( end > someRange.end )
                return 1;
            
            return 0;
        }
        
        /**
         * Returns true if the <CODE>charRange</CODE> specified by someRange is contained within this range.
         *
         *
         * @param someRange The range which must be contained within this range.
         *
         * @return true if the specified range is contained with this range otherwise false.
         *
         */
        public boolean contains( charRange someRange ) {
            return( isValid() &&
            someRange.isValid() &&
            ( start <= someRange.start ) &&
            ( end >= someRange.end ) );
        }
        
        /**
         *  Returns true if the location specified is contained in this range.
         *
         *  @param someLoc the location which is to be tested.
         *  @return true if the location is in this range, otherwise false.
         */
        public boolean contains( int someLoc ) {
            return( isValid() &&
            ( someLoc >= 0 ) &&
            ( start <= someLoc ) &&
            ( end >= someLoc ) );
        }
    };
    
    /**
     * A tagRange is a collection of char ranges useful for describing XML
     * structures. 'startTag' is the range of the opening tag, ie. <tag>
     * 'endTag' is the range of the terminating tag, ie. </tag>. The 'body'
     * range everything between the start and end tags. For empty-element tags
     * the 'startTag' and 'endTag' will be equal and the 'body' tag will be
     * NULL.
     **/
    protected class tagRange {
        public charRange startTag = null;
        public charRange body = null;
        public charRange endTag = null;
        
        public tagRange( ) {
            startTag = new charRange();
            body = new charRange();
            endTag = new charRange();
        }
        
        public tagRange( charRange startTag, charRange body, charRange endTag ) {
            this.startTag = startTag;
            this.body = body;
            this.endTag = endTag;
        }
        
        public boolean isValid() {
            return (null != startTag) && (null != body) && (null != endTag) &&
            startTag.isValid() && body.isValid() && endTag.isValid();
        }
        
        public boolean equals( Object aRange ) {
            if (this == aRange)
                return true;
            
            if( !(aRange instanceof tagRange) )
                return false;
            
            tagRange someRange = (tagRange) aRange;
            
            return( startTag.equals(someRange.startTag) &&
            body.equals(someRange.body) &&
            endTag.equals(someRange.endTag) );
        }
        
        public boolean contains( tagRange someRange ) {
            return( isValid() &&
            someRange.isValid() &&
            ( body.start <= someRange.startTag.start ) &&
            ( body.end >= someRange.endTag.end ) );
        }
    };
    
    /**
     * The document associated with this Element.
     */
    protected LiteXMLDocument doc;
    
    /**
     *  Identifies the element which is the parent of this element. If <code>
     *  this.parent == this</code> then this element is the root of the document.
     *  If <code>null == parent</code> then this element has not yet been
     *  inserted into the document.
     */
    protected Element parent;
    
    /**
     *  The portion of the source XML associated with this node
     */
    protected tagRange loc;
    
    /**
     *  If this node has yet to be inserted into the document then will contain
     *  the String value of this node, otherwise null.
     */
    private StringBuffer  uninserted = null;
    
    /**
     * The child elements associated with this element
     */
    private Vector children = new Vector();
    
    /**
     * Creates new LiteElement
     *
     * @param doc The {@link LiteXMLDocument} which is associated with this element.
     * @param name The name of the element being created.
     */
    protected LiteXMLElement(LiteXMLDocument doc, final String name) {
        this( doc, name, null );
    }
    
    /**
     * Creates new LiteElement
     * @param doc The {@link LiteXMLDocument} which is associated with this element.
     * @param name The name of the element being created.
     * @param val The value of the element being created.
     */
    protected LiteXMLElement(LiteXMLDocument doc, final String name, final String val) {
        this( doc, (tagRange) null );
        loc = new tagRange();
        
        if( null == val )
            uninserted = new StringBuffer( );
        else {
            uninserted = new StringBuffer( val );
            encodeEscaped( uninserted );
        }
        
        uninserted.insert( 0, "<" + name + ">" );
        
        uninserted.append( "</" + name + ">" );
    }
    
    /**
     * Creates new LiteXMLElement
     * @param loc The location of the element within the document.
     * @param doc The {@link LiteXMLDocument} which is associated with this element.
     */
    protected LiteXMLElement(LiteXMLDocument doc, tagRange loc ) {
        this.doc = doc;
        this.loc = loc;
    }
    
    /**
     * Tests two elements for equality.
     *
     * @param element the element to be compared against.
     * @return true if the elements are equal (by whatever definition of
     * equality is appropriate) otherwise false
     */
    public boolean equals( Object element ) {
        if (this == element)
            return true;
        
        if( !(element instanceof LiteXMLElement) )
            return false;
        
        LiteXMLElement liteElement = (LiteXMLElement) element;
        
        if( doc != liteElement.doc )
            return false;
        
        if( !getName().equals( liteElement.getName() ) )
            return false;
        
        String val1;
        if( null != uninserted )
            val1 = uninserted.toString();
        else
            val1 = getTextValue( );
        
        String val2 = liteElement.getTextValue( );
        
        if( (null == val1) && (null == val2) )
            return true;
        
        if( (null == val1) || (null == val2) )
            return false;
        
        return val1.equals( val2 );
    }
    
    /**
     *  Get the root element of the hierarchy this element belongs to.
     *
     *  @return StructuredDocument root of this element's hierarchy.
     *
     *  @since JXTA 1.0 
     */
    public StructuredDocument getRoot() {
        return doc;
    }

    /**
     *  Get the parent of this element. If the element has not been inserted into
     *  the Document then null is returned. If this element is the root of the
     *  Document then it returns itself.
     *
     *  @return Element parent of this element
     *
     *  @since JXTA 1.0
     */
    public Element getParent() {
        return parent;
    }
    
    /**
     * Returns an enumeration of the immediate children of this element
     *
     * @return Enumeration An enumeration containing all of the children of this element.
     *
     * @since JXTA 1.0
     */
    public Enumeration getChildren() {
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        return children.elements();
    }
    
    /**
     * Returns the name associated with this element.
     *
     * @return the name associated with this Element
     */
    public String getName() {
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        
        int current = loc.startTag.start + 1;
        
        while( current < loc.startTag.end )
            if( -1 != "\n\r\t />".indexOf( doc.docContent.charAt(current) ) )
                break;
            else
                current++;
        
        return doc.docContent.substring( loc.startTag.start + 1, current );
    }
    
    /**
     * Get the value (if any) associated with an element.
     *
     * @return A string containing the value of this element, if any, otherwise null.
     */
    public String getTextValue() {
        return getTextValue( false );
    }
    
    /**
     * Add a child element to this element
     *
     * @param element the element to be added as a child
     */
    public void appendChild( TextElement element ) {
        // FIXME  20010218    bondolo@jxta.org This section needs to be rewritten for empty
        // element declarations. ie. <element/> changed to <element>new</element>
        
        if( !LiteXMLElement.class.isInstance( element ) )
            throw new IllegalArgumentException( "element type not supported." );
        
        LiteXMLElement newElement = (LiteXMLElement) element;
        
        if( newElement.doc != doc )
            throw new IllegalArgumentException( "Wrong document" );
        
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        
        if( null != newElement.uninserted ) {
            doc.docContent = doc.docContent.substring( 0, loc.endTag.start ) + newElement.uninserted + doc.docContent.substring( loc.endTag.start );
            
            newElement.loc.startTag.start = loc.endTag.start;
            newElement.loc.startTag.end = doc.docContent.indexOf( '>', newElement.loc.startTag.start );
            
            newElement.loc.body.start =  newElement.loc.startTag.end + 1;
            
            newElement.loc.endTag.end = newElement.loc.startTag.start + newElement.uninserted.length() - 1;
            newElement.loc.endTag.start = doc.docContent.lastIndexOf( '<', newElement.loc.endTag.end );
            
            newElement.loc.body.end = newElement.loc.endTag.start - 1;
            
            if( 0 != loc.body.length() ) {
                doc.adjustLocations( loc.endTag.start, newElement.uninserted.length() );
            }
            else {
                loc.body.start--;
                doc.adjustLocations( loc.endTag.start, newElement.uninserted.length() );
                loc.body.start++;
            }
            
            loc.body.end += newElement.uninserted.length();
            
            newElement.uninserted = null;
        }
        
        if( !loc.contains(newElement.loc) )
            throw new RuntimeException( "Element is not contained by parent" );
        
        newElement.parent = this;
        children.addElement( newElement );
    }
    
    /**
     * Returns an enumeration of the immediate children of this element whose
     * name match the specified string.
     *
     * @param name The name which will be matched against.
     * @return An enumeration containing all of the children of this element.
     */
    public Enumeration getChildren( String name ) {
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        
        Vector result = new Vector();
        for( Enumeration eachChild = children.elements(); eachChild.hasMoreElements(); ) {
            TextElement aChild = (TextElement) eachChild.nextElement();
            
            if( name.equals( aChild.getName() ) )
                result.addElement( aChild );
        }
        
        return result.elements();
    }
    
    /**
     *  Get the value (if any) associated with an element.
     *
     *  @param getEncoded if true then the contents will be encoded such that
     *      the contents will not be interpreted as XML. see
     *      {@link http://www.w3.org/TR/REC-xml#syntax}
     *      ie. < -> &lt; & -> &amp;
     *  @return A string containing the value of this element, if any, otherwise null.
     */
    protected String getTextValue( boolean getEncoded ) {
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        
        if( !loc.isValid() )
            throw new RuntimeException( "Corrupted Node" );
        
        StringBuffer building = new StringBuffer();
        
        Vector ranges = new Vector();
        
         /*
          * insert the ranges of the children in order. insertion method is ok
          * because the number of childre is usually less than 10 or so.
          */
        for( Enumeration eachChild = getChildren(); eachChild.hasMoreElements(); ) {
            LiteXMLElement aChild = (LiteXMLElement) eachChild.nextElement();
            charRange childsRange =
            new charRange( aChild.loc.startTag.start, aChild.loc.endTag.end );
            
            // find where to insert.
            for( int eachRange = 0; eachRange < ranges.size(); eachRange++ ) {
                charRange rangeChild = (charRange) ranges.elementAt( eachRange );
                if( 1 == rangeChild.compareTo( childsRange ) ) {
                    ranges.setElementAt( childsRange, eachRange );
                    childsRange = rangeChild;
                }
            }
            ranges.addElement( childsRange );
        }
        
        int current = loc.body.start;
        Enumeration eachRange = ranges.elements();
        
        // add all the text not part of some child
        while( eachRange.hasMoreElements() ) {
            charRange aRange = (charRange) eachRange.nextElement();
            
            building.append( doc.docContent.substring(current, aRange.start) );
            
            current = aRange.end + 1;
        }
        
        // Add the last bit.
        building.append( doc.docContent.substring( current, loc.endTag.start ) );
        
        if( !getEncoded )
            building = decodeEscaped( building );
        
        // trim (this is what String.trim() claims it does)
        String result = building.toString().trim();
        
        if( 0 == result.length() )
            return null;
        
        return result;
    }
    
    /**
     *  Write the contents of this element and optionally its children. The
     *  writing is done to a provided java.io.Writer. The writing can optionally
     *  be indented
     *
     *  @param into The java.io.Writer that the output will be sent to.
     *  @param indent   the number of tabs which will be inserted before each
     *      line.
     *  @param recurse  if true then also print the children of this element.
     *
     **/
    protected void printNice( Writer into, int indent, boolean recurse ) {
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        
        try {
            // do indent
            for( int eachTab = 0; eachTab < indent; eachTab++ )
                into.write( "\t" );
            
            // print node name
            into.write( doc.docContent.substring(loc.startTag.start, loc.startTag.end + 1) + "\n" );
            
            String itsValue = getTextValue( true );
            // print node value
            if ( null != itsValue ) {
                // do indent
                for( int eachTab = 0; eachTab < indent + 1; eachTab++ )
                    into.write( "\t" );
                
                into.write( itsValue + "\n" );
            }
            
            // recurse as needed
            if ( recurse )
                for( Enumeration childrens = getChildren(); childrens.hasMoreElements(); )
                    ((LiteXMLElement)childrens.nextElement()).printNice( into, indent + 1, recurse );
            
            // do indent
            for( int eachTab = 0; eachTab < indent; eachTab++ )
                into.write( "\t" );
            
            // print node name
            into.write( doc.docContent.substring(loc.endTag.start, loc.endTag.end + 1) + "\n" );
        }
        catch( IOException ignored ) {
            // we ignore any io errors
        }
    }
    
    /**
     * Given a source string, an optional tag and a range with in the source
     * find either the tag specified or the next tag.
     *
     * The search consists of 4 phases :
     *    0.  If no tag was specified, determine if a tag can be found and
     *        learn its name.
     *    1.  Search for the start of the named tag.
     *    2.  Search for the end tag. Each time we think we have found a tag
     *        which might be the end tag we make sure it is not the end tag
     *        of another element with the same name as our tag.
     *    3.  Calculate the position of the body of the tag given the locations
     *        of the start and end.
     *
     * @param source the string to search
     * @param tag the tag to search for in the source string. If this tag is
     *    empty or null then we will search for the next tag.
     * @param range describes the range of character locations in the source
     *    string to which the search will be limited.
     * @return tagRange containing the ranges of the found tag.
     */
    protected tagRange getTagRanges( final String source, String tag, final charRange range ) {
        // FIXME   bondolo@jxta.org 20010327    Does not handle XML comments. ie.  <!-- -->
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added to the document." );
        
        tagRange result = new LiteXMLElement.tagRange();
        int start = range.start;
        int end = source.length() - 1;
        int current;
        boolean foundStartTag = false;
        boolean foundEndTag = false;
        boolean emptyTag = ( (null == tag) || ("".equals(tag)) );
        
        //System.err.println( "Tag : " + tag + " Start : " + start + " End: " + end + " a bit : " + source.substring( start, start + 11 ) );
        
        // check for bogosity
        if( (-1 == start) || (start >= end) )
            return result;
        
        // adjust end of range
        if( (-1 != range.end) && (end > range.end) )
            end = range.end;
        
        // check for empty tag and assign empty string
        if( null == tag )
            tag = "";
        
        current = start;
        
        // Begin Phase 0 : Search for any tag.
        
        if( emptyTag ) {
            int foundTagText = source.indexOf( '<', current );
            int afterTagText;
            
            // was it not found? if not then quit
            if( -1 == foundTagText )
                return result;
            
            // this part is about setting the tag if necessary
            foundTagText++;
            
            afterTagText = foundTagText;
            while( afterTagText <= end ) {
                if( -1 == " \t\n\r/>".indexOf( source.charAt(afterTagText) ) ) {
                    afterTagText++;
                    continue;
                }
                tag = source.substring( foundTagText, afterTagText );
                emptyTag = ( (null == tag) || ("".equals(tag)) );
                
                break;
            }
            
            // it better not be still empty
            if( emptyTag )
                return result;
        };
        
        // Begin Phase 1: Search for the Start Tag
        
        //start looking for the start tag
        while( !foundStartTag && (current < end) ) {
            int foundTagText = source.indexOf( tag, current + 1); // first loc is one past current location
            int foundTagTerminator;
            int foundNextTagStart;
            int afterTagText = foundTagText + tag.length();
            
            // was it not found
            if( ( -1 == foundTagText ) || (afterTagText > end) )
                return result;
            
            // check to see if it is the start tag
            if( ('<' != source.charAt( foundTagText - 1)) || // it has the open tag delimiter before it
            ( -1 == " \t\n\r/>".indexOf( source.charAt(afterTagText) ) ) ) { // is immediately followed by a delimiter
                current = afterTagText;
                continue;
            }
            
            foundTagTerminator = source.indexOf( '>', afterTagText );
            foundNextTagStart = source.indexOf( '<', afterTagText + 1 );
            
            if( (-1 == foundTagTerminator) || // the tag has no terminator
            (foundTagTerminator > end) || //  it is past the valid range
            ( (-1 != foundNextTagStart) &&    // there is another tag start
            (foundNextTagStart < foundTagTerminator) ) ) { // and it is before the terminator we found. very bad
                current = afterTagText;
                continue;
            }
            
            foundStartTag = true;
            result.startTag.start = foundTagText - 1;
            result.startTag.end = foundTagTerminator;
        }
        
        if( !foundStartTag )
            return result;
        
        // is this an empty element declaration?
        if( '/' == source.charAt( result.startTag.end - 1 ) ) {
            result.endTag = result.startTag;        // end is the start and there is no body
            return result;
        }
        
        current =  result.startTag.end + 1;
        tagRange subElement = null;
        
        // if current is past the end then our end tag is not found.
        if( current >= end )
            return result;
        
        // Begin Phase 2 :  Search for the end tag
        
        String endTag = "</" + tag + ">";
        
        while( !foundEndTag && (current < end) ) {
            int foundTagText = source.indexOf( endTag, current );
            
            // was it not found or not in bounds?
            if( (-1 == foundTagText) || ((foundTagText + endTag.length() - 1) > end) )
                break; // it was not found
            
            // We recurse here in order to exclude the end tags of any sub elements with the same name
            
            subElement = getTagRanges( source, tag, new charRange( current,  foundTagText - 1 ) );
            
            // if there was a sub-tag with the same name, skip past it
            if( subElement.startTag.isValid() ) {
           /*      System.err.println( "found a sub-tag with our name (" + tag +
                 ") at : (" + subElement.startTag.start + "," + subElement.endTag.end + ") in range (" +
                 current + "," + (foundTagText - 1) + ")" ); */
                current = foundTagText + endTag.length();
                continue;
            }
            
            foundEndTag = true;
            result.endTag.start = foundTagText;
            result.endTag.end = foundTagText + endTag.length() - 1;
        }
        
        // Begin Phase 3 :  Calculate the location of the body.
        
        result.body.start = result.startTag.end + 1;
        
        if( foundEndTag ) {
            result.body.end = result.endTag.start - 1;
        }
        else {
            result.body.end = end;
        }
        
        return result;
    };
    
    /**
     *  Parse a charRange and add any tags found as content as children of a
     *  specified element. This process is repeated recursivly.
     *
     *  @param  scanRange   the range to be parsed for sub-tags
     *  @param  addTo       the element to add any discovered children to.
     *
     */
    protected void addChildTags( final charRange scanRange, LiteXMLElement addTo ) {
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        
        int current = scanRange.start;
        
        do {
            // scan for any tag.
            tagRange aSubtag = getTagRanges( doc.docContent, null, new charRange( current, scanRange.end ) );
            
            // did we find one?
            if( aSubtag.isValid() ) {
                LiteXMLElement newChild = (LiteXMLElement) doc.createElement( aSubtag );
                addTo.appendChild(newChild);
                addChildTags( aSubtag.body, newChild ); // recurse into the new tag
                
                // all done this tag, move on
                current = aSubtag.endTag.end + 1;
            }
            else
                current = -1; // all done!
        }
        while( (-1 != current) && (current < scanRange.end) );
    };
    
    /**
     *  For this element and all its children adjust the location of its ranges
     *  by the amount specified.
     *
     *  @param beginningAt  adjust all locations which are at or past this
     * location.
     *  @param  by  amount to adjust all matching locations.
     **/
    protected void adjustLocations( final int beginningAt, final int by ) {
        if( null != uninserted )
            throw new RuntimeException( "This element has not been added." );
        
        if( loc.startTag.start >= beginningAt )
            loc.startTag.start += by;
        
        if( loc.startTag.end >= beginningAt )
            loc.startTag.end += by;
        
        if( loc.body.start >= beginningAt )
            loc.body.start += by;
        
        if( loc.body.end >= beginningAt )
            loc.body.end += by;
        
        if( loc.endTag.start >= beginningAt )
            loc.endTag.start += by;
        
        if( loc.endTag.end >= beginningAt )
            loc.endTag.end += by;
        
        for( Enumeration eachChild = getChildren(); eachChild.hasMoreElements(); ) {
            LiteXMLElement aChild = (LiteXMLElement) eachChild.nextElement();
            
            aChild.adjustLocations( beginningAt, by );
        }
    };
    
    /**
     *  Given a StringBuffer find all occurances of escaped characters which
     *  must be decoded and convert them back to their non-escaped equivalents.
     *
     *  @param target The stringbuffer which will be decoded.
     *  @return The decoded version of the stringbuffer.
     *
     **/
    protected StringBuffer decodeEscaped( StringBuffer target ) {
        
        int current = 0;
        
        StringBuffer result = new StringBuffer();
        
        while( current < target.length() ) {
            // FIXME bondolo@jxta.org   20010422    Should process xml comments out here.
            if( '&' != target.charAt( current ) ) {
                result.append( target.charAt( current ) );
                current++;
                continue;
            }
            
            int terminusAt = current + 1;
            
            while(  (terminusAt < target.length()) && // dont go past end
            ((terminusAt - current) < 6 ) &&  // only look 6 chars away.
            (';' != target.charAt(terminusAt)) )   // must be a ;
                terminusAt++;
            
            if(';' != target.charAt(terminusAt)) {
                // if we get here then we didnt find the terminal we needed
                // so we just leave ampersand as it was, the document is
                // ill-formed but why make things worse?
                result.append( target.charAt( current ) );
                current++;
                continue;
            }
            
            char [] sub = new char[ terminusAt - current + 1 ];
            target.getChars(current, terminusAt + 1, sub, 0 );
            String escaped =  new String( sub );
            
            if( "&amp;".equals( escaped ) ) {
                result.append( '&' );
                current += 4;
            }
            else if ( "&#38;".equals( escaped ) ) {
                result.append( '<' );
                current += 3;
            }
            else if ( "&lt;".equals( escaped ) ) {
                result.append( '<' );
                current += 3;
            }
            else if ( "&#60;".equals( escaped ) ) {
                result.append( '<' );
                current += 3;
            }
            else if ( "&gt;".equals( escaped ) ) { // for compatibility with SGML. We dont encode these
                result.append( '>' );
                current += 3;
            }
            else {
                result.append( target.charAt( current ) );
                current++;
                continue;
            }
            
            current++;
        }
        
        return result;
    }
    
    /**
     *  Given a StringBuffer find all occurances of characters which must be
     *  escaped and convert them to their escaped equivalents.
     *
     *  @param target The stringbuffer which will be encoded in place.
     *
     **/
    protected void encodeEscaped( StringBuffer target ) {
        
        int current = 0;
        
        while( current < target.length() ) {
            if( '&' == target.charAt( current ) ) {
                target.insert( current+1, "amp;" );
                current += 5;
                continue;
            }
            else if( '<' == target.charAt( current ) ) {
                target.setCharAt( current, '&' );
                target.insert( current + 1, "lt;" );
                current += 4;
                continue;
            }
            else
                current++;
        }
    }
}
