/*
 * $Id: XMLDocument.java,v 1.13 2001/06/08 23:03:23 bondolo Exp $
 ********************
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 ********************
 */

package net.jxta.impl.document;

import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.String;
import java.util.Enumeration;
import java.util.Vector;

import java.io.IOException;
import java.util.NoSuchElementException;

import javax.xml.parsers.*;

import org.xml.sax.*;
import org.w3c.dom.*;
import org.w3c.dom.traversal.*;

import org.apache.xerces.dom.DocumentImpl;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.Serializer;
import org.apache.xml.serialize.XMLSerializer;

import net.jxta.document.MimeMediaType;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.TextElement;

/**
 * This class is an implementation of the StructuredDocument interface using
 * an XML parser
 * @version     $Revision: 1.13 $
 * @since       JXTA 1.0
 */
public class XMLDocument extends XMLElement implements StructuredTextDocument {
    
    /**
     *  The MIME Media Types which this <CODE>StructuredDocument</CODE> is
     *  capable of emitting.
     */
    private static final MimeMediaType [] myTypes = {
        new MimeMediaType( "Text", "Xml" ),
        new MimeMediaType( "Application", "Xml" )
    };
    
    private MimeMediaType mimeType;
    
    /**
     *  Returns the MIME Media types supported by this this Document per
     *  {@link http://www.ietf.org/rfc/rfc2046.txt}.
     *
     *  Jxta does not currently support the 'Multipart' or 'Message' media types.
     *
     *  @return An array of MimeMediaType objects containing the MIME Media Type
     *  for this Document.
     *
     */
    public static MimeMediaType [] getSupportedMimeTypes() {
        return( myTypes );
    }
    
    /**
     * Constructor for new instances of <CODE>XMLDocument</CODE>
     *
     * @param mimeType This is the MIME Media Type being requested. In general it should be equivalent with
     * the MIME Media Type returned by {@link getMimeType}. A <CODE>StructuredDocument</CODE>
     * sub-class may, however, support more than one MIME Media type so this may be a
     * true parameter. XMLDocument supports additional the MIME Media Type parameters,
     * "charset". The charset parameter must be a value per IETF RFC 2279 or ISO-10646.
     * @param docType Used as the root type of this document. {@link XMLDocument} uses this as the XML
     * <CODE>DOCTYPE</CODE>.
     * @throws Exception Exceptions generated by the underlying implementation.
     */
    public XMLDocument( final MimeMediaType mimeType, final String docType ) throws Exception {
        super( null, null );
        this.root = this;
        
        this.mimeType = (MimeMediaType) mimeType.clone();
        
        try {
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
            docBuilderFactory.setNamespaceAware(true);
            docBuilderFactory.setValidating(false);
            DocumentBuilder dataDocBuilder = docBuilderFactory.newDocumentBuilder();
            org.w3c.dom.Document tempdoc = dataDocBuilder.newDocument();
            
            org.w3c.dom.DOMImplementation domImpl = tempdoc.getImplementation();
            org.w3c.dom.DocumentType doctypeNode = domImpl.createDocumentType( docType, null, null );
            
            domNode = domImpl.createDocument( "http://www.jxta.org/jxta", docType, doctypeNode );
        }
        catch (Exception e) {
            //System.err.println( "Caught Exception in building Document. : " + e.toString() );
            // e.printStackTrace();
            domNode = null;
            throw e;
        }
    }
    
    /**
     * Constructor for existing documents.
     *
     * @param mimeType This is the MIME Media Type being requested. In general it should be equivalent with
     * the MIME Media Type returned by {@link getMimeType}. A <CODE>StructuredDocument</CODE>
     * sub-class may, however, support more than one MIME Media type so this may be a
     * true parameter. XMLDocument supports additional the MIME Media Type parameters,
     * "charset". The charset parameter must be a value per IETF RFC 2279 or ISO-10646.
     * @param stream Contains the input used to construct this object. This stream should be of a type
     * conformant with the type specified by the MIME Media Type "charset" parameter.
     *
     * @throws Exception Propagates exceptions from the underlying implementation.
     *
     */
    public XMLDocument( final MimeMediaType mimeType, final InputStream stream ) throws Exception {
        super( null, null );
        this.root = this;
        
        this.mimeType = (MimeMediaType) mimeType.clone();
        
        try {
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
            docBuilderFactory.setNamespaceAware(true);
            docBuilderFactory.setValidating(false);
            DocumentBuilder dataDocBuilder = docBuilderFactory.newDocumentBuilder();
            domNode = dataDocBuilder.parse( new InputSource(stream) );
        }
        catch (Exception e) {
            //System.err.println( "Failure in processing document" );
            // e.printStackTrace();
            domNode = null;
            throw e;
        }
    }
    
    /**
     * Returns the MIME Media type of this Document per
     * {@link http://www.ietf.org/rfc/rfc2046.txt}. Prefered to {@link getType}
     * because it returns a richer set of information, in particular the parameters
     * are available for inspection.
     *
     * @return The MIME Media Type of this document including any parameter declarations.
     *
     */
    public MimeMediaType getMimeType() {
        return mimeType;
    }
    
    /**
     * Returns a stream of bytes which represent the content of this Document.
     *
     * @return An {@link InputStream} containting the bytes of this Document.
     */
    public InputStream getStream() {
        // XXX bondolo@jxta.org  20010208    This has to be changed. We should be able to use the StringWriter directly.
        try {
            OutputFormat    format  = new OutputFormat( ((org.w3c.dom.Document) domNode) );   //Serialize DOM
            StringWriter  stringOut = new StringWriter();
            // Writer will be a String
            XMLSerializer    serial = new XMLSerializer( stringOut, format );
            serial.asDOMSerializer();                   // As a DOM Serializer
            
            serial.serialize( ((org.w3c.dom.Document) domNode).getDocumentElement() );
            
            return new ByteArrayInputStream( stringOut.toString().getBytes() );
        } catch ( Exception ex ) {
            ex.printStackTrace();
            return null;
        }
    }
    
    public void sendToStream( OutputStream stream ) throws IOException {
        Writer out = new BufferedWriter(new OutputStreamWriter(stream));
        
        sendToWriter( out );
        out.flush();
    }
    
    /**
     * Returns a stream of characters which represent the content of this Document.
     *
     * @return An {@link InputStream} containting the bytes of this Document.
     */
    public Reader getReader() {
        // XXX bondolo@jxta.org  20010208    This has to be changed. We should be able to use the StringWriter directly.
        try {
            OutputFormat    format  = new OutputFormat( ((org.w3c.dom.Document) domNode) );   //Serialize DOM
            StringWriter  stringOut = new StringWriter();
            // Writer will be a String
            XMLSerializer    serial = new XMLSerializer( stringOut, format );
            serial.asDOMSerializer();                   // As a DOM Serializer
            
            serial.serialize( ((org.w3c.dom.Document) domNode).getDocumentElement() );
            
            return new StringReader( stringOut.toString() );
        } catch ( Exception ex ) {
            ex.printStackTrace();
            return null;
        }
    }
    
    public void sendToWriter( Writer stream ) throws IOException {
        // XXX bondolo@jxta.org  20010208    This has to be changed. We should be able to use the StringWriter directly.
        try {
            OutputFormat    format  = new OutputFormat( ((org.w3c.dom.Document) domNode) );   //Serialize DOM
            // Writer will be a String
            XMLSerializer    serial = new XMLSerializer( stream, format );
            serial.asDOMSerializer();                   // As a DOM Serializer
            
            serial.serialize( ((org.w3c.dom.Document) domNode).getDocumentElement() );
        } catch ( IOException ex ) {
            throw ex;
        }
        catch ( Exception ex ) {
            throw new IOException( ex.toString() );
        }
    }
    
    /**
     * create a new element without value
     */
    public net.jxta.document.Element createElement( Object key ) {
        return createElement( key, null );
    }
    
    /**
     * create a new element with value
     */
    public net.jxta.document.Element createElement( Object key, Object val) {
        if( !String.class.isAssignableFrom( key.getClass() ) )
            throw new ClassCastException( key.getClass().getName() + " not supported by createElement." );
        
        if( (null!=val) && !String.class.isAssignableFrom( val.getClass() ) )
            throw new ClassCastException( val.getClass().getName() + " not supported by createElement." );
        
        return (net.jxta.document.Element) createElement( (String) key, (String) val );
    }
    
    // StructuredDocument Methods
    
    /**
     * Create a new element without value
     *
     * @param   name    The name of the element to be created.
     * @return  The new element.
     */
    public TextElement createElement(String name) {
        return (TextElement) new XMLElement( this, ((org.w3c.dom.Document) domNode).createElement(name));
    }
    
    /**
     * Create a new element with value
     *
     * @param   name    The name of the element to be created.
     * @param   value   The value of the element to be created.
     * @return  The new element.
     */
    public TextElement createElement(String name, String value) {
        org.w3c.dom.Element root;
        
        root = ((org.w3c.dom.Document) domNode).createElement(name);
        if( null != value )
            root.appendChild( ((org.w3c.dom.Document) domNode).createTextNode(value));
        return (TextElement) new XMLElement( this, root);
    }
    
    // Element Methods
    
    // Protected & Private Methods
    
    protected Node getAssocNode() {
        return (Node)((org.w3c.dom.Document) domNode).getDocumentElement();
    }
}
