/************************************************************************
 *
 * $Id: Address.java,v 1.6 2001/06/10 01:56:21 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint;

import net.jxta.impl.endpoint.*;
import net.jxta.endpoint.*;

import java.io.*;
 

/** 
 * The Address class implements the JXTA Endpoint address.
 * A JXTA endpoint address is a URI that has the following form:
 *
 * protocol://network address/service name/optional parameters
 *
 * For instance, a JXTA Peer on TCP, will have its PipeService Service address
 * defines as:
 *
 * tcp://192.168.1.10/JxtaPipe
 *
 * A particular pipe on would have the following address:
 *
 * tcp://192.168.1.10/JxtaPipe/000...128 bit UUID of the pipe
 *
 */

public class Address implements EndpointAddress {

    String protocol;
    String protocolAddress;
    String service;
    String serviceParam;

    /**
     * Constructor: builds an empty Address.
     */
    public Address () {

	protocol = null;
	protocolAddress = null;
	service = null;
	serviceParam = null;
    }

    /**
     * Constructor: builds an Address from a string
     */
    public Address (String address) {
	parse (address);
    }

    /**
     * Return a String that contains the canonical representation of the
     * Endpoint Address.
     *
     * @return a String containing the EndpointAddress
     */
    public synchronized String toString() {

        if (service == null) {
	    return protocol + "://" + protocolAddress + "/";
	}
        if (serviceParam != null) {
	    return protocol + "://" + protocolAddress + "/" +
	           service + "/" + serviceParam;
	}

        return protocol + "://" + protocolAddress + "/" + service;
    }

    /**
     * Return a String that contains the name of the protocol
     * contained in the EndpointAddress
     *
     * @return a String containing the protocol name
     * @throw IOExcetion if the EndpointAddress is malformed.
     */
    public synchronized String getProtocolName() throws IOException {
	return protocol;
    }

    /**
     * Return a String that contains the name of the protocol
     * address contained in the EndpointAddress
     *
     * @return a String containing the protocol address
     * @throw IOExcetion if the EndpointAddress is malformed.
     */
    public synchronized String getProtocolAddress() throws IOException {
	return protocolAddress;
    }

    /**
     * Return a String that contains the name of the service name
     * contained in the EndpointAddress
     *
     * @return a String containing the service name
     * @throw IOExcetion if the EndpointAddress is malformed.
     */
    public synchronized String getServiceName() throws IOException {
	return service;
    }

    /**
     * Return a String that contains the name of the service
     * parameter contained in the EndpointAddress
     *
     * @return a String containing the protocol name
     * @throw IOExcetion if the EndpointAddress is malformed, or if
     * there is no parameter associated to the service
     */
    public synchronized String getServiceParameter() throws IOException {
	return serviceParam;
    }

    /**
     * Set the protocol name.
     *
     * @param name String containing the name of the protocol
     */
    public synchronized void setProtocolName(String name) {
	protocol = name;
    }

    /**
     * Set the protocol address.
     *
     * @param address String containing the peer address.
     */
    public synchronized void setProtocolAddress(String address) {
	protocolAddress = address;
    }

    /**
     * Set the service name.
     *
     * @param name String containing the name of the service
     */
    public synchronized void setServiceName(String name) {
	service = name;
    }

    /**
     * Set the service parameter
     *
     * @param name String containing the service parameter
     */
    public synchronized void setServiceParameter(String param) {
	serviceParam = param;
    }

    private void parse (String addr) {
	int index = addr.indexOf ("://");
	String tmp = null;
	if (index == -1) {
	    // The address is not a valid address. An exception
	    // could be returned. XXX: to be checked.
	    return;
	}
	try {
	    protocol = addr.substring (0, index);
	    tmp = addr.substring (index + 3);
	} catch (Exception e) {
	    // The exception should maybe be propagated.
	    // XXX: to be checked
	    return;
	}
	index = tmp.indexOf ("/");
	if (index == -1) {
	    protocolAddress = tmp;
	    return;
	}
	try {
	    protocolAddress = tmp.substring (0, index);
	    tmp = tmp.substring (index + 1);
	} catch (Exception e) {
	    // The exception should maybe be propagated.
	    // XXX: to be checked
	    return;
	}

	index = tmp.indexOf ("/");
	if (index == -1) {
	    service = tmp;
	    return;
	}
	try {
	    service = tmp.substring (0, index);
	    tmp = tmp.substring (index + 1);
	} catch (Exception e) {
	    // The exception should maybe be propagated.
	    // XXX: to be checked
	    return;
	}
        serviceParam = tmp;
    }
    
    /**
     * Return a String that contains the canonical representation of the
     * Endpoint Address.
     *
     * @deprecated Please use toString() instead.
     * @return String a String containing the EndpointAddress
     * @exception IOExcetion if the EndpointAddress is malformed.
     *
     * @since JXTA 1.0
     */
    public String makeString() throws java.io.IOException {
        return toString();
    }
    
}



