/************************************************************************
 *
 * $Id: Endpoint.java,v 1.43 2001/06/10 02:24:41 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint;
import net.jxta.endpoint.*;

import net.jxta.peergroup.*;
import net.jxta.rendezvous.*;
import net.jxta.impl.rendezvous.*;
import net.jxta.document.*;

import java.io.*;
import java.util.*;

/**
 * This class implements the placeholders for all the JXTA
 * transport protocols, as well, as the API for the implementation
 * of the core protocols that use directely the Endpoint
 */

public class Endpoint {

    public static boolean debugFlag = false;

    public static final int NonBlocking = 1;
    public static final int HandChecking = 2;

    private Hashtable protocols = new Hashtable();
    private Hashtable handlers = new Hashtable();
    private Endpoint parentEndpoint = null;

    private boolean runRouter = true;
    private boolean routerInitialized = false;
    public  EndpointRouter router = null;

    private PeerGroup group = null;
    private PeerGroup rootGroup = null;

    private String localPeerId = null;

    private Hashtable inboundTraffic = new Hashtable();
    private Hashtable outboundTraffic = new Hashtable();

    /**
     * Time in seconds since midnight, January 1, 1970 UTC and when
     * this peer last sent a message out
     */
    private int lastMessageSentAt = 0;

    /**
     * Time in seconds since midnight, January 1, 1970 UTC and when
     * this peer last recieved a message
     */
    private int lastMessageRecdAt = 0;

    /** 
     * Traffic recording is disabled by default, in the interest of
     * efficiency. The first PIP request to getTrafficChannels enables
     * it. We can go further and disable it after a timeout if no PIP
     * request has been received for some time (TODO).
     */
    private boolean isTrafficRecordingEnabled = false;

    private static final String EndpointHeaderSrcPeer = "EndpointHeaderSrcPeer";

    public Endpoint (Endpoint parent, PeerGroup g) {

	group = g;
        debugFlag = group.getDebug();
        
	debug ("Endpoint constructor");
        
	if (parent == null) return;

	parentEndpoint = parent;

	// To speed things up, each endpoint has the list of all
	// the protocols at or below its own level. Grab them, now.
	// This way, propagate, send message
	// or getProtocols does not have to be recursive.
	// Limitation: for now, endpoints have no way to notify their
	// descendants of any change.
	// On the other end, registerHandler HAS TO be propagate down the
	// stack, so that demux does not have to be recursive...same idea.

	Enumeration protosBelow = parentEndpoint.getProtocols();
	while (protosBelow.hasMoreElements()) {
	    try {
		addProtocol((TransportProtocol) protosBelow.nextElement());
	    } catch (IOException e) {
		// can't happen, it's just so that we do not have to throw it.
	    }
	}
    }

    public void publishPeerAdv (Advertisement adv) {

	if (runRouter) {
	    router.publishPeerAdv (adv);
	}
    }

    public Enumeration getPeerAdv (String pId) {
	
	if (runRouter) {
	    return router.getPeerAdv (pId);
	}
	return null;
    }

    public synchronized void setRootGroup (PeerGroup g) {

	debug ("setRootGroup");
	rootGroup = g;
	localPeerId = g.getPeerID().toString();
    }

    public synchronized PeerGroup getRootGroup () {
	return rootGroup;
    }

    public synchronized void addProtocol (TransportProtocol proto) 
                       throws IOException {
	// Start the router
	// XXX: this should be configurable. to be fixed
	// lomax@jxta.org
	if ((runRouter) && (!routerInitialized)) {
	    router = new EndpointRouter (group, this);
	    routerInitialized = true;
	}

	try {
	    if (protocols.contains (proto.getProtocolName())) {
		// There is already a protocol of that name.
		// Throw an exception
		throw new IOException();
	    }
	    Object obj = protocols.put (proto.getProtocolName(), proto);
	} catch (Exception e) {
	    throw new IOException();
	}
    }

    public synchronized boolean removeProtocol (TransportProtocol proto) {
	try {
	    Object obj = protocols.remove (proto.getProtocolName());
	    if (obj != null) return true;
	} catch (Exception e) {
	    // XXX: maybe something should be done.. to be checked
	    return false;
	}
	return false;
    }

    public synchronized Enumeration getProtocols () {
	return protocols.elements();
    }

    public Message getMessage () {
	return new MessageImpl();
    }

    public EndpointAddress getEndpointAddress (String Uri) {
        return new Address (Uri);
    }

    public EndpointMessanger getMessanger (EndpointAddress addr,
                                           int type)
	                            throws IOException {

	// First find the appropriate TransportProtocol to use.
        TransportProtocol proto = getProtocolByName (addr.getProtocolName());

	if (proto == null) {
	    throw new IOException();
	}
	EndpointMessanger messanger = proto.getMessanger (addr, null, type);
	if (runRouter) {
	    return router.getMessanger (null,
					messanger,
					this);
	} else {
	    return messanger;
	}
    }

    public EndpointMessanger getMessanger ( String dstPeer,
					    String pName,
					    String pParam,
					    int type)
	                            throws IOException {

	if (dstPeer.equals (localPeerId)) {
	    debug ("getMessanger: return LoopbackMessanger");
	    return new LoopbackMessanger (this, pName, pParam);
	}

	EndpointAddress addr = findAddr (dstPeer);

	if (addr == null) {
	    debug ("getMessanger: no endpoint");
	    return null;
	}

	addr = new Address (addr.toString());

	debug ("getMessanger:");
	debug ("    dest peer= " + dstPeer);
	debug ("    using    = " + addr.toString());

	addr.setServiceName (new String (pName));
	addr.setServiceParameter (new String (pParam));

	// First find the appropriate TransportProtocol to use.
        TransportProtocol proto = getProtocolByName (addr.getProtocolName());

	if (proto == null) {
	    debug ("getMessanger: no protocol");
	    throw new IOException();
	}
	EndpointMessanger messanger = proto.getMessanger (addr, dstPeer, type);
	if (runRouter) {
	    return router.getMessanger (dstPeer,
					messanger,
					this);
	} else {
	    return messanger;
	}
    }

    public void propagate (PeerGroup g,
			   Message srcMsg,
			   String pName,
			   String pParam,
			   int ttl,
			   String prunePeer) throws IOException {


	String gId = null;
	RendezVousImpl rdv = null;

	if (g != null) {
	    gId = g.getID().toString();
	    rdv = (RendezVousImpl) g.getRendezVous();
	}

	debug ("propagate message  group= " + gId);

	// Add our header.
	try {
	    srcMsg.push (EndpointHeaderSrcPeer,
			 new ByteArrayInputStream (localPeerId.getBytes()));
	} catch (Exception e) {
	    debug ("  cannot push Endpoint header onto message");
	}

	TransportProtocol proto = null;
	if (runRouter) {
	    srcMsg = router.processOutgoingPropagateMsg (
				 (MessageImpl) srcMsg,
				 localPeerId,
				 ttl);
	    if (srcMsg == null) {
		// The router has decided to not propagate this message
		debug ("    Router has decided to not propagate this message");
		return;
	    }
	}
	
	MessageImpl m = null;
	Enumeration enum = protocols.elements();

	if ((enum == null) || (!enum.hasMoreElements())) {
	    return;
	}

	while (enum.hasMoreElements()) {
	    try {
		proto = (TransportProtocol) enum.nextElement();
		debug ("propagate onto " + proto.getProtocolName());
		m = (MessageImpl) srcMsg.dup();;		
		m.setDestAddress (((MessageImpl) srcMsg).getDestAddress());
		m.setSrcAddress (((MessageImpl) srcMsg).getSrcAddress());
		proto.propagate (m,
				 new String(pName),
				 new String (pParam),
				 prunePeer);
                recordTraffic(m, false);
	    } catch (Exception e) {
	        e.printStackTrace();
		debug ("Failed to send on protocol");
		continue;
	    }
	}



	// Send to RendezVous and Clients 
	if (rdv != null) {

	    m = (MessageImpl) srcMsg.dup();
	    m.setSrcAddress(((MessageImpl) srcMsg).getSrcAddress());
	    m.setDestAddress(((MessageImpl) srcMsg).getDestAddress());

	    rdv.propagateToRendezVous (m,
				       new String(pName),
				       new String (pParam),
				       prunePeer);

            recordTraffic(m, false);

	    rdv.propagateToClients (srcMsg,
				    new String(pName),
				    new String (pParam),
				    prunePeer);

	    recordTraffic(srcMsg, false);
	}
    }

    synchronized private void registerHandlerSync (String address,
						   EndpointHandler handler)
	throws IOException {

	if (handler == null || handlers.contains (address)) {
	    throw new IOException();
	}

	handlers.put (new String (address), handler);
    }

    public void registerHandler(String address, EndpointHandler handler)
	throws IOException {

	address = new String (address);
	if (parentEndpoint != null)
	    parentEndpoint.registerHandler(address, handler);

	// If parent went through, then let's do it here.
	registerHandlerSync(address, handler);
    }

    // We do not trust services all that much; here, a service of a bad
    // group could screw-up the endpoint of a well behaved ancestor.
    synchronized private
	boolean unregisterHandlerSync (String address,
				       EndpointHandler handler) {

	// Can remove and put back; we're synchonized.
	EndpointHandler h = (EndpointHandler) handlers.remove (address);
	if (h == null) {
            return false;
	}
	if (handler == null) return false;
	if (h == handler) return true;

	// Removed the wrong one; put it back !
	handlers.put (new String(address), h);
	return false;
    }

    public boolean unregisterHandler(String address, EndpointHandler handler) {
	if (unregisterHandlerSync(address, handler) == false) return false;
	if (parentEndpoint == null) return true;
	return parentEndpoint.unregisterHandler(address, handler);
    }

    synchronized public EndpointHandler lookupHandler (String address) {
	return (EndpointHandler) handlers.get (address);
    }

    public void demux (MessageImpl msg) throws IOException {

	debug ("demux starts");
	EndpointAddress addr = msg.getDestAddress();

	if (addr == null) {
	    // No destination address... Just discard
	    debug ("demux: no destination address. Discard");
	    return;
	}

        recordTraffic(msg, true);

	String srcPeer = null;
        InputStream ip = null;
        try {
            ip = msg.pop (EndpointHeaderSrcPeer);
            if (ip != null) {
		byte[] buffer = new byte [ip.available()];
		ip.read (buffer);
		srcPeer = new String (buffer);
            }
	} catch (Exception e) {
            // REVIEW 2001-05-30 <nelson> -- Blanket ignoring of exceptions
            //   is dangerous! But pop() throws an IOException to mean
            //   "tag not found", so this is safe, at least sometimes.
	}

	if ((srcPeer != null) && srcPeer.equals (localPeerId)) {
	    // This is a loopback. Discard.
	    debug ("   discard loopback");
	    return;
	}

	debug ("demux from " + msg.getSrcAddress().toString());
	debug ("demux   to " + msg.getDestAddress().toString());

	// Is there a router enabled ?
	if (runRouter) {
	    debug ("demux: call router");
	    msg = router.processIncomingMsg (msg);
	    if (msg == null) {
		debug ("demux: no message to process after router");
		return;
	    }
	}
	// Find the appropriate destination of this message
	String serviceName = addr.getServiceName();
	String serviceParam = addr.getServiceParameter();
	
	if (serviceName == null) serviceName = "";
	if (serviceParam == null) serviceParam = "";

	debug ("demux:");
	debug ("    service name: " + serviceName);
	debug ("   service param: " + serviceParam);

	EndpointHandler h = lookupHandler(serviceName + serviceParam);

	if (h == null) {
	    debug ("demux: no listener for this kind of message");
	    return; // noone cares
	}
	debug ("demux: send to service handler");
	h.demux(msg);
    }

    public TransportProtocol getProtocolByName (String name) {
	return (TransportProtocol)  protocols.get (name);
    }
   
    private EndpointAddress findAddr (String pId) {
	if (runRouter) {
	    return router.getAddress (pId);
	}
	return null;
    }

    private void debug (String str) {
	if (debugFlag) {
	    System.out.println("[Endpoint] " + str);
	}
    }

    public boolean addPeerRoute (String peer) {
	
	debug ("addPeerRoute " + peer);
	if (runRouter) {
	    EndpointAddress addr = router.getAddress (peer);
	    if (addr == null) {
		return false;
	    } else {
		return true;
	    }
	}
	return true;
    }

    public boolean ping (EndpointAddress addr) {

	if (addr == null) {
	    return false;
	}

	try {
	    String protoName = addr.getProtocolName();
	    TransportProtocol proto = getProtocolByName (protoName);
	    return proto.ping (addr);
	} catch (Exception e) {
	    return false;
	}
    }

    private void recordTraffic(Message msg, boolean inbound)  {

      if (!isTrafficRecordingEnabled)
	  return;

      try {
        int now = (int) (System.currentTimeMillis() / 1000L);
        if (inbound) {
            lastMessageRecdAt = now;
        } else {
            lastMessageSentAt = now;
        }

	MessageImpl msgi = (MessageImpl) msg;
	Address addr = (Address) 
	    (inbound ? msgi.getSrcAddress() : msgi.getDestAddress());
	if (addr == null)
	    return;

        Hashtable traffic = inbound ? inboundTraffic : outboundTraffic;
        Enumeration msgNames = msg.getNames();
        while (msgNames != null && msgNames.hasMoreElements()) {
            String mn = (String) msgNames.nextElement();
            int ms = msg.getSize(mn);
	    mn = addr.getProtocolName() + "://" +
		addr.getProtocolAddress() + "/" +
		mn;
            synchronized (traffic) {
                Long tr = (Long) traffic.get(mn);
                if (tr == null) {
                    tr = new Long(ms);
                    traffic.put(mn, tr);
                } else {
                    long tl = tr.longValue();
                    tl += ms;
                    tr = new Long(tl);
                    traffic.put(mn, tr);
                }
            }
        }
      } catch (Exception ex) {
	  ex.printStackTrace();
      }
    }

    public Enumeration getTrafficChannels(boolean inbound) {
	// enable traffic recording, if not already enabled
	isTrafficRecordingEnabled = true;
        return inbound ? inboundTraffic.keys() : outboundTraffic.keys();
    }

    public Long getTrafficOnChannel(String channel, boolean inbound) {
        return (Long) (inbound ? 
                       inboundTraffic.get(channel) : 
                       outboundTraffic.get(channel));
    }

    public int getLastMessageTime(boolean inbound) {
        return inbound ? lastMessageRecdAt : lastMessageSentAt;
    }
}
