/************************************************************************
 *
 * $Id: EndpointRouter.java,v 1.59 2001/06/14 18:35:06 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint;

import net.jxta.endpoint.*;
import net.jxta.document.*;
import net.jxta.peergroup.*;
import net.jxta.discovery.*;
import net.jxta.impl.discovery.*;
import net.jxta.protocol.*;
import net.jxta.impl.protocol.*;
import net.jxta.resolver.*;
import net.jxta.exception.*;

import java.io.*;
import java.util.*;

public class EndpointRouter implements QueryHandler, Runnable {

    private static boolean debug =  false;

    private Hashtable routedRoutes = null;
    private Hashtable directRoutes = null;
    protected String localPeerId = null;
    private Endpoint endpoint = null;
    // XXX: MaxTTL should be a configurable. For the time being
    // rely on the seven levels of separation.
    // lomax@jxta.org
    private static final int MaxTTL = 7;
    private Discovery discovery = null;
    private PeerGroup group = null;
    private Thread compacterThread = null;
    private int CompacterDelay = 600000; // 10 minutes
    private boolean routedCacheChanged = false;
    private boolean directCacheChanged = false;
    private static final String name = "EndpointRouter";
    private static final String RoutingPeerIdTag = "RoutingPeer";
    private static final String RoutingPeerAdvTag = "RoutingPeerAdv";
    private static final String DestPeerIdTag = "DestPeer";
    private static final String GatewayTag = "Gateway";
    private static final String VersionTag = "Version";
    private static final int    acceptableVersion = 1;
    private static final int    currentVersion = 1;
    private Resolver resolver = null;
    private boolean servicesInitialized = false;
    private int qid = 0;
    private Vector pendingQueries = null;

    public EndpointRouter (PeerGroup g, Endpoint e) {

	group = g;
	debug = g.getDebug();

	localPeerId = group.getPeerID().toString();

	endpoint = e;
	directRoutes = new Hashtable();
	routedRoutes = new Hashtable();
	pendingQueries = new Vector();

	getServices();

	// Start the background thread that compacts and optimizes
	// the cache.

	compacterThread = new Thread (this, "EndpointRouter:Compacter Thread");
	compacterThread.start();
    }


    private synchronized void getServices () {

	if (servicesInitialized) return;
	PeerGroup g = endpoint.getRootGroup();
	if (g == null) return;

	debug ("getServices start");
        discovery = g.getDiscovery();
        resolver = g.getResolver();

        // Register to the Generic ResolverService
        resolver.registerHandler (name, this);
	servicesInitialized = true;
    }


    public void publishPeerAdv (Advertisement adv) {
	
	debug ("publishPeerAdv starts");
	getServices();
	if (discovery == null) {
	    debug ("  no discovery");
	    return;
	}

	try {
	    discovery.publish (adv, Discovery.PEER);
	} catch (Exception e) {
	    debug ("  failed with " + e);
	}
    }

    public Enumeration getPeerAdv (String pId) {

	debug ("getPeerAdv for : " + pId);
	getServices();
	if (discovery == null) {
	    debug ("  no discovery");
	    return null;
	}

	try {
	    return discovery.getLocalAdvertisements(Discovery.PEER,
						    "Pid",
						    pId);
	} catch (Exception e) {
	    debug ("  failed with " + e);
	    return null;
	}
    }

    public EndpointAddress getAddress (String pId) {

	// Compacting the cache first is a good idea if we want to
	// get the best route. Note that compactCaches() can be expensive,
	// but:
	//        - it is saving time and network bandwith ahead of time
	//        - the proces of compacting happens only if there has been
	//          a change in the caches.
	//
	// So, all together, this seems to be a good optimization.
	//

	pId = new String (pId);
	compactCaches();
	debug ("getAddress for ");
	debug ("      peer= " + pId);

	Vector gateways = null;
	int count = 0;
	
	while (true) {

	    // Look if there is a local route to the peer.
	    if (isLocalRoute (pId)) {
		debug ("Found local address");
		return getLocalRoute (pId);
	    }

	    gateways = getRoute (pId);

	    if ((gateways != null) && (gateways.size() > 0)) {

		// The first element of the vector is a peer
		// for which we do have a connection.
		try {
		    EndpointAddress addr =
			getLocalRoute ((String) gateways.elementAt (0));
		    debug ("Found remote address " + addr.toString());
		    return addr;
		} catch (Exception e) {
		    debug ("getAddress for " + pId);
                    debug ("     failed with " + e);
		    return null;
		}
	    }


	    // Trying to recover a route
	    if (count == 0) {
		// this is the first time in the loop.
		findRoute (pId);
	    }

	    // Just wait a bit
	    try {
		Thread.sleep (500);
	    } catch (Exception e) {
		debug ("getAddress got an exception " + e + " while waiting");
	    }
	    if (++count > 240) {
		// Give up after 2 minutes
		debug ("getAddress is giving up for " + pId);
		break;
	    }
	}
	debug ("getAddress did not find a route for " + pId);
	return null;
    }


    public EndpointAddress getBestLocalRoute (String peer, Enumeration addrs) {

	debug ("getBestLocalRoute:");
	debug ("peer= " + peer);
	
	// Check parameters
	if ((peer == null) || (addrs == null) || (!addrs.hasMoreElements())) {
	    return null;
	}

	TransportProtocol proto = null;
	EndpointAddress addr = null;
	EndpointAddress bestAddr = null;
	String protoName = null;
	boolean direct = false;
	boolean fast = false;
	
	while (addrs.hasMoreElements()) {
	    try {
		addr = (EndpointAddress) addrs.nextElement();
		protoName = addr.getProtocolName();
		proto = endpoint.getProtocolByName (protoName);
		debug ("getBestLocalRoute trying: " + addr.toString()
			     + " on protocol " + proto.getProtocolName());
		if (proto == null) {
		    continue;
		}
		if (!proto.ping (addr)) {
		    // This protocol cannot reach the target addr
		    continue;
		}

		if (bestAddr == null) {
		    debug ("Found: " + addr.toString());
		    bestAddr = addr;
		}

		if (proto.isConnected()) {
		    if (direct) {
			// There is already a protocol supporting
			// a direct connection to the target addr.
			if ((!fast) && (proto.isFast())) {
			    return addr;
			}
			continue;
		    } else {
			direct = true;
			if (proto.isFast()) {
			    return addr;
			} else {
			    bestAddr = addr;
			    continue;
			}
		    }
		} else {
		    if (direct) {
			continue;
		    } else {
			bestAddr = addr;
			continue;
		    }
		}
	    } catch (Exception e) {
		continue;
	    }
	}
	
	return bestAddr;
    }


    public synchronized boolean isLocalRoute (String pId) {

	if (directRoutes.containsKey (pId)) {
	    return true;
	}
	return false;
    }
      

    public synchronized EndpointAddress getLocalRoute (String pId) {

	if (directRoutes.containsKey (pId)) {
	    return (EndpointAddress) directRoutes.get (pId);
	}
	return null;
    }
      
    public synchronized void setLocalRoute (String pId, EndpointAddress addr) {

	try {
	    addr = new Address (addr.toString());
	    pId = new String (pId);

	    debug ("setLocalRoute:");
	    debug ("   peerId = " + pId);
	    debug ("   addr   = " + addr.toString());
	    directRoutes.put (pId, addr);
	    directCacheChanged = true;
	} catch (Exception e) {
	}
    }

    public synchronized void removeLocalRoute (String pId) {

	try {
	    directRoutes.remove (pId);
	} catch (Exception e) {
	}
    }

    public synchronized boolean isRoute (String pId) {

	if (routedRoutes.containsKey (pId)) {
	    Vector peers = (Vector) routedRoutes.get (pId);
	    if (peers == null) {
		return false;
	    }
	    return true;
	}
	return false;
    }

    public synchronized Vector getRoute (String pId) {

	if (routedRoutes.containsKey (pId)) {
	    Vector peers = (Vector) routedRoutes.get (pId);
	    if (peers == null) {
		return null;
	    }
	    return peers;
	}
	return null;
    }
	
    public synchronized void setRoute (String pId, Vector v) {

	pId = new String (pId);
	try {
	    debug ("setRoute:");
	    debug ("   pId = " + pId);
	    debug ("  next = " + (String) v.elementAt (0));
	    routedRoutes.put (pId, v);
	    routedCacheChanged = true;
	} catch (Exception e) {
	    debug ("   failed with " + e);
	}
    }

    public synchronized void removeRoute (String pId) {

	try {
	    routedRoutes.remove (pId);
	} catch (Exception e) {
	}
    }

    public MessageImpl processIncomingMsg (MessageImpl msg) {

	debug ("processIncomingMsg starts");
	// We need to initialize the services because that may have failed
	// into the constructor.
	// XXX: to be cleaned up  lomax@jxta.org
	getServices();
	InputStream ip = null;
	try {
	    ip = msg.pop (EndpointRouterMessage.Name);
	    if (ip == null) {
		// The sender did not use this router
		debug ("processIncomingMsg: no routing info");
		return msg;
	    }
	    EndpointRouterMessage routerMsg = null;
	    routerMsg = new EndpointRouterMessage (ip);

	    String srcPeer = routerMsg.getSrcPeer();
	    String destPeer = routerMsg.getDestPeer();
	    Vector gateways = routerMsg.getGateways();
	    if ((gateways != null) && (gateways.size() == 0)) {
		gateways = null;
	    }

	    debug ("     srcPeer= " + srcPeer);
	    debug ("    destPeer= " + destPeer);
	    debug ("nbOfGateways= " + ((gateways == null) ? 0 : gateways.size()));

	    // Is this a loopback ?
	    if ((srcPeer != null) && srcPeer.equals (localPeerId)) {
		debug ("processIncomingMsg: dropped loopback");
		return null;
	    }

	    // Update routing information about that peer.
	    // Do we already know about that peer ?
	    
	    if (srcPeer != null) {
		if (!isLocalRoute (srcPeer) && endpoint.ping (msg.getSrcAddress())) {
		    // We are able to directely talk to this peer. 
		    Vector tmp = getRoute (srcPeer);
		    if ((tmp == null) || (tmp.size() == 0)) {
			// This is the first time we hear about that peer.
			setLocalRoute (srcPeer, msg.getSrcAddress());
		    } else {
			// We already have a non direct route for
			// that peer. This route is direct -- better.
			removeRoute (srcPeer);
			setLocalRoute (srcPeer, msg.getSrcAddress());
		    }
		}
	    }


	    // It is very likely that the first peer in the gateway
	    // vector is the original source of the message. It is
	    // a good optimization, if we are a rendezvous, to memorize
	    // a route to that peer, because we will probably be asked
	    // to route a message to that peer later on.

	    if (gateways != null) {
		try {
		    String tmpPeer = (String) gateways.elementAt (0);

		    // Update routing information about that peer.
		    // Do we already know about that peer ?
	    
		    if (!isLocalRoute (tmpPeer)) {
			Vector tmp = getRoute (tmpPeer);
			if ((tmp == null) || (tmp.size() == 0)) {
			    // This is the first time we hear about that peer.
			    tmp = (Vector) gateways.clone();
			    try {
				// Remove itself
				tmp.removeElementAt (0);
				if (tmp.size() > 0) {
				    setRoute (tmpPeer, reverseGateways(tmp));
				}
			    } catch (Exception e) {
			    }
			} else {
			    // We already have a non direct route for
			    // that peer. Check if this route is better.
			    if (tmp.size() > (gateways.size() - 1)) {
				// This route seems to be better.
				tmp = (Vector) gateways.clone();
				try {
				    // Remove itself
				    tmp.removeElementAt (0);
				    if (tmp.size() > 0) {
					removeRoute (tmpPeer);
					setRoute (tmpPeer, reverseGateways(tmp));
				    }
				} catch (Exception e) {
				}
			    }
			}
		    }
		} catch (Exception e) {
		}
	    }

	    // Update the router message and put it back into the message

	    if (gateways == null) {
		gateways = new Vector();
	    }

	    gateways.addElement (srcPeer);
	    routerMsg.setGateways (gateways);
	    try {
		msg.push (EndpointRouterMessage.Name, routerMsg.getInputStream());
	    } catch (Exception e) {
		debug ("processIncomingMsg cannot push header back in message");
	    }

	    // Do we have to forward this message to another peer ?
	    if ((destPeer != null) && !(destPeer.equals(localPeerId))) {
		// Find the next router
		debug ("processIncomingMsg: forward unicast");

		Vector route = null;
		if (isLocalRoute (destPeer)) {
		    route = new Vector (1);
		    try {
			route.addElement(destPeer);
		    } catch (Exception e) {
			return null;
		    }
		} else {
		    route = getRoute (destPeer);
		}
		
		if (route == null) {
		    // XXX: here we should trigger route recovery
		    // to be fixed. lomax@dioxine.com
		    debug ("processIncoming: no route for peer:");
		    debug ("    " + destPeer);
		    return null;
		}
		EndpointMessanger messanger = null;
		String pName = msg.getDestAddress().getServiceName();
		String pParam = msg.getDestAddress().getServiceParameter();
		try {
		    messanger =
			endpoint.getMessanger ((String) route.elementAt (0),
					       pName,
					       pParam,
					       Endpoint.NonBlocking);
		    if (messanger == null) {
			debug ("processIncomingMsg: no EndpointMessanger");
			return null;
		    }
		    messanger.sendMessage (msg);
		    return null;
		} catch (Exception e) {
		    debug ("processIncomingMsg: error sending msg: " + e);
		    return null;
		}
	    }
	} catch (Exception e) {
	    debug ("   failed with " + e);
	}
	return msg;
    }

    public MessageImpl processOutgoingPropagateMsg (MessageImpl msg,
						    String srcPeer,
						    int ttl) {

	int msgTTL = 0;

        if (srcPeer == null) {
            srcPeer = "UnknownPeer";
	}
        
	debug ("Propagating message from " + srcPeer);
	// Check to see if there already is a EndpointRouterMessage in the message.
	InputStream ip = null;
	EndpointRouterMessage routerMsg = null;
	try {
	    ip = msg.pop (EndpointRouterMessage.Name);
	    if (ip != null) {
		routerMsg = new EndpointRouterMessage (ip);
		msgTTL = routerMsg.getTTL();
		debug ("  TTL= " + msgTTL);
		--msgTTL;
		if (msgTTL == 0) {
		    // We have reached maximum TTL for this request.
		    // Do not propagate it.
		    debug ("   TTL = 0. Message is discarded");
		    return null;
		}
		// Loop detection.
		Vector gateways = routerMsg.getGateways();
		if ((gateways != null) && (gateways.size() > 0)) {
		    for (int i = 0; i < gateways.size(); ++i) {
			try {
			    if (((String) gateways.elementAt (i)).equals (localPeerId)) {
				// This is a loop
				return null;
			    }
			} catch (Exception e) {
			    break;
			}
		    }
		}
	    }
	} catch (Exception e) {
	}

	if (routerMsg == null) {
	    // fill up the source and dest fields.
	
	    routerMsg = new EndpointRouterMessage();
	    routerMsg.setSrcPeer (srcPeer);
	    routerMsg.setTTL (MaxTTL);
	}		

	// Compute TTL
	if (ttl < msgTTL) {
	    msgTTL = ttl;
	}

	if (ttl > MaxTTL) {
	    ttl = MaxTTL;
	}
	
	routerMsg.setTTL (ttl);


	// fill up the gateways section
	if (!isLocalRoute (srcPeer) && !srcPeer.equals (localPeerId)) {
	    Vector gateways = getRoute (srcPeer);
	    routerMsg.setGateways (gateways);
	}

	try {
	    msg.push (EndpointRouterMessage.Name, routerMsg.getInputStream());
	} catch (Exception e) {
	}
	return msg;
    }


    private synchronized void compactCaches () {

	debug ("compactCaches starts");
	// Routed cache
	if (!routedCacheChanged) {
	    // Nothing to do, so do nothing.
	    debug ("compactCaches: nothing to do");
	    return;
	} else {
	    routedCacheChanged = false;

	    // Check if any of the routed peers can, in fact, be
	    // directely reachable. This allows to create shorter
	    // (possibly direct) routes to peers.

	    Enumeration enum = null;
	    enum = routedRoutes.elements();
	    String peer = null;
	    EndpointAddress addr = null;

	    if ((enum != null) && (enum.hasMoreElements())) {
		while (enum.hasMoreElements()) {
		    try {
			peer = (String) enum.nextElement();
			addr = checkPeer (peer);
			if (addr != null) {
			    debug ("compactCaches: found direct route");
			    // Found a direct route for that peer
			    // Remove the entry fron routedCache
			    removeRoute (peer);
			    // Set local route;
			    setLocalRoute (peer, addr);
			}
		    } catch (Exception e) {
			continue;
		    }
		}
	    }
	}

	// Routed cache
	if (!directCacheChanged) {
	    // Nothing to do, so do nothing.
	    return;
	} else {
	    directCacheChanged = false;

	    // Check if any of the peers for which we already have
	    // a direct connection may have a better address to use.

	    Enumeration enum = directRoutes.elements();
	    String peer = null;
	    EndpointAddress addr = null;

	    if ((enum != null) && (enum.hasMoreElements())) {
		while (enum.hasMoreElements()) {
		    try {
			peer = (String) enum.nextElement();
			addr = checkPeer (peer);
			if (addr != null) {
			    // Found a direct route for that peer
			    // Remove the entry fron routedCache
			    debug ("compactCaches: found direct route for remote peer");
			    removeLocalRoute (peer);
			    // Set local route;
			    setLocalRoute (peer, addr);
			}
		    } catch (Exception e) {
			continue;
		    }
		}
	    }
	}
    }

    public EndpointAddress checkPeer (String peer) {

	debug ("checkPeer peer= " + peer);
	    
	// Get (locally) the advertisement of this peer
	Enumeration advs = null;
	PeerGroupAdvertisement adv = null;
	EndpointAdvertisement eadv = null;


	// We need to try to get the services now, because depending on the order on
	// which services are loaded, getServices() may have failed in the constructor
	// XXX: to be cleaned up. lomax@jxta.org
	
	getServices();
	
	try {
	    advs = endpoint.getPeerAdv (peer);

	    if (advs == null) {
		debug ("   no advertisement for " + peer);
		return null;
	    }

	    while (advs.hasMoreElements()) {

		adv = (PeerGroupAdvertisement) advs.nextElement();

		// Get its Endpoint advertisement
		Enumeration endps = null;
		endps = ((Hashtable) adv.getEndpointAdvertisements()).elements();
		if ((endps == null) || (!endps.hasMoreElements())) {
		    debug ("checkPeer: no EndpointAdv for that peer");
		    return null;
		}

		Vector addrs = new Vector();
		while (endps.hasMoreElements()) {
		    try {
			eadv = (EndpointAdvertisement) endps.nextElement();
			addrs.addElement (
					  new Address((String) eadv.getEndpointAddress()));

		    } catch (Exception e) {
			continue;
		    }
		}
		if (addrs.size() > 0) {
		    EndpointAddress bestAddr = null;
		    bestAddr = getBestLocalRoute (peer,
						  addrs.elements());
		    if (bestAddr != null) {
			// Found a direct route. Return it.
			return bestAddr;
		    }
		}
	    }
	} catch (Exception e) {
	    debug ("    failed with " + e);
	    return null;
	}
	debug ("   did not found a good route");
	return null;
    }


    public void run() {
	runCompacter();
    }

    private void runCompacter() {

	while (true) {
	    compactCaches();
	    try {
		Thread.sleep (CompacterDelay);
	    } catch (Exception e) {
	    }
	}
    }

    private void debug (String str) {
	if (debug) {
	    System.out.println ("[EndpointRouter] " + str);
	}
    }

    // This part is for the ResolverService part of the EndpointRouter


    private void findRoute (String peer) {

        debug ("findRoute starts");
	debug ("      peer= " + peer);

	// First check if we can find out by ourself.
	EndpointAddress addr = checkPeer (peer);

	if (addr != null) {
	    setLocalRoute (peer, addr);
	    return;
	}

        try {
            // First create the request message.
            StructuredTextDocument doc = ( StructuredTextDocument )
                StructuredDocumentFactory.newStructuredDocument (
                            new MimeMediaType( "text/xml" ),
                                               "jxta:EndpointRouter");

            Element e = null;
            e = doc.createElement (DestPeerIdTag, peer);
            doc.appendChild (e);

            StringWriter dumped = new StringWriter();
            doc.sendToWriter( dumped );

	    debug ("findRoute sends query " + name);
	    getServices();

            ResolverQuery query = new ResolverQuery (name,
                                                     null,
                                                     null,
                                                     localPeerId,
                                                     dumped.toString(),
						     qid++);

	    // Call get services. This is needed because getServices may have failed
	    // in the constructor, depending on the order the services are loaded.
	    // XXX: to be cleaned up   lomax@jxta.org

            resolver.sendQuery (null, query);
            debug ("getAddress() query sent");

        } catch (Exception ee) {
            debug ("findRoute failed with " + ee);
            return;
        }
    }


    // This is called by the Generic ResolverService when processing a response to
    // a query.
    public void pushResponse(ResolverResponseMsg response) {

        debug ("pushResponse got a response");
        Enumeration enum = null;
        String routingPeer = null;
	String routingPeerAdv = null;
        String destPeer = null;
        MimeMediaType mediaType = null;
        String ipId = null;

        try {
            mediaType = new MimeMediaType ("text/xml");
        } catch (RuntimeException e) {
            debug ("pushResponse: cannot get MimeMediaType " + e);
            return;
        }

        ByteArrayInputStream ip =
            new ByteArrayInputStream (response.getResponse().getBytes());

        StructuredDocument doc = null;
        try {
            doc = StructuredDocumentFactory.newStructuredDocument (mediaType,
                                                                   ip);
        } catch( Exception e) {
            debug ("pushResponse: malformed response - discard");
            return;
        }
        // XXX: we are doing no document sanity check here.
        // Maybe should we be more careful.
        // to be fixed. lomax@jxta.org
	
	Vector gateways = new Vector();
	int version = 0;
        enum = doc.getChildren ();
        while (enum.hasMoreElements()) {
            TextElement elem = (TextElement) enum.nextElement();
            if (elem.getName().equals(DestPeerIdTag)) {
		destPeer = (String) elem.getValue();
                continue;
            }
            if (elem.getName().equals(RoutingPeerIdTag)) {
                routingPeer = (String) elem.getValue();
                continue;
            }

            if (elem.getName().equals(RoutingPeerAdvTag)) {
                routingPeerAdv = (String) elem.getValue();
                continue;
            }

            if (elem.getName().equals(GatewayTag)) {
		debug ("Adding " + ((String) elem.getValue()));
                gateways.addElement ((String) elem.getValue());
                continue;
            }

            if (elem.getName().equals(VersionTag)) {
		try {
		    version = Integer.parseInt ((String) elem.getValue());
		} catch (Exception e) {
		    debug ("cannot decode version");
		    continue;
		}
                continue;
            }
        }

	// Check the version of the router.
	if (version < acceptableVersion) {
	    debug ("reject answer. Wrong version " + version);
	    return;
	}
        if ((routingPeer == null) || (destPeer == null)) {
            debug ("pushResponse: malformed response - discard.");
            // Malformed response. Just discard.
            return;
        }

	// If there was an advertisement in the answer, publish it.
	if (routingPeerAdv != null) {
            Advertisement adv = null;
            try {
                adv = AdvertisementFactory.newAdvertisement (
			 new MimeMediaType ("text/xml"),
                         new ByteArrayInputStream (routingPeerAdv.getBytes()));
		
		endpoint.publishPeerAdv (adv);
            } catch (Exception e) {
                debug ("   no advertisement");
            }
	}

	// Check if we already have a route for that peer.
	// A local route has already been set up in processIncoming.

	if (gateways.size() == 0) {
	    gateways = new Vector (1);
	}
	try {
	    gateways.insertElementAt (routingPeer, 0);
	} catch (Exception e) {
	    debug (" gateways.addElementAt (0) failed " + e);
	    return;
	}

	if (!isLocalRoute (destPeer)) {
	    Vector tmp = getRoute (destPeer);
	    if ((tmp == null) || (tmp.size() == 0)) {
		// This is the first time we hear about that peer.
		setRoute (destPeer, gateways);
	    } else {
		// We already have a non direct route for
		// that peer. Check if this route is better.
		if (tmp.size() > gateways.size()) {
		    // This route seems to be better.
		    removeRoute (destPeer);
		    setRoute (destPeer, gateways);
		}
	    }
	}
	
	// If we are a rendezvous, check that we don't have pending queries
	// that need to be answered.

	if (group.isRendezvous()) {

	    enum = removePendingQuery (destPeer);
	    if ((enum == null) || (!enum.hasMoreElements())) {
		// Nope... nothing to do
		return;
	    }

	    if (isLocalRoute (destPeer)) {
		gateways = null;
	    } else {
		gateways = getRoute (destPeer);
	    }

	    // Build the response.
	    ResolverResponse res = null;
	    try {
		StructuredTextDocument tmpDoc = null;
		tmpDoc = (StructuredTextDocument)
		    StructuredDocumentFactory.newStructuredDocument (
                        new MimeMediaType( "text/xml" ), "jxta:PipeResolver");

		Element e = null;
		e = tmpDoc.createElement (VersionTag, Integer.toString (currentVersion));
		tmpDoc.appendChild (e);
		e = tmpDoc.createElement (DestPeerIdTag, destPeer);
		tmpDoc.appendChild (e);
		e = tmpDoc.createElement (RoutingPeerIdTag, localPeerId);
		tmpDoc.appendChild (e);
		Advertisement adv = group.getAdvertisement();
		String localPeerAdv = null;
		try {
		    InputStream ip2 = null;
		    Document doc2 = null;

		    doc2 = adv.getDocument(new MimeMediaType ("text/xml"));
		    ip2 = doc2.getStream();
		    if (ip2 != null) {
			byte[] buffer = new byte [ip2.available()];
			ip2.read (buffer);
			localPeerAdv = new String (buffer);
		    }
		} catch (Exception ee) {
		    debug ("Cannot get local advertisement " + ee);
		}	
		if (localPeerAdv != null) {
		    e = tmpDoc.createElement (RoutingPeerAdvTag, localPeerAdv);
		    tmpDoc.appendChild (e);
		}

		String gateway = null;
		if ((gateways != null) && (gateways.size() > 0)) {
		    for (int i = 0; i < gateways.size(); ++i) {
			try {
			    gateway = (String) gateways.elementAt (i);
			    e = tmpDoc.createElement (GatewayTag, gateway);
			    tmpDoc.appendChild (e);
			} catch (Exception e1) {
			    continue;
			}
		    }
		}
			
		StringWriter dumped = new StringWriter();
		tmpDoc.sendToWriter( dumped );

		res = new ResolverResponse (name,
					    null,
					    null,
					    0,
					    dumped.toString() );
	    } catch (Exception e) {
		// Too bad.. something failed. Discard
		return;
	    }

	    // Send the response to all of the pending queries
	    PendingQuery pending = null;
	    while (enum.hasMoreElements()) {
		
		try {
		    pending = (PendingQuery) enum.nextElement();
		    debug ("   sending pending response to " +
			   pending.requestingPeer);
		    resolver.sendResponse (pending.requestingPeer, res);
		} catch (Exception e) {
		    continue;
		}
	    }
	}
	return;
    }

    /**
     * Process the Query, and genrate response
     * @param GenericResolverMsg the query to process
     * @return ResolverResponseMsg "Response"
     */

    public ResolverResponseMsg processQuery(ResolverQueryMsg query)
        throws NoResponseException,
	       IOException,
	       ResendQueryException,
               DiscardQueryException {

        debug ("processQuery starts");

        MimeMediaType mediaType = null;
        try {
            mediaType = new MimeMediaType ("text/xml");
        } catch (RuntimeException e) {
	    debug ("Malformed query [1] " + e);
            throw new IOException ();		
        }
        ByteArrayInputStream ip =
                    new ByteArrayInputStream (query.getQuery().getBytes());

        StructuredTextDocument doc = null;
        try {
            doc = (StructuredTextDocument)
                StructuredDocumentFactory.newStructuredDocument (mediaType,
                                                                 ip);
        } catch( Exception e ) {
            debug ("processQuery: malformed request [2] " + e);
            throw new IOException();
        }
        // XXX: we are doing no document sanity check here.
        // Maybe should we be more careful. to be fixed. lomax@jxta.org

        String pId = null;

        Enumeration enum = doc.getChildren ();

        while (enum.hasMoreElements()) {
            TextElement elem = (TextElement) enum.nextElement();
            if (elem.getName().equals(DestPeerIdTag)) {
                pId = (String) elem.getValue();
                break;
            }
        }
        if (pId == null) {
            debug ("processQuery: malformed request, no PeerId.");
            throw new IOException();
        }


        int queryId = query.getQueryId();

	Vector gateways = null;

        // XXX: Only one query at the time...
        // Maybe that should be changed later. lomax@jxta.org

	boolean found = false;
	if (pId.equals (localPeerId)) {
	    found = true;
	} else {
	    if (group.isRendezvous()) {

		if (isLocalRoute (pId)) {
		    found = true;
		} else {
		    gateways = getRoute (pId);
		    if (gateways != null) {
			found = true;
		    }
		}
	    }
	}

	if (!found) {
	    if (!group.isRendezvous()) {
		throw new DiscardQueryException();
	    }
	    if (isPendingQuery (pId)) {
		throw new DiscardQueryException();
	    }
	    if (processNotFound (query.getSrc(), pId)) {
		// if true, processNotFound is asking to resend the
		// request.
		throw new ResendQueryException();
	    } else {
		// if false, just ignore the request
		throw new NoResponseException();
	    }
	}

        try {
            // Build the answer
            doc = (StructuredTextDocument)
            StructuredDocumentFactory.newStructuredDocument (
            new MimeMediaType( "text/xml" ), "jxta:PipeResolver");

            Element e = null;
            e = doc.createElement (VersionTag, Integer.toString (currentVersion));
            doc.appendChild (e);
            e = doc.createElement (DestPeerIdTag, pId);
            doc.appendChild (e);
            e = doc.createElement (RoutingPeerIdTag, localPeerId);
            doc.appendChild (e);

	    Advertisement adv = group.getAdvertisement();
	    String localPeerAdv = null;
	    try {
		InputStream ip2 = null;
		Document doc2 = null;

		doc2 = adv.getDocument(new MimeMediaType ("text/xml"));
		ip2 = doc2.getStream();
		if (ip2 != null) {
		    byte[] buffer = new byte [ip2.available()];
		    ip2.read (buffer);
		    localPeerAdv = new String (buffer);
		}
	    } catch (Exception ee) {
		debug ("Cannot get local advertisement " + ee);
	    }
	    if (localPeerAdv != null) {
		e = doc.createElement (RoutingPeerAdvTag, localPeerAdv);
		doc.appendChild (e);
	    }

	    String gateway = null;
	    if ((gateways != null) && (gateways.size() > 0)) {
		for (int i = 0; i < gateways.size(); ++i) {
		    try {
			gateway = (String) gateways.elementAt (i);
			e = doc.createElement (GatewayTag, gateway);
			doc.appendChild (e);
		    } catch (Exception e1) {
			continue;
		    }
		}
	    }
			
            StringWriter dumped = new StringWriter();
            doc.sendToWriter( dumped );

            ResolverResponse res = new ResolverResponse (name,
                                                         null,
                                                         null,
                                                         queryId,
                                                         dumped.toString() );

            debug ("processQuery: sending answer.");
            return res;

        } catch (Exception ee) {
            debug ("processQuery: error while processing query: " + ee);
            throw new IOException();
        }
    }

    // If this peer is a rendez-vous, cache the query, and ask the ResolverService
    // to resolve that request as for ourselves.

    private boolean processNotFound (String requestingPeer, String requestedPeer) {

	boolean hasToResend = true;

	debug ("processNotFound for " + requestingPeer);
	if (group.isRendezvous()) {
	    if (isPendingQuery (requestedPeer)) {
		// There is already a pending query for that peer. No need to resend
		// the query (that has already been done.
		hasToResend = false;
	    }
	    // Add the current query to the pending queries
	    addPendingQuery (requestingPeer, requestedPeer);
	    return hasToResend;
	} else {
	    // We are not a RendezVous, just ignore.
	    return false;
	}
    }


    protected class PendingQuery {

	protected String requestingPeer = null;
	protected String requestedPeer = null;
	protected long   timeOfQuery = 0;
	protected long   delay;
	private   long InitialDelay = 120000; // 2 Minutes

	public PendingQuery (String requesting, String requested) {

	    requestingPeer = requesting;
	    requestedPeer = requested;
	    timeOfQuery = System.currentTimeMillis();
	    delay = timeOfQuery + InitialDelay;
	}

	public boolean equals (Object obj) {

	    PendingQuery tmp =  (PendingQuery) obj;
	    return ((tmp.requestingPeer.equals (requestingPeer)) &&
		    (tmp.requestedPeer.equals (requestedPeer)));
	}
    }


    private synchronized void addPendingQuery (String requestingPeer,
					       String requestedPeer) {


	if (isPendingQuery (requestingPeer, requestedPeer)) {
	    // already there
	    return;
	}

	debug ("addPendingQuery for " + requestingPeer);
	PendingQuery query = new PendingQuery (requestingPeer,
					       requestedPeer);

	int index = pendingQueries.indexOf (query);
	if (index != -1) {
	    // This query has already been recorded. Just update the time
	    // of the query.
	    try {
		query = (PendingQuery) pendingQueries.elementAt (index);
		query.timeOfQuery = System.currentTimeMillis();
		return;
	    } catch (Exception e) {
		return;
	    }
	}
	// Add this new query
	pendingQueries.addElement (query);

	return;
    }

    private synchronized Enumeration removePendingQuery (String requestedPeer) {

	debug ("removePendingQuery for " + requestedPeer);
	PendingQuery query = null;
	Vector tmp = new Vector();

	for (int i = 0; i < pendingQueries.size(); ++i) {

	    try {
		query = (PendingQuery) pendingQueries.elementAt (i);
		if (query.requestedPeer.equals (requestedPeer)) {
		    // Remove that query
		    try {
			tmp.addElement (query);
			pendingQueries.removeElementAt (i);
			// Decrement i by one, since the vector has been
			// "compacted" by one element due to the removeElementAt(i)
			--i;
		    } catch (Exception e) {
			continue;
		    }
		}
	    } catch (Exception e) {
		continue;
	    }
	}
	if (tmp.size() > 0) {
	    return tmp.elements();
	} else {
	    return null;
	}
    }

    
    private synchronized boolean isPendingQuery (String requestedPeer) {
	
	PendingQuery query = null;
	long currentTime = System.currentTimeMillis();

	for (int i = 0; i < pendingQueries.size(); ++i) {

	    try {
		query = (PendingQuery) pendingQueries.elementAt (i);
		if (query.requestedPeer.equals (requestedPeer)) {
		    if (query.delay <= currentTime) {
			// We allow a new request to be sent
			// Reset delay
			query.delay = currentTime + query.InitialDelay;
			return false;
		    }
		    return true;
		}
	    } catch (Exception e) {
		return false;
	    }
	}
	return false;
    }

    private synchronized boolean isPendingQuery (String requestingPeer, String requestedPeer) {
	
	PendingQuery query = null;

	for (int i = 0; i < pendingQueries.size(); ++i) {

	    try {
		query = (PendingQuery) pendingQueries.elementAt (i);
		if (query.requestedPeer.equals (requestedPeer) &&
		   query.requestingPeer.equals (requestingPeer)) {
		    return true;
		}
	    } catch (Exception e) {
		return false;
	    }
	}
	return false;
    }

    private Vector reverseGateways (Vector src) {
	if ((src == null) || (src.size() == 0)) {
	    return null;
	}
	Vector tmp = new Vector (src.size());
	for (int i = src.size() - 1; i >= 0; --i) {
	    try {
		tmp.addElement (src.elementAt (i));
	    } catch (Exception e) {
		debug ("reverseGateways failed " + e);
	    }
	}
	return tmp;
    }

    // This class is used in order to add routing information
    // to Messangers.

    public class RouterMessanger implements EndpointMessanger {

	protected EndpointMessanger messanger = null;
	protected Endpoint          endpoint = null;
	protected EndpointRouter    router  = null;
	protected String            dstPeer = null;
	protected EndpointRouterMessage routerMsg = null;
	protected String            localPeerId = null;

	public RouterMessanger (EndpointMessanger m,
				String p,
				Endpoint e,
				EndpointRouter r) {

	    messanger = m;
	    endpoint  = e;
	    router    = r;
	    dstPeer   = p;
	    localPeerId = r.localPeerId;

	    routerMsg = new EndpointRouterMessage();
	    debug ("destPeer= " + dstPeer);
	    if (dstPeer != null) {
		routerMsg.setDestPeer (dstPeer);
	    }
	    routerMsg.setSrcPeer (localPeerId);
	}

	public void sendMessage (Message message) throws IOException {
	    
	    EndpointRouterMessage tmp = null;

	    debug ("EndpointRouterMessanger.sendMessage starts");

	    // Add routing info
	    Enumeration enum = message.getNames();
	    if ((enum != null) && (enum.hasMoreElements())) {
		while (enum.hasMoreElements()) {
		    try {
			if (((String) enum.nextElement()).equals
			    (routerMsg.Name)) {
			    // There already is a router message.
			    // Just use it.
			    debug ("There is already a router message");
			    InputStream ip = null;
			    try {
				ip = message.pop (EndpointRouterMessage.Name);
				if (ip == null) {
				    // This should not happen
				    break;
				}
				tmp = new EndpointRouterMessage (ip);
				break;
			    } catch (Exception e) {
				break;
			    }
			}
		    } catch (Exception e) {
			continue;
		    }
		}
	    }

	    if (tmp == null) {
		// No header. Use the default one
		tmp = routerMsg;
	    } else {

		if ((tmp.getDestPeer() == null) ||
		    (tmp.getDestPeer().equals (localPeerId))) {
		    // The router needs to be updated.
		    Vector gateways = tmp.getGateways();
		    if (gateways == null) {
			gateways = new Vector();
		    }
		    gateways.addElement (localPeerId);
		    tmp.setGateways (gateways);

		    debug ("setting destination address to " + dstPeer);
		    if (dstPeer != null) {
			tmp.setDestPeer (dstPeer);
		    }
		}
	    }
	    // Push the router header onto the message.
	    try {
		message.push (tmp.Name, tmp.getInputStream());
	    } catch (Exception e) {
		debug ("Can't push header");
	    }
	    
	    // give the message to the real messanger
	    messanger.sendMessage (message);
	}

	public void close() {
	}

	private void debug (String str) {
	    router.debug ( "[Messanger] " + str);
	}
    }

    public EndpointMessanger getMessanger (String dstPeer,
					   EndpointMessanger messanger,
					   Endpoint          endp) {

	return new RouterMessanger (messanger,
				    dstPeer,
				    endp,
				    this);
    }
					  
}



