/************************************************************************
 *
 * $Id: IncomingMessageTag.java,v 1.7 2001/05/23 01:00:18 lomax Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint;

import java.io.*;
import net.jxta.endpoint.*;

/**
 * An IncomingMessageTag encapsulates the name a chunk of data that has
 * been received.
 * This class also implements the InputStream API so it can directely
 * be used by the message layer to produce the bytes that have been received.
 */

public class IncomingMessageTag implements MessageTag {

    private String name;
    private byte[] srcBytes = null;
    private int startIndex = 0;
    private int endIndex = 0;
    private int length = 0;
    private boolean valid = false;
    private int index = 0;
    private int dataFirstByte = 0;
    private int dataLastByte = 0;
    private int dataSize = 0;

    /**
     * Default constructor
     */
    public IncomingMessageTag(byte[] src, int start, int end) {

	srcBytes = src;
	startIndex = start;
	endIndex = end;
	length = end - start;
	// Decode the data
	buildTag();
    }

    /**
     * set this MessagesTag with the tag. If the tag
     * is invalid, IOException is thrown.
     *
     * @param tag String containing of the tag
     */
    public void setTag (String thisTag)
	throws IOException {
	// The name of an incoming Message tag can not be changed.
	// Just return the exception
	throw new IOException();
    }

    /**
     * Get the String representation of the tag.
     *
     * @return the String representation of the tag.
     */
    public String getTag() {
	return name;
    }

    /**
     * This method needs to be provided in order to implement the MessageTag
     * interface, but since the stream has been set at the construction
     * of this object, it cannot be changed later on. Just throw an
     * exception.
     *
     * @param thisStream InputStream provided by the caller.
     */
    public void setInputStream (InputStream thisStream) throws IOException {
	throw new IOException ();
    }

    /**
     * returns an InputStream to the data of the tag.
     *
     * @return returns an InputStream to the data of the tag.
     */

    public InputStream getInputStream() {

	return new InMsgTagInputStream (srcBytes,
					dataSize,
					startIndex,
					dataFirstByte);
    }


    public InputStream getWireFormatInputStream () {
	// Build an Outgoing Message tag from this tag
	OutgoingMessageTag og = new OutgoingMessageTag (name);
	try {
	    og.setInputStream (getInputStream());
	    return og.getWireFormatInputStream();
	} catch (Exception e) {
	    return null;
	}
    }

    private void buildTag() {

	// This implementation assumes that the various XML tags
	// defining the tag are always defined in the same order:
	// <JxtaMessageTagBody>
	//    <JxtaMessageTagName>
	//              name of the tag
	//    </JxtaMessageTagName>
        //    <JxtaMessageTagData>
	//              data of the tag
	//    </JxtaMessageTagData>
        // </JxtaMessageTagBody>
	//
	// This implementation also assumes that there is no extra space
	// added between tags separator. The reason is that we are
	// not using an XML parser for performance reasons, and this
	// class implements an extremely simple, but rather stupid,
	// parser that would not support anything else.
	//

	valid = false;
	int i = 0;

	// First check this is really a MessageTag.
	String tmp = new String (srcBytes, startIndex + 0,
                                 MessageProto.BodyTagStart.length());
	if (tmp == null) {
	    // Error.. the tag is not a real tag. Just discard
	    return;
	}
	if (!tmp.equals (MessageProto.BodyTagStart)) {
	    // The first tag is not <JxtaMessageTag>. Discard
	    return;
	}
	i += MessageProto.BodyTagStart.length();

	// Get the name of the tag
	tmp = new String (srcBytes, startIndex + i,
			  MessageProto.NameTagStart.length());
	if (tmp == null) {
	    // Error.. the tag is not a real tag. Just discard
	    return;
	}
	if (!tmp.equals (MessageProto.NameTagStart)) {
	    // The first tag is not <JxtaMessageTagName>. Discard
	    return;
	}
	i += MessageProto.NameTagStart.length();
	// Search for the next "<"
	int j;
	for (j = i; j < length; ++j) {
	    if (srcBytes [startIndex + j] == '<') {
		break;
	    }
	}
	if (j == length) {
	    // The termination tag does not exist... Discard the tag
	    return;
	}
	name = new String (srcBytes, startIndex + i, j - i);
	// Check that the next tag is really </JxtaMessageTagName>
	i += name.length();
	tmp = new String (srcBytes, startIndex + i,
			  MessageProto.NameTagEnd.length());
	if (tmp == null) {
	    // Error.. the tag is not a real tag. Just discard
	    return;
	}
	if (!tmp.equals (MessageProto.NameTagEnd)) {
	    // The first tag is not </JxtaMessageTagName>. Discard
	    name = null;
	    return;
	}
	i += MessageProto.NameTagEnd.length();

	// Get the data
	tmp = new String (srcBytes, startIndex + i,
			  MessageProto.DataTagStart.length());
	if (tmp == null) {
	    // Error.. the tag is not a real tag. Just discard
	    return;
	}
	if (!tmp.equals (MessageProto.DataTagStart)) {
	    // The first tag is not <JxtaMessageTagData>. Discard
	    return;
	}
	i += MessageProto.DataTagStart.length();
	dataFirstByte = i;
	// Search for the last instance of </JxtaMessageTagData>
	tmp = new String (srcBytes, startIndex + i, length - i);
	j = tmp.lastIndexOf (MessageProto.DataTagEnd);
	if (j == -1) {
	    // No termination tag is an error.
	    name = null;
	    return;
	}
	dataLastByte = i + j;
	valid = true;
	dataSize = dataLastByte - dataFirstByte;
    }

    public void share() {
	// This implementation is sharable. Nothing has to be done.
    }


    /**
     * This portion is the implementation of the InputStream API.
     */

    private class InMsgTagInputStream extends InputStream {

	private byte[] buffer = null;
	private int dataSize = 0;
	private int index = 0;
	private int startIndex = 0;
	private int dataFirstByte = 0;


	public InMsgTagInputStream (byte[] b, int d, int si, int sf) {

	    this.buffer = b;
	    this.dataSize = d;
	    this.index = 0;
	    this.startIndex = si;
	    this.dataFirstByte = sf;
	}


	public int available() throws IOException {
	    return this.dataSize - this.index;
	}

	public int read () throws IOException {
	    if (this.index >= this.dataSize) {
		// No more to be read. Return -1
		return -1;
	    }
	    int res = this.buffer [this.startIndex + this.dataFirstByte +
				  this.index++];
	    if (res < 0) {
		return 256 + res;
	    } else {
		return res;
	    }
	}
    }
}













