/************************************************************************
 *
 * $Id: MessageImpl.java,v 1.11 2001/06/10 02:24:41 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint;
import net.jxta.endpoint.*;

import java.io.*;
import java.util.*;

/**
 * A MessageImpl contains data and protocol specific information that
 * are exchanged between peer protocols. A MessageImpl consists of a
 * queue of data chunks that are assiciated with a MessageImplTag.
 * A MessageImpl is also associated to the TransportAddress of the
 * sender and the receiver of the message. Note that several different chunks
 * of data associated to the same tag can be pushed. In that case, several pops
 * will be needed to retrieve all the chunks.
 *
 */

public class MessageImpl implements Message {
    
    private static boolean debug = false;
    
    private Vector tags = null;
    private byte[] srcBytes = null;
    private int srcBytesOffset = 0;
    private int srcBytesLen = 0;
    private boolean valid = false;
    private EndpointAddress destAddress = null;
    private EndpointAddress srcAddress = null;
    
    
    /**
     * Constructor for outgoing message. Note that an outgoing message
     * is always valid.
     */
    public MessageImpl() {
        tags = new Vector();
        valid = true;
    }
    
    
    public MessageImpl( byte[] src ) {
        this( src, 0, src.length );
    }
    
    public MessageImpl( byte[] src, int offset ) {
        this( src, offset, src.length - offset );
    }
    
    /**
     * Constructor for incoming messages
     */
    public MessageImpl(byte[] src, int offset, int len) {
        tags = new Vector();
        srcBytes = src;
        
        if( null == src )
            throw new NullPointerException( "src is null!" );
        
        if( len < 0 )
            throw new IllegalArgumentException( "len is negative" );
        
        if( (offset < 0) || (offset >= srcBytes.length) )
            throw new IllegalArgumentException( "offset is not in src" );
        
        // Note: offset or count might be near -1>>>1.
        if (offset > src.length - len)
            throw new IllegalArgumentException( "range does not lie within src" );
        
        srcBytesOffset = offset;
        srcBytesLen = len;
        parse();
        
        if( !valid )
            debug( "[MessageImpl] invalid message!" );
    }
    
    /**
     * Push an a new InputStream onto the message associated with a symbolic name.
     *
     * @param name is the symbolic name associated with the InputStream
     * @param inputStream is the new InputStream to push onto the message
     * @throws IOException when the inputStream is not available.
     */
    public void push(String name, InputStream inputStream)
    throws IOException {
        
        name = new String(name);
        MessageTag tag = new OutgoingMessageTag(name);
        push(tag, inputStream);
    }
    
    
    /**
     * Push a chunk of data associated with its tag onto the message.
     * Throws IOException when the data cannot be read.
     *
     * @param tag tag associated with the data to be pushed
     * @param data InputStream to the data to be pushed
     */
    private synchronized void push(MessageTag tag, InputStream data)
    throws IOException {
        MessageTag newTag = new OutgoingMessageTag();
        try {
            newTag.setTag(tag.getTag());
            newTag.setInputStream(data);
            tags.addElement(newTag);
        } catch (Exception e) {
            throw new IOException();
        }
    }
    
    
    /**
     * Pop an InputStream associated with a symbolic name
     *
     * @param name is the symbolic name associated with the InputStream to get.
     * @return InputStream returns an InputStream associated with name.
     * @throws IOException when the InputStream is not available.
     */
    
    public synchronized InputStream pop(String search)
    throws IOException {
        
        if (!valid) {
            return null;
        }
        
        int nbOfTags = tags.size();
        int i = 0;
        MessageTag foundTag = null;
        
        try {
            for (i = 0; i < nbOfTags; ++i) {
                if (((MessageTag)
                tags.elementAt(i)).getTag().equals(search)) {
                    
                    foundTag = (MessageTag) tags.elementAt(i);
                    break;
                }
            }
            if (foundTag == null) {
                // REVIEW 2001-05-30 <nelson> -- not clear that throwing
                //   an IOException here is right, this is not an IO problem.
                debug(  "Tag not found. : " + search );
                throw new IOException( "Tag not found. : " + search );
            }
            tags.removeElementAt(i);
        } catch (Exception e) {
            throw new IOException(e.getMessage());
        }
        return foundTag.getInputStream();
    }
    
    /**
     * Get the size of a particular tag's data in the message.
     *
     * @param atag the name of the tag whose size to get
     * @return the size of the data in the message for the specified tag
     */
    public int getSize(String atag) {
        
        Enumeration tagenum = tags.elements();
        MessageTag tag = null;
        int size = 0;
        while (tagenum.hasMoreElements()) {
            tag = (MessageTag) tagenum.nextElement();
            if (tag.getTag().equals(atag)) {
                // share() will cause all data to be cached
                tag.share();
                InputStream is = tag.getInputStream();
                if (is == null)
                    continue;
                try {
                    size = is.available();
                } catch (IOException ex) {
                    ex.printStackTrace();
                }
                break;
            }
        }
        return size;
    }
    
    /**
     * Returns an Enumeration containing all the MessageTag names in the
     * message.
     *
     * @return Enumeration of MessageTag objects.
     */
    public Enumeration getNames() {
        
        if (!valid) {
            return null;
        }
        Enumeration tagenum = tags.elements();
        Vector tmp = new Vector();
        
        while (tagenum.hasMoreElements()) {
            try {
                tmp.addElement(((MessageTag) tagenum.nextElement()).getTag());
            } catch (Exception e) {
                break;
            }
        }
        return tmp.elements();
    }
    
    /**
     * Duplicate the message.
     */
    public Message dup() {
        
        MessageImpl msg = new MessageImpl();
        // Make all the tags sharable
        Enumeration tagenum = tags.elements();
        
        MessageTag tag = null;
        while (tagenum.hasMoreElements()) {
            try {
                tag = (MessageTag) tagenum.nextElement();
                tag.share();
                msg.push(tag.getTag(), tag.getInputStream());
            } catch (Exception e) {
                return null;
            }
        }
        return msg;
    }
    
    /**
     * Return an InputStream on the entire message. The bytes returned
     * by the InputStream are formatted in the standard JXTA Endpoint
     * message format.
     *
     * This method is used to get the bytes to be sent (outgoing messages).
     */
    public InputStream getInputStream() {
        if (!valid) {
            return null;
        }
        return new MessageInputStream(this);
    }
    
    /**
     * Parses an incomming message
     */
    private void parse() {
        // FIXME    20010608    bondolo@jxta.org srcBytesOffset & srcBytesLen
        // are ignored in this routine! This really really must be fixed!
        valid = false;
        try {
            int i = -1;
            
            // Find the XML header.
//            for ( int ii = srcBytesOffset; ii < (srcBytesOffset + srcBytesLen); ++ii) {
            for ( int ii = 0; ii < srcBytes.length; ++ii) {
                if (srcBytes [ii] == '<') {
                    i = ii;
                    break;
                }
            }
            
            if ( -1 == i) {
                // There is no XML tag at all. Invalid message
                debug( "message invalid : No XML tag" );
                return;
            }
            
            String tmp = new String(srcBytes, i,
            MessageProto.XmlHeader.length());
            
            if (!tmp.equals(MessageProto.XmlHeader)) {
                // Wrong JXTA Message
                debug( "message invalid : first tag is not header" );
                return;
            }
            i += MessageProto.XmlHeader.length();
            
            // Get the Jxta Message version
            tmp = new String(srcBytes, i,
            MessageProto.JxtaVersion.length());
            
            if (!tmp.equals(MessageProto.JxtaVersion)) {
                // Wrong JXTA Message
                debug( "message invalid : incorrect version" );
                return;
            }
            i += MessageProto.JxtaVersion.length();
            
            // Get the destination address
            tmp = new String(srcBytes, i,
            MessageProto.DestTagStart.length());
            
            if (!tmp.equals(MessageProto.DestTagStart)) {
                // Wrong JXTA Message
                debug( "message invalid : didnt find dest start tag" );
                return;
            }
            i += MessageProto.DestTagStart.length();
            // Find the next "<"
            int j = i;
            for (j = i; j < srcBytes.length; ++j) {
                if (srcBytes[j] == '<') {
                    break;
                }
            }
            
            if (j == srcBytes.length) {
                // There was no termination tag. The message is invalid.
                // Discard.
                debug( "message invalid : didnt find dest end tag (no tags at all)" );
                return;
            }
            destAddress = new Address(new String(srcBytes, i, j - i));
            
            i = j;
            
            tmp = new String(srcBytes, i, MessageProto.DestTagEnd.length());
            
            if (!tmp.equals(MessageProto.DestTagEnd)) {
                // Wrong JXTA Message
                debug( "message invalid : didnt find dest end tag" );
                return;
            }
            i += MessageProto.DestTagEnd.length();
            
            // Get the source address
            tmp = new String(srcBytes, i,
            MessageProto.SrcTagStart.length());
            
            if (!tmp.equals(MessageProto.SrcTagStart)) {
                // Wrong JXTA Message
                debug( "message invalid : didnt find src start tag" );
                return;
            }
            i += MessageProto.SrcTagStart.length();
            // Find the next "<"
            j = i;
            for (j = i; j < srcBytes.length; ++j) {
                if (srcBytes[j] == '<') {
                    break;
                }
            }
            
            if (j == srcBytes.length) {
                // There was no termination tag. The message is invalid.
                // Discard.
                debug( "message invalid : didnt find src end tag (no XML tag)" );
                return;
            }
            srcAddress = new Address(new String(srcBytes, i, j - i));
            i = j;
            
            tmp = new String(srcBytes, i, MessageProto.SrcTagEnd.length());
            
            if (!tmp.equals(MessageProto.SrcTagEnd)) {
                // Wrong JXTA Message
                debug( "message invalid : didnt find src end tag" );
                return;
            }
            i += MessageProto.SrcTagEnd.length();
            
            // The message digest has already been processed. Just skip it
            // now.
            tmp = new String(srcBytes, i, srcBytes.length - i);
            j = tmp.lastIndexOf(MessageProto.DigestTagEnd);
            if (j == -1) {
                // Tag is missing. That should not happen since the
                // the message digest has been already processed
                debug( "message invalid : didnt find digest end tag" );
                return;
            }
            i += j + MessageProto.DigestTagEnd.length();
            
            
            String bodies = new String(srcBytes, i, srcBytes.length - i);
            
            int startIndex = i;
            int endIndex = i;
            j = 0;
            int tmpIndex = 0;
            
            // Parse all the message tags
            while (j < bodies.length()) {
                tmpIndex = bodies.indexOf(MessageProto.BodyTagStart, j);
                
                if (tmpIndex == -1) {
                    // No more tags
                    valid = true;
                    return;
                }
                startIndex = i + tmpIndex;
                tmpIndex = bodies.indexOf(MessageProto.BodyTagEnd, j);
                
                if (tmpIndex == -1) {
                    // The body is not terminated.
                    debug( "message invalid : didnt find a body end tag" );
                    return;
                }
                endIndex = i + tmpIndex + MessageProto.BodyTagEnd.length();
                
                // Create an IncomingMessageTag
                IncomingMessageTag tag = new IncomingMessageTag(srcBytes,
                startIndex,
                endIndex);
                
                tags.addElement(tag);
                j = tmpIndex + MessageProto.BodyTagEnd.length();
            }
            valid = true;
            debug( srcBytes.length + " byte msg parsed. " + tags.size() + " tags" );
        }
        catch( Exception all ) {
            debug( "message invalid : probably an index error" );
            return; // this should mostly occur due to string index errors
        }
    }
    
    /**
     * The following public API is not intended to the application
     * but only to the endpoint implementation. Maybe in the future,
     * we will provide this API to the applications if there is a need
     * for it.
     */
    
    public void setDestAddress(EndpointAddress addr) {
        try {
            destAddress = new Address(addr.toString());
        } catch (Exception e) {
        }
    }
    
    public void setSrcAddress(EndpointAddress addr) {
        try {
            srcAddress = new Address(addr.toString());
        } catch (Exception e) {
        }
    }
    
    public EndpointAddress getDestAddress() {
        return destAddress;
    }
    
    public EndpointAddress getSrcAddress() {
        return srcAddress;
    }
    
    public Vector getTags() {
        return tags;
    }

    private void debug(String str) {
        if (debug) {
            System.out.println("[MessageImpl] " + str);
        }
    }
}
