/************************************************************************
 *
 * $Id: MessageInputStream.java,v 1.10 2001/06/10 02:24:41 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint;
import net.jxta.endpoint.*;

import java.io.*;
import java.util.*;

/**
 * This class implements the formatting of a Message into
 * the Endpoint JXTA Message representation.
 * 
 * This class extends the InputStream interface so it can
 * be used directely in order to get the bytes of the message.
 */

public class MessageInputStream extends InputStream {

    private static String jxtaMessageVersion = "1.0";

    private MessageImpl msg = null;
    private String header = null;
    private EndpointAddress srcUri = null;
    private EndpointAddress destUri = null;
    private Vector msgTags = null;
    private Vector msgTagsIp = null;
    private int messageSize = 0;
    private int index = 0;
    private int currentTagIndex = 0;
    private MessageTag currentTag = null;
    private InputStream currentTagInputStream = null;
    private String digest = null;

    private boolean valid = false;

    /**
     * Constructor. This constructor is intented to be used for
     * outgoing messages.
     */
    public MessageInputStream (MessageImpl m) {

	msg = m;
	msgTags = m.getTags();
	msgTagsIp = new Vector (msgTags.size());
	for (int i = 0; i < msgTags.size(); ++i) {
	    try {
		MessageTag tag = (MessageTag) msgTags.elementAt (i);
		msgTagsIp.addElement (tag.getWireFormatInputStream());
	    } catch (Exception e) {
		// Not much we can do
		System.err.println("MessageInputStream constructor: " + e);
	    }
	}
	setDestAddress (m.getDestAddress());
	setSrcAddress (m.getSrcAddress());
    }

    /**
     * Build the wire format of the message. This has to be done
     * before the InputStream is consumed.
     */
    public void buildWireMessage() {
	// Format the message
	try {
	    makeHeader();
	} catch (Exception e) {
	    // The message cannont be built. This is probably due to
	    // a bug -- just print out an error.
	    System.err.println ("MessageInputStream.buildWireMessage failed");
	    return;
	}
	messageSize = header.length() + computeBodiesSize();
	valid = true;
    }

    /**
     * Sets the destination address.
     */
    public void setDestAddress (EndpointAddress d) {
	try {
	    destUri = new Address (d.toString());
	} catch (Exception e) {
	}
    }
 
    /**
     * Sets the source address.
     */
    public void setSrcAddress (EndpointAddress s) {
	try {
	    srcUri = new Address(s.toString());
	} catch (Exception e) {
	}
    }

    /**
     * Get the destination address
     */
    public EndpointAddress getDestAddress () {
	return destUri;
    }

    /**
     * Get the source address
     */
    public EndpointAddress getSrcAddress () {
	return srcUri;
    }

    /**
     * Builds the message header.
     */
    private void makeHeader () throws IOException {
      // FIXME The digest is not yet implemented. makeDigest is
      // called here to make the compiler happy, but should not be the
      // the right place to invoke it.

	digest = makeDigest();

	header =
	    MessageProto.XmlHeader +
	    MessageProto.JxtaVersion +
	    MessageProto.DestTagStart +
	    destUri.toString() +
	    MessageProto.DestTagEnd +
	    MessageProto.SrcTagStart +
	    srcUri.toString() +
	    MessageProto.SrcTagEnd +
	    MessageProto.DigestTagStart +
	    digest +
	    MessageProto.DigestTagEnd;
    }

    /**
     * Computes the total size of all the bodies contained in the
     * message.
     */
    private int computeBodiesSize () {

	int size = 0;
	int tmpSize = 0;
	int nbOfTags = msgTags.size();

	for (int i=0; i < nbOfTags; ++i) {
	    tmpSize = 0;
	    try {
		MessageTag tag = (MessageTag) msgTags.elementAt (i);
		InputStream ip = (InputStream) msgTagsIp.elementAt (i);
		try {
		    tmpSize += ip.available();
		} catch (IOException e) {
		    // An exception accessing the tag's InputStream
		    // should not happen. If it happens, the tag
		    // is simply ignored (i.e. not sent).
		    continue;
		}
		size += tmpSize;

	    } catch (Exception e) {
		// This should not happen since we are testing
		// in the while loop that there is still elements
		// in the Vector.
		break;
	    }
	}
	return size;
    }

    /**
     * This is the implementation of the InputStream API
     */
    public int available() throws IOException {

	if (!valid) {
	    buildWireMessage();
	}

	if (messageSize == 0) {
	    // This MessageInputStream has not been initialized
	    // using buildWireMessage(). Return an IOException.
	    throw new IOException();
	}
	return messageSize - index;
    }

    public int read() throws IOException {
	// Is the current index still in the header section ?
	if (index < header.length()) {
	    try {
		return header.charAt (index++);
	    } catch (Exception e) {
		throw new IOException();
	    }
	}

	int nbOfTags = msgTags.size();

	// Is the current index in a tag ?
	if (currentTagIndex < nbOfTags) {
	    // Is currentTag up to date ?
	    if (currentTag == null) {
		try {
		    currentTag = (MessageTag)
                          msgTags.elementAt (currentTagIndex);

		    currentTagInputStream = (InputStream)
                          msgTagsIp.elementAt (currentTagIndex);

		} catch (Exception e) {
		    // This should not happen.
		    throw new IOException ();
		}
	    }
	    // Is there still elements in the tag ?
	    try {
		if (currentTagInputStream.available() == 0) {
		    // There is no more bytes to be consummed on that
		    // InputStream. Get the next tag.
		    currentTagIndex++;
		    currentTag = null;
		    currentTagInputStream = null;
		    return read();
		}
		int tmp = currentTagInputStream.read();
		index++;
		return tmp;

	    } catch (Exception e) {
		throw new IOException();
	    }
	}
	// No more bytes.. return -1
	return -1;
    }

    private String makeDigest () {
	// XXX: This is not yet implemented
	return "WARNING DIGEST IS NOT YET IMPLEMENTED";
    }
}





