/************************************************************************
 *
 * $Id: OutgoingMessageTag.java,v 1.7 2001/05/23 01:00:18 lomax Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint;

import java.io.*;
import net.jxta.endpoint.*;

/**
 * An OutgoingMessageTag encapsulates the name a chunk of data that has to
 * be sent with a Message.
 * This class also implements the InputStream API so it can directely
 * be used by the message layer to produce the bytes to be sent. This
 * implementation assumes that the Message layer and/or the Transport
 * layer does either synchronously send the message or is responsible to
 * copy the data into an private buffer, letting the caller re-use the stream
 * if needed. Doing that allows in the best case situation to avoid an
 * extra copy of the user data, which can be very expensive, both for performance
 * and memory utilization.
 */

public class OutgoingMessageTag implements MessageTag {

    private String tagName;
    private InputStream inputStream = null;
    private int tagSize = 0;
    private boolean initialized = false;
    private int footerIndex = 0;
    private int index = 0;
    private String header = null;
    private static String footer = MessageProto.DataTagEnd +
	                           MessageProto.BodyTagEnd;
    private byte[] buffer = null;
    private boolean shared = false;
    private boolean sharable = true;

    /**
     * Default constructor
     */
    public OutgoingMessageTag() {
    }


    /**
     * Constructor. Build a message tag from its symbolic name.
     *
     * @param thisTag contains the symbolic name of this MessageTag.
     */
    public OutgoingMessageTag(String thisTag) {
	tagName = new String (thisTag);
    }

    /**
     * set this MessageTag with the tag. If the tag
     * is invalid, IOException is thrown.
     *
     * @param tag String containing of the tag
     */
    public void setTag (String thisTag)
	throws IOException {
	// We make a copy of the String so the caller can change the content or
	// of the variable holding the parameter.
	tagName = new String (thisTag);
    }

    /**
     * Get the String representation of the tag.
     *
     * @return the String representation of the tag.
     */
    public String getTag() {
	return tagName;
    }

    /**
     * Set the user InputStream. This implementation of the MessageTag
     * makes the assumption that the message will be either sent asynchronously
     * or copied by the Endpoint and transport layer, so there is no need
     * to worry here about sharing the stream with the user.
     *
     * @param thisStream InputStream provided by the caller.
     */
    public void setInputStream (InputStream thisStream) throws IOException {
	inputStream = thisStream;
	tagSize = thisStream.available();
    }

    /**
     * buildTag builds the wire format of the tag. This must be called before
     * actually reading bytes from the inputStream.
     */
    private void buildTag() {

	header = MessageProto.BodyTagStart + MessageProto.NameTagStart +
                 tagName + MessageProto.NameTagEnd + MessageProto.DataTagStart;
	try {
	    tagSize = header.length() + inputStream.available() +
	              footer.length();

	} catch (Exception e) {
	    // An exception here is fatal. The size of the tag is set to zero,
	    // therefore, the tag will not be sent.
	    tagSize = 0;
	}
	sharable = false;
    }

    /**
     * This this tag shareable. This implies buffering the data
     */
    public void share () {

	if (shared) {
	    return;
	}

	if (!sharable) {
	    return;
	}

	// Get the user data
	int size = 0;
	int got = 0;
	try {
	    size = inputStream.available();	
	    buffer = new byte [size];
	    while (got < size) {
		got += inputStream.read (buffer, got, size - got);
	    }
	    shared = true;
	} catch (Exception e) {
	    System.err.println("OutgoingMessageTag: " + e);
	}
    }

    /**
     * returns an InputStream to the data of the tag.
     *
     * @return returns an InputStream to the data of the tag.
     */

    public InputStream getInputStream () {
	if (!sharable) {
	    return null;
	}

	if (shared) {
	    return new ByteArrayInputStream (buffer);
	} else {
	    sharable = false;
	    return inputStream;
	}
    }

    public InputStream getWireFormatInputStream() {

	// If the tag has not yet been built, build it first.
	if (!initialized) {
	    buildTag();
	    initialized = true;
	    sharable = false;
	}
	if (shared) {
	    return new OutMsgTagWireInputStream (tagSize,
						 new ByteArrayInputStream (buffer),
						 header,
						 footer);
	} else {
	    return new OutMsgTagWireInputStream (tagSize,
						 inputStream,
						 header,
						 footer);
	}
    }


    /**
     * This portion is the implementation of the InputStream API.
     */
    public class OutMsgTagWireInputStream extends InputStream {

	private int tagSize = 0;
	private int index = 0;
	private String header = null;
	private InputStream inputStream = null;
	private int footerIndex = 0;
	private String footer = null;

	public OutMsgTagWireInputStream (int ts,
					 InputStream ip,
					 String h,
					 String f) {

	    tagSize = ts;
	    inputStream = ip;
	    header = h;
	    footer = f;
	}

	/**
	 * Return what remains to in the InputStream
	 */

	public int available() throws IOException {
	    return tagSize - index;
	}

	public int read () throws IOException {
	    
	    // Are we still providing the bytes of the header ?
	    if (index < header.length()) {
		try {
		    return header.charAt (index++);
		} catch (Exception e) {
		    throw new IOException();
		}
	    }

	    int tmp = 0;
	    
	    // Is there still bytes in the tag's InputStream ?
	    if (inputStream.available() > 0) {
		tmp = inputStream.read();
		++index;
		return tmp;
	    }

	    // All what remains is the termination tag

	    if (footerIndex < footer.length()) {
		try {
		    tmp = footer.charAt (footerIndex++);
		    ++index;
		    return tmp;
		} catch (Exception e) {
		    // This is not an exception, but rather
		    // the information that this is the end.
		    // Just return -1
		    return -1;
		}
	    }
	    return -1;
	}
    }


}












