/************************************************************************
 *
 * $Id: HttpTransport.java,v 1.71 2001/06/10 02:24:41 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.endpoint.http;

import java.net.*;
import java.io.*;
import java.util.*;

import net.jxta.endpoint.*;
import net.jxta.protocol.*;

import net.jxta.impl.endpoint.*;
import net.jxta.impl.protocol.*;

import net.jxta.impl.cm.*;


/***
 * This class implements the HTTP Transport Protocol
 */

public class HttpTransport implements TransportProtocol, Runnable {
    
    private static final String Register = "reg";
    private static final String Receive = "rec";
    private static final String Send = "snd";
    private static final String Ping = "ack";
    public  static final String HttpSpool = "httpspool";
    
    private static final String HttpContentLength = "Content-length:";
    
    public static final Cm cm = new Cm("HttpTransport");
    private static final String PingOk = "Jxta Ping OK";
    private static final int LingerDelay = 120000;
    private static final int SocketTimeout = 120000;
    
    private static final int pollingDelay = 20000; // 20 seconds
    
    public static final String MagicWord = "JxtaHttpClient";
    
    private static final long MaxMsgPoll = 10;
    
    public  static final int DefaultNbOfUnicastThreads = 1;
    public  static final int MaxNbOfUnicastThreads = 10;
    public  static final int MaxNbOfUnicastKeepAliveThreads = 4;
    public  static final int MaxNbOfPendingSockets = 20;
    
    public static boolean debugFlag = true;
    private static Thread       serverThread = null;
    private static Thread       pollingThread = null;
    public  static int          serverSocketPort;
    
    private static boolean doPropagate = false;
    
    private static String localPeerId = null;
    
    private static Vector routers = null;
    private static String proxyName = null;
    private static String serverName = null;
    
    private static Vector staticHttps = null;
    private static Vector dynamicHttps = null;
    
    public ServerSocket         serverSocket;
    private InetAddress         usingInterfaceAddr;
    
    private Endpoint            endpoint = null;
    private Address localAddress = null;
    private Address localServerAddress = null;
    private Address localClientAddress = null;
    private EndpointAdvertisement epadv= null;
    
    private boolean polling = false;
    
    private IncomingUnicastThreads unicastThreads = null;
    /**
     * This is a portion of the API which might not be useful when
     * the configuration part of the JXTA platform will be completed.
     * XXX: to revisit.
     */
    
    
    public HttpTransport(EndpointAdvertisement a) {
        
        // We receive our initial configuration via an advertisement.
        // If we make adjustments, we must update the advertisement, so that
        // it reflects reality now, rather than configuration data.
        
        epadv = a;
        HTTPAdv adv = (HTTPAdv) epadv.getTransportAdvertisement();
        serverSocketPort = new Integer(adv.getPort()).intValue();
        if (adv.getRouterEnabled()) routers = adv.getRouters();
        if (adv.getProxyEnabled())  proxyName = adv.getProxy();
        if (adv.getServerEnabled()) serverName = adv.getServer();
        
        // Apply our current sanity rules. They are:
        // If we need a router we cannot be a routing server.
        // If we do not need a router, then we cannot need a proxy.
        //
        // If there is a contradiction in the config, we give priority to
        // the need-a-router setting. The reason for this choice is pragmatic:
        // If there is a proxy settig but no router setting, we do not
        // have the info to make-up a router setting, but we can ignore
        // the proxy. As for the server, well, it's a toss-up, and if the
        // user does not seem to know what he's doing; he probably did not
        // realy need to set-up a server.
        
        if (routers == null) {
            if (proxyName != null) {
                System.err.println("HttpTransport: this system does not use "
                + "an external router; ignoring proxy "
                + "settings.");
            }
            proxyName = null;
        } else {
            if (serverName != null) {
                System.err.println("HttpTransport: this system needs an "
                + "external router. It cannot be a routing "
                + "server.");
            }
            serverName = null;
        }
        if (adv.getPropagateEnabled()) {
            System.err.println("HttpTransport: the propagate option is "
            + "deprecated. Setting ignored.");
        }
        
        // Now prune this adv from the useless junk; that is, the items we
        // did not load because they were reported disabled, or that we
        // decide not to use based on our current behaviour.
        
        if (routers == null) adv.setRouters(null);
        if (serverName == null) adv.setServer(null);
        if (proxyName == null) adv.setProxy(null);
        adv.setPropagateEnabled(false);
        
        localPeerId = adv.getPeerId();
        
        debug("Configuring HTTP Transport");
        debug("Port:"+ serverSocketPort);
        
        debug("HTTP server: "
        + (serverName == null ? "none" : serverName));
        debug("HTTP routers: "
        + (routers == null ? "none" : routers.toString()));
        debug("HTTP proxy: "
        + (proxyName == null ? "none" : proxyName));
        
    }
    
    
    public int getPort() {
        return serverSocketPort;
    }
    
    /*
     * XXX: this will need to be revisited when the configuration framework
     * will be completed.
     * For the time being, port numbers are hardcoded.
     */
    
    public void init(Endpoint ep) {
        
        
        debugFlag = ep.debugFlag;
        
        // First create our spool folder
        if (!initDir(HttpSpool)) {
            return;
        }
        
        endpoint = ep;
        
        try {
            usingInterfaceAddr = InetAddress.getLocalHost();
            debug( "Using interface : " +  usingInterfaceAddr.getHostAddress() );
        } catch (Exception e) {
            System.err.println(e);
        }
        
        if (serverName != null) {
            // Server side
            System.out.println("HTTP: Starting server " + serverName);
            
            int port = serverSocketPort;
            
            /* Open the incoming socket. */
            try {
                serverSocket = new ServerSocket(port);
            } catch (IOException e1) {
                System.err.println("[1] Canot create ServerSocket on port " +
                port + ": " + e1.toString());
            } catch (SecurityException e2) {
                System.err.println("[2] Cannot create ServerSocket on port " +
                port + ": " + e2.toString());
            }
            
            // Builds the localAddresses
            localServerAddress = new Address();
            try {
                localServerAddress.setProtocolName("http");
                localServerAddress.setProtocolAddress(serverName);
                if (serverName != null) {
                    epadv.setEndpointAddress(localServerAddress.toString());
                }
            } catch (Exception e) {
                // Not much can be done here. Just print out an error
                debug("HttpTransport init: " + e);
                return;
            }
            localAddress = localServerAddress;
            
            // Cleanup the spooler. All message waiting for clients are removed.
            // Messages can be lost that way, but, that messages can be lost anyway.
            // XXX: something less brutal can be done, but this seems to be reasonable
            // for the time being. lomax@jxta.org
            
            cleanupAllClients();
            
            // Create the incoming thread poll.
            unicastThreads = new IncomingUnicastThreads(this, DefaultNbOfUnicastThreads);
            
            // Create listening Thread
            serverThread = new Thread(this,  "HttpTransport listening thread" );
            serverThread.start();
            Thread.yield();
            
        } else {
            // Client side
            localClientAddress = new Address();
            try {
                localClientAddress.setProtocolName("http");
                localClientAddress.setProtocolAddress(MagicWord +
                localPeerId.substring(localPeerId.indexOf("//") + 2));
                if (serverName == null) {
                    epadv.setEndpointAddress(localClientAddress.toString());
                }
            } catch (Exception e) {
                // Not much can be done here. Just print out an error
                debug("HttpTransport init: " + e);
                return;
            }
            
            // Default address is server address
            localAddress = localClientAddress;
            
            staticHttps = new Vector();
            dynamicHttps = new Vector();
            
            if (routers != null) {
                Enumeration e = routers.elements();
                while (e.hasMoreElements()) {
                    String routerName = (String) e.nextElement();
                    debug("HTTP: Starting connection with " + routerName);
                    addStaticHttp(routerName);
                }
            }
            
            if (proxyName != null) {
                debug("HTTP: Using proxy " + proxyName);
            }
        }
    }
    
    
    public HttpTransport() {
    }
    
    
    public void run() {
        
        
        Thread current = Thread.currentThread();
        
        if (current.equals(serverThread)) {
            runServer();
        }
        
        if (current.equals(pollingThread)) {
            // This thread is polling thread
            polling = true;
            runPolling();
        }
    }
    
    
    public void runServer() {
        
        Socket socket = null;
        
        debug("Server is ready to receive request");
        while (true) {
            try {
                socket = serverSocket.accept();
            } catch (IOException e1) {
                System.err.println("[1] ServerSocket.accept() on port " +
                serverSocketPort +
                " has failed with: " + e1.toString());
                System.out.println("sockets.size= " + unicastThreads.sockets.size());
                try {
                    if (socket != null) {
                        socket.close();
                    }
                } catch (Exception e3) {
                    debug("runServer [1] failed with " + e3);
                }
                continue;
            } catch (SecurityException e2) {
                System.err.println("[2] ServerSocket.accept() on port " +
                serverSocketPort +
                " has failed : " + e2.toString());
                System.out.println("sockets.size= " + unicastThreads.sockets.size());
                try {
                    if (socket != null) {
                        socket.close();
                    }
                } catch (Exception e3) {
                    debug("runServer [1] failed with " + e3);
                }
                continue;
            }
            processReceivingSocket(socket);
        }
    }
    
    public void runReceive(Socket inputSocket) {
        
        InputStream inputStream = null;
        OutputStream outputStream = null;
        
        try {
            inputSocket.setSoLinger(true, LingerDelay);
            inputSocket.setSoTimeout(SocketTimeout);
        } catch (Exception e) {
            debug("runReceive: setting socket failed " + e);
        }
        
        try {
            inputStream = inputSocket.getInputStream();
            
            byte[] buffer = null;
            
            String packet = getResult(inputStream);
            
            if (packet == null) {
                debug("null message - discard");
                inputStream.close();
                inputSocket.close();
                return;
            }
            String cmd = packet.substring(0,3);
            int index = packet.indexOf(" ");
            String uri = packet.substring(index + 1, packet.indexOf("HTTP") -1);
            
            if (cmd.equals("PUT")) {
                // Strip <HTML> tags
                int indexFirst = 0;
                int indexLast = 0;
                
                indexFirst = packet.indexOf("<HTML>\n");
                indexLast = packet.lastIndexOf("</HTML>");
                
                if ((indexFirst >= 0) && (indexLast > 0)) {
                    packet = packet.substring(indexFirst + 7, indexLast);
                } else {
                    debug("Invalid message");
                    inputStream.close();
                    inputSocket.close();
                    return;
                }
            }
            
            if (cmd.equals("GET")) {
                index = uri.indexOf("/reg/");
                int ii = 0;
                String type = null;
                if (index >= 0) {
                    type = "reg";
                    ii = index;
                }
                index = uri.indexOf("/rec/");
                if (index >= 0) {
                    type = "rec";
                    ii = index;
                }
                index = uri.indexOf("/snd/");
                if (index >= 0) {
                    type = "snd";
                    ii = index;
                }
                index = uri.indexOf("/ack/");
                if (index >= 0) {
                    type = "ack";
                    ii = index;
                }
                if (type == null) {
                    inputStream.close();
                    inputSocket.close();
                    return;
                }
                String param = uri.substring(ii + 5);
                
                if (param.indexOf("/") >= 0) {
                    param = param.substring(0, param.length() - 1);
                }
                
                if (type.equals(Register)) {
                    debug("Registering remote peer");
                    buffer = register(param);
                }
                
                if (type.equals(Receive)) {
                    debug("Remote peer polls messages");
                    buffer = processOutgoing(param);
                }
                
                if (type.equals(Ping)) {
                    debug("Processing ping");
                    buffer = processPing();
                }
            }
            
            if (cmd.equals("PUT") || cmd.equals("POST")) {
                // XXX: Note that this implementation heavely
                // uses Strings to parse the request. A more efficient
                // manner would be manipulate directely the array of
                // bytes. The reason for using strings is that it helps
                // debugging. Once the code will be stable enough, this
                // part will be revisited to be more efficient.
                // lomax@jxta.org
                
                debug("PUT request");
                buffer = processIncoming(packet);
            }
            
            
            outputStream = inputSocket.getOutputStream();
            if (outputStream == null) {
                debug("HTTP: outputstream is null");
                inputStream.close();
                inputSocket.close();
                return;
            }
            
            if (buffer != null) {
                if (outputStream == null) {
                    debug("HTTP: outputstream is null");
                } else {
                    outputStream.write("<HTML>\n".getBytes());
                    outputStream.write(buffer);
                    outputStream.write("\n</HTML>\n".getBytes());
                    outputStream.flush();
                    outputStream.close();
                }
            } else {
                outputStream.write("\n<HTML>\n\n</HTML>\n\n".getBytes());
                outputStream.close();
            }
            
            inputStream.close();
            inputSocket.close();
            
        } catch (Exception e) {
            
            try {
                if (inputStream != null) {
                    inputStream.close();
                }
            } catch (Exception e1) {
                debug("runReceive [1] failed " + e1);
            }
            try {
                if (outputStream != null) {
                    outputStream.close();
                }
            } catch (Exception e1) {
                debug("runReceive [2] failed " + e1);
            }
            
            try {
                if (inputSocket != null) {
                    inputSocket.close();
                }
            } catch (Exception e1) {
                debug("runReceive [2] failed " + e1);
            }
            debug("runReceiveError " + e);
        }
    }
    
    public void sendHttpPUT(String url, byte[] buffer)
    throws IOException {
        
        debug("sendHttpPUT to " + url);
        
        int indexStart = url.indexOf("//") + 2;
        int indexStop = url.indexOf("/", indexStart);
        String peerName = url.substring(indexStart, indexStop);
        String hostName = url.substring(0, url.indexOf("/"));
        indexStart = hostName.indexOf(":");
        int port = 80;
        
        if (indexStart >= 0) {
            port = Integer.parseInt(hostName.substring(indexStart + 1));
            hostName = hostName.substring(0, indexStart);
        }
        
        Socket socket = null;
        OutputStream op = null;
        InetAddress addr = null;
        
        if (proxyName != null) {
            url = "http://" + url;
        }
        
        // Builds HTTP header
        // Adding 16 to the size of the buffer is needed in order
        // to add the <HTML> and </HTML> tags.
        String header = "PUT " + url + " HTTP/1.1\n" +
        "User-Agent: Jxta1.0\n" +
        "Host: " + hostName + ":" + port + "\n" +
        "Content-length: " + (buffer.length + 16) + " \n\n";
        
        
        try {
            String hn = null;
            if (proxyName != null) {
                hn = proxyName;
                port = 8080;
                
                indexStart = proxyName.indexOf(":");
                
                if (indexStart >= 0) {
                    port = Integer.parseInt(proxyName.substring(indexStart + 1));
                    hn = proxyName.substring(0, indexStart);
                }
            } else {
                hn = hostName;
            }
            
            addr = InetAddress.getByName(hn);
            
            if (addr == null) {
                throw new IOException( "Cannot resolve address for " + hn );
            }
            socket = new Socket(addr, port, usingInterfaceAddr, 0 );
            try {
                socket.setSoLinger(true, LingerDelay);
                socket.setSoTimeout(SocketTimeout);
            } catch (Exception e) {
                debug("httpPUT: setting socket failed because : " + e);
            }
            op = socket.getOutputStream();
            
            // Send the message
            
            op.write(header.getBytes());
            op.write("<HTML>\n".getBytes());
            op.write(buffer);
            op.write("\n</HTML>\n".getBytes());
            op.flush();
            op.close();
            socket.close();
            debug("Message sent");
        } catch (Exception e) {
            if (op != null) {
                try {
                    op.close();
                } catch (Exception ex) {
                    debug("sendHttpPut [1] " + ex);
                }
            }
            if (socket != null) {
                try {
                    socket.close();
                } catch (Exception ex) {
                    debug("sendHttpPut [2] " + ex);
                }
            }
            debug("sendHttpPUT failed with : " + e);
        }
    }
    
    
    private String sendHttpGET(String url, boolean reply)
    throws IOException {
        
        debug("sendHttpGET to " + url);
        int indexStart = url.indexOf("//") + 2;
        int indexStop = url.indexOf("/", indexStart);
        String peerName = url.substring(indexStart, indexStop);
        String hostName = url.substring(0, url.indexOf("/"));
        indexStart = hostName.indexOf(":");
        int port = 80;
        if (indexStart >= 0) {
            port = Integer.parseInt(hostName.substring(indexStart + 1));
            hostName = hostName.substring(0, indexStart);
        }
        Socket socket = null;
        OutputStream op = null;
        InputStream ip =null;
        InetAddress addr = null;
        
        if (proxyName != null) {
            url = "http://" + url;
        }
        
        // Builds HTTP header
        String header = "GET " + url + " HTTP/1.1\n" +
        "User-Agent: Jxta1.0\n" +
        "Host: " + hostName + ":" + port + "\n" +
        "Content-length: 0\n\n";
        
        
        try {
            String hn = null;
            if (proxyName != null) {
                hn = proxyName;
                port = 8080;
                indexStart = proxyName.indexOf(":");
                
                if (indexStart >= 0) {
                    port = Integer.parseInt(proxyName.substring(indexStart + 1));
                    hn = proxyName.substring(0, indexStart);
                }
            } else {
                hn = hostName;
            }
            
            addr = InetAddress.getByName(hn);
            if (addr == null) {
                debug("cannot resolve (DNS) " + hn);
                throw new IOException();
            }
            socket = new Socket(addr, port, usingInterfaceAddr, 0 );
            try {
                socket.setSoLinger(true, LingerDelay);
                socket.setSoTimeout(SocketTimeout);
            } catch (Exception e) {
                debug("httpGET: setting socket failed " + e);
            }
            op = socket.getOutputStream();
            ip = socket.getInputStream();
            
            // Send the message
            
            op.write(header.getBytes());
            op.write("\n<HTML>\n\n</HTML>\n".getBytes());
            
            if (reply) {
                
                String tmp = getResult(ip);
                
                
                // Strip <HTML> tags
                int indexFirst = 0;
                int indexLast = 0;
                boolean valid = false;
                
                ip.close();
                op.close();
                socket.close();
                
                if (tmp == null) {
                    debug("httpGET no answer");
                    return null;
                }
                
                indexFirst = tmp.indexOf("<HTML>\n");
                indexLast = tmp.lastIndexOf("</HTML>");
                
                if ((indexFirst >= 0) && (indexLast > 0)) {
                    tmp = tmp.substring(indexFirst + 7, indexLast);
                    return tmp;
                }
            } else {
                ip.close();
                op.close();
                socket.close();
                return null;
            }
        } catch (Exception e) {
            try {
                if (ip != null) {
                    ip.close();
                }
            } catch (Exception e1) {
                debug("HttpGET [1] failed with " + e1);
            }
            try {
                if (op != null) {
                    op.close();
                }
            } catch (Exception e1) {
                debug("HttpGET [2] failed with " + e1);
            }
            try {
                if (socket != null) {
                    socket.close();
                }
            } catch (Exception e1) {
                debug("HttpGET [3] failed with " + e1);
            }
            debug("HttpGet to failed with " + e);
        }
        return null;
    }
    
    private byte[] register(String peer) {
        
        debug("Registering peer " + peer);
        // Strips off jxta:// from the PeerId
        String dirName = peer.substring(peer.indexOf("//") + 2);
        
        // First create spool directory for this new client
        // XXX: some security will have to be checked here.
        // That will be done when the general security model
        // will be completed. lomax@jxta.org
        if (initDir(HttpSpool + File.separatorChar + dirName)) {
            return (new HttpMessage(HttpMessage.Ok)).getBytes();
        } else {
            return (new HttpMessage(HttpMessage.Failed)).getBytes();
        }
    }
    
    private byte[] processOutgoing(String p) {
        
        debug("processOutgoing for peer " + p);
        // Strips off jxta:// from the PeerId
        String peer = p.substring(p.indexOf("//") + 2);
        
        String dirName = HttpSpool + File.separatorChar + peer;
        String fn = cm.getOldestFile(dirName);
        
        if (fn == null) {
            // No more messages in the spooler for that peer
            debug("No messages");
            return (new HttpMessage(HttpMessage.Empty)).getBytes();
        }
        // For the time being, just return a message at the time.
        // Maybe a good optimization would be to return a bunch
        // of them...
        // XXX: to revist. lomax@jxta.org
        
        byte[] buffer = (new HttpMessage(dirName, fn)).getBytes();
        
        // Remove the file.
        // XXX: maybe it would be wise to wait for an OK from the other
        // side before removing the file, but that might be too heavy.
        // To be revisited. lomax@jxta.org
        try {
            
            cm.remove(dirName, fn);
            
            while (cm.exists(dirName, fn)) {
                // Wait a little and try again
                try {
                    Thread.sleep(10000);
                } catch (Exception e2) {
                    debug("ex " + e2);
                    
                }
            }
            return buffer;
        } catch (Exception e) {
            System.err.println("HTTP: cannot remove " + dirName
            + File.separatorChar + fn);
            return null;
        }
    }
    
    private byte[] processIncoming(String data) {
        
        try {
            byte[] buffer = data.getBytes();
            MessageImpl msg = new MessageImpl(buffer, 0, buffer.length);
            
            endpoint.demux(msg);
            return null;
            
        } catch (Exception e) {
            // Just discard message
            return null;
        }
    }
    
    private void propagateToClients(byte[] buffer,
    String pName,
    String pParam,
    String prunePeer) {
        
        // XXX: propagate is likely to be removed.
        // For the time being, make sure it is never used.
        if (doPropagate) {
            System.out.println("WARNING: HttpTransport has been configured to propagate.");
            System.out.println("this feature is disabled and is going to be removed.");
            System.out.println("Please send an email to lomax@jxta.org");
            doPropagate = false;
            return;
        }
        
        if (!doPropagate) {
            return;
        }
        
        // Strip off the prefix
        prunePeer = prunePeer.substring("jxta://".length());
        
        // First get the list of clients
        String[] files = cm.getFileNames(HttpSpool);
        if (files == null) {
            return;
        }
        
        String file = null;
        for (int i = 0; i < files.length; ++i) {
            file = files[i];
            try {
                if (file.equals("httpconfig")) {
                    // XXX: this file should be in a different location.
                    //To be fixed lomax@jxta.org
                    continue;
                }
            } catch (Exception e) {
                continue;
            }
            
            try {
                if (file.equals(prunePeer)) {
                    continue;
                }
            } catch (Exception e) {
                continue;
            }
            
            try {
                HttpMessage httpMsg = new HttpMessage(HttpSpool +
                File.separatorChar
                + file,
                cm.createTmpName(),
                buffer);
            } catch (Exception e) {
                continue;
            }
        }
    }
    
    /**
     * Creates and return an EndpointMessanger
     */
    public EndpointMessanger getMessanger(EndpointAddress dst,
    String peer,
    int type)
    throws IOException {
        
        String tmp = dst.getProtocolAddress();
        
        if (tmp.substring(0, MagicWord.length()).equals(MagicWord)) {
            debug("   getting a server messanger");
            return new HttpServerMessanger(dst, peer, this);
        }
        switch (type) {
            case Endpoint.NonBlocking:
                debug("    getting a client messanger");
                return new HttpNonBlockingMessanger(dst, peer, this);
            case Endpoint.HandChecking:
                // XXX: not implemented yet
                System.err.println("HandCheck is not yet implemented on HTTP");
                throw new IOException();
            default:
                System.err.println("HttpTransport: unknown type");
                throw new IOException();
        }
    }
    
    
    /**
     * Propagates a TransportMessage on this TransportProtocol
     *
     * @param msg the TransportMessage to be propagated
     */
    public void propagate(Message message,
    String pName,
    String pParams,
    String prunePeer) throws IOException {
        
        if (!doPropagate) {
            return;
        }
        
        // First build the destination and source addresses
        Address mAddress = new Address();
        mAddress.setProtocolName("http");
        mAddress.setProtocolAddress("jxtamulticast");
        mAddress.setServiceName(pName);
        mAddress.setServiceParameter(pParams);
        
        
        ((MessageImpl) message).setDestAddress(mAddress);
        ((MessageImpl) message).setSrcAddress(localAddress);
        
        byte buffer[] = null;
        InputStream ip = null;
        try {
            // Get the bytes from the message
            ip = ((MessageImpl) message).getInputStream();
            buffer = new byte [ip.available()];
            ip.read(buffer);
        } catch (Exception e) {
            debug("HTTP - send - cannot read outgoing message");
            throw new IOException();
        }
        
        if (serverName != null) {
            propagateToClients(buffer, pName, pParams, prunePeer);
        } else {
            // We are a client
            // prunePeer is not implemented in that case.
            // XXX to be fixed  lomax@jxta.org
            String addr = null;
            for (int i = 0; i < staticHttps.size(); ++i) {
                try {
                    addr = (String) staticHttps.elementAt(i) + "/snd/";
                    sendHttpPUT(addr, buffer);
                } catch (Exception e) {
                    continue;
                }
            }
            
            for (int i = 0; i < dynamicHttps.size(); ++i) {
                try {
                    addr = (String) staticHttps.elementAt(i) + "/snd/";
                    sendHttpPUT(addr, buffer);
                } catch (Exception e) {
                    continue;
                }
            }
        }
    }
    
    
    public String getProtocolName() {
        return "http";
    }
    
    /**
     * closes this TransportProtocol.
     */
    public void close() {
        // XXX: not implemented yet.
    }
    
    public EndpointAddress getLocalAddress() {
        return localAddress;
    }
    
    private boolean initDir(String dn) {
        try {
            cm.createFolder(dn);
            return true;
        } catch (Exception e) {
            System.err.println("HTTP: cannot create spool folder: " + dn);
            return false;
        }
    }
    
    public void addStaticHttp(String url) {
        
        debug("addStaticHttp: " + url);
        url = new String(url);
        
        // Test if the address is already in the table.
        // Don't trust Vector.contains().
        boolean found = false;
        
        for (int i = 0; i < staticHttps.size(); ++i) {
            try {
                if (((String) staticHttps.elementAt(i)).equals(url)) {
                    found = true;
                    break;
                }
            } catch (Exception e) {
                break;
            }
        }
        if (found) {
            return;
        }
        staticHttps.addElement(url);
        
        // Register itself to remote site
        registerWithRemote(url);
        
        // If there is no deamon polling thread yet, create one
        if (pollingThread == null) {
            startPolling();
        }
    }
    
    public void removeStaticHttp(String url) {
        
        // Test if the address is already in the table.
        // Don't trust Vector.contains().
        boolean found = false;
        int i = 0;
        for (i = 0; i < staticHttps.size(); ++i) {
            try {
                if (((String) staticHttps.elementAt(i)).equals(url)) {
                    found = true;
                    break;
                }
            } catch (Exception e) {
                break;
            }
        }
        if (!found) {
            // Nothing to remove
            return;
        }
        try {
            staticHttps.removeElementAt(i);
        } catch (Exception e) {
            // This should not happen.
            System.err.println("HttpTransport: cannot remove static HTTP server");
        }
        
        // Check if we need to terminate polling
        if (staticHttps.isEmpty() && dynamicHttps.isEmpty()) {
            stopPolling();
        }
        return;
    }
    
    public void addDynamicHttp(String url) {
        
        url = new String(url);
        
        // Test if the address is already in the table.
        // Don't trust Vector.contains().
        boolean found = false;
        
        for (int i = 0; i < dynamicHttps.size(); ++i) {
            try {
                if (((String) dynamicHttps.elementAt(i)).equals(url)) {
                    found = true;
                    break;
                }
            } catch (Exception e) {
                break;
            }
        }
        if (found) {
            return;
        }
        dynamicHttps.addElement(url);
        
        // Register itself to remote site
        registerWithRemote(url);
        
        // If there is no deamon polling thread yet, create one
        if (pollingThread == null) {
            startPolling();
        }
    }
    
    public void removeDynamicHttp(String url) {
        
        // Test if the address is already in the table.
        // Don't trust Vector.contains().
        boolean found = false;
        int i = 0;
        for (i = 0; i < dynamicHttps.size(); ++i) {
            try {
                if (((String) dynamicHttps.elementAt(i)).equals(url)) {
                    found = true;
                    break;
                }
            } catch (Exception e) {
                break;
            }
        }
        if (!found) {
            // Nothing to remove
            return;
        }
        try {
            dynamicHttps.removeElementAt(i);
        } catch (Exception e) {
            // This should not happen.
            System.err.println("HttpTransport: cannot remove dynamic HTTP" +
            " server");
        }
        // Check if we need to terminate polling
        if (staticHttps.isEmpty() && dynamicHttps.isEmpty()) {
            stopPolling();
        }
        return;
    }
    
    
    private void startPolling() {
        
        debug("startPolling");
        // Sanity check
        if (pollingThread != null) {
            return;
        }
        
        // Polling means that we are a client
        localAddress = localClientAddress;
        
        // Create listening Thread
        pollingThread = new Thread(this, "HttpTransportPolling" );
        pollingThread.start();
        Thread.yield();
    }
    
    private void stopPolling() {
        
        // Sanity check
        if (pollingThread == null) {
            return;
        }
        pollingThread.interrupt();
        polling = false;
        pollingThread = null;
        
        // We stopped polling, maybe are we server now
        localAddress = localServerAddress;
    }
    
    // runPolling: the purpose of this Thread is to periodically
    // query remote HTTP connection for incoming messages. This
    // is needed if the peer is behind a firewall, and cannot be
    // directely contacted.
    
    
    private void runPolling() {
        
        while (true) {
            
            if (!polling) {
                return;
            }
            
            if (Thread.interrupted()) {
                // we checked already that we are supposed to poll. So, this
                // thread has been interrupted only to force a polling.
                // Just wait a couple of second and start a new poll
                try {
                    Thread.sleep(2000);
                } catch (Exception e) {
                }
            }
            
            // First poll static remote http
            for (int i=0; i < staticHttps.size(); ++i) {
                
                String addr = null;
                try {
                    addr = (String) staticHttps.elementAt(i);
                } catch (Exception e) {
                    continue;
                }
                // Build the proper url
                addr = addr + "/rec/" + localPeerId + "/";
                
                // Polls at least MaxMsgPoll from the server HTTP
                int cnt = 0;
                while ( cnt++ < MaxMsgPoll) {
                    if (!pollRemote(addr)) {
                        // If pollRemote returns false, either something bad
                        // happened, or there is no more messages to get
                        break;
                    }
                }
            }
            
            // poll dynamic remote http
            for (int i=0; i < dynamicHttps.size(); ++i) {
                
                String addr = null;
                try {
                    addr = (String) dynamicHttps.elementAt(i);
                } catch (Exception e) {
                    continue;
                }
                // Build the proper url
                addr = addr + "/rec/" + localPeerId + "/";
                
                // XXX:
                // pollRemote returns false is something bad
                // happened. This allow to optimize for remote
                // connection that are failing. To be implemented
                // later.
                // lomax@jxta.org
                pollRemote(addr);
            }
            try {
                Thread.sleep(pollingDelay);
            } catch (Exception e) {
            }
        }
    }
    
    
    private boolean pollRemote(String addr) {
        
        HttpMessage httpMsg = null;
        MessageImpl msg = null;
        byte[] buffer = null;
        String content = null;
        
        while (true) {
            
            try {
                content = sendHttpGET(addr, true);
                if (content == null) {
                    return false;
                }
                
                httpMsg = new HttpMessage(content);
                
                buffer = httpMsg.getData();
                if (buffer == null) {
                    debug("   no message to poll");
                    return false;
                }
                
                msg = new MessageImpl(buffer, 0, buffer.length);
                
                endpoint.demux(msg);
                debug("   polled one message");
                return true;
                
            } catch (Exception e) {
                System.err.println("HttpTransport.pollRemote " + e);
                return false;
            }
        }
    }
    
    private void registerWithRemote(String addr) {
        
        String result = null;
        try {
            result = sendHttpGET(addr  + "/reg/" + localPeerId + "/", false);
            if (result != null) {
                return;
            }
        } catch (IOException e) {
        }
    }
    /*
    public synchronized void setProxy (String p) {
        proxyName = new String (p);
    }
     
    public synchronized void setRouters (Vector v) {
        routers = (Vector) v.clone();
    }
     
    public synchronized void setServer (String s) {
        serverName = new String (s);
    }
     */
    public synchronized String getProxy() {
        return proxyName;
    }
    
    public synchronized Vector getRouters() {
        return routers;
    }
    
    public synchronized String getServer() {
        return serverName;
    }
    /*
    public synchronized void enablePropagate () {
        doPropagate = true;
    }
     
    public synchronized void disablePropagate() {
        doPropagate = false;
    }
     */
    public synchronized boolean doesPropagate() {
        return doPropagate;
    }
    
    private String getResult(InputStream ip) {
        
        String tmp = "";
        int    cnt = 0;
        byte[] buffer = null;
        int res = 0;
        
        try {
            while (true) {
                if (ip.available() == 0) {
                    try {
                        if (++cnt > 8) break;
                        Thread.sleep(2000);
                    } catch (Exception e) {
                    }
                    continue;
                }
                if (ip.available() == -1) {
                    return null;
                }
                buffer = new byte [ip.available()];
                res = ip.read(buffer);
                if (res == 0) {
                    continue;
                }
                tmp += new String(buffer, 0, res);
                if (tmp.indexOf("GET") == 0) {
                    return tmp;
                }
                if (tmp.lastIndexOf("</HTML>") >= 0) {
                    break;
                }
            }
        } catch (Exception e) {
            return null;
        }
        
        return tmp;
    }
    
    public boolean isFast() {
        return false;
    }
    
    public boolean isConnected() {
        return false;
    }
    
    public boolean ping(EndpointAddress addr) {
        
        if (addr == null) {
            debug("ping: address is null");
            return false;
        }
        
        if (serverName != null) {
            // This is the server side. Just check if the remote
            // peer is registered.
            
            try {
                String dirName = HttpSpool + File.separatorChar +
                addr.getProtocolAddress().substring(MagicWord.length());
                debug("Looking for " + dirName);
                boolean res =  cm.isFolder(dirName);
                
                
                if( !res ) {
                    
                    // the above code may not be needed, we might just
                    
                    // be able to use this code directly.  what needs
                    
                    // to be done is to convert the above code to use
                    
                    // the built in proxy support that java already
                    
                    // has in it
                    
                    
                    
                    // what we do is perform a get on the ping pseudo
                    
                    // address.  if we can read the PingOk off the
                    
                    // stream, then we know it is a valid address
                    
                    
                    
                    // REVIEW 2001-05-30 <raffi> -- this needs to be
                    
                    // optimised to work faster than it currently
                    
                    // works.  if you pass in an invalid address, it
                    
                    // my hang
                    
                    URL url = new URL( "http://" + addr.getProtocolAddress() + "/" + Ping + "/" );
                    
                    debug( "attempting to do a GET on " + url );
                    
                    StringBuffer buffer = new StringBuffer();
                    
                    BufferedReader reader = new BufferedReader( new InputStreamReader( url.openStream() ) );
                    
                    String read = null;
                    
                    while( ( read = reader.readLine() ) != null )
                        
                        buffer.append( read );
                    
                    read = buffer.toString();
                    
                    read = read.substring( read.indexOf( "<HTML>" ) + "<HTML>".length(), read.lastIndexOf( "</HTML>" ) );
                    
                    if( PingOk.equals( read ) )
                        
                        res = true;
                    
                    
                    
                }
                
                
                
                debug("ping returns " + res);
                return res;
                
            } catch (Exception e) {
                debug("ping [1] returns false");
                return false;
            }
            
        } else {
            // This is the client side. Just check if the host is reachable
            // XXX: this implementation isn't complete: for the time being,
            // check if the http address is a client address. If it is, return
            // false, since this peer is also a client, and therefore cannot directely
            // talk to another http client. If the http address is a server address,
            // then return true (this is the incomplete portion: we should check if the
            // remote peer is reachable.
            try {
                String tmp = addr.getProtocolAddress();
                
                if (tmp.substring(0, MagicWord.length()).equals(MagicWord)) {
                    // Client address - return false
                    return false;
                } else {
                    // Server address
                    return true;
                }
            } catch (Exception e) {
                debug("ping failed with " + e);
                return false;
            }
        }
    }
    
    private byte[] processPing() {
        
        return PingOk.getBytes();
    }
    
    private void cleanupAllClients() {
        
        debug("cleanupAllClients starts");
        String[] files = cm.getFileNames(HttpSpool);
        if (files == null) {
            debug("   no client folders");
            return;
        }
        
        for (int i = 0; i < files.length; ++i) {
            debug("   file= " + files[i]);
            if (cm.isFolder(HttpSpool +
            File.separatorChar +
            files [i])) {
                cleanupClient(files[i]);
            }
        }
    }
    
    private void cleanupClient(String peer) {
        
        debug("cleanupClient " + peer);
        
        try {
            String dirName = HttpSpool + File.separatorChar + peer;
            cm.cleanupFolder(dirName);
        } catch (Exception e) {
            debug("cleanup failed for " + peer);
        }
    }
    
    private static void debug(String str) {
        if (debugFlag) {
            System.out.println("[HttpTransport] " + str);
        }
    }
    
    
    public class IncomingUnicastThreads implements Runnable {
        
        protected Vector sockets = null;
        protected HttpTransport tp = null;
        private   int waitingThreads = 0;
        private   int nbOfThreads = 0;
        
        public IncomingUnicastThreads(HttpTransport tp, int initialNumber) {
            
            this.tp = tp;
            sockets = new Vector();
            
            for (int i = 0; i < initialNumber; ++i) {
                ++nbOfThreads;
                Thread t = new Thread(this,
                "HTTP Incoming Unicast Thread " + i);
                t.start();
            }
        }
        
        public void run() {
            
            Socket socket = null;
            while (true) {
                socket = waitForSocket();
                if (socket == null) {
                    // That means that this thread must die
                    decThreads();
                    return;
                }
                tp.runReceive(socket);
            }
        }
        
        synchronized private void decThreads() {
            --nbOfThreads;
        }
        
        synchronized protected Socket waitForSocket() {
            
            Socket socket = null;
            
            while (true) {
                
                if (sockets.size() > 0) {
                    try {
                        socket = (Socket) sockets.elementAt(0);
                    } catch (Exception e) {
                        // Strange error. XXX: to handle better
                        debug("waitForSocket error: " + e);
                    }
                    if (socket != null) {
                        try {
                            sockets.removeElementAt(0);
                        } catch (Exception e) {
                            debug("waitForSocket error: " + e);
                        }
                        return socket;
                    }
                    // We should not get here.
                    debug("waitForSocket cannot get incoming socket");
                }
                
                // No more socket to consume... Just wait.
                // Check if this thread must die or wait for a new socket
                if (mustDie()) {
                    debug("waitForSocket: killing thread");
                    return null;
                }
                try {
                    ++waitingThreads;
                    wait();
                    --waitingThreads;
                } catch (Exception e) {
                }
            }
        }
        
        synchronized private boolean mustDie() {
            
            if ((nbOfThreads > MaxNbOfUnicastThreads) ||
            (waitingThreads > 1)) {
                return true;
            }
            return false;
        }
        
        
        synchronized protected void newSocket(Socket socket) {
            
            if (sockets.size() > MaxNbOfPendingSockets) {
                // Too many request. Just drop it.
                try {
                    socket.close();
                } catch (Exception e) {
                    debug("newSocket socket.close failed with " + e);
                }
                debug("newSocket: too many incoming request, dropped incoming message");
                return;
            }
            
            sockets.addElement(socket);
            // Check if we need to allocate a new Thread
            if (waitingThreads == 0) {
                ++nbOfThreads;
                // Allocate a new thread.
                debug("newSocket: create a new Thread");
                Thread t = new Thread(this,
                "HTTP Incoming Unicast Thread ");
                t.start();
            }
            try {
                notify();
            } catch (Exception e) {
            }
        }
        
        private void debug(String str) {
            tp.debug(str);
        }
    }
    
    
    
    void processReceivingSocket(Socket socket) {
        
        unicastThreads.newSocket(socket);
    }
    
    
}

