/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PasswdMembership.java,v 1.3 2001/05/23 01:00:19 lomax Exp $
 */

package net.jxta.impl.membership;

import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;

import net.jxta.credential.AuthenticationCredential;
import net.jxta.credential.Credential;
import net.jxta.document.Advertisement;
import net.jxta.document.Element;
import net.jxta.document.MimeMediaType;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.id.IDFactory;
import net.jxta.membership.Authenticator;
import net.jxta.membership.Membership;
import net.jxta.peer.PeerID;
import net.jxta.peergroup.PeerGroup;
import net.jxta.peergroup.PeerGroupID;
import net.jxta.protocol.ServiceAdvertisement;
import net.jxta.service.Service;

import net.jxta.exception.ProtocolNotSupportedException;
import net.jxta.exception.PeerGroupException;

/**
 *  The passwd membership service provides a Membership implementation which is
 *  based on a password scheme similar to the unix /etc/passwd system. It is
 *  intended mostly as an example of a simple Membership service and not as
 *  a practical Membership service.
 *
 * @version $Revision: 1.3 $
 * @since JXTA 1.0
 */
public class PasswdMembership extends Membership {
    
    /** This class provides the sub-class of Credential which is associated
     * with the password membership service.
     */    
    private class PasswdCredential implements Credential {
        
        /** The Membership service which generated this credential.
         */        
        PasswdMembership source;
        
        /** The identity associated with this credential
         */        
        String whoami;
        
        /** The peerid associated with this credential.
         */        
        PeerID peerid;
        
        /** The peerid which has been "signed" so that the identity may be verified.
         */        
        String signedPeerID;
        
        protected PasswdCredential( PasswdMembership source, String whoami, String signedPeerID ) {
            
            this.source = (PasswdMembership) source;
            this.whoami = whoami;
            this.peerid = source.getPeerGroup().getPeerID();
            this.signedPeerID = signedPeerID;
        }
        
        protected PasswdCredential( PasswdMembership source, PeerGroupID peergroup, PeerID peer, String whoami, String signedPeerID ) throws PeerGroupException {
            
            this.source = (PasswdMembership) source;
            
            if( !source.getPeerGroup().getID().equals( peergroup ) )
                throw new PeerGroupException( "Cannot credential for a different peer group." );
            
            this.whoami = whoami;
            this.peerid = peer;
            this.signedPeerID = signedPeerID;
        }
        
        /**
         * Returns the service which generated this credential.
         * @return the service which generated this credential.
         */
        public Membership getSourceService() {
            return source;
        }
        
        /** Returns the PeerGroupID associated with this Credential
         * @return the PeerGroupID associated with this Credential
         */        
        public PeerGroupID getPeerGroupID() {
            return source.getPeerGroup().getID();
        }
        
        /** Returns the PeerID assocaited with this credential
         * @return the PeerID assocaited with this credential
         */        
        public PeerID getPeerID() {
            return peerid;
        }
        
        /** Write credential into a document. as is a mime media-type specification and
         * provides the form of the document which is being requested. Two standard
         * document forms are defined. "text/text" encodes the document in a form nice
         * for printing out and "text/xml" which provides an XML format.
         * @param as The mime media type of the encoding format being requested.
         * @throws Exception in the event of errors. this exception spec needs to be narrowed.
         * @return the StructuredDocument which represents this credential.
         */
        public StructuredDocument getDocument(MimeMediaType as) throws Exception {
            StructuredDocument doc =
            StructuredDocumentFactory.newStructuredDocument( as,
            "PasswdCredential" );
            
            Element e = doc.createElement( "PeerGroupID", peergroup.getID().toString() );
            doc.appendChild( e );
            
            e = doc.createElement( "PeerID", peergroup.getPeerID().toString() );
            doc.appendChild( e );
            
            e = doc.createElement( "Identity", whoami );
            doc.appendChild( e );
            
            // FIXME 20010327   Do some kind of signing here based on password.
            e = doc.createElement( "ReallyInsecureSignature", signedPeerID );
            doc.appendChild( e );
            
            return doc;
        }
        
        /* From here on are the methods specific to this credential type */
        
        /** Returns the identity associated with this credential.
         * @return the identity associated with this credential.
         */        
        public String getIdentity() {
            return whoami;
        }
    }
    
    
    /** The authenticator associated with the password membership service.
     */    
    public class PasswdAuthenticator implements Authenticator {
        
        /** The Membership service which generated this authenticator.
         */        
        Membership source;
        /** The Authentication which was provided to the Apply operation of the membership
         * service.
         */        
        AuthenticationCredential application;
        
        /** the identity which is being claimed
         */        
        String whoami = null;
        
        /** the password for that identity.
         */        
        String password = null;
        
        /** Creates an authenticator for the password Membership service. The only method
         * supported is "PasswdAuthentication". Anything entered into the identity info
         * section of the Authentication credential is ignored.
         * @param source The instance of the password membership service which created this
         * authenticator.
         * @param application The Anything entered into the identity info section of the Authentication
         * credential is ignored.
         */        
        PasswdAuthenticator( Membership source, AuthenticationCredential application ) {
            this.source = source;
            this.application = application;
            
            // XXX 20010328 MJD Could do something with the authentication credential here.
        }
        
        /**
         * Returns the policy which generated this authenticator.
         * @return the service which generated this authenticator.
         */
        public Membership getSourceService() {
            return source;
        }
        
        /**
         * Returns true if this Authenticator has been satisfied and is ready
         * for the join operation. Some authenticators may behave in an
         * asynchronously and this method can be used to determine if the
         * authentication process has completed. This method makes no
         * distinction between incomplete authentication and failed authentication.
         * @return true if this Authenticator has been satisfied and is ready
         * for the join operation.
         */
        synchronized public boolean isReadyForJoin() {
            return ( (null != password) && (null != whoami) );
        }
        
        /**
         * Returns the name of this authentication method. This should be the same
         * name which was used in the Authentication credential.
         * @return the name of this authentication method.
         */
        public String getMethodName() {
            return "PasswdAuthentication";
        }
        
        public void setAuth1Identity( String who ) {
            whoami = who;
        }
        
        public String getAuth1Identity() {
            return whoami;
        }
        
        public void setAuth2Password( String secret ) {
            password = secret;
        }
        
        private String getAuth2Password() {
            return password;
        }
        
        /** Returns the AuthenticationCredential which is associated with this authenticator
         * @return  the AuthenticationCredential which is associated with this authenticator
         */        
        public AuthenticationCredential getAuthenticationCredential() {
            return application;
        }
    }
    
    /** if true then the service is running in debug mode and will produce addtional
     * diagnostic information.
     */    
    private boolean debug = false;
    
    /** the peergroup to which this service is associated.
     */    
    private PeerGroup peergroup = null;
    
    /** The current set of principals associated with this peer within this peegroup.
     */    
    private Vector principals = null;
    
    /** The set of AuthenticationCredentials which were used to establish the principals.
     */    
    private Vector authCredentials = null;
    
    /** The ServiceAdvertisement which was used to instantiate this service.
     */    
    private ServiceAdvertisement adv = null;
    
    /** An internal table containing the identity and password pairs as parsed from the
     * the ServiceAdvertisement.
     */    
    private Hashtable logins = null;
    
    /**
     * Initialize the application
     *
     * @param myAdv The Advertisement for this service from the PeerGroupAdvertisement.
     * @param group PeerGroup this service is started from
     * @throws PeerGroupException in the event of errors. this exception spec needs to be narrowed.
     */
    public void init(PeerGroup group, Advertisement myAdv) throws PeerGroupException {
        adv = (ServiceAdvertisement) myAdv;
        peergroup = group;
        
        debug = group.getDebug();
        
        Vector myParams = adv.getParams();
        String etcPasswd = (String) myParams.elementAt( 0 );
        
        logins = new Hashtable();
        
        for( int current = 0; current < etcPasswd.length(); ) {
            int nextDelim = etcPasswd.indexOf( ':', current );
            if( -1 == nextDelim )
                break;
            
            String login = etcPasswd.substring( current, nextDelim ).trim();
            int lastDelim = etcPasswd.indexOf( ':', nextDelim + 1 );
            String passwd = etcPasswd.substring( nextDelim + 1, lastDelim );
            current = lastDelim + 1;
            
            //System.out.println( "Adding login : '" + login + "' with encoded password : '" + passwd + "'" );
            
            logins.put( login, passwd );
        }
        
        // We initialise our set of principals to the resigned state.
        resign();
    }
    
    /** Returns the interface for this service. Some services return proxy versions of
     * themselves to impose security upon the service object or to count references
     * to the service.
     * @return the interface version of this service.
     */    
    public Service getInterface() {
        return this;
    }
    
    /** Returns the ServiceAdvertisement assocaiated with this Service.
     * @return the ServiceAdvertisement assocaiated with this Service.
     */    
    public Advertisement getAdvertisement() {
        return adv;
    }
    
    /**
     * Supply arguments and starts this service if it hadn't started by itself.
     *
     * Currently this service starts by itself and does not expect
     * arguments.
     *
     * @param arg A table of strings arguments.
     * @return int status indication.
     */
    public int startApp(String[] arg) {
        return 0;
    }
    
    /**
     * Ask this service to stop.
     *
     * This request is currently ignored.
     */
    public void stopApp() {
    }
    
    /** Returns the peer group assocaited with this Membership service.
     * @return the peer group assocaited with this Membership service.
     */    
    public PeerGroup getPeerGroup() {
        return peergroup;
    }
    
    /**
     * Request the necessary credentials to join the group with which this
     * policy is associated.
     *
     * @return Authenticator which must be completed before this peergroup may be joined.
     * @param application The AuthenticationCredential containing the authentication method being
     * requested and any identity information which is required to initialize the
     * specified authentication method.
     * @throws PeerGroupException in the event of errors. this exception spec needs to be narrowed.
     * @throws ProtocolNotSupportedException if the authentication method requested
     *     in the application is not supported by this policy.
     */
    public Authenticator apply(AuthenticationCredential application) throws PeerGroupException, ProtocolNotSupportedException {
        
        String method = application.getMethod();
        
        if( (null != method) && !"PasswdAuthentication".equals( method ) )
            throw new ProtocolNotSupportedException(
            "Authentication method not recognized" );
        
        return new PasswdAuthenticator( this, application );
    }
    
    /**
     * Returns an enumeration of the current credentials associated with this peer
     * within in this peergroup.
     * @throws PeerGroupException in the event of errors. this exception spec needs to be narrowed.
     * @return Enumeration of the current credentials associated with this peer
     * within in this peergroup.
     */
    public synchronized Enumeration getCurrentCredentials() throws PeerGroupException {
        return principals.elements();
    }
    
    /**
     * Returns an enumeration of the AuthencticationCredentials which were used as the
     * basis for the current set of credentials.
     *
     * @throws PeerGroupException in the event of errors. this exception spec needs to be narrowed.
     * @return enumeration of the AuthencticationCredentials which were used as the
     * basis for the current set of credentials.
     *
     */
    public synchronized Enumeration getAuthCredentials() throws PeerGroupException {
        return authCredentials.elements();
    }
    
    /**
     * Join the group by virtue of the completed authentication provided.
     *
     * @param authenticated the completed authentication object.
     * @throws PeerGroupException in the event of errors. this exception spec needs to be narrowed.
     * @return The credential created by this join operation.
     */
    public synchronized Credential join(Authenticator authenticated) throws PeerGroupException {
        
        if( !(authenticated instanceof PasswdAuthenticator) )
            throw new ClassCastException( "This is not my authenticator!" );
        
        if( !authenticated.isReadyForJoin() )
            throw new PeerGroupException( "Not Ready to join!" );
        
        if( !checkPasswd(
        ((PasswdAuthenticator)authenticated).getAuth1Identity(),
        ((PasswdAuthenticator)authenticated).getAuth2Password() ) )
            throw new PeerGroupException( "Incorrect Password!" );
        
        // FIXME    20010327    MJD Make up the signed bit.
        
        Credential newCred = new PasswdCredential( this,
        ((PasswdAuthenticator)authenticated).getAuth1Identity(), "blah" );
        
        principals.addElement( newCred );
        
        authCredentials.addElement( authenticated.getAuthenticationCredential() );
        
        return newCred;
    }
    
    /**
     * Leave the group to which this policy is attached.
     * @throws PeerGroupException in the event of errors. this exception spec needs to be narrowed.
     */
    public synchronized void resign() throws PeerGroupException {
        principals = new Vector();
        authCredentials = new Vector();
        
        // FIXME    20010327    MJD Make up the signed bit.
        principals.addElement( new PasswdCredential( this, "nobody", "blah" ) );
    }
    
    /**
     * Given a fragment of a StructuredDocument, reconstruct a Credential object
     * from that fragment.
     *
     * @return Credential which was created from the document fragment.
     * @param element The StructuredDocument fragment to use for building the
     * credential.
     * @throws PeerGroupException in the event of errors. this exception spec needs to be narrowed.
     * @throws Exception in the event of errors. this exception spec needs to be narrowed.
     */
    public Credential makeCredential(Element element) throws PeerGroupException, Exception {
        Object rootIs = element.getKey();
        
        // XXX 20010327 MJD This should use a factory.
        
        if( !"PasswdCredential".equals(rootIs) )
            throw new PeerGroupException( "Element does not contain a recognized credential format" );
        
        Enumeration children = element.getChildren( "PeerGroupID" );
        if( !children.hasMoreElements() )
            throw new RuntimeException( "Missing PeerGroupID Element" );
        
        PeerGroupID peergroup = (PeerGroupID) IDFactory.fromURL( new URL( (String) ((Element) children.nextElement()).getValue() ) );
        if( children.hasMoreElements() )
            throw new RuntimeException( "Extra PeerGroupID Elements" );
        
        children = element.getChildren( "PeerID" );
        if( !children.hasMoreElements() )
            throw new RuntimeException( "Missing PeerID Element" );
        
        PeerID peer = (PeerID) IDFactory.fromURL( new URL( (String) ((Element) children.nextElement()).getValue() ) );
        if( children.hasMoreElements() )
            throw new RuntimeException( "Extra PeerID Elements" );
        
        children = element.getChildren( "Identity" );
        if( !children.hasMoreElements() )
            throw new RuntimeException( "Missing PeerID Element" );
        
        String whoami = (String) ((Element) children.nextElement()).getValue();
        if( children.hasMoreElements() )
            throw new RuntimeException( "Extra Identity Elements" );
        
        children = element.getChildren( "ReallyInsecureSignature" );
        if( !children.hasMoreElements() )
            throw new RuntimeException( "Missing 'ReallyInsecureSignature' Element" );
        
        String signedPeerID = (String) ((Element) children.nextElement()).getValue();
        if( children.hasMoreElements() )
            throw new RuntimeException( "Extra 'ReallyInsecureSignature' Elements" );
        
        return new PasswdCredential( this, peergroup, peer, whoami, signedPeerID );
    }
    
    /**
     * Given an identity and an encoded password
     * @param identity the identity which the user is trying to claim
     * @param passwd the password they used.
     * @return true if the password was correct for the specified identity otherwise false.
     */
    private boolean checkPasswd( String identity, String passwd ) {
        boolean result;
        
        if( !logins.containsKey( identity ) )
            return false;
        
        String encodedPW = makePsswd( passwd );
        
        if( debug )
            System.out.println( "[PasswdMembership] Password '" + passwd +
            "' encodes as: '" + encodedPW + "'");
        
        
        String mustMatch = (String) logins.get( identity );
        
        // if there is a null password for this identity then match everything.
        if( mustMatch.equals( "" ) )
            return true;
        
        result = encodedPW.equals(mustMatch);
        
        return result;
    }
    
    /**
     *  This is the method used to make the password strings. We only provide
     *  one way encoding since we can compare the encoded strings.
     *
     *  FIXME MJD 20010402 : switch to use the standard crypt(3) algorithm for
     *  encoding the passwords. The current algorithm has been breakable since
     *  ancient times, crypt(3) is also weak, but harder to break.
     *
     *   @param source  the string to encode
     *   @return String the encoded version of the password.
     *
     **/
    public static String makePsswd( String source ) {
        /**
         *
         * A->D  B->Q  C->K  D->W  E->H  F->R  G->T  H->E  I->N  J->O  K->G  L->X  M->C
         * N->V  O->Y  P->S  Q->F  R->J  S->P  T->I  U->L  V->Z  W->A  X->B  Y->M  Z->U
         *
         **/
        
        final String xlateTable = "DQKWHRTENOGXCVYSFJPILZABMU";
        
        StringBuffer work = new StringBuffer( source );
        
        for( int eachChar = work.length() - 1; eachChar >=0; eachChar-- ) {
            char aChar = Character.toUpperCase( work.charAt(eachChar) );
            
            int replaceIdx = xlateTable.indexOf( aChar );
            if( -1 != replaceIdx )
                work.setCharAt( eachChar, (char) ('A' + replaceIdx) );
        }
        
        return work.toString();
    }
    
    
}

