/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerInfoService.java,v 1.6 2001/05/23 01:00:19 lomax Exp $
 */

package net.jxta.impl.peer;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URL;
import java.lang.reflect.InvocationTargetException;

import net.jxta.service.Service;
import net.jxta.resolver.Resolver;
import net.jxta.resolver.QueryHandler;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.MimeMediaType;
import net.jxta.document.Element;
import net.jxta.protocol.ResolverQueryMsg;
import net.jxta.protocol.ResolverResponseMsg;
import net.jxta.protocol.PeerGroupAdvertisement;
import net.jxta.protocol.PeerAdvertisement;
import net.jxta.protocol.ServiceAdvertisement;
import net.jxta.protocol.PeerInfoAdvertisement;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.NoResponseException;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.peer.PeerInfo;

import net.jxta.impl.cm.Cm;
import net.jxta.impl.protocol.ResolverQuery;
import net.jxta.impl.protocol.ResolverResponse;
import net.jxta.impl.protocol.PeerInfoAdv;
import net.jxta.impl.peer.PeerInfoInterface;
import net.jxta.impl.peergroup.RefPeerGroup;
import net.jxta.impl.endpoint.Endpoint;

/**
 * Peer Info provides a mechanism to obtain information about peers.
 *
 * @version $Revision: 1.6 $
 * Tag  $Name:  $
 * @since 1.0
 */

public class PeerInfoService implements PeerInfo, QueryHandler {

    /**
     * Directory name for Cm's PeerInfo cache
     */
    private static final String dirname = "PeerInfo";

    /**
     * Time in seconds since midnight, January 1, 1970 UTC and when
     * this peer was started.
     */
    private int startTime = 0;

    /**
     * ResolverService Query Id.
     */
    private volatile int qid = 0;

    private Cm cm = null;
    private Resolver resolver = null;
    private RefPeerGroup pg = null;
    private Endpoint endpoint = null;
    private String localPeerId = null;
    private boolean debug = true;
    private ServiceAdvertisement advertisement = null;

    /**
     * Supply arguments and starts this service if it hadn't started by itself.
     *
     * Currently this service starts by itself and does not expect
     * arguments.
     *
     * @param arg A table of strings arguments.
     * @return int status indication.
     */
    public int startApp(String[] arg) {
        return 0;
    }

    /**
     * Ask this service to stop.
     *
     * This request is currently ignored.
     */
    public void stopApp() {
    }

    /**
     * Service objects are not manipulated directly to protect usage
     * of the service. A Service interface is returned to access the service
     * methods.
     *
     * @return Service public interface of the service
     *
     * @version $Revision: 1.6 $
     * @since JXTA 1.0 
     */
    public Service getInterface() {
        return new PeerInfoInterface(this);
    }

    /**
     * Returns the advertisement for that service.
     *
     * @return Advertisement the advertisement.
     *
     * @version $Revision: 1.6 $
     * @since JXTA 1.0 
     */
    public Advertisement getAdvertisement() {
        return advertisement;
    }

    /**
     * init is called by PeerGroup to init the Service
     * @param pg Peer Group
     * @throws PeerGroupException throw any exception thrown during obtaining
     * Service
     */
    public void init(PeerGroup group, Advertisement sadv)
        throws PeerGroupException
    {

	pg = (RefPeerGroup) group;
        endpoint = pg.getEndpoint();
        advertisement = (ServiceAdvertisement) sadv;
	PeerGroupAdvertisement adv = pg.getAdvertisement();
	localPeerId = pg.getPeerID().toString();
        debug = pg.getDebug();

        resolver = pg.getResolver();
	resolver.registerHandler(advertisement.getName(), this);

	try {
	    URL url = new URL(pg.getID().toString());
            String pgdir = Integer.toHexString((url.getHost()).hashCode());
            cm = new Cm(pgdir);
	    cm.createFolder(dirname);
	} catch (Exception e) {
	    debug("error during creation of local store");
	    if (debug) e.printStackTrace();
            System.out.println(e);
	}

        // record start time at end of successful init
        startTime = (int) (System.currentTimeMillis() / 1000L);
    }

    /**
     * Retrieve Peer Information Advertisement of localhost
     *
     * @return PeerInfoAdvertisement
     */
    public PeerInfoAdvertisement getPeerInfo() {
        int now = ((int) (System.currentTimeMillis() / 1000L));

        Hashtable itraffic = new Hashtable();
        Enumeration ik = endpoint.getTrafficChannels(true);
        while (ik.hasMoreElements()) {
            String k = (String) ik.nextElement();
            itraffic.put(k, endpoint.getTrafficOnChannel(k, true));
        }

        Hashtable otraffic = new Hashtable();
        Enumeration ok = endpoint.getTrafficChannels(false);
        while (ok.hasMoreElements()) {
            String k = (String) ok.nextElement();
            otraffic.put(k, endpoint.getTrafficOnChannel(k, false));
        }

        PeerInfoAdv adv =
            new PeerInfoAdv(localPeerId,        // source pid
                            localPeerId,        // target pid
                            now - startTime,    // uptime
                            now,                // timestamp
                            endpoint.getLastMessageTime(true),  // inbound
                            endpoint.getLastMessageTime(false), // outgoing
                            itraffic,
                            otraffic);
        return adv;
    }

    /**
     * Register a Service in the Peer Advertisement
     * service adversement describes a specific service that is running
     * on the peer (i.e. LDAP, etc)
     *
     * @param advertisement service advertisement to be registered
     * @exception IOException if registeration fails
     */
    public void registerService(ServiceAdvertisement advertisement)
        throws IOException {
        throw new IOException("Not yet implemented");
    }

    /**
     * unRegisters a prevously registered service
     *
     * @param advertisement service advertisement to be unregistered
     * @exception IOException if the un-registeration fails
     */
    public void unRegisterService(ServiceAdvertisement advertisement)
        throws IOException {
        throw new IOException("Not yet implemented");
    }

    /**
     * Get PeerInfo from a remote peer.
     *
     * @param peer Address of a peer, or portal to connect to, if
     * address is null
     */
    public void getRemotePeerInfo(String peer) {
        StructuredTextDocument doc = null;
        try {
            doc = (StructuredTextDocument)
                StructuredDocumentFactory.newStructuredDocument
                (new MimeMediaType("text/xml"),
                 PeerInfoAdvertisement.getAdvertisementType());
        } catch (InvocationTargetException ex) {
            if (debug) ex.printStackTrace();
            System.out.println(ex);
            return;
        }

        Element e = null;

        e = doc.createElement("sourcePid", localPeerId);
        doc.appendChild(e);

        e = doc.createElement("targetPid", peer);
        doc.appendChild(e);

        ResolverQuery query = null;
        String ds = docToString(doc);

        // protect increment of the query id
        synchronized(this) {
            query = new ResolverQuery
                (advertisement.getName(), "JXTACRED", "Token", 
                 localPeerId, ds, qid++);
        }

        debug("sending query " + query.getQueryId() +
              " from " +
              Integer.toHexString(localPeerId.hashCode()) +
              " to " +
              (peer == null ? 
               "null" : Integer.toHexString(peer.hashCode())));

        resolver.sendQuery(peer, query);
    }

    /**
     * Get PeerInfo from local cache.
     *
     * @param peer Address of a peer. Can be null in which case locally
     *  cached PeerInfoAdvertisements of all peers are returned
     * @return An Enumeration of locally cached advertisements for this peer
     */
    public Enumeration getLocalPeerInfo(String peer) throws IOException {

        Vector result = new Vector();
        if (peer == null) {
            // since no peer is specified, return advts for all peers
            String[] fn = cm.getFileNames(dirname);
            for (int i = 0; i < fn.length; i++) {
                FileInputStream is =  cm.getInputStream(dirname, fn[i]);
                Advertisement adv = null;
                try {
                    adv = (Advertisement)
                        AdvertisementFactory.newAdvertisement
                        (new MimeMediaType("text/xml"), is);
                    result.addElement(adv);
                } catch (Exception e) {
                    if (debug) e.printStackTrace();
                    System.out.println(e);
                    continue;
                }
            }
        } else {
            try {
            // peer is specified, get this peer's advt
            String fn = Integer.toHexString(peer.hashCode());
            debug("restoring " + fn + " from " + dirname);
            StructuredDocument doc = null;
            doc = cm.restore(dirname, fn);
            PeerInfoAdv adv = new PeerInfoAdv(doc);
            result.addElement(adv);
            }
            catch( Exception ignored ) {
                
            }
        }
        return result.elements();
    }

    /* methods we must implement by implementing Service, QueryHandler */

    /**
     * Returns the group to which this service is attached.
     * @return PeerGroup the group
     */
    public PeerGroup getGroup () {
        return pg;
    }

    /**
     * Handler API method
     * @return ResolverResponseMsg response
     * @NoResponseException thrown when no response is generated
     * @param query ResolverQueryMsg
     * @throws NoResponseException if no Responses were received
     */
    public ResolverResponseMsg processQuery(ResolverQueryMsg query)
	throws NoResponseException {

        PeerInfoAdvertisement adv = null;
	try {
	    adv = (PeerInfoAdvertisement) AdvertisementFactory.newAdvertisement(
            new MimeMediaType( "text/xml" ),
            new ByteArrayInputStream((query.getQuery()).getBytes()) );
        
	} catch (Exception e) {
            if (debug) e.printStackTrace();
            System.out.println("PeerInfo.processQuery got a bad adv");
            System.out.println(e);
            return null;
	}

        debug("processing query " + query.getQueryId() +
              " at peer " + 
              Integer.toHexString(localPeerId.hashCode()) +
              " from " +
              Integer.toHexString(adv.getSourcePid().hashCode()) + 
              " to " + 
              Integer.toHexString(adv.getTargetPid().hashCode()));

        StructuredTextDocument doc = null;

        // was this query addressed to us?
        if (localPeerId.equals(adv.getTargetPid())) {
            debug("responding to query " + query.getQueryId());
            try {
                doc = (StructuredTextDocument)
                    StructuredDocumentFactory.newStructuredDocument
                    (new MimeMediaType("text/xml"),
                     PeerInfoAdvertisement.getAdvertisementType());
            } catch (InvocationTargetException ex) {
                if (debug) ex.printStackTrace();
                System.out.println(ex);
                return null;
            }

            Element e = null;

            e = doc.createElement("sourcePid", adv.getSourcePid());
            doc.appendChild(e);

            e = doc.createElement("targetPid", adv.getTargetPid());
            doc.appendChild(e);

            int now = ((int) (System.currentTimeMillis() / 1000L));
            
            e = doc.createElement("uptime", Integer.toString(now - startTime));
            doc.appendChild(e);

            e = doc.createElement("timestamp", Integer.toString(now));
            doc.appendChild(e);

            Element traffic = doc.createElement("traffic");
            doc.appendChild(traffic);

            Element itraffic = doc.createElement("in");
            traffic.appendChild(itraffic);

            Element itime = doc.createElement
                ("lastIncomingMessageAt",
                 Integer.toString(endpoint.getLastMessageTime(true)));
            traffic.appendChild(itime);

            Element otraffic = doc.createElement("out");
            traffic.appendChild(otraffic);

            Element otime = doc.createElement
                ("lastOutgoingMessageAt",
                 Integer.toString(endpoint.getLastMessageTime(false)));
            traffic.appendChild(otime);

            Enumeration ichannels = endpoint.getTrafficChannels(true);
            while (ichannels.hasMoreElements()) {
                String ch = (String) ichannels.nextElement();
                Long tr = endpoint.getTrafficOnChannel(ch, true);
                e = doc.createElement(ch, tr.toString());
                itraffic.appendChild(e);
                debug(" inbound traffic on " + ch + " is " + tr);
            }

            Enumeration ochannels = endpoint.getTrafficChannels(false);
            while (ochannels.hasMoreElements()) {
                String ch = (String) ochannels.nextElement();
                Long tr = endpoint.getTrafficOnChannel(ch, false);
                e = doc.createElement(ch, tr.toString());
                otraffic.appendChild(e);
                debug(" outbound traffic on " + ch + " is " + tr);
            }
        } else {
            // just passing thru - do not change the message
            debug("forwarding query " + query.getQueryId());
            try {
                doc = (StructuredTextDocument) 
                    adv.getDocument(new MimeMediaType("text/xml"));
            } catch (Exception e) {
                if (debug) e.printStackTrace();
                System.out.println(e);
                return null;
            }
        }

        ResolverResponse response = null;
	response = new ResolverResponse
            (advertisement.getName(), "JXTACRED", "Token", 
             query.getQueryId(), docToString(doc));
        debug("sending response to " + response.getQueryId());
        return response;
    }

    /**
     * deal with incoming responses
     * @param response
     */
    public void pushResponse(ResolverResponseMsg response) {

	PeerInfoAdvertisement adv = null;
	try {
	    adv = (PeerInfoAdvertisement) AdvertisementFactory.newAdvertisement(
            new MimeMediaType( "text/xml" ),
            new ByteArrayInputStream(response.getResponse().getBytes()) );
	} catch (Exception e) {
	    if (debug) e.printStackTrace();
            System.out.println("PeerInfo.pushResponse got a bad adv");
            System.out.println(e.getMessage());
	    return;
	}

        debug("processing response " + response.getQueryId() +
              " at peer " + 
              Integer.toHexString(localPeerId.hashCode()) +
              " from " +
              Integer.toHexString(adv.getSourcePid().hashCode()) + 
              " to " + 
              Integer.toHexString(adv.getTargetPid().hashCode()));

        debug("publishing response " + response.getQueryId());

        try {
            URL url = new URL(adv.getTargetPid());
            StructuredDocument doc = (StructuredDocument)
                adv.getDocument(new MimeMediaType("text/xml"));

            /* overwrite the status of the peer.  As long as the
               target pid is the same, it should hash to the same
               filename */
            String fn = Integer.toHexString(adv.getTargetPid().hashCode());
            debug("publishing response " + response.getQueryId() +
                  " in " + fn);
            cm.save(dirname, fn, doc);
        } catch (Exception e) {
	    if (debug) e.printStackTrace();
            System.out.println(e);
	    return;
        }
    }

    /**
     * flush cached advertisements.
     *
     * @param id peerId of peer whose locally cached adv is to be
     * deleted. If null, cached advs of all peers are deleted.
     */
    public void flushAdvertisements(String id) throws IOException {
        if (id != null) {
            URL url = new URL(id);
            String fn = Integer.toHexString(id.hashCode());
            debug("flushing advertisements for peer " + 
                  Integer.toHexString(id.hashCode()) + 
                  " fn " + fn);
            cm.remove(dirname, fn);
        } else {
            debug("flushing all advertisements");
            String[] fn = cm.getFileNames(dirname);
            for (int i = 0; i < fn.length; i++) {
                cm.remove(dirname, fn[i]);
            }
        }
    }

    private String docToString(StructuredTextDocument doc) {
        StringWriter out = new StringWriter();
        try {
	    doc.sendToWriter(out);
	    return out.toString();
        } catch (Exception e) {
            System.out.println(e);
            return "";
        }
    }

    private void debug(String str) {
	if (debug) {
	    System.out.println ("[PeerInfo] " + str);
	}
    }
}
