/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: GenericPeerGroup.java,v 1.32 2001/06/05 06:41:11 jice Exp $
 */

package net.jxta.impl.peergroup;

import java.util.*;
import java.io.*;
import java.net.URL;
import net.jxta.id.*;
import net.jxta.peer.*;
import net.jxta.peergroup.*;
import net.jxta.exception.*;
import net.jxta.resolver.*;
import net.jxta.service.*;
import net.jxta.discovery.*;
import net.jxta.protocol.*;
import net.jxta.rendezvous.*;
import net.jxta.pipe.*;
import net.jxta.membership.*;
import net.jxta.endpoint.*;
import net.jxta.platform.*;
import net.jxta.document.*;
import net.jxta.impl.loader.*;
import net.jxta.impl.endpoint.*;
import net.jxta.impl.cm.*;
import net.jxta.impl.protocol.PeerGroupAdv;


public class GenericPeerGroup implements RefPeerGroup {

    /*
     * Shortcuts to well known policies.
     */
    private Resolver resolver;
    private Discovery discovery;
    private Pipe pipe;
    private Membership membership;
    private RendezVous rendezvous;
    private PeerInfo peerinfo;

    // The loader for everyone.
    public static final JxtaLoader loader = new JxtaLoader("./Downloaded");

    // weither this peer is a rendezvous or not.
    private boolean isRendezvous = false;

    // This group's instance advertisement.
    private PeerGroupAdvertisement advertisement;

    // A string that identifies this peer. We essentialy
    // just pass it on to descendants, so that every one has it
    // in the adv.
    private String peerName = "";

    // The ID of this peer. All groups of this peer show the same
    // but it makes life easier to have it here.
    private PeerID peerID;

    // The ID of this group.
    private PeerGroupID groupID = null;

    // This service implements a group but, being a service, it
    // runs inside some group. That's its home group.
    // Exception:The platform itself does not have one. It has to be self
    // sufficient.
    private RefPeerGroup parentGroup = null;

    // All the plug-ins that do the work.
    private Hashtable services = new Hashtable();

    // The endPoint.
    private Endpoint endPoint = null;

    // true after the default net peergroup
    private boolean networkGroupInitialized = false;

    // The initial app object
    private Application mainApp = null;

    // Debug flag
    protected boolean debugStatus = false;

    protected synchronized void initNetworkGroup () {
	if (networkGroupInitialized) {
	    // Nothing to do
	    return;
	}
	endPoint.setRootGroup ((PeerGroup) this);
	networkGroupInitialized = true;
    }

    /*
     * The following is part of the genericPeerGroup implementation.
     * It is only meant to be used by the initializer of derived
     * classes. If you want to expose it to, say, services, then
     * you need to include public versions of these in your
     * own subclass of GenericPeerGroup, and write services that
     * depend on that particular subclass. The main idea is that
     * playing with the set of services is a dangerous thing to do,
     * that should never be exposed to ordinary applications. Also, using
     * services to construct a group should not be imposed by the framework.
     */

    private void setShortCut(String name, Service service) {
        if (name.equals("jxta.service.resolver")) {
            resolver = (Resolver) service;
            return;
        }
        if (name.equals("jxta.service.discovery")) {
            discovery = (Discovery) service;
            return;
        }
        if (name.equals("jxta.service.pipe")) {
            pipe = (Pipe) service;
            return;
        }
        if (name.equals("jxta.service.membership")) {
            membership = (Membership) service;
            return;
        }
        if (name.equals("jxta.service.peerinfo")) {
            peerinfo = (PeerInfo) service;
            return;
        }
        if (name.equals("jxta.service.rendezvous")) {
            rendezvous = (RendezVous) service;
            return;
        }
    }

    private void clearShortCut(String name) {
        if (name.equals("jxta.service.resolver")) {
            resolver = null;
            return;
        }
        if (name.equals("jxta.service.discovery")) {
            discovery = null;
            return;
        }
        if (name.equals("jxta.service.pipe")) {
            pipe = null;
            return;
        }
        if (name.equals("jxta.service.membership")) {
            membership = null;
            return;
        }
        if (name.equals("jxta.service.peerinfo")) {
            peerinfo = null;
            return;
        }
        if (name.equals("jxta.service.rendezvous")) {
            rendezvous = null;
            return;
        }
    }


    /**
     * Adds a service to the set.
     * Returns any pre-existing one.
     */
    private synchronized Service addServiceSync(String name, Service service)

    {
        Service p = (Service) services.remove(name);
        services.put(name, service);
        return p;
    }


    /**
     * Adds a service to the set.
     * removes any pre-existing one with the name name.
     */
    protected void addService(Service service)
    {
        String name =
            ((ServiceAdvertisement) service.getAdvertisement()).getName();
        Service oldp = addServiceSync(name, service);
        setShortCut(name, service);
        // if (oldp != null) oldp.finalize();
    }

    /**
     * Call a service by name.
     * @param name the service name
     * @return Service, the Service registered by that name
     */
    synchronized public Service lookupService(String name)
        throws ServiceNotFoundException
    {
        // Null services are never registered, so we do not need to test that
        // case.
        Service p = (Service) services.get(name);
        if (p == null) {
                System.out.println("Services: " + services.toString());
                throw new ServiceNotFoundException(name);
        }
        return p.getInterface();
    }

    /**
     * check that all required services are there.
     */
    protected void checkServices()
        throws ServiceNotFoundException
    {
        Service ignored;
        ignored = lookupService("jxta.service.resolver");
        ignored = lookupService("jxta.service.discovery");
        ignored = lookupService("jxta.service.pipe");
        ignored = lookupService("jxta.service.rendezvous");
        ignored = lookupService("jxta.service.membership");
        ignored = lookupService("jxta.service.peerinfo");
    }

    /**
     * Ask a group to unregister and unload a service
     * @param service handle to the service to be removed
     */
    synchronized private void removeServiceSync (String name, Service service)
        throws ServiceNotFoundException, ViolationException
    {
        // Weak check: you need to actually posess a handle to the
        // service object you're removing, not its name, but
        // you can fake by doing getService() first.

        Service p = (Service) services.get(name);
        if (p == null) throw new ServiceNotFoundException(name);
        if (p != service) throw new ViolationException(name);
        services.remove(name);
    }


    /**
     * Ask a group to unregister and unload a service
     * FIXME: The finalize() seems like a good thing to do, resourcewise,
     * however, if something was still using it, it'll break it.
     * @param service handle to the service to be removed
     */
    protected void removeService (Service service)
        throws ServiceNotFoundException, ViolationException
    {
        String name =
            ((ServiceAdvertisement) service.getAdvertisement()).getName();
        removeServiceSync(name, service);
        clearShortCut(name);
        // service.finalize(); FIXME. We probably need a terminate()
        //                    method instead.
    }

    /*
     * Now comes the API known only to RefPeerGroup.
     * At the very least, all services have access to it.
     * Not sure about applications yet.
     */

    /**
     * Get this group's parent group.
     */
    public RefPeerGroup getParentGroup() {
        return parentGroup;
    }

    /**
     * Get this group's Endpoint manager.
     * That one should probably never be given to applications, but then
     * some part of the endpoint API might be proxied by the PeerGroup itself,
     * such as "addProtocol", but that's for later.
     */
    public Endpoint getEndpoint()
    {
        return endPoint;
    }

    /*
     * Some minimal init.
     * It is not done at Ctor time because the information might not be
     * available then.
     * We have to assume that we got a RefPeerGroup for parent. That's
     * a rule of the implementation that is not visible to Applications:
     * ANY GROUP IN THE REF IMPLEM IS A REFPEERGROUP.
     */

    public void init(PeerGroup homeGroup, Advertisement baseAdv)
    {
        PeerGroupAdvertisement adv = (PeerGroupAdvertisement) baseAdv;

        parentGroup = (RefPeerGroup) homeGroup; // Our home group is
                                                // our parent (null is ok)

        // Apply a few sanity rules and update the adv accordingly.

        // If we've got a parent, our peerId, peername flag
        // are the same. Ignore whatever the adv says, and actually edit
        // the adv so that it reflects our parent's settings.
        // The platform (no parent) advertisement is the
        // authoritative source of these values.
        if (homeGroup != null) {
	    adv.setPid(homeGroup.getPeerID().toString());
            adv.setPeerName(homeGroup.getPeerName());
	    // If the parent group is a rendez, then this group is also
	    // a rendezvous
	    if (homeGroup.isRendezvous()) {
		adv.setIsRendezvous(true);
	    }
	    debugStatus = homeGroup.getDebug();
        } else {
            // If no-one bothered to set the platform GID, fine, use the
            // standard Platform GID.
            if (adv.getGid() == null || adv.getGid().equals("")) {
                adv.setGid(PeerGroupID.worldPeerGroupID.toString());
            }
            if (adv.getPid() == null || adv.getPid().equals("")) {
               adv.setPid(new PeerID(PeerGroupID.worldPeerGroupID).toString());
            }

            // If the pid has never been set. Create one.
        }


        // If we still did not get a valid group id; make a new one.
        // This is therefore a universaly new group.
        if (adv.getGid() == null || adv.getGid().equals("")) {
            adv.setGid(new PeerGroupID().toString());
        }


        // Now, load our ids from the adv. If they're not valid,
        // make new ones up.
        try {
            groupID = (PeerGroupID) IDFactory.fromURL(new URL( adv.getGid() ));
        } catch (Exception e) {
            System.err.println("Malformed GID in adv: \n"
                               + adv.getGid() + "\nUsing random !");
            groupID = new PeerGroupID();
            adv.setGid(groupID.toString());
        }
        try {
            peerID = (PeerID) IDFactory.fromURL(new URL( adv.getPid() ));
        } catch (Exception e) {
            System.err.println("Malformed PID in adv: \n"
                               + adv.getPid() + "\nUsing random !");
            peerID = new PeerID(groupID);
            adv.setPid(peerID.toString());
        }

        peerName = adv.getPeerName();

        isRendezvous = adv.getIsRendezvous();

        // update the endpoint in adv to include parent's ones.
        if (homeGroup != null) {
          adv.setEndpointAdvertisements(homeGroup.getAdvertisement().
                                        getEndpointAdvertisements());
	}

        // keep the adv. It is *this* group's adv on *this* peer.
        advertisement = adv;

        // Publish the adv. It includes endpoints from our parent.
	// FIXME: That's the way it should be, except that when
	// there will be peergroups (other than platform) that
	// do load endpoints, they'll have to figure out which of
	// such enpoints they already have from their parent and not load
	// them.
        if (homeGroup != null) {
	    // Publish the adv now that DiscoveryService is loaded.
	    // FIXME: we're using the implem here.
	    Discovery disco = homeGroup.getDiscovery();
	    try {
		disco.publish((PeerGroupAdv) adv, Discovery.GROUP);
	    } catch (Exception e) {
		e.printStackTrace();
		System.err.println("Could not publish advertisement");
	    }
        } else {
	    System.out.println("Not publishing the adv (no home group).");
	}

        /*
         * The rest of construction and initialization are left to the
         * implementor.
         * That should include instanciating and setting the endPoint, and
         * finaly supplying it with protocols.
         * That also includes instanciating the appropriate services
         * and registering them.
         * For an example, see the Platform class.
         */
    }

    /**
     * Set this group's Endpoint manager.
     * Separate from init(), because building the endpoint may require the
     * rest to have been set first.
     */
    public void setEndpoint(Endpoint endp)
    {
        endPoint = endp;
    }

    /*
     * Now comes the implementation of the public API.
     */

    /**
     * Returns the main application for that group.
     *
     * @return Application the main application for that group.
     */
    public Application getApp() {
        try {
            if (mainApp == null) {
                ServiceAdvertisement appAdv = advertisement.getApp();
                mainApp =
                    (Application) loader.loadClass(appAdv.getCode(), "",
                                                   true).newInstance();
                if (mainApp == null) {
                    System.out.println("Error loading Initial App:"
                                       + appAdv.getCode());
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return mainApp;
    }


    /**
     * Start the peergroup.
     * In practice, it means starting its main application.
     *
     * The group does not care for start args, and does not come-up
     * with args to pass to its main app. So, until we decide on something
     * more usefull, the args of the group's startApp are passed-on to the
     * group's main app. NB: both the apps init and startApp methods are
     * invoked.
     *
     * @return int Status.
     */
    public int startApp(String[] arg) {

        // FIXME: may be one would like init called separately from startApp.

        ServiceAdvertisement appAdv = advertisement.getApp();
        try { 
            getApp();
            if (mainApp == null) return 1;
            mainApp.init(this, appAdv);
        } catch (PeerGroupException e) {
            return 1;
        }
        return mainApp.startApp(arg);
    }

    /**
     * Stops the group and all its services.
     */
    public void stopApp() {
        // FIXME: should do something here...soon
    }

    /**
     * Ask a group the name of the Peer it is running on.
     */
    public String getPeerName()
    {
        return peerName;
    }

    /**
     * Ask a group its peerId.
     */
    public PeerID getPeerID()
    {
        return (PeerID) peerID.clone();
    }

    /**
     * Ask a group its group id.
     * @return PeerGroupId this Group's ID
     */
    public PeerGroupID getID()
    {
        return (PeerGroupID) groupID.clone();
    }


    /**
     * Ask this group its group advertisement
     * @return Advertisement this Group's advertisement
     */
    public PeerGroupAdvertisement getAdvertisement()
    {
        // FIXME: should clone, maybe.
        return advertisement;
    }


    /**
     * Ask this group its peer advertisement (somewhat backwds compat.)
     * @return Advertisement this Group's advertisement
     */
    public PeerGroupAdvertisement getPeerAdvertisement()
    {
        return advertisement;
    }

    /**
     * Returns the debug status.
     */
    public boolean getDebug() {
	return debugStatus;
    }

    /**
     * Returns the whether the group is s Rendezvous.
     */
    public boolean isRendezvous() {
	return isRendezvous;
    }

  /*
   * shortcuts to the well-known services, in order to avoid calls to lookup.
   */

    /**
     * @return Resolver an object implementing the Resolver service for this
     * group.
     */
    public Resolver getResolver() {
        return resolver;
    }


    /**
     * @return Discovery an object implementing the Discovery service for this
     * group.
     */
    public Discovery getDiscovery() {
        return discovery;
    }

    /**
     * @return PeerInfo an object implementing the PeerInfo service for this
     * group.
     */
    public PeerInfo getPeerInfo() {
        return peerinfo;
    }

    /**
     * @return Membership an object implementing the Membership service for
     * this group.
     */
    public Membership getMembership() {
        return membership;
    }

    /**
     * @return Pipe an object implementing the Pipe service for this
     * group.
     */
    public Pipe getPipe() {
        return pipe;
    };

    /**
     * @return RendezVous an object implementing the RendezVous service for this
     * group.
     */
    public RendezVous getRendezVous() {
        return rendezvous;
    };

}
