/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Platform.java,v 1.54 2001/05/23 01:00:20 lomax Exp $
 */


/**
 * The platform as a whole implements the PeerGroup Interface.
 * Here we implement it as a subclass of the GenericPeerGroup.
 * We do not need a membership service, because we never realy join
 * the group. Peers in the Plaform group just do discovery between them,
 * and the requests made within the group are always accepted.
 * ...we'll see if it takes a dummy membership service to satify the
 * mechanics, or not...
 */

package net.jxta.impl.peergroup;

import java.util.*;
import java.io.*;
import net.jxta.peer.*;
import net.jxta.endpoint.*;
import net.jxta.peergroup.*;
import net.jxta.rendezvous.*;
import net.jxta.protocol.*;

import java.lang.reflect.InvocationTargetException;
import java.util.*;

import net.jxta.impl.discovery.*;
import net.jxta.impl.resolver.*;
import net.jxta.impl.endpoint.*;
import net.jxta.impl.discovery.*;
import net.jxta.impl.pipe.*;
import net.jxta.impl.rendezvous.*;
import net.jxta.impl.document.*;
import net.jxta.impl.protocol.*;
import net.jxta.service.*;
import net.jxta.resolver.*;
import net.jxta.exception.*;
import net.jxta.credential.*;
import net.jxta.document.*;
import net.jxta.impl.protocol.*;
import net.jxta.impl.endpoint.tcp.*;
import net.jxta.impl.endpoint.http.*;
import net.jxta.platform.*;


public class Platform extends GenericPeerGroup {

    Configurator config;

    static {
	try {
	    PeerGroupFactory.setPlatformClass(
                                              Class.forName("net.jxta.impl.peergroup.Platform"));
	} catch(Exception e) {
	    e.printStackTrace();
	}
    }

    private boolean initialized = false;

    protected void normalInit(PeerGroup nullParent,
                              PeerGroupAdvertisement adv) {

	if (nullParent != null) {
	    System.err.println("Platform ignores its parent.");
	}

        // Get the debug flag.
        debugStatus = adv.getDebug();
        
	/*
	 * set the endpoint
	 */

	Endpoint endpoint = new Endpoint (null, this);

	// Set our endpoint. No parent.
	setEndpoint(endpoint);

	// Set-up the minimal GenericPeerGroup
	super.init(null, adv);


	/*
	 * create the TCP Transport if it is there
	 */

	EndpointAdvertisement tadv = (EndpointAdvertisement)
            adv.getEndpointAdvertisements().get("TCP");

	if (tadv != null) {
	    // we have TCP

            if (! tadv.getEnabled()) {
                // Purge the adv of it.
                adv.getEndpointAdvertisements().remove("TCP");
            } else {

                TcpTransport tcp = new TcpTransport(tadv);

                // Init the TcpTransport

                tcp.init(endpoint);

                // Add TcpTransport to the Endpoint
                try {
                    endpoint.addProtocol (tcp);
                } catch (Exception e) {
                    System.out.println("Endpoint.addProtocol for TCP" +
                                       "has thrown an exception :" +
                                       e.toString() );
                }
            }
        }

	/*
	 * Create the HTTP transport if it is there
	 */

	EndpointAdvertisement hadv = (EndpointAdvertisement)
            adv.getEndpointAdvertisements().get("HTTP");

	if (hadv != null) {
	    // we have HTTP

            if (! hadv.getEnabled()) {
                // Purge the adv of it.
                adv.getEndpointAdvertisements().remove("HTTP");
            } else {
                ((HTTPAdv) hadv.getTransportAdvertisement()).
                    setPeerId(getPeerID().toString());

                HttpTransport http = new HttpTransport(hadv);

                // Init the HttpTransport

                http.init(endpoint);

                // Add HTTPTransport to the Endpoint
                try {
                    endpoint.addProtocol (http);
                } catch (Exception e) {
                    System.out.println("Endpoint.addProtocol for HTTP" +
                                       "has thrown an exception :" +
                                       e.toString() );
                }
            }
	}

	try {

	    /*
	     * Load dynamically all the services from the advertisement
	     */

	    Hashtable services = adv.getServiceAdvertisements();
	    ServiceAdvertisement service = null;

	    /*
	     * load the ResolverService first
	     */

	    service =
                (ServiceAdvertisement) services.get("jxta.service.resolver");

	    System.out.println("Loading Service:" + service.getName());
	    Service s = (Service) loader.loadClass(service.getCode(),
                                                   service.getUri(),
                                                   true).newInstance();
	    s.init(this, service);
	    addService(s);


	    /*
	     * Extract the services and load them
	     */

	    for (Enumeration e = services.elements(); e.hasMoreElements();) {

		service = (ServiceAdvertisement) e.nextElement();

                if (!service.getName().equals("jxta.service.resolver")) {
		    System.out.println("Loading Service:" + service.getName());
		    s = (Service)
			loader.loadClass(service.getCode(),
					 service.getUri(), true).newInstance();
		    s.init(this, service);
		    addService(s);
		}
	    }

            // Make sure all the required services are there.
            checkServices();

            // Now that all services had a chance to maw their patch of the
            // lawn, save the platform's adv for next time.
            config.save();

            // Erase the debug flag so that it does not get published.
            adv.setDebug(false);

	} catch (Exception e) {
	    e.printStackTrace();
	    System.out.println("Something botched");
	}
    }

    public void init(PeerGroup nullParent, Advertisement nullAdv)
    {
	if (initialized == true) {
	    System.err.println("You cannot initialize a PeerGroup more"
			       +" than once !");
	    return;
	}
	initialized = true;

	if (nullAdv != null) {
	    System.err.println("Platform has full control over its adv.");
	}

        PrintStream outconf = null;

	if (! new File("jxtaConfig").exists()) { //check if the file exists.
	  try {
            outconf = new PrintStream(new FileOutputStream("jxtaConfig"));

            outconf.println("MembershipAuthenticator=" +

                            "NullAuthenticator");
            outconf.println("MembershipIdentity=" +
                            "somebody");

            outconf.println("InitialNetPeerGroupAppCode=" +
                            "net.jxta.impl.shell.bin.Shell.Shell");
            outconf.println("InitialNetPeerGroupAppCodeURL=" +
                            "http:/www.jxta.org/download/jxta.jar");
            outconf.println("#Default properties. Edit if you want.");

            outconf.close();

	  } catch (Exception  ex1) {
	  }
	}

        config = new Configurator();
        PeerGroupAdvertisement adv = config.get();

        try {
            normalInit(null, adv);
        } catch(Exception e) {
            e.printStackTrace();
            System.err.println("Startup failed, clearing advertisement.");
            config.remove();
            System.exit(1);
        }
	return;
    }
}

