/************************************************************************
 *
 * $Id: NonBlockingOutputPipe.java,v 1.16 2001/06/07 21:54:38 raffi Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.pipe;
import java.io.*;
import java.util.*;

import net.jxta.endpoint.*;
import net.jxta.impl.endpoint.*;
import net.jxta.peergroup.*;
import net.jxta.impl.peergroup.*;
import net.jxta.pipe.*;


/** 
 * This class implements the non blocking Output PipeService.
 *
 */


public class NonBlockingOutputPipe implements OutputPipe {

    public  boolean debugFlag = false;

    private String pipeId = null;
    private RefPeerGroup myGroup = null;
    private PipeResolver myPipeResolver = null;
    private EndpointAddress destAddress = null;
    private EndpointMessanger destMessanger = null;
    private String peer = null;
    private static final long RetryDelay = 5000; // 5 seconds
    private boolean closed = false;

    public NonBlockingOutputPipe (PeerGroup g,
				  PipeResolver r,
				  PipeID opId,
				  String destPeer,
				  long timeout)
	             throws IOException {

	myGroup = (RefPeerGroup) g;
	pipeId = opId.toString();
	myPipeResolver = r;

	debug ("Constructor for " + opId);
	// Search for the pipe

	peer = null;
	if (destPeer == null) {
	    peer = myPipeResolver.find (pipeId, timeout, false);

	    if (peer == null) {
		// PipeResolver was not able to find that pipe.
		debug ("Constructor: no macthing InputPipe for " + opId);
		throw new IOException();
	    }
	} else {
	    peer = destPeer;
	}

	// Send an asynchronous handcheck to the peer that is
	// supposed to have the pipe.
	myPipeResolver.sendVerify (peer, pipeId);

	// Get an endpoint messanger to that address
	Endpoint endpoint = myGroup.getEndpoint();
	destMessanger = endpoint.getMessanger (peer, "PipeService", pipeId,
					       Endpoint.NonBlocking);

	if (destMessanger == null) {
	    debug ("Constructor: no Endpoint Messanger for " + opId);
	    throw new IOException ();
	}
    }

    /**
     * send a message
     *
     * @param msg is the PipeMessage to be sent.
     */

    public void send (Message msg) throws IOException {
	debug ("send");	
	String cachedPeer = myPipeResolver.findLocal (pipeId);
	if (cachedPeer == null) {
	    // The InputPipe is not where it is supposed to be.
	    // Find it, and throw an exception.
	    cachedPeer = myPipeResolver.find (pipeId, RetryDelay, true);
	    if (peer == null) {
		throw new IOException();
	    }
	}
	if (!cachedPeer.equals (peer)) {
	    // The pipe has moved. Get a new Endpoint
	    peer = cachedPeer;
	    destMessanger.close();
	    Endpoint endpoint = myGroup.getEndpoint();
	    destMessanger = endpoint.getMessanger (peer, "PipeService", pipeId,
						   Endpoint.NonBlocking);

	    if (destMessanger == null) {
		debug ("sendMessage: no Endpoint Messanger for " +pipeId);
		throw new IOException ();
	    }
	}
	destMessanger.sendMessage (msg);
    }
	
    /**
     * close the the pipe
     *
     */
    public synchronized void close() {
	// Close the messanger.
	destMessanger.close();
	closed = true;
	
    }

    protected void finalize() throws Throwable {
	if (!closed) {
	    close();
	}
    }

    public void debug (String str) {
	if (debugFlag) {
	    System.out.println ("[NonBlockingOutputPipe] " + str);
	}
    }
}


