/*
 * $Id: PipeResolver.java,v 1.35 2001/06/10 02:09:12 bondolo Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PipeResolver.java,v 1.35 2001/06/10 02:09:12 bondolo Exp $
 */
package net.jxta.impl.pipe;

import java.io.*;
import java.util.*;

import net.jxta.endpoint.*;
import net.jxta.exception.*;

import net.jxta.impl.endpoint.*;
import net.jxta.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.impl.peergroup.*;
import net.jxta.protocol.*;
import net.jxta.resolver.*;
import net.jxta.impl.protocol.*;
import net.jxta.document.*;

import net.jxta.util.*;
import net.jxta.impl.util.*;


// XXX: The following import is need only for the hack described in
// find. This hack MUST disappeear as soon as the generic resolver is done.
import net.jxta.impl.endpoint.tcp.*;

/**
 * This class implements the PipeService ResolverService.
 *
 */


public class PipeResolver implements QueryHandler, JxtaTimerHandler  {

    private boolean debug = false;

    private Resolver resolver = null;

    private static final String PipeResolverName = "JxtaPipeResolver";
    private static final String MsgTypeTag = "MsgType";
    private static final String PipeIdTag = "PipeId";
    private static final String PeerIdTag = "Peer";
    private static final String CachedTag = "Cached";
    private static final String FoundTag = "Found";
    private static final String QueryMsgType = "Query";
    private static final String AnswerMsgType = "Answer";

    private Hashtable pipes = null;
    private Hashtable cachedPipes = null;

    private Vector localAddrs = null;
    private RefPeerGroup myGroup = null;
    private Endpoint  endpoint = null;

    private String localPeerId = null;

    private static int qid=0;
    private Vector pendingQueries = null;
    private static long DefaultPipeLease = 60 * 60 * 1000; // 1 Hour
    private static long GcDelay = 10 * 60 * 1000; // 10 minutes
    private static long MaxPendingQueryDelay = 30 * 60 * 1000; // 30 minutes
    private JxtaTimer gcTimer = null;
    
    public class PipeEntry {

	protected String peer   = null;
	protected long   lease  = 0;

	public PipeEntry (String peer) {

	    this.peer   = peer;
	    this.lease  = DefaultPipeLease;
	}

	public boolean equals (Object obj) {

	    PipeEntry tmp = (PipeEntry) obj;
	    return (peer.equals (tmp.peer));
	}
    }

    private void resetLease (PipeEntry entry) {

	if (entry == null) {
	    return;
	}

	long currentTime = System.currentTimeMillis();
	entry.lease = currentTime + DefaultPipeLease;
    }

    private synchronized void refreshCachedPipes () {

	long currentTime = System.currentTimeMillis();
	Enumeration enum = cachedPipes.keys();
	if ((enum == null) || (!enum.hasMoreElements())) {
	    return;
	}
	PipeEntry entry = null;
	String pipe = null;
	while (enum.hasMoreElements()) {
	    try {
		pipe = (String) enum.nextElement();
		entry = (PipeEntry) cachedPipes.get (pipe);
		if (entry == null) {
		    debug ("refreshCachedPipe: empty entry for pipe " + pipe);
		    continue;
		}
		if (entry.lease < currentTime) {
		    // Remove that entry.
		    cachedPipes.remove (pipe);
		    continue;
		}
	    } catch (Exception e) {
		debug ("refreshCachedPipe: " + e);
		break;
	    }
	}
    }

    public PipeResolver (PeerGroup g) {

	myGroup = (RefPeerGroup) g;
	debug = g.getDebug();
	localPeerId = myGroup.getPeerID().toString();
	endpoint = myGroup.getEndpoint();
	pipes = new Hashtable();
	cachedPipes = new Hashtable();
	pendingQueries = new Vector();

        resolver = g.getResolver();

	// Register to the Generic ResolverService
	resolver.registerHandler (PipeResolverName, this);

	// Create our garbage collection timer
	gcTimer = new JxtaTimer (this, GcDelay, true);
    }


    public void registerCached (String ipId, String peer) {

	debug ("registerCached:");
	debug ("      Id = " + ipId);
	debug ("    peer = " + peer);

	ipId = new String (ipId);
	peer = new String (peer);

	PipeEntry entry = null;
	if (cachedPipes.containsKey (ipId)) {
	    // There already is an entry for this pipe.
	    // Reset the lease on the entry
	    entry = (PipeEntry) cachedPipes.get (ipId);
	    if (entry != null) {
		// That should be the case;
		// Update the peer information and the lease.
		entry.peer = peer;
		resetLease (entry);
	    } else {
		debug ("registerCached. unexpected null pointer");
	    }
	    return;
	}

	entry = new PipeEntry (peer);
        resetLease (entry);
	try {
	    Object obj = cachedPipes.put (ipId, entry);
	} catch (Exception e) {
	    // XXX: is there something we should do here ?
	    debug ("registerCached failed " + e);
	    return;
	}
	notifyWaitingThreads();
    }

    public void register (String ipId, InputPipe ip) {

	debug ("register " + ipId);
	ipId = new String (ipId);
	try {
	    Object obj = pipes.put (ipId, ip);
	} catch (Exception e) {
	    // XXX: is there something we should do here ?
	    return;
	}
	notifyWaitingThreads();
    }


    public void forget (String ipId) {
	debug ("forget " + ipId);
	Object obj = pipes.remove (ipId);
    }


    public void forgetCached (String ipId) {
	debug ("forgetCached " + ipId);
	Object obj = cachedPipes.remove (ipId);
    }


    private String findCached (String ipId) {
	debug ("findCached " + ipId);
	PipeEntry entry = (PipeEntry) cachedPipes.get (ipId);
	if (entry == null) {
	    return null;
	}
	// Something has asked to find this pipe. Update the lease
	resetLease (entry);
	return entry.peer;
    }


    public String findLocal (String ipId) {

	debug ("findLocal " + ipId);
	String peer = null;

	// First look if the pipe is a local InputPipe
	Object obj = pipes.get (ipId);

	if (obj != null) {
	    // Found it. Return the local PeerId
	    debug ("findLocal: found local InputPipe for " + ipId);
	    return localPeerId;
	}

	// Look into our cached set of addresses
	peer = findCached (ipId);
	return peer;
    }


    // nocache is set to true if we want answer only from the peer that
    // has the inputpipe. If nocache is set to false, peers that have
    // cached information about that pipe can answer.
    public String find (String ipId, long timeout, boolean nocache) {

	debug ("find " + ipId);
	// First look into the local various tables to see
	// if we know this pipe.

	String peer = findLocal (ipId);

	if (peer == null) {
	    Enumeration enum = null;
	    // We don't know about this pipe: send a query
	    // Finding a pipe remotely is asynchronos: try again later

	    peer = findRemote (ipId, timeout, nocache);
	    return peer;

	} else {
	    return peer;
	}
    }

    // Send a unicast query to verify that the InputPipe is still
    // where it's supposed to be.

    public void sendVerify (String peer, String ipId) {
	
	debug ("sendVerify:");
	debug ("    pipe= " + ipId);
	debug ("    peer= " + peer);

	try {
	    // First create the request message.
	    StructuredTextDocument doc = ( StructuredTextDocument )
		StructuredDocumentFactory.newStructuredDocument (
			    new MimeMediaType( "text/xml" ),
				               "jxta:PipeResolver");

	    Element e = null;
	    e = doc.createElement (MsgTypeTag, QueryMsgType);
	    doc.appendChild (e);
	    e = doc.createElement (PipeIdTag, ipId);
	    doc.appendChild (e);
	    e = doc.createElement (PeerIdTag, peer);
	    doc.appendChild (e);
	    e = doc.createElement (CachedTag, "false");
	    doc.appendChild (e);
	    
	    StringWriter dumped = new StringWriter();
	    doc.sendToWriter( dumped );

	    ResolverQuery query = new ResolverQuery (PipeResolverName,
						     null,
						     null,
						     localPeerId,
						     dumped.toString(), qid++);

	    resolver.sendQuery (peer, query);

	} catch (Exception e) {
	    debug ("sendVerify failed with " + e);
	    return;
	}
    }

    private String findRemote (String ipId, long timeout, boolean nocache) {

	debug ("findRemote " + ipId);
	try {
	    // First create the request message.
	    StructuredTextDocument doc = ( StructuredTextDocument )
		StructuredDocumentFactory.newStructuredDocument (
			    new MimeMediaType( "text/xml" ),
				               "jxta:PipeResolver");

	    Element e = null;
	    e = doc.createElement (MsgTypeTag, QueryMsgType);
	    doc.appendChild (e);
	    e = doc.createElement (PipeIdTag, ipId);
	    doc.appendChild (e);
	    if (nocache) {
		e = doc.createElement (CachedTag, "false");
		doc.appendChild (e);
	    }
	    StringWriter dumped = new StringWriter();
	    doc.sendToWriter( dumped );

	    ResolverQuery query = new ResolverQuery (PipeResolverName,
						     null,
						     null,
						     localPeerId,
						     dumped.toString(), qid++);

	    resolver.sendQuery (null, query);

	} catch (Exception ee) {
	    debug ("findRemote failed with " + ee);
	    return null;
	}
	// Wait until either the given timeout or something arrived
	if (timeout == 0) {
	    // no timeout, just return
	    debug ("findRemote no timeout");
	    return null;
	}

	long endTime = System.currentTimeMillis() + timeout;

	String peer = null;

	while (true) {
            // check the cache.
            peer = findLocal (ipId);
            if (peer != null) {
                debug ("findRemote found " + peer);
                return peer;
            }
            peer = findCached (ipId);
            if (peer != null) {
                debug ("findRemote found " + peer);
                return peer;
            }

            debug ("findRemote waiting...");
            if (timeout < 0) {
                goWait(Long.MAX_VALUE);
            } else {
                long remain = endTime - System.currentTimeMillis();
                if (remain <= 0) break;
                goWait(remain);
            }
	}
	// Timeout has been reached
	debug ("findRemote timeout.");
	return null;
    }


    public void pushResponse(ResolverResponseMsg response) {

	debug ("pushResponse got a response");
	Enumeration enum = null;
	String peer = null;
	MimeMediaType mediaType = null;
	String ipId = null;

	try {
	    mediaType = new MimeMediaType ("text/xml");
	} catch (RuntimeException e) {
	    debug ("pushResponse: cannot get MimeMediaType " + e);
	    return;
	}

	ByteArrayInputStream ip =
	    new ByteArrayInputStream (response.getResponse().getBytes());

	StructuredDocument doc = null;
	try {
	    doc = StructuredDocumentFactory.newStructuredDocument (mediaType,
								   ip);
	} catch( Exception e) {
	    debug ("pushResponse: malformed response - discard");
	    return;
	}
	// XXX: we are doing no document sanity check here.
	// Maybe should we be more careful.
	// to be fixed. lomax@jxta.org

	enum = doc.getChildren ();
	Vector tmp = new Vector();
	boolean found = true;

	while (enum.hasMoreElements()) {
	    TextElement elem = (TextElement) enum.nextElement();
	    if (elem.getName().equals(PeerIdTag)) {
		tmp.addElement (elem.getValue());
		continue;
	    }
	    if (elem.getName().equals(PipeIdTag)) {
		ipId = (String) elem.getValue();
		continue;
	    }
	    if (elem.getName().equals(FoundTag)) {
		found = ((String) elem.getValue()).equals ("true");
		continue;
	    }
	}

	if ((ipId == null) || (tmp.size() == 0)) {
	    debug ("pushResponse: malformed response - discard.");
	    // Malformed response. Just discard.
	    return;
	}

	enum = tmp.elements();
	if (enum == null) {
	    debug ("pushResponse: empty response");
	    return;
	}

	while (enum.hasMoreElements()) {
	    peer = (String) enum.nextElement();
	    if (!found) {
		// We have received a NACK.
		String cachedPeer = findLocal (ipId);
		if ((cachedPeer != null) && cachedPeer.equals (peer)) {
		    // Remove that entry.
		    forgetCached (ipId);
		}
	    } else {
		// put the answer into the cache.
		registerCached (ipId, peer);
	    }
	}

        // If we are a rendezvous, check that we don't have pending queries
        // that need to be answered.
        if (myGroup.isRendezvous()) {

            PendingQuery query = removePendingQuery (ipId);
	    if (query == null) {
                // Nope... nothing to do
		debug ("No pending query");
                return;
            }

	    Vector requesters = query.getRequesters();
	    if (requesters.size() == 0) {
		debug ("No requester");
		return;
	    }

	    StructuredTextDocument tmpDoc = null;
	    ResolverResponse res = null;
	    try {
		// Build the answer
		tmpDoc = (StructuredTextDocument)
		    StructuredDocumentFactory.newStructuredDocument (
		     new MimeMediaType( "text/xml" ), "jxta:PipeResolver");

		Element e = null;
		e = tmpDoc.createElement (MsgTypeTag, AnswerMsgType);
		tmpDoc.appendChild (e);
		e = tmpDoc.createElement (PipeIdTag, ipId);
		tmpDoc.appendChild (e);
		e = tmpDoc.createElement (PeerIdTag, peer);
		tmpDoc.appendChild (e);


		StringWriter dumped = new StringWriter();
		tmpDoc.sendToWriter( dumped );

		res = new ResolverResponse (PipeResolverName,
					    null,
					    null,
					    0,
					    dumped.toString() );
	    } catch (Exception e) {
		debug ("failed [1] with " + e);
		// Too bad.. something failed. Discard
                return;
            }
            // Send the response to all of the pending queries
	    String requestingPeer = null;
	    if (res != null) {
		for (int i = 0; i < requesters.size(); ++ i) {
		    try {
			requestingPeer = (String) requesters.elementAt (i);
			debug ("   sending pending response to " +
			       requestingPeer);
			resolver.sendResponse (requestingPeer, res);
		    } catch (Exception e) {
			continue;
		    }
		}
	    }
	}
	return;
    }



    /**
     * Process the Query, and genrate response
     * @param GenericResolverMsg the query to process
     * @return ResolverResponseMsg "Response"
     */
    public ResolverResponseMsg processQuery(ResolverQueryMsg query)
	throws NoResponseException,
	       DiscardQueryException,
	       IOException,
	       ResendQueryException {

	debug ("processQuery starts");
	MimeMediaType mediaType = null;
	try {
	    mediaType = new MimeMediaType ("text/xml");
	} catch (RuntimeException e) {
	    throw new IOException();
	}
	ByteArrayInputStream ip =
                    new ByteArrayInputStream (query.getQuery().getBytes());

	StructuredTextDocument doc = null;
	try {
	    doc = (StructuredTextDocument)
		StructuredDocumentFactory.newStructuredDocument (mediaType,
								 ip);
	} catch( Exception e ) {
	    debug ("processQuery: malformed request " + e );
            if( debug )
                e.printStackTrace();
	    throw new IOException();
	}
	// XXX: we are doing no document sanity check here.
        // Maybe should we be more careful. to be fixed. lomax@jxta.org

	String pId = null;
	boolean cached = true;
	String destPeer = null;

	Enumeration enum = doc.getChildren ();

	while (enum.hasMoreElements()) {
	    TextElement elem = (TextElement) enum.nextElement();
	    if (elem.getName().equals(PipeIdTag)) {
		pId = (String) elem.getValue();
		continue;
	    }
	    if (elem.getName().equals(PeerIdTag)) {
		destPeer = (String) elem.getValue();
		continue;
	    }
	    if (elem.getName().equals(CachedTag)) {
		cached = ((String) elem.getValue()).equals ("true");
		continue;
	    }
	}
	if (pId == null) {
	    debug ("processQuery: malformed request, no PipeId.");
	    throw new IOException();
	}

	if ((destPeer != null) &&!destPeer.equals (localPeerId)) {
	    // This request wasn't for this peer.
	    throw new DiscardQueryException ();
	}

	int queryId = query.getQueryId();
	String peer = null;

	// XXX: Only one query at the time...
        // Maybe that should be changed later. lomax@jxta.org

	if (!cached) {
	    // The source of the query has requeried not to receive
	    // cached information.
	    if (pipes.get (pId) != null) {
		peer = localPeerId;
	    } else {
		peer = null;
	    }
	} else {
	    // We can look into our cache.
	    peer = findLocal (pId);
	}

	if ((peer == null) && (destPeer == null)) {
	    // This request was sent to everyone.
	    // Check if we have a pending query for this pipe
	    if (isPendingQuery (pId)) {
		// Best is simply to tell the resolver to ignore
		// this request.
		debug ("  allready got this query. Discard");
		throw new DiscardQueryException ();
	    }
	    // Check if we want to get the answer.
	    if (myGroup.isRendezvous()) {
		processNotFound (query.getSrc(), pId);
		debug ("   resend query");
		throw new ResendQueryException();
	    } else {
		debug ("   forward query");
		throw new NoResponseException ();
	    }
	}
		    
	try {
	    // Build the answer
	    doc = (StructuredTextDocument)
            StructuredDocumentFactory.newStructuredDocument (
            new MimeMediaType( "text/xml" ), "jxta:PipeResolver");

	    Element e = null;
	    e = doc.createElement (MsgTypeTag, AnswerMsgType);
	    doc.appendChild (e);
	    e = doc.createElement (PipeIdTag, pId);
	    doc.appendChild (e);
	    if (peer == null) {
		// We are sending a NACK
		e = doc.createElement (PeerIdTag, localPeerId);
		doc.appendChild (e);
		e = doc.createElement (FoundTag, "false");
		doc.appendChild (e);
	    } else {
		e = doc.createElement (PeerIdTag, peer);
		doc.appendChild (e);
	    }
	    StringWriter dumped = new StringWriter();
	    doc.sendToWriter( dumped );

	    ResolverResponse res = new ResolverResponse (PipeResolverName,
							 null,
							 null,
							 queryId,
							 dumped.toString() );

	    debug ("processQuery: sending answer.");
	    return res;

	} catch (Exception ee) {
	    debug ("processQuery: error while processing query:");
	    if (debug) ee.printStackTrace();
	    throw new IOException();
	}
    }

    private void debug (String str) {
	if (debug) {
	    System.out.println ("[PipeResolver] " + str);
	}
    }


    private synchronized void goWait (long timeout) {
	try {
	    wait(timeout);
	} catch (Exception e) {
	}
    }


    private synchronized void notifyWaitingThreads () {
        notifyAll();
    }

    // If this peer is a rendez-vous, cache the query, and ask the ResolverService
    // to resolve that request as for ourselves.

    private boolean processNotFound (String requestingPeer,
				     String requestedPipe) {

        boolean hasToResend = true;

        debug ("processNotFound for " + requestingPeer);
        if (myGroup.isRendezvous()) {
            if (isPendingQuery (requestedPipe)) {
                // There is already a pending query for that peer. No need to resend
                // the query (that has already been done.
                hasToResend = false;
            }
            // Add the current query to the pending queries
            addPendingQuery (requestingPeer, requestedPipe);
            return hasToResend;
        } else {
            // We are not a RendezVous, just ignore.
            return false;
        }
    }



    protected class PendingQuery {

	protected Vector requesters = null;
        protected String requestedPipe = null;
        protected long   timeOfQuery = 0;
	protected long   delay = 0;
	private long InitialDelay = 120000; // 2 minutes

        public PendingQuery (String requested) {

	    requesters = new Vector();
            requestedPipe = requested;
            timeOfQuery = System.currentTimeMillis();
	    this.delay = this.timeOfQuery + InitialDelay;
        }

        public boolean equals (Object obj) {

            PendingQuery tmp =  (PendingQuery) obj;
            return  (tmp.requestedPipe.equals (requestedPipe));
        }

	public Vector getRequesters() {
	    return requesters;
	}
    }


    private synchronized void addPendingQuery (String requestingPeer,
                                               String requestedPipe) {


        if (isPendingQuery (requestedPipe)) {
            // We already have a pending query of that type

            PendingQuery query = getPendingQuery (requestedPipe);

            if (query.requesters.contains (requestingPeer)) {
                // This requesting peer is already recorded.
                return;
            }
            // add the requesting peer.
            query.requesters.addElement (requestingPeer);

            return;
        }



        debug ("addPendingQuery for " + requestingPeer);
        PendingQuery query = new PendingQuery (requestedPipe);

	query.requesters.addElement (requestingPeer);
        // Add this new query
        pendingQueries.addElement (query);

        return;
    }


    private synchronized PendingQuery getPendingQuery (String requestedPipe) {

        PendingQuery query = null;

        for (int i = 0; i < pendingQueries.size(); ++i) {

            try {
                query = (PendingQuery) pendingQueries.elementAt (i);
                if (query.requestedPipe.equals (requestedPipe)) {
                    return query;
                }
            } catch (Exception e) {
                continue;
            }
        }
        return null;
    }


    private synchronized PendingQuery removePendingQuery (String requestedPipe) {


        debug ("removePendingQuery for " + requestedPipe);
        PendingQuery query = null;

        for (int i = 0; i < pendingQueries.size(); ++i) {

            try {
                query = (PendingQuery) pendingQueries.elementAt (i);
                if (query.requestedPipe.equals (requestedPipe)) {
                    // Remove that query
                    try {
                        pendingQueries.removeElementAt (i);
                    } catch (Exception e) {
			break;
                    }
		    break;
                }
            } catch (Exception e) {
                continue;
            }
        }
	return query;
    }


    private synchronized boolean isPendingQuery (String requestedPipe) {

        PendingQuery query = null;

	debug ("isPendingQuery for " + requestedPipe);

        long currentTime = System.currentTimeMillis();

        for (int i = 0; i < pendingQueries.size(); ++i) {

            try {
                query = (PendingQuery) pendingQueries.elementAt (i);
                if (query.requestedPipe.equals (requestedPipe)) {
                    if (query.delay <= currentTime) {
                        // We allow a new request to be sent
                        // Reset delay
			debug ("   allow resend");
                        query.delay = currentTime + query.InitialDelay;
                        return false;
                    }
		    debug ("   disallow resend");
                    return true;
                }
            } catch (Exception e) {
                return false;
            }
        }
        return false;
    }

    private synchronized void refreshPendingQueries () {

	if (pendingQueries == null) {
	    return;
	}

	PendingQuery query = null;
	long currentTime = System.currentTimeMillis();
	for (int i = 0; i < pendingQueries.size(); ++i) {
	    try {
		query = (PendingQuery) pendingQueries.elementAt (i);
		if (query == null) {
		    debug ("refreshQueries null element");
		    continue;
		}
		if ((query.timeOfQuery + MaxPendingQueryDelay) < currentTime) {
		    pendingQueries.removeElementAt (i);
		    // The index has to be decremented because of the remove
		    --i;
		}
	    } catch (Exception e) {
		debug ("refreshQueries failed " + e);
		break;
	    }
	}
    }

    public void signal (JxtaTimer timer) {

	refreshCachedPipes();
	refreshPendingQueries();
    }
}



