/************************************************************************
 *
 * $Id: PipeService.java,v 1.8 2001/06/07 21:54:38 raffi Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.pipe;

import java.util.*;
import java.io.*;
import java.net.*;

import net.jxta.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.protocol.*;
import net.jxta.endpoint.*;
import net.jxta.service.*;
import net.jxta.peer.*;

import net.jxta.document.Advertisement;
import net.jxta.protocol.ServiceAdvertisement;

import net.jxta.impl.peergroup.*;
import net.jxta.impl.endpoint.*;




/**
 * This class implements the NetPipe interface.
 */

public class PipeService implements Pipe {

    private RefPeerGroup myGroup = null;
    private static Hashtable queueTable = null;
    private Endpoint  endpoint = null;
    private PipeResolver pipeResolver = null;
    private ServiceAdvertisement advertisement = null;

    /**
     * Supply arguments and starts this service if it hadn't started by itself.
     *
     * Currently this service starts by itself and does not expect
     * arguments.
     *
     * @param arg A table of strings arguments.
     * @return int status indication.
     */
    public int startApp(String[] arg) {
        return 0;
    }

    /**
     * Ask this service to stop.
     *
     * This request is currently ignored.
     */
    public void stopApp() {
    }

    /**
     * Service objects are not manipulated directly to protect usage
     * of the service. A Service interface is returned to access the service
     * methods.
     *
     * @return Service public interface of the service
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0 
     */
    public Service getInterface() {
	return new PipeInterface(this);
    }

    /**
     * Returns the advertisement for that service.
     *
     * @return Advertisement the advertisement.
     *
     * @version $Revision: 1.8 $
     * @since JXTA 1.0 
     */
    public Advertisement getAdvertisement() {
        return advertisement;
    }

    /**
     * Init routine
     *
     * @param myGroup The Group associated with this service
     */

    public void init (PeerGroup g, Advertisement sadv)
    {

        advertisement = (ServiceAdvertisement) sadv;
	this.myGroup = (RefPeerGroup) g;

	// Creating the hash table that will contain all the input queues.
	queueTable = new Hashtable();

	// create our resolver handler; it will register itself w/ the resolver.
	pipeResolver= new PipeResolver(g);
	
	// Get the endpoint.
	endpoint = myGroup.getEndpoint();
    }

    /**
     * Constructor: this constructor is only needed in order to create
     * the listener thread of the NodePipe.
     */
    public PipeService() {
    }


    /**
     * Creates an InputPipe of a NetPipe.
     *
     * @param adv is the public advertisement of the NetPipe.
     * @return InputPipe returns a InputPipe associated to the NetPipe.
     */
    public InputPipe createInputPipe (PipeAdvertisement adv)
	throws IOException {

	// Recover the PipeId from the PipeService Advertisement
	PipeID pipeId = (PipeID) adv.getPipeID().clone();

	// create an InputPipe.
	InputPipe ip = new InputPipeImpl (myGroup,
					      pipeResolver,
					      pipeId);

	return ip;
    }


    /**
     * Creates an OutputPipe associated to a NetPipe.
     *
     * @param adv is the public advertisement of the destination NetPipe.
     * @param type the type of the output pipe to create
     * @param timeout the number of milliseconds to wait during pipe creation
     * @return OutputPipe returns an OutputPipe associated to the
     * destination NetPipe.
     */

    public OutputPipe    createOutputPipe (PipeAdvertisement pipeAdv, int type, long timeout)
	throws IOException {

	// Recover the PipeId from the PipeService Advertisement
	PipeID pipeId  = (PipeID) pipeAdv.getPipeID ().clone();

	// create a OutputPipe
	OutputPipe op = null;

	switch (type) {
	case Pipe.HandChecking:
	    // XXX: Not yet implemented.
	    throw new IOException();
	case Pipe.NonBlocking:
	    op = new NonBlockingOutputPipe (myGroup,
					    pipeResolver,
					    pipeId,
					    null,
					    timeout);
	    break;
	case Pipe.Propagate:
	    // XXX: Not yet Implemented
	    throw new IOException();
	default:
	    throw new IOException();
	}
	return op;
    }

    /**
     * create an OutputPipe from the pipe Advertisement giving a PeerId(s)
     * where the corresponding InputPipe is supposed to be.
     *
     * @param adv is the advertisement of the NetPipe.
     * @param type is an int and contains the name of the diffusion mode for
     * this OutputPipe.
     * @param peers is an enumeration of the PeerId of thepeers where to look
     * for the corresponding Pipes
     * @param timeout the number of milliseconds to wait for during pipe creation
     * @return OuputPipe
     * @throws IOException if none of the peers in the enumeration has the
     * corresponding OutputPipe
     */

    public OutputPipe    createOutputPipe (PipeAdvertisement adv,
					   int type,
					   Enumeration peers,
					   long timeout)
	throws IOException {

	// Recover the PipeId from the PipeService Advertisement
	PipeID pipeId  = (PipeID) adv.getPipeID ().clone();

	// create a OutputPipe
	OutputPipe op = null;

	// For the time being, only support one element in the Enumeration
	// XXX: to be fixed when one to many pipe will be implemented. lomax@jxta.org
	String peer = null;
	if ((peers == null) || (!peers.hasMoreElements())) {
	    throw new IOException();
	}

	try {
	    peer = ((PeerID) peers.nextElement()).toString();
	} catch (Exception e) {
	}

	switch (type) {
	case Pipe.HandChecking:
	    // XXX: Not yet implemented.
	    throw new IOException();
	case Pipe.NonBlocking:
	    op = new NonBlockingOutputPipe (myGroup,
					    pipeResolver,
					    pipeId,
					    peer,
					    timeout);
	    break;
	case Pipe.Propagate:
	    // XXX: Not yet Implemented
	    throw new IOException();
	default:
	    throw new IOException();
	}
	return op;
    }

    /**
     * Creates a new Message
     *
     * @return Message returns a newly allocated Message.
     */
    public Message createMessage() {
	return new MessageImpl();
    }

    public void addAddresses (PipeAdvertisement adv, Enumeration addrs) {

	if ((adv == null) || (addrs == null)) {
	    return;
	}

	// Not yet implemented. In fact, this might not be implemented, and
	// instead, removed from the public API.
	// XXX: to be revisited: lomax@jxta.org
    }
}
