/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: HTTPAdv.java,v 1.15 2001/06/09 19:00:39 bondolo Exp $
 */

package net.jxta.impl.protocol;

import java.io.*;
import java.util.*;
import net.jxta.document.*;
import net.jxta.protocol.*;

/**
 * This class implements the HTTPAdvertisement. While being useful,
 * application are not required to use it. However, applications that
 * donot want to use it will still have to provide a class that
 * extends net.jxta.HTTP.HTTPAdvertisement.
 *
 */

public class HTTPAdv extends TransportAdvertisement {

    private String port = null;
    private String protocol = null;
    private String peerId = null;
    private boolean propagateEnabled = false;
    private String proxy  = null;
    private Vector routers = null;
    private String server = null;

    private static final String RouterTag = "Router";
    private static final String ProxyTag = "Proxy";
    private static final String ServerTag = "Server";
    private static final String PropagateTag = "Propagate";

    // These are for configuration; They get saved in the document only if they are
    // off and the correspondig item has a non-null value. So that the value is not lost.
    // When HttpTransport is done initializing, the unused values are set to null, and thus
    // pruned from the published adv.

    private boolean proxyEnabled = true;
    private boolean routerEnabled = true;
    private boolean serverEnabled = true;

    private static final String RouterOffTag = "RouterOff";
    private static final String ProxyOffTag = "ProxyOff";
    private static final String ServerOffTag = "ServerOff";

    
    public static String getAdvertisementType() {
        return "jxta:HTTPTransportAdvertisement";
    }
    
    public HTTPAdv() {
    }
    
    /**
     *  @deprecated Please use the Advertisement factory and accessors
     **/
    public HTTPAdv(String protocol, String port, String id) {
        this.protocol = protocol;
        this.port = port;
        this.peerId = id;
    }
    
    public HTTPAdv( Element root ) throws Exception {
        initialize( root );
    }
    
    public void initialize(Element root) {
        
        if( !TextElement.class.isInstance( root ) )
            throw new IllegalArgumentException( getClass().getName() + " only supports TextElement" );
        
        TextElement doc = (TextElement) root;
        

        if( !getAdvertisementType().equals( doc.getName() ) )
            throw new IllegalArgumentException( "Could not construct : "
            + getClass().getName()
            + "from doc containing a "
            + doc.getName() );
        
        Enumeration elements = doc.getChildren();

        while (elements.hasMoreElements()) {
            TextElement elem = (TextElement) elements.nextElement();
            String tag = elem.getName();
            String value = null;

            if ( tag.equals(RouterOffTag) ) {
                routerEnabled = false;
                continue;
            }
            if ( tag.equals(ProxyOffTag) ) {
                proxyEnabled = false;
                continue;
            }
            if ( tag.equals(ServerOffTag) ) {
                serverEnabled = false;
                continue;
            }

            try {
                value = elem.getTextValue();
            } catch (Exception e) {
                continue; // In case there's no value (not any of the below tags).
            }

            if( tag.equals("Protocol")) {
                setProtocol( value );
                continue;
            }
            if( tag.equals("Port")) {
                setPort( value );
                continue;
            }
            if( tag.equals("PeerId")) {
                setPeerId( value );
                continue;
            }
            
            if ( tag.equals(RouterTag) ) {
                if (routers == null) routers = new Vector();
                routers.addElement(value);
                continue;
            }
            if ( tag.equals(ProxyTag) ) {
                proxy = value;
                continue;
            }
            if ( tag.equals(ServerTag) ) {
                server = value;
                continue;
            }

            // NB we keep that one for now, but http transport ignores it.
            if ( tag.equals (PropagateTag) ) {
                if (value.equals("true") || value.equals("TRUE")) {
                    propagateEnabled = true;
                }
            }
        }

        // For consistency we force the flags to "disabled" for items we do not
        // have data for. However, the flags truely matter only when there is
        // data.
        if (proxy == null) proxyEnabled = false;
        if (server == null) serverEnabled = false;
        if (routers == null) routerEnabled = false;
    }

    // NB: we do not try to enforce dependency rules such as
    // propagate defined only if ServerName defined, or Proxy only
    // when router, because we want to convey the complete configuration,
    // even items corresponding to not currently enabled features.
    // HttpTransport will gracefully disregard items that have no use.
    // in the current context.
    public Document getDocument( MimeMediaType encodeAs ) throws Exception {
        StructuredTextDocument adv = adv = (StructuredTextDocument)
            StructuredDocumentFactory.newStructuredDocument(
            encodeAs, getAdvertisementType() );
        
        Element e1 = adv.createElement("Protocol", getProtocol());
        adv.appendChild(e1);
        
        Element e2 = adv.createElement("Port", getPort());
        adv.appendChild(e2);
        
        Element e3 = adv.createElement("PeerId", getPeerId());
        adv.appendChild(e3);

        Element ext;
        Element detail;
	if (routers != null) {
            Enumeration e = routers.elements();
            while (e.hasMoreElements()) {
                String router = (String) e.nextElement();
                ext = adv.createElement (RouterTag, router);
                adv.appendChild (ext);
            }
            // If disabled, say it; otherwise it is assumed on. In published
            // advs, we only keep data for items that are ON, so we do not
            // have to clutter them with the flag.
            if (! routerEnabled) {
                ext = adv.createElement (RouterOffTag);
                adv.appendChild (ext);
            }
        }
        if (proxy != null) {
            ext = adv.createElement (ProxyTag, proxy);
            adv.appendChild (ext);
            // If disabled, say it; otherwise it is assumed on. In published
            // advs, we only keep data for items that are ON, so we do not
            // have to clutter them with the flag.
            if (! proxyEnabled) {
                ext = adv.createElement (ProxyOffTag);
                adv.appendChild (ext);
            }
        }
	if (server != null) {
	    ext = adv.createElement (ServerTag, server);
            adv.appendChild (ext);
            // If disabled, say it; otherwise it is assumed on. In published
            // advs, we only keep data for items that are ON, so we do not
            // have to clutter them with the flag.
            if (! serverEnabled) {
                ext = adv.createElement (ServerOffTag);
                adv.appendChild (ext);
            }
	}

        // Still save that one for now, but HttpTransport ignores it.
        ext = adv.createElement (PropagateTag, "" + propagateEnabled);
        adv.appendChild (ext);

        return adv;
    }
    
    /**
     * returns the protocol
     *@return string protocol
     */
    
    public String getProtocol() {
        return protocol;
    }
    
    /**
     * returns the protocol
     * @return string protocol
     */
    public void setProtocol(String protocol) {
        this.protocol = protocol;
    }
    
    /**
     * returns the port
     *@return string port
     *
     * @since JXTA 1.0
     */
    
    public String getPort() {
        return port;
    }
    
    /**
     * returns the port
     * @param string port
     *
     * @since JXTA 1.0
     */
    public void setPort(String port) {
        this.port = port;
    }
    
    /**
     * returns the peerId
     *@return string peerId
     *
     * @since JXTA 1.0
     */
    
    public String getPeerId() {
        return peerId;
    }
    
    public void setPeerId(String id) {
        this.peerId = id;
    }

    public String   getProxy(){ return  proxy;  }
    public Vector getRouters(){ return routers; }
    public String  getServer(){ return server;  }
    public boolean  getProxyEnabled(){ return  proxyEnabled; }
    public boolean getRouterEnabled(){ return routerEnabled; }
    public boolean getServerEnabled(){ return serverEnabled; }

    public boolean getPropagateEnabled() { return propagateEnabled; }

    // If one of proxy, server, or router is cleared, the corresponding
    // enabled flag should be false (the opposite is not true).

    public void  setProxy(String name){
        proxy = name;
        if (name == null) proxyEnabled = false;
    }
    public void setRouters(Vector list){
        routers = list;
        if (routers == null) routerEnabled = false;
    }
    public void setServer(String name){
        server = name;
        if (name == null) serverEnabled = false;
    }

    public void  setProxyEnabled(boolean enabled){  proxyEnabled = enabled; }
    public void setRouterEnabled(boolean enabled){ routerEnabled = enabled; }
    public void setServerEnabled(boolean enabled){ serverEnabled = enabled; }

    public void setPropagateEnabled(boolean state){ propagateEnabled = state; }
}
