/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: PeerAdv.java,v 1.12 2001/05/23 01:00:20 lomax Exp $
 */

package net.jxta.impl.protocol;

import java.io.*;
import java.util.*;
import java.lang.reflect.InvocationTargetException;

import net.jxta.document.*;
import net.jxta.protocol.*;


/**
 * This class implements the PeerAdvertisement. While being useful,
 * application are not required to use it. However, applications that
 * donot want to use it will still have to provide a class that
 * extends net.jxta.protocol.PeerAdvertisement.
 *
 */

public class PeerAdv extends PeerAdvertisement {
    
    public PeerAdv() {

super();
    }

    public PeerAdv (String pid, String name, String keywords,
                    Hashtable services, Hashtable endpoints) {
        setPid (pid);
        setName (name);
        setKeywords(keywords);
        setServiceAdvertisements(services);
        setEndpointAdvertisements(endpoints);
    }
    
    public PeerAdv( Element root ) throws Exception {
        initialize( root );
    }
    
    public void initialize(Element root) throws Exception {
        
        if( !TextElement.class.isInstance( root ) )
            throw new IllegalArgumentException( getClass().getName() + " only supports TextElement" );
        
        TextElement doc = (TextElement) root;
        
        // Set up everything empty
        setName( null );
        setKeywords( null );
        setPid( null );
        setApp(null);
        setGroupImpl(null);
        
        // Get local handles to these tables, for convenience.
        Hashtable servs = getServiceAdvertisements();
        Hashtable endps = getEndpointAdvertisements();
        
        // Fill the stuff with what we find in the adv.
        Enumeration elements;
        elements = doc.getChildren( );
        
        while ( elements.hasMoreElements() ) {
            TextElement element = (TextElement) elements.nextElement();
            
            // The basics
            if (element.getName().equals("Name")) {
                setName (element.getTextValue());
                continue;
            }
            
            if (element.getName().equals( "Keywords" )) {
                setKeywords ( element.getTextValue() );
                continue;
            }
            
            if (element.getName().equals( "Pid" )) {
                setPid ( element.getTextValue() );
                continue;
            }
            
            // Services
            if (element.getName().equals( "Service" )) {
                ServiceAdv sadv = new ServiceAdv(element);
                servs.put(sadv.getName(), sadv);
                continue;
            }
            
            
            // Endpoints:
            if (element.getName().equals( "Endpoint" )) {
                EndpointAdv eadv = new EndpointAdv(element);
                endps.put(eadv.getName(), eadv);
                continue;
            }
            
        }
    }
    
    
    public Document getDocument( MimeMediaType encodeAs ) throws Exception {
        StructuredTextDocument adv = null;
        
        adv = (StructuredTextDocument)
        StructuredDocumentFactory.newStructuredDocument(
        encodeAs, getAdvertisementType() );
        
        Element e1 = adv.createElement ("Name", getName());
        adv.appendChild (e1);
        
        Element e2 = adv.createElement ("Keywords", getKeywords());
        adv.appendChild (e2);
        
        Element e = adv.createElement ("Pid", getPid());
        adv.appendChild (e);

        /*
         * create all the services
         *  FIXME MJD 20010308  This really must use the
         * getServiceAdvertisements getDocument
         */

        for (Enumeration p = getServiceAdvertisements().elements();
             p.hasMoreElements();) {

            Element e3 = adv.createElement ("Service");
            ServiceAdvertisement service
                = (ServiceAdvertisement) p.nextElement();

            Element e11 = adv.createElement ("Name", service.getName());
            e3.appendChild (e11);

            Element e12 = adv.createElement ("Keywords",
                                             service.getKeywords());
            e3.appendChild (e12);

            PipeAdvertisement pipe = service.getPipe();

            Element e13 = adv.createElement ("PipeService");

            e3.appendChild (e13);
            
            if (pipe != null) {
                Element e131 = adv.createElement ("id", pipe.getPipeID());
                e13.appendChild (e131);
            }

            Vector params = service.getParams();

            if (params != null) {
                for (Enumeration ep = params.elements();
                     ep.hasMoreElements();
                     ) {

                    Element e134 = adv.createElement
                        ("Params", (String) ep.nextElement());
                    e3.appendChild (e134);
                }
            }

            Element e15 = adv.createElement ("Name", service.getName());
            e3.appendChild (e15);

            Element e16 = adv.createElement ("Version", service.getVersion());
            e3.appendChild (e16);
            Element e17 = adv.createElement ("Uri", service.getUri());
            e3.appendChild (e17);
            Element e18 = adv.createElement ("Provider",
                                             service.getProvider());
            e3.appendChild (e18);
            Element e19 = adv.createElement ("Security",
                                             service.getSecurity());
            e3.appendChild (e19);
            Element e20 = adv.createElement ("Code", service.getCode());
            e3.appendChild (e20);
            adv.appendChild (e3);
        }

        /*
         * create all the endpoints
         *
         *  FIXME MJD 20010308  This really must use the endpoint
         * advertisement getDocument
         */


        for (Enumeration p = getEndpointAdvertisements().elements();
        p.hasMoreElements();) {
            Element e4 = adv.createElement ("Endpoint");
            
            EndpointAdvertisement endpoint =
                (EndpointAdvertisement) p.nextElement();

            Element e21 = adv.createElement ("Name", endpoint.getName());
            e4.appendChild (e21);
            
            Element e22 = adv.createElement ("Keywords", endpoint.getKeywords());
            e4.appendChild (e22);
            
            Element e23 = adv.createElement ("Address",
                                             endpoint.getEndpointAddress());
            e4.appendChild (e23);
            
            Element e44 = adv.createElement ("Transport");
            
            TCPAdv transport= (TCPAdv) endpoint.getTransportAdvertisement();
            
            Element e45 = adv.createElement ("Protocol",
                                             transport.getProtocol());
            e44.appendChild (e45);
            
            Element e46 = adv.createElement ("Port", transport.getPort());
            e44.appendChild (e46);
            Element e47 = adv.createElement ("MulticastAddr",
                                             transport.getMulticastAddr());
            e44.appendChild (e47);
            Element e48 = adv.createElement ("MulticastPort",
                                             transport.getMulticastPort());
            e44.appendChild (e48);
            Element e49 = adv.createElement ("MulticastSize",
                                             transport.getMulticastSize());
            e44.appendChild (e49);
            e4.appendChild (e44);
            
            adv.appendChild (e4);
        }
        
        return adv;
    }
    
    public String toString() {
        throw new RuntimeException();
    }
    
}



