/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TCPAdv.java,v 1.9 2001/05/23 01:00:20 lomax Exp $
 */

package net.jxta.impl.protocol;

import java.io.*;
import java.util.*;
import net.jxta.document.*;
import net.jxta.protocol.*;

/**
 * This class implements the TCPAdvertisement. While being useful,
 * application are not required to use it. However, applications that
 * donot want to use it will still have to provide a class that
 * extends net.jxta.TCP.TCPAdvertisement.
 *
 */

public class TCPAdv extends TransportAdvertisement {
    
    private String port = null;
    private String protocol = null;
    private String multicastaddr = null;
    private String multicastport = null;
    private String multicastsize = null;
    private String publicAddress = null;
    
    public static String getAdvertisementType() {
        return "jxta:TCPTransportAdvertisement";
    }
    
    public TCPAdv() {
    }
    
    public TCPAdv( Element root ) throws Exception {
        initialize( root );
    }
    
    public void initialize(Element root) {
        
        if( !TextElement.class.isInstance( root ) )
            throw new IllegalArgumentException( getClass().getName() + " only supports TextElement" );
        
        TextElement doc = (TextElement) root;
        
        if( !getAdvertisementType().equals( doc.getName() ) )
            throw new IllegalArgumentException( "Could not construct : "
            + getClass().getName()
            + "from doc containing a "
            + doc.getName() );
        
        Enumeration elements = doc.getChildren();
        while (elements.hasMoreElements()) {
            TextElement elem = (TextElement) elements.nextElement();
            if( elem.getName().equals("Protocol")) {
                setProtocol( elem.getTextValue() );
                continue;
            }
            if( elem.getName().equals("Port")) {
                setPort( elem.getTextValue() );
                continue;
            }
            if( elem.getName().equals("MulticastAddr")) {
                setMulticastAddr( elem.getTextValue() );
                continue;
            }
            if( elem.getName().equals("MulticastPort")) {
                setMulticastPort( elem.getTextValue() );
                continue;
            }
            if( elem.getName().equals("MulticastSize")) {
                setMulticastSize( elem.getTextValue() );
                continue;
            }
            if( elem.getName().equals("PublicAddress")) {
                setPublicAddress( elem.getTextValue() );
                continue;
            }
        }
    }
    
    public TCPAdv(String protocol, String port, String multicastaddr,
    String multicastport, String multicastsize, String pa) {
        this.protocol = protocol;
        this.port = port;
        this.multicastaddr = multicastaddr;
        this.multicastport = multicastport;
        this.multicastsize = multicastsize;
        this.publicAddress = pa;
    }
    
    public Document getDocument( MimeMediaType encodeAs ) throws Exception {
        StructuredTextDocument adv = adv = (StructuredTextDocument)
            StructuredDocumentFactory.newStructuredDocument(
            encodeAs, getAdvertisementType() );

        Element e11 = adv.createElement("Protocol", getProtocol());
        adv.appendChild(e11);
        
        Element e12 = adv.createElement("Port", getPort());
        adv.appendChild(e12);
        Element e13 = adv.createElement("MulticastAddr", getMulticastAddr());
        adv.appendChild(e13);
        Element e14 = adv.createElement("MulticastPort", getMulticastPort());
        adv.appendChild(e14);
        Element e15 = adv.createElement("MulticastSize", getMulticastSize());
        adv.appendChild(e15);
        Element e16 = adv.createElement("PublicAddress", getPublicAddress());
        adv.appendChild(e16);
        
        return adv;
    }
    
    /**
     * returns the protocol
     *@return string protocol
     *
     * @since JXTA 1.0
     */
    
    public String getProtocol() {
        return protocol;
    }
    
    /**
     * returns the protocol
     * @return string protocol
     *
     * @since JXTA 1.0
     */
    public void setProtocol(String protocol) {
        this.protocol = protocol;
    }
    
    /**
     * returns the port
     *@return string port
     *
     * @since JXTA 1.0
     */
    
    public String getPort() {
        return port;
    }
    
    /**
     * returns the port
     * @param string port
     *
     * @since JXTA 1.0
     */
    public void setPort(String port) {
        this.port = port;
    }
    
    /**
     * returns the multicastaddr
     * @return string multicastaddr
     *
     * @since JXTA 1.0
     */
    public String getMulticastAddr() {
        return multicastaddr;
    }
    
    /**
     * set the multicastaddr
     * @param string set multicastaddr
     *
     * @since JXTA 1.0
     */
    
    public void setMulticastAddr(String multicastaddr) {
        this.multicastaddr = multicastaddr;
    }
    
    /**
     * returns the multicastport
     * @return string multicastport
     *
     * @since JXTA 1.0
     */
    public String getMulticastPort() {
        return multicastport;
    }
    
    /**
     * set the multicastport
     * @param string set multicastport
     *
     * @since JXTA 1.0
     */
    
    public void setMulticastPort(String multicastport) {
        this.multicastport = multicastport;
    }
    
    /**
     * returns the multicastsize
     * @return string multicastsize
     *
     * @since JXTA 1.0
     */
    public String getMulticastSize() {
        return multicastsize;
    }
    
    /**
     * set the multicastsize
     * @param string set multicastsize
     *
     * @since JXTA 1.0
     */
    
    public void setMulticastSize(String multicastsize) {
        this.multicastsize = multicastsize;
    }
    
    /**
     * returns the public address
     * @return string publicaddress
     *
     * @since JXTA 1.0
     */
    public String getPublicAddress() {
        return publicAddress;
    }
    
    /**
     * set the multicastsize
     * @param string set multicastsize
     *
     * @since JXTA 1.0
     */
    
    public void setPublicAddress(String pa) {
        this.publicAddress = pa;
    }
    
}



