/************************************************************************
 *
 * $Id: RdvMonitor.java,v 1.15 2001/05/23 01:00:21 lomax Exp $
 *
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 *    must not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *********************************************************************************/

package net.jxta.impl.rendezvous;

import java.util.*;

import net.jxta.peer.*;
import net.jxta.protocol.*; 
import net.jxta.document.*;
import net.jxta.rendezvous.*;
import net.jxta.discovery.*;
import net.jxta.peergroup.*;
import net.jxta.util.*;
import net.jxta.impl.util.*;

/** 
 * This class implements a very simple RendezVous Monitor
 */

public class RdvMonitor implements RendezVousMonitor, JxtaTimerHandler {

    private boolean debug = false;
    private PeerGroup group = null;
    private String    gId = null;
    private Discovery discovery = null;
    private long lease = 0;
    private JxtaTimer timer = null;
    private Vector rdvs = null;
    private RendezVousImpl rendezvous = null;

    // LeaseRenewalDelay is the period of the timer that will renew the lease
    // to the rendezvous. If a rendezvous gives a lease less than that delay
    // this client will lose the lease. A lease of less than 5 minutes does not
    // seem to make any sense, since some of the transport protocols have at lease
    // 2 minutes latency, 5 minutes should be fine.

    private static final long LeaseRenewalDelay = 15 * 60 * 1000; // 15 Minutes


    public class RdV {

	protected String peer;
	protected long   lease;

	public RdV (String peer) {

	    this.peer = peer;
	    this.lease = 0;
	}

	public boolean equals (Object obj) {

	    RdV rdv = (RdV) obj;
	    return (rdv.peer.equals (peer));
	}
    }

    public RdvMonitor (PeerGroup g, RendezVousImpl rendezvous) {

	this.debug = g.getDebug();
	this.group = g;
	this.gId = g.getID().toString();
	this.rdvs = new Vector();
	this.timer = null;
	this.rendezvous = rendezvous;
        discovery = group.getDiscovery();
    }


    /**
     * This method is called each time a RendezVous peer is not
     * reachable anymore.
     *
     * @param peer is the PeerId of the RendezVous peer.
     */
    public void lostConnection (PeerID peer) {

	debug ("lostConnection with RendezVous");
	debug ("     peerId = " + peer.toString());
    }

    /**
     * This method is called by the RendezVous service in order
     * to provide advertisement about other RendezVous peers.
     * This is usefull when a RendezVous peer wants to balance its load
     * to other RendezVous
     */
    public void newRendezVous (Advertisement adv) {

	debug ("newRendezVous: got new RendezVous advertisement.");
	// Publish the rendez vous advertisement
        // First publish the advertisement
        if (discovery != null) {
            try {
                discovery.publish ( adv, Discovery.PEER);
            } catch (Exception e) {
            }
        }
    }

    /**
     * This method is called by the RendezVous service to notify the
     * monitor that a new RendezVous has accepted the connection.
     *
     * @param peer is the PeerId of the new connect RendezVous peer.
     * @param lease is the time in millisecond that the RendezVous is
     * ready to serve.
     */
    public void isConnected (PeerID peer, long lease) {

	debug ("isConnected: new connection to a RendezVous");
	debug ("   peer = " + peer.toString());
	debug ("  lease = " + lease);

	addRendezVous (peer.toString(), lease);
    }

    private synchronized void addRendezVous (String peer, long lease) {

	RdV rdv = new RdV (peer);
	if (rdvs.contains (rdv)) {
	    // This is only a confirmation of a renewal of the lease.
	    try {
		int index = rdvs.indexOf (rdv);
		if (index == -1) {
		    // This should not happen
		    debug ("addRendezVous cannot access RdV object");
		    return;
		}
		rdv = (RdV) rdvs.elementAt (index);
	    }  catch (Exception e) {
		debug ("addRendez failed [1] " + e);
	    }
	} else {
	    // This is a new Rendezvous
	    rdvs.addElement (rdv);
	}
	// Set the lease
	long time = System.currentTimeMillis();
	if ((time + lease) < 0) {
	    // The rendez has given an infinite lease.
	    // However, it is a good thing to still renew the lease once in a while
	    lease = 60 * 60 * 1000; // 1 Hour
	}
	rdv.lease = time + lease;
	// Check if the timer thread is already started
	if (timer == null) {
	    timer = new JxtaTimer (this, LeaseRenewalDelay, true);
	}
    }

    public synchronized void signal (JxtaTimer timer) {

	if (rdvs == null) {
	    return;
	}

	RdV rdv = null;
	for (int i = 0; i < rdvs.size(); ++i) {
	    try {
		rdv = (RdV) rdvs.elementAt (i);
		rendezvous.reconnectRendezVous (rdv.peer);
	    } catch (Exception e) {
		continue;
	    }
	}
    }

    private void debug (String str) {
	if (debug) {
	    System.out.println ("[RdvMonitor] group= " + gId);
	    System.out.println ("            " + str);
	}
    }
}



