/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: AdvertisementCache.java,v 1.6 2001/05/23 01:00:21 lomax Exp $
 */


package net.jxta.impl.resolver;

import java.util.*;
import java.io.*;
import java.net.*;
import java.lang.reflect.InvocationTargetException;

import net.jxta.id.*;

import net.jxta.document.*;
import net.jxta.protocol.*;
import net.jxta.document.*;
import net.jxta.impl.protocol.*;

/**
 *
 * @version $Revision: 1.6 $
 * Tag  $Name:  $
 * @since 1.0
 */

public class AdvertisementCache  {

    private final static int INITIAL_ENTRY = 1024;
    private int maxSize = 5*1024;
    private int entry = 0;
    private Hashtable cache = new Hashtable(INITIAL_ENTRY);

 /**
  *
  * AdvertisementCache is used by resolver/discovery to store
  * Advertisements of all types
  * @version $Revision: 1.6 $
  * Tag  $Name:  $
  * @since 1.0
  */

    public AdvertisementCache() {

    }

 /**
  *  Stores the Advertisement in Cache and keys on the hashcode of the
  *  Advertisement's ID
  *  @param PeerAdvertisement adv the advertisement to store
  *
  */

    private synchronized void storeAdvertisement(PeerGroupAdvertisement adv)
    throws IOException, MalformedURLException, InvocationTargetException
    {

        URL url=null;
        ID id=null;
        try {
            url =new URL(adv.getPid());
            id = IDFactory.fromURL(url);
        } catch (MalformedURLException e) {
            throw new MalformedURLException(e.toString() );
        } catch (IOException e) {
            throw new IOException(e.toString() );
        } catch( InvocationTargetException e ) {
            throw e;
        } catch( Exception e ) {
            throw new InvocationTargetException( e );
        }
        if (entry <= maxSize) {
            if (cache.containsKey(id.toString()) ) {
                cache.put((id).toString(),(Object) adv);
                entry++;
            }
        } else {
            throw new IOException( "Reached maxSize "+maxSize );
        }
    }

  /**
   * remove an entry from cache, this useful in invalidating an
   * advertisement
   * @param key
   */

    public synchronized void remove (String key){
        cache.remove(key);
    }

  /**
   * Returns all the keys in the cache
   * @return Enumeration of keys
   */

    public synchronized Enumeration iterate (){
        return (cache.keys());
    }

   /**
    * given the doc ID get returns the document associated with it
    * @param id
    * @return
    */
    public ID get(String id){
        return( (ID) cache.get(id));
    }

  /**
   * flush cache.
   *
   */

    public synchronized void flush () {
        cache.clear();
    }

   /**
    * Checkpoint the cache to disk
    * This method should be used at the very most once, and at shutdown
    * until it is improved it should be utilized differently
    * @param filename
    * @exception IOException
    */

    public synchronized void checkPoint (String filename) throws IOException {

        FileOutputStream out = new FileOutputStream(filename);
        ObjectOutputStream stream = new ObjectOutputStream(out);
        try {
            stream.writeObject(cache);
            stream.flush();
            out.close();
        } catch (IOException e) {
            throw e;
        }
    }

   /**
    * Initialize the cache from a checkpoint
    * this method should only be used during boot
    * @param filename
    * @throws ClassNotFoundException
    * @exception IOException
    */

    public synchronized void initialize(String filename) throws
    ClassNotFoundException,IOException {

        FileInputStream in = new FileInputStream(filename);
        ObjectInputStream stream = new ObjectInputStream(in);
        try {
            cache = (Hashtable)stream.readObject();
            in.close();
            cache.clear();
        } catch (IOException e) {
            throw e;
        } catch (ClassNotFoundException e){
            throw e;
        }
    }

    /**
     * Convenience method to aid in the functionality of reverse discovery
     * One example of reverse discovery is peer invitation where a node would
     * somehow get it's advertisement to another node, via e-mail etc.
     * the receiving end would read in the advertisement and (reverse) discovers the
     * inviting node.  The use of this class is not limited to reverse discovery.
     * @return PeerAdvertisement object which represent the data
     * @param filename
     * @exception IOException
     */

    public synchronized Advertisement
    readAdvertisement(String filename) throws IOException {

        FileInputStream in = new FileInputStream(filename);
        
        try {
        Advertisement adv = AdvertisementFactory.newAdvertisement( new MimeMediaType( "text/xml" ), in );
        return adv;
        }
        catch ( Exception everything ) {
            throw new IOException( everything.toString() );
        }
    }

}
