/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: RendAddrCompactor.java,v 1.25 2001/05/23 01:00:21 lomax Exp $
 */

package net.jxta.impl.resolver;

import java.util.*;
import java.io.IOException;

import net.jxta.resolver.Resolver;
import net.jxta.service.Service;
import net.jxta.protocol.ServiceAdvertisement;
import net.jxta.protocol.PeerGroupAdvertisement;
import net.jxta.endpoint.EndpointAddress;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.impl.endpoint.Address;
import net.jxta.impl.endpoint.EndpointMessanger;
import net.jxta.impl.peergroup.GenericPeerGroup;
import net.jxta.impl.peergroup.RefPeerGroup;
import net.jxta.rendezvous.*;
import net.jxta.impl.rendezvous.*;
import net.jxta.discovery.*;


//import net.jxta.impl.protocol.*;

public class RendAddrCompactor  implements Runnable {

    /**
     *  The task of this thread is to determine which of the
     *  rendezvous points is reachable, and which is not.
     *  the thread is started when a group is initiated.
     *  It starts with what's described in the peer/group
     *  advertisement and add entries one by one to the list
     *  reachable addresses.  The reason behind using Endpoint
     *  and not a direct socket is to ensure we able to discover
     *  routes if the address is not directly reachable
     */


    private PeerGroup myGroup;
    private Vector params;
    private Resolver resolver;
    private Discovery discovery;
    private boolean debug=false;
    private RendezVous rendezvous = null;
    private String pgName = null;
    private String localPeerId = null;

    private static final int MaxNbOfRdv = 3;
    private static final long DiscoNap = 1 * 60 * 1000; // 1 minute
    private static final long HibernateDelay = 30 * 60 * 1000; // 30 minutes

    public RendAddrCompactor(PeerGroup pg) {
            myGroup = pg;
            debug=((RefPeerGroup) myGroup).getDebug();
    }

    /**
     *  to be called during group initialization
     *  different groups can have different rendezvous points
     */
    public void init (){

       debug("Start compacting");
       Thread thread  = new Thread (new RendAddrCompactor(myGroup),
                  "ResolverService:RendezvousCompactor" );
       thread.start();
    }

    public void run() {

	// XXX  lomax@jxta.org
	// We need to wait a little bit until all the groups are initialized
	try {
	    Thread.sleep (4000);
	} catch (Exception e) {
	}
        PeerGroupAdvertisement adv =((RefPeerGroup) myGroup).getAdvertisement();
        ServiceAdvertisement serviceAdv = (ServiceAdvertisement)
                                 (myGroup.getResolver().getAdvertisement());

        params = serviceAdv.getParams();

        resolver = ((GenericPeerGroup) myGroup).getResolver();
	RefPeerGroup parent = ((GenericPeerGroup) myGroup).getParentGroup();

	if (parent != null) {
	    discovery = parent.getDiscovery();
	} else {
	    discovery = myGroup.getDiscovery();
	}

	localPeerId = myGroup.getPeerID().toString();

	if (myGroup.isRendezvous()) {
	    // This peer is a rendezvous. Enable it
	    rendezvous = myGroup.getRendezVous();
	    try {
		rendezvous.startRendezVous (new RdvManager (myGroup));
	    } catch (Exception e) {
         debug ("cannot start Rendezvous :" + e.getMessage());
	    }
	}

	pgName = adv.getName();
	
	// If the group is the PlatformPeerGroup, there is no need to connect
	// to a rendezvous.
	/***
	if (pgName.equals ("PlatformPeerGroup")) {
	    debug ("PlatformPeerGroup Done compacting");
	    return;
	}
	***/
	rendezvous = myGroup.getRendezVous();
	RdvMonitor monitor = new RdvMonitor (myGroup,
					     (RendezVousImpl) rendezvous);
	rendezvous.setMonitor (monitor);

	// If the group is either the Network peer group, get the endpoints,
	// because at this point, it is very likely that this peer cannot find routes,
	// since it is not connected to the Network just yet.

	if (pgName.equals ("NetPeerGroup") ||
	    pgName.equals ("PlatformPeerGroup")){

	    if (params != null) {
		for(int i=0; i < params.size(); i++) {
		    String str = (String)params.elementAt(i);
		    Address address = new Address(str);
		    try {
			rendezvous.connectRendezVous (address);
		    } catch (Exception e) {
			debug ("Cannot connect to Rendezvous at " + str );
		    }
		}
	    }
	    debug("Network peergroup Done compacting");
	    return;
	}
	
	// This is an ordinary group. Search for a few advertisement of this group
	// and find rendezvous.

	debug ("Regular group... Find rendezvous");

	connect (params, null);

	// Now, until we have enough rendezvous, keep updating the group advertisement,
	// looking for new rendezvous, and try to connect to them.
	Enumeration enum = null;
	Enumeration rdvsEnum = null;
	Vector rdvs = null;

	while (true) {

	    search (pgName);

	    rdvsEnum = rendezvous.listRendezVous();
	    if ((rdvsEnum == null) || (!rdvsEnum.hasMoreElements())) {
		rdvs  = null;
	    } else {
		rdvs = new Vector();
		while (rdvsEnum.hasMoreElements()) {
		    try {
			rdvs.addElement (rdvsEnum.nextElement());
		    } catch (Exception e) {
			break;
		    }
		}
	    }

	    if ((rdvs != null) && (rdvs.size() >= MaxNbOfRdv)) {
		// We got enough.. Just sleep for now
		try {
		    if (rdvs.size() == 0) {
			Thread.sleep (DiscoNap);
		    } else {
			Thread.sleep (HibernateDelay);
		    }
		} catch (Exception e) {
		    continue;
		}
		continue;
	    }

	    try {
		enum = discovery.getLocalAdvertisements(Discovery.GROUP,
							"Name",
							pgName);
	    } catch (Exception e) {
		debug ("  failed with " + e);
	    }

	    if ((enum != null) && (enum.hasMoreElements())) {
		PeerGroupAdvertisement rdvAdv = null;

		while (enum.hasMoreElements()) {
		    debug ("Found some local");
		    try {
			rdvAdv = (PeerGroupAdvertisement) enum.nextElement();
			Hashtable sAdvs = null;

			sAdvs = rdvAdv.getServiceAdvertisements ();
			if (sAdvs == null) {
			    // No service advertisement
			    continue;
			}
			// Get the resolver ServiceAdvertisement
			ServiceAdvertisement r = null;
			try {
			    r = (ServiceAdvertisement) sAdvs.get ("jxta.service.resolver");
			    if (r == null) {
				// No resolver service advertisement. Not much we can do.
				continue;
			    }
			} catch (Exception e) {
			    return;
			}

			params = r.getParams();

			if (params == null) {
			    continue;
			}
			connect (params, rdvs);

		    } catch (Exception e) {
			continue;
		    }
		}
	    }
	    try {
		Thread.sleep (DiscoNap);
	    } catch (Exception e) {
	    }
	}
    }


    private void connect (Vector params, Vector connected) {

	if ((params == null) || (params.size() == 0)) {
	    // Nothing to do
	    return;
	}

	boolean found = false;
	String peerId = null;
	String connectedPeer = null;
	for (int i = 0; i < params.size(); ++i) {
	    try {
		peerId = (String) params.elementAt (i);
		if (peerId.equals (localPeerId)) {
		    // this is the local peer. Discard
		    continue;
		}

		// Check if this peer is already connected to this
		// rendezvous
		found = false;
		if ((connected != null) && (connected.size() > 0)) {
		    for (int j = 0; j < connected.size(); ++j) {
			try {
			    connectedPeer = (String) connected.elementAt (i);
			    if (connectedPeer.equals(peerId)) {
				found = true;
				break;
			    }
			} catch (Exception e) {
			    continue;
			}
		    }
		}
		if (found) {
		    // Already connected to that one. Next.
		    continue;
		}
		// Connect to this new rendezvous
		try {
		    debug ("  Connecting...");
		    Address address = new Address(peerId);
		    rendezvous.connectRendezVous (address);
		} catch (Exception e) {
		    debug ("Cannot connect to Rendezvous at " + peerId);
		}
	    } catch (Exception ex) {
		debug ("  failed with " + ex);
		continue;
	    }
	}
    }

    private void search (String name) {

	debug ("searching remote");
	if (name == null) {
	    return;
	}
	// If we are here, we didn't yet had enough rendezvous. Try to find some more.
	discovery.getRemoteAdvertisements(null,
					  Discovery.GROUP,
					  "Name",
					  name,
					  MaxNbOfRdv);
    }
	    
    private void debug (String str) {
	if (debug) {
	    System.out.println ("[RendAddrCompactor] " + str);
	}
    }

}

