/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ResolverInterface.java,v 1.14 2001/06/01 01:45:45 raffi Exp $
 */




package net.jxta.impl.resolver;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.IOException;
import java.io.InputStream;

import net.jxta.service.Service;
import net.jxta.resolver.Resolver;
import net.jxta.resolver.QueryHandler;
import net.jxta.protocol.ResolverQueryMsg;
import net.jxta.protocol.ResolverResponseMsg;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.Document;
import net.jxta.document.MimeMediaType;
import net.jxta.endpoint.EndpointAddress;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.NoResponseException;
import net.jxta.exception.ServiceNotFoundException;

import net.jxta.impl.endpoint.Endpoint;
import net.jxta.impl.endpoint.EndpointMessanger;
import net.jxta.impl.endpoint.Address;
import net.jxta.impl.endpoint.EndpointHandler;
import net.jxta.impl.peergroup.GenericPeerGroup;
import net.jxta.impl.peergroup.RefPeerGroup;
import net.jxta.impl.protocol.PeerGroupAdv;
import net.jxta.impl.protocol.ResolverQuery;
import net.jxta.impl.protocol.ResolverResponse;

/**
 *
 * @version $Revision: 1.14 $
 * Tag  $Name:  $
 * @since JXTA 1.0
 */

public class ResolverInterface implements Resolver {

    ResolverService impl;
    protected Vector rendezvous=null;

    /**
     * returns an interface object that permits to use this
     * service without having access to the real object.
     *
     * Since THIS is already such an object, it returns itself.
     * FIXME: it is kind of absurd to have this method part of the
     * interface but we do not want to define two levels of Service interface
     * just for that.
     * @return Resolver An interface object that implements
     * this service and nothing more.
     */

    public Service getInterface()
    {
	return this;
    }

    /**
     * Returns the advertisment for that service.
     *
     * @return Advertisement the advertisement.
     *
     * @version $Revision: 1.14 $
     * @since JXTA 1.0 
     */
    public Advertisement getAdvertisement() {
        return impl.getAdvertisement();
    }

  /**
   * Only authorized constructor
   */
    public ResolverInterface (ResolverService theRealThing) {
	impl = theRealThing;
    rendezvous = impl.rendezvous;
    }

    /**
     * Registers the given ResolveHandler.
     *
     * @param name The name under which this handler is known.
     * @param handler The handler.
     * @return The previous handler registered under this name
     */
    public QueryHandler registerHandler( String name, QueryHandler handler ) {
        return impl.registerHandler( name, handler );
	
    }

  /**
   * gets the handler registered under the given name.
   * @param name Handler name
   * @return ResolveHandler
   */
    public QueryHandler getHandler(String name) {
        return impl.getHandler(name);
    }

  /**
   * try getting a response for out query
   * @param address
   * @param query
   */

    public void  sendQuery(String rdvPeer,
                 ResolverQueryMsg query)
    {
        impl.sendQuery(rdvPeer, query);
    }

  /**
   * send a response to a peer.
   * @param destPeer is the destination of the response
   * @param response is the response to be sent
   */

    public void  sendResponse (String destPeer,
			       ResolverResponseMsg response)
    {
        impl.sendResponse(destPeer, response);
    }


    /**
     *  Initialize the application
     *  FIXME: This is meaningless for the interface object;
     * it is there only to satisfy the requirements of the
     * interface that we implement. Ultimately, the API should define
     * two levels of interfaces: one for the real service implementation
     * and one for the interface object. Right now it feels a bit heavy
     * to so that since the only different between the two would be
     * init() and may-be getName().
     * @param group PeerGroup this application is started from
     * @since   JXTA 1.0
     */

    public void init(PeerGroup pg, Advertisement a) {
    }


    /**
     * This is here for temporary class hierarchy reasons.
     * it is ALWAYS ignored. By definition, the interface object
     * protects the real object's start/stop methods from being called
     *
     * @param arg A table of strings arguments.
     * @return int status indication.
     */
    public int startApp(String[] arg) {
        return 0;
    }

    /**
     * This is here for temporary class hierarchy reasons.
     * it is ALWAYS ignored. By definition, the interface object
     * protects the real object's start/stop methods from being called
     *
     * This request is currently ignored.
     */
    public void stopApp() {
    }
}
