/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ResolverService.java,v 1.9 2001/06/01 01:45:45 raffi Exp $
 */




package net.jxta.impl.resolver;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.IOException;
import java.io.InputStream;

import net.jxta.service.Service;
import net.jxta.resolver.Resolver;
import net.jxta.resolver.QueryHandler;
import net.jxta.protocol.ResolverQueryMsg;
import net.jxta.protocol.ResolverResponseMsg;
import net.jxta.protocol.ServiceAdvertisement;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.Document;
import net.jxta.document.MimeMediaType;
import net.jxta.endpoint.EndpointAddress;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.NoResponseException;
import net.jxta.exception.ResendQueryException;
import net.jxta.exception.DiscardQueryException;
import net.jxta.exception.ServiceNotFoundException;

import net.jxta.impl.endpoint.Endpoint;
import net.jxta.impl.endpoint.MessageImpl;
import net.jxta.impl.endpoint.EndpointMessanger;
import net.jxta.impl.endpoint.Address;
import net.jxta.impl.endpoint.EndpointHandler;
import net.jxta.impl.peergroup.GenericPeerGroup;
import net.jxta.impl.peergroup.RefPeerGroup;
import net.jxta.impl.protocol.PeerGroupAdv;
import net.jxta.impl.protocol.ResolverQuery;
import net.jxta.impl.protocol.ResolverResponse;


/**
 * ResolverService Service provides a generic mechanism for jxta Services
 * to send "Queries", and receive "Responses".  It removes the burden for
 * registered handlers in deal with :
 *
 *   - Setting message tags, to ensure uniqueness of tags and
 *     ensures that messages are sent to correct address, and group
 *   - Authentication, and Verification of credentials
 *   - drop rogue messages
 *
 * ResolverService Service does not proccess the queries, nor does it not compose
 * reponses. Handling of queries, and composition of responses are left up
 * to the registered handlers. Services that wish to handle queries,
 * and generate reponses must implement <Code>QueryHandler<Code>
 *
 * Message Format :
 *
 * A Query message
 *
 * <?xml version="1.0" standalone='yes'?>
 * <ResolverQuery>
 *   <handlername> name </handlername>
 *   <credentialServiecUri> uri </credentialServiecUri>
 *   <credentialToken> token </credentialToken>
 *   <queryid> id </queryid>
 *   <query> query </query>
 * </ResolverQuery>
 *
 * Note: queryid is unique to the originating node only, it can be utilized to
 * match queries to responses.
 *
 * A Response Message
 *
 * <?xml version="1.0" standalone='yes'?>
 * <ResolverResponse>
 *   <handlername> name </handlername>
 *   <credentialServiecUri> uri </credentialServiecUri>
 *   <credentialToken> token </credentialToken>
 *   <queryid> id </queryid>
 *   <response> response </response>
 * </ResolverResponse>
 *
 * Note: queryid is unique to the originating node only, it can be
 * utilized to match queries to responses
 *
 * @version $Revision: 1.9 $
 * Tag  $Name:  $
 * @see net.jxta.protocol.ResolverQueryMsg;
 * @see net.jxta.protocol.ResolverResponseMsg;
 * @since JXTA 1.0
 * @version $Revision: 1.9 $
 * Tag  $Name:  $
 * @since JXTA 1.0
 */

public class ResolverService implements Resolver, EndpointHandler {

    public  String outQueName="ORes";
    public  String  inQueName="IRes";
    private Endpoint endpoint;
    private Hashtable handlers = new Hashtable(50);
    private MessageImpl propagateMsg;
    private RefPeerGroup myGroup=null;
    protected Vector rendezvous= new Vector();
    private boolean debug=false;
    private String localPeerId = null;

    private ServiceAdvertisement advertisement = null;

  /**
   * Default constructor
   */
    public void ResolverService ( ) {

    }

    /**
     * Supply arguments and starts this service if it hadn't started by itself.
     *
     * Currently this service starts by itself and does not expect
     * arguments.
     *
     * @param arg A table of strings arguments.
     * @return int status indication.
     */
    public int startApp(String[] arg) {
        return 0;
    }

    /**
     * Ask this service to stop.
     *
     * This request is currently ignored.
     */
    public void stopApp() {
    }

    /**
     * Service objects are not manipulated directly to protect usage
     * of the service. A Service interface is returned to access the service
     * methods.
     *
     * @return Service public interface of the service
     *
     * @version $Revision: 1.9 $
     * @since JXTA 1.0
     */
    public Service getInterface() {
	return new ResolverInterface(this);
    }


    /**
     * Returns the advertisement for that service.
     *
     * @return Advertisement the advertisement.
     *
     * @version $Revision: 1.9 $
     * @since JXTA 1.0
     */
    public Advertisement getAdvertisement() {
        return advertisement;
    }

    /**
     * Registers the given ResolveHandler.
     *
     * @param name The name under which this handler is known.
     * @param handler The handler.
     * @return The previous handler registered under this name
     */
    public synchronized QueryHandler registerHandler( String name, QueryHandler handler ) {
        return (QueryHandler)handlers.put( name, handler );

    }

  /**
   * gets the handler registered under the given name.
   * @param name Handler name
   * @return ResolveHandler
   */
    public QueryHandler getHandler(String name) {
        return (QueryHandler) handlers.get(name);
    }


    void propagate (MessageImpl msg, String prunePeer) {

	try {
	    endpoint.propagate(myGroup,
			       msg,
			       advertisement.getName(),
			       outQueName,
			       7,
			       prunePeer);
	} catch (Exception e) {
	}
    }


    /**
     * try getting a response for out query
     * @param address
     * @param query
     * @throws RuntimeException
     */

    public void sendQuery(String rdvPeer,
				       ResolverQueryMsg query)
    {

        MessageImpl recMsg = null;
        ResolverResponse doc=null;
        debug("sending query");

        propagateMsg = new MessageImpl();

	if (rdvPeer == null){
	    try {
		debug("sending a propagate msg");
		propagateMsg.push(outQueName,(InputStream)((Document)
		   (query.getDocument(new MimeMediaType("text/xml")))).getStream());
		propagate (propagateMsg, null);
	    } catch (Exception e) {
                throw new RuntimeException ("Error during endpoint.propagate "
					    +e.toString());
        }
	} else {

	    //unicast instead
	    try {
		respond (rdvPeer,
			 advertisement.getName(),
			 outQueName,
			 outQueName,
			 ((Document)
         (query.getDocument(new MimeMediaType("text/xml")))).getStream());
	    } catch (Exception e) {
		throw new RuntimeException ("Error during endpoint.propagate "
					    +e.toString());
	    }
	}
    }


    /*
     * Returns the group to which this service is attached.
     * @return PeerGroup the group
     */
    public PeerGroup getGroup () {
        return((PeerGroup)myGroup);
    }


    private void processNoResponse (MessageImpl msg) {

	if (!myGroup.isRendezvous()) {
	    // We are not a Rendez vous peer. Just drop the message.
	    return;
	}
	// We are a Rendez vous. Re-propagate the message.
	// It is assumed that the EndpointRouter has loop detectection and TTL.
	// XXX It would be a very good optimization to issue a new request if this
	// peer is the first Rendezvous peer touched by the request, in order to
	// cache the result: it is very likely that other peers "closeby" will issue
	// the same request.
	// For the time being. Just repropagate.
	// lomax@jxta.org

	propagate (msg, localPeerId);
    }


    private void resendQuery (MessageImpl msg) {

	if (!myGroup.isRendezvous()) {
	    // We are not a Rendez vous peer. Just drop the message.
	    return;
	}


	// Rebuild a query based on the existing one
        Enumeration tags       = msg.getNames();
	String tagString = null;

	if (tags == null) {
	    debug ("   no tag");
	    return;
	}

	Vector tmp = new Vector();
	ResolverQuery query = null;

	// We need to change the source peer value of each query.
        while (tags.hasMoreElements()) {

	    tagString=(String)tags.nextElement();
	    if (tagString.indexOf ("Res") == -1) {
		// Tag for something else - just ignore
		continue;
	    }

	    try {
		query = new ResolverQuery(msg.pop(tagString));
		query.setSrc (localPeerId);
		tmp.addElement (query);
	    } catch (Exception e) {
		continue;
	    }
	}
	// At this point, all the queries from the message has been
	// removed, updated, and stored into tmp. Put them back into the
	// the message

	if (tmp.size() == 0) {
	    // There was no query. Discard.
	    return;
	}

	for (int i = 0; i < tmp.size(); ++i) {

	    try {
		query = (ResolverQuery) tmp.elementAt (i);
                msg.push(outQueName,(InputStream)((Document)
			(query.getDocument(new MimeMediaType("text/xml")))).getStream());
	    } catch (Exception e) {
		// Something weird happened. Discard.
		return;
	    }
	}

	// Send the message
	propagate (msg, localPeerId);
    }

    /**
     * This method will be called by the endpoint to handle messages
     * of this type
     * @param message
     */
    public void demux(MessageImpl message) {

        Enumeration tags;
        String tagString;
        Address srcAddress;

        debug("demuxing a query");
        srcAddress = (Address) message.getSrcAddress();
        tags       = message.getNames();
	if (tags == null) {
	    debug ("   no tag");
	    return;
	}

	MessageImpl srcMsg = (MessageImpl) message.dup();
	srcMsg.setDestAddress (message.getDestAddress());
	srcMsg.setSrcAddress (message.getSrcAddress());

	while (tags.hasMoreElements()) {

	    tagString=(String)tags.nextElement();
	    if (tagString.indexOf ("Res") == -1) {
		// Tag for something else - just ignore
		debug ("   Ignore tag " + tagString);
		continue;
	    }

	    try {
		ResolverQuery doc =
		    new ResolverQuery(message.pop(tagString));

		ResolverResponseMsg res =
		    (ResolverResponseMsg) processQuery((ResolverQueryMsg)doc);

		if (res == null) {
		    // There is no response. Nothing to do.
		    return;
		}

		// There is a response to be sent
		srcAddress.setServiceName (advertisement.getName());
		srcAddress.setServiceParameter (inQueName);

		respond (doc.getSrc(),
			 advertisement.getName(),
			 inQueName,
			 tagString,
			 (InputStream)((Document)
	      (res.getDocument(new MimeMediaType("text/xml")))).getStream());


	    } catch (ResendQueryException e2) {
		// The resolver service has no response, but is interested to get
		// a response itself. Resend the query, but just like if that
		// was the source of the query. The resolver service will be
		// reponsible for resending the response to the original sender.
		debug ("   resendQuery");
		resendQuery (srcMsg);
	    } catch (NoResponseException e1) {
		// The resolver service has no response. Process the query to
		// see if there is something else to do.
		debug ("  processNoResponse " + e1);
		processNoResponse (srcMsg);

	    } catch (DiscardQueryException e3) {
		// We are asked to discard this query.
		debug ("   query is discarded upon service resolver decision");
		return;
	    } catch (Exception e4) {
		// The query was invalid somehow. Just discard.
		if (debug) e4.printStackTrace();
		return;
	    }
	}
    }



    /**
     * Inner class to handle the Response queue for outgoing messages
     *
     */

    class RecvDemux implements EndpointHandler {

        public void demux(MessageImpl message){

        if (message == null ) {
		 debug ("RecvDemux.demux: got a null message");
         return;
	    }
        debug("demuxing a response ");
        Enumeration tags = message.getNames();
        if (tags==null) return;
        while (tags.hasMoreElements()) {
         String tagString=(String) tags.nextElement();
         debug ("   tag= " + tagString);
         if (tagString.indexOf ("Res") == -1) {
         // Tag for something else - just ignore
         continue;
        }
        try {
            while (true)
            pushResponse(new
                ResolverResponse(message.pop(tagString)));
        } catch (Exception e){
		    // we'll definitely get this exception when have no more
		    // messages, and this is the desired behavior.
		}
            }
        }
    }


    /**
     * Process the message and respond
     * @return ResolverResponseMsg response
     * @param query
     * @throws NoResponseException
     */

    public ResolverResponseMsg processQuery(ResolverQueryMsg query)
        throws NoResponseException,
               ResendQueryException,
               DiscardQueryException,
               IOException {

        debug("handing a query to  "+((ResolverQuery)query).getHandlerName());

        QueryHandler theHandler = getHandler( ((ResolverQuery)query).getHandlerName());

        if ( theHandler == null) {
            debug("throwing away "+((ResolverQuery)query).getHandlerName());
            return null;
        }
        debug("handing a query to  "+((ResolverQuery)query).getHandlerName());
        ResolverResponseMsg result;

        result = theHandler.processQuery(query);

        return((ResolverResponseMsg) result);
    }


    /**
     * push Response back to the handler
     * @return ResolverResponseMsg response
     * @param query
     * @throws NoResponseException
     */

    public void pushResponse(ResolverResponseMsg resp) {
        debug("pushing a response to  "+((ResolverResponseMsg)resp).getHandlerName());
        QueryHandler theHandler =
           getHandler(((ResolverResponseMsg)resp).getHandlerName());
        theHandler.pushResponse(resp);
    }


    /**
     * Send a response to a peer.
     * This is used by resolver services that are handling query/response in
     * an asynchronous way, which is the case of Rendezvous peers.
     *
     * @param destPeer is the destination of the response
     * @param response is the response to be sent
     */
    public void sendResponse (String destPeer, ResolverResponseMsg response) {

    if (destPeer == null) {
        propagateResponse(response);
    } else {
	try {
	    respond (destPeer,
		     advertisement.getName(),
		     inQueName,
		     inQueName,
		     (InputStream)((Document)
		         (response.getDocument(new MimeMediaType("text/xml")))).
		                          getStream());
	} catch (Exception e) {
	}
    }
    }

    private void propagateResponse(ResolverResponseMsg response) {

        propagateMsg = new MessageImpl();
        try {
	    debug("sending a propagate response");
	    propagateMsg.push(inQueName,(InputStream)((Document)
  	           (response.getDocument(new MimeMediaType("text/xml")))).getStream());

	    endpoint.propagate(myGroup,
			       propagateMsg,
			       advertisement.getName(),
			       inQueName,
			       1,
			       null);
        } catch (Exception e) {
	    if (debug) e.printStackTrace();
        }

    }

    /**
     * respond with a ResolverResponse Message
     * @param address
     * @param tagName
     * @param response
     * @throws RuntimeException
     */
    private void respond (String destPeer,
			 String pName,
			 String pParam,
			 String tagName,
			 InputStream response)
	         throws RuntimeException {




        debug("destPeer :\n"+destPeer);
        debug("responding to "+pName+"  "+pParam+" "+tagName);
        if (response == null) {
          throw new RuntimeException(
              "Attempting to respond with a empty message");
        }

        MessageImpl msg = (MessageImpl) endpoint.getMessage();
        try {
            msg.push(tagName, response);
        } catch (Exception e) {
	    throw new RuntimeException(
                 "msg.push has thrown a "+e.toString());
        }

        EndpointMessanger messenger = null;
        try {
            messenger = endpoint.getMessanger (destPeer,
					       pName,
					       pParam,
					       Endpoint.NonBlocking);

	    if (messenger == null) return;

        } catch (IOException e) {
             throw new RuntimeException(
                 "endpoint.getMessenger has thrown a "+e.toString());
        }

        try {
            messenger.sendMessage (msg);
        } catch (IOException e) {
             throw new RuntimeException(
                  "messenger has thrown a "+e.toString());
        }
    }

    /**
     * init is called by the group running this instance
     * @param g
     * @throws PeerGroupException
     */

    public void init(PeerGroup g, Advertisement sadv)
        throws PeerGroupException
    {

        advertisement     = (ServiceAdvertisement) sadv;
        RefPeerGroup pg   = (RefPeerGroup) g;
        myGroup           = pg;
        debug             = pg.getDebug();
        String uniqueStr  = (pg.getID().getGroupUniqueId()).toString();
        outQueName        = uniqueStr + outQueName;
        inQueName         = uniqueStr + inQueName;
        endpoint          = pg.getEndpoint();
        RecvDemux recvMux = new RecvDemux();
        localPeerId       = myGroup.getPeerID().toString();

        try {
            endpoint.registerHandler(advertisement.getName()
                                     + outQueName,this);
            endpoint.registerHandler(advertisement.getName()
                                     + inQueName,recvMux);
        } catch (Exception e){
	    if (debug) e.printStackTrace();
        }
        propagateMsg= new MessageImpl();
        // run Rendezvous point address compactor thread

        RendAddrCompactor compactor = new RendAddrCompactor(g);
        compactor.init();
    }

    private void debug (String str) {
	if (debug) {
	    System.out.println ("[ResolverService] " + str);
	}
    }

}
