#ifndef SERVENT_H
#define SERVENT_H

#include <qsocket.h>
#include <qthread.h>

#include <vector>
#include <string>
#include <list>
#include <memory>

class Pong;
class Ping;
class Query;
class Host;
class QueryHit;
class Push;
class Address;

class QListViewItem;

//! Function code for PING messages.
#define CMD_PING     0x0
//! Function code for PONG messages.
#define CMD_PONG     0x1
//! Function code for QUERY messages.
#define CMD_QUERY    0x80
//! Function code for QUERY_HIT messages.
#define CMD_QUERYHIT 0x81
//! Function code for PUSH messages.
#define CMD_PUSH     0x40

//! This class is used to communicate to other Gnutella host.
/*! 
    It connects to only one host at once. When data arrives it
    is automatically appended it to the ping, pong, query,
    queryhit and push vectors. These vectors are being precessed by
    the Connection class every 200ms. Since Qtella is mutlithreaded
    we have to call this->mutex.lock() every time we access the
    vectors to prevent conflicts.
    \sa Connection vPing vPong vQuery vQueryHit vPush
*/

class Servent : public QSocket
{
  Q_OBJECT

 public:
  //! Creates an object with the given Gnutella ID.
  /*!
      \param id The Gnutella ID used for this host.
  */
  Servent(string id, QListViewItem *item);

  //! Destructor.
  ~Servent();

  //! Establish a connection to another host at the given address.
  /*!
      This method returns immediately. 
      \param address Address of the other host.
      \sa Address
   */
  void    connectToServent(Address address);

  
  void    closeConnection();

  //! Returns the client's state.
  /*!
      \return serventIdle, serventConnected, serventError serventClosed
  */
  int     serventState();

  //! Sends a PING message to the other host with own Gnutella ID.
  /*!
      \param ttl TTL value.
      \param hops HOPS value.
  */
  void    sendPing(int ttl, int hops);

  //! Sends a PING message to the other host.
  /*!
      \param ttl TTL value.
      \param hops HOPS value.
      \param id Gnutella ID.
  */
  void    sendPing(int ttl, int hops, string id);

  //! Sends a PONG message to the other host.
  /*!
      \param ttl TTL value.
      \param hops HOPS value.
      \param id Gnutella ID.
      \param address Address of host where the PONG message comes from.
      \param files Number of shared files.
      \param kbytes Shared file size in KBytes.
  */
  void    sendPong(int ttl, int hops, string id, Address address, long files, long kbytes);

  //! Sends a QUERY message to the other host.
  /*!
      \param ttl TTL value.
      \param hops HOPS value.
      \param id Gnutella ID.
      \param min_speed Minimum speed.
      \param search Search string.
  */
  void    sendQuery(int ttl, int hops, string id, int min_speed, string search);

  //! Sends QUERY_HIT message to the other host.
  /*!
      \param ttl TTL value.
      \param hops HOPS value.
      \param queryhit The QueryHit object. TTL and HOPS values of this object are ignored.
  */
  void    sendQueryHit(int ttl, int hops, QueryHit queryhit);

  void    sendPush(int ttl, int hops, Push push);
  
  void    writeConnectString();

  void    limitVectors();

  void    sendData(string buffer);


  //! States for connection to other host.
  enum { serventIdle,       /*!< Not connected.                    */
	 serventConnected,  /*!< Connected to other Gnutella host. */
	 serventError,      /*!< An error occurred.                */
	 serventConnecting, /*!< Connecting...                     */
	 serventWaiting,    /*!< Socket is open.                   */
	 serventClosed      /*!< Connection closed                 */
  };

  QListViewItem* _item;
  int            _timeout;
  QMutex         _list_mutex;

 private:

  //! Creates a header.
  /*!
      \param id Gnutella ID.
      \param func Function.
      \param ttl TTL value.
      \param hops HOPS value.
      \param payload Payload length.
  */
  string  createHeader(string id, int func, int ttl, int hops, long payload);

  //! Extracts the data from a header.
  /*!
      \param header Header.
      \return id Gnutella ID of the header.
      \return func Function.
      \return ttl TTL value.
      \return hops HOPS value.
      \return payload Payload length.
  */
  void    checkHeader(string header, string& id, int& func, int& ttl, int& hops, long& payload);


 public slots:
    void    slotDelayedClosed();

    //! Called if an error occured.
    /*!
        State is set to serventError.
    */
    void    slotError(int);

    //! Called if connection has been established.
    /*! 
        State is set to serventConnected.
    */
    void    slotConnect();

    //! Called if new data has been arrived.
    /*!
     */
    void    slotReadyRead();

    //! Called if other client has closed the connection.
    /*!
        State is set to serventIdle.
    */
    void    slotClosed();


 private:

    void    setState(int state);

    //! Holds the current state of the connection.
    int     _state;

    //! Gnutella connect string.
    /*!
        For protocol version 0.4 it is set to "GNUTELLA CONNECT/0.4\n\n".
    */
    string  _connect_str;

    //! Gnutella answer string.
    /*!
        For protocol version 0.4 it is set to "GNUTELLA OK\n\n".
    */
    string  _response_str;

    //! Unique Gnutella ID of this host.
    string  _id;

    //! Buffer which takes all incoming data.
    string  _buffer;

    //! Used for connectToServent(...) to have a valid object.
    string  _strip;

    QMutex  _state_mutex;

 public:
    //! Pong vector for all received PONG messages.
    list<Pong*>     pong_list;

    //! Ping vector for all received PING messages.
    vector<Ping*>     vPing;

    //! Query vector for all received QUERY messages.
    vector<Query*>    vQuery;

    //! QueryHit vector for all received QUERY_HIT messages.
    vector<QueryHit*> vQueryHit;
    
    //! Host vector for active hosts.
    list<Host*>       host_list;

    //! Push vector for all received PUSH messages.
    vector<Push*>     vPush;
    
    //! Vector which stores the IDs of hosts where PING messages came from.
    list  <string>          pingid_list;

    //! Vector which stores the IDs of hosts where QUERY messages came from.
    list<string>            queryid_list;

    //! Number of input messages
    long    _input;
    
    //! Number of output messages
    long    _output;
};

#endif
