/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: HttpResolverClient.java,v 1.1.1.1 2001/06/01 16:49:49 jbeatty Exp $
 */
package net.jxta.search.peer;

import java.io.*;
import java.net.*;
import java.util.*;

import net.jxta.search.util.LogService;
import net.jxta.search.util.UuidService;
import net.jxta.search.util.ValidatingXmlParser;
import net.jxta.search.util.XmlParser;


public class HttpResolverClient implements ResolverClient {

    URL             url;
    LogService      log;
    
    
    public HttpResolverClient (URL             url, 
			       LogService      log) {
	this.url     = url;
	this.log     = log;
    }


    public ProviderInfo[] resolve (Request request)
	throws IOException, XmlParser.Exception
    {
	HttpURLConnection con = makeResolveCall (request);
	
	if (con.getResponseCode () != HttpURLConnection.HTTP_OK)
	    throw new IOException ("Resolver status is " +  
				   con.getResponseCode ());
	

	return parseResolverResponse (con.getInputStream ());
    }

    private HttpURLConnection makeResolveCall (Request request)
	throws IOException
    {
	HttpURLConnection con;
	OutputStream      out;
	
		
	con = (HttpURLConnection) url.openConnection ();
	
	con.setRequestMethod ("POST");
	
	con.setDoInput (true);
	
	con.setDoOutput (true);
	
	con.setRequestProperty ("Content-type", "text/xml");
	
	con.setRequestProperty ("Content-length", 
				String.valueOf (request.getLength ())); 
	
	con.connect ();

	out = con.getOutputStream ();

	request.toStream (out);
	
	out.flush ();
	
	
	return con;
    }

    static ProviderInfo[] parseResolverResponse (InputStream in)
    	throws IOException, XmlParser.Exception
    {
	ResolverResponseHandler handler = new ResolverResponseHandler ();
	byte[] buf = new byte[8192];
	

	ValidatingXmlParser.parse (in, buf, handler, handler.root);
	
	return handler.getProviders ();
    }


    public void deleteRoutes (URL[]    routes, 
			      String[] ids, 
			      String[] reasons)
    {
	HttpURLConnection con = null;
	PrintWriter       writer;
	
	try {
	    con = (HttpURLConnection) url.openConnection ();
	    con.setRequestMethod ("POST");
	    con.setDoOutput (true);
	    con.setDoInput (false);
	    
	    con.setRequestProperty ("Content-type", "text/xml");
	    
	    writer = new PrintWriter(new OutputStreamWriter
		(con.getOutputStream()));
	    writer.println ("<?xml version='1.0'?>");
	    writer.println ("<delete-routes>");
	    for (int i = 0;  i < routes.length;  i++) {
		writer.println ("  <provider>");
		writer.println ("    <url>" + routes[i] + "</url>");
		writer.println ("    <external-id>" + ids[i] + "</external-id>");
		writer.println ("    <reason>" + reasons[i] + "</reason>");
		writer.println ("  </provider>");
	    }
	    writer.println ("</delete-routes>");
	    
	    con.disconnect ();
	} catch (IOException e) {
	    if (con == null)
		if (log == null)
		    e.printStackTrace ();
		else
		    log.log (log.ERROR, "-", "resolver is down: " + 
			     e.getMessage ());
	    else 
		try {
		    if (log == null)
			System.out.println ("Resolver status = " +
					    con.getResponseCode ());
		    else
			log.log (log.ERROR, "-", "resolver status = " + 
				 con.getResponseCode ());
		} catch (IOException ex) {
		    if (log == null)
			ex.printStackTrace ();
		    else
			log.log (log.ERROR, "-", ex.getMessage ());
		}
	    
	}
    }


    /**************************************************
     * Test code
     **************************************************/

    static String usage =
	"java net.jxta.search.router.HttpResolverClient\n" +
	"        resolver              (url)\n" +
	"        query-space           (url)\n" +
	"        query-lifetime        (milliseconds)\n" +
	"        max-fanout            (int count)\n" +
	"        max-hits-per-provider (int count)\n" +
	"        flush-after-providers (int count of providers)\n" +
	"        flush-after-ms        (flush after n milliseconds)\n" +
	"        query                 (free-form)\n";



    public static void main (String[] argv) 
	throws IOException, XmlParser.Exception
    {
	
	if (argv.length == 3) 
	    test (argv);
	else
	    System.out.println (usage);
    }


    
    
    public static void test (String[] argv) 
	throws IOException, XmlParser.Exception
    {
	URL resolverURL = new URL (argv[0]);
	UuidService uuidService = new UuidService ();
	HttpResolverClient  client = new HttpResolverClient (resolverURL, 
							     null);
	Request req = 
	    new Request (argv[1],
			 uuidService.hexify (uuidService.getUuid ()),
			 Integer.parseInt (argv[2]),
			 Integer.parseInt (argv[3]),
			 Integer.parseInt (argv[4]),
			 Integer.parseInt (argv[5]),
			 Integer.parseInt (argv[6]),
			 ("<query>\n"+argv[7]+"</query>\n").getBytes ());
		
	ProviderInfo[] providerInfos = client.resolve (req);
	
	for (int i = 0;  i < providerInfos.length;  i++) {
	    ProviderInfo providerInfo = providerInfos[i];
	    
	    System.out.println (" -> " + providerInfo.getQueryURL () + " " + 
				providerInfo.getId ());
	}
    	
    }
}
