/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: JxtaInfoProviderService.java,v 1.2 2001/06/04 04:30:02 jbeatty Exp $
 */
package net.jxta.search.peer;

import net.jxta.platform.*;
import net.jxta.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.endpoint.Message;
import net.jxta.endpoint.*;
import net.jxta.protocol.*;
import net.jxta.exception.*;
import net.jxta.id.*;
import net.jxta.discovery.*;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.MimeMediaType;

import net.jxta.search.util.JxtaPipeUtil;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Vector;

/** Service class that provides everything you need to become a JXTA
    information provider.  See net.jxta.search.examples.jxtaprovider
    for an example of how to use this class. **/
public class JxtaInfoProviderService implements Runnable {

    /** the name of the pipe advertisement file **/
    private static final String PIPE_ADV_FILENAME = "pipeadv.xml";    
    /** the name of the registration file **/
    private static final String REGISTRATION_FILENAME = "registration.xml";    
    /** the name of the query pipe that the peer exposes **/
    private static final String PIPE_NAME = "Infrasearch Query Pipe";    
    private static final String REG_PIPE_NAME = "Jxtasearch Registration Pipe";
    private static final String REQUEST = "request";
    private static final String RESPONSE = "response";
    private static final String REGISTRATION_REQUEST = "registration-request";
    private static final String REGISTRATION_RESPONSE="registration-response";
    /** we will try to resolve the return pipe for 10 seconds
        before we give up **/    
    private static final int RETURN_PIPE_RESOLVE_TIMEOUT = 10000;
    private static MimeMediaType MIME_TYPE = new MimeMediaType ("text", "xml");


    private boolean isRunning = true;
    private Pipe pipeService;
    private InputPipe inputPipe;
    private PipeAdvertisement pipeadv;
    private PeerGroup peerGroup;    
    private DocumentBuilder builder;
    private RequestHandler requestHandler;

    /** A vector of hub registration pipe advs that I'm registered with **/
    private Vector hubRegPipeAdvs = new Vector();

    public JxtaInfoProviderService( PeerGroup peerGroup,
				    DocumentBuilder builder,
				    RequestHandler requestHandler )
	throws JxtaException, IOException 
    { 

	this.peerGroup = peerGroup;
	this.builder = builder;
	this.requestHandler = requestHandler;

	pipeService = (Pipe) peerGroup.getPipe();	
	// create the adv for the input pipe
	pipeadv = JxtaPipeUtil.createPipeAdv( new File( PIPE_ADV_FILENAME ),
					      PIPE_NAME,
					      peerGroup );
	if ( pipeadv == null ) throw new JxtaException( "No Pipe Adv" );
	inputPipe = pipeService.createInputPipe( pipeadv );
	Discovery disco = (Discovery) peerGroup.getDiscovery();
	disco.publish( pipeadv, Discovery.ADV );
    }

 
    /** the main loop of the info provider.  This simple waits for a message
	and processes it **/
    public void run() {
	try {
	    createRegistration();
	} catch( Exception e ) {
	    System.out.println( "Problem while registering with hubs: " + e );
	    e.printStackTrace();
	}
	
	byte[] buf = new byte[128];
	
	while ( isRunning ) {
	    try {
		Message msg = inputPipe.poll( 1000 );
		if( msg==null ) {
		    if( isRunning )
			continue;
		    else
			break;
		}
		
		// extract response pipe and make output pipe
		InputStream is	       = msg.pop ("responsePipe");
		OutputPipe  outPipe    = createOutputPipe (is);
		Document    requestDoc = builder.parse (msg.pop
							("request"));
		String	    type       = getMessageType (requestDoc);
		Message	    response   = null;
		
		
		if (type.equals (REQUEST)) {		    
		    response = createQueryResponseMessage (requestDoc);
		} else if (type.equals (REGISTRATION_REQUEST)) {
		    response = createRegistrationMessage();
		}
		
		if (response != null) {
		    outPipe.send (response);
		} 		
	    } catch ( SAXException e ) {
		System.err.println( "Problem parsing request message.");
		e.printStackTrace ();
	    } catch ( IOException e ) {
		System.err.println( "Problem reading input pipe while "+
				    "processing request: " + e );
		e.printStackTrace();
	    } catch ( JxtaException e ) {
		System.err.println( "JxtaException processing request:" + e );
		e.printStackTrace();
	    } catch( Exception e ) {
		System.out.println( "Exception while processing request: " + e );
		e.printStackTrace();
	    }
	}
    }

    public void stop() {
	System.out.println( "JxtaInfoProviderService stopping..." );
	isRunning = false;
	try {
	    Object condition = new Object();
	    unregisterWithHubs( condition );
	    synchronized( condition ) {
		try {
		    condition.wait( 1000 * 60 * 2 );
		} catch( InterruptedException e ) {
		    System.out.println( "Unexpected error while stopping" );
		}
	    }
	} catch (Exception e) {
	    System.out.println( "Error while unregistering with hubs." );
	}
    }
    
    Message makeRegistrationMessage() throws IOException {

	Message msg = peerGroup.getPipe().createMessage();

	msg.push( "peerid", new ByteArrayInputStream( peerGroup.getPeerID().toString().getBytes()) );

	msg.push( "requestcode", 
		  new ByteArrayInputStream( "register".getBytes() ) );

	msg.push( "registration", new FileInputStream( new File( REGISTRATION_FILENAME )));
	
	return msg;
    }

    Message makeUnregistrationMessage() throws IOException {

	Message msg = peerGroup.getPipe().createMessage();

	msg.push( "peerid", 
		  new ByteArrayInputStream( peerGroup.getPeerID().toString().getBytes()) );

	msg.push( "requestcode", 
		  new ByteArrayInputStream( "unregister".getBytes() ) );

	return msg;
    }

    /** Registers with any jxtasearch hubs we can discover **/
    public void registerWithHubs() {
	new Thread( new RegistrationWorker() ).start();
    }

    /** Same as unregister with hubs, except that condition is notified
        when the unregistration is complete. nice to coding a graceful
        shutdown.
    **/
    public void unregisterWithHubs( Object condition ) {
	new Thread( new UnregistrationWorker( condition ) ).start();
    }

    /** unregisters with any hubs that we are registered with. **/
    public void unregisterWithHubs() {
	unregisterWithHubs( null );
    }


    /** Writes a new registration if one isn't already present. This has
	the peer's pipe id in it, so it's pretty handy. **/
    void createRegistration() {

	try {
	    File f = new File( REGISTRATION_FILENAME );

	    if( f.exists() )
		return;
	    
	    if( !f.createNewFile() ) {
		System.out.println( "Registration file does not exist and cannot "+
				    "create a new one" );
	    }
	    
	    StringBuffer b = new StringBuffer();
	    b.append( "<?xml version='1.0'?>\n" );
	    b.append( "<register>\n<title>JXTA Test Provider</title><link>http://search.jxta.org</link><description>JXTA Test Provider</description><image><url>http://www.jxta.org/branding/images/pj.gif</url><width>126</width><height>64</height></image>\n<query-server>" );
	    b.append( pipeadv.getPipeID().toString() );
	    b.append( "</query-server>\n<query-space uri='http://search.jxta.org/text'>\n<predicate><query><text></text></query></predicate>\n</query-space>\n</register>" );
	    
	    byte[] arr = b.toString().getBytes();
	    
	    FileOutputStream out = new FileOutputStream( f );
	    out.write( arr );
	    out.close();	
	} catch( IOException e ) {
	    e.printStackTrace();
	}
    }

    private String getMessageType( Document doc ) {
	Element el = doc.getDocumentElement ();
	return el.getTagName ();
    }
    

    class Query {
	
	String uuid;
	String text;
	
	Query (String uuid, String text) {
	    this.uuid = uuid;
	    this.text = text;
	}
    }
    
    private Query extractQuery (Document reqDoc) throws SAXException {
	
	Element  reqEl    = reqDoc.getDocumentElement ();	
	String   uuid     = reqEl.getAttribute ("query-uuid");	
	NodeList children = reqEl.getChildNodes ();	
	String   text     = null;

	for (int i = 0;  i < children.getLength (); i++) {
	    Node n = children.item (i);
	    
	    if (n.getNodeName ().equals ("query")) {
		
		NodeList queryChildren = n.getChildNodes ();
		
		for (int j = 0; j < children.getLength (); j++) {
		    Node m = queryChildren.item (j);

		    if (m.getNodeName ().equals ("text")) {
			text = m.getFirstChild().getNodeValue ();
			break;
		    }
		}
	    }
	    
	    if( text!= null )
		break;
	}
	
	if (uuid == null)
	    throw new SAXException ("No query-uuid attribute in request tag.");
	
	if (text == null)
	    throw new SAXException ("No text tag child in query tag.");
	
	return new Query (uuid, text);
    }
    
    /** Takes a query and returns a jxta Message response **/
    private Message createQueryResponseMessage ( Document doc)
	throws IOException, SAXException 
    {
	Query query = extractQuery (doc);	
	System.out.println( "received query: " + query.text );
	String response = requestHandler.onRequest( query.text, query.uuid );

	Message responsemsg = null;
	if( response!=null ) {
	    InputStream in = new ByteArrayInputStream( response.getBytes ());
	    responsemsg = pipeService.createMessage();
	    if( responsemsg!=null )
		responsemsg.push( "responses", in );
	    else 
		System.out.println( "Pipe service unable to create message" );
	}
	
	return responsemsg;
    }
    
    /** creates a Message containing a registraion **/
    private Message createRegistrationMessage()
	throws IOException, JxtaException {
	System.out.println( "received registration request");
	
	Message responsemsg = null;
	try {
	    File f = new File( REGISTRATION_FILENAME );
	    FileInputStream in = new FileInputStream( f );
	    responsemsg = pipeService.createMessage();
	    responsemsg.push( REGISTRATION_RESPONSE, in );
	} catch( IOException e ) {
	    throw new JxtaException( "Exception pushing registration into message: " + e );
	}
	
	return responsemsg;
    }
    
    
    /** Given an inputstream with a pipe advertisement, returns an OutputPipe
	that can talk to it **/
    private OutputPipe createOutputPipe( InputStream pipeAdvStream )
	throws JxtaException, IOException {
	try {
	    System.out.println( "creating output pipe" );
	    PipeAdvertisement pipeadv = (PipeAdvertisement)
		AdvertisementFactory.newAdvertisement( MIME_TYPE, pipeAdvStream );
	    pipeAdvStream.close();
	    OutputPipe outpipe = pipeService.createOutputPipe (pipeadv,
							       Pipe.NonBlocking,
							       RETURN_PIPE_RESOLVE_TIMEOUT  );
	    
	    if( outpipe == null ) {
		System.out.println( "Pipe resolution timeout" );
		throw new JxtaException( "Could not resolve output pipe" );
	    }
	    
	    return outpipe;
	    
	} catch( InvocationTargetException e ) {
	    System.out.println( "cannot create advertisement: " + e );
	    e.printStackTrace();
	    throw new JxtaException( "Cannot create output pipe");
	}
    }

    class RegistrationWorker implements Runnable {
	public void run() {
	    int success=0;
	    synchronized( hubRegPipeAdvs ) {
		hubRegPipeAdvs = new Vector();
		
		System.out.println( "Registering with all Jxtasearch hubs I can "+
				    "find");
		
		hubRegPipeAdvs = new Vector();
		
		Enumeration enum=null;
		try {
		    enum = JxtaDiscoveryUtil.discoverPipeAdvs( peerGroup,
							       REG_PIPE_NAME,
							       5000 );
		} catch( PeerGroupException e ) {
		    System.out.println( "Error while discovering hubs: " +
					e.getMessage() );
		}
		
		while( enum!=null && enum.hasMoreElements() ) {
		    try {
			PipeAdvertisement adv = (PipeAdvertisement) enum.nextElement();
			OutputPipe outPipe = 
			    peerGroup.getPipe().createOutputPipe (adv,
								  Pipe.NonBlocking,
								  5000 );
			if( outPipe != null ) {
			    try {
				Message msg = makeRegistrationMessage();
				System.out.println( " - Sending registration to a hub" );
				outPipe.send( msg );
				hubRegPipeAdvs.add( adv );
				success++;
			    } catch( IOException e ) {
				System.out.println( " - Problem creating or sending message to hub");
			    }
			}
		    } catch( Exception e ) {
			System.out.println( " - Problem resolving pipe to hub" );
		    }
		}
	    }
	    
	    System.out.println( "Successfully registered with " + success +
				" hub(s)" );
	}
    }

    class UnregistrationWorker implements Runnable {
	private Object condition;
	UnregistrationWorker( Object condition ) {
	    this.condition = condition;
	}

	public void run() {
	    int success=0;
	    synchronized( hubRegPipeAdvs ) {
		System.out.println( "Unregistering with " + 
				    hubRegPipeAdvs.size() + " hubs" );
		Enumeration enum = hubRegPipeAdvs.elements();
		while( enum!=null && enum.hasMoreElements() ) {
		    try {
			PipeAdvertisement adv = 
			    (PipeAdvertisement) enum.nextElement();
			
			OutputPipe outPipe = 
			    peerGroup.getPipe().createOutputPipe (adv,
								  Pipe.NonBlocking,
								  5000 );
			if( outPipe != null ) {
			    try {
				Message msg = makeUnregistrationMessage();
				System.out.println( " - Sending unregistration message to hub" );
				outPipe.send( msg );
				hubRegPipeAdvs.remove( adv );
				success++;
			    } catch( IOException e ) {
				System.out.println( " - Problem sending unregistration message to hub");
			    }
			}
		    } catch( Exception e ) {
			System.out.println( " - Problem resolving pipe to hub" );
		    }
		}
	    }
	    System.out.println( "Successfully unregistered with " + success +
				" hub(s)" );
	    if( condition!=null ) {
		synchronized( condition ) {
		    condition.notifyAll();
		}
	    }
	}
    }
}
