/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: JxtaRouterClient.java,v 1.1.1.1 2001/06/01 16:49:53 jbeatty Exp $
 */
package net.jxta.search.peer;

import net.jxta.platform.*;
import net.jxta.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.endpoint.Message;
import net.jxta.endpoint.*;
import net.jxta.protocol.*;
import net.jxta.exception.*;
import net.jxta.document.*;
import net.jxta.id.*;
import net.jxta.discovery.*;

import org.w3c.dom.Document;
import javax.xml.parsers.DocumentBuilder;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.util.Enumeration;


//debug only
import java.io.InputStream;
import java.io.OutputStream;

/** A Jxta implementation of a RouterClient. Instances are reentrant. **/
public class JxtaRouterClient /* implements RouterClient */ {

    /** map of uuid(String)->callback(QueryResponseCallback) **/
    ResponseCallbackMap callbackMap = new ResponseCallbackMap();

    /** The pipe to the router **/
    OutputPipe requestPipe;

    /** the peer group that this router client belongs in **/
    PeerGroup group;

    /** the pipe advertisement of the router that this client will talk to **/
    PipeAdvertisement routerPipeAdv;

    /** The pipe advertisement for the pipe we expected the responses to
	come in on **/
    PipeAdvertisement responsePipeAdv;

    /** Handle to a JAXP docbuilder **/
    DocumentBuilder builder;

    /** we will try to resolve the return pipe for 10 seconds
        before we give up **/    
    static final int PIPE_RESOLVE_TIMEOUT = 10000;

    static final String QUERY_PIPE_NAME = "Infrasearch Query Pipe";

    static final MimeMediaType MIME_TYPE = new MimeMediaType("text/xml");

    /** Standard constructor for creating JxtaRouterClient instances. Instances
	are reentrant. **/
    public JxtaRouterClient( PeerGroup group,
			     PipeAdvertisement routerPipeAdv,
			     DocumentBuilder builder ) throws IOException {
	this.group = group;
	this.builder = builder;

	try {
	    setRouterPipeAdv( routerPipeAdv );
	} catch( IOException e ) {
	    System.out.println( "Cannot resolve request pipe." );
	}

	// create the response pipe and start the response processor
	//
	try {
	    responsePipeAdv = (PipeAdvertisement)
		AdvertisementFactory.newAdvertisement
		(PipeAdvertisement.getAdvertisementType ());
	} catch( java.lang.reflect.InvocationTargetException e ) {
	    e.printStackTrace();
	    throw new RuntimeException( "Fatal exception creating pipe " +
					"advertisement: " + e );
	}

        responsePipeAdv.setPipeID ( new PipeID( group.getID() ) );
        responsePipeAdv.setName ("Jxtasearch Result Pipe");

	InputPipe responsePipe  =
	    group.getPipe().createInputPipe(responsePipeAdv);
	
	new Thread( new JxtaQueryResponseProcessor( responsePipe,
						    callbackMap,
						    builder )).start();
    }

    /** Returns an enumeration of pipe advertisement for advertised jxtasearch
	routers **/
    public static Enumeration discoverRouterPipeAdvs( PeerGroup group )
    throws PeerGroupException {
	return JxtaDiscoveryUtil.discoverPipeAdvs( group,
						   QUERY_PIPE_NAME,
						   5000 );
    }

    /** Allows the user to change the router pipe advertisement **/
    public void setRouterPipeAdv( PipeAdvertisement pipeAdv )
	throws IOException 
    {
	this.routerPipeAdv = pipeAdv;

	if( pipeAdv != null ) {
	    try {
		// hack. jxta pipe resolution often doesn't work the first time,
		// but will work thereafter. This is to compensate until that
		// is fixed.
		
		requestPipe =
		    group.getPipe().createOutputPipe (routerPipeAdv,
						      Pipe.NonBlocking,
						      PIPE_RESOLVE_TIMEOUT );
	    } catch( IOException e ) {
		// hack. jxta pipe resolution often doesn't work the 1st time,
		// but will work thereafter. This is to compensate until that
		// is fixed.
		requestPipe =
		    group.getPipe().createOutputPipe (routerPipeAdv,
						      Pipe.NonBlocking,
						      PIPE_RESOLVE_TIMEOUT );
	    }
	}
	System.out.println( "JxtaRouterClient: requestPipe="+requestPipe);
   }

    /** Issues an asynchronous search requests and returns the UUID for the
	request.  callback will be called when the response has arrived **/
    public String search( Request request, QueryResponseCallback callback )
	throws IOException 
    {
	
	if( requestPipe == null ) {
	    throw new IOException( "Was not able to resolve output pipe" );
	}

	// construct the jxta message
	//
	Message msg = group.getPipe().createMessage();

	// XXX debug code
	System.out.println( "request" );
	transfer( request.getInputStream(), System.out );
	// XXX end debug code

	msg.push( "request", request.getInputStream() );

	try {
	    msg.push( "responsePipe", responsePipeAdv.getDocument( MIME_TYPE ).getStream() );
	} catch( Exception e ) {
	    // jxta's exception handling is horrible. this isn't much better.
	    throw new IOException( "Exception retrieving document from response pipe advertisement: " + e );
	}

	System.out.println( "JxtaRounterClient: inserting callback: " +
			    callback + " uuid=" + request.queryUuid );
	callbackMap.insertRequest( request.queryUuid, callback );

	try {
	    requestPipe.send( msg );
	} catch( IOException e ) {
	    // in the case that the send is unsuccesful, we need to remove
	    // the entry from the map. Note that we insert the entry into 
	    // the map prior to send() due to possible race conditions (tho
	    // the likelihood of that is near 0).
	    callback.onException( request.queryUuid, e );
	    callbackMap.removeCallback( request.queryUuid );
	    throw e;
	}

    return request.queryUuid;    
    }

    /** Issues a synchronous search request and returns the Document with the
	response. This is a convenience method implemented on top of the 
        other search method **/
    public Document search( Request request ) throws IOException {

	Object condition = new Object();
	NotificationCallback callback = new NotificationCallback( condition );
	search( request, callback );

	try {
	    // wait for the response if its not already arrived
	    if( callback.getResults() == null ) {
		synchronized( condition ) {
		    condition.wait();
		}
	    }
	} catch( InterruptedException e ) {
	    System.out.println( "Interruption while waiting for response.");
	    return null;
	}

	return callback.getResults();
    }

    /** Simple callback implementation that notifies a condition variable
	and holds the results document.  This is useful for implementing
	a sychronous jxta querying mechanism **/
    private static class NotificationCallback implements QueryResponseCallback{
	Object condition;
	Document results;
	NotificationCallback( Object condition ) {
	    this.condition = condition;
	}

	public void onQueryResults( String uuid, Document results ) {
	    this.results = results;
	    synchronized( condition ) {
		condition.notify();
	    }
	}

	public void onException( String uuid, Exception e ) {
	    this.results = null;
	    synchronized( condition ) {
		condition.notify();
	    }
	}

	public Document getResults() {
	    return results;
	}
    }

    /** moves all bytes from is to os **/
    private void transfer( InputStream is, OutputStream os ) {
        int c;
        try {
            while ( (c=is.read())!= -1 ) {
                os.write( c );
            }
        } catch ( IOException e ) {
            System.err.println( "Problem displaying search payload:" + e );
            e.printStackTrace();
        }
    }
}
