/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Index.java,v 1.1.1.1 2001/06/01 16:49:57 jbeatty Exp $
 */

package net.jxta.search.resolver;

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import java.util.*;
import java.io.*;

/** An index of provider predicates, used to accelerate query matching.
 * <p>A separate index is used for each query-space.
 */
class Index {
  private String querySpace;
  private ArrayList predicateIdToPredicateInfo = new ArrayList();
  private ArrayList providerIdToProvider = new ArrayList();
  private ArrayList providerIdToExternalId = new ArrayList();
  private Hashtable externalIdToProviderId = new Hashtable();
  private ProviderSimilarity providerSimilarity;
  private BitSet providerStatus = new BitSet(); //false=provider offline,true=available
  // store the range of predicate ids for each provider in a table
  // Each provider has a sequential and unique set of predicateIds
  private Hashtable providerExternalIdToPredicateIdRange; 
   


  Index(String querySpace)
    {
	this.querySpace = querySpace;
	providerSimilarity = new ProviderSimilarity(this);
	providerExternalIdToPredicateIdRange = new Hashtable();
    }

  class PredicateInfo {
    int clauseCount; 
    int providerId;
    int predicateNumber;			  // within provider, within queryspace

    PredicateInfo(int clauseCount, int providerId, int predicateNumber) {
      this.clauseCount = clauseCount;
      this.providerId = providerId;
      this.predicateNumber = predicateNumber;
    }
  }

  private SortedMap patternToPostings = new TreeMap();
  
  class Posting implements Comparable{
    Posting (int predicateId, int clauseId) {
      this.predicateId = predicateId;
      this.clauseId = clauseId;
      this.score = 1; //scores are initially set to 1
    }
    
    int predicateId;
    int clauseId;
    float score;
    
    public int compareTo(Object object) {
	Posting other = (Posting) object;
	  if (this.predicateId == other.predicateId)
	      return this.clauseId - other.clauseId;
	  else return (this.predicateId - other.predicateId);
      }
	      
    public boolean equals(Object object) {
      Posting other = (Posting) object;
      return ((this.predicateId == other.predicateId)
	&& (this.clauseId == other.clauseId));
    }
      public float getScore() {
	  return score;
      }

      // method to update the posting score
      // (0 <= alpha <= 1)
      // although it is possible to zero a score (alpha=0,input=-score) it is
      // not recommended to do so. It is better to remove the posting from the
      // index (index.removePosting(pattern, Posting)

      private void updateScore(float alpha, float input)
      {
	  score = alpha * score + (1f-alpha) * input;
      }
      
      public String toString()
      {
	  StringBuffer sb = new StringBuffer();
	  sb.append("PredicateId: "+predicateId+"\tclauseId: "+clauseId+"\tscore: "+score);
	  return sb.toString();
      }
    
  }

  

    public String getQuerySpace() {
	return querySpace; 
    }

  public void remove (String externalId) {
      Integer oldId = (Integer) externalIdToProviderId.get(externalId);
      if(oldId != null) {
	  int providerId = oldId.intValue();
	Provider provider = (Provider) providerIdToProvider.get(providerId);
	// perform the reverse steps for adding a provider to the index 

	ArrayList predicates = provider.getPredicates(querySpace);
	for (int predicateNumber = 0; predicateNumber < predicates.size();
	     predicateNumber++) {
	    Predicate predicate = (Predicate)predicates.get(predicateNumber);
	    ArrayList conjunction = predicate.conjunction;
	    int clauseCount = conjunction.size();
	    
	    for (int clauseId = 0; clauseId < clauseCount; clauseId++) {

		Disjunction disjunction = (Disjunction)conjunction.get(clauseId);
		String[] patterns = disjunction.patterns;
		
		for (int k = 0; k < patterns.length; k++) {
		    String pattern = patterns[k];
		    removePattern(externalId, pattern);
		}
	    }
	}
	// remove the added patterns 
	TreeSet ts = provider.getAddedPatterns(querySpace);
	if(ts != null)
	    {
		Iterator it = ts.iterator();
		while(it.hasNext())
		    {
			removePattern(externalId, (String) it.next());
		    }
	    }
      }
  }
  
    public void setProviderStatus(String externalId, boolean online)
    {
	Integer provId = (Integer) externalIdToProviderId.get(externalId);
	if(provId != null)
	    { if(online) providerStatus.set(provId.intValue());
	    else providerStatus.clear(provId.intValue()); }

    }

    public boolean getProviderStatus(int providerId)
    {
	return providerStatus.get(providerId);
    }

    public int[] getMostPopularProviders(int maxProviders)
    {
	return providerSimilarity.getMostPopular(maxProviders);
    }
    
    public int[] getSimilarProviders(int providerId, int maxSimilars) {
	return providerSimilarity.getSimilarProviders(providerId, maxSimilars);
    }

    public void updateSimilarity(int[] similarProviders) {
	providerSimilarity.updateSimilarity(similarProviders);
    }
    
  /** Adds a provider's predicates to this index. */
  public void add(String externalId, Provider provider) {

   int providerId;
   boolean providerExists = false;
   Integer oldId = (Integer) externalIdToProviderId.get(externalId);
   if(oldId != null)     // use the same oldId, but replace the data
       {
	   providerExists = true;
	   remove(externalId);
	   providerId = oldId.intValue();
	   providerIdToProvider.set(providerId, provider);
	   providerIdToExternalId.set(providerId, externalId);
       }
   else { // use a new providerId
       providerId = providerIdToProvider.size();
       providerIdToProvider.add(provider);
       providerIdToExternalId.add(externalId);
       providerSimilarity.addProvider();
   }
   // if provider exists, set older predicateInfos to null preserve space
   if(providerExists) {
       int[] oldRange = (int[])providerExternalIdToPredicateIdRange.get(externalId);
       for(int i = oldRange[0]; i <= oldRange[1]; i++)
	   {
	       PredicateInfo info = (PredicateInfo) predicateIdToPredicateInfo.get(i);
	       info = null;
	   }
       oldRange = null;
   }

   // Store the new predicates
   ArrayList predicates = provider.getPredicates(querySpace);
   int[] range = new int[2];
   range[0] = predicateIdToPredicateInfo.size();
    for (int predicateNumber = 0; predicateNumber < predicates.size();
	 predicateNumber++) {
      Predicate predicate = (Predicate)predicates.get(predicateNumber);
      ArrayList conjunction = predicate.conjunction;
      int clauseCount = conjunction.size();

      PredicateInfo info =
	new PredicateInfo(clauseCount, providerId, predicateNumber);
      int predicateId = predicateIdToPredicateInfo.size();
      predicateIdToPredicateInfo.add(info);

      for (int clauseId = 0; clauseId < clauseCount; clauseId++) {

	Disjunction disjunction = (Disjunction)conjunction.get(clauseId);
	String[] patterns = disjunction.patterns;

	for (int k = 0; k < patterns.length; k++) {
	  String pattern = patterns[k];

	  TreeSet postings = (TreeSet)patternToPostings.get(pattern);
	  if (postings == null)
	    patternToPostings.put(pattern, postings = new TreeSet());

	  postings.add(new Posting(predicateId, clauseId));
	}
      }
    }
    // Add a general predicate where we can add new patterns that are not in the
    // registration. The convention is that the last predicateId is not in the 
    // registration and is added by the index object.
    // It has one clause only
    PredicateInfo info =
	new PredicateInfo(1, providerId, predicates.size());
    int predicateId = predicateIdToPredicateInfo.size();
    predicateIdToPredicateInfo.add(info);
    range[1] = predicateId;
    externalIdToProviderId.put(externalId, new Integer(providerId));
    providerExternalIdToPredicateIdRange.put(externalId,range);
    providerStatus.set(providerId);
  }

  public void close() {}

  /// core search methods, used to compute matches

    /** One greater than the largest predicate id. */
  public int getMaxPredicateId() throws IOException {
    return predicateIdToPredicateInfo.size();
  }

  /** One greater than the largest provider id. */
  public int getMaxProviderId() throws IOException {
    return providerIdToProvider.size();
  }


  ArrayList getPostings(String pattern) {
      TreeSet ts = (TreeSet)patternToPostings.get(pattern);
      if(ts != null)
	  return new ArrayList(ts);
      else return null;
  }


    /** Check if the pattern exists in the index */
    public boolean patternExists(String pattern)
    {
	return (patternToPostings.get(pattern) != null) ;
    }
  
    /** returns the Postings of a particular provider for a particular pattern 
     * Could be more than one posting */
    
    private SortedSet getPostings(String externalId, String pattern)
    {
	int[] range = (int [])providerExternalIdToPredicateIdRange.get(externalId);
	if(range != null)
	    {
		int largestClauseId = 1000000;
		Posting smallestPosting = new Posting(range[0], 0);
		Posting largestPosting = new Posting(range[1], 100000);
		TreeSet ts = (TreeSet)patternToPostings.get(pattern);
		if(ts != null)
		    return ts.subSet(smallestPosting, largestPosting);
		else return new TreeSet();
	    }
	else return new TreeSet();
    }

  int getProviderId(int predicateId) {
    PredicateInfo info =
      (PredicateInfo)predicateIdToPredicateInfo.get(predicateId);
    return info.providerId;
  }

  int getClauseCount(int predicateId) {
	  PredicateInfo info =
	      (PredicateInfo)predicateIdToPredicateInfo.get(predicateId);
	  return info.clauseCount;
  }

  // methods used to display matches

  public String getExternalId(int providerId) {
    return (String)providerIdToExternalId.get(providerId);
  }

  /** Returns the named provider. */
  public Provider getProvider(int providerId) {
    return (Provider)providerIdToProvider.get(providerId);
  }

  /** Returns the named predicate. */
  public Predicate getPredicate(int predicateId) {
    PredicateInfo info =
      (PredicateInfo)predicateIdToPredicateInfo.get(predicateId);
    ArrayList predicates = (ArrayList) getProvider(info.providerId).getPredicates(querySpace);
    return (Predicate)predicates.get(info.predicateNumber);
  }
  
    public boolean changeScore(String externalId, String pattern, float alpha, float input) {
	SortedSet ss = getPostings(externalId, pattern);
	Iterator it = ss.iterator();
	boolean scoreChanged = false;
	while(it.hasNext())
	    {
		Posting p = (Posting) it.next();
		p.updateScore(alpha, input);
		scoreChanged = true;
	    }
	return scoreChanged;
    }

    /** remove all the occurrences of a pattern registered by a provider */
    public void removePattern(String externalId, String pattern)
    {
	SortedSet ss = getPostings(externalId, pattern);
	Iterator it = ss.iterator();
	while(it.hasNext())
	    {
		Posting p = (Posting) it.next();
		removePosting(pattern,p);
	    }
    }

    /** add a pattern to the provider. This pattern is not in the registration. *
     *  It is added to a general predicate */
    
    public void addPattern(String externalId, String pattern)
    {
	SortedSet ss = getPostings(externalId, pattern);
	if(ss.size() > 0) { return; } // pattern is already there, do nothing
	else {
	    int[] range = getPredicateIdRange(externalId);
	    if(range == null) { return; } // externalId does not exist, do not add
	    else {
		int maxPredicateId = range[1];
		Posting p = new Posting(maxPredicateId, 0);
		updatePosting(pattern, p);
	    }
	}
	Integer id = (Integer) externalIdToProviderId.get(externalId);
	if(id != null) {
	    int providerId = id.intValue();
	    Provider provider = (Provider) providerIdToProvider.get(providerId);
	    provider.addToGeneralPredicates(querySpace, pattern);
	}
    }
    
    /** updatePosting updates a specific posting of a pattern   *
     *  could also be used to create a new pattern/posting pair *
     *  if it does not exist */
    private void updatePosting(String pattern, Posting posting)
    {
	TreeSet postings = (TreeSet) patternToPostings.get(pattern);
	if(postings == null) postings = new TreeSet();
	postings.remove(posting);
	postings.add(posting);
	patternToPostings.put(pattern, postings);
    }

    private void removePosting(String pattern, Posting posting)
    {
	TreeSet postings = (TreeSet) patternToPostings.get(pattern);
	if(postings == null) return;
	postings.remove(posting);
	if(postings.size() == 0)  // remove the pattern from the index
	    {
		patternToPostings.remove(pattern);
	    }
    }
    
    private int[] getPredicateIdRange(String externalId)
    {
	int[] range = (int[])providerExternalIdToPredicateIdRange.get(externalId);
	return range;
    }

    
  /** For debugging.  Displays the contents of this index. */
  public String toString() {
    StringBuffer buffer = new StringBuffer();
    Iterator i = patternToPostings.entrySet().iterator();
    while (i.hasNext()) {
      Map.Entry entry = (Map.Entry)i.next();
      buffer.append(entry.getKey());
      buffer.append(": \n");
      ArrayList value = new ArrayList((TreeSet)entry.getValue());
      for (int j = 0; j < value.size(); j++) {
	Posting posting = (Posting)value.get(j);
	buffer.append(posting.toString());
	buffer.append("\tclauseCount: ");
	buffer.append(getClauseCount(posting.predicateId));
	buffer.append("\n");
	buffer.append(getProvider
		      (getProviderId(posting.predicateId)).queryServer);
	buffer.append("\n");
      }
    }
    return buffer.toString();
  }


  private static void testIndex() {

    String testfilestring = "./regs/CNN.com.xml";
    testIndex(testfilestring);

  } /* close testIndex() */



  private static void testIndex(String provider) {

      Index index=null;
    try {
	Provider p = new Provider(new File(provider));
	String qs = (String) ((TreeSet) p.getQuerySpaces()).first(); 
	index = new Index(qs);
	index.add(provider, p);
       
       System.out.println("Index as read from registration:\n");
       System.out.println(index+"\n");
       index.addPattern(provider, "NewWord1");  // added new word to provider
       index.addPattern(provider, "NewWord2");  // added new word to provider
       index.addPattern(provider, "NewWord3"); 
       index.changeScore(provider, "NewWord1", 0.9f, -1f);  // reduce score;
       index.changeScore(provider, "NewWord2",.8f, 2f);  // increase score
       index.removePattern(provider, "NewWord3"); // remove pattern from provider
       
       System.out.println("Index after adding changes:");

    } catch (Exception e) {
       System.out.println(e);
    }

    System.out.println(index);

  } /* close testIndex(String) */

  

  public static void main(String[] args) throws Exception {

    testIndex(); 

  }

} 
