/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: IndexWriter.java,v 1.1.1.1 2001/06/01 16:49:59 jbeatty Exp $
 */

package net.jxta.search.resolver;

import net.jxta.search.resolver.io.*;
import net.jxta.search.util.*;
import java.io.*;
import java.util.*;

/**
  An IndexWriter creates and maintains an index.

  The third argument to the <a href="#IndexWriter"><b>constructor</b></a>
  determines whether a new index is created, or whether an existing index is
  opened for the addition of new providers.

  In either case, providers are added with the <a
  href="#add"><b>add</b></a> method.  When finished adding
  providers, <a href="#close"><b>close</b></a> should be called.

  If an index will not have more providers added for a while and optimal search
  performance is desired, then the <a href="#optimize"><b>optimize</b></a>
  method should be called before the index is closed.
  */

final class IndexWriter {
  private Directory directory;			  // where this index resides

  private SegmentInfos segmentInfos = new SegmentInfos(); // the segments
  private final Directory ramDirectory = new RAMDirectory(); // for temp segs

  /** Constructs an IndexWriter for the index in <code>path</code>.  If
    <code>create</code> is true, then a new, empty index will be created in
    <code>d</code>, replacing the index already there, if any. */
  public IndexWriter(String path, boolean create) throws IOException {
    this(new FSDirectory(path, create), create);
  }

  /** Constructs an IndexWriter for the index in <code>path</code>.  If
    <code>create</code> is true, then a new, empty index will be created in
    <code>d</code>, replacing the index already there, if any. */
  public IndexWriter(File path, boolean create) throws IOException {
    this(new FSDirectory(path, create), create);
  }

  /** Constructs an IndexWriter for the index in <code>d</code>.  If
    <code>create</code> is true, then a new, empty index will be created in
    <code>d</code>, replacing the index already there, if any. */
  public IndexWriter(Directory d, boolean create) throws IOException {
    directory = d;

    if (create)
      segmentInfos.write(directory);
    else
      segmentInfos.read(directory);

  }

  /** Flushes all changes to an index, closes all associated files, and closes
    the directory that the index is stored in. */
  public final synchronized void close() throws IOException {
    flushRamSegments();
    ramDirectory.close();
    directory.close();
  }

  /** Returns the number of providers currently in this index. */
  public final synchronized int providerCount() {
    int count = 0;
    for (int i = 0; i < segmentInfos.size(); i++) {
      SegmentInfo si = segmentInfos.info(i);
      count += si.providerCount;
    }
    return count;
  }

  /** Adds a provider to this index.*/
  public final void add(Provider provider) throws IOException {
    SegmentCreator creator = new SegmentCreator(ramDirectory);
    String segmentName = newSegmentName();
    creator.createSegment(provider, segmentName);
    synchronized (this) {
      segmentInfos.addElement(new SegmentInfo(segmentName, 1, ramDirectory));
      maybeMergeSegments();
    }
  }

  private final synchronized String newSegmentName() {
    return Integer.toString(segmentInfos.counter++, Character.MAX_RADIX);
  }

  /** Determines how often segment indexes are merged by add().  With
   * smaller values, less RAM is used while indexing, and searches on
   * unoptimized indexes are faster, but indexing speed is slower.  With larger
   * values more RAM is used while indexing and searches on unoptimized indexes
   * are slower, but indexing is faster.  Thus larger values (> 10) are best
   * for batched index creation, and smaller values (< 10) for indexes that are
   * interactively maintained.
   *
   * <p>This must never be less than 2.  The default value is 10.*/
  public int mergeFactor = 10;

  /** Determines the largest number of providers ever merged by add().
   * Small values (e.g., less than 10,000) are best for interactive indexing,
   * as this limits the length of pauses while indexing to a few seconds.
   * Larger values are best for batched indexing and speedier searches.
   *
   * <p>The default value is {@link Integer#MAX_VALUE}. */
  public int maxMergeProviders = Integer.MAX_VALUE;

  /** If non-null, information about merges will be printed to this. */
  public PrintStream infoStream = null;

  /** Merges all segments together into a single segment, optimizing an index
      for search. */
  public final synchronized void optimize() throws IOException {
    flushRamSegments();
    while (segmentInfos.size() > 1) {
      int minSegment = segmentInfos.size() - mergeFactor;
      mergeSegments(minSegment < 0 ? 0 : minSegment);
    }
  }

  /** Merges all segments from an array of indexes into this index.
   *
   * <p>This may be used to parallelize batch indexing.  A large provider
   * collection can be broken into sub-collections.  Each sub-collection can be
   * indexed in parallel, on a different thread, process or machine.  The
   * complete index can then be created by merging sub-collection indexes
   * with this method.
   *
   * <p>After this completes, the index is optimized. */
  public final synchronized void addIndexes(Directory[] dirs)
      throws IOException {
    optimize();					  // start with zero or 1 seg
    int minSegment = segmentInfos.size();
    int segmentsAddedSinceMerge = 0;
    for (int i = 0; i < dirs.length; i++) {
      SegmentInfos sis = new SegmentInfos();	  // read infos from dir
      sis.read(dirs[i]);
      for (int j = 0; j < sis.size(); j++) {
	segmentInfos.addElement(sis.info(j));	  // add each info

	// merge whenever mergeFactor segments have been added
	if (++segmentsAddedSinceMerge == mergeFactor) {
	  mergeSegments(minSegment++, false);
	  segmentsAddedSinceMerge = 0;
	}
      }
    }
    optimize();					  // final cleanup
  }

  /** Merges all RAM-resident segments. */
  private final void flushRamSegments() throws IOException {
    int minSegment = segmentInfos.size()-1;
    int providerCount = 0;
    while (minSegment >= 0 &&
	   (segmentInfos.info(minSegment)).dir == ramDirectory) {
      providerCount += segmentInfos.info(minSegment).providerCount;
      minSegment--;
    }
    if (minSegment < 0 ||			  // add one FS segment?
	(providerCount + segmentInfos.info(minSegment).providerCount) > mergeFactor ||
	!(segmentInfos.info(segmentInfos.size()-1).dir == ramDirectory))
      minSegment++;
    if (minSegment >= segmentInfos.size())
      return;					  // none to merge
    mergeSegments(minSegment);
  }

  /** Incremental segment merger.  */
  private final void maybeMergeSegments() throws IOException {
    long targetMergeProviders = mergeFactor;
    while (targetMergeProviders <= maxMergeProviders) {
      // find segments smaller than current target size
      int minSegment = segmentInfos.size();
      int mergeProviders = 0;
      while (--minSegment >= 0) {
	SegmentInfo si = segmentInfos.info(minSegment);
	if (si.providerCount >= targetMergeProviders)
	  break;
	mergeProviders += si.providerCount;
      }

      if (mergeProviders >= targetMergeProviders)		  // found a merge to do
	mergeSegments(minSegment+1);
      else
	break;
      
      targetMergeProviders *= mergeFactor;		  // increase target size
    }
  }

  /** Pops segments off of segmentInfos stack down to minSegment, merges them,
    and pushes the merged index onto the top of the segmentInfos stack. */
  private final void mergeSegments(int minSegment) throws IOException {
    mergeSegments(minSegment, true);
  }

  /** Pops segments off of segmentInfos stack down to minSegment, merges them,
    and pushes the merged index onto the top of the segmentInfos stack. */
  private final void mergeSegments(int minSegment, boolean delete)
      throws IOException {
    String mergedName = newSegmentName();
    int mergedProviderCount = 0;
    if (infoStream != null) infoStream.print("merging segments");
    SegmentMerger merger = new SegmentMerger(directory, mergedName);
    Vector segmentsToDelete = new Vector();
    for (int i = minSegment; i < segmentInfos.size(); i++) {
      SegmentInfo si = segmentInfos.info(i);
      if (infoStream != null)
	infoStream.print(" "+si.name+" ("+si.providerCount+" providers)");
      SegmentReader reader = new SegmentReader(si.dir, si.name);
      merger.add(reader);
      if (delete)
	segmentsToDelete.addElement(reader);	  // queue for deletion
      mergedProviderCount += si.providerCount;
    }
    if (infoStream != null) {
      infoStream.println();
      infoStream.println(" into " + mergedName
			 + " (" + mergedProviderCount + " providers)");
    }
    merger.merge();

    segmentInfos.setSize(minSegment);		  // pop old infos & add new
    segmentInfos.addElement(new SegmentInfo(mergedName, mergedProviderCount,
					    directory));
    
    segmentInfos.write(directory);		  // commit before deleting

    deleteSegments(segmentsToDelete);		  // delete now-unused segments
  }

  /* Some operating systems (e.g. Windows) don't permit a file to be deleted
     while it is opened for read (e.g. by another process or thread).  So we
     assume that when a delete fails it is because the file is open in another
     process, and queue the file for subsequent deletion. */

  private final void deleteSegments(Vector segments) throws IOException {
    Vector deletable = new Vector();

    deleteFiles(readDeleteableFiles(), deletable); // try to delete deleteable
    
    for (int i = 0; i < segments.size(); i++) {
      SegmentReader reader = (SegmentReader)segments.elementAt(i);
      if (reader.directory == this.directory)
	deleteFiles(reader.files(), deletable);	  // try to delete our files
      else
	deleteFiles(reader.files(), reader.directory); // delete, eg, RAM files
    }

    writeDeleteableFiles(deletable);		  // note files we can't delete
  }

  private final void deleteFiles(Vector files, Directory directory)
       throws IOException {
    for (int i = 0; i < files.size(); i++)
      directory.deleteFile((String)files.elementAt(i));
  }

  private final void deleteFiles(Vector files, Vector deletable)
       throws IOException {
    for (int i = 0; i < files.size(); i++) {
      String file = (String)files.elementAt(i);
      try {
	directory.deleteFile(file);		  // try to delete each file
      } catch (IOException e) {			  // if delete fails
	if (directory.fileExists(file)) {
	  if (infoStream != null)
	    infoStream.println(e.getMessage() + "; Will re-try later.");
	  deletable.addElement(file);		  // add to deletable
	}
      }
    }
  }

  private final Vector readDeleteableFiles() throws IOException {
    Vector result = new Vector();
    if (!directory.fileExists("deletable"))
      return result;

    Input input = directory.openFile("deletable");
    try {
      for (int i = input.readInt(); i > 0; i--)	  // read file names
	result.addElement(input.readString());
    } finally {
      input.close();
    }
    return result;
  }

  private final void writeDeleteableFiles(Vector files) throws IOException {
    Output output = directory.createFile("deleteable.new");
    try {
      output.writeInt(files.size());
      for (int i = 0; i < files.size(); i++)
	output.writeString((String)files.elementAt(i));
    } finally {
      output.close();
    }
    directory.renameFile("deleteable.new", "deletable");
  }

    final class SegmentInfo {
    public String name;				  // unique name in dir
    public int providerCount;			  // number of providers in seg
    public Directory dir;			  // where segment resides

    public SegmentInfo(String name, int providerCount, Directory dir) {
      this.name = name;
      this.providerCount = providerCount;
      this.dir = dir;
    }
  }


  final class SegmentInfos extends Vector {
    public int counter = 0;			  // used to name new segments
  
    public final SegmentInfo info(int i) {
      return (SegmentInfo)elementAt(i);
    }

    public final void read(Directory directory) throws IOException {
      Input input = directory.openFile("segments");
      try {
	counter = input.readInt();		  // read counter
	for (int i = input.readInt(); i > 0; i--) { // read segmentInfos
	  SegmentInfo si = new SegmentInfo(input.readString(), input.readInt(),
					   directory);
	  addElement(si);
	}
      } finally {
	input.close();
      }
    }

    public final void write(Directory directory) throws IOException {
      Output output = directory.createFile("segments.new");
      try {
	output.writeInt(counter);		  // write counter
	output.writeInt(size());		  // write infos
	for (int i = 0; i < size(); i++) {
	  SegmentInfo si = info(i);
	  output.writeString(si.name);
	  output.writeInt(si.providerCount);
	}
      } finally {
	output.close();
      }

      // install new segment info
      directory.renameFile("segments.new", "segments");
    }
  }

  public static void main(String[] args) throws Exception {
    IndexWriter writer = new IndexWriter("index", true);
    File[] files = new File("regs").listFiles();
    for (int i = 0; i < files.length; i++)
      if (files[i].getName().endsWith(".xml"))
	  {
	      Provider p = new Provider(files[i]);
	      writer.add(p);
	  }
    writer.optimize();
    writer.close();
  }

}
