/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: JxtaProviderInterface.java,v 1.1.1.1 2001/06/01 16:50:00 jbeatty Exp $
 */
package net.jxta.search.resolver;

import net.jxta.protocol.PipeAdvertisement;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.endpoint.Message;
import net.jxta.pipe.Pipe;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.MimeMediaType;
import net.jxta.discovery.Discovery;

import net.jxta.search.util.ThreadService;
import net.jxta.search.util.LogService;
import net.jxta.search.util.JxtaPipeUtil;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.ByteArrayInputStream;

import org.xml.sax.SAXException;

/** Provides ability to provision/register jxta peers **/
class JxtaProviderInterface implements Runnable {

    /** the peer group that the provider interface lives in **/
    private PeerGroup peerGroup;

    /** the input pipe for receiving registration messages **/
    private InputPipe inPipe;

    /** the filename that stores the input pipe adv **/
    private static final String REGISTER_PIPE_FILENAME =
	"registration_pipe.xml";

    /** input pipe name **/
    private static final String REGISTER_PIPE_NAME =
	"Jxtasearch Registration Pipe";

    /** specifies whether the provider interface should be running
	or not **/
    private boolean isRunning=true;

    /** Thread service for thread pools of reg request processors **/
    private ThreadService threadService;

    /** Thread service parms. TODO: make this configurable **/
    private static final int TS_MIN_THREADS = 10;
    private static final int TS_MAX_THREADS = 20;
    private static final int TS_IDLE_TIMEOUT = 100;

    /** poll timeout value for inpipe **/
    private static final int POLL_TIMEOUT = 1000;

    private ResolverService resolverService;

    private File regDir;

    public JxtaProviderInterface( PeerGroup peerGroup,
				  File regDir ) throws Exception {
	this.peerGroup = peerGroup;
	this.regDir = regDir;

	resolverService = ServiceLoader.getResolverService();

	PipeAdvertisement adv =
	    JxtaPipeUtil.createPipeAdv( new File(REGISTER_PIPE_FILENAME),
					REGISTER_PIPE_NAME,
					peerGroup );

	inPipe = peerGroup.getPipe().createInputPipe(adv);
	peerGroup.getDiscovery().publish (adv, Discovery.ADV);

	threadService = new ThreadService ( TS_MIN_THREADS,
					    TS_MAX_THREADS,
					    TS_IDLE_TIMEOUT,
					    false,
					    false );
    }

    public void stop() {
	isRunning = false;
    }

    public void run() {
	Message msg = null;
	while( isRunning ) {
	    try {
		msg = inPipe.poll( POLL_TIMEOUT );
		if (isRunning && msg != null) {
		    threadService.getThread (new RequestProcessor(msg));
		}
	    } catch( Exception e ) {
		System.out.println( "Exception while processing "+
				    "registration: " + e.getMessage() );
		e.printStackTrace();
	    }
	}
    }

    private class RequestProcessor implements Runnable {
	Message msg;

	public RequestProcessor( Message msg ) {
	    this.msg = msg;
	}

	public void run() {
	    System.out.println("Resolver: Received peer registration message");
	    String failureCode = null;
	    try {
		String externalId = getExternalId( msg.pop( "peerid" ) );
		String requestCode = getRequestCode(msg.pop("requestcode"));
		
		InputStream responsePipeAdvIn=null;
		try {
		    responsePipeAdvIn = msg.pop( "responsePipe" );
		} catch( Exception e ) {
		    // no code intended
		}
		
		if( requestCode.equals( "register" ) )
		    failureCode = 
			processRegistration(externalId,
					    msg.pop( "registration" ));
		else if( requestCode.equals( "unregister" ) )
		    failureCode = processUnregistration(externalId);
		else
		    failureCode = "Unknown request code: " + requestCode;
		  
		String resultCode = (failureCode==null)?"success":failureCode;
		System.out.println("Resolver: " + requestCode + ": " 
				   +resultCode);

		if( responsePipeAdvIn!=null ) {
		    PipeAdvertisement responsePipeAdv = (PipeAdvertisement) 
			AdvertisementFactory.newAdvertisement
			(new MimeMediaType ("text/xml"),
			 msg.pop ("responsePipe") );
		    
		    OutputPipe responsePipe = 
			peerGroup.getPipe().createOutputPipe (responsePipeAdv,
							      Pipe.NonBlocking,
							      10000 );
		    
		    Message msg = peerGroup.getPipe().createMessage();

		    msg.push( "resultCode",
			      new ByteArrayInputStream( resultCode.getBytes() ));
		    
		    responsePipe.send( msg );
		}
	    } catch( Exception e ) {
		System.out.println( "Error processing registration: " + 
				    e.getMessage() );
		e.printStackTrace();
	    }
	}


	private String getRequestCode( InputStream in ) throws IOException {
	    byte[] buf = new byte[10];
	    return streamToString( in, buf, 10 );
	}
	
	private String processRegistration( String externalId,
					    InputStream in ) {
	    try {
		System.out.println( "Resolver: processing registration " +
				    "message from peer " + externalId );
		
		// write the registration to disk
		File regFile = writeRegToFile( externalId, in );
		resolverService.addProvider( externalId, regFile );
	    } catch( SAXException e ) { 
		e.printStackTrace();
		return "Error parsing registration";
	    } catch( IOException e ) {
		e.printStackTrace();
		return "Unknown error processing registration";
	    }
	    
	    return null;
	}

	private String processUnregistration( String externalId ) {
	    resolverService.removeProvider( externalId );
	    String filename = getRegFilename(externalId);
	    System.out.println( "Deleting file: " + filename );
	    File file = new File( regDir, filename );
	    if (file.exists()) {
		try {
		    System.out.println( "File exists. Attempting to delete: "+
					filename );
		    file.delete();
		    System.out.println( "Successfully deleted file " +
					filename );
		} catch( SecurityException e ) {
		    System.out.println( "JxtaProviderInterface: could not "+
					"remove registration for " +
					externalId );
		}
	    }
	    return null;
	}

	private String getExternalId(InputStream in ) throws IOException {
	    byte[] buf = new byte[128];
	    return streamToString( in, buf, 128 );
	}

	private String getRegFilename( String id ) {
	    // strip off the 'jxta://'
	    String base = id.substring(7);
	    return base + ".xml";
	}
	
	private File writeRegToFile( String id, InputStream in )
	throws IOException {
	    File f = new File( regDir, getRegFilename(id) );
	    OutputStream out = new FileOutputStream( f );
	    byte[] buf = new byte[8192];
	    int r;
	    
	    while ((r = in.read (buf)) > 0) {		
		out.write (buf, 0, r);
	    }
	    
	    out.close ();
	    out = null;
	    return f;
	}

	String streamToString (InputStream in, byte[] buf, int maxLen) 
	    throws IOException {
	    int pos = 0;
	    int n;
	    
	    while ((n = in.read (buf, pos, maxLen - pos)) > 0) {
		pos += n;
	    }
		
	    return new String (buf, 0, pos);
	}	    
    }
}
