/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Provider.java,v 1.1.1.1 2001/06/01 16:50:03 jbeatty Exp $
 */

package net.jxta.search.resolver;

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import java.util.*;
import java.io.*;
import java.net.*;
import net.jxta.search.util.KeywordService;

/** A parsed provider registration. */
class Provider {

    private static boolean debug = false;
    private File categorydir = null;

    TreeSet querySpaces = new TreeSet();
    URL    queryServer;
    URL    providerLink;
    URL    providerImageURL;
    int    providerImageWidth;
    int    providerImageHeight;
    String providerTitle;
    String providerDescription;
    TreeMap querySpaceToPredicates = new TreeMap();

    // general predicates are predicates of one clause for each query space
    // they are not part of the registration. They are added by the index
    // as a result of user feedback 
    TreeMap querySpaceToGeneralPredicates = new TreeMap(); 

    public static void setDebug() {
	debug = true;
    }

    public static void clearDebug() {
	debug = false;
    }

    /** The URL of the querySpaces registered. */
    public TreeSet getQuerySpaces() { return querySpaces ; }

    /** The URL that queries should be routed to. */
    public URL getQueryServer() { return queryServer; }

    /** The URL of the providers home */
    public URL getProviderLink () { return providerLink; }

    /** The URL of the providers icon */
    public URL getProviderImageURL () { return providerImageURL; }

    /** The WIDTH attribute for the providers icon */
    public int getProviderImageWidth () { return providerImageWidth; }

    /** The HEIGHT attribute for the providers icon */
    public int getProviderImageHeight () { return providerImageHeight; }
  
    /** The Providers Name/Title */
    public String getProviderTitle () { return providerTitle; }
  
    /** The providers Description/Tagline/motto */
    public String getProviderDescription () { return providerDescription; }


    public void setCategoryDir(File categorydir) {
	this.categorydir = categorydir;
    }
    
    /** Add patterns to a general one clause predicate as a result of positive *
     *  user feedback */
    public void addToGeneralPredicates(String querySpace, String pattern)
    {
	TreeSet ts = (TreeSet) querySpaceToGeneralPredicates.get(querySpace);
	if(ts == null) ts = new TreeSet();
	ts.add(pattern);
	querySpaceToGeneralPredicates.put(querySpace, ts);
    }
    
    /** Returns the patterns added by the index as a result of user feedback. *
     *  These patterns were not in the registration */
    public TreeSet getAddedPatterns(String querySpace)
    {
	return ((TreeSet) querySpaceToGeneralPredicates.get(querySpace));
    }

    /** The predicates, one of which must match in order for queries to be routed
     * to this provider. */
    public ArrayList getPredicates(String querySpace) { 
	ArrayList al = (ArrayList) querySpaceToPredicates.get(querySpace);
	if(al==null) al = new ArrayList();
	return al;
    }
    
    /** return a list of all predicates, regardless of the queryspace */
    public ArrayList getPredicates() {
	Iterator iterator = querySpaces.iterator();
	ArrayList returnedList = new ArrayList();
	while (iterator.hasNext()) {
	    returnedList.addAll((Collection) querySpaceToPredicates.get((String) iterator.next()));
	    
	}
	return returnedList;
    }
	

    private static DocumentBuilder BUILDER;
    static {
	try {
	    BUILDER = DocumentBuilderFactory.newInstance().newDocumentBuilder();
	} catch (Exception e) {
	    throw new RuntimeException(e.toString());
	}
    }

    /** Constructs a provider from its XML description. */
    public Provider(String xml) throws SAXException, IOException {
	this(BUILDER.parse(new ByteArrayInputStream(xml.getBytes())));
    }


    /** Constructs a provider from its XML description. */
    public Provider(File file) throws SAXException, IOException {
	this(BUILDER.parse(file));
    }


    /** Constructs a provider from its XML description. */
    public Provider(Document doc) throws SAXException, MalformedURLException {

	//      if (!"provider".equals(doc.getDoctype().getName()))
	//        throw new SAXException("Document does not use 'provider' DTD.");

	Element root = doc.getDocumentElement();
	if (!"register".equals(root.getTagName()))
	    throw new SAXException("Root tag is not 'register':");

	NodeList nodes = root.getChildNodes(); 
	this.providerTitle = root.getAttribute("name");
	this.providerDescription = root.getAttribute("description");
	String queryServerString = root.getAttribute("query-server");
	if(queryServerString.length() > 0 ) 
	    this.queryServer = new URL(queryServerString);
	String providerLinkString = root.getAttribute("uri");
	if(providerLinkString.length() > 0)
	    this.providerLink = new URL(providerLinkString);
	for (int i = 0; i < nodes.getLength(); i++) {
	    Node node = nodes.item(i);
	    String nodeName = node.getNodeName();

	    if ("query-space".equals(nodeName)) {		  // query-space node
		String querySpaceString = ((Text)((node.getAttributes()).getNamedItem("uri")).getFirstChild()).getData();
		
		querySpaces.add(querySpaceString);
		NodeList nodelist = node.getChildNodes();
		for(int j = 0; j < nodelist.getLength();j++) {
		  Node subNode = nodelist.item(j);
		  String subNodeName = subNode.getNodeName();

		  if ("category".equals(subNodeName)) { // category node
		      // retrieve the predicates of this queryspace
		      ArrayList predicates = (ArrayList) querySpaceToPredicates.get(querySpaceString);
		      if(predicates == null)
			  predicates = new ArrayList();
		      // add predicate to predicates
		      predicates.add(parseCategory(subNode));
		      querySpaceToPredicates.put(querySpaceString, predicates);

		  } else if ("predicate".equals(subNodeName)) {  // predicates node
		      // retrieve the predicates of this queryspace
		      ArrayList predicates = (ArrayList) querySpaceToPredicates.get(querySpaceString);
		      if(predicates == null)
			  predicates = new ArrayList();
		      // add predicate to predicates
		      predicates.add(parsePredicate(subNode));	  // parse & save it
		      querySpaceToPredicates.put(querySpaceString, predicates);
		      
		  }
		}

	    } else if ("query-server".equals(nodeName)) { //deprecated
		this.queryServer = new URL(((Text)node.getFirstChild()).getData());

	    }  else if ("title".equals(nodeName)) { //deprecated 
		this.providerTitle = new String(((Text)node.getFirstChild()).getData());

	    } else if ("description".equals(nodeName)) { //deprecated
		this.providerDescription = new String(((Text)node.getFirstChild()).getData());
        
	    } else if ("link".equals(nodeName)) {  //deprecated
		this.providerLink = new URL(((Text)node.getFirstChild()).getData());

	    } else if ("image".equals(nodeName)) { // Provider Icon Node
		this.parseProviderImage(node);

	    } else if (!isEmptyTextNode(node)) {
		System.err.println("Ignoring node of type: " + nodeName);
	    }
	}
	if(this.queryServer == null) throw new MalformedURLException();
    }

    private void parseProviderImage (Node node) throws SAXException, MalformedURLException {
	NodeList cnodes = node.getChildNodes();

	for (int i = 0; i < cnodes.getLength (); i++) {
	    Node cnode = cnodes.item(i);
	    String cnodeName = cnode.getNodeName();

	    if ("url".equals(cnodeName)) {
		this.providerImageURL = new URL (((Text)cnode.getFirstChild()).getData());

	    } else if ("width".equals(cnodeName)) {
		this.providerImageWidth = Integer.parseInt(((Text)cnode.getFirstChild()).getData());
        
	    } else if ("height".equals(cnodeName)) {
		this.providerImageHeight = Integer.parseInt(((Text)cnode.getFirstChild()).getData());

	    } /* else if (!isEmptyTextNode(node)) {
		 System.err.println("Ignoring node of type: " + cnodeName);
		 }*/

	}

    }


    /** 
     * Category nodes are now expanded in the <predicate>  
     * tags, and contain lists of keywords.
     * 
     * @param Node node is the xml tag containing the 
     *        category name, which is used to construct the 
     *        category file name for the file that contains 
     *        the keywords for the category.
     *
     * @return Predicate consisting of the keywords associated
     *         with a particular category.
     */
    private Predicate parseCategory(Node node)
	throws SAXException  {

  	String category = node.getFirstChild().getNodeValue();
        //Node kwnode = node.getFirstChild();

        if (debug) {
	    System.out.println("Category: " + category);
	}

	if (category == null)
	    throw new SAXException("A category name must appear");

	File categoryfile; 

        if (debug) {
	    categoryfile = new File ("./keywordtest/"
				     + category.toLowerCase()
				     + ".txt");
	    System.out.println("Category File: "
			       + categoryfile);
        } else {
	    categoryfile = new File (categorydir
				     + category.toLowerCase()
				     + ".txt");
        }


        KeywordService kws = new KeywordService(categoryfile);
        String kwstring = kws.readKeywords(categoryfile);

        return makeNewPredicate(kwstring);

       
    } /* close parseCategory() */


    /* FIXME: This should probably go into the Predicate class. */
    private Predicate makeNewPredicate(String kwstring) {

	StringTokenizer st = new StringTokenizer(kwstring);
        int numpatterns = st.countTokens();
        String [] patterns = new String[numpatterns];  
 
        for (int i = 0; i < numpatterns; i++) {
	    patterns[i] = st.nextToken();
        }
            
	Disjunction disjunction = new Disjunction(patterns);
        ArrayList conjunction = new ArrayList();
        conjunction.add(disjunction);
        Predicate p = new Predicate(conjunction);
        return p;

    }  /* close makeNewPredicate() */

  
    private Predicate parsePredicate(Node node) throws SAXException {

	NodeList nodes = node.getChildNodes();

	if (nodes.getLength() == 1) {
	    /** 
             * Handle category tags in here. 
             * We have to do this because of the way Doug
             * wrote the predicate code: when the category
             * tag is part of a predicate, the sax parser 
             * doesn't ever see it.
	     */
	    String value = node.getFirstChild().getNodeName();
            if (value.equals("category")) {
		return parseCategory(node.getFirstChild());
	    }
	    return new Predicate(parseAnd(nodes.item(0)));
	}


	ArrayList conjunction = null;
	for (int i = 0; i < nodes.getLength(); i++) {


	    if (!isEmptyTextNode(nodes.item(i))) {	  // skip empty text nodes
		if (conjunction == null) {
		    conjunction = parseAnd(nodes.item(i));
		} else {
		    throw new SAXException("A 'predicate' must contain 'and' tag " +
					   "or a single pattern:" + node);
		}
	    }
	}

	if (conjunction == null)			  // empty predicate
	    conjunction = new ArrayList();
    
	return new Predicate(conjunction);

    }  /* close parsePredicates() */



	  
    private ArrayList parseAnd(Node node) throws SAXException {
	ArrayList conjunction = new ArrayList();
	if ("and".equals(node.getNodeName())) {	  // explicit conjunction
	    NodeList nodes = node.getChildNodes(); 
	    for (int i = 0; i < nodes.getLength(); i++)
		if (!isEmptyTextNode(nodes.item(i)))
		    conjunction.add(parseOr(nodes.item(i)));
	} else if (!isEmptyTextNode(node)) {	  // one implicit conjunct
	    conjunction.add(parseOr(node));
	}
	return conjunction;
    }

    private static final String[] STRING_ARRAY = new String[0];

    private Disjunction parseOr(Node node) throws SAXException {
	ArrayList patterns = new ArrayList();
	if ("or".equals(node.getNodeName())) {	  // explicit disjunction
	    NodeList nodes = node.getChildNodes();
	    for (int i = 0; i < nodes.getLength(); i++)
		parsePatterns(nodes.item(i), patterns);
	} else {					  // one implicit disjunct
	    parsePatterns(node, patterns);
	}
	return new Disjunction((String[])patterns.toArray(STRING_ARRAY));
    }

    private void parsePatterns(Node node, ArrayList patterns)
	throws SAXException {
	StringBuffer pattern = new StringBuffer();
	Node[] children;
	for (children = getNonEmptyChildren(node); 
	     children.length == 1
		 && !"quote".equals(node.getNodeName());
	     children = getNonEmptyChildren(node)) {

	    if (node.getAttributes().getLength() > 0)
		throw new SAXException("Only terminal nodes may have attributes:"
				       + node);

	    pattern.append(node.getNodeName());	  // add tag to pattern
	    pattern.append(Pattern.TAG);

	    node = children[0];
	}

	int childCount = children.length;

	if ("quote".equals(node.getNodeName())) {	  // a quote
	    if (childCount == 1 && node.getFirstChild() instanceof Text) {
		pattern.append(((Text)node.getFirstChild()).getData()); // add pattern
		patterns.add(pattern.toString());
	    } else if (childCount == 0) {
	    } else {
		throw new SAXException("quote nodes may only contain text: " + node);
	    }
	    return;
	}

	if (childCount > 1)
	    throw new SAXException("Pattern may only contain a single path: "
				   + node);

	if (childCount == 0) {			  // at end of pattern

	    NamedNodeMap attributes = node.getAttributes();

	    if (node instanceof Text) {		  // text patterns
		Text text = (Text)node;
		Enumeration e = TextTokenizer.newTokenizer(text.getData());
		while (e.hasMoreElements()) {		  // tokenize text
	  
		    int savedLength = pattern.length();	  // save pattern length
		    pattern.append((String)e.nextElement()); // add word pattern
		    patterns.add(pattern.toString());
		    pattern.setLength(savedLength);	  // restore pattern
		}
	    } else if (attributes.getLength() > 1) {
		throw new SAXException("Pattern may only have a single attribute: "
				       + node);
	    } else if (attributes.getLength() == 1) {	  // an attribute pattern
		Attr attr = (Attr)attributes.item(0);
		pattern.append(node.getNodeName());	  // add tag to pattern
		pattern.append(Pattern.ATTRIBUTE);
		pattern.append(attr.getName());
		pattern.append(Pattern.VALUE);
		pattern.append(attr.getValue());
		patterns.add(pattern.toString());
	    } else {					  // a tag-only pattern
		pattern.append(node.getNodeName());	  // add tag to pattern
		pattern.append(Pattern.TAG);
		patterns.add(pattern.toString());
	    }
	}
    }

    private static final Node[] NODES_PROTO = {};

    private Node[] getNonEmptyChildren(Node node) {
	NodeList nodes = node.getChildNodes();
	if (nodes == null)
	return NODES_PROTO;
	Vector result = new Vector();
	for (int i = 0; i < nodes.getLength(); i++) {
	    Node child = nodes.item(i);
	    if (!isEmptyTextNode(child))
		result.add(child);
	}
	return (Node[])result.toArray(NODES_PROTO);
    }

    private boolean isEmptyTextNode(Node node) {
	if (!"#text".equals(node.getNodeName()))	  // check that it's text
	    return false;

	String data = ((Text)node).getData();	  // check for whitespace-only
	boolean allWhite = true;
	for (int i = 0; i < data.length(); i++) loop: {
	    switch (data.charAt(i)) {
	    case ' ': case '\t': case '\n': case '\r': case '\f':
		break;
	    default:
		allWhite = false;
		break loop;
	    }
	}
	if (allWhite)
	    return true;

	Enumeration e = TextTokenizer.newTokenizer(data); // tokenize
	if (!e.hasMoreElements())
	    return true;

	return false;
    }

    /** Generate an XML representation of the provider */
    public String toString() {
	StringBuffer buffer = new StringBuffer();
	buffer.append("<register ");
	buffer.append(" query-server=\""+queryServer+"\"");
	if(providerTitle.length() > 0) 
	    buffer.append("  name=\"" + providerTitle + "\" ");
	if(providerLink != null) 
	    buffer.append("  uri=\"" + providerLink + "\" ");
	if(providerDescription.length() > 0)
	    buffer.append("  description=\"" + providerDescription + "\" ");
	buffer.append(">\n");

	buffer.append("  <image>");
	if (providerImageURL != null) { 
	    buffer.append("    <url>" + providerImageURL.toString() + "</url>");
	}
	buffer.append("    <width>" + providerImageWidth + "</width>");
	buffer.append("    <height>" + providerImageHeight + "</height>");
	buffer.append("  </image>\n");
	
	Iterator iterator = querySpaces.iterator();
	while (iterator.hasNext()) {
	    String qs = (String) iterator.next();
	    ArrayList predicates = (ArrayList) querySpaceToPredicates.get(qs);
	
	    buffer.append(" <query-space uri=\""+qs+"\">\n");
	    for (int i = 0; i < predicates.size(); i++) {
		buffer.append("  <predicate>\n");
		buffer.append(predicates.get(i));
		buffer.append("  </predicate>\n");
	    }
	    buffer.append("  <query-space>\n");
	}
	buffer.append("</register>\n");
	return buffer.toString();
    }



    private static void testProvider() {

	String provider = "./testregs/CNN.com.xml";

	testProvider(provider);

    }  /* close testProvider() */


    private static void testProvider(String provider) {

	setDebug();
	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	//factory.setValidating(true);

	try {
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc = builder.parse(new File(provider));
	    System.out.println(new Provider(doc));
	} catch (Exception e) {
	    e.printStackTrace();
	    System.out.println(e);
	}


    }  /* close testProvider(String) */


    /** For debugging. */
    public static void main(String[] args) throws Exception {
 
	setDebug();

	if (args.length >= 1) {
	    testProvider(args[0]);
	} else {
	    testProvider();
	}

    }  /* close main() */


    /// Code purg... delete after 1 June 2001 if not used before then.
    /** 
     * Category nodes are now expanded in the <predicate>  
     * tags, so the xml parser never sees them.
     * 
     * @param Node catnode is the xml tag containing the 
     *        category name, which is used to construct the 
     *        category file name for the file that contains 
     *        the keywords for the category.
     *
     * @return Predicate consisting of the keywords associated
     *         with a particular category.
     */
    /*
    private Predicate handleCategoryNode(Node catnode) {
       
	String category = catnode.getFirstChild().getNodeValue();
    
        Node kwnode = catnode.getFirstChild();
 
        File categoryfile;

        categoryfile = new File ("./keywordtest/"
				 + category.toLowerCase()
				 + ".txt");

        KeywordService kws = new KeywordService(categoryfile);
        String kwstring = kws.readKeywords(categoryfile);

	return makeNewPredicate(kwstring);  // temp for compiling
    }
    */


}



class TestProvider {


    /* For debugging. */
    /* yeah yeah, if you gotta a better way to do this, I am 
     * all ears...
     */
    private void printNodeType(Node node) {

	if (node.getNodeType() == node.ATTRIBUTE_NODE) {
	    System.out.println("Node type is ATTRIBUTE_NODE");
	} else if (node.getNodeType() == node.CDATA_SECTION_NODE) {
	    System.out.println("Node type is CDATA_SECTION_NODE");
	} else if (node.getNodeType() == node.COMMENT_NODE) {
	    System.out.println("Node type is COMMENT_NNODE");
	} else if (node.getNodeType() == node.DOCUMENT_FRAGMENT_NODE) {
	    System.out.println("Node type is DOCUMENT_FRAGMENT_NODE");
	} else if (node.getNodeType() == node.DOCUMENT_NODE) {
	    System.out.println("Node type is DOCUMENT_NODE");
	} else if (node.getNodeType() == node.DOCUMENT_TYPE_NODE) {
	    System.out.println("Node type is DOCUMENT_TYPE_NODE");
	} else if (node.getNodeType() == node.ELEMENT_NODE) {
	    System.out.println("Node type is ELEMENT_NODE");
	} else if (node.getNodeType() == node.ENTITY_NODE) {
	    System.out.println("Node type is ENTITY_NODE");
	} else if (node.getNodeType() == node.ENTITY_REFERENCE_NODE) {
	    System.out.println("Node type is ENTITY_REFERENCE_NODE");
	} else if (node.getNodeType() == node.NOTATION_NODE) {
	    System.out.println("Node type is NOTATION_NODE");
	} else if (node.getNodeType() == node.PROCESSING_INSTRUCTION_NODE) {
	    System.out.println("Node type is PROCESSING_INSTRUCTION_NODE");
	} else if (node.getNodeType() == node.TEXT_NODE) {
	    System.out.println("Node type is TEXT_NODE");
	} else {
	    System.out.println("Unknown node type");
	}

    }  /* close printNodeType() */


}
