/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: QueryHandler.java,v 1.1.1.1 2001/06/01 16:50:05 jbeatty Exp $
 */

package net.jxta.search.resolver;

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.util.*;
import java.io.*;

/** Collects all patterns in an XML query.
 * 
 * Patterns are encoded as strings containing a tag path, and, optionally,
 * either an attribute/value pair or a text word.  Tag path elements are
 * separated by '>'.  Thus the pattern "FOO>BAR>baz" refers to a FOO tag with
 * an embedded BAR tag containing the text word "baz".  The pattern "FOO&A=V"
 * refers to a FOO tag whose "A" attribute has value "V".
 */
class QueryHandler implements ContentHandler {

  // stack containing tags and text buffers
  private int stackPointer = 0;
  private String[] tagStack = {""};
  private StringBuffer[] bufferStack = {new StringBuffer()};

  private StringBuffer pattern = new StringBuffer(); // current pattern
  private TreeSet patterns = new TreeSet();	     // collected patterns
  private String  space    = null;
  private int     maxProviders = 20;
  private boolean buildStack = false;

  private static boolean debug = false;


    public static void setDebug() {
        debug = true;
    }

    public static void clearDebug() {
	debug = false;
    }



  /** Returns all collected patterns. */
  public TreeSet getPatterns() {
    return patterns;
  }

  public String getSpace () {
    return space;
  }


  public int getMaxProviders () {
    return maxProviders;
  }

  public void setDocumentLocator (Locator l) {

  }

  public void startDocument () {

  }

  public void endDocument () {

  }

  public void startPrefixMapping (String a, String b) {

  }

  public void endPrefixMapping (String a) { 

  }

  /** Receive notification of the start of an element.
   */
  public void startElement(String namespaceURI,
			   String tag,
			   String qName,
			   Attributes attributes)  throws SAXException {
      

      if (tag.equals ("request")) {
	
	  String maxProvidersStr = attributes.getValue ("max-fanout");

	  if (maxProvidersStr != null && maxProvidersStr.length () > 0)
	      maxProviders = Integer.parseInt (maxProvidersStr);

	  space = attributes.getValue ("query-space");

      } else {

	  if (tag.equals ("query")) 

	      buildStack = true;
	  
	  if (buildStack) {
	      if (tagStack.length == stackPointer) {	  // grow stack
		  String[] newTagStack = new String[stackPointer * 2];
		  System.arraycopy(tagStack, 0, newTagStack, 0, stackPointer);
		  tagStack = newTagStack;
		  StringBuffer[] newBufferStack = new StringBuffer[stackPointer * 2];
		  System.arraycopy(bufferStack, 0, newBufferStack, 0, stackPointer);
		  bufferStack = newBufferStack;
	      }
	      
	      tagStack[stackPointer] = tag;		  // push tag onto stack
	      stackPointer++;
	      
	      /* The debug code can be removed later if the overhead of testing
	       * the debug flag is too high.
	       */
	      if (debug) {
		  System.out.println(tag);
	      }
	      
	      pattern.append(tag);			  // append tag onto pattern
	      pattern.append(Pattern.TAG);
	      
	      addPatterns();				  // add element patterns
	      
	      for (int i = 0; i < attributes.getLength(); i++) // add attribute patterns
		  addAttribute(attributes.getQName(i), attributes.getValue(i));
	  }
      }
  }


  /** Receive notification of the end of an element. */
  public void endElement (String namespaceURI, String name, String qName)
	throws SAXException {


      if (buildStack) {
	  
	  if (stackPointer == 0) return;		  // outside of tags
	  addText();					  // add patterns for text
	  
	  // pop stack & pattern
	  pattern.setLength(pattern.length()-(tagStack[stackPointer-1].length()+1));
	  stackPointer--; 

	  if (name.equals ("query")) 
	
	      buildStack = false;

      }
  }


  public void ignorableWhitespace (char chars[], int start, int length) {

  }

  public void processingInstruction (String a, String b) {

  }

  public void skippedEntity (String s) { }


  /** Receive notification of character data inside an element. */
  public void characters (char chars[], int start, int length)
    throws SAXException {

    if (! buildStack || stackPointer == 0) return;		  // outside of tags

    StringBuffer buffer = bufferStack[stackPointer-1]; // find correct buffer
    if (buffer == null)
      buffer = bufferStack[stackPointer-1] = new StringBuffer();
    buffer.append(chars, start, length);	  // buffer some text


    /* The debug code can be removed later if the overhead of testing
     * the debug flag is too high.
     */
    if (debug) {
	System.out.println("Query terms: " + new String(chars));
    }

  }


  private void addAttribute(String name, String value) {
    int savedLength = pattern.length();		  // save pattern length
    pattern.append(Pattern.ATTRIBUTE);
    pattern.append(name);			  // add attribute to pattern
    pattern.append(Pattern.VALUE);
    pattern.append(value);			  // add value to pattern
    addPatterns();
    pattern.setLength(savedLength);		  // restore pattern
  }


  private void addText() {
    if (stackPointer == 0) return;		  // outside of tags
    StringBuffer buffer = bufferStack[stackPointer-1];
    if (buffer == null || buffer.length() == 0)
      return;					  // no buffered text
    String text = buffer.toString().trim();
    buffer.setLength(0);			  // clear the buffer
    if (text.length() == 0)
      return;

    Enumeration e = QueryTokenizer.newTokenizer(text);
    while (e.hasMoreElements()) {		  // tokenize text
      int savedLength = pattern.length();	  // save pattern length
      pattern.append((String)e.nextElement());	  // add patterns for word
      addPatterns();	
      pattern.setLength(savedLength);		  // restore pattern
    }
  }


  private void addPatterns() {
    int length = pattern.length();
    int start = 0;

    for (int i = 0; i < stackPointer; i++) {	  // add all suffixes
      patterns.add(pattern.substring(start, length));
      start += tagStack[i].length()+1;
    }
    if (start != length)
      patterns.add(pattern.substring(start, length));
  }


    /** Unit test for the QueryHandler.  The "queryhandletest.xml"
     *	file should be maintained to match the InfraSearch resolver
     *	documentation.
     */
    /** TODO: Add some bogus query testing files that are sure to
     *  fail the handler class, and implement the appropriate
     *  exceptions.
     */
    private static void testQueryHandler() {

        File somefile = new File("./queryhandletest.xml");
	testQueryHandler(somefile);

    }  /* close testQueryHandler() */


    private static void testQueryHandler(File somekindoffile) {

        setDebug();

	try {
	    
	    XMLReader reader = XMLReaderFactory.createXMLReader ();
	    QueryHandler handler = new QueryHandler();
	    reader.setContentHandler (handler);


	    reader.parse (new InputSource
		(new FileInputStream (somekindoffile)));
	    System.out.println("query-space: " + handler.getSpace());
	} catch (Exception e) {
	    System.out.println(e);
	}


    }  /* close testQueryHandler(File) */


  public static void main(String[] args) throws Exception {

      testQueryHandler();

  }  /* close main() */

}
