/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ResolverJxtaApp.java,v 1.1.1.1 2001/06/01 16:50:06 jbeatty Exp $
 */
package net.jxta.search.resolver;

/* jxta imports */
import net.jxta.platform.Application;
import net.jxta.peergroup.PeerGroup;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.pipe.PipeID;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.MimeMediaType;
import net.jxta.discovery.Discovery;
import net.jxta.service.Service;
import net.jxta.endpoint.Message;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.ServiceNotFoundException;


/* java imports */
import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.InvocationTargetException;


/* xml imports */
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.XMLReaderFactory;
import javax.xml.parsers.DocumentBuilder;


/* gonesilent imports */
import net.jxta.search.util.ThreadService;


class ResolverJxtaApp implements Application {

    PeerGroup  peerGroup;
    InputPipe  requestPipe;
    OutputPipe responsePipe;
    ResolverService resolverService;
    ThreadService   threadService;

    DocumentBuilder builder;
    RequestAcceptor requestAcceptor;
    
    static final int MSG_READ_TIMEOUT = 30 * 1000;


    public ResolverJxtaApp (ResolverService resolverService,
			    ThreadService   threadService,
			    DocumentBuilder builder)
			    
    {
	this.resolverService = resolverService;
	this.threadService   = threadService;
	this.builder         = builder;
    }

    public void init (PeerGroup peerGroup, Advertisement serviceAdv) 
	throws PeerGroupException
    {
	PipeAdvertisement adv;

	this.peerGroup = peerGroup;
	
	try {
	    
	    adv = (PipeAdvertisement) AdvertisementFactory.newAdvertisement
		(PipeAdvertisement.getAdvertisementType ());

	    adv.setPipeID (new PipeID (peerGroup.getID ()));

	    adv.setName ("InfraSearch Resolver Request Pipe");

	} catch (InvocationTargetException e) {

	    throw new RuntimeException
		("Cannot create new pipe advertisement: " +
		 e.getTargetException ().getMessage ());
	}

	try {
	    peerGroup.getDiscovery ().publish (adv, Discovery.ADV);

	} catch (IOException e) {

	    throw new PeerGroupException ("Unable to publish pipe " +
					  "advertisement");
	}

	try {
	    requestPipe = peerGroup.getPipe ().createInputPipe (adv);
	} catch (IOException e) {
	    throw new PeerGroupException ("Unable to create input pipe.");
	}

    }

    public int startApp (String[] argv) {
	requestAcceptor = new RequestAcceptor ();

	new Thread (requestAcceptor).start ();

	return 0;
    }

    public void stopApp () {
	requestAcceptor.stop ();

	requestPipe.close ();
    }


    class RequestAcceptor implements Runnable {
	
	boolean running;

	RequestAcceptor () {
	    running = true;
	}

	public synchronized boolean isRunning () {
	    return running;
	}

	public synchronized void stop () {
	    running = false;
	}

	public void run () {

	    Message msg;
	    
	    try {
		
		while (isRunning ()) {
		    msg = requestPipe.poll (MSG_READ_TIMEOUT);

		    if (isRunning ()  &&  msg != null) {

			threadService.getThread (new RequestProcessor (msg));
		    }
		    
		}
	    } catch (IOException e) {
		synchronized (this) {
		    running = false;
		}
	    } catch (InterruptedException e) {
		synchronized (this) {
		    running = false;
		}
	    }
	    
	}

	class RequestProcessor implements Runnable {

	    Message msg;

	    RequestProcessor (Message msg) {
		this.msg = msg;
	    }

	    
	    class Options {
		String schema;
		int    maxProviders;
	    }

	    public void run () {
		MyByteArrayOutputStream out =
			new MyByteArrayOutputStream (1024);
		
		try {
		    Query  query  = 
			new Query (XMLReaderFactory.createXMLReader (), 
				   msg.pop ("request"));
		    		    
		    Searcher.Match[] matches = resolverService.doSearch(query);

		    Searcher.Match   match;
		    
		    
		    out.println("<?xml version='1.0'?>");
		    out.println ("<results>");
		    for (int i = 0; matches != null && i < matches.length; i++) {
			
			match = matches[i];
			
			out.println("<match>");
			out.println("<score>" + match.getScore() + "</score>");
			out.println("<id>" + match.getExternalId() + "</id>");
			out.println("<url>" + match.getQueryServer()+ "</url>");
			out.println("<title>" + match.getProviderTitle() + "</title>");
			out.println("<link>" + match.getProviderLink().toString() + "</link>"); 
			out.println("<description>" + match.getProviderDescription() + "</description>");
			out.println("<image>");
			out.println("<url>" + match.getProviderImageURL().toString() + "</url>");
			out.println("<width>" + match.getProviderImageWidth() + "</width>");
			out.println("<height>" + match.getProviderImageHeight() + "</height>");
			out.println("</image></match>");

		    }

		    out.println ("</results>");
		    
		    /* create jxta results message and send it */
		    msg = peerGroup.getPipe ().createMessage ();
		    msg.push ("results", out.getInputStream ());
		    responsePipe.send (msg);
		    
		} catch (IOException e) {
		    // xxx log
		} catch (SAXException e) {
		    try {
			out.println ("<error>");
			out.println (" Error parsing query: "+e.getMessage());
			out.println ("</error>");
		    } catch (IOException ex) {
			
		    }
		} catch (javax.xml.parsers.ParserConfigurationException e) {
		    try {
			out.println ("<error>");
			out.println (" Internal error.  An administrator " +
				     "has been alerted.");
			out.println ("</error>");
		    } catch (IOException ex) {

		    }
		}
		    

	    }

	    Options getOptions (Message msg) throws IOException, SAXException {
		Options  options = new Options ();
		Document doc   = builder.parse (msg.pop ("query-options"));
		Element  root  = doc.getDocumentElement ();
		NodeList nodes = root.getChildNodes ();
		
		options.maxProviders = 20;
		
		for (int i = 0;  i < nodes.getLength ();  i++) {
		    Element node = (Element) nodes.item (i); 
		    String  name = node.getNodeName ();
		    String  val  = ((Text) node.getFirstChild ()).getData ();
		    
		    if (name.equals ("schema"))
			options.schema = val;
		    else if (name.equals ("maxProviders"))
			options.maxProviders = Integer.parseInt (val);
		}
		
		return options;
	    }
		    
	}
    }
    

    class MyByteArrayOutputStream extends ByteArrayOutputStream {

	int    myPos = 0;
	byte[] newline = "\r\n".getBytes ();
	
	MyByteArrayOutputStream (int size) {
	    super (size);
	}

	void print (String s) throws IOException {
	    super.write (s.getBytes ());
	}

	void println (String s) throws IOException {
	    super.write (s.getBytes ());
	    super.write (newline);
	}

	InputStream getInputStream () {

	    return new InputStream () {

		    public int available () {
			return count - myPos;
		    }
		    
		    public int read () throws IOException {
			if (myPos < count)
			    return buf[myPos++];
			else
			    return -1;
		    }

		    public int read (byte[] toBuf) throws IOException {
			
			return read (toBuf, 0, toBuf.length);
		    }

		    public int read (byte[] toBuf, int off, int len)
			throws IOException
		    {
			if (myPos < count) {

			    int r = (int) Math.min (len, count - myPos);
			    
			    
			    System.arraycopy (buf, myPos, toBuf, off, r);
			    
			    myPos += r;
			    
			    return r;
			} else {
			    return -1;
			}
		    }
		    
		    public void reset () {
			myPos = 0;
		    }
		    
		    public long skip (long n) {
			int r = (int) Math.min ((int) n, count - myPos);

			myPos += n;

			return r;
		    }

		    public boolean markSupported () {
			return false;
		    }
		    
		    public void close () {
			myPos = count;
		    }
		};
	}

    }
    
}
