/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Searcher.java,v 1.1.1.1 2001/06/01 16:50:08 jbeatty Exp $
 */

package net.jxta.search.resolver;

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;
import java.util.*;
import java.io.*;
import java.net.*;
import net.jxta.search.util.*;

/** Searches an index for providers matching a query. */
class Searcher {

  /** A provider's predicate which matched the query. */
  public class Match implements Comparable {
    int providerId;
    int predicateId;
    String[] patterns;
    Index index;
    float score;

    Match(int providerId, int predicateId, String[] patterns, Index index, float score) {
      this.providerId = providerId;
      this.predicateId = predicateId;
      this.patterns = patterns;
      this.index = index;
      this.score = score;
    }
     
   /** Score of the match (highest score normalized to 1
    * Score is based on historical performance of provider/query 
    */
      float getScore() {return score;}
 

    /** The id of a provider that matched the query.
     * @see Index.getProvider(int)
     */
    int getProviderId() { return providerId; }

    /** The predicate within the provider that matched.
     * <p>Useful for debugging predicates.
     * @see Index.getPredicate(int)
     */
    int getPredicateId() { return predicateId; }
    
    /** The patterns within the predicate that matched.
     * <p>Useful for debugging predicates.
     */
    String[] getPatterns() { return patterns; }

    public URL getQueryServer() {
      return index.getProvider(providerId).getQueryServer();
    }

    public String getExternalId() {
      return index.getExternalId(providerId);
    }

    public String getProviderTitle() {
      return index.getProvider(providerId).getProviderTitle();
    }

    public URL getProviderLink() {
      return index.getProvider(providerId).getProviderLink();
    }

    public String getProviderDescription() {
      return index.getProvider(providerId).getProviderDescription();
    }

    public URL getProviderImageURL() {
      return index.getProvider(providerId).getProviderImageURL();
    }

    public int getProviderImageWidth() {
      return index.getProvider(providerId).getProviderImageWidth();
    }

    public int getProviderImageHeight() {
      return index.getProvider(providerId).getProviderImageHeight();
    }
   /* compareTo is based on the Performance Score 
    * If the scores are the same, ranks the two Matches by providerId */
   public int compareTo(Object object) {
      Match other = (Match) object;
      if (this.score == other.score)
	  return this.providerId - other.providerId;
      else return (int) ((other.score - this.score)*100); // orders them descending
   }

  public boolean equals(Object object) {
    Match other = (Match)object;
    return ((this.score == other.score)
	    && (this.providerId == other.providerId));
  }
  }

  /** Searches for providers whose predicates match the query.
   * A maximum of <i>query.maxProviders</i> providers are returned. */
  public Match[] search(Query query, Index index) {
    HashMap providerIdToMatch = new HashMap();
    TreeSet matches = new TreeSet();
    int minClauseCount = 0;
    System.out.println ("Searching with maxMatches = " + query.maxProviders);
    // build priority queue of postings for queryPatterns
    TreeSet queryPatterns = query.patterns;
    ScoreCard sc = new ScoreCard(index);
    Iterator queryIterator = queryPatterns.iterator();
    while (queryIterator.hasNext()) {
      String pattern = (String)queryIterator.next();
      ArrayList postings = index.getPostings(pattern);
      System.out.println ("Pattern " + pattern);
      if (postings != null) {
	  sc.addPostings(pattern, postings);
	  System.out.println ("Got match for pattern " + pattern);
      }
    }


    Match[] result = sc.getOrderedMatches(query.maxProviders);
    return result;   // result may be null if no matches found
  }


/** This class computes the matches to a query and returns an ordered list
 * of the matches together with the total score of each match */

  class ScoreCard {
    TreeMap predicatesToScores;
    Index index;
    int maxProviderId ;
    ScoreCard(Index index) {
        predicatesToScores = new TreeMap();
        this.index = index; 
        try {
	    maxProviderId = index.getMaxProviderId();
	}
	catch(Exception e) {
	    System.err.println("ScoreCard: error when calling index");
	    e.printStackTrace();
	}
    }
    
      /* addPostings add the postings associated with one pattern in the 
       * query to the previous postings */
      public void addPostings(String pattern, ArrayList postings) {
	  int predicateId = 0;
	  int clauseId = 0;
	  int clauseCount = 0;
	  float score = 0;
	  if(postings != null) {
	      for(int i=0; i<postings.size();i++) {
		  Index.Posting posting=(Index.Posting) postings.get(i);
		  predicateId = posting.predicateId;
		  clauseId = posting.clauseId;
		  score = posting.score;
		  Integer predicateInteger = new Integer(predicateId);
		  ScoreAndPatterns[] scoreAndPatterns = (ScoreAndPatterns[]) predicatesToScores.get(predicateInteger);
		  if(scoreAndPatterns != null) { // already exist
		      if(scoreAndPatterns[clauseId] != null)
			  scoreAndPatterns[clauseId].addScorePatternPair(score, pattern);
		      else scoreAndPatterns[clauseId] = new ScoreAndPatterns(score,pattern);
		  }
		  else { // aray does not exist
		      clauseCount = index.getClauseCount(predicateId);
		      scoreAndPatterns = new ScoreAndPatterns[clauseCount];
		      scoreAndPatterns[clauseId]=new ScoreAndPatterns(score,pattern);
		      predicatesToScores.put(predicateInteger,scoreAndPatterns);
		  }
	      }   
	  }
      }
      public Match[] getOrderedMatches(int maxMatches) {
	  TreeMap providerIdsToMatches = new TreeMap();
	  Iterator it = (predicatesToScores.entrySet()).iterator();
	  int nMatched = 0;
	  while (it.hasNext()) {
	      boolean matched = true;
	      int totalScore=0;
	      ArrayList allPatterns = new ArrayList();
	      Map.Entry entry = (Map.Entry) it.next();
	      ScoreAndPatterns[] sp = (ScoreAndPatterns[]) entry.getValue();
	      int predicateId = ((Integer) entry.getKey()).intValue();
	      int providerId;
	      try {
		  providerId = index.getProviderId(predicateId);
		  if(!index.getProviderStatus(providerId)) matched = false;
	      }
	      catch(Exception e) {
		  System.err.println("ScoreCard: error when calling index");
		  e.printStackTrace();
		  return null;
	      }
	      for(int i=0; i < sp.length; i++)    {
		  if(sp[i]==null)  { //the clause did not match
		      matched = false;
		      break;
		  }
		  else  {
		      totalScore += sp[i].getScore();
		      allPatterns.addAll(sp[i].getPatterns());
		  }
	      }
	      if(matched) { //All clauses of a predicate had at least 1 match
		  nMatched++;
		  String [] patterns = new String[allPatterns.size()];
		  for(int cc = 0; cc < allPatterns.size(); cc++)
		      patterns[cc] = (String) allPatterns.get(cc);
		  Integer provInt = new Integer(providerId);
		  Match m = (Match) providerIdsToMatches.get(provInt);
		  if(m !=null) {
		      if(m.getScore() < totalScore)
			  m = new Match(providerId, predicateId, patterns, index, totalScore);
		  }
		  else {
		      m = new Match(providerId, predicateId,patterns, index, totalScore);
		  }
		  providerIdsToMatches.put(provInt, m);
	      }
	  }
	  if(nMatched == 0)
	      return null;
	  else  {
	      Match[] results = new Match[providerIdsToMatches.size()];
	      Iterator iterator = (providerIdsToMatches.values()).iterator();
	      int cc = 0;
	      while(iterator.hasNext())   {
		  results[cc++] = (Match) iterator.next();
	      }
	      Arrays.sort(results);
	      if(nMatched > maxMatches) 
		  {
		      Match[] resultsClipped = new Match[maxMatches];
		      System.arraycopy(results, 0,resultsClipped,0,maxMatches);
		      return resultsClipped;
		  }
	      return results;
	  }
      }
      class ScoreAndPatterns {
	  float score;
	  ArrayList patterns;

	  ScoreAndPatterns(float providerScore, String pattern)
	  {
	      score = 0;
	      patterns = new ArrayList();
	      addScorePatternPair(providerScore, pattern);
	  }
	  
	  void addScorePatternPair(float score, String pattern)
	  {
	      this.score += score;
	      patterns.add(pattern);
	  }
	  float getScore() { return score; }
	  ArrayList getPatterns() { return patterns; }
      }
  }

    /** TODO: Create more testing regs with bad xml to exercise
     * the searcher class.
     */

  public static void testSearcher() {

      String [] args = {"./testregs", "./testregs/sciam.com.xml"};
	
      testSearcher(args);
	
    }  /* close testSearcher() */


  /** For testing and debugging. */
  public static void testSearcher(String[] args) {


    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    //factory.setValidating(true);

    try {
    DocumentBuilder builder = factory.newDocumentBuilder();


    File[] files = new File(args[0]).listFiles();
    Index index = null;
    for (int i = 0; i < files.length; i++) {
      if (!files[i].getName().endsWith(".xml"))
	continue;
      Document doc = builder.parse(files[i]);
      Provider p = new Provider(doc) ;
      TreeSet querySpaces = p.getQuerySpaces();
      String qs = (String) querySpaces.first();
      if (index == null)
	  index = new Index(qs);  //the index that will be created for testing will be 
                                  // that of the first qs of the first file
      
      index.add(files[i].toString(), p);  // will only add p if qs match, otherwise do nothing
    }
    System.out.println(index);

    Query query = new Query(XMLReaderFactory.createXMLReader (), 
                               new FileInputStream(new File(args[1]))); 

    System.out.println("Query patterns: " + query.patterns);

    Searcher searcher = new Searcher();
    Match[] matches = searcher.search(query, index);


    System.out.println();
    for (int i = 0; i < matches.length; i++) {
      System.out.print("Match: ");

      Match match = matches[i];
      System.out.println(match.getQueryServer());
      System.out.println(match.getExternalId());
      System.out.println("Predicate:");
      System.out.print(index.getPredicate(match.predicateId));
      
      System.out.print("Patterns: ");
      for (int j = 0; j < match.patterns.length; j++) {
	System.out.print(match.patterns[j]);
	System.out.print(" ");
      }
      System.out.println("\n");
    }

    } catch (Exception e) {
       System.out.println(e);
    }

  }  /* close testSearcher(String) */


  /** For testing and debugging. */
  public static void main(String[] args) throws Exception {

   testSearcher();

  }  /* close main() */



}
