/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SegmentCreator.java,v 1.1.1.1 2001/06/01 16:50:08 jbeatty Exp $
 */

package net.jxta.search.resolver;

import net.jxta.search.resolver.io.*;
import net.jxta.search.util.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;

import java.io.*;
import java.util.*;

class SegmentCreator {


  private Directory directory;
  
  public SegmentCreator(Directory directory) {
    this.directory = directory;
  }

  class Postings {
    IntVector predicateIds = new IntVector(1);
    IntVector clauseIds = new IntVector(1);
  }

  public void createSegment(Provider provider, String segmentName)
    throws IOException {
    writeProvider(provider, segmentName);
    
    ArrayList predicates = provider.getPredicates();
    
    IntVector clauseCounts = new IntVector(predicates.size());
    IntVector providerIds = new IntVector(predicates.size());
    IntVector predicateNumbers = new IntVector(predicates.size());

    SortedMap patternToPostings = new TreeMap();
    for (int predicateNumber = 0; predicateNumber < predicates.size();
	 predicateNumber++) {
      Predicate predicate = (Predicate)predicates.get(predicateNumber);
      ArrayList conjunction = predicate.conjunction;
      int clauseCount = conjunction.size();
      
      clauseCounts.add(clauseCount);
      providerIds.add(0);
      predicateNumbers.add(predicateNumber);
      
      int predicateId = predicateNumber;

      for (int clauseId = 0; clauseId < clauseCount; clauseId++) {
	
	Disjunction disjunction = (Disjunction)conjunction.get(clauseId);
	String[] patterns = disjunction.patterns;
	
	for (int k = 0; k < patterns.length; k++) {
	  String pattern = patterns[k];
	  
	  Postings postings = (Postings)patternToPostings.get(pattern);
	  if (postings == null)
	    patternToPostings.put(pattern, postings = new Postings());

	  postings.predicateIds.add(predicateId);
	  postings.clauseIds.add(clauseId);
	}
      }
    }

    directory.writeVIntFile(clauseCounts.array(), clauseCounts.length(),
			    segmentName+".pcc", false);
    directory.writeVIntFile(providerIds.array(), providerIds.length(),
			    segmentName+".ppi", true);
    directory.writeVIntFile(predicateNumbers.array(),predicateNumbers.length(),
			    segmentName+".prn", false);

    Output patternClausesOutput = directory.createFile(segmentName+".ptd");
    StringLongWriter patternToPosition
      = new StringLongWriter(directory, segmentName+".pt", true);
    try {
      Iterator i = patternToPostings.entrySet().iterator();
      while (i.hasNext()) {
	Map.Entry entry = (Map.Entry)i.next();
	String pattern = (String)entry.getKey();
	Postings postings = (Postings)entry.getValue();

	patternToPosition.add(pattern, patternClausesOutput.getFilePointer());

	int length = postings.predicateIds.length();
	patternClausesOutput.writeVInt(length);
	patternClausesOutput.writeVInts(postings.predicateIds.array(), 0,
					length, true);
	patternClausesOutput.writeVInts(postings.clauseIds.array(), 0,
					length, false);
      }
    } finally {
      patternClausesOutput.close();
      patternToPosition.close();
    }

  }

  private void writeProvider(Provider provider, String segmentName)
    throws IOException {
    LongVector providerPositions = new LongVector(1);

    Output o = directory.createFile(segmentName+".prd");
    try {
      providerPositions.add(o.getFilePointer());
      o.writeString(provider.toString());
    } finally {
      o.close();
    }

    directory.writeVLongFile(providerPositions.array(),
			     providerPositions.length(),
			     segmentName+".prx", true);
  }


    private static void testSegmentCreator() {
	try {
            Directory directory = new FSDirectory("testindex", true);
	    String filename = "testfilename";
	    testSegmentCreator(directory, filename);
	} catch (IOException ioe) {
	    System.out.println("Failed SegmentCreator testing.\n" + ioe);
	}

    }  /* testSegmentCreator() */


    private static void testSegmentCreator(Directory directory, String filename) {

	try {
               DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
               DocumentBuilder builder = factory.newDocumentBuilder();
               Document doc = builder.parse(new File(filename));
               Provider provider = new Provider(doc); 
               SegmentCreator creator = new SegmentCreator(directory);
               creator.createSegment(provider, "zero");
	} catch (Exception e) {
	    System.out.println("Failed SegmentCreator testing.\n" + e);
	}


    }  /* testSegmentCreator(Directory) */



  /** For debugging. */
  public static void main(String[] args) throws Exception {

      testSegmentCreator();

      /*
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder = factory.newDocumentBuilder();
    Document doc = builder.parse(new File(args[0]));
    Provider provider = new Provider(doc); 
    Directory directory = new FSDirectory("index", true);
    SegmentCreator creator = new SegmentCreator(directory);
    creator.createSegment(provider, "zero");
      */

  }  /* close main() */



}  /* class SegmentCreator */
