/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SegmentMerger.java,v 1.1.1.1 2001/06/01 16:50:09 jbeatty Exp $
 */

package net.jxta.search.resolver;

import net.jxta.search.resolver.io.*;
import net.jxta.search.util.*;
import java.io.*;
import java.util.*;

class SegmentMerger {


  private Directory directory;
  private String segmentName;

  private Vector readers = new Vector();
  
  public SegmentMerger(Directory directory, String segmentName) {
    this.directory = directory;
    this.segmentName = segmentName;
  }

  public void add(SegmentReader reader) {
    readers.addElement(reader);
  }

  public void merge() throws IOException {
    mergeClauseCounts();
    mergeProviderIds();
    mergePredicateNumbers();

    mergeProviders();

    mergePatterns();

    for (int i = 0; i < readers.size(); i++) {
      SegmentReader reader = (SegmentReader)readers.elementAt(i);
      reader.close();
    }
  }

  private void mergeClauseCounts() throws IOException {
    int[][] clauseCounts = new int[readers.size()][];
    for (int i = 0; i < readers.size(); i++) {
      SegmentReader reader = (SegmentReader)readers.elementAt(i);
      clauseCounts[i] = reader.getClauseCounts();
    }
    writeVIntFile(clauseCounts, segmentName+".pcc", false);
  }

  private void mergeProviderIds() throws IOException {
    int[][] providerIds = new int[readers.size()][];
    int base = 0;
    for (int i = 0; i < readers.size(); i++) {
      SegmentReader reader = (SegmentReader)readers.elementAt(i);
      int[] ids = reader.getProviderIds();
      IntVector.add(ids, base);
      providerIds[i] = ids;
      base += reader.getMaxProviderId();
    }
    writeVIntFile(providerIds, segmentName+".ppi", true);
  }

  private void mergePredicateNumbers() throws IOException {
    int[][] predicateNumbers = new int[readers.size()][];
    for (int i = 0; i < readers.size(); i++) {
      SegmentReader reader = (SegmentReader)readers.elementAt(i);
      predicateNumbers[i] = reader.getPredicateNumbers();
    }
    writeVIntFile(predicateNumbers, segmentName+".prn", false);
  }

  private void writeVIntFile(int[][] ints, String fileName,
			     boolean deltaEncode) throws IOException {
    int length = 0;
    for (int i = 0; i < ints.length; i++)
      length += ints[i].length;
    int[] result = new int[length];
    int base = 0;
    for (int i = 0; i < ints.length; i++) {
      int[] sub = ints[i];
      System.arraycopy(sub, 0, result, base, sub.length);
      base += sub.length;
    }
    directory.writeVIntFile(result, length, fileName, deltaEncode);
  }


  private void mergeProviders() throws IOException {
    LongVector providerPositions = new LongVector();
    Output out = directory.createFile(segmentName+".prd");
    try {
      for (int i = 0; i < readers.size(); i++) {
	SegmentReader reader = (SegmentReader)readers.elementAt(i);
	long[] positions = reader.getProviderPositions();
	Input in = reader.getProviderInput();
	for (int j = 0; j < positions.length; j++) {
	  providerPositions.add(out.getFilePointer());
	  synchronized (in) {
	    in.seek(positions[j]);
	    out.writeString(in.readString());
	  }
	}
      }
    } finally {
      out.close();
    }
    directory.writeVLongFile(providerPositions.array(),
			     providerPositions.length(),
			     segmentName+".prx", true);

  }

  private Output patternClausesOutput;
  private StringLongWriter patternToPosition;

  private final void mergePatterns() throws IOException {
    try {
      patternClausesOutput = directory.createFile(segmentName+".ptd");
      patternToPosition
	= new StringLongWriter(directory, segmentName+".pt", true);
      
      mergePatternsInternal();
      
    } finally {
      if (patternClausesOutput != null)	patternClausesOutput.close();
      if (patternToPosition != null) 	patternToPosition.close();
    }
  }

  private final void mergePatternsInternal() throws IOException {
    PatternQueue queue = new PatternQueue(readers.size());
    int base = 0;
    for (int i = 0; i < readers.size(); i++) {
      SegmentReader reader = (SegmentReader)readers.elementAt(i);
      Patterns patterns = new Patterns(base, reader);
      base += reader.getMaxPredicateId();
      if (patterns.next())
	queue.put(patterns);			  // initialize queue
    }

    Patterns[] match = new Patterns[readers.size()];
    
    while (queue.size() > 0) {
      int matchSize = 0;			  // pop matching patterns
      match[matchSize++] = (Patterns)queue.pop();
      String pattern = match[0].pattern;
      Patterns top = (Patterns)queue.top();
      
      while (top != null && pattern.compareTo(top.pattern) == 0) {
	match[matchSize++] = (Patterns)queue.pop();
	top = (Patterns)queue.top();
      }

      mergePattern(match, matchSize);		  // add new PatternInfo
      
      while (matchSize > 0) {
	Patterns patterns = match[--matchSize];
	if (patterns.next())
	  queue.put(patterns);			  // restore queue
      }
    }
  }

  IntVector predicateIds = new IntVector();
  IntVector clauseIds = new IntVector();

  private final void mergePattern(Patterns[] match, int matchSize)
    throws IOException {
    patternToPosition.add(match[0].pattern,
			  patternClausesOutput.getFilePointer());
    
    int base = 0;
    for (int i = 0; i < matchSize; i++) {
      Patterns patterns = match[i];
      Input in = patterns.patternClausesInput;
      int length = in.readVInt();
      int nextBase = base + length;
      predicateIds.setLength(nextBase);
      in.readVInts(predicateIds.array(), base, length, true);
      IntVector.add(predicateIds.array(), base, length, patterns.base);
      clauseIds.setLength(nextBase);
      in.readVInts(clauseIds.array(), base, length, false);
      base = nextBase;
    }

    patternClausesOutput.writeVInt(base);
    patternClausesOutput.writeVInts(predicateIds.array(), 0, base, true);
    patternClausesOutput.writeVInts(clauseIds.array(), 0, base, false);
  }
  

  private final class PatternQueue extends PriorityQueue {
    PatternQueue(int size) {
      initialize(size);
    }

    protected final boolean lessThan(Object a, Object b) {
      Patterns patternsA = (Patterns)a;
      Patterns patternsB = (Patterns)b;
      int comparison = patternsA.pattern.compareTo(patternsB.pattern);
      if (comparison == 0)
	return patternsA.base < patternsB.base; 
      else
	return comparison < 0;
    }
  }

  private final class Patterns {
    String pattern;
    int base;
    StringLongReader patternToPosition;
    int offset;
    Input patternClausesInput;

    public Patterns(int base, SegmentReader reader) throws IOException {
      this.base = base;
      this.patternToPosition = reader.patternToPosition;
      this.patternClausesInput = reader.getPatternClausesInput();
    }

    final boolean next() throws IOException {
      if (offset < patternToPosition.length()) {
	pattern = patternToPosition.getStrings()[offset++];
	return true;
      } else {
	pattern = null;
	return false;
      }
    }
  }
 

    private static void testSegmentMerger() {

        try {
            Directory directory = new FSDirectory("index", false);
  	    testSegmentMerger(directory);
	} catch (IOException ioe) {
	    System.out.println("Failed SegmentMErger testing.\n" + ioe);
	    return;
	}



    }  /* close testSegmentMerger() */

    private static void testSegmentMerger(Directory directory) {

	try {
            SegmentMerger merger = new SegmentMerger(directory, "merged");
            merger.add(new SegmentReader(directory, "zero"));
            merger.add(new SegmentReader(directory, "zero"));
            merger.merge();
            SegmentReader merged = new SegmentReader(directory, "merged");
            merged.dump();
	} catch (IOException ioe) {
	    System.out.println("Failed SegmentMerger testing.\n" + ioe);
	}

    }  /* testSegmentMerger(Directory) */


  public static void main(String[] args) throws Exception {

      testSegmentMerger();

      /*
    Directory directory = new FSDirectory("index", false);
    SegmentMerger merger = new SegmentMerger(directory, "merged");
    merger.add(new SegmentReader(directory, "zero"));
    merger.add(new SegmentReader(directory, "zero"));
    merger.merge();
    SegmentReader merged = new SegmentReader(directory, "merged");
    merged.dump();
      */

  }  /* close main() */


}  /* class SegmentMerger */
