/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SegmentReader.java,v 1.1.1.1 2001/06/01 16:50:10 jbeatty Exp $
 */

package net.jxta.search.resolver;

import net.jxta.search.resolver.io.*;
import java.io.*;
import java.util.*;

class SegmentReader implements IndexReader {
  Directory directory;
  private String segmentName;
  
  private int[] clauseCounts;
  private int[] providerIds;
  private int[] predicateNumbers;

  StringLongReader patternToPosition;

  private Input patternClausesInput;

  private long[] providerPositions;
  private Input providerInput;

  public SegmentReader(Directory directory, String segmentName)
      throws IOException {
    this.directory = directory;
    this.segmentName = segmentName;

    patternToPosition
      = new StringLongReader(directory, segmentName+".pt", true);
  }

  int[] getClauseCounts() throws IOException {
    if (clauseCounts == null)
      clauseCounts = directory.readVIntFile(segmentName+".pcc", false);
    return clauseCounts;
  }

  int[] getProviderIds() throws IOException {
    if (providerIds == null)
      providerIds = directory.readVIntFile(segmentName+".ppi", true);
    return providerIds;
  }

  int[] getPredicateNumbers() throws IOException {
    if (predicateNumbers == null)
      predicateNumbers = directory.readVIntFile(segmentName+".prn", false);
    return predicateNumbers;
  }

  Input getPatternClausesInput() throws IOException {
    if (patternClausesInput == null)
      patternClausesInput = directory.openFile(segmentName+".ptd");
    return patternClausesInput;
  }


  long[] getProviderPositions() throws IOException {
    if (providerPositions == null)
      providerPositions = directory.readVLongFile(segmentName+".prx", true);
    return providerPositions;
  }

  Input getProviderInput() throws IOException {
    if (providerInput == null)
      providerInput = directory.openFile(segmentName+".prd");
    return providerInput;
  }

  public int getMaxPredicateId() throws IOException {
    return getClauseCounts().length;
  }

  public int getMaxProviderId() throws IOException {
    return getProviderPositions().length;
  }

  /// core search methods, used to compute matches

  public PatternClauses getPatternClauses(String pattern) throws IOException {
    long position = patternToPosition.get(pattern);
    if (position < 0) {
      return new PatternClauses(new int[0], new int[0]);
    }

    Input in = getPatternClausesInput();
    synchronized (in) {
      in.seek(position);
      int length = in.readVInt();
      int[] predicateIds = new int[length];
      in.readVInts(predicateIds, 0, length, true);
      int[] clauseIds = new int[length];
      in.readVInts(clauseIds, 0, length, false);
      return new PatternClauses(predicateIds, clauseIds);
    }
  }

  public int getProviderId(int predicateId) throws IOException {
    return getProviderIds()[predicateId];
  }

  public int getClauseCount(int predicateId) throws IOException {
    return getClauseCounts()[predicateId];
  }

  // methods used to display matches

  /** Returns the named provider. */
  public Provider getProvider(int providerId) throws IOException {
    Input in = getProviderInput();
    synchronized (in) { 
      in.seek(getProviderPositions()[providerId]);
      try {
	String providerString = in.readString();
	return new Provider(providerString);
      } catch (org.xml.sax.SAXException e) {
	throw new IOException(e.toString());
      }
    }
  }

  /** Returns the named predicate. */
  public Predicate getPredicate(int predicateId) throws IOException {
    ArrayList predicates = getProvider(getProviderId(predicateId)).getPredicates();
    return (Predicate)predicates.get(getPredicateNumber(predicateId));
  }

  private int getPredicateNumber(int predicateId) throws IOException {
    return getPredicateNumbers()[predicateId];
  }

  final Vector files() throws IOException {
    Vector files = new Vector(16);
    files.addElement(segmentName + ".ptx");
    files.addElement(segmentName + ".pts");
    files.addElement(segmentName + ".pcc");
    files.addElement(segmentName + ".ppi");
    files.addElement(segmentName + ".prn");
    files.addElement(segmentName + ".ptd");
    files.addElement(segmentName + ".prx");
    files.addElement(segmentName + ".prd");
    return files;
  }

  public void close() throws IOException {
    if (patternClausesInput != null)
      patternClausesInput.close();
    if (providerInput != null)
      providerInput.close();
  }

  /** For debugging.  Displays the contents of this index. */
  String dump() throws IOException {
    StringBuffer buffer = new StringBuffer();
    for (int i = 0; i < patternToPosition.length(); i++) {
      String pattern = patternToPosition.getKey(i);
      buffer.append(pattern);
      buffer.append(": \n");
      PatternClauses pc = getPatternClauses(pattern);
      for (int j = 0; j < pc.predicateIds.length; j++) {
	int predicateId = pc.predicateIds[j];
	buffer.append("  ");
	buffer.append(pc.clauseIds[j]);
	buffer.append("/");
	buffer.append(getClauseCount(predicateId));
	buffer.append(" ");
	buffer.append(getProvider(getProviderId(predicateId)).queryServer);
	buffer.append("\n");
      }
    }
    return buffer.toString();
  }


    /** 
     * TODO: Implement the indexing so that this test method actually 
     * has something to work against; it currently fails because there
     * is no index actually on disk right now.
     */
    private static void test() throws Exception {

        String [] args = {"foo", "bar", "baz"};
        test(args); 
    } 


    private static void test(String [] args) throws Exception {

    SegmentReader reader
      = new SegmentReader(new FSDirectory("testindex", false), args[0]);
    System.out.println(reader.dump());

    }


  public static void main(String[] args) throws Exception {
 
      test();

  }

}
