/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: FSDirectory.java,v 1.1.1.1 2001/06/01 16:50:13 jbeatty Exp $
 */

package net.jxta.search.resolver.io;

import java.io.IOException;
import java.io.File;
import java.io.RandomAccessFile;

/**
 * Straightforward implementation of Directory as a directory of files.
 * @see Directory
 */

final public class FSDirectory extends Directory {
  File directory = null;
  boolean isOpen = false;


  public FSDirectory(String path, boolean create) throws IOException {
    this(new File(path), create);
  }


  public FSDirectory(File path, boolean create) throws IOException {
    directory = path;
    if (!directory.exists() && create)
      directory.mkdir();
    if (!directory.isDirectory())
      throw new IOException(path + " not a directory");

    if (create) {				  // clear old files
      String[] files = directory.list();
      for (int i = 0; i < files.length; i++) {
	File file = new File(directory, files[i]);
	if (!file.delete())
	  throw new IOException("couldn't delete " + files[i]);
      }
    }

    isOpen = true;
  }


  /** Returns an array of strings, one for each file in the directory. */
  public final String[] list() throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");
    return directory.list();
  }
       
  /** Returns true iff a file with the given name exists. */
  public final boolean fileExists(String name) throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");
    File file = new File(directory, name);
    return file.exists();
  }
       
  /** Returns the time the named file was last modified. */
  public final long fileModified(String name) throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");
    File file = new File(directory, name);
    return file.lastModified();
  }
       
  /** Returns the time the named file was last modified. */
  public static final long fileModified(File directory, String name)
       throws IOException {
    File file = new File(directory, name);
    return file.lastModified();
  }

  /** Returns the length in bytes of a file in the directory. */
  public final long fileLength(String name) throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");

    File file = new File(directory, name);
    return file.length();
  }

  /** Removes an existing file in the directory. */
  public final void deleteFile(String name) throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");

    File file = new File(directory, name);
    if (!file.delete())
      throw new IOException("couldn't delete " + name);
  }

  /** Renames an existing file in the directory. */
  public final void renameFile(String from, String to) throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");

    File old = new File(directory, from);
    File nu = new File(directory, to);

    /* This is not atomic.  If the program crashes between the call to
       delete() and the call to renameTo() then we're screwed, but I've
       been unable to figure out how else to do this... */

    if (nu.exists())
      if (!nu.delete())
	throw new IOException("couldn't delete " + to);

    if (!old.renameTo(nu))
      throw new IOException("couldn't rename " + from + " to " + to);
  }

  /** Creates a new, empty file in the directory with the given name.
      Returns a stream writing this file. */
  public final Output createFile(String name) throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");

    return new FSOutput(new File(directory, name));
  }

  /** Returns a stream reading an existing file. */
  public final Input openFile(String name) throws IOException {
    if (!isOpen)
      throw new IOException("store is not open");

    return new FSInput(new File(directory, name));
  }

  /** Closes the store to future operations. */
  public final void close() throws IOException {
    if (!isOpen)
      throw new IOException("store is already closed");
    isOpen = false;
  }


    private static void testFSDirectory() {

	System.out.println("FSDirectory testing not yet implemented");

	//Also test FSInput, FSOutput

    }  /* close FSDirectory() */


    public static void main(String [] args) {

	testFSDirectory();

    }  /* close main() */



} /* class FSDirectory */



final class FSInput extends Input {
  RandomAccessFile file = null;

  public FSInput(File path) throws IOException {
    file = new RandomAccessFile(path, "r");
    length = file.length();
  }

  /** Input methods */
  public final void readInternal(byte[] b, int offset, int len)
       throws IOException {
    if (file.read(b, offset, len) != len)
      throw new IOException("refill past EOF");
  }
  public final void close() throws IOException {
    file.close();
  }

  /** Random-access methods */
  public final void seekInternal(long pos) throws IOException {
    file.seek(pos);
  }

  protected final void finalize() throws IOException {
    close();					  // close the file 
  }


}  /* class FSInput */


final class FSOutput extends Output {
  RandomAccessFile file = null;

  public FSOutput(File path) throws IOException {
    if (path.isFile())
      throw new IOException(path + " already exists");
    file = new RandomAccessFile(path, "rw");
  }

  /** output methods: */
  public final void flushBuffer(byte[] b, int size) throws IOException {
    file.write(b, 0, size);
  }
  public final void close() throws IOException {
    super.close();
    file.close();
  }

  /** Random-access methods */
  public final void seek(long pos) throws IOException {
    super.seek(pos);
    file.seek(pos);
  }
  public final long length() throws IOException {
    return file.length();
  }

  protected final void finalize() throws IOException {
    file.close();				  // close the file 
  }

}  /* class FSOutput */
