/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: JxtaRouteConnectionService.java,v 1.1.1.1 2001/06/01 16:50:16 jbeatty Exp $
 */
package net.jxta.search.router;

import java.io.*;
import java.net.*;
import java.util.*;

import net.jxta.search.peer.Request;
import net.jxta.search.peer.ProviderInfo;
import net.jxta.search.util.LogService;
import net.jxta.search.util.MemoryStream;
import net.jxta.search.util.XmlParser;
import net.jxta.search.util.UuidService;
import net.jxta.search.util.JxtaPipeUtil;

import javax.servlet.http.Cookie;
import javax.xml.parsers.DocumentBuilder;

import net.jxta.id.IDFactory;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.JxtaException;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.PipeID;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.document.AdvertisementFactory;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.MimeMediaType;
import net.jxta.document.Element;
import net.jxta.discovery.Discovery;
import net.jxta.endpoint.Message;


class JxtaRouteConnectionService implements Router.ConnectionService {


    LogService            logService;
    UuidService           uuidService;
    XmlParser             parser;
    
    PeerGroup               peerGroup;
    PipeAdvertisement       inPipeAdv;
    InputPipe               inPipe;
    ResponseMessageReceiver responseMessageReceiver;
    Hashtable               routeTable;

    static final int PIPE_RESOLVE_TIMEOUT = 10000;
    static final String RETURNPIPE_ADV_FILENAME = "returnpipeadv.xml";    

    class RouteEntry {
	String        uuid;
	ProviderInfo  provider;
	Router.Merger merger;
	int           maxHits;
	long          startTime;
	
	RouteEntry (String        uuid, 
		    ProviderInfo  provider, 
		    Router.Merger merger,
		    int           maxHits) {
	    this.uuid      = uuid;
	    this.provider  = provider;
	    this.merger    = merger;
	    this.maxHits   = maxHits;
	    this.startTime = System.currentTimeMillis ();
	}

	/*
	boolean expired (long now) {
	    return (startTime + merger.getTimeout () < now);
	}
	*/
    }

    /* taken care of by JxtaRouteConnection.stop 
    void clean () {
	long        now   = System.currentTimeMillis ();
	Enumeration e     = routeTable.elements ();
	RouteEntry  entry;
	
	while (e.hasMoreElements ()) {

	    entry = (RouteEntry) e.nextElement ();

	    if (entry.expired (now))
		routeTable.remove (entry.uuid);
	}
    }
    */
    
    JxtaRouteConnectionService (PeerGroup   peerGroup, 
				LogService  logService,
				UuidService uuidService)
	throws PeerGroupException
    {
	this.peerGroup   = peerGroup;
	this.logService  = logService;
	this.uuidService = uuidService;
	this.parser      = new XmlParser ();
	
	try {
	    inPipeAdv = JxtaPipeUtil.createPipeAdv(
                new File( RETURNPIPE_ADV_FILENAME ),
		"Return pipe for InfraSearch network responses",
		peerGroup );

	    peerGroup.getDiscovery().publish (inPipeAdv, Discovery.ADV);

	    inPipe  = peerGroup.getPipe().createInputPipe (inPipeAdv);

	} catch ( JxtaException e ) {
	    
	    throw new PeerGroupException ("Could not make or publish " +
					  "response pipe advertisement.");
	} catch ( IOException e ) {
	    
	    throw new PeerGroupException ("Could not create response pipe.");
	}

	responseMessageReceiver = new ResponseMessageReceiver (inPipe);

	routeTable = new Hashtable ();
	
	new Thread (responseMessageReceiver).start();
    }

    public Router.Connection openConnection (Router.Merger merger,
                                             ProviderInfo  providerInfo,
					     Cookie[]      cookies,
					     Request       request)	
    {
	String     uuid  = uuidService.hexify (uuidService.getUuid ());
	
	RouteEntry entry = new RouteEntry (uuid, 
					   providerInfo, 
					   merger, 
					   request.getMaxHitsPerProvider ());
	routeTable.put (uuid, entry);
	
	return new JxtaRouteConnection (merger, providerInfo, cookies, 
					request,
					uuid, 
					entry);
    }


    class ResponseMessageReceiver implements Runnable {
	
	InputPipe responsePipe;
	boolean   running=true;
	
	ResponseMessageReceiver (InputPipe responsePipe) {

	    this.responsePipe = responsePipe;
	}

	synchronized void stop () {
	    this.running = false;
	}

	synchronized boolean isRunning () {
	    return running;
	}

	public void run () {

	    Message      msg;
	    String       uuid;
	    InputStream  queryResponse;
	    RouteEntry   entry;
	    byte[]       buf;
	    UuidExtractor uuidExtractor;
	    XmlParser     parser;

	    buf = new byte[8192];
	    uuidExtractor = new UuidExtractor ();
	    parser        = new XmlParser ();

	    try {		
		while ( isRunning () ) {
		    msg = inPipe.poll ( 1000 );
		    
		    if (! isRunning ())
			break;
		    if (msg == null)
			continue;
		    
		    queryResponse = msg.pop ("responses"); 
		    MemoryStream mem = new MemoryStream ();
		    int r = 0;
		    while ((r = queryResponse.read (buf)) > 0)
			mem.write (buf, 0, r);
		    


			try {
			    parser.parse (mem.getInputStream (), 
					  buf, 
					  uuidExtractor);
			} catch (XmlParser.Exception e) {
				if (e.getMessage () != null) {

					// log parse exeption
					e.printStackTrace ();

					continue;
				}
			}

		    entry = (RouteEntry) routeTable.get (uuidExtractor.uuid);
		    System.out.println ("Corresponding entry is " + entry);
		    
		    if (entry != null) {
			ProviderResponse res = 
			    new ProviderResponse (entry.provider);
			MemoryStream out = new MemoryStream ();
			
			try {
			    
			    // queryResponse = msg.pop ("responses"); 
			    res.parse (null, mem.getInputStream (),
				       out, parser, buf);
			    entry.merger.merge (res);
			} catch (XmlParser.Exception e) {
			    e.printStackTrace();
			    logService.log (logService.WARNING,
					    entry.merger.uuid, 
					    "Unable to parse response from " +
					    entry.provider.getQueryURL ());
			}
		    }
		} 
	    } catch (IOException e) {
		e.printStackTrace ();
		stop ();
	    }
	}


    }


    /**
     * Correct and quick way of extracting the uuid out of the
     * responses message.  However, better code would handle 
     * it in ProviderResponse - where the uuid would be extracted,
     * the ProviderInfo would be looked up through some external
     * interface, and parsing would continue.  That's what should
     * be done in the next iteration.
     */
    
    static class UuidExtractor implements XmlParser.ParserCallback {
	String uuid;
	
	
	static final byte[] INSTRUCTION_TAG = "?".getBytes ();
	static final byte[] RESPONSES_TAG   = "RESPONSES".getBytes ();
	static final byte[] QUERY_UUID_ATTR = "QUERY-UUID".getBytes ();
	
	public void startTag (byte[] chars, int start, int len) 
	    throws XmlParser.Exception
	{
	    if (XmlParser.startsWith (INSTRUCTION_TAG, 
				      chars, start+1, len-1)) {
				// that's fine
	    } else if (XmlParser.startsWith (RESPONSES_TAG, 
					     chars, start+1, len-1)) {
		uuid = XmlParser.getAttributeValue (QUERY_UUID_ATTR,
						    chars, start, len);
		System.out.println( "JxtaRouteConnectionService.UuidExtractor: uuid=." + uuid +".");

		if (uuid == null)
		    throw new XmlParser.Exception ("<responses> tag must " +
						   "have a query-uuid " +
						   "attribute.");
		else
		    throw new XmlParser.Exception (null);
	    } else {
		throw new XmlParser.Exception ("Unexpected top-level tag " +
					       new String (chars, start, len) +
					       ", expected <responses>");
	    }
	}
	
	public void endTag (byte[] chars, int start, int len) 
	    throws XmlParser.Exception
	{
	    //throw new XmlParser.Exception ("Unexpected close tag.");
	}
	
	public void chars (byte[] chars, int start, int len) 
	    throws XmlParser.Exception
	{
	    //throw new XmlParser.Exception ("Unexpected characters.");
	}
    }
    
   
/**
 * This class implements the RouteConnection interface with
 * JXTA as the connection transport and XML as the encoding.
 */

class JxtaRouteConnection implements Router.Connection {

    Router.Merger         merger;
    ProviderInfo          providerInfo;
    Cookie[]              cookies;
    Request               request;
    String                uuid;
    MemoryStream          out;
    long                  time;
    int                   hits;
    RouteEntry            routeEntry;

    volatile int          state;

    
    JxtaRouteConnection (Router.Merger merger, 
                         ProviderInfo  providerInfo,
			 Cookie[]      cookies,
			 Request       request,
			 String        uuid,
			 RouteEntry    routeEntry)
    {
	this.merger       = merger;
        this.providerInfo = providerInfo;
	this.cookies      = cookies;
	this.request      = request;
	this.uuid         = uuid;
	this.routeEntry   = routeEntry;
	this.state        = INIT;
    }



    public URL getURL () { return providerInfo.getQueryURL (); }
	
    public int getState () { return state; }
	
    public String getId () { return providerInfo.getId (); }
	
    public void getResponse (OutputStream out) throws IOException { 
	this.out.writeToStream (out); 
    }
	
    public int getResponseLength () { 
	return out.getLength (); 
    }
    
    public long getResponseTime () { 
	return time; 
    }

    public void stop () {
	state = DONE;

	routeTable.remove (routeEntry.uuid);
    }


	
    public void run () {
	PipeAdvertisement outPipeAdv;
	InputPipe         inPipe;
	OutputPipe        outPipe;
	Message           msg;
	InputStream       in;
	long              start = System.currentTimeMillis ();

	// 	merger.start (this);

	
	    
	if (state == DONE) {
	    merger.fail (providerInfo,
			 "Stopped before we had a chance to run.");
	    return;
	}

	try {
	    state = CONNECTING;

            URL  url = this.providerInfo.getQueryURL ();
            Pipe pipeService = peerGroup.getPipe();


	    outPipeAdv = (PipeAdvertisement)
		AdvertisementFactory.newAdvertisement
		(PipeAdvertisement.getAdvertisementType ());
	    
	    outPipeAdv.setPipeID ((PipeID) IDFactory.fromURL (url));
	    outPipeAdv.setName (url.getPath ());
	    
	    	    
	    logService.log (logService.DEBUG, 
			    merger.uuid, 
			    "Connecting to " + url);
	    
	    outPipe = pipeService.createOutputPipe (outPipeAdv, 
						    Pipe.NonBlocking, 
						    PIPE_RESOLVE_TIMEOUT );
	    
            	
	    if (state == DONE) {
		merger.fail (providerInfo, "timed out");
		return;
	    } else {
		logService.log (logService.DEBUG, 
				merger.uuid, 
				"Connected to " + url);
	    }

            // construct message
            // 
	    msg = pipeService.createMessage ();

	    /* make sure the cookies propagate to the provider nodes */
	    if (cookies != null && cookies.length > 0) {
		StructuredDocument doc =
		    StructuredDocumentFactory.newStructuredDocument
		    (new MimeMediaType ("text", "xml"),
		     "infrasearch:cookies");
		
		for (int i = 0;  i < cookies.length;  i++) {
		    if (cookies[i].getName ().equals (this.providerInfo.getId ())) {
			Element e = doc.createElement
			    ("cookie",
			     net.jxta.search.util.TextTools.base64decode
			     (cookies[i].getValue ().getBytes ()).getBytes ());
			doc.appendChild (e);
		    }
		}
	    
		msg.push ("cookies", doc.getStream ());
	    }
	    
	    // push the message type into the message. there are two
            // messagetypes: 'query' and 'regrequest'
            msg.push( "messagetype", new ByteArrayInputStream(
                "query".getBytes() ));

            msg.push ("request", request.getInputStream (uuid));
	    
	    msg.push ("responsePipe", inPipeAdv.getDocument (new MimeMediaType
		("text", "xml")).getStream ());

	    // send message
            //
            outPipe.send (msg);
	    
	    state = DONE;
		
	} catch (Exception e) {
	    System.out.println ("Exception while getting response from " +
				providerInfo.getQueryURL ());
	    e.printStackTrace ();

	    merger.fail (providerInfo, e.getMessage ());

	} 
    }

}

}
