/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Router.java,v 1.1.1.1 2001/06/01 16:50:18 jbeatty Exp $
 */
package net.jxta.search.router;

import java.io.*;
import java.net.*;
import java.util.*;

import org.w3c.dom.*;
import net.jxta.search.util.*;
import javax.servlet.http.Cookie;

import net.jxta.search.peer.ProviderInfo;
import net.jxta.search.peer.Request;

import net.jxta.search.util.HttpConnection;
import net.jxta.search.util.HttpConnectionService;

import net.jxta.pipe.Pipe;
import net.jxta.peergroup.PeerGroup;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.ServiceNotFoundException;




public class Router {

    LogService       logService;
    ThreadService    threadService;
    SchedulerService schedulerService;
    UuidService      uuidService;
    
    PeerGroup        peerGroup;

    HashMap          routeStats;
    HashMap          mergers;
    HashMap          connectionServices;
    
    Manager          manager;
    
    
    
    static int       DELETE_ROUTE_TIMEOUT = 5 * 60 * 1000;

    class Lock {
	Vector threads = new Vector ();
	
	
	synchronized void enter () {
	    System.out.println ("Lock.enter: " + Thread.currentThread ());
	    
	    threads.addElement (Thread.currentThread ());
	}

	synchronized void leave () {
	    System.out.println ("Lock.leave: " + Thread.currentThread ());

	    threads.removeElement (Thread.currentThread ());
	}

	void printThreads () {
	    Enumeration e = threads.elements ();
	    Thread      t;
	    
	    if (e.hasMoreElements ()) {
		System.out.println ("The following threads still own the " +
				    "lock: (if more than one, major bug!");
		
		while (e.hasMoreElements ()) {
		    t = (Thread) e.nextElement ();
		    System.out.println ("    " + t);
		    new Throwable ().printStackTrace ();
		}
	    } else {
		System.out.println ("No threads own the lock.");
	    }
	}
    }


    public static interface Manager {
	
	/**
	 * Whenever the routes become dead (counted as after some
	 * iterations of the exponential back-off algorithm), this
	 * is the callback that should communicate to the resolver
	 * that this route is no longer valid and shouldn't be included
	 * in resolver's responses.
	 */
	void deleteRoutes (URL[] routes, String[] ids, String[] reasons);
    }
    
    public static interface Callback {
	
	/**
	 * This is where your transport code can pre-declare number of
	 * responses or whatever.
	 */
	void setProviderNumbers (int totalProviders, int queriedProviders)
	    throws IOException;
	
	/**
	 * Process the response (probably writing it out to the original
	 * requestor, pre-declaring the length of the response, maybe.
	 */
	void process (ProviderResponse response)
	    throws IOException;
	
	/**
	 * Called according to the heuristics calculated using the parameters
	 * passed to Router.fetch() 
	 */
	void flush () 
	    throws IOException;
    }


    public class RouteStats {
	URL  url;
	String id;
	long last;
	long next;
	int  failed;
  	int  queries;
        long averageTime;

	RouteStats (URL url, String id) {
	    this.url = url;
	    this.id  = id;
	}
	
	void succeeded (long time) {
  	    averageTime = (queries * averageTime + time) / (queries + 1); 
	    queries++;
	    if (failed > 0)
		failed--;
	}

	void failed () {

	    failed++;
	    
	    if (failed < 20) {
		Random r;
		
		last = System.currentTimeMillis ();
		
		r = new Random (last);

		next = last + r.nextInt ((0x2 << failed) & 0x00FFFFFF); 

		
	    }
	}

	boolean shouldTry () {
	    long now = System.currentTimeMillis ();
	    
	    return (failed < 20 && next < now);
	    
	}

	boolean hasExpired () {
	    return failed == 20;
	}

	public int getQueryCount () {
	    return queries;
	}

	public int getFailedAttemptCount () {
	    return failed;
	}

	public long getLastAttemptTime () {
	    return last;
	}

        public long getAverageResponseTime () {
	    return averageTime;
        }
	public URL getURL () {
	    return url;
	}
    }


    public interface Connection extends Runnable {

	public static final int INIT       = 0;
	public static final int CONNECTING = 1;
	public static final int READING    = 2;
	public static final int DONE       = 3;
	
	public int    getState ();
	
	public void   stop ();
    }

    public interface ConnectionService {
	
	Connection openConnection (Merger       merger,
                                   ProviderInfo providerInfo,
				   Cookie[]     cookies,
				   Request      request);
    }


    
    
    public class Merger {
	ArrayList    fetchers;
	Callback     cb;
	int          i;
	int          max;
	int          soFar;
	int          flushAfter;
	long         timeout;
	String       uuid;
	Lock         lock;

	volatile boolean done = false;
	
	Object           pendingAction;
	
	ResponsePriorityQueue responseQueue;
	

	Merger (ArrayList        fetchers,
		Callback         cb, 
		int              max, 
		int              flushAfter,
		long             timeout,
		String           uuid)
	{
	    this.fetchers   = fetchers;
	    this.cb         = cb;
	    this.max        = max;
	    this.flushAfter = flushAfter;
	    this.timeout    = timeout;
	    this.uuid       = uuid;
	    this.responseQueue = new ResponsePriorityQueue (max);
	    this.lock       = new Lock ();
	    
	    this.pendingAction = schedulerService.scheduleAction
		( new SchedulerService.Action () {
			
			public void perform (SchedulerService ss) {
			    
			    logService.log (logService.INFO,
					    Merger.this.uuid, 
					    "Merger alarm expired");
			    lock.printThreads ();
			    /* wake up thread waiting for completion */
			    expire ();
			    			    
			}
			
		    }, timeout );
	    
	    logService.log (logService.DEBUG,
			    Merger.this.uuid, 
			    "Sheduler alarm set for " + timeout + 
			    " in advance");

	}

	public long getTimeout () {
	    return timeout;
	}

	public String getUuid () {
	    return uuid;
	}

	synchronized void start (Connection con) {
	    /*
	    try {
		lock.enter ();
	    */
	    i++;
	    /*
	} finally {
		lock.leave ();
	    }
	    */
	}



	synchronized void merge (ProviderResponse response) {
	    /*
	    try {
		lock.enter ();
	    */
	    ProviderInfo info = response.getProviderInfo ();
	    
	    if (done) {
		
		fail (info, "Merge call after done.");
		
	    } else {
		
		try {
		    responseQueue.put (response);
		    soFar ++;
		    
		    if (flushAfter != -1 && (soFar % flushAfter) == 0) 
			flush ();
		    
		    logService.log (logService.INFO, uuid,
				    "Merger succeeded " + 
				    info.getId () +
				    " " + 
				    info.getQueryURL ());
		    
		    succeeded (info.getQueryURL (), response.getTime ());
		    
		    if (--i == 0) {
			Router.this.mergers.remove (this.uuid);
			notify ();
		    }
				    
		} catch (IOException e) {
		    
		    notify ();
		}
		
	    }
	    /*
	    } finally {
		lock.leave ();
	    }
	    */

	}
	
	synchronized void fail (ProviderInfo info, String reason) {
	    /*
	    try {
		lock.enter ();
	    */
	    logService.log (logService.INFO, uuid, 
			    "Merger failed " + info.getId () + 
			    " " + info.getQueryURL () + " \"" + reason + "\"");
		    
	    failed (info.getQueryURL (), info.getId ());
	    
	    if (--i == 0 || --max == 0) {
		
		Router.this.mergers.remove (this.uuid);
		
		notify ();
		
	    }
	    /*
	    } finally {
		lock.leave ();
	    }
	    */
	}

	synchronized void waitForCompletion () {
	    
	    try {
		
		/* all threads could've completed by the time we launched
		   them all and got to here.
		   also make sure we get ouf of the while loop if the
		   scheduler alarm has expired.
		*/
			
			while (pendingAction != null && i != 0 && max != 0) {
				// System.out.println ("Going into wait.");
				wait ();
			}
			/*
			System.out.println ("pendingAction = " + pendingAction +
								" i = " + i +
								" max = " + max);	
			*/
		// if we got out of wait for a reason other than our
		// own timeout has expired.
		if (pendingAction != null) 
		    schedulerService.cancelAction (pendingAction);
		

		try {
		    flush ();
		} catch (IOException e) {
		    /* not much of a difference now? */
		}

		terminate ();
	    } catch (InterruptedException e) {
		// No code intended
	    } 
	}

	synchronized void expire () {
	    /*
	      try {
	       lock.enter (); 
	    */
	    pendingAction = null;
	    notify ();
	    /*
	    } finally {
		lock.leave ();
	    }
	    */
	}
	
	void terminate () {
	    
	    logService.log (logService.INFO, uuid, "Merger: terminate");
	    
	    Iterator   i = fetchers.iterator ();
	    Connection f = null;
	    
	    
	    
	    done = true;
	    
	    while (i.hasNext ()) {
		f = (Connection) i.next ();

		f.stop ();
	    }

	    Router.this.mergers.remove (this.uuid);

	}

	
	void flush () throws IOException {
	    ProviderResponse response;
	    
	    while (responseQueue.top () != null) {
		response = (ProviderResponse) responseQueue.pop ();

		if (cb != null)
		    cb.process (response);
	    }
	    
	    if (cb != null)
		cb.flush ();
	}


	public Iterator getFetchers () {
	    return fetchers.iterator ();
	}
    }


    public Router (Manager       manager, 
		   ThreadService threadService, 
		   PeerGroup     peerGroup,
		   LogService    logService) 
	throws PeerGroupException 
    {

	this.manager          = manager;
	this.threadService    = threadService;
	this.logService       = logService;
        this.peerGroup        = peerGroup;

	this.uuidService      = new UuidService ();
	this.schedulerService = new SchedulerService ();
	this.routeStats       = new HashMap ();
	this.mergers          = new HashMap ();
	this.connectionServices = new HashMap ();
	
	connectionServices.put ("http", 
				new HttpRouteConnectionService (logService));

	if (peerGroup != null)
	    connectionServices.put ("jxta",
				    new JxtaRouteConnectionService
					(peerGroup, logService, uuidService));
	
			
	new Thread (schedulerService).start ();
	
	
	
	schedulerService.scheduleAction (new SchedulerService.Action () {

		public void perform (SchedulerService ss) {
		    Iterator    i         = routeStats.values ().iterator ();
		    Vector      urlVec    = new Vector ();
		    Vector      idVec     = new Vector (); 
		    Vector      reasonVec = new Vector ();
		    URL[]       urls; 
		    String[]    ids;
		    String[]    reasons;

		    while (i.hasNext ()) {
			RouteStats a = (RouteStats) i.next ();
			
			if (a.hasExpired ()) {
			    urlVec.addElement (a.url);
			    idVec.addElement (a.id);
			    reasonVec.addElement ("Timeout");

			    routeStats.remove (a.url);
			}
		    }

		    urls = new URL[urlVec.size ()];
		    ids  = new String[urlVec.size ()];
		    reasons = new String[urlVec.size ()];
		    
		    urlVec.copyInto (urls);
		    idVec.copyInto (ids);
		    reasonVec.copyInto (reasons);
		    
		    if (Router.this.manager != null)
			Router.this.manager.deleteRoutes (urls, 
							  ids, 
							  reasons);
		
		    
		    schedulerService.scheduleAction (this, 
						     DELETE_ROUTE_TIMEOUT);
		}
	    }, DELETE_ROUTE_TIMEOUT);
    }
				
    
    public void addConnectionService (String protocol, ConnectionService cs) {
	connectionServices.put (protocol, cs);
    }

    
    
    public Iterator getMergers () {
	return mergers.values ().iterator ();
    }
    
    
    public Iterator getRouteStats () {
	return routeStats.values ().iterator ();
    }

    public RouteStats getRouteStats (URL url) {
	return (RouteStats) routeStats.get (url);
    }

    
    void succeeded (URL url, long time) {
	RouteStats a = (RouteStats) routeStats.get (url);

	if (a != null)
	    a.succeeded (time);
    }
    
    void failed (URL url, String id) {
	RouteStats a = (RouteStats) routeStats.get (url);

	if (a == null) {
	    a = new RouteStats (url, id);
	    routeStats.put (url, a);
	}

	a.failed ();
    }

    boolean shouldTry (URL url) {
	RouteStats a = (RouteStats) routeStats.get (url);

	return a == null || a.shouldTry ();
    }
    
    public void route (Request        request,
		       ProviderInfo[] providerInfos, 
		       Cookie[]       cookies,
		       Callback       cb)
	throws IOException
    {
	ArrayList fetchers = new ArrayList (providerInfos.length);
	Merger    merger   = new Merger (fetchers, 
					 cb, 
					 request.getMaxHitsPerProvider (),
					 request.getFlushAfterProviders (), 
					 request.getQueryLifetime (),
					 request.getQueryUuid ());
	int       tried    = 0;
	long      start;
	long      end;
	
	
	mergers.put (request.getQueryUuid (), merger);

	int providerCount = Math.min (providerInfos.length, 
				      request.getMaxFanout ());
	
	for (int i = 0;  i < providerCount;  i++) {
	    
	    if (shouldTry (providerInfos[i].getQueryURL ())) {
		ConnectionService cs = (ConnectionService) 
		    connectionServices.get (providerInfos[i].getQueryURL ().getProtocol ());
		Connection c  = 
		    cs.openConnection (merger, 
				       providerInfos[i], 
				       cookies,
				       request);
				       
		try {
			merger.start (c);
		    threadService.getThread (c);
		    tried ++;
		    fetchers.add (c);
		} catch (InterruptedException e) {
		    // No code intended
		}
	    }
	}
	
	if (cb != null)
	    cb.setProviderNumbers (providerInfos.length, tried);

	
	start = System.currentTimeMillis ();
	merger.waitForCompletion ();
	
	end = System.currentTimeMillis ();
	
	
	try {
	    ProviderResponse response;
	    
	    while (merger.responseQueue.top () != null) {
		response = (ProviderResponse) merger.responseQueue.pop ();
		
		if (cb != null)
		    cb.process (response);
	    }
	    
	    if (cb != null)
		cb.flush ();

	} catch (IOException e) {
	    /* if the connection between an app and the router is gone,
	       then just quit.
	    */
	}
    }


    class ResponsePriorityQueue extends PriorityQueue {
	
	
	ResponsePriorityQueue (int numExpectedResponses) {
	    initialize (numExpectedResponses);
	}
	
	public boolean lessThan (Object o1, Object o2) {
	    ProviderResponse f1 = (ProviderResponse) o1;
	    ProviderResponse f2 = (ProviderResponse) o2;
	    
	    return ((f1.getLength () * f1.getTime () ) < 
		    (f2.getLength () * f2.getTime ()));
	}
    }

    
}
