/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: RouterJxtaApp.java,v 1.1.1.1 2001/06/01 16:50:19 jbeatty Exp $
 */
package net.jxta.search.router;

/* jxta imports */
import net.jxta.platform.Application;
import net.jxta.peergroup.PeerGroup;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.pipe.PipeID;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.MimeMediaType;
import net.jxta.discovery.Discovery;
import net.jxta.service.Service;
import net.jxta.endpoint.Message;
import net.jxta.exception.JxtaException;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.ServiceNotFoundException;


/* java imports */
import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.InvocationTargetException;


/* xml imports */
import org.xml.sax.SAXException;


/* peer search api */
import net.jxta.search.peer.Request;
import net.jxta.search.peer.ResolverClient;
import net.jxta.search.peer.ProviderInfo;

/* util imports */
import net.jxta.search.util.ThreadService;
import net.jxta.search.util.JxtaPipeUtil;
import net.jxta.search.util.LogService;
import net.jxta.search.util.MemoryStream;
import net.jxta.search.util.XmlParser;


class RouterJxtaApp implements Application {

    PeerGroup  peerGroup;
    InputPipe  requestPipe;
    
    Router          router;
    ResolverClient  resolverClient;
    ThreadService   threadService;
    LogService      log;
    
    RequestAcceptor requestAcceptor;

    static final int MAX_REQUEST_SIZE=4096;
    static final int MSG_READ_TIMEOUT = 30 * 1000;
    static final String PIPE_ADV_FILENAME = "querypipeadv.xml";
    static final String PIPE_NAME = "Infrasearch Query Pipe";    
    static final MimeMediaType MIME_TYPE = new MimeMediaType( "text/xml" );
    
    public RouterJxtaApp (Router          router, 
			  ResolverClient  resolverClient,
			  LogService      log,
			  ThreadService   threadService)
    {
	this.router = router;
	this.resolverClient = resolverClient;
	this.threadService = threadService;
    }

    public int startApp(String[] args) {

	requestAcceptor = new RequestAcceptor ();
	
	new Thread (requestAcceptor).start ();
	
        return 0;
    }
    
    public void stopApp() {
        requestAcceptor.stop ();

	requestPipe.close ();
    }
    
    public void init (PeerGroup peerGroup, Advertisement serviceAdv)
	throws PeerGroupException
    {
	this.peerGroup = peerGroup;
        
        PipeAdvertisement adv = null;
        try {
            adv = JxtaPipeUtil.createPipeAdv( new File( PIPE_ADV_FILENAME ),
					      PIPE_NAME,
					      peerGroup );
        } catch( Exception e ) {
            throw new PeerGroupException ("Unable to get pipe " +
                                          "advertisement.");
	}

	try {	    
	    peerGroup.getDiscovery().publish (adv, Discovery.ADV);
	} catch (IOException e) {
	    throw new PeerGroupException ("Unable to publich pipe " +
					  "advertisement.");
	}

	try {
	    requestPipe = peerGroup.getPipe().createInputPipe (adv);
	} catch (IOException e) {
	    throw new PeerGroupException ("Unable to create input pipe.");
	}
	
	new Thread (new RequestAcceptor ()).start ();
    }


    class RequestAcceptor implements Runnable {

	boolean running;

	RequestAcceptor () {
	    running = true;
	}

	public synchronized void stop () {
	    running = false;
	}

	public synchronized boolean isRunning () {
	    return running;
	}

	public void run() {
	    Message     msg;
            	    	    
	    try {

		while (isRunning ()) {
		    
		    msg = requestPipe.poll (MSG_READ_TIMEOUT);
		    
		    if (isRunning () && msg != null)
			threadService.getThread (new RequestProcessor (msg));
                }
	    } catch (IOException e) {
		/* xxx does this really mean the pipe is broken and
		   not readable from again? */
		
	    } catch (InterruptedException e) {
		/* interrupted while waiting for a thread */
		
	    }
	}
	
    }


    class RequestProcessor implements Runnable {

        static final int RETURN_PIPE_RESOLVE_TIMEOUT = 10000;
        	
        Message msg;
	
	RequestProcessor (Message msg) {
	    this.msg = msg;
	}
	
	public void run () {

	    
	    try {
		byte[]  reqBuf = new byte[1024];
		Request req    = new Request (msg.pop ("request"), reqBuf);
                
		// get advertisement of pipe on which to send the response
		PipeAdvertisement responsePipeAdv = (PipeAdvertisement) 
		    AdvertisementFactory.newAdvertisement
		        (new MimeMediaType ("text/xml"),
                         msg.pop ("responsePipe") );
		
		OutputPipe responsePipe = 
		    peerGroup.getPipe().createOutputPipe (responsePipeAdv,
							  Pipe.NonBlocking,
							  RETURN_PIPE_RESOLVE_TIMEOUT );

		System.out.println( "RouterJxtaApp: responsePipe==" +
				    responsePipe );

                if ( responsePipe == null ) {
		    log.log( log.WARNING, 
			     req.getQueryUuid (), 
			     "RouterJxtaApp could not " +
			     "resolve output pipe");
                    return;
                }

		System.out.println( "RouterJxtaApp: Calling resolver client");

		ProviderInfo[] providerInfos = resolverClient.resolve (req);
		System.out.println("RouterJxtaApp: resolved peers: "+
				   providerInfos.length);

		//XXX TODO including this statement deadlocks my thread
		//how strange.
		/* log.log( log.INFO, 
		   req.getQueryUuid (), 
		   "RouterJxtaApp: resolver "+
		   "providers: " + providerInfos.size() );*/
		
		System.out.println( "RouterJxtaApp: router.route() called" );
		router.route (req,
			      providerInfos,
			      null, /* cookies array... tbd how we get 
				       them in the jxta context.. blah. */
			      new RouterCallback (req.getQueryUuid (),
						  responsePipe,
						  peerGroup.getPipe ()));
		System.out.println( "RouterJxtaApp: router.route() returned");
	    } catch (java.lang.reflect.InvocationTargetException e) {
		e.printStackTrace();
	    } catch (IOException e) {
		e.printStackTrace();
		/* errr, need to reorganize above code */
	    } catch (XmlParser.Exception e) {
		e.printStackTrace ();
		/* log bad msg */
	    }
	}
    }
    

    /**
     * This callback object is used by the router to signal
     * when new responses are available and when responses
     * should be sent/flushed back to the waiting client.
     */

    class RouterCallback implements Router.Callback {
	
	String       uuid;         // uuid of current response
	OutputPipe   responsePipe; // pipe to send response msgs to
	Pipe         pipeService;  // need to create new msgs
	Message      msg;    // message currently being constructed
	MemoryStream out;    // in-memory stream for message body
	String       header; // xml document start
	String       footer; // xml document end


	RouterCallback (String     uuid, 
			OutputPipe responsePipe, 
			Pipe       pipeService) {
	    System.out.println( "RouterJxtaApp.RouterCallback()" );
	    
	    this.uuid         = uuid;
	    this.responsePipe = responsePipe;
	    this.pipeService  = pipeService;
	    this.header       = 
		"<?xml version='1.0'?>\n" +
		"<responses xmlns='" + Request.DEFAULT_NAMESPACE + "'\n" +
		"           query-uuid='" + uuid + "'>\n";
	    this.footer       =
		"</responses>\n";
	}

	public void setProviderNumbers (int totalProviders,
					int queriedProviders)
	    throws IOException
	{
	    // kinda meaningless now..  should eliminate this callback anyway
	}

	public synchronized void process (ProviderResponse res) 
	    throws IOException
	{
	    System.out.println( "RouterJxtaApp.RouterCallback.process()" );
	    if (msg == null) {

		msg = peerGroup.getPipe ().createMessage ();
		
		out = new MemoryStream ();
		
		out.write (header.getBytes ());
		
	    }
	    
	    res.writeToStream (out);
	}
	
	public synchronized void flush () {
	    System.out.println( "RouterJxtaApp.RouterCallback.flush()" );
	    try {
		out.write (footer.getBytes ());
		
		//log.log (log.DEBUG, uuid, "RouterJxtaApp: sending response");

		// XXX hack Message does not play nice with MemoryStream-
		// produced InputStream
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		transfer( out.getInputStream(), os );
		byte[] buf = os.toByteArray();		   
		msg.push ("responses", new ByteArrayInputStream(buf));

		responsePipe.send (msg);
		
		out.close ();
		msg = null;
	    } catch (IOException e) {
		e.printStackTrace();
		//log.log (uuid, "Cannot send message: ", e);
	    }
	}
    }

    /** moves all bytes from is to os **/
    private void transfer( InputStream is, OutputStream os ) {
        int c;
	byte[] buf = new byte[128];
        try {
            while ( (c=is.read(buf,0,128))!= -1 ) {
                os.write(buf,0,c);
            }
        } catch ( IOException e ) {
            System.err.println( "Problem displaying search payload:" + e );
            e.printStackTrace();
        }
    }
}
