/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: jxtasearch.java,v 1.1.1.1 2001/06/01 16:51:59 jbeatty Exp $
 */
package net.jxta.impl.shell.bin.jxtasearch;

import net.jxta.platform.*;
import net.jxta.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.endpoint.Message;
import net.jxta.protocol.*;
import net.jxta.document.MimeMediaType;
import net.jxta.document.MimeMediaType;
import net.jxta.exception.*;
import net.jxta.impl.id.*;

import net.jxta.impl.shell.*;

import net.jxta.search.peer.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.BufferedReader;
import java.io.Reader;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.lang.reflect.InvocationTargetException;

import org.w3c.dom.*;
import javax.xml.parsers.*;

/**
 * The infrasearch shell command for searching the infrasearch network
**/
public class jxtasearch extends ShellApp {

    /** we will try to resolve the return pipe for 10 seconds
        before we give up **/    
    private static final int QUERY_PIPE_RESOLVE_TIMEOUT = 10000;

    private static final MimeMediaType MIME_TYPE =
        new MimeMediaType( "text/xml" );

    private static final String DEFAULT_QUERYSPACE =
	"http://search.jxta.org/text";

    private static final int DEFAULT_LIFETIME = 5000;
    private static final int DEFAULT_FANOUT = 100;
    private static final int DEFAULT_RESULTS_PER_PROV = 5;

    private static QueryResults queryResults = new QueryResults();
    private static JxtaRouterClient routerClient;
    private PipeAdvertisement currentQueryPipeAdv;
    private static DocumentBuilder builder;

    private static String DOC_BUILDER_FACTORY =
	"org.apache.xerces.jaxp.DocumentBuilderFactoryImpl";

    private static final String USAGE = "jxtasearch [-f] [-d] [-u <query uuid>] " +
	"[[-r <searchterms> -n <jxtasearch pipe adv> [-q <queryspace>] "+
	"[-t <timeout>] [-m <maxhits>] [-p <maxproviders>]]";


    public jxtasearch() {
    }

    private static void initClass( ShellEnv shellEnv ) throws Exception {
	if( routerClient == null ) {
	    System.setProperty ("javax.xml.parsers.DocumentBuilderFactory",
				DOC_BUILDER_FACTORY );
	    
	    builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
	    PeerGroup group = (PeerGroup) shellEnv.get("stdgroup").getObject();
	    routerClient = new JxtaRouterClient( group,
						 null,
						 builder );
	}
    }

    public void help() {
	println("NAME");
	println("     jxtasearch - queries a jxtasearch peer");
	println(" ");
	println("USAGE");
	println("     " + USAGE);
	println(" ");
	println("SYNOPSIS");
	println("    jxtasearch [-u] view results of query with specified uuid");
	println("               [-d] discover jxtasearch peers");
	println("               [-f] flushes cached search results");
	println("               [-r] Issues a search on the infrasearch network");
	println("               [-q] URI specifying the queryspace");
	println("               [-n] Pipe advertisement of desired jxtasearch node");
	println("               [-t] specifies the query lifetime for the query" );
	println("               [-m] limits the max number of results per provider");
	println("               [-p] limits the fanout value (number of peers to to route query to at each hop)");
	println(" ");
	println("DESCRIPTION");
	println("Use jxtasearch to query an infrasearch peer. When executed with no");
	println("parameters, or with just the -u parameter, the locally stored query");
	println("results are displayed.  The -f parameter can be used to flush the");
	println("local store of search results.");
	println(" ");
	println("Specifying the -d parameter issues a discovery for jxtasearch");
	println("hubs. A list of pipe advertisements to those hubs will be stored");
	println("in shell variables and listed on the console.");
	println("");
	println("When the -r parameter is specified, a query will be issued.  A pipe");
	println("advertisment for a jxtasearch peer must be specified with the -n");
	println("parameter (a list of jxtasearch peers can be discovered with the");
	println("-d parameter described above). The -q parameter can be used to specify");
	println("a \"queryspace\", which is a URI that identies the space to be");
	println("queried.  If not specified, the default queryspace");
	println("\"http://search.jxta.org/text\" will be used. The -t, -m, and -p");
    }

    public int startApp (String[] args) {
	ShellEnv env = getEnv();

        String advdocname = null;
        String query = null;
	String queryuuid = null;
	String queryspace = null;
	String maxFanout = null;
	String maxResultsPerProvider = null;
	String queryLifetime=null;

	boolean flushFlag=false;
	boolean discoverFlag=false;

	boolean illegalOption = false;
	int c;

	if( args != null ) {
	    GetOpt options = new GetOpt(args, "fdn:r:q:u:p:t:m:");
	    
	    while ((c = options.getNextOption()) != -1 && !illegalOption) {
		switch(c) {
		case 'f':
		    flushFlag = true;
		    break;
		case 'd':
		    discoverFlag=true;
		    break;
		case 'n': 
		    advdocname=options.getOptionArg();
		    break;
		case 'r': 
		    query=options.getOptionArg();
		    break;
		case 'u':
		    queryuuid=options.getOptionArg();
		    break;
		case 'q':
		    queryspace=options.getOptionArg();
		    break;
		case 'm':
		    maxResultsPerProvider=options.getOptionArg();
		    break;
		case 'p':
		    maxFanout=options.getOptionArg();
		    break;
		case 't':
		    queryLifetime=options.getOptionArg();
		    break;
		default:  
		    illegalOption = true;
		    break;
		}
	    }
	}
	    
	if( illegalOption ) {
	    return syntaxError();
	}

	if( query==null && (advdocname!=null ||
			    queryspace!=null ||
			    maxResultsPerProvider!=null ||
			    maxFanout!=null ||
			    queryLifetime!=null ) ) {
	    return syntaxError();
	}

	if( flushFlag ) {
	    queryResults.clear();
	}

	if( query==null ) {
	    if( discoverFlag )
		return showDiscovery();
	    else
		return showResults( queryuuid );
	} else {
	    if( advdocname == null ) {
		return syntaxError();
	    }
	    return executeQuery( query, 
				 queryspace,
				 queryLifetime,
				 maxResultsPerProvider,
				 maxFanout,
				 advdocname );
	}
    }

    public void stopApp () {
    }

    private int syntaxError () {
	println( USAGE );
	return ShellApp.appParamError;
    }

    private int showDiscovery() {

	try {
	    Enumeration e = JxtaRouterClient.discoverRouterPipeAdvs( group );
	    int i=0;
	    while( e!=null && e.hasMoreElements() ) {
		PipeAdvertisement adv = (PipeAdvertisement) e.nextElement();
		String desc = "searchpeer"+i;
		getEnv().add( desc, new ShellObject( desc, adv ) );
		println( desc );
		i++;
	    }
	} catch( Exception e ) {
	    e.printStackTrace();
	    return ShellApp.appMiscError;
	}

	return ShellApp.appNoError;
    }

    private int showResults( String uuid ) {
	try {
	    if( uuid == null )
		createShellEnvObjects();
	    else
		createShellEnvObjects( uuid );
	} catch( Exception e ) {
	    println( "infrasearch: problem creating shell env objects" );
	    e.printStackTrace();
	    return ShellApp.appMiscError;
	}

	return ShellApp.appNoError;
    }

    private int executeQuery( String query, 
			      String queryspace,
			      String queryLifetime,
			      String maxResultsPerProvider,
			      String maxFanout,
			      String advenv ) {
	try {
	    initClass( getEnv() );

	    System.out.println( "executeQuery()");

	    PipeAdvertisement adv = getPipeAdvFromEnv( advenv );
	    System.out.println( "executeQuery(): pipeadv: " + adv );
	    if( adv == null ) {
		println( "jxtasearch: invalid pipe advertisement environment variable specified");
		return ShellApp.appMiscError;
	    }

	    System.out.println( "executeQuery(): setRouterPipeAdv()" );
	    routerClient.setRouterPipeAdv( adv );

	    byte[] querybytes =
		("<query><text>"+query+"</text></query>").getBytes();

	    String uuid = UUIDFactory.newUUID().toString();

	    if( queryspace == null )
		queryspace = DEFAULT_QUERYSPACE;

	    Request req = new Request(queryspace,
				      uuid,
				      (queryLifetime==null)? DEFAULT_LIFETIME : Integer.parseInt(queryLifetime),
				      (maxFanout==null)? DEFAULT_FANOUT : Integer.parseInt(maxFanout),
				      (maxResultsPerProvider==null)? DEFAULT_RESULTS_PER_PROV : Integer.parseInt(maxResultsPerProvider),
				      Integer.MAX_VALUE,
				      Integer.MAX_VALUE,
				      querybytes );
				       

	    System.out.println( "excuteQuery(): search()" );
	    String uuid2 = routerClient.search( req,
					       new Callback( queryResults ) );
	    println( "jxtasearch: search message sent: uuid = "+uuid2);

        } catch( Exception e ) {
	    e.printStackTrace();
            return ShellApp.appMiscError;
        }

	return ShellApp.appNoError;
    }

    /** Returns a pipe adv from a StructuredTextDoc environment var **/
    private PipeAdvertisement getPipeAdvFromEnv( String advenv ) {
        
        ShellObject advobj = getEnv().get( advenv );
        if ( advobj == null ) {
            println( "infrasearch: non-existent advertisement variable: " +
		     advenv );
            return null;
        }
        
        return (PipeAdvertisement) advobj.getObject();
    }

    /** creates shell objects and prints out description **/
    private void createShellEnvObjects( String uuid ) throws Exception {
	ShellEnv env = getEnv();
	Enumeration e = queryResults.enumerateHits( uuid );
	if( e == null ) return;
	int i=0;
	while( e.hasMoreElements() ) {
	    Document doc = (Document) e.nextElement();
	    String desc = "jxtasearch result (uuid="+uuid+") search"+i;
	    env.add( "search"+i, new ShellObject( desc, doc ) );
	    println (desc);
	    println (getSummary(doc));
	    println ("");
	    i++;
	}
    }

    private void createShellEnvObjects() throws Exception {
	ShellEnv env = getEnv();
	Enumeration hitvectors = queryResults.enumerateAllHits();
	Enumeration uuidkeys = queryResults.enumerateAllKeys();
	if( hitvectors == null || uuidkeys == null ) return;
	int i=0;
	while( hitvectors.hasMoreElements() ) {
	    Vector v = (Vector) hitvectors.nextElement();
	    String uuid = (String) uuidkeys.nextElement();
	    Enumeration e = v.elements(); 
	    if( e == null ) return;
	    while( e.hasMoreElements() ) {
		Document doc = (Document) e.nextElement();
		String desc = "jxtasearch result (uuid="+uuid+") search"+i;
		env.add( "search"+i, new ShellObject( desc, doc ) );
		println (desc);
		println (getSummary(doc));
		println ("");
		i++;
	    }
	}
    }
    
    private String getSummary( Document doc ) {
	StringBuffer buf = new StringBuffer();
	int hitnum = 0;
	Element root = doc.getDocumentElement();
	NodeList responseList = root.getElementsByTagName("response");
	int responses=responseList.getLength();
	for(int i=0;i<responses;i++) {
	    Element response = (Element) responseList.item(i);
	    NodeList dataList = response.getElementsByTagName("data");
	    // ignore anything but the first
	    if( dataList.getLength() > 0 ) {
		Element data = (Element) dataList.item(0);
		NodeList hitList = data.getElementsByTagName("hit");
		int hits = hitList.getLength();
		for( int j=0; j<hits; j++ ) {
		    Element hit = (Element) hitList.item(j);
		    String titleStr = getChildElementValue(hit,"title");
		    String linkStr = getChildElementValue(hit,"link");
		    buf.append( "  hit#"+ hitnum + "  " + titleStr +
				"  " + linkStr + "\n" );
		    hitnum++;
		}
	    }
	}
	return buf.toString();
    }

    private String getChildElementValue( Element parent, String tagname ) {
	NodeList list = parent.getElementsByTagName( tagname );
	if( list.getLength() == 0 )
	    return null;
	else {
	    Element e = (Element) list.item(0);
	    if( e==null )
		return null;
	    else
		return getElementValue( e );
	}
    }

    private String getElementValue( Element e ) {
	if( e==null )
	    return null;
	Node n = e.getFirstChild();
	if( n==null )
	    return null;
	else
	    return n.getNodeValue();
    }


    private class InfrasearchException extends Exception {
        public InfrasearchException() {
        }
    }


    /** Stores the result documents in a hashtable of vectors.  The hashkey is
	is the query uuid **/
    private static class QueryResults {

	/** Hashtable of Vectors. The vector contain StructuredTextDocs **/
	private Hashtable results = new Hashtable();

	/** adds a search result **/
	synchronized void add( Document result, String uuid ) {
	    System.out.println( "jxtasearch: adding result: " + uuid );
	    Vector v = (Vector) results.get( uuid );
	    if( v==null ) {
		v = new Vector();
		results.put( uuid, v );
	    }
	    v.addElement( result );
	}

	/** clears all search results **/
	synchronized void clear() {
	    results.clear();
	}

	/** returns all search results. Returns an enumeration of Vectors,
	    where the Vectors contain StructuredTextDocuments. **/
	synchronized Enumeration enumerateAllHits() {
	    return results.elements();
	}

	/** returns an enumeration of the keys.  The keys are String uuids. **/
	synchronized Enumeration enumerateAllKeys() {
	    return results.keys();
	}

	/** returns the search results for the given uuid. Returns an
	    enumeration of Elements. Will return null
	    if there are no results for the given uuid.
	**/
	synchronized Enumeration enumerateHits( String uuid ) {
	    Vector v = (Vector) results.get( uuid );
	    if( v==null )
		return null;
	    else
		return v.elements();
	}
    }

    private static class Callback implements QueryResponseCallback {
	QueryResults queryResults;
	public Callback( QueryResults queryResults ) {
	    this.queryResults = queryResults;
	}

	public void onQueryResults(String uuid, Document results) {
	    System.out.println( "Callback: onQueryResults called for uuid="+
				uuid );
	    System.out.println( results.toString() );
	    queryResults.add( results, uuid );
	}

	public void onException( String uuid, Exception e ) {
	    System.out.println( "Callback: onException called for uuid="+
				uuid );
	}
    }
}
