/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: BM.java,v 1.1.1.1 2001/06/01 16:51:13 jbeatty Exp $
 */
package net.jxta.search.util;

import java.io.PrintStream;

/**
 * Boyer-Moore string-matching algorithm.
 * Usage:
 * <pre>
 * StringMatcher bm = new BM ();
 * 
 * StringMatcher.CompiledPattern pattern = bm.compile (searchPhrase);
 * 
 * int index = bm.match (text, 0, pattern);
 * 
 * System.out.println ("First location of " + searchPhrase + " at " + index);
 *
 * </pre>
 */

public class BM implements StringMatcher {


    public static boolean test(PrintStream ps, String [] argv) {

        boolean passing = false;

        if (argv.length > 1) {
	    passing = TestBM.test(ps, argv);
        } else {
            passing = TestBM.test(ps);
        }
      
	return passing;

    }

    public static void main (String[] argv) {

        test(System.out,argv);

    }

    public StringMatcher.CompiledPattern compile (String pattern) {
	return new CompiledPatternImpl (pattern);
    }

    public int indexOf (char[] text, int start, char c) {

	for (int i = start; i < text.length;  i++)
	    if (text[i] == c)
		return i;

	return -1;
    }

    public int match (char[] text, 
		      int    textStart, 
		      StringMatcher.CompiledPattern compiledPattern) 
    {
	CompiledPatternImpl  impl = (CompiledPatternImpl) compiledPattern;

	if (impl.chars.length == 1)
	    return indexOf (text, textStart, impl.chars[0]);

	int[]  lastOccurrenceFunc = impl.lastOccurrenceFunc;
	int[]  goodSuffixFunc     = impl.goodSuffixFunc;
	char[] pattern            = impl.chars;
	int n = text.length;
	int m = pattern.length;
	int s = textStart;
	int j;

	if (textStart + pattern.length > text.length)
		return -1;
	


	while (s <= n - m) {
	    
	    j = m-1;
	    
	    while (j >= 0 && charEquals (pattern[j], text[s+j])) {
		j --;
	    }

	    if (j == -1) {
		// s = s + gamma[0];
		return s;
	    } else {

		s = s + Math.max (goodSuffixFunc[j],
				  j-lastOccurrenceFunc[text[s+j]]);
	    }

	    
	}
	
	return -1;
    }


    private static class CompiledPatternImpl 
	implements StringMatcher.CompiledPattern
    {
	char[] chars;
	int[]  lastOccurrenceFunc;
	int[]  goodSuffixFunc;
	
	CompiledPatternImpl (String pattern) {
	    chars              = pattern.toCharArray ();
	    lastOccurrenceFunc = computeLastOccurrenceFunction (chars);
	    goodSuffixFunc     = computeGoodSuffixFunction (chars);
	}

	private int[] computeLastOccurrenceFunction (char[] pattern) {
	    int[] lambda = new int[256];
	    int m = pattern.length;
	    
	    for (int j = 0;  j < m;  j++) {
		
		lambda[(((int)pattern[j]) & 0x000000FF)] = j;
	    }
	    
	    return lambda;
	}
	
	
	private int[] computeGoodSuffixFunction (char[] pattern) {
	    
	    int[] prefixFunction        = computePrefixFunction(pattern, true);
	    int[] reversePrefixFunction = computePrefixFunction(pattern, false);
	    int   m                     = pattern.length;
	    int[] gamma                 = new int[m];
	    
	    for (int j = 0;  j < m;  j++) {
		gamma[j] = m - prefixFunction[m-1];
	    }
	    
	    for (int l = 0;  l < m;  l++) {
		
		int j = m - reversePrefixFunction[l] - 1;
		
		if (gamma[j] > l - reversePrefixFunction[l]) 
		gamma[j] = l - reversePrefixFunction[l];
		
	    }
	    
	    return gamma;
	}
	
	
	private int[] computePrefixFunction (char[] pattern, boolean forward) {
	    
	    int   m    = pattern.length;
	    int   k    = 0;
	    int[] func = new int[m];
	    
	    func[0] = 0;
	    
	    
	    if (forward ) {
		for (int q = 1;  q < m;  q++) {
		    
		    while (k > 0 &&  ! charEquals (pattern[k], pattern[q])) 
			k = func[k];
		    
		    
		    if (charEquals (pattern[k], pattern[q])) 
			k ++;
		    
		    
		    func[q] = k;
		}
	    } else {
		
		for (int q = m-1;  q > 0;  q--) {
		    
		    while (k > 0 && ! charEquals (pattern[k], pattern[q])) {
			k = func[k];
		    }
		    
		    if (charEquals (pattern[k], pattern[q])) {
			k++;
		    }
		    
		func[q] = k;
		}
	    }
	    
	    return func;
	}
    }


    static final boolean charEquals (char patternChar, char textChar) {

	return ( (patternChar == textChar) ||
		 (isSpace (patternChar) && isSpace (textChar)) );
	
    }
    
    
    static final boolean isSpace (char c) {
	return Character.isWhitespace (c) || c == '>' || c == '<' || 
	    c == '\"' || c == '.' || c == ',';
  }
}



class TestBM {

 
    public static boolean test(PrintStream ps) {

        String doc1 = 
 	   "<html>\n<head>\n<title>Test document 1</title>\n</head>\n"
	   + "<body>\n <h1>Doc 1</h1>\n"
	   + "And as we wind on down the road\n"
	   + "Our shadows taller than our souls\n"
	   + "There walks a lady we all know\n"
	   + "Who shines white light and wants to show\n"
	   + "How everything still turns to gold\n"  
	   + "</body>\n</html>";

        String pattern = ">";
	BM bm = new BM ();
	StringMatcher.CompiledPattern compiledPattern;

	compiledPattern = bm.compile (pattern);
	int s = bm.match (doc1.toCharArray (), 0, compiledPattern);
	    
        if (s == -1) {
	   ps.println ("Pattern doesn't occur in text.");
           return false;
	} else
	   ps.println ("Pattern occurs with shift " + s);


	return true;
    }



    public static boolean test(PrintStream ps, String [] argv) {

        String pattern = argv[0];

	BM bm = new BM ();
	StringMatcher.CompiledPattern compiledPattern;	    
	compiledPattern = bm.compile (pattern);

	for (int i = 1;  i < argv.length;  i++)  {
	    int s = -1;
	    do {
		    
		s = bm.match (argv[i].toCharArray (), s+1, compiledPattern);
		if (s != -1)
		   ps.println ("Pattern occurs with shift " + s);
	    } while (s != -1);
	} 

	return false;
    }

}
