/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Html.java,v 1.1.1.1 2001/06/01 16:51:13 jbeatty Exp $
 */
package net.jxta.search.util;

import java.io.*;
import java.net.*;
import java.util.*;


/**
 * There are three parts to this:
 * <li>
 *   <li>Closing dangerous opened tags.
 *   <li>Removing dangerous dangling close-tags.
 *   <li>Removing irrelevant tags and stuff in-between.
 * </ul>
 */
public class Html {

    static final int NO_TAG = 1;
    static final int PARSING_TAG = 2;
    static Hashtable mustBalanceTable;
    static Hashtable ignoreTagTable;
    
    static {
	mustBalanceTable = new Hashtable (20);
	ignoreTagTable   = new Hashtable (20);

	mustBalanceTable.put ("i",      "/i");
	mustBalanceTable.put ("b",      "/b");
	mustBalanceTable.put ("a",      "/a");

	mustBalanceTable.put ("pre",    "/pre");
	mustBalanceTable.put ("code",   "/code");
	mustBalanceTable.put ("em",     "/em");
	
	mustBalanceTable.put ("table",  "/table");
	/* need to keep track of these because if there's
	   a dangling /tr or /td, then it can fuck up our formatting.*/
	mustBalanceTable.put ("td",     "/td");
	mustBalanceTable.put ("tr",     "/tr");
	
	mustBalanceTable.put ("div",    "/div");
	mustBalanceTable.put ("center", "/center");
	mustBalanceTable.put ("font",   "/font");
	
	
	ignoreTagTable.put ("html",   Boolean.TRUE);
	ignoreTagTable.put ("head",   Boolean.TRUE);
	ignoreTagTable.put ("script", Boolean.TRUE);
    }

    public static String stripHtmlTags (String data) {
      // TODO: ignore select tags that we MAY want to allow such as <i></i> or <b></b>

      StringBuffer clean = new StringBuffer (data.length ());
      int currPos  = 0;
      int tagStart = 0;
      int tagEnd   = 0;

      while ((tagStart >= 0) && (tagEnd >= 0)) {
        tagStart = data.indexOf ("<", currPos);
        if (tagStart > 0) {
          clean.append (data.substring (currPos, tagStart));
          currPos = data.indexOf (">", tagStart) + 1;
        } else {
          // cheap hack
          if ((currPos == 0) && (tagStart < 0)) // no tags to replace present
            clean.append (data);

        }
      }

      return (clean.toString ());

    }

    /**
     * Removes all HTML tags from the supplied string.
     *
     * <p><b>Example</b>: <i>Before stripping</i>: Amazon sells other items too such as 
     * <a href='http://amazon.com/movies'>Movies</a> and more.  <i>After 
     * stripping</i>: Amazon sells other items too such as 
     * movies and more.</p>
     * <p>The method should be able to handle text that has no html
     * markup embedded in it.</p>
     * <p>FIXME: Does this method provably handle unbalanced tags correctly?</p>
     *
     * @param String containing html and text.  
     * @return String with just the text, all html tags stripped out.
     *
     */
    public static String stripTags (String data) {
      boolean done = false;

      StringBuffer clean = new StringBuffer (data.length ());
      final int START        = 0;
      final int SEARCH_OPEN  = 1;
      final int SEARCH_CLOSE = 2;
      final int FINISH       = 3;

      int currPos  = 0;
      int openTag  = 0;
      int closeTag = 0;
      int state    = START;

      while (!done) {
        switch (state) {
          case START:
            closeTag = data.indexOf (">", currPos);
            openTag  = data.indexOf ("<", currPos);

            if ((closeTag < 0) && (openTag < 0)) {
              state = FINISH; // neither found, cleanup
            } else if (((closeTag >= 0) && (closeTag < openTag)) ||
                       ((closeTag >= 0) && (openTag < 0))) {
              // Either there is a > before a < or there is no <
              currPos = closeTag + 1;
              state = SEARCH_OPEN;
            } else if (openTag == 0) {
              state = SEARCH_CLOSE;
            } else {
              state = SEARCH_OPEN;
            }
            break;

          case SEARCH_OPEN:
            if (currPos < data.length ()) {
              openTag = data.indexOf ("<", currPos);
              if (openTag >= 0) {
                clean.append (data.substring (currPos, openTag));
                currPos = openTag + 1;
                state = SEARCH_CLOSE;
              } else {
                state = FINISH;
              }
            } else { 
              state = FINISH;
            }
            break;
           
          case SEARCH_CLOSE:
            if (currPos < data.length()) {
              closeTag = data.indexOf (">", currPos);
              if (closeTag >= 0) {
                currPos = closeTag + 1;
                state = SEARCH_OPEN;
              } else {
                done = true;
              }
            } else {
              state = FINISH;
            }
            break;

          case FINISH:
          
            if (currPos < data.length ())
              clean.append (data.substring (currPos));
            done = true;
            break;
        }

      }
      return (clean.toString());

    }


    public static String modifyLinks (String html, URL base) {
	StringBuffer result = new StringBuffer (html.length () * 2);

	String indexBuffer = html.toUpperCase ();

	final int NO_PARSE   = 1; // looking for a <A or <IMG
	final int PARSING    = 2; // parsing a <A tag (add onclick if prefs allow?)
	final int END_SEARCH = 4; // looking for a closing >

	int state     = NO_PARSE;
	int currPos   = 0;
	int imgStart  = 0;
	int aStart    = 0;
	int tagStart  = 0;
	int tagEnd    = 0;
	int hrefStart = 0;
	int hrefEnd   = 0;
	int i = 0;
	boolean imgDone = false;
	boolean aDone   = false;

	while ( imgDone != true || aDone != true ) {
		switch (state) {

		case NO_PARSE:
		    if (aDone != true) 
			aStart = indexBuffer.indexOf ("<A", currPos);
		    else
			aStart = -1;

		    if (imgDone != true)
			imgStart = indexBuffer.indexOf ("<IMG", currPos);
		    else
			imgStart = -1;

		    if ((aStart > -1) && (imgStart > -1)) {
			tagStart = Math.min (aStart, imgStart);
			state = END_SEARCH;
		    } else if ((aStart > -1) && (imgStart <= -1)) {
			imgDone = true;
			tagStart = aStart;
			state = END_SEARCH;
		    } else if ((aStart <= -1) && (imgStart > -1)) {
			aDone = true;
			tagStart = imgStart;
			state = END_SEARCH;
		    } else if ((aStart <= -1) && (imgStart <= -1)) {
			aDone   = true;
			imgDone = true;
			// we got nuthin, just append from wherever we started to the end
			// we do this here because its the end of our journey
			result.append (html.substring(currPos));
		    }
		    break;

		case PARSING:
		    // take everything from currPos to tagStart and append it
		    // didle with the contents of the tag
		    if ( indexBuffer.startsWith ("<A", tagStart) ) {
			// this is a <A tag, copy it to the buffer, find the href
			// and validate its not / relative, and build a <A/> tag
			String link = getTagAttrib ("href", html.substring (tagStart, tagEnd + 1));
			if (link != null) {
			    result.append("<A HREF=" + relativeLink (base.toString (), link) + ">");
			} else {
			    // somethings wrong, output anyway :)
			    result.append (html.substring (tagStart, tagEnd + 1));
			}
			currPos = tagEnd + 1;
			state = NO_PARSE;
		    } else if ( indexBuffer.startsWith ("<IMG", tagStart) ) {
			// this is a <IMG tag, find the SRC, make sure its not / relative
			// find the acceptable attribute tags, make sure they are not outside
			// of the parameters.  
			String src = getTagAttrib ("src", html.substring (tagStart, tagEnd + 1));
			if (src != null) {
			    result.append ("<IMG SRC=" + relativeLink (base.toString (), src + ">"));
			} else {
			    result.append (html.substring (tagStart, tagEnd + 1));
			}
			currPos = tagEnd + 1;
			state = NO_PARSE;
		    } else {
			// we are fubared here :)
			currPos = tagEnd + 1;
			state = NO_PARSE;
		    }

		    break;

		case END_SEARCH:
		    // we know where the tag starts and we are gonna find the end
		    // append everything before the tag and continue, we do this
		    // so that if the tag is bad we can just skip it
		    if (tagStart > currPos)
			result.append (html.substring(currPos, tagStart) );

		    tagEnd = indexBuffer.indexOf (">", tagStart);

		    if ( (tagEnd > -1) && (tagEnd <= indexBuffer.length ()) ) {
			state = PARSING;
		    } else {
			// something is not right, skip ahead one byte so we don't
			// process this tag again
			System.out.println ("Error : uh oh, endSearch state but no end");
			currPos = tagStart + 1;
			state = NO_PARSE;
		    }
		    break;

		default:
		    break;

		}
	}
	return result.toString ();
    }


    public static String relativeLink (String base, String link) {
	String buff = link.toUpperCase ();
	if (buff.startsWith ("HTTP") ) {
	   return link;
	} else if (buff.startsWith ("/") ) {
	   return base + link.substring (1);
	} else {
	   return base + link;
	}
    }

    /*
     * Finds all instances of the supplied terms and adds HTML highlights to 
     * them in the supplied data.
     */
    public static String highlightTerms (String terms, String data) {
        StringTokenizer termTokens = new StringTokenizer (terms, " ,.+=()|'\"");

        int currPos   = 0;
        int wordStart = 0;
        int wordEnd   = 0;

        boolean hiTermDone = false;
        String  hiTerm     = null;
        String  downData   = null;
        String  workData   = "";
        String  result     = new String (data);

        int numTerms = termTokens.countTokens ();

        for (int i=0; i < numTerms; i++) {
            // For each token in "terms" compare against every token in "data"
            currPos    = 0;
            hiTermDone = false;
            hiTerm     = termTokens.nextToken ().toLowerCase (); // downshift for compare
            workData   = "";
            downData   = result.toLowerCase ();

            while (hiTermDone == false) {
                wordStart = downData.indexOf (hiTerm, currPos);
                if (wordStart >= 0) {
                    workData += result.substring (currPos, wordStart);
                    workData += "<b><font color=red>";
                    workData += result.substring (wordStart, (wordStart + hiTerm.length ()));
                    workData += "</font></b>";
                    currPos = (wordStart + hiTerm.length ());
                } else {
                    // wordStart = -1 
                    workData += result.substring (currPos, result.length ());
                    hiTermDone = true;
                    result = workData;
                }
            }
        }

        return (result);

    } // End highlightTerms


    /*  getAttribFromTag ()
     *  pass a id and a tag and return a string of its value
     *  Examples
     *     <A HREF="http://blah.com" >               -> http://blah.com
     *     <A href="blah.com/blech?value='boo'">     -> blah.com/blech?value='boo'
     *     <a hRef='blah.com/blech?value="boo"'>     -> blah.com/blech?value="boo"
     *     <a hreF=blah.com/blech?value="boo">       -> blah.com/blech?value="boo"
     *     <IMG src=blah.com/logo.gif >              -> blah.com/logo.gif
     *     <IMG src="blah.com/logo.gif" width="100"> -> blah.com/logo.gif || 100
     *     <img sRC = blah.com/logo.gif border=0>    -> blah.com/logo.gif || 0
     */
    public static String getTagAttrib (String id, String source) {
	String indexBuffer = source.toUpperCase ();
	String searchId    = id.toUpperCase ();

	int attStart   = 0;
	int attEnd     = 0;
	int valueStart = 0;
	int valueEnd   = 0;
	int i          = 0;

	attStart = indexBuffer.indexOf (searchId);
	if (attStart > -1) 
	    attStart = indexBuffer.indexOf ("=", attStart);

	if (attStart > -1)
	    attStart++;

	if (attStart > -1) {
	   valueStart = attStart;
	   if (indexBuffer.charAt (valueStart) == ' ')
		valueStart++;

	   if (valueStart < indexBuffer.length ()) {

		switch (indexBuffer.charAt (valueStart)) {

		   case '\'':  // end can only be quote or >
			if ( (valueEnd = indexBuffer.indexOf ("\'", valueStart + 1)) > -1) {
			    // we got it, end with single quote
			    valueStart++;
			} else if ( (valueEnd = indexBuffer.indexOf (">", valueStart + 1)) > -1 ) {
			    // we got it, end with >
			    valueStart++;
			    if ( indexBuffer.charAt(valueEnd - 1) == ' ')
				valueEnd--;
			} else {
			   // no luck, started with ' but could not find end
			}
			break;

		   case '\"':  // end can only be double quote space or >
			if ( (valueEnd = indexBuffer.indexOf ("\"", valueStart + 1)) > -1) {
			   // we got it, end with double quote
			   valueStart++;
			} else if ( (valueEnd = indexBuffer.indexOf (">", valueStart + 1)) > -1) {
			   // we got it, ends with >
			   valueStart++;
			   if (indexBuffer.charAt (valueEnd - 1) == ' ')
				valueEnd--;
			} else {
			   // we are shit out of luck
			}
			break;

		   case '>':  // slim chance but its a link to the sites root
		   case ' ':  // slim chance but its a link to the sites root
			valueEnd = valueStart;
			break;

		   default:    // no quotes can only be space or >
			if ( (valueEnd = indexBuffer.indexOf (" ", valueStart + 1)) > -1) {
			    // we got it, it ends with a space ie. href=blah.com target....
			} else if ( (valueEnd = indexBuffer.indexOf (">", valueStart + 1)) > -1) {
			    // we got it, it ends with a >  ie href=blah.com>
			} else {
			    // we are shit out of luck
			}
			break;
		}

	   }
	   
	} else {

	}

	if ((valueStart > -1) &&
	    (valueEnd   > -1) &&
            (valueStart < source.length()) &&
	    (valueEnd   < source.length()) )
	    return source.substring(valueStart, valueEnd);
	else
	    return (String) null;
    }




    public static String verify (String html) {
	char[]  chars     = html.toCharArray ();
	int     state     = NO_TAG;
	boolean close_tag = false;
	boolean ignore    = false;
	int     startTag  = 0;
	int     endTag    = 0;
	StringBuffer sb = new StringBuffer (chars.length * 2);
	Stack stack = new Stack ();

	for (int i = 0;  i < chars.length;  i++) {

	    switch (state) {

	    case NO_TAG:
		switch (chars[i]) {
		case '<':
		    if (!ignore) 
			sb.append (chars, endTag, i-endTag);
			
		    state = PARSING_TAG;
		    close_tag = false;
		    startTag = i;
		    break;
		case '>':
		    /* bad html. but we won't be vigilant*/
		    // throw new java.text.ParseException ("> must be escaped.");
		    break;
		}
	    
		break;
		
	    case PARSING_TAG:
		switch (chars[i]) {
		case '/':
		    close_tag = true;
		    break;
		case '>':
		    state = NO_TAG;
		    break;
		default: 
		    ignore = false;
		    if (Character.isLetter (chars[i]) || 
			chars[i] == '!' ||
			chars[i] == '-') {
			String name;
			String closeName;
			int j;

			for (j = i;  j < chars.length;  j++)
			    if (! Character.isLetter (chars[j]) && 
				chars[j] != '!' &&
				chars[j] != '-')
				break;
			if (j == chars.length) {
			    i = j;
			    ignore = true;
			    break;
			}
			    
			name = new String (chars, i, j-i).intern ();
			
			for (; j < chars.length;  j++) // scan to end of tag
			    if (chars[j] == '>')
				break;
			if (j == chars.length) {
			    ignore = true;
			    i = j;
			    break;
			}

			if (ignoreTagTable.get (name) != null) {
			    ignore = ! close_tag;
				
			    for (int k = startTag;
				 k < j+1 && k < chars.length;  k++)
				chars[k] = ' ';
			} else if (mustBalanceTable.get (name) == null) {
			    if (close_tag)
				sb.append (chars, i-2, j-i+3);
			    else
				sb.append (chars, i-1, j-i+2);
			} else {
			    if (close_tag) {
				boolean found = false;
				while (! stack.empty ()) {
				    
				    String t =  (String) stack.pop ();
				    String ct = (String) mustBalanceTable.get (t);
				    if (name.equals (ct)) {
					sb.append ('<').append (ct).append (">");
					found = true;
					break;
				    } else {
					sb.append ("<").append (ct).append ('>');
				    }
				}
				
				if (! found) {
				    for (int k = startTag;  
					 k < j+1 && k < chars.length;  k++)
					chars[k] = ' ';
				}
				
			    } else {
				stack.push (name);
				sb.append (chars, i-1, j-i+2);
			    }
			}
			
			i = j;
			state = NO_TAG;
			endTag = i+1;
		    }
		
		    
		}
		
	    }

	}

	if (! ignore) 
	    sb.append (chars, endTag, chars.length-endTag);
	    
	while (! stack.empty ()) {
	    String t = (String) stack.pop ();
	    sb.append ("</").append (t).append (">");
	}

	return sb.toString ();
    }


    private static boolean testGetTagAttrib (String id, String tag, String ref) {
	System.out.println ("testGetTagAttrib : ID  = " + id);
	System.out.println ("                 : Tag = " + tag);
	System.out.println ("                 : Ref = " + ref);
	String result = getTagAttrib (id, tag);
	System.out.println ("                 : Res = " + result);

	if (result.compareTo (ref) != 0) {
	    System.out.println ("                 : FAILED\n");
	    return false;
	} else {
	    System.out.println ("                 : PASSED\n");
	    return true;
	}

    }


    public static void testHtmlClass() {

	//BufferedReader in = 
	//    new BufferedReader (new InputStreamReader (System.in));
	//String line = "";
	//StringBuffer sb = new StringBuffer ();
	
	//while ( (line = in.readLine ()) != null) {
	//    System.out.println (line);
	//    sb.append (line);
	//}

	// System.out.println ("Corrected: " + verify (sb.toString ()));

        // Begin Unit Test for modifyLinks
	// End Unit Test for modifyLinks
	 
	// Begin Unit Test for getTagAttrib
	testGetTagAttrib ("href", 
			  "<a Href=www.bingo.com/something.html>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("href",
			  "<a HrEf=www.bingo.com/something.html target=_blank>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("href",
			  "<a Href = www.bingo.com/something.html target = _blank>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("href",
			  "<A hREF =www.bingo.com/something.html target =_blank>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("href",
			  "<a HreF= www.bingo.com/something.html target= _blank>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("href",
			  "<A HrEF=\"www.bingo.com/something.html\">",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("HrEF",
			  "<A hREF=\"www.bingo.com/something.html>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("hREF",
			  "<a hReF=\"www.bingo.com/something.html >",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("HrEF",
			  "<a Href=\"www.bingo.com/something.html?value='test'\">",
			  "www.bingo.com/something.html?value='test'");

	testGetTagAttrib ("Href",
			  "<a Href=\"www.bingo.com/something.html?value='test'>",
			  "www.bingo.com/something.html?value='test'");

	testGetTagAttrib ("hreF",
			  "<a Href=\"www.bingo.com/something.html?value='test' >",
			  "www.bingo.com/something.html?value='test'");

	testGetTagAttrib ("hrEf",
			  "<a Href=\"www.bingo.com/something.html?value='test >",
			  "www.bingo.com/something.html?value='test");

	testGetTagAttrib ("HrEF",
			  "<a Href=\"www.bingo.com/something.html?value='test'\" target=\"_blank\">",
			  "www.bingo.com/something.html?value='test'");

	testGetTagAttrib ("hREF",
			  "<a Href =\"www.bingo.com/something.html?value='test'\" target =\"_blank\">",
			  "www.bingo.com/something.html?value='test'");

	testGetTagAttrib ("href",
			  "<a hrEf= \"www.bingo.com/something.html?value='test'\" target= \"_blank\">",
			  "www.bingo.com/something.html?value='test'");

	testGetTagAttrib ("HREF",
			  "<a HreF = \"www.bingo.com/something.html?value='test'\" target = \"_blank\">",
			  "www.bingo.com/something.html?value='test'");

	testGetTagAttrib ("href",
			  "<A HrEF='www.bingo.com/something.html'>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("href",
			  "<A hREF='www.bingo.com/something.html>",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("hreF",
			  "<a hReF='www.bingo.com/something.html >",
			  "www.bingo.com/something.html");

	testGetTagAttrib ("hRef",
			  "<a Href='www.bingo.com/something.html?value=\"test\"'>",
			  "www.bingo.com/something.html?value=\"test\"");

	testGetTagAttrib ("HREf",
			  "<a Href='www.bingo.com/something.html?value=\"test\">",
			  "www.bingo.com/something.html?value=\"test\"");

	testGetTagAttrib ("hRef",
			  "<a Href='www.bingo.com/something.html?value=\"test\" >",
			  "www.bingo.com/something.html?value=\"test\"");

	testGetTagAttrib ("HrEf",
			  "<a Href='www.bingo.com/something.html?value=\"test >",
			  "www.bingo.com/something.html?value=\"test");

	testGetTagAttrib ("HrEF",
			  "<a Href=\'www.bingo.com/something.html?value=\"test\"' target=\"_blank\">",
			  "www.bingo.com/something.html?value=\"test\"");

	testGetTagAttrib ("hREF",
			  "<a hRef ='www.bingo.com/something.html?value=\"test\"' target = '_blank'>",
			  "www.bingo.com/something.html?value=\"test\"");

	testGetTagAttrib ("HreF",
			  "<a href= 'www.bingo.com/something.html?value=\"test\"' target = '_blank'>",
			  "www.bingo.com/something.html?value=\"test\"");

	testGetTagAttrib ("href",
			  "<a HreF = 'www.bingo.com/something.html?value=\"test\"' target = '_blank'>",
			  "www.bingo.com/something.html?value=\"test\"");

	testGetTagAttrib ("href",
			  "<a href=>",
			  "");

	testGetTagAttrib ("href",
			  "<a href = >",
			  "");

	// End Unit Test for getTagAttrib
	

	// Begin Unit Test for relativeLink
	String relBase  = "http://www.bingo.com/";
	String relTest1 = "http://www.bingo.com/somewhere.jsp";
	String relRef1  = "http://www.bingo.com/somewhere.jsp";
	System.out.println ("Html.relativeLink - test 1 : BEGIN");
	System.out.println ("                           : Base      = " + relBase);
	System.out.println ("                           : Link      = " + relTest1);
	System.out.println ("                           : Reference = " + relRef1);
	String relTestResult1 = relativeLink (relBase, relTest1);
	System.out.println ("                           : Result    = " + relTestResult1);
	if ( relTestResult1.compareTo (relRef1) != 0) {
	    System.out.println ("       FAILED              :\n\n");
	} else {
	    System.out.println ("       PASSED              :\n\n");
	}	

	
	String relTest2 = "hTTp://www.bingo.com/sq/";
	String relRef2  = "hTTp://www.bingo.com/sq/";
        System.out.println ("Html.relativeLink - test 2 : BEGIN");
        System.out.println ("                           : Base      = " + relBase);
        System.out.println ("                           : Link      = " + relTest2);
        System.out.println ("                           : Reference = " + relRef2);
        String relTestResult2 = relativeLink (relBase, relTest2);
	System.out.println ("                           : Result    = " + relTestResult2);
        if ( relTestResult2.compareTo (relRef2) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

	String relTest3 = "/today/numbers.html";
	String relRef3  = "http://www.bingo.com/today/numbers.html";

        System.out.println ("Html.relativeLink - test 3 : BEGIN");
        System.out.println ("                           : Base      = " + relBase);
        System.out.println ("                           : Link      = " + relTest3);
        System.out.println ("                           : Reference = " + relRef3);
        String relTestResult3 = relativeLink (relBase, relTest3);
	System.out.println ("                           : Result    = " + relTestResult3);
        if ( relTestResult3.compareTo (relRef3) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

	String relTest4 = "today/numbers.html";
	String relRef4  = "http://www.bingo.com/today/numbers.html";
        System.out.println ("Html.relativeLink - test 4 : BEGIN");
        System.out.println ("                           : Base      = " + relBase);
        System.out.println ("                           : Link      = " + relTest4);
        System.out.println ("                           : Reference = " + relRef4);
        String relTestResult4 = relativeLink (relBase, relTest4);
	System.out.println ("                           : Result    = " + relTestResult4);
        if ( relTestResult4.compareTo (relRef4) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String relTest5 = "today/numbers.html";
        String relRef5  = "http://www.yahoo.com/today/numbers.html";
        System.out.println ("Html.relativeLink - test 5 : BEGIN");
        System.out.println ("     FAILURE EXPECTED      : Base      = " + relBase);
        System.out.println ("                           : Link      = " + relTest5);
        System.out.println ("                           : Reference = " + relRef5);
        String relTestResult5 = relativeLink (relBase, relTest5);
        System.out.println ("                           : Result    = " + relTestResult5);
        if ( relTestResult5.compareTo (relRef5) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
	}




        System.out.println ("Html.stripTags () testing\n");

        String stripTest1 = "<A href=blah>stripTest1 test data</a>";
        String stripRef1  = "stripTest1 test data";
        System.out.println ("Html.stripTags - Test 1    : BEGIN");
        System.out.println ("                           : data      = " + stripTest1);
        System.out.println ("                           : reference = " + stripRef1);
        String stripTestResult1 = stripTags (stripTest1);
        System.out.println ("                           : Result    = " + stripTestResult1);
        if ( stripTestResult1.compareTo (stripRef1) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest2 = "<A href=blah>stripTest2 test data</a";
        String stripRef2  = "stripTest2 test data";
        System.out.println ("Html.stripTags - Test 2    : BEGIN");
        System.out.println ("                           : data      = " + stripTest2);
        System.out.println ("                           : reference = " + stripRef2);
        String stripTestResult2 = stripTags (stripTest2);
        System.out.println ("                           : Result    = " + stripTestResult2);
        if ( stripTestResult2.compareTo (stripRef2) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest3 = "<A href=blah>stripTest3 test</a> data";
        String stripRef3  = "stripTest3 test data";
        System.out.println ("Html.stripTags - Test 3    : BEGIN");
        System.out.println ("                           : data      = " + stripTest3);
        System.out.println ("                           : reference = " + stripRef3);
        String stripTestResult3 = stripTags (stripTest3);
        System.out.println ("                           : Result    = " + stripTestResult3);
        if ( stripTestResult3.compareTo (stripRef3) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest4 = "<A href=blah>stripTest4 test data";
        String stripRef4  = "stripTest4 test data";
        System.out.println ("Html.stripTags - Test 4    : BEGIN");
        System.out.println ("                           : data      = " + stripTest4);
        System.out.println ("                           : reference = " + stripRef4);
        String stripTestResult4 = stripTags (stripTest4);
        System.out.println ("                           : Result    = " + stripTestResult4);
        if ( stripTestResult4.compareTo (stripRef4) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest5 = "A href=blah>stripTest5 test data</a>";
        String stripRef5  = "stripTest5 test data";
        System.out.println ("Html.stripTags - Test 5    : BEGIN");
        System.out.println ("                           : data      = " + stripTest5);
        System.out.println ("                           : reference = " + stripRef5);
        String stripTestResult5 = stripTags (stripTest5);
        System.out.println ("                           : Result    = " + stripTestResult5);
        if ( stripTestResult5.compareTo (stripRef5) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest6 = "stripTest6<A href=blah> test </a>data";
        String stripRef6  = "stripTest6 test data";
        System.out.println ("Html.stripTags - Test 6    : BEGIN");
        System.out.println ("                           : data      = " + stripTest6);
        System.out.println ("                           : reference = " + stripRef6);
        String stripTestResult6 = stripTags (stripTest6);
        System.out.println ("                           : Result    = " + stripTestResult6);
        if ( stripTestResult6.compareTo (stripRef6) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest7 = "stripTest7 test data";
        String stripRef7  = "stripTest7 test data";
        System.out.println ("Html.stripTags - Test 7    : BEGIN");
        System.out.println ("                           : data      = " + stripTest7);
        System.out.println ("                           : reference = " + stripRef7);
        String stripTestResult7 = stripTags (stripTest7);
        System.out.println ("                           : Result    = " + stripTestResult7);
        if ( stripTestResult7.compareTo (stripRef7) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest8 = ">stripTest8<A href=blah> test </a>data";
        String stripRef8  = "stripTest8 test data";
        System.out.println ("Html.stripTags - Test 8    : BEGIN");
        System.out.println ("                           : data      = " + stripTest8);
        System.out.println ("                           : reference = " + stripRef8);
        String stripTestResult8 = stripTags (stripTest8);
        System.out.println ("                           : Result    = " + stripTestResult8);
        if ( stripTestResult8.compareTo (stripRef8) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }

        String stripTest9 = "stripTest9<A href=blah> test data<";
        String stripRef9  = "stripTest9 test data";
        System.out.println ("Html.stripTags - Test 9    : BEGIN");
        System.out.println ("                           : data      = " + stripTest9);
        System.out.println ("                           : reference = " + stripRef9);
        String stripTestResult9 = stripTags (stripTest9);
        System.out.println ("                           : Result    = " + stripTestResult9);
        if ( stripTestResult9.compareTo (stripRef9) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }



        String stripTest10 = "<html><head></head><body><A href=blah> test data</body></html>";
        String stripRef10  = " test data";
        System.out.println ("Html.stripTags - Test 10    : BEGIN");
        System.out.println ("                           : data      = " + stripTest10);
        System.out.println ("                           : reference = " + stripRef10);
        String stripTestResult10 = stripTags (stripTest10);
        System.out.println ("                           : Result    = " + stripTestResult10);
        if ( stripTestResult10.compareTo (stripRef10) != 0) {
            System.out.println ("       FAILED              :\n\n");
        } else {
            System.out.println ("       PASSED              :\n\n");
        }



	// End Unit Test for relativeLink
    }

    public static void main(String [] args) throws IOException {

       testHtmlClass();
    }

}
