/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: HttpConnection.java,v 1.4 2001/06/06 01:33:48 tcamarda Exp $
 */
package net.jxta.search.util;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * HTTP 1.1 support: keep-alive and chunked transfer-encoding.
 * <p>
 * Below is a usage example.  However, you are best off using the api
 * of <a href=HttpConnectionService.html>HttpConnectionService</a>
 * <p>
 * <pre>
 * HttpConnectionService service = new HttpConnectionService ();
 * HttpConnection        con     = new HttpConnection (new URL ("http://www.ibm.com/"), 
 *                                                     5 * 1000,
 *                                                     service);
 * 
 * HttpConnection.Request  req = con.beginRequest (new URL ("http://www.ibm.com/products/index.html"), "GET");
 * HttpConnection.Response res;
 * InputStream             in;
 * 
 * req.setHeadder ("Foo", "bar");
 * 
 * res = req.getResponse ();
 * 
 * try {
 *     in = res.getInputStream ();
 * 
 *   // read from the stream
 * } finally {
 *     res.done ();
 * }
 * 
 * </pre>
 */

public class HttpConnection implements LRUCache.Cacheable {


    Socket                s;
    URL                   hostURL;
    boolean               connected;
    String                host;
    int                   port;
    int                   readTimeout;
    InputStream           in;
    OutputStream          sout;
    BufferedOutputStream  out;
    byte[]                buf = new byte[1024 * 8];
    long                  cost; /* the cost of opening this connection */

    boolean keepAlive;
    int     maxRequests;
    int     timeout;
    long    lastRequest;
    boolean inRequest;
    int     waitingRequestorCount;
    

    HttpConnectionService service;
    byte[] newline     = "\r\n".getBytes ();
    byte[] headerSep   = ": ".getBytes ();
    byte[] httpVersion = "HTTP/1.1".getBytes ();

    static final boolean DEBUG = false;


    public class Request {
	URL       url;
	String    method;
	boolean   flushedHeaders;
	boolean   shouldFlushBody;
	
	Request (URL url, String method) throws IOException
	{
	    this.url = url;
	    this.method = method;

	    debug (method + " " + url.getFile () + " " + 
		   new String (httpVersion));
	    
	    sout.write (method.getBytes (), 0, method.length ());
	    sout.write (' ');
	    sout.write (url.getFile ().getBytes (), 0, url.getFile ().length());
	    sout.write (' ');
	    sout.write (httpVersion, 0, httpVersion.length);

	    sout.write (newline, 0, newline.length);
	    sout.flush ();
	    setHeader ("Host", url.getHost () + ":" +
		       (url.getPort () == -1 ? 80 : url.getPort ()));
	    setHeader ("User-Agent", "GoneSilent-1.0");
	    setHeader ("Connection", "Keep-Alive");
	    
	    
	    flushedHeaders  = false;
	    shouldFlushBody = false;
	}

	public void setHeader (String name, String value) throws IOException {
	    debug (name + ": " + value);
	    out.write (name.getBytes (), 0, name.length ());
	    out.write (headerSep, 0, headerSep.length);
	    out.write (value.getBytes (), 0, value.length ());
	    out.write (newline, 0, newline.length);
	}

	public OutputStream getOutputStream () {
	    return new RequestOutputStream (this);
	}

	void flushHeaders () throws IOException {

	    out.write (newline, 0, newline.length);
	    flushedHeaders = true;
	}
	
	void writeBody (int b)
	    throws IOException
	{
	    if (! flushedHeaders)
		flushHeaders ();

	    out.write (b);

	    shouldFlushBody = true;
	}

	void writeBody (byte[] buf, int start, int len) 
	    throws IOException
	{

	    if (! flushedHeaders)
		flushHeaders ();

	    out.write (buf, start, len);

	    shouldFlushBody = true;
	} 


	    
	public Response getResponse () 
	    throws IOException
	{

	    try {
		/*
		if (! flushedHeaders) {
		    
		    flushHeaders ();
		    
		} 
		*/
		out.flush ();
	    
		return new Response (url);
	    } catch (IOException e) {
		// make sure that subsequent requests don't barf
		// because "new Response" that threw an exception
		// left the connection object in an inconsistent state.
		disconnect ();
		throw e;
	    }
	}
	
    }


    public class Response {
	int headerLen;
	int contentLen;
	int readCount;
	int readBody;

	int extraStart;
	int extraLen;

	int[] headerStarts;
	int[] headerLens;
	int   headerCount;

	boolean chunked;
	boolean moreChunks;
	boolean inChunk;
	int chunkLength;
	int chunkRead;

	String protocol;
	int    statusCode;
	String statusMessage;
	
	URL url;
	boolean allRead;

	Response (URL url) throws IOException {
	    int i = 0;
	    
	    this.url = url;

	    headerStarts = new int[64];
	    headerLens   = new int[64];

	    headerLen = readHeaders (buf);

	    debug ("Response from: " + url);
	    debug ("Headers: " + new String (buf, 0, headerLen));

	    extraStart = headerLen;
	    extraLen   = readCount - extraStart;

	    allRead = false;

	    for (i = 0;  i < headerLen;  i++)
		buf[i] = (byte) Character.toLowerCase ((char)buf[i]);
	    
	    
	    for (i = 0;  i < headerLens[0];  i++) {
		if (buf[i] == ' ') {
		    protocol = new String (buf, 0, i);
		    
		    break;
		}
	    }

	    if (i == headerLens[0]) {
		disconnect ();
		throw new IOException ("Status line too short.");
	    }

	    for (i += 1; i < headerLens[0];  i++)
		if (buf[i] == ' ')
		    break;
		else
		    statusCode = statusCode * 10 + (buf[i] - (byte)'0');
	    
	    if (i+1 < headerLens[0])
		statusMessage = new String (buf, i+1, headerLens[0]-i-1);
		
	    
	    contentLen = getIntHeader ("content-length", -1);
	    
	    String transferEncoding = 
		getHeaderValue ("transfer-encoding");
	    String connection =
		getHeaderValue ("connection");
	    	    
	    if (transferEncoding == null) {
		chunked = false;
	    } else if (transferEncoding.startsWith ("chunked")) {
		chunked = true;
		moreChunks = true;
		inChunk = false;
		contentLen = -1;
	    } else {
		throw new IOException ("Unknown TransferEncoding " + 
				       transferEncoding);
	    }
	    
	    if (connection == null) {
		if (protocol.startsWith ("http/1.1")) 
		    keepAlive = true;
		else
		    keepAlive = false; 
		
	    } else if (connection.startsWith ("keep-alive")) {
		String options = getHeaderValue ("keep-alive");

                if (DEBUG)
                    System.out.println ("keep-alive header for " + url + ":");
				    
		keepAlive = true;
		
		if (options == null) {
		    maxRequests = getIntHeaderOption (options, 
						      "max", 
						      Integer.MAX_VALUE);
		    timeout     = getIntHeaderOption (options, 
						      "timeout", 
						      Integer.MAX_VALUE);
		    if (timeout != Integer.MAX_VALUE)
			timeout *= 1000;
                    if (DEBUG) {
		        System.out.println ("maxRequests = " + maxRequests);
		        System.out.println ("timeout = " + timeout);
                    }
		} 
		    
	    } else if (connection.startsWith ("close")) {
		keepAlive = false;
	    }

	    

	}

	int getIntHeaderOption (String headerValue, 
				String optionName, 
				int    deflt)
	{
	    int index = headerValue.indexOf (optionName);
	    int value = deflt;
	    
	    
	    if (index != -1 && headerValue.charAt (index + 1) == '=') {
		int  len = headerValue.length ();
		char c;

		for (int i = index+2;  i < len; i++) {
		    c = headerValue.charAt (i);
		    
		    if ( '0' <= c  &&  c <= '9' )
			value = value * 10 + (c - '0');
		    else
			break;
		}
	    } 

	    return value;

	}

	public int getStatusCode () {
	    return statusCode; 
	}

	public String getStatusMessage () {
	    return statusMessage;
	}

	public int getContentLength () {
	    return contentLen;
	}
	
	public int getIntHeader (String name, int defaultVal) {
	    String val = getHeaderValue (name);
	    
	    return (val == null ? defaultVal : Integer.parseInt (val));
	}
	
	
	class HeaderEnumeration implements Enumeration {
	    byte[] b;
	    int i, j;
	    String next = null;
	    
	    
	    HeaderEnumeration (String name) {
		b = name.toLowerCase ().getBytes ();
		i = 0;
	    }

	    public boolean hasMoreElements () {
		int headerStart;
		int j;

		for (;  i < headerCount;  i++) {

		    headerStart = headerStarts[i];
		    
		    for (j = 0;  j < b.length;  j++) {
			if (b[j] != buf[headerStart+j])
			    break;
		    }

		    if (j == b.length) {
			next = new String (buf, 
					   headerStart+b.length+2,
					   headerLens[i]-b.length-2);
		    }
		}

		return false;
	    }

	    public Object nextElement () {
		return next;
	    }
	}
	
	public Enumeration getHeaderValues (String name) {
	    return new HeaderEnumeration (name);
	}

	public String getHeaderValue (String name) {
	    byte[] b = name.toLowerCase ().getBytes ();
	    int i, j;
	    int headerStart;
	    
	    for (i = 0;  i < headerCount;  i++) {

		headerStart = headerStarts[i];

		for (j = 0;  j < b.length;  j++) {

		    if (b[j] != buf[headerStart + j])
			break;
		}
		
		if (j == b.length) {
		    
		    return new String (buf,
				       headerStart+b.length+2,
				       headerLens[i]-b.length-2);

		}
	    }
	    
	    return null;
	}


	public InputStream getInputStream () {
	    return new ResponseInputStream (this);
	}

	/**
	 * Read headers into headerBuf.  If headers are longer than
	 * headerBuf.length (something like 8k), then truncate them
	 * with a newline.  Headers longer than 8k are gross.
	 */

	int readHeaders (byte[] headerBuf) throws IOException {
	    int r;
	    int pos = 0;
	    int q;
	    int headerStart = 0;
	    int countCR = 0;
	    int countLF = 0;
	    boolean pastHeader = false;


	    while (! pastHeader) {
		
		try {
		    r = in.read (headerBuf, 
				 readCount, 
				 headerBuf.length - readCount);
		    allRead = (r == -1);

		} catch (InterruptedIOException e) {
		    
		    debug ("Timed out trying to read " +
			   (headerBuf.length - readCount) +
			   " from " + url);

                    if (DEBUG)
		        System.out.println ("Read timed out for " + url + ":"
					    + "\n connected " + connected
					    + "\n already read " + readCount 
					    + "\n was trying to read " + 
					    (headerBuf.length - readCount));

		    throw e;
		} catch (IOException e) {
		    
		    		
		    
		    throw e;
		}

		if (r < 0) {
		    System.out.println ("Read " + r);
		    System.out.println ("Was trying to read " +
					(headerBuf.length - readCount));
		    System.out.println ("Headers: " + new String (buf, 0, readCount));
		    throw new IOException ("Received incomplete http header.");
		}
		readCount += r;
		
		while (! pastHeader && pos < readCount) {
		    
		    if (headerBuf[pos] == (byte)'\n') {
			countLF++;
		    } else if (headerBuf[pos] == (byte)'\r') {
			countCR++;
		    } else if (headerBuf[pos] != (byte)' ' && headerBuf[pos] != (byte)'\t') {
			countLF = 0;
			countCR = 0;
		    }
		    
		    if ((countCR == 1 && countLF == 1) &&
			headerCount < headerStarts.length) {
			
			headerStarts[headerCount] = headerStart;
			headerLens[headerCount++] = pos-headerStart-countCR-countLF+1;
			
			
			debug ("Header " + new String (headerBuf,
						       headerStart,
						       pos-headerStart-
						       countCR-countLF+1));
			
			headerStart = pos+1;
						
		    }
		    
		    pastHeader = ( (countCR == 2 && countLF == 0) ||
				   (countCR == 0 && countLF == 2) ||
				   (countCR == 2 && countLF == 2) );
		    
		    pos++;
		    
		}

		debug ("----------");
	    }
	    
	    q = pos - 1;
	    
	    while (((headerBuf[q] == '\n') || (headerBuf[q] == '\r') ||
		    (headerBuf[q] == ' ') || (headerBuf[q] == '\t')) &&
		   (q > 0)) {
		q--;
	    }
	    
	    headerBuf[q+1] = '\n';
	    
	    return pos;
	}

	
	int read (byte[] buf, int start, int len) throws IOException {
	    int r;
	    
	    if (chunked)
		r = readChunked (buf, start, len);
	    else
		r = readPlain (buf, start, len);
	    
	    allRead = (r == -1);

	    return r;
	}

	int read () throws IOException {
	    int r;
	    
	    if (chunked)
		r = readChunked ();
	    else
		r = readPlain ();

	    allRead = (r == -1);

	    return r;
	}

	
	final int readChunked () throws IOException {
	    
	    int r = -1;
	    
	    
	    if (! inChunk && moreChunks ) {

		chunkLength = readHexNumber ();
		debug ("Going to read chunk of length " + 
		       chunkLength + " from " + url);
		// need to also read optional chunk extention, but whatever
		readUntilNewline ();
		
		if (chunkLength == 0) {
		    
		    // read the optional trailer headers
		    // followed by a newline
		    while ( readUntilNewline () > 2)
			;
		    inChunk = false;
		    moreChunks = false;
		    
		    return -1;
		} else {
		    inChunk = true;
		    chunkRead = 0;
		}
	    }


	    if ( inChunk ) {
		
		r = readPlain ();
		
		chunkRead ++;
		
		if (chunkRead == chunkLength) {
		    readUntilNewline ();
		    inChunk = false;
		}
	    }

	    return r;
	}


	final int readPlain () throws IOException {
	    
	    int b;

	    
	    if (extraLen > 0) {
		
		b = buf[extraStart];
		
		extraStart++;
		extraLen--;
	    } else {
		
		b = in.read ();
	    }
	    
	    allRead = (b == -1);
	    
	    // System.out.println ("plain " + url + (b == -1 ? " -1" : " " + (char) b));

	    return b;
	}

	final int readPlain (byte[] toBuf, int start, int len) 
	    throws IOException
	{
	    int copyCount = 0;
	    
	    // System.out.println ("plain " + url);
	    
	    if (contentLen >= 0) {
		if (contentLen == readBody)
		    return -1;
		if (contentLen < len)
		    len = contentLen;
	    }

	    if (extraLen > 0) {
		
		copyCount = Math.min (extraLen, len);
		
		System.arraycopy (buf,   extraStart,
				  toBuf, start,
				  copyCount);

		extraStart += copyCount;
		extraLen   -= copyCount;
		readBody   += copyCount;
	    }
	    
	    
	    int r;
	    
	    if (copyCount < len) {
		
		r = in.read (toBuf, start + copyCount, len - copyCount);
		
		if (r > 0) {
		    readCount += r;
		    readBody += r;
		    return copyCount + r;
		} else {
		    return (copyCount == 0 ? -1 : copyCount);
		}
	    } else {
		return (copyCount == 0 ? -1 : copyCount);
	    }
	    
	}



	int readChunked (byte[] toBuf, int start, int len) throws IOException {
	    int read = -1;
	    int lineLen;
	    debug ("readChunked: start " + len); 
	    if (! inChunk && moreChunks ) {

		chunkLength = readHexNumber ();
		// System.out.println ("chunk " + url);
		// need to also read optional chunk extention, but whatever
		readUntilNewline ();
		debug ("readChunked: found chunk of " + chunkLength);	
		if (chunkLength == 0) {
		    
		    // read the optional trailer headers
		    // followed by a newline
		    while ( readUntilNewline () > 2)
			;
		    inChunk = false;
		    moreChunks = false;
		    
		    return -1;
		} else {
		    inChunk = true;
		    chunkRead = 0;
		}
	    }


	    if ( inChunk ) {
		
		int copyCount = 0;
		int r;
		
		len = Math.min (chunkLength-chunkRead, len);
		
		if (extraLen > 0) {
		    copyCount = Math.min (extraLen, len);
		    System.arraycopy (buf, extraStart, toBuf, start, copyCount);
		    extraStart += copyCount;
		    extraLen -= copyCount;
		}
		debug ("readChunked: copied " + copyCount + " bytes.");	
		r = in.read (toBuf, start+copyCount, len-copyCount);
		
		if (r == -1)
		    read = copyCount;
		else
		    read = copyCount + r;
		chunkRead += read;    
		
		if (chunkRead == chunkLength) {
		    readUntilNewline ();
		    inChunk = false;
		}

		
	    }
	    
	    return read;
	}

	
	/** 
	 * Returns the number of characters read when a newline is found
	 * (up-to and including the newlline).
	 */
	int readUntilNewline () throws IOException {
	    int b;
	    int c = 0;
	    
	    while ( (b = readPlain ()) != -1) {
		c++;
		if (b == '\n')
		    return c;
	    }

	    return -1;
	}

	int readHexNumber () throws IOException {
	    
	    int num = 0;
	    int b;
	    
	    while (( b = readPlain ()) != -1) {
		
		if (Character.isDigit ((char) b)) {
		    num = num * 16 + (b - (byte) '0');
		} else if (b >= 'a' && b <= 'f') {
		    num = num * 16 + 10 + (b - (byte) 'a');
		} else {
		    break;
		}
		
	    }
	 
	    return num;
	} 
	
	public void done () {
	    
	    debug ("Done with " + url);
	    lastRequest = System.currentTimeMillis ();
	    
	    try { // try to read whatever remains of the response if
		// the user code didn't read all of it.
		debug ("Trying to read whatever the user didn't read.");
		while (! allRead) {
		    int r = read (buf, 0, buf.length);
		    if (r == -1)
			break;
		}
		debug ("Done reading whatever user didn't read.");
	    } catch (IOException e) {

	    }

	    if (! keepAlive || maxRequests <= 0) {
		debug ("Disconnecting because keepalive = " + keepAlive +
		       "and max = " + maxRequests); 
		HttpConnection.this.disconnect ();
		
	    }

	    synchronized (HttpConnection.this) {
		HttpConnection.this.inRequest = false;
		debug ("Notifying HttpConnectionService.");
		if (HttpConnection.this.service != null)
		    HttpConnection.this.service.
			doneWithRequest (HttpConnection.this);
		
		HttpConnection.this.notify ();
	    }
	    debug ("Done.");
	}
    }

    
    public HttpConnection (URL url, 
			   int readTimeout, 
			   HttpConnectionService service)
    {
	this.hostURL = url;
	this.host = url.getHost ();
	this.port = (url.getPort () == -1 ? 80 : url.getPort ());
	this.readTimeout = readTimeout;
	this.service = service;
       
	keepAlive   = true;
	maxRequests = Integer.MAX_VALUE;
	timeout     = Integer.MAX_VALUE;
	connected   = false;
	inRequest   = false;
    }
	
	
    public URL getHostURL () {
	return hostURL;
    }

    void connect () throws IOException {
	long start = System.currentTimeMillis ();
	
	s = new Socket (host, port);
	
	try {
	    s.setSoTimeout (readTimeout);
	    s.setSoLinger (false, 0);
	    s.setKeepAlive (true);
	} catch (SocketException e) {
	    try { s.close (); }
	    catch (IOException e2) { }
	    throw e;
	}

	/* if there is a failure creating one of the descriptors,
	   make sure we close everything already opened before
	   we rethrow the exception.
	*/

	try {
	    in  = new BufferedInputStream (s.getInputStream (), 1024);
	    
	} catch (IOException e) {
	    try { s.close (); }
	    catch (IOException e2) { }
	    throw e;
	}

	try {
	    sout = s.getOutputStream ();
	    out  = new BufferedOutputStream (sout, 8192);
	    connected = true;
	} catch (IOException e) {
	    try { in.close (); }
	    catch (IOException e2) { }
	    try { s.close (); }
	    catch (IOException e2) { }
	    throw e;
	}

	cost = System.currentTimeMillis () - start;
	
    }

    /**
     * Get the cost of establishing the connection to this host.
     */

    public int getCost () {
	return (int) cost;
    }

    /**
     * LRUCache.Cacheable implementation - this method is called
     * when the object is deleted from the cache.
     */

    public void deleted () {
	keepAlive = false;
    }

    public void disconnect () {

	synchronized (this) {
	    waitingRequestorCount--;
	    notifyAll ();
	    inRequest = false;
	}

	/* close all the descriptors */
	try {
	    in.close (); 
	} catch (IOException e) {

	}

	try {
	    out.close ();
	} catch (IOException e) {

	}
	
	try {
	    s.close ();
	} catch (IOException e) {

	}

	connected = false;

	if (service != null)
	    service.connectionClosed (this);
    }

    public synchronized boolean isInRequest () {
	return inRequest;
    }

    public synchronized int getWaitingRequestorCount () {
	return waitingRequestorCount;
    }

    public synchronized Request beginRequest (URL url, String method) 
	throws IOException, InterruptedException
    {
	boolean reconnect = false;
	
	debug ("beginRequest: begin.");
	try {
	    waitingRequestorCount++;
	    
	    while (inRequest) {
		debug ("inRequest, so waiting.");
		wait ();
	    }

	    if (keepAlive) {
		long now = System.currentTimeMillis ();
		
		if ( (timeout != -1 && lastRequest + timeout > now) ||
		     maxRequests <= 0)
		    
		    reconnect = true;
		    
	    }

	    inRequest = true;
	    maxRequests --;
	    
	    if (! connected || reconnect) {
		connect ();
	    }

	    try {
		
		return new Request (url, method);
	    
	    } catch (IOException e) {
		e.printStackTrace ();
		debug ("************ new Request failed.");
		connected = false;
		connect ();
		return new Request (url, method);
	    }

	} catch (InterruptedException e) {
	    disconnect ();
	    throw e;
	} finally {
	    waitingRequestorCount--;
	}
	
    }



    class RequestOutputStream extends OutputStream {

	Request r;

	RequestOutputStream (Request r) {
	    this.r = r;
	}

	public void close () throws IOException {
	    out.flush ();
	}

	public void flush () throws IOException {
	    out.flush ();
	}

	public void write (byte[] b) throws IOException {
	    r.writeBody (b, 0, b.length);
	}

	public void write (byte[] b, int start, int len) throws IOException {
	    r.writeBody (b, start, len);
	}

	public void write (int b) throws IOException {
	    r.writeBody (b);
	}
    }

    

    class ResponseInputStream extends InputStream {

	Response r;
		
	ResponseInputStream (Response r) {
	    this.r = r;
	}

	public int available () throws IOException {
	    return r.extraLen + in.available ();
	}

	public void close () throws IOException {
	    r.done ();
	}

	public void mark (int readLimit) {
	    
	}

	public boolean markSupported () {
	    return false;
	}

	public int read () throws IOException {
	    
	    if (DEBUG)
		System.out.println ("want" + r.url);
	    
	    int c = r.read ();

	    if (DEBUG)
		System.out.println ("read " + r.url + ": " + (char) c);
	    
	    return c;
	}

	public int read (byte[] b) throws IOException {
	    
	    return r.read (b, 0, b.length);

	}

	public int read (byte[] b, int start, int len) throws IOException {

	    if (DEBUG)
		System.out.println ("want " + r.url + ": " + len + " bytes.");
	    
	    int readLen = r.read (b, start, len);
	    

	    if (DEBUG)
		System.out.println ("read " + r.url + ": " + readLen + 
				    " bytes.");
	    // new String (b, start, readLen));

	    return readLen;
	}

	public void reset () {

	}
	
    }

    private static final void debug (String msg) {
	if (DEBUG)
	    System.out.println (msg);
    }
    
    public static void main (String[] argv) 
	throws IOException, InterruptedException
    {
	String query = ("<?xml version='1.0'?>\r\n" +
			"<query>\r\n" +
			"  <text>\r\n" +
			"      xcf\r\n" +
			"  </text>\r\n" +
			"</query>");
			
	for (int i = 1;  i < argv.length;  i++) {
	    URL url = new URL (argv[i]);
	    
	    HttpConnection con = new HttpConnection (url, 30*1000, null);
	    
	    Request  req;
	    Response res;
	    byte[] buf = new byte[8192];
	    int r;
	    
	    for (int j = 0;  j < Integer.parseInt (argv[0]);  j++) {
		req = con.beginRequest (url, "POST");
		req.setHeader ("Content-type", "text/xml");
		req.setHeader ("Content-length", 
			       String.valueOf (query.length () + 0));
		// req.flushHeaders ();
		req.getOutputStream ().write (query.getBytes ());
                if (DEBUG) {
		    System.out.print ("\r\n");
		    System.out.println (query);
                }

		res = req.getResponse ();
		if (DEBUG)
		    System.out.println ("Response: " + res.getStatusCode () + " " +
					res.getStatusMessage ());
		
		while ( ( r = res.read (buf, 0, 8192)) != -1) {
		    if (DEBUG)
		        System.out.println (new String (buf, 0, r));
		}
		
		res.done ();
	    }

	    
	    
	}
    }

}
