/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: HttpConnectionService.java,v 1.1.1.1 2001/06/01 16:51:19 jbeatty Exp $
 */
package net.jxta.search.util;


import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;

/**
 * HttpConnectionService maintains a cache of open connections, so
 * that repeated http client connections to other hosts don't result
 * in needless tcp open/closes.  This is possible because the 
 * HttpConnection class is good at implementing the necessary
 * HTTP/1.1 semantics like keep-alive and chunking.  
 * <p>
 * (java.net.HttpURLConnection doesn't support chunked encoding, 
 * while most server-side page-generation tools use it.);
 *
 * Usage:
 * <pre>
 * HttpConnectionService service = new HttpConnectionService (null);
 * HttpConnection.Request req;
 * HttpConnection.Response res;
 * InputStream in;
 *
 * req =  service.beginRequest (new URL ("http://www.ibm.com/products/"),
 *                              "GET",
 *                              5 * 100);
 * req.setHeader ("Foo", "bar");
 *
 * res = req.getResponse ();
 *
 * try {
 *   in = req.getInputStream ();
 *  
 *   // read from the stream 		       
 * } finally {
 *   res.done ();
 * }
 * </pre>
 */

public class HttpConnectionService {

    LRUCache cache;
    HashMap  level2;
    HashMap  dontCacheHosts;

    static final int CLEANUP_INTERVAL = 30 * 1000;
    static final int MAX_SIMULTANEOUS_REQUESTS = 5;

    /**
     * @param HashMap table of hosts connections to which should not
     *                be cached or null if all connections should be cached.
     */
    public HttpConnectionService (HashMap dontCacheHosts) {
	
	this.dontCacheHosts = dontCacheHosts;

	cache = new LRUCache (1024);
	level2 = new HashMap (1024);
	
    }

    
    public HttpConnection.Request beginRequest (URL    url, 
						String method,
						int    readTimeout) 
	throws IOException, InterruptedException
    {
	URL hostUrl = new URL (url.getProtocol (), 
			       url.getHost (), 
			       (url.getPort () == -1 ? 80 : url.getPort ()), 
			       "/");
	HttpConnection con;
	
	if (dontCacheHosts != null && dontCacheHosts.get (hostUrl) != null) {
	    con = new HttpConnection (hostUrl, readTimeout, this);
	} else {

	    con = (HttpConnection) level2.get (hostUrl);
	    
	    /* if it's in level2 cache, then it's busy... */
	    
	    
	    if (con == null) {
		
		con = (HttpConnection) cache.get (hostUrl);
		
		if (con == null) {
		    
		    con = new HttpConnection (hostUrl, readTimeout, this);
		    
		    cache.put (hostUrl, con);
		} else if (con.getWaitingRequestorCount() > MAX_SIMULTANEOUS_REQUESTS) {
		    con = new HttpConnection (hostUrl, readTimeout, this);
		} else {
		    // System.out.println ("Re-using connection to " + url);
		}
		
		level2.put (hostUrl, con);
	    } else {
		// System.out.println ("Reusing the connection from level2.");
	    }
	    
	}
	
	return con.beginRequest (url, method);
    }


    void doneWithRequest (HttpConnection con) {
	/* put it back in the cache */
	level2.remove (con.getHostURL ());
	// cache.put (con.getHostURL (), con);
    }

    void connectionClosed (HttpConnection con) {
	/* remove from the cache */
	level2.remove (con.getHostURL ());
	cache.get (con.getHostURL ());
    }

    

    /*
    class CleanupAction implements SchedulerService.Action {

	public void perform (SchedulerService ss) {
	    
	    Iterator i = connectionTable.entrySet ().iterator ();
	    HttpConnection c;
	    
	    while (i.hasNext ()) {
		c = (HttpConnection) i.next ();
		
		synchronized (c) {
		    if (! c.isInRequest ()) {
			c.disconnect ();
			connectionTable.remove (c.getHostURL ());
		    }

		    
		}

		
		ss.scheduleAction (this, CLEANUP_INTERVAL);
	    }
	}
    }
    
    
    class DisconnectAction implements SchedulerService.Action {

	HttpConnection con;

	DisconnectAction (HttpConnection con) {
	    this.con = con;
	}

	public void perform (SchedulerService ss) {
	    con.disconnect ();
	}
    }
    */

}

