/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: JxtaPipeUtil.java,v 1.1.1.1 2001/06/01 16:51:22 jbeatty Exp $
 */
package net.jxta.search.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.IOException;

import java.lang.reflect.InvocationTargetException;

import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.PipeID;
import net.jxta.pipe.Pipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.MimeMediaType;
import net.jxta.document.Document;
import net.jxta.exception.JxtaException;

import net.jxta.search.jxtastarter.JxtaStarter;

/** Aids in pipe creation **/
public class JxtaPipeUtil {

    /** the mime media type we use for pipe advertisements **/
    private static final MimeMediaType MIME_TYPE =
        new MimeMediaType( "text/xml" );

    /** Given a file containing a pipe advertisement, returns the a
        that advertisement.  If the file does not yet exist,
        a new pipe advertisement is created and written to the file.
        This is very useful for always keeping the same pipe adv for your
        input pipe in your application or service. **/
    public static PipeAdvertisement createPipeAdv( File pipeAdvFile,
						   String pipeName,
						   PeerGroup peerGroup )
    throws JxtaException, IOException {
        
        PipeAdvertisement pipeAdv = readPipeAdvFromFile( pipeAdvFile );
        if( pipeAdv == null ) {
            pipeAdv = createPipeAdv( pipeName, peerGroup );
            writePipeAdvToFile( pipeAdv, pipeAdvFile );
        }

        return pipeAdv;
    }


    /** create a new pipeadv given a pipe name**/
    private static PipeAdvertisement createPipeAdv( String pipeName,
						    PeerGroup peerGroup )
    throws JxtaException {

        PipeAdvertisement pipeadv;
        try {
            pipeadv = (PipeAdvertisement)
                AdvertisementFactory.newAdvertisement(
                    PipeAdvertisement.getAdvertisementType() );
        } catch( InvocationTargetException e ) {
            throw new JxtaException(
                "Cannot create pipe advertisement: " + e );
        }

        PipeID pipeid = new PipeID( peerGroup.getID() );
        pipeadv.setPipeID( pipeid );
        pipeadv.setName( pipeName );

        return pipeadv;
    }


    /** tries reading the pipe advertisement from a file. If it
        does not exist or is invalid, will return null. **/
    private static PipeAdvertisement readPipeAdvFromFile( File file )
    throws IOException, JxtaException {

            PipeAdvertisement pipeadv = null;

            // if the file exists, use it
            if( file.exists() ) {
                InputStream is = new FileInputStream( file );
                
                try {
                    pipeadv = (PipeAdvertisement)
                        AdvertisementFactory.newAdvertisement( MIME_TYPE, is );
                } catch( InvocationTargetException e ) {
                    throw new JxtaException( "Could not create advertisement" + e );
                }

                is.close();
            }

            return pipeadv;
    }


    /** writes the specified pipeadv to a file **/
    private static void writePipeAdvToFile( PipeAdvertisement pipeadv,
					    File pipeAdvFile )
    throws JxtaException, IOException {        
        
        Document pipeadvdoc = null;
        
        try { 
            pipeadvdoc = pipeadv.getDocument( MIME_TYPE ); 
        } catch( Exception e ) {
            throw new JxtaException( "Could not grab document out of " +
                                     "advertisement." );
                                     
        }

        OutputStream os = new FileOutputStream( pipeAdvFile );
        pipeadvdoc.sendToStream( os );
        os.close();
    }

    private static void main( String[] args ) throws Exception {
	PeerGroup group = JxtaStarter.getMyPeerGroup();
	test( group );
    }

    /** Exercises the persistent pipe advertisement methods. Creates
	an adv, writes it to disk, reads it again, and ensures that
	the original pipeid equals the newly read pipe id **/
    private static void test( PeerGroup group ) throws Exception {
	PipeAdvertisement adv = createPipeAdv( "My Pipe", group );
	if( adv == null ) throw new Exception( "Can't create adv" );
	File f = new File( "test.xml" );
	writePipeAdvToFile( adv, f );
	PipeAdvertisement adv2 = readPipeAdvFromFile( f );
	if( adv.getPipeID().equals( adv2.getPipeID() ) )
	    System.out.println( "success" );
	else
	    System.out.println( "fail" );

    }
}
