/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: KMP.java,v 1.1.1.1 2001/06/01 16:51:22 jbeatty Exp $
 */
package net.jxta.search.util;

/**
 * Knuth-Morris-Pratt string-matching algorithm.
 * Usage:
 * <pre>
 * StringMatcher kmp = new KMP ();
 * 
 * StringMatcher.CompiledPattern pattern = kmp.compile (searchPhrase);
 * 
 * int index = bm.match (text, 0, pattern);
 * 
 * System.out.println ("First location of " + searchPhrase + " at " + index);
 * 
 * </pre>
 */
public class KMP implements StringMatcher {


    public static void main (String[] argv) {
	
	String pattern = argv[0];
	KMP kmp = new KMP ();
	StringMatcher.CompiledPattern compiledPattern = kmp.compile (pattern);
		
	for (int i = 1;  i < argv.length;  i++)
	    System.out.println ("Pattern occurs with shift " + 
				kmp.match (argv[i].toCharArray (), 
					   0,
					   compiledPattern) +
				" in: " + argv[i]);;
    }

    public StringMatcher.CompiledPattern compile (String pattern) {
	
	return new CompiledPatternImpl (pattern);
    }

    public int   match (char[]          text, 
			int             textStart, 
			StringMatcher.CompiledPattern compiledPattern)
    {
	CompiledPatternImpl impl = (CompiledPatternImpl) compiledPattern;
	char[] pattern = impl.chars;
	int[]  prefixFunc = impl.prefixFunc;
	int n = text.length;
	int m = pattern.length;
	int q = 0;
	int shift = -1;
	

	for (int i = textStart;  i < n;  i++) {
	    
	    while (q > 0  &&  ! charEquals (pattern[q], text[i])) {
	    // while (q > 0  &&  pattern[q] != text[i]) {

		q = prefixFunc[q];
	    }

	    if (charEquals (pattern[q], text[i])) 
	    // if (pattern[q] == text[i]) 
		q ++;

	    if (q == m) {
		shift = i - m + 1;
		break;
	    }
	}

	return shift;
    }


    
    private static class CompiledPatternImpl 
	implements StringMatcher.CompiledPattern
    {
	char[] chars;
	int[]  prefixFunc;
	
	CompiledPatternImpl (String pattern) {
	    this.chars      = pattern.toCharArray ();
	    this.prefixFunc = computePrefixFunction (chars);
	}
    
	private int[] computePrefixFunction (char[] pattern) {
	    
	    int[] func = new int[pattern.length];
	    
	    int   m    = pattern.length;
	    
	    func[0] = 0;
	    
	    int k = 0;
	    
	    for (int q = 1;  q < m;  q++) {
		
		while (k > 0 &&  ! charEquals (pattern[k], pattern[q])) {
		    // while (k > 0  &&  pattern[k] != pattern[q]) {
		    
		    k = func[k];
		}
		
		if (charEquals (pattern[k], pattern[q])) {
		    // if (pattern[k] == pattern[q]) {
		    k ++;
		}
		
		func[q] = k;
	    }
	    
	    return func;
	}
    
    }
    
    static final boolean charEquals (char patternChar, char textChar) {
	
	return ( (patternChar == textChar) ||
		 (isSpace (patternChar) && isSpace (textChar)) );
	
    }
    
    
    static final boolean isSpace (char c) {
	return Character.isWhitespace (c) || c == '>' || c == '<' || 
	    c == '\"' || c == '.' || c == ',';
    }

}
