/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: LRUCache.java,v 1.2 2001/06/04 02:43:35 yaroslav Exp $
 */
package net.jxta.search.util;


import java.util.*;

/**
 * A cache that replaces items starting with the recently used one.
 * 
 * Objects that are placed in the cache must be of a type that 
 * implements the LRUCache.Cacheable interface.  The interface
 * allows the cache to determine how expensive the object was
 * to make (and hence how expensive it will be to re-make, and
 * also allows the object to be notified when it is being removed
 * from the cache.
 *
 * @see LRUCache.Cacheable
 * @see HttpConnection
 */

class LRUCache {

    HashMap map;
    Queue   lru;
    int     total;
    int     max;
    
    LRUCache (int size) {
	map = new HashMap (size * 2);
	lru = new Queue ();
	max = size;
    }


    /**
     * Objects that are placed in the cache must be of a type that
     * implements this interface.
     */
    
    interface Cacheable {
	
	/**
	 * Return the cost of re-making this object.
	 */
	
	int  getCost ();

	/**
	 * Notify the object that it is being removed from the cache.
	 */

	void deleted ();
    }

    /**
     * Get an object from the cache that corresponds to the given key.
     */

    synchronized Cacheable get (Object key) {
	Item i = (Item) map.get (key);

	if (i == null) {
	    return null;
	} else {

	    total --;
	    map.remove (key);
	    lru.remove (i);

	    return i.o;
	}
    }

    /**
     * Place an object in the cache under a given key.
     * This operation may result in some of the already-stored
     * objects to be deleted.
     */

    synchronized void put (Object key, Cacheable o) {
	Item i = new Item (o);

	if (total == max) {
	    /* make space for it */
	    Item leastCostOne = lru.head;
	    Item cur = lru.head;

	    /* 16 is the number of items to scan before deciding
	       which item is the least costly to replace. */
	    
	    for (int j = 1;  cur != null && j < 16;  cur = cur.next, j++) {
		if (leastCostOne.o.getCost () > cur.o.getCost ())
		    leastCostOne = cur;
	    }

	    /* tell old.o object to close up shop */
	    leastCostOne.o.deleted ();
	    
	}

	lru.enqueue (i);
	
	map.put (key, i);

	total ++;
    }

    class Item {

	Item prev;
	Item next;
    
	Cacheable o;

	Item (Cacheable o) {
	    this.o = o;
	}
    }


    class Queue {

	Item head;
	Item tail;


	void enqueue (Item i) {
	    if (head == null) {
			head = i;
			tail = i;
	    } else {
			tail.next = i;
			i.prev = tail;
			tail = i;
	    }
	}

	Item dequeue () {
	    
	    Item i = head;
	    
	    i.prev = null;
	    i.next = null;
	    
	    head.prev = null;

	    return i;
	}

	void remove (Item i) {
	    
	    
	    if (i == head) {
			head = head.next;
			
			if (head == null)
				tail = null;
			else
				head.prev = null;
	    } else if (i == tail) {
			tail = tail.prev;
			
			tail.next = null;
	    } else {
			i.prev.next = i.next;
			i.next.prev = i.prev;
	    }
		
	    
	    i.prev = null;
	    i.next = null;
	    
	}
    }
}
