/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: LogService.java,v 1.1.1.1 2001/06/01 16:51:23 jbeatty Exp $
 */
package net.jxta.search.util;

import java.io.*;
import java.util.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

public class LogService {

    public static final int ERROR   = 0x1;
    public static final int WARNING = 0x2;
    public static final int DEBUG   = 0x3;
    public static final int INFO    = 0x4;
    public static final int ALL     = ERROR | WARNING | DEBUG | INFO;


    String       id;
    File         file;
    OutputStream os;
    
    int          levelMask;
    long         rolloverPeriod;
    long         lastRolloverTime;
    
    DateFormat   entryTimestamp;
    DateFormat   fileTimestamp;
    Date         date;
    Object       pendingAction;
    
    
    private static byte[] lineSeparator =  
	System.getProperty ("line.separator").getBytes ();
    private static SchedulerService ss = new SchedulerService ();
    
    static {
	new Thread (ss).start ();
    }
    

    

    /**
     * Create a log that will roll over every day.
     */
    
    public LogService (String id, int levelMask) {
	
	this (id, levelMask, 24 * 60 * 60 * 1000, "yyyyMMdd");
    }
    
    
    /**
     * Create a log service.
     */

    public LogService (String id, 
		       int    levelMask, 
		       long   rolloverPeriod,
		       String fileTimestampFormat) 
    {
	File logDir = new File ("log/");
	
	this.id        = id;
	this.levelMask = levelMask;
	this.lastRolloverTime = System.currentTimeMillis ();
	this.date = new Date ();

	if (rolloverPeriod > 0)
	    this.rolloverPeriod = rolloverPeriod;
	else
	    this.rolloverPeriod = -1;

	fileTimestamp  = new SimpleDateFormat (fileTimestampFormat);
	entryTimestamp = new SimpleDateFormat ("MMM_dd_yyyy_HH:mm:ss");

	if (! logDir.exists ())
	    logDir.mkdirs ();
	
	file = new File (logDir, id + ".log");
		
	try {
	    os = new FileOutputStream (file.getAbsolutePath (), true); 
	} catch (IOException e) {
	    e.printStackTrace ();
	}

	
	if (rolloverPeriod > 0) {
	    long next = lastRolloverTime - 
		lastRolloverTime % rolloverPeriod + rolloverPeriod;
	    
	    System.out.println ("> Scheduling first log rollover to " +
				"occur at " + new Date (next));
	    
	    pendingAction = ss.scheduleAction (new RolloverAction (),
					       next - System.currentTimeMillis ());
	    
	 }
    }


    private boolean shouldRollover (long now) {
	return  ( (rolloverPeriod > 0) && 
		  (lastRolloverTime == 0 || 
		   now > lastRolloverTime + rolloverPeriod) );
    }
    
    /**
     * Check to see if we need to rollover the log.
     *
     * @param now current time in milliseconds.
     */
    public synchronized void rollover (boolean force) {
	
	
	Calendar local = Calendar.getInstance ();
	Calendar gmt   = Calendar.getInstance (TimeZone.getTimeZone ("GMT"));
		
	
	if (force || shouldRollover (local.getTime ().getTime ())) {
	
	    lastRolloverTime = local.getTime ().getTime ();


	    try {
		os.close();
		
	    } catch (IOException e) {
		/* this is irrelevant */
	    }
	    
	    os = null;
	    
	    file.renameTo (new File ("log/" + id + ".log." +
				     fileTimestamp.format (new Date
					 (gmt.getTime ().getTime () -
					  rolloverPeriod))));
	    try {
		os = new FileOutputStream (file = new File ("log/" + id + 
							    ".log"));
	    } catch (IOException e) {
		System.out.println ("****** Could not open log file " + file);
		e.printStackTrace ();
	    }
	}
    }
    
    public synchronized void log (int    level, 
				  String msg, 
				  String providerId, 
				  String msgId) 
    {
	
	if ( (levelMask & level) == level ) {
	    
	    String line = /* "" + System.currentTimeMillis () + */ " " + 
		getLevelName (level) + " " + providerId + " " + msgId;

	    
	    log (line);

	    if (level == ERROR)
		System.err.println (line);
	}
    }

    public synchronized void log (int level, String uuid, String msg) {

	if ( (levelMask & level) == level ) {
	    
	    String line = /* "" + System.currentTimeMillis () + */ " " + 
		getLevelName (level) + " " + uuid + " " + msg;

	    log (line);

	    if (level == ERROR) {
		System.err.println (line);
	    }
	}
    }

    
    public synchronized void log (String uuid, String msg, Exception e) {

	CharArrayWriter c = new CharArrayWriter ();
	
	e.printStackTrace (new PrintWriter (c));
	
	log (ERROR, uuid, msg + "\n" + c.toString ());
    }

    

    private static final String getLevelName (int level) {
	String s;

	switch (level) {
	case ERROR:   s = "ERROR"; break;
	case WARNING: s = "WARNING"; break;
	case DEBUG:   s = "DEBUG"; break;
	case INFO:    s = "INFO"; break;
	default:      s = "UNKNOWN"; break;
	}

	return s;
    }

    

    void log (String s) {
	byte[] b = s.getBytes ();
	
	try {
	    write (b, 0, b.length);
	} catch (IOException e) {
	    e.printStackTrace ();
	}
    } 



    void write (byte[] buf, int offset, int length) 
	throws IOException
    {
	// synchronized (os) {
	byte[] timestamp;

	date.setTime (System.currentTimeMillis ());
	timestamp = entryTimestamp.format (date).getBytes ();
	os.write(timestamp, 0, timestamp.length);
	    
	os.write (buf, offset, length); 
	os.write (lineSeparator, 0, lineSeparator.length);
	os.flush ();
	// }
    }


    
    class RolloverAction implements SchedulerService.Action {

	public void perform (SchedulerService ss) {
	    
	    synchronized (LogService.this) {
		long now = System.currentTimeMillis ();
		long next = now - (now % rolloverPeriod) + rolloverPeriod;
		
		
		System.out.println ("> Doing " + id + " log rollover, as " +
				    "scheduled to be done every " + 
				    fileTimestamp.format (new Date
					(rolloverPeriod)));
		
		System.out.println ("> Next " + id + " log rollover will " +
				    " occur at " + 
				    new Date (next));
		
		rollover (false);
		
		pendingAction = ss.scheduleAction (this, next - now);
		
	    }
	}

    }


}
