/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: MemoryStream.java,v 1.3 2001/06/01 21:21:26 tcamarda Exp $
 */
package net.jxta.search.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

public class MemoryStream extends OutputStream {
    
    Chunk head;
    Chunk tail;
    static CircularList freeChunks = new CircularList (256);
    
    static {
	
	/* add chunks to the list of free chunks */
	for (int i = 0;  i < 256;  i++)
	    freeChunks.enqueue (new Chunk ());
	
    }
    
    static class Chunk {
	byte[] buf;
	int    start;
	int    length;
	
	Chunk next;

	Chunk () {
	    buf = new byte[8192];
	}

	void init () {
	    start  = 0;
	    length = 0;
	    next   = null;
	}

	void writeToStream (OutputStream out) throws IOException {
	    out.write (buf, start, length);
	}
    }

    public MemoryStream () {
	
    }

    public void writeToStream (OutputStream out) throws IOException {
	Chunk c = head;

	while (c != null) {
	    c.writeToStream (out);
	    
	    c = c.next;
	}
    }

    public InputStream getInputStream () {
	return new InputStream () {
		Chunk cur = head;
		int   pos = 0;
		
		public int read () throws IOException {
		    int b;
		    
		    if (cur == null) {
			b = -1;
		    } else if (pos == cur.length) {
			cur = cur.next;
			pos = 0;
			b = read ();
		    } else {
			b = (int) (0x000000FF & cur.buf[cur.start+pos]);
			pos++;
		    }

		    return b;
		}

		    
		public int read (byte[] buf) throws IOException {
		    return read (buf, 0, buf.length);
		}

		public int read (byte[] buf, int start, int len) 
		    throws IOException
		{
		    
		    if ( cur == null ) {

			return -1;

		    } else if ( pos == cur.length ) {
			cur = cur.next;
			pos = 0;
			return read (buf, start, len);
			
		    } else {
			int n = (len < cur.length-pos ? len : cur.length-pos);
			
			System.arraycopy (cur.buf, cur.start+pos, 
					  buf, start, n);

			pos += n;
			// we'll geto to next chunk at the next invocation
			return n;
		    }
		    
		}

		
		public int available () {
		    int av = 0;

		    if (cur != null) {
			av = cur.length - pos;
			cur = cur.next;
		    }
		    
		    while (cur != null) {
			
			av += cur.length;
			cur = cur.next;
		    }

		    return av;
		}
		
		public long skip (long toSkip) {
		    long skipped = 0;
		    
		    
		    while (cur != null && skipped < toSkip) {
			
			skipped += (toSkip - skipped > cur.length ? 
				    cur.length : toSkip - skipped);
			cur = cur.next;
		    }

		    return skipped;
		}

		public boolean markSupported () {
		    return false;
		}

		public void mark () { 

		}

		public void reset () {

		}

		public void close () {

		}
	    };
    }

    public void write (int b) throws IOException {
	
	if (tail != null  &&  tail.start + tail.length + 1 < tail.buf.length) {
	    tail.buf[tail.start + tail.length + 1] = (byte) (0x000000FF & b);
	    
	    tail.length += 1;
	} else {
	    addChunk ();

	    write (b);
	}
    }

    public void write (byte[] buf) throws IOException {
	write (buf, 0, buf.length);
    }

    public void write (byte[] buf, int offset, int length) throws IOException {

	if (length < 0)
	    throw new IllegalArgumentException ("length must be positive");
	if (offset < 0)
	    throw new IllegalArgumentException ("starting offset must be " +
						"positive");
	/*
	System.out.println ("" + this + ": " +  
			    new String (buf, offset, length));
	*/
	while (length > 0) {
	    if (tail == null || tail.length >= tail.buf.length)
		addChunk ();

	    int sublen = tail.buf.length - tail.length;

	    if (length < sublen)
		sublen = length;
	    
	    
	    System.arraycopy (buf, offset, tail.buf, tail.length, sublen);

	    offset += sublen;
	    length -= sublen;
	    tail.length += sublen;
	}
    }

    public void flush () throws IOException {
	/* no code intended */
    }


    public void close () throws IOException { 
	Chunk c = head;
	
	while (c != null) {
	    
	    freeChunks.enqueue (c);
	    
	    c = c.next;
	}

	head = tail = null;
    }

    public int getLength () {
	Chunk c = head;
	int   l = 0;
	
	while (c != null) {
	    l += c.length;
	    c = c.next;
	}

	return l;
    }
    
    
    private void addChunk () {
	Chunk c = (Chunk) freeChunks.getHead ();

	if (c == null)
	    c = new Chunk ();
	else
	    c.init ();
	
	if (head == null) {
	    head = tail = c;
	} else {
	    tail.next = c;
	    tail = c;
	}
	
    }


    public void finalize () {
	
	try {
	    close ();
	} catch (IOException e) {
	    /* no code intended */
	}
    }

    
    private static void testWrite (InputStream in)
	throws IOException
    {
	MemoryStream out    = new MemoryStream ();
	int          total = 0;
	int          b;
	
	try {
	    /* copy stream <em>in</em> into memory stream byte by byte */

	    while ((b = in.read ()) != -1) {
		out.write (b);
		total ++;
	    }
	    
	    if (out.getLength () == total)
		System.err.println ("MemoryStream.testWrite part 1: passed.");
	    else
		System.err.println ("MemoryStream.testWrite part 1x: failed.");

	    out.writeToStream (System.out);
	    
	    InputStream in2 = out.getInputStream ();
	    int n = 0;
	    while ((b = in2.read ()) != -1) {
		System.out.write (b);
		n++;
	    }
	    
	    
	    if (out.getLength () == n)
		System.err.println ("MemoryStream.testWrite part 2: passed.");
	    else
		System.err.println ("MemoryStream.testWrite part 2: failed.");
	    
	} finally {
	    out.close ();
	}

    }


    private static void testArrayWrite (InputStream  in, 
					int          bufSize) 
	throws IOException
    {
	MemoryStream out = new MemoryStream ();
	byte[] buf    = new byte[bufSize];
	int    total1 = 0;
	int    total2 = 0;
	int    n;
	
	try {
	    /* copy System.in into memory stream */
	    while ((n = in.read (buf)) != -1) {

		out.write (buf, 0, n);
	
		total1 += n;
	    }

	    /* flush memory stream to out */
	    out.writeToStream (System.out);
	    
	    if (out.getLength () == total1)
		System.err.println ("MemoryStream.testArrayWrite part 1 " +
				    " with block read size " + bufSize + 
				    ": passed.");
	    else
		System.err.println ("MemoryStream.testArrayWrite part 1 " +
				    " with block read size " + bufSize + 
				    ": failed.");

	    InputStream in2 = out.getInputStream ();

	    while ((n = in2.read (buf)) != -1) {
		System.out.write (buf, 0, n);
		total2 += n;
	    }
	    

	    if (out.getLength () == total2)
		System.err.println ("MemoryStream.testArrayWrite part 2 " +
				    " with block read size " + bufSize + 
				    ": passed.");
	    else
		System.err.println ("MemoryStream.testArrayWrite part 2 " +
				    " with block read size " + bufSize + 
				    ": failed.");

	    
	} finally {
	    out.close ();
	}
    }


    /**
     * How to test:
     *
     * Supply data on stdin that would be copied used testWrite
     * and supply a filename of the file that would be copied
     * several times to using testArrayWrite, once for each of
     * the test buffer sizes.
     * Redirect the output to a file that will contain all of
     * data supplied on stdin, plus the data from the file,
     * repeated once for each of the test buffer sizes.
     */
    
    public static void main (String[] argv) throws IOException {

	File file = new File (argv[0]);
	
	int[] testBufferSizes = { 256, 1024, 4092, 8192, 8192 * 2 };

	InputStream in = null;
	

	testWrite (System.in);
	
	for (int i = 0;  i < testBufferSizes.length;  i++) {
	    
	    try {
		in = new FileInputStream (file);
		
		testArrayWrite (in, testBufferSizes[i]);
	    } finally {
		if (in != null)
		    in.close ();
	    }
	}
    }


}
 
