/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: OPMPHF.java,v 1.1.1.1 2001/06/01 16:51:25 jbeatty Exp $
 */
package net.jxta.search.util;


import java.util.Random;


/**
 * Order-preserving minimal perfect hash function generator.
 * Usage:
 *<pre>
 OPMPHF func = new OPMPHF (n);

 do {
    func.beginGenerate ();

    for (int i = 0;  i < n;  i++)
       func.addWord (some word);

  } while (! func.endGenerate ());

  // obtain hash values by doing: 
  func.hash (some word);

 </pre>
 */

public class OPMPHF {

    int   n;
    int   m;
    int[] w1 = new int[128];
    int[] w2 = new int[128];
    int[] g;
    int[] h;
    ConnectionSet edges;
    int h_cur = 0;
    EdgeFuncImpl edgeFunc = new EdgeFuncImpl ();
    int lexiconTotal;
    
    public OPMPHF (int n) {
	
	this.n = n;
	this.m = 2 * n;
	this.edges = new ConnectionSet (m, n);

	// System.out.println ("n = " + n + ", m = " + m);

    }

    public int getLexiconTotal () {
	return lexiconTotal;
    }

    public void beginGenerate () {
	Random r = new Random ();
	
	for (int i = 0;  i < w1.length;  i++) {
	    w1[i] = r.nextInt (128);
	    w2[i] = r.nextInt (128);
	}
	
	
	g = new int[m];
	h = new int[n];

	h_cur = 0;

	edges.start ();
    }
    
    public void addWord (String w) {
	char[] t = w.toCharArray ();

	addWord (t, 0, t.length);
    }

    public void addWord (char[] t, int start, int len) {

	edges.connect (h1(t, start, len), h2(t, start, len), h_cur);
	edges.connect (h2(t, start, len), h1(t, start, len), h_cur);
	h_cur++;
    }


    public boolean endGenerate () {

	lexiconTotal = edges.done ();

	if (findAcyclicGraph ()) {
	    /*
	    for (int i = 0;  i < w1.length;  i++)
		System.out.println ("" + w1[i] + "\t" + w2[i]);
	    
	    for (int i = 0;  i < m;  i++)
		System.out.println ("g["+i+"] = " + g[i]);
	    */
	    return true;
	} else {
	    // System.out.println ("----------------");
	    return false;
	}

	
    }


    boolean findAcyclicGraph () {
	
	for (int v = 0;  v < m;  v++) {
	    g[v] = -1;
	}
	
	for (int v = 0;  v < m;  v++) {
	    if (g[v] == -1)
		if (! labelFrom (v, 0))
		    return false;
	}
	
	return true;
    }

    class EdgeFuncImpl implements ConnectionSet.EdgeFunc {
	
	public boolean invoke (int v, int u, int data) {
	    
	    return labelFrom (u, data - g[v]);
	}
    }
    
    boolean labelFrom (int vertex, int c) {

	// System.out.println ("labelFrom: vertex " + vertex + ", label " + c);
	if (g[vertex] != -1) {
	    
	    if (g[vertex] != c) {
		// System.out.println ("1. Returning false for " + vertex);
		return false; // graph is cyclic
	    } else {
		return true; // vertex was already visited
	    }

	}

	g[vertex] = c;
	
	// System.out.println ("2. g[" + vertex + "]=" + c);

	if (! edges.foreachAdjacentEdge (vertex, edgeFunc))
	    return false;	
	
	return true;
    }

    int h1 (char[] t, int start, int len) {
	
	int sum = 0;

	
	for (int i = start;  i < len;  i++) {
	    
	    sum += ( ((int) t[i]) * w1[i]);
	}

	return sum % m;
    }

    
    int h2 (char[] t, int start, int len) {
	
	int sum = 0;

	
	for (int i = start;  i < len;  i++) {
	    
	    sum += ( ((int) t[i]) * w2[i]);
	}

	return sum % m;
    }

    public int hash (String s) {

	char[] t = s.toCharArray ();
	int    h1 = h1 (t, 0, t.length);
	int    h2 = h2 (t, 0, t.length);
	int    h     = (g[h1] + g[h2]) % n;
	
	// System.out.println ("" + h1 + " " + h2 + " " + h);

	return h;
    }

    public int hash (char[] t, int start, int len) {

        int    h1 = h1 (t, start, len);
        int    h2 = h2 (t, start, len);
	int    h     = (g[h1] + g[h2]) % n;
	
	// System.out.println ("" + h1 + " " + h2 + " " + h);

	return h;
    }




    public static void main (String[] argv) {

	OPMPHF hashFunc = new OPMPHF (argv.length);
	
	int iter = 0;

	do {
	    // System.out.println ("Iteration " + iter++);

	    hashFunc.beginGenerate ();
	    
	    for (int i = 0;  i < argv.length;  i++)
		hashFunc.addWord (argv[i]);
	
	    
	} while (! hashFunc.endGenerate ());
	
	
	for (int i = 0;  i < argv.length;  i++) {
	    
	    
	    System.out.println (argv[i] + " hashes to " + 
				hashFunc.hash (argv[i]));
	}
	
    }
}



    


class ConnectionSet {
    
    int   edgeNo;
    int[] from;
    int[] to;
    int[] weights;
    int[] vertexStarts;
    int[] vertexLengths;

    interface EdgeFunc {
    
	boolean invoke (int u, int v, int data);
    }

    ConnectionSet (int numVerticies, int numEdges) {
	this.from    = new int[numEdges*2];
	this.to      = new int[numEdges*2];
	this.weights = new int[numEdges*2];
	this.vertexStarts = new int[numVerticies];
	this.vertexLengths = new int[numVerticies];
    }

    void start () {
	edgeNo = 0;
    }
    
    void connect (int i, int j, int weight) {
	from[edgeNo]    = i;
	to[edgeNo]      = j;
	weights[edgeNo] = weight;

	// System.out.println ("--> from " + i + " to " + j + 
	// " with weight " + weight);
	edgeNo++;
    }

    int done () {
	sortEdgesByFromVertex ();

	return edgeNo;
    }

    void sortEdgesByFromVertex () {
	/*
	for (int i = 0;  i < from.length;  i++) 
	    System.out.println ("--- from " + from[i] + " to " +
				to[i] + " with weight " + weights[i]);
	*/

	SortService.sort (from, new SortService.Callback () {
		
		public void swap (int i, int j) {

		    int t;

		    t = to[i];      to[i] = to[j];           to[j] = t;
		    t = weights[i]; weights[i] = weights[j]; weights[j] = t;
		    
		}
	    });

	/*
	for (int i = 0;  i < from.length;  i++) 
	    System.out.println ("==== from " + from[i] + " to " +
				to[i] + " with weight " + weights[i]);
	*/

	int curVertex = 0;
	vertexStarts[0] = 0;
	
	for (edgeNo = 0;  edgeNo < from.length;  edgeNo++) {
	    
	    if (from[edgeNo] > curVertex) {
		
		vertexLengths[curVertex] = edgeNo - vertexStarts[curVertex];
		
		curVertex = from[edgeNo];
		vertexStarts[curVertex] = edgeNo;
	    }
	    
	}
	


	vertexLengths[curVertex] = edgeNo - vertexStarts[curVertex];
    } 
    
    
    boolean foreachAdjacentEdge (int vertex, EdgeFunc func) {
		
	for (int i = vertexStarts[vertex];  
	     i < vertexStarts[vertex] + vertexLengths[vertex]; 
	     i++)
	    {
		// System.out.println ("From " + vertex + " to " + to[i]);
		if ( ! func.invoke (vertex, to[i], weights[i]) )
		    return false;
	    }

	return true;
    }
}

/*
class ConnectionSet {

    int    numVerticies;
    int    numEdges;  
    byte[] set;

    int[] rowStarts;
    int[] cols;
    int[] data;
    int[] rowLengths;

    ConnectionSet (int numVerticies, int numEdges) {

	
        this.numVerticies = numVerticies;
	this.set = new byte[numVerticies * numVerticies];

    }


    void start () {
	for (int i = 0;  i < set.length;  i++)
	    set[i] = 0;

    }
	

    
    void connect (int i, int j, int data) {
	// System.out.println ("Connecting " + i + " to " + j);

	int edge = i * numVerticies + j;

	int bit = edge % 8;
	int byt = (int) Math.floor ((edge - bit) / 8);

	set[byt] = (byte) (set[byt] | (((byte)0x1) << bit));
	
	numEdges++;
    }


    boolean isConnected (int i, int j) {
	int edge = i * numVerticies + j;
	int bit  = edge % 8;
	int byt  = (int) Math.floor ((edge - bit) / 8);

	int mask = 0x1 << bit;

	boolean b =  (set[byt] & mask) == mask;

	
	return b;
    }


    int done () {
	int curVertex = 0;
	int off = 0;
	int cnt = 0;
	
	rowStarts = new int[numVerticies];
	cols = new int[numEdges];
	data = new int[numEdges];
	rowLengths = new int[numVerticies];

	// System.out.println ("rowStarts.length = " + rowStarts.length);
	for (int byt = 0;  byt < set.length;  byt++) {
	    for (int bit = 0;  bit < 8;  bit++) {

		int mask = 0x1 << bit;
		if ((set[byt] & mask) == mask) {
		    

		    int n = byt * 8 + bit;
		    int u = n % numVerticies;
		    int v = (int) Math.floor ((n - u) / numVerticies);
		    
		    if (curVertex < v) {
			curVertex = v;
			rowStarts[curVertex] = off;
		    }
		    
		    cols[off] = u;
		    if (v >= u) {
			
			for (int o = rowStarts[u]; 
			     o < rowStarts[u]+rowLengths[u]; 
			     o++)
			    if (cols[o] == v)
				data[o] = cnt;
			
			data[off] = cnt++;
			
		    } 

		    off++;
		    rowLengths[v] = rowLengths[v] + 1;
		}
		    
	    }
	}

	return off;
    }


    boolean foreachAdjacentEdge (int v, EdgeFunc func) {
	int u;
	
	for (int off = rowStarts[v]; off < rowStarts[v]+rowLengths[v]; off++) {
	    
	    if (! func.invoke (v, cols[off], data[off]))
		return false;
	}
	
	return true;
    }
}

*/
