/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SchedulerService.java,v 1.1.1.1 2001/06/01 16:51:28 jbeatty Exp $
 */
package net.jxta.search.util;

public class SchedulerService implements Runnable {

    volatile boolean       done;
    SchedulerPriorityQueue heap;
        
    public static interface Action {

	void perform (SchedulerService ss); // called when alarm expires
    }


    
    
    public SchedulerService () {
	
	this.done = false;
	this.heap = new SchedulerPriorityQueue (1024, 2);
    }
    

    public synchronized Object scheduleAction (Action action, long time) {
	PendingAction el = new PendingAction (action, 
					      System.currentTimeMillis () + 
					      time);

	heap.put (el);

	if (heap.top () == el)
	    notify ();

	return el;
    }

    
    public synchronized void cancelAction (Object pendingAction) {
	PendingAction el = (PendingAction) pendingAction;
	
	el.cancel ();

	if (heap.top () == el)
	    notify ();
    }
    
    public void run () {
	
	PendingAction e;
	
	while (! done) {
	    e = getNextEvent ();

	    if (e == null || done)
		break;
	    
	    try {
		e.perform (this);
	    } catch (Exception ex) {
		
		ex.printStackTrace ();
	    }
	}
    }
    
    private synchronized PendingAction getNextEvent () {
	PendingAction el;
	long          timeout;
	

	if (done) 
	    return null;
	else if ( ( el = (PendingAction) heap.top ()) == null)
	    timeout = Long.MAX_VALUE;
	else
	    timeout = el.time - System.currentTimeMillis ();

	try {
	    while (! done) {

		if ( timeout > 0)
		    wait (timeout);
		
		if (done) {                     // we were stopped

		    break;
		
		} else if (el != heap.top ()) { // new minimum element inserted

		    el = (PendingAction) heap.top ();
		    timeout = el.time - System.currentTimeMillis ();
		    
		} else {                        // we were woken up to act
		    
		    heap.pop ();
		    return el;
		}
	    }
	} catch (InterruptedException e) {
	    done = true;
	}

	return null;
    }

}

class SchedulerPriorityQueue extends PriorityQueue {
    
    SchedulerPriorityQueue (int heapSize, int increase) {
	super.initialize (heapSize, increase);
    }

    protected boolean lessThan (Object o1, Object o2) {
	PendingAction a1 = (PendingAction) o1;
	PendingAction a2 = (PendingAction) o2;

	return a1.time < a2.time;
    }
}

class PendingAction {
    
    SchedulerService.Action action;
    long             time;
    
    
    PendingAction (SchedulerService.Action action, long time) {
	
	this.action = action;
	this.time   = time;
    }

    synchronized void perform (SchedulerService ss) {
	if (action != null) {
	    action.perform (ss);
	    action = null;
	}
    }
    
    synchronized void cancel () {
	action = null;
    }
    
}


