/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SortService.java,v 1.1.1.1 2001/06/01 16:51:28 jbeatty Exp $
 */
package net.jxta.search.util;

/**
 * SortService implements the quick-sort algorithm.
 * The only reason we implemented it is because it has 
 * a callback interface that is notified when elements
 * are swapped.
 *
 * @ee <a href=SortService.Callback.html>SortService.Callback</a>.
 */

class SortService {

    
    /**
     * An instance of a class that implements this interface should
     * be passed to the sort method if you want to be notified of
     * swap operations.  You can construct a visual demonstration of
     * the swapping operations or you can use this feature for a more
     * practical purpose, such as swapping the corresponding elements
     * in a different array than the one being sorted, if you have two
     * arrays with some correspondence between their elements... like
     * if a sparse vector representation, for example.
     */

    interface Callback {
	
	void swap (int i, int j);
    }


    /**
     * Sort an array of integers.
     *
     * @param a  the array of integers
     * @param cb the callback which is notified when elements of <em>a</em> are swapped.
     */

    static void sort (int[] a, Callback cb) {
	sort (a, cb, a.length, 0, a.length-1);
    }


    /**
     * Sort an array of integers.  This is actually a helper method, but
     * is exposed within the package.  No harm in that.
     *
     * @param a  the array of integers
     * @param cb the callback which is notified when elements of <em>a</em> are swapped.
     * @param length the number of elements to work with from array <em>a</em>
     * @param l the index of the left-most element to use
     * @param r the index of the right-most element to use.
     */

    static void sort (int[] a, Callback cb, int length, int l, int r) {
	
	int v, t;
	int i, j;

	if (r > l) {
	    v = a[r];  i = l-1; j = r;
	    
	    for (;;) {

		while ((++i < length) && a[i] < v)
		    ;

		while ((--j > 0) && a[j] > v)
		    ;

		if (i >= j)
		    break;

		t = a[i]; a[i] = a[j]; a[j] = t;
		if (cb != null)
		    cb.swap (i, j);
	    }

	    t = a[i]; a[i] = a[r]; a[r] = t;
	    if (cb != null)
		cb.swap (i, r);
	    
	    sort (a, cb, length, l, i-1);
	    sort (a, cb, length, i+1, r);

	}
    }

    public static void testSortService() {

	String [] args  = { "13", "11", "14", "1", "101" };

	testSortService(args);

    }

    public static void testSortService(String [] argv) {

	int[] a = new int[argv.length];
	
	for (int i = 0;  i < a.length;  i++)
	    a[i] = Integer.parseInt (argv[i]);
	
	testSortService (a);
    }

    public static void testSortService (int[] a) {
	SortService.sort (a, null);

	for (int i = 1;  i < a.length;  i++) {
	    if (a[i] < a[i-1]) {
		System.out.println ("Failed SortService: An element with a higher index " +
				    "is smaller than one with a lower index " +
				    "after sorting.");
		return;
	    }
	}

	System.out.println ("Passed SortService.");
    }


    public static void main (String[] argv) {
	testSortService();
    }


}  
