/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Soundex.java,v 1.1.1.1 2001/06/01 16:51:29 jbeatty Exp $
 */
package net.jxta.search.util;


import java.io.*;
import java.util.NoSuchElementException;

/**
 * Soundex is an algorithm (due to Knuth?) to do spell-checking.
 *
 * Usage:
 <pre>
	Soundex   soundex = new Soundex ();
	int       numSuggestions = 4;

	soundex.insertDictionary (new File ("/usr/dict/words"));
	// or

	soundex.insert ("Birkbeck");
	soundex.insert ("Zbygniewski");
	soundex.insert ("toy");
	soundex.insert ("car");
	soundex.insert ("lorry");
	soundex.insert ("bicycle");
	
	PriorityQueue pq = soundex.getClosest (s, numSuggestions);
	
 	if (pq == null) {
	    System.out.println ("No matches.");
	} else {
	    Soundex.EditWord ew;
	    
	    while ((ew = (EditWord) pq.pop ()) != null) {
		
		System.out.println (ew.s + " with " + ew.distance + " edits.");
	    }
	}
  </pre>
*/
    
public class Soundex {

    Node[] nodes = new Node[26];

    public Soundex () {
	for (int i = 0;  i < 26;  i++)
	    nodes[i] = new Node (0);
    }

    class Node {
	Node[] children;

	Node () { }

	Node (int level) {
	    if (level == 2) {
		children = new LeafNode[10];
		for (int i = 0;  i < 10;  i++)
		    children[i] = new LeafNode ();
	    } else {
		children = new Node[10];
		for (int i = 0;  i < 10;  i++)
		    children[i] = new Node (level+1);
	    }
	}

	void addWord (byte[] hash, int level, String s) {
	    children[ hash[level] ].addWord (hash, level+1, s);
	}

	Word getWords (byte[] hash, int level) {
	    return children[ hash[level] ].getWords (hash, level+1);
	}
    }

    class LeafNode extends Node {
	Word   firstWord;
	Word   lastWord;
	int    total;

	void addWord (byte[] hash, int level, String s) {
	    
	    if (firstWord == null) {
		firstWord = lastWord = new Word (s);
	    } else {
		lastWord.next = new Word (s);
		lastWord = lastWord.next;
	    }
	
	    total++;
	}

	Word getWords (byte[] hash, int level) {
	    return firstWord;
	}
    }
    
    class Word {
	String s;
	Word   next;

	Word (String s) {
	    this.s = s;
	}
    }

    byte[] hash (String s) {
	byte[] buf = new byte[3];
	byte last = 0;
	int j = 0;
	int slen = s.length ();

	for (int i = 1;  j < 3 && i < slen;  i++) {
	    switch (s.charAt (i)) {
	    case 'b': case 'f': case 'p': case 'v':
		if (last != 1)
		    last = buf[j++] = 1;
		break;
	    case 'c': case 'g': case 'j': case 'k': case 'q': case 's': 
	    case 'x': case 'z':
		if (last != 2)
		    last = buf[j++] = 2;
		break;
	    case 'd': case 't':
		if (last != 3)
		    last = buf[j++] = 3;
		break;
	    case 'l':
		if (last != 4)
		    last = buf[j++] = 4;
		break;
	    case 'm': case 'n':
		if (last != 5)
		    last = buf[j++] = 5;
		break;
	    case 'r':
		if (last != 6)
		    last = buf[j++] = 6;
		break;
	    default:
		last = 0;
	    }
	    
	}
	
	
	for (int i = j;  j < 3;  j++)
	buf[j] = 0;
	
	return buf;
    }

    
    public void insert (String s) {

	Node root = nodes[(int)Character.toUpperCase (s.charAt(0)) - (int)'A'];
	
	byte[] hash = hash (s);
	
	// System.out.println (s + " hash " + hashToString (hash));

	root.addWord (hash, 0, s);
	
    }

    Word getWords (String s) {
	char c = s.charAt (0);
	
	if (Character.isLetter (c)) {
	    Node root = nodes[(int)Character.toUpperCase (c) - (int)'A'];
	    
	    byte[] hash = hash (s);
	    
	    return root.getWords (hash, 0);
	} else {
	    return null;
	}
    }

    public void insertDictionary (File dict) throws IOException {
	FileReader     fr   = null;
	BufferedReader br   = null;
	String         line = "";

	if (dict.exists ()) {
	    try {
		fr = new FileReader (dict);
		br = new BufferedReader (fr, 8192);
		
		while ( (line = br.readLine ()) != null) {
		    insert (line);
		}
		
	    } finally {
		if (fr != null) {
		    try { fr.close (); }
		    catch (IOException e) { }
		}
		if (br != null) {
		    try { br.close (); }
		    catch (IOException e) { }
		}
	    }
	}
    }

    int editDistance (char[] s1, char[] s2) {
	/* // Prepare a full memoization matrix for the recursive version.
	int[][] mem = new int[s1.length][s2.length];

	for (int i = 0;  i < s1.length;  i++)
	    for (int j = 0;  j < s2.length;  j++)
		mem[i][j] = -1;
	*/

	return editDistanceIter (s1, s2);
    }

    private static final int min (int i1, int i2, int i3) {
	return (i1 < i2 ? 
		(i2 < i3 ? i1 : (i1 < i3 ? i1 : i3)) :
		(i1 < i3 ? i2 : (i2 < i3 ? i2 : i3))) ;
    }


    /**
     * A recursive version for readability.
     * Requires a matrix of size s1.length x s2.length for memoization.
     */

    private static final int editDistance (char[] s1, char[] s2, 
					      int s1Len, int s2Len,
					      int[][] mem)
    {
	if (s1Len == 0 && s2Len == 0) {
	    return 0;
	} else if (s1Len == 0) {
	    return s2Len;
	} else if (s2Len == 0) {
	    return s1Len;
	} else if (mem[s1Len-1][s2Len-1] != -1) {
	    return mem[s1Len-1][s2Len-1];
	} else {
	    int x = (s1[s1Len-1] == s2[s2Len-1] ? 0 : 1);
	    return mem[s1Len-1][s2Len-1] =
		min (editDistance (s1, s2, s1Len-1, s2Len-1, mem) + x,
		     editDistance (s1, s2, s1Len-1, s2Len,   mem) + 1,
		     editDistance (s1, s2, s1Len,   s2Len-1, mem) + 1);
		
			
	}
    }

    /**
     * An iterative version for speed.
     * It also allocates just two rows, rather than a full matrix
     * for memoization.  Ugh.
     */

    int editDistanceIter (char[] s1, char[] s2) {
	int[] old_row = new int[s2.length];
	int[] new_row = new int[s2.length];
	int[] tmp;
	int i = 0, j = 0;
	int x;
	
	old_row[0] = 0;
	new_row[0] = 1;

	for (j = 0;  j < s2.length;  j++)
	    old_row[j] = j;
	
	for (i = 1;  i < s1.length;  i++) {
	    
	    for (j = 1;  j < s2.length;  j++) {
		x = (s1[i] == s2[j] ? 0 : 1);
		new_row[j] = min (old_row[j-1] + x,
				  old_row[j]   + 1,
				  new_row[j-1] + 1);
	    }
	    
	    tmp = old_row;
	    old_row = new_row;
	    new_row = tmp;
	}

	return old_row[s2.length-1];
    }


    String hashToString (byte[] buf) {
	char[] s = new char[3];

	for (int i = 0;  i < 3;  i++) {
	    s[i] = (char) (buf[i] + 48);
	}

	return new String (s);
    }

    public static class EditWord {
	String s;
	int    distance;

	EditWord (String s, int distance) {
	    this.s = s;
	    this.distance = distance;
	}

	public String getString () { return s; }
	public int    getDistance () { return distance; }
    }

    static class EditQueue extends PriorityQueue {
	int targetLength;

	EditQueue (int targetLength, int size, int growthFactor) {

	    initialize (size, growthFactor);
	    
	    this.targetLength = targetLength;
	}

	public boolean lessThan (Object o1, Object o2) {
	    EditWord ew1 = (EditWord) o1;
	    EditWord ew2 = (EditWord) o2;

	    if (ew1.distance == ew2.distance) { 
		if (ew1.s.charAt (0) == ew2.s.charAt (0))
		    
		    return (Math.abs (ew1.s.length () - targetLength) <
			    Math.abs (ew2.s.length () - targetLength));
			
		else
		    return ew1.s.charAt (0) > ew2.s.charAt (0);
	    } else {
		return ew1.distance < ew2.distance;
	    }
	}
    }

    /**
     * @return PriorityQueue Ordered list of closest-matching words.
     * @return null If there is an exact match.
     * @exception NoSuchElementException If there was no exact match
     * and no words within maxDistance.
     */
    public PriorityQueue getClosest (String  s, int maxDistance) 
	throws NoSuchElementException
    {
	Word      w   = getWords (s);
	Word      w2  = w;
	EditQueue pq  = null;
	char[]    c1;
	char[]    c2;
	int       distance;
	
	if (w == null) {
	    throw new NoSuchElementException ();
	} else {
	    /* check if there is an exact match */
	    while (w != null) {
		if (w.s.equalsIgnoreCase (s))
		    return null;
		else
		    w = w.next;
	    }

	    pq = new EditQueue (s.length (), 16, 2);
	    c1 = s.toCharArray ();

	    w = w2;

	    while (w != null) {
		c2 = w.s.toCharArray ();
		distance = editDistance (c1, c2);
		// System.out.println ("Got " + w.s + 
		//                     " with edit distance " + distance);
		if (distance <= maxDistance)
		    pq.put (new EditWord (w.s, distance));
		
		w = w.next;
	    }
	    
	    if (pq.top () == null)
		throw new NoSuchElementException ();
	}
	
	return pq;
    } 

    public static void main (String[] argv) throws IOException {
	String    s       = argv[0];
	Soundex   soundex = new Soundex ();
	

	soundex.insertDictionary (new File ("/usr/dict/words"));
	
	/*
	soundex.insert ("Birkbeck");
	soundex.insert ("Zbygniewski");
	soundex.insert ("toy");
	soundex.insert ("car");
	soundex.insert ("lorry");
	soundex.insert ("bicycle");
	*/
	
	PriorityQueue pq = soundex.getClosest (s, Integer.parseInt (argv[1]));
	
	if (pq == null) {
	    System.out.println ("No matches.");
	} else {
	    EditWord ew;
	    
	    while ((ew = (EditWord) pq.pop ()) != null) {
		
		System.out.println (ew.s + " with " + ew.distance + " edits.");
	    }
	}

    }
}
