/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TermCountService.java,v 1.3 2001/06/11 22:03:05 ddoolin Exp $
 */

package net.jxta.search.util;

import java.io.*;
import java.net.*;
import java.util.*;
import net.jxta.search.util.BM;
import net.jxta.search.util.ISVector;
import net.jxta.search.util.ISMatrix;
import net.jxta.search.util.StringMatcher;
import java.lang.Math;


/** Used to compute relevance of search results for opensearch. 
 */
public class TermCountService {

    private static boolean debug = false;


    public TermCountService() {

    }


    public void setDebug() {
	debug = true;
    }

    public void clearDebug() {
	debug = false;
    }


    /** 
     * Computes raw stats for relevance routine using 
     * BM method from net.jxta.search.java.util.BM. 
     * NOTE: No range or array checking is done in this
     * method.  All the parameters are assumed to be initialized
     * correctly.
     *
     * @param String queryArray is an array strings each of 
              which is a term in the query string.
     * @param String DocArray is an array of documents, each
              entry of which is a string which is a document.
     * @param int [][] TermCount 
     * @param int [] DocCount 
     */
    public static void compute_counts_bm(String[] queryArray, 
					  String[] docArray, 
					  int[][] TermCount, 
					  int[] DocCount) {

	int query_len = queryArray.length;
	int num_docs = docArray.length;

	StringMatcher matcher = new BM();

	StringMatcher.CompiledPattern [] compiledPatterns = 
	                       new  StringMatcher.CompiledPattern[query_len];

	char[][] docCharArrays = new char[docArray.length][];
	char[][] queryCharArrays = new char[queryArray.length][];
	
	
	for (int i = 0;  i < num_docs;  i++)
	    docCharArrays[i] = docArray[i].toCharArray ();


	for (int i = 0;  i < query_len;  i++) {
	    queryCharArrays[i] = queryArray[i].toCharArray ();

	    if (debug) {
		System.out.println ("query term = " + new String (queryCharArrays[i]) +
	    			", query term length = " + queryCharArrays[i].length);
	    }

            if (queryCharArrays[i].length == 0) {
                continue;
	    }
	    compiledPatterns[i] = matcher.compile (queryArray[i]);
	}


	for(int d = 0; d < num_docs; d ++) {

	    char[] doc = docCharArrays[d];

	    for(int t = 0; t < query_len; t ++) {



		int count = 0;
		char[] q = queryCharArrays[t];
		int index = -1;

           
		if (debug) {
		    System.out.println("Matching "
				       + queryArray[t]
				       + " and "
				       + docArray[d]);
		}

		
		while ((index = matcher.match(doc,
					      index+1,
					      compiledPatterns[t])) != -1) {
		    count++;
		}


		TermCount[d][t] += count;
		if(count > 0)
		    DocCount[t] += 1;

	    }
	}

    }


    /** Compute counts using naive method. 
     */
    public static void compute_counts_naive(String[] queryArray, 
			String[] docArray, 
			int[][] TermCount, 
			int[] DocCount) {

	int query_len = queryArray.length;
	int num_docs = docArray.length;

	for(int d = 0; d < num_docs; d ++) {

	    String doc = docArray[d];
	    StringTokenizer words = new StringTokenizer(doc);

	    while(words.hasMoreTokens()) {

		String word = words.nextToken();
		/* NOTE: Case is not ignored in quick match.
		 * The new code in SearchRelevance is not
		 * stemmed anymore.  Leave this next block in
		 * for the moment.
		 */
		//String orig_word = words.nextToken();
		//String upper = orig_word.toUpperCase();
		//String word = quick_stem(upper);
		//String word = quick_stem(orig_word);

		for(int t = 0; t < query_len; t ++) {

		    String q = queryArray[t];
		    if(quick_match(word, q)) {
			// if the word in doc d matches the query term t
			TermCount[d][t] += 1;
			DocCount[t] += 1;
		    }

		}
	    }
	}

    }  /* close compute_counts_naive() */





   /** 
    * quick_stem purports to implement fast, simple stemming using 
    * common suffixes.  The method assumes (?) that the parameter
    * has already been converted to upper case.
    *
    * @param String orig_string is a term which is assumed to 
    *        converted to upper case.
    *
    * @return String stem_string is supposed to the parameter with 
    *         one of the common suffixes stemmed away.
    *
    * FIXME: Somewhere, there should be a reference in the literature
    *        for how this method works:
    *        <a href="http://whereis.the.reference">reference</a>.
    *
    * FIXME: This method appears to have a bug insofar as stemming 
    *        *ING returns *IN, *ER returns *E, etc.
    */
    public static String quick_stem(String orig_string) {

	String[] stems = {"?", ":","&QUOT;", ",", "'S", "S", "ING", "ED", "ER", "ES"};
	int stem_length = 9;
	String stem_string = orig_string; // fall through value


	for(int s = 0; s < stem_length; s ++) {
	    if(orig_string.endsWith(stems[s])) {
		stem_string = orig_string.substring(0, 
						    orig_string.length() - 1);
		return stem_string; // break if a match
	    }
	}

	return stem_string;

    }



    /** 
     * quick_match performs a match between "word" and query "q"
     * and returns true if there is a match 
     * and false if there is no match, note that case is ignored.
     *
     * @param  String word.  Whitespace matters in this
     *         matching function, so word " foo" will not match
     *         query "foo".
     * @param  String q is a query term.  Same conditions on whitespace
     *         hold as for word.
     * @return boolean true for case-insensitive match, false otherwise.
     * 
     * FIXME: There is a second part to the function besides the 
     *        naive match.  This needs to be documented.
     */
    public static boolean quick_match(String word, String q) {

	if(debug) {
	    System.out.println("quick_match....");
	    System.out.println("query " + q + "l= " + q.length());
	    System.out.println("word " + word + "l = " + word.length());
	    System.out.println(word.equals(q));
	}

	/* The equalsIgnoreCase "fixed" some sort of bug
	 * at one point.  May not be necessary now, but
	 * leave it in as a flag on the previous problem.
	 */
	//if(word.equalsIgnoreCase(q)) { // simple case
        if((word.indexOf(q) != -1)) { // simple case 
	    return true; 
	}

	if((word.indexOf(q) != -1)) { 
	    // kind of fudgy - could match biggles from big etc.
	    return true;
	}

	return false;

    }  /* close quick_match() */

  
   /** 
    * By having each class have a test() method, we can 
    * later wrap all of the testing in a class that hits
    * each test method in succession.
    *
    * TODO: Everything goes to System.out at the moment.
    * It would be smart to be able to specify some arbitrary
    * PrintStream for future automatic testing.
    */
     public static boolean test(PrintStream ps) {

        return TestTermCountService.test(ps);
    
     }


    public static void main (String[] argv) {

       TestTermCountService.test(System.out);

    }  /* close test() */



}  /* class TermCountService */




/** 
 * Having all of the test methods in the same file as the 
 * class to be tested provides example documentation for 
 * anyone reading the code.  Having all of the test methods
 * in a different, private class in this file saves on 
 * footprint because the test class won't be loaded unless it
 * is actually needed.
 *
 * TODO: Add logging support.  Right now everything is going 
 * to System.out, but later we might want to log to a file.
 */

class TestTermCountService  {

  


   public static boolean test(PrintStream ps) {

        boolean passed = true;

        ps.println("-------------  Begin TermCountService unit tests -------");

        if (test_quick_match(ps)) {
            ps.println("Passed quick match unit test.");
        } else {
            passed = false;
            ps.println("Failed quick match unit test.");
        }

        if (test_quick_stem(ps)) {
            ps.println("Passed quick stem unit test.");
        } else {
            passed = false;
            ps.println("Failed quick stem unit test.");
        }

       /* This should call a method that wraps al the count testing
        * methods, because all three count testing methods share 
        * common parameters.
        */
        if (test_compute_counts(ps)) {
            ps.println("Passed compute counts unit tests.");
        } else {
            passed = false;
            ps.println("Failed compute counts unit tests.");
        }

        if (!passed) {
            ps.println("FAILED TermCountService unit testing");
	}

        ps.println("-------------  End TermCountService unit tests -------");        

        return passed; 
            
   }


   /**
    * TODO: Collect the passing examples, abstract common code,
    * then collect failing examples and abstract common code.
    * The result should be similar to the test_quick_stem() 
    * method below.
    */
    private static boolean test_quick_match(PrintStream ps) {
       
        String s1 = "foo";
        String s2 = "foo";
        if (TermCountService.quick_match(s1,s2)) {
            ps.println("Passed quick_match test 1");
        } else {
            ps.println("Failed quick_match test 1");
            return false;
        }
        
        String s3 = "bar";
        if (TermCountService.quick_match(s1,s3)) {
            ps.println("False match, failed quick_match test 2");
            return false;
        } else {
            ps.println("Passed quick_match test 2");
        }

        String s4 = " foo ";
        if (TermCountService.quick_match(s1,s3)) {
            ps.println("False match, failed quick_match test 2");
            return false;
        } else {
            ps.println("Passed quick_match test 3");
        }


      /* TODO: Add more tests as necessary */ 
        
        return true;

     } 


     private static boolean test_quick_stem(PrintStream ps) {
  
	 boolean passed = true;

        if (!test_quick_stem(ps,"BOOMS","BOOM",1)) { 
            passed = false;
        }

        if (!test_quick_stem(ps,"BOOMING","BOOM",2)) { 
            passed = false;
        }

        if (!test_quick_stem(ps,"BOOMER","BOOM",3)) { 
            passed = false;
        }
    
        return passed;

     } 

     private static boolean test_quick_stem(PrintStream ps, String s1, String s2, int test) {

        String stemstring = TermCountService.quick_stem(s1);
        //ps.println(stemstring);

        if (stemstring.equals(s2)) {
            ps.println("Passed quick_stem test " + test);
        } else {
            ps.println("Failed quick_stem test " + test);
            return false;
        }

        return true;

     } 
    

   
    String query_string = "find stuff";
    
   /* Start with just 1 document for now. */
    static int num_docs = 1;
    static int query_len = 2;


   /* See the javadoc comments for meaning of these arrays. */
    static String [] queryArray = {"find", "stuff"};
    static int [][] TermCount = new int[num_docs][query_len];
    static int [] DocCount = new int[query_len];

    static String [] DocArray = {"If you want to find good stuff, "
                            + "you must search the InfraSearch network."}; 

    /* TODO: Test against a null document.*/

   /**
    * FIXME: Abstract common elements to all count methods,
    * then use this as a wrapper to each specific count test
    * in succession.
    */
    private static boolean test_compute_counts(PrintStream ps) {
  
	boolean passed = true;

      if (test_compute_counts_bm(queryArray,
				 DocArray,
				 TermCount,
				 DocCount)) {
	  ps.println("Passed boyer-moore term counting");
      } else {
          ps.println("Failed boyer-moore term counting");
          passed = false;
      }
      printTermArrays(ps);


      if (test_compute_counts_naive(queryArray,
				    DocArray,
				    TermCount,
				    DocCount)) {
	  ps.println("Passed naive term counting");
      } else {
          ps.println("Failed naive term counting");
          passed = false;
      }
      //printTermArrays(ps);


     /* Dummy value for now because test is not fully implemented. */
      return passed;

    }


    private static void printTermArrays(PrintStream ps) {
      ps.println("Counts computed:");
      ISVector.print(queryArray,ps,"queryArray");
      ISVector.print(DocArray,ps,"DocArray");
      ISMatrix.print(TermCount,ps,"TermCount");
      ISVector.print(DocCount,ps,"DocCount");
    }

    private static boolean test_compute_counts_bm(String[] queryArray, 
					  String[] DocArray, 
					  int[][] TermCount, 
					  int[] DocCount) {

	boolean passed = true;


      TermCountService.compute_counts_bm(queryArray,
                                         DocArray,
					 TermCount,
					 DocCount);


      /* TODO: Implement test conditions here. */

        passed = false;  // no tests yet.
	return passed;
    }

    private static boolean test_compute_counts_naive(String[] queryArray, 
					  String[] DocArray, 
					  int[][] TermCount, 
					  int[] DocCount) {

	boolean passed = true;

	ISMatrix.setZero(TermCount);
        ISVector.setZero(DocCount);
        TermCountService.compute_counts_naive(queryArray,
                                              DocArray,
					      TermCount,
					      DocCount);

      /* TODO: Implement test conditions here. */
	int [][] term_count_test = { {1, 1}};

        try {
            if (!ISMatrix.equals(TermCount, term_count_test)) {
                passed = false;
	    }
	} catch (Exception e) {
        }

        int [] doc_count_test = {1,1};

        try {
            if (!ISVector.equals(DocCount, doc_count_test)) {
                passed = false;
	    }
	} catch (Exception e) {
        }

	return passed;
    }


}  /* class TestTermCountService */
