/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ThreadService.java,v 1.1.1.1 2001/06/01 16:51:32 jbeatty Exp $
 */
package net.jxta.search.util;

import java.io.File;
import java.util.Hashtable;


public class ThreadService {

    static final boolean debug = false;
    
    static final byte EVT_DESTROY = 1;
    static final byte EVT_TARGET  = 2;
    static final byte EVT_TIMEOUT = 3;

    private int minThreadCount;
    private int maxThreadCount;
    private int idleTimeout;

    private PoolThread pool[]    = null;
    private PoolThread freeQueue = null;
    
    private int runningThreads  = 0;
    private int existingThreads = 0;

    // Are we shutting down?
    private boolean shutdown           = false;
    // Display uncaught exceptions occuring in an allocated therad?
    private boolean traceRunExceptions = true;
    // Display uncaught errors occuring in an allocated thread?
    private boolean traceRunErrors     = true;

    /**
     * Inner class for managing pooled threads.
     * Instances of that class should NEVER lock the thread pool instance
     * while holding the lock to themselves. This ensure deadlcok avoidance.
     */

    class PoolThread extends Thread {
	// Our id within the thread pool
	int id = -1;
	// Should we shut the thread down ?
	boolean shutdown = false;
	// Next thread in the free queue:
	PoolThread next = null;
	// Our current target (if we're active)
	Runnable target = null;
	// Is that thread going to disappear ?
	boolean wasNotified = false;
	// Has that thread been terminated ?
	boolean finished = false;

	/**
	 * Shut this thread down.
	 */
	synchronized void shutdown () {
	    this.shutdown = true;
	    notifyAll ();
	}

	/**
	 * ThreadPol entry point - Power the given target with that thread.
	 * @param target The target <tt>Runnable</tt> to run.
	 */
	synchronized void launch (Runnable target) {
	    if ( debug )
		System.out.println ("launch: "+target+" with "+this);
	    this.target = target;
	    
	    notifyAll ();
	}

	/**
	 * Wait for that thread to be requested to do something.
	 */
	private synchronized byte waitForSomething (long timeOut) {
	    boolean timed = false;
	    while ( true ) {
		// Are we requested to shutdown ?
		if ( shutdown )
		    return EVT_DESTROY;
		// Do we have a target pending ?
		if ( target != null )
		    return EVT_TARGET;
		// Did we timed out:
		if ( timed )
		    return EVT_TIMEOUT;
		// Wait:
		try {
		    timed = true;
		    wait (timeOut);
		} catch (InterruptedException ex) {
		}
	    }
	}

	/**
	 * Wait until we get a target to execute.
	 */
	public void run () {
	    try {
		while ( true ) {
		    int st = waitForSomething (getTimeOut (this));
		    if ( debug ) 
			System.out.println("waitForSomething: status="+st); 
		    switch (st) {
		      case EVT_DESTROY:
			  // Destroy that thread:
			  return;
		      case EVT_TARGET:
			  // Run that target:
			  try {
			      target.run ();
			  } catch (Exception ex) {
			      if ( traceRunExceptions )
				  ex.printStackTrace ();
			  } catch (Error er) {
			      if ( traceRunErrors )
				  er.printStackTrace();
			      // Errors MUST be propagated:
			      throw er;
			  } finally {
			      target = null;
			      
			      pushIdleThread (this);
			  }
			  break;
		      case EVT_TIMEOUT:
			  // That thread timeout period has ended:
			  if ( ! threadTimedOut (this) )
			      return;
			  break;
		    }
		}
	    } finally {
		
		this.finished = true;
		threadFinished (this);
	    }
	}

	PoolThread (int id) {
	    this.id = id;
	    this.setName ("ThreadServiceImpl-"+id);
	    this.start ();
	}

    }

    /**
     * Configuration - Change the maximum number of threads allowed.
     * @param maxThreadCount The new maximum number of threads.
     */

    void updateMaxThreadCount (int maxThreadCount) {
	if ( pool.length > maxThreadCount ) {
	    // The number of threads will get self regulated
	    return;
	} else {
	    // Resize the pool array
	    synchronized (this) {
		PoolThread newPool[] = new PoolThread [maxThreadCount];
		
		System.arraycopy (pool, 0, newPool, 0, pool.length);

		this.pool = newPool;
	    }
	}
    }

    /**
     * Idle thread queue management.
     * Get an idle thread to run a target.
     * @param timeout The time to wait for a thread to become available.
     * @return An idle <tt>PoolThread</tt> instance, or <strong>null</strong>
     * if the caller doesn't want to wait and no threads are available.
     * @exception InterruptedException If the thread pool is shutting
     * down during request.
     */

    private synchronized PoolThread popIdleThread (long timeout) 
	throws InterruptedException
    {
	// The real time at which we should stop waiting for a thread:
	long stopTime = -1;

	// Loop until something happens:
	while ( true ) {
	    // Are we shutting down ?
	    if ( shutdown )
		throw new InterruptedException ("ThreadService terminating");
	    // Do we have a spare thread ?
	    if ( freeQueue != null ) {
		if ( debug )
		    System.out.println ("popIdleThread: pick idle "+freeQueue);
		runningThreads++;
		PoolThread next = freeQueue;
		freeQueue = freeQueue.next;
		// The thread may have been killed due to timeout, check:
		if ( ! next.finished )
		    return next;
	    }
	    // Have we created all possible threads ?
	    if ( existingThreads < maxThreadCount ) {
		if ( debug )
		    System.out.println ("popIdleThread: new thread");
		for (int i = 0 ; i < pool.length ; i++) {
		    if ( pool[i] == null ) {
			existingThreads++;
			return pool[i] = new PoolThread (i);
		    }
		}
	    }
	    // Wait, or return null:
	    if ( timeout == 0 ) {
		if ( debug )
		    System.out.println ("popIdleThread: returns null");
		return null;
	    } else if ( timeout < 0 ) {
		if ( debug )
		    System.out.println ("popIdleThread: waiting...");
		wait ();
	    } else {
		if ( stopTime < 0 ) {
		    stopTime = System.currentTimeMillis () + timeout;
		} else {
		    if ((timeout = stopTime - System.currentTimeMillis ()) < 0) {
			timeout = 0;
			continue;
		    }
		}
		wait (timeout);
	    }
	}
    }

    /**
     * Push an idle thread into the free queue.
     * @param thread The thread to mark free.
     */

    private synchronized void pushIdleThread (PoolThread thread) {
	if ( debug )
	    System.out.println ("pushIdleThread: "+thread);
	runningThreads--;
	thread.next = freeQueue;
	freeQueue   = thread;
	notifyAll ();
    }

    /**
     * Get the timeout period for a thread to wait.
     * @return A long number of milliseconds for the caller thread to 
     * remain idle.
     */

    private synchronized long getTimeOut (PoolThread thread) {
	return ( (existingThreads > minThreadCount) 
		 ? idleTimeout
		 : 0 );
    }

    /**
     * The given thread has remained idle for its entire timeout period.
     * If too many threads have been created, this method will notify
     * the calling thread to terminate itself by returning <tt>false</tt>.
     * @param thread The thread that has been idle.
     * @return A boolean, <tt>true</tt> if the thread is to remain live,
     * <tt>false</tt> otherwise.
     */


    private synchronized boolean threadTimedOut (PoolThread thread) {
	if ( thread.wasNotified )
	    throw new IllegalStateException("Inconsistent PoolThread state.");
	if ( existingThreads > minThreadCount ) {
	    // Nuke that thread:
	    thread.wasNotified = true;
	    existingThreads--;
	    return false;
	}
	return true;
    }

    /**
     * One of our threads has exited.
     * @param shutdown Is that due to normal shutdown ?
     */

    private synchronized void threadFinished (PoolThread thread) {
	if ( debug )
	    System.out.println ("threadFinished: "+thread);
	pool[thread.id] = null;
	// If it was notifed, we already counted in in notifyIdle
	if ( ! thread.wasNotified )
	    existingThreads--;
    }

    
    /**
     * Entry point.
     * Allocate a thread in all supported mode.
     * @param target The target to run.
     * @param timeout The timeout value, can be negative (block for ever),
     * <strong>0</strong> peek, or positive (wait for given duration).
     * @exception InterruptedException If caller is interrupted while
     * waiting for a thread.
     */

    private Thread allocateThread (Runnable target, long timeout)
	throws InterruptedException
    {
	if ( debug )
	    System.out.println ("getThread ? target="+target);

	// Try to get an idle thread out of the free queue:
	PoolThread thread = popIdleThread (timeout);

	if ( thread == null )
	    return null;

	if ( debug )
	    System.out.println ("getThread: target="+target+" thread="+thread);

	thread.launch(target);

	return thread;
    }

    /**
     * Get a thread from the pool to power the given target.
     * @param target The target <tt>Runnable</tt> to run.
     * @param timeout The number of milliseconds to wait for a thread to
     * become available.
     * @exception IllegalStateException If the thread pool is shutting down.
     * @see com.sun.jes.service.thread.ThreadService#getThread
     */

    public Thread getThread (Runnable target, long timeout)
	throws InterruptedException, IllegalArgumentException
    {
	if ( timeout < 0 )
	    throw new IllegalArgumentException ("negative timeout: " + timeout);
	if ( timeout == 0 )
	    return allocateThread (target, -1L);
	else
	    return allocateThread (target, timeout);
    }

    /**
     * Blocking version of thread allocation.
     * @param target The target object to run.
     * @return A thread instance or <strong>null</strong>.
     * @exception InterruptedException If the caller is interrupted, or
     * the thread service is uninstalled.
     */

    public Thread getThread (Runnable target) 
	throws InterruptedException
    {
	return allocateThread (target, -1L);
    }

    /**
     * Get an available thread.
     * @return A Thread, if one is available right away, or <strong>null
     * </strong>.
     */

    public Thread getAvailableThread (Runnable target)
	throws InterruptedException
    {
	return allocateThread (target, 0L);
    }


    /**
     * Shut the thread pool down.
     * @param waitp If <tt>true</tt> wait for all threads to finish before
     * returning.
     */

    void shutdown (boolean waitp) {
	synchronized (this) {
	    this.shutdown = true;
	    // Notify all pool threads to stop:
	    for (int i = 0 ; i < pool.length ; i++) {
		if ( pool[i] != null )
		    pool[i].shutdown ();
	    }
	}
	// Join each of them if we have to wait:
	if ( waitp ) {
	    for (int i = 0 ; i < pool.length ; i++) {
		try {
		    if ( pool[i] != null )
			pool[i].join ();
		} catch (InterruptedException ex) {
		    // No code intended
		}
	    }
	}
	
    }

    public int getRunningThreads () {
	return runningThreads;
    }

    public int getExistingThreads () {
	return existingThreads;
    }

    public int getMaxThreads () {
	return maxThreadCount;
    }

    public int getMinThreads () {
	return minThreadCount;
    }

    public ThreadService (int     minThreadCount,
			  int     maxThreadCount,
			  int     threadIdleTimeout,
			  boolean traceRunExceptions,
			  boolean traceRunErrors)
    {
	int limit;

	this.minThreadCount = minThreadCount;
	this.maxThreadCount = maxThreadCount;
	this.idleTimeout    = threadIdleTimeout;
	this.traceRunExceptions = traceRunExceptions;
	this.traceRunErrors = traceRunErrors;


	this.pool     = new PoolThread[maxThreadCount];
		
	// We only start the minimum number of threads
	limit     = minThreadCount;

	for (int i = 0 ; i < limit ; i++) {
	    existingThreads++;
	    pushIdleThread (pool[i] = new PoolThread (i));
	}
	
	// pushIdleThread did a runningThreads--
	runningThreads = 0;
    }

}


