/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: UuidService.java,v 1.1.1.1 2001/06/01 16:51:33 jbeatty Exp $
 */
package net.jxta.search.util;

import java.util.Random;

/**
 * This class can create 16-byte globally-unique identifiers, useful for 
 * identifying network messages or tasks that propagate through a network
 * of machines.
 */

public class UuidService {
   
    private static long  lastTime = 0;
    private static long  thisTime;
    private static long  timeOffset;
    private static long  nanosPerDay;  //  Note:  this is really 100-nanos.
    private static short versionNumber;
    private static short variantNumber;
    private static short thisClockSeq;
    private static short lastClockSeq;
    private static short newUuid[] = new short[16];
    private static long  timeMask[] = new long[3];

    /**
     * Create and initialize the uuid-generation variables.
     */

    public UuidService () {
	Random randVal = new Random ();

	// Set constant values:
	versionNumber = (short) 0x0010;
	variantNumber = (short) 0x0080;
	timeMask[0] = (long) 0x00000000FFFFFFFFL;
	timeMask[1] = (long) 0x0000FFFF00000000L;
	timeMask[2] = (long) 0x0FFF000000000000L;

	for (int i = 0; i < 10; i++) {
	    newUuid[i] = 0;
	}

	int rand = 0;

	for (int i = 0; i < 6; i++) {
	    if ((i % 4) == 0)
		rand = randVal.nextInt ();
	    newUuid[i + 10] = (short) (rand & 0xff);
	    rand >>>= 8;
	}

	//  Algorithm is number of 100-nanosecond intervals since 10/15/1582.
	//  Java clocks are based on 1970.
	//  Therefore offset is the number of 100-nanosecond intervals between
	//  10/15/1582 00:00:00 and 01/01/1900 00:00:00
	nanosPerDay =  ((long)86400*10000000);
	timeOffset =  nanosPerDay*(
				   (365*(1970-1583))   // based on years
				   + (24*3)            // based on leap-year days / century
				   + (5)               // leap year days 1583-1600, inclusive
				   + (17)              // leap year days 1900-1970
				   + (17+30+31)        // remaining days in 1582
				   );

	//  Initialize elements that can subsequently change:
	thisClockSeq = (short) ((randVal.nextInt())&0x00003FFF);
	thisTime = timeOffset + (10000 * System.currentTimeMillis());
	lastClockSeq = thisClockSeq;
    }


    /**
     * Get a 16-byte globally-unique identifier.
     */

    public synchronized byte[] getUuid () {
	long nowMillis;
	boolean mustContinue;
	long timeBuff;
	int i;
	byte byteUuid[] = new byte[16];
	short tempUuid[] = new short[16];

	//  This loop will stall until the time changes if we would
	//  issue a duplicate value of timestamp + version.
	do {
	    // Convert to nanoseconds.  To Do: get real 100-nanoseconds.
	    thisTime = timeOffset + (10000 * System.currentTimeMillis());
	    mustContinue = false;

	    if (thisTime > lastTime) { // Most common case
		lastTime = thisTime;
		lastClockSeq = thisClockSeq;
	    } else if (thisTime < lastTime) { // Clock setback
		thisClockSeq++;
		thisClockSeq %= 16384;
		lastClockSeq = thisClockSeq;
		lastTime = thisTime;
	    } else {                 //  Multiple ID's in the same interval
		if (!mustContinue) {
		    thisClockSeq++;
		    thisClockSeq %= 16384;
		}
		if (thisClockSeq == lastClockSeq) {
		    mustContinue = true;
		}
	    }
	} while (mustContinue==true);

	// Now the fun.  We assemble the UUID based on a variety of parts.
	for (i=0;i<16;i++) {
	    tempUuid[i] = newUuid[i];
	}

	timeBuff = thisTime&timeMask[0];
	tempUuid[3] |= (short) (timeBuff%256); timeBuff >>>= 8;
	tempUuid[2] |= (short) (timeBuff%256); timeBuff >>>= 8;
	tempUuid[1] |= (short) (timeBuff%256); timeBuff >>>= 8;
	tempUuid[0] |= (short) (timeBuff%256); timeBuff >>>= 8;

	timeBuff = (thisTime&timeMask[1])>>>32;
	tempUuid[5] |= (short) (timeBuff%256); timeBuff >>>= 8;
	tempUuid[4] |= (short) (timeBuff%256); timeBuff >>>= 8;

	timeBuff = (thisTime&timeMask[2])>>>48;
	tempUuid[7] |= (short) (timeBuff%256); timeBuff >>>= 8;
	tempUuid[6] |= (short) (timeBuff%16); timeBuff >>>= 8;
	tempUuid[6] |= versionNumber;

	timeBuff = thisClockSeq;
	tempUuid[9] |= (short) (timeBuff%256); timeBuff >>>= 8;
	tempUuid[8] |= (short) (timeBuff%64); timeBuff >>>= 8;
	tempUuid[8] |= variantNumber;

	for (i=0;i<16;i++) {
	    byteUuid[i] = (byte)tempUuid[i];
	}

	return byteUuid;
    } 

    /**
     * The getUuid method returns an array of bytes, which is convenient
     * for writing out the uuid to the network, but sometimes it is useful
     * to print it to console or log for debugging purposes or to use it
     * as a key in a hashtable, so this method converts the uuid to a
     * printable string.
     */

    public static String hexify (byte[] buf) {
	String s = new String();
	int hi, lo;
	char hic, loc;

	for (int i = 0; i < buf.length; i++) {
	    hi = ((((int) buf[i]) & 0x000000ff)) / 16;
	    lo = ((((int) buf[i]) & 0x000000ff)) % 16;

	    hic = (char) ((hi > 9) ? (hi - 10 + (int) 'a') : (hi + (int) '0'));
	    loc = (char) ((lo > 9) ? (lo - 10 + (int) 'a') : (lo + (int) '0'));

	    s += hic + "" + loc;
	}

	return s;
    }
}
