/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ValidatingXmlParser.java,v 1.3 2001/06/12 18:09:48 yaroslav Exp $
 */
package net.jxta.search.util;

import java.io.*;


public class ValidatingXmlParser {

    

    public interface ParserCallback {
	
	public void startTag (Tag tag, byte[] chars, int start, int len)
	    throws XmlParser.Exception;
	
	public void chars (Tag tag, String chars)
	    throws XmlParser.Exception;
	
	public void endTag (Tag tag, byte[] chars, int start, int len)
	    throws XmlParser.Exception;
	
	public void cdata (Tag tag, byte[] chars, int start, int len)
	    throws XmlParser.Exception;
			   
	
    }



	static class TagStack {
		
		Tag top;

		void push (Tag t) {
			if (top == null)
				top = t;
			else {
				t.next = top;
				top = t;
			}
		}

		Tag pop () {
			Tag t = top;
			
			if (top != null) 
				
				top = top.next;
			
			return t;
		}
	}

	
	static class ParserCallbackImpl implements XmlParser.ParserCallback {
	    TagStack       stack;
	    Tag            current;
	    Tag[]          potential;
	    ParserCallback callback;
	    boolean        expectEndTag;
	    StringBuffer   sb = new StringBuffer (512);

	    	    
	    ParserCallbackImpl (Tag root, ParserCallback cb) {
		stack     = new TagStack ();
		current   = null;
		potential = new Tag[] { root };
		callback  = cb;
	    }
	    
	    public void startTag (byte[] chars, int start, int len) 
		throws XmlParser.Exception
	    {
		// System.out.println ("cur tag: " + 
		// new String (chars, start, len));
		
		if (potential == null) {
		    throw new XmlParser.Exception ("Not expecting any tag.");
		} else if (potential.length == 0) {
		    callback.cdata (current, chars, start, len);
		    return;
		} else {
		    
		    for (int i = 0;  i < potential.length;  i++) {
			
			if (XmlParser.startsWith (potential[i].name, 
						  chars, start+1, len-1)) {
			    
			    
			    stack.push (current);
			    current = potential[i];
			    potential = current.children;

			    if (potential == null)
				sb.setLength (0);
			    callback.startTag (current, chars, start, len);
						
			    expectEndTag = (chars[start+len-2] != (byte) '/');
			    if (! expectEndTag) {
				callback.endTag (current, chars, start, len);
				current   = stack.pop ();
				if (current != null)
				    potential = current.children; 
			    }
			    
			    return;
			}
		    }
		}
		
		throw new XmlParser.Exception ("Invalid tag " + 
					       new String (chars, start, len) + 
					       ", was expecting " + 
					       concat (potential));
	    }
	    
	    public void chars (byte[] chars, int start, int len) 
		throws XmlParser.Exception
	    {
		if (potential == null) {
		    sb.append (new String (chars, start, len));
		} else if (potential.length == 0) {
		    callback.cdata (current, chars, start, len);
		}
	    }
	    
	    public void endTag (byte[] chars, int start, int len) 
		throws XmlParser.Exception
	    {
		// System.out.println ("end tag " + new String (chars, start, len));
		
		if (XmlParser.startsWith(current.name, chars, start+2, len-2)){
		    

		    if (potential == null)

			callback.chars (current, sb.toString ());
		    

		    callback.endTag (current, chars, start, len);
		    
		    current   = stack.pop ();
		    
		    if (current != null)
			potential = current.children; 
		    
		} else if (potential != null && potential.length == 0) {
		    callback.cdata (current, chars, start, len);
		} else if (expectEndTag) {
		    String encountered = new String (chars, start, len); 
		    throw new XmlParser.Exception ("Encountered: " +
						   encountered +
						   ", but expecting: </"  +
						   new String (current.name) +
						   ">.");
		} else {
				/*
				  throw new XmlParser.Exception ("Tag " +
				  new String (current.name) +
				  " was not expected to have " +
				  " a terminating tag.");
				*/
		} 
	    }
	}

	static String concat (Tag[] tags) {
		if (tags == null) {
			return "none";
		} else if (tags.length == 0) {
			return "any";
		} else {
			String s = tags[0].getName ();
			
			for (int i = 1;  i < tags.length;  i++)
				s += " or " + tags[i].getName ();
			
			return s;
		}
	}

    

	public static void parse (InputStream    reader,
				  byte[]         buf,
				  ParserCallback callback,
				  Tag            root) 
	    throws IOException, XmlParser.Exception
	{
	    // parse using a non-validating parser with
	    // the validating callback

	    XmlParser.parse (reader, 
			     buf,
			     new ParserCallbackImpl (root, callback));
	}
}
