/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: XmlParser.java,v 1.3 2001/06/04 01:50:54 yaroslav Exp $
 */
package net.jxta.search.util;


import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A very light-weight non-validating xml parser.  The call-back mechanism
 * is flexible enough to be used for constructing a stack of tags or more
 * specific tasks like link extraction or indexing of both text and hrefs.
 * <p>
 * Usage:
 * <pre>
 *      XmlParser parser = new XmlParser ();
 *
 *      XmlParser.ParserCallback callback = new XmlParser.ParserCallback () {
 *
 *		public void startTag (byte[] chars, int start, int len) {
 *		    System.out.println ("Start tag: " + 
 *					new String (chars, start, len));
 *		}
 *		
 *		public void chars (byte[] chars, int start, int len) {
 *		    System.out.println ("Chars: ." +
 *					new String (chars, start, len) +
 *					".");
 *		}
 *
 *		public void endTag (byte[] chars, int start, int len) {
 *		    System.out.println ("End tag: " +
 *					new String (chars, start, len));
 *		}
 *	    };
 *
 *	byte[] buf = new char[256];
 *	
 *	InputStream reader = System.in;
 *	
 *	parser.parse (reader, buf, callback);
 * </pre>
 */


public class XmlParser {

    static final int NO_TAG      = 0;
    static final int PARSING_TAG = 1;
    static final int EAT_TAG     = 2;
    static final int EAT_TOKEN   = 3;

	/*
	static final TrieTable escapes = new TrieTable (new String[] {"&amp;",
																  "&lt;",
																  "&gt;",
																  "&quot;",
																  "&apos;" });
	*/
	/**
     * Methods of object implementing this interface will be called
     * at appropriate times during the parsing of the html.
     */

    public interface ParserCallback {
	void startTag (byte[] chars, int start, int len)
	    throws Exception;
	void chars    (byte[] chars, int start, int len)
	    throws Exception;
	void endTag   (byte[] chars, int start, int len)
	    throws Exception;
    }


    public static class Exception extends java.lang.Exception {
	java.lang.Exception nested;
	
	public Exception (String msg) {
	    super (msg);
	}

	public Exception (String msg, java.lang.Exception nested) {
	    super (msg);
	    this.nested = nested;
	}

	public java.lang.Exception getNestedException () {
	    return nested;
	}
    }

    /**
     * Parse the input stream as html.  Please note that tags longer
     * than the character buffer's size will be thrown away.  If your
     * buffer is sufficiently large, say 8k, then this obviously
     * doesn't
     * @param reader   the input stream (unbuffered, hopefully)
     *                   from which to read html
     * @param chars    the character buffer to use when reading
     * @param callback the callback interface implementation
     */
    
    public static void parse (InputStream    reader, 
			      byte[]         chars, 
			      ParserCallback callback) 
	throws IOException, XmlParser.Exception
    {
	int     state      = NO_TAG;
	boolean closeTag   = false;
	boolean potentialCloseTag = false;
	int     startChars = 0;
	int     startTag   = 0;
	int r;
	int startRead = 0;
	int endRead;
	int endToken = 0;
	byte previousChar = -1;
	boolean inAttribute = false;
	boolean potentialEntityReference = false;
	
	while ( ( r = reader.read (chars, startRead, chars.length-startRead) ) > 0) {
	    endRead = startRead + r;
	    
	for (int i = startRead;  i < endRead;  i++) {
		
	    switch (state) {

	    case EAT_TOKEN:
		switch (chars[i]) {
		case ' ': case '\t': case '\n': case '\r':
		    endToken = i;
		    state = NO_TAG;
		    break;
		case '<':
		    state = PARSING_TAG;
		    startTag = i;
		    closeTag = false;
		    potentialCloseTag = false;
		    break;
		} 
		startChars = i + 1;
		break;
	    case NO_TAG:
		
		switch (chars[i]) {
		case ' ':
		case '\t':
		case '\n':
		case '\r':
		    if (potentialEntityReference)
				throw new XmlParser.Exception ("Bare & found.");
			// endToken = i;
		    break;
			
		case '<':
			if (potentialEntityReference)
				throw new XmlParser.Exception ("Bare & found.");

		    if (startChars != i) {
			
			callback.chars (chars, startChars, i - startChars);
		    }
		    state = PARSING_TAG;
		    startTag = i;
		    closeTag = false;
		    
		    break;

		case '&':
			if (potentialEntityReference)
				throw new XmlParser.Exception ("Unescaped & found.");
			else
				potentialEntityReference = true;
			break;
		
		case ';':
			if (potentialEntityReference)
				potentialEntityReference = false;
		}

		break;
	    case PARSING_TAG:
		
		switch (chars[i]) {
		case ' ':
			if (previousChar == '<')
				throw new XmlParser.Exception ("Unescaped < in char data.");
			if (inAttribute && potentialEntityReference)
				throw new XmlParser.Exception ("Bare & found in attribute " +
											   "value.");
			break;
		case '\'':
		case '\"':
			if (! inAttribute && previousChar == '=') {
				inAttribute = true;
			} else if (inAttribute) {
				
				if (potentialEntityReference)
					throw new XmlParser.Exception ("Invalid use of & in " +
												   "attribute value.");
				inAttribute = false;
			}
			break;
		case '<':
			if (inAttribute)
				throw new XmlParser.Exception ("Unescaped " + (char) chars[i]+
											   " character found in " +
											   " attribute value.");
		case '/':
		    if (i == startTag+1) // if </
			closeTag = true;
		    else                 // or maybe it's a />
			potentialCloseTag = true;
		    break;
		case '>':
			
			if (inAttribute)
				throw new XmlParser.Exception ("Attribute value must be " +
											   "terminated by a ' or \".");
		    if (potentialCloseTag) {
			callback.startTag (chars, startTag, i - startTag+1); 
			callback.endTag (chars, startTag, i - startTag+1);
		    } else if (closeTag) {
			callback.endTag (chars, startTag, i - startTag+1);
		    } else {
			callback.startTag (chars, startTag, i - startTag+1); 
		    } 
		    state = NO_TAG;
		    startChars = i + 1;
		    endToken = i+1;
		    break;

		case '&':
			if (inAttribute) {
				
				if (potentialEntityReference)
					throw new XmlParser.Exception ("Invalid & usage.");
				else
					potentialEntityReference = true;
			}
			break;
		case ';':
			if (potentialEntityReference)
				potentialEntityReference = false;
			
			break;
		default:
			if (previousChar == '=') // and current char isn't ' or "
				throw new XmlParser.Exception ("Attribute values must be " +
											   "quoted and the quotes must " +
											   "immediately follow the = " +
											   "sign.");
		    if (potentialCloseTag)
				potentialCloseTag = false;
							
		}
		break;

	    case EAT_TAG:
		if (chars[i] == '>') {
		    state = NO_TAG;
		    startChars = i+1;
		    endToken = i+1;
		}
	    } 
		
		previousChar = chars[i];
	}

	
	switch (state) {
	case NO_TAG:
	    
	    if (startChars < endRead) {
		if (endToken - startChars > 0)
		    callback.chars (chars, startChars, 
				    endToken - startChars);
		/* copy the beginning of the next token down to the
		   beginning of the buffer */
		if (endRead - endToken < chars.length) {
		    
		    System.arraycopy (chars, endToken, 
				      chars, 0, endRead - endToken);
		    startRead = endRead - endToken;
		} else {
		    state = EAT_TOKEN;
		    startRead = 0;
		}
	    }
	    
	    
	    startChars = 0;
	    endToken = 0;
	    startChars = 0;
	    break;
	case PARSING_TAG:
	    
	    
	    if (startTag > endRead / 2) {
		/* if we are in the middle of parsing a tag and it is short,
		   copy it down to the beginning of the buffer. */
		System.arraycopy (chars, startTag, 
				  chars, 0, endRead - startTag);

		startRead = endRead - startTag;
		startTag  = 0;
		
	    } else {
		/* the tag is too long.. forget about it. */
		state = EAT_TAG;
		startTag = 0;
		startRead = 0;
	    }
	    break;
	}

	}

    }


    /**
     * Parse the attribute value of a given attribute name in a tag.
     * @param  attributeName  the name of the attribute
     * @param  tag            the tag in which to look for the attribute
     * @return the attribute's value or null if the tag doesn't contain that attribute.
     */
    public static String getAttributeValue (byte[] attributeName,
					    byte[] tag, int start, int len)
					    
    {
	int valStart = indexOf (attributeName, tag, start, len);
	int valLen   = 0;
	
	if (valStart == -1)
	    return null;
	
	valStart += attributeName.length + 2;
	
	if (tag[valStart] == '\"' || tag[valStart] =='\'')
	    valStart+=1;
	
	for (int i = valStart;  i < start+len;  i++) {
	    if (tag[i] == '\'' || tag[i] == '\"' || tag[i] == '>' ||
		Character.isWhitespace ((char) tag[i])) {
		valLen = i - valStart;
		break;
	    }
	}
	
	return new String (tag, valStart, valLen);
    }


    public static int getIntAttribute (byte[] attributeName,
				       byte[] tag, int start, int len,
				       int    defaultValue)
					    
    {
	int valStart = indexOf (attributeName, tag, start, len);
	int valLen   = 0;
	int val      = 0;

	if (valStart == -1)
	    return defaultValue;
	
	valStart += attributeName.length + 2;
	
	if (tag[valStart] == '\"' || tag[valStart] =='\'')
	    valStart+=1;
	
	for (int i = valStart;  i < start+len;  i++) {
	    if (tag[i] == '\'' || tag[i] == '\"' || tag[i] == '>' ||
		Character.isWhitespace ((char) tag[i])) {
		
		break;
	    } else {
		if (tag[i] <= '9' && tag[i] >= '0')
		    val = val * 10 + (tag[i] - (byte)'0');
		else
		    return defaultValue;
	    }
	}
	
	return val;
    }


    /**
     * Check whether a given character buffer starts with a certain set of characters.
     * @param chars the character buffer
     * @param what  the set of characters the buffer might start with.
     * @return true if the character buffer starts with <em>what</em>
     */
    public static boolean startsWith (byte[] what, 
									  byte[] chars, int start, int len) {
	
		if (what.length <= len) {
			for (int i = 0;  i < what.length;  i++)
				if (Character.toUpperCase ((char) chars[i+start]) != what[i])
					return false;
			
			return true;
		} else {
			return false;
		}
	}
    
    private static int indexOf (byte[] what, byte[] chars, int start, int len) {
		int i, j;
		
		for (i = start; i < start+len;  i++) {
	    
	    for (j = 0;  j < what.length;  j++) {
		
		    if (Character.toUpperCase ((char) chars[i+j]) != what[j])
			break;
	    }
	    
	    if (j == what.length) 
		return i;
	    
	}
	
	return -1;
    }


    public static void main (String[] argv) 
	throws IOException, XmlParser.Exception
    {
	XmlParser parser = new XmlParser ();
	
	XmlParser.ParserCallback callback = new XmlParser.ParserCallback () {
		
		public void startTag (byte[] chars, int start, int len) {
		    System.out.println ("Start: " + 
					new String (chars, start, len));
		}
		
		public void chars (byte[] chars, int start, int len) {
		    System.out.println ("Chars: ." +
					new String (chars, start, len) +
					".");
		}

		public void endTag (byte[] chars, int start, int len) {
		    System.out.println ("End: " +
					new String (chars, start, len));
		}
	    };

	byte[] buf = new byte[64];
	
	InputStream reader = System.in;
	
	parser.parse (reader, buf, callback);
    }




}
