/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Distance.java,v 1.1.1.1 2001/06/01 16:51:35 jbeatty Exp $
 */
/** Utility class for computing various statistical measures of 
 *  interest in information theory.
 *
 *  $Id: Distance.java,v 1.1.1.1 2001/06/01 16:51:35 jbeatty Exp $
 *
 */

package net.jxta.search.util.text;

class Distance {

    private static boolean validation = true;

    /** Default behavior is to validate all arguments to 
     *  functions; running these methods without validation
     *  requires instantiating a Distance and clearing the 
     *  validation flag.
     *
     *  @param flag is boolean to set or clear.
     */
    public void setValidation(boolean flag) {
	validation = flag;
    }


    /**
     * The Kullback Leibler distance D(p||q) between
     * distribution p and q is a measure of distance 
     * between the two distributions.  It measures how
     * many bits we would use extra if we would use the code
     * for distribution q to code objects that are emitted
     * according to distribution p.  It can be used to measure 
     * the distance between the "true" distribution and the
     * statistical model.
     */
    /** FIXME: Need ranges for allowable values of arguments
     *  to be used for validation.
     */
    public static double kullbackLeilberDistance (double[] p, double[] q) {
	double sum = 0;
	
	for (int i = 0;  i < p.length;  i++) {
	    sum += p[i] * log (2, p[i]/q[i]);
	}

	return sum;
    }


    /** FIXME: Specify the validation requirements for 
     *  the input vector arg.
     */
    public static double [] normalizeVector(double [] v) {

	double sum = 0;
	int length = v.length;
	int i;

	for (i = 0; i < length; i++) {
	    sum += Math.abs(v[i]);
	}

	for (i = 0; i < length; i++) {
	    v[i] = v[i]/sum;
	}

	try {
	    validate(v);
	} catch (PException pe) {
	    pe.printStackTrace();
	}

	return v;

    }  // close normalizeVector()


    
    public static double euclideanDistance (double[] v1, double[] v2) {

	int offset = 0;  // start from first element
	int stride = 1;  // See dlamch for details on setting stride
       
	double [] v = new double[v1.length];
 
	if (validation) {
	    try {
	        validate(v1, v2);
	    } catch (PException pe) {
		pe.printStackTrace();
	    }
	}

	for (int i = 0; i < v1.length; i++ ) {
	    v[i] = v1[i] - v2[i];
	}

	return  Dnrm2.dnrm2(v.length,v,offset, stride);
	//return Math.sqrt (squaredEuclideanDistance (v1, v2));
    }


    public static double maxnorm(double [] v) {

	double max = 0.0;
	double entry;

	for (int i = 0; i < v.length; i++) {
	    entry = Math.abs(v[i]);
	    if (max < entry) {
		max = entry;
	    }
	}

	return max;

    }  // close maxnorm()


    /** 
     * @deprecated  Use net.jxta.search.util.text.Dnrm2#dnrm2
     */
    private static double squaredEuclideanDistance (double[] v1, double[] v2) {
	double sum = 0;
	
	for (int i = 0;  i < v1.length;  i++) 
	    
	    sum += (v1[i] - v2[i]) * (v1[i] - v2[i]);
	
	return sum;
    }




    public static double manhattanDistance (double[] v1, double[] v2) {
	double sum = 0;

	for (int i = 0;  i < v1.length; i++) {
	    
	    sum += Math.abs (v1[i] - v2[i]);
	}

	return sum;
    }




   /** Steinhaus (AKA Bray and Curtis) distance
    *  is  a metric is designed and 
    *  recommended for use with count data, and is a 
    *  self-normalizing metric. Since the metric is 
    *  self-normalizing (e.g., it accounts for differences 
    *  in the range of count values), data need not be 
    *  standardized prior to its use.
    */
    /** FIXME:  Specify validation range for arguments.
     */
    public static double steinhausDistance (double[] v1, double[] v2) {

	double sum1 = 0;
	double sum2 = 0;
	double sum3 = 0;

	for (int i = 0;  i < v1.length;  i++) {
	    sum1 += Math.min (v1[i], v2[i]);
	    sum2 += v1[i];
	    sum3 += v2[i];
	}

	return 2 * sum1 / (sum2 + sum3);
    }


    private static double log (int base, double n) {
	double log_base_e_of_n = Math.log (n);
	double log_base_e_of_base = Math.log (base);
	
	return log_base_e_of_n / log_base_e_of_base;
    }




   /** Mutual information is a measure of how much information 
    *  y carries about x.
    */
    public static double getMutualInformation (double[] p_x, 
                                               double[] p_x_y, double [] p_x_bar_y) {
	/* The arguments are validated (if necessary) in the called 
	 * functions.
	 */
	return getEntropy (p_x) - getConditionalEntropy (p_x_y, p_x_bar_y);

    }  // close getMutualInformation()




   /** Entropy is the average uncertainty of a single 
    *  random variable.  Formula taken from 
    *  J. J. Verbeek: An information theoretic approach to
    *  finding word groups for text classification, MS Thesis,
    *  Institute for Logic, Language and Computation, 
    *  Amsterdam, 2000, p. 36.
    *
    * @param p_x  an array of values between 0 and 1 such that 
    *             the values sum to unity.
    */
    public static double getEntropy (double[] p_x) {

        if (validation) {
	    try {
                validate(p_x);
	    } catch (PException pe) {
	        pe.printStackTrace();
            }
	}

	double sum = 0;
	for (int i = 0;  i < p_x.length;  i++) {
	    sum += p_x[i] * log (2, p_x[i]);
	}

	return 0 - sum;
    }



   /** The conditional entropy of a discrete random 
    *  variable Y given another X, for X, Y ~ p(x,y), 
    *  expresses how much extra information you still 
    *  need to supply on average to communicate Y 
    *  given that the other party knows X.  Formula taken from 
    *  J. J. Verbeek: An information theoretic approach to
    *  finding word groups for text classification, MS Thesis,
    *  Institute for Logic, Language and Computation, 
    *  Amsterdam, 2000, p. 36.
    *
    *  @param p_x_y  Probability of x AND y, 
    *                an array of values between 0 and 1 such that 
    *                the values sum to unity.
    *  @param p_x_bar_y Probability of x given y, 
    *                   an array of values between 0 and 1 such that 
    *                   the values sum to unity.
    */
    static double getConditionalEntropy (double[] p_x_y, double[] p_x_bar_y) {

        if (validation) {
	    try {
                validate(p_x_y);
                validate(p_x_bar_y);
	    } catch (PException pe) {
	        pe.printStackTrace();
            }
	}
	
        double sum = 0;

        for (int i = 0; i < p_x_y.length; i++) {
           sum += p_x_y[i] * log(2, p_x_bar_y[i]); 
        }
	return 0 - sum;

    }  // close getConditionalEntropy()



    private static void validate(double [] vals) throws PException {

        double sum = 0.0;
	double eps = 10e-2;

	for (int i = 0; i < vals.length; i++) {
	    if (vals[i]>1.0) {
		throw new PException(vals[i]);
	    }
	    sum += vals[i];
	}

        if (sum < 1.0 - eps || sum > 1.0 + eps) {
            throw new PException("Probability vector does not sum to unity");
	}
    
    }  // close validate()


    private static void validate(double [] v1, 
                                 double [] v2) throws PException {

	if (v1.length != v2.length) {
	    throw new PException("Vectors are different lengths");
	}

    }  // close validate()



    /** FIXME: "Real" example values should be used for this
     *  testing function.
     */
    private static void testDistance() {

	/* These are probably bogus values. */
	double [] p_x = {.3, .7};
        double [] p_x_y = {.5, .5};
        double [] p_x_bar_y = {.4, .6};
	double [] v = {1, 1};


        double entropy;
        System.out.println("Entropy of p_x: " + getEntropy(p_x));

        double conditional_entropy = getConditionalEntropy(p_x_y, p_x_bar_y);
	System.out.println("Conditional Entropy: " + conditional_entropy);

	double mutualinfo = getMutualInformation(p_x, p_x_y, p_x_bar_y);
        System.out.println("Mutual Information: " + mutualinfo);

	double kl = kullbackLeilberDistance(p_x,p_x_y);
        System.out.println("Kullback Leilber Distance: " + kl);

        double sd = steinhausDistance(p_x,p_x_y);
        System.out.println("Steinhaus Distance: " + sd);

        double md = manhattanDistance(p_x,p_x_y);
        System.out.println("Manhattan Distance: " + md);

        double ed = euclideanDistance(p_x, p_x_y);
        System.out.println("Euclidean Distance: " + ed);

        double [] nv = normalizeVector(v);
	System.out.println("nv[0]: " + nv[0] + ", nv[1]: " + nv[1]);
       
	double maxval = maxnorm(p_x);
	System.out.println("Max norm of p_x: " + maxval);

    }



    public static void main(String [] args) {

	testDistance();
    }
    


}  // close class Distance
