/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SparseMatrix.java,v 1.1.1.1 2001/06/01 16:51:37 jbeatty Exp $
 */
package net.jxta.search.util.text;
import java.io.*;

class SparseMatrix {

    int totalElements;
    int totalRows;
    
    int uniqueWordCount;
    
    int[] rows;
    int[] rowStarts;
    int[] rowLengths;
    int[] data;
    int[] colIndicies;

    int[] firstWordLocations;
    int[] lastWordLocations;  /* contains points to last location of words */
    int[] wordLocationsCount;
    int[] wordLocations;

    boolean transposable;
  
    SparseMatrix () {
	rows = new int[1024];
	rowStarts = new int[1024];
	rowLengths = new int[1024];

	data = new int[1024];
	colIndicies = new int[1024];

	firstWordLocations = new int[1024];
	lastWordLocations = new int[1024];
	wordLocationsCount = new int[1024];
	wordLocations = new int[1024];

	transposable = true;
    }

    SparseMatrix (int totalRows, int totalElements, 
		  int[] rows, int[] rowStarts, int[] rowLengths,
		  int[] colIndicies, int[] data)
    {
	this.totalRows = totalRows;
	this.totalElements = totalElements;
	this.rows = rows;
	this.rowStarts = rowStarts;
	this.rowLengths = rowLengths;
	this.colIndicies = colIndicies;
	this.data = data;

	transposable = false;
    }

    void addRow (int newRowIndex, SparseVector newRow) {
	int seqRowId = totalRows++;

	realloc (seqRowId, newRow.len);

	rows[seqRowId] = newRowIndex;
	rowStarts[seqRowId] = totalElements;

	rowLengths[seqRowId] = newRow.len;
    
	/* copy the row data */
	System.arraycopy (newRow.data, 0,
			  data,        rowStarts[seqRowId],
			  newRow.len);
	System.arraycopy (newRow.rowIndicies, 0,
			  colIndicies,        rowStarts[seqRowId],
			  newRow.len);
    
	for (int i = 0;  i < newRow.len;  i++) {
	    int wordId = newRow.rowIndicies[i];
	    int lastLocation = getLastWordLocation (wordId);
	
	    if (lastLocation == -1) {
		firstWordLocations [wordId] = totalElements;
		wordLocations[lastLocation] = totalElements;
		uniqueWordCount++;
	    }
	 
	    lastWordLocations[wordId] = totalElements;
	    wordLocationsCount[wordId] ++;
	    totalElements++;
	}
	
	totalElements += newRow.len;
    }

    SparseMatrix transpose () {
	
	if (! transposable)
	    throw new IllegalStateException
		("Matrix isn't transposable because it has been created " +
		 "with a transpose operation of another sparse matrix. " +
		 "Sorry.");

	int location, nextLocation;
	int[] rows       = new int[uniqueWordCount];
	int[] rowStarts  = new int[uniqueWordCount];
	int[] rowLengths = new int[uniqueWordCount];
	int[] data       = new int[totalElements];
	int[] colIndicies = new int[totalElements];
	
	int cur = 0;
	
	for (int wordId = 0;  wordId < firstWordLocations.length;  wordId++) {
	    rows[wordId] = wordId;
	    rowStarts[wordId] = cur;
	    rowLengths[wordId] = wordLocationsCount[wordId];
	    
	    location = firstWordLocations[wordId];

	    int d = 0;

	    do {
		colIndicies[cur] = wordId;
		data[cur] = data[location];
		nextLocation = wordLocations[location];
	    } while (++d < wordLocationsCount[wordId]);

	    
		
	}

	return new SparseMatrix (uniqueWordCount, totalElements,
				 rows, rowStarts, rowLengths,
				 colIndicies, data);
    }
  
  
    void realloc (int newRowId, int newRowLength) {
    
	if (totalRows <= newRowId) {
	    rows = realloc (rows, rows.length * 2);
	    rowStarts = realloc (rowStarts, rowStarts.length * 2);
	    rowLengths = realloc (rowLengths, rowLengths.length * 2);
	}

	if (totalElements <= totalElements + newRowLength) {
	    data = realloc (data, data.length * 2);
	    colIndicies = realloc (colIndicies, colIndicies.length * 2);
	    wordLocations = realloc (wordLocations, wordLocations.length * 2);
	}
      
    }

    int[] realloc (int[] a, int newLen) {
	int[] newA = new int[newLen];
    
	System.arraycopy (a, 0, newA, 0, a.length);
    
	return newA;
    }

    
    int getLastWordLocation (int wordId) {
	if (lastWordLocations.length <= wordId) {
	    firstWordLocations = realloc (firstWordLocations, wordId + 1024);
	    lastWordLocations = realloc (lastWordLocations, wordId + 1024);
	    wordLocationsCount = realloc (wordLocationsCount, wordId + 1024);

	    for (int i = uniqueWordCount;  i < lastWordLocations.length; i++)
		firstWordLocations[i] = -1;
	    for (int i = uniqueWordCount;  i < lastWordLocations.length;  i++)
		lastWordLocations[i] = -1;
	    

	    return -1;
	} else {
	    return lastWordLocations[wordId];
	}
    }


    /*********************************
     * IO functions
     *********************************/

    void open (DataInputStream in) 
	throws IOException
    {
	byte[] buf = new byte[8192];
    
	totalElements = in.readInt ();
	totalRows     = in.readInt ();
	transposable  = (in.readInt () == 1);
    
	if (transposable)
	    uniqueWordCount = in.readInt ();
    
	rows = new int[totalRows];
	rowStarts = new int[totalRows];
	rowLengths = new int[totalRows];

	data = new int[totalElements];
	colIndicies = new int[totalElements];

	firstWordLocations = new int[uniqueWordCount];
	lastWordLocations  = new int[uniqueWordCount];
	wordLocationsCount = new int[uniqueWordCount];

	wordLocations = new int[totalElements];

	readInts (in, rows, buf);
	readInts (in, rowStarts, buf);
	readInts (in, rowLengths, buf);
   
	readInts (in, data, buf);
	readInts (in, colIndicies, buf);

	if (transposable) {
	    readInts (in, firstWordLocations, buf);
	    readInts (in, lastWordLocations, buf);
	    readInts (in, wordLocationsCount, buf);
	    readInts (in, wordLocations, buf);

	}
    }

    void readInts (InputStream in, int[] arr, byte[] buf) 
	throws IOException
    {
    
	int r = 0, n = 0;

	/* how many chunks of buf.length do we need to read in.
	   the last chunk may be less than buf.length */
	int chunkCount = (int) Math.ceil (arr.length * 4 / buf.length);
	int chunkStart;
	int chunkEnd;
	int byteCount;

	for (int chunk = 0;  chunk < chunkCount;  chunk++) {
      
	    chunkStart = chunk * buf.length / 4;
	    chunkEnd   = Math.min (chunkStart + buf.length / 4, arr.length);
      
	    byteCount   = (chunkEnd - chunkStart) * 4;

	    while ( n < byteCount &&
		    (r = in.read (buf, n, byteCount - n)) > 0)
		n += r;
      
	    if (r == -1 && n < byteCount)
		throw new IOException ("Matrix file is truncated...");
      
	    convertBytesToInts (buf, byteCount, arr, chunkStart);
	}
    }

   
    void save (DataOutputStream out) 
	throws IOException
    {
	byte[] buf = new byte[8192];
    
	out.writeInt (totalElements);
	out.writeInt (totalRows);
    
	out.writeInt (transposable ? 1 : 0);

	if (transposable)
	    out.writeInt (uniqueWordCount);
    
	writeInts (out, rows,       totalRows, buf);
	writeInts (out, rowStarts,  totalRows, buf);
	writeInts (out, rowLengths, totalRows, buf);

	writeInts (out, data,        totalElements, buf);
	writeInts (out, colIndicies, totalElements, buf);

	if (transposable) {
	    writeInts (out, firstWordLocations, uniqueWordCount, buf);
	    writeInts (out, lastWordLocations, uniqueWordCount, buf);
	    writeInts (out, wordLocationsCount, uniqueWordCount, buf);
	    writeInts (out, wordLocations, totalElements, buf);
	}
    }


    void writeInts (OutputStream out, int[] arr, int arrLen, byte[] buf) 
	throws IOException
    {
	int v;
	int j;
	int i = 0;
	int chunkCount = (int) Math.ceil (arrLen * 4 / buf.length);
	int chunkStart;
	int chunkLen;
    
	for (int chunk = 0;  chunk < chunkCount; chunk++){

	    chunkStart = chunk * buf.length / 4;
	    chunkLen   = Math.min (buf.length / 4, arrLen-chunkStart);

	    convertIntsToBytes (arr, chunkStart, chunkLen, buf);
      
	    out.write (buf, 0, chunkLen * 4);
	}

    }

    /**
     * Store the contents of the integer array from start to
     * start+len in the byte buffer.
     */

    void convertIntsToBytes (int[] arr, int start, int len, byte[] buf) {
      
	int v;
      
	for (int i = start, j = 0;  i < start+len;  i++, j+=4) {

	    v = arr[i];
	
	    buf[j]   = (byte) ((v >> 24) & 0x000000FF);
	    buf[j+1] = (byte) ((v >> 16) & 0x000000FF);
	    buf[j+2] = (byte) ((v >>  8) & 0x000000FF);
	    buf[j+3] = (byte) ((v >>  0) & 0x000000FF);
	}
    }

    /**
     * Store the contents of the byte buffer into the integer
     * array, starting with start, ending at start+len.
     */

    void convertBytesToInts (byte[] buf, int len, int[] arr, int start) {

      
	for (int i = start, j = 0;  i < start+len;  i++, j+=4) {
	
	    arr[i] = ( ((buf[j]   << 24) & 0xFF000000) | 
		       ((buf[j+1] << 16) & 0x00FF0000) | 
		       ((buf[j+2] <<  8) & 0x0000FF00) |
		       ((buf[j+3] <<  0) & 0x000000FF));
	
	}
    }
    
}
