/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TFIDFRelevance.java,v 1.1.1.1 2001/06/01 16:50:48 jbeatty Exp $
 */

package net.jxta.search.relevance;

import java.io.*;
import java.net.*;
import java.util.*;
import net.jxta.search.util.TermCountService;
import net.jxta.search.util.StringMatcher;
import net.jxta.search.util.ISVector;
import net.jxta.search.util.ISMatrix;
import java.lang.Math;


/** Used to compute relevance of search results for opensearch. 
 */
public class TFIDFRelevance {

    private static boolean debug = false;
    private static boolean stem = false;
    private static String term_count_type = "naive";    // default naive


    public TFIDFRelevance() {

    }

    public TFIDFRelevance(String count_type) {

    }

    public static void setDebug() {
	debug = true;
    }

    public static void clearDebug() {
	debug = false;
    }

    public static void setStem() {
	stem = true;
    }

    public static void unsetStem() {
	stem = false;
    }

    public static void setTermCountNaive() {
       term_count_type = "naive";
       if (debug) {
          System.out.println("Term count: " + term_count_type);
       }
    }

    public static void setTermCountBM() {
       term_count_type = "bm";
       if (debug) {
          System.out.println("Term count: " + term_count_type);
       }
    }

    public static String getTermCountAlgorithm() {
        return term_count_type;
    }


   /** 
    * Compute inverse document frequency. 
    * See <a href="#baeza-yates">Baeza-Yates</a>, p. 29,
    * Eq. 2.3, where we are using the weights directly as 
    * similarity value.
    * FIXME: This method uses sqrt, but BY uses log.  Why?
    * 
    * @param int num_docs total number of document exclusive
    *        of the query, must be &gt;= 0.  
    * @param int [] DocCount  Each index <i>i</i> of this array points
    *        the number of times the <i>ith</i> query term appears in
    *        a document.  Each entry must be &gt;= 0.
    *
    * @return float [] IDF Inverse Document Frequency of the ith
    *         query term with respect to the all the documents,
    *         computed using sqrt instead of log.
    *
    * NOTE: No range checking is done in this method, and the 
    *       behavior with invalid parameters is unkown.
    */    
    public static float[] compute_idf(int num_docs,int[] DocCount) {

	int query_len = DocCount.length;
	float[] IDF = new float[query_len];               	

    	for(int t = 0; t < query_len; t ++) {
	    if(DocCount[t] != 0) {
		/*  This is usually given as log(N/n_i) */
		IDF[t] = (float) java.lang.Math.sqrt((float)(num_docs/DocCount[t]));
	    }
	    else IDF[t] = 0.0f;
	}

	return IDF;

    }  /* close compute_idf() */



    /** 
     * Compute term frequency using log method.
     * FIXME: Need a reference for this.
     *
     * @param int [][] TermCount represents each document as a row <i>d</i>,
     *         and the number of times some query term <i>i</i>
     *         appears in the ith column for the dth document.
     *         .
     * @return float [][] TF is the term counts weighted by 
     *         1 + the log of the counts of each term 
     *         in some document <i>d</i>.
     */
    public static float[][] compute_tf_log(int[][] TermCount) {

	int num_docs = TermCount.length;
	int query_len = TermCount[0].length;

	float[][] TF = new float[num_docs][query_len];    
	for(int d = 0; d < num_docs; d ++) {
	    for(int t = 0; t < query_len; t ++) {
		if(TermCount[d][t] != 0)
		    TF[d][t] = (1.0f + 
				(float) java.lang.Math.log((float)TermCount[d][t]));
                /* We could just return the TermCounts unweighted.
                 */
		//TF[d][t] = (float)TermCount[d][t];
		else
		    TF[d][t] = 0.0f;
	    }
	}
	return TF;

    }  /* close compute_tf_log */




    /** 
     * Compute term frequency using the max counts of term per doc.
     * See <a href="#baeza-yates">Baeza-Yates</a>, p. 29.
     * @param int [][] TermCount represents each document as a row <i>d</i>,
     *         and the number of times some query term <i>i</i>
     *         appears in the ith column for the dth document.
     *         .
     * @return float [][] TF is the term counts weighted by the count
     *         of the most frequent term in some document <i>d</i>.
     */
     public static float[][] compute_tf_max(int[][] TermCount) {

	int num_docs = TermCount.length;
	int query_len = TermCount[0].length;

	float[][] TF = new float[num_docs][query_len];    
	for(int d = 0; d < num_docs; d ++) {
            int maxcount = 0;

	    // TODO: Should be a Distance.method for this...
            // Use it after changing to new build system.
	    for(int t = 0; t < query_len; t ++) {
 
                if (TermCount[d][t] > maxcount) {
                    maxcount = TermCount[d][t];
		}
	    }

	    /* Normalize for term frequencies for each document. */
	    for (int t = 0; t<query_len; t++) {
                if (maxcount != 0) {
  		    TF[d][t] = (float)(TermCount[d][t]/maxcount);
		} else {  // Should be redundant...?
                    TF[d][t] = 0.0f;
		}
	    }
            
	}
	return TF;

    }  /* close compute_tf() */


    /** 
     * Compute the number of words in each document in the 
     * document array.
     *
     * @param String [] doc_array each entry in the array is a 
     *               document as a single String.
     * @return int [] WordCount  is total number of words in 
     *         some document <i>d</i>.
     */
    private static int [] compute_word_counts(String [] doc_array) {

        int [] word_count = new int[doc_array.length];
        for (int i = 0; i<doc_array.length; i++) {
	    StringTokenizer st = new StringTokenizer(doc_array[i]);
            word_count[i] = st.countTokens();
        }
        return word_count;
    }


    /** 
     * findRel takes a String "queryString" and an ArrayList "docs". 
     *
     * @param String queryString could be a string of terms typed
     *        by a user doing a search.
     * @param ArrayList docs is a list of "documents" whose member
     *        values consist of a string for each document that might
     *        be all or part of a search result.
     *
     * @return float [] Sim is a vector of similarity values computed
     *         using the tf/idf method.  See <a href="#baezayates">
     *         Baeza-Yates</a> for details.
     */

    public static float []  findRel(String queryString, ArrayList docs) {

	int num_docs = docs.size();

	StringTokenizer query = new StringTokenizer(queryString," ,:");
	int query_len = query.countTokens();

	if (debug) {
	    System.out.println("Opensearch: query length "
			       + query_len
			       + " num docs "
			       + num_docs);
	}



	/** TermCount is the 
	 *  count of the number of docs of term 
	 *  t in document d
	 */
	int[][] TermCount = new int[num_docs][query_len]; 

	/** DocCount is the count of the number
	 *   of docs of term t in all documents
	 */
	int[] DocCount = new int[query_len];              

       /* FIXME: Java should initialize arrays to zero when 
        * allocated, find a page number in Gosling to support
        * this.  
        */
        ISMatrix.setZero(TermCount);
        ISVector.setZero(DocCount);


	/** queryArray is an array of the 
	 *  query words (or terms)
         */
	String[] queryArray = makeQueryArray(query);

	/* Each entry in docArray is a String representing the 
         * the entire document.
         */
        String [] docArray = makeDocArray(docs);

       /** 
        * Choose the term counting algorithm, naive by default.
	*/
        if (term_count_type.equals("naive")) {
	    TermCountService.compute_counts_naive(queryArray,
						  docArray,
						  TermCount,
						  DocCount); 
        } else {  // use boyer-moore
	    TermCountService.compute_counts_bm(queryArray,
					       docArray,
					       TermCount,
					       DocCount); 
        }
			    

	/*  TF is the "term frequency" of each 
	 *  term t in document d.
	 */
	float[][] TF = compute_tf_max(TermCount);

	/* IDF is the "inverse document frequency"
	 * of term t over all documents
	 */
	float[] IDF = compute_idf(num_docs, DocCount);

        /* We need to know how many total words in each document
         * to compute similarities between documents.
         */
        int [] WordCount = compute_word_counts(docArray);
        /* Compute similarities between query and documents. */
        float [] Sim = computeTFIDFSimilarity(TF, IDF, WordCount);


	if (debug) {
           printRelevanceStats(DocCount, TF, TermCount, IDF, docs, Sim);
        }
       
        return Sim;

    }  /* close findRel() */



    /** 
     * Turn an ArrayList whose elements are Strings into  
     * an array of Strings.
     *
     * @param ArrayList docs An ArrayList of Strings each representing 
     *        a single document.
     * 
     * @return String [] docArray Each entry in the array is the document
     *         represented as a String.
     */
    private static String [] makeDocArray(ArrayList docs) {

        String [] docArray = new String[docs.size()];

        for(int d = 0; d < docArray.length; d ++) {
        
               docArray[d] = (String)docs.get(d);
        }
        return docArray;    
    }


    /**
     * Takes the raw query terms stored in a StringTokenizer, 
     * and constructs an array of Strings where each String entry
     * is a stemmed, upper-cased query term.
     *
     * @param StringTokenizer st tokenizer of query terms.
     *
     * @return String [] queryArray Each entry is stemmed and 
     *         upper-cased term of query string.
     */
    private static String [] makeQueryArray(StringTokenizer query) {

        int query_len = query.countTokens();
	String[] queryArray = new String[query_len];     
        String upper;
        String stemmed;

	for(int t = 0; t < query_len; t++) {

	    queryArray[t] = query.nextToken();
	    upper = queryArray[t].toUpperCase();


	    /* Do some stemming to help matching.
	     */
            if (stem) {

		if (debug) {
		    System.out.println("Term t = " + t);
		    System.out.println("Unstemmed query term: " + queryArray[t]);
		}   

		stemmed = TermCountService.quick_stem(upper);

		if (queryArray[t].length() == 1  && debug){
		    System.out.println("Stemmed out of existence... length = "
				       + queryArray[t].length());
                
		} else {
		    //upper = queryArray[t].toUpperCase();
		    //stemmed = TermCountService.quick_stem(upper);
		    //queryArray[t] = stemmed;
		}

		queryArray[t] = stemmed;

		if (debug) {
		    System.out.println("Stemmed query term: " + queryArray[t]);
		}

	    } else {
                queryArray[t] = upper;
	    }
        }
        
        return queryArray;
    }

    /* Helper method for the previous. */
    public static String [] makeQueryArray(String queryString) {

	StringTokenizer query = new StringTokenizer(queryString," ,:");
	//int query_len = query.countTokens();
        return makeQueryArray(query);
    }
	

    /**
     * Simple TFIDF weighting is used for a similarity measure.
     *
     * FIXME:  WordCount is used for the summing part of this and 
     * appears that it can be removed from the call.
     */
    public static float [] computeTFIDFSimilarity(
                  float [][] TF, float [] IDF, int [] WordCount) {

        int num_docs = TF.length;

	float[] Sim = new float[num_docs]; 
 
        ISVector.setZero(Sim);

        int query_len = IDF.length;

	for(int d = 0; d < num_docs; d ++) {
	    for(int t = 0; t < query_len; t ++) {
		Sim[d] += IDF[t] * TF[d][t];
	    }
	}
        return Sim;
   }


    /**
     * Normalize a vector.  This code should probably 
     * be in the Distance class.  It was a part of the 
     * similarity method, but was never used in that 
     * method.  Normalization is not always necessary,
     * but certain algorithms (e.g., ephemeral clustering)
     * depend on similarities between 0 and 1.
     *
     * FIXME: Explain why WordCount was needed, reimplement
     * if necessary.
     *
     * @param float [] vector arbitrary vector of floats.
     *
     * @return float array [] normalized input vector, 
     *         this implementation overwrites the input.
     */
    /*
    public static float []  normalize(float [] vector, int [] WordCount) {

	float Sum = 0.0f;

	for(int d = 0; d < num_docs; d ++) {
	    // divide by the number of words in document d
	    Sim[d] /= java.lang.Math.sqrt(WordCount[d]); 
	    Sum += Sim[d];
	}
        return Sim;
    }
    */

    private static void printRelevanceStats(int [] DocCount, float [][] TF,  
                                            int [][] TermCount, float [] IDF,
                                            ArrayList docs, float [] Sim) {

        int query_len = DocCount.length;
        int num_docs = TermCount.length;
        
	for(int t = 0; t < query_len; t ++) {
	    System.out.println("DocCount["+t+"] = "
		                   +DocCount[t]+" IDF = " + IDF[t]);
        }


        for(int d = 0; d < num_docs; d ++) {
	    for(int t = 0; t < query_len; t ++) {
		    System.out.println("TermCount["+d+"]["+t+"] = "
		                       +TermCount[d][t]+" TF = " + TF[d][t]);
            }
	    System.out.println("Sim["+d+"] = "+Sim[d] + " " + docs.get(d));
        }

    }



     private static void test() {

        TestTFIDFRelevance.test(System.out);

     }



    /** Test method  - takes strings as input - 
     *  try java TFIDFRelevance string0 string1 string2 
     *  to find relevance of string0 to string1 and string2
     */
    private static void test(String [] argv) {

	String pattern = argv[0];
	ArrayList docs = new ArrayList(argv.length-1);

	if (debug) {
	    System.out.println("Terms to match: " + argv[0]);
	}


	for(int i = 1; i < argv.length; i ++) {
	    docs.add(argv[i]);
	    if (debug) {
	        System.out.println("Test string (doc): " + argv[i]);
	    }
	}


	//findRel(pattern, docs);


    }  /* close test() */


    public static boolean test(PrintStream ps) {

       return TestTFIDFRelevance.test(ps);

    }

    public static void main (String[] argv) {

       TestTFIDFRelevance.test(System.out);

    }  /* close test() */


}  /* class TFIDFRelevance */



/*
 */
class TestTFIDFRelevance {

    
    public static boolean test(PrintStream ps) {
   
       ps.println("----------- Begin Testing TFIDFRelevance-----------");

       boolean passed = true;

        TFIDFRelevance sr = new TFIDFRelevance();

	String [] strs = { "blah  :  bar",
			   "test1",
			   "blah bar foo baz",
			   "testem",
			   "test1",
			   "teste2",
                           "<html><head></head><body>Test html formatted document.</body></html>" };

       	//sr.setDebug();
       passed = test(strs,ps);

       if (passed) {
           ps.println("Passed TFIDFRelevance unit tests.");
       } else {
           ps.println("FAILED TFIDFRelevance unit tests.");
       }

       ps.println("----------- End Testing TFIDFRelevance-----------");
	
       return passed;

    }


    /** Test method  - takes strings as input - 
     *  try java TFIDFRelevance string0 string1 string2 
     *  to find relevance of string0 to string1 and string2
     */
    private static boolean test(String [] argv, PrintStream ps) {

        boolean passed = true;

	String pattern = argv[0];
	ArrayList docs = new ArrayList(argv.length-1);


	/*
        ISVector.print(TFIDFRelevance.makeQueryArray(pattern),ps,"Query array");
	ps.println("Terms to match: " + argv[0]);
	for(int i = 1; i < argv.length; i ++) {
	    docs.add(argv[i]);
	    ps.println("Test string (doc): " + argv[i]);
	}
	*/

	//findRel(pattern, docs);

        if (testSetTermCounts(ps)) {
            ps.println("Passed term unit test");
        } else {
            ps.println("Failed term unit test");
            passed = false;
        }

        if (testMakeQueryArray(pattern)) {
            ps.println("Passed make query array unit test");
        } else {
            ps.println("Failed make query array unit test");
            passed = false;
        }

        if (testIDF(ps)) {
            ps.println("Passed IFD unit test");
        } else {
            ps.println("Failed IDF unit test");
            passed = false;
        }

        if (testTF(ps)) {
            ps.println("Passed TF unit test");
        } else {
            ps.println("Failed TF unit test");
            passed = false;
        }

        if (testSim(ps)) {
            ps.println("Passed Sim unit test");
        } else {
            ps.println("Failed Sim unit test");
            passed = false;
        }
  
        return passed;


    }  /* close test() */


    private static boolean testMakeQueryArray(String querystring) {
        
        String [] test_query_array = { "BLAH", "BAR" };
        String [] query_array = TFIDFRelevance.makeQueryArray("blah bar");

        boolean passed = true;
   
        try {
            if (!ISVector.equals(query_array,test_query_array))
                passed = false;
	} catch (Exception e) {
        }

        return passed;
    }



    /* Test the get/set for term count method flag. 
     * There is an intermittent bug associated with
     * these methods that is hard to pin down.  It 
     * is NOT here...
     */
     private static boolean testSetTermCounts(PrintStream ps) {
 
         boolean passed = true;

         TFIDFRelevance.setTermCountNaive();
         if ("naive".equals(TFIDFRelevance.getTermCountAlgorithm())) {
            ps.println("Passed term count unit test naive");
         } else {
            ps.println("Failed term count unit test naive");
            passed = false;
         }            

         TFIDFRelevance.setTermCountBM();
         if ("bm".equals(TFIDFRelevance.getTermCountAlgorithm())) {
            ps.println("Passed term count unit test bm");
         } else {
            ps.println("Failed term count unit test bm");
            passed = false;
         }            
   
         return passed;

     }



    /* Comparing floating point numbers for bitwise
      * equality is not usually very productive.
      */
     static float eps = 0.000001f;

    /**
     * FIXME: Make sure the numbers used are "realistic".
     */
     private static boolean testIDF(PrintStream ps) {

	 boolean passed = false;
	 

        /* Say the query is "foo bar", and "foo" appears 
         * twice and "bar" appears once, in some (1) particular
         * document. 
         */
       
         int num_docs = 1;
         int [] DocCount = {2,1};

         float [] idf = TFIDFRelevance.compute_idf(num_docs,DocCount);
  
   
        /* Compare the result with the "true" value... */
        float [] true_idf0 = {0.707107f, 1.0f};

	try {
	    if(!ISVector.equals(idf,true_idf0,eps)) {
                passed = false;
	    }
	} catch (Exception e) {
	}

         //ISVector.print(idf,System.out);

         return true;

     }  /* close testIDF() */



    /**
      * FIXME: Use some "realistic" number for this testing.
      */
    private static boolean testTF(PrintStream ps) {

	boolean passed = true;

	int num_docs = 1;
	int query_len = 2;
	int [][] TermCount = { {2,1} };
        
	float [][] TF = TFIDFRelevance.compute_tf_log(TermCount);

	//ISMatrix.print(TF,ps,"Term frequency");

	float [][] test_TF = { { 1.6931472f,  1.0f} };
         
	try {
	    if(!ISMatrix.equals(TF,test_TF,eps)) {
                passed = false;
	    }
	} catch (Exception e) {
	}
             
	//TODO: Implement the following.
	/* Now check that the max tf works...*/
        // TF = TFIDFRelevance.compute_tf_max(TermCount);
        // test_TF = {{put numbers here.}};
	//try {
	//    if(!ISMatrix.equals(TF,test_TF,eps)) {
        //        passed = false;
	//    }
	//} catch (Exception e) {
	//}

        return passed;
    }


    /**
      * FIXME: The numbers are just dummies, "real" numbers
      * need to be computed and used for the array initializations.
      * Also, range and/or sanity checking needs to be 
      * 1. implemented or 2. explained why not necessary.
      */
    private static boolean testSim(PrintStream ps) {

	boolean passed = false;
        float [][] TF = { {2f, 3f} };
        float [] IDF = {1f, 2f};
        int [] WordCount = {2, 1};

        float [] sim = TFIDFRelevance.computeTFIDFSimilarity(TF,IDF,WordCount);

        float [] test_sim = {1.0f, 2.0f};

	try {
	    if(!ISVector.equals(sim,test_sim,eps)) {
		passed = false;
	    }
	} catch (Exception e) {
	}

	return passed;
    }




}  /* class TestTFIDFRelevance */


