/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: HitList.java,v 1.10 2001/06/12 20:58:53 ddoolin Exp $
 */

package net.jxta.search.websearch;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.servlet.http.*;
import javax.servlet.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.SAXParser;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.w3c.dom.*;

import net.jxta.search.peer.ProviderInfo;
import net.jxta.search.util.Html;
import net.jxta.search.util.PriorityQueue;
import net.jxta.search.util.SchedulerService;
import net.jxta.search.util.UuidService;
import net.jxta.search.util.Soundex;
import net.jxta.search.util.HexTools;
import net.jxta.search.util.LogService;
import net.jxta.search.relevance.TFIDFRelevance;

class Response {
    ProviderInfo[] providers;
    ArrayList      cookies;
    ArrayList      hits;
    int            maxRank;
    int            rank;

    Response (ProviderInfo[] providers) {
        this.providers = providers;
        this.cookies   = new ArrayList ();
        this.hits      = new ArrayList ();
    }
}


/**
 * Container for individual hits
 */
class Hit {
    static final String OP_SPAM = "spam";
    static final String OP_GOOD = "good";

    String   href;
    String   html;
    int      rank;
    float    relevance;
    Response response;
    int      hid;        // static throughout the life of the hit
    int      used = -1;   // this hits place in the display list

    String   title = null;
    String   link = null;
    String   detail;

    String   image;
    int      imageHeight;
    int      imageWidth;


    /**
     *  Validates that the minimum elements are present
     **/

    public boolean validate() throws Exception {
    
       if (title == null) {
          throw new Exception("Null title in Hit");
       }
 
       if (link == null) {
          throw new Exception("Null link in Hit");
       }
 
       if (image != null) {
           if (imageWidth <= 4)
               throw new Exception ("Image width less than 5 pixels");
 
           if (imageHeight <= 4)
               throw new Exception ("Image height less than 5 pixels");
       }

       return true;

    }

    public void adjustImageSize () {
       if (imageHeight >= 80 || imageWidth >= 80) { // too big, adjust size

          float aspect = 0;             // aspect ratio width:height
          aspect = (float) imageWidth / (float) imageHeight;
          System.out.println ("Aspect ratio = " + aspect );

          if (aspect > 1) {              // wider than tall
              System.out.println ("Aspect > 1");
              imageWidth  = 80;
              imageHeight = Math.round (imageWidth / aspect);

          } else if (aspect < 1) {       // taller than wide
              System.out.println ("Aspect < 1");
              imageHeight = 80;
              imageWidth  = Math.round (imageHeight * aspect);

          } else {                       // equal
              System.out.println ("Aspect = 1");
              imageWidth  = 80;
              imageHeight = 80;
          }
       }
       return;        
    }


   /** Compare the size of the image with a 60 x 60 constraint
    * in the html table column.  If height, width or both
    * are larger than 60, rescale the height and width.
    */
    public void setImageSize(int height, int width) {
  
  
       if (height <= 60 && width <= 60) {
          imageHeight = height;
          imageWidth = width;
          return;
       }
 
       double scale;
 
       if (height > width) {
          scale = 60.0/height;
       } else {
          scale = 60.0/width;
       }
 
       this.imageHeight = (int)scale*height;
       this.imageWidth = (int)scale*width;

    }


    /**
     * Returns a String formmatted to suit logging of user feedback activity.
     */
    public String toBinkLog (String opinion) {
      StringBuffer result = new StringBuffer (500);

      result.append ("PROVIDER=" + this.response.providers[0].getId ());
      result.append (",OP=" + opinion);
      result.append (",LINK=" + this.link);
      result.append (",TITLE=" + this.title);
      result.append (",DETAIL=" + this.detail);

      return (result.toString ());
    }

    public String toRedirectLink (String uuid, String redirectRoot) {
      return (redirectRoot + "/" + uuid + "/" + this.link );
      
    }


   /** Each hit gets displayed on the opensearch web page as 
    * a row in an html table.  
    *
    * @see OpensearchServlet
    * @returns String that contains the entire hit as a properly 
    *          fomratted row in html.
    */
    public void toHtmlTableRow (PrintWriter out,
				int         enableImages, 
				String      uuid, 
				String      query,
				String      redirectRoot) 
    {
	
         out.print("<tr height=100>");  // Open the row

         out.print("<td width=110 valign=top>");  // Open the provider info column 

         out.print ("<a href='" + this.response.providers[0].getLink () + "' style='text-decoration:none;'>");

         if (this.response.providers[0].getImageURL () != null && 
             enableImages == Preferences.IMAGES_ENABLED) {
             // User wants images and icon is present
             // TODO: resize icon if too big
             out.print ("<img border=0 width=60 src='" + this.response.providers[0].getImageURL () + "'");

             if (this.response.providers[0].getDescription () != null) 
                 out.print ("alt=\"" + response.providers[0].getDescription () + "\"");
             else if (this.response.providers[0].getTitle () != null)
                 out.print ("alt=\"" + response.providers[0].getTitle () + "\"");

             out.print (">");

         } else {
             // Icon unavailable or user does not want images
             out.print ("<font size=-1 color=green><b>" + response.providers[0].getTitle () + "</b></font>");
         }
         out.print("</a>");

         out.print("<br>");
	 
	 buildRateButton (out, uuid);

         out.print("</td>"); // Close the provider info column

         out.print("<td valign=top>");
         out.print("<a href=\"" + this.toRedirectLink (uuid, redirectRoot)  + "\">");
         if (this.getTitle ().length () <= 60)
             out.print (Html.highlightTerms (query, this.getTitle ()));
         else
             out.print (Html.highlightTerms (query, this.getTitle ().substring (0,60)) + "...");
         out.print("</a>");


         if (this.detail != null) {
             out.print("<br>");
             if (this.detail.length () <= 300)
                 out.print (Html.highlightTerms (query, this.detail));
             else
                 out.print (Html.highlightTerms (query, this.detail.substring (0,300)) + "...");
         }
         out.print("</td>"); // Close the hit detail column

         if (this.image != null) {
             out.print ("<td valign=top>");
             out.print ("<a href=\"" + this.toRedirectLink (uuid, redirectRoot) + "\">");
             out.print ("<img src=" + this.image + " border=0 alt=\"" + this.getTitle () + "\"></a>");
             out.print ("<td>");
         }

         out.print("</tr>"); // Close the row

}
    
    
    static String aStart    = "<a target=\"_blank\" onClick=\"javascript:window.open('/websearch/search?action=b&op=";
    static String aMiddle   = "&hid=";
    static String aEnd      = "', 'ratingWindow', 'toolbar=no,scrollbars=no,resizeable=yes,width=320,height=210')\">";
    static String imgStart  = "<IMG src=\"images/";
    static String imgMiddle = "\" border=0 alt=\"";
    static String imgEnd    = "\"></a>";

    private void buildRateButton (PrintWriter out, String uuid) {

        out.print ("&nbsp;&nbsp;<font size=-3>Rate This Hit:");
	
	// good button
	out.print (aStart);
	out.print (OP_GOOD);
	out.print (aMiddle);
	out.print (hid);
	out.print ("&uuid=");
	out.print (uuid);
	out.print (aEnd);
	out.print (imgStart);
	out.print ("aups.gif");
	out.print (imgMiddle);
	out.print ("Exactly what I was looking for");
	out.print (imgEnd);
	
	// bad button
        out.print (aStart);
	out.print (OP_SPAM);
	out.print (aMiddle);
	out.print (hid);
	out.print ("&uuid=");
	out.print (uuid);
	out.print (aEnd);
	out.print (imgStart);
	out.print ("adwns.gif");
	out.print (imgMiddle);
	out.print ("This hit was spam"); 
	out.print (imgEnd);

    }



    public void setHref(String href) {
       this.href = href;
    }

    public String getHref() {
        return href;
    }

    public void  setHtml(String html) {
       this.html = html;
    }

    public String getHtml() {
       return html;
    }

    public void setRank(int rank) {
        this.rank = rank;
    }

    public int getRank() {
       return rank;
    }

    public void setRelevance(float relevance) {
       this.relevance = relevance;
    }

    public float getRelevance() {
        return relevance;
    }

    public void setTitle(String title) {
       this.title = title;
    }
   
    public String getTitle() {
       return title;
    }
     
    public void setLink(String link) {
        this.link = link;
    }

    public String getLink() {
        return link;
    }

    public void setDetail(String detail) {
       this.detail = detail;
    }

    public void setImage(String image) {
       this.image = image;
    }

    public String getImage() {
       return image;
    }

    /**
     * Returns the contents of the object in comma 
     * deliminated name=value formated String.
     */
    public String toString () {
        return ( "USED=" + used 
                 + ",HREF=" + href 
                 + ",HTML=" + html 
                 + ",RANK=" + rank 
                 + ",RELEVANCE=" + relevance 
                 + "," + response.providers[0].toString () 
                 + ",TITLE=" + title 
                 + ",LINK=" + link );
    }


}  





/**
 * Provides a container and a interface to all hits returned for a query.
 */
class HitList extends ArrayList {
    private int sortType = Preferences.SORT_RELEVANCE;
    private int numUsed;
    private String query;

    // TODO: prevent resorting if the sortType hasn't changed and no new Hits have been
    //       added


    HitList (String query) {
        this.query = query;
    }


    /**
     * Overides add to prevent objects other than of type Hit from entering the list.
     */
    public void add (Hit hit) {
        // Prevent an object other than a Hit from being added
        hit.hid = (this.size () + 1);
        super.add(hit);
    }

/**
     * Returns a hit with a used value equal to the index value or it returns the next 
     * available hit if no hit with the index exists.
     */
    public Hit getHit (int index) {
        Hit hit = null;
        Hit nextHit = null;

    // size is num elements, array is zero rel 
    // TCOK
        for (int i=0; i < this.size (); i++) {
            hit = (Hit) super.get (i);
            if (nextHit == null && hit.used < 0)
                nextHit = hit;
            else if (hit.used == index)
                break;
        }

    // As we search the list for a specific hit, we point nextHit to the first encountered
    // unused hit, if the index passed to the method is not found we make nextHit
    // the hit we return and we set its position used (index) and up the numUsed count.
    // TCOK
        if (hit.used != index) {
            hit = nextHit;
            numUsed++;
            hit.used = numUsed;
        }

        return (hit);
    }


    /**
     * Returns a Hit from the list based upon hit ID (HID). Each hit is tagged with
     * a HID as it is received by the app.  The HID in combination with the
     * GUID create a unique key identifying each individual hit.
     */
    public Hit getHitByHid (int hid) {
        Hit hit     = null;
        Hit currHit = null;

        // TCOK
        for (int i=0; i < this.size (); i++) {
            currHit = (Hit) super.get(i);
            if (currHit.hid == hid) {
                hit = currHit;
                break;
            }
        }
        return (hit);
    }



   /** 
    * Search relevance evaluation has been abstracted out into 
    * its own module.  The original class is retained for 
    * compatibility until the new stuff is fully tested.
    */
    private void handleNewRelevance() {

       /**
        * Step 1:  Set the ArrayList to be just set of strings that 
        * are the title and detail of the documents.
        */ 
       int num_docs = this.size();
 
       if (num_docs == 0) return;

       ArrayList doclist = new ArrayList(num_docs);

       //String[] docArray = new String[num_docs];

       for(int d = 0; d < num_docs; d ++) {

          Hit this_hit = (Hit)this.get(d);
          String s = " ";

          if ((this_hit.title != null) || (this_hit.detail != null)) {
             if (this_hit.title != null)
                s = this_hit.title;
             if (this_hit.detail != null)
                s = s + this_hit.detail;
           } else {
               s = this_hit.html;
           }

	  /* This is to match updated code in SearchRelevance. */
	  if (s == null) {
	      s = "empty";
	  }

	  /* The docArray is a holdover from the previous API,
	   * which now takes an ArrayList, as shown.
	   */
           //docArray[d] = new String(" " + s.toUpperCase());
           doclist.add(new String(" " + s.toUpperCase()));
       }


       // Step 2: Find the similarity array
       String strippedQuery = Html.stripHtmlTags(query);
       String finString = strippedQuery.trim();       
       float [] Sim = TFIDFRelevance.findRel(finString, doclist);
 
       // Step 3: Set the relevance of the hits.
       for(int d = 0; d < num_docs; d ++) {
            Hit this_hit = (Hit)this.get(d);
            this_hit.relevance = Sim[d];
       }

    }


    private void handleRelevance(boolean use_new_relevance) {

       if (use_new_relevance) {
          handleNewRelevance();
       }  else { // use old relevance
	   String strippedQuery = Html.stripHtmlTags(query);
	   SearchRelevance rel = new SearchRelevance();
	   String finString = strippedQuery.trim();
	   rel.findRel(finString, this);
       }

    }


    /**
     * Sorts the hits according to the sortType. 
     */
    public void sort () {
        if (sortType == Preferences.SORT_RELEVANCE) {

            // Sort ALL the available hits based on relevance evaluation            
            //SearchRelevance rel = new SearchRelevance();
            //rel.findRel(query, this);
           /* Eventually, this function will be able to deal with 
            * several types of relevance evaluation.  Right now, the 
            * flag is just used to indicate whether to use the original
            * (false) or the new (true).
            */
            handleRelevance(true);

            Collections.sort (this, new Comparator () {
                    public int compare (Object o1, Object o2) {
                        Hit h1 = (Hit) o1;
                        Hit h2 = (Hit) o2;
                
                        float x1 = h1.relevance;
                        float x2 = h2.relevance;
                        
                        return ((x1 < x2) ? 1 : (x1 == x2 ? 0 : -1));
                     }
                 });
        } else if (sortType == Preferences.SORT_RESP_TIME) {
            // Sort ALL the available hits based on the providers response time
            Collections.sort (this, new Comparator () {
                public int compare (Object o1, Object o2) {
                        Hit h1 = (Hit) o1;
                        Hit h2 = (Hit) o2;
                
                        long x1 = h1.response.providers[0].getTimestamp ();
                        long x2 = h2.response.providers[0].getTimestamp () ;
                        
                        return ((x1 > x2) ? 1 : (x1 == x2 ? 0 : -1));
                     }
                 });
        } else {
            // Since hits are tagged with a Hit ID (HID) as they are received
            // sorting them based on the hid returns them to their original 
            // order.
            Collections.sort (this, new Comparator () {
                    public int compare (Object o1, Object o2) {
                        Hit h1 = (Hit) o1;
                        Hit h2 = (Hit) o2;
                
                        int x1 = h1.hid;
                        int x2 = h2.hid;
                        
                        return ((x1 > x2) ? 1 : (x1 == x2 ? 0 : -1));
                     }
                 });
        }
    }


    /**
     * Returns the number of hits that have been returned by getHit () since 
     * the sort type was last set.
     */
    public int getNumUsed () { return (this.numUsed); }


    /**
     * Returns the number of hits not marked as used
     */
    public int getNumNew () { return (this.size () - this.numUsed); }


    /**
     * Sets the sort type. If the sort type changes, the used field on each
     * hit is reinitialized to indicate that the hit has not been displayed
     * using the current sort method.
     */
    public void setSortType (int sort) {
        if (sort != this.sortType) {
            // if the sort type changed we start displaying from scratch
            Hit hit;
            for (int i=0; i < this.size (); i++) {
                hit = (Hit) super.get(i);
                hit.used = -1;
            }
            this.sortType = sort;
            this.numUsed = 0;
        }
    }


    /**
     * Sets the sort type.
     */
    public int getSortType () { return (this.sortType); }



}

