/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: OpensearchServlet.java,v 1.2 2001/06/06 07:59:10 tcamarda Exp $
 */

package net.jxta.search.websearch;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.servlet.http.*;
import javax.servlet.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.w3c.dom.*;

import net.jxta.search.peer.Request;
import net.jxta.search.peer.RouterClient;
import net.jxta.search.peer.HttpRouterClient;

import net.jxta.search.util.Html;
import net.jxta.search.util.PriorityQueue;
import net.jxta.search.util.SchedulerService;
import net.jxta.search.util.UuidService;
import net.jxta.search.util.Soundex;
import net.jxta.search.util.HexTools;
import net.jxta.search.util.LogService;
import net.jxta.search.util.HttpConnection;

/** 
 *  Provides HTML interface for Opensearch application
 *
 *  @see net.jxta.search.websearch.Preferences
 */
public class OpensearchServlet extends HttpServlet {

    private static Soundex      soundex;
    private static long         DEF_SEARCH_LIFETIME = 5 * 60 * 1000; // 5min
    private static long         DEF_SEARCH_SCAN_INTERVAL = 2 * 60 * 1000; // 2min
    private static String       DEFAULT_QUERY_SPACE = "http://search.jxta.org/text";


    private URL                 selfURL;
    private URL                 routerURL;
    private HttpRouterClient    routerClient;
    private PreferencesService  preferencesService;
    private SchedulerService    schedulerService;
    private UuidService         uuidService;
    private DocumentBuilder     builder;
    private Hashtable           searchTable;

    private int debug = 0;

    long searchLife         = DEF_SEARCH_LIFETIME;
    long searchScanInterval = DEF_SEARCH_SCAN_INTERVAL;

    String redirectRoot;

    LogService websearchLogService;
    LogService accessLogService;
    LogService binkLogService;

    static {
        System.out.println ("Initializing OpensearchServlet class");
    }

    /**
     *  Requires routerURL parameter in config file.<br>
     */   
    public void init() throws ServletException {
        ServletContext context = getServletContext ();
        ServletConfig  config  = getServletConfig ();
        
        searchTable      = new Hashtable ();
        
        preferencesService = PreferencesService.getPreferencesService(context);
        schedulerService   = new SchedulerService ();
        uuidService        = new UuidService ();
        
        try {
            routerURL = new URL (config.getInitParameter ("routerURL"));
        } catch (MalformedURLException e) {
            throw new ServletException (e);
        }

        try {
          searchLife         = Long.parseLong (config.getInitParameter ("searchLife"));
          searchScanInterval = Long.parseLong (config.getInitParameter ("searchScan"));
          System.out.println ("defaultSearchLife =" + searchLife);
          System.out.println ("defaultSearchScanInterval=" + searchScanInterval);

        } catch (NumberFormatException nfe) {
          System.out.println ("Invalid number in parameters using default on one or more items");
        }

        try {
            InetAddress localHost   = InetAddress.getLocalHost ();
            byte[]      addressByte = localHost.getAddress ();

            this.selfURL = new URL ("http://" + 
                                    (addressByte[0] & 0xFF) + "." + 
                                    (addressByte[1] & 0xFF) + "." + 
                                    (addressByte[2] & 0xFF) + "." + 
                                    (addressByte[3] & 0xFF));
            System.out.println ("**WEBUI INIT: SelfURL=" + this.selfURL.toString ());

        } catch (UnknownHostException e) {
            throw new ServletException ("Unable to get ip of this machine");
            
        } catch (MalformedURLException me) {
            throw new ServletException ("Unable to get ip of this machine");
        }

        try {
            String forceURL = config.getInitParameter ("forceSelfURL");
            if (forceURL != null) {
                this.selfURL = new URL (forceURL);
                System.out.println ("**WEBUI INIT: forceSelfURL set to" + this.selfURL.toString ());
            }

        } catch (MalformedURLException me) {
            throw new ServletException ("forceSelfURL set in params but not a valid url");
        }

        websearchLogService = new LogService (config.getInitParameter
                                                  ("websearchLog"), 
                                              LogService.ALL);

        binkLogService      = new LogService (config.getInitParameter
                                                  ("binkLog"), 
                                              LogService.ALL);

        accessLogService    = new LogService (config.getInitParameter
                                                 ("accessLog"),
                                              LogService.ALL);

        redirectRoot = config.getInitParameter ("redirectRoot");

        try {
          debug = Integer.parseInt (config.getInitParameter ("debug"));
        } catch (NumberFormatException nfe) {
          System.out.println ("Unable to get debug level from confg");
        }

        String documentBuilderFactoryClass = getServletConfig ().
            getInitParameter ("DocumentBuilderFactory");
        String xmlReaderClass = getServletConfig ().
            getInitParameter ("XMLReader");
        
        System.setProperty ("javax.xml.parsers.DocumentBuilderFactory",
                            documentBuilderFactoryClass);
        System.setProperty ("org.xml.sax.driver",
                            xmlReaderClass);
        
        try {
            DocumentBuilderFactory factory = 
                DocumentBuilderFactory.newInstance ();

            factory.setValidating (false);

            builder = factory.newDocumentBuilder ();
	    
            routerClient = new HttpRouterClient (routerURL, debug);

        } catch (ParserConfigurationException e) {
            throw new ServletException ("Misconfigured xml parser.");
        }

        schedulerService.scheduleAction (new SchedulerService.Action () {
                
                public void perform (SchedulerService ss) {
                    
                    Enumeration e;
                    
                    e = searchTable.elements ();
                    
                    while (e.hasMoreElements ()) {
                        Search s = (Search) e.nextElement ();
                        if (debug > 1)
                            System.out.println ("Checking if search is too old - uuid: " + s.request.getQueryUuid ());
                        
                        if (s.lastAccess + searchLife < System.currentTimeMillis ()) {
                            if (debug > 1)
                                System.out.println ("Search too old deleting       - uuid: " + s.request.getQueryUuid ());
                            s.terminate ();
                            searchTable.remove (s.request.getQueryUuid ());
                        }
                    }
                    
                    ss.scheduleAction (this, searchScanInterval);
                }
            }, searchScanInterval);
        
        new Thread (schedulerService).start ();
    }

    /**
     * Retrieves user preferences from preferencesService
     *
     * @see net.jxta.search.websearch.PreferencesService#getPreferences
     */    
    Preferences getPreferences (HttpServletRequest req) {
        if (debug > 0) 
            System.out.println ("****OpensearchServlet - Begin getPreferences ()*********************************");
        HttpSession session = req.getSession (true);
        Preferences prefs   =
            (Preferences) session.getAttribute ("preferences");

        if (prefs == null) {
            
            Cookie[] cookies = req.getCookies ();
            for (int i = 0; cookies != null && i < cookies.length;  i++)
                if (cookies[i].getName ().equals ("PREFS")) {
                    if (debug > 0) 
                        System.out.println ("Using cookie preferences");
                    prefs = preferencesService.getPreferences (cookies[i].getValue ());
                }
            
            if (prefs == null) {
                if (debug > 0)
                    System.out.println ("Using default peferences");
                prefs = preferencesService.getDefault ();
            }
            session.setAttribute ("preferences", prefs);
        } else {
            if (debug > 0)
                System.out.println ("Using session preferences");
        }

        if (debug > 0) 
            System.out.println ("****OpensearchServlet - End   getPreferences ()*********************************");
        return prefs;
    }


    /**
     * Handles HTTP POST requests
     * @see net.jxta.search.websearch.OpensearchServlet#doGet
     */
    public void doPost (HttpServletRequest req, HttpServletResponse res)
        throws ServletException, IOException
    {
        doGet (req, res);
    }


    public void doBink (String uuid,
                        HttpServletRequest req, 
                        HttpServletResponse res) 
        throws ServletException, IOException
    {
        String   op  = req.getParameter ("op");
        int      hid = 0;

        Search   s      = null;
        Hit      hit    = null;

        // To make result pages smaller and to prevent users from submitting bogus
        // binks, we assign each hit a id that when combined with the UUID identifies
        // each unique hit.  When a user wants to complain about a particular hit
        // all they need to submit is their opinion, hit id and uuid.  

        if (debug > 0)
            System.out.println ("****OpensearchServlet - Begin doBink ()*****************************************");

        try {
            hid = Integer.parseInt (req.getParameter ("hit"));
        } catch (NumberFormatException ne) {
            // no code intended
        }

        accessLogService.log (accessLogService.INFO,
                              uuid,
                              "ACTION=bink,OPINION=" + op);

        if (uuid != null && hid != 0 && 
            (s = (Search) searchTable.get (uuid)) != null) {

            hit = s.callback.hitList.getHitByHid (hid);

            if (hit != null) {
                // we found the hit in memory
                binkLogService.log (binkLogService.INFO, uuid, hit.toBinkLog (op));
            } 
        }

        // We always output the gee thanks page rather than confronting a user
        // that thinks they are effectively spamming or has responded too late
        ServletContext    contxt = getServletContext ();
        RequestDispatcher header = contxt.getRequestDispatcher ("/rating.jsp");
        header.include (req, res);

        if (debug > 0)
            System.out.println ("****OpensearchServlet - End   doBink ()*****************************************");    
    } // End OpensearchServlet.doBink ()


    public void doQuery (String              uuid, 
                         HttpServletRequest  req, 
                         HttpServletResponse res) 
        throws ServletException, IOException
    {
        HttpConnection.Response routerRes    = null;
        int               start  = 1;
        String            error  = null;
        Search            s      = null;
        String            query  = req.getParameter ("query");
        Preferences       prefs  = getPreferences (req);
        ServletContext    contxt = getServletContext ();
        RequestDispatcher index  = contxt.getRequestDispatcher ("/index.jsp");
        RequestDispatcher header = contxt.getRequestDispatcher ("/header.jsp");
        RequestDispatcher footer = contxt.getRequestDispatcher ("/footer.jsp");

        // "start" is passed in the query string, it indicates the hit to start at
        if (req.getParameter ("start") != null)
            start = Integer.parseInt (req.getParameter ("start"));

        // If the uuid is not present in the query string, or a search object
        // for the uuid specified doesn't exist, we need to validate the query,
        // hit the router, create a search object and spin a thread to wait.
        
        if (uuid == null || (s = (Search) searchTable.get (uuid)) == null) {
            if (validQuery (query)) {
                // Log new query
                accessLogService.log (accessLogService.INFO, 
                                      uuid,
                                      "ACTION=query,QUERY=" + query);

                uuid = HexTools.hexify (uuidService.getUuid ());
                
                // Hit the router 
                try {
                    System.out.println ("making route call");
                    Request request = 
                        makeRequest (query, DEFAULT_QUERY_SPACE, 
                                     prefs, uuid, req.getCookies ());

                    s = new Search (routerClient, 
                                    request, 
                                    builder,
                                    websearchLogService);

                    Thread t = new Thread (s);
                    
                    // add the search object to the table
                    searchTable.put (uuid, s);
                    t.start ();
                    String tempBuff = this.selfURL.toString () +
                                      req.getRequestURI () +
                                      "?action=q&start=1&query=" + 
                                      query + "&uuid=" + uuid;
                    System.out.println ("@@@@@@ Redirecting user to " + tempBuff);
                    res.sendRedirect (tempBuff);
                    return;
                } catch (IOException e) {
                    error = "Router is probably currently down.";
                }
                
            } else {
                // Log invalid query
                accessLogService.log (accessLogService.INFO,
                                      uuid,
                                      "ACTION=bad-query,QUERY=" + query);
                error = "Sorry, your query is invalid.";
            }
        }
        
        req.setAttribute ("totalProviders",  
                          s == null ? new Integer (0) :
                          new Integer (s.totalProviders));
        req.setAttribute ("queriedProviders",  
                          s == null ? new Integer (0) : 
                          new Integer (s.queriedProviders));
        req.setAttribute ("totalTime",    
                          s == null ? new Integer (0) : 
                          new Integer (s.totalTime));
        req.setAttribute ("start", new Integer (start));    

        req.setAttribute ("uuid", uuid);
        req.setAttribute ("preferences", prefs);

        if (req.getParameter ("wap") == null)
            res.setContentType ("text/html");
        else
            res.setContentType ("text/vnd.wap.wml");

        if (error == null) {
            // Log page display
            accessLogService.log (accessLogService.INFO,
                                  uuid,
                                  "ACTION=display,START=" + start);

            s.waitFor ((start - 1), 
                       prefs.getFlushafterProviders (), 
                       prefs.getFlushTime ());
            /*
            setHeaders (res,
                        s.callback.providers,
                        start,
                        prefs.getNumProviders ());
            */
            PrintWriter out  = res.getWriter ();

            header.include (req, res);

            if (start == 0)
                spellCheck (query, out);

            // we either have enough results or the timer popped,
            // display what we have

            int sorting = prefs.getSorting ();  
            if (req.getParameter ("sort") != null)
                sorting = Integer.parseInt (req.getParameter ("sort"));

            int end = 0;
	    boolean done = s.isDone ();
	    
            end = buildHtmlBody (s.callback.hitList, 
                                 start, 
                                 prefs.getResultsPerPage (),
                                 sorting,
                                 prefs.getEnableImages (),
                                 out,
                                 s.request.getQueryUuid (),
                                 s.request.getQuery (), 
                                 s.request.getQuerySpace (),
				 done);

            req.setAttribute ("start", new Integer (end));

	    if (done && end >= s.callback.hitList.size ())
		req.setAttribute ("done", Boolean.TRUE);
	    
            footer.include (req, res);
            
        } else {
            PrintWriter out  = res.getWriter ();
        
            header.include (req, res);
            
            out.println ("Error: " + error);
	    
	    req.setAttribute ("done", Boolean.TRUE);
	    
            footer.include (req, res);
        }
    }

    /**
     * Handles HTTP GET requests
     */
    public void doGet (HttpServletRequest req, HttpServletResponse res)
        throws ServletException, IOException
    {
        String            uuid   = req.getParameter ("uuid");

        if (debug > 0)
            System.out.println ("****OpensearchServlet - Begin doGet ()******************************************");    
 
        if (req.getParameter ("action") != null) {

            if ("q".equals (req.getParameter ("action")))
                doQuery (uuid, req,res);

            else if ("b".equals (req.getParameter ("action")))
                doBink (uuid, req, res);

            return;
        }
        
        if (debug > 0)
            System.out.println ("****OpensearchServlet - End   doGet ()******************************************");

    }


    boolean validQuery (String query) {
        boolean valid = true;

        if (query == null)
            valid = false;
            
        if ((query.trim ().length () <= 0) || 
            (query.trim ().length () > 55))
            valid = false;

        return valid;
    }  // validQuery


    void spellCheck (String query, PrintWriter out) 
        throws IOException
    {
        StringTokenizer  st = new StringTokenizer (query.toLowerCase ());
        PriorityQueue    eq;
        Soundex.EditWord ew;
        String           w;
        long             start = System.currentTimeMillis ();
        
        
        while (st.hasMoreTokens ()) {
            w = (String) st.nextToken ();
            
            try {
                eq = soundex.getClosest (w, 3);
                if (eq == null)
                    ; // out.println ("Word " + w + " is in dictionary.");
                else {
                    out.print ("Did you mean ");
                    if ((ew = (Soundex.EditWord) eq.pop ()) != null)
                        out.print ("<a href=\"" + 
                                   selfURL (ew.getString ()) + 
                                   "\">" +
                                   ew.getString () + 
                                   "</a>");
                    if ((ew = (Soundex.EditWord) eq.pop ()) != null)
                        out.print (" or <a href=\"" + 
                                   selfURL (ew.getString ())
                                   +"\">" +
                                   ew.getString () + 
                                   "</a>");
                    out.print ("?<p>");
                }
            } catch (NoSuchElementException e) {
                ; // out.println ("There are no close matches for " + w);
            }
        }

        out.println ("<!-- Calculation and rendering of " +
                     "spelling suggestions took " + 
                     (System.currentTimeMillis () - start) + 
                     " ms. -->");
    }


    Request makeRequest (String      searchTerms,
             String      querySpace,
             Preferences prefs,
             String      uuid,
             Cookie[]    cookies)
    {
    String query = "  <query>\n<text>" + searchTerms  + "</text>\n";
    if ( ! prefs.getLanguages ().equals ("all"))
        query += "  <language>" + prefs.getLanguages () + "</language>\n";
        query += "  </query>\n";

    return new Request (querySpace, 
                uuid,
                (int) prefs.getQueryLifetime (),
                            Integer.MAX_VALUE,
                            prefs.getResultsPerProvider (),
                prefs.getFlushafterProviders (),
                (int) prefs.getFlushTime (),
                query.getBytes ());
    }
             




    void setHeaders (HttpServletResponse res,
                     ArrayList providers, 
                     int       start,
                     int       len)
    {
        // forget about cookies for now.  grrr.
        /*
        if ((providers != null) && (start < providers.size ())) {
            int end = Math.min (start + len, providers.size ());
              
            for (int i = start;  i < end;  i++) {
                Provider     provider = (Provider) providers.get (i);
                if (provider.cookies != null) {
                    ListIterator cookies  = provider.cookies.listIterator ();       
                  
                    while (cookies.hasNext ()) {
                        String c = (String) cookies.next ();
                        
                        res.addCookie (new Cookie (provider.id, 
                                                   net.jxta.search.util.TextTools.base64encode (c.getBytes ())));
                    }
                }
            }

        }
        */
     }
   
    /** 
     * Builds a HTML body from the HitList supplied, writes it to an 
     * ServletOutputStream and returns the number of the next unused hit.
     *
     * @see net.jxta.search.websearch.SearchRelevance
     * @see net.jxta.search.util.Html#verify
     * @see net.jxta.search.util.Html#modifyLinks
     */
    int buildHtmlBody (HitList      hits,
                       int          start, 
                       int          hitsPerPage,
                       int          sorting,
                       int          enableImages,
                       PrintWriter  out,
                       String       uuid,
                       String       query,
                       String       querySpace,
		       boolean      done)
              throws IOException
    {
        if (debug > 0) {
            System.out.println ("****OpensearchServlet - Begin buildHtmlBody ()**********************************");
            System.out.println ("   Start         : " + start);
            System.out.println ("   HitsPerPage   : " + hitsPerPage);
            System.out.println ("   Sorting       : " + sorting);
            System.out.println ("   Query         : " + query);
	    System.out.println ("   Hits size     : " + hits.size ());
            System.out.println ("   Total hits    : " + hits.size ());
            System.out.println ("   New hits      : " + hits.getNumNew ());
        }

        int numAvailableHits = 0;

	
	if (start > hits.getNumUsed ()) 
            numAvailableHits = Math.min (hitsPerPage, hits.getNumNew () );
        else 
            numAvailableHits = Math.min (hitsPerPage, ((hits.getNumUsed () - start) + 1));


        if (! done && numAvailableHits <= 0) {
            out.println ("<p>Still waiting for more responses.</p>");
            return start;
        }

        hits.setSortType (sorting);
        hits.sort ();

        out.println ("<TABLE WIDTH=100%>"); // Start Hits Table

        for (int i = start;  i < (numAvailableHits + start);  i++) {
            // Get the hit, verify the html, add the onClick for ranking,
            // output the hit then build the rating links.
            Hit hit = hits.getHit (i);

            if (debug > 1)
                System.out.println ("****HIT= " + hit.toString () );

	    hit.toHtmlTableRow (out, enableImages, uuid, query, redirectRoot);

        }  
        out.println ("</TABLE>"); // End Hits Table

        if (debug > 0) 
            System.out.println ("****OpensearchServlet - End   buildHtmlBody ()**********************************");

        return numAvailableHits + start;
    }          

    void xmlToHtml (InputStream inXml, OutputStream outHtml) 
        throws SAXException, IOException
    {
        Document doc = builder.parse (inXml);
        Element  root  = doc.getDocumentElement();
        NodeList nodes;
        

        if ( ! root.getTagName ().equals ("result") )
            throw new SAXException ("Root tag is not 'result':");
        
        nodes = root.getElementsByTagName ("provider"); 
        
        for (int i = 0; i < nodes.getLength (); i++) {

            Element provider = (Element) nodes.item (i);
            Text    urlNode  = (Text) provider.getElementsByTagName ("url")
                .item (0).getFirstChild ();
            Text    timeNode = (Text) provider.getElementsByTagName ("time")
                .item (0).getFirstChild ();
            Text    iconUrlNode = (Text) provider.getElementsByTagName ("icon")
                .item (0).getFirstChild ();
            NodeList hits = provider.getElementsByTagName ("hit");
            
            outHtml.write (("<p>" + iconUrlNode.getData () + " <font color=gray size=-1><a href=more?providerUrl=" + URLEncoder.encode (urlNode.getData()) + ">More responses from this provider</a></font> - <font color=#006600 size=-1>Response Time: " +
                            timeNode.getData () + "</font></p>").getBytes ());
            
            for (int j = 0;  j < hits.getLength ();  j++) {
                Node     hitNode = hits.item (j);
                String   href   = null;
                String   anchor = null;
                String   price  = null;
                String   avail  = null;
                String   html   = null;
                NodeList children = hitNode.getChildNodes ();
                Node     child;
                
                for (int k = 0;  k < children.getLength ();  k++) {
                    child = children.item (k);
                    
                    if (child.getNodeName ().equals ("href"))
                        href = ((Text) child.getFirstChild ()).getData ();
                    else if (child.getNodeName ().equals ("anchor")) {
                        anchor = ((Text) child.getFirstChild ()).getData ();
                    } else if (child.getNodeName ().equals ("price")) {
                        price = ((Text) child.getFirstChild ()).getData ();
                    } else if (child.getNodeName ().equals ("avail")) {
                        avail = ((Text) child.getFirstChild ()).getData ();
                    } else if (child.getNodeName ().equals ("html")) {
                        html = ((Text) child.getFirstChild ()).getData ();
                    }
                }
                
                outHtml.write ( Html.modifyLinks (Html.verify (html), null)
                                .getBytes () );
            }
        }

        
    }


    void xmlToWap (HttpServletResponse res,
                   InputStream         inXml, 
                   OutputStream        outWap) 
        throws SAXException, IOException
    {

        Document doc   = builder.parse (inXml);
        Element  root  = doc.getDocumentElement();
        NodeList nodes = root.getChildNodes(); 
        Element  node;

        
        outWap.write ("<?xml version=\"1.0\"?>\n".getBytes ());
        outWap.write ("<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.1//EN\" \"http://www.wapforum.org/DTD/wml_1.1.xml\">".getBytes ());
        
        outWap.write ("<wml><head></head><card id=card1 title=\"OpenSearch Results\">".getBytes ());
        outWap.write ("<do type=\"prev\" label=\"Back\"<prev/></do>".getBytes ());

        
        for (int i = 0;  i < nodes.getLength ();  i++) {
            Element provider = (Element) nodes.item (i);
            Text    urlNode  = (Text) provider.getElementsByTagName ("url")
                .item (0).getFirstChild ();
            Text    timeNode = (Text) provider.getElementsByTagName ("time")
                .item (0).getFirstChild ();
            NodeList hits = provider.getElementsByTagName ("hit");
        
            for (int j = 0;  j < hits.getLength ();  j++) {
                Node     hitNode = hits.item (j);
                String   href   = null;
                String   anchor = null;
                String   price  = null;
                String   avail  = null;
                String   html   = null;
                NodeList children = hitNode.getChildNodes ();
                Node     child;
                
                for (int k = 0;  k < children.getLength ();  k++) {
                    child = children.item (k);
                    
                    if (child.getNodeName ().equals ("href"))
                        href = ((Text) child.getFirstChild ()).getData ();
                    else if (child.getNodeName ().equals ("anchor")) {
                        anchor = ((Text) child.getFirstChild ()).getData ();
                    } else if (child.getNodeName ().equals ("price")) {
                        price = ((Text) child.getFirstChild ()).getData ();
                    } else if (child.getNodeName ().equals ("avail")) {
                        avail = ((Text) child.getFirstChild ()).getData ();
                    } else if (child.getNodeName ().equals ("html")) {
                        html = ((Text) child.getFirstChild ()).getData ();
                    }
                    
                    outWap.write (("<p><a href=\"" + href + "\">" + 
                                   anchor + "</a></p>").getBytes ());
                }
            }
            
        }
    }

    String selfURL (String query) {
        return "search?query=" + query;
    }

}

