/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: Preferences.java,v 1.1.1.1 2001/06/01 16:51:00 jbeatty Exp $
 */

package net.jxta.search.websearch;

import goodsjpi.Persistent;

/**
 * Stores user preference information using GOODS
 *
 */
public class Preferences extends Persistent {
    static final int SORT_RELEVANCE = 1; // Sort by hit relevance
    static final int SORT_PROVIDER  = 2; // Sort by provider rank
    static final int SORT_RESP_TIME = 3; // Sort by provider response time

    static final int IMAGES_DISABLED = 0; // No images to be displayed in hits
    static final int IMAGES_ENABLED  = 1; // Provider Icons and images enabled

    String  id;
    String  languages;
    int     resultsPerPage;
    int     resultsPerProvider;
    int     sorting;
    int     enableImages;

    long    queryLifetime;
    long    flushTime;

    long    created;
    long    lastUsed;
    

    Preferences (String  id) {
        this.id       = id;
        this.created  = System.currentTimeMillis ();
        this.lastUsed = created;
    }

    void used () {
        lastUsed = System.currentTimeMillis ();
    }

    /**
     * Sets the language, number of providers, 
     * number of hits per provider, grouping and 
     * timout as the user requests.
     * 
     * FIXME: Add some javadoc here with range or value
     * size on these parameters.  For example, we don't
     * want someone sending a 350M String into this
     * constructor.
     */
    public void set (String  languages, 
                     int     resultsPerPage, 
                     int     resultsPerProvider, 
                     int     sorting,
                     int     enableImages,
                     long    queryLifetime,
                     long    flushTime)
    {
     
       /* FIXME: Get the rest of the ranges etc and move all 
        * of these tests into a validate method.
        */
  
        if (resultsPerPage <= 0) {
            throw new IllegalArgumentException("Number of hits "
                      + "per page must be greater than 0");
        }

        if (resultsPerProvider <= 0) {
            throw new IllegalArgumentException("Number of hits "
                      + "per provider must be greater than 0");
        }
  
        if (queryLifetime < 0) {
            throw new IllegalArgumentException("Time out value must be equal"
                          + " to or greater than 0");
        }

        this.id                 = id;
        this.languages          = languages;
        this.resultsPerPage     = resultsPerPage;
        this.resultsPerProvider = resultsPerProvider;
        this.sorting            = sorting;
        this.enableImages       = enableImages;
        this.queryLifetime      = queryLifetime;
        this.flushTime          = flushTime;
    }
    
    /**
     * Returns the unique user ID
     */
    public String getId () { return this.id; }

    /**
     * Returns the languages of responses the user wants considered
     */
    public String getLanguages () { return this.languages; }

    /**
     * Returns the number of providers to wait for before displaying hits
     */
    public int getFlushafterProviders () { 
        return ( (int) Math.ceil ((double) this.resultsPerPage / (double) this.resultsPerProvider) ); 
    }

    /**
     * Returns the number of hits to display per page
     */
     public int getResultsPerPage () { return this.resultsPerPage; }

    /**
     * Returns the number of hits to consider per provider
     */
    public int getResultsPerProvider () { return this.resultsPerProvider; }

    /**
     * Returns TRUE if group by provider else FALSE
     */
    public int getSorting () { return this.sorting; }

   /**
     * Returns TRUE if the user wants Images and icons enabled in results
     */

    public int getEnableImages () { return this.enableImages; }

    /**
     * Returns in milliseconds the user selected time to wait before displaying responses 
     */
    public long getQueryLifetime () { return this.queryLifetime; }

    /**
     * Returns in milliseconds the user selected flush interval time.
     */
     public long getFlushTime () { return this.flushTime; }





}
