/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ResponseCallbackImpl.java,v 1.1.1.1 2001/06/01 16:51:01 jbeatty Exp $
 */

package net.jxta.search.websearch;

import java.io.*;
import java.net.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import javax.xml.parsers.DocumentBuilder;

import net.jxta.search.peer.ProviderInfo;
import net.jxta.search.peer.RouterClient;
import net.jxta.search.util.LogService;
import net.jxta.search.util.Html;


class ResponseCallbackImpl implements RouterClient.ResponseCallback {
	
    DocumentBuilder builder;
    Search          search;
    HitList         hitList;
    LogService      logService;
    int             maxRank;
    int             maxProviderHits;
    
    ResponseCallbackImpl (String          query, 
                          int             maxProviderHits,
			  Search          search, 
			  DocumentBuilder builder,
			  LogService      logService)
    {
	this.search          = search;
        this.maxProviderHits = maxProviderHits;
	this.builder         = builder;
	this.logService      = logService;
	this.hitList         = new HitList (query);
    }
    
    public void response (String         uuid, 
			  ProviderInfo[] providerChain,
			  InputStream    responseStream)
    {
	
	try {
	    
	    Document  responseDoc = builder.parse (responseStream);
	    
	    response (uuid, providerChain, responseDoc);
	} catch (SAXException e) {
	    logService.log (uuid, 
			    "Cannot parse response from " + providerChain[0], 
			    e);
	} catch (IOException e) {
	    logService.log (uuid,
			    "Cannot read response stream from " + 
			    providerChain[0],
			    e);
	}
    }
    
    void response (String         uuid, 
		   ProviderInfo[] providerChain, 
		   Document       responseDoc)
    {
	Element   root        = responseDoc.getDocumentElement ();
	NodeList  hitNodes    = root.getChildNodes ();
	Response  response    = new Response (providerChain);
	int       parsedHits  = 0;

	for (int i = 0;  i < hitNodes.getLength ();  i++) {
	    Node     hitNode  = hitNodes.item (i);
	    
	    if (hitNode.getNodeName ().equals ("hit") &&
                parsedHits < this.maxProviderHits) {
		NodeList children = hitNode.getChildNodes ();
		Hit      hit      = new Hit ();
		    
		hit.response = response;
		
		for (int j = 0;  j < children.getLength ();  j++) {
		    if ( ! (children.item(j) instanceof Element))
			continue;
		    
		    Node node = (Element) children.item (j);
		    String  name = node.getNodeName ();
		    
		    
		    if (name.equals ("html")) {
			hit.html = getText (node);
		    } else if (name.equals ("href")) 
			hit.href = getText (node);
		    else if (name.equals ("title")) 
			hit.title = getText (node);
		    else if (name.equals ("image")) {
			Element el = (Element) node;
			hit.image       = el.getAttributeNode ("src").getValue();
			hit.imageHeight = getIntAttribute (el, "height", 0);
			hit.imageWidth  = getIntAttribute (el, "width", 0);
                        
		    } else if (name.equals ("link")) { 
			hit.link = getText (node);
		    } else if (name.equals ("detail")) { 
			String val = getText (node);
			if (val != null)
			    hit.detail = Html.stripHtmlTags (val);
		    } else if (name.equals ("rank")) {
			String val = getText (node);
			hit.rank = Integer.parseInt (val);
			maxRank  = Math.max (maxRank, hit.rank);
		    }
		}

		parsedHits++;
		hitList.add (hit);
		    
		search.newHit ();
	    } else if (hitNode.getNodeName ().equals ("provider")) {
                NodeList children = hitNode.getChildNodes ();
                ProviderInfo provider = providerChain[0];

                for (int j = 0; j < children.getLength (); j++) {
                    if ( ! (children.item (j) instanceof Element))
                        continue;
                    Node node = (Element) children.item (j);
                    String name = node.getNodeName ();

                    if (name.equals ("link")) {
                        try {
                            provider.setLink (getText (node));
                        } catch (MalformedURLException me) {
                            // invalid == no over ride, no code intended
                        }

                    } else if (name.equals ("description")) {
                        String val = getText (node);
                        if (val != null)
                            provider.setDescription (val);

                    } else if (name.equals ("name")) {
                        String val = getText (node);
                        if (val != null)
                            provider.setTitle (val);

                    } else if (name.equals ("image")) {
                        NodeList imageChildren = node.getChildNodes ();

                        for (int k = 0; k < imageChildren.getLength (); k++) {
                            if ( ! (imageChildren.item (k) instanceof Element))
                                continue;
                            Node imageNode = (Element) imageChildren.item (k);
                            String imageNodeName = imageNode.getNodeName ();

                            if (imageNodeName.equals ("url")) {
                                try {
                                    provider.setImageURL (getText (imageNode));
                                } catch (MalformedURLException me) {
                                    // no code intended
                                }

                            } else if (imageNodeName.equals ("width")) {
                                try {
                                    provider.setImageWidth (getText (imageNode));
                                } catch (NumberFormatException ne) {
                                    // No code intended
                                }

                            } else if (imageNodeName.equals ("height")) {
                                try {
                                    provider.setImageHeight (getText (imageNode));
                                } catch (NumberFormatException ne) {
                                    // No code intended
                                }
                            }
                        }
                        // TODO: make sure we don't try to adjust images with
                        //       a size of zero (divde by zero bad).
                        provider.adjustImageSize (110, 50);
                    }

                }
            }
	}
	    
	search.newProvider ();
    }

    String getText (Node node) {
	Text t = (Text) node.getFirstChild ();

	if (t == null)
	    return null;
	else
	    return t.getData ();
    }
    
    public void exception (String    uuid,
			   Exception exception)
    {
	logService.log (uuid, 
			"Exception while receiving or parsing response.", 
			exception);
    }

    HitList getHitList () {
	return hitList;
    }


    int getIntAttribute (Element el, String name, int deflt) {
	Attr attr = el.getAttributeNode (name);

	return (attr == null) ? deflt : Integer.parseInt (attr.getValue ());
    }

}
