/*
 * Copyright (c) 2001 Sun Microsystems, Inc. All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Sun Microsystems, Inc. for Project JXTA."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA"
 * must not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 * nor may "JXTA" appear in their name, without prior written
 * permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA. For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SearchRelevance.java,v 1.6 2001/06/11 22:03:06 ddoolin Exp $
 */

package net.jxta.search.websearch;

import java.io.*;
import java.net.*;
import java.util.*;
import net.jxta.search.util.BM;
import net.jxta.search.util.StringMatcher;
import java.lang.Math;


/** Used to compute relevance of search results for opensearch. 
 */
public class SearchRelevance {

    private static boolean debug = false;


    public SearchRelevance() {

    }


    public void setDebug() {
        debug = true;
    }

    public void clearDebug() {
        debug = false;
    }


    /** computes raw stats for relevance routine using 
     *  BM method from net.jxta.search.java.util.BM 
     */
    private static void compute_counts_bm(String[] queryArray, 
                                          String[] docArray, 
                                          int[] WordCount, 
                                          int[][] TermCount, 
                                          int[] DocCount) {
        StringMatcher matcher = new BM();

        StringMatcher.CompiledPattern [] compiledPatterns = 
                               new  StringMatcher.CompiledPattern[queryArray.length];

        char[][] docCharArrays = new char[docArray.length][];
        char[][] queryCharArrays = new char[queryArray.length][];
        for (int d = 0;  d < docArray.length;  d++)
            docCharArrays[d] = docArray[d].toCharArray ();

        for (int q = 0;  q < queryArray.length;  q++) {
            queryCharArrays[q] = queryArray[q].toCharArray ();
            //System.out.println ("query term = " + new String (queryCharArrays[i]) +
            //                  ", query term length = " + queryCharArrays[i].length);

            if (queryCharArrays[q].length == 0) continue;
            compiledPatterns[q] = matcher.compile (queryArray[q]);
        }
        for(int d = 0; d < docArray.length; d ++) {
            char[] doc = docCharArrays[d];
            for(int q = 0; q < queryArray.length; q ++) {
                int count = 0;
                char[] thisQuery = queryCharArrays[q];
                int index = -1;
                if (debug) {
                    System.out.println("Matching "
                                       + queryArray[q]
                                       + " and "
                                       + docArray[d]);
                }
                while ((index = matcher.match(doc,
                                              index+1,
                                              compiledPatterns[q])) != -1) {
                    count++;
		    if (index + 1 == doc.length)
			break;
                }
                TermCount[d][q] += count;
                if(count > 0)
                    DocCount[q] += 1;
            }
        }
        for(int d = 0; d < docArray.length; d++) {
            char[] doc = docCharArrays[d];
            int count = 0;
            for (int i = 0;  i < doc.length;  i++) {
                if (doc[i] == ' ')
                    count ++;
            }
            WordCount[d] = count;
            if (debug) {
                System.out.println("wc " + d + " = " + WordCount[d]);
            }
        }
            
    }


    /** Compute counts using hash function method.
     */
    private static void compute_counts_hash(String[] queryArray, 
                                            int query_len,
                                            String[] docArray,
                                            int num_docs,
                                            int[] WordCount,
                                            int[][] TermCount,
                                            int[] DocCount) {
        Map map = new HashMap();
        Map[] dmap = new Map[num_docs];
        Integer ONE = new Integer(1);

        for(int d = 0; d < num_docs; d ++) {

            dmap[d] = new HashMap();
            String doc = docArray[d];
            StringTokenizer words = new StringTokenizer(doc);

            while(words.hasMoreTokens()) {
                WordCount[d] ++;
                String orig_word = words.nextToken();
                String upper = orig_word.toUpperCase();
                String word = quick_stem(upper);
                                
                Integer dfreq = (Integer)dmap[d].get(word);
                if(dfreq == null) {
                    dfreq = ONE;
                } else {
                    int value = dfreq.intValue();
                    dfreq = new Integer(value + 1);
                }
                dmap[d].put(word, dfreq);
                                
                Integer frequency = (Integer)map.get(word);
                if(frequency == null) {
                    frequency = ONE;
                } else {
                    int value = frequency.intValue();
                    frequency = new Integer(value + 1);
                }
                map.put(word, frequency);
            }
        }

        Map sortedMap = new TreeMap(map);

        Map[] dsortmap = new Map[num_docs];
        for(int d = 0; d < num_docs; d ++) {
            dsortmap[d] = new TreeMap(dmap[d]);
        }
        
        for(int d= 0; d < num_docs; d ++) {
            for(int t = 0; t < query_len; t ++) {
                Integer freq = (Integer)dsortmap[d].get(queryArray[t]);
                if(freq != null)
                    TermCount[d][t] = freq.intValue();
                else
                    TermCount[d][t] = 0;
            }
        }
        for(int t = 0; t < query_len; t ++) {
            Integer freq = (Integer)sortedMap.get(queryArray[t]);
            if(freq != null)
                DocCount[t] = freq.intValue();
            else
                DocCount[t] = 0;
        }

    }  /* close compute_counts_hash() */




    /** Compute counts using naive method. 
     */
    private static void compute_counts(String[] queryArray, 
                        String[] docArray, 
                        int[] WordCount, 
                        int[][] TermCount, 
                        int[] DocCount) {

        int query_len = queryArray.length;
        int num_docs = docArray.length;

        for(int d = 0; d < num_docs; d ++) {

            String doc = docArray[d];
            StringTokenizer words = new StringTokenizer(doc);

            while(words.hasMoreTokens()) {

                WordCount[d] ++;
                String word = words.nextToken();

                for(int t = 0; t < query_len; t ++) {
                    if (debug) 
                        System.out.println("Matching "
                                           + queryArray[t]
                                           + " and "
                                           + docArray[d]);
                    String q = queryArray[t];
                    if(quick_match(word, q)) {
                        // if the word in doc d matches the query term t
                        TermCount[d][t] += 1;
                        DocCount[t] += 1;
                    }

                }
            }
        }

    }  /* close compute_counts() */



   /** Compute inverse document frequency. 
    */    
    private static float[] compute_idf(int num_docs,
                                       int[] DocCount,
                                       int query_len) {

        float[] IDF = new float[query_len];                     

        for(int t = 0; t < query_len; t ++) {
            if(DocCount[t] != 0) {
                IDF[t] = (float) java.lang.Math.sqrt((float) 
                                                     (((float) num_docs /
                                                       ((float) DocCount[t]))));

                if (debug) {
                    System.out.println("nhits = "+ num_docs 
                                       + " dc[" + t + "] = " + DocCount[t]);
                }
            }
            else IDF[t] = 0.0f;
        }

        return IDF;

    }  /* close compute_idf() */



    private static float[][] compute_tf(int num_docs,
                                        int[][] TermCount,
                                        int query_len) {

        float[][] TF = new float[num_docs][query_len];    
        for(int d = 0; d < num_docs; d ++) {
            for(int t = 0; t < query_len; t ++) {
                if(TermCount[d][t] != 0)
                    TF[d][t] = (1.0f + 
                                (float) java.lang.Math.log((float)TermCount[d][t]));
                //                  TF[d][t] = (float)TermCount[d][t];
                else
                    TF[d][t] = 0.0f;
            }
        }
        return TF;

    }  /* close compute_tf() */


    private static String quick_stem(String orig_string) {

        String[] stems = {"?", ":","&QUOT;", ",", "'S", "S", "ING", "ED", "ER", "ES"};
        int stem_length = 9;
        String stem_string = orig_string; // fall through value


        for(int s = 0; s < stem_length; s ++) {
            if(orig_string.endsWith(stems[s])) {
                stem_string = orig_string.substring(0, 
                                                    orig_string.length() - 1);
                return stem_string; // break if a match
            }
        }

        return stem_string;

    }



    /** quick_match performs a match between "word" and query "q"
     *  and returns true if there is a match 
     *  and false if there is no match.
     */
    private static boolean quick_match(String word, String q) {
	if(debug) {
	    System.out.println("quick_match....");
	    System.out.println("query " + q + "l= " + q.length());
	    System.out.println("word " + word + "l = " + word.length());
	    System.out.println(word.equals(q));
	}
        if(word.equals(q)) { // simple case
            return true; 
        }
	
        if((word.indexOf(q) != -1)) {
            // kind of fudgy - could match biggles from big etc.
            return true;
        }

        return false;

    }  /* close quick_match() */




    /** findRel takes a String "queryString" and an ArrayList "hits". 
     *  Modifies hits[i].relevance
     */

    public static void findRel(String queryString, ArrayList hits) {
        StringTokenizer query = new StringTokenizer(queryString," ,:");
        int num_docs = hits.size();
	if(num_docs == 0)
	    return;
        int query_len = query.countTokens();

        if (debug) {
            System.out.println("Opensearch: query length "
                               + query_len
                               + " num docs "
                               + num_docs);
        }

        /** WordCount is the count of number of
         *  words in document d
         */
        int[] WordCount = new int[num_docs];          


        /** TermCount is the 
         *  count of the number of hits of term 
         *  t in document d
         */
        int[][] TermCount = new int[num_docs][query_len]; 

        /** DocCount is the count of the number
         *   of hits of term t in all documents
         */
        int[] DocCount = new int[query_len];              


        setZero(WordCount);
        setZero(TermCount);
        setZero(DocCount);


        /** queryArray is an array of the 
         *  query words (or terms)
         */
        String[] queryArray = new String[query_len];     

        for(int t = 0; t < query_len; t ++) {
            queryArray[t] = query.nextToken();
            if (debug) {
               System.out.println("Unstemmed query term: " + queryArray[t]);
            }   
        }
        



        /* Do some stemming to help matching.
         */
        for(int t = 0; t < query_len; t ++) {
            String upper = queryArray[t].toUpperCase();
            String stemmed = quick_stem(upper);

            if (queryArray[t].length() == 1){
                System.out.println("Stemmed out of existence... length = "
                                   + queryArray[t].length());
                
            }

            queryArray[t] = stemmed;

            if (debug) {
               System.out.println("Stemmed query term: " + queryArray[t]);
            }

        }


        /* Loop over all documents, tokenize the documents and find matches 
         * against the query terms.
         */

        String[] docArray = new String[num_docs];
	
	int s_null = 0; // flag to indicate s being null (FIXME)
	
        for(int d = 0; d < num_docs; d ++) {
            Hit this_hit = (Hit) hits.get(d);
            String s = " ";

            // TODO: remove hit.html from here
            if ((this_hit.title != null) || (this_hit.detail != null)) {
              // Build data to do relevance on from title and detail
              if (this_hit.title != null)
                s = this_hit.title;
              if (this_hit.detail != null)
                s = s + this_hit.detail;

            } else {
              // for legacy reasons use the html
              s = this_hit.html;
            }
	    if(s == null) { // FIXME
		docArray[d] = "E8mpty"; // no data in hit so set to dummy string - yeah not that elegant I know...
	    }
	    else{
		docArray[d] = new String(" " + s.toUpperCase());
		if(debug) {
		    System.out.println("Document to be matched");
		    System.out.println(docArray[d]);
		}
	    }
        }

        //compute_counts_hash(queryArray, query_len, docArray, num_docs, WordCount, TermCount, DocCount);

	// System.out.println("I'm about to count... sshhhh");
	
	compute_counts(queryArray, docArray, WordCount, TermCount, DocCount); 
	//  compute_counts_bm(queryArray, docArray, WordCount, TermCount, DocCount); 
                            

        /*  TF is the "term frequency" of each 
         *  term t in document d
         */
        float[][] TF = compute_tf(num_docs, TermCount, query_len);

        /* IDF is the "inverse document frequency"
         * of term t over all documents
         */
        float[] IDF = compute_idf(num_docs, DocCount, query_len);


      /**********   Start compute similarities.  **************/

       /* Compute similarities between query and documents. */
        float[] Sim = new float[num_docs]; 

       /* FIXME: Sim[d] can be initted in the next loop. i
        * But don't do this until a until test is written 
        * to test the results of the change.   Even better,
        * use the setZero method instead.
        */
        // Sim[d] is the similiarity between the query and doc d
        for(int d = 0; d < num_docs; d ++) {
            Sim[d] = 0.0f;
        }

        for(int d = 0; d < num_docs; d ++) {
            for(int t = 0; t < query_len; t ++) {
                float k = IDF[t] * TF[d][t];
                Sim[d] += k;
            }
        }

        float Sum = 0.0f;

        for(int d = 0; d < num_docs; d ++) {
            Sim[d] /= java.lang.Math.sqrt(WordCount[d]); 
            // divide by the number of words in document d
            Sum += Sim[d];
        }

      /***********   End compute similarities.  *******/






        for(int d = 0; d < num_docs; d ++) {
            Hit this_hit = (Hit) hits.get(d);
            this_hit.relevance = Sim[d];
        }

        if (debug) {

            // print some stats
            for(int t = 0; t < query_len; t ++) {
               System.out.println("DocCount["+t+"] = "
                                   +DocCount[t]+" IDF = " + IDF[t]);
            }

            for(int d = 0; d < num_docs; d ++) {
            
                for(int t = 0; t < query_len; t ++) {
                    System.out.println("TermCount["+d+"]["+t+"] = "
                                       +TermCount[d][t]+" TF = " + TF[d][t]);
                }

                Hit this_hit = (Hit) hits.get(d);
                System.out.println("Sim["+d+"] = "+Sim[d] + " " 
                                   + this_hit.href);
            }

        }



    }  /* close findRel() */




    private static void setZero(int [][] array) {
  
        for (int i = 0; i<array.length; i++) {
            setZero(array[i]);
        }
    }  /* close setZero() */


    private static void setZero(int [] array) {
       
        for (int i = 0; i<array.length; i++)
            array[i] = 0;

    }  /* close setZero() */




    /** Test method  - takes strings as input - 
     *  try java SearchRelevance string0 string1 string2 
     *  to find relevance of string0 to string1 and string2
     */
    private static void test(String [] argv) {

        String pattern = argv[0];
        ArrayList hits = new ArrayList(argv.length-1);

        if (debug) {
            System.out.println("Terms to match: " + argv[0]);
        }

        for(int i = 1; i < argv.length; i ++) {
            Hit hit = new Hit();
            hit.html = argv[i];
            hits.add(hit);
            if (debug) {
                System.out.println("Test string (Hit): " + argv[i]);
            }
        }

        findRel(pattern, hits);

    }  /* close test() */



    public static void main (String[] argv) {

        SearchRelevance sr = new SearchRelevance();

        String [] strs = { "blah  :  bar",
                           "test1",
                           "blah bar foo baz",
                           "testem",
                           "test1",
                           "teste2" };

        sr.setDebug();
        sr.test(strs);
        
    }  /* close test() */



}  /* close class SearchRelevance */
