/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: RC4Cipher.java,v 1.5 2001/06/04 19:15:02 bondolo Exp $
 */


 /*
  *
  * Note: The RC4 algorithm is taken directly from "Applied Crytography: by
  *       Bruce Schneier.
  *
  */

package jxta.security.impl.cipher;

import jxta.security.util.Util;
import jxta.security.cipher.Cipher;
import jxta.security.cipher.Key;
import jxta.security.exceptions.CryptoException;

import jxta.security.impl.cipher.SecretKey;
/**
 * This class implements the RC4 stream cipher
 */
final public class RC4Cipher implements jxta.security.cipher.Cipher {
  private byte type;
  private byte mode;
  private byte[] S = null;	// S-boxes
  private int ii, jj;

  private static String myname = "RC4 Cipher";

  // Algorithm name
  public String getAlgorithmName() { return myname; }    

  // for debugging
  private boolean debug = false;
  /**
   * turn on debugging
   */
  public void setDebug() { debug = true; }
  /**
   * turn off debugging
   */
  public void clearDebug() { debug = false; }

  /**
   * Constructor
   */
  public RC4Cipher() {
    type = Cipher.ALG_RC4;
    mode = Cipher.MODE_UNKNOWN;
    S = null;
    ii = jj = 0;
  }
  /**
   * accessor for algorithm
   */
  public byte getAlgorithm() {
    return Cipher.ALG_RC4;
  }
  /** 
   *  Initialize key. RC4 does not require extra parameters.
   */
  public void init(Key theKey, byte theMode, byte[] b, int off, int len) 
    throws CryptoException 
  {
    // RC4 does not require additional parameters (IV, for example)
    throw new CryptoException(CryptoException.ILLEGAL_USE);
  }

  /*
   * Initializes the cipher S-boxes
   * Right out of Applied Cryptograpy!
   */
  static final int buflen = 256;
  public void init(Key theKey, byte theMode) throws CryptoException {
    if ((theKey.getType() != KeyBuilder.TYPE_RC4) ||
	((theMode != Cipher.MODE_ENCRYPT) && 
	 (theMode != Cipher.MODE_DECRYPT))) {
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    }
	
    mode = theMode;
    S = new byte[buflen];
    byte[] key = new byte[(theKey.getSize() + 7) >>> 3];
    SecretKey skey = (SecretKey)theKey;
    skey.getKey(key, (short) 0);
	
    /* Initialize S */
    for (int i=0; i < buflen; i++) S[i] = (byte) i;
	
    /* Initilaize K based on the key */
    byte[] K = new byte[buflen];
    int index = 0;
    while (index < buflen) {	// Fill K with duplicates of key
      for (int i = 0; (i < key.length) && (index < buflen); i++) {
	K[index++] = key[i];
      }
    }

    if (debug) {
      System.out.println("key : " + Util.hexEncode(key));
      System.out.println("K   : " + Util.hexEncode(K));
    }
    int j = 0;
    byte temp;
    for (int i = 0; i < buflen; i++) {
      j = (j + ((S[i] + K[i]) & 0xff)) & 0xff;
      temp = S[i];
      S[i] = S[j];
      S[j] = temp;
    }
    if (debug) {
      System.out.println("S   : " + Util.hexEncode(S));
    }
  }
  /**
   * Encrypt or decrypt the inBuf, storing results in outBuff
   * paying attention to offsets.
   */
  public int transform(byte[] inBuf, int inOff, int inLen,
		       byte[] outBuf, int outOff) {
    byte tmp;
    int t;
    byte kt;
	
    // We don't use loop unrolling since a typical processor for
    // the Kjava platform is unlikely to benefit from it. Conserving
    // memory is likely to be more important

    // System.out.println("\nTransforming " + inLen + " bytes" +
    //	       "\nS   : " + Util.hexEncode(S));

    for (int i = 0; i < inLen; i++) {
      ii = (ii + 1) & 0xff;
      tmp = S[ii];
      jj = (jj + (tmp & 0xff)) & 0xff;
      S[ii] = S[jj];
      S[jj] = tmp;
      t = (S[ii] + tmp) & 0xff;
      kt = S[t];
      outBuf[outOff + i] = (byte) (inBuf[inOff + i] ^ kt);
    }
    return inLen;
  }
  /**
   * Call update only if one has multiple buffers to encrypt/decrypt
   */
  public int update(byte[] inBuf, int inOff, int inLen, 
		      byte[] outBuf, int outOff) 
    throws CryptoException 
  {
    if ((inLen < 0) || 
	(inOff + inLen > inBuf.length) ||
	(outOff + inLen > outBuf.length))
      throw new CryptoException(CryptoException.ILLEGAL_USE);
	
    if (mode == Cipher.MODE_UNKNOWN)
      throw new CryptoException(CryptoException.INVALID_INIT);
	
    int val = transform(inBuf, inOff, inLen, outBuf, outOff);
    return val;
  }
  /**
   * Call here if you only have one buffer, or its the last buffer.
   */
  public int doFinal(byte[] inBuf, int inOff, int inLen, 
			 byte[] outBuf, int outOff) 
    throws CryptoException
  {
    int val = update(inBuf, inOff, inLen, outBuf, outOff);

    // reinit the algorithm
    mode = Cipher.MODE_UNKNOWN;
    S = null;
    ii = jj = 0;
    return (val);
  }
}
