/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SecretKey.java,v 1.3 2001/04/30 22:18:29 yeager Exp $
 */

package jxta.security.impl.cipher;

import jxta.security.cipher.Key;
import jxta.security.exceptions.CryptoException;

// Base interface for all symmetric algorithm keys

public final class SecretKey implements jxta.security.cipher.Key {

  private boolean initDone;	// true if initialized
  private byte keyType;		// Key type
  private short size;		// Key size in bits
  private byte[] secret;	// secret key

  /**
   * Constructor
   */
  public SecretKey(byte type, short len) {
    initDone = false;
    keyType = type;
    size = len;
    secret = null;
  }
  /**
   * Initialization state
   */
  public boolean isInitialized() {
    return initDone;
  }

  /**
   * Clear key values
   */
  public void clearKey() {
    initDone = false;
    size = 0; 
  }

  /**
   * Return Size in bits
   */
  public short getSize() {
    return size;
  }
  /**
   * Length in bytes - Keys SHOULD be multiples of 8
   */
  public int getLength() {
    return ((size + 7) >>> 3);
  }

  /**
   * Return key type code
   */
  public byte getType() {
    return keyType;
  }
 
  /**
   *  Here we copy the secret key into the buffer
   */
   public byte getKey(byte[] buf, int off) {
    if (keyType != KeyBuilder.TYPE_RC4) return 0;
    // Check buffer size
    if (buf.length < secret.length + off) return 0;
    // Copy the key from secret to buf.
    System.arraycopy(secret, 0, buf, off, secret.length);
    return (byte)secret.length;
  }

  /**
   * Set the key into secret
   */
  public void setKey(byte[] buf, int off) 
    throws CryptoException
  {
    // validate type 
    if (keyType != KeyBuilder.TYPE_RC4)
      throw new CryptoException(CryptoException.NO_SUCH_ALGORITHM);

    // Insure input buffer size is consistent with keysize in bytes
    int nbytes = (size + 7) >>> 3;
    if (nbytes + off > buf.length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    // move the data
    secret = new byte[nbytes];
    System.arraycopy(buf, off, secret, 0, nbytes);
  }
}

