/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: JxtaCryptoSuite.java,v 1.7 2001/04/27 21:58:34 yeager Exp $
 */

package jxta.security.impl.crypto;

import jxta.security.cipher.Cipher;
import jxta.security.cipher.Key;
import jxta.security.hash.Hash;
import jxta.security.keyexchange.KeyExchange;
import jxta.security.mac.MAC;
import jxta.security.publickey.PublicKeyAlgorithm;
import jxta.security.signature.Signature;
import jxta.security.exceptions.CryptoException;


import jxta.security.impl.cipher.NullCipher;
import jxta.security.impl.cipher.RC4Cipher;
import jxta.security.impl.hash.NullHash;
import jxta.security.impl.hash.SHA1Hash;
import jxta.security.impl.hash.MD5Hash;
import jxta.security.impl.keyexchange.NullKeyExchange;
import jxta.security.impl.mac.NullMAC;
import jxta.security.impl.mac.MACAlgorithm;
import jxta.security.impl.publickey.NullRSA;
import jxta.security.impl.publickey.RSA;
import jxta.security.impl.publickey.RSAKey;
import jxta.security.impl.signature.NullSignature;
import jxta.security.impl.signature.RSASignature;

// Builds a crypto suite
public class JxtaCryptoSuite implements jxta.security.crypto.JxtaCrypto {
  
  // Implementations
  /**
   * Accessors
   */
  public Cipher getJxtaCipher() { return cipher[0]; } // Cipher 
  /**
   * Only one cipher so far ..
   */
  public Cipher getJxtaCipher(byte type)
    throws CryptoException   
  { 
    return cipher[0]; 
  }
  /**
   * Use the following if just a single hash has been
   * requested.
   */
  public Hash getJxtaHash() { return hash[0]; }

  /**
   * Use the following if just mutiple hashes have been
   * requested.
   */
  public Hash getJxtaHash(byte type)
    throws CryptoException
  { 
    // find digest of type
    for (int i = 0; i < nDigests; i++) {
      if (hash[i].getAlgorithm() == type)
	return hash[i]; 
    }
    throw new CryptoException(CryptoException.NO_SUCH_ALGORITHM);
  }
  public MAC getJxtaMAC() { return mac; } // MAC

  public KeyExchange getJxtaKeyExchange() { 
    return keyExchange;
  } // Key exchange

  public PublicKeyAlgorithm getJxtaPublicKeyAlgorithm() 
  { 
    return pkAlg; 
  } 
  public Signature getJxtaSignature() { return signature; } // Signature

  // Profile used
  private static byte cryptoProfile = 0;
  private static byte crytoSignatureType = 0;
  private static byte cryptoMacType = 0;

  // Our class specific methods
  // Here we have our security objects
  private Cipher[] cipher = new Cipher[nCiphers];
  private Hash[] hash = new Hash[nDigests];
  private KeyExchange keyExchange = null;
  private MAC mac = null;
  private PublicKeyAlgorithm pkAlg = null;
  private Signature signature = null;

  /**
   * Our constructor
   * Supported profiles:
   *    RSA + RC4 + SHA1
   */
  public JxtaCryptoSuite(byte profileType, 
			 RSAKey rKey,
			 byte signatureType,
			 byte macType) 
    throws CryptoException
  {
    // Initialize to null methods
    pkAlg = new NullRSA();
    cipher[0] = new NullCipher();
    for (int i = 0; i < nDigests; ++i) hash[i] = new NullHash();
    keyExchange = new NullKeyExchange();
    mac = new NullMAC();
    signature = new NullSignature(); 

    cryptoProfile = profileType; // set profile for acces
    if (profileType != 0) {
      if ((profileType & MEMBER_RC4) != 0) cipher[0] = new RC4Cipher();
      if ((profileType & MEMBER_RSA) != 0) pkAlg = new RSA(rKey);
      if ((profileType & MEMBER_SHA1) != 0) hash[0] = new SHA1Hash();
      if ((profileType & MEMBER_MD5) != 0) {
	if ((profileType & MEMBER_SHA1) != 0) hash[1] = new MD5Hash();
	else hash[0] = new MD5Hash();
      }
    }
    // check signature
    crytoSignatureType = signatureType;
    if ((signatureType & Signature.ALG_RSA_SHA_PKCS1) != 0 ||
	(signatureType & Signature.ALG_RSA_MD5_PKCS1) != 0) {
      if ((profileType & MEMBER_RSA) == 0) {
	// RSA is required
	throw new CryptoException(CryptoException.ILLEGAL_USE);
      } else if ((profileType & MEMBER_SHA1) != 0) {
	signature = new RSASignature(pkAlg, rKey, signatureType,
				     getJxtaHash(Hash.ALG_SHA1));
      } else if ((profileType & MEMBER_MD5) != 0) {
	signature = new RSASignature(pkAlg, rKey, signatureType,
				     getJxtaHash(Hash.ALG_MD5));
      } else {
	// Valid digest required
	throw new CryptoException(CryptoException.ILLEGAL_USE);
      }
    }
    // Check MAC (none yet)
    cryptoMacType = macType;
    if ((macType & MAC.ALG_RC4_MD5) != 0 ||
	(macType & MAC.ALG_RC4_SHA1) != 0) {
      if ((profileType & MEMBER_RC4) == 0) {
	// RC4 is only cipher and required
	throw new CryptoException(CryptoException.ILLEGAL_USE);
      } else if ((profileType & MEMBER_MD5) != 0) {
	mac = new MACAlgorithm(getJxtaCipher(Cipher.ALG_RC4), MAC.ALG_RC4_MD5, 
			       getJxtaHash(Hash.ALG_MD5));
      } else if ((profileType & MEMBER_SHA1) != 0) {
	mac = new MACAlgorithm(getJxtaCipher(Cipher.ALG_RC4), MAC.ALG_RC4_SHA1, 
			       getJxtaHash(Hash.ALG_SHA1));
      }
    }
  }

  /**
   * access profile
   */
  public static byte getProfile() {
    return cryptoProfile;
  }
}
