/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 */

package jxta.security.impl.hash;

import jxta.security.hash.Hash;

public class MD5Hash implements jxta.security.hash.Hash {
  private static final int S11 = 7;
  private static final int S12 = 12;
  private static final int S13 = 17;
  private static final int S14 = 22;
  private static final int S21 = 5;
  private static final int S22 = 9;
  private static final int S23 = 14;
  private static final int S24 = 20;
  private static final int S31 = 4;
  private static final int S32 = 11;
  private static final int S33 = 16;
  private static final int S34 = 23;
  private static final int S41 = 6;
  private static final int S42 = 10;
  private static final int S43 = 15;
  private static final int S44 = 21;
  
  private static final byte[] padding = {
    (byte) 0x80, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
    (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
  };
  private final int MD5DigestLength = 16;
  /*
   * Constructor
   */
  public MD5Hash() {
    type = ALG_MD5;
    len = MD5DigestLength;
    reset();
  }

  private static String myname = "MD5 Hash";
  /**
   * Algorithm name
   */
  public String getAlgorithmName() { return myname; }

  /**
   * Debugging flag
   */
  private static boolean debug = false;
  /**
   * set debug
   */
  public void setDebug() { debug = true; }
  /**
   * clear debug
   */
  public void clearDebug() { debug = false; }

  /* MD5 parameters */
  private final int bufferLength = 64;

  /* Internal context */
  // state: A, B, C, D
  private int[] state = new int[4];

  // number of bits modulo 2^64 (LSB first)
  private int[] count = new int[2];

  // input buffer
  private byte[] buffer = new byte[bufferLength];
    
  private byte type;  // message digest algorithm
  private byte len;   // length of hash in bytes
    
  
  /*
   * FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
   * Rotation is separate from addition to prevent recomputation.
   */

  private static int FF(int a, int b, int c, int d, int x, int s, int ac) {
    int temp = 0;
	
    temp = ((b & c) | (~b & d));
    a += temp + x + ac;
    a = ((a << s) | (a >>> (32 - s)));
    a += b;
    return a;
  }
  
  private static int GG(int a, int b, int c, int d, int x, int s, int ac) {
    int temp = 0;
    
    temp = (b & d) | (c & ~d);
    a += temp + x + ac;
    a = ((a << s) | (a >>> (32 - s)));
    a += b;
    return a;
  }
  
  private static int HH(int a, int b, int c, int d, int x, int s, int ac) {
    int temp = 0;
    
    temp = (b ^ c ^ d);
    a += temp + x + ac;
    a = ((a << s) | (a >>> (32 - s)));
    a += b;
    return a;
  }
  
  private static int II(int a, int b, int c, int d, int x, int s, int ac) {
    int temp = 0;
    
    temp = (c ^ (b | ~d));
    a += temp + x + ac;
    a = ((a << s) | (a >>> (32 - s)));
    a += b;
    return a;
  }
  
  /* 
   * MD5 basic transformation. Transforms state based on a 64-byte block.
   */
  private void transform (byte[] block, int offset) {
    int a = state[0];
    int b = state[1];
    int c = state[2];
    int d = state[3];
    int[] x = new int [16];
    
    for (int i = 0, j = offset; j < offset + 64; i++, j += 4) {
      x[i] = (block[j ] & 0xff) | ((block[j + 1] & 0xff) << 8) |
	((block[j + 2] & 0xff) << 16) | ((block[j + 3] & 0xff) << 24);
    }
    
    /* Round 1 */
    a = FF(a, b, c, d, x[ 0], S11, 0xd76aa478); /* 1 */
    d = FF(d, a, b, c, x[ 1], S12, 0xe8c7b756); /* 2 */
    c = FF(c, d, a, b, x[ 2], S13, 0x242070db); /* 3 */
    b = FF(b, c, d, a, x[ 3], S14, 0xc1bdceee); /* 4 */
    a = FF(a, b, c, d, x[ 4], S11, 0xf57c0faf); /* 5 */
    d = FF(d, a, b, c, x[ 5], S12, 0x4787c62a); /* 6 */
    c = FF(c, d, a, b, x[ 6], S13, 0xa8304613); /* 7 */
    b = FF(b, c, d, a, x[ 7], S14, 0xfd469501); /* 8 */
    a = FF(a, b, c, d, x[ 8], S11, 0x698098d8); /* 9 */
    d = FF(d, a, b, c, x[ 9], S12, 0x8b44f7af); /* 10 */
    c = FF(c, d, a, b, x[10], S13, 0xffff5bb1); /* 11 */
    b = FF(b, c, d, a, x[11], S14, 0x895cd7be); /* 12 */
    a = FF(a, b, c, d, x[12], S11, 0x6b901122); /* 13 */
    d = FF(d, a, b, c, x[13], S12, 0xfd987193); /* 14 */
    c = FF(c, d, a, b, x[14], S13, 0xa679438e); /* 15 */
    b = FF(b, c, d, a, x[15], S14, 0x49b40821); /* 16 */
    
    /* Round 2 */
    a = GG(a, b, c, d, x[ 1], S21, 0xf61e2562); /* 17 */
    d = GG(d, a, b, c, x[ 6], S22, 0xc040b340); /* 18 */
    c = GG(c, d, a, b, x[11], S23, 0x265e5a51); /* 19 */
    b = GG(b, c, d, a, x[ 0], S24, 0xe9b6c7aa); /* 20 */
    a = GG(a, b, c, d, x[ 5], S21, 0xd62f105d); /* 21 */
    d = GG(d, a, b, c, x[10], S22,  0x2441453); /* 22 */
    c = GG(c, d, a, b, x[15], S23, 0xd8a1e681); /* 23 */
    b = GG(b, c, d, a, x[ 4], S24, 0xe7d3fbc8); /* 24 */
    a = GG(a, b, c, d, x[ 9], S21, 0x21e1cde6); /* 25 */
    d = GG(d, a, b, c, x[14], S22, 0xc33707d6); /* 26 */
    c = GG(c, d, a, b, x[ 3], S23, 0xf4d50d87); /* 27 */
    b = GG(b, c, d, a, x[ 8], S24, 0x455a14ed); /* 28 */
    a = GG(a, b, c, d, x[13], S21, 0xa9e3e905); /* 29 */
    d = GG(d, a, b, c, x[ 2], S22, 0xfcefa3f8); /* 30 */
    c = GG(c, d, a, b, x[ 7], S23, 0x676f02d9); /* 31 */
    b = GG(b, c, d, a, x[12], S24, 0x8d2a4c8a); /* 32 */
    
    /* Round 3 */
    a = HH(a, b, c, d, x[ 5], S31, 0xfffa3942); /* 33 */
    d = HH(d, a, b, c, x[ 8], S32, 0x8771f681); /* 34 */
    c = HH(c, d, a, b, x[11], S33, 0x6d9d6122); /* 35 */
    b = HH(b, c, d, a, x[14], S34, 0xfde5380c); /* 36 */
    a = HH(a, b, c, d, x[ 1], S31, 0xa4beea44); /* 37 */
    d = HH(d, a, b, c, x[ 4], S32, 0x4bdecfa9); /* 38 */
    c = HH(c, d, a, b, x[ 7], S33, 0xf6bb4b60); /* 39 */
    b = HH(b, c, d, a, x[10], S34, 0xbebfbc70); /* 40 */
    a = HH(a, b, c, d, x[13], S31, 0x289b7ec6); /* 41 */
    d = HH(d, a, b, c, x[ 0], S32, 0xeaa127fa); /* 42 */
    c = HH(c, d, a, b, x[ 3], S33, 0xd4ef3085); /* 43 */
    b = HH(b, c, d, a, x[ 6], S34,  0x4881d05); /* 44 */
    a = HH(a, b, c, d, x[ 9], S31, 0xd9d4d039); /* 45 */
    d = HH(d, a, b, c, x[12], S32, 0xe6db99e5); /* 46 */
    c = HH(c, d, a, b, x[15], S33, 0x1fa27cf8); /* 47 */
    b = HH(b, c, d, a, x[ 2], S34, 0xc4ac5665); /* 48 */
    
    /* Round 4 */
    a = II(a, b, c, d, x[ 0], S41, 0xf4292244); /* 49 */
    d = II(d, a, b, c, x[ 7], S42, 0x432aff97); /* 50 */
    c = II(c, d, a, b, x[14], S43, 0xab9423a7); /* 51 */
    b = II(b, c, d, a, x[ 5], S44, 0xfc93a039); /* 52 */
    a = II(a, b, c, d, x[12], S41, 0x655b59c3); /* 53 */
    d = II(d, a, b, c, x[ 3], S42, 0x8f0ccc92); /* 54 */
    c = II(c, d, a, b, x[10], S43, 0xffeff47d); /* 55 */
    b = II(b, c, d, a, x[ 1], S44, 0x85845dd1); /* 56 */
    a = II(a, b, c, d, x[ 8], S41, 0x6fa87e4f); /* 57 */
    d = II(d, a, b, c, x[15], S42, 0xfe2ce6e0); /* 58 */
    c = II(c, d, a, b, x[ 6], S43, 0xa3014314); /* 59 */
    b = II(b, c, d, a, x[13], S44, 0x4e0811a1); /* 60 */
    a = II(a, b, c, d, x[ 4], S41, 0xf7537e82); /* 61 */
    d = II(d, a, b, c, x[11], S42, 0xbd3af235); /* 62 */
    c = II(c, d, a, b, x[ 2], S43, 0x2ad7d2bb); /* 63 */
    b = II(b, c, d, a, x[ 9], S44, 0xeb86d391); /* 64 */
    
    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;
    
    // Zeroize sensitive information.
    for (int i = 0; i < x.length; i++) x[i] = 0;
  }
    
  /* 
   * MD5 finalization. Ends an MD5 message-digest operation, writing the
   * the message digest and zeroizing the context.
   */
  private byte[] digest() {
    byte [] digest = new byte[16];
    byte [] bits = new byte [8];
    int index, padLen;
    
    // Save number of bits 
    bits[0] = (byte) (count[0] & 0xff);
    bits[1] = (byte) ((count[0] >>> 8) & 0xff);
    bits[2] = (byte) ((count[0] >>> 16) & 0xff);
    bits[3] = (byte) ((count[0] >>> 24) & 0xff);
    
    bits[4] = (byte) (count[1] & 0xff);
    bits[5] = (byte) ((count[1] >>> 8) &  0xff);
    bits[6] = (byte) ((count[1] >>> 16) &  0xff);
    bits[7] = (byte) ((count[1] >>> 24) &  0xff);
    
    // Pad out to 56 mod 64. 
    index = (count[0] >>> 3) & 0x3f;
    padLen = (index < 56) ? (56 - index) : (120 - index);
    update(padding, 0, padLen);
    
    // Append length (before padding) 
    update(bits, 0, 8);
    
    // Store state in digest
    digest[0] = (byte) (state[0] & 0xff);
    digest[1] = (byte) ((state[0] >>> 8) & 0xff);
    digest[2] = (byte) ((state[0] >>> 16) & 0xff);
    digest[3] = (byte) ((state[0] >>> 24) & 0xff);
    
    digest[4] = (byte) (state[1] & 0xff);
    digest[5] = (byte) ((state[1] >>> 8) & 0xff);
    digest[6] = (byte) ((state[1] >>> 16) & 0xff);
    digest[7] = (byte) ((state[1] >>> 24) & 0xff);
    
    digest[8] = (byte) (state[2] & 0xff);
    digest[9] = (byte) ((state[2] >>> 8) & 0xff);
    digest[10] = (byte) ((state[2] >>> 16) & 0xff);
    digest[11] = (byte) ((state[2] >>> 24) & 0xff);
    
    digest[12] = (byte) (state[3] & 0xff);
    digest[13] = (byte) ((state[3] >>> 8) & 0xff);
    digest[14] = (byte) ((state[3] >>> 16) & 0xff);
    digest[15] = (byte) ((state[3] >>> 24) & 0xff);
    
    // Zeroize sensitive information
    count[0] = count[1] = 0;
    state[0] = state[1] = state[2] = state[3] = 0;
    
    return(digest);
  }
    
  /**
   * Return algorithm type
   */
  public byte getAlgorithm() { 
    return type;
  }

  /**
   * return digest length
   */
  public byte getDigestLength() {
    return len;
  }
    
  /* MD5 initialization. Begins an MD5 operation, writing a new context */
  public void reset() {
    count[0] = 0;
    count[1] = 0;

    // Load magic initialization constants.
    state[0] = 0x67452301;
    state[1] = 0xefcdab89;
    state[2] = 0x98badcfe;
    state[3] = 0x10325476;
	
    type = ALG_MD5;
    len = MD5DigestLength;
    for (int i = 0; i < bufferLength; i++) buffer[i] = (byte) 0;
  }

  /**
   * Return a MD5 clone of this object's state
   */
  public Object clone() {
    MD5Hash cpy = new MD5Hash();
    
    System.arraycopy(this.state, 0, cpy.state, 0, 4);
    System.arraycopy(this.buffer, 0, cpy.buffer, 0, bufferLength);
    System.arraycopy(this.count, 0, cpy.count, 0, 2);
    cpy.type = this.type;
    cpy.len = this.len;
    return cpy;
  }
  /*
   * MD5 block update operation. Continues an MD5 message-digest
   * operation, processing another message block, and updating internal
   * context.
   *
   * Call do final() on last buffer of data.
   */
  public void update(byte[] input, int offset, int len) {
    int i = 0;
    int index = 0;
    int partLen = 0;
    
    // Compute number of bytes mod 64 
    index = (count[0] >>> 3) & 0x3F;
    // Update number of bits 
    /* TODO: We have a problem here because int is signed */
    int nBits = (len << 3);	// high order 3 bits lost (see below)
    count[0] += nBits;
    if (count[0] < nBits) {	// sign < 0
      count[1]++;		// add carry
    }
    count[1] += (len >>> 29);	// add high order 3 bits to high order word
    
    partLen = bufferLength - index;
    if (len >= partLen) {
      System.arraycopy(input, offset, buffer, index, partLen);
      transform(buffer, 0);
      
      for (i = partLen; (i + 63) < len; i += bufferLength) {
	transform(input, i + offset);
      }
      
      index = 0;
    } else {
      i = 0;
    }
    System.arraycopy(input, i + offset, buffer, index, len - i);
  }
  /**
   * Digest the final buffer. Must be called for last buffer
   * of data to be digested.
   */
  public int doFinal(byte[] inBuf, int inOff,
		       int inLen, byte[] outBuf,
		       int outOff) {
    // check the parameters to prevent a VM crash
    int cnt;
    try {
      cnt = outBuf[outOff] + outBuf[outOff + this.len - 1];
    } catch (ArrayIndexOutOfBoundsException aioob) {
      if (debug) {
	int hlen = this.len - 1;
	System.out.println("MD5 Hash - doFinal: Index out of bounds"
			   + "\nmax = " + outOff + hlen +
			   " outOff = " + outOff + " hash len = " +
			   hlen);
      }
      return 0;
    }     

    // finish the digest
    this.update(inBuf, inOff, inLen);

    // copy to output buffer
    // Note; We verified the end point above.
    System.arraycopy(this.digest(), 0, outBuf, outOff, this.len);

    // clear state
    reset();

    // number of bytes copied (MD5DigestLength);
    return (this.len);
  }    
 }
