/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: SHA1Hash.java,v 1.5 2001/04/23 23:18:36 yeager Exp $
 *
 * Note: 
 * --------------------------------------------------------------
 * This is based on SHA1.java by Chuck McManis:
 * 
 * SHA1.java - An implementation of the SHA-1 Algorithm
 *
 * This version by Chuck McManis (cmcmanis@netcom.com) and
 * still public domain. 
 *
 * Based on the C code that Steve Reid wrote. His header
 * was :
 *      SHA-1 in C
 *      By Steve Reid <steve@edmweb.com>
 *      100% Public Domain
 *
 *      Test Vectors (from FIPS PUB 180-1)
 *      "abc"
 *      A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D
 *      "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
 *      84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1
 *      A million repetitions of "a"
 *      34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F
 *
 */

package jxta.security.impl.hash;

import jxta.security.hash.Hash;

// A test hash algorithm
public class SHA1Hash implements jxta.security.hash.Hash {

  public final byte SHA1DigestLength = 20; // 160 bits

  /**
   * Constructor
   */
  public SHA1Hash() {
    type = ALG_SHA1;
    len = SHA1DigestLength;
    reset();
  }
  
  private static String myname = "SHA-1 Hash";
  /**
   * Algorithm name
   */
  public String getAlgorithmName() { return myname; }

  // for debugging
  private boolean debug = false;
  /**
   * turn on debugging
   */
  public void setDebug() { debug = true; }
  /**
   * turn off debugging
   */
  public void clearDebug() { debug = false; }

  // SHA-1 code parameters
  private final int blockLength = 16;
  private final int nStates = 5;
  private int[] block = new int[blockLength];
  private int blockIndex;
  private int[] dd = new int[nStates];
  private int[] state = new int[nStates];
  private long count = (long) 0;
  private byte type;
  private byte len;
  
  // Temporary variables used in R0-R4
  private int tmp;
  /**
   * Algorithm type
   */
  public byte getAlgorithm() {    
    return type;
  }
  /**
   * Digest length
   */
  public byte getDigestLength() {
    return len;
  }

  /**
   * Reset the calculation
   */
  public void reset() {
    /* SHA1 initialization constants */
    state[0] = 0x67452301;
    state[1] = 0xEFCDAB89;
    state[2] = 0x98BADCFE;
    state[3] = 0x10325476;
    state[4] = 0xC3D2E1F0;
    count = 0;
    blockIndex = 0;
    for (int i = 0; i < blockLength; i++) block[i] = 0;
    for (int i = 0; i < nStates; i++) dd[i] = 0;
    type = ALG_SHA1;
    len = SHA1DigestLength;
  }

  /**
   * Four SHA rounds, R0-R1, R2, R3, R4
   *
   */
  private final void R0(int data[], 
			int v, int w, int x , int y, int z, int i) {
    block[i] = (((block[i] << 24) | (block[i] >>> 8)) & 0xFF00FF00) |
      (((block[i] << 8) | (block[i] >>> 24)) & 0x00FF00FF);
    data[z] += ((data[w] & (data[x] ^ data[y] )) ^ data[y]) +
      block[i] + 0x5A827999 + ((data[v] << 5) | (data[v] >>> 27));
    data[w] =(data[w] << 30) | (data[w] >>> 2);
  }
  
  private final void R1(int data[],
			int v, int w, int x, int y, int z, int i) {
    tmp = block[(i + 13) & 15] ^ block[(i + 8) & 15] ^
      block[(i + 2) & 15] ^ block[i & 15];
    tmp = (tmp << 1) | (tmp >>> 31);
    block[i & 15] = tmp;

    data[z] += ((data[w] & (data[x] ^ data[y])) ^ data[y]) +
      tmp + 0x5A827999 + ((data[v] << 5) | (data[v] >>> 27));
    data[w] =(data[w] << 30) | (data[w] >>> 2);
  }

  private final void R2(int data[],
			int v, int w, int x, int y, int z, int i) {
    tmp = block[(i + 13) & 15] ^ block[(i + 8) & 15] ^
      block[(i + 2) & 15] ^ block[i & 15];
    tmp = (tmp << 1) | (tmp >>> 31);
    block[i & 15] = tmp;

    data[z] += (data[w] ^ data[x] ^ data[y]) +
      tmp + 0x6ED9EBA1 + ((data[v] << 5) | (data[v] >>> 27));
    data[w] =(data[w] << 30) | (data[w] >>> 2);
  }

  private final void R3(int data[],
			int v, int w, int x, int y, int z, int i) {
    tmp = block[(i + 13) & 15] ^ block[(i + 8) & 15] ^
      block[(i + 2) & 15] ^ block[i & 15];
    tmp = (tmp << 1) | (tmp >>> 31);
    block[i & 15] = tmp;

    data[z] += (((data[w] | data[x]) & data[y]) | (data[w] & data[x])) +
      tmp + 0x8F1BBCDC + ((data[v] << 5) | (data[v] >>> 27));
    data[w] =(data[w] << 30) | (data[w] >>> 2);
  }

  private final void R4(int data[],
			int v, int w, int x, int y, int z, int i) {
    tmp = block[(i + 13) & 15] ^ block[(i + 8) & 15] ^
      block[(i + 2) & 15] ^ block[i & 15];
    tmp = (tmp << 1) | (tmp >>> 31);
    block[i & 15] = tmp;

    data[z] += (data[w] ^ data[x] ^ data[y]) +
      tmp + 0xCA62C1D6 + ((data[v] << 5) | (data[v] >>> 27));
    data[w] =(data[w] << 30) | (data[w] >>> 2);
  }
  
  /**
   * Hash a single 512-bit block. This is the core of the algorithm.
   *
   * Note that working with arrays is very inefficent in Java as it
   * does a class cast check each time you store into the array.
   *
   */
  private void transform() {
    /* Copy context->state[] to working vars */
    dd[0] = state[0];
    dd[1] = state[1];
    dd[2] = state[2];
    dd[3] = state[3];
    dd[4] = state[4];

    /* 4 rounds of 20 operations each. Loop unrolled. */
    R0(dd,0,1,2,3,4, 0); R0(dd,4,0,1,2,3, 1); 
    R0(dd,3,4,0,1,2, 2); R0(dd,2,3,4,0,1, 3);
    R0(dd,1,2,3,4,0, 4); R0(dd,0,1,2,3,4, 5); 
    R0(dd,4,0,1,2,3, 6); R0(dd,3,4,0,1,2, 7);
    R0(dd,2,3,4,0,1, 8); R0(dd,1,2,3,4,0, 9);
    R0(dd,0,1,2,3,4,10); R0(dd,4,0,1,2,3,11);
    R0(dd,3,4,0,1,2,12); R0(dd,2,3,4,0,1,13);
    R0(dd,1,2,3,4,0,14); R0(dd,0,1,2,3,4,15);
    R1(dd,4,0,1,2,3,16); R1(dd,3,4,0,1,2,17);
    R1(dd,2,3,4,0,1,18); R1(dd,1,2,3,4,0,19);

    R2(dd,0,1,2,3,4,20); R2(dd,4,0,1,2,3,21);
    R2(dd,3,4,0,1,2,22); R2(dd,2,3,4,0,1,23);
    R2(dd,1,2,3,4,0,24); R2(dd,0,1,2,3,4,25);
    R2(dd,4,0,1,2,3,26); R2(dd,3,4,0,1,2,27);
    R2(dd,2,3,4,0,1,28); R2(dd,1,2,3,4,0,29);
    R2(dd,0,1,2,3,4,30); R2(dd,4,0,1,2,3,31);
    R2(dd,3,4,0,1,2,32); R2(dd,2,3,4,0,1,33);
    R2(dd,1,2,3,4,0,34); R2(dd,0,1,2,3,4,35);
    R2(dd,4,0,1,2,3,36); R2(dd,3,4,0,1,2,37);
    R2(dd,2,3,4,0,1,38); R2(dd,1,2,3,4,0,39);

    R3(dd,0,1,2,3,4,40); R3(dd,4,0,1,2,3,41);
    R3(dd,3,4,0,1,2,42); R3(dd,2,3,4,0,1,43);
    R3(dd,1,2,3,4,0,44); R3(dd,0,1,2,3,4,45);
    R3(dd,4,0,1,2,3,46); R3(dd,3,4,0,1,2,47);
    R3(dd,2,3,4,0,1,48); R3(dd,1,2,3,4,0,49);
    R3(dd,0,1,2,3,4,50); R3(dd,4,0,1,2,3,51);
    R3(dd,3,4,0,1,2,52); R3(dd,2,3,4,0,1,53);
    R3(dd,1,2,3,4,0,54); R3(dd,0,1,2,3,4,55);
    R3(dd,4,0,1,2,3,56); R3(dd,3,4,0,1,2,57);
    R3(dd,2,3,4,0,1,58); R3(dd,1,2,3,4,0,59);

    R4(dd,0,1,2,3,4,60); R4(dd,4,0,1,2,3,61);
    R4(dd,3,4,0,1,2,62); R4(dd,2,3,4,0,1,63);
    R4(dd,1,2,3,4,0,64); R4(dd,0,1,2,3,4,65);
    R4(dd,4,0,1,2,3,66); R4(dd,3,4,0,1,2,67);
    R4(dd,2,3,4,0,1,68); R4(dd,1,2,3,4,0,69);
    R4(dd,0,1,2,3,4,70); R4(dd,4,0,1,2,3,71);
    R4(dd,3,4,0,1,2,72); R4(dd,2,3,4,0,1,73);
    R4(dd,1,2,3,4,0,74); R4(dd,0,1,2,3,4,75);
    R4(dd,4,0,1,2,3,76); R4(dd,3,4,0,1,2,77);
    R4(dd,2,3,4,0,1,78); R4(dd,1,2,3,4,0,79);
    
    /* Add the working vars back into context.state[] */
    state[0] += dd[0];
    state[1] += dd[1];
    state[2] += dd[2];
    state[3] += dd[3];
    state[4] += dd[4];
  }
  
  
  
  /**
   * Complete processing on the message digest.
   */
  private byte[] digest() {
    byte[] result = new byte[SHA1DigestLength];
    byte bits[] = new byte[8];
    int i, j;
    
    for (i = 0; i < 8; i++) {
      bits[i] = (byte)((count >>> (((7 - i) * 8))) & 0xff);
    }
    
    int padLen = 64 - 9 - blockIndex;
    if (padLen < 0) padLen += 64;
    int padSize = padLen + 1;
    byte[] pad = new byte[padSize];
    pad[0] = (byte) 128;
    for (i = 1; i < padSize; i++) {
      pad[i] = (byte) 0;
    }
    update(pad, 0, padSize);
    update(bits, 0, 8);
    
    for (i = 0; i < SHA1DigestLength; i++) {
      result[i] = (byte)
	((state[i>>2] >> ((3-(i & 3)) * 8)) & 0xff);
    }
    return result;
  }
  
  // public methods
  /**
   * Clone "this"
   */
  public Object clone() {
    SHA1Hash cpy = new SHA1Hash();
    
    System.arraycopy(this.block, 0, cpy.block, 0, blockLength);
    System.arraycopy(this.dd, 0, cpy.dd, 0, 5);
    System.arraycopy(this.state, 0, cpy.state, 0, 5);
    cpy.blockIndex = this.blockIndex;
    cpy.count = this.count;
    cpy.type = this.type;
    cpy.len = this.len;
    return cpy;
  }

  /**
   * Call update for all but the last buffer of data
   * to be digested.
   *
   * Call doFinal() for the last buffer of data
   */
  public void update(byte[] buf, int off, int len) {
    int mask = 0;
    
    for (int i = off; i < (off + len); i++) {
      mask = (blockIndex & 3) << 3;
      
      block[blockIndex >> 2] &= ~(0xff << mask);
      block[blockIndex >> 2] |= (buf[i] & 0xff) << mask;
      blockIndex++;
      if (blockIndex == 64) {
	transform();
	blockIndex = 0;
      }
    }
    count += (len << 3);
  }
  /**
   *  If only one buffer is to be digested, then call doFinal
   *  only.
   *  
   *
   */
  public int doFinal(byte[] inBuf, int inOff,
		     int inLen, byte[] outBuf,
		     int outOff) {
    // check parameters to avoid a VM crash
    try {
      int test = outBuf[outOff] + outBuf[outOff + this.len - 1];
    } catch (ArrayIndexOutOfBoundsException aioob) {
      if (debug) {
	int hlen = this.len - 1;
	System.out.println("SHA1Hash - doFinal: Index out of bounds"
			   + "\nmax = " + outOff + hlen +
			   " outOff = " + outOff + " hash len = " +
			   hlen);
      }
      return 0;
    }
    
    this.update(inBuf, inOff, inLen);
    System.arraycopy(this.digest(), 0, outBuf, outOff, this.len);
    reset();
    return SHA1DigestLength;
  }
}
