/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * 
 */

/**
 * Here we have APIs for the following:
 *
 *   1) Initialize an RSA key pair for use in exchanging 
 *      key-material between two or more peers.
 *
 *   2) Provide access to the public key for public key exchange.
 *
 *   3) Generate a random session key for symmetric key encryption.
 *      Return the key's value encrypted with a provided RSA public
 *      key. 
 * 
 *   4) Encrypt data with the session key.
 * 
 *   5) Reset the session, clearing all keys.
 *
 *   Note: All data can be returned as either a binary or base64
 *   encoded byte array.
 *
 *   The idea is to support a public key exchange protocol to enable
 *   the secure exchange of key-material which can be used to secure
 *   a pipe.
 *
 */
package jxta.security.impl.jxtasec;

import jxta.security.exceptions.CryptoException;
import jxta.security.util.Util;
import jxta.security.util.URLBase64;
import jxta.security.cipher.Cipher;
import jxta.security.cipher.Key;
import jxta.security.crypto.JxtaCrypto;

import jxta.security.impl.publickey.RSAKey;
import jxta.security.impl.publickey.RSA;
import jxta.security.impl.publickey.RSAPublickeyData;
import jxta.security.impl.publickey.RSAPrivatekeyData;
import jxta.security.impl.cipher.KeyBuilder;
import jxta.security.impl.cipher.RC4Cipher;
import jxta.security.impl.cipher.SecretKey;
import jxta.security.impl.random.JRandom;
import jxta.security.impl.crypto.JxtaCryptoSuite;

public class JxtaSec1 {

  public JxtaSec1() {;}
  private boolean debug = false;
  public void setDebug() { debug = true; }
  public void clearDebug() { debug = false; }

  // suite of algorithms
  private static JxtaCrypto suite = null; 

  // RC4 variables
  private static SecretKey secret = null;
  private static byte[] sessionKey = null;
  private static Cipher rc4 = null;

  // RSA variables
  private static RSAKey rKey = null;
  private static RSA rsa = null;
  private static RSAPublickeyData ourPublicD = null;
  private static RSAPrivatekeyData ourPrivateD = null;
  private static RSAPublickeyData peerPublicD =  null;
  private static boolean initialized = false;
  private static boolean sessionInitialized = false;

  // Random number generator
  private static JRandom jran = null;

  // our crypto profile
  private static byte profile = (byte)(JxtaCrypto.MEMBER_RC4|JxtaCrypto.MEMBER_RSA);

  /**
   * Initialize security variables
   */
  public boolean jxtaSecInit()
  {
    if (!commonInit()) return false;

    // RSA algorithm: creates both public and private keys
    // and stashes them internally.
    if (!setRSAAlgorithm()) return false;

    // everything initialized successfully
    initialized = true;
    return true;
  }

  /** 
   * Initialize with previous RSA public/private key values
   */
  public boolean jxtaSecInit(Object publicKeyData, Object privateKeyData)
  {
    if (!commonInit()) return false;

    // RSA algorithm: creates both public and private keys
    // and stashes them internally.
    if (!setRSAAlgorithm(publicKeyData, privateKeyData)) return false;

    // everything initialized successfully
    initialized = true;
    return true;
  }

  /**
   * Common initialization calls.
   */
  private boolean commonInit()
  {
    // Random number generator
    try {
      jran = new JRandom();
    } catch (CryptoException cex) {
      System.out.println("Random init failure" + cex.toString());
      return false;
    }    

    // RSAkey and RSA algorithm
    try {
      rKey = (RSAKey)KeyBuilder.buildKey(KeyBuilder.TYPE_RSA,
					 KeyBuilder.LENGTH_RSA_512,
					 false);
    }  catch (CryptoException ce) {
      System.out.println("RSA init failure" + ce.toString());
      return false;
    }
    // Get our crypto suite (can be expanded for more members)
    try {
      suite = new JxtaCryptoSuite(profile, rKey,
				  (byte)0, (byte)0);      
    } catch (CryptoException cex) {
      System.out.println("Crypto Suite: " + cex.toString());
      return false;      
    }

    // RC4 SecretKey and Algorithm
    //    RC4 algorithm:
    rc4 = suite.getJxtaCipher();
    try {
      // build a secret key
      //   Note: We will later set the session key before
      //   initializing our rc4 object.
      secret = (SecretKey)KeyBuilder.buildKey(KeyBuilder.TYPE_RC4, 
					      KeyBuilder.LENGTH_RC4, 
					      false);
    } catch (CryptoException ce) {
      System.out.println("RC4 Key init failed: " + ce.toString());
      return false;
    }
    return true;
  }
  /**
   * Set RSA Algorithm parameters by computing the private, public
   * keys first.
   */
  private boolean setRSAAlgorithm()
  {
    rsa = (RSA)suite.getJxtaPublicKeyAlgorithm();
    
    // Set rsa public and private keys, and
    // save their values
    try {    
      rsa.setPublicKey();
      rsa.setPrivateKey();
      ourPublicD = (RSAPublickeyData)rsa.getPublickey();
      ourPrivateD = (RSAPrivatekeyData)rsa.getPrivatekey();      
    } catch (CryptoException ce) {
      System.out.println("RSA Key Error: " + ce.toString());
      return false;
    }

    // Reset RSA object's internal key state
    //  Later use will requiring this code setting the appropriate
    //  key, or the  public key of another peer.
    rsa.reset();
    return true;
  }

  /**
   * Set RSA Algorithm parameters from saved values
   */
  private boolean setRSAAlgorithm(Object publicKeyData, Object privateKeyData)
  {
    rsa = (RSA)suite.getJxtaPublicKeyAlgorithm();
    
    // Set rsa public and private keys, and
    // save their values
    try {    
      rsa.setPublicKey(publicKeyData);
      rsa.setPrivateKey(privateKeyData);
      ourPublicD = (RSAPublickeyData)publicKeyData;
      ourPrivateD = (RSAPrivatekeyData)privateKeyData;
    } catch (CryptoException ce) {
      System.out.println("RSA Key Error: " + ce.toString());
      return false;
    }

    // Reset RSA object's internal key state
    //  Later use will requiring this code setting the appropriate
    //  key, or the  public key of another peer.
    rsa.reset();
    return true;
  }

  /**
   *  Here we return the public key as byte array whose internal structure
   *  is (it may certainly be placed in an XML object):
   *
   *      m=base64 modulus:e=base64 exponent
   *      
   */
  private static final byte[] estart = {(byte)':', (byte)'e', (byte)'='};
  private static final byte[] mstart = {(byte)'m', (byte)'='};
  public byte[] getOurRSAPublicKey()
    throws CryptoException
  {
    // Make sure we have been initialized
    if (!initialized) {
      throw new CryptoException(CryptoException.INVALID_INIT);
    }
    byte[] eB64 = URLBase64.encode(ourPublicD.publicExponent);
    byte[] mB64 = URLBase64.encode(ourPublicD.modulus);
    byte[] pk = new byte[eB64.length+mB64.length+estart.length+mstart.length];

    // little preamble for modulus
    for (int i = 0; i < mstart.length; i++) pk[i] = mstart[i];
    // copy the modulus
    System.arraycopy(mB64, 0, pk, mstart.length, mB64.length);
    // little preamble for exponent
    int x = mstart.length + mB64.length;
    for (int i = x, j = 0; j < estart.length; j++) pk[i++] = estart[j];
    // copy the exponent
    x += estart.length;		// start of exponent data
    System.arraycopy(eB64, 0, pk, x, eB64.length);
    return pk;
  }
  /** 
   * Here we set the remote peer's RSA public key.
   * It must be in the above format. Otherwise a CryptoException is thrown.
   */
  private void setOtherPeerRSAPublicKey(byte[] otherRSAPublicKey)
    throws CryptoException
  {
    // Make sure we have been initialized
    if (!initialized) {
      throw new CryptoException(CryptoException.INVALID_INIT);
    }
    /*
     * Verify the format.
     */
    int i;
    for (i = 0; i < mstart.length; i++) {
      if (otherRSAPublicKey[i] != mstart[i]) {
	throw new CryptoException(CryptoException.ILLEGAL_VALUE);
      }
    }
    /**
     * Find the ":"
     */
    int modulusLength, modulusOffset = i;
    for (modulusLength = 0; i < otherRSAPublicKey.length; i++, modulusLength++) {
      if (otherRSAPublicKey[i] == estart[0]) break;
    }
    /**
     * validate the exponent preamble
     */
    if (i == otherRSAPublicKey.length || otherRSAPublicKey[i] != estart[0]) {
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    } 
    i += 1;			// skip the first delimeter
    for (int j = 1; j < estart.length && i < otherRSAPublicKey.length; j++, i++) {
      if (estart[j] != otherRSAPublicKey[i]) {
	throw new CryptoException(CryptoException.ILLEGAL_VALUE);
      }
    }
    if (i == otherRSAPublicKey.length) {
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    }

    /**
     * OK, that is valid. Let's b64 decode the key.
     */
    int exponentOffset = i;
    int exponentLength = otherRSAPublicKey.length - i;
    RSAPublickeyData rpub = new RSAPublickeyData();

    if (debug) {
      String modStr = new String(otherRSAPublicKey, modulusOffset, modulusLength);
      String expStr = new String(otherRSAPublicKey, exponentOffset,exponentLength);
      System.out.println("\n*****************************" +
			 "\n  modulus = " + modStr +
			 "\n exponent = " + expStr +
			 "\n*****************************");
    }

    rpub.modulus = URLBase64.decode(otherRSAPublicKey,  modulusOffset, modulusLength);
    rpub.publicExponent = URLBase64.decode(otherRSAPublicKey, exponentOffset, exponentLength);
    peerPublicD = rpub;
  }

  /**
   * We have received an encoded RSA Public Key from
   * the other peer:
   *
   * We return the session key as a base64 string
   * encrpted in the other peers public key.
   * The remote peer will use the session key we send
   * for encrypting data.
   */
  public synchronized byte[] getSessionKey(byte[] encodedRSApublicKey)
    throws CryptoException
  {
    /*
     * Set RSA object public key
     */
    setOtherPeerRSAPublicKey(encodedRSApublicKey);
    rsa.setPublicKey((Object)peerPublicD);

    // create a random session key, and set it for this session
    sessionKey = new byte[secret.getLength()];
    jran.nextBytes(sessionKey);
    secret.setKey(sessionKey, 0);
    sessionInitialized = true;
    
    // encrypt the session key with the RSA Public Key
    byte[] encryptedSkey = rsa.Algorithm(sessionKey, 
					 0,
					 sessionKey.length,
					 KeyBuilder.TYPE_RSA_PUBLIC,
					 true);	// encrypt
    rsa.reset();		// clear public key
    byte[] b64Eskey = URLBase64.encode(encryptedSkey);
    return b64Eskey;
  }
  /**
   * We have received a session key encrypted in our
   * RSA public key and base64 encoded. We decrypt it, 
   * and set it as our symmetric session key for data.
   */
  public synchronized void setSessionKey(byte[] encryptedSKey)
    throws CryptoException 
  {
     // Make sure we have been initialized
    if (!initialized) {
      throw new CryptoException(CryptoException.INVALID_INIT);
    }
    // OK, base64 decode
    byte[] pubEncKey = URLBase64.decode(encryptedSKey, 0, encryptedSKey.length);
    rsa.setPrivateKey(ourPrivateD); // set private key
    // decrypt the session key
    sessionKey = rsa.Algorithm(pubEncKey, 
			       0,
			       pubEncKey.length,
			       KeyBuilder.TYPE_RSA_PRIVATE,
			       false); // decrypt
    rsa.reset();		// clear private key
    // initialize the session
    secret.setKey(sessionKey, 0);
    sessionInitialized = true;
  }

  /**
   * Return data URLBase64 decoded, and then
   * decrypted with the session key.
   */
  public byte[] decryptData(byte[] data)
    throws CryptoException
  {
    if (!sessionInitialized) {
      throw new CryptoException(CryptoException.INVALID_INIT);
    }
    // N bytes in == N bytes out with RC4
    byte[] decoded = URLBase64.decode(data, 0, data.length);
    byte[] decrypted = new byte[decoded.length];

    // place in decrypt mode
    rc4.init(secret,Cipher.MODE_DECRYPT);

    // decrypt the data in the buffer
    rc4.doFinal(decoded, 0, decoded.length, decrypted, 0);
    return decrypted;
  }

  /**
   * Return data encrypted with the session key in URLBase64.
   */
  public byte[] encryptData(byte[] data)
    throws CryptoException
  {
    if (!sessionInitialized) {
      throw new CryptoException(CryptoException.INVALID_INIT);
    }
    // RC4 is byte stream cipher. N bytes in == N out
    byte[] encrypted = new byte[data.length];
    rc4.init(secret,Cipher.MODE_ENCRYPT);
    rc4.doFinal(data, 0, data.length, encrypted, 0);
    byte[] b64 = URLBase64.encode(encrypted);
    return b64;
  }

  /**
   * Get rsa public key data for local storage
   */
  public Object getRSAPublicKeyData()
    throws CryptoException
  {
    if (!initialized) {
      throw new CryptoException(CryptoException.INVALID_INIT);
    } else {
      return (Object)ourPublicD;
    }
  }

  /**
   * Get rsa private key data for local storage
   */
  public Object getRSAPrivateKeyData()
    throws CryptoException
  {
    if (!initialized) {
      throw new CryptoException(CryptoException.INVALID_INIT);
    } else {
      return (Object)ourPrivateD;
    }
  }

  /**
   * Reset the object
   */
  public void reset()
  {
    rKey = null;
    rsa = null;
    ourPublicD = null;
    ourPrivateD = null;
    peerPublicD =  null;
    initialized = false;
    sessionInitialized = false;
    jran = null;
  }
}

 
