/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 */

package jxta.security.impl.mac;

import jxta.security.mac.MAC;
import jxta.security.hash.Hash;
import jxta.security.cipher.Cipher;
import jxta.security.cipher.Key;
import jxta.security.exceptions.CryptoException;
import jxta.security.util.Util;

import jxta.security.impl.cipher.KeyBuilder;
import jxta.security.impl.cipher.SecretKey;

// For RSA digital signature code 
public class MACAlgorithm implements jxta.security.mac.MAC {

  private static String myname = "MAC";

  private static int macLength = 0;       // length of returned MAC
  private static byte type = 0;           // Algorithm type
  private static Hash digest = null;      // Hash object
  private static Cipher cipher = null;    // rsa object
  private static SecretKey rKey = null;   // rsa Key
  private static byte mode = 0;       	  // encrypt or verify
  /**
   *   // Algorithm name
   */
  public String getAlgorithmName() {
    return myname;
  }
  /**
   * constructor
   */
  public MACAlgorithm(Cipher cipherAlg,
		      byte theType,
		      Hash theDigest)
    throws CryptoException
  {
    cipher = cipherAlg;
    digest = theDigest;
    type = theType;
    macLength = calculateMacLength();
    myname += " " + cipher.getAlgorithmName() + " " + digest.getAlgorithmName();
  }
  private int calculateMacLength()
    throws CryptoException
  {
    // for byte stream ciphers the encrypted output
    // has the same length as the digest input
    // We only have RC4 right now
    if (cipher.getAlgorithm() ==  Cipher.ALG_RC4) {
      macLength = digest.getDigestLength();
    } else {
      // Otherwise calculate the cipher output length
      // as a function of the digest length, and the
      // cipher block sizes.
      macLength = 0;
      throw new CryptoException(CryptoException.ILLEGAL_USE);
    }
    return macLength;
  }
  /**
   *   // Algorithm type
   */
  public byte getAlgorithm() {
    return type;
  }
  // for debugging
  private boolean debug = false;
  /**
   * turn on debugging
   */
  public void setDebug() { debug = true; }
  /**
   * turn off debugging
   */
  public void clearDebug() { debug = false; }
  /**
   * get the length of the signature
   */
  public int getLength() {
    return macLength;
  }
  /**
   * initialize the Signature mode - sign or verify
   */
  public void init(byte theMode, Key theKey, byte[] privateKey)
    throws CryptoException
  {
    rKey = (SecretKey)theKey;
    if (theMode != MODE_ENCRYPT && theMode != MODE_VERIFY)
      throw new CryptoException(CryptoException.INVALID_INIT);
    else
      mode = theMode;
    // set the private encryption key
    rKey.setKey(privateKey, 0);
    // set the cipher mode
    byte cipherMode = (byte)(mode == MODE_ENCRYPT ? Cipher.MODE_ENCRYPT :
			     Cipher.MODE_DECRYPT);
    cipher.init(rKey, cipherMode);
  }
  /**
   * updates data to be encrypted. Here, the data in inbuf
   * is hashed with the digest.
   */
  public void update(byte[] inbuf, 
		     int offset, 
		     int length)
    throws CryptoException
  {
    if (inbuf.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    digest.update(inbuf, offset, length);
  }
  /**
   * Signs the hash generated from the last(only) buffer of data.
   */
  public int encrypt(byte[] inbuf, 
		     int offset, 
		     int inLength,
		     byte[] macBuf, 
		     int macOffset)
    throws CryptoException 
  {
    // Make sure we have room for the digest
    if (macBuf.length < macLength + macOffset)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    else if (mode != MODE_ENCRYPT)	// correct mode check
      throw new CryptoException(CryptoException.INVALID_INIT);
 
    // last buffer
    byte[] digestBuf = new byte[digest.getDigestLength()];
    digest.doFinal(inbuf, offset, inLength, digestBuf, 0);

    // The private key, and mode have been initialized in init(..)
    // Encrypted MAC returned in macBuff starting at macOffset
    cipher.doFinal(digestBuf, 0, digestBuf.length, macBuf, macOffset);
    return macLength;
  }

  /**
   * Verifies a MAC after updating last/only inBuf against
   * the MAC in macBuf. One can use update as with
   * encrypt().
   */
  public boolean verify(byte[] inBuf,
			int inOffset,
			int inLength,
			byte[] macBuf,
			int macOffset,
			int macLength)
      throws CryptoException
  {
    if (mode != MODE_VERIFY)	// correct mode check
      throw new CryptoException(CryptoException.INVALID_INIT);
    // update the last buffer of data
    byte[] digestBuf = new byte[digest.getDigestLength()];
    digest.doFinal(inBuf, inOffset, inLength, digestBuf, 0);

    // Decrypt the MAC in macBuf, yielding the original
    // digest in digestBuf1.
    byte[] digestBuf1 = new byte[macLength];
    cipher.doFinal(macBuf, macOffset, macBuf.length, digestBuf1, 0); 

    if (debug) {
      System.out.println("Digest [" + digestBuf.length + "] =\n" +
			 Util.hexEncode(digestBuf) +
			 "\nDecrypted MAC [" + digestBuf1.length + "] =\n" +
			 Util.hexEncode(digestBuf1));
    }
    // MUST have used the same digests (we hope)
    // Compare digest to decrypted digest ...
    if (digestBuf1.length == digestBuf.length) {
      for (int i = 0; i < digestBuf.length; i++)
	if (digestBuf1[i] != digestBuf[i]) return false;
    } else return false;
    return true;
  }  
}
