/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: RSA.java,v 1.8 2001/04/30 22:18:30 yeager Exp $
 */

/**
 * Note: This code uses excerpts of the RSA code from The Cryptix Foundation.
 * In particular, the Chinese Remainder Theorem Private Key, and initial
 * prime calculations.
 *
 **************************************************************************
 *
 * Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000 
 * The Cryptix Foundation Limited. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * 1. Redistributions of source code must retain the copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE CRYPTIX FOUNDATION LIMITED ``AS IS'' 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR 
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package jxta.security.impl.publickey;

import java.math.BigInteger;
import java.util.Random;
import jxta.security.exceptions.CryptoException;
import jxta.security.publickey.PublicKeyAlgorithm;
import jxta.security.util.Util;

import jxta.security.impl.random.JRandom;
import jxta.security.impl.cipher.KeyBuilder;
import jxta.security.impl.publickey.RSAPublickeyData;
import jxta.security.impl.publickey.RSAPrivatekeyData;

// For RSA Public Key code
public class RSA implements jxta.security.publickey.PublicKeyAlgorithm {

  static final int CERTAINTY = 80;    // Certainty of primality
  static final BigInteger ONE = BigInteger.valueOf((int)0x1),
    F4 = BigInteger.valueOf((int)0x10001L); // 65,537

  private static short keySize = KeyBuilder.LENGTH_RSA_512;
  private static short keyLength = 0;
  private static RSAKey rsaKey = null;
  private static JRandom random = null;
  private static boolean publicSet = false;
  private static BigInteger p, q, n, e;	// public key requirements
  private static BigInteger d;	// private key requirement
  private static BigInteger pExpP, pExpQ, crtCoef; // Chinese remainder theorem
  private static int maxDataLength; // <= (keySize/8) - 11 to allow for PKS#1 padding
  static private boolean canUsePublic = false,
    publicCalculated = false,
    canUsePrivate = false;

  /**
   *  Note: KeyBuilder must be called prior to instantiating this
   *        class to set the appropriate size.
   */
  public RSA(RSAKey rKey)
    throws CryptoException
  {
    keySize = rKey.getSize();	// key size in bits (multiple of 8)
    keyLength = (short)(keySize >>> 3);	// in bytes
    rsaKey = rKey;		// we set the keys
    random = new JRandom();	// random number generator
    reset();
    maxDataLength = keyLength - 11; // rfc2313
  }

  // for debugging
  private boolean debug = false;
  /**
   * turn on debugging
   */
  public void setDebug() { debug = true; }
  /**
   * turn off debugging
   */
  public void clearDebug() { debug = false; }

  public void reset()
  {
    p = BigInteger.valueOf(0L);
    q = BigInteger.valueOf(0L);
    n = BigInteger.valueOf(0L);
    e = BigInteger.valueOf(0L);
    d = BigInteger.valueOf(0L);
    pExpP = BigInteger.valueOf(0L);
    pExpQ = BigInteger.valueOf(0L);
    crtCoef = BigInteger.valueOf(0L);
    canUsePublic = false;
    publicCalculated = false;
    canUsePrivate = false;
  }

  private static String myname = "RSA Algorithm";

  /**
   *   // Algorithm name
   */
  public String getAlgorithmName() { return myname; }

  /**
   * RSA algorithm: BigInteger implementation
   *
   * Given m, an exponent, e, and the product of two primes, n,
   * compute ciper text, c = m**e mod(n)
   * Here, the pair {e, n} is the RSA public key. Note that where
   *    d = 1/e mod((p-1)(q-1)), d is the private key, and that encryption
   * and decryption are symmetric mod(n).
   */
  private BigInteger rsa(BigInteger m, 
			 BigInteger exp, 
			 BigInteger n)
  {
    BigInteger c = m.modPow(exp, n); // m**e mod(n)
    return c;
  }
  /**
   * CRT encrypt/decrypt X
   * 
   */
  private BigInteger crt(BigInteger X)
  {
    // Factors are known and usable by our CRT code.
    BigInteger p2 = X.mod(p).modPow(pExpP, p);
    BigInteger q2 = X.mod(q).modPow(pExpQ, q);

    if (debug) {
      System.out.println("crt(X):" +
			 "\np2.length = " + p2.bitLength() + "\np2 = " + p2 +
			 "\nq2.length = " + q2.bitLength() + "\nq2 = " + q2); 
    }
    if (p2.equals(q2))
      return q2;

    BigInteger k = (p2.subtract(q2).mod(p));
    BigInteger l = k.multiply(crtCoef).mod(p);

    if (debug) {
      System.out.println("crt(X):\n" +
			 "coef.length = " + crtCoef.bitLength() + "\ncoef = " + crtCoef +
			 "\nk.length = " + k.bitLength() + "\nk = " + k +
			 "\nl.length = " + l.bitLength() + "\nl = " + l +
			 "\nq.length = " + q.bitLength() + "\nq = " + q); 
    }

    return q.multiply(l).add(q2);    
  }

  /**
   * max input data block length
   */
  public int getMaxInputDataBlockLength() {
    return maxDataLength;
  }

  /**
   * length 0f Datablock + padding (Modulus in bytes)
   */
  public int getEncryptionBlockLength() {
    return keyLength;
  }

  /**
   * Encrypt/decrypt byte array. 
   * o Decrypt:
   *   1) Public key decryption of private key encryption.
   *      CRT private key encrpytion used: length <= modulus length
   *   2) Private key decryption of public key encryption
   *      (modulus length) <= length <= (modulus length + 1)
   * o Encrypt:
   *    If size <=  (modulus length - 11), then this is a legal encrypt, and we
   *    add the required padding before encryption as per PKCS #1 (rfc 2313)
   *    In this case the size MUST be <= than the modulus length
   *    in bytes - 11. If not, an illegal size error is thrown.
   * 
   */
  public byte[] Algorithm(byte[] data, 
			  int offset, 
			  int length,
			  byte type,
			  boolean encrypt)
    throws CryptoException
  {
    if (data.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    else if (!(canUsePublic && (type == KeyBuilder.TYPE_RSA_PUBLIC)) &&
	     !(canUsePrivate && (type == KeyBuilder.TYPE_RSA_PRIVATE)))
      throw new CryptoException(CryptoException.UNINITIALIZED_KEY);

    // See if we add padding 
    byte[] forBig = null;

    // See if the data has been encrypted, ie, we have an
    // encrypted DB.
    if (!encrypt) {
      // take the encrypted data as is
      forBig = new byte[length];
      for (int i = 0, k = offset; i < length; i++) forBig[i] = data[k++];
    } else if (length > maxDataLength) { // illegal data size
      throw new CryptoException(CryptoException.INVALID_RSA_DATALENGTH);
    } else {
      // create RSA Encrypted Block (byte[0] will be 0)
      forBig = addRSApadding(data, offset, length, type); 
    }

    BigInteger X = new BigInteger(forBig), Y;
    // Private key or Public key test
    String algType = null;
    if (canUsePrivate && 
	(type == KeyBuilder.TYPE_RSA_PRIVATE)) {
      algType = "Private";
      Y = crt(X);
   } else {
     algType = "Public";
      Y = rsa(X, e, n);
    }
    if (debug) {
      System.out.println("RSAAlgorithm, " + algType + ": encrypt = " + encrypt +
			 ", Y has " + Y.bitLength() + " bits");
    }
    byte[] rv = Y.toByteArray(); // the data
    if (debug) {
      System.out.println("Result: byte[] Y =\n" + Util.hexEncode(rv));
    }
    // If decrypting, then we need to remove the padding
    if (!encrypt) rv = removePadding(rv);
    return rv;
  }
  /**
   * Add PKCS #1 Encrypted Block (EB)
   * This padds the data to make its length
   * EXACTLY the modulus length BEFORE encrypting.
   *
   * Rules are simple:
   * 0x00, 0x01 or 0x02, followed by  N bytes of
   *   Private Key: 0xFF
   *    Public Key: random non-zero bytes
   * and terminated with 0x00
   *   N = K - 3 - data.length, K == bytes in modulus
   * See RFC 2313 as usual.
   */
  private static byte PRBT = (byte)0x01; // private key block tag
  private static byte PUBT = (byte)0x02; // public key block tag

  private byte[] addRSApadding(byte[] D, int offset, int len, byte type)
  {
    int i = 0;
    int k = offset + 2;		// start byte of padding 
    int N = keyLength - 3 - len; // bytes of padding
    byte[] out = new byte[keyLength];

    out[0] = (byte)0x00;	// start of EB
    if (type == KeyBuilder.TYPE_RSA_PRIVATE) {
      out[1] = PRBT;		// private key mark
      for (i = 0; i < N; ++i) out[k++] = (byte)0xFF;
    } else {
      out[1] = PUBT;		// public key mark
      // get N random bytes
      byte[] padding = new byte[N];
      random.nextBytes(padding);
      // make sure non-zero, and copy to buffer
      for (i = 0; i < N; i++) {
	if (padding[i] == (byte)0x0) padding[i] = (byte)0xFF;
	out[k++] = padding[i];
      }
    }
    // Terminate padding, and copy data
    out[k++] = (byte)0x00;	// termination
    // append data
    int j;
    for (i = k, j = 0; j < len; j++) out[i++] = D[j];

    // Fully compliant RSA Encrypted Block
    if (debug) {
      System.out.println("Padded: RSA EB(" + out.length + ":PS " + N + 
			 ":D " + len + ")=\n" + Util.hexEncode(out));
    }
    return out;
  }
  /**
   * Remove PKCS #1 padding (rfc2313)
   *     after decryption. 
   * Leading 0x00 byte will not be there because
   * the big int is always positive.
   */
  private byte[] removePadding(byte[] data)
    throws CryptoException
  {
    if (debug) {
      System.out.println("\nremovePadding, EB(" + data.length + ")=\n" +
			 Util.hexEncode(data));
    }

    // There are always at least 11 bytes of padding
    // We validate the format
    boolean badFormat = true;
    int length = data.length;
    int i = 0, j = 0, k = 0, nPadding = 0, 
      maxToScan = (int)(length - 2);
    if (data[0] == PRBT) {
      // A string off 0xFF for at most length - 4 bytes
      // You must encrypt at least one byte of data
      for (j = 1, k = 0; k < maxToScan; j++, k++) {
	if (data[j] != (byte)0xFF) break;
	nPadding += 1;	// count padding bytes
      }
      // check endpoint
      if (data[j] == (byte)0x00) {
	badFormat = false;	// padding is correct
      }
    } else if (data[0] == PUBT) {
      // A string of random NON-ZERO bytes for at most
      // length - 3 bytes as above.
      for (j = 1, k = 0; k < maxToScan; j++, k++) {
	if (data[j] == (byte)0x00) break;
	nPadding += 1;
      }
      if (nPadding >= 8)	// make sure we have at least 8 bytes
	badFormat = false;
      // Copy back into the byte array if ok
    }
    if (!badFormat) {
      // We only have the BT and 00 Marks, so
      int nData = length - nPadding - 2;
      byte[] rv = new byte[nData];
      for (i = 0, k = j + 1;  i < nData; i++)
	rv[i] = data[k++];

      if (debug) {
	System.out.println("Without padding =\n" + Util.hexEncode(rv));
      }

      return rv;
    }
    throw new CryptoException(CryptoException.INVALID_RSA_PADDING);
  }
  
  /**
   * Compute the private key from the public key
   * Here, we compute d = 1/e mod((p-1)(q-1)),
   * and store the remainder of the CRT parameters
   */
  public void setPrivateKey()
    throws CryptoException
  {
    if (!publicCalculated)	// must calculate public key first
      throw new CryptoException(CryptoException.INVALID_RSA_INIT);

    // p, q, and e have been calculated
    BigInteger pm1 = p.subtract(ONE);
    BigInteger qm1 = q.subtract(ONE);
    BigInteger phi = qm1.multiply(pm1);	
    d = e.modInverse(phi);	// calculate private key

    // CRT values
    pExpP = d.mod(pm1);
    pExpQ = d.mod(qm1);
    crtCoef = q.modInverse(p);	// CRT coefficient

    // copy to RSAkey
    byte[] bd = d.toByteArray();
    byte[] bpExpP = pExpP.toByteArray();
    byte[] bpExpQ = pExpQ.toByteArray();
    byte[] bcrt = crtCoef.toByteArray();
    byte[] primeP = p.toByteArray();
    byte[] primeQ = q.toByteArray();
    byte[] modulus = n.toByteArray();

    // save required private key values
    saveRSAPrivateKey(bd, bpExpP, bpExpQ, bcrt, primeP, primeQ, modulus);
    canUsePrivate = true;
  }
  /**
   * Set private key from saved values
   */
  public void setPrivateKey(Object privatekeyData)
    throws CryptoException
  {
    RSAPrivatekeyData pkeyData = null;
    try {
      pkeyData = (RSAPrivatekeyData)privatekeyData;
    } catch (java.lang.ClassCastException ex) {
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    }
    p = new BigInteger(pkeyData.primeP);
    q = new BigInteger(pkeyData.primeQ);
    n = new BigInteger(pkeyData.modulus);
    d = new BigInteger(pkeyData.privateExponentD);
    e = F4;			// a constant
    pExpP = new BigInteger(pkeyData.primeExponentP);
    pExpQ = new BigInteger(pkeyData.primeExponentQ);
    crtCoef = new BigInteger(pkeyData.crtCoefficient);
    /**
     * OK, save the private key
     */
    saveRSAPrivateKey(pkeyData.privateExponentD, 
		      pkeyData.primeExponentP,
		      pkeyData.primeExponentQ, 
		      pkeyData.crtCoefficient,
		      pkeyData.primeP,
		      pkeyData.primeQ, 
		      pkeyData.modulus);
    canUsePrivate = true;	// we can use private key
  }
  /**
   * Store all values associated with the private key encryption/decryption
   */
  private void saveRSAPrivateKey(byte[] pExponentD, 
				 byte[] primeExpP, 
				 byte[] primeExpQ, 
				 byte[] crtCoefficient,
				 byte[] primeP, 
				 byte[] primeQ,
				 byte[] modulus)
    throws CryptoException
  {
    rsaKey.setPrivateExponent(pExponentD, (short)0, (short)pExponentD.length);
    rsaKey.setPrimeExponentP(primeExpP, (short)0, (short)primeExpP.length);
    rsaKey.setPrimeExponentQ(primeExpQ, (short)0, (short)primeExpQ.length);
    rsaKey.setCrtCoefficient(crtCoefficient, (short)0, (short)crtCoefficient.length);
    rsaKey.setPrimeP(primeP, (short)0, (short)primeP.length);
    rsaKey.setPrimeQ(primeQ, (short)0, (short)primeQ.length);
    rsaKey.setModulus(modulus, (short)0, (short)modulus.length);
  }
  /**
   * Compute the public key given the key size in bits
   * Here, we compute p, q, and n = pq.
   *
   */
  public void setPublicKey() throws CryptoException {
    int pLen = keySize / 2;
    int qLen = keySize - pLen;	// lengths of p and q

    // Compute our big primes
    do {
      p = new BigInteger(pLen, CERTAINTY, (Random)random);
      q = new BigInteger(qLen, CERTAINTY, (Random)random); 
      n = p.multiply(q);
    }
    while((p.compareTo(q) == 0) || (n.bitLength() != keySize));

    if (debug) {
      System.out.println("plen = " + p.bitLength() + "\np = " + p +
			 "\nqlen = " + q.bitLength() + "\nq = " + q +
			 "\nnlen = " + n.bitLength() + "\nn = " + n);
    }

    // The public exponent is a constant
    e = F4;			// Fermat's 4th prime
    // OK, copy them to our RSAkey as byte arrays.
    // Note: BigInteger to byte array may add one extra byte
    // to keep the sign positive.
    byte[] bp = p.toByteArray();// all are positive
    byte[] bq = q.toByteArray();
    byte[] bn = n.toByteArray();
    byte[] be = e.toByteArray();
    rsaKey.setPublicExponent(be, (short)0, (short)be.length);
    rsaKey.setModulus(bn, (short)0, (short)bn.length);

    // Save these as part of the private key
    rsaKey.setPrimeP(bp, (short)0, (short)bp.length);
    rsaKey.setPrimeQ(bq, (short)0, (short)bq.length);
    publicCalculated = true;
    canUsePublic = true;
  }
  /**
   * Set the public key from a saved value for public key
   */
  public void setPublicKey(byte[] nModulus)
    throws CryptoException
  {
    n = new BigInteger(nModulus);
    e = F4;			// always constant

    byte[] be = e.toByteArray();  
    rsaKey.setPublicExponent(be, (short)0, (short)be.length);
    rsaKey.setModulus(nModulus, (short)0, (short)nModulus.length);  
    canUsePublic = true;
  }

  /**
   * Set the public key from a saved value for public key
   */
  public void setPublicKey(Object publickeyData)
    throws CryptoException
  {
    RSAPublickeyData pkeyData = null;
    try {
      pkeyData = (RSAPublickeyData)publickeyData;
    } catch (java.lang.ClassCastException ex) {
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    }
    n = new BigInteger(pkeyData.modulus);
    e = F4;			// always constant

    byte[] be = e.toByteArray();  
    rsaKey.setPublicExponent(be, (short)0, (short)be.length);
    rsaKey.setModulus(pkeyData.modulus, (short)0, (short)pkeyData.modulus.length);  
    canUsePublic = true;
  }

  /**
   * get the RSA public key data
   */
  public Object getPublickey()
    throws CryptoException
  {
    if (!canUsePublic)		// values not set
      throw new CryptoException(CryptoException.INVALID_RSA_INIT);

    // allocate the object + slots
    RSAPublickeyData rpkd = new RSAPublickeyData();
    rpkd.modulus = new byte[rsaKey.getModulusLength()];
    rpkd.publicExponent = new byte[rsaKey.getPublicExponentLength()];

    // get the values
    rsaKey.getModulus(rpkd.modulus, (short)0);
    rsaKey.getPublicExponent(rpkd.publicExponent, (short)0);

    return (Object)rpkd;
  }

  /**
   * get the RSA private key data
   */
  public Object getPrivatekey()
    throws CryptoException
  {
    if (!canUsePrivate)		// values not set
      throw new CryptoException(CryptoException.INVALID_RSA_INIT);

    // allocate the object and its slots 
    RSAPrivatekeyData rpkd = new RSAPrivatekeyData();

    rpkd.privateExponentD = new byte[rsaKey.getPrivateExponentLength()];
    rpkd.primeExponentP = new byte[rsaKey.getPrimeExponentPLength()];
    rpkd.primeExponentQ = new byte[rsaKey.getPrimeExponentQLength()];
    rpkd.crtCoefficient = new byte[rsaKey.getCrtCoefficientLength()];
    rpkd.primeP = new byte[rsaKey.getPrimePLength()];
    rpkd.primeQ = new byte[rsaKey.getPrimeQLength()];
    rpkd.modulus = new byte[rsaKey.getModulusLength()];

    // now the copies
    rsaKey.getPrivateExponent(rpkd.privateExponentD, (short)0);
    rsaKey.getPrimeExponentP(rpkd.primeExponentP, (short)0);
    rsaKey.getPrimeExponentQ(rpkd.primeExponentQ, (short)0);
    rsaKey.getCrtCoefficient(rpkd.crtCoefficient, (short)0);
    rsaKey.getPrimeP(rpkd.primeP, (short)0);
    rsaKey.getPrimeQ(rpkd.primeQ, (short)0);
    rsaKey.getModulus(rpkd.modulus, (short)0);

    return (Object)rpkd;
  }
}
