/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: RSAKey.java,v 1.6 2001/04/30 22:18:30 yeager Exp $
 */

package jxta.security.impl.publickey;

import jxta.security.cipher.Key;
import jxta.security.publickey.RSAPublicKey;
import jxta.security.publickey.RSAPrivateKey;
import jxta.security.util.Util;
import jxta.security.exceptions.CryptoException;
import jxta.security.impl.cipher.KeyBuilder;

// Base interface for all RSA public/private algorithms keys

public final class RSAKey implements 
jxta.security.publickey.RSAPublicKey,jxta.security.publickey.RSAPrivateKey {

  private boolean initDone;	// true if initialized
  private byte keyType;		// Key type
  private int size;		// Key size in bits (modulus limitation)
  private int modBytes;		// size / 8

  // for public or private key algorithms
  // All of these values are byte array representations of Big integers
  // All are positive.
  private byte[] publicExponent = null; // e
  private byte[] privateExponent = null; // d
  private byte[] modulus = null; // pq as discussed below

  // For private key Chinese remainder theorem algorithm
  private byte[] primeP = null;	// p in n = pq
  private byte[] primeQ = null;	// q in n = pq
  private byte[] primeExponentP = null; // d mod (p-1)
  private byte[] primeExponentQ = null; // d mod (q-1)
  private byte[] crtCoefficient = null;	// 1/q mod p
  public String getAlgorithmName() {
    return("RSA Key");
  }

  // for debugging
  private boolean debug = false;
  /**
   * turn on debugging
   */
  public void setDebug() { debug = true; }
  /**
   * turn off debugging
   */
  public void clearDebug() { debug = false; }

  /**
   *   Given two primes, p and q, and (e,(p-1)(q-1)) = 1, n = pq
   *       Public key: e, n
   *           c = m**e mod n, e is exponent, n modulus
   *       Private key: d = 1/e mod ((p-1)(q-1))
   *           b = c**d mod n, d is exponent, n modulus
   *   Note: By restricting the modulus, pq, we restrict d to be somewhat
   *         less. Modulus is always restricted to a power of 8. The
   *         default is 8**3 (512 bits) which is exportable.
   *  For the CRT private algorithm we additionally require:
   *       p, q, d mod(p - 1), d mod(q - 1), 1/q mod p
   *
   */

  /**
   * Constructor
   */
  public RSAKey (byte type, short len) {
    initDone = false;
    keyType = type;
    if (len != KeyBuilder.LENGTH_RSA_MIN ||
	len != KeyBuilder.LENGTH_RSA_512) 
      size = KeyBuilder.LENGTH_RSA_512;
    else
      size = (int)len;
    modBytes = (int)(size >>> 3);
  }
  /**
   * Check initialization state of key
   */
  public boolean isInitialized() {
    return initDone;
  }

  /**
   * Clear key values
   */
  public void clearKey() {
    initDone = false;
    size = 0; 
  }
  /**
   * get Key length in bits
   */
  public short getSize() {
    return (short)size;
  }
  /**
   * get Key length in bytes
   */
  public int getLength() {
    return (int)((size + 7) >>> 3);
  }

  /**
   * Get key type
   */
  public byte getType() {
    return keyType;
  }
  // interface methods
  public boolean equals(RSAPrivateKey key) {
    return RSAEquals((RSAKey)key, KeyBuilder.TYPE_RSA_PRIVATE);
  }
  
  public boolean equals(RSAPublicKey key) {
    return RSAEquals((RSAKey)key, KeyBuilder.TYPE_RSA_PUBLIC);
  }
  /**
   * Return true if key is equal to "this" key. Otherwise false.
   * Here we can compare key type, lengths of the exponents and moduli,
   * and finally the byte arrays themselves.
   */
  public boolean RSAEquals(RSAKey k, byte type) {
    byte[] exp = null, locExp = null;
    byte[] mod = new byte[modulus.length]; // use "this" limitations
    int length = 0, kLength = 0;
    if (type == KeyBuilder.TYPE_RSA_PRIVATE) {
      exp = new byte[privateExponent.length];
      length = privateExponent.length;
      locExp = privateExponent;
      kLength = k.getPrivateExponent(exp, 0);
    } else {
      exp = new byte[publicExponent.length];
      length = publicExponent.length;
      locExp = publicExponent;
      kLength = k.getPublicExponent(exp, 0);
    }
    // compare lengths of key parts
    if (k.getModulus(mod, 0) != modulus.length) return false;
    if (kLength != length) return false;
    // compare key parts themselves
    for (int i = 0; i < modulus.length; i++)
      if (mod[i] != modulus[i]) return false;
    for (int i = 0; i < length; i++)
      if (exp[i] != locExp[i]) return false;
    return true;
  }

  /**
   * Analgous routines to the secret key. We need to be able to set,
   * and get the RSA key stuff.
   */

  /**
   * get public exponent length
   *   e
   */
  public int getPublicExponentLength()
  {
    return publicExponent.length;
  }

  /**
   * get private exponent length
   *  d = 1/e mod (p-1)(q-1)
   */
  public int getPrivateExponentLength()
  {
    return privateExponent.length;
  }

  /**
   * get modulus length, n
   * Note: This may exceed the keysize in bytes by one
   *       byte. This is because big integer arithmetic
   *       will prepend a (byte)0x00 if the sign bit is
   *       set. Similarly for p, and q.
   */
  public int getModulusLength()
  {
    return modulus.length;
  }

  /**
   * Prime p
   */
  public int getPrimePLength()
  {
    return primeP.length;
  }

  /**
   * Prime q
   */
  public int getPrimeQLength()
  {
    return primeQ.length;
  }

  /**
   * prime exponent, d mod (p-1)
   */
  public int getPrimeExponentPLength() 
  {
    return primeExponentP.length;
  }

  /**
   * prime exponent, d mod (q-1)
   */
  public int getPrimeExponentQLength() 
  {
    return primeExponentQ.length;
  }

  /**
   * Chinese remainder theorem coefficient
   *     1/q mod p
   */
  public int getCrtCoefficientLength()
  {
    return crtCoefficient.length;
  }

  
  /**
   * Store the exponent as restricted by our size.
   */
  public void setPublicExponent(byte[] exp, int offset, int len) 
    throws CryptoException 
  {
    // Assure we have initialized the keyType first
    if ((keyType & KeyBuilder.TYPE_RSA) == 0)
      throw new CryptoException(CryptoException.NO_SUCH_ALGORITHM);

    // Make sure the length is correct
    if (len + offset > exp.length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    // Copy into a new array
    publicExponent = new byte[len];
    System.arraycopy(exp, offset, publicExponent, 0, len);
    if (modulus != null) initDone = true; // key is fully set
  }
  public void setPrivateExponent(byte[] exp, int offset, int len) 
    throws CryptoException 
  {
    // Assure we have initialized the keyType first
    if ((keyType & KeyBuilder.TYPE_RSA) == 0)
      throw new CryptoException(CryptoException.NO_SUCH_ALGORITHM);

    // Make sure the length is correct
    if (len + offset > exp.length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    // Copy into a new array
    privateExponent = new byte[len];
    System.arraycopy(exp, offset, privateExponent, 0, len);
    if (modulus != null) initDone = true; // key is fully set
  }

  /**
   * Store the modulus as restricted by our size.
   */
  public void setModulus(byte[] mod, int offset, int len) 
    throws CryptoException 
  {
    // Assure we have initialized the keyType first
    if ((keyType & KeyBuilder.TYPE_RSA) == 0)
      throw new CryptoException(CryptoException.NO_SUCH_ALGORITHM);

    // Make sure the length is correct
    if (debug) {
      System.out.println("SetModulus, modbytes = " + modBytes + " len = " + len);
    }
    if (len + offset > mod.length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);

    if (debug) {
      System.out.println("modulus =\n" + Util.hexEncode(mod));
    }

    // Copy into a new array
    if (modBytes < len) modBytes = len;
    modulus = new byte[modBytes];// limit this size
    // We copy at most modBytes of modulus, and prefill the modulus
    // with 0's by offsetting the difference.
    System.arraycopy(mod, offset, modulus, (modBytes - len), modBytes);
    if (publicExponent != null || privateExponent != null)
      initDone = true; // exponent and modulus are set
  }

  // Chinese remainder theorem private key algorithm parameters
  // We require modulus and exponent to be set FIRST.
  public void setPrimeP(byte[] prP, int offset, int length)
    throws CryptoException     
  {
    // Private key only
    if (!initDone || prP.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);

    // pq == modulus(we assume this)
    primeP = new byte[length];
    System.arraycopy(prP, offset, primeP, 0, length); 
  }

  public void setPrimeQ(byte[] prQ, int offset, int length)
    throws CryptoException     
  {
    if (!initDone || prQ.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);

    // pq == modulus(we assume this)
    primeQ = new byte[length];
    System.arraycopy(prQ, offset, primeQ, 0, length); 
  }
  /**
   * d mod(p - 1)
   */
  public void setPrimeExponentP(byte[] pExpP, int offset, int length)
    throws CryptoException     
  {
    // Private key only
    if (!initDone || pExpP.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);

    // pq == modulus(we assume this)
    primeExponentP = new byte[length];
    System.arraycopy(pExpP, offset, primeExponentP, 0, length); 
  }
  /**
   * d mod(q - 1)
   */
  public void setPrimeExponentQ(byte[] pExpQ, int offset, int length)
    throws CryptoException     
  {
    // Private key only
    if (!initDone || pExpQ.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);

    primeExponentQ = new byte[length];
    System.arraycopy(pExpQ, offset, primeExponentQ, 0, length); 
  }

  /**
   * 1/q mod p
   */ 
  public void setCrtCoefficient(byte[] crtCo, int offset, int length)
    throws CryptoException     
  {
    // Private key only
    if (!initDone || crtCo.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);

    crtCoefficient = new byte[length];
    System.arraycopy(crtCo, offset, crtCoefficient, 0, length); 
  }

  /**
   * copy the exponent into the byte array and return the exponent
   * length in bytes.
   */
  public int getPublicExponent(byte[] exp, int offset) {
    int ebytes = publicExponent.length;
    if (exp.length < offset + ebytes) 
      return 0;
    System.arraycopy(publicExponent, 0, exp, offset, ebytes);
    return ebytes;
  }
  public int getPrivateExponent(byte[] exp, int offset) {
    int ebytes = privateExponent.length;
    if (exp.length < offset + ebytes) 
      return 0;
    System.arraycopy(privateExponent, 0, exp, offset, ebytes);
    return ebytes;
  }

  public int getModulus(byte[] mod, int offset) {
    if (mod.length < offset + modBytes) 
      return 0;
    System.arraycopy(modulus, 0, mod, offset, modBytes);
    return modBytes;
  }

  // access methods for CRT private key algorithm parameters
  public int getPrimeP(byte[] prP, int offset) {
    // length consistent
    if (primeP == null || prP.length < offset + primeP.length) 
      return 0;
    System.arraycopy(primeP, 0, prP, offset, primeP.length);
    return primeP.length;
  }
						  
  public int getPrimeQ(byte[] prQ, int offset) {
    // length consistent
    if (primeQ == null || prQ.length < offset + primeQ.length) 
      return 0;
    System.arraycopy(primeQ, 0, prQ, offset, primeQ.length);
    return primeQ.length;
  }

  public int getPrimeExponentP(byte[] pExpP, int offset) {
    // length consistent
    if (primeExponentP == null || pExpP.length < offset + primeExponentP.length) 
      return 0;
    System.arraycopy(primeExponentP, 0, pExpP, offset, primeExponentP.length);
    return primeExponentP.length;
  }

  public int getPrimeExponentQ(byte[] pExpQ, int offset) {
    // length consistent
    if (primeExponentQ == null || pExpQ.length < offset + primeExponentQ.length) 
      return 0;
    System.arraycopy(primeExponentQ, 0, pExpQ, offset, primeExponentQ.length);
    return primeExponentQ.length;
  }

  public int getCrtCoefficient(byte[] crtCo, int offset) {
    // length consistent
    if (crtCoefficient == null || crtCo.length < offset + crtCoefficient.length) 
      return 0;
    System.arraycopy(crtCoefficient, 0, crtCo, offset, crtCoefficient.length);
    return crtCoefficient.length;
  }  
}
