/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 */
package jxta.security.impl.random;

import java.util.Random;

import jxta.security.exceptions.CryptoException;
import jxta.security.crypto.JxtaCrypto;
import jxta.security.hash.Hash;

import jxta.security.impl.crypto.JxtaCryptoSuite;

// Generate pseudo random data 
// To generate secure random data a much better initial seed must
// be used.
// 
// Test using "Random" class as required by BigInteger
//
public class JRandom extends java.util.Random {
  /*
   * Here we use the recommendation of rfc2412 for a choice of a seed.
   * These are the first 160 bits of the binary expansion of pi
   */
  private static byte[] seed = {
    (byte) 0xC9, (byte) 0x0F, (byte) 0xDA, (byte) 0xA2,
    (byte) 0x21, (byte) 0x68, (byte) 0xC2, (byte) 0x34,
    (byte) 0xC4, (byte) 0xC6, (byte) 0x62, (byte) 0x8B,
    (byte) 0x80, (byte) 0xDC, (byte) 0x1C, (byte) 0xD1,
    (byte) 0x29, (byte) 0x02, (byte) 0x4E, (byte) 0x08
  };
  private static final byte seedLength = 20; // length in bytes
  private Hash sha1 = null;	// SHA-1 hash algorithm
  private long count = 0;	// for some salt
  private final int timeLength = 8;
  Runtime rtime = null;		// java runtime
  /**
   * get instance of the SHA-1 hash algorithm
   */
  private void PRandom() throws CryptoException {
    JxtaCrypto suite = new JxtaCryptoSuite(JxtaCrypto.MEMBER_SHA1, null,
					   (byte)0, (byte)0);
    short exception = 0;
    if (suite != null) {
      sha1 = suite.getJxtaHash(Hash.ALG_SHA1);
      if (sha1 != null) {
	byte digestLength = sha1.getDigestLength();
	// this is very paranoid
	if (digestLength == seedLength) {
	  rtime = Runtime.getRuntime();
	  count = 0;		// could be more clever here
	  return;
	} else			// something badly broken
	  exception = CryptoException.ILLEGAL_VALUE;
      }
    } else
      exception = CryptoException.NO_SUCH_ALGORITHM;
    // error ..
    throw new CryptoException(exception);
  }
  /**
   * Required constructors
   */
  public JRandom() throws CryptoException {
    PRandom();
  }

  static final int longSize = 8;
  public JRandom(long lseed) throws CryptoException {
    // long into leading 8 bytes of seed
    for (int i = 0; i < longSize; i++)
      seed[i] = (byte)((lseed >>> (8*(7 - i))) & 0xFF);
    PRandom();
  }

  /**
   * An identifier
   */
  public String whoami() { return "Jxta Random"; }
  /**
   * generate some random value. 
   * Note: system time is not that great as we all know but will
   * have to do for small mobile devices.
   *
   * We add a little randomness by XORing a few bits of the amount
   * of free memory.
   *
   * To do: implement a much better churn.
   */
  private synchronized byte[] getChurn()
  {
    long msecs = System.currentTimeMillis() + count; // 64 bits
    long fmem = rtime.freeMemory();
    msecs ^= fmem & 0xFF;

    byte[] tmp = new byte[timeLength];
    // copy into byte array
    for (int i = 0; i < timeLength; i++) {
      tmp[i] = (byte)((msecs >>> (8*(7 - i))) & 0xFF); // s-right 56, 84, ..., 0
    }
    count += 1;			// too slice up a millisecond.
    return tmp;  
  }
  /**
   * generate requested bytes of random data
   * See "Applied Cryptograpy" on "Distilling Randomness."
   * We use SHA-1 rather than MD5.
   */
  public void generateData(byte[] buffer, int offset, int length)
  {
    // first hash our churned random info
    byte[] churn = getChurn();

    sha1.update(seed, 0, seedLength); // hash the seed
    sha1.doFinal(churn, 0, timeLength, seed, 0);
    
    // OK. Copy seed into buffer as required
    if (buffer.length > offset && (offset + length <= buffer.length)) {
      int j = 0;
      for (int i = 0; i < length; i++) {
	buffer[i + offset] = seed[j++]; 
	// see if we've reached the seed's end
	if (j == seedLength) {
	  j = 0;			// beginning of seed
	  // remix the seed
	  sha1.update(seed, 0, seedLength);
	  byte r = (byte)(seed[0] & 0x0F); // 4 random bits
	  churn[0] ^= r;		// change churn randomly
	  sha1.doFinal(churn, 0, timeLength, seed, 0); 
	}
      }
    }
    // seed is our random number and generates our next random
    // number. Make sure that cannot be guessed.
    sha1.update(seed, 0, seedLength);
    byte s = (byte)(seed[0] ^ 0x0F); // 4 random bits
    churn[0] ^= s;		// change churn randomly
    sha1.doFinal(churn, 0, timeLength, seed, 0); 
  }

  /**
   * override our default seed
   */
  public void setSeed(byte buffer[], int offset, int length) 
  {
    if (offset + length <= buffer.length) {
      int nbytes = (int)(length < (int)seedLength ? length : seedLength);
      System.arraycopy(buffer, (int)offset, seed, 0, nbytes);
    }
  }
  /**
   * Random class methods
   */
  public synchronized void setSeed(long lseed) {
    // long into leading 8 bytes of seed
    for (int i = 0; i < longSize; i++)
      seed[i] = (byte)((lseed >>> (8*(7 - i))) & 0xFF);
  }

  /**
   * some bits of an int
   */
  protected synchronized int next(int bits) {
    int i = nextInt();
    i >>>= (32 - bits);		// normalize to bits
    return i;
  }

  /**
   * generate buff.length random bytes
   */
  public void nextBytes(byte buff[]) {
    generateData(buff, 0, buff.length);
  }
  /**
   * Next pseudo random, uniformily distributed int value.
   */
  static final int intSize = 4;

  public int nextInt() {
    byte[] ibyte = new byte[intSize];
    generateData(ibyte, 0, intSize);
    int i = 0, k = intSize - 1;
    for (int j = 0; j < intSize; j++)
      i |= (((int)ibyte[j]) & 0xFF) << 8*(k - j);
    return i & 0x7FFFFFFF;
  }
  public long nextLong() {
    byte[] ibyte = new byte[longSize];
    generateData(ibyte, 0, longSize); 
    long l = 0, m = longSize - 1;
    ibyte[0] &= 0x7F;		// clear any sign bit
    for (int i = 0; i < longSize; i++)
      l |= (((long)ibyte[i]) & 0xFF) << 8*(m - i);
    return l;
  }
}


