/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: RSASignature.java,v 1.5 2001/05/03 00:58:53 yeager Exp $
 */

package jxta.security.impl.signature;

import jxta.security.signature.Signature;
import jxta.security.hash.Hash;
import jxta.security.publickey.PublicKeyAlgorithm;
import jxta.security.exceptions.CryptoException;
import jxta.security.signature.Signature;
import jxta.security.util.Util;

import jxta.security.impl.cipher.KeyBuilder;
import jxta.security.impl.publickey.RSA;
import jxta.security.impl.publickey.RSAKey;

// For RSA digital signature code 
public class RSASignature implements jxta.security.signature.Signature {

  /**
   *   // constructor
   */
  public RSASignature() {}

  private static String myname = "Signature";

  private static byte type = 0;           // Algorithm type
  private static Hash digest = null;      // Hash object
  private static PublicKeyAlgorithm rsa;  // rsa object
  private static RSAKey rKey = null;      // rsa Key
  private static byte mode;       	  // verify or sign
  /**
   *   // Algorithm name
   */
  public String getAlgorithmName() {
    return myname;
  }
  /**
   * constructor
   */
  public RSASignature(PublicKeyAlgorithm rsaAlg,
		      RSAKey theKey,
		      byte theType,
		      Hash theDigest) {
    rsa = rsaAlg;
    rKey = theKey;
    digest = theDigest;
    type = theType;
  }
  /**
   *   // Algorithm type
   */
  public byte getAlgorithm() {
    return type;
  }
  // for debugging
  private boolean debug = false;
  /**
   * turn on debugging
   */
  public void setDebug() { debug = true; }
  /**
   * turn off debugging
   */
  public void clearDebug() { debug = false; }
  /**
   * initialize the Signature mode - sign or verify
   */
  public void init(byte theMode)
    throws CryptoException
  {
    if (theMode != MODE_SIGN && theMode != MODE_VERIFY)
      throw new CryptoException(CryptoException.INVALID_INIT);
    else
      mode = theMode;
  }
  /**
   * updates data to be signed. Here, the data in inbuf
   * is hashed with the digest.
   */
  public void update(byte[] inbuf, 
		     int offset, 
		     int length)
    throws CryptoException
  {
    if (inbuf.length < offset + length)
      throw new CryptoException(CryptoException.ILLEGAL_VALUE);
    digest.update(inbuf, offset, length);
  }
  /**
   * Signs the hash generated from the last(only) buffer of data.
   */
  public byte[] sign(byte[] inbuf, 
		     int offset, 
		     int inLength)
    throws CryptoException 
  {
    // Verify mode
    if (mode != MODE_SIGN)	// correct mode check
      throw new CryptoException(CryptoException.INVALID_INIT);
 
    // last buffer
    byte[] digestBuf = new byte[digest.getDigestLength()];
    digest.doFinal(inbuf, offset, inLength, digestBuf, 0);

    // OK. Sign with RSA Private key
    // The caller MUST have initialized the private key. Otherwise,
    // an UNITIALIZED_KEY CryptoException will be thrown there.
    byte[] localSig = rsa.Algorithm(digestBuf, 0, digestBuf.length, 
				    KeyBuilder.TYPE_RSA_PRIVATE,
				    true);
    return localSig;
  }

  /**
   * Verifies a signature for last/only inBuf against
   * the signature in sigBuf. One can use update as with
   * sign().
   */
  public boolean verify(byte[] inBuf,
			int inOffset,
			int inLength,
			byte[] sigBuf,
			int sigOffset,
			int sigLength)
      throws CryptoException
  {
    if (mode != MODE_VERIFY)	// correct mode check
      throw new CryptoException(CryptoException.INVALID_INIT);
    // update the last buffer of data
    byte[] digestBuf = new byte[digest.getDigestLength()];
    digest.doFinal(inBuf, inOffset, inLength, digestBuf, 0);

    // Decrypt the signature in sigBuf, yielding the original
    // digest
    byte[] digestBuf1 = rsa.Algorithm(sigBuf, sigOffset, sigLength, 
				      KeyBuilder.TYPE_RSA_PUBLIC,
				      false); 

    if (debug) {
      System.out.println("Digest [" + digestBuf.length + "] =\n" +
			 Util.hexEncode(digestBuf) +
			 "\nDecrypted Signature [" + digestBuf1.length + "] =\n" +
			 Util.hexEncode(digestBuf1));
    }
    // MUST have used the same digests (we hope)
    if (digestBuf1.length == digestBuf.length) {
      for (int i = 0; i < digestBuf.length; i++)
	if (digestBuf1[i] != digestBuf[i]) return false;
    } else return false;
    return true;
  }  
}
