/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TestDigest.java,v 1.1 2001/05/10 20:39:04 yeager Exp $
 */

package jxta.security.impl.test;

import jxta.security.hash.Hash;
import jxta.security.exceptions.CryptoException;
import jxta.security.crypto.JxtaCrypto;
import jxta.security.util.Util;

import jxta.security.impl.crypto.JxtaCryptoSuite;

// A general test program for security algorithms we use

public class TestDigest {
  // Test vectors with known results
    private static final String[] shatext  = {
	"abc",
	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
	"-"
    };

    private static final String[] shadigest  = {
	"A9993E364706816ABA3E25717850C26C9CD0D89D",
	"84983E441C3BD26EBAAE4AA1F95129E5E54670F1",
	"34AA973CD4C4DAA4F61EEB2BDBAD27316534016F"
    };
    
    private static final String[] md5text  = {
	"",
	"a",
	"abc",
	"message digest",
	"abcdefghijklmnopqrstuvwxyz",
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
	"12345678901234567890123456789012345678901234567890" +
	    "123456789012345678901234567890"
    };
    
    private static final String[] md5digest = {
	"D41D8CD98F00B204E9800998ECF8427E",
	"0CC175B9C0F1B6A831C399E269772661",
	"900150983CD24FB0D6963F7D28E17F72",
	"F96B697D7CB7938D525A2F31AAF161D0",
	"C3FCD3D76192E4007DFB496CCA67E13B",
	"D174AB98D277D9F5A5611C2C9F419D9F",
	"57EDF4A22BE3C955AC49DA2E2107B67A"
    };  

  private static final byte digestsAlgorithms = JxtaCrypto.MEMBER_SHA1 +
    JxtaCrypto.MEMBER_MD5;

  public static void main(String argv[]) {
    // Select a suite that contains SHA1 and MD5
    JxtaCrypto suite = null;
    try {
      suite = new JxtaCryptoSuite(digestsAlgorithms, null, (byte)0, (byte)0);
    } catch (CryptoException cex) {
      System.out.println("Suite Failure: " + cex.toString());
      System.exit(1);
    }

    Hash sha1 = null;
    Hash md5 = null;
    try {
      sha1 = suite.getJxtaHash(Hash.ALG_SHA1);
      md5 = suite.getJxtaHash(Hash.ALG_MD5);
    } catch (CryptoException cex) {
      System.out.println("Hash Init: " + cex.toString());
      System.exit(1);
    }

    System.out.println("******************************************");
    System.out.println("Digest test of " + sha1.getAlgorithmName());
    System.out.println("Running test vectors through the code.");
    System.out.println("******************************************");

    int i;
    byte[] data = null;
    byte[] digest = new byte[sha1.getDigestLength()];
    long start, stop;    // Get sha1 object
    
    for (i = 0; i < 2; i++) {
      try {
	System.out.println("Test is   <" + shatext[i] + ">");
	data = shatext[i].getBytes();
	int digestLength = sha1.doFinal(data, 0, data.length,
					digest, 0);
	System.out.println("Result:   <" + Util.hexEncode(digest) + ">");
	System.out.println("Expected: <" + shadigest[i] + ">");
	System.out.println("------------------------------------------");	
      } catch (ArrayIndexOutOfBoundsException cex) {
	System.exit(1);
      }
    }

    // do the 1,000,000 "a" test too
    System.out.println("\nVerifying and Timing 1,000,000 a's");

    // build a 1000 byte buffer
    data = new byte[1000];
    for (i = 0; i < 1000; ++i) data[i] = (byte)'a';
    start = System.currentTimeMillis();
    try {
      for (i = 0; i < 999; ++i) {
	sha1.update(data, 0, data.length);
      }
      // last buffer full
      sha1.doFinal(data, 0, data.length, digest, 0);
    } catch (ArrayIndexOutOfBoundsException cex) {
      System.exit(1);
    }
    stop = System.currentTimeMillis();  
    System.out.println("Result:   <" + Util.hexEncode(digest) + ">");
    System.out.println("Expected: <" + shadigest[2] + ">");
    System.out.println("Speed   : <" + 8*1000000/(stop - start) + " Kbits/s>");
    System.out.println("------------------------------------------");    
    System.out.println("\n******************************************");
    System.out.println("Digest test of " + md5.getAlgorithmName());
    System.out.println("Running test vectors through the code.");
    System.out.println("******************************************");

    // Test MD5 with same data array. We don't know the result
    digest = new byte[md5.getDigestLength()];
    try {
      for (i = 0; i < 7; i++) {
	System.out.println("Test is   <" + md5text[i] + ">");
	data = md5text[i].getBytes();
	md5.doFinal(data, 0, data.length, digest, 0);
	System.out.println("Result:   <" + Util.hexEncode(digest) + ">");
	System.out.println("Expected: <" + md5digest[i] + ">");
	System.out.println("------------------------------------------");
      }
    } catch (ArrayIndexOutOfBoundsException cex) {
      System.exit(1);
    }
    System.out.println("\nTest is 1 million 'a' characters");
    start = System.currentTimeMillis();
    try {
      for (i = 0; i < 999; ++i) {
	md5.update(data, 0, data.length);
      }
      // last buffer full
      md5.doFinal(data, 0, data.length, digest, 0);
    } catch (ArrayIndexOutOfBoundsException cex) {
      System.exit(1);
    }
    stop = System.currentTimeMillis();
    if (start == stop) stop += 1; // PC is too fast
    System.out.println("Result:   <" + Util.hexEncode(digest) + ">");
    System.out.println("Expected: <" + "unknown" + ">");
    System.out.println("Speed   : <" + 8*1000000/(stop - start) + " Kbits/s>");
    System.out.println("------------------------------------------");    
    System.exit(0);
  }
}

