/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * 
 */

/**
 * Methods for base64 encode/decode
 */

package jxta.security.impl.test;

import jxta.security.exceptions.CryptoException;
import jxta.security.util.Util;
import jxta.security.util.URLBase64;
import jxta.security.util.GetOpt;

import jxta.security.impl.random.JRandom;
import jxta.security.impl.jxtasec.JxtaSec1;

// Here we fake out having two peers:
// @Peer1 = At Peer1
// @Peer2 = At Peer2
//
// They can exchange public keys and session keys in pipe advertisements
//
public class TestJSEC {
  static private boolean debug = false;
  static private boolean testB64 = false;
  static private boolean testKeyRestore = false;

  static private Object rPublicKey = null;
  static private Object rPrivateKey = null;
  public static void main(String argv[]) {
    getOptions(argv);
    // If URL Base64 testing ...
    if (testB64) {
      dotestB64();
      System.exit(0);
    }
    JxtaSec1 js1 = new JxtaSec1(); // security object

    if (debug) js1.setDebug();	// debugging output

    // @Peer1 and @Peer2: Both initialize object
    if (!js1.jxtaSecInit()) {
      System.out.println("JXTA SEC Init() failure");
      System.exit(1);
    }
    // See if we want to pretend that we got our keys from
    // local storage
    //
    if (testKeyRestore) {
      try {
	rPublicKey = js1.getRSAPublicKeyData();
	rPrivateKey = js1.getRSAPrivateKeyData();
	js1.reset();		// reset the session
	if (!js1.jxtaSecInit(rPublicKey, rPrivateKey)) {
	  System.out.println("JXTA SEC Init(Keys) failure");
	  System.exit(1);	  
	}
      } catch (CryptoException cex) {
	System.out.println("GET Public Key: " + cex.toString());
	System.exit(1);     
      }
    }
    // All generated byte arrays are URL-Base64 encoded strings
    // @Peer1:
    // Get our public key to place it in our pipe 
    // advertisement:
    byte[] advPublicKey = null;
    try {
      advPublicKey = js1.getOurRSAPublicKey();
    } catch (CryptoException cex) {
      System.out.println("GET Public Key: " + cex.toString());
      System.exit(1);     
    }

    if (debug) {
      String adv = new String(advPublicKey); // URL-Base64
      System.out.println("PK Encrypted:\n" + "          " + adv);
    } 

    // @Peer2 initiates the talk session:
    //
    // We now assume we have received the above 
    // public key in a pipe advertisement, have 
    // resolved the pipe, and want to send a symmetric session
    // key in our pipe advertisement for a secure talk session.
    //
    // The following call returns the symmetric session key encrypted in
    // the public key received from @Peer1. 
    byte[] sessionKey = null;
    try {
      sessionKey = js1.getSessionKey(advPublicKey);
    } catch (CryptoException cex) {
      System.out.println("GET Session Key: " + cex.toString());
      System.exit(1);  
    }

    if (debug) {
      String skey = new String(sessionKey); // URL-Base64
      System.out.println("**Encrypted generated session Key: " + skey);
    }

    // @Peer1
    // We have received a pipe advertisement from @Peer2. It contains
    // the symmetric session key we will use encrypted in our advertised
    // pipe. 
    try {
      js1.setSessionKey(sessionKey);
    } catch (CryptoException cex) {
      System.out.println("SET Session Key: " + cex.toString());
      System.exit(1);      
    }
    // @Peer2
    // Send some text to @Peer1
    // All sent data is encrypted or decrypted with the shared
    // symmetric session key the two peers have exchanged:
    //
    String hello = new String("Hello, Bonjour, Nay ho, Ni how, Buenas Dias" +
			      " Konichi-wa mes chers amis!");
    byte[] salut = new byte[hello.length()];
    int N = salut.length;
    for (int i = 0; i < N; i++) salut[i] = (byte)hello.charAt(i);
    byte[] sent = null;
    try {
      sent = js1.encryptData(salut);
    } catch (CryptoException cex) {
      System.out.println("Encrypt Data: " + cex.toString());
      System.exit(1);   
    }
    System.out.println("\n    Send: " + hello);
    if (debug) {
      String str = new String(sent); // URL-Base64
      System.out.println("**Sent Encrypted: " + str);
    }
    //@Peer1
    // Pretend we received the data from @Peer2
    byte[] rcvd = null;
    try {
      rcvd = js1.decryptData(sent);
    } catch (CryptoException cex) {
      System.out.println("Decrypt Data: " + cex.toString());
      System.exit(1);   
    }
    String rstr = new String(rcvd); // this is text in this case
    System.out.println("\nReceived: " + rstr);

    System.exit(0);
  }
  public static void dotestB64() {
    // generate some byte random byte arrays
    //   Encode and decode
    JRandom ran = null;
    try {
      ran = new JRandom();
    } catch (CryptoException cex) {
      System.out.println("Init JRandom: " + cex.toString());
      System.exit(1);
    }
    if (debug) {
      URLBase64.setDebug();
    }
    for (int i = 0; i < 65; i++) {
      byte[] data = new byte[i+1];
      ran.nextBytes(data);
      byte[] be64 = URLBase64.encode(data);
      String str = new String(be64);
      System.out.println("\n   data[" + data.length + "]=" + Util.hexEncode(data) +
			 "\nencoded[" + be64.length + "]=" + Util.hexEncode(be64) +
			 "(" + str + ")");
      byte[] bd64 = null;
      try {
	bd64 = URLBase64.decode(be64, 0, be64.length);
      } catch (CryptoException cex) {
	System.out.println("Base 64 decode: " + cex.toString());
	continue;
      }
      // compare input/output
      boolean success = true;
      if (data.length == bd64.length) {
	for (int j = 0; j < data.length; j++) {
	  if (bd64[i] != data[i]) {
	    success = false;
	    break;
	  }
	}
      } else success = false;
      if (success) {
	System.out.println("Successful decode!");
      } else {
	System.out.println("Bad decode, decoded[" + bd64.length + "]=" + Util.hexEncode(bd64)); 
      }
    }
  }
  public static void getOptions(String[] argv)
  {
    GetOpt options = new GetOpt(argv, "dbk");
    try {
      int c = 0;
      while ((c = options.getNextOption()) != -1) {
	switch (c) {
	case 'b':
	  testB64 = true;
	  break;
	case 'd':
	  debug = true; 
	  break;
	case 't':
	  testKeyRestore = true;
	  break;
	default: break;
	}
      }
    } catch (IllegalArgumentException iae) {
      System.out.println("Usage: TestJSEC [-d -b -t]" +
			 "\n       -d debugging ON" +
			 "\n       -b Test URL base 64" +
			 "\n       -t Test key restore");
    }
  }  
}
