/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TestMAC.java,v 1.1 2001/05/10 20:39:04 yeager Exp $
 */

package jxta.security.impl.test;

import jxta.security.cipher.Key;
import jxta.security.exceptions.CryptoException;
import jxta.security.crypto.JxtaCrypto;
import jxta.security.util.Util;
import jxta.security.util.GetOpt;
import jxta.security.mac.MAC;
import java.lang.IllegalArgumentException;

import jxta.security.impl.cipher.KeyBuilder;
import jxta.security.impl.cipher.SecretKey;
import jxta.security.impl.random.JRandom;
import jxta.security.impl.crypto.JxtaCryptoSuite;

// A general test program for security algorithms we use

public class TestMAC {
  private static int N = 1024;	// number of random bytes we hash
  private static boolean debug = false;
  private static int level = 0; // debugging level default
  private static byte[] macBuf = null;
  private static byte[] data = null;
  private static SecretKey sKey = null;
  private static byte[]privateKey = null;
  private static MAC mac = null;
  private static boolean useMD5 = false;
  private static byte suiteProfile = JxtaCrypto.PROFILE_RC4_SHA1;
  private static byte macProfile = MAC.ALG_RC4_SHA1;

  public static void main(String argv[]) {
    // We need to generate random bytes to sign
    getOptions(argv);
    JRandom ran = null;
    sKey = null;
    // Get random number generator, and a Secret key
    try {
      ran = new JRandom();
      sKey = (SecretKey)KeyBuilder.buildKey(KeyBuilder.TYPE_RC4,
					    KeyBuilder.LENGTH_RC4,
					    false);
    } catch (CryptoException cex) {
      System.out.println("Random# or SecretKey: " + cex.toString());
      System.exit(1);
    }

    // Get a suite to do RC4 + SHA-1 Macs
    JxtaCrypto suite = null;
    try {
      suite = new JxtaCryptoSuite(suiteProfile,
				  null, 
				  (byte)0,
  				  macProfile); // MAC profile
    } catch (CryptoException cex) {
      System.out.println("Crypto suite: " + cex.toString());
    }

    // get our MAC algorithm as specified in the MAC profile above
    mac = suite.getJxtaMAC();
    System.out.println("***** MAC test of " + mac.getAlgorithmName());
    // Set debugging if necessary
    if (debug) {
      mac.setDebug();
    }

    // Make our symmetric key from random data
    privateKey = new byte[sKey.getLength()];
    ran.nextBytes(privateKey);	// get random key

    // get N random bytes, and the signature.
    byte[] ibuf = new byte[N];
    ran.nextBytes(ibuf);

    if (debug) {
      System.out.println("Private Key =\n" + Util.hexEncode(privateKey));
    }

    // For our computed MAC
    System.out.println("\n==================================" + 
		       "\nMAC of " + N + " bytes of random data ...");
    macBuf = new byte[mac.getLength()];
    int macLength = 0;
    try {
      System.out.println("Init MAC algorithm");
      mac.init(MAC.MODE_ENCRYPT, (Key)sKey, privateKey);
      // just one buffer so use sign.
      macLength = mac.encrypt(ibuf, 0, ibuf.length, macBuf, 0);
    } catch (CryptoException cex) {
      System.out.println("MAC failed: " + cex.getMessage());
      System.exit(1);
    }
   
    if (debug) {
      System.out.println("MAC is " + macLength + " bytes");
    }

    // OK. Now we pretend we have received the data and the MAC,
    // done the same key initialization with the shared secret, privateKey
    System.out.println("\nVerifying MAC ...");
    boolean verified = false;
    try {
      mac.init(MAC.MODE_VERIFY, (Key)sKey, privateKey);
      verified = mac.verify(ibuf, 0, ibuf.length, macBuf, 0, macBuf.length);
    } catch (CryptoException cex) {
      System.out.println("MAC Verfication failed: " + cex.getMessage());
      System.exit(1);
    }    

    // What happened?
    System.out.println("\nVerified = " + verified +
		       "\n==================================");

    // now MAC and verify 1,000,000 a characters
    // Set private key and encrypt
    System.out.println("\n==================================" + 
		       "\nInititializing MAC Private Key ...");
    ran.nextBytes(privateKey);	// different private key

    if (debug) {
      System.out.println("Private Key =\n" + Util.hexEncode(privateKey));
    }

    System.out.println("MAC of 1,000,000 a's");   

    try {
      // multiple buffers will use update and sign
      oneMillion(MAC.MODE_ENCRYPT, true); 
    } catch (CryptoException cex) {
      System.out.println("MAC error: " + cex.toString());
      System.exit(1);
    }   
    // set verify mode and verify
    System.out.println("Verify 1,000,000 a's");   
    verified = false;
    // do the 1,000,000
    try {
      verified = oneMillion(MAC.MODE_VERIFY, false);
    } catch (CryptoException cex) {
      System.out.println("Verify error: " + cex.toString());
      System.exit(1);
    }  
    System.out.println("\nVerified = " + verified +
		       "\n==================================");
    System.exit(0);
  }

  // encrypt or verify 1,000,000 a characters
  public static boolean oneMillion(byte mode, boolean init)
    throws CryptoException
  {
    boolean success = false;
    int i = 0;
    if (init) {
      // build a 1000 byte buffer
      data = new byte[1000];
      for (i = 0; i < 1000; ++i) data[i] = (byte)'a';
    }
    try {
      mac.init(mode, (Key)sKey, privateKey);
      for (i = 0; i < 999; ++i) {
	mac.update(data, 0, data.length);
      }
      // last buffer full
      if (mode == MAC.MODE_ENCRYPT) {
	// sign and return the signature
	mac.encrypt(data, 0, data.length, macBuf, 0);
      } else {
	// verify with saved signature
	mac.verify(data, 0, data.length, macBuf, 0, macBuf.length);
      }
      success = true;
    } catch (ArrayIndexOutOfBoundsException cex) {
      System.exit(1);
    }    
    return success;
  }

  public static void getOptions(String[] argv)
  {
    GetOpt options = new GetOpt(argv, "dn:m");
    try {
      int c;
      String multiplier;
      Integer vx;
      while ((c = options.getNextOption()) != -1) {
	switch (c) {
	case 'd':
	  debug = true; 
	  break;
	case 'n': 
	  String NStr = options.getOptionArg();
	  Integer iN = new Integer(NStr);
	  N = iN.intValue();
	  if (N < 0 || N > 4096) {
	    System.out.println("N = " + N + ", illegal. B set to 2048");
	    N = 2048;
	  }
	  break;
	case 'm': 
	  // use MD5 in lieu of SHA-1
	  suiteProfile = JxtaCrypto.PROFILE_RC4_MD5;
	  macProfile = MAC.ALG_RC4_MD5;
	  break;
	default: break;
	}
      }
    } catch (IllegalArgumentException iae) {
      System.out.println("Usage: TestSigning [-d L -n N]" +
			 "\n       -d debugging ON" +
			 "\n       -n N Random bytes." +
			 "\n            1 <= N <= 4096");
    }
  }
}

