/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: TestSigning.java,v 1.1 2001/05/10 20:39:04 yeager Exp $
 */

package jxta.security.impl.test;

import jxta.security.cipher.Key;
import jxta.security.exceptions.CryptoException;
import jxta.security.crypto.JxtaCrypto;
import jxta.security.util.Util;
import jxta.security.signature.Signature;
import jxta.security.util.GetOpt;
import jxta.security.publickey.PublicKeyAlgorithm;
import java.lang.IllegalArgumentException;

import jxta.security.impl.publickey.RSA;
import jxta.security.impl.publickey.RSAKey;
import jxta.security.impl.publickey.RSAPublickeyData;
import jxta.security.impl.publickey.RSAPrivatekeyData;
import jxta.security.impl.cipher.KeyBuilder;
import jxta.security.impl.cipher.SecretKey;
import jxta.security.impl.random.JRandom;
import jxta.security.impl.crypto.JxtaCryptoSuite;

// A general test program for security algorithms we use

public class TestSigning {
  private static int N = 1024;	// number of random bytes we sign
  private static boolean debug = false;
  private static int level = 0; // debugging level default
  private static byte[] sigBuf = null;
  private static byte[] data = null;
  private static Signature sig = null;

  public static void main(String argv[]) {
    // We need to generate random bytes to sign
    System.out.println("Signature test");
    getOptions(argv);
    JRandom ran = null;
    RSAKey rsaKey = null;
    // Get random number generator, and a RSA key
    try {
      ran = new JRandom();
      rsaKey = (RSAKey)KeyBuilder.buildKey(KeyBuilder.TYPE_RSA,
					   KeyBuilder.LENGTH_RSA_512,
					   false);
    } catch (CryptoException cex) {
      System.out.println("Random# or RSAKey: " + cex.toString());
      System.exit(1);
    }

    // Get a suite to do RSA, SHA1, and RSA_SHA1 signatures 
    JxtaCrypto suite = null;
    try {
      suite = new JxtaCryptoSuite(JxtaCrypto.PROFILE_RSA_SHA1,
				  rsaKey, 
				  Signature.ALG_RSA_SHA_PKCS1, 
				  (byte)0);
    } catch (CryptoException cex) {
      System.out.println("Crypto suite: " + cex.toString());
    }

    // Get our rsa and signature objects
    PublicKeyAlgorithm rsa = suite.getJxtaPublicKeyAlgorithm();
    sig = suite.getJxtaSignature();
    // Set debugging if necessary
    if (debug) {
      sig.setDebug();
      if (level == 1) rsa.setDebug();
    }

    // initialize our public and private keys, and save the data
    RSAPublickeyData rPublicD = null;
    RSAPrivatekeyData rPrivateD = null;
    System.out.println("Inititializing RSA Keys ...");
    try {
      rsa.setPublicKey();
      rsa.setPrivateKey();
      rPublicD = (RSAPublickeyData)rsa.getPublickey();
      rPrivateD = (RSAPrivatekeyData)rsa.getPrivatekey();      
    } catch (CryptoException cex) {
      System.out.println("RSA startup failure: " + cex.toString());
      System.exit(1);
    }

    // get N random bytes, and the signature.
    byte[] ibuf = new byte[N];
    ran.nextBytes(ibuf);
    System.out.println("\n==================================" + 
		       "\nSigning " + N + " bytes of random data ...");
    try {
      sig.init(Signature.MODE_SIGN);
      // just one buffer so use sign.
      sigBuf = sig.sign(ibuf, 0, ibuf.length);
    } catch (CryptoException cex) {
      System.out.println("Signature failed: " + cex.getMessage());
      System.exit(1);
    }
   
    if (debug) {
      System.out.println("Signed " + N + " bytes, signature is " + sigBuf.length +
			 " bytes");
    }

    // OK. Now we pretend we have received the data and the signature,
    // and we verify the signature.
    System.out.println("Verifying signature ...");
    boolean verified = false;
    try {
      sig.init(Signature.MODE_VERIFY);
      verified = sig.verify(ibuf, 0, ibuf.length, 
			    sigBuf, 0, sigBuf.length);
    } catch (CryptoException cex) {
      System.out.println("Verfication failed: " + cex.getMessage());
      System.exit(1);
    }    

    // What happened?
    System.out.println("\nVerified = " + verified +
		       "\n==================================");
    // now sign and verify 1,000,000 a characters
    rsa.reset();		// use saved keys

    // Set private key and sign
    System.out.println("\n==================================" + 
		       "\nInititializing RSA Private Key ...");
    try {
      rsa.setPrivateKey(rPrivateD);
    } catch (CryptoException cex) {
      System.out.println("RSA Key init: " + cex.toString());
      System.exit(1);
    }
    // Set signing mode
    try {
      sig.init(Signature.MODE_SIGN);
    } catch (CryptoException cex) {
      System.out.println("Signature init: " + cex.toString());
      System.exit(1);
    }  
    System.out.println("Sign 1,000,000 a's");   
    try {
      // mutiple buffers will use update and sign
      oneMillion(Signature.MODE_SIGN, true); 
    } catch (CryptoException cex) {
      System.out.println("Verify error: " + cex.toString());
      System.exit(1);
    }   
    // reset the RSA key, and set the public key
    rsa.reset();
    System.out.println("\n**********************************" + 
		       "\nSet RSA Public key");
    try {
      rsa.setPublicKey(rPublicD);
    } catch (CryptoException cex) {
      System.out.println("RSA Public Key init: " + cex.toString());
      System.exit(1);
    } 
    // set verify mode and verify
    System.out.println("Verify 1,000,000 a's");   
    verified = false;
    try {
      sig.init(Signature.MODE_VERIFY);
    } catch (CryptoException cex) {
      System.out.println("Verify INIT failed: " + cex.getMessage());
      System.exit(1);
    }
    // do the 1,000,000
    try {
      verified = oneMillion(Signature.MODE_VERIFY, false);
    } catch (CryptoException cex) {
      System.out.println("Verify error: " + cex.toString());
      System.exit(1);
    }  
    System.out.println("\nVerified = " + verified +
		       "\n==================================");
    System.exit(0);
  }

  // sign or verify 1,000,000 a characters
  public static boolean oneMillion(byte mode, boolean init)
    throws CryptoException
  {
    boolean success = false;
    int i = 0;
    if (init) {
      // build a 1000 byte buffer
      data = new byte[1000];
      for (i = 0; i < 1000; ++i) data[i] = (byte)'a';
    }
    try {
      for (i = 0; i < 999; ++i) {
	sig.update(data, 0, data.length);
      }
      // last buffer full
      if (mode == Signature.MODE_SIGN) {
	// sign and return the signature
	sigBuf = sig.sign(data, 0, data.length);
      } else {
	// verify with saved signature
	sig.verify(data, 0, data.length, sigBuf, 0, sigBuf.length);
      }
      success = true;
    } catch (ArrayIndexOutOfBoundsException cex) {
      System.exit(1);
    }    
    return success;
  }

  public static void getOptions(String[] argv)
  {
    GetOpt options = new GetOpt(argv, "d:n:");
    try {
      int c;
      String multiplier;
      Integer vx;
      while ((c = options.getNextOption()) != -1) {
	switch (c) {
	case 'd':
	  debug = true; 
	  String LStr = options.getOptionArg();
	  Integer ilevel =  new Integer(LStr);
	  level = ilevel.intValue();
	  if (level != 0 && level != 1) {
	    System.out.println("L = " + level + ", illegal. L set to 0");
	    level = 0;
	  }
	  break;
	case 'n': 
	  String NStr = options.getOptionArg();
	  Integer iN = new Integer(NStr);
	  N = iN.intValue();
	  if (N < 0 || N > 4096) {
	    System.out.println("N = " + N + ", illegal. B set to 2048");
	    N = 2048;
	  }
	  break;
	default: break;
	}
      }
    } catch (IllegalArgumentException iae) {
      System.out.println("Usage: TestSigning [-d L -n N]" +
			 "\n       -d L   Signature debugging ON" +
			 "\n          L = 0 RSA debugging OFF" +
			 "\n          L = 1 RSA debugging ON" +
			 "\n       -n N Random bytes." +
			 "\n            1 <= N <= 4096");
    }
  }
}

