/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ShellApp.java,v 1.13 2001/06/21 21:28:41 ddoolin Exp $
 */


package net.jxta.impl.shell;

import net.jxta.platform.*;
import net.jxta.document.*;
import net.jxta.peergroup.*;
import net.jxta.exception.*;
import net.jxta.endpoint.*;
import net.jxta.pipe.*;

import java.io.*;
import java.util.*;

/**
 * This class is the base class any JXTA Shell application must
 * extend.
 */
public abstract class ShellApp implements Application {
    
    /**
     *  The command is still running.
     */
    static final public int appSpawned = -1;
    
    /**
     *  The command completed successfully.
     */
    static final public int appNoError = 0;
    
    /**
     *  An error occurred resulting from incorrect or missing parameters.
     */
    static final public int appParamError = 1;
    
    /**
     * Something bad happened. Don't know what it means, don't care why it
     * happened.
     */
    static final public int appMiscError = Integer.MAX_VALUE;
    
    /**
     *  The default peergroup associated with this app.
     **/
    protected PeerGroup group = null;
    
    /**
     *  The pipe service for our default peer group.
     **/
    protected Pipe  pipes = null;
    
    /**
     *  The advertisement for this application if any. nor
     **/
    protected Advertisement adv = null;
    
    /**
     *  The environment for this application. Copied from the host shell.
     **/
    private ShellEnv env = null;
    
    /**
     *  The "stdin" input pipe for this command.
     **/
    protected InputPipe inputPipe = null;
    
    /**
     *  The "stdout" output pipe for this command.
     **/
    protected OutputPipe outputPipe = null;;
    
    /**
     *  The console input. Differs from the stdin if stdin has been
     *  redirected. consin is not normally redirected.
     **/
    protected InputPipe consin = null;;
    
    /**
     *  The console output. Differs from the stdout if stdout has been
     *  redirected. consout is not normally redirected.
     **/
    protected OutputPipe  consout = null;;
    
    /**
     *  if the result is a shell object then store it using this name.
     **/
    protected String returnVar = null;
    
    /**
     *  Has this app begun running?
     **/
    protected volatile boolean started = false;
    
    /**
     *  Has this app in the process of quitting?
     *
     **/
    protected volatile boolean stopped = false;

    /** 
     *  We might need the actual command line for
     *  custom parsing in one of the command classes
     *  located in the impl.shell.bin.* packages.
     *  (This does not seem to work.)
     */
    protected String currentcmdline = null;
    
    /**
     *  private buffer for input pipe
     **/
    private Vector buffered = new Vector();
    
    /**
     *  private buffer for console input pipe.
     **/
    private Vector consbuffer = new Vector();
    
    /**
     *  If this thread is enabled, then this is the root thread of another
     *  command which this command ca
     **/
    public Thread dependsOn = null;
    
    public void init(PeerGroup pg, Advertisement adv) {
        // For now, just do the minimum; real apps can overload it.
        // if desired.
        
        setGroup( pg );
        setAdvertisement( adv );
    }
    
    public int startApp(String[] args) {
        started  = true;
        
        return ShellApp.appSpawned;
    }
        
    public void stopApp() {
        stopped = true;
    }


    /* Override as necessary.  We need something like this
     * because ShellApp may be subclassed several times.
     * Thus every subclass needs to handle whether or not any
     * particular instance of itself is a child shell, no
     * matter what class the parent shell may be.  Returning
     * "true" (hopefully) provides the least astonishing default
     * default behavior, viz., the entire application exits
     * when exit is issued.
     */
    public boolean isRootShell() {
        return true;
    }
    


    // Accessor Methods
    
    final public ShellEnv getEnv() {
        return env;
    }
    
    final public ShellEnv setEnv(ShellEnv e) {
        ShellEnv old = env;
        env = e;
        return old;
    }
    
    final public PeerGroup getGroup() {
        return group;
    }
    
    final public PeerGroup setGroup(PeerGroup g) {
        PeerGroup old = group;
        group = g;
        pipes = group.getPipe();
        return old;
    }
    
    public Advertisement getAdvertisement() {
        return this.adv;
    }
    
    public Advertisement setAdvertisement(Advertisement adv) {
        Advertisement old = this.adv;
        this.adv = adv;
        return old;
    }
    
    final public InputPipe getInputPipe () {
        return inputPipe;
    }
    
    /**
     *
     *  Set the input pipe to the provided pipe. The previous
     *  pipe is returned.
     *
     **/
    final public InputPipe setInputPipe(InputPipe ip) {
        InputPipe old = inputPipe;

        inputPipe = ip;

        return old;
    }
    
    final public OutputPipe getOutputPipe () {
        return outputPipe;
    }
    
    final public OutputPipe setOutputPipe(OutputPipe op) {
        OutputPipe old = outputPipe;
        outputPipe = op;
        return old;
    }
    
    final public InputPipe getInputConsPipe () {
        return consin;
    }
    
    final public InputPipe setInputConsPipe(InputPipe ip) {
        InputPipe old = consin;
        consin = ip;
        return old;
    }
    
    final public OutputPipe getOutputConsPipe() {
        return consout;
    }
    
    final public OutputPipe setOutputConsPipe(OutputPipe op) {
        OutputPipe old = consout;
        consout = op;
        return old;
    }
    
    public void setReturnVariable(String v) {
        returnVar = v;
    }
    
    public String getReturnVariable() {
        return returnVar;
    }
    
    // IO Operations
    
    final protected void println(String line) {
        pipePrintln( outputPipe, line );
    }
    
    final protected void print(String line) {
        pipePrint( outputPipe, line );
    }
    
    final protected String pollInput() throws IOException {
        return pipePollInput( inputPipe, buffered );
    }
    
    final protected String waitForInput() throws IOException {
        return pipeWaitForInput( inputPipe, buffered, true );
    }
    
    final protected void consprint(String line) {
        pipePrint( consout, line );
    }
    
    final protected void consprintln(String line) {
        pipePrintln( consout, line );
    }
    
    final protected String consPollInput() throws IOException {
        return pipePollInput( consin, consbuffer );
    }
    
    final protected String consWaitForInput() throws IOException {
        return pipeWaitForInput( consin, consbuffer, false );
    }
    
    public void help() {
        println("No help available for this ShellApp.");
    }
    
    // Private implementations
    
    private void pipePrint( OutputPipe pipe, String line ) {
        if (null == pipe) {
            return;
        }
        
        String lineToSend = null;
        if (line.length() == 0) {
            lineToSend = " ";
        } else {
            lineToSend = line;
        }
        //Create a message off this string.
        Message msg = pipes.createMessage();
        try {
            msg.push("ShellOutputPipe",
            new ByteArrayInputStream(lineToSend.getBytes()));
            
            pipe.send(msg);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    private void pipePrintln( OutputPipe pipe, String line ) {
        pipePrint( pipe, line + "\n" );
    }
    
    /*
     *  Return a string containing one line of input from the specified pipe.
     *
     *  @param input the pipe to poll.
     *  @param lineBuffer a vector containing buffered lines associated with
     *  this pipe. If not present then it is VERY LIKELY YOU WILL USE MESSAGES.
     *  @return String containing the new line or null if no line was available.
     *  @throws IOException if the input pipe somehow becomes uninitialized.
     */
    private String pipePollInput( InputPipe input, Vector lineBuffer ) throws IOException {
        boolean onePoll = false;
        
        do {
            if( null != lineBuffer ) {
                synchronized( lineBuffer ) {
                    // Check in the buffer first
                    if (lineBuffer.size() > 0) {
                        try {
                            String val = (String) lineBuffer.elementAt(0);
                            lineBuffer.removeElementAt(0);
                            return val;
                        } catch (Exception e) {
                            // This is a very strange case, but if that happens, let just
                            // wait for a message on the InputPipe.
                        }
                    }
                }
            }
            
            if (input == null) {
                throw new IOException();
            }
            
            if( onePoll )
                return null;
            
            // InputPipe.poll() has a different semantic then this method uses
            // Specifically, it will wait a specified an amount of time before
            // returning whether or not there is a message. This method,
            // pipePollInput() is really supposed to be instantaneous so we just
            // calll InputPipe.poll() with a short interval.
            Message msg = input.poll(1);
            
            // no message was waiting.
            if( null == msg )
                return null;
            
            onePoll = true;
            
            // Get all the chunks of data in the message
            for ( Enumeration names = msg.getNames(); names.hasMoreElements();) {
                try {
                    String name = (String) names.nextElement();
                    InputStream inputStream = msg.pop(name);
                    BufferedReader inputReader =
                    new BufferedReader( new InputStreamReader( inputStream ) );
                    
                    do {
                        String command = inputReader.readLine();
                        
                        if( null != command )
                            if( null != lineBuffer )
                                synchronized( lineBuffer ) {
                                    lineBuffer.addElement( command );
                                }
                            else
                                return command;
                        else
                            break;
                    }
                    while( true );
                } catch (Exception e) {
                    e.printStackTrace();
                    continue;
                }
            }
        }
        while( true );
    }
    
    private String pipeWaitForInput( InputPipe input, Vector lineBuffer, boolean join )
    throws IOException {
        String msg = null;
        
        while( true ) {
            msg = pipePollInput( input, lineBuffer );
            
            if( (null != msg) || stopped )
                break;
            
            if( join && (null != dependsOn) && (!dependsOn.isAlive()) )
                break;
            
            try {
                Thread.sleep( 2 ); // not long at all
            }
            catch( InterruptedException woken ) {
            }
        }
        
        return msg;
    }
}
