/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: ShellConsole.java,v 1.23 2001/06/21 21:28:41 ddoolin Exp $
 */

package net.jxta.impl.shell;

import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.io.*;

import net.jxta.impl.cm.*;

public class ShellConsole implements Runnable, KeyListener, MouseListener {
    
    private static final int MAX_LINE = 256;

    /** The number of terminal windows currently open */
    private static volatile int wNumber = 0;
    
    // the name of the file to put in the Cm's shell config private
    // directory to force the shell to come up in terminal mode
    private static final String NOWINDOWFILENAME = "nowindow";

    // the name of the system property that will force the shell to
    // come up in terminal mode
    private static final String NOWINDOWPROPNAME = "SHELLNOWINDOW";

   /** The TextArea object that displays the data */
    private TextArea text = null;
    
    /**
     *  If in Window mode then things send on this go to "in"
     **/
    private OutputStream opipe;
    
    /**
     *  If in Window mode then things read from this come from "out"
     **/
    private InputStream ipipe;
    
    /**
     *  We read from here
     **/
    private BufferedReader stdin = null;
    
    /**
     *  We write to here
     **/
    private BufferedWriter stdout = null;
    
     /** The shell command currently being typed */
    private StringBuffer line = new StringBuffer();

    /** The length of the text displayed, excluding the shell
    *  command currently typed,
    * i.e. the real text length is textLength + line.length() */
    private int textLength = 0;
    
    /** Where in the shell command is the next character added */
    private int lineInsert = 0;

     /** The frame that holds the TextArea object making up the terminal */
    private Frame f = null;

    /** The thread in which this console runs */
    private Thread thread = null;

    /** The name of this console */
    private String consoleName = null;
    
     /**
     * Creates a new terminal window with rows row and cols columns to display.
     * @param rows  the number of rows displayed 
     * @param cols the number of columns displayed 
     */      
    public ShellConsole( int rows, int cols ) {
        this( "JXTA Shell", rows, cols );
    }
    
   
     /**
     * Creates a new terminal window with rows row and cols columns to
     *  display. 
     * @param useName the  name of the console
     * @param rows  the number of rows displayed 
     * @param cols the number of columns displayed 
     */      
    public ShellConsole( String useName, int rows, int cols) {
        InputStream in = null;
        PrintStream out = null;
        
        consoleName = useName;
        
        boolean windowMode = true;
        
        // The following is a trick that allows to start the Shell in
        // text mode not windowed. If a file private/nowindow does
        // exist, then start in text mode
        try {
            Cm cm = new Cm( "ShellConfig" );
            windowMode = ( null == cm.restore( Cm.PrivateDirName, NOWINDOWFILENAME ) );
            if( !windowMode )
                System.err.println( "Shell: the cm/ShellConfig/private/" + NOWINDOWFILENAME +
                " technique is obsolete. please use the " + NOWINDOWPROPNAME + " property" );

        } catch( IOException ignored ) {

        }
        
	// check to see if a system property has been set to run the
	// shell in a graphics free way
        if( windowMode )
            windowMode = !Boolean.getBoolean( NOWINDOWPROPNAME );
        
        // if we are to open graphics, then we need to verify that it
        // may be opened -- this will have the side effect of setting
        // the "text" variable if we can go into graphics mode
        if( windowMode )
            try {
                text = new TextArea( null, rows, cols, TextArea.SCROLLBARS_VERTICAL_ONLY );

		// starting up graphics mode
		wNumber ++;
		
		int fontsize=12;
		String fontsizeProp = System.getProperty( "SHELLFONTSIZE" );
		if( fontsizeProp != null ) {
		    try {
			fontsize = Integer.valueOf( fontsizeProp ).intValue();
		    } catch( NumberFormatException e ) {
		    }
		}
		text.setFont(new Font("DialogInput", Font.PLAIN, fontsize) );
		text.setEditable(true);
		text.addKeyListener(this);
                text.addMouseListener(this);
		
		// Connect the pipes
		try {
		    opipe = new PipedOutputStream();
		    in = new PipedInputStream((PipedOutputStream)opipe);
		    PipedOutputStream pout = new PipedOutputStream();
		    out = new PrintStream( pout );
		    ipipe = new PipedInputStream(pout);
		} catch ( IOException e ) {
		}
		
		// Start the thing
		
		thread = new Thread(this, "JXTA Console-" + consoleName );
		thread.start();
		text.requestFocus();
		
		/*
		 * create the frame
		 */
		
		f = new Frame( consoleName );
		f.add(text, "Center");
		f.pack();
		
		// Offset the frames
		f.setLocation( wNumber * 15, wNumber * 15 );
		f.setVisible(true);
		
		f.addWindowListener( new WindowAdapter() {
			public void windowClosing( WindowEvent e ) {
			    f.dispose();
			    wNumber--;
			    if (wNumber <= 0 ) System.exit(0);
			}
		    } );
                
            } catch( InternalError error ) {
                // this will occur is AWT is not available or can't be
                // initted (most commonly on unix machines in a
                // terminal window)
                windowMode = false;

            }
        
	// there are two specific paths that can be taken to get here
	// -- either the windowMode flag is set to false by the
	// existance of nowindow or the system property, or we went
	// into window mode, got caught by the internal error catch,
	// and then fell through to here
	if( !windowMode ) {
            in = System.in;
            out = System.out;

        }
        
        stdin = new BufferedReader( new InputStreamReader( in ) );
        stdout = new BufferedWriter( new OutputStreamWriter( out ) );

    }
    
    public String getConsoleName() {
        return consoleName;
    }

   /**
    *  Terminates the current terminal window. If this was 
    * the last window open, the program is terminated.
    */    
    public void destroy() {
        thread.interrupt();
        f.dispose();
        
    }
     /**
    *  Handles KeyEvent events.s
    * <P>
    * The events currrently handled are:
    * <ul>
    *   <li> Ctrl-C - allows to get this event to go through to the 
    *                 TextArea object, so it can to the copying
    *   <li> Ctrl-V - call paste to insert the text from the ClipBoard
    *   <li> Ctrl-D - Close shell and exit.</li>
    *   <li> Ctrl-L - Clear the text area.</li>
    *   <li> For all other cases call handling 
    * </ul>
    * All events (except for the ones mentioned above) are consumed, so that
    * they are not handled by the TextArea object.
    * @param e the triggering KeyEvent object
    */   
    public void keyPressed(KeyEvent e ) {
        
        int  val = e.getKeyCode();
        char ch  = e.getKeyChar();
        int  mod = e.getModifiers();
        
         text.setEditable(true);
        // Check for copy/paste hotkeys.  Following Windows convention,
        // copy is Ctrl-C, paste is Ctrl-V.
        //
        // There may be user confusion about Ctrl-C: since this is a shell, users
        // might expect Ctrl-C to terminate the currently running command.
        // At this writing, we don't have command termination, so we'll go ahead
        // and grab Ctrl-C for copy.  (Suggest Esc for termination...)
        //
	 /* Since we are running a shell, and even the cmd shell in windows
	  * responds appropriately to Control-C (amazing), Control-C probably
	  * ought to be changed to terminate the currently running 
	  * command.  Unfortunately, this does conflict with MS Word, 
	  * which does use (intuitively IMO) Control-C for copy.  There
	  * are several alternatives here.  One is to support emacs 
	  * binding with Control-Insert for copy.  Another is to make
	  * the binding context sensitive (yuck?).  Vi bindings could 
	  * be supported.  Or it could be left alone and some other 
	  * key could be used for killing commands.  Two questions really
	  * should be answered before going much further:  1.  Who are 
	  * the targeted users of the jxta shell? 2. What kinds of things
	  * are these users going to do with the shell?  These questions
	  * may be difficult to answer, but once they are handled, key
	  * handling questions go away.  
	  */
        boolean ctrlDown = false;
        if ((mod & InputEvent.CTRL_MASK) > 0)
            ctrlDown = true;
        
        // For copy, just return and let the TextArea control handle it.
        if (ctrlDown == true  &&  val == KeyEvent.VK_C)
            return;
        
        // Paste is more complicated...
        if (ctrlDown == true  &&  val == KeyEvent.VK_V) {
            paste();
            return;
        }
               
        // Let's try a ^D quit...
        if (ctrlDown == true  &&  val == KeyEvent.VK_D) {
            System.out.println("Control-D pressed, bye bye, have a nice day!");
            // Call the exit command from here...
            line.append("exit");
            readme(true);
            return;
        }

        // Control-L is supposed to clear the screen...
        if (ctrlDown == true  &&  val == KeyEvent.VK_L) {
            line.append("clear");
            readme(true);
            return;
        }


        handling(val, ch, mod);
        
        // Eat the event so that it doesn't get processed by the TextArea control.
        e.consume();
    }
    
    /**
    * Handles the editing of the command line.
    * Handling is as follows:
    * <ul>
    *   <li> backspace - delete the character a lineInsert and 
    *        update the terminal text
    *   <li> enter     - call readme
    *   <li> Ctrl-U    - delete the entire currently typed command line
    *   <li> Cursor keys left and right - move the carot and update the 
    *                                     current insertLine value
    *   <li> for all else call process
    * </ul>
    *  
    * @param val the KeyCode value of the key pressed
    * @param ch the  character associated with the key pressed
    * @param modifiers  any modifiers that might have been pressed 
    */
    public void handling(int val, char ch, int modifiers ) {
        int len;

        switch (val) {
            case (KeyEvent.VK_BACK_SPACE ):
                if (line.length() > 0) {
                    if( lineInsert >= 1 && lineInsert <= line.length()){
		       //line.deleteCharAt(lineInsert - 1);
                       // XXX 20010619 bondolo@jxta.org hack to replace deletCharAt
                       String tmpLine = new String( line );
                       line.setLength(lineInsert - 1);
                       line.append(tmpLine.substring(lineInsert));
                       lineInsert--;
                       if( lineInsert < 0) lineInsert = 0;
                       text.replaceRange( line.toString(), textLength, 
					  textLength + line.length()+1 );
                       text.setCaretPosition(textLength+lineInsert);
		    }
                }
                break;
            case (KeyEvent.VK_DELETE ):
                if (line.length() > 0) {
                    if( lineInsert < line.length()){
		       //line.deleteCharAt(lineInsert - 1);
                       // XXX 20010619 bondolo@jxta.org hack to replace deletCharAt
                       String tmpLine = new String( line );
                       line.setLength(lineInsert);
                       line.append(tmpLine.substring(lineInsert + 1));
                       text.replaceRange( line.toString(), textLength, 
					  textLength + line.length()+1 );
                       text.setCaretPosition(textLength+lineInsert);
		    }
                }
                break;
            case ( KeyEvent.VK_ENTER ):
                readme(true);
                break;
            case ( KeyEvent.VK_U ):
                if ((modifiers & InputEvent.CTRL_MASK) > 0 ) {
                    len = line.length();
                    text.replaceRange( "", textLength,
                                           textLength + len);
                    line.setLength( 0 );
                    textLength = text.getText().length();
                    lineInsert = 0;
                } else
                    process(ch);
                break;
	    case (KeyEvent.VK_KP_LEFT):
            case (KeyEvent.VK_LEFT ):
		lineInsert--;
                if( lineInsert < 0) lineInsert = 0;
                text.setCaretPosition(textLength+lineInsert);
                break;
	    case (KeyEvent.VK_KP_RIGHT):
            case (KeyEvent.VK_RIGHT ):
		lineInsert++;
                if( lineInsert > line.length()) lineInsert = line.length();
                text.setCaretPosition(textLength+lineInsert);
                break;
	    case (KeyEvent.VK_KP_UP):
            case (KeyEvent.VK_UP): 
                len = line.length();
                line.setLength(0);
                text.replaceRange(line.toString(),textLength,
				  textLength + len);
                line.append(KeyEvent.getKeyText(KeyEvent.VK_UP));
                readme(false);
                break;
	    case (KeyEvent.VK_KP_DOWN):
            case (KeyEvent.VK_DOWN): 
                len = line.length();
                line.setLength(0);
                text.replaceRange(line.toString(),textLength,
 				                  textLength + len);
                line.append(KeyEvent.getKeyText(KeyEvent.VK_DOWN));
                readme(false);
                break;
            default:
                process(ch);
        }
    }

    /**
    *  Insert the character ch (if it is a valid character) into the
    * current command line at the current insertion point
    * @param ch the character to insert
    */    
    private void process(char ch) {
        if ((ch >= ' ') && (ch <= '~')) {
            int max = text.getText().length();

            line.insert( lineInsert, ch );
            lineInsert++;
            text.replaceRange( line.toString(), textLength,max); 
            text.setCaretPosition(textLength+lineInsert);
       }
    }
    
    private void process(String str) {
        int max = text.getText().length();

        line.insert(lineInsert, str);
        lineInsert += str.length();
        text.replaceRange( line.toString(), textLength, max);
        text.setCaretPosition(textLength+lineInsert);
    }

    /**
    * Replaces the current command line with the command indicated
    *
    * @param prompt if not null, this is set as the new prompt
    * @param cmd the command replacing the typed command
    */
    public synchronized void setCommandLine(String prompt,String cmd){
        if( text != null){     // we are running in windowed mode
           int len = line.length();
           int max = text.getText().length();

           if( prompt != null){
  	         text.replaceRange( prompt, textLength,max); 
                 textLength = text.getText().length();
                 max = textLength;
                 len = 0;
           }
           line.setLength(0);
           line.append(cmd);
           lineInsert = line.length();
           text.replaceRange( line.toString(), textLength, max); 
           text.setCaretPosition(textLength+lineInsert);       
	}
    }
    
    private synchronized void readme(boolean appendNewLine) {
        int len = line.length();

        line.append( '\n' );
        try {
            opipe.write( line.toString().getBytes() );
            opipe.flush();
        }
        catch( IOException ignored ) {
        }

        line.setLength( 0 );
        lineInsert = 0;
        if( appendNewLine) {
             text.append("\n");
             textLength = text.getText().length();
	}
        text.setCaretPosition(textLength + lineInsert);
     }
    
    /**
     * Paste text from the clipboard into the shell.
     * Text is added to at the end of the current command line.
     */
    private void paste() {
        
        // Get text from the clipboard.  If the clipboard contents is non-text,
        // we'll bail out silently.
        Clipboard cb = text.getToolkit().getSystemClipboard();
        Transferable trans = cb.getContents(this);
        if (trans == null)
            return;
        
        String cbText = null;
        try {
            cbText = (String)trans.getTransferData(DataFlavor.stringFlavor);
        }
        catch (Exception e) {
            return;
        }
        
        if (cbText == null)
            return;
        
        
        // Add the clipboard text to the end of the current command line.
        // If there are multiple lines in the clipboard, we paste and
        // execute each line as if the user entered it and and hit return.
        char ch;
        String strLine = "";
        for (int i = 0; i < cbText.length(); i++) {
            ch = cbText.charAt(i);
            if (ch == '\n') {
                
                // Add the line to the end of the current shell text display.
                process(strLine);
                strLine = "";
                
                // Tell the shell to read the current line.  It will go to the
                // shell as a command, unless a command is already running.
                // In that case, the running command may get the text if it reads
                // from the console.
                //
                // After the read, wait a little bit, to give the system time
                // to react to the command.  This is a kludge, but without
                // it the synchronization between commands, prompts, and
                // responses gets messy.  (So messy that I considered allowing
                // only one line of paste text.  But let's try this for awhile.
                // multiline paste seems like a valuable thing.)
                //
                readme(true);
                if (i + 1 < cbText.length())  // There's more to paste...
                    try {Thread.sleep(1000);} catch (InterruptedException e) {}
                    continue;
            }
            
            strLine += String.valueOf(ch);
        }
        
        // If there's anything in strLine at this point, it's text that didn't
        // end in a newline char.  So paste it into the shell, but don't call
        // readme() to execute it.
        //
        if (strLine.length() > 0)
            process(strLine);
    }
    
    synchronized public void print( String s ) {
        text.append(s);
        textLength = text.getText().length(); // sync for safety
        text.setCaretPosition(Integer.MAX_VALUE);
    }
    
    synchronized public void println( String s ) {
        print(s+"\n");
    }
    
    public void run() {
        try {
            byte [] buff = new byte [MAX_LINE];
            int read;
            while ((read = ipipe.read(buff)) != -1 ) {
                print(new String(buff, 0, read) );
            }
        } catch ( IOException e ) {
        } catch ( Exception e) {
            return;
        }
    }
   
    public String read() {
        try {
            return stdin.readLine();
        } catch (IOException ex) {
            return null;
        }
    }
    
    public void write(String msg) {
        try {
            stdout.write( msg );
            stdout.flush();
        } catch (Exception ex) {
            System.err.println("Error=" + textLength + msg.length());
        }
    }
    
    public void keyTyped(KeyEvent e) {
    }
    
    public void keyReleased(KeyEvent e) {
    }
    
  /*
   * Clear the shell window.
   */
    synchronized public void clear() {
        
        // If we're not using the AWT window, then don't attempt to clear.
        if (text == null)
            return;
        
        // Clear the AWT TextArea control.
        //
        // You'd think we could use setText(null), or setText("").  But, nooooooo.
        // It causes later confusion in the control as to what needs to be painted.
        // The symptom is that new typed text appended to control after the clear
        // is not visible.  Weirdly, the invisible text problem doesn't happen
        // immediately after the clear.  It happens after a new command is run
        // after the clear.  (Symptoms seen on WinMe Explorer VM).
        // Anyway, the problem is avoided if replaceRange() is used instead of setText().
        //
        text.replaceRange("", 0, text.getText().length());
        text.setCaretPosition(Integer.MAX_VALUE);
         textLength = 0;
        line.setLength(0);
        lineInsert = 0;
        text.setCaretPosition(textLength + lineInsert);
   }

    // The mouse listener methods
   
    /**
     *    Invoked when the mouse has been clicked on a component.
     *    Changes the isEditable status of text to false.
     *    @param e triggering mouse event
     */
    public void mouseClicked(MouseEvent e){
	text.setEditable(false);
    }

    /**
     *    Invoked when the mouse has entered the  component.
     *    Does nothing
     *    @param e triggering mouse event
     */
    public void mouseEntered(MouseEvent e){ }


    /**
     *    Invoked when the mouse has exited component.
     *    Does nothing
     *    @param e triggering mouse event
     */
    public void mouseExited(MouseEvent e){ }


    /**
     *    Invoked when the mouse has been pressed on a component.
     *    Changes the isEditable status of text to false.
     *    @param e triggering mouse event
     */
    public void mousePressed(MouseEvent e){  
	text.setEditable(false);
    }


    /**
     *    Invoked when the mouse has been released on a component.
     *    Changes the isEditable status of text to false.
     *    @param e triggering mouse event
     */
    public void mouseReleased(MouseEvent e){ 
	text.setEditable(false);
    }

    // define the key names that are send if we use cursor up ar cursor down
    /** 
    *  Returns the command this console send if the cursor up key was pressed.
    * @return command this console send if the cursor up key was pressed
    */ 
    public String getCursorUpName(){
	if( text != null) {       // we are running in window mode
  	   return KeyEvent.getKeyText(KeyEvent.VK_UP);
	}
        else return null;
    }

    /** 
    *  Returns the command this console send if the cursor down
    *  key was pressed.
    * @return command this console send if the cursor down key was pressed
    */ 
    public String getCursorDownName(){
        if( text != null){        // we are running in window mode
	    return KeyEvent.getKeyText(KeyEvent.VK_DOWN);
	}
        else return null;
    }
}

