/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: chpgrp.java,v 1.8 2001/06/20 00:59:45 wiarda Exp $
 */

package net.jxta.impl.shell.bin.chpgrp;

import net.jxta.impl.shell.*;

import net.jxta.peergroup.*;

/**
 * This command changes the ShellObject associated with
 * stdgroup to a new PeerGroup.
 * This is only meaningful, if the shell user has already joined the 
 * group. The join operation itself cahnges the value of stdgroup. However,
 * since  the shell user may want to join more than one group, the
 * chpgrp command can use to reflect the peer group to/from which
 * data are currently piped
 */

public class chpgrp extends ShellApp {

    /** The ShellEnv environment in which to execute the command */
  private ShellEnv env;
 
    /** The name of the group to join - the first command line argument */
  private String newname = null;
 

  /**
  * Constructor - allows to instanciate this class with no parameters
  */
  public chpgrp() {
  }


  /**
  * Start the application and write the output to the appropriate OutputPipe.
  * We expect one command line argument: the name of the group to join.
  * <p>
  * The complete environment variable name is PG#&lt;name&gt; if
  * the group was no yet joined and  PG@&lt;name&gt; otherwise. 
  * The only time we want to change stdgroup is if the group was joined but
  * is not currently stored in stdgroup.
  * @param args the command line arguments. In our case the length of the 
  *             array should be 1 and the value at index 0 should be 
  *              the name of the group to save into stdgroup
  */
  public int startApp (String[] args) {

    env = getEnv();
    ShellObject obj;
    int i;
    if (args == null || args.length > 1)
      return syntaxError();
    
    newname = args[0];


    // check if the new group exist

    if (!exist(newname)) {
      return ShellApp.appParamError;
    }
    
    /**
     * ok switch to the new group
     */

    obj = env.get("PG#"+ newname);
    if( obj == null) obj = env.get("PG@"+ newname);
    if( obj == null){
        println("The peergroup "+newname+" does not exist!");
    }
    PeerGroup npg = (PeerGroup) obj.getObject();
    
    env.add("stdgroup", new ShellObject("PeerGroup", npg));

    return ShellApp.appNoError;
  }

  /**
  * Determines whether the group exist in the shell environment.
  * The complete environment variable name is PG#&lt;name&gt; if
  * the group was no yet joined and  PG@&lt;name&gt; otherwise. 
  * The only time we want to change stdgroup is if the group was joined but
  * is not currently stored in stdgroup.
  * Thus an error message is printed in any case where we do not want
  * to do anything.
  * @return true if the group exists, was joined by the shell user and is
  *         not currently stored in stdgroup; false otherwise
  */
  private boolean exist(String name) {
    ShellObject obj = env.get("PG#"+name);
    if (obj == null) { 
      obj = env.get("PG@"+name);
      if (obj == null) {
	println("There is not group of this name or you have not joined this group yet ");
	return false;
      } else {
	ShellObject sg =  env.get("stdgroup");
	if (sg.getObject() == obj.getObject()) {
	  println("This is already the stdgroup of this shell");
	  return false;
	} else
	  return true;
      } 
    } else
      println("Please join the group first"); 
      return false;
  }
  
 /**
  * Stops the application - does nothing.
  */ 
  public void stopApp() {
  }


 /**
  * Prints out an error message, if the input command line parameters are wrong
  */ 
  private int syntaxError() {
    println ("Usage: chgrp name]");
    return ShellApp.appParamError;
  }
  
 /**
  * Prints out a man type help message
  */ 
  public void help() {
    println("NAME");
    println("     chpgrp  - change the current peer group ");
    println(" ");
    println("SYNOPSIS");
    println(" ");
    println("     chpgrp <name>");
    println(" ");
    println("DESCRIPTION");
    println(" ");
    println("The 'chpgrp' command is used to switch the default Shell peer");
    println("group 'stdgroup' variable to another group that was previously");
    println("joined via a 'join' command. The 'join' command is used to join");
    println("a peergroup.");
    println("After changing group, the Shell 'stdgroup' variable");
    println("is set to the value of the new peer group joined.");
    println(" ");
    println("EXAMPLE");
    println(" ");
    println("       JXTA>mygroupadv = mkadv -g mygroup");
    println("       JXTA>mkpgrp -d mygroupadv mygroup");
    println("       JXTA>join mygroup");
    println("       JXTA>chpgrp moi");
    println(" ");
    println(" This creates a new peergroup which is cloning the services");
    println(" of the parent peer group. You can find the services of the");
    println(" current peer group via the command 'whoami -g'. The new group");
    println(" is given the name 'mygroup'. Before you can do anything with");
    println(" the group you need to join the group via the 'join' command.");
    println(" The 'chpgrp' command is used to change the default group");
    println(" to the new group 'moi'.");
    println(" ");
    println("SEE ALSO");
    println("     mkadv mkpgrp join");
  }
}



