/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: exportfile.java,v 1.8 2001/05/29 16:51:36 msrdanovic Exp $
 */


package net.jxta.impl.shell.bin.exportfile;

import net.jxta.impl.shell.*;

import java.io.*;
import java.util.*;

import net.jxta.peergroup.*;
import net.jxta.peer.*;
import net.jxta.document.*;
import net.jxta.id.*;
import net.jxta.protocol.*;


/**
 * the 'exportfile' shell commands allows to export a StructuredDocument
 * store in a Shell environment variable into an external object 
 * (i.e. file). The same object can be later
 * imported via the 'import' command.
 *
 *
 */

public class exportfile extends ShellApp {

    ShellEnv env;
    boolean stdout = false;

    public exportfile() {
    }

    public void stopApp () {
    }

    private int syntaxError () {
	    println ("Usage: exportfile -f <filename> <variable>");
	    return ShellApp.appParamError;
    }

    public int startApp (String[] args) {

        String filename = null;
	int i = 0;
	String name = null;

	if (args == null || args.length > 3) {
	    return syntaxError();
	}

        env = getEnv();

	if (group == null) {
	    println("exportfile: group is null");
	    return ShellApp.appMiscError;
	}

	if (args[0].equals("-f")) {	  
	  filename = args[1];
	  i=2;
	} else
	  return syntaxError();

	if (args.length != i+1) { // no env, so use stdin
	  return exportfileStdin(filename);
	} else {
	  name = args[i];
	  ShellObject obj = env.get (name);
	  if (obj == null) {
	    println ("exportfile: cannot access " + name);
	    return ShellApp.appMiscError;
	  }

	  try {

	    if (obj.getObject() instanceof StructuredTextDocument)
	      return exportfileStructuredDocument((StructuredTextDocument)
						  obj.getObject(), filename);
	    else 				     
	      return exportfileStructuredDocument((StructuredTextDocument) 
		  ((Advertisement) obj.getObject())
		  .getDocument( new MimeMediaType("text/xml")),filename);

	  } catch (Exception e) {
	    // This is neither an Advertisement or a  Structured 
	    // document
	    println("exportfile: cannot export this kind of object");
	    return ShellApp.appMiscError;
	  }
	}
    }

    private File openFile (String fn) {
	try {
	    return new File (fn);
	} catch (Exception e) {
	    println ("exportfile: cannot open file " + fn);
	    return null;
	}
    }

    private OutputStream getOutput (File f) {

	try {
	    return new FileOutputStream (f);
	} catch (Exception e) {
	    println ("exportfile: cannot open the file for writing");
	    return null;
	}
    }

    private int exportfileStructuredDocument (StructuredTextDocument doc,
					      String fn) {
      File file = openFile(fn);
      if (file == null) {
	return ShellApp.appNoError;
      }

      OutputStream op = getOutput (file);
      if (op == null) {
	return ShellApp.appNoError;
      }
	
      try {
	doc.sendToStream( op );    
	op.close();
      } catch (Exception e) {
	println ("exportfile: I/O error");
	return ShellApp.appMiscError;
      }
      
      return ShellApp.appNoError;

    }
  
    private int exportfileStdin(String fn) {

      File file = openFile(fn);
      if (file == null) {
	return ShellApp.appNoError;
      }

      OutputStream op = getOutput (file);
      if (op == null) {
	return ShellApp.appNoError;
      }

      try {

	String entry = waitForInput() ;
	  
	while(true) {
	    op.write(entry.getBytes());
	    entry =  pollInput();
	    if (entry == null) {
	      op.close();
	      return ShellApp.appNoError;
	    }
	}
      } catch (Exception ex) {
      }
      return ShellApp.appNoError;
    }

    public void help() {
      println("NAME");
      println("     exportfile - export to an external file ");
      println(" ");
      println("SYNOPSIS");
      println(" ");
      println("     exportfile -f <filename> [<env>]");
      println(" ");
      println("DESCRIPTION");
      println(" ");
      println("'exportfile' exports the content of a Shell environment variable");
      println("into an external file. The exported object is stored in the");
      println("supplied filename argument.");
      println("If no variable name is given, the 'stdin' pipe is used to read");
      println("data and saved them into the file.");
      println("'exportfile' is the reverse operation of 'importfile'");
      println("The Shell environment variable is not deleted after the");
      println("the file is created");
      println(" ");
      println("EXAMPLE");
      println("    JXTA> exportfile -f /home/tra/myfile myfile");
      println(" ");
      println("This command saves into the file '/home/tra/myfile' the content");
      println("of the 'myfile' environment variable");
      println(" ");
      println("SEE ALSO");
      println("        importfile ");
    }
}
