/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: importfile.java,v 1.6 2001/05/22 23:42:58 lomax Exp $
 */


package net.jxta.impl.shell.bin.importfile;

import net.jxta.impl.shell.*;

import java.io.*;
import java.util.*;

import net.jxta.pipe.*;
import net.jxta.peergroup.*;
import net.jxta.peer.*;
import net.jxta.document.*;
import net.jxta.id.*;


/**
 * 'importfile' Shell command to load a new file object into a
 * Structured document content environment variable
 *
 */

public class importfile extends ShellApp {

    ShellEnv env;
    String envName;
    boolean noenv = false;

    public importfile() {
    }

    public void stopApp () {
    }

    private int syntaxError () {
	    println ("Usage: importfile -f <filename> [<env>]");
	    return ShellApp.appParamError;
    }

    public int startApp (String[] args) {
 
        int i = 0;

	if ((args == null) || (args.length > 3)) {
	    return syntaxError();
	}

        env = getEnv();

	String filename = null;
	if (args[i].equals("-f")) {
	  filename = args[i+1];
	  i=2;
	} else {
	  return syntaxError();
	}

	if (args.length != i+1) { // no env, so use stdout
	  noenv = true;
	} else {
	  envName = args[i];
	}
	return importfileFile (filename);
    }

    private File openFile (String fn) {
	try {
	    return new File (fn);
	} catch (Exception e) {
	    println ("importfile: cannot open file " + fn);
	    return null;
	}
    }

    private InputStream getInput (File f) {

	try {
	    return new FileInputStream (f);
	} catch (Exception e) {
	    println ("importfile: cannot open the file for reading");
	    return null;
	}
    }


    private int importfileFile (String fn) {

	File file = openFile(fn);
	if (file == null) {
	    return ShellApp.appNoError;
	}

	InputStream ip = getInput (file);
	if (ip == null) {
	    return ShellApp.appNoError;
	}
	byte[] buffer = null;
	try {
	    buffer = new byte [ip.available()];
	    int res = ip.read (buffer);
	    ip.close();

	    if (buffer.length > 64*1024) {
	      println ("importfile: file too big > 64Kb");
	      return ShellApp.appMiscError;
	    }
	} catch (Exception e) {
	    println ("importfile: cannot read the file");
	    return ShellApp.appMiscError;
	}

	ip = new ByteArrayInputStream (buffer);
	StructuredDocument doc = null;


	try {

	  /*
	   * Check if the Document is an XML document
	   */

	  if (buffer[0] == '<' &&
	      buffer[1] == '?' &&
	      buffer[2] == 'x' &&
	      buffer[3] == 'm' &&
	      buffer[4] == 'l') { //OK enough this looks like XML

	    doc = StructuredDocumentFactory.newStructuredDocument (
				   new MimeMediaType ("text/xml"), ip);
	    ip.close();
	     
	  } else {

	      // We may have got an exception because the file is
	      // not an XML Document. Encapsulate it with fake tags
	      String newDoc = 
		"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
		"<ShellDoc>\n<Item>\n" +  new String (buffer) +
		"\n</Item>\n</ShellDoc>";

	      ip = new ByteArrayInputStream (newDoc.getBytes());

	      doc = StructuredDocumentFactory.newStructuredDocument (
		    new MimeMediaType ("text/xml"), ip);
	      ip.close();
	    }
	} catch (Exception ee) {
	  // This time.. There's nothing we can do...
	  ee.printStackTrace();
	  return ShellApp.appMiscError;
	}
    
	
	if (noenv) { // dump it to stdout
	  try {
	    StringWriter    out = new StringWriter();
	    ((StructuredTextDocument)doc).sendToWriter( out );
	    println(out.toString() );
	    out.close();
	  } catch (Exception ex) {
	    ex.printStackTrace();
	  }
	} else {
	  ShellObject old = env.add(envName,
			   new ShellObject("StructuredDocument", doc));
	}

	return ShellApp.appNoError;
    }

    public void help() {
      println("NAME");
      println("     importfile - import an external file ");
      println(" "); 
      println("SYNOPSIS");
      println("     importfile -f <filename> [<env>]");
      println(" ");
      println("DESCRIPTION");
      println("'importfile' imports an external file into a StructuredDocument");
      println("object stored in a Shell environment variable. The name of the");
      println("environment variable is specified as an argument.");
      println("'importfile' is the reverse operation of 'exportfile'");
      println(" ");
      println("EXAMPLE");
      println("    JXTA> importfile -f /home/tra/myfile myfile");
      println("    JXTA> cat myfile");
      println(" ");
      println("This command imports the file '/home/tra/myfile' into the");
      println("'myfile' environment variable");
      println(" ");
      println("SEE ALSO");
      println("        exportfile ");
    }
}
