/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: peerinfo.java,v 1.5 2001/05/22 23:42:59 lomax Exp $
 */

package net.jxta.impl.shell.bin.peerinfo;

import java.util.Enumeration;
import net.jxta.peer.PeerInfo;
import net.jxta.exception.ServiceNotFoundException;
import net.jxta.protocol.PeerInfoAdvertisement;
import net.jxta.impl.shell.ShellEnv;
import net.jxta.impl.shell.ShellApp;
import net.jxta.impl.shell.GetOpt;
import net.jxta.impl.shell.ShellObject;

/**
 * peerinfo command gets information about peers locally and remote
 *
 * @version
 * @since       JXTA 1.0
 */
public class peerinfo extends ShellApp  {
    private ShellEnv env;
    private PeerInfo pip = null;
    private String pid = null;
    private boolean pflag = false;
    private boolean rflag = false;
    private boolean lflag = false;
    private boolean fflag = false;

    public peerinfo() {
    }

    public void stopApp() {
    }

    public int startApp(String[] args) {

        env = getEnv();
        pip =  group.getPeerInfo();

        if (args == null) {
            return getLocal(null);
        }

        GetOpt getopt = new GetOpt(args, "flrp:");

        int c;
        try {
            while ((c = getopt.getNextOption()) != -1) {
                switch (c) {
                case 'p': 
                    pid = getopt.getOptionArg();
                    pflag = true;
                    break;

                case 'r': 
                    rflag = true;
                    break;

                case 'f': 
                    fflag = true;
                    break;

                case 'l': 
                    lflag = true;
                    break;

                default:
                    println("Error: peerinfo this should have never happened");

                }
            }

            if (rflag) {
                // pid will be null if pflag is false
                return getRemote(pid);
            } else if (lflag) {
                return getSelf();
            } else if (fflag) {
                // pid will be null if pflag is false
                return flush(pid);
            } else {
                return getLocal(pid);
            }
        } catch (Exception ex) {
            println("Illegal option");
            help();
        }
        return ShellApp.appNoError;
    }

    // address can be null
    private int flush(String address) {
        try {
            pip.flushAdvertisements(address);
        } catch (Exception ex) {
            return ShellApp.appMiscError;
        }
        return ShellApp.appNoError;
    }

    private int getRemote(String address) {
        pip.getRemotePeerInfo(address);
        println ("peer info message sent");

        return ShellApp.appNoError;
    }

    private int getSelf() {
        PeerInfoAdvertisement adv = null;
        try {
            adv = pip.getPeerInfo();
        } catch (Exception e) {
            println("nothing stored");
            return ShellApp.appMiscError;
        }

        // peerinfo0 is always localhost
        env.add("peerinfo0", 
                new ShellObject("stored local peerinfo advertisement", 
                                adv));
        println("stored peerinfo0");

        return ShellApp.appNoError;
    }

    // pid can be null, in that case, return info for all local advts
    private int getLocal(String pid) {
        Enumeration res;
        try {
            res = pip.getLocalPeerInfo(pid);
        } catch (Exception e) {
            println("nothing stored");
            return ShellApp.appMiscError;
        }

        // we start at index 1 since peerinfo0 is always localhost
        int j = 1;
        while (res.hasMoreElements()) {
            PeerInfoAdvertisement peer = 
                (PeerInfoAdvertisement) res.nextElement();
            String pname = "peerinfo"+j;
            env.add(pname,
                    new ShellObject("stored local peerinfo advertisement", 
                                    peer));
            println("stored " + pname);
            j++;
        }
        return ShellApp.appNoError;
    }

    public void help() {
        println("NAME");
        println("     peerinfo - get information about peers ");
        println(" ");
        println("SYNOPSIS");
        println("     peerinfo ");
        println("           [-p peerid] ");
        println("           [-r] gets information about peers remotely");
        println("           [-l] return local peer info");
        println("           [-f] flush peer info");



        println(" ");
        println("DESCRIPTION");
        println(" ");
        println("use to get information about other peers within a peer group or at a");
        println("specified peer location. The default peerinfo option lists");
        println("only the peerinfo already known by the peer. The '-r' option is used");
        println("to send a propagate request to find information about new peers.");
        println("peerinfo stores results in the local cache, and inserts ");
        println("advertisement(s) into the environment, using the default");
        println("naming: peerinfoX where X is a growing integer number.");
        println(" ");
        println("OPTIONS");
        println("-p peerid");
        println("     gets peer info at a given peer location");
        println("-r");
        println("     gets info about remote peers");
        println("-l");
        println("     return local peer info");
        println("-f");
        println("     flush locally cached peer info");
        println(" ");
        println("EXAMPLE");
        println(" ");
        println("    JXTA>peerinfo -r");
        println(" ");
        println("    refresh info about the peers in the current peer group");
        println(" ");
        println("SEE ALSO");
        println("    ");


    }
}
