/*
 * Copyright (c) 2001 Sun Microsystems, Inc.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *       Sun Microsystems, Inc. for Project JXTA."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Sun", "Sun Microsystems, Inc.", "JXTA" and "Project JXTA" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Project JXTA at http://www.jxta.org.
 *
 * 5. Products derived from this software may not be called "JXTA",
 *    nor may "JXTA" appear in their name, without prior written
 *    permission of Sun.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SUN MICROSYSTEMS OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of Project JXTA.  For more
 * information on Project JXTA, please see
 * <http://www.jxta.org/>.
 *
 * This license is based on the BSD license adopted by the Apache Foundation.
 *
 * $Id: recv.java,v 1.5 2001/05/22 23:43:00 lomax Exp $
 */


package net.jxta.impl.shell.bin.recv;

import net.jxta.impl.shell.*;



import java.io.*;
import java.util.*;

import net.jxta.peergroup.*;
import net.jxta.endpoint.*;
import net.jxta.pipe.*;

/**
 * recv: receive a message from an InputPipe
 */
public class recv extends ShellApp {
  ShellEnv env;

  public recv() {
  }

  public void stopApp () {
  }

  private int syntaxError () {
      println ("Usage: recv [-t timeout] <InputPipe>");
      return ShellApp.appParamError;
  }

  public int startApp (String[] args) {
    if ((args == null) || (args.length > 3)) {
	return syntaxError();
    }

    env = getEnv();

    // Is there a timeout ?
    int timeout = -1;

    if (args[0].equals ("-t")) {
	try {
	    timeout = Integer.parseInt(args[1]);
	} catch (Exception e) {
	    println ("Timeout must be a number");
	    return ShellApp.appParamError;
	}
    }

    String pipe = null;
    if (timeout >= 0) {
	pipe = args[2];
    } else {
	pipe = args[0];
    }

    ShellObject obj = env.get (pipe);
    if (obj == null) {
      println ("recv: cannot access " + pipe);
      return ShellApp.appMiscError;
    }

    InputPipe ip = null;
    try {
      ip = (InputPipe) obj.getObject();
    } catch (Exception e) {
      println("wait: " + pipe + " is not an InputPipe");
      println(e.toString());
      return ShellApp.appMiscError;
    }

    Message msg = null;
    int counter = 0;
    try {
	if (timeout == -1) {
	    msg = ip.poll(0);
	} else {
	    msg = ip.poll(timeout);
	}
    } catch (IOException e) {
	// The pipe has been closed. Return
	//        println("wait: pipe " + pipe + " is closed");
	//        return ShellApp.appMiscError;
	return ShellApp.appNoError;
    }

    if (msg == null) {
	println ("recv has not received any message");
	return ShellApp.appMiscError;
    } else {
	println ("recv has received a message");
    }

    ShellObject old = env.add (getReturnVariable(),
                 new ShellObject ("Message from  " + pipe, msg));
    return ShellApp.appNoError;
  }

  public void help() {
      println("NAME");
      println("     recv - receive a message from a pipe");
      println(" ");
      println("SYNOPSIS");
      println("     recv [-t timeout] <input pipe>");
      println(" ");
      println("DESCRIPTION");
      println(" ");
      println("'recv' receives a message from an input pipe. The input pipe needs");
      println("to have been previously created. A timeout in seconds can be given.");
      println("If no timeout is given, the call will block until a message is");
      println("received. A timeout of zero corresponds to a non-blocking call.");
      println(" ");
      println("EXAMPLE");
      println(" ");
      println("    JXTA> pipeadv = mkadv -p");
      println("    JXTA> inpipe = mkpipe -i pipeadv");
      println("    JXTA> msg = recv inpipe");
      println("    JXTA> data = get msg mytag");
      println(" ");
      println("This example creates a pipe advertisement 'pipeadv'");
      println("create an input pipe 'inpipe' and receive a message 'msg'");
      println("The body of the message associated with the tag 'mytag' is");
      println("retreived from the message via the 'get' command.");
      println(" ");
      println("SEE ALSO");
      println("    mkmsg put get send mkadv mkpipe");
  }
}
